--[[----------------------------------------------------------------------------

LmAlbumAssetModel.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSReactive = require 'KSReactive'
local KSPromises = require 'KSPromises'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzComments = require 'WFOzComments'
local WFOzFavorites = require 'WFOzFavorites'
	
	-- WF.oz_client.catalog
local OzCatSharingHelper = require 'OzCatSharingHelper'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark LmAlbumModel

local LmAlbumAssetModel = {}

--------------------------------------------------------------------------------

local function translateCommentsOrFavorites( docs )
	
	local translatedDocs = {}
	
	for _, doc in ipairs( docs ) do
		local name = KSTableUtils.derefOpt( doc, 'author', 'fullName' )
		local authorId = KSTableUtils.derefOpt( doc, 'author', 'id' )
		local text = doc.body or nil
		local isLike = not doc.body
		local id = doc.id
		local created = KSDate.iso8601ToTimeIntervalSinceReference( doc.updated or doc.created )
		
		local translatedDoc = {
								  authorId = authorId,
								  commentRealname = name,
								  comment = text,
								  is_like = isLike,
								  dateCreated = created,
								  remoteId = id,
								  avatars = doc.avatars
								}
		
		table.insert( translatedDocs, translatedDoc )
	end

	return translatedDocs
end

--------------------------------------------------------------------------------

local function combineCommentsAndFavorites( comments, favorites )
	
	local combined = {}
	
	for i = 1, #comments do
		combined[ #combined + 1 ] = comments[i]
	end
	
  	for i = 1, #favorites do
        combined[ #combined + 1 ] = favorites[i]
    end

	return combined
end

--------------------------------------------------------------------------------

local function translateAvatarUrlsFromP( wfOzCatalog, avatarsP )
		
	return avatarsP
			:mapDataP( function( status, links )
				local avatars = {}
				
				if links then
					local baseUrl = wfOzCatalog:ozCatalogSpec():serverBaseUrl()
					for k, link in pairs( links ) do
						local slashPos = k:match( ".*/()" ) or 1
					
						local key = k:sub( slashPos )
						avatars[ key ] = {
							url = baseUrl .. link.href,
							size = link.size,
						}
					end
				end
				
				return status, avatars
			
			end )
end

--------------------------------------------------------------------------------

function LmAlbumAssetModel.modelForOzAlbumAndAssetId( lmOzCatalog, ozAlbumId, ozAssetId )

	check( 'LmOzCatalog', '?uuid', '?uuid' )
	
	local wfOzCatalog = lmOzCatalog:wfOzCatalog()
	
	local avatarUrlsForCommentP = function( model, ozCommentId )
		-- this promise is based of the avatar URLs behavior and will not return until some
		-- avatars are actually available for the comment
		local avatarsB = WFOzComments.getAvatarUrlsForCommentB( wfOzCatalog, ozCommentId )
		local avatarsP = require 'KSReactivePromiseBridge'.behaviorToPromise( avatarsB, function( v ) return v and next( v ) ~= nil end )
		
		return translateAvatarUrlsFromP( wfOzCatalog, avatarsP )
			
	end
	
	local favoritesB = WFOzFavorites.getFavoritesOnAlbumAssetB( wfOzCatalog, ozAlbumId, ozAssetId )
							:closingMapSwitchB( function( favoritesArray )
								if WFUnknown.isUnknown( favoritesArray ) then
									return KSReactive.constantB( WFUnknown.value() )
								end
								
								local addAvatarsP = KSPromises.mapArrayP( function( favorite )
								
								-- attach whatever avatars are currently in the doc store to the favorites list
								local avatarsP = WFOzFavorites.getAvatarUrlsForFavoriteP( wfOzCatalog, favorite.id )
								
								return translateAvatarUrlsFromP( wfOzCatalog, avatarsP )
									:mapP( function( status, avatars )
										
										if status == 'data' and type( avatars ) == 'table' then
											favorite.avatars = avatars
										end
										
										return 'data', favorite
									end )
									
								end, favoritesArray )
									
								return KSReactivePromiseBridge.promiseToBehavior( addAvatarsP, {} )

							end )
	
	local commentsB = WFOzComments.getCommentsOnAlbumAssetB( wfOzCatalog, ozAlbumId, ozAssetId )
				:closingMapSwitchB( function( commentsArray )
					if WFUnknown.isUnknown( commentsArray ) then
						return KSReactive.constantB( WFUnknown.value() )
					end
		
					local addAvatarsP = KSPromises.mapArrayP( function( comment )
							-- attach whatever avatars are currently in the doc store to the comments list
							local avatarsP = WFOzComments.getAvatarUrlsForCommentP( wfOzCatalog, comment.id )
							
							return translateAvatarUrlsFromP( wfOzCatalog, avatarsP )
								:mapP( function( status, avatars )
									
									if status == 'data' and type( avatars ) == 'table' then
										comment.avatars = avatars
									end
									
									return 'data', comment
								end )
								
						end, commentsArray )
						
					return KSReactivePromiseBridge.promiseToBehavior( addAvatarsP, {} )
				end )

	
	local isSharedB = OzCatSharingHelper.isAssetOrAlbumSharedB( wfOzCatalog:indexStore(),
																wfOzCatalog:docStore(),
																'album',
																ozAlbumId )
	
	local commentsAndFavoritesB = KSReactive.closingDataCallB(
		function( comments, favorites )
			if WFUnknown.isUnknown( comments ) or WFUnknown.isUnknown( favorites ) then
				-- return an empty table until we have real data for both comments and favorites
				return {}
				
			else
			 	
				comments = translateCommentsOrFavorites( comments )
				favorites = translateCommentsOrFavorites( favorites )
				local commentsAndFavorites = combineCommentsAndFavorites( comments, favorites )
				
				return {
						 comments = comments,
						 favorites = favorites,
						 commentsAndFavorites = commentsAndFavorites,
						}
			end
		end,
		commentsB, favoritesB )
		
	local addCommentP = function( model, comment )
	
		return WFOzComments.addCommentP( wfOzCatalog, ozAlbumId, ozAssetId, comment )
	
	end

	local deleteCommentP = function( model, commentId )
	
		return WFOzComments.deleteCommentP( wfOzCatalog, commentId )
	
	end
			
	return {

		value = {
			
			favorites = favoritesB,
			comments = commentsB,
			commentsAndFavorites = commentsAndFavoritesB,
			isShared = isSharedB,
		},
				
		models = {
			addComment = addCommentP,
			deleteComment = deleteCommentP,
			avatarUrlsForComment = avatarUrlsForCommentP,
		},
	}

end

--------------------------------------------------------------------------------

return LmAlbumAssetModel
