--[[----------------------------------------------------------------------------

LmAlbumModel.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.catalog
local WFOzSharing = require 'WFOzSharing'

	-- WF.oz_client.sync_engine
local WFOzSyncStatus = require 'WFOzSyncStatus'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark LmAlbumModel

local LmAlbumModel = {}

--------------------------------------------------------------------------------

function LmAlbumModel.modelForOzAlbumId( lmOzCatalog, ozAlbumId )

	check( 'LmOzCatalog', 'uuid' )

	local docStore = lmOzCatalog:docStore()
	local indexStore = lmOzCatalog:indexStore()
	
	local albumNameB = docStore:fieldValueCheckTypeB( 'album', ozAlbumId, 'name' )

	local albumSizeB = WFOzSyncStatus.albumSizeB( indexStore, ozAlbumId )
	local uploadedCountB = WFOzSyncStatus.uploadedAssetsForAlbumB( indexStore, ozAlbumId )
	
	local uploadedPercentB = KSReactive.dataCallB(
		function( albumSize, uploadedCount )
		
			if type( albumSize ) == 'number'
				and type( uploadedCount ) == 'number'
				and albumSize > 0
			then
				return uploadedCount / albumSize
			else
				return 0
			end
		
		end,
		albumSizeB, uploadedCountB )
	
	local spaceIdB = WFOzSharing.spaceIdB( lmOzCatalog:wfOzCatalog(), ozAlbumId )
	
	local isAlbumSharedPubliclyB = WFOzSharing.isLoginRequiredForShareB( lmOzCatalog:wfOzCatalog(), ozAlbumId )
		:mapB( function( isLoginRequired )
			return isLoginRequired == false
		end )
		
	local publicAlbumShareUrlB = KSReactive.closingDataCallB( function( isLoginRequired, url )
	
			if isLoginRequired ~= nil and WFUnknown.isKnown( isLoginRequired ) then
				return not isLoginRequired and url or nil
			else
				return nil
			end
		end,
		WFOzSharing.isLoginRequiredForShareB( lmOzCatalog:wfOzCatalog(), ozAlbumId ),
		WFOzSharing.publicShareUrlB( lmOzCatalog:wfOzCatalog(), ozAlbumId ) )
	
	-- FIXME [scouten 2014-04-29]: This should come from Oz. Waiting for Matt
	-- to tell me it's been deployed.

	local privateAlbumShareUrlB = KSReactive.constantB( string.format(
					'https://lightroom.adobe.com/libraries/%s/albums/%s',
					lmOzCatalog:ozCatalogId(),
					ozAlbumId ) )

	local shareAlbumP = function( model )
			return WFOzSharing.makePublicShareP( lmOzCatalog:wfOzCatalog(), ozAlbumId )
		end
																
	local unshareAlbumP = function( model )
			return WFOzSharing.makePrivateShareP( lmOzCatalog:wfOzCatalog(), ozAlbumId )
		end
		
	local removeAlbumShareP = function( model )
			return WFOzSharing.removeShareP( lmOzCatalog:wfOzCatalog(), ozAlbumId )
		end
	
	return {

		value = {

			ozAlbumId = ozAlbumId,
			
			name = albumNameB,
			
			size = albumSizeB,
			uploadedCount = uploadedCountB,
			uploadedPercent = uploadedPercentB,
			
			isAlbumSharedPublicly = isAlbumSharedPubliclyB,
			publicAlbumShareUrl = publicAlbumShareUrlB,
			privateAlbumShareUrl = privateAlbumShareUrlB,
			spaceId = spaceIdB,
		},
		
		close = function( self )
			-- TODO [matthewj 2014-05-29]: It would be nice to not have
			-- to explicitly enumerate the behaviors that need to be closed
			-- here and have more WF-ish close targeting in models
			albumNameB:close()
			albumSizeB:close()
			uploadedCountB:close()
			uploadedPercentB:close()
			publicAlbumShareUrlB:close()
			privateAlbumShareUrlB:close()
			isAlbumSharedPubliclyB:close()
			spaceIdB:close()

		end,
		
		models = {
			shareAlbum = shareAlbumP,
			unshareAlbum = unshareAlbumP,
			removeAlbumShare = removeAlbumShareP,
		},
	}

end

--------------------------------------------------------------------------------

return LmAlbumModel
