--[[----------------------------------------------------------------------------

LmServiceModel.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.os
local WFLocalizationUtils = require 'WFLocalizationUtils'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'	
local KSReactive = require 'KSReactive'

	-- WF.core.rxlua
local WFRxPromiseBridge = require 'WFRxPromiseBridge'

	-- WF.oz_client.account
local WFOzClient = require 'WFOzClient'

	-- WF.oz_client.catalog
local AgLrOzEndpoints = require 'AgLrOzEndpoints'

	-- WF.oz_client.ims
local WFImsAppIdentity = require 'WFImsAppIdentity'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

	-- WF.web.http_client
local WFHttpClient = require 'WFHttpClient'
local WFReachability = require 'WFReachability'

	-- WF.web.utils
local WFURI = require 'WFURI'

--------------------------------------------------------------------------------

local asKSPromise = WFRxPromiseBridge

--------------------------------------------------------------------------------

local sAuthorizationScope = 'openid,AdobeID'

local sHttpTimeout = 30 -- seconds

local sRedirectUri = 'http://stophere.adobe.com'


--============================================================================--

-- @mark LmServiceModel

local LmServiceModel = {}

--------------------------------------------------------------------------------

local function imsAuthUrlP( overrides )

	return WFOzClient.getImsAuthUrl( overrides )
		:call( asKSPromise )
	
end

--------------------------------------------------------------------------------

local function imsRequestP( args )

	local paramsStr = WFURI.buildQueryString( args.params )

	local overrides = {
		baseServerUrl = AgLrOzEndpoints.syncAndGeneralBaseUrl,
		languageCode = args.languageCode,
		regionCode = args.regionCode,
		apiVersion = "v2",
			-- WARNING [scouten 2016-02-03]: DO NOT integrate this change "apiVersion = 'v2'"
			-- to Lightroom 6.x! This is hopefully temporary until we can strip the Nimbus UI
			-- app of any attempt to communicate with Oz.
	}

	return imsAuthUrlP( overrides )

		:mapDataP( function( _, _, imsDomain )

			local requestUrl = string.format( '%s/%s', imsDomain, args.imsUrlPath )

			return WFHttpClient.putP( requestUrl, paramsStr,
								{
									requestMethod = 'POST',
									requestHeaders = {
										{
											header = 'Content-Type',
											value = 'application/x-www-form-urlencoded',
										},
									},
									timeout = sHttpTimeout,
								} )
		
		end )

		:mapP( function( status, rawJson )
			
			local _, jsonResult = nil, {}
			
			if status ~= 'error' then
				_, jsonResult = pcall( KSJsonDecoder.decodeWebSafeJson, rawJson, true )
			end
			
			if type( jsonResult ) ~= 'table' then
				jsonResult = {}
			end

			if jsonResult.jump then

				return 'data', { jumpUrl = jsonResult.jump }
			
			end

			local err

			if status == 'data' then
								
				local success, result = pcall( args.parseExpectedJson, jsonResult )
				
				if success and result then
					if KSPromises.isPromise( result ) then
						return result
					else
						return 'data', result
					end
				else
					err = result
				end
			
			end
			
			-- Something went wrong. See if we can construct a reasonable
			-- error message.
			
			err = err or jsonResult.error

			if ( not err ) and status == 'error' and type( jsonResult ) == 'string' then
				err = jsonResult
			end
			
			return 'error', err
		
		end )
		
end

--------------------------------------------------------------------------------

local function deviceTokenToAccessAndRefreshTokenP( model, deviceToken )
	
	return imsRequestP
		{

			imsUrlPath = 'token/v1',
					
			params = {
				grant_type = 'device',
				device_id = model.deviceId,
				client_id = model.clientId,
				client_secret = model.clientSecret,
				scope = sAuthorizationScope,
				locale = model.locale,
				languageCode = model.languageCode,
				regionCode = model.regionCode,
				device_token = deviceToken,
			},
			
			parseExpectedJson = function( jsonResult )
			
				if jsonResult.access_token and jsonResult.refresh_token then
					return
						{
							accessToken = jsonResult.access_token,
							refreshToken = jsonResult.refresh_token,
							deviceToken = deviceToken,
							userId = jsonResult.userId
						}
				end
	
			end,
			
		}	
end

--------------------------------------------------------------------------------

local function fetchContinueTokenP( model, url, accessToken, targetClientId )

	return imsRequestP
		{

			imsUrlPath = 'jumptoken/v1',
					
			params = {
				target_client_id = targetClientId,
				target_redirect_uri = url,
				target_scope = model.scope,
				locale = model.locale,
				bearer_token = accessToken,
			},
			
		}

		:mapDataP( function( _, jumpInfo )

			if type( jumpInfo ) ~= 'table' or not jumpInfo.jumpUrl then
				return 'error', "Incomplete jump info from IMS", jumpInfo
			end

			
			return 'data', jumpInfo.jumpUrl
			
		end )
	
end

--------------------------------------------------------------------------------

-- [bzibble 2014-06-13]: fix #3772741: when the app language is set to automatic,
-- we need to translate the language code that we get from Mac OS to make it match
-- the format expected by IMS.

local function sanitizeLanguageCode( code )

	local replacements = {
		[ "zh-Hans" ] = "zh_cn", -- Chinese simplified
		[ "zh-Hant" ] = "zh_tw", -- Chinese traditional
	}
	
	return replacements[ code ] or code
end

--------------------------------------------------------------------------------

function LmServiceModel.serviceModel( systemInfo )
	
	local hostIsReachableB = WFReachability.hostIsReachableB( AgLrOzEndpoints.syncAndGeneralBaseUrl )
	
	-- TODO [scouten 2016-01-13]: Some of these values can now also be computed by WFOzClient.getImsAuthUrl
	-- but they are local to that function (rightly so IMHO). Other parts of LmServiceModel need access,
	-- so duplicating effort was deemed the lesser of evils.
	
	local clientId, clientSecret = WFImsAppIdentity.getClientIdAndSecret()
	
	local wfDeviceId = WFImsAppIdentity.generateDeviceId( systemInfo )
	WFImsAppIdentity.setImsDeviceId( wfDeviceId )
	
	local languageCode = sanitizeLanguageCode( systemInfo.languageCode or WFLocalizationUtils.getLanguageCode() )
	local regionCode = systemInfo.regionCode or WFLocalizationUtils.getRegionCode()
	local locale = string.format( '%s_%s', languageCode, regionCode )
	
	local imsAuthUrlP = WFOzClient.getImsAuthUrl {
			apiVersion = "v2",
				-- WARNING [scouten 2016-02-03]: DO NOT integrate this change "apiVersion = 'v2'"
				-- to Lightroom 6.x! This is hopefully temporary until we can strip the Nimbus UI
				-- app of any attempt to communicate with Oz.
			baseServerUrl = AgLrOzEndpoints.syncAndGeneralBaseUrl,
			languageCode = systemInfo.languageCode,
			regionCode = systemInfo.regionCode,
			systemUserName = systemInfo.systemUserName,
			macAddress = systemInfo.macAddress,
			queryParamOverrides = {
				response_type = 'device',
				client_id = clientId,
				client_type = 'desktop',
				scope = sAuthorizationScope,
				idp_flow = 'login',
				redirect_uri = sRedirectUri,
				ctx_id = 'lr5_c1',
				locale = locale,
				device_id = wfDeviceId,
			}
		}
		:call( asKSPromise )
		:mapP( function( status, ... )
			if status == 'error' then
				return 'data', nil 
			else
				return status, ... 
			end
		end )
		:runUntilDoneP()
	
	local loginUrlForSusiB = hostIsReachableB
				:mapSwitchB( function( isReachable ) 
					if isReachable == 'unreachable' then
						return KSReactive.constantB( nil )
					else 
						return imsAuthUrlP
							:untilResolvedB( 'data', nil )
					end
				end )
				
	
	return {

		value = {

			-- host system status
			hostIsReachable = hostIsReachableB,
			
			-- SPAM [matthewj 2015-07-20]: Ideally these live somewhere else to provide 
			-- the SUSI login functionality or, ideally, not at all
			-- in the event that all login/account interaction could 
			-- happen in AMT dialogs and we don't have to host a web
			-- view
			
			-- service info for SUSI login
			loginUrlForSUSI = loginUrlForSusiB,
			
			redirectUri = sRedirectUri,
		},
		
		models = {
			-- SPAM [matthewj 2015-07-20]: This really should be moved to WFOzUserLogin but so far we need to be 
			-- able to handle error responses from IMS and display them in LoginDialog. As 
			-- events in WFOzUserLogin this isn't possible
			deviceTokenToAccessAndRefreshToken = deviceTokenToAccessAndRefreshTokenP,
			
			-- SPAM [matthewj 2015-08-06]: This should be moved to WFOzUserModel/WFOzUser
			fetchContinueToken = fetchContinueTokenP,
		},
		
		deviceId = wfDeviceId,
		locale = locale,
		machineName = systemInfo.machineName or "",
		languageCode = languageCode,
		regionCode = regionCode,
		clientId = clientId,
		clientSecret = clientSecret,
		
		close = function( self )
		
			hostIsReachableB:close()
		end,
		
	}
end

--------------------------------------------------------------------------------

return LmServiceModel
