--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
local WFPlatformInfo = require 'WFPlatformInfo'
local LrmAllPhotosQueryBuilder = {}
local WFLogger = require 'WFLogger'
local logger = WFLogger 'PeopleLogger'
 :setLevel 'debug'


--------------------------------------------------------------------------------

local function getCollectionInfoFields (activeAlbum, isSearch)

	local	fields = [[wasDeveloped, thumbnailMD5, captureDate, contentType, duration, 
		displayDate, timestamp, copyTimestamp, flag, croppedWidth, croppedHeight, 
		fileName, width, height, apertureValue, isoSpeedRatings, exposureTime, 
		cameraMaker, cameraModel, rating, modifiedDate, isVideo, localUrl, fnumber, isMergedHDR, hasThumbnail]]
	
	if WFPlatformInfo.isAndroid then
		fields = [[croppedWidth, croppedHeight, captureDate,
		fileName, modifiedDate, thumbnailMD5, timestamp as importDate, isMergedHDR, hasThumbnail]]
	end
	fields = "distinct(coreInfo.docId) as assetId, "	.. fields
	
	if isSearch then
		fields = fields .. ", TEMP_search.sortOrder"
	end
	
	return fields
	
end

--------------------------------------------------------------------------------

local function isRatingPredicateRequired( ratingFilter, ratingComparisonOperator )

	if ratingComparisonOperator == ">=" and ratingFilter == 0 then
		return false
	elseif ratingComparisonOperator == "<=" and ratingFilter == 5 then
		return false
	end
	return true

end

--------------------------------------------------------------------------------

local function addRatingPredicate( ratingFilter, ratingComparisonOperator, predicate, queryParams )

	predicate = predicate .. " rating " .. ratingComparisonOperator .. " ?"
	table.insert( queryParams, ratingFilter )
	
	return predicate, queryParams

end

--------------------------------------------------------------------------------

local function makeAllPhotoViewPredicate( assetFilter, assetTypeFilter, ratingFilter, ratingComparisonOperator )
	
	local queryParams		= {}
	local predicate 		= ""
	local predicateStart	= [[WHERE ]]
	
	if #assetFilter > 0 then
		
		predicate = predicateStart .. "("
	
		for i = 1, #assetFilter do
		
			if i > 1 then
				predicate = predicate .. " OR "
			end
			
			local filter = assetFilter[ i ]
			
			if filter == 'pick' or filter == 'reject' or filter == 'unflagged' then
				predicate = predicate .. "flag == ?"
				table.insert( queryParams, filter )
			elseif filter == 'favorite' then
				predicate = predicate .. "favoriteCount > 0"
			elseif filter == 'comment' then
				predicate = predicate .. "commentCount > 0"
			end
			
		end
		
		predicate = predicate .. ")"
	end
	
	if assetTypeFilter and #assetTypeFilter > 0 then
	
		if predicate == "" then
			predicate = predicateStart .. "("
		else
			predicate = predicate .. " AND ("
		end
		
		for i = 1, #assetTypeFilter do
		
			if i > 1 then
				predicate = predicate .. " OR "
			end
			
			local assetType = assetTypeFilter[ i ]
			
			if assetType == 'image' then
				predicate = predicate .. "isVideo == ?"
				table.insert( queryParams, 0 )
			elseif assetType == 'video' then
				predicate = predicate .. "isVideo == ?"
				table.insert( queryParams, 1 )
			end
			
		end
		
		predicate = predicate .. ")"
	
	end
	
	if isRatingPredicateRequired( ratingFilter, ratingComparisonOperator ) then
		if predicate == "" then
			predicate = predicateStart
		else
			predicate = predicate .. " AND"
		end
		predicate, queryParams = addRatingPredicate( ratingFilter, ratingComparisonOperator, predicate, queryParams )
	end
	
	return predicate, queryParams
end

--------------------------------------------------------------------------------

function LrmAllPhotosQueryBuilder.buildCoreInfoQuery( activeAlbum, ascending, assetFilter, assetTypeFilter, sortCriteria, 
														ratingFilter, ratingComparisonOperator, isSearch, personId )

    local order = 'DESC'
    
    if ascending or isSearch then
        order = 'ASC'
    end

	local predicate = ""
	local queryParams = {}
	local selectSubqueries = ""
    local personAssetFilter = {}
    local personRatingFilter = 0
	
	if activeAlbum then

		if personId then
			predicate, queryParams = makeAllPhotoViewPredicate( personAssetFilter, assetTypeFilter, personRatingFilter, ratingComparisonOperator )
		else
			predicate, queryParams = makeAllPhotoViewPredicate( assetFilter, assetTypeFilter, ratingFilter, ratingComparisonOperator )
		end


	end
	
	-- INFO [pkamin 2013-12-11]: we map the filename criteria to the table column with lowercase names and apply the correct collation
	if sortCriteria == 'fileName' then
		sortCriteria = 'fileNameLC COLLATE i18n_collate '
	end
	
	local secondOrder = 'coreInfo.captureDate'
	
	if isSearch then
		sortCriteria = 'TEMP_search.sortOrder'
	elseif sortCriteria == 'userDefined' then
		sortCriteria = 'sortOrder'
	elseif sortCriteria == 'captureDate' then
		secondOrder = 'coreInfo.docId'
	elseif sortCriteria == 'importDate' then
		sortCriteria = "timestamp"
	elseif sortCriteria == 'mostFavorited' then
		sortCriteria = "favoriteCount "
	end
	
	-- Build sub queries for comment count and favorite count
	
		selectSubqueries = [[, 0 AS commentCount, 0 AS favoriteCount ]]
	
	-- Build final SQL statement
	local collectionFields = getCollectionInfoFields (activeAlbum)
	
	local sql = [[SELECT ]] .. collectionFields .. selectSubqueries .. [[ FROM coreInfo ]]
	
	if isSearch then
		sql = sql .. [[ INNER JOIN TEMP_search ON coreInfo.docId = TEMP_search.docId ]]
	end

	logger:debugf("Person Id is %s",personId)

	sql = sql	.. predicate

    if personId then
        -- continuation of where clause
        -- TODO agk : revisit this when we have video support as where clause can become empty for people incase of it
        sql = sql .. [[AND coreInfo.docId in
        (SELECT faces.assetId from faces join persons p on p.docId = faces.personId
        where p.docId = ? or p.primaryId == ?)
            ]]

        table.insert(queryParams, personId)
        table.insert(queryParams, personId)

    end

    sql = sql ..
		[[ ORDER BY ]] .. sortCriteria .. [[ ]] .. order .. [[, ]] .. secondOrder .. [[ ]] .. order
		
	if sortCriteria == 'modifiedDate' or sortCriteria == 'rating' then
		sql = sql .. [[, fileNameLC COLLATE i18n_collate ]] .. order .. [[, coreInfo.docId ]] .. order
	end

	logger:debugf('sql is : %s , %s', sql, queryParams)

	-- Fire query
	
	return sql, queryParams
						
end

--------------------------------------------------------------------------------

return LrmAllPhotosQueryBuilder

