--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require "Rx"

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSIdGenerator = require 'KSIdGenerator'
local WFCatUtils = require 'WFCatUtils'
local WFTableMerge = require 'WFTableMerge'

	-- WF.core.os
local KSDate = require 'KSDate'

local WFCatCollaboration = require 'WFCatCollaboration'

local WFRxPromiseBridge = require 'WFRxPromiseBridge'

local WFRxRunUntilDone = require "WFRxRunUntilDone"

--==============================================================================

local asKSPromise = WFRxPromiseBridge

local runUntilDone = WFRxRunUntilDone

--==============================================================================

local LrmCatAlbum = {}

--------------------------------------------------------------------------------
--[[!
	\brief Creates an album in this catalog.
	
	\param name (string) The name of the new album.
	\param parentId (string) The parentId of the new album
				(could be 'root' or 'empty' which is synonym for top level album)
	\param isFolder (boolean) The new album is of subtype collection_set
	\param hideInOrganize (boolean) The new album is an adhoc share
	\return <code>P< string ></code> A Promise object that resolves to the ID of the
		newly-created album when the operation is complete.
]]--

function LrmCatAlbum.createAlbumP( catalog, name, parentId, isFolder, hideInOrganize )

	WFArguments.check( 'OzCatalogService', 'string', 'string', 'boolean', 'boolean' )
	
	if not ( #name > 0 ) then
		error( "OzCatalogService:createAlbumP: The album name must be a non-empty string", 2 )
	end
	
	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )
	local subtype = isFolder and 'collection_set' or 'collection'

	local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( albumId )

	local albumDocValue = {
		type = 'album',
		subtype = subtype,
		name = name,
		_localOnly = {
		updateTimeStamp = updateTimeStamp,
		}
	}

	albumDocValue.hideInOrganize = hideInOrganize

	if #parentId > 0 and parentId ~= 'root' then
		albumDocValue.parent = {
			id = parentId
		}
	end
	
	WFTableMerge.applyOverlay( albumDocValue, baseDocValue )

	local resultP = catalog:docStore():applyOverlayP( albumDocValue )
	
	:runUntilDoneP()

	return resultP:mapDataP( function( status,_ )
		return status, albumId
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Reparent an album in this catalog.

	\param albumId (string) The albumId of the album to reparent.
	\param newParentId (string) The new parentId

	\return <code>P< string ></code> A Promise object that resolves to true when the operation is complete.
]]--

function LrmCatAlbum.reparentAlbumP( catalog, albumId, newParentId )

	WFArguments.check( 'OzCatalogService', 'string', 'string' )

	if not ( #albumId > 0 ) then
		error( "OzCatalogService:reparentAlbumP: The albumId must be a non-empty string", 2 )
	end


	if newParentId ~= albumId then
	
		local overlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( albumId )
		
		overlay.type = 'album'
		overlay.parent = {
			id = newParentId
		}

		if not (#newParentId > 0) or newParentId == 'root' then
			overlay.parent = {
				_delete = true
			}
		end

		local resultP = catalog:docStore():applyOverlayP( overlay )
		:runUntilDoneP()

		return resultP:mapDataP( function ()
			return 'data', true
		end )

	else

		return KSPromises.resolvedPromiseP( 'error', "Album can not be its own parent" )

	end

end

--------------------------------------------------------------------------------

--[[!
	\brief Is an albumId valid for this catalog
	
	\param albumId (string) The unique ID of the album to modify.
	
	\return <code>P< boolean ></code> A Promise object that resolves to a boolean
		if the album id exists and has not been deleted for this catalog.
]]--

function LrmCatAlbum.isValidAlbumP( catalog, albumId )

	WFArguments.check( 'OzCatalogService', 'string' )
	
	return catalog:indexStore():queryP( [[ SELECT 1 FROM albums WHERE docId = ? LIMIT 1 ]], albumId )
		:mapDataP( function( status, albumList )
			
			return 'data', not not albumList[ 1 ]

		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Renames an album in this catalog.
	
	It is an error to rename an album to match the name of an existing album,
	regardless of case.
	
	\param albumId (string) The unique ID of the album to rename.
	\param name (string) The new name for the album.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function LrmCatAlbum.renameAlbumP( catalog, albumId, name )

	WFArguments.check( 'OzCatalogService', 'string', 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:renameAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	if not ( #name > 0 ) then
		error( "OzCatalogService:renameAlbumP: The album name must be a non-empty string", 2 )
	end
	
	local baseDocValue = WFCatUtils.createBaseOverlayWithUserUpdatedField( albumId )

	local albumDocValue = {
		type = 'album',
		name = name,
		_localOnly = {
			updateTimeStamp = KSDate.timeInIso8601Format( nil, false ),
		}
	}

	WFTableMerge.applyOverlay( albumDocValue, baseDocValue )

	local resultP = catalog:docStore():applyOverlayP( albumDocValue )

	resultP:runUntilDone()

	return resultP:mapDataP( function ()
		return 'data', true
	end )
	
end

--------------------------------------------------------------------------------

function LrmCatAlbum.unshareAlbum( catalog, albumId )

	WFArguments.check( 'OzCatalogService', 'string' )

	return WFCatCollaboration.deleteSharedAlbum( catalog, albumId )

end

--------------------------------------------------------------------------------
--[[!
	\brief Unshares an album in this catalog, deleting the associated space
		and space_album documents.

	\param albumId (string) The unique ID of the album.

	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function LrmCatAlbum.unshareAlbumP( catalog, albumId )

	return LrmCatAlbum.deleteSharedAlbum( catalog, albumId )
		:call ( asKSPromise )
	
end

--------------------------------------------------------------------------------

function LrmCatAlbum.deleteAlbum( catalog, albumId )

	WFArguments.check( 'OzCatalogService', 'string' )

	if not ( #albumId > 0 ) then
		error( "OzCatalogService:deleteAlbumP: The albumId must be a non-empty string", 2 )
	end

	-- Remove any album share first, then delete the album, too.
	-- Regardless of success/failure of unsharing the album, we're still going
	-- to delete it, hence the catch()

	return LrmCatAlbum.unshareAlbum( catalog, albumId )
		:catch( Rx.Observable.empty() )
		:concat( Rx.Observable.just( true ) )
		:takeLast( 1 )
		:flatMapLatest( function()

			return catalog:docStore():applyOverlay {
				docType = 'album',
				_id = albumId,
				_deleted = true,
			}

		end )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

--[[!
	\brief Deletes an album from this catalog.

	\param albumId (string) The unique ID of the album to delete.

	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function LrmCatAlbum.deleteAlbumP( catalog, albumId )

	return LrmCatAlbum.deleteAlbum( catalog, albumId )
		:call( asKSPromise )

end

--------------------------------------------------------------------------------
--[[!
	\brief Updates the timestamp for an album in this catalog.
	
	\param albumId (string) The unique ID of the album to touch.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function LrmCatAlbum.touchAlbumP( catalog, albumId )

	WFArguments.check( 'OzCatalogService', 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:touchAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	local baseDocValue = WFCatUtils.createBaseOverlayWithUserUpdatedField( albumId )

	local albumDocValue = {
		type = 'album',
		_localOnly = {
			updateTimeStamp = KSDate.timeInIso8601Format( nil, false ),
		}
	}

	WFTableMerge.applyOverlay( albumDocValue, baseDocValue )

	local resultP = catalog:docStore():applyOverlayP( albumDocValue )
						
	resultP:runUntilDone()
	
	return resultP:mapDataP( function ()
		return 'data', true
	end )

end

function LrmCatAlbum.convertAlbumToAdhoc(catalog , albumId)
	WFArguments.check( 'OzCatalogService', 'string' )

	if not ( #albumId > 0 ) then
		error( "OzCatalogService:touchAlbumP: The albumId must be a non-empty string", 2 )
	end

	local result = {}
	local overlay = {
		_id = albumId,
		hideInOrganize = true,
		parent = {
			_delete = true
		},
		_localOnly = {
			updateTimeStamp = KSDate.timeInIso8601Format( nil, false ),
		}
	}

	result = catalog:docStore():applyOverlay( overlay )

	return {
		value = result
	}


end

function LrmCatAlbum.convertAdhocToAlbum(catalog , albumId)
	WFArguments.check( 'OzCatalogService', 'string' )

	if not ( #albumId > 0 ) then
		error( "OzCatalogService:touchAlbumP: The albumId must be a non-empty string", 2 )
	end

	local result = {}
	local overlay = {
		_id = albumId,
		hideInOrganize = false,
		_localOnly = {
			updateTimeStamp = KSDate.timeInIso8601Format( nil, false ),
		}
	}

	result = catalog:docStore():applyOverlay( overlay )

	return {
		value = result
	}

end

--------------------------------------------------------------------------------

return LrmCatAlbum
