--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSTableUtils = require "KSTableUtils"
local KSStringUtils = require 'KSStringUtils'
local WFPlatformInfo = require 'WFPlatformInfo'
local LrmAssetUtils = require 'LrmAssetUtils'

local LrmCoreInfoIndexDefinitions = require "LrmCoreInfoIndexDefinitions"

local LrmOzRenditionServerUtils = require "LrmOzRenditionServerUtils"

local WFRxIndexStoreSchema = require 'WFRxIndexStoreSchema'

local WFOzContentTypes = require 'WFOzContentTypes'

--------------------------------------------------------------------------------

local getLocalUrl = assert( LrmAssetUtils.getLocalUrl )

local derefOpt = assert( KSTableUtils.derefOpt )

local assetIsDisplayable = assert( LrmCoreInfoIndexDefinitions.assetIsDisplayable )

local toLower = assert( KSStringUtils.lower )

local getContentTypeNumberForContentType = WFOzContentTypes.getContentTypeNumberForContentType

--==============================================================================

local LrmCatalogIndexDefinitions = {}

--------------------------------------------------------------------------------
-- @mark assetProfileLinks
--------------------------------------------------------------------------------

local kAssetProfileLinks = {

	sql = [[CREATE TABLE assetProfileLinks( subtype, fileName, masterUrl, uniqueId )]],

	updateLazily = WFPlatformInfo.isAndroid and false or nil, -- don't wait for the table in Android to be used. Do indexing ASAP

	map = function( docValue, emit )
		
		if docValue._deleted then
			return
		end
		if not assetIsDisplayable( docValue ) then
			return
		end
		
		local cameraProfileInfo = LrmAssetUtils.getCameraProfileInfo( docValue)
		if cameraProfileInfo then
			emit( "camera", cameraProfileInfo.filename, cameraProfileInfo.href, cameraProfileInfo.uniqueId )
		end
		
		local lensProfileInfo = LrmAssetUtils.getLensProfileInfo( docValue )
		if lensProfileInfo then
			emit( "lens", lensProfileInfo.filename, lensProfileInfo.href, lensProfileInfo.uniqueId )
		end
	end,
}

local kAssetProfileLinks_uniqueId = [[CREATE INDEX assetProfilesLink_uniqueId ON assetProfileLinks( uniqueId )]]

--------------------------------------------------------------------------------
-- @mark profileRegistration
--------------------------------------------------------------------------------

local kProfileRegistration = {

	sql = [[CREATE TABLE profileRegistration( subtype, fileName, uniqueId )]],

	updateLazily = WFPlatformInfo.isAndroid and false or nil, -- don't wait for the table in Android to be used. Do indexing ASAP

	map = function( docValue, emit )

		if (docValue.type == 'asset' and docValue.subtype == 'camera_profile') or docValue.type == 'profile'
			and not docValue._deleted then

			local fileName
			local uniqueId
			local subtype

			if docValue.type == 'asset' and not WFPlatformInfo.isAndroid then
				-- Custom DCPs

				subtype = 'camera'
				fileName = docValue.fileName
				uniqueId = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )

			else
				-- Adobe standard camera matching profiles

				if WFPlatformInfo.isAndroid and docValue.type ~= 'profile' then
					return
				end

				local localOnly = docValue._localOnly

				subtype = localOnly.subtype
				fileName = localOnly.fileName
				uniqueId = localOnly.uniqueId

			end

			if uniqueId then
				emit( subtype, fileName, uniqueId )
			end
		end
	end,
}

local kProfileRegistration_uniqueId = [[CREATE INDEX profileRegistration_uniqueId ON profileRegistration( uniqueId )]]
local kProfileRegistration_fileName = [[CREATE INDEX profileRegistration_fileName ON profileRegistration( fileName COLLATE NOCASE )]]

--------------------------------------------------------------------------------
-- @mark importSource
--------------------------------------------------------------------------------

local kImportSource =  {

	sql = [[CREATE TABLE importSource( sha256, localUrl )]],

	map = function( docValue, emit )

		if docValue.type == 'asset' and (docValue.subtype == 'image' or docValue.subtype == 'video') 
		  and not docValue._deleted then
		
			local sha256 = docValue.importSource
								and docValue.importSource.sha256
			
			local localUrl = getLocalUrl( docValue )
			
			if sha256 then
				sha256 = toLower( sha256 )
			end
			
			if sha256 or localUrl then
				emit( sha256, localUrl )
			end
		
		end

	end,

}

--------------------------------------------------------------------------------
-- @mark TEMP_searchFilterCount
--------------------------------------------------------------------------------

local kTEMP_searchFilterCount =  {

	sql = [[CREATE TEMPORARY TABLE TEMP_searchFilterCount( filterType, filterCount )]],
	
	isManualTable = true
	
}

--------------------------------------------------------------------------------
-- @mark renditionRevisions
--------------------------------------------------------------------------------

local function isVideo( docValue )

	if docValue and docValue.video then
		
		return true
		
	end
	
	local contentType = derefOpt( docValue, 'importSource', 'contentType' )
								or derefOpt( docValue, '_serverProvidedMetadata', 'links', 
												'/rels/master', 'content_type' )
				
	if contentType then
	
		local contentTypeNumber = getContentTypeNumberForContentType( contentType )
		
		if LrmAssetUtils.isVideoContentTypeNumber( contentTypeNumber ) then
	
			return true
			
		end
		
	end
	
	return false
		
end

--------------------------------------------------------------------------------

local kRenditionRevisionsDefinition = {

	sql = [[CREATE TABLE renditionRevisions( renditionName, timestamp, size )]],
		
	map = function( docValue, emit )

		if docValue._deleted then
			return
		end
		
		if not assetIsDisplayable( docValue ) then return end
		
		local timestamp = docValue
			and docValue.importSource
			and docValue.importSource.importTimestamp
		
		if LrmOzRenditionServerUtils.shouldDownloadOzRendition( docValue, "thumbnail2x" ) then
			emit( 'thumbnail2x', timestamp, 22528 )
		end
		
		if LrmOzRenditionServerUtils.shouldDownloadOzRendition( docValue, "preview" ) then
			emit( 'preview', timestamp, 461824 )
		end
		
		local hasProxyOnOz = LrmAssetUtils.hasProxyOnOz( docValue )
		local hasProxyLocally = LrmAssetUtils.hasProxyLocally( docValue )
		local hasMasterOnOz = LrmAssetUtils.hasMasterOnOz( docValue )
		local hasMasterLocally = LrmAssetUtils.hasMasterLocally( docValue )
		local hasExternalXmpDevelop = LrmAssetUtils.hasExternalXmpDevelop( docValue )
		
		local isVideo = isVideo( docValue )
		
		if not isVideo then
		
			if hasMasterOnOz and not hasMasterLocally then
			
				local fileSize = derefOpt( docValue, 'importSource', 'fileSize' )
		
				if not fileSize or fileSize <= 0 then
					fileSize = 20971520
				end
		
				emit( 'master', timestamp, fileSize )
				
			end
			
			if hasProxyOnOz and not hasProxyLocally then
				emit( 'proxy', timestamp, 1331692 )
			end
			
			if not hasMasterOnOz and not hasMasterLocally and not hasProxyLocally then
				emit( 'proxyWithoutMaster', timestamp, 1331692 )
			end
		
			if hasExternalXmpDevelop and not LrmAssetUtils.hasExternalXmpDevelopLocally( docValue ) then
			
				local fileSize = derefOpt(docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'fileSize' )
			
				emit( 'xmp_develop', timestamp, fileSize )
			end

		end
						
	end
}


local kImportSource_sha256 =
	[[CREATE INDEX importSource_sha256 ON importSource( sha256 )]]

local kImportSource_localUrl =
	[[CREATE INDEX importSource_localUrl ON importSource( localUrl )]]


function LrmCatalogIndexDefinitions.tableDefinitions( tableDefinitions )
	
	tableDefinitions.renditionRevisions = kRenditionRevisionsDefinition
	tableDefinitions.assetProfileLinks = kAssetProfileLinks
	tableDefinitions.profileRegistration = kProfileRegistration
	tableDefinitions.importSource = kImportSource
	tableDefinitions.TEMP_searchFilterCount = kTEMP_searchFilterCount
	
	tableDefinitions.new_asset_activity = nil
	tableDefinitions.tags = nil
	tableDefinitions.favorite_activity = nil
	tableDefinitions.selectedPhotos = nil
	tableDefinitions.space_user = nil
	
	return tableDefinitions
end

--------------------------------------------------------------------------------

function LrmCatalogIndexDefinitions.indexDefinitions( indexDefinitions )

	indexDefinitions.profileRegistration_uniqueId = kProfileRegistration_uniqueId
	indexDefinitions.profileRegistration_fileName = kProfileRegistration_fileName
	indexDefinitions.assetProfileLinks_uniqueId = kAssetProfileLinks_uniqueId
	
	indexDefinitions.importSource_sha256 = kImportSource_sha256
	indexDefinitions.importSource_localUrl = kImportSource_localUrl
		
	return indexDefinitions
	
end

--==============================================================================

local tableDefinitions = {}
local indexDefinitions = {}

--------------------------------------------------------------------------------

local indexStoreSchema = WFRxIndexStoreSchema( )

indexStoreSchema:addDatabaseInitRoutine( require 'WFSQLiteSetLogicImp'.registerSetModule )

require 'OzBinaryUploadUtils'.filterIndexDefinitions( tableDefinitions, indexDefinitions )

--------------------------------------------------------------------------------

LrmCatalogIndexDefinitions.tableDefinitions( tableDefinitions )
LrmCatalogIndexDefinitions.indexDefinitions( indexDefinitions )

--------------------------------------------------------------------------------

for k, definition in pairs( tableDefinitions ) do
	indexStoreSchema:addTable( definition )
end

for k, definition in pairs( indexDefinitions ) do
	indexStoreSchema:addIndex( definition )
end

local function getIndexSchemaGenerators(  )
	return { 
		'LrmCoreInfoIndexDefinitions',
		'LrmAlbumsIndexDefinitions',
		'LrmCacheIndexDefinitions',
		'LrmSpaceIndexDefinitions',
		'LrmCommentIndexDefinitions',
		'LrmKeywordsIndexDefinitions',
		'LrmPresetsProfilesIndexDefinitions',
		'LrmPeopleIndexDefinitions'
		}
end

--------------------------------------------------------------------------------

function LrmCatalogIndexDefinitions.generateIndexSchema( indexStoreSchema )

	local indexSchemaGenerators = getIndexSchemaGenerators()
	
	for k, generator in pairs( indexSchemaGenerators ) do
		require ( generator ).generateIndexSchema( indexStoreSchema )
	end
	
end

--------------------------------------------------------------------------------

LrmCatalogIndexDefinitions.generateIndexSchema ( indexStoreSchema )

--------------------------------------------------------------------------------

require 'WFCatConflictResolver'.generateIndexSchema( indexStoreSchema )
require 'WFCatErrorHandler'.generateIndexSchema( indexStoreSchema )
require 'WFCatStorage'.generateIndexSchema( indexStoreSchema )

--------------------------------------------------------------------------------

local tablesToReindexByTag = {}

tablesToReindexByTag[ '4.0' ] = { -- LRMA - 4195285
	'coreInfo',
	'persons',
	'similar_person',
	'flags',
	'faces',
	'album_asset',
	'albums',
}

LrmCatalogIndexDefinitions.tablesToReindexByTag = tablesToReindexByTag

LrmCatalogIndexDefinitions.indexStoreSchema = indexStoreSchema

--------------------------------------------------------------------------------

return LrmCatalogIndexDefinitions
