--[[----------------------------------------------------------------------------

LrmCommentIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local KSTableUtils = require "KSTableUtils"

local WFArguments = require 'WFArguments'

local WFOzSyncConstants = require "WFOzSyncConstants"

local LrmCoreInfoIndexDefinitions = require "LrmCoreInfoIndexDefinitions"

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local derefOpt = assert( KSTableUtils.derefOpt )

local assetIsDisplayable = assert( LrmCoreInfoIndexDefinitions.assetIsDisplayable )

--============================================================================--

-- @mark LrmCommentIndexDefinitions namespace

local LrmCommentIndexDefinitions = {}

--------------------------------------------------------------------------------

function LrmCommentIndexDefinitions.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	-- table comment

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE comment( assetId, spaceId, body, created, author_first_name, author_last_name, author_full_name, author_id )]],

		map = function( docValue, emit )

			if docValue and not docValue._deleted
			  and docValue.type == 'comment'
			then
				
				local spm = docValue._serverProvidedMetadata or {}
				local localOnly = docValue._localOnly or {}
				local spaceId = ( docValue.space and docValue.space.id ) or ( localOnly.spaceId )
				local assetId = ( docValue.asset and docValue.asset.id ) or ( localOnly.assetId )
				local created = spm.created or ( localOnly.created )
					-- check _localOnly.created, to allow indexing of comments that we've created locally
					-- but haven't yet gotten synced back from Oz

				local body = docValue.body
				local authorInfo = spm.author or ( localOnly.author )
				local author_first_name, author_last_name, author_full_name, author_id
				if authorInfo then
					author_first_name, author_last_name, author_full_name, author_id =
						authorInfo.first_name,
						authorInfo.last_name,
						authorInfo.full_name,
						authorInfo.id
				
				end
				
				if assetId and created then
					
					if not spaceId then
						
						spaceId = WFOzSyncConstants.kGlobalSpace
					
					end
					
					emit( assetId, spaceId, body, created, author_first_name, author_last_name, author_full_name, author_id )
					
				end
					
			end
						
		end,

	}


	indexStoreSchema:addIndex [[CREATE INDEX comment_asset_space ON comment( assetId, spaceId )]]
	indexStoreSchema:addIndex [[CREATE INDEX comment_space ON comment( spaceId )]]

	-- table flags

	indexStoreSchema:addTable {
		sql = [[CREATE TABLE flags( userId, date, flag )]],
		
		map = function( docValue, emit )
			
			if docValue._deleted then
				return
			end
			if not assetIsDisplayable( docValue ) then
				return
			end
			
			local reviews = derefOpt( docValue, 'reviews' );
			local hasReview = false
			if reviews and type( reviews ) == 'table' then
				for k, v in pairs( reviews ) do
					if v  and type( v ) == 'table' then
						local flag = v.flag
						local userId = k
						local date = v.date
						
						if #userId == 32 and flag ~= nil then
							hasReview = true
							emit( userId, date, flag )
						end
					end
				end
			end
			
			if not hasReview then
				emit( "", "", "unflagged" )
			end
		end
	}

	indexStoreSchema:addIndex [[CREATE INDEX flags_flag ON flags( flag )]]

end

--------------------------------------------------------------------------------

return LrmCommentIndexDefinitions
