--[[----------------------------------------------------------------------------

LrmCoreInfoIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local KSStringUtils = require "KSStringUtils"
local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'

--------------------------------------------------------------------------------

local LrmAssetUtils = require "LrmAssetUtils"

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--============================================================================--

-- @mark LrmCoreInfoIndexDefinitions namespace

local LrmCoreInfoIndexDefinitions = {}

--------------------------------------------------------------------------------

function LrmCoreInfoIndexDefinitions.assetIsDisplayable( docValue )

	local docType = docValue.type
	local subtype = docValue.subtype

	if WFPlatformInfo.isAndroid then
		local video = derefOpt( docValue, 'video' )
		if video then return false end
	end

	if docType ~= 'asset' then return false end
	
	if not ( subtype == 'image' or subtype == 'video' ) then return false end
	
	-- Include if we can get a thumbnail from Oz.
					
	local hasProxy =  derefOpt( docValue, '_serverProvidedMetadata', 'links',
				 '/rels/proxy_type/2560', 'href' )
				
	local hasMaster =  derefOpt( docValue, '_serverProvidedMetadata', 'links',
				 '/rels/master', 'href' )
				
	
	local localOnly = docValue._localOnly
		
	local localProxy = derefOpt( localOnly, 'proxy' )
						
	if localProxy then
		hasProxy = true
	end
	
	local localOriginal = docValue
						and docValue._localOnly
						and docValue._localOnly.original
	
	-- INFO [pkamin 2015-07-08]: we also allow assets with referenced 
	-- only originals
	
	local localUrl = derefOpt( localOnly, 'localUrl' )
	
	if localOriginal or localUrl then
		hasMaster = true
	end
	
	if hasProxy or hasMaster then
		return true
	end
	
	return false

end

--------------------------------------------------------------------------------

function LrmCoreInfoIndexDefinitions.hasThumbnail( docValue )

	if derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/rendition_type/thumbnail2x', 'href' ) then
		return true
	end
                               	  
	if derefOpt( docValue, '_localOnly', 'thumbnail2x', 'md5' ) then
		return true
	end

	return false
	
end   

--------------------------------------------------------------------------------

function LrmCoreInfoIndexDefinitions.coreInfoFromDocValue( docValue )

	local importSource = derefOpt( docValue, 'importSource' )
	local width = importSource and importSource.originalWidth
	local height = importSource and importSource.originalHeight
	
	-- Cropped width and height win.
	
	local develop = docValue.develop
	
	if develop and develop.croppedWidth and develop.croppedHeight then
		width = develop.croppedWidth
		height = develop.croppedHeight
	end
	
	-- Local only thumbnail wins.
	-- No, because this causes a lot of extra computation as large
	-- grid view results are re-sent when one thumbnail is loaded
	-- arahn 05/25/2013
	local thumbnail = derefOpt( docValue, '_localOnly', 'thumbnail2x' )
	
	if width == nil and height == nil and thumbnail and thumbnail.referenceWidth and thumbnail.referenceHeight then
		width = thumbnail.referenceWidth
		height = thumbnail.referenceHeight
	end

	local aspectRatio

	if not docValue._deleted
		and type( width ) == 'number' and width > 0
		and type( height ) == 'number' and height > 0
	then
		aspectRatio = width / height
	end
	
	local thumbnailMD5 = thumbnail and thumbnail.md5
	
	-- FIXME [ngautam 2011-12-13]: Need to replace string.lower by
	-- localization-friendly string conversion.

	local favoritedCount = 0
	local favorites = docValue and docValue.favorites

	if favorites then
		for k, v in pairs( favorites ) do
			if v then
				favoritedCount = favoritedCount + 1
			end
		end
	end

	return aspectRatio, thumbnailMD5, favoritedCount

end

function LrmCoreInfoIndexDefinitions.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	--[[ TODO [immisch 2018-05-25]: for the next big reindexing (after the BlobProvider is activated)
		these fields should be removed:
		- thumbnailMD5
		- hasThumbnail
	]]--

	indexStoreSchema:addTable {
	
		-- Instead of specifying CREATE TABLE SQL, in order to use
		-- niceToHave, you must specify the field names and let the index
		-- store generate the SQL
		tableName = "coreInfo",
		
		fields = {
			"aspectRatio", "captureDate", "displayDate", "timestamp",
			"copyTimestamp", "thumbnailMD5", "favoritedCount", "importedBy", "contentType", "duration", 
			"wasDeveloped", "croppedWidth", "croppedHeight", "fileName", "modifiedDate", "userDefined", 
			"width", "height", "apertureValue", "isoSpeedRatings", "exposureTime", "cameraMaker", 
			"cameraModel", "flag", "fileNameLC", "rating", "copyright", "isVideo", "localUrl", "fnumber",
			"isMergedHDR", 'isCapturedWithAdobeCamera', 'hasThumbnail' },
		
		COLLATE = {
			eventName = 'i18n_collate',
		},
		
		-- query won't send updates for changes that only affect niceToHave
		-- columns
		niceToHave = {
			"thumbnailMD5",
		},

		map = function( docValue, emit )

			if docValue._deleted then
				return
			end
			if not LrmCoreInfoIndexDefinitions.assetIsDisplayable( docValue ) then
				return
			end
			
			local aspectRatio, thumbnailMD5, favoritedCount = LrmCoreInfoIndexDefinitions.coreInfoFromDocValue( docValue )

			local hasThumb = LrmCoreInfoIndexDefinitions.hasThumbnail( docValue )

			local displayDate
			local userDefined
			local timestamp
			local fileName
			local width
			local height
			local copyTimestamp
			local importedBy
			
			local importSource = docValue.importSource
			
			if importSource then
				
				fileName = importSource.fileName
				timestamp = importSource.importTimestamp
				copyTimestamp = importSource.copyTimestamp
				importedBy = importSource.importedBy or ''
				width = importSource.originalWidth
				height = importSource.originalHeight

			end
			
			-- Capture date may have a "Z" appended to the end, signifying that the
			-- timestamp is in UTC (20 chars). Otherwise, the timestamp is treated as
			-- local time (19 chars). Either way, we just care about getting the
			-- YYYY-MM-DD from the beginning of the string.

			if docValue.captureDate and string.len( docValue.captureDate ) > 10 then
				displayDate = string.replace( string.sub( docValue.captureDate, 1, 10 ), ":", "-" )
			end

			local modifiedDate = docValue.userUpdated
			local captureDate = docValue.captureDate
			
			if not modifiedDate then
			
				modifiedDate = derefOpt( docValue, '_serverProvidedMetadata', 'updated' )
	 			local changedAtTime = docValue.changedAtTime
	
				if changedAtTime then
					if modifiedDate then
						if changedAtTime > modifiedDate then
							modifiedDate = changedAtTime
						end
					else
						modifiedDate = changedAtTime
	 				end
	 			end
				
				-- use importTimestamp as fallback for not-yet synced assets
				if not modifiedDate and timestamp then
					modifiedDate = timestamp
				end
			
			end
			
			local fileNameLC = ""
			
			if fileName then
			
				fileNameLC = KSStringUtils.lower( fileName )
				fileNameLC = KSStringUtils.trim( fileNameLC )
				
			end
			
			local contentTypeNumber = LrmAssetUtils.getContentTypeNumber( docValue )
			
			-- Cropped width and height win.
			local croppedWidth = width
			local croppedHeight = height
			
			local develop = derefOpt( docValue, 'develop' )
			if develop and develop.croppedWidth and develop.croppedHeight then
				croppedWidth = develop.croppedWidth
				croppedHeight = develop.croppedHeight
			end
			
			local isVideo = docValue.subtype == 'video'
		
			local durationRatio = derefOpt( docValue, 'video', 'duration' )
			local duration = nil
			if type( durationRatio ) == 'table' then
				duration = math.ceil( ( durationRatio[ 1 ] or 0 ) / ( durationRatio[ 2 ] or 1 ) )
			end
			
			local rating = LrmAssetUtils.getDefaultAssetRating( docValue )
			local flag = LrmAssetUtils.getDefaultAssetReview( docValue )
			local xmp = LrmAssetUtils.getXmp( docValue )
			
			local wasDeveloped = LrmAssetUtils.isDeveloped( docValue )
			local localUrl = derefOpt( docValue, '_localOnly', 'localUrl' )

			local isMergedHDR = derefOpt( docValue, 'xmp', 'aux', 'IsMergedHDR' )
			
			local isCapturedWithAdobeCamera = derefOpt( docValue, 'xmp', 'xmp', 'CreatorTool' )
			isCapturedWithAdobeCamera = isCapturedWithAdobeCamera and type( isCapturedWithAdobeCamera ) == 'string' and
				string.find( isCapturedWithAdobeCamera, 'Adobe Camera' ) == 1 or false

			emit( 	aspectRatio, captureDate, displayDate, timestamp,
					copyTimestamp, thumbnailMD5, favoritedCount, importedBy, contentTypeNumber,
					duration, wasDeveloped, croppedWidth, croppedHeight, fileName, modifiedDate, userDefined, width, height,
					xmp.apertureValue, xmp.isoSpeedRatings, xmp.exposureTime, xmp.cameraMake, xmp.cameraModel, flag, fileNameLC, rating,
					xmp.copyright, isVideo, localUrl, xmp.fnumber, isMergedHDR, isCapturedWithAdobeCamera, hasThumb
			 )
		end,

	}

	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_importedBy ON coreInfo( importedBy, captureDate, docId )]]

	-- Grid sorting criteria which can benefit from from an index.
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_captureDate ON coreInfo( captureDate )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_timestamp ON coreInfo( timestamp )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_modifiedDate ON coreInfo( modifiedDate, captureDate, fileNameLC COLLATE i18n_collate, docId )]]
	indexStoreSchema:addIndex [[CREATE INDEX coreInfo_fileNameLC ON coreInfo( fileNameLC COLLATE i18n_collate  )]]
	-- Note: an index on column 'rating' actually increases sorting time significantly.
	-- This is probably due to it being a "low quality index" - one with more than 10-20 rows having the same value in left most column.

end

--------------------------------------------------------------------------------

return LrmCoreInfoIndexDefinitions
