--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

LrmLibraryCacheModelImp.lua

----------------------------------------------------------------------------]]--

local Rx = require "Rx"

local KSData = require "KSData"
local KSTableUtils = require "KSTableUtils"
local KSPathUtils = require "KSPathUtils"

local WFFileUtils = require "WFFileUtils"
local WFDigest = require "WFDigest"
local WFTableArchiving = require "WFTableArchiving"
local WFArguments = require 'WFArguments'
local WFLogger = require "WFLogger"
local WFRxRunUntilDone = require "WFRxRunUntilDone"

--------------------------------------------------------------------------------

local Model = {}

--------------------------------------------------------------------------------

local logger = WFLogger "LrmLibraryCacheModel"
 :setLevel 'info'

local check = assert( WFArguments.check )
local derefOpt = assert( KSTableUtils.derefOpt )
local runUntilDone = WFRxRunUntilDone

--------------------------------------------------------------------------------

function Model:purgeableCacheSizeForAlbum( albumId )

	check( 'table', 'string' )
	
	local sql = [[ 
		SELECT 
			SUM(size)
		FROM 
			cacheReferences 
		WHERE 
			cacheReferences.hasUploaded == 1 AND
			cacheReferences.renditionName != 'thumbnail2x' AND 
			docId 
		IN ( 
			SELECT 
				album_asset.assetId 
			FROM 
				album_asset
			INNER JOIN 
				albums 
			ON 
				album_asset.albumId == albums.docId
			WHERE 
				album_asset.albumId == ?
			
			EXCEPT 
			
			SELECT 
				album_asset.assetId
			FROM 
				album_asset 
			INNER JOIN 
				albums 
			ON 
				album_asset.albumId == albums.docId
			WHERE 
				albums.useOffline == 1 AND
				album_asset.albumId != ?
		)  ]]
	
	return {
		value =	self.catalog:indexStore():queryP( sql, albumId, albumId )
			:mapDataP_new( function( size )
			
				local WFPlatformInfo = require 'WFPlatformInfo'
				if WFPlatformInfo.isAndroid then
                    return {
						size = size and size[ 1 ] and size[ 1 ][ 1 ] or 0,
						albumId = albumId,		-- android expects albumid as it does not maintain state
					}
				end
				return size and size[ 1 ] and size[ 1 ][ 1 ] or 0
				
			end )
	}
	
end

--------------------------------------------------------------------------------

function Model:purgeableCacheSize()

	check( 'table' )
	
	local sql = [[ 
		SELECT 
			SUM(size)
		FROM 
			cacheReferences 
		WHERE 
			cacheReferences.renditionName != 'thumbnail2x' AND 
			docId 
		IN ( 
			SELECT 
				album_asset.assetId
			FROM 
				album_asset 
			INNER JOIN 
				albums 
			ON 
				album_asset.albumId == albums.docId
			WHERE 
				albums.useOffline == 0 
		) ]]
	
	return {
		value =	self.catalog:indexStore():queryP( sql )
			:mapDataP_new( function( size )
			
				return size and size[ 1 ] and size[ 1 ][ 1 ] or 0
				
			end )
	}
	
end

--------------------------------------------------------------------------------

local function enumerate( dataCache, batchSize )

    local keyStreamSubject = Rx.ReplaySubject(1)
    keyStreamSubject:onNext(nil)

    return keyStreamSubject:flatMap(function(nextKey)
    
        return dataCache:enumerateKeys( nextKey, batchSize )

    end )
    :tap( function(keys)
    
        if not next(keys) then
            keyStreamSubject:onCompleted()
        else
            keyStreamSubject:onNext(keys[ #keys ])
        end
        
    end )

end

--------------------------------------------------------------------------------

local function isValidXmpCameraRaw( xmpCameraRaw )
	-- SPAM [matthewj 2016-08-08]: This isn't a rigorous check of the xmpCameraRaw, but it should catch
	-- the most egregious incorrect xmpCameraRaw. xmpCameraRaw can be either a string with the xmp in it
	-- or a table containing the sha256 of the xmp string for external xmp handling
	local xmpCameraRawType = type( xmpCameraRaw )

	if xmpCameraRawType == 'table' then

		return xmpCameraRaw.sha256 ~= nil

	end

	return xmpCameraRawType == 'string'

end

--------------------------------------------------------------------------------

local function renditionContentId( docValue )

	local xmpCameraRaw = derefOpt( docValue, "develop", "xmpCameraRaw" )

	local userOrientation = derefOpt( docValue, "develop", "userOrientation" )

	-- SPAM [matthewj 2015-12-21]: Ran into issues where the xmpCameraRaw
	-- contained an invalid, non-string value (the image's width value, in
	-- fact) and the md5 calculation would then blow up. So if it's nil or not
	-- a string, then use the default <xmp />
	xmpCameraRaw = isValidXmpCameraRaw( xmpCameraRaw ) and xmpCameraRaw or "<xmp />"

	local xmpCameraRawForDigest = type( xmpCameraRaw ) == 'table'
								and xmpCameraRaw.sha256
								or xmpCameraRaw

	-- Include SHA256 of the parent binary used to create the renders so that we will
	-- generate new renders when the binary's SHA256 changes due to destructive edits.

	local assetSha256 = derefOpt( docValue, 'importSource', 'sha256' )

	-- Include the orientation in the digest so changes to it will kick off new renders.

	local compositeDevSettingsForDigest =
			string.format( "%s%s%s",
					xmpCameraRawForDigest,
					userOrientation or 1,
					assetSha256 )

	return WFDigest.MD5.digest( compositeDevSettingsForDigest )

end

--------------------------------------------------------------------------------

function Model:migrateThumbnailCache()

	-- EXPLAIN QUERY PLAN SELECT docId FROM coreMD5 WHERE thumbnail2x IN ('f165206261368b265588996fc02ede67', 'f72e780ca66a35122524a8ed6e944047' )

	local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'

	local function writeThumbnail( catalogId, strippedDoc, data )
	
		if not strippedDoc then
			return false
		end

		local contentId = renditionContentId( strippedDoc )
		local docId = strippedDoc._id

		if not contentId or not docId then
			logger.errorf( "thumbnail: no contentId found for %s", docId or "unknown")
			return false
		end
			
		local p = KSPathUtils.child( documentsDir, catalogId )
		
		p = KSPathUtils.child( p, 'blobs' )
		p = KSPathUtils.child( p, string.sub( docId, 1, 2 ) )
		p = KSPathUtils.child( p, docId )
	
		if not pcall( WFFileUtils.mkdir, p ) then
			logger.errorf( "could not create asset directory %s for %s", p, docId )
			return false
		end
	
		p = KSPathUtils.child( p, 'th_' .. contentId .. '.jpg' )
	
		if not pcall( KSData.saveToFile, data, p ) then
			logger.errorf( "could not save thumbnail for %s", docId )
		end

		return true

	end

	local createIndex = "CREATE INDEX coreMD5_thumbnail2x ON coreMD5( thumbnail2x ASC )" 
	local dropIndex = "DROP INDEX IF EXISTS coreMD5_thumbnail2x" 

	local fieldSpec = {
		_id = true,
		develop = {
			xmpCameraRaw = true,
			userOrientation = true
		},
		importSource = {
			sha256 = true
		}
	}

	local migratePreviews = self.catalog:indexStore():query { sql = createIndex }
		:take( 1 )
		:flatMap( function()
			return Rx.Observable.combineLatest( function( dataCache, wfCatalog )
				return dataCache, wfCatalog:ozCatalogSpec():catalogId()
			end,
			self.catalog:getDataCache(), self._wfCatalogStream )
		end )
		:flatMap( function( dataCache, catalogId )
			return enumerate( dataCache, 10 )
				:takeWhile( function( keys ) 
					return next( keys )
				end )
				:flatMap( function( keys )
	
					local sql = 'SELECT docId from coreMD5 where thumbnail2x in ' .. WFTableArchiving.tableToString( keys )
				
					-- create a string representation of a list suitable for an SQL IN statement
					sql = string.gsub( sql, '{', '(' )
					sql = string.gsub( sql, ',?%s*}', ')' )					
	
					return self.catalog:indexStore():query { sql = sql }
						:take( 1 )
						:map( KSTableUtils.flatten )
						:flatMap( function( docIds )
							return self.catalog:docStore():bulkDocValue( docIds, fieldSpec )
						end )
						:flatMap( function( strippedDocValues )
						
							local keysAndStrippedDocValues = {}
							for i, k in ipairs( keys ) do
								table.insert(keysAndStrippedDocValues, { k, strippedDocValues[ i ] } )
							end
						
							return Rx.Observable.from( keysAndStrippedDocValues )
								:flatMap( function( keyAndStrippedDocValue )
									return dataCache:get( keyAndStrippedDocValue[ 1 ] )
										:map( function( data )
											return { keyAndStrippedDocValue[ 2 ], data }
										end )
								end )
						end )
						:bufferWithCount( #keys )
						:map( function( strippedDocValuesAndData )
						
							local count = 0
						
							for _, strippedAndData in ipairs(strippedDocValuesAndData) do
							
								if writeThumbnail( catalogId, unpack( strippedAndData ) ) then
									count = count + 1
								end
							
							end
							
							return count
							
						end )

				end )
				:sum()
				:map( function( count )
					logger:infof( "%d thumbnails migrated", count )
				end )
		end )
		:finally( function()
			self.catalog:indexStore():query { sql = dropIndex }
				:flatMap( function()
					return self.catalog:getDataCache()
						-- ultimately, we will drop the data cache entirely here
						:flatMap( function( dataCache )
							return dataCache:removeAll()
								:flatMap( function() 
									return dataCache:vacuum()
								end )
						end )
				end )
				:call( runUntilDone )
		end )

	return {
		value = migratePreviews
	}

end

--------------------------------------------------------------------------------

return Model
