--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local WFUnknown = require 'WFUnknown'

	-- WF.core.os
local WFLocalizationUtils = require 'WFLocalizationUtils'
local KSDate = require 'KSDate'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.oz_client.account
local OzAccountService = require 'OzAccountService'

	-- WF.oz_client.ims
local WFImsAppIdentity = require 'WFImsAppIdentity'

	-- WF.oz_client.sync_engine.user
local WFOzCurrentUser = require 'WFOzCurrentUser'

	-- WF.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- WF.web.utils
local WFURI = require 'WFURI'

	-- WF.web.http_client
local WFHttpClient = require 'WFHttpClient'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

--------------------------------------------------------------------------------

local logger = WFLogger 'LrmLoginModel'
--	:setLevel 'debug'

local loginEventsForDiagnosticLog = {}

local TIMEOUT = 30 -- seconds

local kScope = 'openid,AdobeID'
local kRedirectUri = 'https://fake.domain.adobe.com'

--============================================================================--

-- @mark LrmLoginModel

local LrmLoginModel = {}

--------------------------------------------------------------------------------

local function addLoginEvent( initialData )

	local loginEvent = {}
	
	for k, v in pairs( initialData ) do
		loginEvent[ k ] = v
	end
	
	loginEvent[ ' timeAtStart' ] = KSDate.timeInIso8601Format( nil, true )
		-- Leading space to make this sort to top of list.
	
	loginEventsForDiagnosticLog[ #loginEventsForDiagnosticLog + 1 ] = loginEvent
	
	return loginEvent

end

--------------------------------------------------------------------------------

local function addPromiseResolutionToLoginEvent( loginEvent, name, status, info )

	local resolutionLog = {
		[ ' timeResolved' ] = KSDate.timeInIso8601Format( nil, true ),
		[ ' status' ] = status,
	}
	
	if type( info ) == 'table' then
		for k, v in pairs( info ) do
			resolutionLog[ k ] = v
		end
	end
	
	loginEvent[ name ] = resolutionLog
	
end
--------------------------------------------------------------------------------


local function imsRequestP( args )

	local loginEvent = assert( args.loginEvent )
	--local imsAuthUrlP = assert( args.imsAuthUrlP ) -- unused
	local hasLoginErrorsB = assert( args.hasLoginErrorsB )
	
	local paramsStr = WFURI.buildQueryString( args.params )

	local languageCode = WFLocalizationUtils.getLanguageCode()
	local regionCode = WFLocalizationUtils.getRegionCode()
	
	local overrides = {
		languageCode = languageCode,
		regionCode = regionCode,
	}

	return OzAccountService.getImsAuthUrlP( overrides )
	
		:observeP( function( status, _, imsDomain )

			addPromiseResolutionToLoginEvent( loginEvent, 'getImsAuthUrl', status,
				{
					imsDomain = imsDomain,
				} )
			
		end )

		:mapDataP( function( _, _, imsDomain )

			local requestUrl = string.format( '%s/%s', imsDomain, args.imsUrlPath )

			logger:debugf( "POST %s\n\nparams = %s\n\n", requestUrl, paramsStr )

			return WFHttpClient.putP( requestUrl, paramsStr,
								{
									requestMethod = 'POST',
									requestHeaders = {
										{
											header = 'Content-Type',
											value = 'application/x-www-form-urlencoded',
										},
									},
									timeout = TIMEOUT,
								} )
		
		end )

		:mapP( function( status, rawJson )

			logger:debugf( "HTTP response was %s\n\nbody = %s", status, rawJson )

			local parsingSuccess, jsonResult = pcall( KSJsonDecoder.decodeWebSafeJson, rawJson, true )

			if (parsingSuccess) then
				hasLoginErrorsB:set( false )
			end
			
			if type( jsonResult ) ~= 'table' then
			
				jsonResult = {}
				
				if (parsingSuccess == false) then
				
					hasLoginErrorsB:set( true )
					jsonResult.error = "Unable to login successfully. Please retry"
					
				end
				
			end
			
			addPromiseResolutionToLoginEvent( loginEvent, 'imsPost', status,
				{
					countryCode = jsonResult.countryCode,
					displayName = jsonResult.displayName,
					email = jsonResult.email,
					emailVerified = jsonResult.emailVerified,
					expires_in = jsonResult.expires_in,
					preferredLanguages = jsonResult.preferredLanguages,
					token_type = jsonResult.token_type,
				} )

			if jsonResult.jump then
		
				logger:infof( "received jump URL %s from request to URL %s",
							  jsonResult.jump, args.imsUrlPath )

				return 'data', { jumpUrl = jsonResult.jump }
			
			end

			local err

			if status == 'data' then
								
				local success, result = pcall( args.parseExpectedJson, jsonResult )
				
				if success and result then
					if KSPromises.isPromise( result ) then
						return result
					else
						return 'data', result
					end
				else
					err = result
				end
			
			end
			
			-- Something went wrong. See if we can construct a reasonable
			-- error message.
			
			err = err or jsonResult.error

			if ( not err ) and status == 'error' and type( jsonResult ) == 'string' then
				err = jsonResult
			end
			
			loginEvent.error = err

			logger:infof( "received error %s from request to URL %s",
						  err, args.imsUrlPath )
			
			return 'error', err
		
		end )
		
end

--------------------------------------------------------------------------------

local function generateDiagnosticLogDataForModel( model )

	WFDiagnosticLog.addSectionToLog
		{
			name = "Lightroom mobile Login",
			id = 'LrmLoginModel',
			weight = 4025,
	
			body = {
			
				modelData = {
					deviceId = model.deviceId,
					scope = model.scope,
					locale = model.locale,
					languageCode = model.languageCode,
					regionCode = model.regionCode,
					clientId = model.clientId,
				},
		
				loginEvents = loginEventsForDiagnosticLog,
			
			},
		}

end

--------------------------------------------------------------------------------

local function loginP( loginInfo, loginEvent )

	local accessToken = assert( loginInfo.accessToken )
	local refreshToken = loginInfo.refreshToken
	local deviceToken = loginInfo.deviceToken
	
	return WFOzCurrentUser.loginWithNewTokensP( OzCatalogUrls.baseServerUrl,
												accessToken, refreshToken, deviceToken )

		:observeP( function( status, message )

			addPromiseResolutionToLoginEvent( loginEvent, 'ozLogin', status,
				{
					message = status ~= 'data' and message or nil,
				} )
			
		end )

end

--------------------------------------------------------------------------------

local function signInWithDeviceTokenP( self, model, deviceToken )

  WFImsAppIdentity.setImsDeviceId( deviceToken )
  
	local loginEvent = addLoginEvent {
		event = 'signInWithDeviceTokenP',
	}

	return imsRequestP
		{
			hasLoginErrorsB = self._hasLoginErrorsB,
			imsUrlPath = 'token/v1',
			loginEvent = loginEvent,
			imsAuthUrlP = model.imsAuthUrlP,
			params = {
				grant_type = 'device',
				device_id = model.deviceId,
				client_id = model.clientId,
				client_secret = model.clientSecret,
				scope = model.scope,
				locale = model.locale,
				languageCode = model.languageCode,
				regionCode = model.regionCode,
				device_token = deviceToken,
			},
			
			parseExpectedJson = function( jsonResult )
			
				if jsonResult.access_token and jsonResult.refresh_token then
					return
						{
							accessToken = jsonResult.access_token,
							refreshToken = jsonResult.refresh_token,
							deviceToken = deviceToken,
							userId = jsonResult.userId
						}
				end
				
				if jsonResult.error then
					return
						{
							error = jsonResult.error
						}
				end

	
			end,
			
		}

		-- If we got valid token here, attempt to log in to Oz now.

		:mapDataP( function( _, loginInfo )

			logger:debugf( "processing loginInfo = %s", loginInfo )

			if type( loginInfo ) ~= 'table' or not loginInfo.accessToken then
				return 'error', "Incomplete login info from IMS", loginInfo
			end

			return loginP( loginInfo, loginEvent )
				:mapDataP( function()
				
					return 'data', loginInfo
				
				end )
		
		end )
		
end

--------------------------------------------------------------------------------

function LrmLoginModel.make()

	local clientId, clientSecret = WFImsAppIdentity.getClientIdAndSecret()

	local hasLoginErrorsB = KSReactive.variableB( false )
	
	local languageCode = WFLocalizationUtils.getLanguageCode()
	local regionCode = WFLocalizationUtils.getRegionCode()
	local locale = string.format( '%s_%s', languageCode, regionCode )
	
	local deviceId = WFImsAppIdentity.getImsDeviceId()
			
	local loginParams =
		{
			response_type = 'device',
			device_id = deviceId,
			client_id = clientId,
			scope = kScope,
			locale = locale,
			idp_flow = 'login',
			redirect_uri = kRedirectUri,
		}
		
	local imsAuthUrlP = OzAccountService.getImsAuthUrlP
		{
			languageCode = languageCode,
			regionCode = regionCode,
		}
		:runUntilDoneP()
		
	local modelAttributes =
		{
			deviceId = deviceId,
			scope = kScope,
			locale = locale,
			languageCode = languageCode,
			regionCode = regionCode,
			clientId = clientId,
			clientSecret = clientSecret,
			imsAuthUrlP = imsAuthUrlP,
		}
	
	local modelResult = {
	
		_hasLoginErrorsB = hasLoginErrorsB,
		
		value = {
		
			loginUrl = imsAuthUrlP
				:untilResolvedB( 'data', WFUnknown.value() )
				:closingMapB( function( authUrl )
				
					if type( authUrl ) == 'string' then
						authUrl = authUrl:match( "^(.*)%?" ) -- strip existing query args
						return authUrl .. '/?'
							.. WFURI.buildQueryString( loginParams )
					end
					
					return WFUnknown.value()
				
				end ),
				
			redirectUrl = kRedirectUri,
			
			hasLoginError = hasLoginErrorsB,
				
		},
		
		events = {
		
			signInWithDeviceToken = function( self, deviceToken )
				signInWithDeviceTokenP( self, modelAttributes, deviceToken )
					:runUntilDone()
			end,
		},
		
		attributes = modelAttributes,
	}
	
	WFDiagnosticLog.addLogSectionFunc( function()
		generateDiagnosticLogDataForModel( modelAttributes )
	end )
	
	return modelResult
	
end

return LrmLoginModel
