--[[----------------------------------------------------------------------------

LrmOzPurgeAgent.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class monitors the content of a single catalog's data cache and periodically
purges it.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local ReplaySubject = Rx.ReplaySubject

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua
local KSData = require 'KSData'
local KSLuaUtils = require "KSLuaUtils"

	-- WF.core.os
local WFFileUtils = require 'WFFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- WF.core.rxlua
-- local WFRxScheduler = require 'WFRxScheduler'

local WFArguments = require 'WFArguments'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local KSTableUtils = require 'KSTableUtils'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local KSIdGenerator = require "KSIdGenerator"

	-- WF.os.prefs
-- local KSPrefs = require 'KSPrefs'
local WFPlatformInfo = require 'WFPlatformInfo'

local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'

	-- WF.oz.purge_agent
local WFOzPurgeSettings = require 'WFOzPurgeSettings'

	-- WF.oz_client.sync_engine
-- local WFOzBinaryUploadUtils = require 'WFOzBinaryUploadUtils'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

local LrmConfig = require "LrmConfig"

local LrmTHLibrary
if not WFPlatformInfo.isAndroid then
    local _, LrmTHLibrary = KSLuaUtils.requireIfExists( 'LrmTHLibrary' )
end
--------------------------------------------------------------------------------

LrmConfig.registerBinaryStrategyCheck()

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

local cachePollingInterval = 40	-- seconds; feel free to tune

local minimumFreeSpace = WFOzPurgeSettings.minimumFreeSpace()

local logger = WFLogger 'LrmOzPurgeAgent'
	-- :setLevel 'DEBUG'

local derefOpt = KSTableUtils.derefOpt

--------------------------------------------------------------------------------

local getDiskSpaceNeeded = function( self )
	return self._dataCachePathStream
		:map( function( path )
				if WFPlatformInfo.isAndroid then
					path = KSPathUtils.parent(path)
				end

				local fileSystemFreeSpace, _ -- fileSystemSize
				= WFFileUtils.freeSpaceAtPath( path )

			local requiredSpace = math.max(0, minimumFreeSpace - fileSystemFreeSpace)
			return requiredSpace
		end )
end	


--============================================================================--
--[[!
	\brief A \c LrmOzPurgeAgent monitors the content of a single catalog's data
	cache and periodically purges it to make additional disk space available.
	
	It should be used only by LrmSyncClientImp.	
--]]

-- @mark LrmOzPurgeAgent

local purgeAgentMethods = {
	
	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (OzIndexStore)
	
	_dataCacheStream = false,
		-- (Observable< KSDataCache >) data cache for this doc store
	
	_dataCachePathStream = false,
		-- (Observable< string >) path to cache
	
	_purgeEpoch = false,
		-- An IndexStore query listing purgeable assets by epoch
	
	_purgeAll = false,
		-- An IndexStore query listing all purgeable	
	
	_purgeOnDeleteChain = false,
		-- the reactive chain that is executed on purgeOnDelete (exposed for testing)
	
	_purgeOnDelete = false,
		-- Subscription to the docStore's document deleter
	
	_checkCacheKeysFrom = false,
		-- (false or string) where do we start cross-checking cache keys?

	_warnedMissingEntries = false,
		-- (table) entries that we've warned about (so we don't keep bleating about them)

	_binaryStrategy = false,
		-- (string) LrmConfig

	_purgeSub = false,
		-- A subscription to _purgeEpoch used to continously purge unused renditions.

	_catalogId = false,
		-- (string) CatalogId

	_deleteFeedCallback = false,
		-- A subscription to send callback once delete operation is successful.

	private = {
		
		-- The following are methods are internal, but are exposed for testing.
		_getDiskSpaceNeeded = getDiskSpaceNeeded
	},

	WFCloseSupport.addCloseSupportMethods,

}


--------------------------------------------------------------------------------

function purgeAgentMethods:sendDiagnosticLogData()

	return {

		name = "Lrm Purge Agent",
		id = 'LrmOzPurgeAgent',
		weight = 521,

		body = {
			binaryStrategy = self._binaryStrategy,	
		}
	}
	
end

--------------------------------------------------------------------------------

local function purgeableAlbumAssetListSql()

	return [[
		SELECT docId, renditionName, md5, size 
		FROM cacheReferences 
		WHERE 
			cacheReferences.hasUploaded == 1
			AND cacheReferences.renditionName != ?
			AND docId IN ( 
				SELECT 
					album_asset.assetId 
				FROM 
					album_asset
				INNER JOIN 
					albums 
				ON 
					album_asset.albumId == albums.docId 
				WHERE 
					album_asset.albumId == ?)
			AND (docId NOT IN ( 
					SELECT 
					   album_asset.assetId 
					FROM 
					   album_asset 
					INNER JOIN 
					   albums 
					ON 
						   album_asset.albumId == albums.docId 
					WHERE 
						   albums.useOffline == 1 
					)
				)
		   LIMIT 100 ]]
end
 
 --------------------------------------------------------------------------------

 local function purgeEpochSql()
 
	 return [[ 
		SELECT docId, renditionName, md5, size 
		FROM cacheReferences
			WHERE epoch < ( SELECT max( epoch ) FROM cacheReferences )
				AND cacheReferences.hasUploaded == 1
				AND cacheReferences.renditionName != 'thumbnail2x'
				AND (docId NOT IN (
					SELECT 
						album_asset.assetId 
					FROM 
						album_asset 
					INNER JOIN 
						albums 
					ON 
						album_asset.albumId == albums.docId 
					WHERE 
						albums.useOffline == 1 
				)
			)
			ORDER BY epoch ASC, size DESC LIMIT 100 ]]
end

 --------------------------------------------------------------------------------

local function purgeAllSql()

	return [[ 
		SELECT docId, renditionName, md5, size 
		FROM cacheReferences 
			WHERE cacheReferences.hasUploaded == 1
				AND cacheReferences.renditionName != ?
				AND (docId NOT IN (
					SELECT 
						album_asset.assetId 
					FROM 
						album_asset 
					INNER JOIN 
						albums 
					ON 
						album_asset.albumId == albums.docId 
					WHERE 
						albums.useOffline == 1 
				)
			)
			LIMIT 100 ]]

end

--------------------------------------------------------------------------------

local function setBinaryStrategy( self, binaryStrategy )

	if self._binaryStrategy ~= binaryStrategy then

		self._binaryStrategy = binaryStrategy

		self._purgeEpoch = self._indexStore:query {
			sql = purgeEpochSql(),
			quiesceUntilUpToDate = true,
			calm = cachePollingInterval
		}

	        local cacheReferenceNotToConsider = ''
	        if WFPlatformInfo.isAndroid then
	            cacheReferenceNotToConsider = 'thumbnail2x'
	        end
		self._purgeAll = self._indexStore:query {
			sql = purgeAllSql(),
			queryId = 'purgeAll',
			quiesceUntilUpToDate = true,
            		cacheReferenceNotToConsider
		}

	end

end

--------------------------------------------------------------------------------

local function cacheEntriesForDocument( docValue )

	local renditionNames = { 'thumbnail', 'thumbnail2x', 'preview' }
	local keys = {}
	
	for _, renditionName in ipairs( renditionNames ) do
	
		local md5 = derefOpt( docValue, '_localOnly', renditionName, 'md5' )
	
		if md5 then
			table.insert( keys, md5 )
		end
	end

	return keys

end

--------------------------------------------------------------------------------

local function removeAssets( self, sql, ... )

	local queryId = 'removeAssets' .. KSIdGenerator.localCounter()

	return self._indexStore:query{
			sql = sql,
			quiesceUntilUpToDate = true,
			queryId = queryId,
			...
		}
		:takeWhile( function( docIds )
			return next( docIds )
		end )
		:flatMap( function( docIds )
			
			local bulkOverlay = { label = 'oz' }

			for i, r in ipairs( docIds ) do
		
				table.insert(bulkOverlay, {
					_id = r[ 1 ],
					_deleted = true
				} )
			end
					
			return self._docStore:bulkApplyOverlay( bulkOverlay )
				:map( function( revs )
					return #revs
				end )	
		end )
		:tap(function()
		
			self._indexStore:requestFromQuery( queryId )

		end )
		:sum()
end 

--------------------------------------------------------------------------------

local function setupPurgeOnDelete( self )

	self._purgeOnDeleteChain = self._docStore:deletedDocStream()

		:filter( function( docValue )
			-- deleted assets from Oz have no type information by this point, so
			-- pass those through for deletion just in case they were assets/albums/spaces

			return docValue.type == "asset" or docValue.type == 'album' or docValue.type == 'space' or docValue.type == nil
		end )
		
		:flatMap( function( docValue )
			-- Immediately purge all files associated with that document
			-- This operation is harmless if the document isn't an asset
			self:purgeDeletedDocument( docValue )

			-- If the LrmTHLibrary is available, tell it a document was deleted
			if LrmTHLibrary and LrmTHLibrary.purgeDocument then
				LrmTHLibrary.purgeDocument( docValue )
			end

			-- Now assume it is an album
			return self:purgeForAlbum( docValue._id )
				:flatMap( function ()
			
					return removeAssets( self,
						[[ SELECT docId FROM album_asset WHERE albumID = ? LIMIT 100]],
						docValue._id )

				end )
				:defaultIfEmpty( docValue )
				:map( function()
					return docValue
				end )
		end )

		:flatMap( function( docValue )

			return removeAssets( self,
				[[SELECT docId FROM comment WHERE spaceId = ?
					UNION SELECT docId FROM spacefavorite where spaceId = ? LIMIT 100]],
					docValue._id,
					docValue._id)
				:map( function()
					return docValue
				end )

		end )
		:flatMap( function( docValue )

			local cacheEntries = cacheEntriesForDocument( docValue )

			if ( next( cacheEntries ) ) then

				return self._dataCacheStream
					:flatMap( function( dataCache )

						-- remove the cache entries
						return dataCache:removeBulk( cacheEntries )
					
					end )
			end

			return Rx.Observable.just( 0 )

		end )
		:shareReplay( 1 )

	self._purgeOnDelete = self._purgeOnDeleteChain:subscribe()
	
	-- value produced is the number of entries removed from the cache
end 

--------------------------------------------------------------------------------

--[[!
	\brief Initialize a purge agent for a specific catalog.
	
	\param catalogService (OzCatalogServiceViaWFCatalog) catalog service
	\param binaryStrategy (LrmConfig) prefer proxy or master
--]]

function purgeAgentMethods:__init( catalogService, binaryStrategy )

	check( 'LrmOzPurgeAgent', 'OzCatalogService', '?binaryStrategy' )
	
	self._docStore = catalogService:docStore()
	self._indexStore = catalogService:indexStore()
	
	self._dataCacheStream = catalogService:getDataCache()
		:map(function (dataCache)
			return dataCache
		end )
		:shareReplay( 1 )

	self._dataCachePathStream = catalogService:getDataCachePath()
	self._dataCachePathStream:call( runUntilDone )

    catalogService:getCatalogIdP()
    :observeP( function( _, catalogId )

        self._catalogId = catalogId

    end )
	:runUntilDoneP()

    self._deleteFeedCallback = ReplaySubject( 1 )

	self._warnedMissingEntries = {}
	
	if binaryStrategy then
		setBinaryStrategy( self, binaryStrategy )
	end

	setupPurgeOnDelete( self )

	WFDiagnosticLog.addLogProvider( self )
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:getDeleteFeedCallback( )

    return self._deleteFeedCallback

end

--------------------------------------------------------------------------------

function purgeAgentMethods:close()

	local purgeOnDelete = self._purgeOnDelete
	self._purgeOnDelete = nil
	
	if purgeOnDelete then
		purgeOnDelete:dispose()
	end

	self:stopContinuousPurge()

end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeDeletedDocument( docValue )

	local original = derefOpt( docValue, '_localOnly', 'originalFilePath' )
	if original then
		WFFileUtils.delete( original )
	end

	local proxy = derefOpt( docValue, '_localOnly', 'proxy', 'filePath' )
	if proxy then
		WFFileUtils.delete( proxy )
	end
	
	local externalXmp = derefOpt( docValue, '_localOnly', 'files', 'xmp_develop', 'path' )
	if externalXmp then
	
		local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
	
		WFFileUtils.delete( KSPathUtils.child( documentsDir, externalXmp ) ) 
	end

	local presetProfile = KSTableUtils.derefOpt( docValue, "_localOnly", "files", "presetProfile", "path" )

	if WFPlatformInfo.isAndroid and presetProfile then

		local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'

		WFFileUtils.delete( KSPathUtils.child( KSPathUtils.child( documentsDir, self._catalogId) , presetProfile ) )

		if self._deleteFeedCallback then
			self._deleteFeedCallback:onNext( true )
		end

	end

end

--------------------------------------------------------------------------------

--[[

\brief prepare purging of documents

\param victims as an array of { docId, renditionName, md5, size }
	   (the output of an index store query)
\param spaceNeeded: spaceNeeded to satisfy minimumFreeSpace threshold or
		-1 for no purge limit

\returns a table with
	- docs:  changed documents
	- removeFiles: files that can be removed
	- removeCache: cache entries to purge
	- spaceReclaimed: space reclaimed
	- purged: number of purged documents
]]--

local function preparePurge( victims, spaceNeeded )

	logger:infof( "purgeVictims\nvictims = %s\nspaceNeeded = %s",
				   victims, spaceNeeded )

	local result = {
		docs = {},
		removeFiles = {},
		removeCache = {},
		spaceReclaimed = 0,
		purged = 0
	}

	if spaceNeeded == 0 or not next(victims) then
		return result -- we're done
	end

	for i, victim in ipairs( victims ) do

		local renditionName = victim[ 2 ]
			-- Hard-coded indices are based on indexStore queries purgeAllSql, purgeEpochSql 
			-- and purgeableAlbumAssetListSql
		
		local overlay = {
			type = 'asset',
			_id = victim[ 1 ]
		}

		overlay._localOnly = {
			[ renditionName ] = { _delete = true }
		}

		
		-- Also remove the local path so that the original can be downloaded again
		-- if the user turns store originals on.

		if renditionName == 'original' then
			overlay._localOnly.originalFilePath = { _delete = true }
		end
		
		if renditionName == 'proxy' then
			overlay._localOnly.proxy = { _delete = true }
		end

		if renditionName == 'original' or renditionName == 'latestOriginal' or renditionName == 'proxy' then			
			result.removeFiles[ #result.removeFiles + 1 ] = victim[ 3 ]
		else
			result.removeCache[ #result.removeCache + 1 ] = victim[ 3 ]
		end
	
		result.docs[ i ] = overlay
	
		result.spaceReclaimed = result.spaceReclaimed + victim[ 4 ]
		result.purged = result.purged + 1

		if ( spaceNeeded ~= -1 ) and ( result.spaceReclaimed >= spaceNeeded ) then
			return result
		end

	end
			
	return result

end


--------------------------------------------------------------------------------

local collectUnreferencedKeys -- forward declaration

--------------------------------------------------------------------------------

local function purgeVictims(self, victims, spaceNeeded )

	local purged = 0

	local batch = preparePurge( victims, spaceNeeded )
	
	return self._docStore:bulkApplyOverlay( batch.docs )
		:flatMap( function()
		
			purged = purged + batch.purged
			if spaceNeeded ~= -1 then
				spaceNeeded = math.max(0, spaceNeeded - batch.spaceReclaimed)
			end
		
			-- remove the files
		
			for i, fileName in ipairs(batch.removeFiles) do
				WFFileUtils.delete( OzCatalogUploadStatus.checkAndCorrectLocalPath( fileName )  )
			end
			
			if next( batch.removeCache ) then
			
				return self._dataCacheStream
					:flatMap( function( dataCache )
					
						-- remove the cache entries

						return dataCache:removeBulk( batch.removeCache )

					end )
					:map( function ()
					
						return purged, spaceNeeded
					
					end )
	
			end
			
			return Rx.Observable.just( purged, spaceNeeded )
									
		end )

end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeUntilDone( query, spaceNeeded, queryId )

	return query
	
		:takeWhile( function( victims )
		
			return next( victims )
		
		end )
	
		:flatMap( function( victims )
						
			return purgeVictims( self, victims, spaceNeeded )
			
		end )

		:tap(function()
		
			self._indexStore:requestFromQuery( queryId )

		end )
		
		:map( function (_purged, _spaceNeeded)

			spaceNeeded = _spaceNeeded
	
			return _purged
			
		end )
		
		:sum()
		
		:flatMap( function( sum )
		
			return collectUnreferencedKeys( self )
				:map( function ()
					return sum
				end )

		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Triggers an immediate purge cycle.
]]--

function purgeAgentMethods:purgeImmediately( binaryStrategy, noPurgeLimit )

	check( 'LrmOzPurgeAgent', '?binaryStrategy', '?boolean' )

	if binaryStrategy then
		setBinaryStrategy( self, binaryStrategy )
	end

	local spaceRequired = minimumFreeSpace
	if noPurgeLimit then
		spaceRequired = -1
	end

	return self:purgeUntilDone( self._purgeAll, spaceRequired, 'purgeAll' )
		
end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeForAlbum( albumId, binaryStrategy, noPurgeLimit )
	
	check( 'LrmOzPurgeAgent', 'uuid', '?binaryStrategy', '?boolean' )

	logger:debugf( "purgeForAlbum, %s", albumId )
	
	local indexStore = self._indexStore
	local docStore = self._docStore
	
	local albumDoc = docStore:docValue( albumId )
	
	local spaceRequired = minimumFreeSpace
	if noPurgeLimit then
		spaceRequired = -1
	end

	-- IMPORTANT!!
	--
	-- Verify offline use is still disabled for the album otherwise we are gonna ddos Oz
	-- as the PurgeAgent will try to remove files locally whereas the BinaryDownloader 
	-- will download again
	
	local cacheReferenceNotToConsider = ''
	if WFPlatformInfo.isAndroid then
		cacheReferenceNotToConsider = 'thumbnail2x'
	end
	local result = albumDoc
		:flatMap( function( docValue )
	
			local useOffline = derefOpt( docValue, '_localOnly', 'useOffline' )
			
			if useOffline == false then
			
				return indexStore:query { 
						sql = purgeableAlbumAssetListSql(),
						quiesceUntilUpToDate = true,
						queryId = 'purgeAlbum',
						cacheReferenceNotToConsider,
						albumId
					}
				
			else
			
				return Rx.Observable.just( {} )
				
			end
			
		end )
	
	return self:purgeUntilDone( result, spaceRequired, 'purgeAlbum' )
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:stopContinuousPurge()

	check( 'LrmOzPurgeAgent' )

	if self._purgeSub then
		self._purgeSub:dispose()
		self._purgeSub = nil
	end
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:startContinuousPurge( binaryStrategy )

	check( 'LrmOzPurgeAgent', 'binaryStrategy' )

	if self._purgeSub and binaryStrategy == self._binaryStrategy then
		-- if we're already running and the binary strategy hasn't changed, we're done
		return
	end

	if self._purgeSub then
		self:stopContinuousPurge()
	end

	setBinaryStrategy( self, binaryStrategy )

	self._purgeSub = Rx.Observable.combineLatest(function( potentialVictims, spaceNeeded)
		purgeVictims( self, potentialVictims, spaceNeeded )
			:call( runUntilDone )
	end,
	self._purgeEpoch,
	self.private._getDiskSpaceNeeded( self ))
		:subscribe()
		
	print( self._purgeSub )

end

--------------------------------------------------------------------------------

-- FIXME [scouten 2011-08-12]: We should also sniff for unreferenced originals
-- and missing originals.

--------------------------------------------------------------------------------

-- write asset with thumbnail, then delete asset

local function discoveredMissingEntry( self, k, photoId )

	-- FIXME [scouten 2011-08-29]: Would like to know why the enumerate API
	-- is causing false alarms just after importing photos, but since this is
	-- the only place we enumerate the data cache, it can wait IMHO.
	
	logger:debugf( "looks like cache entry %s for photo %s is missing ... checking",
					k, photoId )


	self._dataCacheStream
		:flatMap( function( dataCache )

			return dataCache:get( k )
			
		end )
	
		:map( function( data )
				
			if data and KSData.length( data ) > 0 then
					
				logger:infof( "checked cache entry %s for photo %s and it was OK",
							   k, photoId )

				data:close()
			end
				
		end )
		:tapOnError(function( message )
			logger:warnf( "checked cache entry %s for photo %s and it was missing with message = %s",
								   k, photoId, message )
					
			self._warnedMissingEntries[ k ] = true

			-- FIXME [scouten 2011-08-29]: Yank the entry from the
			-- doc DB entry.
		end )
						
		:call( runUntilDone )
end

--------------------------------------------------------------------------------



function collectUnreferencedKeys( self )

	local optStartPath = self._checkCacheKeysFrom or nil
	
	local indexStore = self._indexStore
	
	local keysFromCacheAsSet = {} -- filled in below
	
	local keysFromCache = self._dataCacheStream
		:flatMap( function( dataCache )
			
			return dataCache:enumerateKeys( optStartPath )			
			
		end )
	
		:map( function( data )
		
			-- If we didn't get any keys from this query, then we've cycled
			-- through all keys.

			if not data or #data == 0 then
				self._checkCacheKeysFrom = false
				return false, optStartPath
			end
			
			local minKey = optStartPath
			local maxKey = data[ #data ]
			self._checkCacheKeysFrom = maxKey
			
			logger:infof( "\n\nminKey = %s" ..
						   "\n\nmaxKey = %s\n\n",
								minKey, maxKey )

			for i = 1, #data do
				keysFromCacheAsSet[ data[ i ] ] = true
			end
			
			return true, minKey, maxKey
			 
		end )
		
	local keysFromCatalog = keysFromCache
		:flatMap( function( hadKeys, minKey, maxKey )

			local renditions = "( 'thumbnail', 'thumbnail2x', 'preview' )"

			if hadKeys then
			
				logger:infof( "hadKeys == true" )
			
				return indexStore:query{ 
						sql = [[SELECT md5, docId FROM cacheReferences
							WHERE md5 >= ? AND md5 <= ? AND renditionName IN ]] .. renditions, 
						minKey or 0, 
						maxKey, 
						quiesceUntilUpToDate = true 
					}
					:take( 1 )
			
			elseif optStartPath then

				logger:infof( "hadKeys == false; finishing range from %s", optStartPath )
			
				return indexStore:query { 
						sql = [[SELECT md5, docId FROM cacheReferences
							WHERE md5 > ? AND renditionName IN ]] .. renditions, 
						optStartPath, 
						quiesceUntilUpToDate = true 
					}
					:take( 1 )
				
			else

				logger:infof( "looks like cache was completely empty" )

				return indexStore:query{ 
						sql = [[SELECT md5, docId FROM cacheReferences WHERE renditionName IN ]] .. renditions,
						quiesceUntilUpToDate = true
					}
					:take( 1 )
			
			end
			
		end )
		
		:map( function( data )
	
			local keysFromCatalogAsSet = {}
			for i = 1, #data do
				keysFromCatalogAsSet[ data[ i ][ 1 ] ] = data[ i ][ 2 ]
			end
			
			return keysFromCatalogAsSet
		
		end )

	-- Now that we have both key sets, cross-reference them.
	
	local unreferencedKeys= {} -- filled in below

	local unfulfilledKeysFromCatalogAsSet = {} -- filled in below
	
	local crossReferenceKeys = keysFromCatalog
		:map( function( keysFromCatalogAsSet )
	
			for k in pairs( keysFromCacheAsSet ) do
				if not keysFromCatalogAsSet[ k ] then
					logger:debugf( "flagging cache entry %s for deletion", k )
					unreferencedKeys[ #unreferencedKeys + 1 ] = k
				end
			end
			
			for k, photoId in pairs( keysFromCatalogAsSet ) do
				if not keysFromCacheAsSet[ k ] then
					if not self._warnedMissingEntries[ k ] then
						discoveredMissingEntry( self, k, photoId )
					end
					unfulfilledKeysFromCatalogAsSet[ k ] = photoId
				end
			end
				
			logger:infof( "\n\nunreferencedKeys = %s" ..
						   "\n\nunfulfilledKeysFromCatalogAsSet = %s\n\n",
						   unreferencedKeys,
						   unfulfilledKeysFromCatalogAsSet )

			return
		
		end )
	
	-- Now go kill off cache entries that are unreferenced.
	
		:flatMap( function()
	
			if next( unreferencedKeys ) then
	
				return self._dataCacheStream
					:flatMap( function( dataCache )

						return dataCache:removeBulk( unreferencedKeys )
		
					end )
			end
			
			return Rx.Observable.just()
			
		end )

	-- FIXME [scouten 2011-08-11]: Should also fix up references to
	-- missing data in cache. Be careful not to toss references to
	-- originals when doing so.
	
	return crossReferenceKeys

end

--------------------------------------------------------------------------------


return Rx.Class 'LrmOzPurgeAgent'( purgeAgentMethods )
