--[[----------------------------------------------------------------------------

WFOzUserLogin.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFLocalizationUtils = require 'WFLocalizationUtils'

	-- WF.core.rxlua
local WFRxPromiseBridge = require 'WFRxPromiseBridge'
local WFRxRetryObservable = require 'WFRxRetryObservable'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCapabilities = require 'WFCapabilities'

	-- WF.oz_client.account
local WFOzClient = require 'WFOzClient'

	-- WF.oz_client.ims
local WFImsAppIdentity = require 'WFImsAppIdentity'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

	-- WF.web.http_client
local WFRxHttpClient = require 'WFRxHttpClient'

	-- WF.web.utils
local WFURI = require 'WFURI'


--------------------------------------------------------------------------------

local userLoginLog = WFLogger 'WFOzUserLogin'
--	:setLevel( 'DEBUG' )

local check = assert( WFArguments.check )

local asKSPromise = WFRxPromiseBridge

local runUntilDone = WFRxRunUntilDone


--============================================================================--

-- @mark WFOzUserLogin namespace

local WFOzUserLogin = {}

--------------------------------------------------------------------------------

local function loginFromTokens_imp( baseServerUrl, accessToken,
									refreshToken, optDeviceToken )

	userLoginLog:infof( "loginFromTokens( accessToken = %s, refreshToken = %s, "
						.. "optDeviceToken = %s )",
						accessToken, refreshToken, optDeviceToken )


	if optDeviceToken then
		require 'KSKeychainService'.writeP( 'ims_device_token', optDeviceToken )
	end

	local registrationPostParams = {
		requestMethod = 'POST',
		requestHeaders = {
			{ header = 'Authorization', value = 'Bearer ' .. accessToken },
			{ header = 'Content-Type', value = 'application/json' },
		}
	}

	local authHeader = {
		requestHeaders = {
			{ header = 'Authorization', value = 'Bearer ' .. accessToken },
--			{ header = 'x-token-ttl', value = '30' },
				-- [scouten 2011-07-01]: Normally this should be commented out.
				-- Useful when testing the refresh logic. This specifies the lifetime
				-- of the access token in seconds.
		}
	}

	local registrationUrl = WFURI.resolve( baseServerUrl, '/v1.0/registrations' )

	return WFRxHttpClient.put( registrationUrl, '{}', registrationPostParams )

		:tap( function( message )
			userLoginLog:infof( "registration call succeeded with message = %s", message )
		end,
		function( message )
			userLoginLog:errorf( "registration call failed with message = %s", message )
		end )

		:map( function()

			-- TODO [scouten 2015-07-17]: We should ignore status of registration call
			-- but for now we don't. (Rx gurus: Advice?)

			return WFRxHttpClient.get( WFURI.resolve( baseServerUrl,
													  '/v1.0/accounts/00000000000000000000000000000000' ),
									   authHeader )

		end )

		:switch()

		:map( KSJsonDecoder.decodeWebSafeJson_stripNull )

		:map( function( userInfo )

			local userId = userInfo and userInfo.id and userInfo.id:sub( -32 )

			if userId then
				userLoginLog:infof( "token corresponds to user %s", userInfo )
			else
				userLoginLog:errorf( "info packed from Oz was missing user ID\n\n%s\n\n", userInfo )
				error "info packet from Oz was missing user ID"
			end

			return userId, userInfo, baseServerUrl, accessToken, refreshToken, optDeviceToken

		end )

end

--------------------------------------------------------------------------------

local function returnRxUserAccount( userId,
									userInfo,
									baseServerUrl,
									accessToken,
									refreshToken,
									optDeviceToken )

	-- Spin up WFCatUser account and record the initial user info packet.

	local catalogType = WFCapabilities.getCapability( "WFOzCatalog.subtype" )
		or 'lightroom_v2'

	local wfCatUser = require 'WFCatUser'( baseServerUrl, catalogType, userId )

	local userInfoCache = wfCatUser:userInfoCache()

	userInfo._id = userId
	userInfo.type = 'user'

	userLoginLog:debugf( "\n\nuserInfoCachePath = %s\n\n", wfCatUser:userInfoCachePath() )
	userLoginLog:debugf( "\n\nwriting info packet to userInfoCache %s\n\n", userInfo )

	-- TODO [scouten 2015-07-18]: Use :setDocValue instead?
	-- Might need to add a _force option if so.

	return userInfoCache:applyOverlay( userInfo )
		:map( function()
			return wfCatUser, userInfo, accessToken, refreshToken, optDeviceToken
				-- TODO [scouten 2015-07-22]: Five args is a bit unwieldy.
				-- Change to a table?
		end )

end

--------------------------------------------------------------------------------

local function returnClassicUserAccountP( status,
										  userId,
										  userInfo,
										  baseServerUrl,
										  accessToken,
										  refreshToken,
										  optDeviceToken )

	-- Spin up WFOzUser account and record the initial user info packet.

	local wfOzUser = require 'WFOzUser'( baseServerUrl, userId )

	local userInfoCache = wfOzUser:userInfoCache()

	userInfo._id = userId
	userInfo._slot = 'main'
	userInfo.type = 'user'

	userLoginLog:debugf( "\n\nuserInfoCachePath = %s\n\n", wfOzUser:userInfoCachePath() )
	userLoginLog:debugf( "\n\nwriting info packet to userInfoCache %s\n\n", userInfo )

	return userInfoCache:setDocValueP( userInfo )
		:mapDataP( function()

			return 'data', wfOzUser, userInfo

		end )

		:observeP( function( status, wfOzUser )

			if status == 'data' then

				-- Store new tokens from account.

				local tempOAuth2Session = require 'KSOAuth2Session'( wfOzUser:userId() )

				tempOAuth2Session:storeTokensP( accessToken, refreshToken, optDeviceToken )
					:whenResolvedDo( function()
						tempOAuth2Session:close()
					end )
					:runUntilDone()

			end

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Given a set of IMS tokens, establishes a connection to Oz.

	Call when a user has successfully authenticated against the IMS server.
	Triggers a call to the Oz server to obtain further user information
	(such as user name, subscription status, and so on).

	\param baseServerUrl (string) The base URL for the Oz server to use.
	\param accessToken (string) The new access token.
	\param refreshToken (string) The new refresh token.
	\param optDeviceToken (string or nil) The device token, if applicable

	\return Observable< WFCatUser, info packet > user object and table containing user info
]]--

function WFOzUserLogin.loginFromTokens( baseServerUrl, accessToken,
										refreshToken, optDeviceToken )

	check( '+string', '+string', '+string', '?string' )

	return loginFromTokens_imp( baseServerUrl, accessToken, refreshToken, optDeviceToken )

		:map( returnRxUserAccount )
		:switch()

end

--------------------------------------------------------------------------------
--[[!
	\brief Given a set of IMS tokens, establishes a connection to Oz.

	Call when a user has successfully authenticated against the IMS server.
	Triggers a call to the Oz server to obtain further user information
	(such as user name, subscription status, and so on).

	\param baseServerUrl (string) The base URL for the Oz server to use.
	\param accessToken (string) The new access token.
	\param refreshToken (string) The new refresh token.
	\param optDeviceToken (string or nil) The device token, if applicable

	\return P< WFOzUser, info packet > user object and table containing user info
]]--

function WFOzUserLogin.loginFromTokensP( baseServerUrl, accessToken,
										 refreshToken, optDeviceToken )

	check( '+string', '+string', '+string', '?string' )

	return loginFromTokens_imp( baseServerUrl, accessToken, refreshToken, optDeviceToken )
		:call( asKSPromise )
		:mapDataP( returnClassicUserAccountP )

end

--------------------------------------------------------------------------------
--[[!
	\brief Given an IMS email and password pair, obtain login tokens.

	Should be used only for testing.

	\param baseServerUrl (string) The base URL for the Oz server to use.
	\param email (string) user's e-mail address
	\param password (string) user's password

	\return Observable< baseServerUrl, accessToken, refreshToken >
]]--

function WFOzUserLogin.tokensFromEmailAndPassword( baseServerUrl, email, password )

	return WFRxRetryObservable {

		maxAttempts = 3,
		retryDelay = 5,

		factory = function()

			local language_code = WFLocalizationUtils.getLanguageCode()
			local region_code = WFLocalizationUtils.getRegionCode()

			local locale = string.format( "%s_%s", language_code, region_code )

			local clientId, clientSecret = WFImsAppIdentity.getClientIdAndSecret()

			local httpParams = {
				username = email,
				password = password,
				redirect_uri = 'air-auth://carousel',
				scope = 'openid,AdobeID',
				locale = locale,
				client_id = clientId,
				client_secret = clientSecret,
			}

			return WFOzClient.getImsAuthUrl()

				:selectMany( function( imsAuthUrl, imsBaseUrl )

					local url = string.format( '%s/login/code', imsBaseUrl )
					local postBody = WFURI.buildQueryString( httpParams )

					local postArgs = {
						requestMethod = 'POST',
						requestHeaders = {
							{ header = 'Content-Type', value = 'application/x-www-form-urlencoded' },
						},
					}

					return WFRxHttpClient.put( url, postBody, postArgs )

				end )

				:map( KSJsonDecoder.decodeWebSafeJson )

				:map( function( json )

					if json.error then
						error( "IMS error: " .. json.error )
					end

					if json.access_token and json.refresh_token then
						return baseServerUrl, json.access_token, json.refresh_token
					else
						error( "Tokens missing", json )
					end

				end )

		end,

	}

	:take( 1 )

end

--------------------------------------------------------------------------------

local function loginFromEmailAndPassword_imp( baseServerUrl, email, password )

	userLoginLog:infof( "Attempting login for user %s", email )

	return WFOzUserLogin.tokensFromEmailAndPassword( baseServerUrl, email, password )

		:selectMany( function( baseServerUrl, accessToken, refreshToken )
			return loginFromTokens_imp( baseServerUrl, accessToken, refreshToken )
		end )

		:call( runUntilDone )

end

--------------------------------------------------------------------------------
--[[!
	\brief Login using an e-mail address and password.

	Should be used only for testing.

	\param baseServerUrl (string) The base URL for the Oz server to use.
	\param email (string) user's e-mail address
	\param password (string) user's password

	\return Observable< WFOzUser, info packet > user object and table containing user info
]]--

function WFOzUserLogin.loginFromEmailAndPassword( baseServerUrl, email, password )

	return loginFromEmailAndPassword_imp( baseServerUrl, email, password )

		:selectMany( returnRxUserAccount )

end

--------------------------------------------------------------------------------
--[[!
	\brief Login using an e-mail address and password.

	Should be used only for testing.

	\param baseServerUrl (string) The base URL for the Oz server to use.
	\param email (string) user's e-mail address
	\param password (string) user's password

	\return P< WFOzUser, info packet > user object and table containing user info
]]--

function WFOzUserLogin.loginFromEmailAndPasswordP( baseServerUrl, email, password )

	return loginFromEmailAndPassword_imp( baseServerUrl, email, password )
		:call( asKSPromise )
		:mapDataP( returnClassicUserAccountP )

end

--------------------------------------------------------------------------------

return WFOzUserLogin
