--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'
local KSTableUtils = require 'KSTableUtils'
local WFDeviceInfo = require "WFDeviceInfo"
local WFAnalytics = require "WFAnalytics"

	-- WF.core.lua.threading
local WFRequestDispatcher = require "WFRequestDispatcher"

	-- WF.oz_client.managed_catalog
local WFCatUser = require 'WFCatUser'
local WFCatSpec = require 'WFCatSpec'
local WFCatalog = require 'WFCatalog'
local WFCatMetadataPuller = require 'WFCatMetadataPuller'
local WFCatMetadataPusher = require 'WFCatMetadataPusher'
local WFCatConflictResolver = require 'WFCatConflictResolver'
local WFCatErrorHandler = require 'WFCatErrorHandler'
local WFCatUserSynchronizer = require 'WFCatUserSynchronizer'
local WFCatStorage = require 'WFCatStorage'

	-- WF.oz_client.lrmobile
local OzBinaryUploader = require 'OzBinaryUploader'
local WFOzSyncConstants = require 'WFOzSyncConstants'
local OzSyncMetadataMapper = require 'OzSyncMetadataMapper'

	-- WF.oz_client.doc_store

local OzCatalogServiceViaWFCatalog = require "OzCatalogServiceViaWFCatalog"

local LrmConfig = require "LrmConfig"

local LrmOzPurgeAgent = require "LrmOzPurgeAgent"
local LrmCatAsset = require "LrmCatAsset"
local LrmCatAlbum = require "LrmCatAlbum"

local WFLogger = require 'WFLogger'

local logger = WFLogger( 'LrmSyncClientImp' )

--------------------------------------------------------------------------------

LrmConfig.registerBinaryStrategyCheck()

local check = assert( WFArguments.check)
local derefOpt = assert( KSTableUtils.derefOpt )

local Observable = Rx.Observable

--------------------------------------------------------------------------------

local function remapAssetDocForCreateAndUpdate( self, localDoc, payload )

	local userId = self._catalog:ozUserId()
	
	OzSyncMetadataMapper.remapGlobalDocumentId( userId, payload )
	
end

--------------------------------------------------------------------------------

local function trackSha256MasterUploadError( puller, docValue )

	if docValue then
	
		local docType = derefOpt( docValue, 'type' )
		
		if docType == 'asset' then
		
			local hasSha256Error = derefOpt( docValue, 'importSource', 'status', 'error', 'all', 'reason', 'sha256' ) == "does not match"
			
			if hasSha256Error then
			
				local uniqueDeviceId = derefOpt( docValue, 'importSource', 'uniqueDeviceId' )
			
				if uniqueDeviceId and ( uniqueDeviceId == WFDeviceInfo.uniqueDeviceId() ) then
			
					local errorEvent = {}
				
					errorEvent['event.type'] = 'error'
					errorEvent['event.workflow'] = 'SYNC'
					errorEvent['event.subcategory'] = 'error_handler'
					errorEvent['event.error_type'] = 'Sha256MismatchError'
					
					errorEvent['content.id'] = docValue._id
					errorEvent['content.type'] = derefOpt( docValue, 'subtype' )
					errorEvent['content.mimetype'] = derefOpt( docValue, 'importSource', 'contentType' )
					errorEvent['content.size'] = derefOpt( docValue, 'importSource', 'fileSize' )
					errorEvent['event.error_desc'] = derefOpt( docValue, 'importSource', 'sha256' )
					
					WFAnalytics.trackEvent( errorEvent )
					
				end
				
			end
			
		end
		
	end
	
	
end

--------------------------------------------------------------------------------

local function albumAssetsForFilteredAssetsQueryFn( catalog, assetIds )

	return catalog:indexStore():query {
				sql = [[SELECT album_asset.docId, album_asset.assetId FROM TEMP_ids JOIN album_asset ON 
 							TEMP_ids.docId = album_asset.assetId]],
				tempIds = assetIds,
		}
		:take( 1 )
end

--------------------------------------------------------------------------------

--==============================================================================

-- @mark LrmSyncClientImp namespace

--------------------------------------------------------------------------------

local syncClientImpMethods = {

	_user = false,
		-- (WFCatUser)
		
	_catSpec = false,
		-- (WFCatSpec)
		
	_catalog = false,
		-- (WFCatalog)
		
	_puller = false,
		-- (WFCatMetadataPuller)
		
	_pusher = false,
		-- (WFCatMetadataPusher)
	
	_storageMonitor = false,
		-- (WFCatStorage)
	
	_binaryUploader = false,
		-- (OzBinaryUploader)
	
	_purgeAgent = false,
		-- (LrmOzPurgeAgent)
		
	_errorHandler = false,
		-- (WFCatErrorHandler)
		
	_userSynchronizer = false,
		-- (WFCatUserSynchronizer)
	
	_conflictResolver = false,
		-- (WFCatConflictResolver)

	_catalogServiceViaWFCatalog = false,
		-- (OzCatalogServiceViaWFCatalog)
		
	attachMessageHandlers = false,
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function syncClientImpMethods:__init( options )

	check( 'LrmSyncClientImp', 'table' )
	
	local serverBase = assert( options.serverBase )
	local userId = assert( options.userId )
	local userInfoCacheFolder = assert( options.userInfoCacheFolder )
	local catalogId = assert( options.catalogId )
	local rootCatalogPath = assert( options.rootCatalogPath )
	
	local user = WFCatUser( serverBase, userId, userInfoCacheFolder )
	
	local catSpec = WFCatSpec( user, catalogId )
	
	local catArgs = {
		docStorePath = assert( options.docStorePath ),
		indexSpecScript = assert( options.indexSpecScript ),
		noFileAccessDb = true,
		noCatalogSearch = true,
	}
	
	local catalog = WFCatalog( catSpec, rootCatalogPath, catArgs )
	
	-- ConflictResolver
	local conflictResolver = WFCatConflictResolver( catalog:docStore(), catalog:indexStore() )
	
	-- Quota Monitoring
	local storageMonitor = WFCatStorage( catalog )
	
	-- Error Handler
	
	local errorHandler = WFCatErrorHandler( catalog, {
		deleteAssets = LrmCatAsset.deleteAssets,
		deleteIndividualAlbum = LrmCatAlbum.deleteAlbum } )
	
	-- Metadata Sync
	local puller = WFCatMetadataPuller( catalog )
	local pusher = WFCatMetadataPusher( catalog, {
		delayBeforeSendingUpdates = 3, -- sec
		documentBatchCount = 100, -- batch 100 documents max before sending to Oz
		albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn
		} )

	pusher:addCreateDocRemapFunction( 'asset', remapAssetDocForCreateAndUpdate )
	pusher:addUpdateDocRemapFunction( 'asset', remapAssetDocForCreateAndUpdate )
	
	puller:addAssetTypeHandler( 'asset', trackSha256MasterUploadError )
	
	-- User Synchronizer
	local userSynchronizer = WFCatUserSynchronizer( user )
	
	-- Binary Uploader

	local ozCatalog = OzCatalogServiceViaWFCatalog( Rx.Observable.just( catalog ) )
	
	local deleteOriginalAfterUpload = false
	
	local binaryUploader = OzBinaryUploader {
		ozUser = user,
		docStore = ozCatalog:docStore(),
		indexStore = catalog:indexStore(),
		catalogId = options.catalogId,
		deleteOriginalAfterUpload = deleteOriginalAfterUpload,
		noScheduler = true,
	}
	
	-- monitor quota
	local quotaMonitorSubscription = storageMonitor:isOverStorage()
		:distinctUntilChanged()
		:subscribe( function( isOverStorage )
			pusher:setUserIsOverQuota( isOverStorage )
		end )
	
	-- monitor expiration
	local userExpiredSubscription = user:isSubscriptionExpired()
		:distinctUntilChanged()
		:subscribe( function( isExpired )
			pusher:setUserIsExpired( isExpired )
		end )
	
	local purgeAgent = LrmOzPurgeAgent( ozCatalog )

	ozCatalog:addCloseAction( function()
	
		quotaMonitorSubscription:dispose()
		userExpiredSubscription:dispose()
		
	end )
	
	self:setOwned {
		_user = user,
		_catalog = catalog,
		_puller = puller,
		_pusher = pusher,
		_storageMonitor = storageMonitor,
		_binaryUploader = binaryUploader,
		_purgeAgent = purgeAgent,
		_errorHandler = errorHandler,
		_userSynchronizer = userSynchronizer,
		_conflictResolver = conflictResolver,
		_catalogServiceViaWFCatalog = ozCatalog
	}
	
		-- WFCatSyncClientMessageProcessor.lua uses rawget, so we have to stash 
		-- attachMessageHandlers here
	self.attachMessageHandlers = function( self )

		WFRequestDispatcher.onMessageDo {
			message = 'purgeImmediately',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

		WFRequestDispatcher.onMessageDo {
			message = 'purgeForAlbum',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

		WFRequestDispatcher.onMessageDo {
			message = 'startContinuousPurge',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

		WFRequestDispatcher.onMessageDo {
			message = 'stopContinuousPurge',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

	end
	
	self._catSpec = catSpec
	
	-- INFO [pkamin 2017-01-27]: we always run the user synchronizer for 
	-- authenticated user to refresh the account document even if sync is stopped
	-- ( e.g. trial expired )
	
	-- [pkamin 2017-09-11]: We only manage the UserSynchronizer on this message
	-- processor for Android. See LrmCatUserModel:userSynchronizer for iOS
	
	local isAuthenticatedUser = userId ~= WFOzSyncConstants.kGlobalDocumentId
	
	if isAuthenticatedUser and not WFPlatformInfo.isiOS then
		self._userSynchronizer:startContinuousUpdates()
	end
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getDeleteFeedCallback( )

    check( 'LrmSyncClientImp' )

    if self._purgeAgent then
        return self._purgeAgent:getDeleteFeedCallback()
    end

    return Observable.empty()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:startSync( binaryStrategy )

	if WFPlatformInfo.isAndroid then
		check( 'LrmSyncClientImp' )
	else
		check( 'LrmSyncClientImp', 'binaryStrategy'  )
    end

	logger:debugf( 'starting sync' )
	
	self._puller:getUpdatesContinuously()
	self._pusher:sendUpdatesContinuously()
	self._binaryUploader:startUploading()
	
	return Observable.empty()
	
end


--------------------------------------------------------------------------------

function syncClientImpMethods:startAccountsSync( binaryStrategy )

	logger:debugf('starting accounts sync')

	self._userSynchronizer:startContinuousUpdates()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:stopSync()

	check( 'LrmSyncClientImp' )
	
	logger:debugf( 'stopping sync' )
	
	self._puller:stopContinuousUpdates()
	self._pusher:stopContinuousUpdates()
	self._binaryUploader:stopUploading()
	
	return Observable.empty()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:stopAccountsSync()

	logger:debugf('stopping accounts sync')

	self._userSynchronizer:stopContinuousUpdates()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:isPushingMetadata()

	check( 'LrmSyncClientImp' )
	
	return self._pusher:isDoingWork()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isPullingMetadata()

	check( 'LrmSyncClientImp' )
	
	return self._puller:isDoingWork()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getRequiredUpdateStream()

	check( 'LrmSyncClientImp' )
	
	return self._user:getRequiredUpdateStream()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isOzCatalogDeleted()

	check( 'LrmSyncClientImp' )
	
	return self._puller:isOzCatalogDeleted()
	
end

--------------------------------------------------------------------------------

return Rx.Class( 'LrmSyncClientImp', syncClientImpMethods )
