--[[----------------------------------------------------------------------------

NmbAlbumsCache.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2017 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- Rx
local Rx = require 'Rx'

	-- WF.core
local WFCloseSupport = require 'WFCloseSupport'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.oz_client.managed_catalog
local WFCatAlbums = require 'WFCatAlbums'
local WFCatSyncStatus = require 'WFCatSyncStatus'
local WFCatCollaboration = require 'WFCatCollaboration'

local WFLogger = require 'WFLogger'

--------------------------------------------------------------------------------

local logger = WFLogger( 'NmbAlbumsCache' )

local runUntilDone = WFRxRunUntilDone

--============================================================================--

-- @mark NmbAlbumsCache namespace

local NmbAlbumsCache = {}

--------------------------------------------------------------------------------

local SimpleAlbumsCache = Rx.Class 'WF.internals.SimpleAlbumsCache' {
	
	_albumsCacheObservable = false,
	
	_albumsSubscription = false,
	
	WFCloseSupport.addCloseSupportMethods,
	
	__init = function( self, catalog )
		
		self._albumsCacheObservable = Rx.ReplaySubject( 1 )
			
		self._albumsSubscription = WFCatAlbums._flattenedAlbumHierarchy( catalog, false )
			:subscribe( function( flattenedAlbums )
				
				self:maybeUpdateCacheWithHierarchy( flattenedAlbums )
				
			end )
	end,
	
	close = function( self )
		if self._albumsSubscription then
			self._albumsSubscription:dispose()
		end
		
		self._albumsSubscription = nil
	end,
	
	maybeUpdateCacheWithHierarchy = function( self, hierarchy )
		self._albumsCacheObservable:onNext( hierarchy )
	end,
	
	addAlbumWithSubtype = function( self, localId, albumName, subtype, optParentId, optHidden )
	
		-- SPAM [matthewj 2017-02-03]: Because the creation in the cache on adds happens *after*
		-- the creation in the doc store, we are not doing validation on whether it's ok to add 
		-- to the cache
		return self:getCache()
			:take( 1 )
			:map( function( cachedHierarchy )
			
				if not WFCatAlbums._getAlbumFromHierarchyById( cachedHierarchy, localId ) then
					
					local album = {
						id = localId,
						name = albumName,
						parentId = optParentId,
						subtype = subtype,
						hidden = optHidden,
					}
					
					WFCatAlbums._addAlbumIntoHierarchy( album, cachedHierarchy )
					
					self:maybeUpdateCacheWithHierarchy( cachedHierarchy )
				end
				
			end )
	end,
		
	addFolder = function( self, localId, folderName, optParentId )
		
		return self:addAlbumWithSubtype( localId, folderName, 'collection_set', optParentId )
	
	end,
	
	addAlbum = function( self, localId, albumName, optParentId, optHidden )
		
		return self:addAlbumWithSubtype( localId, albumName, nil, optParentId, optHidden )
	
	end,
	
	renameAlbumOrFolder = function( self, albumId, albumName )
		return  WFCatAlbums._validateRenameAlbumInHierarchy( self:getCache():take( 1 ), albumId, albumName )
			:map( function( cachedHierarchy )
				
				local album = WFCatAlbums._getAlbumFromHierarchyById( cachedHierarchy, albumId )
				
				if album then
					album.name = albumName
					self:maybeUpdateCacheWithHierarchy( cachedHierarchy )
				end
				
			end )
			:catch( Rx.Observable.empty() )
	end,
	
	reparentAlbumOrFolder = function( self, albumId, optNewParentId, optOrder )
		return WFCatAlbums._validateReparentAlbumInHierarchy( self:getCache():take( 1 ), albumId, optNewParentId, optOrder )
			:map( function( hierarchy )

				hierarchy = WFCatAlbums._reparentAlbumInHierarchy( hierarchy, albumId, optNewParentId, optOrder )
				self:maybeUpdateCacheWithHierarchy( hierarchy )
								
			end )
	end,
	
	reparentMultipleAlbumsOrFolders = function( self, albumIds, optNewParents )
		return WFCatAlbums._validateReparentMultipleAlbumsInHierarchy( self:getCache():take( 1 ), albumIds, optNewParents )
			:map( function( hierarchy, maybeErrors )
				-- Related to NMB-6314: if attempting to reparent multiple albums and there is an error, we 
				-- reject the full reparent even if some would succeed.
				if not maybeErrors then
					
					for _, albumId in ipairs( albumIds ) do
						local maybeParentId 
						if type( optNewParents ) == 'number' or type( optNewParents ) == 'string' then
							maybeParentId = optNewParents
						elseif type( optNewParents ) == 'table' then
							maybeParentId = optNewParents[ albumId ]
						end
				
						hierarchy = WFCatAlbums._reparentAlbumInHierarchy( hierarchy, albumId, maybeParentId )
					end
					
					self:maybeUpdateCacheWithHierarchy( hierarchy )
					
				end
				
			end )
	end,
	
	deleteAlbumOrFolder = function( self, albumId )
	
		return self:getCache()
			:take( 1 )
			:map( function ( cachedHierarchy )
				
				local idsRemoved
				cachedHierarchy, idsRemoved = WFCatAlbums._deleteAlbumFromHierarchy( cachedHierarchy, albumId )
			
				self:maybeUpdateCacheWithHierarchy( cachedHierarchy )
				
				return idsRemoved
			end )
	end,
	
	shareAlbum = function( self, albumId )
		return self:getCache()
			:take( 1 )
			:map( function( cachedHierarchy )
				local album = WFCatAlbums._getAlbumFromHierarchyById( cachedHierarchy, albumId )
				if album then
					album.isShared = true
					self:maybeUpdateCacheWithHierarchy( cachedHierarchy )
				end
			end )
	end,
	
	unshareAlbum = function( self, albumId )
		return self:getCache()
			:take( 1 )
			:map( function( cachedHierarchy )
				local album = WFCatAlbums._getAlbumFromHierarchyById( cachedHierarchy, albumId )
				if album then
					album.isShared = nil
					self:maybeUpdateCacheWithHierarchy( cachedHierarchy )
				end
			end )
	end,
	
	getCache = function( self )
		return self._albumsCacheObservable
	end,
	
	getAlbumsFromCache = function( self )
		return self._albumsCacheObservable
			:map( function( hierarchy ) 
				if hierarchy and hierarchy.albums then
					return hierarchy.albums
				else 
					return {}
				end
			end )
	end,

}

--------------------------------------------------------------------------------

local function createFolderWithCache( wfCatalog, albumsCache, folderName, optParentId )
		
	return WFCatAlbums.createFolder( wfCatalog, folderName, optParentId )
			:flatMap( function( localDocId, fullDocId, docValue )
			
			return albumsCache:addFolder( localDocId, docValue.name, optParentId )
				:map( function()
					return localDocId, fullDocId, docValue
				end )
		end )
end

--------------------------------------------------------------------------------

local function createAlbumWithCache( wfCatalog, albumsCache, albumName, optParentId, optHidden )
	
	return WFCatAlbums.createAlbum( wfCatalog, albumName, optParentId, optHidden )
			:flatMap( function( localDocId, fullDocId, docValue )
			
			return albumsCache:addAlbum( localDocId, docValue.name, optParentId, optHidden )
				:map( function()
					return localDocId, fullDocId, docValue
				end )
		end )
end

--------------------------------------------------------------------------------

local function renameAlbumOrFolderWithCache( wfCatalog, albumsCache, albumId, albumName )
		
	return Rx.Observable.combineLatest(
		albumsCache:renameAlbumOrFolder( albumId, albumName ),
		WFCatAlbums.renameAlbumOrFolder( wfCatalog, albumId, albumName ) )
		:map( function( _, result ) return result end )
	
end

--------------------------------------------------------------------------------

local function reparentAlbumOrFolderWithCache( wfCatalog, albumsCache, albumId, optNewParentId, optOrder )
	 
	return Rx.Observable.combineLatest(
		albumsCache:reparentAlbumOrFolder( albumId, optNewParentId, optOrder ),
		WFCatAlbums.reparentAlbumOrFolder( wfCatalog, albumId, optNewParentId, optOrder) )
		:map( function( _, result ) return result end )
end

--------------------------------------------------------------------------------

local function reparentMultipleAlbumsOrFoldersWithCache( wfCatalog, albumsCache, albumIds, optNewParentId, optOrder )
	
	return Rx.Observable.combineLatest(
		albumsCache:reparentMultipleAlbumsOrFolders( albumIds, optNewParentId, optOrder ),
		WFCatAlbums.reparentMultipleAlbumsOrFolders( wfCatalog, albumIds, optNewParentId, optOrder) )
		:map( function( _, result ) return result end )

end

--------------------------------------------------------------------------------

local function deleteAlbumOrFolderWithCache( wfCatalog, albumsCache, albumId )
	
	-- SPAM [matthewj 2018-01-24]: NMB-7104 Note that deletion is different within the cache
	-- here.  Since the cache is what is displayed to users, we need to make sure that we only
	-- delete from the real doc store those folders/albums we removed from the cache.  Otherwise
	-- we can delete things that the user does not realize they are deleting.
	return albumsCache:deleteAlbumOrFolder( albumId )
		:flatMap( function( folderAndAlbumIdsRemovedFromCache )
			return WFCatAlbums.deleteIndividualAlbums( wfCatalog, folderAndAlbumIdsRemovedFromCache )
		end )
		:takeLast( 0 )

end

--------------------------------------------------------------------------------

local function shareAlbumWithCache( wfCatalog, albumsCache, albumId, ... )

	return Rx.Observable.combineLatest(
		albumsCache:shareAlbum( albumId ),
		WFCatCollaboration.shareAlbum( wfCatalog, albumId, ... ) )
		:map( function( _, result ) return result end ) 

end

--------------------------------------------------------------------------------

local function unshareAlbumWithCache( wfCatalog, albumsCache, albumId )

	return Rx.Observable.combineLatest(
		albumsCache:unshareAlbum( albumId ),
		WFCatCollaboration.unshareAlbum( wfCatalog, albumId ) )
		:map( function( _, result ) return result end ) 

end

--------------------------------------------------------------------------------

local function makeCatalogAndCacheBasedObservableIntoSubModel( fn, optObservableType, maybeRunUntilDone )
	
	local deliverOneResult = optObservableType == 'deliverOneResult'
	
	return function( model, ... )
		local params = { ... }
		local nParams = select( '#', ... )
		local value

		local catalogStream = deliverOneResult and model._wfCatalog or model._wfCatalog:take( 1 )
		local subModelStream = catalogStream
                             	:mapSwitch( function( wfCatalog, cache )
									
									if maybeRunUntilDone then
                             			return fn( wfCatalog, cache, unpack( params, 1, nParams ) )
											:call( runUntilDone )
									else
										return fn( wfCatalog, cache, unpack( params, 1, nParams ) )
									end
									
                               	end )
								
								
		value = deliverOneResult and subModelStream:take( 1 ) or subModelStream

		return {
			_wfCatalog = model._wfCatalog,

			value = value,
		}
		
	end
	
end

--------------------------------------------------------------------------------

local function makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( fn, optObservableType )

	return makeCatalogAndCacheBasedObservableIntoSubModel( fn, optObservableType, true )

end

--------------------------------------------------------------------------------

local function makeCatalogBasedObservableIntoSubModel( fn, optObservableType, maybeRunUntilDone )
	
	local deliverOneResult = optObservableType == 'deliverOneResult'
	
	return function( model, ... )
		local params = { ... }
		local nParams = select( '#', ... )
		local value
		
		local catalogStream = deliverOneResult and model._wfCatalog or model._wfCatalog:take( 1 )
		local subModelStream = catalogStream
                             	:mapSwitch( function( wfCatalog )
								
									if maybeRunUntilDone then
                             			return fn( wfCatalog, unpack( params, 1, nParams ) )
											:call( runUntilDone )
									else
										return fn( wfCatalog, unpack( params, 1, nParams ) )
									end
									
                               	end )
								
		value = deliverOneResult and subModelStream:take( 1 ) or subModelStream

		return {
			_wfCatalog = model._wfCatalog,

			value = value,
		}
	end
end

--------------------------------------------------------------------------------

local function makeCatalogBasedObservableIntoRunUntilDoneSubModel( fn, optObservableType )

	return makeCatalogBasedObservableIntoSubModel( fn, optObservableType, true )
	
end

--------------------------------------------------------------------------------

local function sharingStatusByFullAlbumId( catalog )
	
	return WFCatCollaboration.getAllSharedAlbums( catalog )
		:map( function( sharedAlbums )
			local sharesByAlbumId = {}
			for _, share in ipairs( sharedAlbums ) do
				sharesByAlbumId[ share.albumId ] = not share.isPrivate
			end
			return sharesByAlbumId
		end )
end

--------------------------------------------------------------------------------

local function addAdditionalAlbumInfoToAlbumsList( albums, counts, covers, statuses, shares )
	
	local albumsWithMoreData = {}
	
	for _, album in ipairs( albums ) do
		album.syncStatus = statuses[ album.fullDocId ]
		album.count = counts[ album.fullDocId ]
		album.cover = covers[ album.id ]
		album.isShared = shares[ album.fullDocId ]
		
		table.insert( albumsWithMoreData, album )
	end
	
	return albumsWithMoreData
end

--------------------------------------------------------------------------------

local function combineAdditionalAlbumInfoWithAlbumList( wfCatalog, albumListObservable )
	
	return Rx.Observable.combineLatest( 
		function( albums, counts, covers, statuses, shares )
			
			if next( counts ) or next( covers ) or next( statuses ) or next( shares ) then
				
				albums = addAdditionalAlbumInfoToAlbumsList( albums, counts, covers, statuses, shares )
							
			end
			
			return albums
			
		end,
		
		albumListObservable,
		WFCatAlbums.getAlbumCounts( wfCatalog ):startWith( {} ),
		WFCatAlbums.getAlbumCovers( wfCatalog ):startWith( {} ),
		WFCatSyncStatus.getSyncStatusForAlbums( wfCatalog, true --[[ignore album status]] ):startWith( {} ),
		sharingStatusByFullAlbumId( wfCatalog ):startWith( {} ) )

end

--------------------------------------------------------------------------------

--[[
	\param wfCatalog an Observable( WFCatalog )
	Returns a model whose value is an Observable array of top-level albums, each containing fields:
		id : (int) the uuid of the album
		name : (user string) the name of the album
		path : (user string) The name of the parents and album separated by /
		parentId : (uuid, or nil) the id of the albums parent
		children : (array) array of child albums
--]]
function NmbAlbumsCache.makeAlbumsCacheModel( wfCatalog )
	
	local albumsCaches = {}
	
	local wfCatalogAndCache = wfCatalog
		:filter( function( catalog ) return catalog ~= nil end )
		:map( function( catalog )
		
			local cache = albumsCaches[ catalog ]
		
			if not cache then
				cache = SimpleAlbumsCache( catalog )
				albumsCaches[ catalog ] = cache
			end
			
			return catalog, cache
			
		end )
	
	local albums = wfCatalogAndCache
		:filter( function( catalog ) return catalog ~= nil end )
		:flatMapLatest( function( catalog, cache )
			
			return combineAdditionalAlbumInfoWithAlbumList( 
						catalog, 
						cache:getAlbumsFromCache( ) )
			
		end )

	return {
		_wfCatalog = wfCatalogAndCache,
		
		priority = 'high',

		value = {
			albums = albums,
		},

		models = {
			-- use cache
			createFolder = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( createFolderWithCache, 'deliverOneResult' ),
			createAlbum = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( createAlbumWithCache, 'deliverOneResult' ),
			deleteAlbum = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( deleteAlbumOrFolderWithCache, 'deliverOneResult' ),
			renameAlbum = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( renameAlbumOrFolderWithCache, 'deliverOneResult' ),
			reparentAlbum = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( reparentAlbumOrFolderWithCache, 'deliverOneResult' ),
			reparentMultipleAlbums = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( reparentMultipleAlbumsOrFoldersWithCache, 'deliverOneResult' ),
			
			shareAlbum = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( shareAlbumWithCache, 'deliverOneResult' ),
			unshareAlbum = makeCatalogAndCacheBasedObservableIntoRunUntilDoneSubModel( unshareAlbumWithCache, 'deliverOneResult' ),
			
			-- do not use cache
			addAssetsToAlbum = makeCatalogBasedObservableIntoRunUntilDoneSubModel( WFCatAlbums.addAssetsToAlbum, 'deliverOneResult' ),
			removeAssetsFromAlbum = makeCatalogBasedObservableIntoRunUntilDoneSubModel( WFCatAlbums.removeAssetsFromAlbum, 'deliverOneResult' ),
			setAlbumCover = makeCatalogBasedObservableIntoRunUntilDoneSubModel( WFCatAlbums.setAlbumCover, 'deliverOneResult' ),
			
			sharedAlbumUrl = makeCatalogBasedObservableIntoSubModel( WFCatCollaboration.sharedAlbumUrl ),
			albumCover = makeCatalogBasedObservableIntoSubModel( WFCatAlbums.albumCover ),
			getSharedAlbumAttributes = makeCatalogBasedObservableIntoSubModel( WFCatCollaboration.getSharedAlbumAttributes ),
			setSharedAlbumAttributes = makeCatalogBasedObservableIntoSubModel( WFCatCollaboration.setSharedAlbumAttributes ),
			
		},

		close = function()
			
			logger:info( "Closing" )
			
			for _, cache in pairs( albumsCaches ) do
				cache:close()
			end
			
		end,

	}

end

--------------------------------------------------------------------------------

return NmbAlbumsCache
