--[[----------------------------------------------------------------------------

NmbCatalog.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

local WFArguments = require 'WFArguments'
local WFCatalogBase = require 'WFCatalogBase'
local NmbFileAccessDatabase = require 'NmbFileAccessDatabase'
local WFObjectProxy = require 'WFObjectProxy'
local NmbFileMover = require 'NmbFileMover'
local WFCatStacksProcessor = require 'WFCatStacksProcessor'
local NmbDeletionBacklog = require 'NmbDeletionBacklog'

local NmbSimpleDocStoreUpdater = require 'NmbSimpleDocStoreUpdater'
local NmbFileManagementSettings = require 'NmbFileManagementSettings'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local nmbCatalogPrototype = {
	
	WFCatalogBase,
		--superclass
	
	_initOptArgs = false,
		-- (table) optional args passed in at __init (needed for __inited)
		
	_fileAccessDb = false,
		-- (NmbFileAccessDatabase) Keep track of recently accessed files.
		
	_deletionBacklog = false,
		-- (NmbDeletionBacklog) A database of assets that need to be deleted from disk.
		
	_volumeAgents = false,
		-- (table<NmbVolumeAgent>) The Nimbus volume agents.
	
	_fileMover = false,
		-- (NmbFileMover) The file mover responsible for migrating assets
		-- when the original storage location changes.
	
	_docStoreUpdater = false,
		-- (NmbSimpleDocStoreUpdater) Allows for simple updating (format changes) of documents
		
	_stacksProcessor = false,
		-- (WFCatStacksProcessor) Processes incoming stacks to ensure they're formatted correctly
		-- for display
		
	class = {},
	
	_searchSchema = false,
	
}

--------------------------------------------------------------------------------

--[[!
	\brief Creates a NmbCatalog.
	
	\param ozCatalogSpec (WFCatSpec) specification for the catalog user and catalogID
	\param rootCatalogPath (string) root directory under which to store catalog related files
	\param optArgs (table) optional args for doc store index store creation. These are in addition
			to any optional args allowed by WFCatalogBase
		- fileAccessDbPath -- path to files access db
]]--

function nmbCatalogPrototype:__init( ozCatalogSpec, rootCatalogPath, optArgs )
	
	check( 'NmbCatalog', 'WFCatSpec', 'string', '?table' )
	
	self._initOptArgs = optArgs or {}
	
	self._volumeAgents = {}
	
	-- these arguments are returned to be used in the superclass (WFCatalog) init
	return ozCatalogSpec, rootCatalogPath, optArgs
	
end

--------------------------------------------------------------------------------

--[[ 
	\brief Completes initialization of the NmbCatalog after the superclass has been initialized
]]--
function nmbCatalogPrototype:__inited( )

	check( 'NmbCatalog' )
					
	local searchSchema = require 'WFCatSearchSchema'()
	
	require 'WFCatSearchHandler'.addSearchToSearchSchema( searchSchema ) 
	
	self._searchSchema = searchSchema

	require 'WFCatIdleDownloader'.addSearchToSearchSchema( self._searchSchema )
	
	local optFileAccessDbPath = self._initOptArgs.fileAccessDbPath
	
	local fileAccessDb = NmbFileAccessDatabase( self._docStore, 
											self._indexStore,
											optFileAccessDbPath or self._wfCatFileSpec:fileAccessDbPath() )
	
	local deletionBacklog = NmbDeletionBacklog( self._docStore, 
			optFileAccessDbPath or self._wfCatFileSpec:fileAccessDbPath() )
	
	local fileMover = NmbFileMover( self._docStore, self._indexStore, self._wfCatFileSpec, deletionBacklog, 
			NmbFileManagementSettings.keepAllOriginals(), NmbFileManagementSettings.keepAllProxies() )
	
	local docStoreUpdater = NmbSimpleDocStoreUpdater( self._docStore, self._indexStore )
	
	local stacksProcessor = WFCatStacksProcessor( self._docStore, self._indexStore )
		
	self:setOwned {
		_fileAccessDb = fileAccessDb,
		_deletionBacklog = deletionBacklog,
		_docStoreUpdater = docStoreUpdater,
		_fileMover = fileMover,
		_stacksProcessor = stacksProcessor
	}
	
end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:searchSchema()

	check( 'NmbCatalog' )
	return self._searchSchema

end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:fileAccessDb()

	check( 'NmbCatalog' )
	return self._fileAccessDb

end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:getDeletionBacklog()

	check( 'NmbCatalog' )
	return self._deletionBacklog

end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:getVolumeAgentByVolumeGuid( volumeGuid )

	check( 'NmbCatalog', 'string' )
	return self._volumeAgents[ volumeGuid ]
	
end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:getFileMover()

	check( 'NmbCatalog' )
	return self._fileMover

end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:getDocStoreUpdater()

	check( 'NmbCatalog' )
	return self._docStoreUpdater

end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:addVolumeAgent( volumeAgent )

	check( 'NmbCatalog', 'NmbVolumeAgent' )
	self._volumeAgents[ volumeAgent:getVolumeGuid() ] = volumeAgent

end

--------------------------------------------------------------------------------

function nmbCatalogPrototype:getAllVolumeAgents()

	check( 'NmbCatalog' )
	return self._volumeAgents

end

--------------------------------------------------------------------------------

local NmbCatalog_class = Rx.Class( 'NmbCatalog', nmbCatalogPrototype )

return WFObjectProxy( 'WFCatalog', NmbCatalog_class, nil, true )
