--[[----------------------------------------------------------------------------

NmbDocValueToRenderInfoStream.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

NmbDocValueToRenderInfoStream consumes a sequence of array-valued tables of asset IDs
and reports the "render info" of each corresponding document either as it first
appears in the array, or as it changes in value while still in the array.

Asset IDs may be either numeric (local) or string (global) document IDs.

Render info is expressed as a table similar to the following:

	{
		_id = "(uuid / global ID)",
		_localId = (number),
		pathToOriginal = "(path)",
		pathToProxy = "(path)",
		develop = {
			-- develop settings as from Oz (http://belt.sea.adobe.com/asset/develop.html)
		}
	}

------------------------------------------------------------------------------]]

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFTableArchiving = require 'WFTableArchiving'
local WFFunctionUtils = require 'WFFunctionUtils'

	-- WF.oz_client
local WFLocalDocUtils = require "WFLocalDocUtils"

	-- WF.sqlite.doc_store
local WFCatFileSpec = require 'WFCatFileSpec'

	-- WF.oz_client.managed_catalog.develop
local NmbDocValueDevelopSettingsUtils = 
		require "NmbDocValueDevelopSettingsUtils"

	-- pixel provider
local WFCatRenderCapabilities = require 'WFCatRenderCapabilities'

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local WFFunctionUtils_reduce = assert( WFFunctionUtils.reduce,
			"WFFunctionUtils.reduce does not exist" )

--------------------------------------------------------------------------------

local KSTableUtils_mapArrayToSet = assert( KSTableUtils.mapArrayToSet,
			"KSTableUtils.mapArrayToSet does not exist" )

--------------------------------------------------------------------------------

local WFLocalDocUtils_currentRevIdIsOnOz =
		assert( WFLocalDocUtils.currentRevIdIsOnOz, 
				"WFLocalDocUtils.currentRevIdIsOnOz does not exist" )

--------------------------------------------------------------------------------

local logger = require 'WFLogger'( 'NmbDocValueToRenderInfoStream' )

--------------------------------------------------------------------------------

local function maybeGetSize( maybeSizeUri )

	if type( maybeSizeUri ) ~= "string" then

		return

	end
	
	local maybeSize = maybeSizeUri:gsub( ".*/(.*)", "%1" )

	if maybeSize == "" then

		return

	end

	return maybeSize

end

--------------------------------------------------------------------------------

local function getBinaryInfo( docValue )

	local renditionsList = {}
	local proxiesList = {}
	local master
	
	local links = derefOpt( docValue, '_serverProvidedMetadata', 'links' ) or {}

	local mayHaveUnsyncedLocalChanges =
			not WFLocalDocUtils_currentRevIdIsOnOz( docValue )
	
	for sizeKey, link in pairs( links ) do

		-- make sure we have a link table or else we're useless
		local linkTable = type( link ) == 'table' and link or {}
		local isVideo = docValue.subtype == 'video'

		local isRendition = not not sizeKey:find( 'rendition_type' )
		local isProxy = not isRendition and not not sizeKey:find( 'proxy_type' )
		local isMaster = sizeKey == '/rels/master'

		if isVideo then
			-- Instead of using a regex, actual type is being used
			-- since we are supporting only 720p as of now.
			isProxy = not not sizeKey:find( 'rendition_type/720p' )
			isRendition = not isProxy and not not sizeKey:find( 'rendition_type' )
		end
		
		local isInvalid = derefOpt( linkTable, 'invalid' ) or false
		
		if not isInvalid then
			-- SPAM [matthewj 2015-12-03]: We're currently ignoring all Oz binaries
			-- when they are marked as invalid. This may not be the right 
			-- decision and we may want to ensure we send pixel information
			-- if we have it, even if it's stale. If we do, we'll need to mark
			-- it as stale and treat it appropriately further down the 
			-- pipeline
			
			local href = derefOpt( linkTable, 'href' )

			local size = not isMaster and maybeGetSize( sizeKey )

			if isRendition and href and size and 
					not mayHaveUnsyncedLocalChanges then
				
				table.insert( renditionsList, { size = size, href = href } )

			elseif isProxy and href and size then
					
				table.insert( proxiesList, { size = size, href = href } )
				
			elseif isMaster then 
								
				master = href	

			end
		
		end

	end
	
	return renditionsList, proxiesList, master

end

--------------------------------------------------------------------------------

local function maybeGetCustomProfileInfo( docValue )

	local customCameraProfileDigest = derefOpt( docValue, "_localOnly", "customCameraProfile", "profileDigest" )
	customCameraProfileDigest = customCameraProfileDigest ~= "NA" and customCameraProfileDigest or nil
	
	local customCameraProfileMissing = derefOpt( docValue, "_localOnly", "customCameraProfile", "isMissing" )
	
	return customCameraProfileDigest, customCameraProfileMissing
end

--------------------------------------------------------------------------------

local NmbDocValueDevelopSettingsUtils_getSanitizedDocValueDevelopStanza =
		assert( NmbDocValueDevelopSettingsUtils.getSanitizedDocValueDevelopStanza,
				"missing NmbDocValueDevelopSettingsUtils.getSanitizedDocValueDevelopStanza" )

--------------------------------------------------------------------------------

local function extractRenderInfoFromDocValue( docValue )
	
	local renditions, proxies, master = getBinaryInfo( docValue )
	
	local ozProxyLink, ozProxySize = WFCatFileSpec.getProxyLinkFromDocValue( docValue )
	
	local xmpDevelopInfo = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue ) or {}
	
	local customCameraProfileDigest, customCameraProfileMissing = maybeGetCustomProfileInfo( docValue )
	
	return {

		_id = docValue._id,

		_localId = docValue._localId,

		_revId =  docValue._revId,
		
		_changeCounter = docValue._changeCounter,
		
		subtype = docValue.subtype,

		pathToOriginal = WFCatFileSpec.buildOriginalPathFromDocValue( docValue ),

		pathToProxy = WFCatFileSpec.buildProxyPathFromDocValue( docValue ),
		
		pathToXmpDevelop = xmpDevelopInfo.path,
		
		pathToXmpDevelopSha256 = xmpDevelopInfo.sha256,

		pathToXmpSidecar = WFCatFileSpec.getXmpSidecarPathFromDocValue( docValue ),

		xmpSidecarExists = WFCatFileSpec.getXmpSidecarExistsFromDocValue( docValue ),

		sha256 = derefOpt( docValue, 'importSource', 'sha256' ),

		localOriginalSha256 = WFCatFileSpec.getLocalOriginalSha256( docValue ),

		ozMasterLink = WFCatFileSpec.getOriginalLinkFromDocValue( docValue ),
		
		localProxySha256 = WFCatFileSpec.getLocalProxySha256( docValue ),
		
		ozProxyLink = ozProxyLink,
		
		ozProxySize = ozProxySize,
		
		ozXmpDevelopLink = WFCatFileSpec.getXmpDevelopLinkFromDocValue( docValue ),

		ozXmpLink = WFCatFileSpec.getXmpLinkFromDocValue( docValue ),

		develop = NmbDocValueDevelopSettingsUtils_getSanitizedDocValueDevelopStanza(
					docValue ),
		
		renditions = renditions,
		
		proxies = proxies,
		
		master = master,
		
		customCameraProfileDigest = customCameraProfileDigest,
		
		customCameraProfileMissing = customCameraProfileMissing,
	}

end

--------------------------------------------------------------------------------

local function maybeExtractRenderInfoFromDocValue( docValue )

	if not WFCatRenderCapabilities.isRenderableAsset( docValue ) then

		if logger.will_info then

			logger:infof( "not extracting render info for docValue: %s", 
							WFTableArchiving.tableToString( docValue ) )

		end

		return

	end

	return extractRenderInfoFromDocValue( docValue )

end

--------------------------------------------------------------------------------

local function extractRenderInfoOrReportFailure( packedRenderInfosResults, 
													docValue )

	local renderInfos, missingRenderInfoIds =
			packedRenderInfosResults[ 1 ], packedRenderInfosResults[ 2 ]

	local maybeRenderInfo = maybeExtractRenderInfoFromDocValue( docValue )

	if maybeRenderInfo then

		renderInfos[ #renderInfos + 1 ] = maybeRenderInfo

	else

		missingRenderInfoIds[ #missingRenderInfoIds + 1 ] = docValue._localId

	end

	return packedRenderInfosResults

end

--------------------------------------------------------------------------------

local function extractRenderInfos( docValues )

	local packedRenderInfosResults = { {}, {} }
	
	return unpack( WFFunctionUtils_reduce( 
						packedRenderInfosResults,
						extractRenderInfoOrReportFailure, docValues ) )

end

--------------------------------------------------------------------------------

local function renderInfoStream( docStore, localAssetIds )

	if logger.will_info then

		logger:infof( "Calling bulkDocValue on { %s }",  
						table.concat( localAssetIds, ", " ) )

	end

	local origAssetIdCount = #localAssetIds

	local renderInfosStream = 
			docStore
				:bulkDocValue( localAssetIds )
				:map( extractRenderInfos )

	return logger.will_info and
			renderInfosStream	
				:tap(
					function( renderInfos, missingRenderInfoIds )

						logger:infof( "Got %s render infos and %s errorIds after requesting %s",  
										#renderInfos, #missingRenderInfoIds, origAssetIdCount )

					end ) or
			renderInfosStream

end

--------------------------------------------------------------------------------

local function maybeReportRenderInfoUpdate( watchedLocalAssetIdsSet, docValue )

	local localAssetId = docValue._localId 

	if not localAssetId then

		return watchedLocalAssetIdsSet

	end

	if not WFCatRenderCapabilities.isRenderableAsset( docValue ) then

		return watchedLocalAssetIdsSet

	end

	if not watchedLocalAssetIdsSet[ localAssetId ] then

		return watchedLocalAssetIdsSet

	end

	-- TODO we could improve performance by saving render info
	-- in watchedLocalAssetIdsSet when it is generated then
	-- compare it with a previous value.
	return watchedLocalAssetIdsSet, { extractRenderInfoFromDocValue( docValue ) }

end

--------------------------------------------------------------------------------

local function changedRenderInfoStream( docStore, localAssetIds )

	local watchedLocalAssetIdsSet = KSTableUtils_mapArrayToSet( localAssetIds )

	return docStore
			:scannedFilteredMappedChangesFeed( 
				"NmbDocValueToRenderInfoStream", "maybeReportRenderInfoUpdate",
				watchedLocalAssetIdsSet )

end

--==============================================================================

return {

	renderInfoStream = renderInfoStream,

	changedRenderInfoStream = changedRenderInfoStream,

	maybeExtractRenderInfoFromDocValue = maybeExtractRenderInfoFromDocValue,

	maybeReportRenderInfoUpdate = maybeReportRenderInfoUpdate,

}
