--[[----------------------------------------------------------------------------

NmbFileAccessDatabase.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.
s
------------------------------------------------------------------------------]]

--[[
	AdobePatentID="P7167-US"
]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSDate = require 'KSDate'
local WFCloseSupport = require 'WFCloseSupport'
local WFRxScheduler = require 'WFRxScheduler'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'
local WFDigest = require 'WFDigest'

	-- WF.sqlite.database
local WFSQLiteDatabase = require 'WFSQLiteDatabase'
local WFSQLiteDatabaseSchema = require 'WFSQLiteDatabaseSchema'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local md5 = assert( WFDigest.MD5, "Cannot find MD5 digest" )

local logger = WFLogger 'NmbFileAccessDatabase'

local KSDate_timeNow = assert( KSDate.timeNow, "KSDate.timeNow does not exist" )

local ACCESS_QUERY_DEFAULT_LIMIT = 100

local runUntilDone = WFRxRunUntilDone

--------------------------------------------------------------------------------

local function fromBool( bool )

	return bool and 1 or 0

end

--============================================================================--

-- @mark NmbFileAccessDatabase namespace

--[[!

	\class NmbFileAccessDatabase

	\brief The file access database is a traditional persistent SQLite database which
	can be used to track binary data access on disk. This database can be easily queried to
	identify candidates for purging when we need to make more disk space available.
	
--]]

local NmbFileAccessDatabase = {

	_docStore = false,
		-- (WFRxDocStore)

	_fileAccessDatabase = false,
		-- (WFSQLiteDatabase)
	
	_touchCache = false,
		-- (table) Used to track updates that haven't yet been written to the db.
		-- key: local id
		-- value: table {
		--    touchDate = <timestamp>
		--    hasProxy = <boolean>
		--    hasOriginal = <boolean>
		-- }
		
	_cacheSize = false,
		-- (number) The number of entries in the touch cache. If the cache grows too large,
		-- we commit the data to the db.
		
	_writeStreamSubscription = false,
		-- ( Rx.Subscription )
		
	_changeCounterStream = false,
		-- ( Rx.ReplaySubject ) Fires with the local change counter.
		
	_lastTempIds = false,
		
	_isDebugMode = false,
		-- (boolean) If true, updates to assets are immeadiately commited to db.
		
	WFCloseSupport.addCloseSupportMethods,	
	
}

--------------------------------------------------------------------------------

local function commitTouchTimesToDb( self )

	if not next( self._touchCache ) then
		return
	end

	local sql = 'SAVEPOINT updateLastAccess'

	self._fileAccessDatabase:exec( sql )

	local statement

	for localId, entry in pairs( self._touchCache ) do
	
		if type( entry ) == 'table' then
			
			if entry.delete then
		
				statement = [[DELETE FROM lastAccess WHERE localId = ?]]
				self._fileAccessDatabase:exec( statement, localId )
							
			else
		
				if entry.newEntryTouchDate then
				
					-- Insert a new entry into the DB if it does't exist.
					
					statement = [[INSERT OR IGNORE INTO lastAccess( localId, accessTimestamp ) VALUES( ?, ? )]]
					self._fileAccessDatabase:exec( statement, localId, entry.newEntryTouchDate )
				
				end
			
				if entry.touchDate then
				
					-- Insert a new entry into the DB if it does't exist.
					
					statement = [[INSERT OR IGNORE INTO lastAccess( localId, accessTimestamp ) VALUES( ?, ? )]]
					self._fileAccessDatabase:exec( statement, localId, entry.touchDate )
					
					-- If the entry exists, update the touch date. We're not using INSERT or REPLACE
					-- in the above SQL because we want to maintain the values of the other fields in
					-- the row; REPLACE would squash them.
					
					statement = [[UPDATE lastAccess SET accessTimestamp = ? WHERE localId = ?]]
					self._fileAccessDatabase:exec( statement, entry.touchDate, localId )
					
				end
				
				local currentTime = KSDate_timeNow()
				
				for field, value in pairs( entry.simple or {} ) do
					
					statement = [[INSERT OR IGNORE INTO lastAccess( localId, accessTimestamp ) VALUES( ?, ? )]]
					self._fileAccessDatabase:exec( statement, localId, currentTime )
						-- If the asset doesn't exist in the db, insert it.
						
					if value == 'null' then
						statement = string.format( "UPDATE lastAccess SET %s = null WHERE localId = ?", field )
						self._fileAccessDatabase:exec( statement, localId )
					else
						statement = string.format( "UPDATE lastAccess SET %s = ? WHERE localId = ?", field )
						self._fileAccessDatabase:exec( statement, value, localId )
					end
					
				end
			
			end
			
		end
		
	end
	
	local continueFromChangeCounter = self._touchCache[ 'changeCounter' ]
	
	if continueFromChangeCounter then
		self._fileAccessDatabase:exec( 
				[[UPDATE tableStatus SET needsUpdateFrom = ? WHERE tableName = 'lastAccess']], continueFromChangeCounter )
	end
	
	self._touchCache = {}
	self._cacheSize = 0
	
	sql = 'RELEASE SAVEPOINT updateLastAccess'
	
	self._fileAccessDatabase:exec( sql )

end

--------------------------------------------------------------------------------

local function maybeResetTable( self, tableName, tableSqlSchema )

	-- Similar to the index store, we track a hash of the sql schema used
	-- to create the tables in the MRU. If the hash of the sql schema changes,
	-- we dump the data and regather it.

	local db = self._fileAccessDatabase

	db:exec( [[DROP TABLE IF EXISTS lastAccessV2]] )
	db:exec( [[DROP TABLE IF EXISTS lastAccessV3]] )
		-- SPAM [jandrew 2017-10-02]: These tables were used prior to switching to
		-- the new sql schema hash approach for table versioning, so we should 
		-- manually delete them if they happen to exist.

	local rows = self:gatherArray( 
			[[SELECT sqlSchemaHash 
				FROM tableStatus
				WHERE tableName = ?]], 
			tableName )

	local cachedSqlSchemaHash  = rows[ 1 ]
	
	local dbUpdaterVersion = 2
		-- SPAM [jandrew 2017-10-25]: 
		-- If NmbPurgeDbUpdater changes, we may need to increment dbUpdaterVersion so
		-- that the MRU table is rebuilt.
	
	local sqlSchemaHash = md5.digest( string.format( '%s%d', tableSqlSchema, dbUpdaterVersion ) )

	if sqlSchemaHash ~= cachedSqlSchemaHash then

		local sql = 'SAVEPOINT resetTable'

		db:exec( sql )
		
		db:exec( [[DELETE FROM tableStatus WHERE tableName == ?]],
				tableName )
		
		db:exec( [[INSERT INTO tableStatus VALUES( ?, ?, ? )]],
				tableName,
				0,
				sqlSchemaHash )
		
		db:exec( string.format( [[DELETE FROM %s]], tableName ) )
		
		sql = 'RELEASE SAVEPOINT resetTable'
			
		db:exec( sql )
			
	end

	local changeCounter = self:gatherArray( 
			[[SELECT needsUpdateFrom FROM tableStatus WHERE tableName = 'lastAccess']] )[ 1 ] or 0
					
	self._changeCounterStream:onNext( changeCounter )

end

--------------------------------------------------------------------------------

--[[!

	\brief Creates a persistent SQLite database for tracking binary data access.

	\param dbPath (string) The path to the db.

	\return NmbFileAccessDatabase instance object
	
--]]


function NmbFileAccessDatabase:__init( docStore, indexStore, dbPath, optOverrides )

	check( 'NmbFileAccessDatabase', 'WFRxDocStore', 'WFRxIndexStore', '+string', '?table' )

	logger:debugf( "Creating NmbFileAccessDatabase at path: %s", dbPath )

	local schema = WFSQLiteDatabaseSchema( '1.0' )

	-- Create a table to track the update status and schema of the tables.

	schema:addTable
			[[CREATE TABLE tableStatus( 
				tableName UNIQUE,
				needsUpdateFrom,
				sqlSchemaHash )]]
		:onConflict( 'replace' )

	-- Create the main table to track last access times, and other important data.

	local lastAccessSqlSchema =
			[[CREATE TABLE lastAccess( 
					localId PRIMARY KEY,
					accessTimestamp NOT NULL,
					volumeGuidUsedByProxy,
					ozHasProxy,
					proxyFileSize,
					volumeGuidUsedByOriginal,
					ozHasOriginal,
					originalFileSize,
					isImportant )]]

	schema:addTable( lastAccessSqlSchema )
		:onConflict 'replace'
		
	-- Add some indexes
		
	schema:addIndex
			[[CREATE INDEX lastAccess_proxySearchIndex
			ON lastAccess(
					isImportant,
					volumeGuidUsedByProxy, 
					ozHasProxy, 
					accessTimestamp )]]
	
	schema:addIndex
			[[CREATE INDEX lastAccess_originalSearchIndex
			ON lastAccess(
					isImportant,
					volumeGuidUsedByOriginal, 
					ozHasOriginal, 
					accessTimestamp )]]
	
	schema:addIndex
			[[CREATE INDEX lastAccess_accessTimestamp
			ON lastAccess( accessTimestamp )]]
	
	schema:addTable [[CREATE TEMPORARY TABLE TEMP_ids( docId )]]
	
	if dbPath ~= ':memory:' then
		local parentDir = KSPathUtils.parent( dbPath )
		WFFileUtils.mkdir( parentDir )
	end

	optOverrides = optOverrides or {}

	self._isDebugMode = optOverrides.isDebugMode
	
	self._fileAccessDatabase = WFSQLiteDatabase( dbPath, schema, true )
	
	self._docStore = docStore
	
	self._changeCounterStream = Rx.ReplaySubject( 1 )
	
	self._touchCache = {}
	
	self._cacheSize = 0
	
	maybeResetTable( self, 'lastAccess', lastAccessSqlSchema )

	self._writeStreamSubscription = Observable.timer( 0, optOverrides.writeInterval or 60, WFRxScheduler.sharedScheduler() )
		:subscribe(
			function()
				commitTouchTimesToDb( self )
			end )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Closes the database.
	
	\return void
	
--]]

function NmbFileAccessDatabase:close()
										
	check( 'NmbFileAccessDatabase' )
	
	local writeStreamSubscription = self._writeStreamSubscription
	self._writeStreamSubscription = nil
	
	if writeStreamSubscription then
		writeStreamSubscription:dispose()
	end
	
	-- Clear the cache
	commitTouchTimesToDb( self )
	
	local fileAccessDatabase = self._fileAccessDatabase
	self._fileAccessDatabase = nil
	
	if fileAccessDatabase then
		fileAccessDatabase:close()
	end
	
	local changeCounterStream = self._changeCounterStream
	self._changeCounterStream = nil
	
	if changeCounterStream then
		changeCounterStream:dispose()
	end
	
end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:getContinueFromChangeCounter()

	check( 'NmbFileAccessDatabase' )

	return self._changeCounterStream
	
		:take( 1 )
	
end

--------------------------------------------------------------------------------

local function getLocalIdStream( self, localIdOrDocId )

	local localIdStream = Observable.just( localIdOrDocId )

	if type( localIdOrDocId ) == 'string' then
	
		localIdStream = self._docStore:docValue( localIdOrDocId ):pluck( '_localId' )
	
	end
	
	return localIdStream
	
		:filter( function( localId )
	
			if not localId or localId <= 0 then
	
				logger:warnf( "Invalid localId: %s", localId )
				return false
			
			end
			
			return true
			
		end )
		
end

--------------------------------------------------------------------------------

local function addToTouchCache( self, localIdOrDocId, key, value, maybeSimpleUpdate )

	logger:debugf( "Updating %s status of asset %s to %s.", key, localIdOrDocId, value )

	return getLocalIdStream( self, localIdOrDocId )
	
		:map( function( localId )
			
			local entry = self._touchCache[ localId ]
			
			if not entry then
			
				self._cacheSize = self._cacheSize + 1
				entry = {}
			
			end
			
			if maybeSimpleUpdate then
			
				local simpleUpdates = entry[ 'simple' ] or {}
				simpleUpdates[ key ] = value or 'null'
				entry[ 'simple' ] = simpleUpdates
				
			else
			
				entry[ key ] = value	
			
			end
			
			self._touchCache[ localId ] = entry
			
			if self._cacheSize > 100 or self._isDebugMode then
				commitTouchTimesToDb( self )
			end
			
			return
			
		end )
		
		:flatMapLatest( function()
		
			return Observable.empty()
		
		end )
		
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

--[[!

	\brief Add a new entry, or update an existing entry in the db.

	\param localId (number) Local ID of the asset on disk.
	\param optDate (timestamp) optionally override the last access date.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:touch( localIdOrDocId, optDate )

	check( 'NmbFileAccessDatabase', 'number|string', '?number' )

	local touchDate = optDate or KSDate_timeNow()

	return addToTouchCache( self, localIdOrDocId, 'touchDate', touchDate )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Add multiple new entries, or ignore the updates if they already exists.

	\param localIdsOrDocIds (table) table of local ids or doc ids
	\param optDate (timestamp) optionally override the last access date.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:touchMultipleIfNew( localIdsOrDocIds, optDate )

	check( 'NmbFileAccessDatabase', 'table', '?number' )

	logger:debugf( "touchIfNew for %d assets.", #localIdsOrDocIds )
	
	if #localIdsOrDocIds == 0 then
		return Observable.empty()
	end
	
	local touchDate = optDate or KSDate_timeNow()

	local localIdStream = Observable.just( localIdsOrDocIds )
		:flatMap( function( localIds )
			return Rx.Observable.from( localIds )
		end )
		
	if type( localIdsOrDocIds[ 1 ] ) == 'string' then
		
		localIdStream = self._docStore:bulkDocValue( localIdsOrDocIds )
			:flatMap( function( docValues )
				return Rx.Observable.from( docValues )
			end )
			:pluck( '_localId' )
			
	end
		
	return localIdStream
		:tap( function( localId )
			addToTouchCache( self, localId, 'newEntryTouchDate', touchDate )
		end )
		:takeLast( 0 )
		:call( runUntilDone )
		
end

--------------------------------------------------------------------------------

--[[!

	\brief Update the change counter for this DB.

	\param changerCounter (number) The change counter at the time of the update.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateChangeCounter( changeCounter )

	check( 'NmbFileAccessDatabase', 'number' )

	local key = 'changeCounter'
	
	local counter = self._touchCache[ key ] or 0
	
	local newChangeCounter = math.max( counter, changeCounter )
	
	self._changeCounterStream:onNext( newChangeCounter )
	
	self._touchCache[ key ] = newChangeCounter

	return Observable.empty()
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's isImportant status in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param isImportant (boolean) Indicates if the asset is important.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateImportance( localIdOrDocId, isImportant )

	check( 'NmbFileAccessDatabase', 'number|string', 'boolean' )

	return addToTouchCache( self, localIdOrDocId, 'isImportant', fromBool( isImportant ), true )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's ozHasOriginal status in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param ozHasOriginal (boolean) Does oz have the original.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateOzOriginalStatus( localIdOrDocId, ozHasOriginal )

	check( 'NmbFileAccessDatabase', 'number|string', 'boolean' )

	return addToTouchCache( self, localIdOrDocId, 'ozHasOriginal', fromBool( ozHasOriginal ), true )

end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's ozHasProxy status in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param ozHasProxy (boolean) Does oz have the proxy.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateOzProxyStatus( localIdOrDocId, ozHasProxy )

	check( 'NmbFileAccessDatabase', 'number|string', 'boolean' )

	return addToTouchCache( self, localIdOrDocId, 'ozHasProxy', fromBool( ozHasProxy ), true )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's proxy size in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param proxyFileSize (number) The size of the proxy.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateProxySize( localIdOrDocId, proxyFileSize )

	check( 'NmbFileAccessDatabase', 'number|string', 'number' )

	return addToTouchCache( self, localIdOrDocId, 'proxyFileSize', proxyFileSize, true )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's original size in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param originalFileSize (number) The size of the original.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateOriginalSize( localIdOrDocId, originalFileSize )

	check( 'NmbFileAccessDatabase', 'number|string', 'number' )

	return addToTouchCache( self, localIdOrDocId, 'originalFileSize', originalFileSize, true )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's local proxy status in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param volumeGuidUsedByProxy (optional string) The volume guid where the proxy is stored. 
		Nil if no local proxy.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:updateLocalProxyStatus( localIdOrDocId, volumeGuidUsedByProxy )

	check( 'NmbFileAccessDatabase', 'number|string', '?string' )

	return addToTouchCache( self, localIdOrDocId, 'volumeGuidUsedByProxy', volumeGuidUsedByProxy, true )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Update an existing entry's local original status in the db.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	\param volumeGuidUsedByOriginal (optional string) The volume guid where the original is stored. 
		Nil if no local original.
		
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]


function NmbFileAccessDatabase:updateLocalOriginalStatus( localIdOrDocId, volumeGuidUsedByOriginal )

	check( 'NmbFileAccessDatabase', 'number|string', '?string' )

	return addToTouchCache( self, localIdOrDocId, 'volumeGuidUsedByOriginal', volumeGuidUsedByOriginal, true )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Delete an asset from the database for a given local Id.

	\param localIdOrDocId (number|string) localId or docId of the asset.
	
	\return <code>Observable< nil ></code> that completes on success. There's
  		no need to subscribe to this observable; it's only used by unit tests.
	
--]]

function NmbFileAccessDatabase:delete( localIdOrDocId )

	check( 'NmbFileAccessDatabase', 'number|string' )
	
	logger:debugf( "Deleting asset %s.", localIdOrDocId )
	
	return addToTouchCache( self, localIdOrDocId, 'delete', true )
	
		:takeLast( 0 )
	
		:call( runUntilDone )
	
end

--------------------------------------------------------------------------------

local function convertSqlRowsToUsefulLuaTables( arr )

	local usefulArr = {}

	for idx, row in pairs( arr ) do

		local localId = row[ 1 ]
		local lastAccessed = row[ 2 ]
		local volumeGuidUsedByProxy = row[ 3 ]
		local ozHasProxy = row[ 4 ]
		local proxyFileSize = row[ 5 ]
		local volumeGuidUsedByOriginal = row[ 6 ]
		local ozHasOriginal = row[ 7 ]
		local originalFileSize = row[ 8 ]
		local isImportant = row[ 9 ]

		usefulArr[ idx ] = {
			localId = localId,
			lastAccessed = lastAccessed,
			volumeGuidUsedByProxy = volumeGuidUsedByProxy,
			ozHasProxy = ozHasProxy,
			proxyFileSize = proxyFileSize,
			volumeGuidUsedByOriginal = volumeGuidUsedByOriginal,
			ozHasOriginal = ozHasOriginal,
			originalFileSize = originalFileSize,
			isImportant = isImportant
		}

	end

	return usefulArr

end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:gatherArrayOfArray( ... )

	if not WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		commitTouchTimesToDb( self )
  		return self._fileAccessDatabase:gatherArrayOfArray( ... )
	else
		return {}
	end
	
end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:gatherArray( ... )

	if not WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		commitTouchTimesToDb( self )
  		return self._fileAccessDatabase:gatherArray( ... )
	else
		return {}
	end
end

--------------------------------------------------------------------------------

--[[!

	\brief TEST USE ONLY - dump the contents of the db.

	\param void
	
	\return Contents of the db
	
--]]

function NmbFileAccessDatabase:DEBUG_dumpContents()

	check( 'NmbFileAccessDatabase' )

	return convertSqlRowsToUsefulLuaTables( self:gatherArrayOfArray( [[SELECT * FROM lastAccess]] ) )

end

--------------------------------------------------------------------------------

local function _getRecentlyAccessed( self, limit, sort, skip )

	sort = sort or 'ASC'

	skip = skip or 0

	local function getNextBatch()

		skip = skip + limit
		return _getRecentlyAccessed( self, limit, sort, skip ), getNextBatch
	
	end

	return self:gatherArray( 
			[[SELECT * 
			FROM lastAccess
			ORDER BY accessTimestamp ]] .. sort .. -- SPAM: '?' Doesn't work with sort order.
			[[ LIMIT ? OFFSET ?]],
			limit,
			skip ), getNextBatch
	
end

--------------------------------------------------------------------------------

local function _getLeastRecentlyAccessedIdsInWindow( self, volumeGuid, binaryTypeFilter, timeWindow, inactiveAssetsOnly, limit, skipNMostRecent, offset )

	offset = offset or 0

	local function getNextBatch()

		offset = offset + limit
		return _getLeastRecentlyAccessedIdsInWindow( self, volumeGuid, binaryTypeFilter, timeWindow, inactiveAssetsOnly, limit, skipNMostRecent, offset ), getNextBatch
	
	end

	local daysToSeconds = function( days )
		return days * 24 * 60 * 60
	end

	local beforeNDaysAgo = timeWindow.beforeNDaysAgo
	local beforeNDaysAgoTimestamp = KSDate_timeNow() - daysToSeconds( beforeNDaysAgo )

	-- Build the binary status query filter and the ORDER BY statement
	
	local binaryFilter = " "
	local orderByStatement = " "
	
	if binaryTypeFilter == 'proxy' then
	
		binaryFilter = string.format( " AND volumeGuidUsedByProxy == '%s' AND ozHasProxy = 1 ", volumeGuid )
		orderByStatement = " ORDER BY isImportant ASC, proxyFileSize DESC, accessTimestamp ASC "
	
	elseif binaryTypeFilter == 'original' then
	
		binaryFilter = string.format( " AND volumeGuidUsedByOriginal == '%s' AND ozHasOriginal = 1 ", volumeGuid )
		orderByStatement = " ORDER BY isImportant ASC, originalFileSize DESC, accessTimestamp ASC "
		
	else
	
		binaryFilter = string.format( 
				[[ AND ( ( volumeGuidUsedByProxy == '%s' AND ozHasProxy = 1 ) 
					OR ( volumeGuidUsedByOriginal = '%s' AND ozHasOriginal = 1 ) ) ]],
				volumeGuid,
				volumeGuid )
				
		orderByStatement = " ORDER BY isImportant ASC, originalFileSize DESC, proxyFileSize DESC, accessTimestamp ASC "
		
	end

	-- Maybe add an inactive asset filter
	
	local inactiveFilter = " "
	
	if inactiveAssetsOnly then
		inactiveFilter = " AND isImportant = 0 "
	end

	-- Fetch the most recently accessed timestamps
	
	local mostRecentlyAccessed = self:gatherArray( 
			[[ SELECT accessTimestamp 
				FROM lastAccess 
				ORDER BY accessTimestamp DESC LIMIT ? ]],
				skipNMostRecent )
 
 	-- Calculate the lower time bound
 
	local lowerTimeBound = mostRecentlyAccessed[ #mostRecentlyAccessed ] or math.huge
	
	lowerTimeBound = math.min( beforeNDaysAgoTimestamp, lowerTimeBound )
	
	-- Calculate the upper time bound, if there is one.
	
	local upperTimeBoundFilter = " "
	
	if timeWindow.afterNDaysAgo then
	
		local afterNDaysAgo = timeWindow.afterNDaysAgo
		local afterNDaysAgoTimestamp = KSDate_timeNow() - daysToSeconds( afterNDaysAgo )
	
		upperTimeBoundFilter = string.format( " AND accessTimestamp >= %d ", afterNDaysAgoTimestamp )
	
	end
	
	-- Run the query
	
	return self:gatherArray( 
			[[ SELECT * 
				FROM lastAccess
				WHERE accessTimestamp < ? ]]
				.. upperTimeBoundFilter .. inactiveFilter .. binaryFilter .. orderByStatement ..
				[[ LIMIT ? OFFSET ? ]],
			lowerTimeBound,
			limit,
			offset ), getNextBatch

end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:waitUntilUpToDate()

	return Observable.combineLatest( self._docStore:changeCounter(), self._changeCounterStream )
		
		:filter( function( docStoreChangeCounter, localChangeCounter )
		
			-- Wait until the NmbFileAccessDatabase is up to date with the Doc Store. 
		
			return localChangeCounter >= docStoreChangeCounter
		
		end )
		
		:take( 1 )
		
		:map( function()
		
			commitTouchTimesToDb( self )
			
			return true
		
		end )

end

--------------------------------------------------------------------------------

local function _getLeastRecentlyAccessedIdsInWindowStream( self, volumeGuid, binaryTypeFilter, timeWindow, limit, skipNMostRecent, offset )

	return self:waitUntilUpToDate()
		
		:map( function()
		
			limit = limit or ACCESS_QUERY_DEFAULT_LIMIT
	
			skipNMostRecent = skipNMostRecent or 0
		
			local inactiveAssetsOnly = false
		
			return _getLeastRecentlyAccessedIdsInWindow( self, volumeGuid, binaryTypeFilter, timeWindow, inactiveAssetsOnly, limit, skipNMostRecent, offset )
		
		end )

end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve the current list of purge candidates from the db, sorted
	by time of last access, but only for assets that haven't been accessed within
	N days.

	\param volumeGuid (string) Only return assets that live on this specific volume.
	\param timeWindow (table)
	\param limit (optional number) Limit the number of assets to search for.
	\param skipNMostRecent (optional number) Skip the N most recent assets from the query.

	\return <code>Observable< array ></code> An array of purge candidates, along with a function for getting the next
	batch of limited data. Example:
	
		local fiveRecentlyAccessed, getNextBatchFunc = NmbFileAccessDatabase.getLeastRecentlyAccessedInWindow( 10, 5 )
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		
		and so on.
	
--]]

function NmbFileAccessDatabase:getLeastRecentlyAccessedInWindow( volumeGuid, timeWindow, limit, skipNMostRecent )

	check( 'NmbFileAccessDatabase', 'string', 'table', '?number', '?number' )

	return _getLeastRecentlyAccessedIdsInWindowStream( self, volumeGuid, nil, timeWindow, limit, skipNMostRecent )

end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:getLeastRecentlyAccessedProxiesInWindow( volumeGuid, timeWindow, limit, skipNMostRecent )

	check( 'NmbFileAccessDatabase', 'string', 'table', '?number', '?number' )
	
	return _getLeastRecentlyAccessedIdsInWindowStream( self, volumeGuid, 'proxy', timeWindow, limit, skipNMostRecent )

end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:getLeastRecentlyAccessedOriginalsInWindow( volumeGuid, timeWindow, limit, skipNMostRecent )

	check( 'NmbFileAccessDatabase', 'string', 'table', '?number', '?number' )
	
	return _getLeastRecentlyAccessedIdsInWindowStream( self, volumeGuid, 'original', timeWindow, limit, skipNMostRecent )

end

--------------------------------------------------------------------------------

local function _getInactiveAssetsInWindowStream( self, volumeGuid, binaryTypeFilter, timeWindow, limit, skipNMostRecent, offset )

	return self:waitUntilUpToDate()
		
		:map( function()
		
			limit = limit or ACCESS_QUERY_DEFAULT_LIMIT
	
			skipNMostRecent = skipNMostRecent or 0
		
			local inactiveAssetsOnly = true
		
			return _getLeastRecentlyAccessedIdsInWindow( 
					self, volumeGuid, binaryTypeFilter, timeWindow, inactiveAssetsOnly, limit, skipNMostRecent, offset )
		
		end )

end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve the current list of inactive, proactive purge candidates from the 
	db, sorted by time of last access, but only for assets that haven't been accessed 
	within N days.

	\param volumeGuid (string) Only return assets that live on this specific volume.
	\param timeWindow (table)
	\param limit (optional number) Limit the number of assets to search for.
	\param skipNMostRecent (optional number) Skip the N most recent assets from the query.

	\return <code>Observable< array ></code> An array of purge candidates, along with a function for getting the next
	batch of limited data. Example:
	
		local fiveRecentlyAccessed, getNextBatchFunc = NmbFileAccessDatabase.getInactiveAssetsInWindow( ... )
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		
		and so on.
	
--]]

function NmbFileAccessDatabase:getInactiveAssetsInWindow( volumeGuid, timeWindow, limit, skipNMostRecent )

	check( 'NmbFileAccessDatabase', 'string', 'table', '?number', '?number' )
	
	return _getInactiveAssetsInWindowStream( self, volumeGuid, nil, timeWindow, limit, skipNMostRecent )

end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:getInactiveProxiesInWindow( volumeGuid, timeWindow, limit, skipNMostRecent )

	check( 'NmbFileAccessDatabase', 'string', 'table', '?number', '?number' )
	
	return _getInactiveAssetsInWindowStream( self, volumeGuid, 'proxy', timeWindow, limit, skipNMostRecent )

end

--------------------------------------------------------------------------------

function NmbFileAccessDatabase:getInactiveOriginalsInWindow( volumeGuid, timeWindow, limit, skipNMostRecent )

	check( 'NmbFileAccessDatabase', 'string', 'table', '?number', '?number' )
	
	return _getInactiveAssetsInWindowStream( self, volumeGuid, 'original', timeWindow, limit, skipNMostRecent )

end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve the current list of purge candidates from the db, sorted
	by time of last access.

	\param limit (optional number) Limit the number of assets to search for.
	
	\return An array of purge candidates, along with a function for getting the next
	batch of limited data. Example:
	
		local fiveRecentlyAccessed, getNextBatchFunc = NmbFileAccessDatabase.getLeastRecentlyAccessed( 5 )
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		fiveRecentlyAccessed, getNextBatchFunc = getNextBatchFunc()
		...
		
		and so on.
	
--]]

function NmbFileAccessDatabase:getLeastRecentlyAccessed( limit )

	check( 'NmbFileAccessDatabase', '?number' )
	
	limit = limit or ACCESS_QUERY_DEFAULT_LIMIT
	
	return _getRecentlyAccessed( self, limit )

end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve the N most recently accessed assets.

	\param limit (optional number) Limit the number of assets to search for.
	
	\return An array of purge candidates, along with a function for getting the next
	batch of limited data.
	
--]]

function NmbFileAccessDatabase:getMostRecentlyAccessed( limit )

	check( 'NmbFileAccessDatabase', '?number' )
	
	limit = limit or ACCESS_QUERY_DEFAULT_LIMIT
	
	return _getRecentlyAccessed( self, limit, 'DESC' )

end

--------------------------------------------------------------------------------

return Rx.Class 'NmbFileAccessDatabase'( NmbFileAccessDatabase )
