--[[----------------------------------------------------------------------------

NmbLocalRenditionDigest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]


	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFDigest = require 'WFDigest'

local NmbDocValueToRenderInfoStream = require "NmbDocValueToRenderInfoStream"


local logger = require 'WFLogger'( 'NmbLocalRenditionDigest' )
--------------------------------------------------------------------------------

local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt, 
										"KSTableUtils.derefOpt not defined" )

--------------------------------------------------------------------------------

local md5 = assert( KSTableUtils_derefOpt( WFDigest, "MD5", "digest" ),
						"WFDigest.MD5.digest does not exist" )

--------------------------------------------------------------------------------

local function isValidXmpCameraRaw( xmpCameraRaw ) 
	-- SPAM [matthewj 2016-08-08]: This isn't a rigorous check of the xmpCameraRaw, but it should catch 
	-- the most egregious incorrect xmpCameraRaw. xmpCameraRaw can be either a string with the xmp in it
	-- or a table containing the sha256 of the xmp string for external xmp handling
	local xmpCameraRawType = type( xmpCameraRaw )
	
	if xmpCameraRawType == 'table' then
	
		return xmpCameraRaw.sha256 ~= nil
	
	end
	
	return xmpCameraRawType == 'string'
	
end

--------------------------------------------------------------------------------

local function generateDigest( renditionSha256, xmpCameraRawForDigest, 
								maybeUserOrientation,
								customCameraProfileDigest,
								customCameraProfileMissing )
		
	-- Include the orientation in the digest so changes to it will kick off new renders.
	
	-- all images except custom camera profiles use the digest below
	local compositeDevSettingsForDigest =
			string.format( "%s%s%s",
					xmpCameraRawForDigest,
					maybeUserOrientation or 1,
					renditionSha256 )

	-- if an image has a custom camera profile, the dev settings digest gets
	-- two more seeds, the custom camera profile digest and an "empty
	-- unless missing" status string

	if customCameraProfileDigest then

		local maybeMissingCustomCameraProfileStatus =
				customCameraProfileMissing and "missing" or ""

		compositeDevSettingsForDigest =
				string.format( "%s%s%s",
						compositeDevSettingsForDigest,
						customCameraProfileDigest,
						maybeMissingCustomCameraProfileStatus )

	end

    return md5( compositeDevSettingsForDigest )

end

--------------------------------------------------------------------------------

local function computeTemporaryDefaultForOrientation( 
					ozAssetId, temporaryXmpCameraRaw, userOrientation,
					customProfileDigest, customProfileMissing )

	return generateDigest( ozAssetId, temporaryXmpCameraRaw, userOrientation,
		customProfileDigest, customProfileMissing )

end

--------------------------------------------------------------------------------
-- renderInfo comes from NmbDocValueToRenderInfoStream
--------------------------------------------------------------------------------

local function computeFromRenderInfo( renderInfo )

	local xmpCameraRaw = KSTableUtils_derefOpt( 
							renderInfo, "develop", "xmpCameraRaw" ) 
	
	local userOrientation = KSTableUtils_derefOpt( 
							renderInfo, "develop", "userOrientation" )
	
	-- SPAM [matthewj 2015-12-21]: Ran into issues where the xmpCameraRaw
	-- contained an invalid, non-string value (the image's width value, in
	-- fact) and the md5 calculation would then blow up. So if it's nil or not
	-- a string, then use the default <xmp />
	xmpCameraRaw = isValidXmpCameraRaw( xmpCameraRaw ) and xmpCameraRaw or "<xmp />"

	local xmpCameraRawForDigest = type( xmpCameraRaw ) == 'table' 
								and xmpCameraRaw.sha256 
								or xmpCameraRaw

	-- Include SHA256 of the parent binary used to create the renders so that we will 
	-- generate new renders when the binary's SHA256 changes due to destructive edits. 
	
	local hasLocalOriginal = not not KSTableUtils_derefOpt( renderInfo, "pathToOriginal" )
	local hasLocalProxy = not not KSTableUtils_derefOpt( renderInfo, "pathToProxy" )
	
	local assetSha256 = KSTableUtils_derefOpt( renderInfo, "sha256" )
	local localOriginalSha256 = KSTableUtils_derefOpt( renderInfo, "localOriginalSha256" )
	local localProxySha256 = KSTableUtils_derefOpt( renderInfo, "localProxySha256" )

	local renditionSha256 = assetSha256

	-- If the asset has a local original, that the renders are generated from it.
	-- If it doesn't, but it has a local proxy, that the renders are generated from it.
	-- Otherwise we download something.
	
	if hasLocalOriginal and localOriginalSha256 then
	
		renditionSha256 = localOriginalSha256
		
	elseif hasLocalProxy and localProxySha256 then
	
		renditionSha256 = localProxySha256
		
	end
	
	logger:debugf( "LocalId:%s: profileDigest:%s isMissing:%s",
		renderInfo._localId, 
		renderInfo.customCameraProfileDigest,
		renderInfo.customCameraProfileMissing )

	local digest =
			generateDigest( renditionSha256, xmpCameraRawForDigest,
							userOrientation, 
							renderInfo.customCameraProfileDigest,
							renderInfo.customCameraProfileMissing )

	return digest, xmpCameraRaw, userOrientation

end

--------------------------------------------------------------------------------

local function maybeComputeFromDocValue( docValue )

	local renderInfo = 
		NmbDocValueToRenderInfoStream.maybeExtractRenderInfoFromDocValue( 
			docValue )

	if not renderInfo then

		return nil

	end

	return computeFromRenderInfo( renderInfo )

end

--==============================================================================

return {

	-- use Now suffix for synchronous API, and Stream suffix for async API
	-- 
	-- the Now API don't check for camera profile availability, so should
	-- only be used when accuracy isn't important.
	-- 
	-- the Stream API check profile availabity, so they are more accurate
	computeTemporaryDefaultForOrientation = 
		computeTemporaryDefaultForOrientation,

	computeFromRenderInfo = computeFromRenderInfo,
	
	maybeComputeFromDocValue = maybeComputeFromDocValue,

}
