--[[----------------------------------------------------------------------------

NmbProfileCatalogUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require 'Rx'
local Observable = Rx.Observable

local KSTableUtils = require 'KSTableUtils'
local WFDigest = require 'WFDigest'
local WFCatFileSpec = require 'WFCatFileSpec'
--------------------------------------------------------------------------------

local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt, 
										"KSTableUtils.derefOpt not defined" )

local md5 = assert( KSTableUtils_derefOpt( WFDigest, "MD5", "digest" ),
						"WFDigest.MD5.digest does not exist" )
--============================================================================--

local NmbProfileCatalogUtils = {}

--------------------------------------------------------------------------------

local profileCache = {}

--------------------------------------------------------------------------------

local function startsWith( str, startStr )

   return string.sub( str, 1, string.len( startStr ) ) == startStr
   
end

--------------------------------------------------------------------------------

local function endsWith( str, endStr )

   return string.sub( str, -string.len( endStr ) ) == endStr
   
end

--------------------------------------------------------------------------------

local function isCustomProfile( catalogId, href, maybeIsFallbackProfile, maybeCameraProfileName )
	
	-- Has missing profile case
	local isMissingProfile = maybeIsFallbackProfile 
	
	-- has a profile link which points to user's catalog
	local hasCustomProfileLink = href and href:match( catalogId )
	
	-- has a profile link which doesn't point to user's catalog
	local hasStandardProfileLink = href and not hasCustomProfileLink
	
	-- Has embedded or internal profile
	local isInternalProfile = not href 
		and maybeCameraProfileName 
		and NmbProfileCatalogUtils.isInternalProfile( string.lower( maybeCameraProfileName ))
	
	return isMissingProfile or hasCustomProfileLink or 
		not ( hasStandardProfileLink or isInternalProfile )
end

--------------------------------------------------------------------------------

local function cacheFilenameAndHrefFn( normalizedDigest )
	
	return function( filename, href )
		
		profileCache[ normalizedDigest ] = { filename = filename, href = href } 
	
	end
	
end

--------------------------------------------------------------------------------

local function checkProfileCache( normalizedDigest )
	
	return Observable.just( profileCache[ normalizedDigest ] )
			:map( function( maybeCacheResult )
				if maybeCacheResult then
					return maybeCacheResult.filename, maybeCacheResult.href
				end
				return nil, nil -- cache fail
			end )

end
--------------------------------------------------------------------------------

local function getFilenameAndHrefForDigest( profileCatalog, digest, wfCatalog, optCameraModel )
	
	if not digest then
		return Observable.just( nil )
	end

	local normalizedDigest = digest:lower()
	
	return checkProfileCache( normalizedDigest )
			:flatMap( function( filename, href )
						
				if filename and href then
					return Observable.just( filename, href )
				else
					
					return profileCatalog:indexStore():query {
							sql = [[SELECT filename, href FROM profiles WHERE digest = ?]], 
							quiesceUntilUpToDate = true,
							normalizedDigest
						}
						:take( 1 )
						:flatMap( function( results )
							
							if #results == 0 then
							
								if not wfCatalog then
									return Observable.just( nil )
								end
							
								local sql = [[ SELECT fileName, href from customCameraProfiles WHERE cameraProfileUniqueID = ? ]]
								
								if optCameraModel then
									if type( optCameraModel ) == 'string' then
										optCameraModel = optCameraModel:lower()
									end
									
									-- Caller has camera model and digest, not fingerprint.
									sql = [[ SELECT fileName, href from customCameraProfiles WHERE cameraProfileDigest = ?
												AND cameraModel = ? ]]
								end
								
								return wfCatalog:indexStore():query {
									sql = sql,
									quiesceUntilUpToDate = true,
									normalizedDigest,
									optCameraModel,
								}
								:take( 1 )
							end

							return Observable.just( results )
						end )
						:map( function( results )
							
							local filename = KSTableUtils_derefOpt( results, 1, 1 )
							local href = KSTableUtils_derefOpt( results, 1, 2 )
							
							return filename, href
						
						end )
						:tap( cacheFilenameAndHrefFn( normalizedDigest ) )
				
				end
			end )
end

--------------------------------------------------------------------------------

local function addCameraProfileToProfileHash( profileCatalog, wfCatalog, maybeCameraProfileDigest, maybeProfileHash )
	
	return getFilenameAndHrefForDigest( profileCatalog, maybeCameraProfileDigest, wfCatalog )
			:map( function( filename, href )
				
				if filename and href then
					maybeProfileHash = maybeProfileHash or {}
					maybeProfileHash.camera = {
						filename = filename,
						href = href,
					}
				end
					
				return maybeProfileHash
			end )
end

--------------------------------------------------------------------------------

local function addLensProfileToProfileHash( profileCatalog, wfCatalog, maybeLensProfileDigest, maybeProfileHash )

	return getFilenameAndHrefForDigest( profileCatalog, maybeLensProfileDigest, wfCatalog )
			:map( function( filename, href )
				
				if filename and href then
				
					maybeProfileHash = maybeProfileHash or {}
					maybeProfileHash.lens = {
						filename = filename,
						href = href,
					}
					
				end
					
				return maybeProfileHash
				
			end )

end

--------------------------------------------------------------------------------


function NmbProfileCatalogUtils.addCameraAndLensProfilesToDoc( profileCatalog, wfCatalog, assetDocValue, maybeCameraProfileDigest, maybeLensProfileDigest )
	
	local currentProfileHash = KSTableUtils_derefOpt( assetDocValue, 'develop', 'profiles' )
	
	return Observable.just( currentProfileHash )
		:flatMap( function( maybeHash )
			return addCameraProfileToProfileHash( profileCatalog, wfCatalog, maybeCameraProfileDigest, maybeHash )
		end )
		:flatMap( function( maybeHash )
			return addLensProfileToProfileHash( profileCatalog, wfCatalog, maybeLensProfileDigest, maybeHash )
		end )
		:map( function( maybeHash )
			
			if maybeHash and next( maybeHash ) then -- don't add an empty profiles hash
				assetDocValue.develop = assetDocValue.develop or {}
				assetDocValue.develop.profiles = maybeHash
			end
			
			return assetDocValue
		end )
end

--------------------------------------------------------------------------------

function NmbProfileCatalogUtils.lensOrCameraProfileOverlay( wfCatalog, profileCatalog, imageId, maybeProfileFingerprint, typeKey, optCameraModel )

	if not wfCatalog then
		return
	end
	
	return getFilenameAndHrefForDigest( profileCatalog, maybeProfileFingerprint, wfCatalog, optCameraModel )
		:flatMap( function( filename, href )
			
			if filename and href then
				
				return Observable.just( 
						{
							develop = {
								profiles = { 
									[ typeKey ] = {
										filename = filename,
										href = href,
									}
								}
							}
						} )
				
			else 
				
				-- check to see if we need to remove profile overlay
				return wfCatalog:docStore():docValue( imageId )
					
					:map( function( docValue )
						
						local section = KSTableUtils.derefOpt( docValue, 'develop', 'profiles', typeKey )
						

						if not section then
							-- if there's no section no need to remove it, provide an empty overlay
							return {}
						end
						
						-- if there is a section, we have to remove it. but calculating the overlay 
						-- here is a bit annoying, removing the section may require us to also 
						-- remove the profiles section
						local profilesSection = KSTableUtils_derefOpt( docValue, 'develop', 'profiles' )
						
						
						-- try remove the section, then see if there's anything left in profiles, if not
						-- we want to remove the whole profiles section
						local profilesCopy = KSTableUtils.deepCopy( profilesSection )
						profilesCopy[ typeKey ] = nil
						
						local newProfiles
						if not next( profilesCopy ) then
							newProfiles = { _delete = true }
						else
							newProfiles = { [ typeKey ] = { _delete = true } }
						end
						
						return { 
								develop = { 
									profiles = newProfiles 
								}
							}
						
					end )
			end
			
		end )
end
--------------------------------------------------------------------------------
	-- isFallbackProfile - Whether the cameraProfileUniqueID corresponds to a fallback profile
	-- cameraProfileDigest - The profile digest of the actual camera profile( not the fallback one )
	-- cameraProfileName - The name of the actual camera profile( not the fallback one )
--[[!
	\brief Returns the custom profile localOnly overlay for the asset
	
	\param catalogId (UUID) Id of the current catalog
	\param docValueOverride (table) Table corresponding to the overlay which is to be applied.
	\param maybeIsFallbackProfile (boolean) Whether the cameraProfileUniqueID corresponds to a fallback profile
	\param maybeCameraProfileDigest (string) The profile digest of the actual camera profile( not the fallback one )
	\param maybeCameraProfileName (string) The name of the actual camera profile( not the fallback one )

	\return (table) localOnly overlay
]]--
function NmbProfileCatalogUtils.getCustomProfileOverlay( catalogId, 
	docValueOverride, maybeIsFallbackProfile, maybeCameraProfileDigest,
	maybeCameraProfileName )

	-- There is no straight forward way of knowing
	-- if the profile is custom or not. We are using
	-- the following.
	-- Profile is a custom profile, if
	-- 		There is no href, then its a custom profile
	-- 		There is an href and its it points to user's
	-- 			catalog, then its a custom profile
	--		There is an href, which points to global
	--			catalog, but fallbackProfile is true.
	--
	
	local href = WFCatFileSpec.getCameraProfileLinkFromDocValue( docValueOverride )
	
	local customCameraProfileDigest = nil
	local isCustomCameraProfile
	
	if isCustomProfile( catalogId, href, maybeIsFallbackProfile, maybeCameraProfileName ) then
		
		isCustomCameraProfile = true
		customCameraProfileDigest = maybeCameraProfileDigest and maybeCameraProfileDigest:lower()
	end
	
	local isMissing = customCameraProfileDigest and maybeIsFallbackProfile or nil
	local xmpCameraRawProfileDigest = 
		NmbProfileCatalogUtils.getXmpCameraRawProfileDigest( nil, docValueOverride )
		
	local profileLinkCheckDoneOverlay
	if isMissing and isCustomCameraProfile then
		profileLinkCheckDoneOverlay = { _delete = true }
	end
	
	local localOnly = {
		_localOnly = {
			customCameraProfile = {
				profileDigest = customCameraProfileDigest or "NA",
				isMissing = isMissing or { _delete = true },
				xmpCameraRawProfileDigest = xmpCameraRawProfileDigest,
				profileLinkCheckDone = profileLinkCheckDoneOverlay,
			}
		},
	}
	
	return localOnly

end

--------------------------------------------------------------------------------

function NmbProfileCatalogUtils.lensProfileOverlay( wfCatalog, profileCatalog, imageId, maybeProfileDigest )

	return NmbProfileCatalogUtils.lensOrCameraProfileOverlay( wfCatalog, profileCatalog, imageId, maybeProfileDigest, "lens" )

end

--------------------------------------------------------------------------------

function NmbProfileCatalogUtils.cameraProfileOverlay( wfCatalog, profileCatalog, imageId, maybeProfileDigest )

	return NmbProfileCatalogUtils.lensOrCameraProfileOverlay( wfCatalog, profileCatalog, imageId, maybeProfileDigest, "camera" )
end
--------------------------------------------------------------------------------

function NmbProfileCatalogUtils.getXmpCameraRawProfileDigest( docValue, docValueOverride )

	docValue = docValueOverride or docValue
	local xmpCameraRaw = KSTableUtils_derefOpt( 
				docValue, "develop", "xmpCameraRaw" )
	local xmpCameraRawProfileDigest = type( xmpCameraRaw ) == 'table' 
					and xmpCameraRaw.sha256 
					or xmpCameraRaw
	local xmpDevelopInfo = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue ) or {}
	local cameraProfileLink = WFCatFileSpec.getCameraProfileLinkFromDocValue( docValue )
					
	xmpCameraRawProfileDigest = string.format( "%s_%s_%s", xmpCameraRawProfileDigest, 
		xmpDevelopInfo.path, cameraProfileLink )

	return md5( xmpCameraRawProfileDigest )
end


--------------------------------------------------------------------------------

function NmbProfileCatalogUtils.isInternalProfile( cameraProfileName )

	-- For certain images, the profile info can be
	-- embedded in the image itself, or it might be
	-- available as part of ACR code. Hence in those
	-- cases, we shouldn't report them as custom profile.
	
	-- SPAM [ashj 2018-05-16]: I would really prefer, 
	-- if this info can come from ACR. But ACR, right
	-- now needs negative for any such query; we dont
	-- have that luxury.
	-- Adapted from cr_params, missingProfile validation.
	
	local isInternal = cameraProfileName == "adobe standard"
		or cameraProfileName == "embedded"
		or cameraProfileName == "matrix"
		or cameraProfileName =="beta"
		or startsWith( cameraProfileName, "acr" )
		or endsWith( cameraProfileName, " beta" )
	
	return isInternal
	
end

--------------------------------------------------------------------------------

return NmbProfileCatalogUtils
