--[[----------------------------------------------------------------------------

NmbPurgeDbUpdater.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2017 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.managed_catalog
local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatPhotoImportance = require 'WFCatPhotoImportance'

--------------------------------------------------------------------------------

local derefOpt = KSTableUtils.derefOpt

local check = assert( WFArguments.check )

local isValidAssetDocument = assert( WFCatIndexUtils.isValidAssetDocument )

--------------------------------------------------------------------------------

local function docValueHasLocalOriginal( docValue )

	local localOriginalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
	local localOriginalRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue )
	return localOriginalVolumeGuid and localOriginalRelativePath

end

--------------------------------------------------------------------------------

local function didLocalNegativeStatusChange( assetInfo, docValue )

	local previousLocalOriginalGuid = assetInfo.localOriginalVolumeGuid
	local previousLocalProxyGuid = assetInfo.localProxyVolumeGuid
	
	local localNegativeInfo = {

		originalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue ),
		originalRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue ),
		proxyVolumeGuid = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue ),
		proxyRelativePath = WFCatFileSpec.getProxyRelativePathFromDocValue( docValue )
		
	}

	assetInfo.localOriginalVolumeGuid = localNegativeInfo.originalVolumeGuid
	assetInfo.localProxyVolumeGuid = localNegativeInfo.proxyVolumeGuid
	
	local hasLocalOriginalUpdate = localNegativeInfo.originalVolumeGuid ~= previousLocalOriginalGuid 
	local hasLocalProxyUpdate = localNegativeInfo.proxyVolumeGuid ~= previousLocalProxyGuid
	
	return assetInfo, ( hasLocalOriginalUpdate or hasLocalProxyUpdate ) and localNegativeInfo

end

--------------------------------------------------------------------------------

local function didImportanceChange( assetInfo, docValue )

	local previousImportance = assetInfo.importance

	local currentImportanceInfo = {
		isImportant = WFCatPhotoImportance.isPhotoImportant( docValue )
	}
	
	assetInfo.importance = currentImportanceInfo.isImportant
	
	local hasImportanceUpdate = currentImportanceInfo.isImportant ~= previousImportance 
	
	return assetInfo, hasImportanceUpdate and currentImportanceInfo

end

--------------------------------------------------------------------------------

local function didSizeChange( assetInfo, docValue )

	local previousProxySize = assetInfo.proxySize
	local previousOriginalSize = assetInfo.originalSize

	local currentSizeInfo = {
		proxySize = WFCatFileSpec.getProxySizeFromDocValue( docValue ) or 0,
		originalSize = WFCatFileSpec.getOriginalSizeFromDocValue( docValue ) or 0
	}
	
	assetInfo.proxySize = currentSizeInfo.proxySize
	assetInfo.originalSize = currentSizeInfo.originalSize
	
	local hasProxyUpdate = currentSizeInfo.proxySize ~= previousProxySize
	
	if not hasProxyUpdate then
		currentSizeInfo.proxySize = nil
			-- Remove proxy size since it doesn't need to be updated.
	end
	
	local hasOriginalUpdate = currentSizeInfo.originalSize ~= previousOriginalSize
	
	if not hasOriginalUpdate then
		currentSizeInfo.originalSize = nil
			-- Remove original size since it doesn't need to be updated.
	end
	
	return assetInfo, ( hasProxyUpdate or hasOriginalUpdate ) and currentSizeInfo

end

--------------------------------------------------------------------------------

local function didOzStatusChange( assetInfo, docValue )

	local previousOzOriginalStatus = assetInfo.ozHasOriginal
	local previousOzProxyStatus = assetInfo.ozHasProxy
	
	local _, isOzOriginalInvalid = WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
	local validSha256 = derefOpt( docValue, 'importSource', 'sha256' )
	local hasLocalOriginal = docValueHasLocalOriginal( docValue )
	local localOriginalSha256 = hasLocalOriginal and WFCatFileSpec.getLocalOriginalSha256( docValue )
	
	local isOriginalSafeToPurge = false
	
	-- If the original on Oz is invalid, then we should only purge the local original
	-- if it's also invalid. If the local original is valid, we shouldn't purge it
	-- in case it's the only copy of the valid original.

	if isOzOriginalInvalid then
		
		local localOriginalIsInvalid = validSha256 ~= localOriginalSha256

		if localOriginalIsInvalid then
			isOriginalSafeToPurge = true
		end
		
	else
		isOriginalSafeToPurge = true
	end
	
	local currentOzStatusInfo = {
		ozHasOriginal = isOriginalSafeToPurge and not not WFCatFileSpec.getOriginalLinkFromDocValue( docValue ),
		ozHasProxy = not not WFCatFileSpec.getProxyLinkFromDocValue( docValue )
	}
	
	assetInfo.ozHasOriginal = currentOzStatusInfo.ozHasOriginal
	assetInfo.ozHasProxy = currentOzStatusInfo.ozHasProxy
	
	local hasOzOriginalStatusUpdate = currentOzStatusInfo.ozHasOriginal ~= previousOzOriginalStatus 

	if not hasOzOriginalStatusUpdate then
		currentOzStatusInfo.ozHasOriginal = nil
			-- Remove oz original status since it doesn't need to be updated.
	end

	local hasOzProxyStatusUpdate = currentOzStatusInfo.ozHasProxy ~= previousOzProxyStatus

	if not hasOzProxyStatusUpdate then
		currentOzStatusInfo.ozHasProxy = nil
			-- Remove oz proxy status since it doesn't need to be updated.
	end

	return assetInfo, ( hasOzOriginalStatusUpdate or hasOzProxyStatusUpdate ) and currentOzStatusInfo

end


--============================================================================--
--[[!

	\namespace NmbPurgeDbUpdater
	
	\brief This namespace is used to monitor the doc store and update the
		NmbFileAccessDatabase db used by the purge agent.
	
--]]

-- @mark NmbPurgeDbUpdater namespace

--------------------------------------------------------------------------------

local NmbPurgeDbUpdater = {}

--------------------------------------------------------------------------------

function NmbPurgeDbUpdater.scanForAssetUpdates( assetInfoCache, docValue )
	
	-- WARNING [jandrew 2017-10-25]:
	-- If the implemenation changes in this function, or in any of the called
	-- functions, we may need to manually increment dbUpdaterVersion found in
	-- NmbFileAccessDatabase.lua so that the MRU table is rebuilt.
	
	local assetId = docValue._localId
	local changeCounter = docValue._changeCounter
	
	if not isValidAssetDocument( docValue ) then
		return assetInfoCache, assetId, changeCounter
	end
	
	local assetInfo = assetInfoCache[ assetId ] or {
		localOriginalVolumeGuid = false,
		localProxyVolumeGuid  = false,
			-- `nil` is a valid value for these fields. Init'ing to false ensures the fields update on
			-- the first value change.
	}
	
	-- Check to see if there have been any changes to the local negative status.
	
	local maybeLocalNegativeInfo

	assetInfo, maybeLocalNegativeInfo = didLocalNegativeStatusChange( assetInfo, docValue )

	-- Check to see if there have been any changes to the asset importance.
	
	local maybeImportanceInfo
	
	assetInfo, maybeImportanceInfo = didImportanceChange( assetInfo, docValue )
	
	-- Check to see if there have been any changes to file size.
	
	local maybeSizeInfo
	
	assetInfo, maybeSizeInfo = didSizeChange( assetInfo, docValue )
	
	-- Check to see if there have been changes to the oz status.
	
	local maybeOzStatusInfo
	
	assetInfo, maybeOzStatusInfo = didOzStatusChange( assetInfo, docValue )
	
	assetInfoCache[ assetId ] = assetInfo

	return assetInfoCache, assetId, changeCounter,
			maybeLocalNegativeInfo, maybeImportanceInfo,
			maybeSizeInfo, maybeOzStatusInfo

end

----------------------------------------------------------------------------

function NmbPurgeDbUpdater.trackAssetUpdates( docStore, fileAccessDb, deletionBacklog, optBufferTime )

	check( 'WFRxDocStore', 'NmbFileAccessDatabase', 'NmbDeletionBacklog', '?number' )

	local assetInfoCache = {}

	local changeCounterStream = fileAccessDb:getContinueFromChangeCounter()

	local updateFileAccessDatabaseFileStatuses = changeCounterStream
	
		:flatMapLatest( function( changeCounter )
			return docStore
				:scannedFilteredMappedChangesFeedSince( 
						'NmbPurgeDbUpdater', 
						'scanForAssetUpdates', 
						assetInfoCache,
						changeCounter )
			
		end )
		:map( function( localAssetId, changeCounter, maybeLocalNegativeInfo, maybeImportanceInfo, 
				maybeSizeInfo, maybeOzStatusInfo )
			
			return {
				localAssetId = localAssetId,
				changeCounter = changeCounter,
				maybeLocalNegativeInfo = maybeLocalNegativeInfo,
				maybeImportanceInfo = maybeImportanceInfo,
				maybeSizeInfo = maybeSizeInfo,
				maybeOzStatusInfo = maybeOzStatusInfo
			}
			
		end )
		:bufferWithTimeOrCount( optBufferTime or 5, 100, require 'WFRxScheduler'.sharedScheduler() )
		:subscribe( function( bufferedResults ) 
			
			local latestChangeCounter
			for _, row in ipairs( bufferedResults ) do
				
				local localAssetId = row.localAssetId
				local changeCounter = row.changeCounter
				local maybeLocalNegativeInfo = row.maybeLocalNegativeInfo
				local maybeImportanceInfo = row.maybeImportanceInfo
				local maybeSizeInfo = row.maybeSizeInfo
				local maybeOzStatusInfo = row.maybeOzStatusInfo
				
				if maybeLocalNegativeInfo then
				
					fileAccessDb:updateLocalOriginalStatus( localAssetId, maybeLocalNegativeInfo.originalVolumeGuid )
				
					if maybeLocalNegativeInfo.originalVolumeGuid and maybeLocalNegativeInfo.originalRelativePath then
						
						deletionBacklog:removeFromDeletionBacklog( maybeLocalNegativeInfo.originalVolumeGuid, maybeLocalNegativeInfo.originalRelativePath )
								-- If the original has been previously added to the deletion backlog, clear it out
								-- so we don't delete it.
								
					end
					
					fileAccessDb:updateLocalProxyStatus( localAssetId, maybeLocalNegativeInfo.proxyVolumeGuid )
					
					if maybeLocalNegativeInfo.proxyVolumeGuid and maybeLocalNegativeInfo.proxyRelativePath then
					
						deletionBacklog:removeFromDeletionBacklog( maybeLocalNegativeInfo.proxyVolumeGuid, maybeLocalNegativeInfo.proxyRelativePath )
								-- If the proxy has been previously added to the deletion backlog, clear it out
								-- so we don't delete it.
								
					end
				
				end
				
				if maybeImportanceInfo then
				
					fileAccessDb:updateImportance( localAssetId, maybeImportanceInfo.isImportant )
				
				end
				
				if maybeSizeInfo then
				
					if maybeSizeInfo.proxySize then
						fileAccessDb:updateProxySize( localAssetId, maybeSizeInfo.proxySize )
					end
					
					if maybeSizeInfo.originalSize then
						fileAccessDb:updateOriginalSize( localAssetId, maybeSizeInfo.originalSize )
					end
					
				end
				
				if maybeOzStatusInfo then
				
					if maybeOzStatusInfo.ozHasOriginal ~= nil then
						fileAccessDb:updateOzOriginalStatus( localAssetId, maybeOzStatusInfo.ozHasOriginal )
					end
					
					if maybeOzStatusInfo.ozHasProxy ~= nil then
						fileAccessDb:updateOzProxyStatus( localAssetId, maybeOzStatusInfo.ozHasProxy )
					end
					
				end
			
				latestChangeCounter = changeCounter
			end
			
			if latestChangeCounter then
				fileAccessDb:updateChangeCounter( latestChangeCounter )
			end
		end )
		
	fileAccessDb:addCloseAction( function()
		updateFileAccessDatabaseFileStatuses:dispose()
		updateFileAccessDatabaseFileStatuses = nil
	end )
	
end

----------------------------------------------------------------------------

return NmbPurgeDbUpdater
