--[[----------------------------------------------------------------------------

NmbUserModel.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'
local BehaviorSubject = Rx.BehaviorSubject
local ReplaySubject = Rx.ReplaySubject
local Observable = Rx.Observable

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'
local WFInsomnia = require 'WFInsomnia'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFRxLowFidelityTimer = require 'WFRxLowFidelityTimer'
local WFConfiguration = require 'WFConfiguration'
local KSIdGenerator = require "KSIdGenerator"
local KSPrefs = require 'KSPrefs'
local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.managed_catalog
local NmbCatalog = require 'NmbCatalog'
local NmbFileManagementSettings = require 'NmbFileManagementSettings'
local WFCatalog = require 'WFCatalog'
local WFCatSpec = require 'WFCatSpec'
local WFCatStorage = require 'WFCatStorage'
local WFCatSynchronizer = require 'WFCatSynchronizer'
local WFCatSyncStatus = require 'WFCatSyncStatus'
local WFCatUser = require 'WFCatUser'
local WFCatUserSynchronizer = require 'WFCatUserSynchronizer'
local NmbAppMessaging = require 'NmbAppMessaging'
local WFCatAvatar = require 'WFCatAvatar'
local NmbUserModel_LocalStorage = require 'NmbUserModel_LocalStorage'
local NmbCrPathOverrides = require 'NmbCrPathOverrides'
local WFCatAlbums = require 'WFCatAlbums'

	-- WF.oz_client.ims
local WFImsAppIdentity = require 'WFImsAppIdentity'

	-- WF.oz_client.managed_catalog.pixel_provider
local PixelProviderProcessor = require 'PixelProviderProcessor'
	
	-- WF.oz_client.sync_engine
local WFOzUserLogin = require 'WFOzUserLogin'

	-- WF.web.http_client
local WFReachability = require 'WFReachability'

	-- WF.web
local KSJsonDecoder = require "KSJsonDecoder"

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

local logger = WFLogger 'NmbUserModel'
	
local POSTPONE_DURATION = WFConfiguration.FileMover.postponeDurationInSeconds

local DEBUG_OZ_IS_DOWN = WFConfiguration.Debug.debugOzIsDown

local USE_NON_CC_CREDENTIALS = WFConfiguration.LrMobile.useNonCCCredentials

local kClientUpgradeRequired = "upgrade_required"

--------------------------------------------------------------------------------

local function setupPixelProccessor( wfCatUser, wfCatalogSpec,
									rootCatalogDataDir, catalogOptions,
									pixelProviderSettings )

	local pixelProviderProcessor = PixelProviderProcessor( wfCatUser,
																wfCatalogSpec, 
																rootCatalogDataDir, 
																catalogOptions,
																pixelProviderSettings)
	
	local scanAndPurgeDelay = WFConfiguration.PixelProvider.RenditionCachePurgeDelay

	scanAndPurgeDelay = type( scanAndPurgeDelay ) == "number" and scanAndPurgeDelay > 0 and
							scanAndPurgeDelay or 5 * 60

	local delayedScanAndPurgeRenditionCache =
			WFRxLowFidelityTimer.timer( scanAndPurgeDelay )
				:take( 1 )
				:flatMap( 
					function()

						return pixelProviderProcessor:scanAndPurge()

					end )

	local scanAndPurgeRenditionCacheSubscription =
				delayedScanAndPurgeRenditionCache:subscribe(
					function()

						logger:info( "Scan and purge requested" )

					end )

	pixelProviderProcessor:addCloseAction( 
		function() 

			scanAndPurgeRenditionCacheSubscription:dispose()

		end )

	return pixelProviderProcessor

end

--------------------------------------------------------------------------------

local function addModelValueToDiagnosticLog( modelValue )
	
	WFDiagnosticLog.addLogSectionFunc( function()

		local modelValue_copy = {}

		for k, v in pairs( modelValue ) do
			if type( v ) == 'function' then
				modelValue_copy[ k ] = v()
			else
				modelValue_copy[ k ] = v
			end
		end

		modelValue_copy.email = nil
		modelValue_copy.fullName = nil
		modelValue_copy.firstName = nil
		modelValue_copy.lastName = nil

		WFDiagnosticLog.addSectionToLog
		{
			name = "Oz logged in user",
			id = 'NmbUserModel',
			weight = -100,

			body = modelValue_copy,
		}

	end )
	
end

--------------------------------------------------------------------------------

local function takeOneValueAndRunFunc( observable, functionToRun )

	observable
		:take( 1 )
		:tap( functionToRun )
		:call( runUntilDone )
		
end

--------------------------------------------------------------------------------

local function logoutUser( closingUser )

	closingUser:logout()

		:map( function()
			return closingUser:close()
		end )

		:call( runUntilDone )

end

--------------------------------------------------------------------------------

local wfCatUser_now

local function createCurrentUserStream( serverBaseUrlStream, userIdStream, userInfoCachePath )
	
	local previousUser
	
	return Observable.combineLatest( function( serverBaseUrl, userId )

		logger:debugf( "currentUser from serverBaseUrl %s, userId %s via %s",
					   serverBaseUrl, userId, ':traceback:' )

		local newUser
		
		if serverBaseUrl and userId then
			newUser = WFCatUser( serverBaseUrl, userId, userInfoCachePath )
		end

		if previousUser and previousUser ~= newUser then
			logoutUser( previousUser )
		end

		previousUser = newUser

		if not newUser then
			KSPrefs.lastDaemonUser = nil
		end

		wfCatUser_now = newUser

		return newUser

	end,
	serverBaseUrlStream:distinctUntilChanged(), userIdStream:distinctUntilChanged() )
	:replay( nil, 1 )
	:refCount()
		
end

--------------------------------------------------------------------------------

local function createHasValidConnectionToOz( currentUserStream )

	return currentUserStream
	
		:flatMap( function( currentUser )

			if currentUser then
				return currentUser:hasValidConnectionToOz()
			else
				return Observable.just( false )
			end

		end )
		
end

--------------------------------------------------------------------------------

local function getUserIdStream( processId, imsOverrides )

	local lastUser = KSPrefs.lastDaemonUser
	local lastUserId, prevProcessId
	
	if lastUser then
		logger:infof( "Found previous user ID %s in prefs", lastUser )
	end

	if lastUser and processId then
		local equalsIndex = string.find( lastUser, '=' )
		if equalsIndex and equalsIndex > 0 then
			prevProcessId = string.sub( lastUser, 1, equalsIndex - 1 )
			lastUserId = string.sub( lastUser, equalsIndex + 1 )
		end
	end

	if imsOverrides and imsOverrides.imsClientId and imsOverrides.imsClientSecret then
		WFImsAppIdentity.setClientIdAndSecret( imsOverrides.imsClientId,
														 imsOverrides.imsClientSecret )
	end

	local userIdStream = ReplaySubject( 1 )

	if prevProcessId == processId then
		userIdStream:onNext( lastUserId )
	end

	userIdStream
		:tap( function( ozUserId )
			if ozUserId and processId then
				KSPrefs.lastDaemonUser = processId .. "=" .. ozUserId
			else
				KSPrefs.lastDaemonUser = nil
			end
		end )
		:call( runUntilDone )

	return userIdStream, lastUserId

end

--------------------------------------------------------------------------------

local wfCatSynchronizer

local function startSyncingUserAndCatalog( modelData )

	modelData._shouldSyncUserInfoStream:onNext( true )

	if wfCatSynchronizer then
		wfCatSynchronizer:startSyncing()
	end

end

--------------------------------------------------------------------------------

local function stopSyncingUserAndCatalog( modelData )

	modelData._shouldSyncUserInfoStream:onNext( false )

	if wfCatSynchronizer then
		wfCatSynchronizer:stopSyncing()
	end

end

--------------------------------------------------------------------------------

local function activateOnlineMode( modelData )

	logger:info( "Online mode activated" )

	modelData._isOnlineStream:onNext( true )

	if not KSPrefs.syncPaused then
		startSyncingUserAndCatalog( modelData )
	end

end

--------------------------------------------------------------------------------

local function activateOfflineMode( modelData )

	modelData._isOnlineStream:onNext( false )

	stopSyncingUserAndCatalog( modelData )

	if not modelData._lastUserId then
		logger:infof( "Offline mode requested, but no stored user ID. Ignoring." )
		return
	end

	logger:warnf( "Offline Mode. Using %s", modelData._lastUserId )

	modelData._userIdStream:onNext( modelData._lastUserId )

end

--------------------------------------------------------------------------------

local function monitorLoginTask( modelData, loginTaskObservable )

	modelData._loginStatusStream:onNext( { status = 'pending' } )

	local observable = loginTaskObservable

	if DEBUG_OZ_IS_DOWN then
		observable = Observable.throw( 'Oz is down!', 'server_error' )
	end

	observable

		:flatMapLatest( function( wfCatUser, userInfo, accessToken, refreshToken, optDeviceToken )

			logger:debugf( "Storing tokens for new user %s", userInfo )

			return wfCatUser:storeTokens( accessToken, refreshToken, optDeviceToken )
				:concat( Observable.just( wfCatUser ) )

		end )

		:tap(

			function( wfCatUser )

				local _userId = wfCatUser:userId()

				logger:infof( "Successfully logged in to Oz with user ID %s",
							  _userId  )

				modelData._userIdStream:onNext( _userId )

				modelData._loginStatusStream:onNext( { status = "success" } )

			end,

			function( message, errCode, isRetryable )

				logger:errorf( "Failed to log in to Oz: %s, ( code: %s)", message, errCode )

				if isRetryable == false then
					
					local errorMessage = message
		
					local ok, errorInfo = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, message )
						
					if ok then
						errorMessage = KSTableUtils.derefOpt( errorInfo, "code" ) or message
					end
					
					modelData._loginStatusStream:onNext( { status = "error", errorMessage = errorMessage } )

				end

				if errCode == 'server_error' then
					activateOfflineMode( modelData )
				elseif isRetryable == false then
					modelData._userIdStream:onNext( nil )
				end

			end )
			
		:call( runUntilDone )

	-- TODO [scouten 2015-07-22]: This subscription should probably
	-- have a lifetime so we can cancel it.

end

--------------------------------------------------------------------------------

local function loginFromUsernameAndPassword( modelData, _, baseServerUrl, email, password )

	modelData._isOnlineStream
	
		:take( 1 )
		
		:tap( function( isOnline )

			if isOnline then

				logger:infof( "loginFromUsernameAndPassword url = %s, email = %s",
							  baseServerUrl, email )

				modelData._loginMechanismStream:onNext( 'usernameAndPassword' )

				local loginResults = WFOzUserLogin.loginFromEmailAndPassword( modelData._userInfoCachePath, baseServerUrl, email, password )
				monitorLoginTask( modelData, loginResults )

			end

		end )

		:call( runUntilDone )

end

--------------------------------------------------------------------------------

local function loginFromTokens( modelData, _,
								baseServerUrl,
								accessToken,
								refreshToken,
								optDeviceToken )

	modelData._isOnlineStream
	
		:take( 1 )
		
		:tap( function( isOnline )

			if isOnline then
	
				logger:infof( "loginFromTokens url = %s", baseServerUrl )

				modelData._loginMechanismStream:onNext( 'tokens' )

				if optDeviceToken and #optDeviceToken == 0 then
					optDeviceToken = nil
				end

				local loginResult = WFOzUserLogin.loginFromTokens( modelData._userInfoCachePath, 
																   baseServerUrl,
																   accessToken,
																   refreshToken,
																   optDeviceToken )

				monitorLoginTask( modelData, loginResult )

			end

		end )

		:call( runUntilDone )

end

--------------------------------------------------------------------------------

local function currentUserFieldValueForPath( modelData, memberFieldPath )

	return modelData._currentUserStream and modelData._currentUserStream

		:map( function( wfCatUser )

			if wfCatUser then
				return wfCatUser:userInfoCache():ongoingFieldValue( wfCatUser:userId(), memberFieldPath )
			else
				return Observable.just( nil )
			end

		end )

		:switch()

end

--------------------------------------------------------------------------------

local function useLastUser( modelData )

	local lastUserId = modelData._lastUserId

	if not lastUserId then
		logger:infof( "Last user mode requested, but no stored user ID. Ignoring." )
		return
	end

	logger:warnf( "Last user mode. Using %s", lastUserId )

	modelData._userIdStream:onNext( lastUserId )

end

--------------------------------------------------------------------------------

-- When new assets are synced from Oz, add them to the file access database.

local function trackNewlySyncedFileAccess( fileAccessDb, latestDownloadedAssetIdsStream )

	local trackNewlySyncedFileAccessSub = latestDownloadedAssetIdsStream
		:subscribe( function( assetIds )
			fileAccessDb:touchMultipleIfNew( assetIds )
		end )

	fileAccessDb:addCloseAction( function()
		trackNewlySyncedFileAccessSub:dispose()
		trackNewlySyncedFileAccessSub = nil
	end )

end

----------------------------------------------------------------------------

local function mapVolumeGuidToName( volumesToMap )

	local mappedVolumes = {}

	for volumeGuid, value in pairs( volumesToMap ) do
	
		local volumeIdentifier

		if WFPlatformInfo.isMac then
			volumeIdentifier = WFFileUtils.getVolumeNameFromGuid( volumeGuid )
				-- On Mac, use the volume name.
		else 
			volumeIdentifier = WFFileUtils.getVolumeFromGuid( volumeGuid )
				-- On Windows, use the drive letter.
		end

		if not volumeIdentifier then
	
			-- If the drive identifier cannot be obtained ( likely because the drive is offline ),
			-- use the cached value.
	
			volumeIdentifier = NmbFileManagementSettings.getReadableIdentifierForVolumeGuid( volumeGuid )
					or volumeGuid
			
		end
		
		mappedVolumes[ volumeIdentifier ] = value
		
	end

	return mappedVolumes

end

----------------------------------------------------------------------------

local function watchForRequiredUpdate( modelData, wfCatSynchronizer )

	local requiredUpdateStreamSub = wfCatSynchronizer:getRequiredUpdateStream()
		:subscribe( function( updateRequired )
		
			if updateRequired then
			
				modelData._isUpdateRequiredStream:onNext( true )
				
				if wfCatSynchronizer then
					KSPrefs.syncPaused = true
					stopSyncingUserAndCatalog( modelData )
				end
				
			end
		
		end )

	wfCatSynchronizer:addCloseAction( function()
		requiredUpdateStreamSub:dispose()
		requiredUpdateStreamSub = nil
	end )

end

----------------------------------------------------------------------------

local function restrictedCountryMode( modelData )

	KSPrefs.restrictedCountryMode = true

	modelData._restrictedCountryModeStream:onNext( true )

	wfCatSynchronizer:enableRestrictedCountryMode()

	wfCatSynchronizer:stopSyncing()
		
	if KSPrefs.syncPaused then
		KSPrefs.syncPaused = false
	end

end

----------------------------------------------------------------------------

local function checkForRestrictedCountry( wfCatUser, modelData )

	local chineseAccountCheckSub = wfCatUser:getAccountDocFromOz()
			:catch( Observable.just() )
			:pluck( 'country' )
			:take( 1 )
			:subscribe( function( countryCode )
			
				local chinaCountryCode = 'CN'
				
				if countryCode == chinaCountryCode then
					restrictedCountryMode( modelData )
				else
					KSPrefs.restrictedCountryMode = false
				end
				
			end )
						
	wfCatUser:addCloseAction( function()
		chineseAccountCheckSub:dispose()
		chineseAccountCheckSub = nil
	end )

end

----------------------------------------------------------------------------

local function createOzCatalog( wfCatUser, wfCatalogSpec, maybeProfileCatalogSpec, 
		rootCatalogDataDir, pixelProviderSettings, modelData )

	local catalogOptions = {
		indexSpecScript = 'WFCatIndexDefinitions.lua',
	}

	local wfCatalog = NmbCatalog( wfCatalogSpec, rootCatalogDataDir, catalogOptions )
	NmbCrPathOverrides.set( wfCatalog:fileSpec():catalogPath() )
	
	local pixelProviderProcessor = 
			setupPixelProccessor( wfCatUser, wfCatalogSpec,
									rootCatalogDataDir, catalogOptions, pixelProviderSettings )

	wfCatalog:addCloseTarget( pixelProviderProcessor )

	local maybeProfileCatalogOptions
	local maybeProfileCatalogIdOrDoc
	local maybeProfileCatalog
	if maybeProfileCatalogSpec then
		
		local profilesDir = KSPathUtils.child( rootCatalogDataDir, 'profiles' )
		local profileCatalogPath = KSPathUtils.child( profilesDir, maybeProfileCatalogSpec:catalogId() .. '.mcat' )
		maybeProfileCatalogOptions = { 
			docStorePath = profileCatalogPath,
			indexSpecScript = 'NmbProfilesIndexDefinitions.lua',
			noIdleDownloader = true,
			noCatalogSearch = true,
		}
		
		maybeProfileCatalog = WFCatalog( maybeProfileCatalogSpec, rootCatalogDataDir, 
							maybeProfileCatalogOptions )
		maybeProfileCatalogIdOrDoc = maybeProfileCatalogSpec:catalogDoc() or maybeProfileCatalogSpec:ozCatalogId()
	end

	
	local workerThreadArgs = { wfCatUser:serverBaseUrl(),
								wfCatUser:userId(),
								wfCatUser:userInfoCacheFolder(),
								wfCatalogSpec:catalogDoc() or wfCatalogSpec:catalogId(),
								rootCatalogDataDir, 
								catalogOptions,
								maybeProfileCatalogIdOrDoc,
								maybeProfileCatalogOptions,
							}

	wfCatSynchronizer = WFCatSynchronizer( wfCatalog, maybeProfileCatalog, pixelProviderProcessor, workerThreadArgs,
		not KSPrefs.syncPaused, KSPrefs.residualMode )

	wfCatalog:addCloseAction( function()
	
		wfCatSynchronizer:close()
		wfCatSynchronizer = nil
		
	end )

	local docStoreUpdater = wfCatalog:getDocStoreUpdater()
	docStoreUpdater:setSynchronizer( wfCatSynchronizer )
	docStoreUpdater:start()
	
	local currentDaemonDownloadStatusSub = modelData._currentDaemonDownloadStatusStream
		:subscribe( function( statusTable )
			pixelProviderProcessor:runUntilDone( 'setDaemonDownloadStatus', statusTable )
		end )

	local isUploadingMetadataSub = wfCatSynchronizer:isUploadingMetadata()
		:subscribe( function( isUploading )
			modelData._isSyncUploadingMetadataStream:onNext( isUploading )
		end )

	local isDownloadingMetadataSub = wfCatSynchronizer:isDownloadingMetadata()
		:subscribe( function( isDownloading )
			modelData._isSyncDownloadingMetadataStream:onNext( isDownloading )
		end )

	local isUploadingBinariesSub = wfCatSynchronizer:isUploadingBinaries()
		:subscribe( function( isUploading )
			modelData._isSyncUploadingBinariesStream:onNext( isUploading )
		end )

	local isDownloadingOriginalsOrProxiesSub = wfCatSynchronizer:isDownloadingRequiredOriginalsOrProxies()
		:subscribe( function( isDownloading )
			modelData._isSyncDownloadingOriginalsOrProxiesStream:onNext( isDownloading )
		end )
	
	local isDownloadingStylesSub = wfCatSynchronizer:isDownloadingStyles()
		:subscribe( function( isDownloadingStyles )
			modelData._isSyncDownloadingStylesStream:onNext( isDownloadingStyles )
		end )

	local isSyncPendingSub = WFCatSyncStatus.syncPendingForAnyAsset( wfCatalog )
		:subscribe( function( isPending )
			modelData._isSyncPendingStream:onNext( isPending )
		end )

	local havePersistentSyncErrorSub = WFCatSyncStatus.persistentSyncErrorForAnyAsset( wfCatalog )
		:subscribe( function( haveError )
			modelData._havePersistentSyncErrorStream:onNext( haveError )
		end )

	local isInitialSyncSub = WFCatSyncStatus.isDoingInitialSync( wfCatalog )
		:subscribe( function( isInitialSyncActive )
			modelData._isInitialSyncStream:onNext( isInitialSyncActive )
		end )
	
	local isOutOfStorageSub = WFCatStorage.overStorageQuota( wfCatalog )
		:subscribe( function( isOverQuota )
			modelData._isOutOfOzStorageStream:onNext( isOverQuota )
		end )

	local isLowOnStorageSub = WFCatStorage.lowOnStorage( wfCatalog )
		:subscribe( function( isLow )
			modelData._isLowOnOzStorageStream:onNext( isLow )
		end )

	local deletedCatalogSub = wfCatSynchronizer:isOzCatalogDeleted()
		:filter( function( isDeleted )
			return isDeleted
		end )
		:subscribe( function( isDeleted )
			modelData._deletedCatalogStream:onNext( isDeleted )
		end )

	local sleepPreventionSubscription = Observable.combineLatest( 
			modelData._isActivelySyncingStream, modelData._shouldPreventSleepWhileSyncingStream )
		:filter( function( isSyncing, shouldPreventSleepWhileSyncing )
			return isSyncing and shouldPreventSleepWhileSyncing ~= false
		end )
		:subscribe( function( isSyncing )
			WFInsomnia.preventSystemSleep( "Syncing" )
				-- Prevent the system from sleeping when syncing while
				-- on AC Power. This will not prevent sleep if the user
				-- is on Battery Power.
				-- preventSystemSleep prevents the system from
				-- sleeping for ~30 seconds.
		end )

	wfCatSynchronizer:addCloseAction( function()
		
		currentDaemonDownloadStatusSub:dispose()
		currentDaemonDownloadStatusSub = nil
		   
		isUploadingMetadataSub:dispose()
		isUploadingMetadataSub = nil

		isDownloadingMetadataSub:dispose()
		isDownloadingMetadataSub = nil

		isUploadingBinariesSub:dispose()
		isUploadingBinariesSub = nil

		isDownloadingOriginalsOrProxiesSub:dispose()
		isDownloadingOriginalsOrProxiesSub = nil

		isDownloadingStylesSub:dispose()
		isDownloadingStylesSub = nil
		
		isSyncPendingSub:dispose()
		isSyncPendingSub = nil

		havePersistentSyncErrorSub:dispose()
		havePersistentSyncErrorSub = nil

		isInitialSyncSub:dispose()
		isInitialSyncSub = nil
		
		isOutOfStorageSub:dispose()
		isOutOfStorageSub = nil

		isLowOnStorageSub:dispose()
		isLowOnStorageSub = nil
		
		deletedCatalogSub:dispose()
		deletedCatalogSub = nil
		
		sleepPreventionSubscription:dispose()
		sleepPreventionSubscription = nil
		
		modelData:onSynchronizerClose()
		
	end )

	local fileMover = wfCatalog:getFileMover()

	fileMover:setSynchronizer( wfCatSynchronizer )

	local mainVolumeGuid = modelData._userModelLocalStorage:setupVolumeAgents( wfCatalog, 
			wfCatSynchronizer:getIdleDownloader(), fileMover, 
			wfCatSynchronizer:getBinaryUploader(), 
			wfCatSynchronizer:getStyleSynchronizer(),
			modelData._isOnlineStream, modelData._isSyncPausedStream )

	modelData._mainVolumeGuidStream:onNext( mainVolumeGuid )

	trackNewlySyncedFileAccess( wfCatalog:fileAccessDb(), wfCatSynchronizer:getLatestDownloadedAssetIdsStream() )
	
	require 'NmbPurgeDbUpdater'.trackAssetUpdates( wfCatalog:docStore(), wfCatalog:fileAccessDb(), wfCatalog:getDeletionBacklog() )
	
	watchForRequiredUpdate( modelData, wfCatSynchronizer )
	
	NmbAppMessaging.setupUpdateAvailableCheck( wfCatUser, modelData._isUpdateAvailableStream )

	checkForRestrictedCountry( wfCatUser, modelData )

	modelData._originalStorageLocationInfoStream:onNext( NmbFileManagementSettings.originalStorageLocationInfo() )
	modelData._proxyStorageLocationInfoStream:onNext( NmbFileManagementSettings.proxyStorageLocationInfo() )

	WFCatAlbums.addDiagnosticLogSectionFunc( wfCatalog )
		-- Track album cycles in the diagnostic log.
		
	return wfCatalog, pixelProviderProcessor, maybeProfileCatalog, wfCatSynchronizer

end

----------------------------------------------------------------------------

local wfOzCatalogStream, wfOzCatalog_now

local function getWFCatalog( rootCatalogDir, pixelProviderSettings, modelData )

	if not wfOzCatalogStream then

		wfOzCatalogStream = Observable.combineLatest( modelData._currentUserStream, modelData._deletedCatalogStream )

			:filter( function( wfCatUser )
				return wfCatUser
			end )
			
			:flatMapLatest( function( wfCatUser )

				logger:infof( "Got new user %s, creating / finding catalog",
									wfCatUser
										and tostring( wfCatUser )
										or ":unknown:" )

				if wfOzCatalog_now then
					wfOzCatalog_now:close()
					wfOzCatalog_now = nil
				end

				if wfCatUser then

					-- TODO [scouten 2015-08-25]: If we have the cached catalog ID
					-- available locally, we'll launch faster. Consider using that before
					-- going out to the network.

					return Observable.combineLatest(
							wfCatUser:findOrCreateCatalog( "lightroom" ),
							wfCatUser:findFirstCatalogWithSubtype( 'profile', 'any' ) )
						:take( 1 )
						
						:flatMapLatest( function( catalogSpec, profileSpec )
							
							local managedCatalogDoc = catalogSpec:catalogDoc()
							if managedCatalogDoc then
								managedCatalogDoc._replace = true
							end
							
							local profileCatalogDoc = profileSpec:catalogDoc()
							if profileCatalogDoc then
								profileCatalogDoc._replace = true
							end
							
							return wfCatUser._userInfoCache:applyOverlay
								{
									_id = wfCatUser:userId(),
									_localOnly = {
										managedCatalog = managedCatalogDoc,
										managedCatalogId = catalogSpec:catalogId(),
										profileCatalog = profileCatalogDoc,
										profileCatalogId = profileSpec and profileSpec:catalogId()
									}
								}

								:map( function( ... )

									return createOzCatalog( wfCatUser,
															catalogSpec,
															profileSpec, 
															rootCatalogDir,
															pixelProviderSettings,
															modelData )

								end )

						end )

						:catch( function( whatWentWrong )

							logger:warnf( "Failed to get catalog ID from Oz( %s ), trying cached reference",
										  whatWentWrong )

							return wfCatUser:hasValidConnectionToOz()
							
								:take( 1 )
							
								:flatMapLatest( function( hasValidConnection )

									if hasValidConnection then

										return wfCatUser._userInfoCache:docValue( wfCatUser:userId() )
											:map( function( userDoc )
												
												local managedCatalog = KSTableUtils.derefOpt( userDoc, '_localOnly', 'managedCatalog' )
												local managedCatalogId = KSTableUtils.derefOpt( userDoc, '_localOnly', 'managedCatalogId' )
												local profileCatalog = KSTableUtils.derefOpt( userDoc, '_localOnly', 'profileCatalog' )
												local profileCatalogId = KSTableUtils.derefOpt( userDoc, '_localOnly', 'profileCatalogId' )
													
												logger:infof( "Found cached catalog ID %s", managedCatalogId )

												if managedCatalog or managedCatalogId then

													local catalogSpec = WFCatSpec( wfCatUser, managedCatalog or managedCatalogId )
																											
													local profileSpec
													if profileCatalog or profileCatalogId then
														profileSpec = WFCatSpec( wfCatUser, profileCatalog or profileCatalogId )
													end
													
													return createOzCatalog( wfCatUser, catalogSpec,
																			profileSpec, rootCatalogDir,
																			pixelProviderSettings,
																			modelData )

												else

													modelData._userIdStream:onNext( nil ) -- force logout
													error( whatWentWrong )

												end

											end )

									else

										modelData._userIdStream:onNext( nil ) -- force logout
										error( whatWentWrong )

									end

							end )

						end )

				else

					return Observable.just( nil )

				end

			end )
			:map( function( wfOzCatalog, pixelProviderProcessor, profileCatalog, wfCatSynchronizer )
			
				wfOzCatalog_now = wfOzCatalog
				
				modelData._catalogIdStream:onNext( wfOzCatalog:ozCatalogId() )
				
				return wfOzCatalog, pixelProviderProcessor, profileCatalog, wfCatSynchronizer
				
			end )
			:replay( nil, 1 )
			:refCount()

	end

	return wfOzCatalogStream

end

--------------------------------------------------------------------------------

local function getAppModel( rootCatalogDir, pixelProviderSettings, modelData, _, namespace, func, isTiedToCatalog, ... )

	local packedN = select( '#', ... )
	local packed = { ... }

	local wfCatalogObservable = getWFCatalog( rootCatalogDir, pixelProviderSettings, modelData )

	if isTiedToCatalog == nil then
		isTiedToCatalog = true
	end

	if isTiedToCatalog then
		wfCatalogObservable = wfCatalogObservable:take( 1 )
	end

	local value = wfCatalogObservable
		:flatMapLatest( function( wfOzCatalog )
			if wfOzCatalog then
				return require( namespace )[ func ]( wfOzCatalog,
													 unpack( packed, 1, packedN ) )
			else
				return Observable.just( {} )
			end
		end )

	return {
	
		value = value,

		close = function()
			logger:debugf( "Model %s.%s CLOSED", namespace, func )
		end,
		
	}

end


--============================================================================--

-- @mark NmbModelData class

--------------------------------------------------------------------------------

local NmbModelDataImp = {

	-- User Info
	
	_lastUserId = false,
	_userInfoCachePath = false,
	_userIdStream = false,
	_catalogIdStream = false,
	_deletedCatalogStream = false,
	_currentUserStream = false,
	_loginStatusStream = false,
	_loginMechanismStream = false,
	_isLoggedInStream = false,
	_userInfoStream = false,
	_shouldSyncUserInfoStream = false,
	_isSyncingUserInfoStream = false,
	
	-- Local Storage
	
	_userModelLocalStorage = false,
	_volumesOutOfSpaceStream = false,
	_offlineVolumesStream = false,
	_volumesRequiringSyncingStream = false,
	_readOnlyVolumesStream = false,
	
	-- Sync

	_serverBaseUrlStream = false,
	_hasValidConnectionToOzStream = false,
	_isOnlineStream = false,
	_isSyncPendingStream = false,
	_havePersistentSyncErrorStream = false,
	_isInitialSyncStream = false,
	_isSyncUploadingMetadataStream = false,
	_isSyncDownloadingMetadataStream = false,
	_isSyncUploadingBinariesStream = false,
	_isSyncDownloadingOriginalsOrProxiesStream = false,
	_isSyncDownloadingStylesStream = false,
	_shouldPreventSleepWhileSyncingStream = false,
	_isSyncPausedStream = false,
	_isActivelySyncingStream = false,
	_currentSyncStatusStream = false,
	_hostIsReachableStream = false,
	_currentDaemonDownloadStatusStream = false,
	
	-- Cloud Storage
	
	_isOutOfOzStorageStream = false,
	_isLowOnOzStorageStream = false,

	-- File Management

	_mainVolumeGuidStream = false,
	_keepAllOriginalsStream = false,
	_keepAllProxiesStream = false,
	_targetDiskSpacePercentageStream = false,
	_originalStorageLocationInfoStream = false,
	_proxyStorageLocationInfoStream = false,
	_fileMoverPostponedUntilDateStream = false,
	
	-- App Updates / Residual / Trial Mode

	_isUpdateAvailableStream = false,
	_isUpdateRequiredStream = false,
	_residualModeStream = false,
	_restrictedCountryModeStream = false,
	_trialEndDateStream = false,
	_deletionDateStream = false,
	
	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

local function createUserInfoStream( currentUserStream )
	
	return currentUserStream
		:flatMap( function( wfCatUser )

			if wfCatUser then
				return wfCatUser:userInfoCache():docValue( wfCatUser:userId() )
			else
				return Observable.just( nil )
			end

		end )
		:replay( nil, 1 )
		:refCount()

end

--------------------------------------------------------------------------------

local userInfoSynchronizer_now

local function createIsSyncingUserInfoStream( shouldSyncUserInfoStream, currentUserStream )

	return Observable.combineLatest( function( shouldSync, wfCatUser )
	
		if not wfCatUser then 
			if userInfoSynchronizer_now then
				userInfoSynchronizer_now:close()
				userInfoSynchronizer_now = nil
			end
			return false
		end
		
		if not userInfoSynchronizer_now or 
			( userInfoSynchronizer_now and 
			( userInfoSynchronizer_now:userId() ~= wfCatUser:userId() ) ) then 
			
			if userInfoSynchronizer_now then
				userInfoSynchronizer_now:close()
			end
			
			userInfoSynchronizer_now = WFCatUserSynchronizer( wfCatUser )
		
		end
			
		if shouldSync then
			userInfoSynchronizer_now:startContinuousUpdates()
		else
			userInfoSynchronizer_now:stopContinuousUpdates()
		end
		
		return shouldSync
		
	end,
	shouldSyncUserInfoStream, currentUserStream )

end

--------------------------------------------------------------------------------

local function getTrialEndDateStream( userInfoStream )

	return userInfoStream
		:map( function( userInfo )

			local status = KSTableUtils.derefOpt( userInfo,
												  'entitlement',
												  'status' )

			local trialEnd = KSTableUtils.derefOpt( userInfo,
													'entitlement',
													'trial',
													'end' )

			if status == 'trial' then
				return trialEnd
			else
				return nil
			end

		end )

end

--------------------------------------------------------------------------------

local function getDeletionDateStream( userInfoStream )

	return userInfoStream
		:map( function( userInfo )

			local deletionDate = KSTableUtils.derefOpt( userInfo,
													'entitlement',
													'deletion_date' )

			return deletionDate

		end )
	
end

--------------------------------------------------------------------------------

local function getIsSyncPausedStream( shouldSyncUserInfoStream )

	return shouldSyncUserInfoStream
		:map( function()

			-- Could use shouldSyncUserInfo, but re-inforce that the pref
			-- controls the paused state.
			-- Using shouldSyncUserInfo also updates observable.
			return KSPrefs.syncPaused

		end )

end

----------------------------------------------------------------------------

local function getCurrentDaemonDownloadStatusStream( userModelLocalStorage,
		isSyncPausedStream, isOnlineStream, volumesOutOfSpaceStream, 
		offlineVolumesStream )

	local kOkay = 'okay'
	local kOutOfDiskSpace = 'out_of_disk_space'
	local kVolumeOffline = 'volume_offline'

	return Observable.combineLatest(
	
		function( syncPaused, online, volumesOutOfSpace, offlineVolumes )

			local outOfLocalStorageForOriginals, outOfLocalStorageForProxies, outOfLocalStorageOnMainVolume
					= userModelLocalStorage:whichVolumes( volumesOutOfSpace )

			local offlineVolumeForOriginals, offlineVolumeForProxies = 
					userModelLocalStorage:whichVolumes( offlineVolumes )

			local statusTable = {
				syncStatus = kOkay,
				mainVolumeStatus = kOkay,
				originalVolumeStatus = kOkay,
				proxyVolumeStatus = kOkay
			}
			
			if syncPaused then
				statusTable.syncStatus = 'sync_paused'
			elseif not online then
				statusTable.syncStatus = 'offline'
			end
			
			if outOfLocalStorageOnMainVolume then
				statusTable.mainVolumeStatus = kOutOfDiskSpace
			end
			
			if outOfLocalStorageForOriginals then
				statusTable.originalVolumeStatus = kOutOfDiskSpace
			elseif offlineVolumeForOriginals then
				statusTable.originalVolumeStatus = kVolumeOffline
			end
			
			if outOfLocalStorageForProxies then
				statusTable.proxyVolumeStatus = kOutOfDiskSpace
			elseif offlineVolumeForProxies then
				statusTable.proxyVolumeStatus = kVolumeOffline
			end

			return statusTable

		end,
		isSyncPausedStream, isOnlineStream, 
		volumesOutOfSpaceStream, offlineVolumesStream 
		
	)

end

--------------------------------------------------------------------------------

local function getIsActivelySyncingStream( modelData )
	
	local self = modelData
	
	-- This observable emits a boolean at least once every 15 seconds that
	-- indicates whether or not the system is actively syncing.
	
	return Observable.combineLatest(
	
		function( syncPaused, online, loggedIn, 
			syncMetaUp, syncMetaDown, 
			syncOrigUp, syncOrigOrProxyDown,
			syncStylesDown )

			if not syncPaused and online and loggedIn then
				
				if syncMetaUp or
					syncMetaDown or
					syncOrigUp or
					syncOrigOrProxyDown or
					syncStylesDown then
					
					return true
					
				end
				
			end
			
			return false
			
		end,
		self._isSyncPausedStream, self._isOnlineStream, self._isLoggedInStream,
		self._isSyncUploadingMetadataStream, self._isSyncDownloadingMetadataStream,
		self._isSyncUploadingBinariesStream, self._isSyncDownloadingOriginalsOrProxiesStream,
		self._isSyncDownloadingStylesStream,
		Observable.timer( 0, 15, WFRxScheduler.sharedScheduler() )
		
	)

end

--------------------------------------------------------------------------------

local function getCurrentSyncStatusStream( modelData )
	
	local self = modelData
	
	return Observable.combineLatest(
	
		function( syncPaused, online, loggedIn, syncPending,
			initialSync, syncMetaUp, syncMetaDown,
			syncOrigUp, syncOrigOrProxyDown, syncStylesDown,
			outOfStorage, lowOnStorage,
			volumesOutOfSpace, offlineVolumes, volumesRequiringSyncing,
			updateRequired, persistentSyncError )

			local outOfLocalStorageForOriginals, outOfLocalStorageForProxies, outOfLocalStorageOnMainVolume
					= self._userModelLocalStorage:whichVolumes( volumesOutOfSpace )

			local offlineVolumeForOriginals, offlineVolumeForProxies = self._userModelLocalStorage:whichVolumes( offlineVolumes )
	
			local offlineVolumeRequiresSync = self._userModelLocalStorage:offlineVolumeRequiresSyncing( offlineVolumes, volumesRequiringSyncing )
	
			local status

			if not loggedIn then
				status = nil -- no status available when not signed in
			elseif updateRequired then
				status = 'SyncPausedUpdateRequired'
			elseif syncPaused then
				status = 'SyncPaused'
			elseif not online then
				status = 'SyncDisconnectedNoNetwork'
			elseif offlineVolumeRequiresSync then
				status = 'SyncVolumeOfflineRequiresSyncing'
			elseif outOfLocalStorageOnMainVolume then
				status = 'SyncErrorNoDiskSpaceAvailableOnMainVolume'
			elseif outOfLocalStorageForOriginals then
				status = 'SyncErrorNoDiskSpaceAvailableForOriginals'
			elseif outOfLocalStorageForProxies then
				status = 'SyncErrorNoDiskSpaceAvailableForProxies'
			elseif outOfStorage then
				status = 'SyncErrorOutOfStorage'
			elseif initialSync and ( syncMetaUp or syncOrigUp ) then
				status = 'SyncInitialAndUp'
			elseif initialSync then
				status = 'SyncingInitial'
			elseif ( syncMetaDown or syncOrigOrProxyDown ) and ( syncMetaUp or syncOrigUp ) then
				status = 'SyncingUpAndDown'
			elseif syncMetaUp or syncOrigUp then
				status = 'SyncingUp'
			elseif syncMetaDown or syncOrigOrProxyDown or syncStylesDown then
				status = 'SyncingDown'
			elseif syncPending then
				status = 'SyncPending'
			elseif lowOnStorage then
				status = 'SyncLowOnStorage'
			elseif offlineVolumeForOriginals then
				status = 'SyncVolumeOfflineForOriginals'
			elseif offlineVolumeForProxies then
				status = 'SyncVolumeOfflineForProxies'
			elseif persistentSyncError then
				status = 'SyncCompletedWithError'
			else
				status = 'SyncNull'
			end

			return status

		end,
		self._isSyncPausedStream, self._isOnlineStream, self._isLoggedInStream, self._isSyncPendingStream,
		self._isInitialSyncStream, self._isSyncUploadingMetadataStream, self._isSyncDownloadingMetadataStream,
		self._isSyncUploadingBinariesStream, self._isSyncDownloadingOriginalsOrProxiesStream,
		self._isSyncDownloadingStylesStream,
		self._isOutOfOzStorageStream, self._isLowOnOzStorageStream, self._volumesOutOfSpaceStream, 
		self._offlineVolumesStream, self._volumesRequiringSyncingStream, 
		self._isUpdateRequiredStream, self._havePersistentSyncErrorStream
		
	)
	
end

--------------------------------------------------------------------------------

function NmbModelDataImp:__init( serverBase, rootCatalogDir, processId, imsOverrides )

	check( 'NmbModelData', 'string', 'string', '?string', '?table', '?table' )


	-- User Info
	
	local userIdStream, lastUserId = getUserIdStream( processId, imsOverrides )

	self._userIdStream = userIdStream
	self._lastUserId = lastUserId

	self._catalogIdStream = ReplaySubject( 1 )

	self._deletedCatalogStream = BehaviorSubject( false )

	self._userInfoCachePath = rootCatalogDir == ':memory:' and rootCatalogDir or 
										KSPathUtils.child( rootCatalogDir, 'user' )

	self._serverBaseUrlStream = BehaviorSubject( serverBase )

	self._currentUserStream = createCurrentUserStream( self._serverBaseUrlStream, 
			self._userIdStream, self._userInfoCachePath )

	self._userInfoStream = createUserInfoStream( self._currentUserStream )

	self._shouldSyncUserInfoStream = BehaviorSubject( false )

	self._isSyncingUserInfoStream = createIsSyncingUserInfoStream( 
			self._shouldSyncUserInfoStream, self._currentUserStream )

	self._loginStatusStream = BehaviorSubject( { status = "init" } )

	self._loginMechanismStream = BehaviorSubject( nil )

	self._isLoggedInStream = self._userIdStream
		:map( function( userId )
			return not not userId
		end )

	-- Local Storage
	
	local userModelLocalStorage = NmbUserModel_LocalStorage()
	self._userModelLocalStorage = userModelLocalStorage
	self._volumesOutOfSpaceStream = userModelLocalStorage:getVolumesOutOfSpaceStream()
	self._offlineVolumesStream = userModelLocalStorage:getOfflineVolumesStream()
	self._volumesRequiringSyncingStream = userModelLocalStorage:getVolumesRequiringSyncingStream()
	self._readOnlyVolumesStream = userModelLocalStorage:getReadOnlyVolumesStream()
	
	-- Sync

	self._hasValidConnectionToOzStream = createHasValidConnectionToOz( self._currentUserStream )
	
	if USE_NON_CC_CREDENTIALS then
	
		-- For automation, start the daemon as if it's online.
		self._isOnlineStream = BehaviorSubject( true )
		
	else
	
		-- For standard thor login, start the daemon as if it's offline.
		self._isOnlineStream = BehaviorSubject( false )
		
	end
	
	self._isSyncPendingStream = BehaviorSubject( false )
	self._havePersistentSyncErrorStream = BehaviorSubject( false )
	self._isInitialSyncStream = BehaviorSubject( true )
	self._isSyncUploadingMetadataStream = BehaviorSubject( false )
	self._isSyncDownloadingMetadataStream = BehaviorSubject( false )
	self._isSyncUploadingBinariesStream = BehaviorSubject( false )
	self._isSyncDownloadingOriginalsOrProxiesStream = BehaviorSubject( false )
	self._isSyncDownloadingStylesStream = BehaviorSubject( false )
	self._shouldPreventSleepWhileSyncingStream = BehaviorSubject( KSPrefs.shouldPreventSleepWhileSyncing )
	self._isSyncPausedStream = getIsSyncPausedStream( self._shouldSyncUserInfoStream )
	self._isActivelySyncingStream = getIsActivelySyncingStream( self )
	
	self._hostIsReachableStream = self._serverBaseUrlStream
		:flatMap(function( baseUrl )
			return WFReachability.hostIsReachable( baseUrl ) 
		end)
	
	-- Cloud Storage
	
	self._isOutOfOzStorageStream = BehaviorSubject( false )
	self._isLowOnOzStorageStream = BehaviorSubject( false )

	-- File Management

	self._mainVolumeGuidStream = BehaviorSubject( false )
	self._keepAllOriginalsStream = BehaviorSubject( NmbFileManagementSettings.keepAllOriginals() )
	self._keepAllProxiesStream = BehaviorSubject( NmbFileManagementSettings.keepAllProxies() )
	self._targetDiskSpacePercentageStream = BehaviorSubject( NmbFileManagementSettings.targetNimbusDiskSpacePercentage() )
	self._originalStorageLocationInfoStream = BehaviorSubject( false )
	self._proxyStorageLocationInfoStream = BehaviorSubject( false )
	self._fileMoverPostponedUntilDateStream = BehaviorSubject( KSPrefs.postponeFileMoveUntilDate )
	
	-- App Updates / Residual / Trial Mode

	self._isUpdateAvailableStream = BehaviorSubject( false )
	self._isUpdateRequiredStream = BehaviorSubject( false )
	self._residualModeStream = BehaviorSubject( not not KSPrefs.residualMode )
	self._restrictedCountryModeStream = BehaviorSubject( false )
	self._trialEndDateStream = getTrialEndDateStream( self._userInfoStream )
	self._deletionDateStream = getDeletionDateStream( self._userInfoStream )
		
	-- The following streams rely on data from both file management and sync streams.

	self._currentDaemonDownloadStatusStream = getCurrentDaemonDownloadStatusStream( 
			userModelLocalStorage, self._isSyncPausedStream, self._isOnlineStream, 
			self._volumesOutOfSpaceStream, self._offlineVolumesStream )

	self._currentSyncStatusStream = getCurrentSyncStatusStream( self )
	
end

--------------------------------------------------------------------------------

function NmbModelDataImp:onSynchronizerClose()

	self._isSyncUploadingMetadataStream:onNext( false )

	self._isSyncDownloadingMetadataStream:onNext( false )

	self._isSyncUploadingBinariesStream:onNext( false )

	self._isSyncDownloadingOriginalsOrProxiesStream:onNext( false )
	
	self._isSyncDownloadingStylesStream:onNext( false )

	self._isSyncPendingStream:onNext( false )

	self._havePersistentSyncErrorStream:onNext( false )

	self._isInitialSyncStream:onNext( false )
	
	self._isOutOfOzStorageStream:onNext( false )

	self._isLowOnOzStorageStream:onNext( false )

	self._volumesOutOfSpaceStream:onNext( {} )
	self._offlineVolumesStream:onNext( {} )
	self._volumesRequiringSyncingStream:onNext( {} )
	self._readOnlyVolumesStream:onNext( {} )

end

--------------------------------------------------------------------------------

function NmbModelDataImp:close()

	self._userIdStream:dispose()

	self._catalogIdStream:dispose()

end

--------------------------------------------------------------------------------

local NmbModelData = Rx.Class 'NmbModelData'( NmbModelDataImp )


--============================================================================--

-- @mark NmbUserModel

local NmbUserModel = {}

--------------------------------------------------------------------------------

function NmbUserModel.makeUserModel( serverBase, rootCatalogDir, processId, imsOverrides, pixelProviderSettings )

	check( 'string', '?string', '?string', '?table', '?table' )

	pixelProviderSettings = pixelProviderSettings or {}
	
	imsOverrides = imsOverrides or {}
	
	logger:infof( "NmbUserModel.makeUserModel( %s, %s, %s, %s, %s, %s )",
				  serverBase, rootCatalogDir, processId, imsOverrides, 
				  pixelProviderSettings.maxSaves, pixelProviderSettings.maxRenders )

	local modelData = NmbModelData( serverBase, rootCatalogDir, processId, imsOverrides )

	----------------------------------------------------------------------------

	local modelValue = {

		-- model info
		
		userModelId = Observable.just( KSIdGenerator.uuidInLowercaseHex() ),

		-- basic user info

		ozUserId = modelData._userIdStream,
		catalogId = modelData._catalogIdStream:startWith( nil ),
		isLoggedIn = modelData._isLoggedInStream,
		loginStatus = modelData._loginStatusStream,
		loginMechanism = modelData._loginMechanismStream,
		email = currentUserFieldValueForPath( modelData, 'email' ),
		fullName = currentUserFieldValueForPath( modelData, 'full_name' ),
		firstName = currentUserFieldValueForPath( modelData, 'first_name' ),
		lastName = currentUserFieldValueForPath( modelData, 'last_name' ),
		
		-- sync
		
		isSyncingUserInfo = modelData._isSyncingUserInfoStream,
		syncPaused = modelData._isSyncPausedStream,
		isOnline = modelData._isOnlineStream,
		hasValidConnectionToOz = modelData._hasValidConnectionToOzStream,
		syncStatus = modelData._currentSyncStatusStream,
		syncUploadingMetadata = modelData._isSyncUploadingMetadataStream,
		syncDownloadingMetadata = modelData._isSyncDownloadingMetadataStream,
		syncUploadingBinaries = modelData._isSyncUploadingBinariesStream,
		syncDownloadingOriginalsOrProxies = modelData._isSyncDownloadingOriginalsOrProxiesStream,
		syncDownloadingStyles = modelData._isSyncDownloadingStylesStream,
		syncDownloadingMetadataInitial = modelData._isInitialSyncStream,
		syncPending = modelData._isSyncPendingStream,
		
		-- oz storage

		isOutOfOzStorage = modelData._isOutOfOzStorageStream,
		isLowOnOzStorage = modelData._isLowOnOzStorageStream,

		-- local storage

		volumesOutOfSpace = modelData._volumesOutOfSpaceStream:map( mapVolumeGuidToName ),
		offlineVolumes = modelData._offlineVolumesStream:map( mapVolumeGuidToName ),
		volumesRequiringSyncing = modelData._volumesRequiringSyncingStream:map( mapVolumeGuidToName ),
		readOnlyVolumes = modelData._readOnlyVolumesStream:map( mapVolumeGuidToName ),
		
		-- file management settings

		fmMainVolumeGuid = modelData._mainVolumeGuidStream,
		fmKeepAllOriginals = modelData._keepAllOriginalsStream,
		fmKeepAllProxies = modelData._keepAllProxiesStream,
		fmDiskSpacePercentage = modelData._targetDiskSpacePercentageStream,
		fmOriginalStorageLocationInfo = modelData._originalStorageLocationInfoStream,
		fmProxyStorageLocationInfo = modelData._proxyStorageLocationInfoStream,
		fmFileMoverPostponedUntilDate = modelData._fileMoverPostponedUntilDateStream,
		
		-- account status
		
		isInResidualMode = modelData._residualModeStream,
		isInRestrictedCountryMode = modelData._restrictedCountryModeStream,
		accountStatus = currentUserFieldValueForPath( modelData, 'entitlement/status' ),
		trialEndDate = modelData._trialEndDateStream,
		deletionDate = modelData._deletionDateStream,
		
		-- upgrades
		
		isLightroomCCUpdateAvailable = modelData._isUpdateAvailableStream,
		isLightroomCCUpdateRequired = modelData._isUpdateRequiredStream,
		
	}

	----------------------------------------------------------------------------

	addModelValueToDiagnosticLog( modelValue )

	----------------------------------------------------------------------------

	return {

		_imsOverrides = imsOverrides,

		value = modelValue,

		models = {

			--------------------------------------------------------------------

			appModel = function( ... )
				return getAppModel( rootCatalogDir, pixelProviderSettings, modelData, ... )
			end,
			
			--------------------------------------------------------------------
			
			indexStatus = function( self )
			
				return require 'NmbForceReindexing'.makeModel( getWFCatalog( rootCatalogDir, pixelProviderSettings, modelData ) )
				
			end,
			
			--------------------------------------------------------------------
			
			catalog = function( self )
			
				self._catalogModel = require 'NmbRootModel'.makeModel( getWFCatalog( rootCatalogDir, pixelProviderSettings, modelData ) )
				return self._catalogModel
				
			end,
			
			--------------------------------------------------------------------
			
			lastUserId = function()
				return {
					value = Observable.just( modelData._lastUserId )
				}
			end,
			
			--------------------------------------------------------------------
			
			hostIsReachable = function()
				return {
					value = modelData._hostIsReachableStream
				}
			end,
			
			--------------------------------------------------------------------
			
			appNeedsRestart = function()
				return {
					value = modelData._deletedCatalogStream
				}
			end,

			--------------------------------------------------------------------

			tempFolderPathForOriginals = function()
				return {
					value = Observable.combineLatest( getWFCatalog( rootCatalogDir, pixelProviderSettings, modelData ), 
							modelData._originalStorageLocationInfoStream )
						:map( function( wfCatalog, locationInfo )
						
							if locationInfo and locationInfo.volumeGuid then
								return wfCatalog:fileSpec():tempFolderPathForVolume( locationInfo.volumeGuid )
							end
							
							return false
							
						end )
						:take( 1 )
				}
			end,
			
			--------------------------------------------------------------------
			
			diagnosticLog = function( _, logName, logFileName )
				return require 'WFDiagnosticLog'.makeSimpleModel( logName, logFileName )
			end,

			--------------------------------------------------------------------

			getImsAuthUrl = function( _, optOverrides )

				logger:infof( "getImsAuthUrl original serverBase = %s, optOverrides = %s",
							  serverBase, optOverrides )

				optOverrides = optOverrides or {}

				if not optOverrides.baseServerUrl then
					optOverrides.baseServerUrl = serverBase
				end

				optOverrides.apiVersion = 'v2'

				if optOverrides.imsClientId and optOverrides.imsClientSecret then
					require 'WFImsAppIdentity'.setClientIdAndSecret( optOverrides.imsClientId,
																	 optOverrides.imsClientSecret )
				end

				return {
					value = require 'WFOzClient'.getImsAuthUrl( optOverrides )
				}

			end,

			--------------------------------------------------------------------

			getJumpUrlForUrl = function( _, url, targetClientId, optOverrides )

				logger:infof( "getJumpUrlForUrl url = %s, webCid = %s",
							  url, targetClientId )

				optOverrides = optOverrides or {}

				if not optOverrides.baseServerUrl then
					optOverrides.baseServerUrl = serverBase
				end

				return {
				
					value = getWFCatalog( rootCatalogDir, pixelProviderSettings, modelData )

						:flatMap( function( catalog )
							return catalog:ozUser():fetchTokens()
						end )

						:flatMap( function( accessToken )
							return require 'WFOzClient'.getJumpUrlForUrl( url, targetClientId, accessToken, optOverrides )
						end )
					
				}

			end,

			--------------------------------------------------------------------

			loginFromTokens = function( ... )
				loginFromTokens( modelData, ... )
				return {
					value = true,
				}
			end,

			--------------------------------------------------------------------

			avatar = function( self, avatarDir, adobeId )
			
				return {
					value = modelData._currentUserStream
								:flatMapLatest( function( wfCatUser )
									if wfCatUser then
										return WFCatAvatar.retrieveAvatar( wfCatUser, adobeId, avatarDir )
									else
										return Observable.empty()
									end
								end )
				}
				
			end,
			
			--------------------------------------------------------------------
			
			startProxyAuthErrorMonitoring = function( self )
			
				return {
					value = require "NmbProxyAuthErrorMonitor".startProxyAuthErrorMonitor()
				}
				
			end,

		},

		events = {

			--------------------------------------------------------------------

			activateOfflineMode = function( ... )
				activateOfflineMode( modelData, ... )
			end,

			--------------------------------------------------------------------

			activateOnlineMode = function( ... )
				activateOnlineMode( modelData, ... )
			end,
			
			--------------------------------------------------------------------
			
			useLastUser = function( ... )
				useLastUser( modelData, ... )
			end,
			
			--------------------------------------------------------------------
			
			loginFromUsernameAndPassword = function( ...)
				loginFromUsernameAndPassword( modelData, ... )
			end,

			--------------------------------------------------------------------

			loginFromTokens = function( ...)
				loginFromTokens( modelData, ... )
			end,
			
			--------------------------------------------------------------------
			
			logout = function( self, message )
				modelData._userIdStream:onNext( nil )
				modelData._loginStatusStream:onNext( { status = "logout" } )
				modelData._loginMechanismStream:onNext( nil )
			end,

			--------------------------------------------------------------------

			startSync = function()
				takeOneValueAndRunFunc( modelData._isUpdateRequiredStream, function( updateRequired )
					if updateRequired then
						logger:info( "Not starting sync, because a client upgrade is required." )
						return
					elseif not KSPrefs.syncPaused or KSPrefs.syncPaused == kClientUpgradeRequired then
					
						-- When KSPrefs.syncPaused is "upgrade_required", we attempt to resume sync,
						-- because the user didn't choose to pause it. If an upgrade is still required,
						-- that will become clear soon enough, and sync will be paused with the
						-- the corresponding notification presented to the user.
					
						logger:infof( "Resuming sync with (previous) syncPaused preference of %s", KSPrefs.syncPaused )
						KSPrefs.syncPaused = false
						startSyncingUserAndCatalog( modelData )
					end
				end )
			end,

			--------------------------------------------------------------------

			stopSync = function()
				stopSyncingUserAndCatalog( modelData )
			end,

			--------------------------------------------------------------------

			pauseSync = function()
				KSPrefs.syncPaused = true
				stopSyncingUserAndCatalog( modelData )
			end,
			
			--------------------------------------------------------------------
			
			preventSleep = function( self, reason, optKeepDisplayAwake )
				WFInsomnia.preventSystemSleep( reason, optKeepDisplayAwake )
			end,
			
			--------------------------------------------------------------------
			
			residualMode = function()
				
				if not KSPrefs.residualMode and not KSPrefs.restrictedCountryMode then
				
					KSPrefs.residualMode = true
				
					modelData._residualModeStream:onNext( KSPrefs.residualMode )
				
					-- If wfCatSynchronizer doesn't yet exist,
					-- it will be created using the current value
					-- in KSPrefs.residualMode
				
					if wfCatSynchronizer then
					
						wfCatSynchronizer:stopSyncing()
							-- Stop the normal mode sync.
							
						wfCatSynchronizer:enableResidualMode()
						
						startSyncingUserAndCatalog( modelData )
						
					end
					
					if KSPrefs.syncPaused then
						KSPrefs.syncPaused = false
					end
					
				end
				
			end,
			
			--------------------------------------------------------------------
			
			normalMode = function()
			
				if KSPrefs.residualMode and not KSPrefs.restrictedCountryMode then
			
					KSPrefs.residualMode = false
			
					modelData._residualModeStream:onNext( KSPrefs.residualMode )
			
					-- If wfCatSynchronizer doesn't yet exist,
					-- it will be created using the current value
					-- in KSPrefs.residualMode
			
					if wfCatSynchronizer then
					
						wfCatSynchronizer:stopSyncing()
							-- Stop the residual mode sync.
							
						wfCatSynchronizer:disableResidualMode()
						
						startSyncingUserAndCatalog( modelData )
						
					end
					
					if KSPrefs.syncPaused then
						KSPrefs.syncPaused = false
					end
				
				end
				
			end,

			--------------------------------------------------------------------

			resumeSync = function()
				takeOneValueAndRunFunc( modelData._isUpdateRequiredStream, function( updateRequired )
					if updateRequired then
						logger:info( "Not resuming sync, because a client upgrade is required." )
						return
					end
					KSPrefs.syncPaused = false
					startSyncingUserAndCatalog( modelData )
				end )
			end,

			--------------------------------------------------------------------

			resetFileManagementPreferences = function( self ) 
				
				NmbFileManagementSettings.reset()
				
				KSPrefs.postponeFileMoveUntilDate = nil
				modelData._fileMoverPostponedUntilDateStream:onNext( KSPrefs.postponeFileMoveUntilDate )
				
				local keepAllOriginals = NmbFileManagementSettings.keepAllOriginals()
				local keepAllProxies = NmbFileManagementSettings.keepAllProxies()
				local targetDiskSpace = NmbFileManagementSettings.targetNimbusDiskSpacePercentage()
				local newOriginalStorageLocationInfo = NmbFileManagementSettings.originalStorageLocationInfo()
				local newProxyStorageLocationInfo = NmbFileManagementSettings.proxyStorageLocationInfo()
				
				modelData._keepAllOriginalsStream:onNext( keepAllOriginals )
				modelData._keepAllProxiesStream:onNext( keepAllProxies )
				modelData._targetDiskSpacePercentageStream:onNext( targetDiskSpace )
				modelData._originalStorageLocationInfoStream:onNext( newOriginalStorageLocationInfo )
				modelData._proxyStorageLocationInfoStream:onNext( newProxyStorageLocationInfo )
				
			end,

			--------------------------------------------------------------------

			updateFileManagementPreferences = function( self, targetDiskSpace, keepAllOriginals, keepAllProxies ) 
				
				if keepAllOriginals ~= nil then
					NmbFileManagementSettings.setKeepAllOriginals( keepAllOriginals )
					modelData._keepAllOriginalsStream:onNext( keepAllOriginals )
				end

				if keepAllProxies ~= nil then
					NmbFileManagementSettings.setKeepAllProxies( keepAllProxies )
					modelData._keepAllProxiesStream:onNext( keepAllProxies )
				end
				
				if targetDiskSpace then
					NmbFileManagementSettings.setTargetDiskSpacePercentage( targetDiskSpace )
					modelData._targetDiskSpacePercentageStream:onNext( targetDiskSpace )
				end
				
			end,

			--------------------------------------------------------------------

			changeOriginalStorageLocation = function( self, newOriginalStorageLocation ) 
				
				KSPrefs.postponeFileMoveUntilDate = nil
				modelData._fileMoverPostponedUntilDateStream:onNext( KSPrefs.postponeFileMoveUntilDate )
				
				NmbFileManagementSettings.setOriginalStorageLocation( newOriginalStorageLocation )
				modelData._originalStorageLocationInfoStream:onNext( NmbFileManagementSettings.originalStorageLocationInfo() )
				
			end,

			--------------------------------------------------------------------

			changeProxyStorageLocation = function( self, newProxyStorageLocation ) 
				
				KSPrefs.postponeFileMoveUntilDate = nil
				modelData._fileMoverPostponedUntilDateStream:onNext( KSPrefs.postponeFileMoveUntilDate )
				
				NmbFileManagementSettings.setProxyStorageLocation( newProxyStorageLocation )
				modelData._proxyStorageLocationInfoStream:onNext( NmbFileManagementSettings.proxyStorageLocationInfo() )
				
			end,

			--------------------------------------------------------------------

			updateVolumeStorageName = function( self, volumeGuid, readableIdentifier ) 
				
				NmbFileManagementSettings.setVolumeReadableIdentifier( volumeGuid, readableIdentifier )
				
				modelData._originalStorageLocationInfoStream:onNext( NmbFileManagementSettings.originalStorageLocationInfo() )
				modelData._proxyStorageLocationInfoStream:onNext( NmbFileManagementSettings.proxyStorageLocationInfo() )
				
			end,
			
			--------------------------------------------------------------------
			
			togglePreventSleepWhileSyncing = function( self, shouldPreventSleepWhileSyncing ) 
				
				modelData._shouldPreventSleepWhileSyncingStream:onNext( shouldPreventSleepWhileSyncing )
				KSPrefs.shouldPreventSleepWhileSyncing = shouldPreventSleepWhileSyncing
		
			end,
			
			--------------------------------------------------------------------
			
			postponeFileMove = function( self, maybePostponeDurationInSeconds ) 
				
				local daysToSeconds = function( days )
					return days * 24 * 60 * 60
				end
				
				local postponeDelayInSeconds = POSTPONE_DURATION or maybePostponeDurationInSeconds or daysToSeconds( 6 )
					-- Intentionally delaying 6 days instead of 7

				if postponeDelayInSeconds <= 0 then
					KSPrefs.postponeFileMoveUntilDate = nil
				else
					KSPrefs.postponeFileMoveUntilDate = require 'KSDate'.timeNow() + postponeDelayInSeconds
				end
				
				modelData._fileMoverPostponedUntilDateStream:onNext( KSPrefs.postponeFileMoveUntilDate )
				
			end,
			
			--------------------------------------------------------------------
			
			updateAvatar = function( self, newAvatarFilePath, ext )
				
				return modelData._currentUserStream
					:take( 1 )
					:flatMap( function( wfCatUser )
						if wfCatUser then
							return WFCatAvatar.updateAvatar( wfCatUser, newAvatarFilePath, ext )
						else
							return Observable.empty()
						end
					end )
					:call( WFRxRunUntilDone )
						
			end,
			
			--------------------------------------------------------------------
			
			triggerForceLightroomClassicUpgradeFlag = function( self )
				if userInfoSynchronizer_now then
					userInfoSynchronizer_now:triggerForceLightroomClassicUpgradeFlag()
				end
			end,
			
			--------------------------------------------------------------------
			
			setProxyCredentials = function( self, username, password )
				require "WFRxHttpClient".setProxyCredentials( username, password )
				modelData._serverBaseUrlStream
					:tap( function( baseUrl )
						require "WFReachability".refreshReachabilityStatus( baseUrl )
					end )
					:call( WFRxRunUntilDone )
				
			end,
			
			--------------------------------------------------------------------
			
			DEBUG_dumpCoverage = function( _, path )
				require 'SetupCoverage'.dumpCoverage( path )
			end,

			--------------------------------------------------------------------

			DEBUG_incStorageUsed = function( _, amountUsed )
				WFCatStorage.debugStorageUsed( amountUsed )
			end,

		},

		close = function()

			logger:info( "Closing" )

			if wfCatUser_now then
				wfCatUser_now:close()
				wfCatUser_now = nil
			end
			
			if modelData then
				modelData:close()
				modelData = nil
			end
			
			if userInfoSynchronizer_now then
				userInfoSynchronizer_now:close()
				userInfoSynchronizer_now = nil
			end
			
			if wfOzCatalog_now then
				wfOzCatalog_now:close()
				wfOzCatalog_now = nil
			end

		end,

	}
	
end

--------------------------------------------------------------------------------

return NmbUserModel
