--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFURI = require 'WFURI'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCapabilities = require 'WFCapabilities'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSDate = require 'KSDate'
local WFDeviceInfo = require 'WFDeviceInfo'

	-- WF.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'
local WFOzContentTypes = require 'WFOzContentTypes'

	-- WF.oz_client.sync_agent
local OzCentralChunkedUpload = require "OzCentralChunkedUpload"

	-- WF.oz_client.video
local WFVideoTranscodingOptions = require 'WFVideoTranscodingOptions'

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

local check = assert( WFArguments.check )

local binaryUploadLogger = WFLogger( 'OzBinaryUploadUtils' )
--	:setLevel( 'DEBUG' )

--------------------------------------------------------------------------------

local MIN_SIZE_FOR_CHUNKING = WFCapabilities.getCapability "OzCentralChunkedUpload.minSizeForChunkedUpload" or ( 5 * 1024 * 1024 )

local string_match = string.match

--==============================================================================

local OzBinaryUploadUtils = {}

--------------------------------------------------------------------------------

local function isGuid( arg )

	if not arg then
		return
	end
	
	return not not ( type( arg ) == 'string'
		and #arg == 32
		and string_match( arg, '^[0-9a-f]+$' ) )
		
end

--------------------------------------------------------------------------------

local function verifyGuid( guid )

	if not isGuid( guid ) then
		return
	end
	
	return guid
	
end

--------------------------------------------------------------------------------

local function isRevisionUploaded_imp( docValue )
	
	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' ) or {}
	local revId = docValue._revId
	
	local count = #revisionIds
			
	if count == 0 then
		return false
	end
			
	for i = count, 1, -1 do
		
		local ozRevId = revisionIds[ i ]
		if revId == ozRevId then
			return true
		end
		
	end
	
	return false
end

--------------------------------------------------------------------------------

local function getMasterUploadRevision_imp( docValue, localOnly )

	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' )
								  
	local firstRevisionId = revisionIds and revisionIds[ 1 ]
	
	return verifyGuid( firstRevisionId )
	
end

--------------------------------------------------------------------------------

local function getLastUploadRevision_imp( docValue, localOnly )

	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' )
								  
	local lastRevisionId = revisionIds and revisionIds[ #revisionIds ]
	
	return verifyGuid( lastRevisionId )
	
end

--------------------------------------------------------------------------------


local function shouldUploadMaster_imp( docValue )

	if docValue.type ~= 'asset' then
		return false, 'notAsset'
	end
	
	local localOnly = docValue._localOnly or {}
	
	local importSource = docValue.importSource or {}
	
	local localAssetId = importSource.localAssetId
	local uniqueDeviceId = importSource.uniqueDeviceId
	
	local localUrl = localOnly.localUrl or ( localAssetId and ( uniqueDeviceId == WFDeviceInfo.uniqueDeviceId() ) )
	
	if not localOnly.originalFilePath and not localUrl then
		return false, 'itemMissingLocally'	
	end
	
	local chunkedUploadRestartCount = derefOpt( localOnly, 'chunkedUploadRestartCount' ) or 0
	if chunkedUploadRestartCount > 5 then
		return false, 'restartExceedMaxLimit'
	end
	
	local serverLinks = derefOpt( docValue, '_serverProvidedMetadata', 'links' ) or {}
	local masterUrl = derefOpt( serverLinks, '/rels/master', 'href' )
	local isValidMasterUrl = not OzCentralChunkedUpload.getSessionId( masterUrl )
	
	-- INFO [pkamin 2017-10-02]: disable this code for now as it isn't implemented appropriately for all binary cases
	local ozSha256Error = false --derefOpt( docValue, 'importSource', 'status', 'error', 'all', 'reason', 'sha256' ) == "does not match"
	
	-- If there's a sha256 error that we can fix, there should be a binaryUploadError for the current ozMasterUrl
	-- If so, we can actually clear the sha256error and try to re-upload locally.
	local localSha256Ok = masterUrl and derefOpt( docValue, '_localOnly', 'binaryUploadErrors', '/rels/master', masterUrl )
	local blockedBySha256Error = not not ( ozSha256Error and ( not localSha256Ok ) )
	
	local ozUrl = derefOpt( localOnly, 'original', 'ozUrl' )
	local ozUrlUploadedTime = derefOpt( localOnly, 'original', 'uploaded' )
	local ozMultipartCompleteError = derefOpt( localOnly, 'ozError', 'subtype' ) == 'MultipartCompleteError'
	
	if ozUrl and ozUrlUploadedTime and not ozMultipartCompleteError and not ( blockedBySha256Error or ( masterUrl and ozSha256Error and not blockedBySha256Error ) ) then
		return false, 'itemIsCurrent'
	end
	
	local originalStatus = derefOpt( docValue, 'importSource', 'originalStatus' )
	
	if originalStatus == 'lost' then
		return false, 'originalIsLost'
	end
	
	local localOriginalHasError = derefOpt( localOnly, 'localOriginalError' )
	
	if localOriginalHasError then
		return false, 'originalError'
	end
	
	if ( (localOnly.originalFilePath and localOnly.originalFilePath ~= 'missing') or localUrl )
			and ( not masterUrl or ( ozUrl and not isValidMasterUrl ) or blockedBySha256Error or ( masterUrl and ozSha256Error and not blockedBySha256Error ) ) 
			then
	
		return true, 'itemNotUploaded'
		
	end
	
	return false, 'itemIsCurrent'
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.shouldUploadMaster( docValue )

	check( 'table' )
	
	return shouldUploadMaster_imp( docValue )
	
end

--------------------------------------------------------------------------------

--------------------------------------------------------------------------------

local function shouldUploadXmpDevelop_imp( docValue )
	
	if docValue.type ~= 'asset' then
		return false, 'notAsset'
	end

	-- uploading a changed external xmp develop settings is triggered by Oz: it sends us
	-- an updated document that indicates that links[ '/rels/xmp/develop' ] is invalid 

	local linkXmpDevelop = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop' )

	local xmpFiles = derefOpt( docValue, '_localOnly', 'files', 'xmp_develop')
	
	local xmpDevelopSha256 = derefOpt( docValue, 'develop', 'xmpCameraRaw', 'sha256' )
	
	if linkXmpDevelop and linkXmpDevelop.invalid and xmpFiles and xmpDevelopSha256 == xmpFiles.sha256 
	  and not xmpFiles.uploadStatus then
	  	-- file was invalidated by Oz
		return true, 'itemNotUploaded'
	end
	
	if xmpFiles and not xmpFiles.uploadStatus and not linkXmpDevelop and xmpDevelopSha256 then

		if isRevisionUploaded_imp( docValue ) then
			-- not uploaded to Oz yet - first creation
			return true, 'itemNotUploaded'
		end
	
	end
	
	return false, 'itemIsCurrent'
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.shouldUploadXmpDevelop( docValue )

	check( 'table' )
	
	return shouldUploadXmpDevelop_imp( docValue )
	
end

--------------------------------------------------------------------------------

local function shouldUploadPresetProfile_imp( docValue )

	if docValue.type ~= 'asset' then
		return false, 'notAsset'
	end

	local presetProfileFile = derefOpt( docValue, '_localOnly', 'files', 'presetProfile')

	local masterLink = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master' )

	local importSourceSha256 = derefOpt( docValue, 'importSource', 'sha256' )

	if masterLink and masterLink.invalid and presetProfileFile and importSourceSha256 == presetProfileFile.sha256
	  and not presetProfileFile.uploadStatus then
		-- file was invalidated by Oz
		return true, 'itemNotUploaded'
	end


	if presetProfileFile and not presetProfileFile.uploadStatus and not masterLink and importSourceSha256 then

		if isRevisionUploaded_imp( docValue ) then
			-- not uploaded to Oz yet - first creation
			return true, 'itemNotUploaded'
		end

	end

	return false, 'itemIsCurrent'

end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.shouldUploadPresetProfile( docValue )

	check( 'table' )

	return shouldUploadPresetProfile_imp( docValue )

end

--------------------------------------------------------------------------------

-- jrechtie 2018/0/24 copied/modified here from WFCatIndex.lua to remove dependency.
function OzBinaryUploadUtils.isValidUploadableAssetDocument( docValue )

	return docValue
		and not docValue._deleted
		and docValue.type == 'asset'
		and ( docValue.subtype == 'image'
				or docValue.subtype == 'video'
				or docValue.subtype == 'preset'
				or ( docValue.subtype == 'profile' and not ( docValue._localOnly and docValue._localOnly.uniqueId ) ) -- do not try to upload cameraProfiles, lensProfiles
			)
end

--------------------------------------------------------------------------------

local function missing_binary_on_oz_table()

	return {
		
		sql = [[CREATE TABLE missingBinariesOnOz( 
					importDate, readyToUpload, binaryType, uploadSequence, chunkType, part, 
					bytesNeedToUpload, fileSize, filePath, url, contentType, subtype )]],
					
		
		map = function( docValue, emit )
		
			if not OzBinaryUploadUtils.isValidUploadableAssetDocument( docValue ) then
				return
			end
		
			-- TODO [pkamin 2015-07-23]: if video give lower upload priority
			
			if shouldUploadMaster_imp( docValue ) then
			
				local docId = docValue._id
				
				local timestamp = derefOpt( docValue, 'importSource', 'importTimestamp' )
			
				if not timestamp then
					timestamp = docValue.captureDate
				end
				
				if not timestamp then
					timestamp = KSDate.timeInIso8601Format( nil, false )
				end
				
				local localOnly = docValue._localOnly or {}
				
				local ozUrl = derefOpt( localOnly, 'original', 'ozUrl' )
				local ozMultipartCompleteError = derefOpt( localOnly, 'ozError', 'subtype' ) == 'MultipartCompleteError'
				
				local whichRevision = getMasterUploadRevision_imp( docValue, localOnly )
				local masterCreateUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master_create', 'href' )
				
				if masterCreateUrl and whichRevision then
					masterCreateUrl = masterCreateUrl:replace( '{revision_id}', whichRevision )
				end
				
				local readyToUpload = masterCreateUrl and 1 or 0
				
				local masterUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
				
				-- INFO [pkamin 2017-10-02]: disable this code for now as it isn't implemented appropriately for all binary cases
				local ozSha256Error = false --derefOpt( docValue, 'importSource', 'status', 'error', 'all', 'reason', 'sha256' ) == "does not match"
	
				local localSha256Ok = masterUrl and derefOpt( docValue, '_localOnly', 'binaryUploadErrors', '/rels/master', masterUrl )
				local blockedBySha256Error = not not ( ozSha256Error and ( not localSha256Ok ) )
				
				local chunkedUpload = derefOpt( localOnly, 'chunkedUpload' )
				
				local originalFilePath = derefOpt( localOnly, 'originalFilePath' )
				local contentType = derefOpt( docValue, "importSource", "contentType" )
				local subtype = derefOpt( docValue, 'subtype' )	
				local fileSize = derefOpt( docValue, 'importSource', 'fileSize' )
				
				if not masterCreateUrl 
					or ozUrl
					or ozMultipartCompleteError
					or blockedBySha256Error or ( masterUrl and ozSha256Error and not blockedBySha256Error ) then
				
					binaryUploadLogger:debugf( 'emit docStore upload master error handling %s', docId )
					
					emit( timestamp, readyToUpload, 'master', 1, nil, -1, 
						fileSize, fileSize )
				
				elseif fileSize >= 2 * MIN_SIZE_FOR_CHUNKING and chunkedUpload and chunkedUpload.parts and masterCreateUrl then

					local finishUrl = derefOpt( chunkedUpload, "finish_href" )
					local restart = derefOpt( chunkedUpload, "restart" )
					
					if restart then
						
						if finishUrl then
							finishUrl = WFURI.resolve( masterCreateUrl, finishUrl )
							
							emit( timestamp, readyToUpload, 'master', 1, 'restart', -1, 
								fileSize, fileSize, nil, finishUrl, contentType, subtype )
						else
							emit( timestamp, readyToUpload, 'master', 1, 'clearRestartAndFinish', -1, 
								fileSize, fileSize, nil, nil, contentType, subtype )
						end
						
						return
					end
					
					if not finishUrl then
						
						-- if we don't have finishUrl, we should restart chunked upload
						-- sending master-finish with nil finishUrl will restart the upload
						
						binaryUploadLogger:debugf( 'emit chunk finish %s', docId )
						
						emit( timestamp, readyToUpload, 'master', 1, 'finish', 0, 
							fileSize, fileSize, nil, nil, contentType, subtype )
						
						return
					end
					
					
					local part_template = derefOpt( chunkedUpload, "part_template" )
					local totalChunkNumber = math.floor( fileSize / MIN_SIZE_FOR_CHUNKING )
					finishUrl = WFURI.resolve( masterCreateUrl, finishUrl )
					
					local finishChunks = true
					for chunkPart = 1, totalChunkNumber do
					
						local shouldUploadChunkPart = not ( chunkedUpload.parts[ chunkPart ] and chunkedUpload.parts[ chunkPart ].sessionId )
					
						if shouldUploadChunkPart then
						
							binaryUploadLogger:debugf( 'emit chunk %s chunkPart: %s', docId, chunkPart )
							
							local partUrl = string.replace( part_template, "{part_number}", tostring( chunkPart ) )

							partUrl = WFURI.resolve( masterCreateUrl, partUrl )
						
							if chunkPart < totalChunkNumber then
							
								emit( timestamp, readyToUpload, 'master', 1, 'chunk', chunkPart, 
									MIN_SIZE_FOR_CHUNKING, fileSize, originalFilePath, partUrl, contentType, subtype )
							
							else
								
								emit( timestamp, readyToUpload, 'master', 1, 'chunk', chunkPart, 
									MIN_SIZE_FOR_CHUNKING + (fileSize % MIN_SIZE_FOR_CHUNKING), fileSize, originalFilePath, partUrl, contentType, subtype )
								
							end
							
							finishChunks = false
						
						end
						
					end
				
					if finishChunks then
					
						binaryUploadLogger:debugf( 'emit chunk finish %s', docId )
					
						emit( timestamp, readyToUpload, 'master', 1, 'finish', totalChunkNumber + 1, 
							0, fileSize, nil, finishUrl, contentType, subtype )
					
					end
					
					
				-- OzBinaryUploader:_uploadMasterP would handle if not fileSize or fileSize <= 0
				elseif not originalFilePath or not fileSize or fileSize <= 0 or fileSize >= 2 * MIN_SIZE_FOR_CHUNKING then
				
					binaryUploadLogger:debugf( 'emit docStore upload master %s', docId )
					
					emit( timestamp, readyToUpload, 'master', 1, nil, -1, 
						fileSize, fileSize )
				
				else
				
					binaryUploadLogger:debugf( 'emit indexStore upload master %s', docId )
			
					emit( timestamp, readyToUpload, 'master', 1, nil, -1, 
						fileSize, fileSize, originalFilePath, masterCreateUrl, contentType, subtype )

				end
				
			end
			
			if shouldUploadXmpDevelop_imp( docValue ) then
				
				local localOnly = docValue._localOnly or {}
				
				local whichRevision = getLastUploadRevision_imp( docValue, localOnly )
				local xmpDevelopUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp_develop_create', 'href' )
				
				if not xmpDevelopUrl or not whichRevision then
					return
				end

				xmpDevelopUrl = xmpDevelopUrl:replace( '{revision_id}', whichRevision )
				
				local relPath = derefOpt( docValue, '_localOnly', 'files', 'xmp_develop', 'path' )
			
				emit( KSDate.timeInIso8601Format( nil, false ), 1, 'xmp_develop', 1, nil, -1, 
					0, 0, relPath, xmpDevelopUrl )
			
			end
			
			if shouldUploadPresetProfile_imp( docValue ) then

				local localOnly = docValue._localOnly or {}

				local whichRevision = getLastUploadRevision_imp( docValue, localOnly )
				local presetProfileCreateUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master_create', 'href' )

				if not presetProfileCreateUrl or not whichRevision then
					return
				end

				presetProfileCreateUrl = presetProfileCreateUrl:gsub( '{revision_id}', whichRevision )

				local localPath = derefOpt( docValue, '_localOnly', 'files', 'presetProfile', 'path' )
				local fileSize = derefOpt( docValue, '_localOnly', 'files', 'presetProfile', 'fileSize' )

				local contentType = 'application/rdf+xml'
				local subtype = derefOpt( docValue, 'subtype' )

				emit( KSDate.timeInIso8601Format( nil, false ), 1, 'presetProfile', 1, nil, -1,
					fileSize, fileSize, localPath, presetProfileCreateUrl, contentType, subtype )

			end

		end,
		
	}
	
end

--------------------------------------------------------------------------------

local function missing_binary_index()
	
	return [[ CREATE INDEX missingBinariesOnOz_uploadSequence ON missingBinariesOnOz (
		readyToUpload, importDate, binaryType, part ASC )]]
		
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.filterIndexDefinitions( tableDefinitions, indexDefinitions )

	tableDefinitions.missingBinaryOnOz = missing_binary_on_oz_table()
	
	indexDefinitions.missingBinaryOnOzIndex_UploadSequence = missing_binary_index()
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getUploadableAssetsSQL( maxConcurrentTasks )

	maxConcurrentTasks = maxConcurrentTasks or 100

	return [[
		SELECT 
			docId, part, binaryType, chunkType, filePath, url, fileSize, contentType, subtype
		FROM 
			missingBinariesOnOz 
		WHERE 
			readyToUpload == 1 
		ORDER BY 
			importDate, part ASC 
		LIMIT ]] .. tostring( maxConcurrentTasks )
	
end

--------------------------------------------------------------------------------

-- @mark Utility functions

--------------------------------------------------------------------------------

local function getImageUploadBaseUrl()

	return WFCapabilities.getCapability( "OzCatalogUrls.uploadImagesBaseUrl" )
		or OzCatalogUrls.baseServerUrl
		
end

--------------------------------------------------------------------------------

local function getVideoUploadBaseUrl()

	return WFCapabilities.getCapability( "OzCatalogUrls.uploadVideoBaseUrl" )
		or OzCatalogUrls.baseServerUrl
		
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getUploadBaseUrlForContentType( docValue )

	local contentType = derefOpt( docValue, 'importSource', 'contentType' )
	
	if contentType then
	
		local videoLow, videoHigh = WFOzContentTypes.getContentTypeNumberRangeForSubType( 'video' )
		
		local contentTypeNumber = WFOzContentTypes.getContentTypeNumberForContentType( contentType )
		
		if contentTypeNumber >= videoLow and contentTypeNumber <= videoHigh then
		
			return getVideoUploadBaseUrl()
			
		end
		
	end
	
	return getImageUploadBaseUrl()
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getMasterUploadRevision( docValue, localOnly )

	check( 'table', 'table' )
	
	return getMasterUploadRevision_imp( docValue, localOnly )
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getLastUploadRevision( docValue, localOnly )

	check( 'table', 'table' )
	
	return getLastUploadRevision_imp( docValue, localOnly )
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getMasterUploadHeaders( docValue, localOnly )

	local contentType = derefOpt( docValue, "importSource", "contentType" )
	local subtype = derefOpt( docValue, "subtype" )
	
	if not contentType then
		if subtype == "video" then
			contentType = "application/octet-stream;video/*"
		else
			contentType = "application/octet-stream"
		end
	end
	
	local isVideo = string.find( contentType, "video/")
	
	local getVideoHeaders = function()
		if isVideo then

			local duration = KSTableUtils.derefOpt( docValue, 'video', 'duration' )
			local needHls = duration and duration[ 1 ] / ( duration[ 2 ] or 1 ) < 600
				-- only request HLS for content >= 10 minutes
			
			return WFVideoTranscodingOptions.ozTranscodingHeaders( needHls )
			
		end
	end
	
	local getImageHeaders = function()
		if not isVideo then
			return { header = 'X-Generate-Proxies', value = 'all' }
		end
	end
	
	local masterPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = contentType
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
			getVideoHeaders() or getImageHeaders()
		},
		timeout = 60,
		returnHttpErrorsAsData = true,
		bypassRateLimiter = true,
	}
	
	table.insert( masterPutArgs.requestHeaders, { header = 'X-Generate-Renditions', value = 'all' } )

	return masterPutArgs
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getMasterUploadWithFilePathHeaders( contentType, subtype )

	if not contentType then
		if subtype == "video" then
			contentType = "application/octet-stream;video/*"
		else
			contentType = "application/octet-stream"
		end
	end
	
	local isVideo = string.find( contentType, "video/" )
	
	local getVideoHeaders = function()
		if isVideo then
		
			-- TODO: [ylo 2016-10-24] make sure we can pass false here
			return WFVideoTranscodingOptions.ozTranscodingHeaders( false )
			
		end
	end
	
	local getImageHeaders = function()
		if not isVideo then
			return { header = 'X-Generate-Proxies', value = 'all' }
		end
	end
	
	local masterPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = contentType
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
			getVideoHeaders() or getImageHeaders()
		},
		timeout = 60,
		returnHttpErrorsAsData = true,
		bypassRateLimiter = true,
	}
	
	table.insert( masterPutArgs.requestHeaders, { header = 'X-Generate-Renditions', value = 'all' } )

	return masterPutArgs
	
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getXmpUploadHeaders()
	local masterPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = 'application/rdf+xml'
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
			{ 
				header = 'X-Generate-Renditions', 
				value = 'all' 
			},
		},
		timeout = 60,
		returnHttpErrorsAsData = true,
		bypassRateLimiter = true,
	}

	return masterPutArgs
end

--------------------------------------------------------------------------------

function OzBinaryUploadUtils.getPresetProfileUploadHeaders()
	local masterPutArgs = {
		requestHeaders = {
			{
				header = 'Content-Type',
				value = 'application/rdf+xml'
			},
			{
				header = 'Accept',
				value = 'application/json'
			},
		},
		timeout = 60,
		returnHttpErrorsAsData = true,
		bypassRateLimiter = true,
	}

	return masterPutArgs
end

--------------------------------------------------------------------------------

return OzBinaryUploadUtils
