--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.rxlua
local WFRxPromiseBridgeWithResolver = require 'WFRxPromiseBridgeWithResolver'
local WFRxPromiseBridge = require 'WFRxPromiseBridge'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local KSTableUtils = require 'KSTableUtils'
local WFAsyncDigest = require 'WFAsyncDigest'
local WFAnalytics = require 'WFAnalytics'
local WFPlatformInfo = require 'WFPlatformInfo'

local LrmImageDataConversion = require 'LrmImageDataConversion'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.os
local KSDate = require 'KSDate'
local WFFileUtils = require 'WFFileUtils'
local KSPathUtils = require "KSPathUtils"
local WFDeviceInfo = require 'WFDeviceInfo'

	-- WF.oz_client.index_store
local WFQueryDrivenWorkQueueScheduler = require 'WFQueryDrivenWorkQueueScheduler'

	-- WF.oz_client.sync_agent
local OzCentralChunkedUpload = require "OzCentralChunkedUpload"
local OzBinaryUploadUtils = require 'OzBinaryUploadUtils'
local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'
local OzBinaryUploadTaskManager = require 'OzBinaryUploadTaskManager'
local WFOzSyncConstants = require 'WFOzSyncConstants'

	-- WF.oz_client.catalog
local OzFilePaths = require 'OzFilePaths'
local OzCatHelperFns = require 'OzCatHelperFns'

	-- WF.imaging.camera_roll
local WFCameraRollService = require 'WFCameraRollService'

	-- WF.web.http_client
local WFRxHttpClient = require 'WFRxHttpClient'
local KSJsonDecoder = require "KSJsonDecoder"

	-- WF.core.performance
local WFPerformanceReport = require 'WFPerformanceReport'

--------------------------------------------------------------------------------

local binaryUploaderLog = WFLogger 'OzBinaryUploader'
--	:setLevel( 'DEBUG' )
	
local now = assert( KSDate.timeNowForAnimation )

local derefOpt = assert( KSTableUtils.derefOpt )

local asKSPromiseWithResolver = WFRxPromiseBridgeWithResolver

local getMasterUploadRevision = assert( OzBinaryUploadUtils.getMasterUploadRevision )

local getMasterUploadHeaders = assert( OzBinaryUploadUtils.getMasterUploadHeaders )

local getMasterUploadWithFilePathHeaders = assert( OzBinaryUploadUtils.getMasterUploadWithFilePathHeaders )

local getXmpUploadHeaders = assert( OzBinaryUploadUtils.getXmpUploadHeaders )

local getPresetProfileUploadHeaders = assert( OzBinaryUploadUtils.getPresetProfileUploadHeaders )

--------------------------------------------------------------------------------

local MIN_SIZE_FOR_CHUNKING = OzCentralChunkedUpload.MIN_SIZE_FOR_CHUNKING

--==============================================================================

local function addBinaryStatsToDiagnosticLog( self, bucket, time, size )
	
	local sizeInKB = size / 1000
	
	local bucketStats = self._uploadSpeedStats.buckets[ bucket ]
	local uploadSpeed = sizeInKB / time
	
	if not bucketStats then

		bucketStats = {
			count = 0,
			timings = WFDiagnosticLog.makeRingBuffer { "Time (sec)" },
			sizes = WFDiagnosticLog.makeRingBuffer { "Size (KB)" },
			speed = WFDiagnosticLog.makeRingBuffer { "Upload speed (Kbps)" },
		}
									
		self._uploadSpeedStats.buckets[ bucket ] = bucketStats
	
	end
	
	bucketStats.count = bucketStats.count + 1
	bucketStats.timings:add( time )
	bucketStats.sizes:add( sizeInKB )
	bucketStats.speed:add( uploadSpeed )
	
	local totalBinarySize = self._uploadSpeedStats.totalSize or 0
	self._uploadSpeedStats.totalSize = totalBinarySize + sizeInKB
	
end

--------------------------------------------------------------------------------

local function trackHttpError( errorLog, docInfo, binaryType, masterUrl, chunkPart, uploadSource, headers, responseData, errorInfo )

	-- Special case for HTTP transit failure.
	
	local errorType = 'BinaryUpload.Error'
	if binaryType ~= 'master' then
		errorType = 'XmpExternalUpload.Error'
	end
	
	local eventSubcategory
	if chunkPart then
		eventSubcategory = 'binary_upload_chunk'
	else
		eventSubcategory = 'binary_upload_data'
	end
	
	local errorEvent = {}
			
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = eventSubcategory
	errorEvent['event.url'] = masterUrl
	errorEvent['event.error_type'] = errorType
	errorEvent['event.context_sequence'] = chunkPart
	
	errorEvent['content.id'] = docInfo.docId
	errorEvent['content.type'] = docInfo.subtype
	errorEvent['content.mimetype'] = docInfo.contentType
	errorEvent['content.size'] = docInfo.fileSize
	errorEvent['content.source'] = uploadSource
	
	if not headers then
	
		local errorCode = KSTableUtils.derefOpt( errorInfo, 'errorCode' ) or 'Unknown'
		local annotatedReason = string.format( "%s", tostring( responseData ) )
		
		errorEvent['event.subtype'] = 'OS'
		errorEvent['event.error_code'] = errorCode
		errorEvent['event.error_desc'] = annotatedReason
		
		local logFmtStr = "Unable to communicate with Oz URL %s\n\nreason = %s"
		binaryUploaderLog:warnf( logFmtStr, masterUrl, annotatedReason )
	
	else
	
		local ozErrorResponse = string.format( "%s",
									   tostring( responseData ) )
		
		local ozStatusCode = headers.status_code or 'nil'
		local ozRequestId = headers['X-Request-Id']

		errorEvent['event.subtype'] = 'API'
		errorEvent['event.error_code'] = ozStatusCode
		errorEvent['event.error_desc'] = ozErrorResponse
		errorEvent['event.context_guid'] = ozRequestId
		
		local logFmtStr =  "Got status %s back from Oz\n\n"
		local logParams = { headers.status_code or 'nil' }
	
		if masterUrl then
			logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = masterUrl
		end
	
		logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
		logParams[ #logParams + 1 ] = responseData
		
		local errorString = string.format( logFmtStr, unpack( logParams ) )
		errorLog[ #errorLog + 1 ] = {
				
			time = KSDate.timeInIso8601FormatWithTimeZone(),
			
			id = docInfo.docId,
			binaryType = 'master',
			error = errorString,
		
		}
			
		if #errorLog > 10 then
			table.remove( errorLog, 1 )
		end
	
		binaryUploaderLog:warnf( logFmtStr, unpack( logParams ) )
		
	end
	
	WFAnalytics.trackEvent( errorEvent )
	
end

--------------------------------------------------------------------------------

local function makeUploadTaskManagerKey( docId, binaryType )

	return docId .. '\t' .. binaryType

end

--------------------------------------------------------------------------------

local function makeAddToUploadTaskManagerFn( self )
	
	return function( uploadTask, uploadTaskManagerKey )
	
		if self._uploadTaskManagerItems[ uploadTaskManagerKey ] then
	
			self._uploadTaskManagerItems[ uploadTaskManagerKey ]:addCloseTarget( uploadTask )
				
		else
			binaryUploaderLog:warnf( "Attach to nil uploadTaskManager with doc: " .. uploadTaskManagerKey )
		end
	
	end
	
end

--------------------------------------------------------------------------------

local function makeRemoveUploadTaskManagerFn( self )

	return function( uploadTaskManagerKey )
	
		if self._uploadTaskManagerItems[ uploadTaskManagerKey ] then
	
			self._uploadTaskManagerItems[ uploadTaskManagerKey ]:close()
			self._uploadTaskManagerItems[ uploadTaskManagerKey ] = nil
		
		else
			binaryUploaderLog:warnf( "Remove from nil uploadTaskManager with doc: " .. uploadTaskManagerKey )
		end
	
	end

end

--------------------------------------------------------------------------------

local function makeUploadMasterResponse( self, docValue, masterPutUrl, startTime )

	return function( status, lastChunk )
	
		if lastChunk == 'lastChunk' then
			
			local _, cutoff = masterPutUrl:find("master")
			local partialMasterUrl = masterPutUrl:sub(1, cutoff)
			
			local endTime = KSDate.timeInIso8601Format()
				
			WFAnalytics.trackEvent {
				['event.type'] = 'master_uploaded',
				['event.workflow'] = 'SYNC',
				['event.subcategory'] = 'binary_upload_success',
				['content.id'] = docValue._id,
				['content.type'] = derefOpt( docValue, 'subtype' ),
				['content.mimetype'] = derefOpt( docValue, 'importSource', 'contentType' ),
				['content.size'] = derefOpt( docValue, 'importSource', 'fileSize' ),
				['content.url'] = masterPutUrl,
			}
			
			local docStatusOverlay = {
				_localOnly = {
					original = {
						ozUrl = partialMasterUrl,
						doNotBackUpFlagSet = true,
						uploaded = endTime,
					},
					chunkedUpload = {
						_delete = true,
					},
				},
			}
			
			local subtype = derefOpt( docValue, 'subtype' )
			local isVideo = docValue.video ~= nil or subtype == 'video'
			
			if self._deleteOriginalAfterUpload or isVideo then
			
				local originalFilePath = derefOpt( docValue, '_localOnly', 'originalFilePath' )
				
				originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )
		
				if originalFilePath and WFFileUtils.exists( originalFilePath ) then
				
					pcall( WFFileUtils.delete, originalFilePath )
					
				end
				
				docStatusOverlay._localOnly.originalFilePath = { _delete = true }
				
			end
	
			local resultP = self._docStore:applyOverlayToDocumentP {
				docType = docValue._docType,
				docId = docValue._docId,
				overlay = docStatusOverlay,
			}
			
			if resultP then
				return resultP:mapDataP( function()
					return 'data', 'removeUploadTaskManager'
				end )
			else
				return 'data', 'removeUploadTaskManager'
			end
		
		elseif status == 'error' then
		
			return 'data', 'removeUploadTaskManager'
		
		end
		
		return 'data', true
			
	end
	
end

--------------------------------------------------------------------------------

local function makeUploadMasterWithFilePathResponse( self, assetId, filePath, masterPutUrl, subtype, startTime )

	return function( status, lastChunk )
	
		if lastChunk == 'lastChunk' then
			
			local partialMasterUrl = masterPutUrl:gsub( "^.*//([^/]+", "" )
			
			local endTime = KSDate.timeInIso8601Format()

			WFAnalytics.trackEvent {
				['event.type'] = 'master_uploaded',
				['event.workflow'] = 'SYNC',
				['event.subcategory'] = 'binary_upload_success',
				['content.id'] = assetId,
				['content.type'] = subtype,
				['content.url'] = masterPutUrl,
			}
					
			local docStatusOverlay = {
				_localOnly = {
					original = {
						ozUrl = partialMasterUrl,
						doNotBackUpFlagSet = true,
						uploaded = endTime,
					},
					chunkedUpload = {
						_delete = true,
					},
				},
			}
			
			local isVideo = subtype == 'video'
			if self._deleteOriginalAfterUpload or isVideo then
			
				local originalFilePath = filePath
				
				originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )
		
				if originalFilePath and WFFileUtils.exists( originalFilePath ) then
				
					pcall( WFFileUtils.delete, originalFilePath )
					
				end
				
				docStatusOverlay._localOnly.originalFilePath = { _delete = true }
				
			end
	
			local resultP = self._docStore:applyOverlayToDocumentP {
				docType = 'asset',
				docId = assetId,
				overlay = docStatusOverlay,
			}
			
			if resultP then
				return resultP:mapDataP( function()
					return 'data', 'removeUploadTaskManager'
				end )
			else
				return 'data', 'removeUploadTaskManager'
			end
		
		elseif status == 'error' then
		
			return 'data', 'removeUploadTaskManager'
		
		end
		
		return 'data', true
			
	end
	
end

--------------------------------------------------------------------------------

local function makeUploadXmpResponse( self, docId, xmpPutUrl, startTime )

	return function( status, lastChunk )
	
		if lastChunk == 'lastChunk' then

			WFAnalytics.trackEvent {
				['event.type'] = 'xmp_uploaded',
				['event.workflow'] = 'SYNC',
				['event.subcategory'] = 'binary_upload_success',
				['content.id'] = docId,
				['content.type'] = 'xmp_develop',
				['content.url'] = xmpPutUrl,
			}
			
			local docStatusOverlay = {
				_localOnly = {
					files = {
						xmp_develop = {
							uploadStatus = "success"
						}
					}
				},
				_serverProvidedMetadata = {
					links = {
						[ '/rels/xmp/develop' ] = {
							invalid = { _delete = true }
						}
					}
				}
			}
	
			local resultP =  self._docStore:applyOverlayToDocumentP {
				docType = "asset",
				docId = docId,
				overlay = docStatusOverlay,
			}
	
			return resultP
		
		end
		
		return 'data', true
			
	end
	
end

--------------------------------------------------------------------------------

local function makeUploadPresetProfileResponse( self, docId, presetProfilePutUrl, startTime )

	return function( status, lastChunk )

		if lastChunk == 'lastChunk' then

			WFAnalytics.trackEvent {
				['event.type'] = 'presetProfile_uploaded',
				['event.workflow'] = 'SYNC',
				['event.subcategory'] = 'binary_upload_success',
				['content.id'] = docId,
				['content.type'] = 'application/rdf+xml',
				['content.url'] = presetProfilePutUrl,
			}
			local presetProfile = 'presetsProfile'

			if WFPlatformInfo.isAndroid then
				presetProfile = 'presetProfile'
			end

			local docStatusOverlay = {
				_localOnly = {
					files = {
						[ presetProfile ] = {
							uploadStatus = "success"
						}
					}
				},
				_serverProvidedMetadata = {
					links = {
						[ '/rels/master' ] = {
							invalid = { _delete = true }
						}
					}
				}

			}

			local resultP =  self._docStore:applyOverlayToDocumentP {
				docType = "asset",
				docId = docId,
				overlay = docStatusOverlay,
			}

			return resultP

		end

		return 'data', true

	end

end

--------------------------------------------------------------------------------

local function uploadDataP( self, args )

	local uploadP
	local observablePromiseResolver
	
	local addToUploadTaskManagerFn = makeAddToUploadTaskManagerFn( self )
	local removeUploadTaskManagerFn = makeRemoveUploadTaskManagerFn( self )
	
	local docFileSize = assert( args.docFileSize )
	local url = assert( args.url )
	local filePathOrData = assert( args.filePathOrData )
	local docId = assert( args.docId )
	local docValue = args.docValue
	local putArgs = args.putArgs or {}
	local binaryType = assert( args.binaryType )
	local chunkType = args.chunkType or 'chunk'
	local chunkPart = args.chunkPart or '-1'
	local chunkedUpload = args.chunkedUpload
	local contentType = args.contentType
	
	local uploadTaskManagerKey = makeUploadTaskManagerKey( docId, binaryType )
	
	local startTime = now()
	
	local size
	local uploadSource
	if KSData.isData( filePathOrData ) then
		size = KSData.length( filePathOrData )
		uploadSource = "data"
	else
		size = 0
		if type( filePathOrData ) == 'number' then
			uploadSource = "descriptor"
		else
			uploadSource = "file"
		end
	end

	if args.docFileSize >= 2 * MIN_SIZE_FOR_CHUNKING and chunkedUpload then

		binaryUploaderLog:debugf( "Chunk upload of %s for asset: %s", binaryType, docId )
		
		local context = {
			docStore = self._docStore,
			docType = 'asset',
			docId = docId,
			ozUser = self._ozUser,
		}
		
		local errorLog = self._uploadErrorLog
		
		context.maybeAddToErrorLog = function( status, headers, responseData, errorInfo )
		
			local statusCode = type( headers ) == 'table' and tostring( headers.status_code )
			
			if statusCode then
			
				local statusCodeKey = statusCode:sub( 1, -3 )
				
				if status == 'data' and statusCodeKey == '2' then
					return
				end
				
			end
			
			if status == 'error' and responseData == 'cancelled' then
				return
			end
			
			local docInfo = {
				docId = docId,
				subtype = derefOpt( docValue, 'subtype' ),
				contentType = contentType,
				fileSize = docFileSize,
			}

			trackHttpError( errorLog, docInfo, binaryType, url, chunkPart, uploadSource, headers, responseData, errorInfo  )
			
		end
		
		uploadP = OzCentralChunkedUpload.doChunkedUploadP( context, docValue, 
															addToUploadTaskManagerFn, removeUploadTaskManagerFn, 
															binaryType, chunkType, chunkPart, url, putArgs, filePathOrData, docFileSize )
		
			:mapP( function( ... )
			
				-- Release the data object ASAP.
			
				if KSData.isData( filePathOrData ) then
					filePathOrData:close()
				end
				
				local elapsedTime = now() - startTime
				
				addBinaryStatsToDiagnosticLog( self, 'chunk', elapsedTime, MIN_SIZE_FOR_CHUNKING )
				
				return ...
				
			end )

	end
	
	if not uploadP then
		
		binaryUploaderLog:debugf( "Direct upload for asset: %s", docId )
		
		if not KSData.isData( filePathOrData ) then
		
			putArgs.dataFormat = 'filePath'
			putArgs.length = docFileSize
			
		end
		
		putArgs.timeout = 240
		
		uploadP, observablePromiseResolver = WFPerformanceReport.wrapObservable( 'OzBinaryUploader non-chunked',
				self._ozUser:putBinary( url, filePathOrData, putArgs ),
				{
					url = url,
				},
				WFRxHttpClient.extractHttpTimingForPerformanceReport
			)
			:call( asKSPromiseWithResolver )
		
		addToUploadTaskManagerFn( observablePromiseResolver, uploadTaskManagerKey )
		
		
		uploadP = uploadP:mapP( function( status, responseData, headers, errorInfo, ... )
		
				-- This upload has been removed from uploadTaskManager
				if not status and not headers and not responseData and not errorInfo then
					return 'error', 'cancelled request'
				end
			
				local statusCode = type( headers ) == 'table' and headers.status_code

				binaryUploaderLog:debugf( "Direct upload completed %s", statusCode )
				
				-- Release the data object ASAP.
			
				if KSData.isData( filePathOrData ) then
					filePathOrData:close()
				end
				
				-- Calculate elapsed time.
				
				local elapsedTime = now() - startTime
				local roundedElapsedTime = math.floor( elapsedTime + .5 )
			
				local ozRequestId = derefOpt( headers, 'X-Request-Id' )
				
				-- Only a 201 response is acceptable.
			
				if status == 'data' and statusCode == '201' then
	
					local perfEvent = {}
					perfEvent['event.type'] = 'performance'
					perfEvent['event.workflow'] = 'SYNC'
					perfEvent['event.subcategory'] = 'binary_upload_data'
					perfEvent['event.subtype'] = binaryType
					perfEvent['event.url'] = url
					perfEvent['content.id'] = docId
					perfEvent['content.type'] = derefOpt( docValue, 'subtype' )
					perfEvent['content.mimetype'] = contentType
					perfEvent['content.size'] = docFileSize
					perfEvent['content.source'] = uploadSource
					perfEvent['event.value'] = elapsedTime
					perfEvent['event.context_guid'] = ozRequestId
					
					WFAnalytics.trackEvent( perfEvent )
					
					addBinaryStatsToDiagnosticLog( self, binaryType, elapsedTime, size )
	
					return status, responseData, headers, errorInfo, ...
	
				end
				
				if status == 'error' and responseData == 'cancelled' then
					
					return status, responseData
					
				end
				
				local docInfo = {
					docId = docId,
					subtype = derefOpt( docValue, 'subtype' ),
					contentType = contentType,
					fileSize = docFileSize,
				}
				
				trackHttpError( self._uploadErrorLog, docInfo, binaryType, url, nil, uploadSource, headers, responseData, errorInfo  )
				
				-- Special case for HTTP transit failure.
				
				if statusCode == '403' then
				
					local responseJson = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( responseData )
					
					local responseSubtype = KSTableUtils.derefOpt( responseJson, 'subtype' )
					
					if responseSubtype == 'UploadMasterWhenNotAllowedError' then
					
						return self._docStore:applyOverlayP( {
							_id = docId,
							importSource = {
								willNotUploadMaster = { _delete = true },
							}
						} )
						:mapP( function()
						
							return 'error', 'removed importSource.willNotUploadMaster'
							
						end )
						
					end

				elseif statusCode == '415' and string.lower(contentType) == 'image/heic' then
				
					local conversionStatus, jpegFileInfo = LrmImageDataConversion.convertHEICImageFileToJPEGImageFile( filePathOrData )
					
					if conversionStatus then
					
						return self._docStore
							:applyOverlayToDocumentP {
								docType = 'asset',
								docId = docId,
								overlay = {
									_localOnly = {
										original = {
											size = jpegFileInfo.fileSize,
										},
										originalFilePath = jpegFileInfo.filePath, 
									},
									importSource = {
										contentType = 'image/jpeg',
										sha256 = jpegFileInfo.sha256,
										fileName = jpegFileInfo.fileName,
										fileSize = jpegFileInfo.fileSize,
									}
								}
							}
							:mapP( function()
								return 'error', 'wrong mimeType'
							end )
					else
						
						local errorType = 'BinaryUpload.Error'
						if binaryType ~= 'master' then
							errorType = 'XmpExternalUpload.Error'
						end
						
						local conversionErrorEvent = {}
						
						conversionErrorEvent['event.type'] = 'error'
						conversionErrorEvent['event.workflow'] = 'SYNC'
						conversionErrorEvent['event.subcategory'] = 'heic_conversion_fail'
						conversionErrorEvent['event.url'] = url
						conversionErrorEvent['event.error_type'] = errorType
						
						conversionErrorEvent['content.id'] = docId
						conversionErrorEvent['content.type'] = derefOpt( docValue, 'subtype' )
						conversionErrorEvent['content.mimetype'] = contentType
						conversionErrorEvent['content.size'] = docFileSize
						conversionErrorEvent['content.source'] = uploadSource
						
						WFAnalytics.trackEvent( conversionErrorEvent )
					end
					
				end

				if not headers then
				
					local annotatedReason = string.format( "%s (%d seconds)",
														   tostring( responseData ),
														   roundedElapsedTime )
				
					return 'error', annotatedReason
				
				end
				
				-- We were able to communicate with Oz, but Oz didn't like the response.

				return 'error', 'bad'
				
				-- If we got a 2xx, 3xx, or 4xx status code, we probably won't
				-- succeed if we try again with the same data. Flag this asset
				-- document so that we don't retry without explicitly clearing
				-- the error flag.
				
			end )
			
			:mapDataP( function()
				return 'data', 'lastChunk'
			end )
		
	end
	
	return uploadP
	
end

--------------------------------------------------------------------------------

local function uploadMasterP( self, docValue, binaryType, chunkType, chunkPart, url )

--	binaryUploaderLog:debugf( "uploadMasterP docValue = %s", docValue )
	
	-- INFO [pkamin 2016-02-12]: in case we hit the race condition where the doc 
	-- got deleted after we marked it for upload in the index store
	-- we check the doc for the deletion attribute
	
	if docValue._deleted then
		return KSPromises.resolvedPromiseP( 'data', true )
	end
	
	local localOnly = docValue._localOnly or {}
	
	local whichRevision = getMasterUploadRevision( docValue, localOnly )
			
	local docId = docValue._docId
	
	binaryUploaderLog:debugf( "uploadMasterP docId = %s for revId = %s", docId, whichRevision )
	
	if not url then
		url = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master_create', 'href' )
		url = url:replace( '{revision_id}', whichRevision )
	end
	
	local startTime = now()
	
	local originalFilePath = localOnly.originalFilePath
	
	originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )
	
	local args = {
		url =  url,
		putArgs = getMasterUploadHeaders( docValue, localOnly ),
		docValue = docValue,
		docId = docValue._docId,
		binaryType = binaryType,
		chunkType = chunkType,
		docFileSize = derefOpt( docValue, 'importSource', 'fileSize' ),
		filePathOrData = originalFilePath,
		chunkPart = chunkPart,
		chunkedUpload = true,
		contentType = derefOpt( docValue, 'importSource', 'contentType' ),
	}

	return uploadDataP( self, args )
		:mapP( makeUploadMasterResponse( self, docValue, url, startTime ) )
		
end

--------------------------------------------------------------------------------

local function uploadMasterWithFilePathP( self, docId, binaryType, chunkType, chunkPart, filePath, url, fileSize, contentType, subtype )

	binaryUploaderLog:debugf( "uploadWithFilePathP docId = %s for masterFilePath = %s", docId, filePath )	
	
	local startTime = now()
	
	local putArgs = getMasterUploadWithFilePathHeaders( contentType, subtype ) or {}
	
	local args = {
		url =  url,
		putArgs = putArgs,
		docId = docId,
		docFileSize = fileSize,
		filePathOrData = filePath,
		binaryType = binaryType,
		chunkType = chunkType,
		chunkPart = chunkPart,
		chunkedUpload = true,
		contentType = contentType,
	}
	
	return uploadDataP( self, args )
		:mapP( makeUploadMasterWithFilePathResponse( self, docId, filePath, url, subtype, startTime ) )
		
end

--------------------------------------------------------------------------------

local function uploadXmpDevelopP( self, docId, relFilePath, xmpDevelopUrl )

--	binaryUploaderLog:debugf( "uploadXmpDevelopP docValue = %s", docValue )
	
	-- INFO [pkamin 2016-02-12]: in case we hit the race condition where the doc 
	-- got deleted after we marked it for upload in the index store
	-- we check the doc for the deletion attribute
	
	binaryUploaderLog:debugf( "uploadXmpDevelopP docId = %s for xmpDevelopUrl = %s", docId, xmpDevelopUrl )
	
	-- [ylo 2016-10-26]: no xmpDevelopUrl should not happen
	if not xmpDevelopUrl then
	
		return KSPromises.resolvedPromiseP( 'error', 'Xmp develop has no upload URL' )
	
	end
	
	local startTime = now()
	
	-- MEH [immisch 2016-09-27]: this needs to be in sync with LrmOzDevelopSession 	
	local absPath = KSPathUtils.child( KSPathUtils.getStandardFilePath( 'carouselDocuments' ), relFilePath )

	local fileAttributes = WFFileUtils.fileAttributes( absPath )
	
	if not relFilePath or not fileAttributes or not fileAttributes.fileSize or fileAttributes.fileSize == 0 then
	
		-- TODO [immisch 2016-09-20]: what to do here?
		-- This is is an annoying error, but not fatal - the settings got lost.
		
		-- We might mark the asset and display a status for the user to make transparent what happened

		WFAnalytics.trackEvent {
			['event.type'] = 'error',
			['event.workflow'] = 'SYNC',
			['event.error_type'] = 'XmpExternalUpload.AssetNotFound',
			['content.id'] = docId,
			['content.url'] = xmpDevelopUrl,
			['content.type'] = 'xmp_develop',
			['content.source'] = relFilePath,
		}
					
		binaryUploaderLog:warnf( 'Xmp not found with assetId %s', docId )

		return self._docStore:applyOverlayP( {
				_id = docId,
				_localOnly = {
					files = {
						xmp_develop = {
							uploadStatus = 'errorFileMissing'
						}
					}
				} 
			} )
			:mapP(function( status, data )
				return 'error', 'Xmp develop was lost' 
			end )

	end
	
	local putArgs = getXmpUploadHeaders()

	local args = {
		url =  xmpDevelopUrl,
		putArgs = putArgs,
		docId = docId,
		docFileSize = fileAttributes.fileSize,
		filePathOrData = absPath,
		binaryType = 'xmp_develop'
	}
		
	return uploadDataP( self, args )
		:mapP( makeUploadXmpResponse( self, docId, xmpDevelopUrl, startTime ) )
		
end

--------------------------------------------------------------------------------


local function uploadPresetProfileP( self, docId, relFilePath, presetProfileUrl )

	binaryUploaderLog:debugf( "uploadPresetProfileP docId = %s for presetProfilesUrl = %s", docId, presetProfileUrl )

	if not presetProfileUrl then

		return KSPromises.resolvedPromiseP( 'error', 'presetProfiles has no upload URL' )

	end

	local startTime = now()

	local catalogPath = KSPathUtils.child( KSPathUtils.getStandardFilePath( 'carouselDocuments' ), self._catalogId )

	local absPath = KSPathUtils.child( catalogPath, relFilePath )

	local fileAttributes = WFFileUtils.fileAttributes( absPath )

	if not relFilePath or not fileAttributes or not fileAttributes.fileSize or fileAttributes.fileSize == 0 then


		WFAnalytics.trackEvent {
			['event.type'] = 'error',
			['event.workflow'] = 'SYNC',
			['event.error_type'] = 'PresetProfileUpload.AssetNotFound',
			['content.id'] = docId,
			['content.url'] = presetProfileUrl,
			['content.type'] = 'application/rdf+xml',
			['content.source'] = relFilePath,
		}

		binaryUploaderLog:warnf( 'PresetProfile not found with assetId %s', docId )

		return KSPromises.resolvedPromiseP( 'error', 'Preset/profile  was lost' )

	end

	local putArgs = getPresetProfileUploadHeaders()

	local args = {
		url =  presetProfileUrl,
		putArgs = putArgs,
		docId = docId,
		docFileSize = fileAttributes.fileSize,
		filePathOrData = absPath,
		binaryType = 'presetProfile'
	}

	return uploadDataP( self, args )
		:mapP( makeUploadPresetProfileResponse( self, docId, presetProfileUrl, startTime ) )

end

--------------------------------------------------------------------------------

local function deleteAssetP( self, docValue )

	return self._docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docId = docValue._docId,
		overlay = {
			_deleted = true
		}
	}
	
end

--------------------------------------------------------------------------------

local function _uploadMasterP( self, docValue, binaryType, chunkType, chunkPart, url )

    local function markAssetAsDeletedIfNotFound(self)

        binaryUploaderLog:warnf( 'Asset not found with assetId %s', docValue._docId )

        local errorEvent = {}

        errorEvent['event.type'] = 'error'
        errorEvent['event.workflow'] = 'SYNC'
        errorEvent['event.subcategory'] = 'binary_upload_missing_original'
        errorEvent['event.url'] = url
        errorEvent['event.error_type'] = 'BinaryUpload.Error'

        errorEvent['content.id'] = docValue._docId
        errorEvent['content.type'] = derefOpt( docValue, 'subtype' )
        errorEvent['content.mimetype'] = derefOpt( docValue, 'importSource', 'contentType' )
        errorEvent['content.size'] = derefOpt( docValue, 'importSource', 'fileSize' )

        WFAnalytics.trackEvent( errorEvent )

        return deleteAssetP( self, docValue )
        :mapP( function()

            local uploadTaskManagerKey = makeUploadTaskManagerKey( docValue._docId, binaryType )

            self._uploadTaskManagerItems[ uploadTaskManagerKey ]:close()
            self._uploadTaskManagerItems[ uploadTaskManagerKey ] = nil

        end )
        :runUntilDoneP()

    end

	if docValue._deleted then
		return KSPromises.resolvedPromiseP( 'data', true )
	end
	
	local originalFilePath = derefOpt( docValue, '_localOnly', 'originalFilePath' )
	originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( originalFilePath )

	if not originalFilePath then

		local localUrl = derefOpt( docValue, '_localOnly', 'localUrl' )
		local localAssetId = derefOpt( docValue, 'importSource', 'localAssetId' )
		local uniqueDeviceId = derefOpt( docValue, 'importSource', 'uniqueDeviceId' )

		if localAssetId and ( uniqueDeviceId == WFDeviceInfo.uniqueDeviceId() ) then
			localUrl = localAssetId
		end
		
		if localUrl == nil then
			return KSPromises.resolvedPromiseP( 'data', true )
		end

		if WFPlatformInfo.isAndroid then

			if not WFFileUtils.exists( localUrl ) then -- marking asset as deleted as file locally was deleted
				return markAssetAsDeletedIfNotFound( self )
			end

			local sha256 = docValue.importSource and docValue.importSource.sha256
			sha256 = sha256 and string.upper(sha256)

			local WFDigest = require 'WFDigest'
            local sha256Digest = WFDigest.digestFile(localUrl, 'SHA256')
			local expectedSHA256 = sha256Digest and string.upper(sha256Digest)

			if expectedSHA256 and sha256 ~= expectedSHA256 then -- marking asset as deleted as file locally was modified
				return markAssetAsDeletedIfNotFound( self )
			end

			return self._docStore:applyOverlayP( {
				_id = docValue._docId,
				_localOnly = {
					originalFilePath = localUrl,
					chunkedUpload = { _delete = true },
				}
			} )
		end

		return WFCameraRollService.getAssetData( localUrl )
			:call( WFRxPromiseBridge )
			:mapP( function( status, assetInfo )
			
				if status == 'data' and KSData.isData( assetInfo.data ) then
					
					local documentsPath = OzFilePaths.getStandardOriginalFolderDocumentsPath( self._catalogId )
					local folder = OzFilePaths.originalsFolderForDate( documentsPath, docValue.captureDate )
		
					WFFileUtils.mkdir( folder )

					local leafName = OzCatHelperFns.importFileNameFromPhotoInfo( docValue )
					local destPath = KSPathUtils.child( folder, leafName )
					
					local originalData = assetInfo.data
					local originalSize = KSData.length( originalData )

					local saveDataFn = function()
						destPath = KSData.saveToUniquelyNamedFile( originalData, destPath )
					end
							
					local success, result = pcall( saveDataFn )
					
					
					if not success then
						return 'error', result
					end
					
					return WFAsyncDigest.digestFile( destPath, 'SHA256' )
						:call( WFRxPromiseBridge )
						:mapDataP_new( function( localSha256 )
						
							return self._docStore:applyOverlayP( {
								_id = docValue._docId,
								_localOnly = {
									originalFilePath = destPath,
									chunkedUpload = { _delete = true },
								},
								importSource = {
									fileSize = originalSize,
									sha256 = localSha256
								}
							} )
						
						end )
					
				end
				
				return 'error', assetInfo
				
			end )
			
			:mapP( function( status, info, ... ) 
				
				if( status == 'error' ) then
				
					binaryUploaderLog:warnf( 'Asset cannot be saved into originalFilePath with assetId %s', docValue._docId )
					
					local errorEvent = {}
				
					errorEvent['event.type'] = 'error'
					errorEvent['event.workflow'] = 'SYNC'
					errorEvent['event.subcategory'] = 'binary_upload_fail_writing_original'
					errorEvent['event.url'] = url
					errorEvent['event.error_type'] = 'BinaryUpload.Error'
					errorEvent['event.error_desc'] = info
					
					errorEvent['content.id'] = docValue._docId
					errorEvent['content.type'] = derefOpt( docValue, 'subtype' )
					errorEvent['content.mimetype'] = derefOpt( docValue, 'importSource', 'contentType' )
					errorEvent['content.size'] = derefOpt( docValue, 'importSource', 'fileSize' )
					
					WFAnalytics.trackEvent( errorEvent )
			
					return deleteAssetP( self, docValue )
						:mapP( function()
						
							local uploadTaskManagerKey = makeUploadTaskManagerKey( docValue._docId, binaryType )
						
							self._uploadTaskManagerItems[ uploadTaskManagerKey ]:close()
							self._uploadTaskManagerItems[ uploadTaskManagerKey ] = nil
							
						end )
						:runUntilDoneP()
				
				end
				
				return status, info, ...
			end )

	elseif not originalFilePath or not WFFileUtils.exists( originalFilePath ) then

		return markAssetAsDeletedIfNotFound(self)

	end
	
	
	-- do sanity check if fileSize is less or equal than zero
	-- if not try to recover by either getting size from file system or CameraRoll
	
	local docFileSize = derefOpt( docValue, 'importSource', 'fileSize' )
	local fileAttrs = WFFileUtils.fileAttributes( originalFilePath )
	local originalFileSize = fileAttrs and fileAttrs.fileSize
	
	if docFileSize == nil or docFileSize <= 0 or docFileSize ~= originalFileSize then
				
		if originalFileSize then
		
			if originalFileSize > 0 then
		
				return self._docStore:applyOverlayToDocumentP {
					docType = 'asset',
					docId = docValue._docId,
					overlay = {
						importSource = {
							fileSize = originalFileSize,
						}
					}
				}
				:runUntilDoneP()
			
			else
			
				return deleteAssetP( self, docValue )
					:mapP( function()
		
						local uploadTaskManagerKey = makeUploadTaskManagerKey( docValue._docId, binaryType )
					
						self._uploadTaskManagerItems[ uploadTaskManagerKey ]:close()
						self._uploadTaskManagerItems[ uploadTaskManagerKey ] = nil
	
					end )
					:runUntilDoneP()
			end
			
		end
		
	end
	
	local ozUrl = derefOpt( docValue, '_localOnly', 'original', 'ozUrl' )
	local ozMultipartCompleteError = derefOpt( docValue, '_localOnly', 'ozError', 'subtype' ) == 'MultipartCompleteError'
	local ozUrlUploadedTime = derefOpt( docValue, '_localOnly', 'original', 'uploaded' )
	
	-- need to recover from failed chunked uploads
	if ( ozUrl and not ozUrlUploadedTime ) or ozMultipartCompleteError then
	
		local ozUpdatedDeltaInDays = 0
		local ozUpdatedTime = derefOpt( docValue, '_serverProvidedMetadata', 'updated' )
		
		if not ozMultipartCompleteError and ozUpdatedTime then
			local currentTime = KSDate.timeInIso8601Format()
			ozUpdatedDeltaInDays = KSDate.differenceInDays( ozUpdatedTime, currentTime )
		end
		
		if ozMultipartCompleteError or ozUpdatedDeltaInDays >= 1 then
		
			local chunkedUploadRestartCount = derefOpt( docValue, '_localOnly', 'chunkedUploadRestartCount' ) or 0
			chunkedUploadRestartCount = chunkedUploadRestartCount + 1
			
			local docOverlay = {
				docType = 'asset',
				docId = docValue._docId,
				overlay = {
					_localOnly = {
						original = {
							ozUrl = { _delete = true },
							uploaded = { _delete = true },
	--						doNotBackUpFlagSet = true,
						},
						chunkedUpload = {
							_delete = true,
						},
						chunkedUploadRestartCount = chunkedUploadRestartCount,
					},
					_serverProvidedMetadata = {
						links = {
							[ '/rels/master' ] = { _delete = true },
						},
					},
				}
			}
			
			if ozMultipartCompleteError then
				docOverlay.overlay._localOnly.ozError = { _delete = true }
			end
			
			
			local errorEvent = {}
			
			errorEvent['event.type'] = 'error'
			errorEvent['event.workflow'] = 'SYNC'
			errorEvent['event.subcategory'] = 'binary_upload_chunk_multipartError_restart'
			errorEvent['event.url'] = url
			errorEvent['event.error_type'] = 'BinaryUpload.Error'
			errorEvent['event.value'] = chunkedUploadRestartCount
			
			errorEvent['content.id'] = docValue._docId
			errorEvent['content.type'] = derefOpt( docValue, 'subtype' )
			errorEvent['content.mimetype'] = derefOpt( docValue, 'importSource', 'contentType' )
			errorEvent['content.size'] = docFileSize
			
			WFAnalytics.trackEvent( errorEvent )
			
			
			return self._docStore:applyOverlayToDocumentP( docOverlay )
				:runUntilDoneP()
		
		end
	
	end
	
	
	local docId = docValue._docId
	
	local masterUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
	local ozSha256Error = derefOpt( docValue, 'importSource', 'status', 'error', 'all', 'reason', 'sha256' ) == "does not match"
	local localSha256Ok = masterUrl and derefOpt( docValue, '_localOnly', 'binaryUploadErrors', '/rels/master', masterUrl )
	local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )

	local blockedBySha256Error = not not ( ozSha256Error and ( not localSha256Ok ) )
	
	-- INFO [pkamin 2017-10-02]: disable this code for now as it isn't implemented appropriately for all binary cases
	
	if false and blockedBySha256Error then
 
		-- Recalculate sha256 to verify it matches the sha256 in the importSource to
		-- verify we have the correct file to upload.
	
		local asyncDigest, _ =  WFAsyncDigest.digestFile( originalFilePath, 'SHA256' )
			:call( asKSPromiseWithResolver )
		
		return asyncDigest
			:mapDataP_new( function( localSha256 )
		
				if localSha256 == imageSha256 then
		
					-- Update the doc so the binary uploader will try to upload this file again.
		
					binaryUploaderLog:debugf( "re-uploading corrupted original for docId %s path %s", docId, originalFilePath )
							
					local currentMasterUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
			
					return self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = docValue._docId,
						overlay = { 
							_localOnly = {
							-- Add a local annotation for the sha error since we're going to
							-- otherwise remove all traces that it happened.  This is necessary
							-- to prevent us from accidentally deleting the local master because
							-- we think the master is uploaded and fine (NMB-4862)
								binaryUploadErrors = {
									[ '/rels/master' ] = {
										[ currentMasterUrl ]  = "corrupt upload",
									}
								}
							}
						}
					}
					
				else
		
					return self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = docValue._docId,
						overlay = { 
							_localOnly = {
								localOriginalError = 'InvalidOriginal',
							}
						}
					}
		
				end
				
			end )
		
	end
	
	-- doc value passed our sanity checks - start upload
	
	return uploadMasterP( self, docValue, binaryType, chunkType, chunkPart, url )
	
end

--------------------------------------------------------------------------------

local function makeUploadOneBinaryFunc( self )

	return function( uploadQueryRow )
		
		local binaryUploadTask
		local removeUploadTaskManagerFn = makeRemoveUploadTaskManagerFn( self )
		
		local docId = uploadQueryRow[ 1 ]
		local chunkPart = uploadQueryRow[ 2 ]
		local binaryType = uploadQueryRow[ 3 ]
		local chunkType = uploadQueryRow[ 4 ]
		local filePath = uploadQueryRow[ 5 ]
		local url = uploadQueryRow[ 6 ]
		
		local uploadTaskManagerKey = makeUploadTaskManagerKey( docId, binaryType )

		if not self._uploadTaskManagerItems[ uploadTaskManagerKey ] then
			self._uploadTaskManagerItems[ uploadTaskManagerKey ] = OzBinaryUploadTaskManager{ 
																		docId = docId,
																		binaryType = binaryType }
		end
		
		if not filePath then
			
			binaryUploadTask =  self._docStore:docValueOnceP( 'asset', docId )
				
				:mapDataP_new( function( docValue )
				
					return _uploadMasterP( self, docValue, binaryType, chunkType, chunkPart, url )
					
				end )
		else
		
			local fileSize = uploadQueryRow[ 7 ]
			local contentType = uploadQueryRow[ 8 ]
			local subtype = uploadQueryRow[ 9 ]
			
			if binaryType == 'xmp_develop' then

				binaryUploadTask =  uploadXmpDevelopP( self, docId, filePath, url )

			elseif binaryType == 'presetProfile' then

				binaryUploadTask =  uploadPresetProfileP( self, docId, filePath, url )

			else
			
				filePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( filePath )
				
				if not WFFileUtils.exists( filePath ) then
				
					binaryUploaderLog:warnf( 'Asset not found with assetId %s', docId )
					
					local errorEvent = {}
			
					errorEvent['event.type'] = 'error'
					errorEvent['event.workflow'] = 'SYNC'
					errorEvent['event.subcategory'] = 'binary_upload_missing_original'
					errorEvent['event.url'] = url
					errorEvent['event.error_type'] = 'BinaryUpload.Error'
					
					errorEvent['content.id'] = docId
					errorEvent['content.type'] = subtype
					errorEvent['content.mimetype'] = contentType
					errorEvent['content.size'] = fileSize
					
					WFAnalytics.trackEvent( errorEvent )
					
					local resultP = self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = docId,
						overlay = {
							_deleted = true
						}
					}
					
					return resultP
						:mapP( function()
							removeUploadTaskManagerFn( uploadTaskManagerKey )
							return KSPromises.resolvedPromiseP( 'error', 'Asset marked as lost' )
						end )
						
				end
				
				local fileAttrs = WFFileUtils.fileAttributes( filePath )
				local originalFileSize = fileAttrs and fileAttrs.fileSize
				
				if originalFileSize and fileSize ~= originalFileSize then
					
					if originalFileSize > 0 then
			
						return self._docStore:applyOverlayToDocumentP {
							docType = 'asset',
							docId = docId,
							overlay = {
								importSource = {
									fileSize = originalFileSize,
								}
							}
						}
						:runUntilDoneP()
					
					else
					
						return self._docStore
							:applyOverlayToDocumentP {
								docType = 'asset',
								docId = docId,
								overlay = {
									_deleted = true
								}
							}
							:mapP( function()
				
								local uploadTaskManagerKey = makeUploadTaskManagerKey( docId, binaryType )
							
								self._uploadTaskManagerItems[ uploadTaskManagerKey ]:close()
								self._uploadTaskManagerItems[ uploadTaskManagerKey ] = nil
		
							end )
							:runUntilDoneP()
					end
					
				end
			
				binaryUploadTask = uploadMasterWithFilePathP( self, docId, binaryType, chunkType, chunkPart, filePath, url, fileSize, contentType, subtype )
			
			end
			
		end
		
		return binaryUploadTask
			:mapP( function( status, removeUploadTaskManager )

				if( removeUploadTaskManager == 'removeUploadTaskManager' ) then
				
					return removeUploadTaskManagerFn( uploadTaskManagerKey )
				
				end
				
			end )
		
	end
	
end

--------------------------------------------------------------------------------

local ozBinaryUploaderMethods = {

	_docStore = false,
		-- (OzDocStore)
		
	_indexStore = false,
		-- (WFIndexStore)
		
	_ozUser = false,
		-- (OzUser)
	
	_catalogId = false,
		-- (string) catalog id
		
	_workScheduler = false,
		-- (WFQueryDrivenWorkQueueScheduler)
		
	_uploadOneBinaryFn = false,
		-- (function) upload a single binary object (used primarily for testing)
		
	_uploadTaskManagerItems = false,
		-- (table) attach every task to corresponding assetId uploadTaskManager
		
	_uploadStatusBins = false,
		-- (table) record of upload status/counts for each binary rendition type
			
	_uploadErrorLog = false,
		-- (table) record of recent upload errors for the diagnostic log
	
	_uploadSpeedStats = false,
		-- (table) record of recent upload speeds
		
	_uploadQueueOptions = false,
		-- options to control the upload queue behavior
		
	_deleteOriginalAfterUpload = false,
		-- (boolean) whether we delete the originals after the upload to Oz
		
	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:__init( args )

	self._docStore = assert( args.docStore )
	
	self._indexStore = assert( args.indexStore )
	
	self._ozUser = assert( args.ozUser )
	
	self._catalogId = assert( args.catalogId )
	
	self._uploadQueueOptions = args.uploadQueueOptions or {}
	
	self._deleteOriginalAfterUpload = args.deleteOriginalAfterUpload
	
	if self._deleteOriginalAfterUpload == nil then
		self._deleteOriginalAfterUpload = true
	end
	
	local uploadOneItemFn = makeUploadOneBinaryFunc( self )
	self._uploadOneBinaryFn = uploadOneItemFn

	-- Typically, we'll create our work scheduler right away,
	-- but in some test scenarios we want to control its creation.
	
	if not args.noScheduler then
		self:startUploading()
	end
	
	self._uploadStatusBins = {}
	self._uploadErrorLog = {}

	self._uploadSpeedStats = {
		buckets = {},
		totalSize = 0,
	}
	
	WFCloseSupport.mustClose( self )
	
	WFDiagnosticLog.addLogProvider( self )
	
end

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:startUploading( sequence, videoIncluded )

	if not self._uploadTaskManagerItems then
		self._uploadTaskManagerItems = {}
	end

	if not self._workScheduler then
	
		local systemStats = WFPlatformInfo.getSystemStats()
		
		if systemStats and systemStats.freeSystemSize and systemStats.freeSystemSize > 1024 * 1024 * 1024 then
			self._uploadQueueOptions.concurrentTasks = 100
		else
			self._uploadQueueOptions.concurrentTasks = 5
		end
	
		local sql = OzBinaryUploadUtils.getUploadableAssetsSQL( self._uploadQueueOptions.concurrentTasks )
		
		local queryB = self._indexStore:queryB{ 
			sql = sql,
			quiesceUntilUpToDate = true
		}

		local concurrentTasks = 100

		if WFPlatformInfo.isAndroid then
			concurrentTasks = 5
		end

		self:setOwned {
			_workScheduler = WFQueryDrivenWorkQueueScheduler( queryB,
															   self._uploadQueueOptions.concurrentTasks or concurrentTasks,
															   self._uploadQueueOptions.blackout or 40,
															   self._uploadOneBinaryFn )
		}
		
	end
	
end

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:stopUploading()

	if self._workScheduler then
	
		self._workScheduler:close()
		self._workScheduler = nil
		
	end
	
	if self._uploadTaskManagerItems then

		local removeUploadTaskManagerFn = makeRemoveUploadTaskManagerFn( self )
	
		for uploadTaskManagerKey, binaryUploadTaskManager in pairs( self._uploadTaskManagerItems ) do
	  		
			removeUploadTaskManagerFn( uploadTaskManagerKey )
			
		end
		
		self._uploadTaskManagerItems = nil
	
	end
	
end

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:uploadMasterP( docValue )

	local docId = docValue._docId
	local uploadTaskManagerKey = makeUploadTaskManagerKey( docId, 'master' )

	self._uploadTaskManagerItems = {}
	self._uploadTaskManagerItems[ uploadTaskManagerKey ] = OzBinaryUploadTaskManager{ 
																docId = docId,
																binaryType = 'master' }
	return _uploadMasterP( self, docValue, "master" )
	
end

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:uploadXmpDevelopP( docValue, relFilePath, xmpDevelopUrl )

	local docId = docValue
	local uploadTaskManagerKey = makeUploadTaskManagerKey( docId, 'xmp' )

	self._uploadTaskManagerItems = {}
	self._uploadTaskManagerItems[ uploadTaskManagerKey ] = OzBinaryUploadTaskManager{ 
																docId = docId,
																binaryType = 'xmp' }

	return uploadXmpDevelopP( self, docValue, relFilePath, xmpDevelopUrl )
	
end

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:uploadPresetProfileP( docValue, relFilePath, presetProfileUrl )

	local docId = docValue
	local uploadTaskManagerKey = makeUploadTaskManagerKey( docId, 'presetProfile' )

	self._uploadTaskManagerItems = {}
	self._uploadTaskManagerItems[ uploadTaskManagerKey ] = OzBinaryUploadTaskManager{
																docId = docId,
																binaryType = 'presetProfile' }

	return uploadPresetProfileP( self, docValue, relFilePath, presetProfileUrl )

end

-- @mark diagnostic log

--------------------------------------------------------------------------------

function ozBinaryUploaderMethods:sendDiagnosticLogData()

	local ozUser = self._ozUser
	local docStore = self._docStore
	local userId = ozUser:userId()
	
	local finalStats = {}
	
	for bucket, stats in pairs( self._uploadSpeedStats.buckets ) do
		finalStats[ bucket ] = {
			count = stats.count,
			timings = { stats.timings:visualize 'boxAndWhisker' },
			size = { stats.sizes:visualize 'boxAndWhisker' },
			speed = { stats.speed:visualize 'boxAndWhisker' },
		}
	end

	local isRunning = not ( self._workScheduler == false or self._workScheduler == nil )
	
	local assetDocValuesInUploadQueue = Rx.Observable.empty()
	
	if not WFOzSyncConstants.isGlobalDocumentId( userId ) then
	
		assetDocValuesInUploadQueue = self._indexStore:query( {
				sql = [[
						SELECT 
							docId
						FROM 
							missingBinariesOnOz 
				]],
				quiesceUntilUpToDate = true } )
			:flatMap( function( rows )
			
				local docIds = {}
				
				for i, row in ipairs( rows ) do
				
					docIds[ #docIds + 1 ] = row[ 1 ]
					
				end
				
				return docStore:bulkDocValue( docIds )
				
			end )
			
	end
			
	return {

		name = "Oz Binary Uploader",
		id = 'OzBinaryUploader',
		weight = 510,

		body = {
		
			ozUserId = userId,
			isRunning = isRunning,
			binaryRenditionStatus = self._uploadStatusBins,
			
			recentErrors = self._uploadErrorLog,
			
			uploadSpeedStats = finalStats,
			totalBinarySize = self._uploadSpeedStats.totalSize,
			assetIdsInUploadQueue = self._indexStore:queryP( 
				[[
					SELECT 
						importDate, readyToUpload, binaryType, uploadSequence, part, docId
					FROM 
						missingBinariesOnOz 
					ORDER BY
						importDate, part ASC
			]] )
			:mapDataP_new( function( rows )
			
				local result = {}
				
				for i, row in ipairs( rows ) do
				
					result[ #result + 1 ] = {
						importDate = row[ 1 ], 
						hasUploadRevision = row[ 2 ], 
						binaryType = row[ 3 ], 
						uploadSequence = row[ 4 ], 
						part = row[ 5 ],
						docId = row[ 6 ],
					}
					
				end
				
				return result
				
			end )
			:runUntilDoneP(),
			assetDocValuesInUploadQueue = assetDocValuesInUploadQueue,
		},
	
	}
	
end

--------------------------------------------------------------------------------

return Rx.Class( 'OzBinaryUploader', ozBinaryUploaderMethods )
