--[[----------------------------------------------------------------------------

OzCatAlbums.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local _ = require 'WFArguments_uuid'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'


--==============================================================================

-- @mark OzCatalogService class

local AlbumMethods = {}

local OzCatalogServiceArg = 'OzCatalogServiceImp'

--------------------------------------------------------------------------------

local logger = WFLogger 'OzCatalogService :: albums '
--  :setLevel 'debug'

--------------------------------------------------------------------------------
--[[!
	\brief Creates an album in this catalog.
	
	\param name (string) The name of the new album.
	
	\return <code>P< string ></code> A Promise object that resolves to the ID of the
		newly-created album when the operation is complete.
]]--

function AlbumMethods:createAlbumP( --[[string]] name )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	if not ( #name > 0 ) then
		error( "OzCatalogService:createAlbumP: The album name must be a non-empty string", 2 )
	end
	
	local resultP = self._indexStore:queryP( [[SELECT DISTINCT nameLC FROM albums WHERE nameLC == ?]], KSStringUtils.lower( name ) )
		:mapP( function( status, rows )
		
			local nameIsUnique = true
			if status == 'data' and rows[ 1 ] and rows[ 1 ][ 1 ] then
				nameIsUnique = false
			end
			
			if nameIsUnique then
			
				local albumId = KSIdGenerator.uuidInLowercaseHex()
				local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'album',
					docId = albumId,
					overlay = {
						name = name,
						_localOnly = {
							updateTimeStamp = updateTimeStamp,
						},
					},
				}
				
				return innerResultP:mapDataP( function( status,_ )
					return status, albumId
				end )
				
			else
			
				return 'error', 'NameExists'
				
			end
						
		end )

	
	resultP:runUntilDone()
	
	return resultP
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes an album from this catalog.
	
	\param albumId (string) The unique ID of the album to delete.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:deleteAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:deleteAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	-- PR #3312056: Unshare the album first, then delete the album, too.
	-- Regardless of success/failure of unsharing the album, we're still going
	-- to delete it, thus the use of :mapP instead of :mapDataP.
	
	return self:unshareAlbumP( albumId )
		:mapP( function()
	
			local resultP = self._docStore:applyOverlayToDocumentP {
				docType = 'album',
				docId = albumId,
				overlay = {
						_deleted = true,
					},
				}
								
			return resultP:mapDataP( function ()
				return 'data', true
			end )
		
		end )
		:runUntilDoneP()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Renames an album in this catalog.
	
	It is an error to rename an album to match the name of an existing album,
	regardless of case.
	
	\param albumId (string) The unique ID of the album to rename.
	\param name (string) The new name for the album.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:renameAlbumP( --[[string]] albumId,
											 --[[string]] name )

	WFArguments.check( OzCatalogServiceArg, 'string', 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:renameAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	if not ( #name > 0 ) then
		error( "OzCatalogService:renameAlbumP: The album name must be a non-empty string", 2 )
	end
	
	local resultP = self._indexStore:queryP( [[SELECT DISTINCT nameLC FROM albums WHERE id != ? AND nameLC == ? ]], albumId, KSStringUtils.lower( name ) )
		:mapP( function( status, rows )
		
			local nameIsUnique = true
			if status == 'data' and rows[ 1 ] and rows[ 1 ][ 1 ] then
				nameIsUnique = false
			end
			
			if nameIsUnique then
			
				local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'album',
					docId = albumId,
					overlay = {
						name = name,
						_localOnly = {
							updateTimeStamp = updateTimeStamp,
						},
					},
				}
				
				return innerResultP
				
			else
				return 'error', 'NameExists', name
			end
						
		end )
						
	resultP:runUntilDone()

	return resultP:mapDataP( function ()
		return 'data', true
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Updates the timestamp for an album in this catalog.
	
	\param albumId (string) The unique ID of the album to touch.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:touchAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	if not ( #albumId > 0 ) then
		error( "OzCatalogService:touchAlbumP: The albumId must be a non-empty string", 2 )
	end
	
	local updateTimeStamp = KSDate.timeInIso8601Format( nil, false )

	local resultP = self._docStore:applyOverlayToDocumentP {
		docType = 'album',
		docId = albumId,
		overlay = {
			_localOnly = {
				updateTimeStamp = updateTimeStamp,
			},
		},
	}
						
	resultP:runUntilDone()
	
	return resultP:mapDataP( function ()
		return 'data', true
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the name of an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< string | nil ></code> A Behavior object whose value is the name
		or nil if the album is not found.
]]--

function AlbumMethods:getAlbumNameB( albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )

	if not albumId then return KSReactive.constantB( nil ) end
	
	return self._indexStore:queryB(
		[[SELECT name FROM albums WHERE id==?]], albumId )
	
	:closingMapB( function( data )

		if not data[ 1 ] then return nil end
		
		return data[ 1 ][ 1 ]
	
	end )
			
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a list of albums for this catalog, sorted by last update date
		(latest first).
	
	\return <code>B< table ></code> A Behavior object whose value is an array of album
		descriptions where each entry in the array is a table containing
		 \c id and \c name and \c updateTimeStamp members.
]]--

function AlbumMethods:albumsDateSortedB()

	if ( not self._albumsDateSortedB )
			or WFCloseSupport.hasBeenOrWillBeClosed( self._albumsDateSortedB ) then
		self:setOwned {
			_albumsDateSortedB = self._indexStore:queryB(
				[[SELECT DISTINCT a.id as id, a.name as name, MAX(
					CASE WHEN a.updateTimeStamp IS NULL THEN 0 ELSE a.updateTimeStamp END,
					CASE WHEN a.localTimeStamp IS NULL THEN 0 ELSE a.localTimeStamp END,
					CASE WHEN b.maxUpdateTimeStamp IS NULL THEN 0 ELSE b.maxUpdateTimeStamp END,
					CASE WHEN c.maxUpdateTimeStamp IS NULL THEN 0 ELSE c.maxUpdateTimeStamp END) as timeStamp FROM
						(SELECT id, name, updateTimeStamp, localTimeStamp FROM albums) a
						LEFT JOIN
						(SELECT albumId, MAX(updateTimeStamp) as maxUpdateTimeStamp FROM album_asset GROUP BY albumId) b
							ON a.id == b.albumId
						LEFT JOIN
						(SELECT albumId, MAX(updateTimeStamp) as maxUpdateTimeStamp FROM space_album GROUP BY albumId) c
							ON a.id == c.albumId
								ORDER BY timeStamp DESC
				]] )
				:closingMapB( function( data )
					if WFUnknown.isUnknown( data) then
						return WFUnknown.value()
					end
					logger:debugf( "Albums date sorted: %s", data )
					local albums = {}
					for i = 1, #data do
						albums[ i ] = {}
						albums[ i ].id = data[ i ][ 1 ]
						albums[ i ].name = data[ i ][ 2 ]
						albums[ i ].updateTimeStamp = data[ i ][ 3 ]
					end
					return albums
				end )
		}
	end
	
	return self._albumsDateSortedB:autoCloseRelayB()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a list of albums for this catalog, sorted by the most recent captureDate
	of any assets in each album. if there are no assets in an album, the updateTimestamp is used.
	
	\return <code>B< table ></code> A Behavior object whose value is an array of album
		descriptions where each entry in the array is a table containing
		 \c id and \c name and \c updateTimeStamp \c captureDate members.
		The MAX captureDate will be nil if there are no assets in an album.
]]--

function AlbumMethods:albumsCaptureDateSortedB()

	if not self._albumsCaptureDateSortedB then
		self:setOwned {
			_albumsCaptureDateSortedB = self._indexStore:queryB( [[
				SELECT albums.id as id, albums.name as name, albums.updateTimestamp as timestamp, a.captureDate FROM albums
					LEFT OUTER JOIN
						( SELECT album_asset.albumId as albumId, MAX(coreInfo.captureDate) as captureDate FROM coreInfo
							INNER JOIN album_asset ON coreInfo.docId == album_asset.assetId
								GROUP BY album_asset.albumId ) as a
					ON a.albumId == albums.id
						ORDER BY CASE WHEN a.captureDate IS NULL THEN albums.updateTimestamp ELSE a.captureDate END DESC
					]] )
				:mapB( function( data )
					if WFUnknown.isUnknown( data) then
						return WFUnknown.value()
					end
					logger:debugf( "Albums sorted by max captureDate: %s", data )
					local albums = {}
					for i = 1, #data do
						albums[ i ] = {}
						albums[ i ].id = data[ i ][ 1 ]
						albums[ i ].name = data[ i ][ 2 ]
						albums[ i ].updateTimeStamp = data[ i ][ 3 ]
						albums[ i ].captureDate = data[ i ][ 4 ]
					end
					return albums
				end )
		}
	end
	
	return self._albumsCaptureDateSortedB:relayB()
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves an unsorted list of albums for this catalog.
	
	\return <code>P< table ></code> A Promise object that resolves to a table of album
		descriptions where each entry in the table is a table containing keys
		 \c id and \c name.
]]--

function AlbumMethods:albumsP()

	return self._indexStore:queryP( [[ SELECT id, name FROM albums ]] )
		:mapDataP( function( status, data )
		
			logger:debugf( "AlbumsP: %s", data )
			local albums = {}
			for i = 1, #data do
				albums[ i ] = {}
				albums[ i ].id = data[ i ][ 1 ]
				albums[ i ].name = data[ i ][ 2 ]
			end
			return 'data', albums
			
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Adds assets to an album in this catalog.
	
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings) An array of unique IDs for the assets to add.
	
	TODO: The return results from this function are weird.  It should consistently
	return a boolean, or a list of ids, not a mix of the two as it does currently.
	
	\return <code>P< true | array ></code>
		A Promise object that resolves to \c true if the assets were already in the album,
		or else to the list of asset IDs that were actually added to the album.
		
]]--

function AlbumMethods:addAssetsToAlbumP( --[[string]] albumId,
					--[[string]] assetIds )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table' )

	local undoRedoAssetIds = {}
	
	local assetsPresent = {}
	
	local function checkForExistingAssetsP()
		
		local arguments = table.concat( assetIds, '\',\'' )
		arguments = string.format( '\'%s\'', arguments )
		
		local sqlQuery = [[ SELECT assetId FROM album_asset WHERE albumId == ? AND assetId IN ( ]] .. arguments .. [[ ) ]]
		return self._indexStore:queryP( sqlQuery, albumId )
		:mapDataP( function( _, idRows )
			for i = 1, #idRows do
				assetsPresent[ idRows[ i ][ 1 ] ] = true
			end
			return 'data', assetsPresent
		end )

	end
	
	local function createAlbumAssetDocument( assetId )
		
		if assetsPresent[ assetId ] then
			return 'data', true
		else
			local albumAssetId = KSIdGenerator.uuidInLowercaseHex()
			local innerResultP = self._docStore:applyOverlayToDocumentP {
				docType = 'album_asset',
				docIds = albumAssetId,
				overlay = {
					album = {
						_id = albumId,
					},
					asset = {
						_id = assetId,
					},
				},
			}

			table.insert( undoRedoAssetIds, assetId )
		
			return innerResultP
			
		end
					
	end
	
	local resultP = self._indexStore:queryP( [[SELECT docId FROM albums WHERE docId == ? ]], albumId )
		:mapP( function ( status, data )

			if status == 'data' and data and data [ 1 ] and data[ 1 ][ 1 ] then
				return checkForExistingAssetsP()
				:mapDataP( function ()
					return KSPromises.allP( createAlbumAssetDocument, assetIds )
				end )
			else
				return 'error', 'invalidAlbum'
			end
			
		end )

	resultP:mapDataP( function()
		self:touchAlbumP( albumId ):runUntilDone()
	end ):runUntilDone()

	return resultP:mapDataP( function( _, assetIds )
		return 'data', assetIds, undoRedoAssetIds
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes assets from an album in this catalog.
	
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings) An array of unique IDs for the assets to delete.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:removeAssetsFromAlbumP( albumId, assetIds )

	WFArguments.check( OzCatalogServiceArg, 'string', 'table' )
		
	local albumAssetsPresent = {}

	local function getAlbumAssetIdsP()
		
		-- TODO [arahn 2013-11-05]: This is a SQL injection vulnerability.
		local arguments = table.concat( assetIds, '\',\'' )
		arguments = string.format( '\'%s\'', arguments )
		
		local sqlQuery = [[ SELECT docId, assetId FROM album_asset WHERE albumId == ? AND assetId IN ( ]] .. arguments .. [[ ) ]]
		return self._indexStore:queryP( sqlQuery, albumId )
		:mapDataP( function( _, idRows )
			for i = 1, #idRows do
				albumAssetsPresent[ idRows[ i ][ 2 ] ] = idRows[ i ][ 1 ]
			end
			return 'data', albumAssetsPresent
		end )

	end
	
	local function deleteAlbumAssetDocument( assetId )
	
		local albumAssetId = albumAssetsPresent[ assetId ]
		
		if albumAssetId then
		
			local localOnly = {
				album = {
					_id = albumId,
				},
				asset = {
					_id = assetId,
				},
			}
					
			local innerResultP = self._docStore:applyOverlayToDocumentP {
				docType = 'album_asset',
				docIds = albumAssetId,
				overlay = {
					_deleted = true,
					_localOnly = localOnly,
				},
			}
						
			return innerResultP

		else
			return 'data', true
		end

	end
	
	local resultP
	
	if albumId == ( 'favorites' .. self._ozUserId ) then
		resultP = self._docStore:applyOverlayToDocumentP {
						docType = 'asset',
						docId = assetIds,
						overlay = {
							favorites = {
								[ self._ozUserId ] = { _delete = true },
							},
						},
					}
	else
		resultP = self._indexStore:queryP( [[SELECT docId FROM albums WHERE docId == ? ]], albumId )
		:mapP( function ( status, data )

			if status == 'data' and data and data [ 1 ] and data[ 1 ][ 1 ] then
				return getAlbumAssetIdsP()
				:mapDataP( function ()
					return KSPromises.allP( deleteAlbumAssetDocument, assetIds )
				end )
			else
				return 'error', 'invalidAlbum'
			end
		
		end )
		:mapP( function( status, data )
			return self._docStore:docValueOnceP( 'album', albumId )
		end )
		:mapP( function( status, album )
			local coverId = KSTableUtils.derefOpt( album, 'cover', '_id' )
			if coverId and albumAssetsPresent[ coverId ] then
				return self._docStore:applyOverlayToDocumentP {
					docType = 'album',
					docId = albumId,
					overlay = {
						cover = {
							_delete = true,
						},
					},
				}
			end
		end )
		
		resultP:mapDataP( function()
			self:touchAlbumP( albumId ):runUntilDone()
		end ):runUntilDone()
	end
	
	return resultP
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Is an albumId valid for this catalog
	
	\param albumId (string) The unique ID of the album to modify.
	
	\return <code>P< boolean ></code> A Promise object that resolves to a boolean
		if the album id exists and has not been deleted for this catalog.
]]--

function AlbumMethods:isValidAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )
	
	return self._indexStore:queryP( [[ SELECT 1 FROM albums WHERE id = ? LIMIT 1 ]], albumId )
		:mapDataP( function( status, albumList )
			
			return 'data', not not albumList[ 1 ]

		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Is an albumId valid for this catalog
	
	\param albumId (string) The unique ID of the album to modify.
	
	\return <code>B< boolean ></code> A Behavior object that resolves to a boolean
		if the album id exists and has not been deleted for this catalog.
]]--

function AlbumMethods:isValidAlbumB( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string|boolean' )
	
	-- use a count so we can differentiate the initial empty results from
	-- the none-exists result.
	return self._indexStore:queryB( [[ SELECT count( id ) FROM albums WHERE id = ? LIMIT 1 ]], albumId )
		:closingMapB( function( albumList )
			
			return ( albumList[ 1 ] and albumList[ 1 ][ 1 ] ) ~= 0

		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Modifies the 'Allow Downloads' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	\param allowDownloads (boolean) The new value for the setting.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function AlbumMethods:setAllowDownloadsForAlbumP( --[[string]] albumId,
						--[[boolean]] allowDownloads )

	WFArguments.check( OzCatalogServiceArg, 'string', 'boolean' )
	
	local resultP = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function ( status, rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
				
				local innerResultP = self._docStore:applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						download = allowDownloads,
					},
				}
								
				return innerResultP
					:mapDataP( function ()
						return 'data', true
					end )
			else
				return 'error', 'AlbumNotShared'
			end
			
		end )
			
	return resultP
 
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports the current 'Allow Downloads' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< boolean ></code> A Behavior object whose value is the current value of the setting.

 ]]--

function AlbumMethods:getAllowDownloadsForAlbumB( --[[string]] albumId )
	
	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryB( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
	
		:closingMapSwitchB( function ( rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
			
				return self._docStore:fieldValueB( 'space', spaceId,
							'download' )
				:mapB( function ( data )
					return data
				end )
			else
				return KSReactive.constantB( nil )
			end
			
		end )
		
	return resultB
		
end

--------------------------------------------------------------------------------
--[[!
	Reports the current 'Allow Downloads' setting for an album in this catalog.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< boolean ></code> A Behavior object whose value is the current value of the setting.

 ]]--

function AlbumMethods:getAllowDownloadsForAlbumP( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	local resultB = self._indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
	
		:mapDataP( function ( status, rows )
		
			local spaceId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			if spaceId then
			
				return self._docStore:docValueOnceP( 'space', spaceId )
					:mapDataP( function( _, docValue )
					
					local downloadsForAlbum = docValue and docValue.download
					return 'data', downloadsForAlbum
				end )
			
			else
				return 'data', nil
			end
			
		end )
		
	return resultB
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the assetId for the cover image for an album in this catalog.
	Rules governing the selection of the photo:
	If a cover image has been set, and we have the asset in the catalog.
	Select the first photo in the album in chronological order.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< string ></code> A Behavior object whose value is the assetId
	of the current cover photo for this album.
	
	Behavior value will be initially set to <code>WFUnknown.value()</code> until
	the assetId is determined, or nil if there are no assets in the album.

 ]]--

function AlbumMethods:albumCoverImageB( albumId )

	WFArguments.check( OzCatalogServiceArg, 'uuid' )
	
	local explicitCoverFieldB = self._docStore:fieldValueB( 'album', albumId,
							'cover/_id' )

	local resultB = explicitCoverFieldB:closingMapSwitchB( function( coverId, isExisting )
		if isExisting == nil then
			return KSReactive.constantB( WFUnknown.value() )
		end
		local queryB = self._indexStore:queryB( [[
				SELECT coreInfo.docId
				FROM coreInfo
				INNER JOIN album_asset ON album_asset.assetId == coreInfo.docId
				WHERE
					albumId == ?
				AND
					coreInfo.docId = ?
			UNION ALL
				SELECT ( SELECT coreInfo.docId
				FROM coreInfo
				INNER JOIN album_asset ON album_asset.assetId == coreInfo.docId
				WHERE
					albumId == ?
				ORDER BY
					case when album_asset.sortOrder is null then 1 else 0 end,
					album_asset.sortOrder, captureDate ASC )
			LIMIT 1]]
				, albumId, coverId, albumId )
		
			-- GRUNGE [dtull 2014-05-27]: This where clause is effectively duplicated
			-- in OzCatCollections.makeAlbumPredicate...
		
		local calmQueryB = queryB
			:closingCalmB( 'later' )
		
		return calmQueryB
		:closingMapB( function ( rows )

			if WFUnknown.isUnknown( rows ) then
				return rows
			end

			local coverAssetId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
			
			return coverAssetId
			
		end )
	end )
		
	return resultB
		
end

--------------------------------------------------------------------------------
function AlbumMethods:allPhotosAlbumCoverImageB()
	local sql = [[
		SELECT coreInfo.docId 
		FROM coreInfo 
		ORDER BY captureDate DESC 
		LIMIT 1]]

		
	local resultB = self._indexStore

	:queryB( { 
		sql = sql,
		calm = 30
	} )
		
	:closingMapB( function ( rows )

		if WFUnknown.isUnknown( rows ) then
			return rows
		end

		local coverAssetId = rows and rows[ 1 ] and rows[ 1 ][ 1 ]
		
		return coverAssetId
		
	end )

	return resultB
end

--------------------------------------------------------------------------------
--[[!
	\brief Set the assetId for a cover image
]]--

function AlbumMethods:setAlbumCoverImageP( albumId, assetId )

	WFArguments.check( OzCatalogServiceArg, 'uuid', 'uuid' )
	
	return self._docStore:applyOverlayToDocumentP {
		docType = 'album',
		docId = albumId,
		overlay = {
			cover = {
				_id = assetId,
			},
		}
	}
	:mapDataP( function ()
		return 'data', true
	end )

	:runUntilDoneP()

end

--------------------------------------------------------------------------------

local function createIndex( array )
	local index = { }
	for i, asset in ipairs( array ) do
		index[ asset ] = i
	end
	
	return index
end

--------------------------------------------------------------------------------

local function createMarkedAssetArray( assets, touched )
	local markedAssets = { }
	
	for i, assetId in ipairs( assets ) do
		markedAssets[ i ] = {
			assetId = assetId,
			touched = not not touched[ assetId ] -- toboolean
		}
	end
	
	return markedAssets
end

--------------------------------------------------------------------------------

--[[!
	\brief Apply an ordering to a set of assets in the album.
	
	All assets whose ids are in the list will have a persisted ordering when this
	is complete. Any members of the album not in the list will be untouched.  The
	optional set of "touched" ids will be used as a hint to avoid unnecessary
	changes to the order of other items.
	
	Assets in the list that turn out to not actually be in the album for any
	reason (either they aren't really in there or have been deleted since the list
	was generated) will be ignored 	rather than interfere with the persistence of
	other ordering adjustments.
	
	If the assets are already in the specified order, no changes be made even if
	they appear in the touched id set (i.e., superfluous touched id entries should
	not trigger unnecessary changes).
	
	If changes are required, this routine will prefer to modify the order attribute
	of items that appear in the touched id set.
	
	\param albumId (string) The unique ID of the album.
	\param assetArray (table) An array of asset id strings in the desired order.
	\param optTouchedIds (table) a set of "touched" assets (asset id string keys)
	
	\return <code>P< true ></code> An indication that the order was saved.
 ]]--

function AlbumMethods:updateAssetOrderingP( albumId, assetArray, optTouchedIds )

	return self._indexStore:queryP( [[select assetId, docId, sortOrder
			from album_asset where albumId = ?]], albumId )
		:mapDataP( function( status, results )
		
			local assetIdToIndex = createIndex( assetArray )
			local markedAssetArray = createMarkedAssetArray( assetArray, optTouchedIds or { } )
		
			for i, result in ipairs( results ) do
				local index = assetIdToIndex[ result[ 1 ] ]
				if index then
					markedAssetArray[ index ].albumAssetId = result[ 2 ]
					markedAssetArray[ index ].order = result[ 3 ]
				end
			end
			
			local WFAlbumOrdering = require 'WFAlbumOrdering'

			local updates = WFAlbumOrdering.adjustOrderFields( markedAssetArray )
			
			local docIds, overlays = { }, { }
			
			for i, update in pairs( updates ) do
				local docId = markedAssetArray[ i ].albumAssetId
				
				if docId then
					docIds[ #docIds + 1 ] = docId
					overlays[ #overlays + 1 ] = {
						order = update,
					}
				end
			end

			if #docIds > 0 then
				return self._docStore:applyOverlayToDocumentP {
					docType = 'album_asset',
					docIds = docIds,
					overlays = overlays
				}
			else
				return KSPromises.resolvedPromiseP( 'data', true )
			end
		end ):runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief Determine if an album has any custom ordered assets.

	Returns true if any asset in this album has a custom sort order applied and
	false if none do.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>P< boolean ></code> True if this album has customized ordering.
 ]]--

function AlbumMethods:hasCustomizedOrderingP( albumId )

	return self._indexStore:queryP( [[select sortOrder from album_asset
			where albumId = ? and sortOrder is not null limit 1]], albumId )
		:mapDataP( function( status, rows )
			return status, #rows > 0
		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Determine if an album has any custom ordered assets.

	Returns true if any asset in this album has a custom sort order applied and
	false if none do.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< boolean ></code> True if this album has customized ordering.
 ]]--

function AlbumMethods:hasCustomizedOrderingB( albumId )

	return self._indexStore:queryB( [[select sortOrder from album_asset
			where albumId = ? and sortOrder is not null limit 1]], albumId )
		:mapB( function( rows )
			return rows and #rows > 0
		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Clear ordering from set of assets in the album.
	
	All assets whose ids are in the set will have their order field removed.
	
	Assets in the set that turn out to not actually be in the album for any
	reason (either they aren't really in there or have been deleted since the set
	was generated) will be ignored 	rather than interfere with the persistence of
	other adjustments.

	If the set is not provided, the order field of all assets in the album will
	be cleared instead.
	
	\param albumId (string) The unique ID of the album.
	\param optAssetIds (table) a set of assets to clear (asset id string keys)
	
	\return <code>P< true ></code> An indication that the order was saved.
 ]]--

function AlbumMethods:clearAssetOrderingP( albumId, optAssetIds )

	return self._indexStore:queryP( [[select assetId, docId, sortOrder
			from album_asset where albumId = ?]], albumId )
		:mapDataP( function( status, results )

			local toClear = { }
		
			for i, result in ipairs( results ) do
				local assetId = result[ 1 ]
				local albumAssetId = result[ 2 ]
				local order = result[ 3 ]
			
				-- apply to all assets if no set is provided, but only clear
				-- assets that currently have an ordering applied
				if not optAssetIds or optAssetIds[ assetId ] and order then
					toClear[ albumAssetId ] = true
				end
			end
			
			local docIds = { }
			for id in pairs( toClear ) do
				docIds[ #docIds + 1 ] = id
			end

			return self._docStore:applyOverlayToDocumentP {
				docType = 'album_asset',
				docIds = docIds,
				overlay = {
					order = {
						_delete = true
					}
				},
			}
		end ):runUntilDoneP()

end

--------------------------------------------------------------------------------

--[[!
	\brief Return the update timestamp for an album.
	
	\param albumId (string) The unique ID of the album.
	
	\return <code>B< string ></code> The server date for the updated time, or nil
	if none is known.
	
	Behavior value will be initially set to <code>WFUnknown.value()</code> until
	the assetId is determined, or nil if there are no assets in the album.

 ]]--

function AlbumMethods:albumTimeStampB( albumId )

	WFArguments.check( OzCatalogServiceArg, 'uuid' )
	
	return self:albumsDateSortedB()
	:closingMapB( function( rows )
		if WFUnknown.isUnknown( rows ) or rows == nil then
			return rows
		end
		local t
		for _, album in ipairs( rows ) do
			if album.id == albumId then
				t = album.timeStamp or album.updateTimeStamp
			end
		end
		return t
	end )

end

--==============================================================================

return {
	AlbumMethods = AlbumMethods,
}
