--[[----------------------------------------------------------------------------

OzCatInit.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.os
local KSDate = require 'KSDate'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local _ = require 'WFArguments_uuid'
local WFCapabilities = require 'WFCapabilities'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'
local OzFilePaths = require "OzFilePaths"

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.lrmobile_only
local OzUser = require 'OzUser'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzCatalogSpec = require 'WFOzCatalogSpec'

	-- WF.web.http_client
local WFReachability = require 'WFReachability'


--============================================================================--

-- @mark OzCatalogService class

local InitMethods = {}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a catalog instance for an Oz user.
	
	\param catalogId (string) The UUID for the catalog
	\param ozUser (OzUser) The user who is accessing the catalog.
]]--

function InitMethods:__init( catalogId, ozUser, minimized )

	WFArguments.check( 'OzCatalogServiceImp', 'uuid', 'OzUser' )

	WFCloseSupport.mustClose( self )

	self._catalogId = catalogId
	self._refsToThisCatalog = KSTableUtils.makeFullyWeak()
	
	self._ozUserId = ozUser:id()

	self:setOwned { _ozUser = OzUser.make( self._ozUserId ) }

	-- Spin up doc store and index store.

	if WFCapabilities.getCapability 'OzDocStore.disable' then
	
		local ozCatalogSpec = WFOzCatalogSpec( ozUser:wfOzUser(), catalogId )
		local ozCatalog = require 'WFOzCatalog'( ozCatalogSpec )
		
		local docStore = ozCatalog:docStore()
		local indexStore = ozCatalog:indexStore()
		
		self:setOwned { _ozCatalog = ozCatalog }
		
		self._docStoreWF = docStore
		
		-- Separate initialization because the second depends on the first.
		
		self:setOwned {
			_docStore = require 'OzDocStoreViaWFDocStore'( self._docStoreWF, catalogId ),
		}
		
		self._indexStore = indexStore
	
	else
	
		self:setOwned { _docStore = require 'OzDocStore'( catalogId ) }
		self:setOwned {
			_indexStore = WFIndexStore( self._docStore, 'OzCatalogIndexDefinitions.lua' )
		}

	end
	
	-- GRUNGE [scouten 2011-05-31]: Should really have a centralized location
	-- where we calculate and return catalog paths. This code is approximately
	-- duplicated from OzDocStoreImp.

	self._originalsFolderDocuments = OzFilePaths.getStandardOriginalFolderDocumentsPath( catalogId )
										
	self._originalsFolderCaches = OzFilePaths.getStandardOriginalFolderCachesPath( catalogId )
	
	WFFileUtils.mkdir( self._originalsFolderDocuments )
	WFFileUtils.mkdir( self._originalsFolderCaches )
	
	self._minimized = not not minimized
	
	self:createReactiveVariable( '_accessErrorB', false )
	
	-- Track reachability status notifications.
	
	self:setOwned {
		_ozHostIsReachableB = WFReachability.hostIsReachableB( OzCatalogUrls.baseServerUrl )
	}
	
	self._ozHostReachabilityLog = {}
	
	self._ozHostIsReachableB:ownedOnDataB( self, function( newStatus )
	
		local logSize = #( self._ozHostReachabilityLog )
		
		self._ozHostReachabilityLog[ logSize + 1 ] = {
			status = newStatus or 'nil',
			time = KSDate.timeInIso8601FormatWithTimeZone(),
		}
		
		if logSize >= 10 then
			table.remove( self._ozHostReachabilityLog, 1 )
		end
	
	end )
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves the Oz user who owns this catalog.
	
 	\return (OzUser) The Oz user instance.
]]--

function InitMethods:ozUser()

	return self._ozUser

end

--[[!
	\brief Retrieves the Oz user id.
	
 	\return (string) The Oz user id.
]]--


function InitMethods:ozUserId()

	return self._ozUser:id()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the Oz catalog
	
 	\return (WFOzCatalog) The Oz catalog instance.
]]--

function InitMethods:ozCatalog()

	return self._ozCatalog
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the unique ID of this catalog.
	
 	\return (string) The UUID.
]]--

function InitMethods:getCatalogId()

	return self._catalogId

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the path of the documents folder for this catalog.
	
 	\return (string) The UUID.
]]--

function InitMethods:getOriginalsDocumentsFolder()

	return self._originalsFolderDocuments
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the name of this catalog.
	
 	\return <code>B< string ></code> A Behavior object whose value is the catalog name.
]]--

function InitMethods:nameB()

	if not self._nameB then
	
		self:setOwned {
			_nameB = self._docStore:fieldValueB( 'catalog', self._catalogId, 'name' )
		}
	
	end
	
	return self._nameB:relayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Enables some basic APIs, such as nameB and ownerIdB by manually
	filling in the root catalog document.
	
	\return (resultP<bool>) a promise that returns true when finished.
--]]

function InitMethods:bootstrapBasicInfoP( catalogInfo )

	return self._docStore:docValueOnceP( 'catalog', catalogInfo._id )
	:mapDataP( function( _, value )

		-- [arahn 2014-02-06] Thanks to this FIXME in WFOzDownloadTask:
		
			-- FIXME [scouten 2012-04-17]: Split catalog users out into a separate
			-- document?
			
		-- I cannot simply boostrap catalog info with a clean version
		-- from Oz, because WFOzDownloadTask adds extra fields to the
		-- "server" value.
		
		-- FIXME_SNOWLESS : Don't repeat this mistake: Make info about users
		-- in a catalog separate 'user' documents.
		
		if value._deleted then
			return 'data', true
		end
		
		local merged = {}
		for k,v in pairs( value ) do
			merged[ k ] = v
		end
		for k,v in pairs( catalogInfo ) do
			merged[ k ] = v
		end
		self._docStore:setServerValueP( 'catalog', catalogInfo._id, merged )
	end )
	:mapDataP( function()
		return 'data', true
	end )
	:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the document store instance that backs this catalog.
	
	\warning Do not close this document store; doing so causes
		any other clients of this catalog instance to fail.
	
	\return (OzDocStore) The document store instance.
--]]

function InitMethods:docStore()

	return self._docStore

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the index store instance that backs this catalog.
	
	\warning Do not close this index store; doing so causes
		any other clients of this catalog instance to fail.
	
	\return (WFIndexStore) The index store instance.
--]]

function InitMethods:indexStore()

	return self._indexStore

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieve the data cache.
	
	\return (KSDataCache) The data cache.
--]]

function InitMethods:dataCache()

	return self._docStore:dataCache()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the number of documents left to index in this catalog.
	
	This can be used to approximate the time left until the index store
	catches up.
	
	\return <code>B< number ></code>A Behavior object whose value is the number
		of documents left to index.
--]]

function InitMethods:sizeOfIndexQueueB()

	return self._indexStore:requestE( 'sizeOfIndexQueueB' ):closingHoldB( nil )

end

--------------------------------------------------------------------------------

-- These are the tables that are considered "critical" enough to merit
-- up front re-indexing if they are sufficiently far behind.
--
-- In this case, "critical" is defined as: out of date queries produce
-- an unsettling user experience (for example, out of date results in
-- some of these tables will result in the appearance of missing most
-- of the photos in your Library and then slowing re-syncing them, which
-- freaks people out; it is better in that case to bite the bullet and
-- bring things up to date than lazily rebuild)

local essentials = {
	albums = true,
	album_asset = true,
	coreInfo = true,
	coreMD5 = true,
	tags = true,
}

--[[!
	\brief Checks the readiness of the index store, returning a promise
	that resolves to an indication of whether it is out of date.
		
	\return <code>P< boolean ></code>A Promise object whose value is true
	if the index store is far enough behind it is worth waiting to bring
	it up to date
--]]

function InitMethods:shouldWaitForIndexingP()
	
	-- this threshold is arbitrary; we want there to be a big enough
	-- backlog that it's worth the time and transitions to show the
	-- progress UI during the process
	return self._indexStore:shouldWaitForIndexingP( 2500, essentials )

end

--[[!
	\brief Brings the index store fully up to date and provides a
	progress value to track the completion of the work.

	\return <code>B< number ></code>A Behavior object the provides a
	progress value starting at 0 and moving toward 1. It may have the
	value WFUnknown.value() initially, but should be numeric
--]]


function InitMethods:indexingProgressB()

	-- policy decision: we decide whether to force re-indexing based on
	-- the status of critical tables; but once we decide to re-index we
	-- bring everything up to date while we're at it
	return self._indexStore:indexingProgressB()

end

--------------------------------------------------------------------------------

return {
	InitMethods = InitMethods,
}
