--[[----------------------------------------------------------------------------

OzCatSharing.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFArguments = require 'WFArguments'
local OzCatSharingHelper = require 'OzCatSharingHelper'

--==============================================================================

-- @mark OzCatalogService class

local OzCatalogServiceArg = 'OzCatalogServiceImp'

local SharingMethods = {
}

--------------------------------------------------------------------------------

function SharingMethods:resolveAlbumFromSpaceP( spaceId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	return OzCatSharingHelper.resolveAlbumFromSpaceP( self._indexStore, spaceId )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Shares an album in this catalog, creating an associated space document.
	
	\param albumId (string) The unique ID of the album.
	\param spaceSubtype (string) The subtype for the album's space; usually nil.
	
	\return <code>P< string ></code> A Promise object that resolves to the ID of
		the new space if the request succeeded.
]]--

function SharingMethods:shareAlbumP( albumId, spaceSubtype )
	
	return OzCatSharingHelper.shareAssetOrAlbum( self._indexStore, self._docStore, 'album', albumId, spaceSubtype )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Unshares an album in this catalog, deleting the associated space
		and space_album documents.

	\param albumId (string) The unique ID of the album.

	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function SharingMethods:unshareAlbumP( albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	return OzCatSharingHelper.unshareAssetOrAlbumP( self._indexStore, self._docStore, 'album', albumId )
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Checks whether an album in this catalog is shared.
	
	\param albumId (string) The unique ID of the album.
		
	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		if the album is shared, \c false if not
]]--

function SharingMethods:isAlbumSharedB( --[[string]] albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	return OzCatSharingHelper.isAssetOrAlbumSharedB( self._indexStore, self._docStore, 'album', albumId )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether an album in this catalog is shared.
	
	\param albumId (string) The unique ID of the album.
		
	\return <code>P< boolean ></code> A Promise object that resolves to \c true
		if the album is shared, \c false if not
]]--

function SharingMethods:isAlbumSharedP( albumId )

	WFArguments.check( OzCatalogServiceArg, 'string' )
	
	return OzCatSharingHelper.isAssetOrAlbumSharedP( self._indexStore, 'album', albumId )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the public URL for an album in this catalog, if the album is shared.
	
	\param albumId (string) The unique ID of the album.
	\param optSpaceId (optional string) The ID of the sharing space record for
		this album. If provided, can help speed the query.
	
	\return <code>B< string ></code> A Behavior object whose value is the public URL
		for the shared album.
]]--

function SharingMethods:sharedAlbumUrlB( albumId, optSpaceId )

	WFArguments.check( OzCatalogServiceArg, 'string', '?string' )
	
	return OzCatSharingHelper.sharedAssetOrAlbumUrlB( self._indexStore, self._docStore, 'album', albumId, optSpaceId )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the public URL for an album in this catalog, if the album is shared.
 	
	\param albumId (string) The unique ID of the album.
	
	\param nonNilResult (boolean) When \c true, the returned promise does not resolve
		until there is a non-nil value; when \c false, the promise resolves to
		whatever value is present in the local database.
	
	\return <code>P< string ></code> A Promise object that resolves to the public URL
		for the shared album, if any.
 ]]--

function SharingMethods:sharedAlbumUrlP( albumId, nonNilResult )

	WFArguments.check( OzCatalogServiceArg, 'string', '?boolean' )
	
	return OzCatSharingHelper.sharedAssetOrAlbumUrlP( self._indexStore, self._docStore, 'album', albumId, nonNilResult )

end

--==============================================================================

return {
	SharingMethods = SharingMethods,
}
