--[[----------------------------------------------------------------------------

OzCatSyncStatus.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSReactive = require "KSReactive"
local WFWorkerThread = require "WFWorkerThread"

local WFCloseSupport = require 'WFCloseSupport'


--==============================================================================

local BabluEP = WFWorkerThread( "BabluImp.lua" )

-- @mark OzCatalogService class

local SyncStatusMethods = {
}

--==============================================================================

local serviceAvailableB = WFCloseSupport.memoizeFn( function( catalogId )

		return BabluEP:requestB {
			message = 'serviceAvailableB',
			arguments = catalogId,
			startingValue = true
		}
		
end )

--------------------------------------------------------------------------------
--[[!
     \brief Reports whether Oz service is available.

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
	Oz service is available. If Oz is unavailable, also returns a 503 error
	with the following response to signal unavailability:
	<code>
	while (1) {}	{"code":9999,"description":"Service is temporarily unavailable"}
	</code>
	
	((is this described correctly?))
]]--

function SyncStatusMethods:serviceAvailableB()

	return serviceAvailableB( self._catalogId ):autoCloseRelayB()

end

--------------------------------------------------------------------------------

local raw_isUploadingToOzB = WFCloseSupport.memoizeCloseTargetFn( function( self )

	local resultB = BabluEP:requestB {
		message = 'isUploadingToOzB',
		arguments = self._catalogId,
		-- allow startingValue to be nil so that clients can (if they choose)
		-- distinguish between "nothing to upload" and "do not know yet"
	}
	
	return resultB

end )

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether the sync engine is currently uploading changes to Oz.
	
	This version returns the real, immediate status, which is probably too twitchy
	for most circumstances; compare \c isUploadingToOzB()..

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
		uploading is in progress.

]]--

function SyncStatusMethods:raw_isUploadingToOzB()

	return raw_isUploadingToOzB( self ):autoCloseRelayB()

end

--------------------------------------------------------------------------------

local isUploadingToOzB = WFCloseSupport.memoizeCloseTargetFn( function( self )

	local rawUploadingB = raw_isUploadingToOzB( self ):autoCloseRelayB()

	local resultB = rawUploadingB:closingMapSwitchE( function( status )
		if status then
			return KSReactive.timerE( 1 ):constantE( true )
		else
			return KSReactive.timerE( 3 ):constantE( status )
		end
	end ):closingHoldValueNowB( rawUploadingB )
	
	return resultB
	
end )

--------------------------------------------------------------------------------

--[[!
	\brief Reports whether the sync engine is currently uploading changes to Oz.
	
	This version has a delay that prevents over-reporting;
	compare \c raw_isUploadingToOzB().

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
		uploading is in progress.

]]--

function SyncStatusMethods:isUploadingToOzB()

	return isUploadingToOzB( self ):autoCloseRelayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether the sync engine is currently pulling changes from Oz.

	\return <code>B< boolean ></code> A Behavior object whose value is \c true if
		downloading is in progress.
]]--

function SyncStatusMethods:isDownloadingFromOzB()

	return BabluEP:requestB {
		message = 'isDownloadingFromOzB',
		arguments = self._catalogId,
		startingValue = false
	}

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether this catalog has ever fully synced with Oz. Note that this
	does \b not signal that the catalog is \b currently synced with Oz,
	but whether it has fully completed the \b first sync with Oz.
		
	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		once the catalog has fully synced with Oz, \c false before then.
--]]

function SyncStatusMethods:hasCompletedInitialSyncB()

	local isDoingInitialSyncB = self._docStore:localStateB( 'isDoingInitialSync' )
	local ozNextUpdatesLinkB = self._docStore:localStateB( 'ozNextUpdatesLink' )

	return KSReactive.closingDataCallB( function( isDoingInitialSync, ozNextUpdatesLink )
		return (not isDoingInitialSync) and ozNextUpdatesLink ~= nil
		end,
		isDoingInitialSyncB,
		ozNextUpdatesLinkB
		)

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether this catalog is performing it's initial sync
		
	\return <code>B< boolean ></code> A Behavior object whose value is \c true
		if the catalog is on it's initial sync, \c false otherwise.
--]]

function SyncStatusMethods:isDoingInitialSyncB()

	return self._docStore:localStateB( 'isDoingInitialSync' )

end

--------------------------------------------------------------------------------
--[[!
	\brief Triggers an immediate sync cycle with Oz.

	\return <code>P< true, boolean ></code> A Promise object that resolves to a pair
		of booleans when the sync cycle is done. The first will be \c true, indicating
		completion. The second will be \c true if new data was downloaded during the sync,
		\c false otherwise.
--]]

function SyncStatusMethods:syncNowP()

	return BabluEP:requestP( 'syncNowP', self._catalogId )

end

--------------------------------------------------------------------------------
--[[!
	\brief Enables or disables automatic synchronization for this catalog.

	\param enabled (boolean) \c true to enable sync, \c false to disable.
]]--

function SyncStatusMethods:setSyncEnabled( --[[boolean]] enabled )

	assert( type( enabled ) == 'boolean' )

--	debug.print( string.format( "User Id: %s switching to Catalog Id: %s", self._ozUserId, self._catalogId ) )

	--debug.print( "Setting syncEnabled for catalog ", self._catalogId, " to ", enabled )

	if enabled then
		self._syncDisabledBy = false
	else
		self._syncDisabledBy = debug.traceback()
	end

	BabluEP:requestP( enabled and 'startSync' or 'stopSync', self._catalogId )
		:runUntilDone()

end

--------------------------------------------------------------------------------

local getSyncEnabledB = WFCloseSupport.memoizeFn( function( catalogId )

	return BabluEP:requestB {
		message = 'syncEnabledB',
		arguments = catalogId,
		startingValue = false
	}
	
end )

--------------------------------------------------------------------------------

--[[!
	\brief Reports whether automatic synchronization is currently enabled for this catalog.
	
	/return <code>B< boolean ></code> A Behavior object whose value is
		\c true if sync is enabled, \c false otherwise.
	 
 ]]--
function SyncStatusMethods:syncEnabledB()

	return getSyncEnabledB( self._catalogId ):autoCloseRelayB()

end


--==============================================================================

return {
	SyncStatusMethods = SyncStatusMethods,
}
