--[[----------------------------------------------------------------------------

OzCentralChunkedUpload.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFCapabilities = require 'WFCapabilities'
local KSTableUtils = require 'KSTableUtils'

local KSPromises = require 'KSPromises'
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.rxlua
local WFRxPromiseBridgeWithResolver = require 'WFRxPromiseBridgeWithResolver'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.os
local KSDate = require 'KSDate'
local WFAnalytics = require "WFAnalytics"

local WFLogger = require 'WFLogger'

	-- WF.core.performance
local WFPerformanceReport = require 'WFPerformanceReport'

	-- WF.web.http_client
local WFRxHttpClient = require 'WFRxHttpClient'


local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt )

local kDoubleQuote = '"' -- "

local asKSPromiseWithResolver = WFRxPromiseBridgeWithResolver

--==============================================================================

local OzCentralChunkedUpload = {}

local logger = WFLogger "OzCentralChunkedUpload"
--:setLevel 'DEBUG'

local MIN_SIZE_FOR_CHUNKING = WFCapabilities.getCapability "OzCentralChunkedUpload.minSizeForChunkedUpload" or ( 5 * 1024 * 1024 )

local kRestartCurrentChunk = "RestartCurrentChunk"
local kRestartChunkedUpload = "RestartChunkedUpload"
local kCannotUploadMaster = "CannotUploadMaster"
local kFileDoesNotExist = "FileDoesNotExist"

--------------------------------------------------------------------------------

local function GenerateNamedError( type, msg, ... )
	return "handledError", {
		errorType = type,
		message = select( '#', ... ) == 0 and msg or string.format( msg, ... ),
	}
end

--------------------------------------------------------------------------------

local function RestartCurrentChunkError( msg, ... )
	return GenerateNamedError( kRestartCurrentChunk, msg, ... )
end

--------------------------------------------------------------------------------

local function RestartChunkedUploadError( msg, url, docId, ... )

	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'binary_upload_chunk_restart'
	errorEvent['event.url'] = url
	errorEvent['event.error_type'] = 'BinaryUpload.Error'
	errorEvent['event.error_desc'] = msg
	
	errorEvent['content.id'] = docId
	
	WFAnalytics.trackEvent( errorEvent )

	return GenerateNamedError( kRestartChunkedUpload, msg, ... )
end

--------------------------------------------------------------------------------

local function CannotUploadMasterError( msg, ... )
	return GenerateNamedError( kCannotUploadMaster, msg, ... )
end

--------------------------------------------------------------------------------

local function convertHttpStatusToError( codesAndErrorFns, url, docId )
	
	return function( status, data, headers, ... )
		
		local statusCode = headers and headers.status_code
		if not statusCode then
			return
		end
		
		local statusCodeKey = statusCode
		local errFn = codesAndErrorFns[ statusCodeKey ]
		if not errFn then
			statusCodeKey = statusCodeKey:sub( 1, -2 ) .. 'x'
			errFn = codesAndErrorFns[ statusCodeKey ]
		end
		if not errFn then
			statusCodeKey = statusCodeKey:sub( 1, -3 ) .. 'xx'
			errFn = codesAndErrorFns[ statusCodeKey ]
		end
		
		if errFn == nil then
			-- do nothing
		elseif type( errFn ) == 'table' then
			-- Handle case where we need to introspect data to see what to do
			local info = data and KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
			local errors = info and info.errors
			if not errors then
				errors = { default = { "default error" } }
			end
			for _, errorCase in ipairs( errFn ) do
				local key = errorCase[ 1 ]
				local errFn2 = errors[ key ]
				if errFn2 then
					return 'error', errFn2( string.format("status_code_%s_error_key_%s", statusCode, key) )
				end
			end
		elseif type( errFn ) == 'function' then
			return 'error', errFn( string.format("status_code_%s", statusCode), url, docId )
		end
		
	end
end

--==============================================================================

local function replaceOrInsertHeader( headers, name, value )
	
	local found = false
	for index, header in ipairs( headers ) do
		if header.header == name then
			if value ~= nil then
				header.value = value
			else
				table.remove( headers, index )
			end
			found = true
			break
		end
	end
	if not found and value ~= nil then
		headers[ #headers + 1 ] = {
			header = name,
			value = value,
		}
	end

end

--------------------------------------------------------------------------------

local function deleted()
	return {
		_delete = true
	}
end

--------------------------------------------------------------------------------

function OzCentralChunkedUpload.getSessionId( url )
	
	if url then
		return url:match( '=(.*)' )
	end
	
	return nil

end

--------------------------------------------------------------------------------

local function makePutP( context, addToUploadTaskManagerFn, uploadTaskManagerKey )

	return function( url, ... )
		url = url:replace( 'v1.0/', 'v2/')
		local chunkedUploadP, observablePromiseResolver = WFPerformanceReport.wrapObservable( 'OzBinaryUploader chunked putBinary',
				context.ozUser:putBinary( url, ... ),
				{
					url = url,
				},
				WFRxHttpClient.extractHttpTimingForPerformanceReport
			)
			:call( asKSPromiseWithResolver )

		addToUploadTaskManagerFn( observablePromiseResolver, uploadTaskManagerKey )
			
		chunkedUploadP = chunkedUploadP:observeP( function( status, responseData, headers, errorInfo )
		
				-- This upload has been removed from uploadTaskManager
				if not status and not headers and not responseData and not errorInfo then
					return 'error', 'cancelled request'
				end
		
				if context.maybeAddToErrorLog then
				
					context.maybeAddToErrorLog( status, headers, responseData, errorInfo )
					
				end
				
			end )
			
		return chunkedUploadP
	end
	
end

--------------------------------------------------------------------------------

local function makePutMetadataP( context, addToUploadTaskManagerFn, uploadTaskManagerKey )

	return function( url, ... )
		url = url:replace( 'v1.0/', 'v2/')
		local metadataUploadP, observablePromiseResolver = WFPerformanceReport.wrapObservable( 'OzBinaryUploader chunked putMetadata',
			context.ozUser:put( url, ... ),
				{
					url = url,
				},
				WFRxHttpClient.extractHttpTimingForPerformanceReport
			)
			:call( asKSPromiseWithResolver )
		
		addToUploadTaskManagerFn( observablePromiseResolver, uploadTaskManagerKey )
		
		return metadataUploadP
			
	end
		
end

--------------------------------------------------------------------------------

local function filterByStatus( httpStatus )

	return function( key, uploadUrl, uploadBody )
	
		return function( status, data, headers, ... )
	
			-- If not a 201, something's borked.
		
			if not headers or headers.status_code ~= httpStatus then
				
				if not headers then
					headers = {}
				end
				headers.errorCaptureKey = key
				
				local logFmtStr =  "Got status %s back from Oz\n\n"
				local logParams = { headers.status_code }
				if uploadUrl then
					logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
					logParams[ #logParams + 1 ] = uploadUrl
				end
				if uploadBody then
					logFmtStr = logFmtStr .. "what we uploaded in body was:\n\n%s\n\n"
					logParams[ #logParams + 1 ] = tostring(uploadBody)
				end
				logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
				logParams[ #logParams + 1 ] = data
				
				logger:errorf( logFmtStr, unpack( logParams ) )
	
				return 'error', data, headers
	
			end
	
			return status, data, headers, ...
	
		end
	
	end
		
end

--------------------------------------------------------------------------------

local filter201 = filterByStatus "201"
local filter202 = filterByStatus "202"

--------------------------------------------------------------------------------

OzCentralChunkedUpload.MIN_SIZE_FOR_CHUNKING = MIN_SIZE_FOR_CHUNKING

--==============================================================================

local function initiateUploadP( putMetadataP, docValue, url, putArgs )

	assert( docValue )

	local parts = KSTableUtils_derefOpt( docValue, "_localOnly", "chunkedUpload", "parts" )
	
	if parts ~= nil then
		KSPromises.resolvedPromiseP( 'data', false, "multipartAlreadyInitiated" )
	end
	
	local masterUrl = url

	-- initiate multi-part upload
	local initiateUrl = masterUrl .. '/parts'
	return putMetadataP( initiateUrl, '', putArgs )
	:mapDataP( convertHttpStatusToError( {
		[ "4xx" ] = CannotUploadMasterError,
	}, initiateUrl, docValue._docId ) )
	:mapDataP( filter201( "initiate master multipart", initiateUrl ) )
	:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmptyP )
	:mapDataP( function( status, results, headers )
		local docStatusOverlay = {
			_localOnly = {
				chunkedUpload = {
					parts = {
					
					},
					part_template = KSTableUtils_derefOpt( results, "links", "/rels/create_part", "href" ),
					finish_href = KSTableUtils_derefOpt( results, "links", "self", "href" ),
				},
			},
		}

		return "data", false, "applyOverlay", docStatusOverlay
		
	end )

end

--------------------------------------------------------------------------------

local function doChunkP( putP, docId, url, putArgs, filePath, chunkPart, startOffset, length )

	putArgs.dataFormat = 'filePath'
	putArgs.offset = startOffset
	putArgs.length = length
	
	local chunkedUploadSource
	if KSData.isData( filePath ) then
		chunkedUploadSource = "data"
	else
		chunkedUploadSource = "file"
	end

	replaceOrInsertHeader( putArgs.requestHeaders, 'X-Generate-Renditions', nil )
	
	local startTime = KSDate.timeNowForAnimation()
	
	logger:debugf( "Start chunk upload %s", chunkPart )
	
	putArgs.timeout = 240
	return putP( url, filePath, putArgs )
	:mapDataP( convertHttpStatusToError( {
		[ "4xx" ] = RestartChunkedUploadError,
	}, url, docId ) )
	:mapDataP( filter201( "master part " .. tostring( chunkPart ), url, filePath ) )
	:mapDataP( function( status, results, headers )

		if not headers.Etag then
			return 'error', RestartCurrentChunkError( "Upload did not include Etag" )
		end
		if headers.Etag:sub( 1, 1 ) ~= kDoubleQuote then
			return 'error', RestartCurrentChunkError( string.format("Etag <<%s>> did not start with double quote", headers.Etag) )
		end
		if headers.Etag:sub( -1, -1 ) ~= kDoubleQuote then
			return 'error', RestartCurrentChunkError( string.format("Etag <<%s>> did not end with double quote", headers.Etag) )
		end
		
		local etag = string.sub( headers.Etag, 2, -2 )
		local sessionId = OzCentralChunkedUpload.getSessionId( url )
		
		local docStatusOverlay = {
			_localOnly = {
				chunkedUpload = {
					parts = {
						[ chunkPart ] = {
							nextPartStart = startOffset + length,
							ETag = etag,
							sessionId = sessionId,
						},
					},
				},
			},
		}
		
		local elapsedTime = KSDate.timeNowForAnimation() - startTime
		local roundedElapsedTime = math.floor( elapsedTime + .5 )
		local ozRequestId = headers['X-Request-Id']
		
		local perfEvent = {}
		perfEvent['event.type'] = 'performance'
		perfEvent['event.workflow'] = 'SYNC'
		perfEvent['event.subcategory'] = 'binary_upload_chunk'
		perfEvent['event.url'] = url
		perfEvent['content.id'] = docId
		perfEvent['content.size'] = length
		perfEvent['content.source'] = chunkedUploadSource
		perfEvent['event.value'] = elapsedTime
		perfEvent['event.context_guid'] = ozRequestId
		
		WFAnalytics.trackEvent( perfEvent )
		
		logger:debugf( "Uploaded chunk %s - %s bytes total - took %s sec", 
						chunkPart, startOffset + length - 1, roundedElapsedTime )
			
		return "data", false, "applyOverlay", docStatusOverlay
		
	end )

end

--------------------------------------------------------------------------------

local function finishUploadP( putP, docValue, url, putArgs )

	assert( docValue )
	
	local finishUrlSessionId = OzCentralChunkedUpload.getSessionId( url )

	local parts = KSTableUtils_derefOpt( docValue, "_localOnly", "chunkedUpload", "parts" )
	local partData = {}
	for partNumber, info in ipairs( parts ) do
	
		if finishUrlSessionId ~= info.sessionId then
		
			local chunkedUploadRestartCount = KSTableUtils_derefOpt( docValue, "_localOnly", "chunkedUploadRestartCount" ) or 0
			chunkedUploadRestartCount = chunkedUploadRestartCount + 1
			
			local docStatusOverlay = {
				_localOnly = {
					chunkedUploadRestartCount = chunkedUploadRestartCount
				},
			}
			
			local errorEvent = {}
			errorEvent['event.type'] = 'error'
			errorEvent['event.workflow'] = 'SYNC'
			errorEvent['event.subcategory'] = 'binary_upload_chunk_multipartError_restart'
			errorEvent['event.url'] = url
			errorEvent['event.error_type'] = 'BinaryUpload.Error'
			errorEvent['event.context_sequence'] = 'binary_upload_chunk_finish'
			errorEvent['event.value'] = chunkedUploadRestartCount
			
			errorEvent['content.id'] = docValue._docId
			errorEvent['content.size'] =  KSTableUtils_derefOpt( docValue, 'importSource', 'fileSize' )
			errorEvent['content.type'] = KSTableUtils_derefOpt( docValue, 'subtype' )
			errorEvent['content.mimetype'] = KSTableUtils_derefOpt( docValue, "importSource", "contentType" )
		
			WFAnalytics.trackEvent( errorEvent )
		
			logger:debugf( "Uploaded chunk sessionId mismatch. Chunk %d sessionId: %s, Finish sessionId:%s", 
							partNumber, info.sessionId, finishUrlSessionId )
							
			RestartChunkedUploadError( 'sessionId_mismatch', url, docValue._docId )
		
			return KSPromises.resolvedPromiseP( 'data', false, kRestartChunkedUpload, docStatusOverlay )
		end
	
		partData[ partNumber ] = {
			part_number = partNumber,
			etag = info.ETag,
		}
	end

	local jsonData = KSJsonEncoder.encodeOzUploadJson( { parts = partData } )
	putArgs.requestMethod = "POST"

	local finishUrl = url
	replaceOrInsertHeader( putArgs.requestHeaders, 'Content-Type', 'application/json' )
	
	return putP( finishUrl, jsonData, putArgs )
	:mapDataP( convertHttpStatusToError( {
		[ "4xx" ] = RestartChunkedUploadError,
	}, url, docValue._docId ) )
	:mapDataP( filter202( "finish multipart", finishUrl ) )
	:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmptyP )
	:mapDataP( function( status, results, headers )
	
		logger:debugf( 'Finished chunked.' )
		
		return "data", true
	end )
	
end

--------------------------------------------------------------------------------

local function restartChunkedP( putMetadataP, url, putArgs )
	local finishUrl = url
	
	putArgs.requestMethod = "DELETE"
	putArgs.requestHeaders = {}
	
	return putMetadataP( finishUrl, '', putArgs )
	:mapDataP( function( status, results, headers )
		local is2x = headers and string.find( headers.status_code, "^2" )
		local is4x = headers and string.find( headers.status_code, "^4" )
		if is2x or is4x then
			local docStatusOverlay = {
				_localOnly = {
					chunkedUpload = {
						_delete = true,
					},
				},
			}
			return "data", false, "applyOverlay", docStatusOverlay
		end
	end )
end

--------------------------------------------------------------------------------

local function clearRestartAndFinishP()
	return KSPromises.resolvedPromiseP( 'data', false, 'applyOverlay', {
		_localOnly = {
			chunkedUpload = {
				_delete = true,
			},
		},
	} )
end

--------------------------------------------------------------------------------

local function sourceFileLooksUnchanged( filePathOrData, docFileSize )

--	local docFileSize = KSTableUtils_derefOpt( docValue, "importSource", "fileSize" )
	
	local fileSize = 0
	
	if KSData.isData( filePathOrData ) then
	
		fileSize = KSData.length( filePathOrData )
		
	else
		
		if type( filePathOrData ) == 'string' and WFFileUtils.exists( filePathOrData ) then
		
			local fileAttrs = WFFileUtils.fileAttributes( filePathOrData )
		
			fileSize = fileAttrs.fileSize
		
		else
		
			fileSize = docFileSize
			
		end
	
	end
	
	
	
	if docFileSize ~= fileSize then
		-- update doc and restart upload
		local docStatusOverlay = {
			importSource = {
				fileSize = fileSize,
			},
		}

		return true, docStatusOverlay
	end
	return false
	
end

--------------------------------------------------------------------------------

local function doChunkedUploadImpP( context, docValue, addToUploadTaskManagerFn, binaryType, chunkType, chunkPart, url, putArgs, filePath, docFileSize )
	
	local uploadTaskManagerKey = context.docId .. '\t' .. binaryType
	
	local putP = makePutP( context, addToUploadTaskManagerFn, uploadTaskManagerKey )
	
	local putMetadataP = makePutMetadataP( context, addToUploadTaskManagerFn, uploadTaskManagerKey )

	local fileHasChanged, overlay = sourceFileLooksUnchanged( filePath, docFileSize )
	
	if fileHasChanged then
		RestartChunkedUploadError( 'file_has_changed', url, context.docId )
		return KSPromises.resolvedPromiseP( 'data', false, kRestartChunkedUpload, overlay )
	end
	
	if chunkType == 'restart' then
	
		return restartChunkedP( putMetadataP, url, putArgs )

	elseif chunkType == 'clearRestartAndFinish' then
	
		return clearRestartAndFinishP()

	elseif chunkType == 'finish' then
	
		if chunkPart == 0 then
		
			-- Uh oh.  We thought we were finished uploading chunks, but don't have the complete
			-- url.  Bail.
			RestartChunkedUploadError( 'no_finish_url', url, context.docId )
			
			return KSPromises.resolvedPromiseP( 'data', false, kRestartChunkedUpload )
		end
	
		return finishUploadP( putP, docValue, url, putArgs )
	end
	
	if chunkPart < 0 then
		return initiateUploadP( putMetadataP, docValue, url, putArgs )
	end
	
	local startOffset = ( chunkPart - 1 ) * MIN_SIZE_FOR_CHUNKING + 1
	local length = MIN_SIZE_FOR_CHUNKING
	
	if ( length + startOffset ) > docFileSize or ( docFileSize - length - startOffset ) < MIN_SIZE_FOR_CHUNKING then
		length = 1 + docFileSize - startOffset
	end
	
	return doChunkP( putP, context.docId, url, putArgs, filePath, chunkPart, startOffset, length )
end

--------------------------------------------------------------------------------

function OzCentralChunkedUpload.doChunkedUploadP( context, docValue, addToUploadTaskManagerFn, removeUploadTaskManagerFn, binaryType, chunkType, chunkPart, url, putArgs, filePath, docFileSize )

	-- docValue only required for finishUpload
	return doChunkedUploadImpP( context, docValue, addToUploadTaskManagerFn, binaryType, chunkType, chunkPart, url, putArgs, filePath, docFileSize )
	:mapP {
		error = function( status, maybeHandledError, maybeErrorInfo )
			if maybeHandledError == "handledError" then
				logger:errorf( "Chunked upload encountered error: %s", maybeErrorInfo.message or "NO message" )
				return 'data', false, maybeErrorInfo.errorType
			end
		end
	}
	:mapDataP( function( status, isLastChunk, action, docStatusOverlay )
	
		local resultP
		local needRemoveUploadTask = false
		
		if action == kRestartChunkedUpload then
		
			docStatusOverlay = docStatusOverlay or {}
			local localOnly = docStatusOverlay._localOnly
			if not localOnly then
				localOnly = {}
				docStatusOverlay._localOnly = localOnly
			end
			
			localOnly.chunkedUpload = {
				restart = true,
				part_template = deleted(),
			}
			
			action = 'applyOverlay'
			needRemoveUploadTask = true
			
		end
		
		if action == kRestartCurrentChunk then
			
		end
		
		if action == kFileDoesNotExist then
			needRemoveUploadTask = true
		end
		
		if action == 'applyOverlay' then
			resultP = context.docStore:applyOverlayToDocumentP {
				docType = context.docType,
				docId = context.docId,
				overlay = docStatusOverlay,
			}
		end
		
		if needRemoveUploadTask then
		
			local uploadTaskManagerKey = context.docId .. '\t' .. binaryType
			removeUploadTaskManagerFn( uploadTaskManagerKey )
		end
		
		if isLastChunk then
			if resultP then
				return resultP:mapDataP( function()
					return 'data', 'lastChunk'
				end )
			else
				return 'data', 'lastChunk'
			end
		end
		
		if resultP then
			return resultP
		else
			return 'data', true
		end

	end )
end

--==============================================================================

return OzCentralChunkedUpload
