--[[----------------------------------------------------------------------------

OzUser.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

DEPRECATION NOTICE: This class is used by Lightroom Mobile, but no new clients
should use it.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua.threading
local WFTimer = require 'WFTimer'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFObjectProxy = require 'WFObjectProxy'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.sync_engine
local WFOzUser = require 'WFOzUser'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- WF.web.oauth2_client
local KSOAuth2Session = require 'KSOAuth2Session'
local WFOzUserSynchronizer = require 'WFOzUserSynchronizer'

	-- WF.web.utils
local WFURI = require 'WFURI'

--------------------------------------------------------------------------------

local logger = WFLogger 'OzUser'
--	:setLevel 'debug'

local _WFHttpClient -- lazily loaded

local WFCloseSupport_fetchCached = assert( WFCloseSupport.fetchCached )

KSData.registerArgumentChecker()

local OzUser
	-- forward declaration

--============================================================================--

-- @mark OzUser class

--------------------------------------------------------------------------------
--[[!
       \class OzUser

      \brief This class manages the information about a single user in the Oz/IMS ecosystem.
       
       You can call the functions from any thread. Note that for operations on catalogs,
       you might want to access information about the catalog's owner through that
       catalog object.

       \sa OzCatUsers
]]--

local userMethods = {
	
	_ozBaseUrl = false,
	
	_ozUserId = false,
		-- (string) Adobe user ID

	_userInfoCache = false,
		-- (WFDocStore) this doc store retains info about this user

	_userInfoIndex = false,
		-- (WFIndexStore) index store for user info doc store

	_oAuth2Session = false,
		-- (KSOAuth2Session) OAuth2 session for talking to Oz via IMS

	_wfOzUser = true,
		-- (WFOzUser) new user object
		
	_userSynchronizer = false,
		-- (WFOzUserSynchronizer) the synchronizer (sync agent and peer)

	_refreshTriggerE = true,
		-- (E< boolean >) Oz user sync peer uses this to trigger new sync

	_lastInternalProgressCount = false,
	_lastExternalProgressCount = false,
		-- (number) values for tracking download progress
	
	_pendingUploadP = false,
		-- (promise) the last request to PendingUpload; used to chain
		-- multiple calls together to ensure we don't skip over values
	
	_forceUpgradeB = false,
		-- cache of forceUpgradeB
		
	_behaviorCache = false,
		-- cache used by catalogsB
	
	WFCloseSupport.addCloseSupportMethods,
	
	KSReactive.addReactiveSupportMethods,
	
}

--------------------------------------------------------------------------------
--[[!
      \brief Constructs an OzUser from the database user ID.

      \param ozUserId (string) The unique user ID for the Oz database.
]]--

function userMethods:__init( ozUserId, optBaseUrl )

	if type( ozUserId ) ~= 'string' then
		error( "OzUser: must initialize with a valid user UUID", 4 )
			-- FIXME [scouten 2011-08-05]: This should really enforce that it's a UUID,
			-- not just a string.
	end

	-- OzUser constructs objects that must be explicitly closed, ergo
	-- it must itself be explicitly closed.

	WFCloseSupport.mustClose( self )
	
	self._behaviorCache = KSTableUtils.makeFullyWeak()
	self._forceUpgradeB = KSReactive.variableB( false )
	
	optBaseUrl = optBaseUrl or require 'OzCatalogUrls'.baseServerUrl
		-- bottleneck access to global/static url info this one time only,
		-- which ensures that if a baseUrl is provided we always use that
		-- instead of accidentally referring to a (possibly different)
		-- global static.  That's why the require is done here, so we don't
		-- create a file-level global OzCatalogUrls that can be accessed
		-- anywhere else in this file.
	
	self:setOwned { _wfOzUser = WFOzUser( optBaseUrl, ozUserId ) }

	self._userInfoCache, self._userInfoIndex = self._wfOzUser:userInfoCache()
	assert( not WFCloseSupport.hasBeenOrWillBeClosed( self._userInfoCache ) )

	self._ozUserId = ozUserId

	self:setOwned { _oAuth2Session = KSOAuth2Session( ozUserId ) }

	self:setOwned { _refreshTriggerE = KSReactive.sourceE() }
	
	self._refreshTriggerE:closingCalmE( 1 ):onDataE( function()
		
		local userSynchronizer = self:getOrCreateUserSynchronizer()
	
		userSynchronizer:syncUntilDoneP()
			-- The synchronizer's syncUntilDoneP method calls runUntilDone()

	end )

end

--------------------------------------------------------------------------------

function userMethods:close( )
	debug.print( "closing OzUser" )
end

--------------------------------------------------------------------------------

function userMethods:storeTokensP( accessToken, refreshToken, deviceToken )

	-- WARNING [scouten 2015-07-20]: The parameter sequence here was reversed
	-- today for consistency with other parts of the framework. If you see something
	-- new that's flaky, this might be why.

	return self._oAuth2Session:storeTokensP( accessToken, refreshToken, deviceToken )

end

--------------------------------------------------------------------------------

function userMethods:getOrCreateUserSynchronizer()
	
	if not self._userSynchronizer then
		
		local wfOzUser = self._wfOzUser
		local userSynchronizer = WFOzUserSynchronizer( wfOzUser:serverBaseUrl(), wfOzUser:userId() )
		
		self:setOwned {
			_userSynchronizer = userSynchronizer
		}
	
	end
	
	return self._userSynchronizer
	
end

--------------------------------------------------------------------------------

function userMethods:serverBaseUrl()
	return self._wfOzUser:serverBaseUrl()
end

--------------------------------------------------------------------------------

function userMethods:serverUrl( pathWithFormat, ... )
	return WFURI.resolve( self:serverBaseUrl(),
				string.format( pathWithFormat, ... ) )
end

--------------------------------------------------------------------------------

function userMethods:catalogUrl( catalogId, optAdditionalPath )
	return self:serverUrl( "/v1.0/catalogs/%s%s", catalogId, optAdditionalPath or "" )
end

--------------------------------------------------------------------------------

-- @mark basic info about user

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the unique ID of this Oz user.

	\return (string) The ID.
]]--

function userMethods:id()

	return self._ozUserId

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the \c WFOzUser object for this user.
	
	\return (WFOzUser)
]]--

function userMethods:wfOzUser()

	return self._wfOzUser

end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves (as a Behavior) the email address of this Oz user.

	\return <code>B < string ></code> A Behavior whose value is
		the email address.
]]--

function userMethods:emailB()

	return self._userInfoCache:fieldValueCheckTypeB( 'user', self._ozUserId, 'email' )
		:closingMapB( function( rawEmailValue )

			return rawEmailValue or ""

		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief Retrieves (as a Behavior) the entitlement of this Oz user.

	\return <code>B < string ></code> A Behavior whose value is
		the entitlement.
]]--

function userMethods:entitlementB()

	return self._userInfoCache:fieldValueCheckTypeB( 'user', self._ozUserId, "entitlement" )
		
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Promise) information about this Oz user.

	\return <code>P < table ></code> A Promise that resolves to
		the user information.
]]--


function userMethods:userInfoP( optUserId )

	return self._userInfoCache:docValueCheckTypeP( 'user', optUserId or self._ozUserId )

end

--------------------------------------------------------------------------------

--[[!
	\brief Simulates logout from the current authorization session.

]]--

function userMethods:simulateForcedLogout()

	return self._oAuth2Session:simulateForcedLogoutP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a Behavior) whether this user has a current valid connection
	to the Oz server.
	
	\return <code>B< boolean ></code> A Behavior whose value is
		\c true while there is a valid connection, and becomes \c false
		if the OAuth2 refresh token for this user is ever rejected.
]]--

function userMethods:hasValidConnectionToOzB()

	return self._oAuth2Session:hasValidLoginB():autoCloseRelayB()

end

function userMethods:isWaitingForTokensB()

	return self._oAuth2Session:isWaitingForTokensB():autoCloseRelayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP GET request, signing the request	with the
	current access token of the current user and handles refreshing the access
	token if needed. Returns the response via a promise.

	\param url (string) The URL to fetch.
	\param optArgs (optional, table) Options to modify the request.
	\param optSequencing (optional, string) 'LIFO' or 'FIFO'. FIFO by default.

	\return <code>P< data, headers ></code> A Promise that resolves to the response from HTTP server

	\sa WFHttpClient.getP
]]--

function userMethods:getP( --[[string]] url, --[[table]] optArgs, --[[string]] optSequencing )

	return self._oAuth2Session:getP( url, optArgs, optSequencing )

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP PUT request, signing the request with the
	current access token of the current user and handles refreshing the access
	token if needed. Returns the response via a promise.

	\param url (string) the URL to fetch
	\param data (\c KSData or table) data to send with the request
		(use a Lua table for form data)
	\param optArgs (optional, table) options to modify the request

	\return <code>P< data, headers ></code> A Promise that resolves to the response from HTTP server
	
	\sa HttpClient.putP
]]--

function userMethods:putP( --[[string]] url, --[[KSData or table]] data, --[[table]] optArgs )

	return self._oAuth2Session:putP( url, data, optArgs )

end

--------------------------------------------------------------------------------
--[[!
	\brief Logs this user out of the current Oz session.
]]--

function userMethods:logoutP()

	logger:errorf( "OzUser:logoutP is now deprecated\n\n"
				   .. "called in thread %s via %s", ':thread:', ':traceback:' )

	return self._wfOzUser:logoutP()

end

--------------------------------------------------------------------------------
--[[!
 	\brief Requests an update from Oz for this user's information.
	Use after a significant event (such as a purchase) occurs.

	\return <code>P< table ></code> A Promise that resolves to
		the new user information packet when the reload has finished.
		Callers need not attach an action to the promise.
]]--

function userMethods:refreshP()

	logger:error( "OzUser:refreshP will still trigger an update, but promise is not precise" )
		-- Avoid using this call, or call through to WFSyncAgent:syncNowP().
	
	self._refreshTriggerE:post( true )
	
	return WFTimer.afterDelayP( 1 )

end

--------------------------------------------------------------------------------

local mapCatalogQuery = function( rows )
	if WFUnknown.isUnknown( rows ) then
		return rows
	end
	local result = {}
	for i, row in ipairs( rows ) do
		result[ i ] = {
			id = row[ 1 ],
			name = row[ 2 ],
			ownerId = row[ 3 ],
			ownerFullName = row[ 4 ],
			url = row[ 5 ],
			sharedWith = row[ 6 ],
			acceptedInvite = row[ 7 ] == 'accepted',
			subtype = row[ 8 ],
		}
	end
	return result
end

--------------------------------------------------------------------------------

local mapCatalogQueryForP = function( _, ... )
	return 'data', mapCatalogQuery( ... )
end

--------------------------------------------------------------------------------

local function catalogQueryString( shared )
	local key = shared and "ownerId IS NOT ?" or "ownerId IS ?"
	return string.format( [[
		SELECT
			docId, catalogName, ownerId, ownerFullName, url, userCount, sharingState, subtype
		FROM
			catalogs
		WHERE
			%s
		ORDER BY
			catalogName COLLATE i18n_collate, docId
	]], key )
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) all catalogs this user has created.
 
 	\return <code>B< nil | array of strings ></code> A Behavior whose
		value is initially nil, then the list of catalog IDs.
		The list is updated as catalogs are added or removed.
]]--

function userMethods:catalogsB()

	-- Force a reload from Oz (but not more than one per second or so)
	-- whenever somebody requests this behavior.

	self._refreshTriggerE:post( true )

	-- Build the behavior if we don't have it already.

	local resultB = WFCloseSupport_fetchCached( self._behaviorCache, 'catalogsB' )

	if not resultB then

		resultB = self._userInfoIndex:queryB {
			sql = catalogQueryString( false ),
			self._ozUserId
		}
		:closingMapB( mapCatalogQuery )

		self._behaviorCache.catalogsB = resultB

	end

	return resultB:autoCloseRelayB()

end

--------------------------------------------------------------------------------
--[[
	B< boolean > = ozUser:forceUpgradeB()

	Return a boolean value indicating if the server is forcing an app version
	upgrade.
	
]]--

function userMethods:forceUpgradeB()

	return self._forceUpgradeB:autoCloseRelayB()
	
end

--------------------------------------------------------------------------------

-- FIXME [ arahn 2014/02/08 ] I wonder if we should cache this, so that if
-- multiple calls are made while one is in flight, we don't double our effort.

local function getCatalogListP( self, shared )

	local userSynchronizer = self:getOrCreateUserSynchronizer()

	-- If the user is the current user, then wait for the sync to finish
	-- before we return the query results.
	
	return userSynchronizer:syncUntilDoneP()
	:mapP( function()
		-- In the event the this user is the current user, we've made
		-- extra effort to sync right now so that the list returned is up-to-date.
		-- However, as with many Promise API's, we will return what-is-in-the-docstore
		-- and consider that good enough for a promise.  The fact that this API
		-- takes the time to try and update first is a bit of an odd behavior...
		-- a legacy requirement from how some users were calling this during
		-- first run / first time logged in.
		-- Thus, if this API is called when offline, I found it is better
		-- to fall back to the cached data in the user info cache rather than
		-- to resolve to error.
		return 'data', true
	end )
	:mapDataP( function( )
		return self._userInfoIndex:queryP {
			sql = catalogQueryString( shared ),
			self._ozUserId
		}
	end )
	:mapDataP( mapCatalogQueryForP )
	:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
 	\brief Retrieves (as a Promise) a list of all catalogs to which this user has
	access.
	
 	\return <code>P< table ></code> A Promise that resolves to the list of
 		catalog URLs.
]]--

function userMethods:catalogsP( subtype )

	return getCatalogListP( self, false )

end

--------------------------------------------------------------------------------

local function lazy_WFHttpClient()

	if not _WFHttpClient then
		_WFHttpClient = require 'WFHttpClient'
	end
	
	return _WFHttpClient

end

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new catalog on the Oz server.

	\param args (table) The description of the new catalog, a table with
		these keys:
	       <ul>
		<li>id (optional, string): Value is the UUID (in lowercase hex format) of the new
			catalog; if missing, one is created,</li>

		<li>name (required, non-empty string): Value is the user-visible name for the new
			catalog.</li>

		<li>users (optional, table): Value is the list of user emails to be invited
			to the catalog. </li>
		</ul>

	\return A Promise. On success, resolves to:
		'data', catalogUrl, headers, catalogId, didUserInvitationSucceed (boolean)
		<ul><li>If catalog creation succeeded, but only the user invitation failed, the promise returns:
		'data', catalogUrl, headers, catalogId, false</li>
		<li>If the catalog creation completely failed, the promise resolves in error.</li></ul>

	Any catalog list behaviors are reloaded upon resolution of the promise.
	\sa :catalogB
]]--

function userMethods:createCatalogP( args )

	if type( args ) ~= 'table' then
		error( "OzCatalogList.createCatalog: args must be a table", 2 )
	end

	if type( args.name ) ~= 'string' or #args.name == 0 then
		error( "OzCatalogList.createCatalog: name must be a non-empty string", 2 )
	end

	if args.id and type( args.id ) ~= 'string' and #args.id ~= 32 then
		error( "OzCatalogList.createCatalog: id, if present, must be a UUID string", 2 )
	end

	-- Process users to be invited.

	if args.users then
		for i, v in ipairs( args.users ) do
			args.users[ i ] = "+mailto:" .. args.users[ i ]
		end

	end

	local catalogId = args.id or KSIdGenerator.uuidInLowercaseHex()

	local newCatalogUrl = self:catalogUrl( catalogId )

	local newCatalogData
	
	local catalogData = { name = args.name }
		
	if args.subtype then
		catalogData[ 'subtype' ] = args.subtype
	end
		
	if args.payload then
		catalogData[ 'payload' ] = args.payload
	end
		
	newCatalogData = KSJsonEncoder.encodeOzUploadJson( catalogData )

	local putArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
		},
	}

	return self:putP( newCatalogUrl, newCatalogData, putArgs )
		:mapP( function( status, data, headers )

			if status == 'error' then

				if lazy_WFHttpClient().responseContentTypeIsJson( headers )
					and data
				then

					local result, eData = pcall( KSJsonDecoder.decodeWebSafeJson, data )
					return status, ( result and eData and eData.description )
								   or "unknown error"

				else
					return status, data, headers
				end

			end

			-- The only response we treat as success is 201 Created
			-- with a Location: headers. Anything else is an error.

			local location = headers
				and headers.status_code == '201'
				and headers.Location

			if location then

				self._refreshTriggerE:post( true )

				local OzCatalogService = require 'OzCatalogService'
				
				local tempCatalog = OzCatalogService( catalogId, self )

				local ownerId = self._ozUserId

				-- As a fix for WATSON #2989758 (Spurious conflict reported in debug
				-- log for catalog owner ID) we check if the user id has the urn prefix,
				-- if not we attach it to avoid spurious conflicts when the sync agent
				-- for this catalog fetches the ownerId from Oz and writes it to the doc
				-- store.

				if not string.match( ownerId, 'urn:carousel:account:' ) then
					ownerId = 'urn:carousel:account:' .. ownerId
				end

				-- As a fix for #3049221, we only write the owner status to the doc
				-- store if it is not "created". This is to avoid spurious conflicts
				-- in the case when a new user is created and the default catalog is
				-- created for this new user since the server value returned by Oz
				-- is "trial." This can lead to import being disallowed for this catalog
				-- because the catalog docStore says the owner status is "created."

				-- #3176709	- Write the account status after making sure that the
				-- user info we have is refreshed from Oz. This is to make sure we
				-- don't end up with spurious conflicts if the account status has
				-- changed on Oz in the meantime.

				local refreshedUserInfoP = self:refreshP()

				return refreshedUserInfoP
				
					:mapDataP( function()
					
						return self:userInfoP()
					
					end )

					:mapDataP( function( status, userInfo )

						local ownerOverlay = {}
						ownerOverlay.id = ownerId

						local ownerStatus = userInfo.status
						if ownerStatus ~= 'created' then
							ownerOverlay.status = ownerStatus
						end

						return tempCatalog._docStore:applyOverlayToDocumentP
							{
								docId = catalogId,
								docType = 'catalog',
								overlay = {
									owner = ownerOverlay,
								},
							}

					end )
					
					:observeP( function()
					
						tempCatalog:close()
						
					end )

					:mapDataP( function( status, data )
	
						-- Send a request to invite users.

						if args.users ~= nil then
							local usersData = KSJsonEncoder.encodeOzUploadJson
								{ users = args.users  }
	
							local postArgs = {
								requestHeaders = {
									{ header = 'Content-Type', value = 'application/json' },
								},
								requestMethod = 'POST',
							}
	
							return self:putP( newCatalogUrl, usersData, postArgs )
								:mapP( function ( status, data, inviteHeaders )
	
									if status == 'error' then
										return 'data', location, headers, catalogId, false
									else
										return 'data', location, headers, catalogId, true
									end
								end )
	
						else

							return 'data', location, headers, catalogId, true
	
						end
	
					end )

				:runUntilDoneP()

			else

				return 'error', false

			end

		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief HTTP login call to IMS to get access token and refresh token for test
	cases.

	This is based on the following protocol:
	https://zerowing.corp.adobe.com/display/ims/IMS+API+-+login#IMSAPI-login-Usagefromanativeappwiththeauthorizationcodegranttypeenabled

	Intended for test cases where we need access token and refresh token
	and we want to bypass IMS login through webview/UI.

	\param imsLoginUrl (string) IMS login url
	\param postParams (string) Parameters to pass according to IMS documentation in above link, for example:
	 	client_id=RevelMacTest1&redirect_uri=air-auth://carousel&locale=en_US&scope=openid,AdobeID&
		client_secret=12d5de23-e638-41f5-815e-04a10a5e13e1&username=testUser&password=testPassword

	\return <code>P< string, string ></code> A Promise that resolves to the access token and refresh token.
]]--

function userMethods:getAccessTokenRefreshTokenFromImsP( --[[ string ]] imsLoginUrl,
							--[[ string ]] postParams )

	local postArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/x-www-form-urlencoded' },
			{ header = 'Accept', value = '*/*' },
		},
		requestMethod = 'POST',
	}

	return lazy_WFHttpClient().putP( imsLoginUrl, postParams, postArgs )
		:mapDataP( function( status, data, headers )

			if headers.status_code ~= "200" then

				local responseString = tostring( data )
				return 'error', responseString

			else

				local decodedJson = KSJsonDecoder.decodeWebSafeJson_stripNull( data )
				return 'data', decodedJson.access_token, decodedJson.refresh_token

			end

		end )

end

--------------------------------------------------------------------------------

-- Method internal to this file, for deleting any device tokens understood to be registered
-- with Oz.
-- compareTokenUrl if passed in will be compared with the device token url locally saved
-- and the deletion will not take place if the two are the same.

local function deleteDeviceTokenP( self, compareTokenUrl, app_id, aps_env )

	return self._userInfoCache:docValueCheckTypeP( 'user', self._ozUserId )
		:mapDataP( function ( status, docValue )
					
			local deleteTokenUrl = KSTableUtils.derefOpt( docValue, '_localOnly', 'device_token' )

			if deleteTokenUrl then
				
				if compareTokenUrl and compareTokenUrl == deleteTokenUrl then
					return 'data', true
				end
				
				local queryString = WFURI.buildQueryString {
					app_id = app_id,
					aps_env = aps_env,
				}
				
				if queryString ~= "" and string.sub( deleteTokenUrl, -1, -1 ) ~= '?' then
					deleteTokenUrl = deleteTokenUrl .. "?"
				end
				
				if queryString then
					deleteTokenUrl = deleteTokenUrl .. queryString
				end
						
				local deleteArgs = {
					requestMethod = 'DELETE',
					returnHttpErrorsAsData = true,
					requestHeaders = {
						{ header = 'Content-Type', value = 'application/json' },
					},
				}
				
				return self:getP( deleteTokenUrl, deleteArgs )
					:mapP( function( status, data, headers )
						
						local status_code = KSTableUtils.derefOpt( headers, 'status_code' )
						
						-- If the status_code returned by Oz is '204' or '404' we interpret it as a success.
						if not ( status_code == "204" or status_code == "404" ) then
							
							if status == 'error' and data == 'No refresh/access token stored in keychain' then
								return 'error', 'no_token'
							end
							
							logger:errorf( "deleteDeviceToken: status = %s, data = %s, headers = %s",
										   status or '(nil)', data or '(nil)', headers or '(nil)' )
							
							local responseString = 'deleteDeviceToken: error deleting token from Oz status_code: ' .. (status_code or '(nil)')
							return 'error', responseString
		
						else

							-- Delete the token location locally.

							return self._userInfoCache:applyOverlayP
								{
									type = 'user',
									_id = self._ozUserId,
									_slot = 'main',
									_localOnly = {
										device_token = { _delete = true },
									}
								}
								:mapDataP( function()
									return 'data', true
								end )
							
						end

				end )
				
			else
				-- No token location, return true
				return 'data', true
			end
			
		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief API to update the device token for this account in Oz.

	Oz API documentation:
	http://ps-sdk.corp.adobe.com/oz_public/multi_page/index.html#resources_accounts

	\param app_id (string) The app bundle id.
	\param aps_env (string) The aps-environment entitlement value.
	It should be either "development" or "production".
	\param token (string or KSData) The device Token value as received from
	Apple.

	\return <code>P< string|boolean ></code> Promise that resolves to the value
	'true' if the request succeeds and the error string, if it fails.
]]--

-- TODO_SNOWLESS [scouten 2013-12-18]: Create new mechanism for managing
-- device tokens. Note that this is *NOT* the IMS device token, but the Apple push
-- notification service device token.

function userMethods:registerDeviceTokenP( app_id, aps_env, token )

	WFArguments.check( 'OzUser', 'string', 'string', 'KSData|string' )
	
	local postArgs = {
		requestMethod = 'POST',
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
		},
	}

	local registerTokenUrl = self:serverUrl( '/v1.0/accounts/%s/devices', self._ozUserId )

	-- Convert the token to hex
	
	local hexToken = KSStringUtils.convertToHex( token )
	
	local postParams = KSJsonEncoder.encodeOzUploadJson
		{
			device_token = hexToken,
			app_id = app_id,
			aps_env = aps_env,
		}
				
	return self:putP( registerTokenUrl, postParams, postArgs )
		:mapDataP( function( status, data, headers )
			
			if headers.status_code ~= "201" or not headers.Location then

				local responseString = tostring( data )
				return 'error', responseString

			else
				-- Delete any old token first, if it's NOT the same as the
				-- one we just registered. This prevents us deleting the token
				-- in the case that we try to register the same token twice.
				return deleteDeviceTokenP( self, headers.Location, app_id, aps_env )
					:mapDataP( function ( _,_ )
				
						-- Then save the location of new token

						return self._userInfoCache:applyOverlayP
							{
								type = 'user',
								_id = self._ozUserId,
								_slot = 'main',
								_localOnly = {
									device_token = headers.Location,
								},
							}
							:mapDataP( function()
								return 'data', true
							end )
					
				end )

			end

		end )

end

--------------------------------------------------------------------------------

--[[!
	\brief API to delete the device token for this account from Oz.

	Oz API documentation:
	http://ps-sdk.corp.adobe.com/oz_public/multi_page/index.html#resources_accounts

	\return <code>P< string|boolean ></code> Promise that resolves to the value
	'true' if the request succeeds and the error string, if it fails.
]]--

function userMethods:deleteDeviceTokenP( app_id, aps_env )

	WFArguments.check( 'OzUser' )
	
	return deleteDeviceTokenP( self, nil, app_id, aps_env )
	
end

--------------------------------------------------------------------------------

local OzUser_class = Rx.Class 'OzUser'( userMethods )


--============================================================================--

-- @mark OzUser namespace

OzUser = {}

--------------------------------------------------------------------------------

local function validateCached( cache, userId, optEntryPoint )

	local cachedUser = WFCloseSupport_fetchCached( cache, userId )
	
	local oldEntryPoint = cachedUser and cachedUser:serverBaseUrl()
	local newEntryPoint = optEntryPoint or require 'OzCatalogUrls'.baseServerUrl
	
	if cachedUser
			and ( oldEntryPoint ~= newEntryPoint ) then
			
			-- Don't reuse a user that doesn't match all arguments
		cache[ userId ] = nil
		cachedUser = nil
	end
	
	return userId, cachedUser
	
end

--------------------------------------------------------------------------------

local OzUserProxy = WFObjectProxy( "OzUser", OzUser_class, validateCached, true )

function OzUser.make( userId, optEntryPoint )

	return OzUserProxy( userId, optEntryPoint )
	
end

--------------------------------------------------------------------------------

return OzUser
