--[[----------------------------------------------------------------------------

PixelProviderSendDataTask.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- wf.core
local KSTableUtils = require 'KSTableUtils'

	-- pixel provider
local PixelProviderSharedTaskUtils = require 'PixelProviderSharedTaskUtils'
local PixelPathProviderStateMachine = require 'PixelPathProviderStateMachine'
local PixelPathProviderSchedulerPrivateMethods =
		require "PixelPathProviderSchedulerPrivateMethods"

--------------------------------------------------------------------------------

local logger = require 'WFLogger'( 'PixelProviderSendDataTask' )

--------------------------------------------------------------------------------

local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt, 
										"KSTableUtils.derefOpt not defined" )

--------------------------------------------------------------------------------

local PixelProviderSharedTaskUtils_getRenderSize =
		assert( PixelProviderSharedTaskUtils.getRenderSize,
				"PixelProviderSharedTaskUtils.getRenderSize not defined" )

--------------------------------------------------------------------------------

local PixelProviderSharedTaskUtils_getBestPossibleRenderSizeData =
		assert( PixelProviderSharedTaskUtils.getBestPossibleRenderSizeData,
				"PixelProviderSharedTaskUtils.getBestPossibleRenderSizeData not defined" )

--------------------------------------------------------------------------------

local PixelProviderSharedTaskUtils_getBestPossibleRenderSizes =
		assert( PixelProviderSharedTaskUtils.getBestPossibleRenderSizes,
				"PixelProviderSharedTaskUtils.getBestPossibleRenderSizes not defined" )


--------------------------------------------------------------------------------

local PixelProviderSharedTaskUtils_makeNonSparseStringArray =
		assert( PixelProviderSharedTaskUtils.makeNonSparseStringArray,
				"PixelProviderSharedTaskUtils.makeNonSparseStringArray not defined" )

--------------------------------------------------------------------------------

local SendDataStates = PixelPathProviderStateMachine.SendDataStates

--==============================================================================
-- choose pixel strategy 
--==============================================================================

local function maybeGetCurrRenderData( baseRenderInfoCache, localAssetId )

	local maybeBaseRenderInfo =
			baseRenderInfoCache:peekCacheItem( localAssetId )

	return KSTableUtils_derefOpt( maybeBaseRenderInfo, 'croppedLongSide' ),
			KSTableUtils_derefOpt( maybeBaseRenderInfo, 'digest' )

end

--------------------------------------------------------------------------------

local PixelPathProviderSchedulerPrivateMethods_sendRenditionPathToPixelStream =
		PixelPathProviderSchedulerPrivateMethods.sendRenditionPathToPixelStream

--------------------------------------------------------------------------------

local function sendRenditionPathForRequest( request, scheduler, 
											renditionDiskStatusCache,
											baseRenderInfoCache )

	local pixStreamId, localAssetId, longSide = unpack( request )

	local croppedLongSide, currDigest =
			maybeGetCurrRenderData( baseRenderInfoCache, localAssetId ) 

	if not croppedLongSide then

		logger:warnf( "Can't send pixel path for %s.%s: missing expected croppedLongSide",
						pixStreamId, localAssetId )

		scheduler:_applyTransitionToRequest( request, 'missing_data' )

		return

	end

	local neededRenderSize = PixelProviderSharedTaskUtils_getRenderSize( 
								croppedLongSide, longSide )

	local path, renderSize, renditionDigest = renditionDiskStatusCache:maybeGetPath( 
										localAssetId, neededRenderSize )

	if not path then

		logger:warnf( "Missing expected pixel path for %s.%s",
						pixStreamId, localAssetId )

		scheduler:_applyTransitionToRequest( request, 'missing_expected_path_computed_in_previous_step' )

		return

	end
	
	PixelPathProviderSchedulerPrivateMethods_sendRenditionPathToPixelStream( 
		scheduler,  pixStreamId, localAssetId, path, renditionDigest,
		renderSize, croppedLongSide )

	local transition = renditionDigest == currDigest and 
							'sent' or 'stale_data_sent'

	scheduler:_applyTransitionToRequest( request, transition )

end

--------------------------------------------------------------------------------

local function maybeGetBestPath( renditionDiskStatusCache, baseRenderInfoCache, 
									localAssetId, longSide )

	local maybeCroppedLongSide, currDigest =
			maybeGetCurrRenderData( baseRenderInfoCache, localAssetId ) 

	local _, acceptableUpperBound = 
				PixelProviderSharedTaskUtils_getBestPossibleRenderSizeData( 
					maybeCroppedLongSide, longSide )

	local path, renderSize, renditionDigest = 
			renditionDiskStatusCache:maybeGetBestPath( localAssetId,
														acceptableUpperBound )

	return path, renderSize, renditionDigest, maybeCroppedLongSide, currDigest,
			acceptableUpperBound

end

--------------------------------------------------------------------------------

local function sendInitialBestRenditionPathForRequest( 
				request, scheduler, renditionDiskStatusCache,
				baseRenderInfoCache )

	local pixStreamId, localAssetId, longSide = unpack( request )

	local path, renderSize, renditionDigest, maybeCroppedLongSide = 
				maybeGetBestPath( renditionDiskStatusCache,
									baseRenderInfoCache, localAssetId, 
									longSide )

	if path then
		
		PixelPathProviderSchedulerPrivateMethods_sendRenditionPathToPixelStream( 
			scheduler,  pixStreamId, localAssetId, path, renditionDigest,
			renderSize, maybeCroppedLongSide )

	end

	local transition = 'sent'

	if maybeCroppedLongSide then

		local acceptableSizes = 
			PixelProviderSharedTaskUtils_getBestPossibleRenderSizes(
				maybeCroppedLongSide, longSide )

		if not renditionDiskStatusCache:haveCheckedAllRenditionSizes(
					localAssetId, acceptableSizes ) then

			transition = 'need_further_check'

		end

	end

	logger:infof( "sendInitialBestRenditionPathForRequest for %s.%s",
						pixStreamId, localAssetId )

	scheduler:_sendTestData( 'sendInitialBestRenditionPathForRequest',
								pixStreamId, localAssetId, transition )

	scheduler:_applyTransitionToRequest( request, transition )

end

--------------------------------------------------------------------------------

local function sendFinalBestRenditionPathForRequest( 
				request, scheduler, renditionDiskStatusCache, 
				baseRenderInfoCache )

	local pixStreamId, localAssetId, longSide = unpack( request )

	local path, renderSize, renditionDigest, maybeCroppedLongSide = 
				maybeGetBestPath( renditionDiskStatusCache,
									baseRenderInfoCache, localAssetId, 
									longSide )

	if path then
		
		PixelPathProviderSchedulerPrivateMethods_sendRenditionPathToPixelStream( 
			scheduler,  pixStreamId, localAssetId, path, renditionDigest,
			renderSize, maybeCroppedLongSide )

	end

	logger:infof( "sendFinalBestRenditionPathForRequest for %s.%s path %s", pixStreamId, localAssetId, path )

	scheduler:_sendTestData( 'sendFinalBestRenditionPathForRequest',
								pixStreamId, localAssetId, path )

	scheduler:_applyTransitionToRequest( request, 'sent' )

end

--------------------------------------------------------------------------------

local PixelPathProviderSchedulerPrivateMethods_sendLocalBinaryPathToPixelStream =
	PixelPathProviderSchedulerPrivateMethods.sendLocalBinaryPathToPixelStream

--------------------------------------------------------------------------------

local function sendLocalBinaryPathForRequest( request, scheduler, 
												localBinaryPathCache )

	local pixStreamId, localAssetId = unpack( request )

	local maybeNegativePathInfo = 
			localBinaryPathCache:popCacheItem( pixStreamId, localAssetId ) 

	local path, ozBinaryType = unpack( maybeNegativePathInfo or {} )

	if not path then

		logger:warnf( "Can't send local negative path for %s.%s: missing expected data",
						pixStreamId, localAssetId )

		scheduler:_applyTransitionToRequest( request, 'missing_expected_path_computed_in_previous_step' )

		return

	end
	
	logger:infof( "sendLocalBinaryPathForRequest for %s.%s path %s", pixStreamId, localAssetId, path )
	
	PixelPathProviderSchedulerPrivateMethods_sendLocalBinaryPathToPixelStream( 
		scheduler,  pixStreamId, localAssetId, path, ozBinaryType )

	scheduler:_applyTransitionToRequest( request, 'sent' )

end

--------------------------------------------------------------------------------

local function sendNotXmpSidecarForRequest( request, scheduler,
												localBinaryPathCache )

	local pixStreamId, localAssetId = unpack( request )

	local maybeNegativePathInfo =
			localBinaryPathCache:popCacheItem( pixStreamId, localAssetId )

	local path, ozBinaryType = unpack( maybeNegativePathInfo or {} )

	if path ~= nil then
		logger:warnf( "Expecting nil path for %s.%s.", pixStreamId, localAssetId )
	end

	PixelPathProviderSchedulerPrivateMethods_sendLocalBinaryPathToPixelStream(
		scheduler,  pixStreamId, localAssetId, nil, ozBinaryType )

	scheduler:_applyTransitionToRequest( request, 'sent' )

end

--------------------------------------------------------------------------------

local function sendProxyForBestPossibleLocalNegativePath( 
					request, scheduler, localBinaryPathCache,
					baseRenderInfoCache )

	local pixStreamId, localAssetId = unpack( request )

	local maybeNegativePathInfo = 
			localBinaryPathCache:popCacheItem( pixStreamId, localAssetId ) 

	local path, ozBinaryType = unpack( maybeNegativePathInfo or {} )

	if not path then

		logger:warnf( "Can't send local negative path for %s.%s: missing expected data",
						pixStreamId, localAssetId )

		scheduler:_applyTransitionToRequest( request, 'missing_expected_path_computed_in_previous_step' )

		return

	end
	
	PixelPathProviderSchedulerPrivateMethods_sendLocalBinaryPathToPixelStream( 
		scheduler,  pixStreamId, localAssetId, path, ozBinaryType )

	local baseRenderInfo = baseRenderInfoCache:peekCacheItem( localAssetId )

	logger:infof( "sendProxyForBestPossibleLocalNegativePath for %s.%s path %s", pixStreamId, localAssetId, path )
	

	-- we've sent a proxy path, now let's see if we can send an original path
	
	if type( baseRenderInfo ) ~= "table" then

		-- can't figure out if we have original, enter the wait state which will
		-- be updated when docstore is updated
		return scheduler:_applyTransitionToRequest( request, 
											'sent_cannot_download_original' )

	end
	
	if baseRenderInfo.originalPath then

		-- we have a local path to the original, too. transition to a state machine loop that
		-- will check the local original path and send it
		return scheduler:_applyTransitionToRequest( request, 
											'sent_have_local_original' )

	end

	if baseRenderInfo.master then

		-- we have a downloadable original, too. transition to a state machine loop that
		-- will check the download the original 
		return scheduler:_applyTransitionToRequest( request, 
											'sent_can_download_original' )

	end

	-- No original available, enter the wait state which will be updated when
	-- docstore is updated. So if Oz gets the original at some point, we
	-- we get a docstore update which will add a original download URL that will
	-- in turn trigger the download
	return scheduler:_applyTransitionToRequest( request, 
										'sent_cannot_download_original' )

end

--------------------------------------------------------------------------------

local function sendErrorForRequest( request, scheduler,
										activeRenditionInfoCache,
										localBinaryPathCache )

	local pixStreamId, localAssetId, _, _, errorState = unpack( request )

	localBinaryPathCache:destroyCacheItem( pixStreamId, localAssetId )

	activeRenditionInfoCache:destroyCacheItem( pixStreamId, localAssetId )

	logger:infof( "Sending error for %s.%s: %s", pixStreamId, localAssetId, 
													errorState )
	
	PixelPathProviderSchedulerPrivateMethods.sendErrorToPixelStream( 
		scheduler, pixStreamId, localAssetId, errorState )

	scheduler:_applyTransitionToRequest( request, 'sent' )

end


--------------------------------------------------------------------------------

local _haveLocalBinaryStates = 
	PixelProviderSharedTaskUtils_makeNonSparseStringArray(
		SendDataStates.HaveLocalNegativePath,
		SendDataStates.HaveLocalXmpDevelopPath,
		SendDataStates.HaveLocalXmpSidecarPath,
		SendDataStates.HaveOriginalForBestPossibleLocalNegativePath
	)

--------------------------------------------------------------------------------

local _sendErrorStates =
	PixelProviderSharedTaskUtils_makeNonSparseStringArray(
		SendDataStates.SendNoDownloadableRenderOrBinary,
		SendDataStates.SendNoDownloadableOriginal,
		SendDataStates.SendNoDownloadableBinary,
		SendDataStates.SendNoDownloadableBinaryForRender,
		SendDataStates.SendNoDownloadableXmpDevelop,
		SendDataStates.SendNoDownloadableXmpSidecar,
		SendDataStates.SendPermanentDownloadError,
		SendDataStates.SendPermanentError,
		SendDataStates.SendPermanentBadAssetIdError,
		SendDataStates.SendPermanentDownloadErrorForBestPossibleLocalNegativePath,
		SendDataStates.SendPermanentErrorForBestPossibleLocalNegativePath )

--------------------------------------------------------------------------------

local function addOrphanedStates( orphanedSendDataStateCounts, request )

	local _, _, _, _, state = unpack( request )

	orphanedSendDataStateCounts[ state ] =
		( orphanedSendDataStateCounts[ state ] or 0 ) + 1

	return orphanedSendDataStateCounts

end

--------------------------------------------------------------------------------

local function processRequests( scheduler, 
									renditionDiskStatusCache,
									activeRenditionInfoCache,
									baseRenderInfoCache,
									localBinaryPathCache )

	-- send rendition paths
	scheduler:_mapAllRequestsWithState( SendDataStates.HaveRenditionPath,
			sendRenditionPathForRequest,
				scheduler, renditionDiskStatusCache,
				baseRenderInfoCache )

	scheduler:_mapAllRequestsWithState( SendDataStates.SendInitialBestRenditionPath,
			sendInitialBestRenditionPathForRequest,
				scheduler, renditionDiskStatusCache,
				baseRenderInfoCache )

	scheduler:_mapAllRequestsWithState( SendDataStates.SendFinalBestRenditionPath,
			sendFinalBestRenditionPathForRequest,
				scheduler, renditionDiskStatusCache,
				baseRenderInfoCache )

	-- send local negative paths
	scheduler:_mapAllRequestsWithStates( _haveLocalBinaryStates,
			sendLocalBinaryPathForRequest,
				scheduler, localBinaryPathCache )

	scheduler:_mapAllRequestsWithState( SendDataStates.NotXmpSidecar,
			sendNotXmpSidecarForRequest,
				scheduler, localBinaryPathCache )

	scheduler:_mapAllRequestsWithState( SendDataStates.HaveProxyForBestPossibleLocalNegativePath,
			sendProxyForBestPossibleLocalNegativePath,
				scheduler, localBinaryPathCache, baseRenderInfoCache )
				
	scheduler:_mapAllRequestsWithStates( _sendErrorStates,
			sendErrorForRequest,
				scheduler, activeRenditionInfoCache, localBinaryPathCache )

	if logger.will_warn then

		local sendDataStateArray =
				KSTableUtils.extractKeys( SendDataStates )

		local orphanedSendDataStateCounts = 
				PixelPathProviderSchedulerPrivateMethods.reduceAllRequestsWithStates( 
							scheduler, {}, sendDataStateArray, addOrphanedStates )

		for state, count in pairs( orphanedSendDataStateCounts ) do

			-- This warning fires if we have a state in the SendDataStates
			-- state group that does not send data in the above code.
			--
			-- This means if code is waiting for a request to send data, it
			-- will hang forever.
			logger:warnf( "There are %s orphaned %s requests, data will never be sent!",
							count, state )

		end

	end

end

--==============================================================================

return {

	processRequests = processRequests,

}
