--[[----------------------------------------------------------------------------

PixelProviderSharedTaskUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- pixel provider
local WFCatRenderSize = require 'WFCatRenderSize'
local PixelProviderTaskCache = require 'PixelProviderTaskCache'
local WFFunctionUtils = require 'WFFunctionUtils'
local WFCatUser = require 'WFCatUser'
local WFCatSpec = require 'WFCatSpec'
local WFCatalog = require 'WFCatalog'

--------------------------------------------------------------------------------

local WFFunctionUtils_reduce = assert( WFFunctionUtils.reduce,
			"WFFunctionUtils.reduce does not exist" )

--------------------------------------------------------------------------------
-- key:
--
-- 	localAssetId
--
-- value:
-- 
-- { ozAssetId = assetId,
-- 		changeCounter = newChangeCounter,
-- 		revId = newRevId,
-- 		digest = newDigest,
-- 		croppedLongSide = newCroppedLongSide,
-- 		originalPath = newOriginalPath,
-- 		proxyPath = newProxyPath, 
-- 		renditions = newRenditions }
--
-- lifetime:
--
--  stays in cache as long as there is at least one localAssetId request in
--  scheduler queue
--------------------------------------------------------------------------------
	
local function makeBaseRenderInfoCache()

	return PixelProviderTaskCache.makeLocalAssetSimpleDataCache()

end

--------------------------------------------------------------------------------
-- key:
--
-- 	localAssetId
--
-- value:
--
-- 	current dev settings for localAssetId
--
-- lifetime:
--
--  stays in cache as long as there is at least one localAssetId request in
--  scheduler queue.
--
--  TODO: bkruse 2015.11.03 we should consider refining this lifetime
-- 
--------------------------------------------------------------------------------
local function makeCurrRenderSettingsCache()

	return PixelProviderTaskCache.makeLocalAssetSimpleDataCache()

end

--------------------------------------------------------------------------------
-- key:
--
-- 	pixStream, localAssetId
--
-- value:
--
-- 	{ renditionPath, digest, renderSize, croppedLongSide, 
-- 		maybeDevSettings }
--
-- lifetime:
--
--  stays in cache until we send the renditionPath or enter one of EndStates
--  for pixStreamId, localAssetId
-- 
--------------------------------------------------------------------------------
local function makeActiveRenditionInfoCache()

	return PixelProviderTaskCache.makePerClientLocalAssetSimpleDataCache()

end

--------------------------------------------------------------------------------
-- key:
--
-- 	pixStream, localAssetId
--
-- value:
--
-- 	{ binaryPath, binarySize }
--
-- lifetime:
--
--  stays in cache until we send the binaryPath or enter one of EndStates
--  for pixStreamId, localAssetId
-- 
--------------------------------------------------------------------------------
local function makeLocalBinaryPathCache()

	return PixelProviderTaskCache.makePerClientLocalAssetSimpleDataCache()

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the proxy sub-type matching the asset info provided.

	\param baseRenditionInfo (table)

	\return proxy sub-type (string) Can be any of the following
		"2560"
		"720p"

	Currently it doesn't return 'panorama' in any case. This logic
	will have to evolve to take into account 'panorama' and any
	other case which is not handled right now.
]]--

local function getBinarySubtypeForProxy( baseRenditionInfo )

	-- TODO bkruse 2016.03.14 
	--
	-- what about panoramas?
	local isVideo = baseRenditionInfo and baseRenditionInfo.subtype == 'video' or false

	return isVideo and '720p' or '2560'
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the sub-type matching the binary type
	and asset info provided.

	\param localAssetId (string)

	\param ozBinaryType (string) Can be any of the
	following types
		"original"
		"proxy"
		"xmp_develop"

	\param baseRenditionInfoCache (table)

	\return binary sub-type (string) 

]]--

local function getBinarySubtypeForBinaryType( localAssetId, 
												ozBinaryType,
												baseRenderInfoCache )

	if ozBinaryType == 'xmp_develop' then

		return 'xmp_develop'

	end
	
	if ozBinaryType == 'xmp_sidecar' then

		return 'xmp_sidecar'

	end

	if ozBinaryType == 'original' then

		return 'master'

	end

	local baseRenditionInfo = baseRenderInfoCache:peekCacheItem( localAssetId )

	return getBinarySubtypeForProxy( baseRenditionInfo )

end

--------------------------------------------------------------------------------

local WFCatRenderSize_defaultRenderSizes = WFCatRenderSize.defaultRenderSizes 

--------------------------------------------------------------------------------

local WFCatRenderSize_getBoundedRenderSize =
		WFCatRenderSize.getBoundedRenderSize

--------------------------------------------------------------------------------

local function getBestPossibleRenderSizeData( maybeCroppedLongSide, 
														longSide )

	local targetRenderSize = 
			type( maybeCroppedLongSide ) == "number" and
				WFCatRenderSize_getBoundedRenderSize( maybeCroppedLongSide, 
														longSide ) or
				longSide

	if type( targetRenderSize ) == "number" then

		local upperBound = 4 * targetRenderSize

		local numDefaults = #WFCatRenderSize_defaultRenderSizes

		if upperBound >= WFCatRenderSize_defaultRenderSizes[ 1 ] and
				upperBound <= WFCatRenderSize_defaultRenderSizes[ numDefaults ] then

			local normalizedUpperBound = WFCatRenderSize_defaultRenderSizes[ 1 ] 

			for i = 2, numDefaults do

				if upperBound < WFCatRenderSize_defaultRenderSizes[ i ] then

					return targetRenderSize, normalizedUpperBound

				end

				normalizedUpperBound = WFCatRenderSize_defaultRenderSizes[ i ] 

			end

		end

	end

	return longSide, "full"

end

--------------------------------------------------------------------------------

local function maybeAddSuitableSize( suitableSizes, size, wantedSize, 
										acceptableUpperBound )

	if size == wantedSize then

		-- should already be at front of suitableSizes
		return suitableSizes

	end

	if size < acceptableUpperBound then

		suitableSizes[ #suitableSizes + 1 ] = size

	end

	return suitableSizes

end

--------------------------------------------------------------------------------

local function getBestPossibleRenderSizes( croppedLongSide, longSide )

	local wantedRenderSize, acceptableUpperBound  = 
			getBestPossibleRenderSizeData( croppedLongSide, longSide )

	if acceptableUpperBound == "full" then

		return WFCatRenderSize_defaultRenderSizes

	end

	return WFFunctionUtils_reduce( 
			{ wantedRenderSize }, 
				maybeAddSuitableSize,
					WFCatRenderSize_defaultRenderSizes, 
						wantedRenderSize, acceptableUpperBound )

end

--------------------------------------------------------------------------------

local function makeNonSparseStringArray( ... )

	local ret = {}

	for i = 1, select( '#', ... ) do

		local maybeStr = select( i, ... )


		if type( maybeStr ) ~= "string" then

			error( string.format( 
					"parameter at position %s is not a string", i ) )

		end

		ret[ #ret + 1 ] = maybeStr

	end

	return ret

end

--------------------------------------------------------------------------------

local function makeWFCatalog( serverBaseUrl, userId, userInfoCachePath,
								managedCatalogIdOrDoc, rootCatalogDataDir,
								catalogOptions )

	local wfCatUser = WFCatUser( serverBaseUrl, userId, userInfoCachePath )

	local wfCatSpec = WFCatSpec( wfCatUser, managedCatalogIdOrDoc )

	return WFCatalog( wfCatSpec, rootCatalogDataDir, catalogOptions )

end

--==============================================================================

return {

	makeBaseRenderInfoCache = makeBaseRenderInfoCache,

	makeCurrRenderSettingsCache = makeCurrRenderSettingsCache,

	makeActiveRenditionInfoCache = makeActiveRenditionInfoCache,

	makeLocalBinaryPathCache = makeLocalBinaryPathCache,

	getRenderSize = WFCatRenderSize_getBoundedRenderSize,

	getBestPossibleRenderSizeData = getBestPossibleRenderSizeData,

	getBestPossibleRenderSizes = getBestPossibleRenderSizes,

	getBinarySubtypeForBinaryType = getBinarySubtypeForBinaryType,

	getBinarySubtypeForProxy = getBinarySubtypeForProxy,

	makeNonSparseStringArray = makeNonSparseStringArray,
	
	makeWFCatalog = makeWFCatalog,
}

