--[[----------------------------------------------------------------------------

WFCatAlbums.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'
local WFSortOrderStrings = require 'WFSortOrderStrings'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCatCollaboration = require 'WFCatCollaboration'
local WFCatIdUtils = require 'WFCatIdUtils'
local WFCatOzErrors = require 'WFCatOzErrors'
local WFAlbumOrdering = require 'WFAlbumOrdering'
local WFRxBatchedBulkApplyOverlay = require 'WFRxBatchedBulkApplyOverlay'
local WFRxBatchedBulkSetDocValue = require 'WFRxBatchedBulkSetDocValue'
local WFLogger = require 'WFLogger'
local WFTableMerge = require 'WFTableMerge'
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.oz_client.managed_catalog
local WFCatDelete = require 'WFCatDelete'
local WFCatUtils = require 'WFCatUtils'
local WFCatSyncStatus = require 'WFCatSyncStatus'

--------------------------------------------------------------------------------

local logger = WFLogger( 'WFCatAlbums' )

--------------------------------------------------------------------------------

local OZ_MAX_ALBUM_NAME_CHARS = 250

local ROOT_ALBUM_PARENT_ID = 'root'

local SORT_PREV_FOR_INSERT_AS_FIRST = ''

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local derefOpt = assert( KSTableUtils.derefOpt )

local runUntilDone = WFRxRunUntilDone
	
--============================================================================--

-- @mark WFCatAlbums namespace

local WFCatAlbums = {}

--------------------------------------------------------------------------------

local function hasParent( query, match, ... )
	if not match then
		return false
	end
	if query == match then
		return true
	end
	return hasParent( query, ... )
end

--------------------------------------------------------------------------------

local function findCycles( rows )

	local allParents = {}
		-- each entry is an array of all the parents for that album
		
	local cycles = {}
		-- array of arrays, each containing the docIds of albums in a cycle

	for _, row in ipairs( rows ) do
	
		local parentId = row[ 7 ]
		local docId = row[ 1 ]
		
		if parentId then
		
			local parentParents = allParents[ parentId ]
			local myParents = { parentParents and unpack( parentParents ) }
			
			myParents[ #myParents + 1 ] = parentId
			
			if hasParent( docId, unpack( myParents ) ) then
				cycles[ #cycles + 1 ] = myParents
			end
			
			allParents[ docId ] = myParents
			
		end
		
	end

	return cycles
	
end

--------------------------------------------------------------------------------

local function breakCycles( rows )

	local cycles = findCycles( rows )

	-- If any loops are found (usually none will be) we
	-- need to break them

	if next( cycles ) then
	
		local rowsById = {}
		
		for _, row in ipairs( rows ) do
			rowsById[ row[ 1 ] ] = row
		end
		
		for _, docIds in ipairs( cycles ) do
		
			table.sort( docIds, function( a, b )
				local rowA = rowsById[ a ]
				local rowB = rowsById[ b ]
				local tieBreakerA = rowA[ 5 ] or rowA[ 6 ] or rowA[ 1 ]
				local tieBreakerB = rowB[ 5 ] or rowB[ 6 ] or rowB[ 1 ]
				return tieBreakerA < tieBreakerB
			end )
			
			-- Nil out the parent field on the oldest node
			rowsById[ docIds[ 1 ] ][ 7 ] = nil
			
		end
		
	end
	
end

WFCatAlbums._breakCycles = breakCycles
	-- For unit tests

--------------------------------------------------------------------------------

local function createAlbumAssetDocument( albumId, assetId )

	local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( KSIdGenerator.uuidInLowercaseHex() )

	local albumAssetDocValue = {
		type = 'album_asset',
		album = {
			id = albumId,
		},
		asset = {
			id = assetId,
		},
	}
	
	WFTableMerge.applyOverlay( albumAssetDocValue, baseDocValue )
	
	return albumAssetDocValue
	
end

WFCatAlbums._createAlbumAssetDocument = createAlbumAssetDocument

--------------------------------------------------------------------------------

local function isFolder( docSubtype )

	return docSubtype == 'collection_set'
	
end

--------------------------------------------------------------------------------

local function basicAlbumsList( wfCatalog, quiesceUntilUpToDate )
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT albums.docId, albumId, name, updateTimestamp, localTimestamp, parentId, sortOrder, subtype, hidden
				FROM albums]],
		quiesceUntilUpToDate = quiesceUntilUpToDate,
	}
	:map( function( rows )
		
		local albums = {}
		local docIdsToLocalId = {}
		
		for i, row in ipairs( rows ) do
			
			local album = {}
			album.id = row[ 1 ]
			album.fullDocId = row[ 2 ]
			album.name = row[ 3 ]
			album.updateTimestamp = row[ 4 ]
			album.localTimestamp = row[ 5 ]
			album.parentDocId = row[ 6 ]
			album.sortOrder = row[ 7 ]
			album.subtype = isFolder( row[ 8 ] ) and row[ 8 ] or nil
			album.hidden = row[ 9 ] ~= nil and row[ 9 ] ~= 0
			
			docIdsToLocalId[ album.fullDocId ] = album.id 
			
			albums[ #albums + 1 ] = album
		end
	
		return albums, docIdsToLocalId
	end )
end

--------------------------------------------------------------------------------

local function addAlbumIntoHierarchy( album, hierarchy )
	
	local albums = hierarchy.albums or {}
	local localIdtoAlbumData = hierarchy.localIdtoAlbumData or {}
	local docIdToAlbumData = hierarchy.docIdToAlbumData or {}
	local localIdToChildrenAlbumData = hierarchy.localIdToChildrenAlbumData or {}
	
	albums[ #albums + 1 ] = album
	
	localIdtoAlbumData[ album.id ] = album
	
	if album.fullDocId then
		docIdToAlbumData[ album.fullDocId ] = album
	end
		
	local parentId = album.parentId or ROOT_ALBUM_PARENT_ID
		
	local albumsInParent = localIdToChildrenAlbumData[ parentId ] or {}
	albumsInParent[ #albumsInParent + 1 ] = album
	
	localIdToChildrenAlbumData[ parentId ] = albumsInParent
	
	hierarchy.albums = albums
	hierarchy.localIdtoAlbumData = localIdtoAlbumData
	hierarchy.localIdToChildrenAlbumData = localIdToChildrenAlbumData
	hierarchy.docIdToAlbumData = docIdToAlbumData
	
	
	return hierarchy
end

WFCatAlbums._addAlbumIntoHierarchy = addAlbumIntoHierarchy

--------------------------------------------------------------------------------

local function reparentAlbumInHierarchy( hierarchy, albumId, optNewParentId, optOrder )
	
	local album = WFCatAlbums._getAlbumFromHierarchyById( hierarchy, albumId )
	local parent = WFCatAlbums._getAlbumFromHierarchyById( hierarchy, optNewParentId )
	
	local oldParentId = album and album.parentId or 'root'
	local newParentId = optNewParentId or 'root'
		
	if album and oldParentId ~= newParentId then 
		-- if we have an album and the parent has changed, move it around in the cache

		-- remove from old parent in cache
		local oldSiblings = hierarchy.localIdToChildrenAlbumData[ oldParentId ] or {}
		for i = #oldSiblings, 1, -1 do -- iterate backwards so we can use table.remove safely
			local sibling = oldSiblings[ i ]
			if sibling.id == album.id then
				table.remove( oldSiblings, i )
				break 
			end
		end
		hierarchy.localIdToChildrenAlbumData[ oldParentId ] = oldSiblings
		
		-- switch album's parent id
		if parent then
			album.parentId = optNewParentId
		else
			album.parentId = nil
		end
		
		-- add to new parent in cache
		local newSiblings = hierarchy.localIdToChildrenAlbumData[ newParentId ] or {}
		newSiblings[ #newSiblings + 1] = album
		hierarchy.localIdToChildrenAlbumData[ newParentId ] = newSiblings
		
	end
	
	return hierarchy

end

WFCatAlbums._reparentAlbumInHierarchy = reparentAlbumInHierarchy

--------------------------------------------------------------------------------

local function deleteAlbumFromHierarchy( hierarchy, albumId )

	local album = WFCatAlbums._getAlbumFromHierarchyById( hierarchy, albumId )
	
	local idsRemoved = {}
	if album then
		local children = {}
		WFCatAlbums._addAllChildrenOfAlbumTo( hierarchy, album, children )
		
		local idsToRemove = {}
		idsToRemove[ album.id ] = true
		for _, child in ipairs( children ) do
			idsToRemove[ child.id ] = true
		end
		
		local allAlbums = hierarchy.albums
		
		for i= #allAlbums, 1, -1 do
		
			local anAlbum = allAlbums[ i ]
			if idsToRemove[ anAlbum.id ] then
			
				table.remove( allAlbums, i )
				hierarchy.localIdToChildrenAlbumData[ anAlbum.id ] = nil
				hierarchy.localIdtoAlbumData[ anAlbum.id ] = nil
				
				if anAlbum.fullDocId then
					hierarchy.docIdToAlbumData[ anAlbum.fullDocId ] = nil
				end
				
				idsRemoved[ #idsRemoved + 1 ] = anAlbum.id
			end
		end
	end
	
	return hierarchy, idsRemoved
	
end

WFCatAlbums._deleteAlbumFromHierarchy = deleteAlbumFromHierarchy

--------------------------------------------------------------------------------

-- return observable containing the rationalized structure of the
-- album index, breaking any loops in hierarchy based on the oldest-modified
-- album is promoted to root when a loop is introduced
-- Pretty much all album APIs operate by computing the full album structure
-- first, and then doing simple lua-data structure interrogation.
local function flattenedAlbumHierarchy( wfCatalog, quiesceUntilUpToDate )

	return basicAlbumsList( wfCatalog, quiesceUntilUpToDate )
		:map( function( albums, docIdsToLocalId )
	
			local flattenedHierarchy = {
				albums = {},
				localIdToChildrenAlbumData = {},
				localIdtoAlbumData = {},
				docIdToAlbumData = {},
			}
			
			-- SPAM [matthewj 2017-01-27]: Need to properly implement breakCycles for albums
			-- not rows (and it doesn't really work for cycles > 2 nodes currently anyway)
			-- breakCycles( albums )
	
			for _, album in ipairs( albums ) do
				
				album.parentId = docIdsToLocalId[ album.parentDocId ] 
				album.parentDocId = nil -- this just cleans up the full doc id of the parent in 
										-- the album record. passing only local ids to the UI
							
				addAlbumIntoHierarchy( album, flattenedHierarchy )
		
			end
	
			return flattenedHierarchy
	
		end )
end

WFCatAlbums._flattenedAlbumHierarchy = flattenedAlbumHierarchy

--------------------------------------------------------------------------------

local function getAlbumFromHierarchyById( hierarchy, albumId )
	
	local album
	if type( albumId ) == 'number' then
		album = hierarchy.localIdtoAlbumData[ albumId ]
	else
		album = hierarchy.docIdToAlbumData[ albumId ]
	end

	return album
end

WFCatAlbums._getAlbumFromHierarchyById = getAlbumFromHierarchyById

--------------------------------------------------------------------------------

local function addAllChildrenOfAlbumTo( hierarchy, album, albumList )
	
	local children = hierarchy.localIdToChildrenAlbumData[ album.id ] or {}
	
	for _, child in ipairs( children ) do
		albumList[ #albumList + 1  ] = child
		addAllChildrenOfAlbumTo( hierarchy, child, albumList )
	end
end

WFCatAlbums._addAllChildrenOfAlbumTo = addAllChildrenOfAlbumTo

--------------------------------------------------------------------------------

local function getAlbumAndChildren( wfCatalog, albumId )

	return flattenedAlbumHierarchy( wfCatalog, true )
		:take( 1 )
		:map( function( hierarchy )
			
			
			local album = getAlbumFromHierarchyById( hierarchy, albumId )
			
			local albumAndChildren = {}
			
			if album then
				albumAndChildren[ #albumAndChildren + 1 ] = album
				
				addAllChildrenOfAlbumTo( hierarchy, album, albumAndChildren )
			end
				
			return albumAndChildren
		end )
		:mapSwitch( function( array )
			return Observable.from( array )
		end )

end

--------------------------------------------------------------------------------

local function isAlbumNameUniqueAmongSiblings( albumName, albumSubtype, siblings, optIgnoreAlbumId )
	
	local nameIsUnique = true
	local lcAlbumName = KSStringUtils.lower( albumName )
	
	for i, sibling in ipairs( siblings or {} ) do

		-- if optIgnoreAlbumId is given, don't check for name collisions with it

		if optIgnoreAlbumId == nil or sibling.id ~= optIgnoreAlbumId then

			if KSStringUtils.lower( sibling.name ) == lcAlbumName 
				and sibling.subtype == albumSubtype then
		
				nameIsUnique = false
				break
			end
		end
	end

	return nameIsUnique

end

--------------------------------------------------------------------------------

local function maybeMakeAlbumNameValid( albumName )

	if string.len( albumName ) > OZ_MAX_ALBUM_NAME_CHARS then
		logger:warnf( 'Album names are limited to %d chars, truncating: %s', OZ_MAX_ALBUM_NAME_CHARS, albumName )
		
		albumName = KSStringUtils.truncateUTF8( albumName, OZ_MAX_ALBUM_NAME_CHARS )
	end
	
	return albumName

end

--------------------------------------------------------------------------------

local function isValidParent( album )

	if not album then
		return true -- assumes parent will be converted to root, 
					-- if no parent actually exists
	end
	
	if album.subtype == 'collection_set' then
		return true
	end
	
	return false
end

--------------------------------------------------------------------------------

function WFCatAlbums.getAlbumCounts( wfCatalog )
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT albumId, 
					( SELECT COUNT( DISTINCT album_asset_v2.assetId )
							FROM album_asset_v2 
							INNER JOIN assets USING( assetId )
							WHERE albumId = a.albumId
								AND assets.isMedia = 1 )
					FROM albums as a
					WHERE a.subtype = 'collection']],
			calm = 2,
		}
		:distinctUntilChanged()
		:map( function( albumAssetCounts )
			
			local albumCounts = {}
			
			for _, albumAsset in ipairs( albumAssetCounts ) do
			
				local albumId = albumAsset[ 1 ]
				local assetCount = albumAsset[ 2 ]
				albumCounts[ albumId ] = assetCount
	
			end
			
			return albumCounts
		
		end )

end

--------------------------------------------------------------------------------

function WFCatAlbums.getAlbumCovers( wfCatalog )

    WFArguments.check( 'WFCatalog' )
	
	local explicitCoversQuery = wfCatalog:indexStore():query {
            sql = [[SELECT albums.docId, albums.albumId, assets.docId
                        FROM albums 
                        LEFT JOIN assets
                        ON albums.coverId == assets.assetId ]],
            calm = 0.5
        }
		
	local implicitCoversSortOrderQuery = wfCatalog:indexStore():query {
		sql = [[SELECT albumId, 
					( SELECT assets.docId 
							FROM assets 
							INNER JOIN album_asset_v2 USING( assetId )
							WHERE albumId = a.albumId
								AND assets.isMedia = 1
								AND sortOrder > ''
							ORDER BY sortOrder ASC )
					FROM albums as a
					WHERE a.subtype = 'collection']],
		calm = 2,
	}
	:map( function( rows )
		local implicitCoversSortOrderByAlbumId = {}
		
		for _, row in ipairs( rows ) do
			local albumId = row[ 1 ]
			local implicitCoverId = row[ 2 ]
			implicitCoversSortOrderByAlbumId[ albumId ] = implicitCoverId
		end
		
		return implicitCoversSortOrderByAlbumId
	end )
	
	local implicitCoversCaptureDateQuery = wfCatalog:indexStore():query {
		sql = [[SELECT albumId, 
					( SELECT assets.docId 
							FROM assets 
							INNER JOIN album_asset_v2 USING( assetId )
							WHERE albumId = a.albumId
								AND assets.isMedia = 1
							ORDER BY captureDate ASC )
					FROM albums as a
					WHERE a.subtype = 'collection']],
		calm = 2,
	}
	:map( function( rows )
		local implicitCoversCaptureDateByAlbumId = {}
		
		for _, row in ipairs( rows ) do
			local albumId = row[ 1 ]
			local implicitCoverId = row[ 2 ]
			implicitCoversCaptureDateByAlbumId[ albumId ] = implicitCoverId
		end
		
		return implicitCoversCaptureDateByAlbumId
	end )
		
	return Observable.combineLatest(
		explicitCoversQuery,
		implicitCoversSortOrderQuery,
		implicitCoversCaptureDateQuery )
		:map( function( explicitCoversRows, implicitCoversSortOrderByAlbumId, implicitCoversCaptureDateByAlbumId )
			local covers = {}
			
			for _, row in ipairs( explicitCoversRows ) do
				local albumDocId = row[ 1 ]
				local albumId = row[ 2 ]
				local explicitCoverId = row[ 3 ]
				
				-- prefer explicit cover id
				local coverId = explicitCoverId
				
				-- then implicit by sort order
				if not coverId then
					coverId = implicitCoversSortOrderByAlbumId[ albumId ]
				end
				
				-- then implicit by capture date
				if not coverId then
					coverId = implicitCoversCaptureDateByAlbumId[ albumId ]
				end
				
				covers[ albumDocId ] = coverId
				
			end
			
			return covers
		end )
		:distinctUntilChanged()
   
end

--------------------------------------------------------------------------------

--[[
	\brief Fetches the list of albums

	\param wfCatalog (WFCatalog) The catalog.

	\return <code>Observable< ></code> An Observable object sends a value whenever
		the albums change.  The result is an array of album nodes, each containing
		the id (as a local integer id), the album name, the parent album (if any),
		subtype, and cover
--]]

function WFCatAlbums.getAlbumsList( wfCatalog )

	return flattenedAlbumHierarchy( wfCatalog, false )
		:pluck( 'albums' )

end

--------------------------------------------------------------------------------

local function reorderNewAlbumAndSiblings( newAlbumId, newAlbumName, siblings )
	
	-- compute order field
	local ordering = {}

	for i, album in ipairs( siblings or {} ) do
		ordering[ i ] = {
			id = album.id,
			-- in this case it's the local id, but that's ok
			touched = false,
			order = album.sortOrder,
		}
	end

	ordering[ #ordering + 1 ] = {
		id = newAlbumId,
			-- in this case it's the uuid but that's ok too
		touched = true,
	}

	local updatedOrders = WFAlbumOrdering.adjustOrderFields( ordering )
	
	local newAlbumOrder
	local overlays = {}
	for i, updatedOrder in pairs( updatedOrders ) do
		if ordering[ i ].id == newAlbumId then
			newAlbumOrder = updatedOrder
		else
		
			local overlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( ordering[ i ].id )
			overlay.order = updatedOrder
			overlays[ #overlays + 1 ] = overlay
			
		end
	end
	
	return newAlbumOrder, overlays
end

--------------------------------------------------------------------------------

local function createAlbum_imp( wfCatalog, albumSubtype, albumName, optParentId, optHidden, optImportSource, optDocId, optManifestSourceId )
	
	albumName = maybeMakeAlbumNameValid( albumName )
	
	local docStore = wfCatalog:docStore()
	
	local maybeGetParentDoc = Observable.just( {} )
	
	if optParentId then
	
		maybeGetParentDoc = docStore:docValue( optParentId )
	
	end
	
	return maybeGetParentDoc
		:take( 1 )
		:flatMapLatest( function( parentDoc )

			local optParentDocId = parentDoc._id

			if optParentId and not isValidParent( parentDoc ) then 
				return Observable.throw( 'InvalidParent' )
			end

			local albumId = optDocId or KSIdGenerator.uuidInLowercaseHex()
			local updateTimestamp = KSDate.timeInIso8601Format( nil, false )
			
			local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( albumId )

			local albumDocValue = {
				type = 'album',
				subtype = albumSubtype,
				name = albumName,
				hideInOrganize = optHidden, -- use 'not not' if we want ths field to always exist
				parent = optParentDocId and {
					id = optParentDocId,
				},
				importSource = optImportSource,
				_localOnly = {
					updateTimeStamp = updateTimestamp,
					manifest = optManifestSourceId and {
						[ optManifestSourceId ] = true
					},
				},
			}
	
			WFTableMerge.applyOverlay( albumDocValue, baseDocValue )
	
			local result = docStore:setDocValue( albumDocValue )
	
			-- Wait until all of the above operations have completed,
			-- then extract the local doc ID form doc store.
	
			return result
				:takeLast( 0 )
				:concat( docStore:docValue( albumId ) )
				:map( function( docValue )
					return docValue._localId, albumId, docValue
				end )

	end )
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Creates an album in the provided catalog.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumName (string) The name of the new album, will truncate to 250 chars as Oz does not allow longer names.
	\param optParentId (string|int|nil) The id of the parent.
	\param optHidden (boolean) true if this album should be hidden from the albums list
	\param optUpdateTimestamp (iso8601date|nil) The date to put in _localOnly.updateTimeStamp.
	\param optDocId (string) The id of the new document if we already know what it should be
	\param optManifestSourceId (string) The id of the source in case request is coming via manifest processing

	\return <code>Observable< ></code> An Observable object that signals the albumId
		and then completes when operation is completed; or resolves to an error
		if an error occurs.
]]--
function WFCatAlbums.createAlbum( wfCatalog, albumName, optParentId, optHidden, optImportSource, optDocId, optManifestSourceId )

	check( 'WFCatalog', '+string', '?+string|number', '?boolean', "?table", "?uuid", '?+string' )
	
	return createAlbum_imp( wfCatalog, 'collection', albumName, optParentId, optHidden, optImportSource, optDocId, optManifestSourceId )
end

--------------------------------------------------------------------------------

--[[!
	\brief Creates a folder (album of subtype 'collection_set') in the provided catalog.

	\param wfCatalog (WFCatalog) The catalog.
	\param folderName (string) The name of the new folder, will truncate to 250 chars as Oz does not allow longer names.
	\param optParentId (string|int|nil) The id of the parent.
	\param optUpdateTimestamp (iso8601date|nil) The date to put in _localOnly.updateTimestamp.
	\param optDocId (string) The id of the new document if we already know what it should be
	\param optManifestSourceId (string) The id of the source in case request is coming via manifest processing

	\return <code>Observable< ></code> An Observable object that signals the albumId
		and then completes when operation is completed; or resolves to an error
		if an error occurs.
]]--
function WFCatAlbums.createFolder( wfCatalog, folderName, optParentId, optImportSource, optDocId, optManifestSourceId )
	
	check( 'WFCatalog', '+string', '?+string|number', "?table", "?uuid", '?+string' )
	
	return createAlbum_imp( wfCatalog, 'collection_set', folderName, optParentId, false, optImportSource, optDocId, optManifestSourceId )

end

--------------------------------------------------------------------------------

--[[!
	\brief Deletes an album or folder, but ignores all child album/folders. This should generally
		be used only internally unless you really know you want it.  WFCatAlbums.deleteAlbumOrFolder
		is probably more appropriate for most uses.
		
	\param wfCatalog (WFCatalog) The catalog
	\param albumDocId (string) the UUID of the album/folder to delete.
	
	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatAlbums.deleteIndividualAlbum( wfCatalog, albumDocId )
	check( "WFCatalog", "+string" )
	
	-- PR #3312056: Remove any album share first, then delete the album, too.
	-- Regardless of success/failure of unsharing the album, we're still going
	-- to delete it, hence the catch()
	
	return WFCatCollaboration.deleteSharedAlbum( wfCatalog, albumDocId )
				:catch( Observable.empty() )
				:concat( Observable.just( true ) )
				:takeLast( 1 )

					-- Remove all album assets from the album
				:flatMapLatest( function()
					return wfCatalog:indexStore():query {
						-- get all the album_assets
						sql = [[SELECT docId FROM album_asset_v2 WHERE albumId = ?]],
						albumDocId,
						quiesceUntilUpToDate = true,
					}
					:take( 1 )
				end )
				:flatMapLatest( function( rows )
					
					return WFCatDelete.deleteDocsFromQueryResult( wfCatalog:docStore(), rows, true )
					
				end )
				:takeLast( 1 )

				:flatMapLatest( function()
					return wfCatalog:docStore():deleteDoc( albumDocId )
					:takeLast( 0 )
				end )
end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes an album/folder and all child albums/folders from this catalog.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string|int) The unique or local ID of the album/folder to delete.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--

function WFCatAlbums.deleteAlbumOrFolder( wfCatalog, albumId )

	check( "WFCatalog", '+string|number' )

	if type( albumId ) == 'string' and #albumId < 32 then
		return Observable.throw( "WFCatAlbums:deleteAlbumOrFolder: The albumId must be a 32-char uuid", 2 )
	end

	return getAlbumAndChildren( wfCatalog, albumId )
		:selectMany( function( album )
		
			local albumDocId = album.fullDocId
			return WFCatAlbums.deleteIndividualAlbum( wfCatalog, albumDocId )
		end )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes only the albums/folders specified by id in the params.  It will not 
		remove any children from this catalog that are not explicitly specified in the list
		of albums to remove so you should fully consider
	
	\param wfCatalog (WFCatalog) The catalog.
	\param albumIds (table) Table of local ids of the albums and/or folders to delete.
	
	\return <code>Observable<></code> An Observable object completes when the delete
		operation is completed; or resolves to an error if an error occurs.
]]--
function WFCatAlbums.deleteIndividualAlbums( wfCatalog, folderAndAlbumIds )
	
	check( "WFCatalog", "table" )
	
	return Observable.from( folderAndAlbumIds )
		:flatMap( function( id )
			return WFCatIdUtils.albumUuidForMaybeLocalId( wfCatalog, id )
				:flatMap( function( albumId )
					if albumId then
						return WFCatAlbums.deleteIndividualAlbum( wfCatalog, albumId )
					else 
						return Observable.just()
					end
				end )
		end )
		:toArray()
		:call( runUntilDone )
end

--------------------------------------------------------------------------------

function WFCatAlbums._validateRenameAlbumInHierarchy( hierarchyObservable, albumId, albumName )
	
	return hierarchyObservable
		:flatMap( function( hierarchy )
			
			local album = getAlbumFromHierarchyById( hierarchy, albumId )
	
			if not album then
				return Observable.throw( "AlbumDoesNotExist" )
			end
	
			local parentId = album.parentId or ROOT_ALBUM_PARENT_ID
	
			-- ignore collisions with this album's own name so the user can change its capitalization.
			local ignoreAlbumId = album.id

			local nameIsUnique = isAlbumNameUniqueAmongSiblings( albumName, album.subtype, hierarchy.localIdToChildrenAlbumData[ parentId ], ignoreAlbumId )

			if not nameIsUnique then
				return Observable.throw( "NameExists" )
			end
	
			return Observable.just( hierarchy )
	
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Renames an album or folder in this catalog.

	It is an error to rename an album or folder to match the name of an existing album,
	regardless of case.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string) The unique or local ID of the album/folder to rename.
	\param name (string) The new name for the album/folder.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--

function WFCatAlbums.renameAlbumOrFolder( wfCatalog, albumId, albumName )

	check( "WFCatalog", '+string|number', '+string' )

	albumName = maybeMakeAlbumNameValid( albumName )
	
	local docStore = wfCatalog:docStore()

	return WFCatAlbums._validateRenameAlbumInHierarchy( flattenedAlbumHierarchy( wfCatalog, true ):take( 1 ), albumId, albumName )
		:flatMapLatest( function( hierarchy )
			
			local baseOverlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( albumId )
			
			baseOverlay.name = albumName
			
			return docStore:applyOverlay( baseOverlay )
			:takeLast( 0 )
			
		end )

end

--------------------------------------------------------------------------------

local function wouldMoveCreateACycle( hierarchy, albumId, parent )

	if not parent then
		return false -- if we're moving to the top level, it can't create a loop
	end
	
	local visitedNodes = {}
	
	local node = parent
	while node do
		if node.id == albumId then
			-- if the node we're looking at now is the album we're moving
			-- we've got a loop because we started at its future parent
			return true
		end
		
		if visitedNodes[ node ] then
			-- SPAM [matthewj 2017-01-30]: In the edge case that there's already a 
			-- cycle in the parentage but the moved node isn't the one creating it
			-- we return true here to be overly cautious
			return true
		end
		visitedNodes[ node ] = true
		
		-- get the next node to look at
		if node.parentId then
			node = getAlbumFromHierarchyById( hierarchy, node.parentId )
		else
			node = nil
		end
		
	end
	
	return false
end

--------------------------------------------------------------------------------

local function performReparentValidation( hierarchy, albumId, optNewParentId )
	
	local album = getAlbumFromHierarchyById( hierarchy, albumId )
			
	if not album then
		return false, "AlbumDoesNotExist"
	end

	local parent = getAlbumFromHierarchyById( hierarchy, optNewParentId )
	
	if not isValidParent( parent ) then
		return false, "InvalidParent"
	end
	
	if wouldMoveCreateACycle( hierarchy, album.id, parent ) then
		return false, "CycleDetected"
	end
	
	local albumName = album.name
	
	local parentId = parent and parent.id or ROOT_ALBUM_PARENT_ID
	local siblingAlbums = hierarchy.localIdToChildrenAlbumData[ parentId ]
	
	local nameIsUnique = isAlbumNameUniqueAmongSiblings( albumName, album.subtype, siblingAlbums, albumName )
	
	if not nameIsUnique then
		return false, "NameExists"
	end
	
	return true
end

--------------------------------------------------------------------------------

function WFCatAlbums._validateReparentAlbumInHierarchy( hierarchyObservable, albumId, optNewParentId )
	
	return hierarchyObservable
		:flatMap( function( hierarchy )
			
			local isValid, maybeError = performReparentValidation( hierarchy, albumId, optNewParentId )
			
			if not isValid and maybeError then
				return Observable.throw( maybeError )
			end
			
			return Observable.just( hierarchy )
		end )

end

--------------------------------------------------------------------------------

function WFCatAlbums._validateReparentMultipleAlbumsInHierarchy( hierarchyObservable, albumIds, optNewParents )
	
	return hierarchyObservable
		:flatMap( function( hierarchy )
			
			local maybeErrors = {} 
			for _, albumId in ipairs( albumIds ) do
				
				local maybeParentId 
				if type( optNewParents ) == 'number' or type( optNewParents ) == 'string' then
					maybeParentId = optNewParents
				elseif type( optNewParents ) == 'table' then
					maybeParentId = optNewParents[ albumId ]
				end
				
				local isValid, maybeError = performReparentValidation( hierarchy, albumId, maybeParentId )
				
				if not isValid then
					maybeErrors[ albumId ] = maybeError
				end
			end
			
			maybeErrors = next( maybeErrors ) and maybeErrors or nil
			
			return Observable.just( hierarchy, maybeErrors )
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Moves an album or folder onto a different parent

	It is an error to move an album/folder a parent that already has an album with that
	same name. Only folders are currently allowed to contain children so trying to reparent 
	to an album will throw an error.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string) The unique or local ID of the album/folder to rename.
	\param optNewParentId (string|int|nil) The new parent or nil to make this a root album/folder
	\param optOrder (array|nil) Array of album ids (local or global) in the order they should
		be in that new parent.  Any album ids that are specified that are not in that parent
		album are ignored.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--

function WFCatAlbums.reparentAlbumOrFolder( wfCatalog, albumId, optNewParentId, optOrder )

	check( 'WFCatalog', '+string|number', '?+string|number' )

	return WFCatAlbums._validateReparentAlbumInHierarchy( flattenedAlbumHierarchy( wfCatalog, true ):take( 1 ), 
															albumId, optNewParentId )
		:flatMap( function( hierarchy )
			local album = getAlbumFromHierarchyById( hierarchy, albumId )
			local parent = getAlbumFromHierarchyById( hierarchy, optNewParentId )
			
			local newParentOverlay
			if parent then
				newParentOverlay = {
					id = parent.fullDocId,
				}
			else
				newParentOverlay = {
					_delete = true,
				}
			end

			local parentId = parent and parent.id or ROOT_ALBUM_PARENT_ID
			local siblingAlbums = hierarchy.localIdToChildrenAlbumData[ parentId ]
			
			local newAlbumOrder, siblingAlbumOverlays = reorderNewAlbumAndSiblings( albumId, album.albumName, siblingAlbums )
	
			siblingAlbumOverlays[ #siblingAlbumOverlays + 1 ] = {
				_id = albumId,
				parent = newParentOverlay,
				order = newAlbumOrder,
			}

			return WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), siblingAlbumOverlays )

		end )
		:takeLast( 0 )

end

--------------------------------------------------------------------------------

function WFCatAlbums.reparentMultipleAlbumsOrFolders( wfCatalog, albumIds, optNewParents )
	
	check( 'WFCatalog', 'table' )
	
	return WFCatAlbums._validateReparentMultipleAlbumsInHierarchy( flattenedAlbumHierarchy( wfCatalog, true ):take( 1 ),
																	albumIds, optNewParents )
		:flatMap( function( hierarchy, maybeErrors )
			
			local albumParentOverlays = {}
			
			-- Related to NMB-6314: if attempting to reparent multiple albums and there is an error, we 
			-- reject the full reparent even if some would succeed. Unfortunately, models can currently 
			-- only pass a string for an error event so we need to return the errors table as a data
			-- event and if nil, this should be interpreted as a success.
			if maybeErrors then 
				return Observable.just( maybeErrors )
			end
			
			-- there isn't any validation error so we should be able to add overlays to move the albums
			for _, albumId in ipairs( albumIds ) do
				
				local maybeParentId 
				if type( optNewParents ) == 'number' or type( optNewParents ) == 'string' then
					maybeParentId = optNewParents
				elseif type( optNewParents ) == 'table' then
					maybeParentId = optNewParents[ albumId ]
				end
				
				local parent = getAlbumFromHierarchyById( hierarchy, maybeParentId )
		
				local overlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( albumId )
				
				if parent then
					overlay.parent = {
						id = parent.fullDocId,
					}
				else
					overlay.parent = {
						_delete = true,
					}
				end
			
				albumParentOverlays[ albumId ] = overlay
			end
				
			local overlayObservable
			if next( albumParentOverlays ) then
				local finalOverlays = {}
				for _, overlay in pairs( albumParentOverlays ) do
					finalOverlays[ #finalOverlays + 1 ] = overlay
				end
				overlayObservable = WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), finalOverlays )
			else
				overlayObservable = Observable.just()
			end
			
			-- make sure we return nil (which will be interpreted as 'success')
			return overlayObservable:map( function() return nil end )
		end )
		:takeLast( 1 )
end

--------------------------------------------------------------------------------

local function addAdditionalAlbumInfoToAlbumsList( albums, counts, covers, statuses )
 
	local albumsWithMoreData = {}
	
	for _, album in ipairs( albums ) do
		album.syncStatus = statuses[ album.fullDocId ]
		album.count = counts[ album.fullDocId ]
		album.cover = covers[ album.id ]
		
		albumsWithMoreData[ #albumsWithMoreData + 1 ] = album
	end
	
	return albumsWithMoreData
end 

 --------------------------------------------------------------------------------
 
local function combineAdditionalAlbumInfoWithAlbumList( wfCatalog, albumListObservable )
     
	return Rx.Observable.combineLatest( 
		function( albums, counts, covers, statuses )
		
			if next( counts ) or next( covers ) or next( statuses ) then
			
				albums = addAdditionalAlbumInfoToAlbumsList( albums, counts, covers, statuses )
			
			end
			
			return albums
		
		end,
	
	albumListObservable,
	WFCatAlbums.getAlbumCounts( wfCatalog ):startWith( {} ),
	WFCatAlbums.getAlbumCovers( wfCatalog ):startWith( {} ),
	WFCatSyncStatus.getSyncStatusForAlbums( wfCatalog, true --[[ignore album status]] ):startWith( {} ) )
 
 end
 
--------------------------------------------------------------------------------

 --[[
     \brief Returns an Observable producing array of top-level albums, each containing fields:
         id : (int) the uuid of the album
         name : (user string) the name of the album
         path : (user string) The name of the parents and album separated by /
         parentId : (uuid, or nil) the id of the albums parent
         children : (array) array of child albums
         
     \param wfCatalog an Observable( WFCatalog )
--]]     
 function WFCatAlbums.albumsWithSyncStatus( wfCatalog )
 
     -- Fallback in case we fail to log in and get a catalog reference.
 
     if wfCatalog == nil then
         return Rx.Observable.just {}
     end
 
     return combineAdditionalAlbumInfoWithAlbumList( wfCatalog, WFCatAlbums.getAlbumsList( wfCatalog ) )
         
 end
 
--------------------------------------------------------------------------------

local function assetUuidsForMaybeLocalAlbumId( wfCatalog, maybeLocalAlbumId )
	
	check( "WFCatalog", '+string|number' )
	
	if type( maybeLocalAlbumId ) == 'number' then
		
		return wfCatalog:indexStore():query {
				sql = [[SELECT DISTINCT album_asset_v2.assetId, album_asset_v2.docId 
							FROM album_asset_v2 
							JOIN albums ON album_asset_v2.albumId == albums.albumId
							WHERE albums.docId = ? ]], 
				maybeLocalAlbumId,
				quiesceUntilUpToDate = true,
			}
			:take( 1 )
			:map( function( rows )
				
				logger:debugf( 'assetUuidsForMaybeLocalAlbumId for localId %s complete w/ %d rows', maybeLocalAlbumId, #rows )
			
				local assetIds = {}
				
				for _, albumAsset in ipairs( rows ) do
				
					local assetId = albumAsset[ 1 ]
					local docId = albumAsset[ 2 ]
		
					assetIds[ assetId ] = docId
		
				end
				
				return assetIds
			
			end )
	end
	
	-- maybeLocalAlbumId is already string
	return wfCatalog:docStore():docValue( maybeLocalAlbumId )
		:flatMapLatest( function( docValue )
			if docValue._empty then
				return Observable.throw( "No such album: " .. tostring( maybeLocalAlbumId ) )
			end
			
			assert( type( docValue._localId ) == 'number' )
			return assetUuidsForMaybeLocalAlbumId( wfCatalog, docValue._localId )
		end )
		
end

--------------------------------------------------------------------------------

local function assetsWithQuotaErrors( wfCatalog, assetIds )
	
	if not next( assetIds ) then
		return Observable.just( {} )
	end
	
	local idColumn
	if type( assetIds[ 1 ] ) == 'string' then
		idColumn = 'assets.assetId'
	else
		idColumn = 'assets.docId'
	end
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT assets.assetId FROM quota_exceeded 
				JOIN assets USING( docId )
				JOIN TEMP_ids ON TEMP_ids.docId == ]] .. idColumn,
		tempIds = assetIds,
	}
	:map( function( assetIdRows )
		-- turn into a map for easy lookup
		local assetIdMap = {}
		for _, row in ipairs( assetIdRows ) do
			assetIdMap[ row[ 1 ]  ] = WFCatOzErrors.Errors.kStorageExceededError
		end
		
		return assetIdMap
	end )
	:take( 1 )

end

--------------------------------------------------------------------------------

local function highestOrderForAlbum( wfCatalog, albumId )

	return wfCatalog:indexStore():query( {
		sql = [[SELECT assetId, sortOrder FROM album_asset_v2 
				WHERE albumId == ? and sortOrder IS NOT NULL 
				ORDER BY sortOrder DESC limit 1]],
			albumId,
		quiesceUntilUpToDate = true,
	} )
		
		:map( function( row )
		
			if row and #row >= 1 then
				return row[ 1 ][ 2 ]
			end

		end )
end

--------------------------------------------------------------------------------

local function getAssetOrderingObservable( wfCatalog, albumId, assetIds )

	-- Docs will be inserted in order after the last doc with
	-- order. Docs without order will not be affected.
	
	-- Fetch the order of last doc with order.
	return highestOrderForAlbum( wfCatalog, albumId )
		:take( 1 )
		:map( function( highestOrder )
		
			local ordering = {}
			local formerOrder = highestOrder or SORT_PREV_FOR_INSERT_AS_FIRST
			
			for i, assetId in ipairs( assetIds ) do
				ordering[ assetId ] = WFSortOrderStrings.insertAfter_lex64( formerOrder, 'low' )
				formerOrder = ordering[ assetId ]
			end
			
			return ordering
		end )
end

--------------------------------------------------------------------------------

--[[!
	\brief Adds assets to an album in this catalog.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string|number) The unique ID of the album to modify.
	\param assetIds (array of strings or integers) An array of unique or local
		IDs for the assets to add.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--

function WFCatAlbums.addAssetsToAlbum( wfCatalog, maybeLocalAlbumId, maybeLocalAssetIds )

	check( "WFCatalog", '+string|number', 'table' )

	local albumUuid = WFCatIdUtils.albumUuidForMaybeLocalId( wfCatalog, maybeLocalAlbumId )
	local assetUuids = WFCatIdUtils.getOzAssetIdsForLocalIds( wfCatalog, maybeLocalAssetIds )
	local assetsWithQuotaErrors = assetsWithQuotaErrors( wfCatalog, maybeLocalAssetIds )
	local existingAssetUuids = assetUuidsForMaybeLocalAlbumId( wfCatalog, maybeLocalAlbumId )
	
	return Observable.combineLatest( function( albumId, assetIds, existingAssetIds, assetsWithQuotaErrors )
		return Observable.from( assetIds )
			:filter( function( assetId )
				return not existingAssetIds[ assetId ]
			end )
			:map( function( assetId )
				local aaDoc = createAlbumAssetDocument( albumId, assetId )
				if assetsWithQuotaErrors[ assetId ] then
					aaDoc = WFCatOzErrors.addOverlayForOzErrorSubtype( aaDoc, WFCatOzErrors.Errors.kStorageExceededError )
				end
				return aaDoc
			end )
	end, albumUuid, assetUuids, existingAssetUuids, assetsWithQuotaErrors )
	:switch()
	:toArray()
	:flatMapLatest( function( docs )
	
		if next( docs ) then
		
			logger:debugf( 'addAssetsToAlbum WFRxBatchedBulkSetDocValue for %d docs', #docs )
			
			return WFRxBatchedBulkSetDocValue( wfCatalog:docStore(), docs )
			
		else
			return Observable.just()
		end
		
	end )
	:tap( function()
		logger:debugf( 'addAssetsToAlbum WFRxBatchedBulkSetDocValue COMPLETE' )
	end )
	:takeLast( 0 )

end

--------------------------------------------------------------------------------

--[[!
	\brief Adds assets to an album taking care of ids of album_asset documents already
		on Oz. (Example: while migrating a synced album from Lightroom)

		If the asset is already in the album, do not create a new entry. 
		Simply return the id of the existing album_asset document irrespective
		of whether the client sent a doc-id or not.
		If the client provides doc-id of an existing album-asset document in Oz, 
		then create a document with the same id. Otherwise, generate one. 
		Return the new document's id.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (list of strings) 
	\param albumAssetIdsForAsset (table of strings) A table in the following format
		{
			[ assetId3 ] = guid3, -- If Oz already has this album_asset document with guid3
			[ assetId4 ] = guid4, -- If Oz already has this album_asset document with guid4
		}
	\param obeyOrder (boolean)  If provided, the sort order will be set
		obeying the order of the assets in the input list.

	\return <code>Observable< ></code> An Observable object sends the ids of the album_asset
		documents in the following format.
		{
			[ assetId1 ] = guid1,
			[ assetId2 ] = guid2,
			[ assetId3 ] = guid3,
			[ assetId4 ] = false, -- if there was an error validating the given albumAssetId
		}
]]--

function WFCatAlbums.addAssetsToAlbumGivenDocIds( wfCatalog, albumId, assetIds, albumAssetIdsForAsset, obeyOrder )
	check( "WFCatalog", '+string', 'table', 'table', '?boolean' )
	
	local existingAssetUuidsObservable = assetUuidsForMaybeLocalAlbumId( wfCatalog, albumId )
	
	local existingAlbumAssetIdList = {}
	for _, guid in pairs( albumAssetIdsForAsset ) do
		existingAlbumAssetIdList[ #existingAlbumAssetIdList + 1 ] = guid
	end
	
	local existingAlbumAssetDocsObservable = #existingAlbumAssetIdList > 0
		and wfCatalog:docStore():bulkDocValue( existingAlbumAssetIdList )
		or Observable.just( {} )
	
	local assetOrderingObservable
	
	if obeyOrder then
		-- Docs will be inserted in order after the last doc with
		-- order. Docs without order will not be affected.
		
		assetOrderingObservable = getAssetOrderingObservable( wfCatalog, albumId, assetIds )
	else
		assetOrderingObservable = Observable.just( {} )
	end
	
	return Observable.combineLatest( existingAssetUuidsObservable, existingAlbumAssetDocsObservable, assetOrderingObservable )
	
		:flatMap( function( assetsInAlbum, existingAlbumAssetDocs, ordering )
		
			local existingAlbumAssetLocalIds = {}
			local assetIdsToAdd = {}
			local validationFailedAssets = {}
			local deletedAlbumAssets = {}
			
			local existingAlbumAssetDocMap = {}
			for _, docValue in ipairs( existingAlbumAssetDocs ) do
				if docValue._id then
					
					if docValue._deleted then
						-- Album asset doc has been deleted
						-- Treat this case same as the 
						-- non-existing album-asset case
						deletedAlbumAssets[ docValue._id ] = true
					
					elseif not docValue._empty then
						-- Album asset doc exists
						existingAlbumAssetDocMap[ docValue._id ] = docValue
					end
				end
			end
			
			-- Sort the assets that we want to add into three buckets:
			-- those that already are part of the album, those that are not,
			-- and a degenerate case where an existing album_asset with the
			-- same id points to an unrelated asset or album
			for i, assetId in ipairs( assetIds ) do
			
				local albumAssetId = albumAssetIdsForAsset[ assetId ]
				local isValid = true
				
				if albumAssetId then
				
					local albumAssetDoc = existingAlbumAssetDocMap[ albumAssetId ]
					
					if albumAssetDoc then
						if KSTableUtils.derefOpt( albumAssetDoc, "asset", "id" ) ~= assetId or 
							KSTableUtils.derefOpt( albumAssetDoc, "album", "id" ) ~= albumId then
							
							isValid = false
							validationFailedAssets[ #validationFailedAssets + 1 ] = assetId
						end
					end
				end
				
				if isValid then
					if assetsInAlbum[ assetId ] then
						existingAlbumAssetLocalIds[ #existingAlbumAssetLocalIds + 1 ] = assetsInAlbum[ assetId ]
					else
						assetIdsToAdd[ #assetIdsToAdd + 1] = assetId
					end
				end
			end
			
			-- Create album_asset documents for new assets
			local newDocs, newDocsRet = {}, {}
			for i, assetId in ipairs( assetIdsToAdd ) do
				
				-- Deleted album-assets should be treated similar to
				-- new album-assets. Create the album-asset with a 
				-- new doc id.
				
				local providedDocId = albumAssetIdsForAsset[ assetId ]
				local isDeleted = providedDocId and deletedAlbumAssets[ providedDocId ]
				
				local docId = not isDeleted and providedDocId
					or KSIdGenerator.uuidInLowercaseHex()

				local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( docId )

				local docValue = {
					type = 'album_asset',
					order = ordering[ assetId ],
					album = {
						id = albumId,
					},
					asset = {
						id = assetId,
					},
				}
				
				WFTableMerge.applyOverlay( docValue, baseDocValue )
				
				newDocsRet[ assetId ] = docId
				newDocs[ i ] = docValue

			end
			
			-- Create album_asset documents for existing assets
			-- to alter ordering
			
			local newAlbumAssetsDocsObservable = #newDocs > 0
				and WFRxBatchedBulkSetDocValue( wfCatalog:docStore(), newDocs )
						:map( function()
							return newDocsRet
						end )
				or Observable.just( {} )
			
			-- Fetch the album_asset doc ids of those assets already in the album
			local existingAlbumAssetDocsObservable
			
			if obeyOrder and #existingAlbumAssetLocalIds > 0 then
			
				-- Already existing album-assets in the input list 
				-- needs to be modified to respect the order
				existingAlbumAssetDocsObservable = wfCatalog:docStore():bulkDocValue( existingAlbumAssetLocalIds )
				 	:flatMap( function( docs )
						
						local existingDocs = {}
						for i, doc in ipairs( docs ) do
							
							local overlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( doc._id )
							
							overlay.type = 'album_asset'
							overlay.order = ordering[ doc.asset.id ]

							existingDocs[ #existingDocs + 1 ] = overlay
							
						end
						
						if #existingDocs > 0 then
						
							return WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), existingDocs )
								:map( function()
									return docs
								end )
								
						else
							return Observable.just( {} )
						end
					end )
			else
				existingAlbumAssetDocsObservable = wfCatalog:docStore():bulkDocValue( existingAlbumAssetLocalIds )
			end
			
			
			local skippedFailedAlbumAssets = Observable.just( validationFailedAssets )
			
			-- Merge the existing docs and new docs stream
			return Observable.combineLatest( function( existingDocs, newDocRevs, failedAssets )
				
				local ret = {}
				
				for assetId, docId in pairs( newDocRevs ) do
					ret[ assetId ] = docId
				end
				
				for i, doc in ipairs( existingDocs ) do
					ret[ doc.asset.id ] = doc._id
				end
				
				for _, assetId in ipairs( failedAssets ) do
					ret[ assetId ] = false
				end
				
				return ret
				
			end, 
			existingAlbumAssetDocsObservable, 
			newAlbumAssetsDocsObservable,
			skippedFailedAlbumAssets )
			
		end )
end

--------------------------------------------------------------------------------

local function assetIdsInStacksAsMap( wfCatalog, stackIds, quiesce )
	
	return wfCatalog:indexStore():query {
				sql = [[SELECT assetId FROM stack_assets JOIN TEMP_ids AS t ON t.docId == stack_assets.stackId]],
				quiesceUntilUpToDate = quiesce or false,
				tempIds = stackIds,
			}
			:take( 1 )
			:map( function( assetIdRows )
				local assetIds = {}
				for _, row in ipairs( assetIdRows ) do
					assetIds[ row[ 1 ] ] = true
				end
				return assetIds
			end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Removes assets from an album in this catalog.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings or integers) An array of unique or local
		IDs for the assets to remove.  It is not an error to remove assets
		that are not in the album -- those asset ids are simply ignored.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--

function WFCatAlbums.removeAssetsFromAlbum( wfCatalog, maybeLocalAlbumId, maybeLocalAssetIds )

	check( "WFCatalog", '+string|number', 'table' )

	local albumUuid = WFCatIdUtils.albumUuidForMaybeLocalId( wfCatalog, maybeLocalAlbumId )
	local assetUuidsAndTypes = WFCatIdUtils.getOzAssetIdsAndTypeForIds( wfCatalog, maybeLocalAssetIds )

	return Observable.combineLatest( albumUuid, assetUuidsAndTypes )
		:map( function( albumId, assetUuidsAndTypes )
		
		local assetIds = {}
		local stacksForRemoval = {}
		-- process the assetIds - we need a list of all asset ids as well as a 
		-- map of which asset ids are for stacks which we'll use later
		for _, assetUuidAndType in pairs( assetUuidsAndTypes ) do
			local assetId = assetUuidAndType[ 1 ]
			local subtype = assetUuidAndType[ 2 ]
			if subtype == 'stack' then
				stacksForRemoval[ #stacksForRemoval + 1 ] = assetId
			end
			assetIds[ #assetIds + 1 ] = assetId
		end
		
		-- fetch all album_assets for the assetIds
		local albumAssets = wfCatalog:indexStore():query {
				-- get all the album_assets
				sql = [[SELECT aa.docId, aa.assetId FROM album_asset_v2 AS aa JOIN TEMP_ids AS t ON t.docId == aa.assetId WHERE aa.albumId = ?]],
				quiesceUntilUpToDate = true,
				tempIds = assetIds,
				albumId,
			}
			:take( 1 )
		
		-- look for the cover asset id as we'll need to know if we delete the cover asset
		local albumCoverAssetUuid = WFCatAlbums.albumCover( wfCatalog, albumId )
					:map( function( docId, assetId )
						return assetId 
					end )
		
		local assetIdsForRemovedStacks = assetIdsInStacksAsMap( wfCatalog, stacksForRemoval, true )
		
		-- which stacks (and their stack_assets) are not being removed
		local assetsIdsForStacksRemaining = wfCatalog:indexStore():query {
				-- get stacks in album that are not being deleted
				sql = [[SELECT aa.assetId FROM album_asset_v2 as aa JOIN assets USING( assetId ) 
							WHERE aa.albumId = ? AND assets.subtype = 'stack' 
						EXCEPT 
							SELECT docId FROM TEMP_ids]],
				quiesceUntilUpToDate = true,
				tempIds = stacksForRemoval,
				albumId,
			}
			:take( 1 )
			:map( function( stackAssetRows )
				local stackIdsRemaining = {}
				for _, row in ipairs( stackAssetRows ) do
					stackIdsRemaining[ #stackIdsRemaining + 1 ] = row[ 1 ]
				end
				
				return stackIdsRemaining
			end )
			:flatMap( function( stackIds )
				return assetIdsInStacksAsMap( wfCatalog, stackIds, true )
			end )		
		
		return Observable.combineLatest( albumAssets, albumCoverAssetUuid, assetIdsForRemovedStacks, assetsIdsForStacksRemaining )
			:take( 1 )
			:map( function( albumAssetRows, albumCoverId, assetIdsForRemovedStacks, assetIdsForStacksRemaining )
				
				local deletedAlbumCover = false
				local overlays = {}
				local overlaysToKeepFromOz = {
						label = 'oz'
					}
				
				for i, row in ipairs( albumAssetRows ) do
					
					local albumAssetId = row[ 1 ]
					local assetId = row[ 2 ]
					
					if assetId  == albumCoverId then
						deletedAlbumCover = true
					end
					
					-- check to see if the asset is in a stack that's remaining. if it is, then it should actually stay.
					if not assetIdsForStacksRemaining[ assetId ] then 
						
						local overlay = {
								_id = albumAssetId,
								_deleted = true
							}
						
						-- asset is not in a stack that is sticking around, so if it's in a stack that's being removed,
						-- give it a label of 'oz' so that we don't send it
						if assetIdsForRemovedStacks[ assetId ] then
							overlaysToKeepFromOz[ #overlaysToKeepFromOz + 1 ] = overlay
						else
							overlays[ #overlays + 1 ] = overlay
						end
						
					end
				end
				
				return deletedAlbumCover, overlays, overlaysToKeepFromOz
			end )
			:flatMap( function( deletedAlbumCover, overlays, overlaysToKeepFromOz )
				
				return WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), overlays )
					:flatMap( function()
						return WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), overlaysToKeepFromOz )
					end )
					:map( function()
						return deletedAlbumCover, albumId
					end )
			end )
			
		end )
	:switch()
	:flatMapLatest( function( deletedAlbumCover, albumId )

		-- Not sure if it can happen but delete album cover at the end
		-- so the UI doesn't try defaulting it to a photo that will
		-- subsequently be deleted when deleting multiple assets.

		if deletedAlbumCover then
			return WFCatAlbums.setAlbumCover( wfCatalog, albumId, nil )
		end

		return Observable.just( true )

	end )
	:takeLast( 0 )
	:call( runUntilDone )


end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the cover image for an album in this catalog.

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string or number) The id of the album to modify.
	\param assetId (string or number or nil) The id of the asset to set as cover.
		If nil, any existing cover setting is removed from the album.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatAlbums.setAlbumCover( wfCatalog, albumId, maybeAssetId )
	
	check( "WFCatalog", 'string|number', '?string|number' )
	
	local globalAssetId = Observable.just( nil )
	
	if type( maybeAssetId ) == 'number' then
		-- get the assetId for a local id
		globalAssetId = wfCatalog:indexStore():query {
				sql = [[ SELECT assetId FROM assets where docId = ?]],
				maybeAssetId,
			}
			:take( 1 )
			:map( function( rows )
				return derefOpt( rows, 1, 1 )
			end )
	
	elseif type( maybeAssetId ) == 'string' then
		
		globalAssetId = Observable.just( maybeAssetId )
		
	end
		
	
	return globalAssetId
			:take( 1 )
			:map( function( maybeGlobalAssetId )
				
				local overlay = WFCatUtils.createBaseOverlayWithUserUpdatedField( albumId )
				
				if maybeGlobalAssetId then
					overlay.cover = { id = maybeGlobalAssetId }	
				else
					overlay.cover = { _delete = true }
				end
				
				return wfCatalog:docStore():applyOverlay( overlay )

			end )
			:switch()
			:takeLast( 0 )
	 
end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the local asset id for the cover image for an album in this catalog.
	Rules governing the selection of the photo:
	If a cover image has been set, and we have the asset in the catalog use that asset
	Select the first photo in the album in chronological order.
	
	\param maybeLocalAlbumId (string|number) The unique ID of the album. This can either be the local 
	id or the global uuid of the album.
	
	\return <code>Observable< number, string ></code> An Observable object whose value is the docId, assetId
	of the current cover photo for this album.
	
	Observable value may be nil if there are no assets in the album or we do not yet have 
	local information for the asset id referred to as the cover.

 ]]--
function WFCatAlbums.albumCover( wfCatalog, maybeLocalAlbumId )

	WFArguments.check( 'WFCatalog', 'string|number' )
	
	local albumIdObservable

	if type( maybeLocalAlbumId ) == 'number' then
		albumIdObservable = wfCatalog:indexStore():query {
								sql = [[SELECT albumId FROM albums WHERE docId = ?]],
								maybeLocalAlbumId
							}
							:map( function( rows )
								return derefOpt( rows, 1, 1 )
							end )
	else
		albumIdObservable = Observable.just( maybeLocalAlbumId )
	end
	
	return albumIdObservable
		:take( 1 )
		:map( function( albumId )	
			return wfCatalog:indexStore():query {
	
				sql = [[SELECT assets.docId, assets.assetId
							FROM assets
							INNER JOIN albums ON albums.coverId == assets.assetId
							WHERE albums.albumId = ?
					UNION ALL
						SELECT docId, assetId FROM 
						( SELECT assets.docId, assets.assetId
							FROM assets
							INNER JOIN album_asset_v2 ON album_asset_v2.assetId == assets.assetId
							WHERE album_asset_v2.albumId = ?
							ORDER BY
								case when album_asset_v2.sortOrder is null then 1 else 0 end,
								album_asset_v2.sortOrder, captureDate ASC )
					LIMIT 1]],
				albumId, albumId
			}
			:map( function( rows )
				-- pull out local cover id 
				return derefOpt( rows, 1, 1 ), derefOpt( rows, 1, 2 )
			end )
			
		end )
		:switch()
	
end

--------------------------------------------------------------------------------


function WFCatAlbums.addDiagnosticLogSectionFunc( catalog )

	WFDiagnosticLog.addLogSectionFunc( function()

		return {

			name = "WFCatAlbums",
			id = 'WFCatAlbums',
			weight = 510,

			body = {
			
				albumCycles = basicAlbumsList( catalog, true )
				
					:map( function( albums )
						
						return findCycles( albums )
						
					end )
					
			},

		}

	end )

end

--------------------------------------------------------------------------------

return WFCatAlbums
