--[[----------------------------------------------------------------------------

WFCatAssets.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- wf.core.utils
local WFFunctionUtils = require "WFFunctionUtils"

	-- wf.oz_client.managed_catalog.pixel_provider
local WFCatRenderCapabilities = require "WFCatRenderCapabilities"

	-- rxlua
local Rx = require 'Rx'

--------------------------------------------------------------------------------

local WFFunctionUtils_reduce = assert( WFFunctionUtils.reduce,
			"WFFunctionUtils.reduce does not exist" )

--------------------------------------------------------------------------------

local function addLocalAssetIdToSet( set, localAssetIdArray )

	local localAssetId = localAssetIdArray[ 1 ]

	if not localAssetId then

		return set

	end

	set[ localAssetId ] = true

	return set

end

--------------------------------------------------------------------------------

local function convertLocalAssetIdTupleArrayToSet( localAssetIdTupleArrays )

	return WFFunctionUtils_reduce( {}, addLocalAssetIdToSet,
										localAssetIdTupleArrays )

end

--------------------------------------------------------------------------------

local function getAllLocalAssetIdsAsSetStream( wfCatalog )

	local changeCounter = wfCatalog:docStore():changeCounter():take( 1 )

	local allLocalAssetIdsSet =
			wfCatalog
				:indexStore()
				:query( { sql = [[SELECT docId FROM assets]],
						  quiesceUntilUpToDate = true } )
				:take( 1 )
				:map( convertLocalAssetIdTupleArrayToSet )

	return Rx.Observable.concat( changeCounter, allLocalAssetIdsSet )
				:toArray()
				:map( unpack )	

end

--------------------------------------------------------------------------------

local function makeAccumulatedLocalAssetIdUpdateStream( ozCatalogId, stream,
															interval )

	local scheduler = require 'WFRxScheduler'.sharedScheduler()

	local adds = {}

	local deletes = {}

	local kIgnoredFirstResult = {}

	-- kIgnoredFirstResult is used to make sure that :calm doesn't emit until interval
	-- passes
	return Rx.Observable.concat( Rx.Observable.just( kIgnoredFirstResult ), stream )
			:map(
				function( updateType, localAssetId )

					if updateType == kIgnoredFirstResult then

						adds[ kIgnoredFirstResult ] = true

					elseif updateType == "add" then

						if not deletes[ localAssetId ] then

							adds[ localAssetId ] = true

						end

					elseif updateType == "delete" then

						adds[ localAssetId ] = nil

						deletes[ localAssetId ] = true

					end

				end )
			:calm( interval, scheduler )
			:map(
				function()

					local currAdds, currDeletes = adds, deletes

					adds, deletes = {}, {}

					currAdds[ kIgnoredFirstResult ] = nil

					return ozCatalogId, currAdds, currDeletes

				end )
			:where(
				function( _, currAdds, currDeletes )

					return next( currAdds ) ~= nil or next( currDeletes ) ~= nil

				end )

end

--------------------------------------------------------------------------------
-- maybeReportAllLocalAssetIdChange is called from
-- 	docStore:scannedFilteredMappedChangesFeedSince on a separate worker thread.
--------------------------------------------------------------------------------

local function maybeReportAllLocalAssetIdChange( allLocalAssetIds, docValue )

	if not WFCatRenderCapabilities.isRenderableAsset( docValue ) then

		return allLocalAssetIds

	end

	local localAssetId = docValue._localId

	if not localAssetId then

		return allLocalAssetIds

	end

	if docValue._deleted then

		allLocalAssetIds[ localAssetId ] = nil

		return allLocalAssetIds, "delete", localAssetId

	elseif not allLocalAssetIds[ localAssetId ] then

		allLocalAssetIds[ localAssetId ] = true

		return allLocalAssetIds, "add", localAssetId

	end

	return allLocalAssetIds

end

--------------------------------------------------------------------------------

local function makeLocalAssetIdsDeltaStream( wfCatalog, ozCatalogId,
												changeCounter,
												allLocalAssetIdsSet,
												bufferInterval )

	local unbufferedLocalAssetIdUpdateStream = 
				wfCatalog
					:docStore()
					:scannedFilteredMappedChangesFeedSince( 
						"WFCatAssets", "maybeReportAllLocalAssetIdChange", 
						allLocalAssetIdsSet, changeCounter )

	return makeAccumulatedLocalAssetIdUpdateStream( ozCatalogId,
				unbufferedLocalAssetIdUpdateStream, bufferInterval )

end

--------------------------------------------------------------------------------

local function allLocalAssetIdsStreamImp( wfCatalog, bufferInterval )

	bufferInterval = type( bufferInterval ) == "number" and 
						bufferInterval > 0 and bufferInterval or 0.5

	local ozCatalogId = wfCatalog:ozCatalogId()

	return getAllLocalAssetIdsAsSetStream( wfCatalog )
			:take( 1 )
			:flatMapLatest(
				function( changeCounter, allLocalAssetIdsSet )

					local initialDeltaStream =
							Rx.Observable.just( ozCatalogId,
												allLocalAssetIdsSet, {} )

					local ongoingLocalAssetIdsDeltaStream =
							makeLocalAssetIdsDeltaStream( 
								wfCatalog, ozCatalogId, changeCounter + 1,
								allLocalAssetIdsSet, bufferInterval )

					return Rx.Observable.concat( 
								initialDeltaStream,
								ongoingLocalAssetIdsDeltaStream )

				end )

end

--------------------------------------------------------------------------------

local function pack( ... )

	return { ... }

end

--------------------------------------------------------------------------------
-- allLocalAssetIdsStream
--
-- returns a stream of lua tuples, 
--
-- 	{ ozCatalogId, addedLocalAssetIds, deletedLocalAssetIds }, 
-- 	{ ozCatalogId, addedLocalAssetIds, deletedLocalAssetIds }, 
-- 	...
--
-- 	The first pair will have
--
-- 		addedLocalAssetIds = set of all localAssetIds in catalog at
-- 		                     that point in time
--
-- 		deletedLocalAssetIds = {}
--
-- 	All subsequent pairs will be the deltas since the previous pair
-- 	was sent. i.e.
--
-- 		addedLocalAssetIds = set of all new localAssetIds since the
-- 		                     last pair
--
-- 		deletedLocalAssetIds = set of all removed localAssetIds
-- 		                       since the last pair
--------------------------------------------------------------------------------

local function allLocalAssetIdsStream( wfCatalog, bufferInterval )

	return Rx.Observable.create(

		function( observer )

			return allLocalAssetIdsStreamImp( wfCatalog, bufferInterval )
					:map( pack )
					:subscribe( observer )

		end )

end


--==============================================================================

return {

	allLocalAssetIdsStream = allLocalAssetIdsStream,

	maybeReportAllLocalAssetIdChange = maybeReportAllLocalAssetIdChange,

	privateUnitTestAPI = {

		getAllLocalAssetIdsAsSetStream = getAllLocalAssetIdsAsSetStream

	}

}
