--[[----------------------------------------------------------------------------

WFCatBinaryUploader.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable
local SingleAssignmentDisposable = Rx.SingleAssignmentDisposable

	-- WF.oz_client.managed_catalog
local WFCatBlacklistHelper = require 'WFCatBlacklistHelper'

local WFCatImportDuplicateFiles = require 'WFCatImportDuplicateFiles'
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatPhotoImportance = require 'WFCatPhotoImportance'
local WFCatOzErrors = require 'WFCatOzErrors'
local WFOzContentTypes = require 'WFOzContentTypes'
local NmbTrackedDataBroadcaster = require 'NmbTrackedDataBroadcaster'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.os
local KSDate = require 'KSDate'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.performance
local WFPerformanceReport = require 'WFPerformanceReport'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSJsonDecoder = require 'KSJsonDecoder'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFAsyncDigest = require 'WFAsyncDigest'
local WFCloseSupport = require 'WFCloseSupport'
local WFConfiguration = require 'WFConfiguration'
local WFRxContinuousTask = require 'WFRxContinuousTask'
local WFRxLowFidelityTimer = require 'WFRxLowFidelityTimer'

	-- WF.web.http_client
local WFRxHttpClient = require 'WFRxHttpClient'

local WFCatChunkedUploaderClassName = 'WFCatContentRangeUploader'
if WFConfiguration.ChunkedUploader.useChunkedUploader then
	WFCatChunkedUploaderClassName = 'WFCatChunkedUploader'
end
local WFCatChunkedUploader = require( WFCatChunkedUploaderClassName )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local derefOpt = assert( KSTableUtils.derefOpt )
local now = assert( KSDate.timeNowForAnimation )

local doesAssetDocumentHaveLocalBinary = assert( WFCatIndexUtils.doesAssetDocumentHaveLocalBinary )
local isValidAssetDocument = assert( WFCatIndexUtils.isValidAssetDocument )
local isNonDeletedStyleDocument = assert( WFCatIndexUtils.isNonDeletedStyleDocument )

local runUntilDone = WFRxRunUntilDone

local logger = WFLogger 'WFCatBinaryUploader'

--------------------------------------------------------------------------------

local maxScheduledUploads = 20
	-- SPAM [jandrew 2016-06-09]: The maximum number of uploads to schedule at once. 
	-- The WFRxRateLimiter used in the HTTP implementation limits the number of concurrent
	-- HTTP requests to 5. However, it's good to queue up more than the max number of 
	-- concurrent requests so that we can quickly process the next batch without needing
	-- to wait for the next batch to be scheduled.
	
local uploadBlackoutWindowInSeconds = 60
	-- Number of seconds we disable the re-uploading of a binary after a successful upload.

local numberOfTriesBeforeBlacklistingUpload = 3
	-- Number of attempts to upload an original before it is blacklisted.

local failedUploadBlackoutWindowInSeconds = 300
	-- Number of seconds an original remains blacklisted after it has failed to upload n times.

local kContentTypeXML = 'application/rdf+xml'
local kContentTypeDng = 'image/dng'
local kContentTypeOctetStream = 'application/octet-stream'

local uploadHttpRequestCost = WFRxHttpClient.requestCostUploadNegative

--------------------------------------------------------------------------------

local function intToBool( zeroOrOne )
	return zeroOrOne ~= 0
end

--------------------------------------------------------------------------------

local function boolToInt( isTrue )
	return isTrue and 1 or 0
end

--------------------------------------------------------------------------------

local function checkIfBlacklisted( docValue, binaryType )

	return derefOpt( docValue, '_localOnly', 'blacklistedBy', 'WFCatBinaryUploader' ) and 1 or 0
	
end

--------------------------------------------------------------------------------

local function getDatesFromDocValue( docValue )

	local importDate = derefOpt( docValue, 'importSource', 'importTimestamp' )
	
	local captureDate = derefOpt( docValue, 'captureDate' )
	
	return importDate or 1, captureDate or 1

end

--------------------------------------------------------------------------------

local function getOriginalLinkFromDocValue_IgnoringShaError( docValue )

	-- The binary uploader uses the rels/master link only to indicate whether
	-- an asset has completed upload, so return the field's contents even if
	-- there was an error in the import source status field.

	return derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )

end

--------------------------------------------------------------------------------

local function isOriginalLinkInvalid( docValue )

	return derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'invalid' )

end

--------------------------------------------------------------------------------

local function isRevisionOnOz( revisionId, allRevisionIds )
	
	allRevisionIds = allRevisionIds or {}
	
	local count = #allRevisionIds
			
	if count == 0 then
		return false	
	end
			
	for i = count, 1, -1 do
		
		local ozRevId = allRevisionIds[ i ]
		if revisionId == ozRevId then
			return true
		end
		
	end
	
	return false
end

--------------------------------------------------------------------------------

local function isAssetDocumentACopy( docValue )

	return isValidAssetDocument( docValue ) and derefOpt( docValue, 'copy', 'copy_id' )

end

--------------------------------------------------------------------------------

local function maybeEmitMasterUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
	
	local localPath = WFCatFileSpec.buildOriginalPathFromDocValue( docValue )
	
	local fileMaybeOnOfflineVolume = not localPath
	
	local localVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
	
	local localRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue )
	
	local ozMasterUrl = getOriginalLinkFromDocValue_IgnoringShaError( docValue )

	local ozMasterCreateTemplate = derefOpt( docValue,
											 '_serverProvidedMetadata',
											 'links',
											 '/rels/master_create',
											 'href' )

	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' )
		
	local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]
	
	
	local isOzOriginalInvalid = ozMasterUrl and isOriginalLinkInvalid( docValue )	
	local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )
	local localOriginalSha256 = localVolumeGuid and localRelativePath and WFCatFileSpec.getLocalOriginalSha256( docValue )
	local isLocalOriginalValid = localOriginalSha256 == imageSha256
	
	-- NMB-5779: If we have a master file then we should upload if there is no master/proxy at all. But we 
	-- may have to wait until Oz marks it invalid so we don't upload it too soon. Import(migration) should 
	-- flag this case in the localOnly section.
	
	local isInvalid
	local waitForInvalid = derefOpt( docValue, '_localOnly', 'binaryUpload', 'waitForInvalid' )
	if waitForInvalid then
		-- does master url exist and is invalid? if it doesn't exist, then we shouldn't consider it invalid
		isInvalid = isOzOriginalInvalid
		
		-- has no master, but proxy url exist and is invalid?
		local ozProxyUrl, _, isProxyInvalid = WFCatFileSpec.getProxyLinkFromDocValue( docValue )
		isInvalid = isInvalid or ( ozProxyUrl and isProxyInvalid )
		
		-- is there no master or proxy at all? this is an edge case we'll consider invalid even if it's 
		-- technically not
		isInvalid = isInvalid or ( not ozMasterUrl and not ozProxyUrl )
		
		if not isInvalid then return end
	end
	
	-- SPAM [matthewj 2017-08-29]: re: the above 'invalid' handling - technically, if the image is marked
	-- as invalid, but we have the appropriate binary based on our local sha256 matching the server sha256 
	-- then we *should* be able to upload it
	
	local ozSha256Error = derefOpt( docValue,
								  'importSource',
								  'status',
								  'error',
								  'all',
								  'reason',
								  'sha256' ) == "does not match"
								  
	-- If there's a sha256 error that we can fix, there should be a binaryUploadError for the current ozMasterUrl
	-- If so, we can actually clear the sha256error and try to re-upload locally.
	local localSha256Ok = ozMasterUrl and derefOpt( docValue, '_localOnly', 'binaryUploadErrors', '/rels/master', ozMasterUrl )

	local blockedBySha256Error = not not ( ozSha256Error and ( not localSha256Ok ) )

	local localOriginalHasError = derefOpt( docValue, '_localOnly', 'localOriginalError' )
	
	local trialOrSubscriptionExpiredError = not not WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue )

	local copyId = derefOpt( docValue, 'copy', 'copy_id' )

	local willNotUploadMaster = KSTableUtils.derefOpt( docValue,
													   'importSource',
													   'willNotUploadMaster' )

	local binaryType = 'master'
	if docValue.subtype == "preset" or docValue.subtype == "profile" then
		binaryType = 'preset_master'
	elseif docValue.subtype == "camera_profile" then
		binaryType = 'camera_profile'
	end
	
	local shouldUpload = false
	
	if not ozMasterUrl or blockedBySha256Error then
	
		-- there's no master url on Oz, or there is but it has a sha256 error (which we need to see if we can clear)
		
		shouldUpload = true
	
	elseif ozMasterUrl and ozSha256Error and not blockedBySha256Error then
	
		-- there's a master url and we already know we can clear it.
	
		shouldUpload = true
	
	elseif ozMasterUrl and isOzOriginalInvalid and isLocalOriginalValid then
	
		-- the master on Oz is invalid, and our local original is valid.
		
		shouldUpload = true
	
	end

	if ( ( localPath or ( localVolumeGuid and localRelativePath ) ) or ( copyId and not willNotUploadMaster ) )
		and ozMasterCreateTemplate
		and latestRevisionId
		and shouldUpload
		and not localOriginalHasError
		and not trialOrSubscriptionExpiredError
	then

		local contentType = WFCatFileSpec.getContentTypeFromDocValue( docValue )

		local fileSize = WFCatFileSpec.getOriginalSizeFromDocValue( docValue )

		local masterCreateUrl = ozMasterCreateTemplate:replace( '{revision_id}', latestRevisionId )

		local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceOriginalLink' )

		logger:debugf( "emit %s, %s", localPath, masterCreateUrl )
		emit( localPath, localVolumeGuid, localRelativePath, boolToInt( fileMaybeOnOfflineVolume ), masterCreateUrl, binaryType, contentType, importDate,
				captureDate, isBlacklisted, imageSha256, fileSize, boolToInt( blockedBySha256Error ), priorityForUpload, sourceLinkForCopy )

	end
		
end

--------------------------------------------------------------------------------

local function maybeEmitXmpDevelopUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
	
	local xmpCameraRaw = derefOpt( docValue, 'develop', 'xmpCameraRaw' )
	
	if not ( type( xmpCameraRaw ) == 'table' ) then
		-- if the xmpCameraRaw is not external then there's also nothing to upload
		return
	end
	
	local xmpCameraRawSha256 = xmpCameraRaw.sha256

	
	local localXmpDevelopInfo = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue )

	local localSha256, localPath, localFileSize
	if type( localXmpDevelopInfo ) == 'table' then
		localSha256 = localXmpDevelopInfo.sha256
		localPath = localXmpDevelopInfo.path
		localFileSize = localXmpDevelopInfo.fileSize
	end

	local ozXmpDevelopCreateTemplate = derefOpt( docValue,
											 '_serverProvidedMetadata',
											 'links',
											 '/rels/xmp_develop_create',
											 'href' )
	
	local xmpDevelopUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'href' )
	
	local isXmpDevelopInvalid = not not derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'invalid' )
		
	local revisionIds = derefOpt( docValue, '_serverProvidedMetadata', 'revision_ids' )

	local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceXmpDevelopLink' )

	if ( ( localPath and localFileSize and localSha256 and localSha256 == xmpCameraRawSha256 ) or sourceLinkForCopy ) and
		xmpCameraRawSha256 and
		ozXmpDevelopCreateTemplate and
		isRevisionOnOz( docValue._revId, revisionIds ) and
		( isXmpDevelopInvalid or not xmpDevelopUrl )
		then
		
		local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]
	
		local xmpDevelopCreateUrl = ozXmpDevelopCreateTemplate:replace( '{revision_id}', latestRevisionId )

		logger:debugf( "emit %s, %s", localPath, xmpDevelopCreateUrl )
		emit( localPath, nil, nil, boolToInt( false ), xmpDevelopCreateUrl, 'xmp_develop', kContentTypeXML, importDate, 
				captureDate, isBlacklisted, localSha256, localFileSize, boolToInt( false ), priorityForUpload, sourceLinkForCopy )

	end

end

--------------------------------------------------------------------------------

local function maybeEmitXmpSidecarUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )

	local ozMasterUrl = WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
	if ozMasterUrl == nil then
		-- Upload after the master is on oz as xmp may be overwritten by it.
		return
	end

	local xmpLocalPath = WFCatFileSpec.getXmpSidecarPathFromDocValue( docValue )
	local xmpSha256 = nil -- not needed for xmp sidecar

	local xmpCreateUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp', 'href' )

	local xmpNeedsUpload = WFCatFileSpec.getXmpSidecarNeedsUploadFromDocValue( docValue )
	local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceXmpLink' )

	if xmpNeedsUpload and
		( xmpLocalPath or sourceLinkForCopy ) and
		xmpCreateUrl
	then

		local xmpFileSize = WFCatFileSpec.getXmpSidecarFileSizeFromDocValue( docValue )

		logger:debugf( "emit %s, %s", xmpLocalPath, xmpCreateUrl )
		emit( xmpLocalPath, nil, nil, boolToInt( false ), xmpCreateUrl, 'xmp_sidecar', kContentTypeXML, importDate,
				captureDate, isBlacklisted, xmpSha256, xmpFileSize, boolToInt( false ), priorityForUpload, sourceLinkForCopy )

	end

end

--------------------------------------------------------------------------------

local function maybeEmitProxyUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )

	local localPath = WFCatFileSpec.buildProxyPathFromDocValue( docValue )

	local fileMaybeOnOfflineVolume = not localPath

	local localVolumeGuid, proxySize = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
	
	local localRelativePath = WFCatFileSpec.getProxyRelativePathFromDocValue( docValue )

	local ozProxyUrl = WFCatFileSpec.getProxyLinkFromDocValue( docValue, proxySize )

	local ozProxyCreateTemplate = derefOpt( docValue,
											 '_serverProvidedMetadata',
											 'links',
											 '/rels/proxy_create',
											 'href' )

	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' )

	local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]

	local copyId = derefOpt( docValue, 'copy', 'copy_id' )

	local willNotUploadMaster = KSTableUtils.derefOpt( docValue,
													   'importSource',
													   'willNotUploadMaster' )

	local masterPath = WFCatFileSpec.buildOriginalPathFromDocValue( docValue )

	-- Only try to upload a proxy if it doesn't have a proxy or master link.
	-- and the asset indicates it was imported from lightroom without a master.
	-- Also assume that asset has to have a proxy locally.

	if willNotUploadMaster
		and not masterPath
		and ( localPath or ( localVolumeGuid and localRelativePath ) or copyId )
		and ozProxyCreateTemplate
		and latestRevisionId
		and not ozProxyUrl
	then

		local contentType = kContentTypeDng

		local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )

		local fileSize = WFCatFileSpec.getOriginalSizeFromDocValue( docValue )

		local proxyCreateUrl = ozProxyCreateTemplate:replace( '{revision_id}', latestRevisionId )

		-- Make a new guid for the proxyId, Nimbus doesn't use this guid outside
		-- of the pre-built string in the proxy link (typically it's created by oz).
		local proxyId = KSIdGenerator.uuidInLowercaseHex()
		proxyCreateUrl = proxyCreateUrl:replace( '{proxy_id}', proxyId )

		local binaryType = 'proxy'
		if proxySize == 'panorama' then
			binaryType = 'proxy_pano'
		end

		local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceProxyLink' )

		logger:debugf( "emit %s, %s", localPath, proxyCreateUrl )
		emit( localPath, localVolumeGuid, localRelativePath, boolToInt( fileMaybeOnOfflineVolume ), proxyCreateUrl, binaryType, contentType, importDate,
				captureDate, isBlacklisted, imageSha256, fileSize, boolToInt( false ), priorityForUpload, sourceLinkForCopy )

	end

end

--------------------------------------------------------------------------------

local function generateIndexSchema( _, indexStoreSchema )

	check( '_', 'WFRxIndexStoreSchema' )

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE needsBinaryUpload(
				localFilePath,
				localVolumeGuid,
				localRelativePath,
				fileMaybeOnOfflineVolume,
				uploadUrl,
				binaryType,
				contentType,
				importDate,
				captureDate,
				isBlacklisted,
				sha256,
				fileSize,
				sha256Error,
				priorityForUpload,
				sourceLinkForCopy )]],
		
		docValueFilter = function( docValue )
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )

			--logger:debugf( "indexer reviewing %s", docValue )
			
			if doesAssetDocumentHaveLocalBinary( docValue ) or isAssetDocumentACopy( docValue ) then
				
				local isBlacklisted = checkIfBlacklisted( docValue )

				local importDate, captureDate = getDatesFromDocValue( docValue )
	
				local priorityForUpload = WFCatPhotoImportance.getPriorityForUpload( docValue )
	
				-- maybe add row for xmp/develop
				maybeEmitXmpDevelopUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )

				-- maybe add row for full xmp sidecar
				maybeEmitXmpSidecarUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )

				-- maybe add row for proxy but only if no master exists (copied Lightroom asset)
				maybeEmitProxyUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
				
				-- maybe add row for master
				maybeEmitMasterUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload, 26 )

			end

		end,

	}

	indexStoreSchema:addIndex( 
		[[CREATE INDEX needsBinaryUpload_isBlistedImportCapture 
			ON needsBinaryUpload( fileMaybeOnOfflineVolume, isBlacklisted, priorityForUpload DESC, importDate, captureDate )]] )

end

--------------------------------------------------------------------------------

local function isMaster( binaryType )

	return binaryType == 'master' or binaryType == 'preset_master' or binaryType == 'camera_profile'

end

--------------------------------------------------------------------------------

local function uploadFinished( self )

	if self._uploadFinished then

		-- Schedule the call to _uploadFinished to prevent stack overflow.
		WFRxScheduler.sharedScheduler():schedule( function()
			-- the value here is unimportant, but making it unique can help in debugging
			self._uploadFinished:onNext( KSDate.timeInIso8601Format() )
		end )

	end

end

--------------------------------------------------------------------------------

local function putAndTrackPerformance( self, masterUrl, binaryType, filePath, fileSize, options, docId )

	local putFile

	if isMaster( binaryType ) and fileSize >= self._chunkedUploader:chunkFileSize() then
		putFile = self._chunkedUploader:doChunkedUpload( masterUrl, filePath, options, docId )
	else
		putFile = self._catalogSpec:putBinary( masterUrl, filePath, options )
	end

	return WFPerformanceReport.wrapObservable( string.format( 'WFCatBinaryUploader: %s', binaryType ), 
		putFile, { fileSize = fileSize } )

end

--------------------------------------------------------------------------------

local function getRecentlyUploadedSha256( self, docId, binaryType )
	
	return derefOpt( self._recentlyUploaded, docId, binaryType )

end

--------------------------------------------------------------------------------

local function getUploadingAssetsWithSha256( self, sha256 )

	return sha256 and self._uploadingAssetWithSha256[ sha256 ] or 0

end

--------------------------------------------------------------------------------

local function addToRecentlyUploaded( self, docId, binaryType, sha256 )

	local recentlyUploadedForDocId = self._recentlyUploaded[ docId ] or {}

	if recentlyUploadedForDocId[ binaryType ] then
		logger:errorf( "addToRecentlyUploaded recentlyUploadedForDocId ALREADY SET for %s %s", binaryType, docId )
	end

	recentlyUploadedForDocId[ binaryType ] = sha256 

	self._recentlyUploaded[ docId ] = recentlyUploadedForDocId

	if isMaster( binaryType ) then
		local countOfAssets = self._uploadingAssetWithSha256[ sha256 ] or 0
		self._uploadingAssetWithSha256[ sha256 ] = countOfAssets + 1
	end

end

--------------------------------------------------------------------------------

local function clearFromRecentlyUploaded( self, reason, docId, binaryType )

	logger:infof( "clearFromRecentlyUploaded %s %s because %s", binaryType, docId, reason )

	local recentlyUploadedForDocId = self._recentlyUploaded[ docId ] or {}

	if isMaster( binaryType ) then

		local sha256 = getRecentlyUploadedSha256( self, docId, binaryType )

		if self._uploadingAssetWithSha256[ sha256 ] then

			local newCountOfAssets = self._uploadingAssetWithSha256[ sha256 ] - 1

			if newCountOfAssets > 0 then
				self._uploadingAssetWithSha256[ sha256 ] = newCountOfAssets
			else
				self._uploadingAssetWithSha256[ sha256 ] = nil
			end

		end

	end

	recentlyUploadedForDocId[ binaryType ] = nil

	if next( recentlyUploadedForDocId ) then 
		self._recentlyUploaded[ docId ] = recentlyUploadedForDocId
	else
		self._recentlyUploaded[ docId ] = nil
	end

end

--------------------------------------------------------------------------------

local function delayedClearFromRecentlyUploaded( self, reason, docId, binaryType, optDelay )

	-- Because the index store takes a few seconds to update, and the doc store
	-- could have potentially just updated, we will give the index store some time
	-- to bring itself up to date.
	
	-- TODO: Is there a more deterministic way to do this? An index store API that fires
	-- when it's up to date?

	local indexStoreUpdateBuffer = optDelay or 30

	logger:infof( "Calling clearFromRecentlyUploaded in %d seconds for doc %s type %s because %s.", indexStoreUpdateBuffer, docId, binaryType, reason )

	return self._timerAborter
		:flatMapLatest( function()
			return WFRxLowFidelityTimer.timer( indexStoreUpdateBuffer )
		end )
		:tap( function()

			clearFromRecentlyUploaded( self, reason, docId, binaryType )

			-- Trigger combineLatest in uploadSomeOriginals as we may need to try uploading docId again
			uploadFinished( self )

		end )
		:call( runUntilDone )
		
end

--------------------------------------------------------------------------------

local function addToActiveUploads( self, docId, binaryType, subscription )

	self._binariesInFlight = self._binariesInFlight + 1

	local activeUploadsForDocId = self._activeUploads[ docId ] or {}

	if activeUploadsForDocId[ binaryType ] then
		logger:errorf( "Already have subscription for uploading %s binaryType = %s", docId, binaryType )
	end

	activeUploadsForDocId[ binaryType ] = subscription
	
	self._activeUploads[ docId ] = activeUploadsForDocId
	
end

--------------------------------------------------------------------------------

local function removeFromActiveUploads( self, docId, binaryType )

	if derefOpt( self._activeUploads, docId, binaryType ) == nil then
		-- Already removed by completed or error, finally will call a second
		-- time but is needed in case the subscription is disposed before
		-- completed or error.
		return
	end

	local activeUploadsForDocId = self._activeUploads[ docId ] or {}
	
	activeUploadsForDocId[ binaryType ] = nil
	
	if next( activeUploadsForDocId ) then
		self._activeUploads[ docId ] = activeUploadsForDocId
	else
		self._activeUploads[ docId ] = nil
	end

	self._binariesInFlight = self._binariesInFlight - 1

	if self._binariesInFlight < 0 then
		logger:errorf( "_binariesInFlight went negative == %d", self._binariesInFlight )
		self._binariesInFlight = 0
	end
	if self._binariesInFlight > 0 and next( self._activeUploads ) == nil then
		logger:errorf( "_binariesInFlight == %d and _activeUploads is empty", self._binariesInFlight )
	end

end

--------------------------------------------------------------------------------

local function eventuallyClearBlackoutWindow( self, docId, binaryType, waitCount )

	-- Don't let this original be uploaded again

	waitCount = waitCount or 0

	self._timerAborter
		:flatMapLatest( function()
			return WFRxLowFidelityTimer.docValueTimer( uploadBlackoutWindowInSeconds, self._docStore, docId )
		end )
		:tap( function( docValue )

			local uploadedSha256, currentSha256
			local uploadConfirmed

			if not isValidAssetDocument( docValue ) and not isNonDeletedStyleDocument( docValue )
					and not WFCatIndexUtils.isNonDeletedDCPDocument( docValue ) then

				-- Clear out RecentlyUploaded for deleted documents.
				logger:warnf( "Asset deleted prior to upload completion for %s asset %s.", binaryType, docId )

				uploadConfirmed = true

			elseif isMaster( binaryType ) then
				
				uploadConfirmed = getOriginalLinkFromDocValue_IgnoringShaError( docValue ) ~= nil
				uploadedSha256 = getRecentlyUploadedSha256( self, docId, binaryType )
				currentSha256 = derefOpt( docValue, 'importSource', 'sha256' )

			elseif binaryType == 'xmp_develop' then
				
				uploadConfirmed = WFCatFileSpec.getXmpDevelopLinkFromDocValue( docValue ) ~= nil
				uploadedSha256 = getRecentlyUploadedSha256( self, docId, binaryType )
				currentSha256 = derefOpt( docValue, 'develop', 'xmpCameraRaw', 'sha256' )

			elseif binaryType == 'xmp_sidecar' then

				-- Verify _localyOnly.xmpSidecarNeedsUpload entry is removed from the docValue.
				uploadConfirmed = WFCatFileSpec.getXmpSidecarNeedsUploadFromDocValue( docValue ) == nil
				uploadedSha256 = nil
				currentSha256 = nil

			end

			if uploadedSha256 ~= currentSha256 then
			
				-- If there was a destructive edit that caused the original to change,
				-- resulting in a new sha256, we clear the asset from the list so that 
				-- we may upload the new original. It should be noted that we don't yet
				-- support destructive edits in many other parts of sync.
				
				-- This probably happens when xmp_develop is modified during upload.

				logger:warnf( "sha no longer matches for %s type %s", docId, binaryType )

				delayedClearFromRecentlyUploaded( self, "sha no longer matches", docId, binaryType )
				
			elseif uploadConfirmed then
			
				-- Once the docValue contains an Oz master href, we can safely remove
				-- the asset from the recently uploaded table since we won't try to reupload
				-- assets that have oz master links.

				delayedClearFromRecentlyUploaded( self, "upload confirmed", docId, binaryType )

			else
			
				waitCount = waitCount + 1
			
				logger:warnf( "Doc store hasn't updated with the presence of uploaded %s for %s after %d seconds",
						binaryType, docId, uploadBlackoutWindowInSeconds * waitCount )
						
				if waitCount > 6 then
					
					logger:warnf( "Assuming upload failed for asset %s. Retrying...", docId )
					
					delayedClearFromRecentlyUploaded( self, "assuming upload failed", docId, binaryType )
					
				else
				
					-- Keep waiting
					
					eventuallyClearBlackoutWindow( self, docId, binaryType, waitCount )
					
				end
			end
			
		end )
		
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

local function uploadCompleted( self, docId, binaryType, uploadSha256 )

	eventuallyClearBlackoutWindow( self, docId, binaryType )
	
	local docStore = self._docStore
	
	if isMaster( binaryType ) then
		
		-- NMB-5779: We may have received the master via migration or another
		-- change which required us to wait until Oz marked the previous binaries
		-- as invalid.  If this is the case, remove the annotation that tells us 
		-- we must wait
		docStore:docValue( docId )
			:flatMap( function(docValue )
				
				local waitForInvalid = derefOpt( docValue, '_localOnly', 'binaryUpload', 'waitForInvalid' )
				
				-- NMB-7626: Previously wait for invalid was simply a boolean true. With the introduction of presets
				-- which are both small and the file is changeable, we 
				if waitForInvalid == true -- old case (< 1.4) which may exist in doc store before upgrade
				    or ( waitForInvalid == uploadSha256 ) then -- 1.4+
					return docStore:applyOverlay {
						_id = docId,
						_localOnly = {
							binaryUpload = {
								waitForInvalid = { _delete = true },
							},
						},
					}
					
				else
					return Observable.just()
				end
			end )
			:call( runUntilDone )
			
	elseif binaryType == 'xmp_sidecar' then

		-- Originals and develop settings have links set in the _serverProvidedMetadata
		-- for the binary's location on oz, whereas xmp is generated when
		-- an original is received.  So there is no way to detect the upload has
		-- completed by watch the sync feed, so update _localyOnly.xmpSidecarNeedsUpload
		-- when completed so that eventuallyClearBlackoutWindow can clear the upload
		-- from the blackout window when it's timer fires.

		docStore:applyOverlay  {
			_id = docId,
			_localOnly = {
				xmpSidecarNeedsUpload = { _delete = true },
			},
		}
		:call( runUntilDone )
		
	end

end

--------------------------------------------------------------------------------

local function addTransferStatsToDiagnosticLog( self, bucket, time, size )

	local sizeInMB = size / (1024 * 1024)

	local bucketStats = self._uploadSpeedStats.buckets[ bucket ]
	local uploadSpeed = sizeInMB / time

	if not bucketStats then

		bucketStats = {
			count = 0,
			timings = WFDiagnosticLog.makeRingBuffer { "Time (sec)" },
			sizes = WFDiagnosticLog.makeRingBuffer { "Size (MB)" },
			speed = WFDiagnosticLog.makeRingBuffer { "Upload speed (MBps)" },
		}

		self._uploadSpeedStats.buckets[ bucket ] = bucketStats

	end

	bucketStats.count = bucketStats.count + 1
	bucketStats.timings:add( time )
	bucketStats.sizes:add( sizeInMB )
	bucketStats.speed:add( uploadSpeed )

	local totalBinarySize = self._uploadSpeedStats.totalBinarySize or 0
	self._uploadSpeedStats.totalBinarySize = totalBinarySize + sizeInMB

end

--------------------------------------------------------------------------------

local function addHttpErrorToLog( errorLog, docId, masterPutUrl, responseData, headers )

	-- Special case for HTTP transit failure.

	if not headers then

		local annotatedReason = string.format( "%s", tostring( responseData ) )

		local logFmtStr = "Unable to communicate with Oz URL %s\n\nreason = %s"

		logger:errorf( logFmtStr, masterPutUrl, annotatedReason )

	else

		-- We were able to communicate with Oz, but Oz didn't like the response. Log this.

		if not headers then
			headers = {}
		end

		local logFmtStr = "Got status %s back from Oz\n\n"
		local logParams = { headers.status_code or 'nil' }

		if masterPutUrl then
			logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = masterPutUrl
		end

		logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
		logParams[ #logParams + 1 ] = responseData

		errorLog[ #errorLog + 1 ] = {

			time = KSDate.timeInIso8601FormatWithTimeZone(),

			id = docId,
			binaryType = 'master',
			error = string.format( logFmtStr, unpack( logParams ) ),

		}

		if #errorLog > 10 then
			table.remove( errorLog, 1 )
		end

		logger:errorf( logFmtStr, unpack( logParams ) )

	end

end

--------------------------------------------------------------------------------

local function addInvalidOriginalErrorToLog( errorLog, docId, filePath, reason )

	errorLog[ #errorLog + 1 ] = {

		time = KSDate.timeInIso8601FormatWithTimeZone(),

		id = docId,
		binaryType = 'master',
		error = reason,
		filePath = filePath,
	}

	logger:errorf( "Marking original as invalid for docId %s path %s", docId, filePath )

end

--------------------------------------------------------------------------------

local function getRequestOptionsFor( binaryType, contentType )
	
	local options = {
		dataFormat = 'filePath',
		requestPriority = WFRxHttpClient.requestPriorityMedium,
		requestCost = uploadHttpRequestCost
	}
	
	if binaryType == 'master' then
		
		local requestHeaders = {
			{
				header = 'X-Generate-Renditions',
				value = 'all',
			},
			{
				header = 'Content-Type',
				value = contentType or 'image/jpeg',
			},
			{
				header = "Accept",
				value = "application/json",
			},
		}
		
		if not WFOzContentTypes.isVideo( contentType ) then
			-- don't request proxy generation for videos, but 
			-- do for image/* and raw files (application/octet-stream)
			table.insert( requestHeaders, 
				{
					header = 'X-Generate-Proxies',
					value = 'all',
				} )
		end
		
		options.requestHeaders = requestHeaders

	elseif binaryType == 'preset_master' then
		-- master for preset/profile
		
		local requestHeaders = {
			{
				header = 'Accept',
				value = 'application/json',
			},
			{
				header = 'Content-Type',
				value = contentType or kContentTypeXML,
			},
		}
		
		options.requestHeaders = requestHeaders
		
	elseif binaryType == 'camera_profile' then
		-- master for camera_profile
		local requestHeaders = {
			{
				header = 'Accept',
				value = 'application/json',
			},
			{
				header = 'Content-Type',
				value = contentType or kContentTypeOctetStream,
			},
		}
		options.requestHeaders = requestHeaders
		
	elseif binaryType == 'proxy' or binaryType == 'proxy_pano' then

		local proxyType = '2560'
		if binaryType == 'proxy_pano' then
			proxyType = 'panorama'
		end

		options.requestHeaders = {
				{
					header = 'X-Proxy-Type',
					value = proxyType,
				},
				{
					header = 'Content-Type',
					value = contentType or kContentTypeDng,
				},
				{
					header = 'Accept',
					value = 'application/json',
				},
			}
		
	elseif binaryType == 'xmp_develop' then
		
		options.requestHeaders = {
				{
					header = 'X-Generate-Renditions',
					value = 'all',
				},
				{
					header = 'X-Generate-Proxies',
					value = 'all',
				},
				{
					header = 'Content-Type',
					value = contentType or kContentTypeXML,
				},
				{
					header = 'Accept',
					value = 'application/json',
				},
			}

	elseif binaryType == 'xmp_sidecar' then
		
		options.requestHeaders = {
				{
					header = 'Content-Type',
					value = contentType or kContentTypeXML,
				},
				{
					header = 'Accept',
					value = 'application/json',
				},
			}
			
	end
	
	return options

end

--------------------------------------------------------------------------------

local function recordSuccess( self, docId, startTime, uploadUrl, binaryType, localSize, response, isCopy )
	
	local bucket = binaryType
	if localSize >= self._chunkedUploader:chunkFileSize() then
		bucket = bucket .. "-chunked"
	end

	if isCopy then
		bucket = bucket .. "-copy"
	end

	addTransferStatsToDiagnosticLog( self, bucket, now() - startTime, localSize )
	
	if not isCopy and binaryType ~= 'preset_master' then
	
		local broadcastData = {
			docId = docId,
			binaryType = binaryType,
			size = localSize,
		}
	
		NmbTrackedDataBroadcaster.broadcastMessage( 'UploadSuccess', broadcastData )
		
	end
	
	logger:infof( "Uploaded %s to %s, response = %s", binaryType, uploadUrl, response )

end

--------------------------------------------------------------------------------

local function removeSidecarFileAnnotation( self, docId )

	local binarySizes = {}
	binarySizes[ 'xmp' ] = { _delete = true }
	
	local files = {}
	files[ 'original' ] = binarySizes

	return self._docStore:applyOverlay {
			_id = docId,
			_localOnly = {
				files = files,
				xmpSidecarNeedsUpload = { _delete = true },
				xmpSidecarExists = false,
			}
		}
		
end

--------------------------------------------------------------------------------

local function applyInvalidOriginalOverlay( self, docId, filePath, reason )

	reason = reason or 'InvalidLocalOriginal'

	local shouldApplyOverlay = true
	
	local volumeIsOnline = filePath and WFFileUtils.getVolumeGuidForPath( filePath )
	
	if filePath and not volumeIsOnline then
	
		-- If the file is on a volume which was disconnected
		-- in the middle of the upload, then we shouldn't apply the
		-- error overlay.
	
		shouldApplyOverlay = false

	end

	if shouldApplyOverlay then
	
		addInvalidOriginalErrorToLog( self._uploadErrorLog, docId, filePath, reason )

		-- Mark the original as invalid since the sha256 no longer matches what we imported.

		-- This can happen because of disk corruption or from an invalid copy during import.
		-- (since sha may be calculated prior to importing by the Import UI).

		-- It could also happen if the user copied a new file into the ~/Pictures/Nimbus folder
		-- manually (which they should not do).

		-- Adding localOriginalError will cause this file to no longer be considered for uploaded.

		return self._docStore:applyOverlay {
				_id = docId,
				_localOnly = {
					localOriginalError = reason,
				}
			}

	else
	
		return Observable.just( false )
	
	end

end

--------------------------------------------------------------------------------

local function resolveFailedChunkedUpload( self, masterUploadUrl, filePath, docId, imageSha256 )
 
	-- Recalculate sha256 to verify it matches the sha256 in the importSource to
	-- verify we have the correct file to upload.

	return WFAsyncDigest.digestFile( filePath, 'SHA256' )

	:flatMap( function( localSha256 )

		if localSha256 == imageSha256 then

			-- Update the doc so the binary uploader will try to upload this file again.

			logger:debugf( "re-uploading corrupted original for docId %s path %s", docId, filePath )

			return self._docStore:docValue( docId )
				:flatMap( function( docValue )
					
					local currentMasterUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
	
					return self._docStore:applyOverlay {
						_id = docId,
						_localOnly = { 
							-- Add a local annotation for the sha error since we're going to
							-- otherwise remove all traces that it happened.  This is necessary
							-- to prevent us from accidentally deleting the local master because
							-- we think the master is uploaded and fine (NMB-4862)
							binaryUploadErrors = {
								[ '/rels/master' ] = {
									[ currentMasterUrl ]  = "corrupt upload",
								}
							}
						},
					}
			end )
			
		else

			return applyInvalidOriginalOverlay( self, docId, filePath )

		end

	end )

end

--------------------------------------------------------------------------------

local function disposeActiveUploads( self )

	logger:debugf( "disposeActiveUploads started" )

	-- Abort the timers we have for clearing the recently uploaded fields
	-- since we are clearing them out immediately below.  This prevents
	-- a timer from a prior upload request from being issued which can cause
	-- a file to be uploaded multiple times simultaneously.	
	self._timerAborter:onError( 'Abort' )

	-- Create a new _timerAborter that isn't in the terminal state for the
	-- next set of timers.
	self._timerAborter = Rx.ReplaySubject( 1 )
	self._timerAborter:onNext( true )

	for docId, uploadsForDocId in pairs( self._activeUploads ) do
		
		for binaryType, subscription in pairs( uploadsForDocId ) do 
		
			logger:debugf( "Aborting upload for docId %s", docId )
	
			subscription:dispose()
	
			clearFromRecentlyUploaded( self, "disposed", docId, binaryType )
			
		end
		
	end

	if self._binariesInFlight > 0 then
		logger:errorf( "_binariesInFlight == %d should be 0", self._binariesInFlight )
	end

	self._binariesInFlight = 0

	self._recentlyUploaded = {}

	self._activeUploads = {}

	self._uploadingAssetWithSha256 = {}

end

--------------------------------------------------------------------------------

local function handleUploadFailure( self, docId, binaryType, uploadUrl, responseData, headers, invalidOriginal )

	self._errorStatus:onNext( docId, responseData )

	addHttpErrorToLog( self._uploadErrorLog, docId, uploadUrl, responseData, headers )

	if invalidOriginal then
		delayedClearFromRecentlyUploaded( self, "wait for docstore update as invalidOriginal", docId, binaryType )
		return
	end
	
	if self._failedUploadBlacklistHelper:maybeBlacklist( docId ) then

		-- The asset is now blacklisted. It will be automatically removed from the
		-- blacklist after failedUploadBlackoutWindowInSeconds seconds.

		WFPerformanceReport.recordSingleValue( 'WFCatBinaryUploader:Blacklist', {
			docId = docId,
			numberOfFailedAttempts = numberOfTriesBeforeBlacklistingUpload,
			blacklistDuration = failedUploadBlackoutWindowInSeconds
		} )

		delayedClearFromRecentlyUploaded( self, "wait for docstore update as blacklisted", docId, binaryType )

	else
		
		local statusCode = headers and headers.status_code or 'unknown'
		if not headers and string.match( responseData, "timed out" ) then
			statusCode = "timeout"
		end
		
		local broadcastData = {
			description = "An upload failure has occurred.",
			docId = docId,
			binaryType = binaryType,
			uploadUrl = uploadUrl,
			statusCode = statusCode
		}
	
		NmbTrackedDataBroadcaster.broadcastMessage( 'UploadFailure', broadcastData )

		-- Upload wasn't blacklisted so try to upload the original again.

		local uploadErrorDelay = 10
		delayedClearFromRecentlyUploaded( self, "upload failed, delay to let other assets upload first", docId, binaryType, uploadErrorDelay )

	end

end

--------------------------------------------------------------------------------

local function errorShaDoesNotMatch( binaryType, responseData, headers )

	if isMaster( binaryType ) and headers and headers.status_code == '400' and headers[ 'X-Result-Code' ] == '1005' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'errors', 'sha256', 1 ) == "does not match"

	end

	return false

end

--------------------------------------------------------------------------------

local function errorTrialOrSubscriptionExpired( binaryType, responseData, headers )

	if headers and headers.status_code == '403' and headers[ 'X-Result-Code' ] == '4300' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'subtype' ) == WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError

	end

	return false

end

--------------------------------------------------------------------------------

local function errorUploadMasterWhenNotAllowed( binaryType, responseData, headers )

	if isMaster( binaryType ) and headers and headers.status_code == '403' and headers[ 'X-Result-Code' ] == 'UploadMasterWhenNotAllowedError' then

		return true

	end

	return false

end

--------------------------------------------------------------------------------

local function errorPayloadMustExist( binaryType, responseData, headers )

	if not isMaster( binaryType ) then
		return false
	end

	if responseData == 'payload must exist' then
		-- Chunked uploader check file size up front as it can't be 0.
		return true
	end

	if headers and headers.status_code == '400' and headers[ 'X-Result-Code' ] == '1005' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'errors', 'payload', 1 ) == "must exist"

	end

	return false

end

--------------------------------------------------------------------------------

local function applyRemoveWillNotUploadMasterOverlay( self, docId )

	return self._docStore:applyOverlay {
			_id = docId,
			importSource = {
				willNotUploadMaster = { _delete = true },
			},
		}

end

--------------------------------------------------------------------------------

local function applyTrialOrSubscriptionExpiredOverlay( self, docId )
	
	local overlay = WFCatOzErrors.createOverlayForOzErrorSubtype( docId, WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError )
	return self._docStore:applyOverlay( overlay )
	
end

--------------------------------------------------------------------------------

local function errorMasterOrProxyResourceDoesNotExist( binaryType, responseData, headers )

	if isMaster( binaryType ) or binaryType == 'proxy' and headers and headers.status_code == '404' and headers[ 'X-Result-Code' ] == '1000' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'errors', 'revision', 1 ) == "does not exist"

	end

	return false

end

--------------------------------------------------------------------------------

local function generateRenditionsOnOz( self, generateFilesUrl )

	local generateFilesOptions = {
		requestMethod = 'POST',
		requestHeaders = {
			{
				header = 'X-Generate-Renditions',
				value = 'all',
			},
		},
	}

	return self._catalogSpec:getBinary( generateFilesUrl, generateFilesOptions )

end

--------------------------------------------------------------------------------

local function generateRenditionsIfProxy( self, binaryType, binaryUrl, response, headers )

	if binaryType == 'proxy' or binaryType == 'proxy_pano' then

		-- We only upload a proxy if no master exists for the asset, in this case
		-- we need to ask oz to generate renditions for the proxy otherwise it won't
		-- show up correctly in Lightroom Web.
		local generateRenditionsUrl = string.gsub( binaryUrl, "proxies/(.*)$", "renditions" )
		return generateRenditionsOnOz( self, generateRenditionsUrl )
			:map( function()
				-- pass on successful results from the proxy upload instead
				return response, headers
			end)

	else

		return Observable.just( response, headers )

	end

end

--------------------------------------------------------------------------------

local function uploadOneBinary( self, docId, localPath, localSize, uploadUrl, binaryType, contentType, sha256 )

	logger:debugf( "Uploading %s for docId %s from %s to %s", binaryType, docId, localPath, uploadUrl )

	local options = getRequestOptionsFor( binaryType, contentType )
	options.requestDescription = string.format( '%s : Uploading %s', docId, binaryType )
	
	local startTime = now()

	return putAndTrackPerformance( self, uploadUrl, binaryType, localPath, localSize, options, docId )

	:flatMap( function( response, headers )
		return generateRenditionsIfProxy( self, binaryType, uploadUrl, response, headers )
	end )

	:catch( function( responseData, headers )

		if errorShaDoesNotMatch( binaryType, responseData, headers ) then

			-- Recalculate sha256 to verify it matches the sha256 in the importSource to
			-- verify we have the correct file to upload.

			return WFAsyncDigest.digestFile( localPath, 'SHA256' )
				:flatMap( function( localSha256 )

					if localSha256 == sha256 then

						return Observable.throw( responseData, headers, false )

					else

						return applyInvalidOriginalOverlay( self, docId, localPath )
							:flatMap( function()
								return Observable.throw( responseData, headers, true )
							end )

					end
				end )

		end


		if errorUploadMasterWhenNotAllowed( binaryType, responseData, headers ) then

			return applyRemoveWillNotUploadMasterOverlay( self, docId )

				:flatMap( function()

					-- Throw an error so we retry the upload after the doc store
					-- has had time to push the removal of the original through
					-- the changes feed.

					return Observable.throw( responseData, headers, true )

				end )

		end

		if errorPayloadMustExist( binaryType, responseData, headers ) then

			return applyInvalidOriginalOverlay( self, docId, localPath )
				:flatMap( function()
					return Observable.throw( responseData, headers, true )
				end )

		end
		
		if errorTrialOrSubscriptionExpired( binaryType, responseData, headers ) then
			
			return applyTrialOrSubscriptionExpiredOverlay( self, docId )
				:flatMap( function()
					return Observable.throw( responseData, headers, true )
				end )
		
		end

		return Observable.throw( responseData, headers, false )

	end )

	:tap(
		function( response )
			recordSuccess( self, docId, startTime, uploadUrl, binaryType, localSize, response )
		end,
		nil,
		function()
			uploadCompleted( self, docId, binaryType, sha256 )
		end )
		
	:catch( function( responseData, headers, invalidOriginal )

			-- On binary uploads of the xmp_develop where we request rendition generation we
			-- may get a 404 with a SourceDoesNotExist error from Oz if a master/proxy does
			-- not yet exist. In these cases, the upload succeeds, but the rendition generation
			-- does not. This failure can be treated as a success as we will request renditions
			-- again when the master/proxy is uploaded
			
			local uploadFailed = true
			
			if binaryType == 'xmp_develop' and headers.status_code == '404' then
				local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )
				
				if ok and type( response ) == 'table' and response.code == "SourceDoesNotExist" then
				
					recordSuccess( self, docId, startTime, uploadUrl, binaryType, localSize, responseData )
					uploadCompleted( self, docId, binaryType, sha256 )
					
					uploadFailed = false
					
				end
				
			end

			if uploadFailed then
				handleUploadFailure( self, docId, binaryType, uploadUrl, responseData, headers, invalidOriginal )
			end
			
			return Observable.just( false )

	end )


end

--------------------------------------------------------------------------------

local function copyMaster( self, docId, uploadUrl, binaryType, contentType, sourceAssetLink )

	logger:debugf( "Copying %s for docId %s from %s to %s", binaryType, docId, sourceAssetLink, uploadUrl )

	local data = '{ "file_href": "' .. sourceAssetLink .. '" }'

	local options = getRequestOptionsFor( binaryType, 'application/json' )
	table.insert( options.requestHeaders, {
			header = 'X-Async-Copy',
			value = 'true'
		} )
	options.requestCost = WFRxHttpClient.requestCostLow
	options.dataFormat = nil

	return self._catalogSpec:putBinary( uploadUrl, data, options )
		:map( function( response, headers )
			return #data, response, headers
		end )
		:catch( function( responseData, headers )
			
			if errorUploadMasterWhenNotAllowed( binaryType, responseData, headers ) then

				return applyRemoveWillNotUploadMasterOverlay( self, docId )

					:flatMap( function()
	
						-- Throw an error so we retry the upload after the doc store
						-- has had time to push the removal of the original through
						-- the changes feed.
	
						return Observable.throw( responseData, headers, true )
	
					end )

			end
	
			if errorTrialOrSubscriptionExpired( binaryType, responseData, headers ) then
				
				return applyTrialOrSubscriptionExpiredOverlay( self, docId )
					:flatMap( function()
						return Observable.throw( responseData, headers, true )
					end )
			
			end
	
			return Observable.throw( responseData, headers, false )
			
		end )

end

--------------------------------------------------------------------------------

local function copyOneBinary( self, docId, binaryUrl, binaryType, sha256, contentType, sourceLinkForCopy )

	local copyBinaryObservable

	local startTime = now()

	if isMaster( binaryType ) then

		copyBinaryObservable = copyMaster( self, docId, binaryUrl, binaryType, contentType, sourceLinkForCopy )

	else

		copyBinaryObservable = self._catalogSpec:getBinary( sourceLinkForCopy )
		:flatMap( function( data, headers )
			local options = getRequestOptionsFor( binaryType )
			options.dataFormat = nil
			return Observable.combineLatest(
				Observable.just( KSData.length( data ) ),
				self._catalogSpec:putBinary( binaryUrl, data, options )
				:flatMap( function( response, headers )
					return generateRenditionsIfProxy( self, binaryType, binaryUrl, response, headers )
				end )
			)
		end )

	end

	copyBinaryObservable = copyBinaryObservable:tap( function( dataLength, response, headers )
		recordSuccess( self, docId, startTime, binaryUrl, binaryType, dataLength, response, true )
	end,
	nil, -- caller must issue call to handleUploadFailure
	function()
		uploadCompleted( self, docId, binaryType, sha256 )
	end )

	return copyBinaryObservable

end

--------------------------------------------------------------------------------

local function uploadOrCopyOneMasterOrProxy( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256, isFirstUploadForSha256 )

	return WFCatImportDuplicateFiles.findUploadedAssetWithSha256HintNotDocId( self._catalog, sha256, docId )
		:flatMap( function( sourceAssetId, sourceHasNoMaster )

			if sourceAssetId == nil then

				-- Upload the asset.

				if isFirstUploadForSha256 then

					if localFilePath then

						return uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256 )

					else

						-- Copied an asset that didn't have a master and source was deleted before we could upload it.

						return applyInvalidOriginalOverlay( self, docId, nil, "CopyFailedSourceMissing" )
							:finally( function()
								delayedClearFromRecentlyUploaded( self, "upload failed, wait for docstore update as CopyFailedSourceMissing", docId, binaryType )
							end )

					end

				else

					delayedClearFromRecentlyUploaded( self, "Uploading first asset with same sha256, delay so we can copy later", docId, binaryType, 10 )
	
					-- Complete right away to let other assets in, as uploadFinished will get called.
					return Observable.just( true )

				end

			end

			-- Copy the asset.
			return self._docStore:docValue( sourceAssetId )
			:flatMap( function( sourceDocValue )

				-- SourceAssetLink may be nil if the source asset has an invalid original.
				local sourceAssetLink
				if isMaster( binaryType ) then
					sourceAssetLink = WFCatFileSpec.getOriginalLinkFromDocValue( sourceDocValue )
				else
					sourceAssetLink = WFCatFileSpec.getProxyLinkFromDocValue( sourceDocValue )
				end

				if sourceAssetLink == nil then

					if localFilePath then

						return uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256 )

					else

						-- Copied an asset that didn't have a master that was deleted before we could upload it.
						return applyInvalidOriginalOverlay( self, docId, nil, "CopyFailedSourceMissing" )
							:finally( function()
								delayedClearFromRecentlyUploaded( self, "copy failed, wait for docstore update CopyFailedSourceMissing", docId, binaryType )
							end )

					end

				else

					return copyOneBinary( self, docId, binaryUrl, binaryType, sha256, contentType, sourceAssetLink )
					:tapOnError( function( responseData, headers )
						handleUploadFailure( self, docId, binaryType, binaryUrl, responseData, headers, false )
					end )

				end

			end )

		end )

end

--------------------------------------------------------------------------------

local function maybePerformanceWrap( id, observable )

	id = id or 'unknown'
	
	return WFPerformanceReport.wrapObservable( 'WFCatBinaryUploader: ' .. id, observable )

end 

--------------------------------------------------------------------------------

local function uploadSomeBinaries( self )
	
	local needsUploadQueryStr = 
		[[SELECT docId, localFilePath, uploadUrl, binaryType, contentType, 
				isBlacklisted, sha256, fileSize, sha256Error, sourceLinkForCopy,
				localVolumeGuid, localRelativePath
			FROM needsBinaryUpload
			ORDER BY fileMaybeOnOfflineVolume ASC, isBlacklisted ASC, priorityForUpload DESC, importDate ASC, captureDate ASC 
			LIMIT 100]]

	local needsUploadQuery =
			self._catalog:indexStore():query {
				sql = needsUploadQueryStr,
				quiesceUntilUpToDate = true,
			}

	needsUploadQuery = maybePerformanceWrap( 'query', needsUploadQuery )

	-- There's potentially for a race condition where the query might be refreshed
	-- before we've determined on this end that the upload has finished. When this happens,
	-- the below rx chain will do no work since we still think we have the maximum
	-- number of uploads in flight. This can potentialy cause sync to freeze. In
	-- order to remedy this, we use combineLatest to cause the rx chain to fire on
	-- query refresh, as well as when an upload finishes.
	
	return Observable.combineLatest(
			needsUploadQuery,
			self._uploadFinished,
			self._volumeStatusHasChanged )
			
	:flatMapLatest( function( docsAndUrls )

		return self._continuousUploadTask:throwIfNotRunning( docsAndUrls )
			-- Need to throw as uploadSomeBinaries doesn't ever complete.

	end )

	:tap( function( docsAndUrls )

--		logger:debugf( "uploadSomeBinaries raw query result = %s", docsAndUrls )
		logger:debugf( "uploadSomeBinaries raw query result = %s", #docsAndUrls )

		if #docsAndUrls == 0 then
			self._continuousUploadTask:taskDidWork( false )
				-- Since we don't typically send onCompleted events we need to
				-- indicate when we don't do work.
		end

		local didWork = false

		for _, docAndUrlPair in ipairs( docsAndUrls ) do	

			if not self._continuousUploadTask:isRunning() then break end

			local docId = docAndUrlPair[ 1 ]
			local localFilePath = docAndUrlPair[ 2 ]
			local binaryUrl = docAndUrlPair[ 3 ]
			local binaryType = docAndUrlPair[ 4 ]
			local contentType = docAndUrlPair[ 5 ]
			local isBlacklisted = intToBool( docAndUrlPair[ 6 ] )
			local sha256 = docAndUrlPair[ 7 ]
			local localFileSize = docAndUrlPair[ 8 ]
			local sha256Error = intToBool( docAndUrlPair[ 9 ] )
			local sourceLinkForCopy = docAndUrlPair[ 10 ]
			local localFileVolumeGuid = docAndUrlPair[ 11 ]
			local localFileRelativePath = docAndUrlPair[ 12 ]
			
			if isBlacklisted then
			
				-- Check if the asset is actually blacklisted.
				isBlacklisted = self._failedUploadBlacklistHelper:maybeResetBlacklistStatus( docId )
			
			end

			local assetIsOffline = false
			
			-- ensure localFilePath is a string and not bad data
			
			localFilePath = type( localFilePath ) == 'string' and localFilePath or nil
			
			if localFileVolumeGuid and localFileRelativePath then
			
				local success, filePath = pcall( 
					WFFileUtils.createFullPathFromRelativePathAndGuid,
					localFileRelativePath, localFileVolumeGuid )
					
				if success and filePath then
					localFilePath = localFilePath or filePath
				else
					assetIsOffline = true
				end
				
			end
			
			local binaryExistsLocally = localFilePath and WFFileUtils.exists( localFilePath )
				-- Though we could check that the original exists in the indexer, it's
				-- more efficient to check for it here since it won't run all items,
				-- but only the items before we've reached maxOriginalUploadsInFlight.
				
			if self._binariesInFlight < maxScheduledUploads and
				not getRecentlyUploadedSha256( self, docId, binaryType ) and
				not isBlacklisted 
				and ( binaryExistsLocally or -- Standard Uploads 
					( localFilePath == nil and not assetIsOffline ) ) -- Copies
				and getUploadingAssetsWithSha256( self, sha256 ) < maxScheduledUploads -- Prevent creating a large blacklist while uploading first asset if many copies exist, but allow parallel copies to occur.
				then
				
				local isFirstUploadForSha256 = getUploadingAssetsWithSha256( self, sha256 ) == 0

				addToRecentlyUploaded( self, docId, binaryType, sha256 or true )

				local subscription = SingleAssignmentDisposable()

				addToActiveUploads( self, docId, binaryType, subscription )

				local function removeUploadSubscription()

					-- Don't call uploadFinished from finally as it could be the result of
					-- dispose being called from disposeActiveUploads and we don't want to
					-- start other uploads in this case.
					removeFromActiveUploads( self, docId, binaryType )

				end

				local function uploadCompletedOrError()

					-- Call removeFromActiveUploads here (in addition to the finally) so that _binariesInFlight
					-- is reduced so the following uploadFinished call can trigger another upload as 
					-- finally is called after the completed not before it.
					removeFromActiveUploads( self, docId, binaryType )

					-- Trigger combineLatest above to start other uploads (in case _binariesInFlight reached maxScheduledUploads).
					uploadFinished( self )

					if self._binariesInFlight == 0 then
						self._idleStatus:onNext( true )
					end

				end

				didWork = true
				
				local action

				if sha256Error then

					local function delayedClearFromRecentlyUploadedFn()
						delayedClearFromRecentlyUploaded( self, "waiting for resolveFailedChunkedUpload doc update", docId, binaryType )
					end

					-- Handle asynchronous error from a prior chunked upload that is sent via catalog metadata.
					action = resolveFailedChunkedUpload( self, binaryUrl, localFilePath, docId, sha256 )
						:finally( delayedClearFromRecentlyUploadedFn )

				elseif sourceLinkForCopy ~= nil then

					-- Asset was copied and the original is already up on oz, determine whether to copy or upload the binary.

					-- Copy master even if we have a local binary as it should be faster than uploading,
					-- otherwise upload the file if we have a local path to avoid having to first
					-- download the source content (oz doesn't provide api's to copy xmp or xmp_develop).

					if isMaster( binaryType ) or localFilePath == nil then

						action = copyOneBinary( self, docId, binaryUrl, binaryType, sha256, contentType, sourceLinkForCopy )
						:tap( function( dataLength, response, headers )

							-- User made a copy of an asset that was already on oz so we know the oz copy master API will
							-- have successfully copied the sidecar xmp as well so we don't need to upload/copy the xmp.

							-- For copies made to assets where the source asset isn't the same as the copied asset
							-- (which means we found another asset with matching sha256 hash to make a copy of)
							-- we don't clear the xmpSideCarNeedsUpload flag as it could potentially be different.

							if binaryType == 'master' then
								self._docStore:applyOverlay {
									_id = docId,
									_localOnly = {
										xmpSidecarNeedsUpload = { _delete = true },
									}
								}
								:call( runUntilDone )
							end

						end )
						:catch( function( responseData, headers )

							if errorMasterOrProxyResourceDoesNotExist( binaryType, responseData, headers ) then

								-- The source asset was deleted, see if there is another asset with the same sha and copy that instead.
								return uploadOrCopyOneMasterOrProxy( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256, isFirstUploadForSha256 )

							end

							handleUploadFailure( self, docId, binaryType, binaryUrl, responseData, headers, false )
							return Observable.throw( responseData, headers )

						end )

					else

						action = uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256 )

					end

				elseif isMaster( binaryType ) or binaryType == 'proxy' then -- proxy addresses: lightroom asset is copied while paused, then the copy is copied (results in no sourceLinkForCopy)

					action = uploadOrCopyOneMasterOrProxy( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256, isFirstUploadForSha256 )

				else

					action = uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, sha256 )

				end

				subscription:setDisposable(
					action
						:finally( removeUploadSubscription )
						:subscribe( nil, uploadCompletedOrError, uploadCompletedOrError )
				)

				if self._binariesInFlight >= maxScheduledUploads then break end

			elseif not binaryExistsLocally then

				if localFilePath then -- localFilePath can be nil doing a copy.

					logger:warnf( "Uh oh. Trying to upload %s for %s but it doesn't exist at path %s.", binaryType, docId, localFilePath )

					if isMaster( binaryType ) and not assetIsOffline then

						applyInvalidOriginalOverlay( self, docId, localFilePath, 'UploadFailedMissingOriginal' )
							:finally( function( appliedOverlay )
							
								if appliedOverlay then
								
									addToRecentlyUploaded( self, docId, binaryType, sha256 or true )
									delayedClearFromRecentlyUploaded( self, "binary missing, marking invalid", docId, binaryType )
								
								end
								
							end )
							:call( runUntilDone )
					end

				end

				if binaryType == 'xmp_sidecar' and not sourceLinkForCopy then
				
					-- Xmp sidecar files are always stored on the main volume. If the
					-- xmp sidecar file doesn't exist, then we need to delete the
					-- file annotation.

					addToRecentlyUploaded( self, docId, binaryType, sha256 or true )

					removeSidecarFileAnnotation( self, docId, binaryType )
						:finally( function()
							delayedClearFromRecentlyUploaded( self, "removing sidecar annotation", docId, binaryType )
						end )
						:call( runUntilDone )

				end

			end

		end
		
		if didWork then
			self._continuousUploadTask:taskDidWork()
		end
		
	end )

	:catch( function( errorMessage )

		logger:debugf( "uploadSomeBinaries %s", errorMessage )

		if self._binariesInFlight > 0 then

			-- Wait until all the in-flight requests have finished.

			return self._idleStatus
				:take( 1 )
				:flatMap( function()

					self._continuousUploadTask:taskDidWork( false )
					return self._continuousUploadTask:throwIfNotStopped( errorMessage )

				end )

		else

			self._continuousUploadTask:taskDidWork( false )
			return self._continuousUploadTask:throwIfNotStopped( errorMessage )

		end

	end )

end

--============================================================================--

local WFCatBinaryUploader = {

	_catalog = false,
		-- (WFCatalog)

	_docStore = false,
		-- (WFRxDocStore)

	_catalogSpec = false,
		-- (WFCatSpec)

	_binariesInFlight = false,
		-- (number) Number of uploads that are in progress.
		
	_recentlyUploaded = false,
		-- (table) set of document IDs for which we've recently uploaded

	_activeUploads = false,
		-- (table) Table of subscriptions of each active upload.

	_uploadingAssetWithSha256 = false,
		-- (table) Table of counts of the number of assets attempted to be uploaded indexed by sha256.

	_continuousUploadTask = false,
		-- (WFRxContinuousTask) A continuous task to handle continuous uploads

	_errorStatus = false,
		-- (Subject) A Rx.Subject used to relay upload error status

	_idleStatus = false,
		-- (Subject) A Rx.Subject used to indicate idle status.

	_uploadFinished = false,
		-- (Rx BehaviorSubject) A local stream that fires when an upload completes.
	
	_volumeStatusHasChanged = false,
		-- (Rx BehaviorSubject) A local stream that fires when a volumes status changes.
	
	_failedUploadBlacklistHelper = false,
		-- (WFCatBlacklistHelper) A helper classed used to temporarily blacklist failed uploads.

	_chunkedUploader = false,
		-- (WFCatChunkedUploader) Used to upload large files in chunks.

	_timerAborter = false,
		-- (ReplaySubject) Used to indicate the uploader is being closed by sending an error, used to clear out blackout timers.

	_uploadErrorLog = false,
		-- (table) record of recent upload errors for the diagnostic log

	_uploadSpeedStats = false,
		-- (table) record of recent upload speeds

	class = {
		generateIndexSchema = generateIndexSchema,
	},

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function WFCatBinaryUploader:__init( catalog, chunkedUploader, optTaskDelays, optCatSpec )

	check( 'WFCatBinaryUploader', 'WFCatalog', '?table', '?table', '?WFCatSpec' )

	self:setOwned {
		_catalog = catalog:relayObject()
	}
	
	self._docStore = catalog:docStore()

	self._catalogSpec = optCatSpec or catalog:ozCatalogSpec()

	self._failedUploadBlacklistHelper = WFCatBlacklistHelper( 
												self._docStore, 
												'WFCatBinaryUploader',
												numberOfTriesBeforeBlacklistingUpload,
												failedUploadBlackoutWindowInSeconds,
												true )
			
	self._chunkedUploader = chunkedUploader or
		WFCatChunkedUploader( catalog:docStore(), self._catalogSpec )

	self._recentlyUploaded = {}

	self._activeUploads = {}

	self._binariesInFlight = 0

	self._uploadingAssetWithSha256 = {}

	local continuousTaskDelays = optTaskDelays or {
		progressiveBackoff = true,
		errorDelay = 10,
		maxDelay = 60,
	}
	
	local continuousTask = function() 
		return uploadSomeBinaries( self ) 
	end

	self._continuousUploadTask = WFRxContinuousTask( "WFCatBinaryUploader", continuousTask, continuousTaskDelays )

	self._errorStatus = Rx.Subject()

	self._idleStatus = Rx.Subject()

	self._uploadFinished = Rx.BehaviorSubject()
	
	self._volumeStatusHasChanged = Rx.BehaviorSubject()
	
	self._timerAborter = Rx.ReplaySubject( 1 )
	self._timerAborter:onNext( true )

	self._uploadErrorLog = {}

	self._uploadSpeedStats = {
		buckets = {},
		totalSize = 0,
	}

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:close()

	check( 'WFCatBinaryUploader' )

	disposeActiveUploads( self )

	local errorStatus = self._errorStatus
	self._errorStatus = nil
	errorStatus:dispose()

	local continuousUploadTask = self._continuousUploadTask
	self._continuousUploadTask = nil
	
	if continuousUploadTask then
		continuousUploadTask:close()
	end

	local failedUploadBlacklistHelper = self._failedUploadBlacklistHelper
	self._failedUploadBlacklistHelper = nil
	
	if failedUploadBlacklistHelper then
		failedUploadBlacklistHelper:close()
	end

end

--------------------------------------------------------------------------------	

function WFCatBinaryUploader:stopContinuousUploads()

	self._continuousUploadTask:stopContinuousTask()

	disposeActiveUploads( self )

	uploadFinished( self )
		-- trigger combineLatest in uploadSomeBinaries so it can see it is stopped.

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:startContinuousUploads()

	self._continuousUploadTask:startContinuousTask()

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:errorStatus()

	return self._errorStatus

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:isDoingWork()

	return self._continuousUploadTask:isTaskDoingWork()

end

--------------------------------------------------------------------------------

--[[!

	\brief Report the status of the Nimbus Volumes
	
	\return (void)
	
--]]

function WFCatBinaryUploader:reportStatusOfVolumes( statusOfVolumes )

	check( 'WFCatBinaryUploader', 'table' )
	
	-- When a volume's status changes, namely when a volume is reconnected
	-- to Nimbus, there's a chance that we'll be able to upload binaries
	-- that were previously not accessible. 
	
	-- The query WFCatBinaryUploader uses does not filter out assets that
	-- are on offline volumes, unlike WFCatIdleDownloader, because we also
	-- support uploading "copies" that have no local file.
	
	-- Instead, we prioritize uploading the files that are known to be connected,
	-- as well as copies, and then check to see if the possibly-disconnected
	-- files can be accessed and uploaded.
	
	-- The following statement guarentees we process the disconnected -> connected
	-- files once a volume comes online. The value here is unimportant, but making
	-- it unique can help in debugging.
	
	self._volumeStatusHasChanged:onNext( KSDate.timeInIso8601Format() )
	
end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:DEBUG_getDiagnosticLogErrors()
	return logger:DEBUG_makeDiagnosticLogErrorsHtmlText()
end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:DEBUG_resetDiagnosticLogErrors()
	return logger:DEBUG_resetDiagnosticLogErrors()
end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:sendDiagnosticLogData()

	local finalStats = {}

	local activeUploads = {}
	for docId, uploadsForDocId in pairs( self._activeUploads ) do
		for binaryType, subscription in pairs( uploadsForDocId ) do
			table.insert( activeUploads, { id = docId, binaryType = binaryType } )
		end
	end

	for bucket, stats in pairs( self._uploadSpeedStats.buckets ) do
		finalStats[ bucket ] = {
			count = stats.count,
			timings = { stats.timings:visualize 'boxAndWhisker' },
			size = { stats.sizes:visualize 'boxAndWhisker' },
			speed = { stats.speed:visualize 'boxAndWhisker' },
		}
	end

	return {

		name = "Binary Uploader",
		id = 'WFCatBinaryUploader',
		weight = 510,

		body = {
			activeUploads = activeUploads,
			binariesInFlight = self._binariesInFlight,
			recentErrors = self._uploadErrorLog,
			uploadHttpRequestCost = uploadHttpRequestCost,
			uploadSpeedStats = finalStats,
			totalBinarySize = self._uploadSpeedStats.totalBinarySize,
			chunkFileAtSize = self._chunkedUploader:chunkFileSize(),
			chunkSize = self._chunkedUploader:chunkSize(),
		},

	}

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatBinaryUploader'( WFCatBinaryUploader )
