--[[----------------------------------------------------------------------------

WFCatChunkedUploader.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.rxlua
local WFRxScheduler = require 'WFRxScheduler'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFConfiguration = require 'WFConfiguration'
local WFTableMerge = require 'WFTableMerge'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- WF.web.http
local WFRxHttpClient = require 'WFRxHttpClient'

	-- WF.web.utils
local WFURI = require 'WFURI'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local KSTableUtils_derefOpt = assert( KSTableUtils.derefOpt )
local KSTableUtils_deepCopy = assert( KSTableUtils.deepCopy )

local kDoubleQuote = '"'

local logger = WFLogger 'WFCatChunkedUploader'

local OZ_CHUNK_FILE_SIZE_RECOMMENDED = ( 5 * 1024 * 1024 )
local OZ_CHUNK_FILE_SIZE_REQUIRED = ( 200 * 1024 * 1024 )
local OZ_MIN_CHUNK_SIZE = ( 5 * 1024 * 1024 )

local kRestartCurrentChunk = 'RestartCurrentChunk'
local kRestartChunkedUpload = 'RestartChunkedUpload'

--------------------------------------------------------------------------------

local function GenerateNamedError( type, msg, ... )
	return 'handledError', {
		errorType = type,
		message = select( '#', ... ) == 0 and msg or string.format( msg, ... ),
	}
end

--------------------------------------------------------------------------------

local function RestartCurrentChunkError( msg, ... )
	return GenerateNamedError( kRestartCurrentChunk, msg, ... )
end

--------------------------------------------------------------------------------

local function RestartChunkedUploadError( msg, ... )
	return GenerateNamedError( kRestartChunkedUpload, msg, ... )
end

--------------------------------------------------------------------------------

local function throwIfHttpStatusIs4xx( data, headers, errFn )

	local statusCode = headers and headers.status_code or ""

	-- Don't consider a 429 a 4xx error as it means retryAfterDelay and we
	-- wouldn't want to start the entire chunked upload over in that case.

	if statusCode:sub( 1, -3 ) == "4" and statusCode ~= '429' then
		return Observable.throw( errFn( "Status Code: %s", statusCode ) )
	end

	return Observable.just( data, headers )

end

--------------------------------------------------------------------------------

local function throwIfHttpStatusNot( expectedStatusCode, data, headers, key, uploadUrl, uploadBody )

	local statusCode = headers and headers.status_code

	if statusCode ~= expectedStatusCode then

		headers = headers or {}
		headers.errorCaptureKey = key

		local logFmtStr =  "Got status %s back from Oz\n\n"
		local logParams = { statusCode }

		if uploadUrl then
			logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = uploadUrl
		end

		if uploadBody then
			logFmtStr = logFmtStr .. "what we uploaded in body was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = tostring(uploadBody)
		end

		logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
		logParams[ #logParams + 1 ] = data

		logger:errorf( logFmtStr, unpack( logParams ) )

		return Observable.throw( data, headers )

	end

	return Observable.just( data, headers )

end

--==============================================================================

local function replaceOrInsertHeader( headers, name, value )

	local found = false
	for index, header in ipairs( headers ) do
		if header.header == name then
			if value ~= nil then
				header.value = value
			else
				table.remove( headers, index )
			end
			found = true
			break
		end
	end
	if not found and value ~= nil then
		headers[ #headers + 1 ] = {
			header = name,
			value = value,
		}
	end

end

--==============================================================================

local WFCatChunkedUploader = {

	_docStore = false,
		-- (WFRxDocStore)

	_catalogSpec = false,
		-- (WFCatSpec)

	_chunkFileSize = OZ_CHUNK_FILE_SIZE_RECOMMENDED,
		-- (number) Files this size and larger should be uploaded via chunked upload.

	_chunkSize = OZ_CHUNK_FILE_SIZE_RECOMMENDED,
		-- (number) Size of chunks used during chunked upload.

	_chunkSubscriptionsByDocId = false,
		-- (table) Table of in-flight chunk requests.
		-- key: docId
		-- value: (table) Table of in-flight requests for the doc Id.

	_activeOverlays = false,
		-- (table) Table of the chunkedUpload portion of
		-- documents that are in flight.
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function WFCatChunkedUploader:__init( docStore, catalogSpec, chunkFileSize, chunkSize )

	check( 'WFCatChunkedUploader', 'WFRxDocStore', 'WFCatSpec', '?number', '?number' )

	self._docStore = docStore

	self._catalogSpec = catalogSpec

	self._chunkSubscriptionsByDocId = {}

	chunkFileSize = chunkFileSize or WFConfiguration.ChunkedUploader.chunkFileSize

	if chunkFileSize == nil
			or chunkFileSize >= OZ_CHUNK_FILE_SIZE_REQUIRED
			or chunkFileSize < OZ_MIN_CHUNK_SIZE then
		chunkFileSize = OZ_CHUNK_FILE_SIZE_RECOMMENDED
	end

	self._chunkFileSize = chunkFileSize

	chunkSize = chunkSize or WFConfiguration.ChunkedUploader.chunkSize

	if chunkSize == nil or chunkSize >= OZ_CHUNK_FILE_SIZE_REQUIRED then
		chunkSize = OZ_CHUNK_FILE_SIZE_RECOMMENDED
	elseif chunkSize < OZ_MIN_CHUNK_SIZE then
		chunkSize = OZ_MIN_CHUNK_SIZE
	end

	self._chunkSize = chunkSize

	self._activeOverlays = {}

end

--------------------------------------------------------------------------------

local function chunkedUploader_activeOverlay( self, docId )

	return self._activeOverlays[ docId ]

end

--------------------------------------------------------------------------------

local function chunkedUploader_docValue( self, docId )

	return self._docStore:docValue( docId )
		:map( function( docValue )

			local activeOverlay = chunkedUploader_activeOverlay( self, docId )

			if activeOverlay then

				if docValue._localOnly then
					docValue._localOnly.chunkedUpload = nil
				end

				WFTableMerge.applyOverlay( docValue, activeOverlay )

			else

				-- Grab persistent chunkedUpload if it exists, it will be replaced entirely with our in memory version.

				local chunkedUpload = KSTableUtils_derefOpt( docValue, '_localOnly', 'chunkedUpload' )

				if chunkedUpload then
					self._activeOverlays[ docId ] = {
						_id = docId,
						_localOnly = {
							chunkedUpload = chunkedUpload,
							docHadPersistentInfo = true, -- This field only exists in memory
						},
					}
				end

			end

			return docValue

		end )

end

--------------------------------------------------------------------------------

local function chunkedUploader_makeOverlayPersistentIfNeeded( self, docId, activeOverlay )

	self._activeOverlays[ docId ] = nil

	local docHasNewPersistentInfo = KSTableUtils_derefOpt( activeOverlay, '_localOnly', 'chunkedUpload' )
	local docHadPersistentInfo = KSTableUtils_derefOpt( activeOverlay, '_localOnly', 'docHadPersistentInfo' )

	if docHadPersistentInfo then
		-- This field doesn't exist in persistent storage remove it.
		activeOverlay._localOnly.docHadPersistentInfo = nil
	end

	if docHadPersistentInfo or docHasNewPersistentInfo then

		if docHasNewPersistentInfo then

			logger:infof( "Persisting chunkedUpload info to docId = %s", docId )

			activeOverlay._localOnly.chunkedUpload._replace = true

		else -- docHadPersistentInfo

			logger:infof( "Removing persistent chunkedUpload info from docId = %s", docId )

			activeOverlay = {
				_id = docId,
				_localOnly = {
					chunkedUpload = {
						_delete = true,
					},
				},
			}

		end

		return self._docStore:applyOverlay( activeOverlay )

	end

	return Observable.just()

end

--------------------------------------------------------------------------------

local function chunkedUploader_applyOverlay( self, docStatusOverlay, makePersistentIfNeeded )

	local docId = docStatusOverlay._id
	local activeOverlay = chunkedUploader_activeOverlay( self, docId ) or {}

	WFTableMerge.applyOverlay( activeOverlay, docStatusOverlay )

	self._activeOverlays[ docId ] = activeOverlay
	
	if makePersistentIfNeeded then
		return chunkedUploader_makeOverlayPersistentIfNeeded( self, docId, activeOverlay )
	end

	return Observable.just()

end

--------------------------------------------------------------------------------

function WFCatChunkedUploader:chunkSize()
	return self._chunkSize
end

--------------------------------------------------------------------------------

function WFCatChunkedUploader:setChunkSize( chunkSize )

	check( 'WFCatChunkedUploader', 'number' )

	chunkSize = chunkSize or WFConfiguration.ChunkedUploader.chunkSize

	if chunkSize >= OZ_CHUNK_FILE_SIZE_REQUIRED then
		chunkSize = OZ_CHUNK_FILE_SIZE_RECOMMENDED
	elseif chunkSize < OZ_MIN_CHUNK_SIZE then
		chunkSize = OZ_MIN_CHUNK_SIZE
	end

	self._chunkSize = chunkSize

end

--------------------------------------------------------------------------------

function WFCatChunkedUploader:chunkFileSize()
	return self._chunkFileSize
end

--------------------------------------------------------------------------------

function WFCatChunkedUploader:minimumChunkSize()
	return OZ_MIN_CHUNK_SIZE
end

--------------------------------------------------------------------------------

local function put( self, ... )

	return self._catalogSpec:putBinary( ... )

-- TODO [kcarlson 2016-03-21]: add error logging here?

--	:tapOnError( function( responseData, headers )
--		self:maybeAddToErrorLog( headers, responseData )
--	end )

end

--------------------------------------------------------------------------------

local function initiateUpload( self, docValue, masterUrl, putArgs )

	-- initiate multi-part upload
	local initiateUrl = masterUrl .. '/parts'

	replaceOrInsertHeader( putArgs.requestHeaders, 'X-Generate-Renditions', nil )
	replaceOrInsertHeader( putArgs.requestHeaders, 'X-Generate-Proxies', nil )

	putArgs.requestCost = WFRxHttpClient.requestCostLow
	putArgs.requestDescription = string.format( '%s : Initiate Chunked Upload', docValue._id )
	
	return put( self, initiateUrl, '', putArgs )
	:flatMap( function( results, headers )
		return throwIfHttpStatusNot( '201', results, headers, 'initiate master multipart', initiateUrl )
	end )
	:map( function( results, headers )
		return KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( results ), headers
	end )
	:map( function( results, headers )

		local docStatusOverlay = {
			_localOnly = {
				chunkedUpload = {
					parts = {},
					part_template = KSTableUtils_derefOpt( results, 'links', '/rels/create_part', 'href' ),
					finish_href = KSTableUtils_derefOpt( results, 'links', 'self', 'href' ),
				},
			},
		}

		return false, 'applyOverlay', docStatusOverlay

	end )

end

--------------------------------------------------------------------------------

local function doChunk( self, docValue, masterUrl, putArgs, filePath, startOffset, length, partNumber )

	local part_template = KSTableUtils_derefOpt( docValue, '_localOnly', 'chunkedUpload', 'part_template' )
	part_template = string.replace( part_template, '{part_number}', tostring( partNumber ) )

	part_template = WFURI.resolve( masterUrl, part_template )
	putArgs.dataFormat = 'filePath'
	putArgs.offset = startOffset
	putArgs.length = length
	putArgs.requestDescription = string.format( '%s : Uploading Chunk %d', docValue._id, partNumber )
	
	replaceOrInsertHeader( putArgs.requestHeaders, 'X-Generate-Renditions', nil )
	replaceOrInsertHeader( putArgs.requestHeaders, 'X-Generate-Proxies', nil )
	replaceOrInsertHeader( putArgs.requestHeaders, 'Content-Type', nil )

	logger:infof( "Uploading chunk %d (offset %d chunkSize %d) of %s", partNumber, startOffset, length, docValue._id )
	
	return put( self, part_template, filePath, putArgs )
	:flatMap( function( data, headers, ... )
		return throwIfHttpStatusIs4xx( data, headers, RestartChunkedUploadError )
	end )
	:flatMap( function( data, headers )
		local errorKey = 'master part ' .. tostring( partNumber )
		return throwIfHttpStatusNot( '201', data, headers, errorKey, part_template, filePath )
	end )
	:flatMap( function( results, headers )

		local headerETag = headers.ETag or headers.Etag
			-- Mac has lower case t

		if not headerETag then
			return Observable.throw( RestartCurrentChunkError( "Upload did not include ETag" ) )
		end
		if headerETag:sub( 1, 1 ) ~= kDoubleQuote then
			return Observable.throw( RestartCurrentChunkError( "ETag <<%s>> did not start with double quote", headerETag ) )
		end
		if headerETag:sub( -1, -1 ) ~= kDoubleQuote then
			return Observable.throw( RestartCurrentChunkError( "ETag <<%s>> did not end with double quote", headerETag ) )
		end

		local etag = string.sub( headerETag, 2, -2 )

		local docStatusOverlay = {
			_id = docValue._id,
			_localOnly = {
				chunkedUpload = {
					parts = {
						[ partNumber ] = {
							nextPartStart = startOffset + length,
							ETag = etag,
						},
					},
				},
			},
		}

		logger:infof( "Uploaded chunk %d (offset %d chunkSize %d) of %s", partNumber, startOffset, length, docValue._id )
		return chunkedUploader_applyOverlay( self, docStatusOverlay, false )

	end )

end

--------------------------------------------------------------------------------

local function finishUpload( self, docValue, masterUrl, putArgs )

	local parts = KSTableUtils_derefOpt( docValue, '_localOnly', 'chunkedUpload', 'parts' )
	local partData = {}
	for partNumber, info in ipairs( parts ) do
		partData[ partNumber ] = {
			part_number = partNumber,
			etag = info.ETag,
		}
	end

	local jsonData = KSJsonEncoder.encodeOzUploadJson( { parts = partData } )
	putArgs.requestMethod = 'POST'

	putArgs.requestPriority = WFRxHttpClient.requestPriorityHigh
		-- Because all parts have been uploaded, sending this final request with
		-- a high priority allows it to jump ahead of other chunked upload
		-- requests on the WFRxRateLimiter queue.
		
	putArgs.requestCost = WFRxHttpClient.requestCostLow
		-- Very little data is being uploaded, so the request cost here is low.
		
	putArgs.requestDescription = string.format( '%s : Finished Chunk Upload', docValue._id )
		
	local finishUrl = KSTableUtils_derefOpt( docValue, '_localOnly', 'chunkedUpload', 'finish_href' )
	finishUrl = WFURI.resolve( masterUrl, finishUrl )
	replaceOrInsertHeader( putArgs.requestHeaders, 'Content-Type', 'application/json' )

	return put( self, finishUrl, jsonData, putArgs )
	:flatMap( function( results, headers )
		return throwIfHttpStatusIs4xx( results, headers, RestartChunkedUploadError )
	end )
	:flatMap( function( results, headers )
		return throwIfHttpStatusNot( '202', results, headers, 'finish multipart', finishUrl )
	end )
	:map( function( results, headers )

		logger:infof( 'Finished chunked upload.' )

		-- Assembly will happen async on oz, assume success as we
		-- don't know how long to wait before asking if it exists.

		-- Don't set _serverProvidedMetadata.links[ '/rels/master' ] to headers.Location here
		-- as we want to get it with the doc update along with the importSource.status
		-- in case oz finds out the sha does not match.

		local docStatusOverlay = {
			_localOnly = {
				chunkedUpload = {
					_delete = true,
				},
			},
		}

		return true, 'applyOverlay', docStatusOverlay

	end )

end

--------------------------------------------------------------------------------

local function restartChunked( self, docValue, masterUrl, putArgs )

	local finishUrl = KSTableUtils_derefOpt( docValue, '_localOnly', 'chunkedUpload', 'finish_href' )
	finishUrl = WFURI.resolve( masterUrl, finishUrl )

	putArgs.requestMethod = 'DELETE'
	putArgs.requestHeaders = {}

	return put( self, finishUrl, '', putArgs )
	:flatMap( function( results, headers )

		local statusCode = headers and headers.status_code or ""

		local is2x = string.find( statusCode, '^2' )
		local is4x = string.find( statusCode, '^4' ) and statusCode ~= '429'

		if is2x or is4x then

			local docStatusOverlay = {
				_localOnly = {
					chunkedUpload = {
						_delete = true,
					},
				},
			}

			return Observable.just( false, 'applyOverlay', docStatusOverlay )

		else

			-- Most likely a 503 service unavailable error.

			-- Abort out and let the asset get black listed before trying again.
			headers.errorCaptureKey = "Abort multipart upload failed"
			return Observable.throw( results, headers )

		end

	end )

end

--------------------------------------------------------------------------------

local function clearRestartAndFinish()
	return Observable.just( false, 'applyOverlay', {
		_localOnly = {
			chunkedUpload = {
				_delete = true,
			},
		},
	} )
end

--------------------------------------------------------------------------------

local function doUploadImp( self, docValue, masterUrl, putArgs, filePath )

	local docFileSize = WFFileUtils.fileAttributes( filePath ).fileSize

	if docFileSize == nil or docFileSize == 0 then
		return Observable.throw( 'payload must exist', {} )
	end

	local docId = docValue._id

	putArgs.returnHttpErrorsAsData = true
		-- Handle the errors as values to be able to restart chunks
		-- or restart the entire chunked upload.

	putArgs.dataFormat = nil
	putArgs.offset = nil
	putArgs.length = nil

	local chunkedUpload = KSTableUtils_derefOpt( docValue, '_localOnly', 'chunkedUpload' )
	local part_template = KSTableUtils_derefOpt( chunkedUpload, 'part_template' )
	local finish_href = KSTableUtils_derefOpt( chunkedUpload, 'finish_href' )
	local parts = KSTableUtils_derefOpt( chunkedUpload, 'parts' )
	local restart = KSTableUtils_derefOpt( chunkedUpload, 'restart' )

	if restart then
		if finish_href then
			return restartChunked( self, docValue, masterUrl, putArgs )
		else
			return clearRestartAndFinish()
		end
	end

	if not part_template then
		return initiateUpload( self, docValue, masterUrl, putArgs )
	end

	local startOffset = 1
	local length = self._chunkSize

	local numberOfChunks = math.ceil( docFileSize / self._chunkSize )

	local chunksHaveFinishedUploadingStream = Rx.ReplaySubject( 1 )
	local completedChunks = 0
	local haveChunksToUpload = false
	
	local encounteredError

	local function chunkCompleted()

		completedChunks = completedChunks + 1

		if completedChunks == numberOfChunks then

			if encounteredError then
				chunksHaveFinishedUploadingStream:onError( unpack( encounteredError ) )
			else
				chunksHaveFinishedUploadingStream:onNext( false )
				chunksHaveFinishedUploadingStream:onCompleted()
			end

		end

	end

	local chunkSubscriptions = {}
	
	for i = 1, numberOfChunks, 1 do

		if length + startOffset > docFileSize then
			length = 1 + docFileSize - startOffset
		end

		if KSTableUtils_derefOpt( parts, i, 'nextPartStart' ) == startOffset + length then
			
			-- The chunk has already been uploaded with the correct chunkSize.
			chunkCompleted()
			
		else
	
			haveChunksToUpload = true
	
			local putArgsCopy = KSTableUtils_deepCopy( putArgs )
		
			chunkSubscriptions[ i ] = doChunk( self, docValue, masterUrl, putArgsCopy, filePath, startOffset, length, i )
	
				:catch( function( ... )

					-- If an error occurs while uploading a chunk, allow the other
					-- other chunks to complete before bubbling up the error which
					-- may force a retry of the lone chunk or of the entire asset.

					encounteredError = arg
					return Observable.just()
					
				end )
	
				:finally( function()
			
					logger:infof( '%s CHUNK %d COMPLETE', docId, i )
			
					local subscription = chunkSubscriptions[ i ]
					chunkSubscriptions[ i ] = nil
					
					if subscription then
						subscription:dispose()
					end
					
					chunkCompleted()
					
				end )
				
				:subscribe()
	
			chunkSubscriptions[ i ]:disposeWith( chunksHaveFinishedUploadingStream )

		end

		startOffset = startOffset + length

	end

	self._chunkSubscriptionsByDocId[ docId ] = chunkSubscriptions

	if haveChunksToUpload then
		return chunksHaveFinishedUploadingStream
	end
	
	if finish_href then
		return finishUpload( self, docValue, masterUrl, putArgs )
	end
	 
	return Observable.just( false, kRestartChunkedUpload )
		--Uh oh.  We thought we were finished uploading chunks, but don't have
		-- the complete url. Bail.
		
end

--------------------------------------------------------------------------------

local function uploadAndUpdateDoc( self, masterUrl, filePath, putArgs, docId )

	return chunkedUploader_docValue( self, docId )

	:flatMap( function( docValue )

		if docValue._empty or docValue.type ~= 'asset' then
			return Observable.throw( "asset has been deleted" )
		end

		return doUploadImp( self, docValue, masterUrl, putArgs, filePath )

	end )

	:catch( function( maybeHandledError, maybeErrorInfo, ... )

		if maybeHandledError == 'handledError' then

			logger:warnf( "Chunked upload encountered error: %s", maybeErrorInfo.message or "NO message" )
			return Observable.just( false, maybeErrorInfo.errorType )

		else

			return Observable.throw( maybeHandledError, maybeErrorInfo, ... )

		end

	end )

	:flatMap( function( isLastChunk, action, docStatusOverlay )

		local result

		if action == kRestartChunkedUpload then

			docStatusOverlay = {
				_localOnly = {
					chunkedUpload = {
						restart = true,
						parts = { _delete = true },
						part_template = { _delete = true },
						-- don't delete finish_href as restartChunked uses it
					}
				}
			}

			action = 'applyOverlay'

		end

		if action == 'applyOverlay' then

			docStatusOverlay._id = docId

			result = chunkedUploader_applyOverlay( self, docStatusOverlay, isLastChunk )
				:map( function()
					return isLastChunk
				end )

		else

			result = Observable.just( isLastChunk )

		end

		return result

	end )

end

--------------------------------------------------------------------------------

--[[

	doChunkedUpload recursively calls uploadAndUpdateDoc as part of the chunked
	upload process. For the happy path, uploadAndUpdateDoc is called 3 times:
	
	1. The first time uploadAndUpdateDoc is called, we initiate the multi-part upload with Oz.
	2. The second time uploadAndUpdateDoc is called, ALL chunks are scheduled for upload.
	3. The third time uploadAndUpdateDoc is called, we tell Oz we're done with the multi-part upload.
	
	If a handled error occurs in any of the above steps, we'll call uploadAndUpdateDoc an additional
	time to process that error and retry parts of, or the whole, chunked upload process.
	
]]--

function WFCatChunkedUploader:doChunkedUpload( masterUrl, filePath, putArgs, docId )

	local scheduler = WFRxScheduler.sharedScheduler()

	local isDisposed = false
	
    return Observable.create( function( observer )

        return scheduler:scheduleRecursive( 0, function( idx, recurseFn )

			local putArgsCopy = KSTableUtils_deepCopy( putArgs )

			if isDisposed then
				-- SPAM [jandrew 2017-05-23]:  There's a slight race condition that
				-- can occur when using scheduleRecursive where the parent disposable
				-- is disposed right after the recurseFn is called. This blocks
				-- against that.
				return
			end
			
			-- return the subscription to the scheduled item callback, it will update 
			-- the disposable associated with the scheduledItem allowing us to dispose
			-- the uploadChunkAndUpdateDoc subscription when doChunkedUpload's subscription
			-- is disposed of.
			
			-- FIXME [jandrew 2017-05-23]: -> KCarlson: ^^^ This doesn't appear to work properly.
			-- When doChunkedUpload's subscription is disposed of, the uploadChunkAndUpdateDoc
			-- subscription is _not_ immediately disposed. 
			-- This appears to be an Rx Issue, as I've been able to reproduce the issue with a 
			-- simple Rx-only unit test.
			
			return uploadAndUpdateDoc( self, masterUrl, filePath, putArgsCopy, docId )

				:catch( function( ... )

					local caughtError = { n = select( '#', ... ), ... }

					local activeOverlay = chunkedUploader_activeOverlay( self, docId )
					if activeOverlay then

						-- Persist any chunkedUpload info when errors occur that we can't handle immediately.

						return chunkedUploader_makeOverlayPersistentIfNeeded( self, docId, activeOverlay )
						:flatMap( function()
							return Observable.throw( unpack( caughtError, 1, caughtError.n ) )
						end )

					else

						return Observable.throw( unpack( caughtError, 1, caughtError.n ) )

					end

				end )
				:subscribe( function( isLastChunk )

					if isLastChunk then

						observer:onNext( 'chunkedupload done' )
							-- send onNext for WFPerformanceReport.wrapObservable

						observer:onCompleted()

					else
						recurseFn( idx + 1 )
					end

				end,
				
				function( ... )

					observer:onError( ... )

				end )

		end )

	end )
	
	:finally( function()
	
		isDisposed = true
		
		-- SPAM [jandrew 2017-05-23]: Because the current active item in the scheduleRecursive
		-- block above is not being properly disposed when doChunkedUpload's subscription
		-- is disposed of, we're now manually tracking and disposing of the child requests.
		-- This makes it so pausing sync causes all in-flight chunked uploads to immediately stop.

		local activeOverlay = chunkedUploader_activeOverlay( self, docId )
		if activeOverlay then

			-- Persist chunkedUploadInfo when disposed to the docstore, this happens
			-- when app is quit, or sync is paused.
			chunkedUploader_makeOverlayPersistentIfNeeded( self, docId, activeOverlay )
				:call( WFRxRunUntilDone )

		end

		local chunkSubscriptions = self._chunkSubscriptionsByDocId[ docId ] or {} 
		
		for _, subscription in pairs( chunkSubscriptions ) do
			subscription:dispose()
		end
		
	end )

end

--==============================================================================

return Rx.Class 'WFCatChunkedUploader'( WFCatChunkedUploader )
