--[[----------------------------------------------------------------------------

WFCatCollaboration.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2017 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
local Rx = require 'Rx'
local Observable = Rx.Observable

local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFTableMerge = require 'WFTableMerge'

local WFCatUtils = require 'WFCatUtils'

--------------------------------------------------------------------------------

local check = WFArguments.check
local derefOpt = KSTableUtils.derefOpt

--==============================================================================

-- @mark WFCatCollaboration namespace

--------------------------------------------------------------------------------

local WFCatCollaboration = {}

--------------------------------------------------------------------------------

local function intToBool( zeroOrOne )
	return zeroOrOne ~= 0
end

--------------------------------------------------------------------------------

function WFCatCollaboration.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	-- table space

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE space( created, updateTimeStamp, private, download )]],
		docValueFilter = function( docValue )
			return docValue.subtype == 'shared' and docValue.type == 'space'
		end,
		hasOneRowPerDocument = true,
		
		map = function( docValue, emit )

			if docValue and not docValue._deleted
			  and docValue.type == 'space' and docValue.subtype == 'shared'
			then

				local localOnly = docValue._localOnly or {}
				
				local isPrivate = docValue.private
				local canDownload = docValue.download
				local updateTimeStamp = docValue.updated
				local created = docValue.created or localOnly.created or
								derefOpt( docValue, '_serverProvidedMetadata', 'created' )
								
				emit( created, updateTimeStamp, isPrivate and 1 or 0, canDownload and 1 or 0 )
					
			end
								
		end,

	}

	-- table space_album

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE space_album( spaceId, albumId, updateTimeStamp )]],
		docValueFilter = function( docValue )
			return docValue.type == 'space_album'
		end,
		map = function( docValue, emit )

			if docValue and not docValue._deleted
			  and docValue.type == 'space_album'
			then

				local albumId = docValue.album and docValue.album.id
				local spaceId = docValue.space and docValue.space.id
				local updateTimeStamp = docValue.updated
				
				if spaceId and albumId then
					
					emit( spaceId, albumId, updateTimeStamp )
					
				end
					
			end
						
		end,

	}

	indexStoreSchema:addIndex [[CREATE INDEX space_album_index ON space_album( spaceId, albumId )]]
	indexStoreSchema:addIndex [[CREATE INDEX space_album_album ON space_album( albumId )]]
	
end

--------------------------------------------------------------------------------

function WFCatCollaboration.getAllSharedAlbums( catalog ) 
	
	return catalog:indexStore():query {
			sql = [[ SELECT albumId, spaceId, private from space_album JOIN space ON space_album.spaceId == space.docId ]]
		}
		:map( function( sharedAlbums )
			local results = {}
			for _, sharedAlbum in ipairs( sharedAlbums ) do 
				local result = {
					albumId = sharedAlbum[ 1 ],
					spaceId = sharedAlbum[ 2 ],
					isPrivate = intToBool( sharedAlbum[ 3 ] ),
				}
				table.insert( results, result )
			end
			return results
		end )
end

--------------------------------------------------------------------------------

local function getSyncTerms( docType )

	if docType == 'asset' then
		return {
			fetchSql = [[SELECT docId, spaceId FROM space_asset WHERE assetId == ?]],
			isShared = [[SELECT
							space_asset.docId, spaceId FROM space_asset
						INNER JOIN
							space
						ON
							space_asset.spaceId == space.docId
						WHERE
							assetId == ? AND space.private == 0]],
			docId = 'assetId',
			space_doc_id = 'space_asset_id',
			space_doc = 'space_asset',
			
		}
	elseif docType == 'album' then
		return {
			fetchSql = [[SELECT docId, spaceId FROM space_album WHERE albumId == ?]],
			isShared = [[SELECT
							space_album.docId, spaceId FROM space_album
						INNER JOIN
							space
						ON
							space_album.spaceId == space.docId
						WHERE
							albumId == ? AND space.private == 0]],
			docId = 'albumId',
			space_doc_id = 'space_album_id',
			space_doc = 'space_album',
		}
	else
		error "docType must be 'asset' or 'album'";
	end
end

--------------------------------------------------------------------------------

local function setSpaceAttributes( space, optAttributes )
	local attr = optAttributes or {}
	space.download = attr.download or false
	space.metadata = attr.metadata or false
	space.location = attr.location or false
	space.private = attr.private or false
	return space
end

--------------------------------------------------------------------------------

local function createSpaceDocument( spaceSubtype, optAttributes )

	local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( KSIdGenerator.uuidInLowercaseHex() )

	local spaceDocument = {
		type = 'space',
		subtype = spaceSubtype,
		private = false,
	}
	
	WFTableMerge.applyOverlay( spaceDocument, baseDocValue )
	
	return setSpaceAttributes ( spaceDocument, optAttributes )
end

--------------------------------------------------------------------------------

local function createSpaceAssetOrAlbumDocument( spaceId, spaceDocType, docType, fullDocId ) 
	
	local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( KSIdGenerator.uuidInLowercaseHex() )
	
	local spaceAssetOrAlbumDoc = {
		type = spaceDocType,
		space = {
			id = spaceId,
		},
		[ docType ] = {
			id = fullDocId,
		}
	}
	
	WFTableMerge.applyOverlay( spaceAssetOrAlbumDoc, baseDocValue )
	
	return spaceAssetOrAlbumDoc

end

--------------------------------------------------------------------------------

local function createShareWithOz( catalog, fullDocId, optAttributes )
	-- if we don't already have a URL, request it directly from oz
	local KSJsonEncoder = require "KSJsonEncoder"
	local KSJsonDecoder = require "KSJsonDecoder"
	local spaceId = KSIdGenerator.uuidInLowercaseHex()
	local spaceAlbumId = KSIdGenerator.uuidInLowercaseHex()
	local payload = setSpaceAttributes( { private = false } , optAttributes )
	local data = KSJsonEncoder.encodeOzUploadJson(
		{
			payload= payload,
			subtype = "shared",
			resources = {
				{
					id = fullDocId,
					type = "album",
					space_album_id = spaceAlbumId,
				}
			}
		}
	)
	
	--  create space locally when online call fails
	local function createSpaceLocally()
			-- if we fail at calling oz, fall back to creating locally and syncing when we get back online ( URL will not update )
		local spaceDocument = createSpaceDocument( "shared", optAttributes )

		local spaceAssetOrAlbumDocument = createSpaceAssetOrAlbumDocument( spaceDocument._id, 'space_album', "album", fullDocId )

		return catalog:docStore():bulkSetDocValue {
			spaceDocument,
			spaceAssetOrAlbumDocument,
		}
		:flatMap( function()
			return  Observable.just ( setSpaceAttributes( { id = spaceDocument._id }, spaceDocument ) )
		end )
	end

	local catalogSpec = catalog:ozCatalogSpec()
	if not catalogSpec then  -- if test environment doesn't support ozCatalogSpec
		return createSpaceLocally()
	end

	local requestUrl = string.format( '%s/spaces/%s', catalogSpec:catalogHref(), spaceId )
	return catalogSpec:put( requestUrl, data,
			{ 	requestMethod = 'PUT',
				requestHeaders = {
					{ header = 'Content-Type', value = 'application/json' },
				},
			} )
	:flatMap( function( data, headers )

		if headers and headers.status_code == "201" then

			return catalogSpec:get( requestUrl, { } )
			:flatMap( function( data, headers )

				if headers and headers.status_code == "200" then
					local decodedSpaceJson = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( data )
					local spaceDocument = require 'WFOzDocUtils'.remapBasicOzDocument( decodedSpaceJson, decodedSpaceJson.base )
					
					local href = derefOpt( spaceDocument, '_serverProvidedMetadata', 'links', '/rels/web_space', 'href'  )

					local spaceAlbumDocument = {
						type = "space_album",
						_id = spaceAlbumId,
						space = {
							id = spaceId,
						},
						album = {
							id = fullDocId,
						}
					}
					return catalog:docStore():bulkSetDocValue {
						label = 'oz',
						spaceDocument,
						spaceAlbumDocument,
					}
					:flatMap( function()
						return Observable.just ( setSpaceAttributes( { id = spaceId, url = href }, spaceDocument ) )
					end )
				else
					return createSpaceLocally()
				end
			end)
		else
			return createSpaceLocally()
		end
	end)
end

--------------------------------------------------------------------------------

-- keeping this around because it is useful for testing and debugging
--local function getShareFromOz( catalog, spaceId )
--	-- if we don't already have a URL, request it directly from oz
--	local KSJsonDecoder = require "KSJsonDecoder" 
--	local catalogSpec = catalog:ozCatalogSpec()
--	local requestUrl = string.format( '%sspaces/%s', catalogSpec:baseHref(), spaceId )
--	return catalogSpec:get( requestUrl, { } )
--	:map( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty )
--	:map( function( docValue )
--		local href = derefOpt( docValue, 'links', '/rels/web_space', 'href'  )
--		return setSpaceAttributes ( { id = spaceId, url = href }, docValue )
--	end )
--end

--------------------------------------------------------------------------------

local function shareAssetOrAlbum( catalog, docType, fullDocId, spaceSubtype, optAttributes )
	
	-- The current sharing model for albums calls for a 1-1 mapping from an 'album' document to
	-- a 'space'. In order to share the document, a 'space' needs to be created and the document
	-- added to it.

	-- This function creates a space and adds the provided album to it by creating a
	-- 'space_album' document as appropriate. If a space already exists the existing space is used
	-- if optional attributes are given, they will be used in the new space or updated in the existing one
	-- This returns the attributes of the space
	
	local syncTerms = getSyncTerms( docType )
		
	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			quiescUntilUpToDate = true,
			fullDocId
		}
		:take( 1 )
		:flatMap( function( idRows )

			-- If there is a space for this album, get the ID for that space.
			local spaceId = derefOpt( idRows, 1, 2 )

			if spaceId then

				local overlay = WFCatUtils.createBaseOverlayWithUserCreatedFields( spaceId )
				overlay.private = false
				setSpaceAttributes ( overlay, optAttributes )
				return catalog:docStore():applyOverlay( overlay )
				:flatMap( function()
					-- now read the updated space and return it
					return catalog:docStore():docValue( spaceId )
					:flatMap( function (docValue )
						local href = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/web_space', 'href'  )
						return Observable.just ( setSpaceAttributes ( { id = spaceId, url = href }, docValue ) )
					end )
				end )
			else  -- no space exists, so create one
				return createShareWithOz( catalog, fullDocId, optAttributes )
			end
		end )
end

--------------------------------------------------------------------------------

local function unshareAssetOrAlbum( catalog, docType, fullDocId )

	local syncTerms = getSyncTerms( docType )
		
	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			quiescUntilUpToDate = true,
			fullDocId,
		}
		:take( 1 )
		:flatMap( function( idRows )
		
			local spaceId = derefOpt( idRows, 1, 2 )
			
			if not spaceId then
				return 'error', docType .. " is not shared"
			end
			
			local overlay = WFCatUtils.createBaseOverlayWithUserCreatedFields( spaceId )
			overlay.private = true
				
			return catalog:docStore():applyOverlay( overlay )
				:map( function()
					return true 
				end )
					
		end )
end

--------------------------------------------------------------------------------

local function deleteAlbumOrAssetShare( catalog, docType, fullDocId )

	local syncTerms = getSyncTerms( docType )
	
	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			quiesceUntilUpToDate = true,
			fullDocId,
		}
		:take( 1 )
		:flatMap( function( idRows )
			
			-- SPAM [matthewj 2017-06-02]: Currently the sharing model is a
			-- 1-1 asset/album to space relationship (not including adhoc shares
			-- which we currently ignore). Because of this when an album is 
			-- deleted, we know the space can be. As can all (only 1)
			-- space_album/space_asset records. If this changes or additional
			-- space types are included in these shares this deletion needs
			-- to be rethought.
			local spaceResourceId = derefOpt( idRows, 1, 1 )
			local spaceId = derefOpt( idRows, 1, 2 )
			
			local deleteSpaceResource
			if spaceResourceId then
				deleteSpaceResource = catalog:docStore():bulkApplyOverlay( 
				{
					label = 'oz', -- don't send space resource deletes to Oz
					{ 
						_id = spaceResourceId, 
						_deleted = true,
					}
				} )
			else
				deleteSpaceResource = Observable.empty()
			end
			
			local deleteSpace
			if spaceId then			
				deleteSpace = catalog:docStore():deleteDoc( spaceId )
			else
				deleteSpace = Observable.empty()
			end

			return Observable.combineLatest( deleteSpace, deleteSpaceResource )
			
		end )
		:takeLast( 0 ) 

end

--------------------------------------------------------------------------------

local function sharedAssetOrAlbumUrl( catalog, docType, fullDocId )

	local syncTerms = getSyncTerms( docType )
	
	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			fullDocId
		}
		:pluck( 1, 2 )
		:flatMapLatest( function( spaceId )
			
			if spaceId then
				return catalog:docStore():ongoingDocValue( spaceId )
					:pluck( '_serverProvidedMetadata', 'links', 
							'/rels/web_space', 'href' )
			end
			
			return Observable.just( nil )
		
		end )
		:distinctUntilChanged()
end

--------------------------------------------------------------------------------

local function getShareAttributes( catalog, docType, fullDocId )

	local syncTerms = getSyncTerms( docType )

	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			fullDocId,
		}
		:pluck( 1, 2 ) -- Extract spaceId
		:filter( function( spaceId )
			return spaceId -- Only continue the Rx Chain if spaceId is not null.
		end )
		:take( 1 ) -- this take( 1 ) effectively closes down the index store query.
		:flatMap( function( spaceId )
			return catalog:docStore():ongoingDocValue( spaceId )
			:flatMap( function (docValue )
				local href = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/web_space', 'href'  )
--				if href == nil then
--					return getShareFromOz( catalog, spaceId )
--				end
				return Observable.just ( setSpaceAttributes ( { id = spaceId, url = href }, docValue ) )
			end )
		end )
		:distinctUntilChanged()
end

--------------------------------------------------------------------------------

local function setShareAttributes( catalog, docType, fullDocId, attributes )

	local syncTerms = getSyncTerms( docType )

	return catalog:indexStore():query {
			sql = syncTerms.fetchSql,
			quiescUntilUpToDate = true,
			fullDocId,
		}
		:take( 1 )
		:flatMap( function( idRows )
			local spaceId = derefOpt( idRows, 1, 2 )
			if not spaceId then
				return Observable.throw( docType .. " is not shared" )
			end
			return catalog:docStore():applyOverlay( setSpaceAttributes ( { _id = spaceId, }, attributes ) )
				:map( function()
					return true
				end )
		end )
end

--------------------------------------------------------------------------------

function WFCatCollaboration.sharedAlbumUrl( catalog, albumId )
	
	return catalog:docStore():docValue( albumId )
		:pluck( '_id' )
		:flatMap( function( fullAlbumId )
			return sharedAssetOrAlbumUrl( catalog, 'album', fullAlbumId )
		end )
end				


--------------------------------------------------------------------------------

--[[!
	\brief deletes an album share. Generally only use this when the album itself is being 
		deleted. In most cases, this is not what you want as the share is
		permanently deleted. Unsharing an album removes the share from public view but
		keeps it's metadata (particularly it's URL) which allows share/unshare/share to
		be a non-destructive workflow. 
		
	\param albumDocId (string) The unique id of the album to remove shares for.
	
	\return <code>Observable< ></code> An Observable object that completes when 
		the operation is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.deleteSharedAlbum( catalog, albumDocId )
	
	return deleteAlbumOrAssetShare( catalog, 'album', albumDocId )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief shares an album, creating the associated space and space_album documents.

	\param albumId (string|int) The unique or local ID of the album.
	\param optAttributes ( table ) optional attributes for creating the shared space

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.shareAlbum( catalog, albumId, optAttributes )

	
	return catalog:docStore():docValue( albumId )
		:pluck( '_id' )
		:flatMap( function( fullAlbumId )
			return shareAssetOrAlbum( catalog, 'album', fullAlbumId, 'shared' , optAttributes )
		end )
end

--------------------------------------------------------------------------------
--[[!
	\brief Unshares an album, deleting the associated space and space_album documents.

	\param albumId (string|int) The unique or local ID of the album.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.unshareAlbum( catalog, albumId )
	
	return catalog:docStore():docValue( albumId )
		:pluck( '_id' )
		:flatMap( function( fullAlbumId )
			return unshareAssetOrAlbum( catalog, 'album', fullAlbumId )
		end )

end

-------------------------------------------------------------------------------
--[[!
	\brief get shared album attributes.

	\param albumId (string|int) The unique or local ID of the album.

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.getSharedAlbumAttributes( catalog, albumId )

	return catalog:docStore():docValue( albumId )
		:pluck( '_id' )
		:flatMap( function( fullAlbumId )
			return getShareAttributes( catalog, 'album', fullAlbumId )
		end )
end

-------------------------------------------------------------------------------
--[[!
	\brief set shared album attributes.

	\param albumId (string|int) The unique or local ID of the album.
	\param attributes ( table ) updated attributes for the shared space

	\return <code>Observable< ></code> An Observable object completes when operation
		is completed; or resolves to an error if an error occurs.
]]--
function WFCatCollaboration.setSharedAlbumAttributes( catalog, albumId, attributes )

	return catalog:docStore():docValue( albumId )
		:pluck( '_id' )
		:flatMap( function( fullAlbumId )
			return setShareAttributes( catalog, 'album', fullAlbumId, attributes )
		end )
end
--------------------------------------------------------------------------------

return WFCatCollaboration
