--[[----------------------------------------------------------------------------

WFCatConflictResolver.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFTableMerge = require 'WFTableMerge'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFCatConflictResolver'
--	:setLevel( 'debug' )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--============================================================================--

-- @mark WFCatConflictResolver namespace

local function generateIndexSchema( _, indexStoreSchema )

	check( '_', 'WFRxIndexStoreSchema' )

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE conflicts( sequence )]],

		map = function( docValue, emit )

			if docValue._conflict then
				emit( docValue._changeCounter )
			end

		end,

		hasOneRowPerDocument = true,

	}

	indexStoreSchema:addIndex [[CREATE INDEX conflicts_sequence ON conflicts( sequence )]]

end

--------------------------------------------------------------------------------

local function stripNoiseFromRev( docValue )

	docValue._changeCounter = nil
	docValue._conflict = nil
	docValue._revId = nil

end

--------------------------------------------------------------------------------

local function resolveConflictingRevs( docStore, localRev, remoteRev )

	-- Note: The variable naming here reflects the *likely* scenario given the
	-- doc store's strategy of making bulk changes (i.e. typically changes received
	-- from Oz) be the losing revision. Not guaranteed, but highly likely.

	local localHistory = docStore:docRevisionHistory( localRev._id, localRev._revId )
	local remoteHistory = docStore:docRevisionHistory( remoteRev._id, remoteRev._revId )

	local remoteRevId = remoteRev._revId
	
	return Rx.Observable.combineLatest(
		function( localRevList, remoteRevList )

			logger:debugf( "localRevList = %s", localRevList )
			logger:debugf( "remoteRevList = %s", remoteRevList )

			-- Identify the first available revision that occurs in both lists
			-- and use it as base for three-way merge.

			local localRevIds = {}
			for _, info in ipairs( localRevList ) do
				if info.status == 'available' then
					localRevIds[ info.revId ] = true
				end
			end

			for _, info in ipairs( remoteRevList ) do
				if info.status == 'available' and localRevIds[ info.revId ] then
					return info.revId
				end
			end

		end, localHistory, remoteHistory )

		:take( 1 )
		:flatMapLatest( function( baseRevId )

			if baseRevId then
				return docStore:docValue( localRev._id, baseRevId )
			else
				return Rx.Observable.just( nil )
			end

		end )

		:flatMapLatest( function( baseRev )

			logger:debugf( "localRev = %s", localRev )
			logger:debugf( "remoteRev = %s", remoteRev )
			logger:debugf( "baseRev = %s", baseRev )

			if localRev._deleted or remoteRev._deleted then
				-- if either revision has been deleted, that's the change
				-- that should win so explicity delete the doc
				return docStore:deleteDoc( localRev._id )
				
			end
			
			if baseRev then
				
				stripNoiseFromRev( localRev )
				stripNoiseFromRev( remoteRev )
				stripNoiseFromRev( baseRev )

				local mergedRev = WFTableMerge.threeWayMerge( baseRev, localRev, remoteRev )

				logger:infof( "mergedRev = %s", mergedRev )

				local overlay = WFTableMerge.createOverlay( localRev, mergedRev )
				overlay._id = localRev._id

				return docStore:applyOverlay( overlay )

			else

				-- We can't find a suitable base revision on which to do a three-way
				-- merge.

				if localRev._changeCounter and remoteRev._changeCounter then

					-- Use the changeCounter to determine which conflicting rev should win.
					-- The reasoning is, that a higher changeCounter usually means a "newer" revision.

					if remoteRev._changeCounter > localRev._changeCounter then

						-- Modify localRev to the content of remoteRev.
						stripNoiseFromRev( localRev )
						stripNoiseFromRev( remoteRev )

						local overlay = WFTableMerge.createOverlay( localRev, remoteRev )
						overlay._id = localRev._id

						return docStore:applyOverlay( overlay )

					else

						-- localRev is automatically winning by deleting remoteRev in next step.

						return Rx.Observable.just( nil )

					end

				end

				-- Can we reach this point?
				-- Give up and arbitrary solve conflict by deleting remoteRev in next step.

				logger:errorf( "Arbitrarily resolving conflict for doc ID %s", localRev._id )

				return Rx.Observable.just( nil )

				-- TODO [scouten 2016-03-02]: We should probably log this event
				-- for diagnostic log.

			end

		end )

		:flatMapLatest( function()

			return docStore:deleteRevisionAsNonCurrent( remoteRev._id, remoteRevId )

		end )
		
end

--------------------------------------------------------------------------------

local function resolveOneConflict( docStore, docId )

	return docStore:currentRevisionsForDoc( docId, true )
		:tapOnError( function( message )
			logger:errorf( "docStore:currentRevisionsForDoc( %s ) failed with message %s",
							   docId, message )
		end )
		:flatMap( function( currentRevs )
				
				-- If there are more than two current revs, ignore third and successive
				-- revs. We'll catch those on a successive round. Without P2P sync, this
				-- scenario is highly unlikely but I can't completely rule it out.

				if type( currentRevs ) == 'table' and currentRevs[ 2 ] then
					return resolveConflictingRevs( docStore, currentRevs[ 1 ], currentRevs[ 2 ] )
				else
					return Rx.Observable.empty()
				end

		end )

end

--------------------------------------------------------------------------------

local WFCatConflictResolver = {

	_conflictQuerySubscription = false,
		-- (Subscription) holds the conflicts query subscription
	
	_currentConflictResolutions = {},
		-- (table) 
		
	class = {
		generateIndexSchema = generateIndexSchema,
	},
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function WFCatConflictResolver:__init( docStore, indexStore )

	check( 'WFCatConflictResolver', 'WFRxDocStore', 'WFRxIndexStore' )

	local query = indexStore:query {
				sql = [[SELECT docId, sequence FROM conflicts ORDER BY sequence ASC LIMIT 10]],
				calm = 1,
			}
	
	self._conflictQuerySubscription = query:subscribe(
		function( docIdRows )
			if #docIdRows > 0 then
				logger:infof( "Conflict resolver identified %d matching documents", #docIdRows )
			end
			for _, docIdRow in ipairs( docIdRows ) do
				local docId = docIdRow[ 1 ]
				
				if not self._currentConflictResolutions[ docId ] then
						logger:infof( "Resolving conflict for %s", docId )
						
						local clearResolutionSubscription = function() 
							self._currentConflictResolutions[ docId ] = nil 
						end
						
						self._currentConflictResolutions[ docId ] = resolveOneConflict( docStore, docId )
																:subscribe( nil,
																			clearResolutionSubscription, 
																			clearResolutionSubscription )
							
				else
					logger:warnf( "Skipping resolution for %s because one is in progress", docId )
				end
				
			end
		end,
		function( message )
			logger:errorf( "Conflict resolver query failed with message: %s", message )
			self._conflictQuerySubscription = nil
		end,
		function()
			logger:debugf( "Conflict resolver query notified as done" )
			self._conflictQuerySubscription = nil
		end )

end

--------------------------------------------------------------------------------

function WFCatConflictResolver:close()
	
	if self._conflictQuerySubscription then
		self._conflictQuerySubscription:dispose()
		self._conflictQuerySubscription = nil
	end
end

--------------------------------------------------------------------------------

-- TODO [scouten 2015-08-05]: Diagnostic log reporting on # conflicts resolved.

--------------------------------------------------------------------------------

return Rx.Class 'WFCatConflictResolver'( WFCatConflictResolver )
