--[[----------------------------------------------------------------------------

WFCatCopies.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.os
local WFAsyncFileUtils = require 'WFAsyncFileUtils'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCatIdUtils = require 'WFCatIdUtils'
local WFTableMerge = require 'WFTableMerge'

	-- WF.oz_client.managed_catalog
local WFCatStacks = require 'WFCatStacks'

	-- WF.oz_client.managed_catalog
local WFCatFileSpec = require 'WFCatFileSpec'
local NmbLocalRenditionDigest = require 'NmbLocalRenditionDigest'
local WFCatUtils = require 'WFCatUtils'

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )
local deepCopy = assert( KSTableUtils.deepCopy )
local check = assert( WFArguments.check )

local kBatchSize = 40

local underscoreByte = string.byte( '_' )

--==============================================================================

local WFCatCopies = {}

--------------------------------------------------------------------------------

--[[!
	\brief Set the given set of assets as a copy of each other. If any one of the
	assets already has a copy_id, it is propagated through to all the assets. If 
	there is a conflict, throw an error

	\param wfCatalog the catalog
	\param assetIdArray An array of asset-ids

	\return Rx.Observable( array of strings ) Array of revision IDs for each document.
--]]

function WFCatCopies.setAssetsAsCopies( wfCatalog, assetIdArray )

	check( 'WFCatalog', 'table' )

	if #assetIdArray < 2 then
		return Observable.empty()
	end
	
	local docStore = wfCatalog:docStore()
	
	return docStore:bulkDocValue( assetIdArray )
		:flatMap( function( docValueArray )
			
			local copyId
			local docOverlays = {}
			
			-- Check if these assets are already a part of a copy
			for _, docValue in ipairs( docValueArray ) do
				
				local docCopyId = docValue and docValue.copy and docValue.copy.copy_id

				if docCopyId and copyId == nil then
					copyId = docCopyId
					break
				end
			end
			
			-- Create a copy id if none of them are already a copy
			if not copyId then
				copyId = KSIdGenerator.uuidInLowercaseHex( )
			end
			
			-- Form the overlays
			for _, docValue in ipairs( docValueArray ) do
			
				if docValue.copy and docValue.copy.copy_id then
					assert( docValue.copy.copy_id == copyId, "Trying to change the copy-id of an asset" )
				else
					local overlay = {
						_id = docValue._id,
						copy = {
							copy_id = copyId,
						}
					}
					
					table.insert( docOverlays, overlay )
				end
			end
			
			if #docOverlays == 0 then
				-- If all copy ids are already applied
				-- then nop
				return Observable.just()
			end
			
			return docStore:bulkApplyOverlay( docOverlays )
			
		end )
	
end

--------------------------------------------------------------------------------

local function createCopyOfDocValue( docValue )

	local docValueOverlay = nil

	local copyDocValue = deepCopy( docValue )

	for k, _ in pairs( copyDocValue ) do
		if string.byte( k, 1 ) == underscoreByte then
			copyDocValue[ k ] = nil
		end
	end

	copyDocValue.label = nil

	-- Clear out importSource.uniqueDeviceId to work around a problem where copies
	-- are deleted by lightroom mobile if it sees them before the copy is completed.
	if copyDocValue.importSource then
		copyDocValue.importSource.uniqueDeviceId = nil
	end

	copyDocValue._id = KSIdGenerator.uuidInLowercaseHex()

	-- Copy the info about the xmp sidecar as it may hint that a sidecar needs to be
	-- uploaded if we just copied a file with one that wasn't uploaded yet.
	local copyDocLocalOnly = {
		xmpSidecarExists = derefOpt( docValue, '_localOnly', 'xmpSidecarExists' ),
		xmpSidecarNeedsUpload = derefOpt( docValue, '_localOnly', 'xmpSidecarNeedsUpload' ),
	}

	local copyId = derefOpt( docValue, 'copy', 'copy_id')
	if not copyId then

		local copyInfo = { copy_id = KSIdGenerator.uuidInLowercaseHex() }
		copyDocValue.copy = copyInfo

		-- Add a copy.copy_id to the original document as well.
		docValueOverlay = {
			_id = docValue._id,
			copy = copyInfo
		}

	end

	-- If the source exists on oz copy the links for the copy operations that will be
	-- issued by the uploader.  This is needed to get the correct develop settings
	-- if an asset doesn't have them downloaded locally when copied as they are
	-- usually modified and end up different from other copied assets.
	copyDocLocalOnly.sourceOriginalLink = WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
	copyDocLocalOnly.sourceXmpLink = WFCatFileSpec.getXmpLinkFromDocValue( docValue )
	copyDocLocalOnly.sourceXmpDevelopLink = WFCatFileSpec.getXmpDevelopLinkFromDocValue( docValue )
	copyDocLocalOnly.sourceProxyLink = WFCatFileSpec.getProxyLinkFromDocValue( docValue )

	copyDocLocalOnly.sourceAssetId = docValue._id

	if next( copyDocLocalOnly ) ~= nil then
		copyDocValue._localOnly = copyDocLocalOnly
	end

	return copyDocValue, docValueOverlay

end

--------------------------------------------------------------------------------

local function maybeCopyFile( sourceFilePath, copyFilePath )

	if not sourceFilePath then
		return Observable.just( nil )
	end

	return WFAsyncFileUtils.copyToUniquelyNamedFile( sourceFilePath, copyFilePath )

end

--------------------------------------------------------------------------------

local function copyFilesToNewDoc( wfCatalog, docValue, copyDocValue, docValueOverlay )

	local originalPath = WFCatFileSpec.buildOriginalPathFromDocValue( docValue )
	local originalCopy = originalPath and WFFileUtils.ensureFileNameIsUnique( originalPath )

	local proxyPath = WFCatFileSpec.buildProxyPathFromDocValue( docValue )
	local proxySize, proxyType = WFCatFileSpec.getProxySizeFromDocValue( docValue )
	local proxyCopy = proxyType and wfCatalog:fileSpec():pathForProxy( copyDocValue._id, proxyType )

	local xmpSidecarPath = WFCatFileSpec.getXmpSidecarPathFromDocValue( docValue )
	local xmpSidecarCopy = wfCatalog:fileSpec():pathForXmpSidecar( copyDocValue._id )

	local xmpDevelopInfo = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue )
	-- Don't bother copying devSettings as it shares the same path for all
	-- assets that have the same settings sha256.

	local renditionPaths = {}

	local md5XmpCameraRawSrc = NmbLocalRenditionDigest.maybeComputeFromDocValue( docValue )

	local md5XmpCameraRawCpy = NmbLocalRenditionDigest.maybeComputeFromDocValue( copyDocValue )

	if md5XmpCameraRawSrc and md5XmpCameraRawCpy then

		for _, longEdge in ipairs( { 320, 640, 1280, 2048 } ) do
			local path = wfCatalog:fileSpec():pathForPreviewDigestAndLongEdge( docValue._id, md5XmpCameraRawSrc, longEdge, false )
			if WFFileUtils.exists( path ) == "file" then
				local copy = wfCatalog:fileSpec():pathForPreviewDigestAndLongEdge( copyDocValue._id, md5XmpCameraRawCpy, longEdge, true )
				table.insert( renditionPaths, { srcPath = path, cpyPath = copy } )
			end
		end

	end

	if #renditionPaths == 0 then
		-- Think this will only ever happen in unit tests.
		table.insert( renditionPaths, {} ) -- add an empty value so takeLast( 1 ) below will always get something.
	end

	return Observable.from( renditionPaths )
	:flatMap( function( copyRendition )

		-- Copy the renditions so the image appears immediately and copy will
		-- work without a proxy or original when offline.
		return maybeCopyFile( copyRendition.srcPath, copyRendition.cpyPath )

	end )
	:takeLast( 1 )
	:flatMap( function()

		return Observable.combineLatest(
			maybeCopyFile( originalPath, originalCopy ),
			maybeCopyFile( proxyPath, proxyCopy ),
			maybeCopyFile( xmpSidecarPath, xmpSidecarCopy ) )

	end )
	:map( function( originalCopy, proxyCopy, xmpSidecarCopy )

		local originalInfo = {
			master = originalCopy and {
				relativePath = WFFileUtils.removeVolumeFromPath( originalCopy ),
				volumeGuid = WFFileUtils.getVolumeGuidForPath( originalCopy ),
				size = derefOpt( docValue, '_localOnly', 'files', 'original', 'master', 'size' ),
					-- deref docValue directly as getOriginalSizeFromDocValue also checks the importSource.
			} or nil,
			xmp = xmpSidecarCopy and {
				path = xmpSidecarCopy,
				fileSize = WFCatFileSpec.getXmpSidecarFileSizeFromDocValue( docValue ),
			} or nil,
		}

		if next( originalInfo ) == nil then
			originalInfo = nil
		end

		local proxyInfo = nil

		if proxyCopy then
			proxyInfo = {
				[ proxyType ] = {
					relativePath = WFFileUtils.removeVolumeFromPath( proxyCopy ),
					volumeGuid = WFFileUtils.getVolumeGuidForPath( proxyCopy ),
					size = proxySize,
				}
			}
		end

		local xmpSidecarExists = WFCatFileSpec.getXmpSidecarExistsFromDocValue( docValue )

		WFTableMerge.applyOverlay( copyDocValue, {
			_localOnly = {
				files = {
					original = originalInfo,
					xmp_develop = xmpDevelopInfo,
					proxy = proxyInfo,
				},
				xmpSidecarNeedsUpload = xmpSidecarExists ~= false, -- true or nil
					-- Upload or copy the xmp sidecar if we know one exists or if
					-- we don't know. Don't bother uploading xmp if we know
					-- one doesn't exist.
				xmpSidecarExists = xmpSidecarExists,
			}
		} )

		return copyDocValue, docValueOverlay

	end )

end

--------------------------------------------------------------------------------

local function getAllAlbumsForAssets( wfCatalog, assetIdArray )
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT album_asset_v2.assetId, album_asset_v2.albumId FROM TEMP_ids JOIN assets ON TEMP_ids.docId == assets.docId
					JOIN album_asset_v2 ON assets.assetId == album_asset_v2.assetId]],
		tempIds = assetIdArray,
	}


end

--------------------------------------------------------------------------------

local function getAlbumAssetMappings( wfCatalog, assetIdArray, albumIdArray )
	if ( not albumIdArray ) or 
		( not next( albumIdArray ) ) then
		return Observable.just( {} )
	end
	
	-- get the albums that each asset is in 
	return Observable.combineLatest( getAllAlbumsForAssets( wfCatalog, assetIdArray ),
						WFCatIdUtils.getOzAlbumIdsForLocalId( wfCatalog, albumIdArray ) )
		:take( 1 )
		:map( function( albumAssetRows, selectedAlbumIds )
			-- build the albums we care about so we can easily filter
			local selectedAlbumMap = {}
			for _, albumId in ipairs( selectedAlbumIds ) do
				selectedAlbumMap[ albumId ] = true
			end
			
			local assetToAlbumMap = {}
			for _, albumAsset in ipairs( albumAssetRows ) do
				local asset = albumAsset[ 1 ]
				local album = albumAsset[ 2 ]
				
				if selectedAlbumMap[ album ] then
					local albums = assetToAlbumMap[ asset ] or {}
					table.insert( albums, album )
					assetToAlbumMap[ asset ] = albums
				end
			end
			
			return assetToAlbumMap
		end )
	
end

--------------------------------------------------------------------------------

--[[!
	\brief Create a copy of each of the assets in the provided assetIdArray

	\param wfCatalog the catalog
	\param assetIdArray An array of asset-ids
	\param albumIdArray An array of album-ids
	\param stackIdsByAlbumIds A table of stack-ids whithin which copies are
		being created, keyed by ids of assets having copies created

	\return Rx.Observable() that completes when done.
--]]

function WFCatCopies.createCopiesOfAssets( wfCatalog, assetIdArray, albumIdArray, stackIdsByAssetIds )

	check( 'WFCatalog', 'table' )

	local copiedAssetIds = {}
	
	local stackIds = {}
	local stackAssetIds = {}
	
	for assetId, stackId in pairs( stackIdsByAssetIds or {} ) do
		table.insert( stackIds, stackId )
		table.insert( stackAssetIds, assetId )
	end

	local docStore = wfCatalog:docStore()

	return Observable.from( assetIdArray )
	:bufferWithCount( kBatchSize )
	:flatMap( function( assetIdArray )

		return Observable.combineLatest( docStore:bulkDocValue( assetIdArray ),
										getAlbumAssetMappings( wfCatalog, assetIdArray, albumIdArray ),
										WFCatIdUtils.getOzAssetIdsForLocalIds( wfCatalog, stackIds ),
										WFCatIdUtils.getOzAssetIdsForLocalIds( wfCatalog, stackAssetIds )
										)
		:flatMap( function( docValueArray, assetToAlbumMap, ozStackIds, ozAssetIds )
		
			-- Construct stackIdsByAssetIds' equivalent with Oz ids.
			local ozStackIdsByOzAssetIds = {}
			for i, localAssetId in ipairs( stackAssetIds ) do
				if ozAssetIds[ i ] and ozStackIds[ i ] then
					ozStackIdsByOzAssetIds[ ozAssetIds[ i ] ] = ozStackIds[ i ]
				end
			end

			return Observable.from( docValueArray )
			:map( function( docValue )
				local copyValue, docValueOverlay = createCopyOfDocValue( docValue )
				return docValue, copyValue, docValueOverlay
			end )
			:flatMap( function( docValue, copyDocValue, docValueOverlay )
				return copyFilesToNewDoc( wfCatalog, docValue, copyDocValue, docValueOverlay )
			end )
			:scan( function( newDocValues, existingDocOverlays, copyDocValue, existingDocOverlay )
				table.insert( copiedAssetIds, copyDocValue._id )
				table.insert( newDocValues, copyDocValue )
				
				local sourceAssetId = KSTableUtils.derefOpt( copyDocValue, "_localOnly", "sourceAssetId" )

				if sourceAssetId then
					local albumsForCopiedAsset = assetToAlbumMap[ sourceAssetId ]
					if albumsForCopiedAsset then
						
						for _, albumId in ipairs( albumsForCopiedAsset ) do
						
							local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( KSIdGenerator.uuidInLowercaseHex() )
						
							-- for the album, create an album_asset record for the copy
							
							local albumAssetDocValue = { 
								type = 'album_asset',
								album = {
									id = albumId,
								},
								asset = {
									id = copyDocValue._id,
								}
							}
									
							WFTableMerge.applyOverlay( albumAssetDocValue, baseDocValue )
									
							table.insert( newDocValues, albumAssetDocValue )
									
						end
					end

					local ozStackId = ozStackIdsByOzAssetIds[ sourceAssetId ]
					if ozStackId then
						table.insert( newDocValues, WFCatStacks.createStackAssetDocument( ozStackId, copyDocValue._id ) )
					end
				end
				
				if existingDocOverlay then
					table.insert( existingDocOverlays, existingDocOverlay )
				end
				return newDocValues, existingDocOverlays
			end, {}, {} )
			:takeLast( 1 )
			:flatMap( function( copyDocValues, existingDocOverlays )

				local result
				if #existingDocOverlays > 0 then
					result = docStore:bulkApplyOverlay( existingDocOverlays )
				else
					result = Observable.empty()
				end
				
				return result
				:merge( docStore:bulkSetDocValue( copyDocValues ) )
				:takeLast( 1 )
			end )

		end )
		:call( WFRxRunUntilDone )
			-- Using runUntilDone within the flatMap so each buffer of
			-- kBatchSize runs to completion to ensure corresponding files
			-- are copied. This is to handle the case if the model was
			-- closed to terminate the action.

	end )
	:takeLast( 1 )
	:map( function()
		return copiedAssetIds
	end )

end

--------------------------------------------------------------------------------

function WFCatCopies.createCopiesOfAssetsModel( wfCatalogStream, imageIds, maybeAlbumIdArray, maybeStackIdsByImageIds )

	check( '_', 'table' )

	local value = wfCatalogStream
		:filter( function( wfCatalog )
			return wfCatalog ~= nil
		end )
		:take( 1 )
		:flatMap( function( wfCatalog )
			return WFCatCopies.createCopiesOfAssets( wfCatalog, imageIds, maybeAlbumIdArray, maybeStackIdsByImageIds )
		end )
		:call( WFRxRunUntilDone )

	return {
		value = value,
	}

end

--==============================================================================

return WFCatCopies
