--[[----------------------------------------------------------------------------

WFCatDevelopSettings.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.os
local KSData = require 'KSData'
local KSDate = require 'KSDate'
local WFArguments = require 'WFArguments'
local WFFunctionUtils = require 'WFFunctionUtils'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFAsyncDigest = require 'WFAsyncDigest'
local WFAsyncFileUtils = require 'WFAsyncFileUtils'
local WFConfiguration = require 'WFConfiguration'
local WFLogger = require 'WFLogger'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFFileUtils = require 'WFFileUtils'

	-- wf.acr
local WFCrInfo = require 'WFCrInfo'

	-- WF.oz_client.managed_catalog
local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatUtils = require 'WFCatUtils'

	-- WF.oz_client.managed_catalog.develop
local NmbDocValueDevelopSettingsUtils = 
		require "NmbDocValueDevelopSettingsUtils"

local NmbProfileCatalogUtils = require 'NmbProfileCatalogUtils'

--------------------------------------------------------------------------------

local INLINE_XMP_CAMERA_RAW_MAX_LENGTH = WFConfiguration.Nimbus.inlineXmpMaxLength
										or 10 * 1024 --10KB

--------------------------------------------------------------------------------

local logger = WFLogger( 'WFCatDevelopSettings' )

local derefOpt = assert( KSTableUtils.derefOpt )

local KSDate_timeInIso8601FormatNearestSecond = assert( KSDate.timeInIso8601FormatNearestSecond, "KSDate.timeInIso8601FormatNearestSecond does not exist" )

local runUntilDone = WFRxRunUntilDone

--============================================================================--

local function matchesPixStreamId( pixStreamIdToMatch )

	return function( pixStreamId, ... )
	
		return pixStreamId == pixStreamIdToMatch

	end

end

--------------------------------------------------------------------------------

local function createKey( pixStreamId, assetId )

	return string.format( "%s:%d", pixStreamId, assetId )

end

--------------------------------------------------------------------------------

local observersWaitingForDevelopSettings = {}

local retainedDevSettingsPathObservers = {}

local xmpDevelopSettingsPathStreamSubscription

--------------------------------------------------------------------------------

local function bootstrapXMPDevelopSettingsPathStreamSubscription( pixelProviderProcessor, pixStreamIdToMatch )

	if not xmpDevelopSettingsPathStreamSubscription then
	
		xmpDevelopSettingsPathStreamSubscription = pixelProviderProcessor
				:relayStream( 'pixelPathStream' )
				:where( matchesPixStreamId( pixStreamIdToMatch ) )
				:subscribe( function( pixStreamId, assetId, success, xmpDevelopPath )
				
					local key = createKey( pixStreamId, assetId )
					
					local observers = observersWaitingForDevelopSettings[ key ] or {}
					
					for _, observer in pairs( observers ) do
					
						observer:onNext( success, xmpDevelopPath )
					
					end
				
					observersWaitingForDevelopSettings[ key ] = nil
				
					pixelProviderProcessor:runUntilDone( 'removeRequest',
															pixStreamId, assetId )
				
				end )

	end

end

--------------------------------------------------------------------------------

local function getXmpDevelopPath( pixelProviderProcessor, imageId, needsASAP )

	local pixStreamId = 'wfCatDevelopSettings'
	
	return Rx.Observable.create( 

		function( observer )

			bootstrapXMPDevelopSettingsPathStreamSubscription( 
				pixelProviderProcessor, pixStreamId )

			local key = createKey( pixStreamId, imageId )

			local observers = observersWaitingForDevelopSettings[ key ] or KSTableUtils.makeFullyWeak()
			
			observers[ observer ] = observer
			
			observersWaitingForDevelopSettings[ key ] = observers
			
			retainedDevSettingsPathObservers[ observer ] = observer
			
			pixelProviderProcessor:runUntilDone( 
									"requestXmpDevelopForDevelop", 
										pixStreamId,
										imageId,
										needsASAP )

			return Rx.Disposable.create( function() 
			
				retainedDevSettingsPathObservers[ observer ] = nil
			
			end )

		end )

end

--------------------------------------------------------------------------------

local function getExternalFileXMPSettingsStream( 
				pixelProviderProcessor, imageId, docValue, xmpCameraRaw,
				baseSettings, needsASAP )

	local xmpPathObservable = 
			getXmpDevelopPath( pixelProviderProcessor, imageId, needsASAP )
			:flatMap( 
				function( success, xmpDevelopPath ) 
					
					if not success or not xmpDevelopPath then
						logger:warnf( "Failed to fetch xmp_develop settings for asset %s: %s", imageId, xmpDevelopPath )
						
						baseSettings.settingsXmpStatus = 'failed'
						
						return Observable.just( baseSettings )
					end
					
					return WFAsyncFileUtils.readFile( xmpDevelopPath )
							:map( function( fileData ) 
								
								if fileData then

									baseSettings.settingsXmp = tostring( fileData )

									baseSettings.settingsXmpStatus = 'complete'

									fileData:close()

								else

									logger:warnf( "failed to read xmp from external develop settings file: %s",
													xmpDevelopPath )

									baseSettings.settingsXmpStatus = 'failed'

								end
								
								return baseSettings

							end )
					
				end )
		
	local xmpDevelopInfo = WFCatFileSpec.getXmpDevelopInfoFromDocValue( 
											docValue )

	if not xmpDevelopInfo or xmpDevelopInfo.sha256 ~= xmpCameraRaw.sha256 then

		-- if we don't think we have this locally yet, mark it pending as it
		-- may take a bit and clients might need to know that we're fetching
		-- the settings
		baseSettings.settingsXmpStatus = 'pending'
		
		xmpPathObservable = xmpPathObservable:startWith( baseSettings )
		
	end
	
	return xmpPathObservable

end

--------------------------------------------------------------------------------

local NmbDocValueDevelopSettingsUtils_getSanitizedDocValueDevelopStanza =
		assert( NmbDocValueDevelopSettingsUtils.getSanitizedDocValueDevelopStanza,
				"missing NmbDocValueDevelopSettingsUtils.getSanitizedDocValueDevelopStanza" )

--------------------------------------------------------------------------------

local function getValidatedXMPSettingsStream( pixelProviderProcessor, imageId,
												docValue, baseSettings, needsASAP )

	local developStanza = 
			NmbDocValueDevelopSettingsUtils_getSanitizedDocValueDevelopStanza(
				docValue )

	local xmpCameraRaw = developStanza.xmpCameraRaw
		
	if type( xmpCameraRaw ) == 'table' and xmpCameraRaw.sha256 then
		
		return getExternalFileXMPSettingsStream( 
					pixelProviderProcessor, imageId, docValue, xmpCameraRaw,
					baseSettings, needsASAP )

	end
	
	if xmpCameraRaw == nil or type( xmpCameraRaw ) == 'string' then

		baseSettings.settingsXmp = xmpCameraRaw

		baseSettings.settingsXmpStatus = 'complete'
	
		return Observable.just( baseSettings )

	end

	logger:warnf( "bad docValue.develop.xmpCameraRaw value in docstore: %s",
					type( xmpCameraRaw ) ~= "table" and
						type( xmpCameraRaw ) or
						"xmp was a table without a sha256" )

	baseSettings.settingsXmpStatus = 'failed'
	
	return Observable.just( baseSettings )

end

--------------------------------------------------------------------------------

local function docValueHasLocalOriginal( docValue )

	local localOriginalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
	local localOriginalRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue )
	return localOriginalVolumeGuid and localOriginalRelativePath

end

--------------------------------------------------------------------------------

local function developContentStreamImp( pixelProviderProcessor, docValue, needsASAP )
	
	if docValue._empty or docValue.type ~= 'asset' then

		logger:errorf( "Asset does not exist, or has wrong type: id = %s",
								docValue._localId )

		return Observable.empty()

	end
	
	local imageId = docValue._localId

	local originalPath = WFCatFileSpec.buildOriginalPathFromDocValue( docValue )
	
	if originalPath and not WFFileUtils.exists( originalPath ) then
		originalPath = nil
	end
	
	local proxyPath = WFCatFileSpec.buildProxyPathFromDocValue( docValue )
	
	if proxyPath and not WFFileUtils.exists( proxyPath ) then
		proxyPath = nil
	end
	
	local isUnsupportedFormat = WFCatUtils.isUnsupportedFormat( docValue )
	
	-- 2017.07.28 bkruse
	-- 
	-- Nimbus has one asset id per file semantics, use the asset's id as the
	-- fileUUId. This might change in the future, but for now, it's true and
	-- filUUId is used to cache negative paths so we don't want to
	-- assets to have the same fileUUId unless the files are indeed unique.
	-- 
	-- It is possible via catalog migration to have two files with the same
	-- sha256, so we can't rely on the sha256 as unique fileUUId
	local fileUuid = docValue._id -- derefOpt( docValue, 'importSource', 'sha256' ) or docValue._id
	
	local baseSettings = {

			changeCounter = tonumber( docValue._changeCounter ),

			fileUuid = fileUuid,

			historySettingsID = KSIdGenerator.uuid(),
	
			fileFormat = derefOpt( docValue, 'importSource', 'contentType' ) == 'image/jpeg' and "JPG" or "RAW",
				-- TODO [arahn 2015-08-27]: is "RAW" correct?
	
			payloadHeight = derefOpt( docValue, 'importSource', 'originalHeight' ),
			payloadWidth = derefOpt( docValue, 'importSource', 'originalWidth' ),
			noOriginalOnOz = derefOpt( docValue, 'importSource', 'willNotUploadMaster' ),
			id_local = docValue._localId,
			isVideo = derefOpt( docValue, 'subtype') == 'video',
			numUnlockedSnapshots = 0,
			numVirtualCopies = 0,
	
			tiffOrientation = derefOpt( docValue, 'develop', 'userOrientation' ) or 1,
			
			defaultCrop = derefOpt( docValue, 'importSource', 'originalCrop' ),
			
			rootFile = not isUnsupportedFormat and {
				filePath = originalPath or proxyPath,
				originalPath = originalPath,
				proxyPath = proxyPath,
			} or nil,
			
			settingsID = "default",
			
			virtualCopy = false,
			virtualCopyMap = {},

			hasProxyLink = WFCatFileSpec.getProxyLinkFromDocValue( docValue ) ~= nil,
			hasOriginalLink = WFCatFileSpec.getOriginalLinkFromDocValue( docValue ) ~= nil,
			
			hasLocalOriginal = docValueHasLocalOriginal( docValue ),
			
			isUnsupportedFormat = isUnsupportedFormat,
			
		}
		
	return getValidatedXMPSettingsStream( 
			pixelProviderProcessor, imageId, docValue, baseSettings,
			needsASAP )
	
end

--------------------------------------------------------------------------------

local function docIsNonEmpty( docValue )

	if not docValue._empty then

		return docValue

	end

end

--------------------------------------------------------------------------------

local function nonEmptyDocValues( docValues )

	return WFFunctionUtils.filter( docIsNonEmpty, docValues )

end

--------------------------------------------------------------------------------

local function addToSet( set, assetId )

	set[ assetId ] = true

	return set

end

--------------------------------------------------------------------------------
-- isUpdateWanted is called from 
-- 	docStore:scannedFilteredMappedChangesFeed( ... )
-- 	
-- 	it's used to scan *all* ongoing changes to the doc store and wantUpdatesSet
-- 	is used to filter out only those docValues that we are interested in.
-- 	
-- it should always return wantUpdatesSet unchanged. wantUpdatesSet will always
-- hold the set of assetIds for which we want updateds
-- 
-- it should return the docValue as a second value if the docValue is of interest
--------------------------------------------------------------------------------

local function isUpdateWanted( wantUpdatesSet, docValue )

	if docValue._empty or docValue._deleted then

		return wantUpdatesSet

	end

	local localAssetId = docValue._localId

	if not localAssetId then

		return wantUpdatesSet

	end

	if not wantUpdatesSet[ localAssetId ] then

		return wantUpdatesSet

	end

	return wantUpdatesSet, docValue

end

--------------------------------------------------------------------------------

local function addToBatch( batches, assetId, batchSize )

	local currBatchSize = #( batches[ #batches ] )

	if currBatchSize == batchSize then

		currBatchSize = 0

		batches[ #batches + 1 ] = {}

	end

	batches[ #batches ][ currBatchSize + 1 ] = assetId

	return batches

end

--------------------------------------------------------------------------------

local function batchedDocValuesStream( desc, wfCatalog, assetIds, batchSize )

	local batchedAssetIds = 
			WFFunctionUtils.reduce( { {} }, addToBatch, assetIds, batchSize )

	return Rx.Observable.from( batchedAssetIds )
			:map( 
				function( assetIdsBatch )

					return wfCatalog
							:docStore()
							:benchmarkedBulkDocValue( desc, assetIdsBatch )
							:filter( nonEmptyDocValues )
							:flatMapLatest( Rx.Observable.from )

				end )
			-- flatMerge( 1 ) calls bulkDocValue only after previous completes
			:flatMerge( 1 ) 
					
end

--------------------------------------------------------------------------------

local _batchSize = 5

--------------------------------------------------------------------------------

local function ongoingDevelopContentsStream( 
					desc, wfCatalog, pixelProviderProcessor,
					needsInitialValueAndUpdatesAssetIds,
					needsJustUpdateAssetIds, needsASAP )

	if #needsInitialValueAndUpdatesAssetIds == 0 and
			#needsJustUpdateAssetIds == 0 then

		return Rx.Observable.empty()

	end

	local initialDocValuesStream =
			#needsInitialValueAndUpdatesAssetIds == 0 and
				Rx.Observable.empty() or
				batchedDocValuesStream( 
					desc, wfCatalog, needsInitialValueAndUpdatesAssetIds, 
					_batchSize )

	local wantUpdatesSet = WFFunctionUtils.reduce( 
							{}, addToSet, needsInitialValueAndUpdatesAssetIds )

	wantUpdatesSet = WFFunctionUtils.reduce( 
							wantUpdatesSet, addToSet, needsJustUpdateAssetIds )

	local updatesStream =
				wfCatalog
					:docStore()
					:scannedFilteredMappedChangesFeed( 
						"WFCatDevelopSettings", "isUpdateWanted", 
						wantUpdatesSet )
	
	local docValueStream = 
			Rx.Observable.concat( initialDocValuesStream, updatesStream )

	return docValueStream
			:flatMap(
				function( docValue )

					return developContentStreamImp( pixelProviderProcessor, 
													docValue, needsASAP )

				end )

end

--------------------------------------------------------------------------------

local function makeOngoingDevelopContentsStream(
					desc, wfCatalog, pixelProviderProcessor,
					needDevelopContentStream )

	return needDevelopContentStream
			:flatMapLatest(
				function( needsInitialValueAndUpdatesAssetIds,
							needsJustUpdateAssetIds, needsASAP )

					return ongoingDevelopContentsStream( 
							desc, wfCatalog, pixelProviderProcessor,
							needsInitialValueAndUpdatesAssetIds,
							needsJustUpdateAssetIds, needsASAP )

				end )

end

--------------------------------------------------------------------------------

local function updateFileAccessTimestampStream( 
				wfCatalog, needDevelopContentStream )

	return needDevelopContentStream
			:filter( 
				function( needsInitialValueAndUpdatesAssetIds,
							needsJustUpdateAssetIds )

					return  #needsInitialValueAndUpdatesAssetIds > 0  or
								#needsJustUpdateAssetIds > 0

				end )
			:map( 
				function( needsInitialValueAndUpdatesAssetIds,
							needsJustUpdateAssetIds )

					local localId = needsInitialValueAndUpdatesAssetIds[ 1 ] or
										needsJustUpdateAssetIds[ 1 ]

					wfCatalog:fileAccessDb():touch( localId )

					return string.format( "did update file access for %s", localId )

				end )
	
end

--------------------------------------------------------------------------------

local function makeOngoingDevelopContentsStreamWithUpdateFileAccess(
					wfCatalog, pixelProviderProcessor,
					needDevelopContentStream )

	return Rx.Observable.create(
			function( observer )

				local updateFileAccessTimestampSubscription =
						updateFileAccessTimestampStream( 
								wfCatalog, needDevelopContentStream )
							:subscribe(
								function( msg )

									logger:info( msg )

								end )

				local devContentsStream = 
					makeOngoingDevelopContentsStream(
						"ongoingDevelopContent", wfCatalog,
						pixelProviderProcessor, needDevelopContentStream )

				return Rx.CompositeDisposable(
							updateFileAccessTimestampSubscription,
							devContentsStream:subscribe( observer ) )


			end )

end
 
--------------------------------------------------------------------------------

local function ongoingDevelopContentModel( wfCatalogStream )

	local needDevelopContentStream = Rx.Subject()

	return {
		
		value = wfCatalogStream
					:mapSwitch(
						function( wfCatalog, pixelProviderProcessor )

							return makeOngoingDevelopContentsStreamWithUpdateFileAccess( 
									wfCatalog, pixelProviderProcessor,
									needDevelopContentStream )

						end ),

		events = {

			setInitialPlusUpdatesAssetId = function( _, assetId )

				needDevelopContentStream:onNext( { assetId }, {}, true )

			end,

			setUpdatesOnlyAssetId = function( _, assetId )

				WFArguments.check( "_", "number" )

				needDevelopContentStream:onNext( {}, { assetId }, true )

			end,

		}

	}
	
end
 
--------------------------------------------------------------------------------

local function developContentStream( wfCatalog, pixelProviderProcessor, 
										assetId, needsASAP )

	return wfCatalog
			:docStore()
			:docValue( assetId )
			:tap(
				function( docValue )

					logger:infof( "developContentStream got %sdoc for %s",
									not docIsNonEmpty( docValue ) and
										"empty" or "",
									assetId )
										
				end )
			:filter( docIsNonEmpty )
			:flatMap(
				function( docValue )

					if docValue._localId then

						wfCatalog
							:fileAccessDb()
							:touch( docValue._localId )

					end

					return developContentStreamImp( 
								pixelProviderProcessor, docValue, needsASAP )

				end )

end
 
--------------------------------------------------------------------------------

local function developContentModel( wfCatalogStream, assetId )

	return {
		
		value = wfCatalogStream
					:mapSwitch(
						function( wfCatalog, pixelProviderProcessor )

							return developContentStream( 
									wfCatalog, pixelProviderProcessor, 
									assetId )

						end )

	}
	
end

--------------------------------------------------------------------------------

local _developContentWithLocalNegativePathModelId = 0

--------------------------------------------------------------------------------

local function matches( ... )

	local matchN = select( '#', ... )

	local matchArgs = { ... }

	return function( ... )

		local receivedN = select( '#', ... )

		if receivedN < matchN then

			return false

		end

		for i = 1, matchN do

			if matchArgs[ i ] ~= select( i, ... ) then

				return false

			end

		end

		return true

	end

end

--------------------------------------------------------------------------------

local function settingsXmpStatusIsNotPending( baseSettings )
	
	return baseSettings.settingsXmpStatus ~= "pending"

end

-----------------------------------------------------------------------------------------

local function baseSettingsWithLocalNegativePathStream(
				wfCatalog, pixelProviderProcessor, pixStreamId, assetId ) 

	return pixelProviderProcessor
			:relayStream( "pixelPathStream" )
			:where( matches( pixStreamId, assetId ) )
			:tap(
				function( thePixStreamId, localAssetId, havePath )

					logger:infof( "%s.%s state: %s",
									thePixStreamId, localAssetId, 
									havePath and 
										"have local path" or
										"has no local path" )

				end )
			:flatMap(
				function( _, localAssetId, havePath )

					if not havePath then

						return Rx.Observable.throw( 
								string.format( "no local negative path for: %s",
												localAssetId ) )

					end

					return developContentStream( 
							wfCatalog, pixelProviderProcessor, 
							localAssetId, 'asap' )

				end )
			:tap(
				function( baseSettings )

					logger:infof( "%s.%s got preliminary settings with settingsXmpStatus: %s",
									pixStreamId, baseSettings.id_local, 
									baseSettings.settingsXmpStatus )

				end )
			:filter( settingsXmpStatusIsNotPending )
			:tap(
				function( baseSettings )

					logger:infof( "%s.%s got final settings with settingsXmpStatus: %s",
									pixStreamId, baseSettings.id_local, 
									baseSettings.settingsXmpStatus )

				end )
			:take( 1 )

end

--------------------------------------------------------------------------------

local function developContentWithLocalNegativePathModel( 
				wfCatalogStream, assetId )

	_developContentWithLocalNegativePathModelId = 
		_developContentWithLocalNegativePathModelId + 1

	local pixStreamId = 
			string.format( "developContentWithLocalNegativePathModel%s", 
							_developContentWithLocalNegativePathModelId )

	return {
		
		value = wfCatalogStream
					:mapSwitch(
						function( wfCatalog, pixelProviderProcessor )

							return Rx.Observable.create( 
								function( observer )

									-- avoid race condition subscribe to
									-- stream before we request a local
									-- negative path
									local baseSettingsWithLocalNegativePathSubscription =
										baseSettingsWithLocalNegativePathStream(
												wfCatalog,
												pixelProviderProcessor,
												pixStreamId, assetId ) 
											:subscribe( observer )

									local negPathSubscription =
										pixelProviderProcessor
											:dispatch( 
												"getLocalNegativePath", 
													'asap', pixStreamId, assetId )
											:subscribeOnCompleted(

												function()

													logger:infof( "finished getLocalNegativePath request for %s.%s",
																	pixStreamId, assetId )

												end )

									local disposePixelProviderRequest =
										Rx.Disposable.create(
											function()

												logger:infof( "Cleaning up request: %s",
																pixStreamId )

												pixelProviderProcessor
													:runUntilDone( 
														"abandonAllPixStreamRequests", 
															pixStreamId )

											end )

									return Rx.CompositeDisposable(
												baseSettingsWithLocalNegativePathSubscription,
												negPathSubscription,
												disposePixelProviderRequest )

								end )

						end )

	}
	
end

--------------------------------------------------------------------------------

local function ongoingDevelopContentsModel( wfCatalogStream )
	
	local needDevelopContentStream = Rx.Subject()

	return {
		
		value = wfCatalogStream
					:mapSwitch(
						function( wfCatalog, pixelProviderProcessor )

							return makeOngoingDevelopContentsStream( 
									"ongoingDevelopContents", wfCatalog,
									pixelProviderProcessor,
									needDevelopContentStream )

						end ),

		events = {

			setAssetIds = function( _, needsInitialValueAndUpdatesAssetIds, 
										needsJustUpdateAssetIds )

				WFArguments.check( "_", "table", "table" )

				needDevelopContentStream
					:onNext( needsInitialValueAndUpdatesAssetIds,
								needsJustUpdateAssetIds )

			end,

		}

	}
	
end

--------------------------------------------------------------------------------

local function xmpCameraRawOverlay( wfCatalog, cameraRawXmp, baseOverlay, forceExternalization, maybeCamerRawLimitOverride )
	
	local settingsLength = cameraRawXmp and string.len( cameraRawXmp ) or 0
	
	if maybeCamerRawLimitOverride and maybeCamerRawLimitOverride < INLINE_XMP_CAMERA_RAW_MAX_LENGTH then
		logger:infof( "Overriding max ACR settings length. Was %s, now %s", INLINE_XMP_CAMERA_RAW_MAX_LENGTH, maybeCamerRawLimitOverride )
		INLINE_XMP_CAMERA_RAW_MAX_LENGTH = maybeCamerRawLimitOverride
	end
	
	-- ensure baseOverlay meets the minimum requirements
	baseOverlay = baseOverlay or {}
	baseOverlay.develop = baseOverlay.develop or {}
	
	if settingsLength < INLINE_XMP_CAMERA_RAW_MAX_LENGTH and ( not forceExternalization ) then
		
		baseOverlay.develop.xmpCameraRaw = cameraRawXmp
		
		return Observable.just( baseOverlay )

	end
		
	return Observable.just( baseOverlay )
		:flatMap( function( overlay )
			
			-- too big to send to Oz inline, save to local file, 
			-- send sha256 of develop settings and let the upload 
			-- process attach the file to the asset record on Oz
		
			local settingsAsKSData = KSData.make( cameraRawXmp )
			
			return WFAsyncDigest.digestData( settingsAsKSData, 'SHA256' )
				:map( function( sha256 )
					
					local pathForSettingsXmp = wfCatalog:fileSpec():pathForSettingsXmp( sha256 )
					
					overlay.develop.xmpCameraRaw = { sha256 = sha256 }
		
					overlay._localOnly = { 
						files = {
							xmp_develop = {
								path = pathForSettingsXmp,
								fileSize = settingsLength,
								sha256 = sha256,
							}
						}
					}
					
					return overlay, pathForSettingsXmp
					
				end )
				:flatMap( function( overlayWithSha256, path )
					
					return WFAsyncFileUtils.calculatePathSize( path )
						:flatMap( function( pathSize )
							local settingsLen = KSData.length( settingsAsKSData )
							
							if pathSize == settingsLen then -- settings already exist at this location, don't resave
								return Observable.just( overlayWithSha256 )
							end
							
							return WFAsyncFileUtils.saveToFile( settingsAsKSData, path )
								:tapOnError( function( stdError, platformError )
									logger:errorf( "failed to write external develop settings to %s: %s (%s)", path, stdError, platformError )
								end )
								:catch( function()
									return Observable.just( nil )
								end )
								:map( function()
									return overlayWithSha256
								end )
						end )
				:tap( function()
					settingsAsKSData:close()
				end )
				
			end )

		end )

end

--------------------------------------------------------------------------------

local crVersion

local function developSettingsOverlay( wfCatalog, imageId, orientation, croppedWidth, croppedHeight, settingsXmp, dateChanged )
	
	if not crVersion then
		local _, _, _, crVersionString = WFCrInfo.getACRVersion()
		crVersion = crVersionString
	end
	
	local overlayWithoutXMPCameraRaw = {
		_id = imageId,
		
		develop = {
			processingModel = 'lightroom',
			crsVersion = crVersion,
			userOrientation = orientation,
			croppedWidth = croppedWidth,
			croppedHeight = croppedHeight,
			userUpdated = dateChanged,
		}
	}
	
	return xmpCameraRawOverlay( wfCatalog, settingsXmp, overlayWithoutXMPCameraRaw )
	
end

--------------------------------------------------------------------------------

local function externalizeDevelopSettings( wfCatalog, imageId, maybeMaxLength )

	return wfCatalog:docStore():docValue( imageId )
		:flatMap( function( docValue )
			if docValue._empty then
				-- doc has been deleted, ignore externalizing develop settings
				return Observable.just( true )
			end
			
			local currentSettingsXmp = derefOpt( docValue, 'develop', 'xmpCameraRaw' )
			
			if type( currentSettingsXmp ) ~= 'string' then
				-- doc already has settings externalized, no need to redo
				return Observable.just( true )
			end
			
			local baseOverlay = {
				_id = docValue._localId,
				develop = {
				}
			}
			
			return xmpCameraRawOverlay( wfCatalog, currentSettingsXmp, baseOverlay, true, maybeMaxLength )
				:flatMap( function( overlay )
					return wfCatalog:docStore():applyOverlay( overlay )
						:map( true )
				end )
		end )
	
end

--------------------------------------------------------------------------------

local function setCameraRawSettingsStream( 
					wfCatalog, profileCatalog, saveSequence, assetId,
					croppedWidth, croppedHeight, settingsXmp,
					maybeLensProfileDigest, 
					maybeCameraProfileFingerprint, --UUID
					maybeIsFallbackProfile, maybeCameraProfileDigest,
					maybeCameraProfileName )
	
	local devSettingsDate = KSDate_timeInIso8601FormatNearestSecond()

	local developSettingsOverlayStream =
			developSettingsOverlay( 
				wfCatalog, assetId, nil, croppedWidth, croppedHeight,
				settingsXmp, devSettingsDate )

	local lensProfileOverlayStream =
			NmbProfileCatalogUtils.lensProfileOverlay( 
				wfCatalog, profileCatalog, assetId, maybeLensProfileDigest )
	
	local cameraProfileOverlayStream =
			NmbProfileCatalogUtils.cameraProfileOverlay( 
				wfCatalog, profileCatalog, assetId, 
				maybeCameraProfileFingerprint,
				maybeIsFallbackProfile, maybeCameraProfileDigest )

	return Observable.combineLatest( 
				developSettingsOverlayStream, 
				lensProfileOverlayStream,
				cameraProfileOverlayStream )
			:map( 
				function( baseDevelopOverlay, lensProfileOverlay, cameraProfileOverlay )
			
					local baseOverlay = WFCatUtils.createBaseOverlayWithUserCreatedFields()
			
					local baseDevelopOverlayWithUserCreatedFields = 
						KSTableUtils.merge( 
								{},  baseOverlay, baseDevelopOverlay )
			
					local developSettingsOverlayWithLensProfile = 
							KSTableUtils.merge( 
								{},  baseDevelopOverlayWithUserCreatedFields, lensProfileOverlay )
								
					local developSettingsOverlayWithCameraAndLensProfile = 
							KSTableUtils.merge( 
								{},  developSettingsOverlayWithLensProfile, cameraProfileOverlay )
								
					
					return developSettingsOverlayWithCameraAndLensProfile
					
				end )
			:map( 
				function( settingsOverlay )
					
					local customProfileOverlay = 
						NmbProfileCatalogUtils.getCustomProfileOverlay( 
							wfCatalog:ozCatalogId(), 
							settingsOverlay, 
							maybeIsFallbackProfile, 
							maybeCameraProfileDigest,
							maybeCameraProfileName )
							
					-- In the case where a custom camera profile which does NOT yet
					-- exist on Oz (and maybe not locally yet either; e.g. pasted settings),
					-- is replacing a custom camera profile which DOES exist on Oz,
					-- we may be looking at a situation here where the profile link
					-- still in the asset payload is referring to the latter (the previously
					-- applied custom camera profile). One could argue that the profile link
					-- should point to a 'fallback' profile in this case. If that is so,
					-- then the fact that it doesn't is a bug in code that already
					-- executed upstream in this Rx chain - NmbProfileCatalogUtils.cameraProfileOverlay.
					
					-- It shouldn't matter too much whether the link left behind points to
					-- the already-on-Oz formerly-referred-to custom camera profile or
					-- and Adobe provided 'fallback' profile, as long as it gets fixed
					-- properly if/when the currently-referred-to custom camera profile
					-- becomes available on Oz. (Fixed by the Lr CC instance which uploaded
					-- the profile.)
							
					return KSTableUtils.merge( 
								{},  settingsOverlay, customProfileOverlay )
				end )
			:flatMap( 
				function( settingsOverlay )
			
					return wfCatalog:docStore():applyOverlay( settingsOverlay )
				
				end )
			:call( runUntilDone )
			:map(
				function()

					return saveSequence, assetId

				end )
		
end

--------------------------------------------------------------------------------

local function setCameraRawSettings( wfCatalogStream )

	local cameraRawSettingsStream = Rx.Subject()

	return  {

		value = wfCatalogStream
					:mapSwitch(
						function( wfCatalog, _, profileCatalog )

							return cameraRawSettingsStream
									:flatMap(
										function( saveSequence, assetId,
													croppedWidth,
													croppedHeight, xmpSettings,
													maybeLensProfileDigest,
													maybeCameraProfileFingerprint,
													maybeIsFallbackProfile, 
													maybeCameraProfileDigest,
													maybeCameraProfileName )

											logger:infof( "starting save of imageId %s", assetId )
					
											return 
												setCameraRawSettingsStream(
													wfCatalog, profileCatalog,
													saveSequence, assetId,
													croppedWidth,
													croppedHeight, xmpSettings,
													maybeLensProfileDigest,
													maybeCameraProfileFingerprint,
													maybeIsFallbackProfile, 
													maybeCameraProfileDigest,
													maybeCameraProfileName )

										end )
									:map(
										function( saveSequence, savedAssetId )

											logger:infof( "Saved settings for imageId %s with sequence", 
															savedAssetId, saveSequence )
										
											return { saveSequence, savedAssetId }

										end )

						end ),

		events = {

			save = function( _, saveSequence, assetId, croppedWidth,
								croppedHeight, xmpSettings,
								maybeLensProfileDigest,
								maybeCameraProfileFingerprint,
								maybeIsFallbackProfile,
								maybeCameraProfileDigest,
								maybeCameraProfileName )

				cameraRawSettingsStream
					:onNext( saveSequence, assetId, croppedWidth,
								croppedHeight, xmpSettings,
								maybeLensProfileDigest,
								maybeCameraProfileFingerprint,
								maybeIsFallbackProfile,
								maybeCameraProfileDigest,
								maybeCameraProfileName )

			end,
				
		}
	}

end

--==============================================================================

return {

	isUpdateWanted = isUpdateWanted,

	ongoingDevelopContentModel = ongoingDevelopContentModel,

	developContentModel = developContentModel,

	developContentWithLocalNegativePathModel = 
		developContentWithLocalNegativePathModel,

	ongoingDevelopContentsModel = ongoingDevelopContentsModel,

	developSettingsOverlay = developSettingsOverlay,
	
	externalizeDevelopSettings = externalizeDevelopSettings,

	setCameraRawSettings = setCameraRawSettings,

}
