--[[----------------------------------------------------------------------------

WFCatErrorHandler.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	--Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

local KSDate = require 'KSDate'
local KSJsonDecoder = require 'KSJsonDecoder'
local KSIdGenerator = require 'KSIdGenerator'
local KSStringUtils = require 'KSStringUtils'
local KSTableUtils = require 'KSTableUtils'
local WFAnalytics = require 'WFAnalytics'
local WFArguments = require 'WFArguments'
local WFCatOzErrors = require 'WFCatOzErrors'
local WFCloseSupport = require 'WFCloseSupport'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local WFLogger = require 'WFLogger'
local WFOzDocUtils = require 'WFOzDocUtils'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'
local WFTableArchiving = require 'WFTableArchiving'
local WFURI = require 'WFURI'

--------------------------------------------------------------------------------

local logger = WFLogger 'WFCatErrorHandler'

--------------------------------------------------------------------------------

local derefOpt = KSTableUtils.derefOpt
local check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local function errors_table( )

	return {

		sql = [[CREATE TABLE errors( type )]],
		
		docValueFilter = function( docValue )
			return docValue.type == 'error'
		end,
	
		map = function( docValue, emit )
			
			if docValue.type == 'error' and docValue._deleted ~= true then
				
				local isUnknownError = derefOpt( docValue, '_localOnly', 'unknownError' )
				
				if not isUnknownError then
					emit( docValue.subtype )
				else
					
					if docValue.subtype == WFCatOzErrors.Errors.kStorageExceededError then
						emit( docValue.subtype )
					end
					
					-- Here is where we try to look at previously unknown errors and see if it's a type
					-- that we now know about.  If so, we should be able to emit to the errors
					-- table and handle it.
					if docValue.subtype == WFCatOzErrors.Errors.kInputValidationError then
					
						local previouslyFailedHandling = derefOpt( docValue, '_localOnly', 'localError', WFCatOzErrors.Errors.kInputValidationError, 'handlingFailed' )
						
						-- missing 'download' field? there was a short time where Lightroom CC wasn't sending 
						-- this correctly in some cases.
						local wasMissingDownloadValue = derefOpt( docValue, '_serverProvidedMetadata', 'errors', 'download', 1 )
						
						if wasMissingDownloadValue and not previouslyFailedHandling then
							emit( docValue.subtype )
							return
						end
						
						-- missing 'if_match'? there appear to be some cases where we are missing the server provided metadata
						-- and the revision_ids and so while we send an update, we don't send an if_match revision.  Not yet
						-- sure why this occurs, but it appears to happen mainly to deleted assets. See NMB-6190
						local wasIfMatchMissing = derefOpt( docValue, '_serverProvidedMetadata', 'errors', 'if_match', 1 )
						if wasIfMatchMissing == 'is required' and not previouslyFailedHandling then
							emit( docValue.subtype )
							return
						end
					end
					
					-- NMB-6352: Any errors previously marked unknown that have a status code of 5XX should be automatically 
					-- retried by the default handler which now should handle these appropriately
					local maybeStatusCode = tonumber( derefOpt( docValue, '_serverProvidedMetadata', 'status' ) )
					if maybeStatusCode and ( 500 <= maybeStatusCode and maybeStatusCode < 600 ) then
						emit( docValue.subtype )
						return
					end
					
				end
			end
		end,

		hasOneRowPerDocument = true,
	}
	
end

--------------------------------------------------------------------------------

local function blocked_uploads_table( )
	
	return {
	
		sql = [[CREATE TABLE blocked_uploads( reason )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		map = function( docValue, emit )
			
			if WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue ) and not docValue._deleted then
				emit( WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError )				
			end
			
		end,
	}

end

--------------------------------------------------------------------------------

local function generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )

	indexStoreSchema:addTable( errors_table( ) )

	indexStoreSchema:addIndex [[CREATE INDEX error_type ON errors( type )]]
	
	indexStoreSchema:addTable( blocked_uploads_table( ) )
	
	indexStoreSchema:addIndex [[CREATE INDEX blocked_uploads_reason ON blocked_uploads( reason )]]

end

--------------------------------------------------------------------------------

local function getErrorHandler( self, errSubtype )

	return self._errorDocumentHandlerFns[ errSubtype ]
	
end

--------------------------------------------------------------------------------

local function isDependentDocumentType( docType )

	return docType == 'album_asset' or docType == 'stack_asset'
	
end

--------------------------------------------------------------------------------

local function mimicOzDeletionEvent( docStore, docId )
	
	-- Rather than just 'deleteDoc' we create a deletion document that looks like the 
	-- deletion we would get from Oz in the sync feed.  This ensures that we clean up the
	-- document in the same way we would if we saw the deletion via the sync feed.
	local deletionDoc = {
			_id = docId,
			_deleted = true,
			_revIds = {
				KSIdGenerator.uuidInLowercaseHex(), -- Oz does not send revision ids on deletions
				force = true,
			},
			_serverProvidedMetadata = {
				deleted_by = 'WFCatErrorHandler'
			}
		}
	
	return docStore:bulkSetDocValue( { label = 'oz', deletionDoc } )

end

--------------------------------------------------------------------------------

local function getOrCreateSubtablesForPath( overlay, path )

	local sub = overlay
	
	for i, v in ipairs( path ) do
	
		local d = sub[ v ]
		if not d then
			d = {}
			sub[ v ] = d
		end
		
		sub = d
		
	end
	
	return overlay, sub
	
end

--------------------------------------------------------------------------------

local function createValueAtPath( overlay, path, value )

	local key = path[ #path ]

	table.remove( path, #path )

	local sub
	overlay, sub = getOrCreateSubtablesForPath( overlay, path )
	
	sub[ key ] = value
	
	
	return overlay

end

--------------------------------------------------------------------------------

local function removeOffendingKey( overlay, docPath )
					
						
	-- INFO [pkamin 2016-07-25]: validate errorCode for correct handling
	-- ( disallowed, invalid, required )
	-- Right now we delete the value to unblock sync which is very aggressive.
	-- We might be able to act smarter if the logic can be driven by the Data Model 
	-- Schema JSON - escpecially for the `required` error code
		
	createValueAtPath( overlay, docPath, { _delete = true } )
	
end

--------------------------------------------------------------------------------

local function fixTimestamp( overlay, docPath, value, errorCode )
	
	local date = KSDate.recoverDateForOz( value )
	if not date then
		date = KSDate.timeInIso8601Format( nil, false )
	end
	
	createValueAtPath( overlay, docPath, date )

end

--------------------------------------------------------------------------------

local function fixCaptureDate( overlay, docPath, value, errorCode )
	
	local date = KSDate.recoverDateForOz( value )
	if not date then
		date = "0000-00-00T00:00:00"
	end
	
	createValueAtPath( overlay, docPath, date )

end

--------------------------------------------------------------------------------

local function fixTruncate( overlay, docPath, value, errorCode )

	-- if it's a number, truncate it, otherwise, remove the field
	if type( value ) == 'number' then
		createValueAtPath( overlay, docPath, math.floor( value ) )		
	else
		createValueAtPath( overlay, docPath, { _delete = true } )
	end

end

--------------------------------------------------------------------------------

local function fixFaceDimension( overlay, docPath, value, errorCode )
	
	if not value or value < 0 then
		createValueAtPath( overlay, docPath, 0 )
	elseif value > 1 then
		createValueAtPath( overlay, docPath, 1 )
	end
	
end

--------------------------------------------------------------------------------

local function fixSpaceDocument( catalog, docValue, errorKey, errorValue )
	
	local overlay = {}
	
	if errorKey == 'download' and docValue.download == nil then
		overlay.download = false
	end
	
	if next( overlay ) then
		overlay._id = docValue._id
		return catalog:docStore():applyOverlay( overlay )
	else
		return Observable.just( true )
	end
	
end

--------------------------------------------------------------------------------

local pathComponentBySubtype = {
	image = 'assets',
	video = 'assets',
}

--------------------------------------------------------------------------------

local function fixIfMatchIssue( catalog, docValue, errorKey, errorValue )
	
	-- happy path for fixing this error assumes internet access
	local docId = docValue._id
	
	-- get link from docValue's self href if it's there
	-- otherwise construct the url manually
	local documentUrl = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', 'self', 'href' )
	if not documentUrl then
		local href = catalog:ozCatalogSpec():catalogHref()
		
		local pathComponent = pathComponentBySubtype[ docValue.subtype ] 
		if pathComponent then
			href = WFURI.stringByAppendingPathComponents( href, pathComponent )
			href = WFURI.stringByAppendingPathComponents( href, docId )
			documentUrl = href
		end
		
	end
	
	if not documentUrl then
		logger:warnf( "No URL, unable to handle InputValidationError for %s", docId )
		return Observable.just() 
	end
	
	return catalog:ozCatalogSpec():get( documentUrl, { returnHttpErrorsAsData = true } )
		:flatMap( function( result, headers )
			if headers.status_code == '200' then
				-- convert JSON document and insert from Oz
				local decodedDoc = KSJsonDecoder.decodeWebSafeJson_stripNull( result )
				
				local docValue = WFOzDocUtils.remapBasicOzDocument( decodedDoc, decodedDoc.base )
							
				if docValue and not docValue._deleted then
					return catalog:docStore():bulkSetDocValue( { label = 'oz',
																docValue } )
				end
				
			elseif headers.status_code == '404' then
				
				-- consider this a deletion from Oz
				return mimicOzDeletionEvent( catalog:docStore(), docId )
							
			end
			
			return Observable.just( )
			
		end )
		:catch( Observable.just( ) )
		
	
end

--------------------------------------------------------------------------------

-- these functions just add to an overlay
local payloadValidationFixups = {
	-- fixups for asset documents
	[ 'payload.importSource.importTimestamp' ] = fixTimestamp,
	[ 'payload.captureDate' ] = fixCaptureDate,
	[ 'payload.develop.userUpdated' ] = fixTimestamp,
	[ 'payload.importSource.originalWidth' ] = fixTruncate,
	[ 'payload.importSource.originalHeight' ] = fixTruncate,
	[ 'payload.xmp.mwg-rs.Regions.RegionList.uuid.Area.x' ] = fixFaceDimension,
	[ 'payload.xmp.mwg-rs.Regions.RegionList.uuid.Area.y' ] = fixFaceDimension,
}

local exifPayloadValidationFields = {
	[ 'FNumber' ] = true,
	[ 'ApertureValue' ] = true,
	[ 'ExposureTime' ] = true,
	[ 'FocalLength' ] = true,
}

-- these functions must return an Observable that fixes the document
local inputValidationFixups = {
	-- fixups for space documents
	[ 'download' ] = fixSpaceDocument,
	[ 'if_match' ] = fixIfMatchIssue,
}

--------------------------------------------------------------------------------

local function maybeGetFixFunctionForKey( key )
	
	-- some keys are at variable places in the data model 
	-- e.g. payload.xmp.mwg-rs.Regions.RegionList.0185a47f32354e3d9cf56634a8b20109.Area.x
	-- so we need to possible replace it before looking up the fixFn
	local canonicalKey = string.gsub( key, 'RegionList%..*%.Area', 'RegionList.uuid.Area' )
	
	canonicalKey = canonicalKey or key
		
	return payloadValidationFixups[ canonicalKey ]
	
end

--------------------------------------------------------------------------------

local function parseDocPathFromKey( key, docValueThatIncurredError )
	
	local docPath = KSStringUtils.split( key, '%.' )
						
	if #docPath < 1 then
		return nil
	end

	if #docPath == 1 then
		-- Something below payload level...this generally makes no sense and is not fixable, but it may make sense 
		-- for some doc types (spaces) so if we *know* it's fixable, use it, otherwise don't consider the docPath 
		-- fixable.  Add future cases to this if
		if docValueThatIncurredError.type == 'space' then
			return docPath
		else 
			return nil
		end
	end
	
	-- remove 'payload' level
	table.remove( docPath, 1 )
	
	local actualKey = docPath[ #docPath ]
	
	-- remove final key from path (we'll add it back)
	table.remove( docPath, #docPath )
	
	-- If key is a number, then we should probably do the following:
	-- Check to make sure it being considered a string isn't correct by seeing
	-- whether the docValue has something at that key. 
	-- If so, assume it should be treated as a string for the delete
	-- If not, treat it as a number index into an array but we need to add 1 
	-- since Oz indexes are 0 based and lua is 1 based
	local maybeNumber = tonumber( actualKey )
	if maybeNumber then
		
		local valueAtStringPath 
		if #docPath > 0 then 
			valueAtStringPath = derefOpt( docValueThatIncurredError, unpack( docPath ), actualKey )
		else 
			valueAtStringPath = derefOpt( docValueThatIncurredError, actualKey )
		end
									
		if valueAtStringPath then
			-- it looks like a number, but it quacks like a string so treat it as such
			table.insert( docPath, actualKey )
		
		else
			-- it looks like a number and it doesn't quack like a string so treat it as a number
			-- (but add 1 because our arrays are 1 based)
			local parentOfActualKey = docPath[ #docPath ] -- parent key is same as #docPath because last value was removed previously.

			if not exifPayloadValidationFields [ parentOfActualKey ] then
				table.insert( docPath, maybeNumber + 1 )
			end

		end
		
	else
		-- not even close to a number
		table.insert( docPath, actualKey )
		
	end
	
	return docPath
	
end

--------------------------------------------------------------------------------

local function deleteErrorDoc( docStore, errorDocId )
	return docStore:deleteDoc( errorDocId )
end

--------------------------------------------------------------------------------

local function markErrorAsUnknown( docStore, errorDocId )
	return docStore:applyOverlay {
			_id = errorDocId,
			_localOnly = {
				unknownError = true,
			}
		}
end

--------------------------------------------------------------------------------

local function lastRetryIsOlderThanXHours( maybeDate, hours )
	
	if not maybeDate then
		return true
	end
	
	local olderThanInSeconds = hours * 60 * 60
	local now = KSDate.timeInIso8601Format()
	
	local differenceInSeconds = KSDate.differenceInSeconds( maybeDate, now )
	
	return differenceInSeconds > olderThanInSeconds
end

--------------------------------------------------------------------------------

local function defaultHandler( self, errorDoc, docValue )
	
	local docStore = self._catalog:docStore()
	
	local stringOfErrorDoc = WFTableArchiving.tableToString( errorDoc )
	
 	local spm =  derefOpt( errorDoc, '_serverProvidedMetadata' )
	local statusCode = tonumber( derefOpt( spm, 'status' ) )
 	local idOfDocThatIncurredError = derefOpt( spm, 'source', 'id' )
 	local actionOnDocThatIncurredError = derefOpt( spm, 'source', 'action' )
 
 	local errorEvent = {}
 	errorEvent['event.type'] = 'error'
 	errorEvent['event.workflow'] = 'SYNC'
 	errorEvent['event.subcategory'] = 'error_handler'
 	errorEvent['event.error_type'] = tostring( errorDoc.subtype )
 	errorEvent['event.error_desc'] = stringOfErrorDoc
	errorEvent['event.error_code'] = statusCode
 	errorEvent['content.id'] = idOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError
	
 	WFAnalytics.trackEvent( errorEvent )
		
	-- on 5XX errors - regarless of the error type, we should be able to retry the document 
	-- update on Oz. So annotate the document with some error info which will also update the
	-- sequence and cause it to try to be resent	
	if statusCode and ( 500 <= statusCode and statusCode < 600 ) then
		
		local existingRetries = tonumber( derefOpt( docValue, '_localOnly', 'ozError', 'retries' ) ) or 0
		local maybeLastRetry = derefOpt( docValue, '_localOnly', 'ozError', 'lastRetry' )
		
		local maybeOverlayObservable
		-- if we've already retried 3 times in the last hour, don't do it again as it seems likely 
		-- to not really be a retryable error and instead something more persistent. a manual change
		-- may be required.
		if existingRetries < 3 or lastRetryIsOlderThanXHours( maybeLastRetry, 1 ) then
			if existingRetries >= 3 then
				-- if we'd already hit 3 retries from a previous series of errors then we got
				-- here because the last retry was old, so reset the retry count and try again
				existingRetries = 0
			end
			
			maybeOverlayObservable = docStore:applyOverlay {
					_id = docValue._id,
					_localOnly = {
						ozError = {
							subtype = errorDoc.subtype,
							status = statusCode,
							retries = existingRetries + 1,
							lastRetry = KSDate.timeInIso8601Format()
						}
					}
				}
		else
			maybeOverlayObservable = Observable.just()
		end
		
		return maybeOverlayObservable
			:catch( Observable.just() )
			:flatMap( function()
				return deleteErrorDoc( docStore, errorDoc._id)
			end )
	else
		-- Log an error that we received an error document from Oz that we are not sure how to handle.
		logger:errorf( "Unhandled error document type %s: %s", tostring( errorDoc.subtype ), WFTableArchiving.tableToString( errorDoc ) )
	
		-- mark it so we're not stuck on it.
		return markErrorAsUnknown( docStore, errorDoc._id )
	
	end	
end

--------------------------------------------------------------------------------

local function payloadValidationHandler( self, errorDoc, docValueThatIncurredError )
	
	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	local spm =  derefOpt( errorDoc, '_serverProvidedMetadata' )
	local resourceId = derefOpt( spm, 'source', 'id' )
	local resourceRevId = derefOpt( spm, 'source', 'revision_id' )
	local typeOfDocThatIncurredError = derefOpt( spm, 'source', 'type' )
	local actionOnDocThatIncurredError = derefOpt( spm, 'source', 'action' )
	
	local errors = derefOpt( spm, 'errors' )
	
	logger:warnf( "payload validation error for %s revId: %s", resourceId, resourceRevId )
	
	local overlay = {}
	
	local hasUnknownError = false
	
	if errors and type( errors ) == 'table' then
		
		for key, errorCode in pairs( errors ) do
			
			-- errorCode comes from Oz as an array of a single item - the actual errorCode
			errorCode = type( errorCode ) == 'table' and derefOpt( errorCode, 1 ) or errorCode
			
			logger:warnf( "payload validation error on '%s' for %s revId: %s", key, resourceId, resourceRevId )
			
			local docPath = parseDocPathFromKey( key, docValueThatIncurredError )
			
			if docPath then
				
				local existingValue = derefOpt( docValueThatIncurredError, unpack( docPath ) )
				
				local fixFn = maybeGetFixFunctionForKey( key )
				if fixFn then
				
					fixFn( overlay, docPath, existingValue, errorCode )
					
				else
					
					if errorCode ~= 'required' then 
						-- if it's required and we remove it again, we've created a looper, so don't
						removeOffendingKey( overlay, docPath )
					end
					
				end
			
			else
				
				hasUnknownError = true
				
			end
			
		end
		
	end
	
	local errorEvent = {}
			
	local interestingSourceDocSections = KSTableUtils.filter( 
		function( key, value )
			return key == 'importSource'
		end, docValueThatIncurredError )
	
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = 'PayloadValidationError'
	errorEvent['event.error_desc'] = WFTableArchiving.tableToString( errorDoc )
	errorEvent['event.error_code' ] = hasUnknownError and 'UnknownError'
	errorEvent['event.value'] = WFTableArchiving.tableToString( overlay )
	
	errorEvent['content.id'] = resourceId
	errorEvent['content.revision_id'] = resourceRevId
	errorEvent['content.type'] = typeOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError
	errorEvent['content.value'] = WFTableArchiving.tableToString( interestingSourceDocSections )
	
	WFAnalytics.trackEvent( errorEvent )
	
	local handlerObservable
	
	if hasUnknownError and not next( overlay ) then 
		-- there's nothing we fixed and at least one error seen was unknown
		return markErrorAsUnknown( docStore, errorDoc._id )
	end
	
	if next( overlay ) then
		overlay._id = resourceId
		handlerObservable = docStore:applyOverlay( overlay )
	else
		handlerObservable = Rx.Observable.just()
	end
	
	return handlerObservable
		:catch( Observable.just() )
		:flatMap( function()
			-- delete error doc once we've attempted to fix the resource
			return deleteErrorDoc( docStore, errorDoc._localId )
				
		end )
		
end

--------------------------------------------------------------------------------

local function inputValidationHandler( self, errorDoc, docValue ) 

	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	local spm =  derefOpt( errorDoc, '_serverProvidedMetadata' )
	local resourceId = derefOpt( spm, 'source', 'id' )
	local resourceRevId = derefOpt( spm, 'source', 'revision_id' )
	local typeOfDocThatIncurredError = derefOpt( spm, 'source', 'type' )
	local actionOnDocThatIncurredError = derefOpt( spm, 'source', 'action' )
	
	local errors = derefOpt( spm, 'errors' )
	local errorKey, errorValue
	if errors and type( errors ) == 'table' then
		
		-- the input validation table is in an odd form like 
		-- errors = {
		--     download = {
		--         "must be true or false" 
		--     }
		-- } 
		-- so parse the last level like an array
		errorKey = next( errors )
		if errorKey then
			errorValue = derefOpt( errors, errorKey, 1 )
		end
		
	end
	
	logger:warnf( "InputValidationError for %s revId: %s", resourceId, resourceRevId )
	
	return Observable.just( docValue )
		
		:flatMap( function( docValueThatIncurredError )
			
			if errorKey and inputValidationFixups[ errorKey ] then
				
				local fixFn = inputValidationFixups[ errorKey ]
						
				return fixFn( catalog, docValueThatIncurredError, errorKey, errorValue )
						:map( function()
							return true
						end )				
			else
				-- this error is for us, but we don't know what to do with it
				-- mark the error document as unable to be handled right now so 
				-- we don't try to process it anymore
				return self._catalog:docStore():applyOverlay( {
					_id = errorDoc._id,
					_localOnly = {
						localError = {
							[ errorDoc.subtype ] = {
								handlingFailed = true
							}
						},
					}
				} )
				:map( function()
					-- don't delete it, we may be able to process in the future
					return false 
				end )
				
			end
			
		end )
		:flatMap( function( wasHandled )
			if wasHandled then
			
				-- fire an analytics event that we should be able to fix the error in the document
				local errorEvent = {}
				errorEvent['event.type'] = 'error'
				errorEvent['event.workflow'] = 'SYNC'
				errorEvent['event.subcategory'] = 'error_handler'
				errorEvent['event.error_type'] = WFCatOzErrors.Errors.kInputValidationError
				errorEvent['event.error_desc'] = errorValue
				errorEvent['event.error_code'] = 'fixed'
				
				errorEvent['content.id'] = resourceId
				errorEvent['content.revision_id'] = resourceRevId
				errorEvent['content.type'] = typeOfDocThatIncurredError
				errorEvent['content.action'] = actionOnDocThatIncurredError
				errorEvent['content.url'] = errorKey
				
				-- track the error event
				WFAnalytics.trackEvent( errorEvent )
												
				-- delete error doc once we've marked the resource as storage exceeded
				return deleteErrorDoc( docStore, errorDoc._localId )
			else
				-- if we don't know how to handle a specific InputValidationError, use the default handler to mark it as 
				-- unhandled rather than deleting it as a future update might be able to help here.
				return defaultHandler( self, errorDoc, docValue )
			end
		end )
		
end

--------------------------------------------------------------------------------

local function resourceExistsHandler( self, errorDoc, docThatIncurredError )

	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	local spmSource =  derefOpt( errorDoc, '_serverProvidedMetadata', 'source' )
	
	local idOfDocThatIncurredError = derefOpt( spmSource, 'id' )
	local typeOfDocThatIncurredError = derefOpt( spmSource, 'type' )
	local actionOnDocThatIncurredError = derefOpt( spmSource, 'action' )
	
	local documentUrl = derefOpt( spmSource, 'links', 'self', 'href' )

	local eventErrorCode

	local handleErrorDoc
	if not documentUrl then
									
		logger:warnf( "No document url for ResourceExistsError for resource id of %s, ignoring error doc", idOfDocThatIncurredError )
		handleErrorDoc = Observable.just()
		
	else 
	
		-- let's handle the happy path...we're online...so check whether Oz has the asset by
		-- directly asking for it. If it does, we can ignore the ResourceExistsError - sync
		-- should resolve this eventually. If it does not, then it's a deleted asset and so 
		-- we should delete it locally
		handleErrorDoc = catalog:ozCatalogSpec():get( documentUrl )
			:flatMap( function( result, headers )
									
				if headers.status_code == '200' then
					
					local ok, resultJson = pcall( KSJsonDecoder.decodeWebSafeJson, result )
					resultJson = ok and resultJson or {}
					
					local isDeleted = derefOpt( resultJson, 'deleted' ) == true
					
					if docThatIncurredError.type == 'asset' and isDeleted then
						
						if not self._deleteAssetsFn then
							logger:errorf( "Setup error: no deleteAssets implementation" )
							
							return Observable.just()
						end

						-- if it's an asset document, we need to clean up any associated local files
						-- NOTE: We use concat here because deleteAssets impls are provided by the client and we
						-- can't assume that it will give us only a single event. So take only the completion with 
						-- takeLast( 0 ) and then concat with Observable.just so we do get a single value for our
						-- purposes
						eventErrorCode = 'deleted'
						return Observable.concat( 
								self._deleteAssetsFn( catalog, { idOfDocThatIncurredError } ):takeLast( 0 ),
								Observable.just() )
					
					elseif docThatIncurredError.type == 'album' and isDeleted then
						
						if not self._deleteIndividualAlbumFn then
							logger:errorf( "Setup error: no deleteIndividualAlbum implementation" )
						
							return Observable.just()
						end
						
						-- if it's an album we need to clean up and associated album_assets)
						-- NOTE: We use concat here because deleteIndividualAlbum impls are provided by the 
						-- client and we can't assume that it will give us only a single event. So take only 
						-- the completion with takeLast( 0 ) and then concat with Observable.just so we do 
						-- get a single value for our purposes
						eventErrorCode = 'deleted'
						return Observable.concat( 
								self._deleteIndividualAlbumFn(  catalog, idOfDocThatIncurredError ):takeLast( 0 ),
								Observable.just() )
					
					elseif docThatIncurredError.type == 'album_asset' and ( isDeleted or resultJson._id ~= idOfDocThatIncurredError ) then
					
						
						-- album_assets may get a ResourceExistsError if the same album_asset association is
						-- made with a different album_asset_id than is already being used. so also delete the 
						-- album_asset doc in that case
						if resultJson._id ~= idOfDocThatIncurredError then 
							eventErrorCode = 'duplicate'
						else
							eventErrorCode = 'deleted'
						end
						
						return docStore:deleteDoc( idOfDocThatIncurredError )
						
					elseif isDeleted then
						
						-- other doc types just delete the doc itself
						eventErrorCode = 'deleted'
						return docStore:deleteDoc( idOfDocThatIncurredError )
						
					else
						
						eventErrorCode = 'missing confirmation'
						
						-- don't alter the existing document based on the error doc received, trust the sync process
						-- to resolve the issue that caused the ResourceExistsError
						return Observable.just( KSJsonDecoder.decodeWebSafeJson_stripNull( result ) )
							:map( function( decodedDocumentJson )
								
								local docValue = WFOzDocUtils.remapBasicOzDocument( decodedDocumentJson, decodedDocumentJson.base )
								
								return docValue						
							end )
							:flatMap( function( docValue )
								
								if not docValue or docValue._deleted then
									return Observable.just()
								end
								
								return self._catalog:docStore():bulkSetDocValue( { label = 'oz',
                                                                                     docValue } )
							end )
							
					end
				end
				
				return Observable.just( false )
				
			end )
	end
	
	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = WFCatOzErrors.Errors.kResourceExistsError
	errorEvent['event.error_code'] = eventErrorCode
	errorEvent['content.id'] = idOfDocThatIncurredError
	errorEvent['content.type'] = typeOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError

	-- track the error event
	WFAnalytics.trackEvent( errorEvent )
	
	return handleErrorDoc
		:catch( Observable.just() )
		:flatMap( function()
			-- delete the error doc once we've handled the error document
			return deleteErrorDoc( docStore, errorDoc._localId )
		end )

end

--------------------------------------------------------------------------------

local function inlineSettingsHandler( self, errorDoc, docValue )
	
	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	local maybeLimit = derefOpt( errorDoc, '_serverProvidedMetadata', 'limit' )
	
	local spmSource =  derefOpt( errorDoc, '_serverProvidedMetadata', 'source' )	
	local idOfDocThatIncurredError = derefOpt( spmSource, 'id' )
	local typeOfDocThatIncurredError = derefOpt( spmSource, 'type' )
	local actionOnDocThatIncurredError = derefOpt( spmSource, 'action' )

	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = WFCatOzErrors.Errors.kInlineXmpDevelopSettingTooBigError
	
	errorEvent['content.id'] = idOfDocThatIncurredError
	errorEvent['content.type'] = typeOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError
	
	WFAnalytics.trackEvent( errorEvent )
	
	return self._externalizeSettingsFn( catalog, idOfDocThatIncurredError, tonumber( maybeLimit ) )
		:catch( Observable.just() )
		:flatMap( function()
			-- delete the error doc once we've handled the error document
			return deleteErrorDoc( docStore, errorDoc._localId )
		end )
end

--------------------------------------------------------------------------------

local function resourceNotFoundHandler( self, errorDoc, docValue )
	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	local spm =  derefOpt( errorDoc, '_serverProvidedMetadata' )
	local idOfDocThatIncurredError = derefOpt( spm, 'source', 'id' )
	local typeOfDocThatIncurredError = derefOpt( spm, 'source', 'type' )
	local actionOnDocThatIncurredError = derefOpt( spm, 'source', 'action' )
	
	local missingResourceId = derefOpt( spm, 'missing_resource_id' ) or idOfDocThatIncurredError
	logger:infof( "received resource not found error for %s", missingResourceId )
	
	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = WFCatOzErrors.Errors.kResourceNotFoundError
	
	errorEvent['content.id'] = idOfDocThatIncurredError
	errorEvent['content.type'] = typeOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError
	errorEvent['content.missing_resource_id'] = missingResourceId

	WFAnalytics.trackEvent( errorEvent )

	return docStore:docValue( missingResourceId )
		:flatMap( function( missingResourceDocValue )
			
			
			if idOfDocThatIncurredError == missingResourceId then
			
				-- Is the missing resource id the same as the id of the document we were trying to send?
				-- if so, then we either have a create/update to a doc that doesn't exist on Oz yet
				-- (ignore the error, it should clear up eventually) or an update to a doc that doesn't
				-- exist on Oz any more (remove the document locally)
			
				local didPreviouslyExistOnOz = not not derefOpt( missingResourceDocValue, '_serverProvidedMetadata', 'created' )
								
				if missingResourceDocValue._empty or didPreviouslyExistOnOz then
					
					return mimicOzDeletionEvent( docStore, missingResourceId )
						
				end
			
			else 
			
				-- If the missing resource id is not the same as the id of the document we were trying to send, 
				-- then we are looking at doc that incurred the error as a dependent document of the missing 
				-- resource. In that case we'll need to determine why the missing resource is missing and 
				-- mark the dependent document with the same error message. This will allow us to clear the
				-- dependent documents error at the same time as it's 'parent' resource
				
				if missingResourceDocValue._empty and isDependentDocumentType( typeOfDocThatIncurredError ) then
				
					-- the missing resource is missing becaue it's deleted. so we should delete the dependent
					-- document too
					return docStore:deleteDoc( idOfDocThatIncurredError )
					
				else
				
					local ozErrorForMissingResource = WFCatOzErrors.maybeGetOzErrorHash( missingResourceDocValue )
					
					local errorOverlay
					if ozErrorForMissingResource then 
						
						errorOverlay = {
								_id = idOfDocThatIncurredError,
								_localOnly = { 
									ozError = ozErrorForMissingResource
								}
							}
								
					else 
						
						errorOverlay = WFCatOzErrors.createOverlayForOzErrorSubtype( idOfDocThatIncurredError, errorDoc.subtype )
						
					end
					
					return docStore:applyOverlay( errorOverlay )
				
				end				
			
			end
			
			-- If we receive a resource not found error for other reasons than those handled explicitly we are 
			-- hopefully in a situation that should clear itself up. Log that we received this error, but there
			-- is not much more we should do to handle the ResourceNotFound error locally.
			logger:warnf( "ignoring %s error from Oz for %s", errorDoc.subtype, idOfDocThatIncurredError )
			
			return Observable.just( true )
			
		end )
		:finally( function()
			-- delete error document once we've dealt with it
			return docStore:deleteDoc( errorDoc._localId )
		end )
		
end

--------------------------------------------------------------------------------

local function storageExceededHandler( self, errorDoc, storageExceededDoc )
	 
	local resourceId = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'id' )
	local typeOfDocThatIncurredError = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'type' )
	local actionOnDocThatIncurredError = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'action' )
	
	logger:infof( "received storage exceeded error for %s", resourceId )
	
	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = WFCatOzErrors.Errors.kStorageExceededError
	
	errorEvent['content.id'] = resourceId
	errorEvent['content.type'] = typeOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError
	
	WFAnalytics.trackEvent( errorEvent )
	
	local errorOverlay = WFCatOzErrors.createOverlayForOzErrorSubtype( resourceId, errorDoc.subtype )
	
	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	return docStore:applyOverlay( errorOverlay )
		:catch( Observable.just() )
		:flatMap( function()
			-- delete error doc once we've marked the resource as storage exceeded
			return deleteErrorDoc( docStore, errorDoc._localId )
		end )
	
end

--------------------------------------------------------------------------------

local function multipartCompleteHandler( self, errorDoc, multipartCompleteErrorDoc )
	
	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	local resourceId = KSTableUtils.derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'asset_id' )
	
	if not resourceId then
		resourceId = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'id' )
	end
	
	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = WFCatOzErrors.Errors.kMultipartCompleteError
	
	errorEvent['content.id'] = resourceId

	WFAnalytics.trackEvent( errorEvent )
	
	if not multipartCompleteErrorDoc._empty then
	
		logger:infof( "received multipart complete error for %s", resourceId )
		
		local errorOverlay = WFCatOzErrors.createOverlayForOzErrorSubtype( resourceId, errorDoc.subtype )
		
		return docStore:applyOverlay( errorOverlay )
			:catch( Observable.just() )
			:flatMap( function()
				-- delete error doc once we've marked the resource as multipart error
				return deleteErrorDoc( docStore, errorDoc._localId )
			end )

	end
	
	return Observable.just()
		
end

--------------------------------------------------------------------------------

local function trialOrSubscriptionExpiredHandler( self, errorDoc, subscriptionExpiredDoc )

	local resourceId = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'id' )
	local typeOfDocThatIncurredError = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'type' )
	local actionOnDocThatIncurredError = derefOpt( errorDoc, '_serverProvidedMetadata', 'source', 'action' )
	
	logger:infof( "received trial/subscription expired error for %s", resourceId )
	
	local errorEvent = {}
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'error_handler'
	errorEvent['event.error_type'] = WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError
	
	errorEvent['content.id'] = resourceId
	errorEvent['content.type'] = typeOfDocThatIncurredError
	errorEvent['content.action'] = actionOnDocThatIncurredError
	
	WFAnalytics.trackEvent( errorEvent )
	
	local errorOverlay = WFCatOzErrors.createOverlayForOzErrorSubtype( resourceId, errorDoc.subtype )
	
	local catalog = self._catalog
	local docStore = catalog:docStore()
	
	return docStore:applyOverlay( errorOverlay )
		:catch( Observable.just() )
		:flatMap( function()
			-- delete error doc once we've marked the resource as trial/sub expired
			return deleteErrorDoc( docStore, errorDoc._localId )
		end )

end

--------------------------------------------------------------------------------

local function addErrorDocToDiagnosticLog( self, errorDoc )
	
	-- for each error subtype record
	-- subtype
	--    * subtype count
	--    * status
	--    * source type (e.g. asset)
	--    * source action (e.g. create, update)
	--    * errors section (parse differently for most errors and PVE)
	--    * source ids (for easier troubleshooting)???

	local spm = derefOpt( errorDoc, '_serverProvidedMetadata' )
	local sourceDocType = derefOpt( spm, 'source', 'type' )
	local action = derefOpt( spm, 'source', 'action' )
	local status = derefOpt( spm, 'status' )
	local sourceDocId = derefOpt( spm, 'source', 'id' )
	
	local subtype = errorDoc.subtype
	local summaryForSubtype = self._errorDocSummary[ subtype ] or {}
	
	-- increment the count for all documents which hit this error subtype
	summaryForSubtype.count = ( summaryForSubtype.count or 0 ) + 1
	
	-- get the errors that this error doc represents (may be multiple)
	local errorDetails = {}
	local errors = derefOpt( spm, 'errors' )
	for err, detailTable in pairs( errors ) do
		errorDetails[ #errorDetails + 1 ] = string.format( '%s:%s', err, detailTable[ 1 ] ) 
	end
	
	if not next( errorDetails ) then
		-- if, for some reason, there's no errors table then provide a default message so that
		-- it's still logged
		errorDetails[ #errorDetails + 1 ] = 'missing error info' 
	end
	
	for _, errorDetail in ipairs( errorDetails ) do
		
		local errorKey = string.format( '%s:%s:%s:%s', sourceDocType, action, status, errorDetail )
	
		local errorSummary = summaryForSubtype[ errorKey ] or {}
		
		errorSummary.count = ( errorSummary.count or 0 ) + 1
		
		errorSummary.sourceDocIds = errorSummary.sourceDocIds or {}
		
		errorSummary.sourceDocIds[ #errorSummary.sourceDocIds + 1 ] = sourceDocId
		
		summaryForSubtype[ errorKey ] = errorSummary
	
	end
	
	self._errorDocSummary[ subtype ] = summaryForSubtype
	
end

--------------------------------------------------------------------------------

local function processOneErrorDocument( self, errorDocId )
	
	local docStore = self._catalog:docStore()
	
	return self._catalog:docStore():docValue( errorDocId )
		:flatMap( function( errorDoc )

			if errorDoc._empty then
				return Observable.just()
			end
			
			-- all error documents should have the id of the document that caused the error
			-- make sure we have that document before sending it to a handler since clients
			-- can receive error documents generated by other ecosystem clients
			local spm =  derefOpt( errorDoc, '_serverProvidedMetadata', 'source' )
			local idOfDocThatIncurredError = derefOpt( spm, 'asset_id' ) or derefOpt( spm, 'id' )
			local revIdThatIncurredError = derefOpt( spm, 'revision_id' )
			
			if not idOfDocThatIncurredError then
				-- this error doesn't match the form we expect, mark it unknown for now
				logger:errorf( "Unexpected format in error document, missing source document id : %s", WFTableArchiving.tableToString( errorDoc ) )
				return markErrorAsUnknown( docStore, errorDocId )
			end
			
			return docStore:docValue( idOfDocThatIncurredError )
				:flatMap( function( docThatIncurredError )
					
					-- Check that the document is not empty/deleted and that the revision matches the latest revision.
					--
					-- This is intended to filter out the error documents that may have been generated by changes from
					-- other clients. However, it also means that it filters out any error documents generated from this
					-- client, but at an earlier revision than what is current.
					-- 
					-- In these cases we expect that the later revision will be sent to Oz and, if that again results
					-- in an error, the error will be handled with an error document pertaining to that revision.
					if docThatIncurredError._empty or docThatIncurredError._revId ~= revIdThatIncurredError then
						return deleteErrorDoc( docStore, errorDocId )
					end
					
					
					local subtype = errorDoc.subtype
					
					-- don't let any unexpected error formats that fail to get to the diagnostic loc cause any issues 
					-- with processing more error documents
					pcall( addErrorDocToDiagnosticLog, self, errorDoc )
			
					local handler = getErrorHandler( self, subtype ) or defaultHandler
				
					return handler( self, errorDoc, docThatIncurredError )
			
			end )
				
		end )
		:call( WFRxRunUntilDone )

end

--------------------------------------------------------------------------------

local function monitorBlockedUploads( self )
	
	-- SPAM [matthewj 2018-01-12]: This is currently specific to TrialOrSubscriptionExpired, but 
	-- that's only because it's the only thing in the table. If storage errors or other errors are
	-- moved in, this will need to be generalized
	
	local expiredDocIdRows = self._indexStore:query {
			sql = [[SELECT docId FROM blocked_uploads WHERE reason = ?]],
			calm = 5,
			enqueue = 'slow',
			WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError
		}
	
	local isSubscriptionExpired = self._catalog:ozUser():isSubscriptionExpired()
	
	return Observable.combineLatest( expiredDocIdRows, 
							isSubscriptionExpired, 
							Observable.timer( 10, WFRxScheduler.sharedScheduler() ) )
		:filter( function( expiredDocIdRows, isExpired ) 
				return ( #expiredDocIdRows > 0 ) and not isExpired 
		end )
		:flatMap( function( expiredDocIdRows, isExpired )
			-- we should only be here if we have ids blocked because of subscription expiry and 
			-- we are not in an expired state so we can go ahead and clear all of these errors
			
			local bulkOverlay = {}
			
			for _, docIdRow in ipairs( expiredDocIdRows ) do
				table.insert( bulkOverlay, WFCatOzErrors.createOverlayToRemoveOzError( docIdRow[ 1 ] ) )
			end
			
			return self._catalog:docStore():bulkApplyOverlay( bulkOverlay )
			
		end )
		:subscribe( function( )
			logger:info( "Cleared errors on documents previously blocked from upload" )
		end )
end

--------------------------------------------------------------------------------

local function wrapFnForClass( fn )
	return function( _, ... )
		return fn( ... )
	end
end

--============================================================================--

-- @mark WFCatErrorHandler namespace

local WFCatErrorHandler = {

	_catalog = false,
		-- (WFCatalog)
		
	_indexStore = false,
		-- (WFRxIndexStore)
	
	_errorHandlerSubscription = false,
		-- (Subscription) Subscription to the Observable that will watch for process error documents
	
	_errorDocumentHandlerFns = false,
		-- (table) table mapping error document subtypes to error document handler functions
		
	_externalizeSettingsFn = false,
		-- (function) function that takes a catalog, document id, and optional enforced limit and returns 
		-- an Observable that will externalize the develop settings and return true when complete
	
	_deleteAssetsFn = false,
		-- (function) function that takes a catalog and list of document ids, and returns 
		-- an Observable that will delete the assets when complete

	_deleteIndividualAlbumFn = false,
		-- (function) function that takes a catalog and document id, and returns 
		-- an Observable that will delete the individual album when complete
 
	_errorDocSummary = {},
		-- (table) summary of the error documents received by the client
	
	_blockedUploadSubscription = false,
		-- (Subscription) watches uploads blocked by Oz and the conditions that would unblock them
	
	class = {
		generateIndexSchema = wrapFnForClass( generateIndexSchema ),
	},
	
	WFCloseSupport.addCloseSupportMethods,

}


--------------------------------------------------------------------------------
--[[
	\brief
	
	
]]--
function WFCatErrorHandler:__init( catalog, optClientHandlerFnTbl )
	
	check( '_', 'WFCatalog' )
	
	self._errorDocumentHandlerFns = {}
	
	self:setOwned {
		_catalog = catalog:relayObject()
	}
	
	self._indexStore = catalog:indexStore()
	
	-- set up error document handlers

	self:registerErrorHandler( WFCatOzErrors.Errors.kResourceNotFoundError, resourceNotFoundHandler )
	self:registerErrorHandler( WFCatOzErrors.Errors.kStorageExceededError, storageExceededHandler )
	self:registerErrorHandler( WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError, trialOrSubscriptionExpiredHandler )
	self:registerErrorHandler( WFCatOzErrors.Errors.kPayloadValidationError, payloadValidationHandler )
	self:registerErrorHandler( WFCatOzErrors.Errors.kInputValidationError, inputValidationHandler )
	self:registerErrorHandler( WFCatOzErrors.Errors.kResourceExistsError, resourceExistsHandler )
	self:registerErrorHandler( WFCatOzErrors.Errors.kMultipartCompleteError, multipartCompleteHandler )
	
	-- Some handlers need client specific implementations which can be passed in the fn
	-- table. If the table exists and the fns are provided, add them to the error handler
	if optClientHandlerFnTbl then
		self._externalizeSettingsFn = optClientHandlerFnTbl.externalizeSettings
		self._deleteAssetsFn = optClientHandlerFnTbl.deleteAssets
		self._deleteIndividualAlbumFn = optClientHandlerFnTbl.deleteIndividualAlbum
	end
	
	if self._externalizeSettingsFn then
		self:registerErrorHandler( WFCatOzErrors.Errors.kInlineXmpDevelopSettingTooBigError, inlineSettingsHandler )
	end
	
	--------------------------------------------------------------------------------

	local errorQuery = self._indexStore:query {
		sql = [[SELECT docId FROM errors ORDER BY docId LIMIT 10]], 
		calm = 1,
		enqueue = 'slow',
	}
	
	self._errorHandlerSubscription = errorQuery:subscribe(
			function( docIdRows )
				if #docIdRows > 0 then
					logger:infof( "Error document handler identified %d error documents", #docIdRows )
				end
				for _, docIdRow in ipairs( docIdRows ) do
					processOneErrorDocument( self, docIdRow[ 1 ] )
				end
			end,
			function( message )
				logger:errorf( "Error document handler query failed with message: %s", message )
			end,
			function()
				logger:debugf( "Error document handler query notified as done" )
			end )
	
	self._blockedUploadSubscription = monitorBlockedUploads( self )
	
	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

function WFCatErrorHandler:registerErrorHandler( errSubtype, handlerFn )
	
	self._errorDocumentHandlerFns[ errSubtype ] = handlerFn
	
end

--------------------------------------------------------------------------------

function WFCatErrorHandler:getDefaultErrorHandlerFn()
	return defaultHandler
end

--------------------------------------------------------------------------------

function WFCatErrorHandler:sendDiagnosticLogData()
	
	WFDiagnosticLog.addSectionToLog
		{

			id = string.format( 'WFCatErrorHandler' ),

			name = string.format( "Sync Error Handler" ),
			weight = 510,
			
			body = self._errorDocSummary

		}
			
end

--------------------------------------------------------------------------------

function WFCatErrorHandler:close()
	
	self._errorHandlerSubscription:dispose()
	self._blockedUploadSubscription:dispose()
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatErrorHandler'( WFCatErrorHandler )
