--[[----------------------------------------------------------------------------

WFCatFaces.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2017 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local WFArguments = require 'WFArguments'
local KSDate = require 'KSDate'
local WFTableMerge = require 'WFTableMerge'
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.managed_catalog
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatIdUtils = require 'WFCatIdUtils'
local WFCatUtils = require 'WFCatUtils'

--------------------------------------------------------------------------------

local check = WFArguments.check
local derefOpt = KSTableUtils.derefOpt

--==============================================================================

local WFCatFaces = {}

--------------------------------------------------------------------------------

--[[!
	\brief 

	\param 
	\param 

	\return 
--]]

function WFCatFaces.createAddFaceOverlay( faceInfo, faceArea, optPeople )
	
	check( 'table', 'table', '?table' )
	
	local area = {
		x = assert( faceArea.x ),
		y = assert( faceArea.y ),
		w = assert( faceArea.w ),
		h = assert( faceArea.h ),
		unit = "normalized"
	}
	
	
	local regionId = faceInfo.uuid or KSIdGenerator.uuidInLowercaseHex( )
	local name  = faceInfo.name 
	local desc = faceInfo.desc
	local timestamp = faceInfo.timestamp or KSDate.formatDateTimeAsUTC( KSDate.timeNow() ) .. "Z"
	local confidenceLevel = faceInfo.confidenceLevel
	
	local people 
	if optPeople then
		
		people = {}
		for i, person in ipairs( optPeople ) do
			people[ person.id ] = {
				Confirmed = person.confirmed,
				ConfidenceLevel = person.confidenceLevel,
			}
		end
	end
	
	local extensions = {}
	extensions[ "adobe-fi" ] = {
		Timestamp = timestamp,
		ConfidenceLevel = confidenceLevel,
		People = people,
	}
	
	return {
		xmp = {
			["mwg-rs"] = {
				Regions = {
					RegionList = {
						[ regionId ] = {
							Area = area,
							Name = name,
							Description = desc,
							Type = "Face",
							Extensions = extensions,
						}
					}
				}
			}
		}
	}
	
end

--------------------------------------------------------------------------------

--[[!
	\brief 

	\param 
	\param 

	\return 
--]]

function WFCatFaces.createAddFaceOverlaysForAsset( assetId, appliedToDimensions, faceDetailsArray )
	
	check( '+string|number', 'table', 'table' )
	
	local overlay = WFCatUtils.createBaseOverlayWithUserCreatedFields( assetId )

	for i, faceDetails in ipairs( faceDetailsArray ) do
		WFTableMerge.applyOverlay( overlay, 
			WFCatFaces.createAddFaceOverlay( faceDetails.faceInfo or {}, faceDetails.faceArea, faceDetails.people ) )
	end
	
	local dimensionOverlay = {
		xmp = {
			["mwg-rs"] = {
				Regions = {
					AppliedToDimensions = {
						w = assert( appliedToDimensions.w ),
						h = assert( appliedToDimensions.h ),
						unit = "pixel",
					}
				}
			}
		}
	}
	
	
	return WFTableMerge.applyOverlay( overlay, dimensionOverlay )
	
end



--------------------------------------------------------------------------------

--[[!
	\brief 

	\param 
	\param 

	\return 
--]]

function WFCatFaces.addFacesToAsset( catalog, assetId, appliedToDimensions, faceDetailsArray )
	
	check( 'WFCatalog', '+string|number', 'table', 'table' )
	
	local ok, overlay = pcall( WFCatFaces.createAddFaceOverlaysForAsset, assetId, appliedToDimensions, faceDetailsArray )
	if not ok then
		return Observable.throw( "face_overlay_error" )
	end
	
	return catalog:docStore():applyOverlay( overlay )
	
end
--------------------------------------------------------------------------------

function WFCatFaces.getPeople( wfCatalog, imageIds )

	check( 'WFCatalog', 'table' )
	
	return WFCatIdUtils.getOzAssetIdsForLocalIds( wfCatalog, imageIds, true )
		:flatMap( function( assetIds )
		
			return wfCatalog:indexStore():query {
				sql = [[SELECT name FROM person, faces INNER JOIN TEMP_ids ON TEMP_ids.docId == faces.assetId and person.personId == faces.personId and person_confirmed = 1]],
				quiesceUntilUpToDate = true,
				tempIds = assetIds,
			}
			:map( function( rows )
				local result = {}
				for i, row in ipairs( rows ) do
					result[ i ] = row[ 1 ] -- Just add the name for now
				end
				return result
			end )
		end )
		
end
--------------------------------------------------------------------------------

function WFCatFaces.generateIndexSchema( indexStoreSchema )

	check( 'WFRxIndexStoreSchema' )
	
	indexStoreSchema:addTable {

		sql = [[CREATE TABLE faces( faceId, assetId, personId, person_confirmed )]],

		useLocalId = true,
		hasOneRowPerDocument = false,
		docValueFilter = function( docValue )
			return ( docValue.subtype == 'image' or docValue.subtype == 'video' ) and docValue.type == 'asset'
		end,
		map = function( docValue, emit )

			if WFCatIndexUtils.isValidAssetDocument( docValue ) 
				and ( docValue.subtype == 'image' or docValue.subtype == 'video' ) then
					-- Stacks do have same properties as the main asset.
					-- Hence consider only normal assets here.
				
				local regionList = derefOpt( docValue, 'xmp','mwg-rs', 'Regions', 'RegionList' )
				
				for k, region in pairs( regionList or {} ) do
					
					if region.Type == "Face" then
						
						local people = derefOpt( region, 'Extensions', 'adobe-fi', 'People')
						
						for personId, person in pairs( people or {} ) do
							emit( k, docValue._id, personId, person.Confirmed and true or false)
						end
					end
				end
				
			end

		end,

	}
	
	indexStoreSchema:addIndex [[CREATE INDEX faces_confirmed_asset_person ON faces( person_confirmed, assetId, personId )]]
	
end

--==============================================================================

return WFCatFaces
