--[[----------------------------------------------------------------------------

WFCatIdUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2017 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'
local WFLogger = require 'WFLogger'


--============================================================================--

local WFCatIdUtils = {}

--------------------------------------------------------------------------------

local logger = WFLogger( 'WFCatIdUtils' )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local albumUuidByLocalId = {}
	-- key: album local Id
	-- value: album uuid

local assetUuidByLocalId = {}
	-- key: asset local Id
	-- value: asset uuid
	
local assetSubtypeByLocalId = {}
	-- key: asset local Id
	-- value: asset subtype
	
local assetSubtypeByAssetUuid = {}
	-- key: asset uuid
	-- value: asset subtype
	
local debug_queryCount = 0 -- DEBUG
	
local incrementQueryCount = function() -- DEBUG
	debug_queryCount = debug_queryCount + 1
end
	
--------------------------------------------------------------------------------
	
local function checkCache( cacheToCheck, localIds )
	
	local cachedResults = {}
		
	for i, localId in ipairs( localIds ) do
	
		local cachedValue = cacheToCheck[ localId ]
			
		if cachedValue then
			table.insert( cachedResults, cachedValue )
		else
			cachedResults = false
			break
		end
	
	end

	return cachedResults

end
	
--------------------------------------------------------------------------------
--[[!
	\brief Fetches the oz-asset-ids corresponding to the local-ids
		passed. If the ones passed as local-ids were already uuids, 
		then we will return the same.

	\param wfCatalog (WFCatalog) catalog
	\param maybeLocalIds (table) list of localids.
	\param maybeIgnoreStackType (boolean) true if the asset id translation should ignore any stack documents

	\return Observable which fires only once with the list of
		oz-asset-ids.
--]]

function WFCatIdUtils.getOzAssetIdsForLocalIds( wfCatalog, maybeLocalIds, maybeIgnoreStackType )
	
	maybeLocalIds = maybeLocalIds or {}
	local idsAreNumbers = type( maybeLocalIds[ 1 ] ) == 'number' -- assume they're all the same type
	
	local sql = [[SELECT TEMP_ids.docId, assetId, subtype FROM assets INNER JOIN TEMP_ids ON TEMP_ids.docId == assets.docId]]
	
	-- Check the cache to see if we need to run the query or not
	
	local cachedResults = {}
	
	if maybeIgnoreStackType then
	
		for i, localId in ipairs( maybeLocalIds ) do
		
			local cachedAssetUuid = assetUuidByLocalId[ localId ]
			local cachedSubtype = assetSubtypeByLocalId[ localId ]
			
			if cachedAssetUuid and cachedSubtype then
				
				if cachedSubtype ~= 'stack' then
					table.insert( cachedResults, cachedAssetUuid )
				end
				
			else
				cachedResults = false
				break
			end
		
		end
	
	else
	
		cachedResults = checkCache( assetUuidByLocalId, maybeLocalIds )
		
	end
	
	if cachedResults then
		return Observable.just( cachedResults )
	end
	
	-- First run the query without quiesceUntilUpToDate. If we get all the Uuids, we're done. 
	-- If not rerun the query WITH quiesceUntilUpToDate.
	
	local assetIdsQueryFn = function( quiesceUntilUpToDate )
		
		quiesceUntilUpToDate = quiesceUntilUpToDate or false
		
		return wfCatalog:indexStore():query {
				sql = sql,
				quiesceUntilUpToDate = quiesceUntilUpToDate,
				tempIds = maybeLocalIds,
			}
			:take( 1 )
	
	end
	
	if idsAreNumbers then 
		
		incrementQueryCount()
		
		return assetIdsQueryFn( false or maybeIgnoreStackType )
		
			:flatMap( function( rows )
			
				if not maybeIgnoreStackType and #rows ~= #maybeLocalIds then
					return assetIdsQueryFn( true )
				end
			
				return Observable.just( rows )
			
			end )
			
			:map( function( rows )
				local result = {}
				for i, row in ipairs( rows ) do
				
					local localId = row[ 1 ]
					local assetId = row[ 2 ]
					local subtype = row[ 3 ]
				
					if not maybeIgnoreStackType or subtype ~= 'stack' then
						table.insert( result, assetId )
					end
					
					assetUuidByLocalId[ localId ] = assetId
					assetSubtypeByLocalId[ localId ] = subtype
					assetSubtypeByAssetUuid[ assetId ] = subtype
					
				end
				return result
			end )
	
	else
	
		return Observable.just( maybeLocalIds or {} )
	
	end
end

--------------------------------------------------------------------------------
--[[!	
	\brief Fetches the oz-asset-ids and asset type corresponding to the local-ids
		passed. If the ones passed as local-ids were already uuids, 
		then we will return the same.

	\param wfCatalog (WFCatalog) catalog
	\param maybeLocalIds (table) list of localids. 

	\return Observable which fires only once with the list of
		oz-asset-ids and asset type pairs
--]]

function WFCatIdUtils.getOzAssetIdsAndTypeForIds( wfCatalog, maybeLocalIds )
	
	maybeLocalIds = maybeLocalIds or {}
	local idsAreNumbers = type( maybeLocalIds[ 1 ] ) == 'number' -- assume they're all the same type
	
	local sql
	
	local cachedResults = {}
	
	if idsAreNumbers then 
	
		for i, localId in ipairs( maybeLocalIds ) do
	
			local cachedAssetUuid = assetUuidByLocalId[ localId ]
			local cachedSubtype = assetSubtypeByLocalId[ localId ]
			
			if cachedAssetUuid and cachedSubtype then
				
				local row = {
					cachedAssetUuid,
					cachedSubtype
				}
				
				table.insert( cachedResults, row )
				
			else
				cachedResults = false
				break
			end
		
		end
	
		sql = [[SELECT assets.docId, assetId, subtype FROM assets INNER JOIN TEMP_ids ON TEMP_ids.docId == assets.docId]]
		
	else
	
		for i, assetUuid in ipairs( maybeLocalIds ) do
	
			local cachedSubtype = assetSubtypeByAssetUuid[ assetUuid ]
			
			if assetUuid and cachedSubtype then
				
				local row = {
					assetUuid,
					cachedSubtype
				}
				
				table.insert( cachedResults, row )
				
			else
				cachedResults = false
				break
			end
		
		end
	
		sql = [[SELECT assets.docId, assetId, subtype FROM assets INNER JOIN TEMP_ids ON TEMP_ids.docId == assets.assetId]]
		
	end

	if cachedResults then
		return Observable.just( cachedResults )
	end
	
	incrementQueryCount()
	
	-- First run the query without quiesceUntilUpToDate. If we get all the Uuids, we're done. 
	-- If not rerun the query WITH quiesceUntilUpToDate.
	
	local getAssetIdsAndTypesQuery = function( quiesceUntilUpToDate )
		
		quiesceUntilUpToDate = quiesceUntilUpToDate or false
		
		return wfCatalog:indexStore():query {
				sql = sql,
				quiesceUntilUpToDate = quiesceUntilUpToDate,
				tempIds = maybeLocalIds,
			}
			:take( 1 )
	
	end
	
	return getAssetIdsAndTypesQuery( false )
	
		:flatMap( function( rows )
		
			if #rows ~= #maybeLocalIds then
				logger:warn( 'WFCatIdUtils.getOzAssetIdsAndTypeForIds retrying query with quiesceUntilUpToDate' )
					-- Maybe record this in highbeam? Or diagnostic log?
				return getAssetIdsAndTypesQuery( true )
			end
		
			return Observable.just( rows )
		
		end )

		:map( function( rows )

			local result = {}

			for i, row in ipairs( rows ) do
			
				local localId = row[ 1 ]
				local assetId = row[ 2 ]
				local subtype = row[ 3 ]
				
				table.insert( result, {
					assetId,
					subtype
				} )
				
				assetUuidByLocalId[ localId ] = assetId
				assetSubtypeByLocalId[ localId ] = subtype
				assetSubtypeByAssetUuid[ assetId ] = subtype
				
			end

			return result

		end )
		
end

--[[!
	\brief Fetches the oz album ids corresponding to the ids passed in. If the ids passed in 
		are already uuids, then we return the same.
		
	\param wfCatalog( WFCatalog) catalog
	\param maybeLocalIds (table) list of ids
	
	\return Observable which fires only once with the list of oz album ids
--]]
function WFCatIdUtils.getOzAlbumIdsForLocalId( wfCatalog, maybeLocalIds )

	maybeLocalIds = maybeLocalIds or {}
	local idsAreNumbers = type( maybeLocalIds[ 1 ] ) == 'number' -- assume they're all the same type
	
	if idsAreNumbers then
	
		-- Check the cache to see if we need to run the query or not
		
		local cachedResults = checkCache( albumUuidByLocalId, maybeLocalIds )

		if cachedResults then
			return Observable.just( cachedResults )
		end
	
		incrementQueryCount()
	
		-- Else, run the query.
	
		-- First run the query without quiesceUntilUpToDate. If we get all the Uuids, we're done. 
		-- If not rerun the query WITH quiesceUntilUpToDate.
	
		local getAlbumIdsQuery = function( quiesceUntilUpToDate )
		
			quiesceUntilUpToDate = quiesceUntilUpToDate or false
			
			return wfCatalog:indexStore():query {
					sql = [[SELECT TEMP_ids.docId, albumId FROM albums INNER JOIN TEMP_ids ON TEMP_ids.docId == albums.docId]],
					quiesceUntilUpToDate = quiesceUntilUpToDate,
					tempIds = maybeLocalIds,
				}
				:take( 1 )
		
		end
	
		return getAlbumIdsQuery()
		
			:flatMap( function( rows )
			
				if #rows ~= #maybeLocalIds then
					logger:warn( 'WFCatIdUtils.getOzAlbumIdsForLocalId retrying query with quiesceUntilUpToDate' )
						-- Maybe record this in highbeam? Or diagnostic log?
					return getAlbumIdsQuery( true )
				end
			
				return Observable.just( rows )
			
			end )
		
			:map( function( rows )
				local result = {}
				for i, row in ipairs( rows ) do
				
					local localId = row[ 1 ]
					local albumId = row[ 2 ]
				
					table.insert( result, albumId )
					albumUuidByLocalId[ localId ] = albumId
					
				end
				return result
			end )
	
	else
	
		return Observable.just( maybeLocalIds or {} )
	
	end

end

--------------------------------------------------------------------------------

--[[!
	\brief Finds the uuid for an album's local-id

	\param wfCatalog (WFCatalog) The catalog.
	\param albumId (string) The local-id of the album

	\return <code>Observable< ></code> An Observable that emits the uuid of the album.
]]--

function WFCatIdUtils.albumUuidForMaybeLocalId( wfCatalog, maybeLocalAlbumId )
	
	check( "WFCatalog", '+string|number' )

	local cachedAlbumUuid = albumUuidByLocalId[ maybeLocalAlbumId ]
		
	if cachedAlbumUuid then
		return Observable.just( cachedAlbumUuid )
	end

	if type( maybeLocalAlbumId ) == 'number' then
	
		incrementQueryCount()
	
		-- First run the query without quiesceUntilUpToDate. If we get all the Uuids, we're done. 
		-- If not rerun the query WITH quiesceUntilUpToDate.
	
		local getAlbumIdQuery = function( quiesceUntilUpToDate )
		
			quiesceUntilUpToDate = quiesceUntilUpToDate or false
			
			return wfCatalog:indexStore():query {
					sql = "SELECT albumId FROM albums WHERE docId = ?",
					maybeLocalAlbumId,
					quiesceUntilUpToDate = quiesceUntilUpToDate,
				}
				:take( 1 )
		
		end
		
		local albumUuidObservable = getAlbumIdQuery( false )
		
			:flatMap( function( rows )
			
				local maybeAlbumUuid = derefOpt( rows, 1, 1 ) 
			
				if not maybeAlbumUuid then
					logger:warn( 'WFCatIdUtils.albumUuidForMaybeLocalId retrying query with quiesceUntilUpToDate' )
						-- Maybe record this in highbeam? Or diagnostic log?
					return getAlbumIdQuery( true )
				end
			
				return Observable.just( rows )
			
			end )
			
			:map( function( rows )
			
				logger:debugf( 'WFCatIdUtils.albumUuidForMaybeLocalId for localId %s complete w/ %d rows', maybeLocalAlbumId, #rows )
			
				local maybeAlbumUuid = derefOpt( rows, 1, 1 )
				
				if not maybeAlbumUuid then
					return error( "No such album: " .. tostring( maybeLocalAlbumId ) )
				else
					albumUuidByLocalId[ maybeLocalAlbumId ] = maybeAlbumUuid
				end
				
				return maybeAlbumUuid
				
			end )
		
		return albumUuidObservable
	
	end

	-- maybeLocalAlbumId is already string
	return Observable.just( maybeLocalAlbumId )
	
end

--------------------------------------------------------------------------------

--[[!
	\brief DEBUG - Only used in unit tests. Clear the WFCatIdUtils caches used to
		speed up Id translations.

	\return void
]]--

function WFCatIdUtils.debug_clearCaches()

	albumUuidByLocalId = {}
	assetUuidByLocalId = {}
	assetSubtypeByLocalId = {}
	assetSubtypeByAssetUuid = {}

end

--------------------------------------------------------------------------------

--[[!
	\brief DEBUG - Only used in unit tests.
	
	\return (numter) The number of times this class has run index store queries.
]]--

function WFCatIdUtils.debug_getQueryCount()

	return debug_queryCount

end

--============================================================================--

return WFCatIdUtils
