--[[----------------------------------------------------------------------------

WFCatIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local KSStringUtils = require 'KSStringUtils'

	-- WF.oz_client.managed_catalog
local WFCatIndexUtils = require 'WFCatIndexUtils'

	-- WF.sqlite.index_store
local WFRxIndexStoreSchema = require 'WFRxIndexStoreSchema'

--------------------------------------------------------------------------------

local derefOpt = KSTableUtils.derefOpt
local isValidDisplayableAssetDocument = WFCatIndexUtils.isValidDisplayableAssetDocument

--==============================================================================

local function getRating( docValue )

	local ratings = derefOpt( docValue, 'ratings' )
	
	if ratings then

		-- Due to legacy issues, ratings are keyed by user id. We can now assume
		-- only one user (the current user) is able to rate a photo.

		local _, v = next( ratings )

		if v then
			return v.rating or 0
		end
	end
	
	return 0
end

--------------------------------------------------------------------------------

local flagNameToInteger = {
	reject = -1,
	unflagged = 0,
	pick = 1,
}

--------------------------------------------------------------------------------

local function getFlag( docValue )

	local reviews = derefOpt( docValue, 'reviews' )
	
	if reviews then

		-- Due to legacy issues, flags are keyed by user id. We can now assume
		-- only one user (the current user) is able to flag a photo.

		local _, v = next( reviews )

		if v then
			local flagName = v.flag
			
			if flagName then
				return flagNameToInteger[ flagName ] or 0
			end

		end
	end
	
	return 0
end

--------------------------------------------------------------------------------

local function getDuration( docValue )

	local subtype = derefOpt( docValue, 'subtype' )

	if subtype ~= 'video' then
		return
	end

	local videoDuration = derefOpt( docValue, 'video', 'duration' )

	if videoDuration and #videoDuration == 2 then
		return videoDuration[ 2 ] == 0 and 0
			or videoDuration[ 1 ] / videoDuration[ 2 ]
	end

	return
end

--============================================================================--

local WFCatIndexDefinitions = {}

--------------------------------------------------------------------------------

local tableDefinitions = {}
local indexDefinitions = {}

--------------------------------------------------------------------------------

tableDefinitions.assets = {

	sql = [[CREATE TABLE assets( width INTEGER, height INTEGER, orientation INTEGER, captureDate, filename, filenameLC, assetId, subtype, rating INTEGER, flag INTEGER, duration INTEGER, isMedia )]],
	
	docValueFilter = function( docValue )
		return docValue.type == 'asset'
	end,
	
	map = function( docValue, emit )
		
		local subtype = docValue.subtype
		
		local isMedia = subtype == 'image' or subtype == 'video'
		
		local isDisplayable, displayableDocValue = isValidDisplayableAssetDocument( docValue )
		if isDisplayable then

			-- If the develop stanza is present it's required contain croppedWidth and croppedHeight and they
			-- should be used. If the develop stanza is not present, the payload width/height are used instead.
			-- Both of these will have the orientation already baked into them: width/height will be 
			-- flipped at import time based on the TIFF camera orientation, and croppedWidth/croppedHeight 
			-- will be populated and flipped as needed each time the userOrientation changes.
			
			local width = derefOpt( displayableDocValue, 'develop', 'croppedWidth' )
							or derefOpt( displayableDocValue, 'importSource', 'originalWidth' )
	
			local height = derefOpt( displayableDocValue, 'develop', 'croppedHeight' )
							or derefOpt( displayableDocValue, 'importSource', 'originalHeight' )
			
			if not width or not height then
				-- NMB-7361: No width/height? then set width/height values that will work (I hope)
				-- to define the most common aspect ratio
				if subtype == 'image' then
					width = 3 
					height = 2
				elseif subtype == 'video' then
					width = 16
					height = 9
				else
					width = 1
					height = 1
				end
			end
			
			-- This describes the transformation that should be applied to the original file dimensions, 
			-- which are not currently recorded anywhere in Oz. The payload width/height might be swapped,
			-- depending on the camera orientation at import time.  You have to open the negative and get its
			-- dimensions to know how to apply this orientation.
			
			local filename = derefOpt( displayableDocValue, 'importSource', 'fileName' )
			
			local filenameLC = ""
			
			if filename and type( filename ) == "string" then
			
				filenameLC = KSStringUtils.lower( filename )
				filenameLC = KSStringUtils.trim( filenameLC )
				
			end
			
			local orientation = derefOpt( displayableDocValue, 'develop', 'userOrientation' ) or 1

			local captureDate = derefOpt( displayableDocValue, 'captureDate' )
			
			local rating = getRating( displayableDocValue )

			local flag = getFlag( displayableDocValue )

			local duration = getDuration( displayableDocValue )

			isMedia = isMedia and 1 or 0

			if width and height and captureDate then

				emit( width, height, orientation, captureDate, filename, filenameLC, docValue._id, subtype, rating, flag, duration, isMedia )

			end

		end

	end,

	useLocalId = true,
	hasOneRowPerDocument = true,

}

indexDefinitions.assets_assetIddocId =
	[[CREATE INDEX assets_assetIddocId ON assets( assetId, docId )]]
indexDefinitions.assets_assetIdsubtype = 
	[[CREATE INDEX assets_assetIdsubtype ON assets( assetId, subtype )]]
indexDefinitions.assets_subtype = 
	[[CREATE INDEX assets_subtype ON assets( subtype )]]
indexDefinitions.assets_captureDatefilename =
	[[CREATE INDEX assets_captureDatefilename ON assets( captureDate, filenameLC COLLATE i18n_collate )]]
indexDefinitions.assets_assetIdIsMediaCaptureDate =
	[[CREATE INDEX assets_assetIdIsMediaCaptureDate ON assets( assetId, isMedia, captureDate )]]
	
--============================================================================--

local sqlitePageSize = 4096

local indexStoreSchema = WFRxIndexStoreSchema( sqlitePageSize )

indexStoreSchema:addDatabaseInitRoutine( require 'WFSQLiteSetLogicImp'.registerSetModule )

--------------------------------------------------------------------------------

for k, definition in pairs( tableDefinitions ) do
	indexStoreSchema:addTable( definition )
end

for k, definition in pairs( indexDefinitions ) do
	indexStoreSchema:addIndex( definition )
end


local function getIndexSchemaGenerators(  )
	return { 
		'NmbSimpleDocStoreUpdater',
		'NmbPresetDocumentProcessor',
		'NmbPathRestorer',
		'NmbFileMover',
		'NmbExternalDevelopSettings',
		'WFCatBinaryUploader',
		'WFCatConflictResolver',
		'WFCatCollaboration',
		'NmbDCPDocumentProcessor',
		'WFCatErrorHandler',
		'WFCatIdleDownloader',
		'WFCatImportDuplicateFiles',
		'WFCatSearchHandler',
		'WFCatSortHandler',
		'WFCatStacks',
		'WFCatStaleBinaryRemover',
		'WFCatStorage',
		'WFCatSyncStatus',
		'WFCatPerson',
		'WFCatFaces',
		require 'WFConfiguration'.Nimbus.Nmb5582InvalidOriginals and 'WFCatInvalidOriginalHandler' or nil,
		}
end


--------------------------------------------------------------------------------
local function generateIndexSchema( indexStoreSchema )

	local indexSchemaGenerators = getIndexSchemaGenerators()
	
	for k, generator in pairs( indexSchemaGenerators ) do
		require ( generator ).generateIndexSchema( indexStoreSchema )
	end
	
end

generateIndexSchema ( indexStoreSchema )

--------------------------------------------------------------------------------

WFCatIndexDefinitions.indexStoreSchema = indexStoreSchema

-------------------------------------------------------------------------------

--[[ 
	When the index store map server is initialized, we force update a list of 
	index store tables associated with that version tag. This only happens once
	per version tag. The mapping between version tag and the tables-to-update is 
	stored within tablesToReindexByTag, the format of which is:

	<key>: a version tag
	<value>: a list of tables to update

	Format of version tag:
	
		<major>.<minor>.<patch>.<dateEdited>

	There should only be one version tag per LrCC release; if there are multiple
	updates to a version tag per release, then the <dateEdited> field should be
	updated accordingly.
	
	If possible, please include the bug number in a comment that corresponds to
	the reason the tables are being reindexed for the specified version.

	Example:

	{ 
		'1.4.0.20180515' = { 
			'assets', -- NMB-123
			'albums' -- NMB-5829
		},
			
		'1.4.1.20180702' = { 
			'localStatusAndOzStatus' -- NMB-8080
		},
			
		...
	}
]]--

local tablesToReindexByTag = {}
	
tablesToReindexByTag[ '1.4.0.20180523' ] = { 
	'assets', -- NMB-7361 
	'albums', 
	'album_asset_v2',
	'assetsNeedingCameraProfileLinks',
	'keywords', -- NMB-8306
	'cameras' -- NMB-8306
}
	
WFCatIndexDefinitions.tablesToReindexByTag = tablesToReindexByTag

--------------------------------------------------------------------------------

return WFCatIndexDefinitions
