--[[----------------------------------------------------------------------------

WFCatIndexUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSTableUtils = require 'KSTableUtils'
local WFCatFileSpec = require 'WFCatFileSpec'

--============================================================================--

local derefOpt = KSTableUtils.derefOpt

--============================================================================--

local WFCatIndexUtils = {}

--------------------------------------------------------------------------------

local function maybeGetDisplayableDoc( docValue )

	if docValue.subtype == 'stack' then
		return derefOpt( docValue, '_localOnly', 'cover' )
	else
		return docValue
	end
	
end

--------------------------------------------------------------------------------

local function maybeGetSearchableDoc( docValue )
	-- SPAM: Yes this is the same as maybeGetDisplayableDoc for now, but these
	-- may diverge in the future a bit as stacks and other asset types may be 
	-- displayable but not searchable.
	return maybeGetDisplayableDoc( docValue )
end

--------------------------------------------------------------------------------

local function hasLocalOriginal( docValue )

	local localOriginalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
	local localOriginalRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue )
	return localOriginalVolumeGuid and localOriginalRelativePath

end

--------------------------------------------------------------------------------

local function haveLocalProxy( docValue )

	local localProxyVolumeGuid = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
	local localProxyRelativePath = WFCatFileSpec.getProxyRelativePathFromDocValue( docValue )
	return localProxyVolumeGuid and localProxyRelativePath

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.hasLocalProxyOrOriginal( docValue )
	return hasLocalOriginal( docValue ) or 
		haveLocalProxy( docValue )
end

--------------------------------------------------------------------------------

local function haveLocalThumbnail( docValue )

	-- Locally copied assets will have a _localOnly.sourceAssetId set,
	-- and any thumbnails which it has (which it must as it had to be displayble
	-- in order to be copied) will be copied to the copied asset.  This is
	-- needed in order for copy to work when no proxy or original exists and
	-- sync is paused.  It also improves percieved performance of the copy operation.
	return derefOpt( docValue, '_localOnly', 'sourceAssetId' )

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.assetHasThumbnail( docValue )
	
	local assetDoc = maybeGetDisplayableDoc( docValue )
	
	-- This filter should be applied to all indexes that drive user-visible content so that
	-- we don't show assets for which we don't have, or can't produce, a thumbnail
	
	if ( not assetDoc ) or ( assetDoc.type ~= 'asset' ) then return false end
	
	-- Include if we could generate a thumbnail locally
	
	if hasLocalOriginal( assetDoc ) or haveLocalProxy( assetDoc ) then
		return true
	end

	if haveLocalThumbnail( assetDoc ) then
		return true
	end

	-- Include if we can get a thumbnail from Oz.
	local links = derefOpt( assetDoc, '_serverProvidedMetadata', 'links' )
	
	if derefOpt( links, '/rels/rendition_type/144h', 'href' ) or
		derefOpt( links, '/rels/rendition_type/thumbnail2x', 'href' )
	then
		return true
	end
	
	-- Include if we can get an original/proxy from Oz in order to render locally
	if derefOpt( links, '/rels/proxy_type/2560', 'href' ) or
		derefOpt( links, '/rels/master', 'href' ) then
		return true
	end
	
	-- No thumbnail? Pretend it doesn't exist yet.
	
	return false

end

--------------------------------------------------------------------------------

-- TODO: This function should probably return true for subtypes 'preset' and
-- 'profile' too, but that would lead to the need for touching many places in
-- the code that can otherwise be left alone at this point, to avoid problems
-- with file mover code etc. Punting on that, at least for now.
-- 
-- Another option is to rename this function to represent what it is actually
-- used for, e.g. 'isVideoImageOrStackDocument'. That would need to be coupled
-- with renaming of things like 'isValidAssetDocument', though, which turns
-- it into quite the snowball.

function WFCatIndexUtils.isAssetDocument( docValue )
	
	return docValue
		and docValue.type == 'asset'
		and ( docValue.subtype == 'image' or docValue.subtype == 'video' or docValue.subtype == 'stack' )

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.isValidAssetDocument( docValue )
	
	return docValue
		and WFCatIndexUtils.isAssetDocument( docValue )
		and not docValue._deleted

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.isStyleDocument( docValue )
	
	return docValue
		and docValue.type == 'asset'
		and ( docValue.subtype == 'preset' or docValue.subtype == 'profile' )

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.isNonDeletedStyleDocument( docValue )

	return WFCatIndexUtils.isStyleDocument( docValue )
		and not docValue._deleted

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.isDCPDocument( docValue )
	
	return docValue
		and docValue.type == 'asset'
		and ( docValue.subtype == 'camera_profile' )

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.isNonDeletedDCPDocument( docValue )

	return WFCatIndexUtils.isDCPDocument( docValue )
		and not docValue._deleted

end

--------------------------------------------------------------------------------

function WFCatIndexUtils.doesAssetDocumentHaveLocalBinary( docValue )

	return
			( WFCatIndexUtils.isValidAssetDocument( docValue ) and
			WFCatIndexUtils.hasLocalProxyOrOriginal( docValue ) )
		or
			( WFCatIndexUtils.isNonDeletedStyleDocument( docValue ) and
			hasLocalOriginal( docValue ) )
		or
			( WFCatIndexUtils.isNonDeletedDCPDocument( docValue ) and
			hasLocalOriginal( docValue ) )

end
		
--------------------------------------------------------------------------------

function WFCatIndexUtils.isValidSearchableAssetDocument( docValue )
	local isSearchable = WFCatIndexUtils.isValidAssetDocument( docValue )
		and WFCatIndexUtils.assetHasThumbnail( docValue )
	
	return isSearchable,
		isSearchable and maybeGetSearchableDoc( docValue ) or nil
end

--------------------------------------------------------------------------------

function WFCatIndexUtils.isValidDisplayableAssetDocument( docValue )
	
	local isDisplayable = WFCatIndexUtils.isValidAssetDocument( docValue )
		and WFCatIndexUtils.assetHasThumbnail( docValue )
		
	return isDisplayable,
		isDisplayable and maybeGetDisplayableDoc( docValue ) or nil

end

--------------------------------------------------------------------------------

return WFCatIndexUtils
