--[[----------------------------------------------------------------------------

WFCatMetadataPuller.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable
local ReplaySubject = Rx.ReplaySubject

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.rxlua
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFOzDocUtils = require 'WFOzDocUtils'
local WFRxContinuousTask = require 'WFRxContinuousTask'
local WFPerformanceReport = require 'WFPerformanceReport'
local WFURI = require 'WFURI'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

	-- WF.oz_client.managed_catalog.sync
local WFCatUrlUtils = require 'WFCatUrlUtils'

--------------------------------------------------------------------------------

local resolveToAbsoluteUrl = assert( WFCatUrlUtils.resolveToAbsoluteUrl )

local isValidAssetDocument = assert( require 'WFCatIndexUtils'.isValidAssetDocument )

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

local timeNowForAnimation = assert( KSDate.timeNowForAnimation )

local logger = WFLogger 'WFCatMetadataPuller'
--:setLevel( 'debug' )

--------------------------------------------------------------------------------

local function remapOzDocumentData( self, ozDocument, baseUrl )

	local result = WFOzDocUtils.remapBasicOzDocument( ozDocument, baseUrl )
	
	-- Remap deletion to our flavor of deletion.
	if ozDocument.deleted then

		ozDocument.deleted = nil
		result._deleted = true

		self._deletedDocumentIdStream:onNext( result._id )

		for k in pairs( result ) do
			if not k:match( '^_' ) then
				result[ k ] = nil
			end
		end

	end
	
	-- In case Oz fails to send revision_ids, squawk about it, but force
	-- push the doc anyway.
	if result and ( not result._revIds or #result._revIds == 0 ) then

		-- No need to squawk if it's a deletion. No revision_ids expected.
		
		if not result._deleted then
			logger:infof( "Oz did not provide revision_ids in this document: %s", result )
		end
		
		result._revIds = { KSIdGenerator.uuidInLowercaseHex(), force = true }

	end
	
	local assetTypeHandler = self._assetTypeHandler[ result.type ] 

	if assetTypeHandler then
		assetTypeHandler( self, result )
		
		-- Assert _localOnly as it would overwrite existing data in the DocStore
		-- due to the Puller using :setDocValue()
		
		assert( result._localOnly == nil, 
			'_localOnly is not permitted here, as it would overwrite existing data' )
			
	end

	return result

end

--------------------------------------------------------------------------------

local function convertUpdatesToDocValues( self, updates, baseUrl )

	if type( updates ) ~= 'table' then
		return updates
	end

	-- Read an Oz updates feed and convert to input for :bulkSetDocValue.

	local result = { label = 'oz' }
	
	for i, updateDoc in ipairs( updates ) do

		logger:debugf( "about to remap doc from Oz: %s", updateDoc )

		local remappedDoc = remapOzDocumentData( self, updateDoc, baseUrl )
		
		result[ i ] = remappedDoc

		logger:debugf( "remapped doc: %s\n\n", remappedDoc )

		local docType = remappedDoc.type

		if docType then
			self._documentsReceivedByType[ docType ]
				= ( self._documentsReceivedByType[ docType ] or 0 ) + 1
		end

	end

	return result

end

--------------------------------------------------------------------------------

-- Needed to avoid storing the 'force_upgrade_error' parameter, if present.

local function stripForcedUpgrade( str )

	local pattern = "force_upgrade_error=1[&]*"

	return str:gsub( pattern, "" ):gsub( "%?$", "" ):gsub( "&$", "" )

end

--------------------------------------------------------------------------------

local function updateFeedLink( docStore, newLink )

	WFPerformanceReport.wrapObservable( 'WFCatMetadataPuller : updateFeedLink', 
		docStore:setLocalDocValue {
			_id = 'Oz-nextUpdateFeedLink',
			link = stripForcedUpgrade( newLink ),
		}, 
		{ 
			newUpdateFeedLink = newLink,
		} )

	:call( runUntilDone )

end

--------------------------------------------------------------------------------

local function updateLocalFeedLink( self, newLink )

	self._localNextUpdateFeedLink:onNext( newLink )

end

--------------------------------------------------------------------------------

local function checkForInitialSync( self, docStore )
	
	if not self._hasCheckedForInitialSync then 
		
		self._hasCheckedForInitialSync = true
		
		return docStore:localDocValue( 'Oz-isDoingInitialSync' )
			:map( function( docValue )
				
				if docValue.isDoingInitialSync == nil then 
					docStore:setLocalDocValue {
						_id = 'Oz-isDoingInitialSync',
						isDoingInitialSync = true,
						beginTimestamp = KSDate.timeNow(),
					}
					:call( runUntilDone )
				end
				
				-- if the isDoingInitialSync value hadn't yet been set, consider it true
				self._isDoingInitialSync = docValue.isDoingInitialSync == nil or docValue.isDoingInitialSync
				
				return self._isDoingInitialSync
			end )
			
	else 
		
		return Observable.just( self._isDoingInitialSync )
		
	end
end

--------------------------------------------------------------------------------

local function markInitialSyncCompleteIfNecessary( self, docStore )
	
	if self._isDoingInitialSync then 
		self._isDoingInitialSync = false
		
		docStore:applyOverlayLocalDocValue {
			_id = 'Oz-isDoingInitialSync',
			isDoingInitialSync = false,
			endTimestamp = KSDate.timeNow(),
		}
		:call( runUntilDone )
	
	end
end 

--------------------------------------------------------------------------------

local function retryDelayRequested( headers )
	return headers and headers.status_code == '200' and tonumber( headers[ 'Retry-After' ] )
end

--------------------------------------------------------------------------------

local function maybeAddDelayFromResponseHeader( headers )

	local retryDelay = retryDelayRequested( headers )

	if retryDelay then
		return Observable.timer( retryDelay, WFRxScheduler.sharedScheduler() )
	end

	return Observable.just()

end

--------------------------------------------------------------------------------

local function maybeSetRetryDelayRequestByOz( self, headers )

	local retryDelay = retryDelayRequested( headers )

	if retryDelay then

		if self._retryDelayRequestByOz then
			-- Don't think this can happen.
			logger:error( "Trying to add a retryDelay when one is already set, using new delay" )
		end

		self._retryDelayRequestByOz = retryDelay
		self._retryDelayRequestByOzAtTime = timeNowForAnimation()

	end

end

--------------------------------------------------------------------------------

local function maybeAddDelayRequestByOz( self )

	if self._retryDelayRequestByOz then

		local timeSinceDelayRequested = timeNowForAnimation() - self._retryDelayRequestByOzAtTime

		local retryDelay = self._retryDelayRequestByOz - timeSinceDelayRequested

		self._retryDelayRequestByOz = nil
		self._retryDelayRequestByOzAtTime = nil

		if retryDelay > 0 then
			return Observable.timer( retryDelay, WFRxScheduler.sharedScheduler() )
		end

	end

	return Observable.just(true)

end

--------------------------------------------------------------------------------

local function prefetchAlbumHierarchyImp( self, albumsLink, iterationCount )
	
	if not albumsLink or iterationCount > 20 then
		return Observable.just()
	end
	
	return self._catalogSpec:get( albumsLink, {
			returnHttpErrorsAsData = true
		} )
		:map( function( albumsJson, headers )
			
			if not headers or headers.status_code ~= '200' then
				return Observable.just()
			end
			
			-- decode the albums json into docs and splat into the doc store
			local decodedAlbumJson = KSJsonDecoder.decodeWebSafeJson_stripNull( albumsJson )
						
			local ozAlbums = KSTableUtils.derefOpt( decodedAlbumJson, "resources" )
			local baseUrl = KSTableUtils.derefOpt( decodedAlbumJson, 'base' )
			local albumDocs = convertUpdatesToDocValues( self, ozAlbums, baseUrl )
			
			local maybeNextLink = KSTableUtils.derefOpt( decodedAlbumJson, 'links', 'next', 'href' )
			
			if maybeNextLink then
				maybeNextLink = WFURI.resolve( baseUrl, maybeNextLink )
			end
						
			return albumDocs, maybeNextLink, headers
			
		end )
		:flatMap( function( albumDocs, maybeNextLink, headers )
		
			if albumDocs and #albumDocs > 0 then
				return self._catalog:docStore():bulkSetDocValue( albumDocs )
					:map( function()
						return maybeNextLink, headers
					end )
			else
				return Observable.just()
			end
		
		end )
		:flatMap( function( maybeNextLink, maybeHeaders )
			
			if not maybeNextLink then
				return Observable.just()
			end
			
			return prefetchAlbumHierarchyImp( self, maybeNextLink, iterationCount + 1 )
				:delay( function()
					return maybeAddDelayFromResponseHeader( maybeHeaders )
				end )
		
		end )
		:catch( Observable.just() )
		
end

--------------------------------------------------------------------------------

local function prefetchAlbumHierarchy( self )
	
	local catSpec = self._catalogSpec
	
	local albumsHref = string.format( '%s/albums?subtype=collection%%3Bcollection_set', catSpec:catalogHref() )
	
	return prefetchAlbumHierarchyImp( self, albumsHref, 0 )
		:takeLast( 1 )
end

--------------------------------------------------------------------------------

local function remapCatalogDocument( catalogDoc )
	
	catalogDoc = WFOzDocUtils.remapBasicOzDocument( catalogDoc, catalogDoc.base )
	
	-- Map as if it came from oz as part of sync feed
	catalogDoc._labels = { oz = true }
	
	return catalogDoc
	
end

--------------------------------------------------------------------------------

local function getCatalogDocFromOzAndRemap( catalog )
	
	local catalogSpec = catalog:ozCatalogSpec()
	
	return catalogSpec:get( catalogSpec:catalogHref() )
		
		:map( function( catalogJson, headers )
			
			if not headers or headers.status_code ~= '200' then
				error( "Request failed" )
			end
			
			local catalogDoc = KSJsonDecoder.decodeWebSafeJson_stripNull( catalogJson )
			catalogDoc = remapCatalogDocument( catalogDoc )
			return catalogDoc
		end )
end

--------------------------------------------------------------------------------

local function ensureCatalogDocIsAvailableLocally( self )

	if self._hasEnsuredCatalogExists then
		return Rx.Observable.just()
	end

	return self._catalog:docStore():docValue( self._catalog:ozCatalogId() )
	
		:flatMap( function( docValue )
			
			if docValue and not docValue._empty then
				-- Catalog doc is present. Return.
				return Rx.Observable.just()
			end
			
			return getCatalogDocFromOzAndRemap( self._catalog )
				:flatMap( function( catalogDoc )
					return self._catalog:docStore():setDocValue( catalogDoc )
				end )
			
		end )
		
		:tap( function()
			
			-- Turn on the variable so that we dont do this task again.
			
			self._hasEnsuredCatalogExists = true
			
		end )
		
		:tapOnError( function( err )
		
			logger:errorf( "Adding catalog document failed. Error:%s", err )
		end )

end

--------------------------------------------------------------------------------

local function getLatestUpdates( self, onlyGetUpdatesOnce )

	check( 'WFCatMetadataPuller' )

	-- TODO [scouten 2015-07-20]: Implement connection status reporting.
	
	if self._gettingLatestUpdates then
		return self._gettingLatestUpdates
	end
	
	if WFCloseSupport.hasBeenOrWillBeClosed( self._catalog ) then
		-- SPAM [matthewj 2016-08-26]: If we are shutting down and the continuous task runs
		-- we can stop it from sloppily attempting additional work by returning an Observable
		-- that won't complete 
		return Observable.never()	
	end
	
	local docStore = assert( self._catalog:docStore() )
	
	-- Find out where we left off before.
	self._gettingLatestUpdates = checkForInitialSync( self, docStore )
		:flatMap( function( isInitialSync ) 
			
			return docStore:localDocValue( 'Oz-nextUpdateFeedLink' )
				:map( function( docValue )
					local maybeUpdatesUrl = KSTableUtils.derefOpt( docValue, 'link' )
					return isInitialSync, maybeUpdatesUrl
				end )
			
		end )
		:flatMap( function( isInitialSync, maybeUpdatesUrl )

			if ( not isInitialSync ) or maybeUpdatesUrl then
				return Observable.just( maybeUpdatesUrl )
			end
			
			-- first time sync for the catalog, let's prime the pump with some albums
			return prefetchAlbumHierarchy( self )
				:map( function()
					return maybeUpdatesUrl
				end )
		
		end )
		:flatMap( function( maybeUpdatesUrl )

			-- Store the catalog document, if it is not
			-- preset already
			return ensureCatalogDocIsAvailableLocally( self )
				:map( function()
					return maybeUpdatesUrl
				end )
		
		end )
		:map( function( maybeUpdatesUrl )

			local updateUrl = maybeUpdatesUrl or self._catalogSpec:catalogUpdatesUrl()

			return updateUrl
			
		end )

		-- Switch over to a stream of update feed links that we gather when
		-- processing updates.
	
		:flatMap( function( updateUrl )
		
			updateLocalFeedLink( self, updateUrl )
			return self._localNextUpdateFeedLink
			
		end )

		:distinctUntilChanged()

		:flatMap( function( updateUrl )
			-- Using flatMap instead of delay to avoid ordering issues in unit tests caused by using delay.
			return maybeAddDelayRequestByOz( self )
			:map( function()
				return updateUrl
			end )
		end )

		-- Get next page of update feed from Oz.

		:flatMap( function( updateUrl )
		
			local options = {
				bypassRateLimiter = true,
			}
		
			return WFPerformanceReport.wrapObservable( 'WFCatMetadataPuller : get update feed', 
			self._catalogSpec:get( updateUrl, options ), 
			{ 
				requestUrl = updateUrl,
			} )
			
		end )

		-- Look for requested Retry-After and Log results.

		:tap( function( rawJson, headers )

			maybeSetRetryDelayRequestByOz( self, headers )

			logger:debugf( "Update feed pull\n\nGET %s\n\n%s\n\n", headers, rawJson )

		end )
		
		:tapOnError( function( rawErrorJson, headers )
			local WFPlatformInfo = require 'WFPlatformInfo'
			if headers and headers.status_code == '404' then
				if WFPlatformInfo.isAndroid and ((string.find(tostring(rawErrorJson), 'description":"Resource not found')))
						and ((string.find(tostring(rawErrorJson), 'errors":."catalog":."does not exist'))) then
					-- handling case for no catalog exists but a signed in user
				else
					self._isOzCatalogDeleted:onNext( true )
				end
			end
			
		end )

		-- Convert to JSON.

		:map( KSJsonDecoder.decodeWebSafeJson_stripNull )

		-- Convert Oz-flavored documents to WF-flavored documents. 

		:map( function( json )
		
			local ozUpdateDocs = KSTableUtils.derefOpt( json, 'updates' )
			local baseUrl = KSTableUtils.derefOpt( json, 'base' )
			local wfUpdateDocs = convertUpdatesToDocValues( self, ozUpdateDocs, baseUrl )
			local nextLink = KSTableUtils.derefOpt( json, 'links', 'next', 'href' )
			
			if nextLink and nextLink:len() > 0 then
			
				nextLink = resolveToAbsoluteUrl( baseUrl, nextLink )
				
				wfUpdateDocs.nextFeedLink = nextLink
				
				-- We update the local Oz update feed link, which allows us to pre-fetch 
				-- the next HTTP request while the current request continues to process.
				
				-- Note: We don't update the link tracked by the doc store until the http results
				-- from the previous request have been processed and the doc store has been
				-- updated. Order matters.

				if not onlyGetUpdatesOnce then
				
					updateLocalFeedLink( self, nextLink )
				
				end
			
			end
			
			return wfUpdateDocs
			
		end )
		
		:concatMap( function( wfUpdateDocs )

			local labelUpdateObservables = {}

			for i, wfUpdateDoc in ipairs( wfUpdateDocs ) do
				if wfUpdateDoc.type == 'warning' and wfUpdateDoc.subtype == 'RevisionCompressed' then

					-- jrechtie: 2017/12/05
					-- Compressed revisions can generally be ignored.

					-- One notable exception though:
					-- a revision which has been uploaded to Oz, but not
					-- received back from Oz might get compressed away by Oz. Such
					-- a revision might be the correct three-way-merge base revision to
					-- resolve further local and remote edits but wouldn't be considered
					-- because it doesn't appear in the remote edit parent revision list.

					-- This can be fixed in docStore:setDocValue() by consulting the 'oz'
					-- label when choosing a local parentRevision for a new revision which is
					-- incoming from remote. For this to work, we need to update the
					-- 'oz' label for a compressed away revision if it's also available
					-- locally and should carry the label but doesn't because it never
					-- came back from Oz.

					local spm = wfUpdateDoc._serverProvidedMetadata
					if spm and spm.source and spm.source.id then
						local removed_revisions = KSTableUtils.slice( wfUpdateDoc.original_revisions, 2, #wfUpdateDoc.original_revisions-1, 1 )
						local labelUpdateObservable = self._catalog:docStore():updateLabelFromRevisionList( spm.source.id, 'oz', removed_revisions )
														:catch( Observable.empty() )
														:defaultIfEmpty( 'done' )
						table.insert( labelUpdateObservables, labelUpdateObservable )
						wfUpdateDoc._deleted = true
					end

				end
			end

			if #labelUpdateObservables > 0 then
				local  combinedLabelUpdates = Observable.combineLatest( unpack ( labelUpdateObservables ) )
				return Observable.combineLatest( function( wfUpdateDocs, resultFromLabelUpdates )
							return wfUpdateDocs
						end,
						Observable.just(wfUpdateDocs), combinedLabelUpdates )
			else
				return Observable.just( wfUpdateDocs )
			end
		end)

		-- Write documents to database. We use concatMap here to make sure
		-- that we're processing the results from Oz in the order that the HTTP
		-- requests were issued.

		:concatMap( function( wfUpdateDocs )

			local updateDocCount = #wfUpdateDocs

			if type( wfUpdateDocs ) == 'table' then
			
				local nextLink = KSTableUtils.derefOpt( wfUpdateDocs, 'nextFeedLink' ) or false
			
				if updateDocCount > 0 then
			
					logger:debugf( "About to write new docs\n\n%s\n\n", wfUpdateDocs )
					
					self._lastUpdateCycleWithData = KSDate.timeInIso8601FormatWithTimeZone()
					
					self._continuousPullTask:taskDidWork()

					self._totalDocumentsReceived = self._totalDocumentsReceived + updateDocCount
				
					local bulkSetDocValueWithNextLink = Observable.combineLatest(
						function( bulkUpdateStatus, nextFeedLink )
							return bulkUpdateStatus, nextFeedLink
						end,
						docStore:bulkSetDocValue( wfUpdateDocs ), 
						Observable.just( nextLink ) )
				
					return WFPerformanceReport.wrapObservable( 'WFCatMetadataPuller : updateDocStore', 
						bulkSetDocValueWithNextLink, 
						{ 
							docsToUpdate = updateDocCount,
						} )
						:take( 1 )
						:tapOnCompleted( function()
							self._latestUpdatesStream:onNext( wfUpdateDocs )
						end )
				
				else

                    if self._isDoingInitialSync then
                        local sql = [[SELECT COUNT(docId) FROM flags]]
                        self._catalog:indexStore()
                        :queryP(sql)
                        :mapP (
                            function( _, rows )

                                if self._isDoingInitialSync then
                                    markInitialSyncCompleteIfNecessary( self, docStore )
                                end

                            end )
                        :runUntilDone()
                    end

                    -- Placeholder that says "no documents."
                    logger:debugf( "No documents to write" )
                    return Observable.just( false, nextLink )

				end
			else
				return Observable.throw( "Expected table, received " .. type( wfUpdateDocs ) )
			end

		end )

		:tap(

			function( bulkUpdateStatus, maybeNextLink )

				-- Log any successes.
 
				logger:debugf( "Bulk Update Status %s", bulkUpdateStatus )

				self._successfulUpdateCycles = self._successfulUpdateCycles + 1

				-- Now that an update has suceeded, we can update the next feed link
				-- that's tracked by the doc store.

				if maybeNextLink then
					updateFeedLink( docStore, maybeNextLink )
				end
				
			end,

			function( rawJsonOrError, headers, ... )

				-- Log any failures.
				logger:errorf( "ERROR on update feed pull\n\nGET %s\n\n%s\n\n", headers, rawJsonOrError )

			end )

		-- If bulkUpdateStatus has content, we received updates from Oz.
		-- If we didn't receive any updates, we want to shut down this observable
		-- and then check for updates again in a second or two.
		
		:filter( function( bulkUpdateStatus )
		
			local noMoreUpdates = not bulkUpdateStatus
			
			return noMoreUpdates or onlyGetUpdatesOnce or self._syncIsStopped
			
		end )

		:take( 1 )

		:flatMap( function()
			self._gettingLatestUpdates = nil
			return Observable.empty()	
		end )
		
		:tapOnError( function()
			self._gettingLatestUpdates = nil
		end )
		
		return self._gettingLatestUpdates

end

--============================================================================--

local WFCatMetadataPuller = {

	_catalog = false,
		-- (WFCatalog)

	_catalogSpec = false,
		-- (WFCatSpec)

	_documentsReceivedByType = false,
		-- (table)
	
	_lastUpdateCycleWithData = false,
		-- (string) Timestamp of the last pull from Oz that included data. Used only for debugging purposes.
		
	_totalDocumentsReceived = 0,
		-- (number)
		
	_successfulUpdateCycles = 0,
		-- (number)
	
	_assetTypeHandler = false,
		-- (table)
		-- key: asset type string
		-- value: a user specified function
	
	_localNextUpdateFeedLink = false,
		-- (Rx ReplaySubject) A local stream that fires with Oz update feed links.
		
	_latestUpdatesStream = false,
		-- (Rx ReplaySubject) A stream that fires with new the latest updates
		-- that are synced from Oz.
	
	_deletedDocumentIdStream = false,
		-- (rx ReplaySubject) A stream that fires with the latest document deletions from Oz.
	
	_gettingLatestUpdates = false,
		-- (Rx Observable) The observable being used to pull metadata updates.
	
	_continuousPullTask = false,
		-- (WFRxContinuousTask) A continuous task to handle continuous metadata updates
	
	_isDoingInitialSync = false,
		-- (boolean) indicates whether the puller is still processing the initial sync

	_hasCheckedForInitialSync = false,
		-- (boolean) indicates whether we've checked for initial sync
	
	_hasEnsuredCatalogExists = false,
		-- (boolean) indicates whether we've already made sure catalog doc exists locally
	
	_retryDelayRequestByOz = false,
		-- (number) when set indicates seconds since _retryDelayRequestByOzAtTime to delay before requesting updates.

	_retryDelayRequestByOzAtTime = false,
		-- (number) the value of timeNowForAnimation clock when _retryDelayRequestByOz was last set.

	_syncIsStopped = false,
		-- (boolean) indicates whether metadata sync should be running, for internal use only
		
	_isOzCatalogDeleted = false,
		-- (Rx.ReplaySubject) Indicates whether the Oz catalog got deleted
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function WFCatMetadataPuller:__init( catalog, optTaskDelays )

	check( 'WFCatMetadataPuller', 'WFCatalog', '?table' )

	self:setOwned {
		_catalog = catalog:relayObject()
	}
	
	self._syncIsStopped = true
	
	self._catalogSpec = catalog:ozCatalogSpec()

	self._documentsReceivedByType = {}

	self._assetTypeHandler = {}

	self._localNextUpdateFeedLink = ReplaySubject( 1 )

	self._latestUpdatesStream = ReplaySubject( 1 )

	self._deletedDocumentIdStream = ReplaySubject( 1 )
	
	self._isOzCatalogDeleted = ReplaySubject( 1 )
	self._isOzCatalogDeleted:onNext( false )

	WFDiagnosticLog.addLogProvider( self )

	-- TODO [scouten 2015-07-21]: Inspect the error and network
	-- conditions. Maybe we need to wait longer?

	local continuousTaskDelays = optTaskDelays or {
		progressiveBackoff = true,
		noWorkDelay = 2,
		maxNoWorkDelay = 30,
		errorDelay = 10,
		maxDelay = 60,
	}
	
	local continuousTask = function() return getLatestUpdates( self ) end

	self._continuousPullTask = WFRxContinuousTask( "WFCatMetadataPuller_"..catalog:ozCatalogId(), continuousTask, continuousTaskDelays )

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:close()

	check( 'WFCatMetadataPuller' )
	
	local continuousPullTask = self._continuousPullTask
	self._continuousPullTask = nil
	
	if continuousPullTask then
		continuousPullTask:close()
	end

	local localNextUpdateFeedLink = self._localNextUpdateFeedLink
	self._localNextUpdateFeedLink = nil
	
	if localNextUpdateFeedLink then
		localNextUpdateFeedLink:dispose()
	end

	local latestUpdatesStream = self._latestUpdatesStream
	self._latestUpdatesStream = nil
	
	if latestUpdatesStream then
		latestUpdatesStream:dispose()
	end

	local deletedDocumentIdStream = self._deletedDocumentIdStream
	self._deletedDocumentIdStream = nil
	
	if deletedDocumentIdStream then
		deletedDocumentIdStream:dispose()
	end
	
	local isOzCatalogDeleted = self._isOzCatalogDeleted
	self._isOzCatalogDeleted = nil
	
	if isOzCatalogDeleted then
		isOzCatalogDeleted:dispose()
	end

end

--------------------------------------------------------------------------------

-- SPAM [scouten 2015-07-21]: This is really only exposed for testing purposes.
-- Hide later?

function WFCatMetadataPuller:getUpdatesOnce()

	return getLatestUpdates( self, true )

end

--------------------------------------------------------------------------------

-- SPAM [jandrew 2016-04-05]: This is only exposed for testing.

function WFCatMetadataPuller:getLatestUpdates()

	return getLatestUpdates( self )

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:stopContinuousUpdates()

	self._syncIsStopped = true
	self._continuousPullTask:stopContinuousTask()

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:getUpdatesContinuously()

	self._syncIsStopped = false
	self._continuousPullTask:startContinuousTask()

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:isDoingWork()

	return self._continuousPullTask:isTaskDoingWork()

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:isOzCatalogDeleted()

	check( 'WFCatMetadataPuller' )
	
	return self._isOzCatalogDeleted
		:distinctUntilChanged()
		
end

--------------------------------------------------------------------------------

--[[!

	\brief Add a custom asset type handler to the WFCatMetadataPuller

	\param assetType (string) The oz asset type the custom handler will act upon.
	
	\param handlerFn( (function) A function that will be called on each asset
	of the given assetType.
	
	handlerFn takes two arguments:
		1. A WFCatMetadataPuller instance.
		2. The constructed doc value that has yet to be inserted into the doc store.
	
	Example handlerFn:
	
		local function addOpinionOnPieAnnotation( wfCatMetadataPuller, docValue )

			docValue._localOnly = { 
				opinionOnPie = "Pretty good."
			}

		end
	
	\return void
	
--]]

function WFCatMetadataPuller:addAssetTypeHandler( assetType, handlerFn )

	check( 'WFCatMetadataPuller', 'string', 'function' )

	if self._assetTypeHandler[ assetType ] then
	
		logger:warnf( "Now overwriting the existing asset type handler" )
	
	end
	
	self._assetTypeHandler[ assetType ] = handlerFn

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:getLatestDownloadedAssetIdsStream()

	return self._latestUpdatesStream
	
		:map( function( updates )
			
			local assetIds = {}
			
			for _, updateDoc in ipairs( updates ) do
			
				if isValidAssetDocument( updateDoc ) then
					table.insert( assetIds, updateDoc._id )
				end
			
			end
			
			return assetIds
			
		end )
		
end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:getDeletedDocumentIdStream()

	return self._deletedDocumentIdStream
  
end

--------------------------------------------------------------------------------
function WFCatMetadataPuller:getDiagnosticLogData()

	return {
		isCurrentlySyncing = not self._syncIsStopped,
		documentsReceivedByType = self._documentsReceivedByType,
		successfulUpdateCycles = self._successfulUpdateCycles,
		nextUpdateFeedLink = self._catalog:docStore():localDocValue( 'Oz-nextUpdateFeedLink' ),
		totalDocumentsReceived = self._totalDocumentsReceived,
		lastUpdateCycleWithData = self._lastUpdateCycleWithData,
		isOzCatalogDeleted = self._isOzCatalogDeleted,
	}

end

--------------------------------------------------------------------------------

function WFCatMetadataPuller:sendDiagnosticLogData()

	return {

		name = 'Metadata Sync Down: ' .. self._catalogSpec:catalogId(),
		id = 'WFCatMetadataPuller-' .. self._catalogSpec:catalogId(),
		weight = 500,
		body = self:getDiagnosticLogData()

	}

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatMetadataPuller'( WFCatMetadataPuller )
