--[[----------------------------------------------------------------------------

WFCatMockCatalog.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.self_test
local WFTestFileUtils = require 'WFTestFileUtils'

	-- WF.core.utils
local KSPathUtils = require 'KSPathUtils'
local NmbDeletionBacklog = require 'NmbDeletionBacklog'
local WFCatPixelPathGenerator = require 'WFCatPixelPathGenerator'
local WFCloseSupport = require 'WFCloseSupport'
local WFFileUtils = require 'WFFileUtils'

--==============================================================================

local WFCatMockCatalog = {}

--------------------------------------------------------------------------------

local kOzUserId = 'abad1dea00000000abad1dea00000000'
local kOzCatalogId = 'c1737ec7bfe942dd95d8c733f682b5df'

--------------------------------------------------------------------------------

local WFCatFileSpecMock = {
	_basePath = false,
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function WFCatFileSpecMock:__init( basePath )
	self._basePath = basePath
	WFFileUtils.mkdir( basePath )
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:pixelsPath() 
	local path =  KSPathUtils.child( self._basePath, 'renditionsPath' )
	WFFileUtils.mkdir( path )
	return path
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:catalogPath()
	local path = KSPathUtils.child( self._basePath, 'catalogPath' )
	WFFileUtils.mkdir( path )
	return path
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:docStorePath()
	return ':memory:'
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:fileAccessDbPath()
	return ':memory:'
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:rootPath( )
	return self._basePath
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:catalogPathInfo()
	
	local volumeGuid = WFFileUtils.getVolumeGuidForPath( self._basePath )
	local relativePath = WFFileUtils.removeVolumeFromPath( self._basePath )
	
	return {
		volumeGuid = volumeGuid,
		relativePath = relativePath,
	}
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:tempFolderPathForVolume()
	
	local tempFolder = KSPathUtils.getStandardFilePath( "temp" )
	pcall( WFFileUtils.mkdir, tempFolder )
	
	return tempFolder
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:originalsPathInfo()
	
	local volumeGuid = WFFileUtils.getVolumeGuidForPath( self._basePath )
	local relativePath = WFFileUtils.removeVolumeFromPath( self._basePath )
	
	return {
		volumeGuid = volumeGuid,
		relativePath = relativePath,
	}
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:proxiesPathInfo()

	local volumeGuid = WFFileUtils.getVolumeGuidForPath( self._basePath )
	local relativePath = WFFileUtils.removeVolumeFromPath( self._basePath )
	
	return {
		volumeGuid = volumeGuid,
		relativePath = relativePath,
	}
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:pathForOriginal( fileName, captureDate )
	
	return WFCatPixelPathGenerator.pathForOriginal( self._basePath, fileName, captureDate )
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:pathForProxy( assetId, proxyType )

	return WFCatPixelPathGenerator.pathForProxyTypeAndUuid( self._basePath, proxyType, assetId )
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:pathForXmpSidecar( assetId )

	return WFCatPixelPathGenerator.pathForXmpSidecar( self._basePath, assetId )

end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:folderForSettingsXmp( )

	return WFCatPixelPathGenerator.folderForSettingsXmp( self:pixelsPath() )

end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:pathForStyle( assetId )
	
	return WFCatPixelPathGenerator.pathForStyle( self:pixelsPath(), assetId )
	
end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:maybeStylePath( path )
	
	return WFCatPixelPathGenerator( self:pixelsPath(), path )

end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:pathForPreviewDigestAndLongEdge( uuid, digest, longEdge, optCreateDirectory )

	return WFCatPixelPathGenerator.pathForPreviewDigestAndLongEdge( self._basePath,
		uuid, digest, longEdge, optCreateDirectory )

end

--------------------------------------------------------------------------------

function WFCatFileSpecMock:close() 
	WFFileUtils.delete( self._basePath )
end

--------------------------------------------------------------------------------

WFCatFileSpecMock = Rx.Class( 'WFCatFileSpec', WFCatFileSpecMock )

--------------------------------------------------------------------------------

local function getUniqueBasePath( rxTest )

	return KSPathUtils.child(
				KSPathUtils.child( 
					KSPathUtils.getStandardFilePath( 'temp' ), 
					require 'KSIdGenerator'.uuidInLowercaseHex() ),
				rxTest:testName() )

end

--------------------------------------------------------------------------------

local fakeWFCatalog = Rx.Class 'WFCatalog' {
	_docStore = false,
	_indexStore = false,
	_fileAccessDb = false,
	_ozUserId = false,
	_searchSchema = false,
	_fileSpec = false,
	_catSpec = false,
	_volumeAgents = false,
	_deletionBacklog = false,
	
	WFCloseSupport.addCloseSupportMethods,

	__init = function( self, rxTest, ozUserId, optDocStorePath, optRequestResponsePairs )
		local WFRxDocStore = require 'WFRxDocStore'
		local WFRxIndexStore = require 'WFRxIndexStore'
		local NmbFileAccessDatabase = require 'NmbFileAccessDatabase'
		
		local basePath = getUniqueBasePath( rxTest )
		local fileSpec = WFCatFileSpecMock( basePath )
		
		local docStorePath = optDocStorePath or ':memory:'
		local docStore = WFRxDocStore( docStorePath )
		local indexStore = WFRxIndexStore( docStore, 'WFCatIndexDefinitions.lua' )
		local fileAccessDb = NmbFileAccessDatabase( docStore, indexStore, fileSpec:fileAccessDbPath(), { 
				writeInterval = 0.1,
				doNotInitialize = true,
				isDebugMode = true,
			} )
		
		local deletionBacklog = NmbDeletionBacklog( docStore, fileSpec:fileAccessDbPath() )
		
		local searchSchema = require 'WFCatSearchSchema'()
		require 'WFCatSearchHandler'.addSearchToSearchSchema( searchSchema )
		require 'WFCatIdleDownloader'.addSearchToSearchSchema( searchSchema )
		
		self._ozUserId = ozUserId
		self._searchSchema = searchSchema
		self._volumeAgents = {}
		
		self._catSpec = require 'WFCatSpecMock'( '98db98bfd554470dab2f862a4b853c40', '04c358b4a268f2bad074455dfb89bd89', rxTest, optRequestResponsePairs or {} )

		WFTestFileUtils.addTestFolder( basePath )

		self:setOwned {
			_docStore = docStore,
			_indexStore = indexStore,
			_fileAccessDb = fileAccessDb,
			_fileSpec = fileSpec,
			_deletionBacklog = deletionBacklog
		}

	end,
	docStore = function( self )
		return self._docStore
	end,
	indexStore = function( self )
		return self._indexStore
	end,
	fileAccessDb = function( self )
		return self._fileAccessDb
	end,
	ozUserId = function( self )
		return self._ozUserId
	end,
	ozCatalogId = function( self )
		return kOzCatalogId
	end,
	searchSchema = function( self )
		return self._searchSchema
	end,
	fileSpec = function( self )
		return self._fileSpec
	end,
	ozCatalogSpec = function( self )
		return self._catSpec
	end,
	relayObject = function( self )
		return self
	end,
	getAllVolumeAgents = function( self )
		return self._volumeAgents
	end,
	getDeletionBacklog = function( self )
		return self._deletionBacklog
	end
}

--------------------------------------------------------------------------------

function WFCatMockCatalog.createFakeWFCatalog( rxTest, optDocStorePath, optRequestResponsePairs )

	local result = fakeWFCatalog( rxTest, kOzUserId, optDocStorePath, optRequestResponsePairs )

	rxTest:finally( function()
		result:close()
	end )

	return result

end

--------------------------------------------------------------------------------

function WFCatMockCatalog.createWFCatalogForTestWithSpec( rxTest, wfCatalogSpec, optCatalogClassType )
	
	local basePath = getUniqueBasePath( rxTest )
	
	local optArgs = {
		docStorePath = ':memory:',
		fileAccessDbPath = ':memory:',
	}
	
	local catalog = require ( optCatalogClassType or 'WFCatalog' )( wfCatalogSpec, basePath, optArgs )

	rxTest:finally( function()
		catalog:close() -- closing the catalog will delete the basePath
	end )
	
	return catalog
end

--==============================================================================

return WFCatMockCatalog
