--[[----------------------------------------------------------------------------

WFCatOfflineAssets.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require 'Rx'

local Observable = Rx.Observable
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatDelete = require 'WFCatDelete'

local WFFileUtils = require 'WFFileUtils'
local WFRxBatchedBulkApplyOverlay = require "WFRxBatchedBulkApplyOverlay"

local runUntilDone = WFRxRunUntilDone


--==============================================================================

local WFCatOfflineAssets = {}

--------------------------------------------------------------------------------

local function validateImageIds( imageIds )

	local assetUuids
	
	if type( imageIds[ 1 ] ) == 'number' then

		for _, imageId in ipairs( imageIds ) do
			assert( type( imageId ) == 'number', 'All asset ids must be either local or global ids but not a mix' )
		end

		assetUuids = Observable.just( imageIds )

	elseif type( imageIds[ 1 ] ) == 'string' then

		for _, imageId in ipairs( imageIds ) do
			assert( type( imageId ) == 'string', 'All asset ids must be either local or global ids but not a mix' )
		end

		assetUuids = Observable.just( imageIds )

	else

		assert( false, 'Asset ids must all be uuid strings, or all be non-zero positive integers' )

	end

	return assetUuids

end

--------------------------------------------------------------------------------

function WFCatOfflineAssets.markAssetsForOfflineUse( wfCatalog, imageIds )

	local assetUuids = validateImageIds( imageIds )

	local docStore = wfCatalog:docStore()
	
	local deletionBacklog = wfCatalog:getDeletionBacklog()

	return assetUuids
	
		:flatMapLatest( function( assetUuids )
		
			return wfCatalog:docStore():bulkDocValue( assetUuids )
		
		end )
	
		:flatMapLatest( function( docValues )
			return Observable.from( docValues )
		end )
	
		:flatMapLatest( function( docValue )
		
			local hasLocalOriginalPath = WFCatFileSpec.buildOriginalPathFromDocValue( docValue )
			
			if hasLocalOriginalPath and not WFFileUtils.exists( hasLocalOriginalPath ) then
			
				-- If the local path is valid, but the asset doesn't exist, then delete
				-- that path annotation. This is to cover the case where the asset is
				-- manually deleted from disk and then the  user chooses to mark the asset
				-- for local use.

				return WFCatDelete.deleteLocalOriginalsByDocValues( docStore, { docValue }, deletionBacklog )
						:map( function()
							return docValue._id
						end )
				
			end
		
			return Observable.just( docValue._id )
		
		end )
		
		:map( function( docId )
		
			local baseOverlay = { _id = docId }
		
			baseOverlay._localOnly = {
				markedForOfflineUse = true,
			}
			
			return baseOverlay
			
		end )
		
		:toArray()
		
		:flatMapLatest( function( overlays )
			return WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), overlays )
		end )
		
		:takeLast( 0 )

end

--------------------------------------------------------------------------------

function WFCatOfflineAssets.disableOfflineUse( wfCatalog, imageIds )

	local assetUuids = validateImageIds( imageIds )

	return assetUuids:flatMapLatest( function( assetUuids )
			return Observable.from( assetUuids )
		end )
		:map( function( docId )
		
			local baseOverlay = { _id = docId }
		
			baseOverlay._localOnly = {
				markedForOfflineUse = { _delete = true },
			}
		
			return baseOverlay
			
		end )
		:toArray()
		:flatMapLatest( function( overlays )
			return WFRxBatchedBulkApplyOverlay( wfCatalog:docStore(), overlays )
		end )
		:takeLast( 0 )

end

--------------------------------------------------------------------------------

function WFCatOfflineAssets.offlineAssetsModel( wfCatalog, imageIds, enableOfflineMode )

	local value = wfCatalog
		:filter( function( wfCatalog )
			return wfCatalog ~= nil
		end )
		:take( 1 )
		:selectMany( function( wfCatalog )
			if enableOfflineMode then
				return WFCatOfflineAssets.markAssetsForOfflineUse( wfCatalog, imageIds )
			else
				return WFCatOfflineAssets.disableOfflineUse( wfCatalog, imageIds )
			end
		end )
		:call( runUntilDone )

	return {
		value = value,
	}

end

--==============================================================================

return WFCatOfflineAssets
