--[[----------------------------------------------------------------------------

WFCatSearch.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.rxlua
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.managed_catalog.search
local WFCatSearchQuery = require 'WFCatSearchQuery'

--==============================================================================

local WFCatSearch = {}

--------------------------------------------------------------------------------

function WFCatSearch.search( wfCatalog )

	local indexStore = wfCatalog:indexStore()
	return indexStore:query {
		sql = "SELECT docId FROM assets",
	}
	:map( function( rows )
		local assetIds = {}
		for i, row in ipairs( rows ) do
			assetIds[ i ] = row[ 1 ]
		end
		return {
			content = assetIds,
		}
	end )

end


local requiredFields =  require 'WFCatAsset'.getRequiredFields()


--------------------------------------------------------------------------------

function WFCatSearch.executeSearch( wfCatalog, delay, searchDescription, currentRequestedSortOrder, optInitialDelay )
	
	if currentRequestedSortOrder == nil then
		currentRequestedSortOrder = "DESC"
	end

 	local initialDelay = optInitialDelay or 0
	
	if wfCatalog == nil or ( delay == 1 and initialDelay == 0 ) then
		-- A delay of 1 and initialDelay of 0 means the caller does not want to 
		-- wait for the results and delay of 2 will trigger 'immediately' so 
		-- don't run a limit query now. Wait until the delay is 2 and we run the
		-- full query
		return Observable.never()
	end
	local docStore = wfCatalog:docStore()
	local indexStore = wfCatalog:indexStore()
	local searchSchema = wfCatalog:searchSchema()
	
	-- Rani [TODO] : https://jira.corp.adobe.com/browse/NMB-718
	-- Separate Asset Table Data and other filtering data
	
	local terms = require 'WFCatSearchQuery'.getSearchTerms( docStore, indexStore, searchSchema, searchDescription )

	-- Remove stacks from search results if the user has any non-collection filter applied.
	
 	local filterOutStacks = false

 	local predicateTerms = KSTableUtils.derefOpt( searchDescription, 'predicates', 'terms' ) 
			or { KSTableUtils.derefOpt( searchDescription, 'predicates' ) }
	
	for _, term in ipairs( predicateTerms ) do
		if not term.album then
			filterOutStacks = true
			break
		end
	end
	
	local maybeFilterOutStacksClause = ""
	if filterOutStacks then
		maybeFilterOutStacksClause = "WHERE subtype == 'image' or subtype == 'video'"
	end
	
	-- Search Limit
	
	local maybeLimit = ""
	if delay == 1 then
		maybeLimit = "LIMIT 200"
	end

	local sql = string.format( 
			[[SELECT %s 
				FROM assets 
				@JOIN
				%s
				ORDER BY captureDate %s, filenameLC COLLATE i18n_collate %s, docId %s %s]],
			requiredFields,
			maybeFilterOutStacksClause,
			currentRequestedSortOrder,
			currentRequestedSortOrder,
			currentRequestedSortOrder,
			maybeLimit )
			
	local assetsObservable = WFCatSearchQuery.query( indexStore, {
			sql = sql,
			join = terms,
			left = 'assets.docId',
		} )

	return assetsObservable

end

--------------------------------------------------------------------------------
--[[
-- 'basicSearch' submodel
function WFCatSearch.makeModel( wfCatalog, searchDescription, optInitialDelay )
	-- For performance reasons, you may want to send a limited number of assets matching 
	-- the search initially, then scale up to all results after initialDelay.	
	local initialDelay = optInitialDelay or 0

	local delayer
	if initialDelay > 0 then
		delayer = Observable.timer( 0, initialDelay, WFRxScheduler.sharedScheduler() )
	else
		delayer = Observable.from { 1, 2 }
	end

	local search = wfCatalog
	:combineLatest( delayer:take( 2 ) )
	:map( function( wfCatalog, delay )
		
		if wfCatalog == nil or ( delay == 1 and initialDelay == 0 ) then
			-- A delay of 1 and initialDelay of 0 means the caller does not want to 
			-- wait for the results and delay of 2 will trigger 'immediately' so 
			-- don't run a limit query now. Wait until the delay is 2 and we run the
			-- full query
			return Observable.never()
		end
		local docStore = wfCatalog:docStore()
		local indexStore = wfCatalog:indexStore()
		local searchSchema = wfCatalog:searchSchema()

		local terms = require 'WFCatSearchQuery'.getSearchTerms( docStore, indexStore, searchSchema, searchDescription )

		local limit = ""
		if delay == 1 then
			limit = " LIMIT 200"
		end

		return WFCatSearchQuery.query( indexStore, {
			sql = "SELECT " .. requiredFields .." FROM assets @JOIN ORDER BY captureDate DESC" .. limit,
			join = terms,
			left = 'assets.docId',
		} )
		:map( function( rows )
			local assetInfos = {}

			for i, row in ipairs( rows ) do
				assetInfos[ i ] = WFCatSearch.getAssetInfoDetails( row )
			end

			return assetInfos

		end )

	end )

	:switch()

	return {
		value = search,
	}

end
--]]
--------------------------------------------------------------------------------

function WFCatSearch.makeCountModel( wfCatalog, searchDescription, optInitialDelay )
	local initialDelay = optInitialDelay or 0
		-- Avoid hammering index store with count queries initially

	local delayer
	if initialDelay > 0 then
		delayer = Observable.timer( 0, initialDelay, WFRxScheduler.sharedScheduler() )
	else
		delayer = Observable.from { 1, 2 }
	end
		
	local search = wfCatalog
	:combineLatest( delayer:take( 2 ) )
	:map( function( wfCatalog, delay )
		if delay == 1 then
			return Observable.never()
		end
		if wfCatalog == nil then
			return Observable.just( 0 )
		end
		if searchDescription and ( searchDescription.operator or searchDescription.type ) then
			-- it wasn't wrapped as a searchDescription, this is just an expression
			searchDescription = {
				predicates = searchDescription
			}
		end

		local docStore = wfCatalog:docStore()
		local indexStore = wfCatalog:indexStore()
		local searchSchema = wfCatalog:searchSchema()

		local terms = require 'WFCatSearchQuery'.getSearchTerms( docStore, indexStore, searchSchema, searchDescription )

		return WFCatSearchQuery.count( indexStore, terms )

	end )

	:switch()

	return {
		value = search,
	}
end

--------------------------------------------------------------------------------

function WFCatSearch.makeQueryTermCategoriesModel( wfOzCatalog )
	return {
		value = {
			{
				term = 'general',
				ztitle = "$$$/WFCatSearch/TermCategories/General=General",
			},
			{
				term = 'picks',
				ztitle = "$$$/WFCatSearch/TermCategories/Flags=Flags",
			},
			{
				term = 'ratings',
				ztitle = "$$$/WFCatSearch/TermCategories/Ratings=Ratings",
			},
--			{
--				term = 'albums',
--				ztitle = "$$$/WFCatSearch/TermCategories/Albums=Collections",
--			},
--			{
--				term = 'dates',
--				ztitle = "$$$/WFCatSearch/TermCategories/Date=Dates",
--			},
--			{
--				term = 'keywords',
--				ztitle = "$$$/WFCatSearch/TermCategories/Keywords=Keywords",
--			},
--			{
--				term = 'locations',
--				ztitle = "$$$/WFCatSearch/TermCategories/Places=Places",
--			},
		}
	}
end

--------------------------------------------------------------------------------

local function _makeQueryTermsValue( wfCatalog, termCategory, searchDescription )
	local KSJsonEncoder = require 'KSJsonEncoder'

	local json = KSJsonEncoder.encodeWebSafeJson( searchDescription )
	local len = string.len( json ) + 1

	-- use a hyperbolic function based on the length of the search query
	-- to do build a fake count
	local count = math.floor( ( 999999 + len ) / len )

	local abortCode
		-- magic number used to figure out how many fake entries to include

	local terms = {}

	if termCategory == 'general' then
		terms[ #terms + 1 ] = {
			type = 'all',
			ztitle = "$$$/WFCatSearch/TermCategories/General/SubItem/All=All",
			expression = {
				type = 'all',
			}
		}
--		terms[ #terms + 1 ] = {
--			type = 'all',
--			ztitle = "$$$/WFCatSearch/TermCategories/General/SubItem/Photos=Photos",
--			expression = {
--				type = 'subtype',
--				subtype = 'jpeg',
--			}
--		}
--		terms[ #terms + 1 ] = {
--			type = 'all',
--			ztitle = "$$$/WFCatSearch/TermCategories/General/SubItem/Videos=Videos",
--			expression = {
--				type = 'subtype',
--				subtype = 'video',
--			}
--		}
--		terms[ #terms + 1 ] = {
--			type = 'all',
--			ztitle = "$$$/WFCatSearch/TermCategories/General/SubItem/LastImport=Last Import",
--			expression = {
--				type = 'importIndex',
--				importIndex = -1,
--			}
--		}
	end

	if termCategory == 'albums' then
		terms[ #terms + 1 ] = {
			type = 'album',
			label = 'My Album',
			expression = {
				type = 'album',
				album = 'abcd1234abcd1234abcd1234abcd1234',
			}
		}
		terms[ #terms + 1 ] = {
			type = 'album',
			label = 'My OtherAlbum',
			expression = {
				type = 'album',
				album = '1234abcd1234abcd1234abcd1234abcd',
			}
		}
	end

	if termCategory == 'ratings' and wfCatalog then
		local userId = wfCatalog:ozUserId()

		terms[ #terms + 1 ] = {
			type = 'rating',
			rating = 0,
			ztitle = "$$$/WFCatSearch/TermCategories/Ratings/Rating/0=No Stars",
			expression = {
				type = 'all',
			}
		}
		terms[ #terms + 1 ] = {
			type = 'rating',
			rating = 1,
			ztitle = "$$$/WFCatSearch/TermCategories/Ratings/Rating/1=One Star",
			expression = {
				type = 'rating',
				rating = 1,
				operator = '>=',
				user = userId,
			}
		}
		terms[ #terms + 1 ] = {
			type = 'rating',
			rating = 2,
			ztitle = "$$$/WFCatSearch/TermCategories/Ratings/Rating/2=Two Stars",
			expression = {
				type = 'rating',
				rating = 2,
				operator = '>=',
				user = userId,
			}
		}
		terms[ #terms + 1 ] = {
			type = 'rating',
			rating = 3,
			ztitle = "$$$/WFCatSearch/TermCategories/Ratings/Rating/3=Three Stars",
			expression = {
				type = 'rating',
				rating = 3,
				operator = '>=',
				user = userId,
			}
		}
		terms[ #terms + 1 ] = {
			type = 'rating',
			rating = 4,
			ztitle = "$$$/WFCatSearch/TermCategories/Ratings/Rating/4=Four Stars",
			expression = {
				type = 'rating',
				rating = 4,
				operator = '>=',
				user = userId,
			}
		}
		terms[ #terms + 1 ] = {
			type = 'rating',
			rating = 5,
			ztitle = "$$$/WFCatSearch/TermCategories/Ratings/Rating/5=Five Stars",
			expression = {
				type = 'rating',
				rating = 5,
				operator = '>=',
				user = userId,
			}
		}

	end

	if termCategory == 'keyword' then
		local keywords = {
			"Beach", "Bird", "Black", "Blue", "Cat", "Dog", "House", "Red",
			"Sunset", "Tree", "Treehouse", "White", "Yellow",
		}

		abortCode = 60000
		for i, keyword in ipairs( keywords ) do
			if count > abortCode then
				terms[ #terms + 1 ] = {
					type = 'keyword',
					label = keyword,
					expression = {
						type = 'keyword',
						keyword = keyword,
					}
				}
			end
			abortCode = abortCode / 2
		end
	end

	if termCategory == 'locations' then
		local WFZString = require 'WFZString'

		local places = {
			"St. Paul, MN", "San Fransisco, CA", "Seattle, WA"
		}

		abortCode = 20000
		for i, place in ipairs( places ) do
			if count > abortCode then
				terms[ #terms + 1 ] = {
					type = 'location',
					label = place,
					expression = {
						type = 'location',
						geotag = place,
					}
				}
			end
			abortCode = abortCode / 2
		end

		if count > 8000 then
			terms[ #terms + 1 ] = {
				type = 'location',
				label = WFZString.fromZString( "$$$/private/Search/Location=^0^d^1 ^2^d^3",
					45, 'N', '93', 'W' ),
				expression = {
					type = 'location',
					geohash = 'cbjb',
				}
			}
		end
	end

	if termCategory == 'picks' and wfCatalog then
		local searchFlags = {
			Flagged = 'pick',
			Unflagged = 'unflagged',
			Rejected = 'rejected'
		}
		local userId = wfCatalog:ozUserId()
		for _, flag in ipairs { "Flagged", "Unflagged", "Rejected" } do
			terms[ #terms + 1 ] = {
				type = 'picks',
				label = flag,
				expression = {
					type = 'review',
					flag = searchFlags[ flag ],
					operator = '==',
					user = userId,
				}
			}
		end
	end

	if termCategory == 'dates' then
		if count > 60000 then
			terms[ #terms + 1 ] = {
				type = 'dates',
				subtype = 'years',
				years = {
					2006, 2007, 2008, 2010, 2012, 2014, 2015
				}
			}
		elseif count > 2000 then
			terms[ #terms + 1 ] = {
				type = 'dates',
				subtype = 'months',
				year = 2015,
				months = {
					1, 2, 3, 5,
				}
			}
		elseif count > 100 then
			terms[ #terms + 1 ] = {
				type = 'dates',
				subtype = 'days',
				year = 2015,
				month = 3,
				days = {
					8, 9, 20, 21, 22
				}
			}
		end
	end

	local WFDigest = require 'WFDigest'

	for i, term in ipairs( terms ) do

		local termJson = KSJsonEncoder.encodeWebSafeJson( term )
		local digest = WFDigest.SHA1.digest( termJson )
		term.clientId = digest

	end

	return terms
end

function WFCatSearch.makeQueryTermsModel( wfCatalog, termCategory, searchDescription )

	return {
		value = wfCatalog:map( function( wfCatalog )
			return _makeQueryTermsValue( wfCatalog, termCategory, searchDescription )
		end ),
	}

end

--==============================================================================

return WFCatSearch
