--[[----------------------------------------------------------------------------

WFCatSearchByLocation.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSTableUtils = require 'KSTableUtils'
local KSStringUtils = require 'KSStringUtils'
--------------------------------------------------------------------------------

local derefOpt = KSTableUtils.derefOpt

local isValidSearchableAssetDocument = assert( require 'WFCatIndexUtils'.isValidSearchableAssetDocument )


--==============================================================================

local WFCatSearchByLocation = {}

--[[
	Index definitions for finding assets by location.  The SQLite index only;
--]]
--------------------------------------------------------------------------------
local function addLocation( currentItemMap, requiredItemsList, item )
	if not item then
		return
	end
	
	item  = KSStringUtils.trim( item )
	if string.len( item ) == 0 then
		return
	end
	
	local itemLc =  string.lower( item )
	if not currentItemMap[ itemLc ] then
		currentItemMap[ itemLc ] = true
		table.insert( requiredItemsList, item )
	end
end
--------------------------------------------------------------------------------

function WFCatSearchByLocation.generateIndexSchema( indexStoreSchema )

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE locations( location TEXT, lowerCaseLocation TEXT )]],

		useLocalId = true,
		docValueFilter = function( docValue )
			-- can probably be slightly more selective here
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )

			local isSearchable, searchableDoc = isValidSearchableAssetDocument( docValue )
			if isSearchable and searchableDoc then
				local requiredLocations = {}
				local currentValueMap = {}
				
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'xmp', 'iptcCore', 'Location' ) )
				
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'xmp', 'photoshop', 'Country' ) )
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'xmp', 'photoshop', 'State' ) )
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'xmp', 'photoshop', 'City' ) )
				
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'location', 'country' ) )
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'location', 'state' ) )
				addLocation( currentValueMap, requiredLocations, derefOpt( searchableDoc, 'location', 'city' ) )
				
				for _, location in ipairs( requiredLocations ) do
					
					local lowerCaseLocation = string.lower( location )
					emit( location, lowerCaseLocation )
					
				end
			end

		end,

	}

	indexStoreSchema:addIndex
		[[CREATE INDEX locations_location ON locations( location, docId )]]

	indexStoreSchema:addIndex
		[[CREATE INDEX locations_lowerCaseLocation ON locations( lowerCaseLocation, location )]]

end

--------------------------------------------------------------------------------

function WFCatSearchByLocation.addSearchToSearchSchema( searchSchema )

	searchSchema:addSearchableData {
		term = 'locations',
		dataType = 'oneToMany',
		sql = [[SELECT docId FROM locations WHERE lowerCaseLocation = LOWER(?) ORDER BY docId]],
		query_params = {
			'location',
		},
	}

end

--------------------------------------------------------------------------------

function WFCatSearchByLocation.collectLocationDetailsOfRequiredAssets( wfCatalog, assetIds )
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT location
                    FROM  TEMP_ids  
                    LEFT JOIN locations 
                    ON TEMP_ids.docId = locations.docId GROUP BY lowerCaseLocation
                    ORDER BY lowerCaseLocation ASC]],
		tempIds = assetIds,
		calm = 5,
	}
	
end

--------------------------------------------------------------------------------

function WFCatSearchByLocation.collectLocationDetailsOfAllAssets( wfCatalog )
	
	return wfCatalog:indexStore():query {
		sql = [[ SELECT location
			FROM locations GROUP BY lowerCaseLocation
			ORDER BY lowerCaseLocation ASC]],
		calm = 5,
	}

end

--==============================================================================

return WFCatSearchByLocation
