--[[----------------------------------------------------------------------------

WFCatSearchOnlineRequest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require "Rx"
local Disposable = Rx.Disposable
local Subject = Rx.Subject
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

local KSJsonEncoder = require "KSJsonEncoder"
local KSJsonDecoder = require "KSJsonDecoder"
local KSTableUtils = require "KSTableUtils"
local WFTableArchiving = require 'WFTableArchiving'

local WFCatSearchAndSortConstants = require "WFCatSearchAndSortConstants"
local WFCatOnlineSearchResultTableHandler = require 'WFCatOnlineSearchResultTableHandler'
local WFConfiguration = require "WFConfiguration"

	-- WF.core.logging
local WFLogger = require 'WFLogger'
local logger = WFLogger( 'WFCatSearchOnlineRequest' )
local WFSearchDebugLogging = require 'WFSearchDebugLogging'


--------------------------------------------------------------------------------
local CURRENT_ONLINE_ASSET_SOURCE = "lightroom"
local PAGE_SIZE = WFCatSearchAndSortConstants.PAGE_SIZE
--------------------------------------------------------------------------------
local OnlineSearchStatus = WFCatSearchAndSortConstants.OnlineSearchStatus

--============ Logging =======================
local _latestRequestCounter = 0
-- not sure of checking a logger active stauts, hence to prevent the unwanted logging process, this flag is used
local onlineRequestLoggingNeeded =  WFConfiguration.searchConfiguration.onlineRequestLoggingNeeded
local useV3Provider = WFConfiguration.searchConfiguration.useV3Provider

--------------------------------------------------------------------------------
local function search_debugPrint( messageType, ... )
	if messageType == "e" or onlineRequestLoggingNeeded then
		WFSearchDebugLogging.search_debugPrint( "WFCatSearchOnlineRequest", logger, messageType, ... )
	end
end

--------------------------------------------------------------------------------
local function getFormattedTimeForSearchDebug( )
	if onlineRequestLoggingNeeded then
		return WFSearchDebugLogging.getFormattedTimeForSearchDebug( "WFCatSearchAndSort" )
	else
		return ""
	end
end

--------------------------------------------------------------------------------

local WFCatSearchOnlineRequest = {
	_catalog = false,
	_catUser = false,
	
	_url = false,
	_searchOptions = false,
	_onlineSearchTerms = false,

	_isSearchActive = false,
	_currentStart = false,
	_limit = false,
	
	_totalAvailable = false,
	_totalReceived = false,
	_recievedAssets = false,
	_recievedAssetsMap = false,
	_currentReceivedCount = false,
	
	_searchResultSubject = false,
	_executionStatus = false,
	
	Disposable,
}
--============== Local methods ================--
-------------------- getLightroomDataFromSearchResult : V1 ---------------------------
local function getLightroomDataFromSearchResultV1( searchResult )
	local directories = searchResult and searchResult.directories
	if directories then
		for _,directory in pairs( directories ) do
			local name = directory.name
			if name == CURRENT_ONLINE_ASSET_SOURCE then
			
				if onlineRequestLoggingNeeded then	
					local searchResultData = directory.total_children
					search_debugPrint( "d", "Search : WFCatSearchOnlineRequest: result  total_items in current page V1" .. searchResultData , 
									_latestRequestCounter, getFormattedTimeForSearchDebug() , "\n" ) 
				end
				return directory.children, directory.total_children
			end
		end
	else
		if onlineRequestLoggingNeeded then	
			search_debugPrint( "d", "Search : WFCatSearchOnlineRequest: no directories in V1" , 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
		end
	end
end
-------------------- getLightroomDataFromSearchResult : V2 ---------------------------
local function getLightroomDataFromSearchResultV2( searchResult )
	local resultSets = searchResult and searchResult.result_sets
	if resultSets then
		for _,resultSet in pairs( resultSets ) do
			local name = resultSet.name
			if name == CURRENT_ONLINE_ASSET_SOURCE then
			
				local curentItemsCount = ( resultSet.items and #resultSet.items ) or 0
				local totalCount = resultSet.total_items
					
				if onlineRequestLoggingNeeded then	
					
					local message = string.format( 
						" Search : WFCatSearchOnlineRequest: result  total_items in current page V2 : ( % s ) , total items for the search V2 : ( %s )", 
							curentItemsCount, totalCount )
					search_debugPrint( "d", message , 
									_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n") 
				end
		
				return resultSet.items,  curentItemsCount, totalCount
			end
		end
	else
		if onlineRequestLoggingNeeded then	
			search_debugPrint( "d", "Search : WFCatSearchOnlineRequest: no result set V2" , 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
		end
	end
end

-------------------- getLightRoomDataFromSearchResult ---------------------------
local function getLightroomDataFromSearchResult( searchResult )
	if WFCatSearchAndSortConstants.SEARCH_VERSION ==  WFCatSearchAndSortConstants.SEARCH_VERSION_V2 then
		return getLightroomDataFromSearchResultV2( searchResult )
	end
	return getLightroomDataFromSearchResultV1( searchResult )
end



------------------------- getAssetId : V1 --------------------
local function getAssetIdV1( assetDetails )
	local assetType = assetDetails.type
	assert( assetType, "type detials is expected in online search results analyseSearchResultData " )
	
	local infoObj = assetDetails[ assetType ]
	assert( infoObj, "infoObj  is expected in online search results analyseSearchResultData " )
	
	return infoObj.id
end

------------------------- getAssetId : V2 --------------------
local function getAssetIdV2( assetDetails )
	local assetId =  assetDetails[ "asset_id" ]
	assert( assetId, "assetId is expected to be present in the V2 result as asset_id")
	return assetId
end


-------------------- getAssetId ---------------------------
local function getAssetId( assetDetails )
	if WFCatSearchAndSortConstants.SEARCH_VERSION ==  WFCatSearchAndSortConstants.SEARCH_VERSION_V2 then
		return getAssetIdV2( assetDetails )
	end
	return getAssetIdV1( assetDetails )
end

-------------------- addNextDataRangeTerms V2 ---------------------------
local function addNextDataRangeTermsV2( self )
	self._onlineSearchTerms["start_index" ] = self._currentStart
	self._onlineSearchTerms["limit" ] = self._limit 
end

-------------------- addNextDataRangeTerms : V1 ---------------------------
local function addNextDataRangeTermsV1( self )
	self._onlineSearchTerms[ "display" ] = { start_index = self._currentStart, limit = self._limit }
end


-------------------- addNextDataRangeTerms ---------------------------
local function addNextDataRangeTerms( self )
	if WFCatSearchAndSortConstants.SEARCH_VERSION ==  WFCatSearchAndSortConstants.SEARCH_VERSION_V2 then
		return addNextDataRangeTermsV2( self )
	end
	return addNextDataRangeTermsV1( self )
end


			
-------------------- initializeCommonTerms : V2 ---------------------------
local function initializeCommonTermsV2( self )
	self._onlineSearchTerms[ "scope" ] 				= 	{ CURRENT_ONLINE_ASSET_SOURCE  }
	self._onlineSearchTerms[ "sort_order" ] 		=  	"desc"
	self._onlineSearchTerms[ "sort_orderby" ] 		=  	"relevancy"
	self._onlineSearchTerms[ "fields_excluded" ] 	= 	{ "links", "tags" }
	
	-- explicitly redirect to LightroomProvider3 for additional facet keys support
	if useV3Provider then
		local lightroomVersion = { }
		lightroomVersion["lightroom_version"] = "3"
		self._onlineSearchTerms[ "hints" ]				=	self._onlineSearchTerms[ "hints" ] or { }
		table.insert(self._onlineSearchTerms[ "hints" ], lightroomVersion)
	end
	
end

-------------------- initializeCommonTerms : V1 ---------------------------
local function initializeCommonTermsV1( self )
	self._onlineSearchTerms[ "sources" ] = { CURRENT_ONLINE_ASSET_SOURCE  }
	self._onlineSearchTerms[ "fields_excluded" ] = { "links", "tags" }
end

-------------------- initializeCommonTerms ---------------------------
local function initializeCommonTerms( self )
	if WFCatSearchAndSortConstants.SEARCH_VERSION ==  WFCatSearchAndSortConstants.SEARCH_VERSION_V2 then
		return initializeCommonTermsV2( self )
	end
	return initializeCommonTermsV1( self )
end



-------------------- analyseSearchResultData ---------------------------
local function analyseSearchResultData( self, searchResult, requestTimeCounter )
	local ignoreCurrentResult = false
	if requestTimeCounter ~= _latestRequestCounter then
		-- we are not the latest request
		ignoreCurrentResult = true
		return nil, ignoreCurrentResult
	end
	
	local dataReceived  = false
	local metrics = searchResult.metrics
	if metrics then
	
		if onlineRequestLoggingNeeded then	
			local searchResultMetricsData = WFTableArchiving.tableToString( metrics )
			search_debugPrint( "d", "Search : WFCatSearchOnlineRequest: result  metrics" .. searchResultMetricsData , 
							_latestRequestCounter, getFormattedTimeForSearchDebug() ,"\n") 
		end
		
		-- since we have only one search scope,  metrics.total_hits is our total_hits
		-- hence not changing this to the page's totalCount
		if self._totalAvailable == false then
			self._totalAvailable = metrics.total_hits
		end
	end
	
	
	if self._totalAvailable > 0 then
		local lightroomResults, currentCount = getLightroomDataFromSearchResult( searchResult )
		currentCount = currentCount or 0
		self._currentReceivedCount = currentCount
		
		if currentCount > 0 then
			dataReceived = true
			self._totalReceived = self._totalReceived + currentCount
			local currentAssets = {}
			if not self._recievedAssets then
				self._recievedAssets = {}
			end
			
			if not self._recievedAssetsMap then
				-- currently there are some issues with Federer results
				-- in the successive page requests, we get the repeated data
				-- this results in empty grid.
				-- we are peventing the same here, by checking against present
				-- itemsMap
				self._recievedAssetsMap = {}
			end
			for _,assetDetails in ipairs( lightroomResults ) do
			
				local assetId = getAssetId( assetDetails )
				if not  self._recievedAssetsMap[ assetId ]  then
					table.insert( self._recievedAssets, assetId )
					table.insert( currentAssets, assetId )
					self._recievedAssetsMap[ assetId ] = true
				else
					search_debugPrint( "d", "Search : WFCatSearchOnlineRequest: received duplicate asset ".. assetId , 
							_latestRequestCounter, getFormattedTimeForSearchDebug() ,"\n") 
				end
				
			end
		
			if #currentAssets > 0 then
				WFCatOnlineSearchResultTableHandler.updateSearchTable( self._catalog, self._currentStart, 
					currentAssets )
			end
			
		end
		
	end
	-- if current count is 0, we report nil and handle this as completedUnexpectedly case 
	-- if total received is not same as total available
	return dataReceived, ignoreCurrentResult
end

-------------------- isFurtherRequestPossible ---------------------------
local function isFurtherRequestPossible( self )
	
	local totalReceived = self._totalReceived
	if ( self._currentReceivedCount % PAGE_SIZE ~= 0 ) then
		-- we have an issue of one item missing
		-- so temp fixing
		totalReceived = totalReceived + 1
	end
	
	
	return  totalReceived < self._totalAvailable 
end

-------------------- setNewSearchDataRange ---------------------------
local function setNewSearchDataRange( self )
	if self._totalAvailable  == false then
		-- we have not recieved the total count
		-- so let us initialize the detais
		self._currentStart = 0
		self._limit = PAGE_SIZE
		self._totalReceived = 0
	else
		if isFurtherRequestPossible( self ) then
			-- we have more elements to obtain
			self._currentStart = self._totalReceived  
			local endLimit = PAGE_SIZE -- math.min( self._totalReceived+PAGE_SIZE, self._totalAvailable )
			self._limit = endLimit
		end
	end
	self._currentReceivedCount  = 0
	addNextDataRangeTerms( self ) 
end

-------------------------- handleSearchCompletion -------------------------------------------
local function sendSearchCompletion( self, status )
	-- TBD handle the status
	self._isSearchActive = false
	self._searchResultSubject:onNext( status ) 
	self._searchResultSubject:onCompleted()
end
	
-------------------------- handleSearchCompletion -------------------------------------------
local function sendSearchStatus( self, status, isCompleted )
	self._executionStatus = status
	if isCompleted then
		self._isSearchActive = false
	end
	
	self._searchResultSubject:onNext(  )
	
	if isCompleted then
		self._searchResultSubject:onCompleted(  )
	end
end
		

-------------------------- startNextSetSearch -------------------------------------------
local function startNextSetSearch( self )

	local currentCounter = _latestRequestCounter
	if not self._searchResultSubject then
		self._searchResultSubject  = Subject( )
	end
	
	if self._catUser then
		setNewSearchDataRange( self )
		
		local searchOptions = KSTableUtils.deepCopy( self._searchOptions )
		
		local searchTerms = KSTableUtils.deepCopy( self._onlineSearchTerms )
		-- this is an internal data field so, let us remove before sending to USS
		searchTerms.isNetworkAvailable = nil
		
		if onlineRequestLoggingNeeded then
			local startString = "\n----------------------------------\n"
			
			if _latestRequestCounter == 1 then
				search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: url \n" .. self._url .. "\n", 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
			end
			
							
			local searchInputData = WFTableArchiving.tableToString( searchTerms )
			
			search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: requestData \n" .. searchInputData .. "\n", 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
							
		end
		
		local jsonData = KSJsonEncoder.prettyPrint( searchTerms )
						
		local dataObservable = self._catUser:put( self._url, jsonData, searchOptions )
		
		dataObservable : map( function( searchResult )
			local didResultAnalysisSucceed, message = pcall ( function ()
				if  self._isSearchActive then
					local success, searchResultTable = pcall( KSJsonDecoder.decodeWebSafeJson, searchResult )
					local ignoreCurrentResult = false
					if success then
						local dataReceived
						dataReceived , ignoreCurrentResult = analyseSearchResultData( self, searchResultTable, currentCounter ) 
						if dataReceived then
							
							-- we received the data, so sending progress
							sendSearchStatus( self, OnlineSearchStatus.InProgress , false )
							
							if isFurtherRequestPossible( self ) then
								-- we have more data, so place the next page request
								startNextSetSearch( self )
							else
								-- all data receieved completly
								sendSearchStatus( self, OnlineSearchStatus.Completed , true )
							end
						else
							if not ignoreCurrentResult then
								-- no data received
								-- if available vs received if there is an issue, let us send CompletedUnexpectedly
								if self._totalAvailable ~= self._totalReceived then
									sendSearchStatus( self, OnlineSearchStatus.CompletedUnexpectedly , true )
								else
									sendSearchStatus( self, OnlineSearchStatus.Completed , true )
								end
							end
						end
						
					else
						sendSearchCompletion( self, false )
					end
				end
			end )
			if not didResultAnalysisSucceed then
				search_debugPrint( "d","\n********** IMP ERROR IN PROCESSING ***************\n Search : WFCatSearchOnlineRequest: processing result error \n" .. message .. "\n", 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
			end
			
		end)
		:tapOnError( function (response, headers )
			local decoded, responseTable = pcall( KSJsonDecoder.decodeWebSafeJson, response )
			
			if decoded then
				-- TODO [anmurark 2018-05-17]: Take action based on the response error code
				
				local searchResponseData = WFTableArchiving.tableToString( responseTable )
				local startString = "\n----------------------------------\n"
				search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: responseError \n" .. searchResponseData .. "\n", 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 

			end
			
			sendSearchStatus( self, OnlineSearchStatus.CompletedUnexpectedly , true )

		end)
		:call( WFRxRunUntilDone ) 
	end
	
	return self._searchResultSubject
end

-------------------------- startFacetValuesSearch -------------------------------------------
local function startFacetValuesSearch( self )
	
	self._searchResultSubject = Subject( )
	
	if self._catUser then
	
		-- Set limit to 0 to ignore items in the response
		self._limit = 0	
		self._currentStart = 0
		addNextDataRangeTerms( self )
		
		local searchOptions = KSTableUtils.deepCopy( self._searchOptions )
		local searchTerms = KSTableUtils.deepCopy( self._onlineSearchTerms )
		
		if onlineRequestLoggingNeeded then
			local startString = "\n----------------------------------\n"
			
			search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: url \n" .. self._url .. "\n", 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
			
			local searchInputData = WFTableArchiving.tableToString( searchTerms )
			
			search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: requestFacetsData \n" .. searchInputData .. "\n", 
							_latestRequestCounter, getFormattedTimeForSearchDebug(), "\n" ) 
		
		end
		
		local jsonData = KSJsonEncoder.prettyPrint( searchTerms )
		
		local dataObservable = self._catUser:put( self._url, jsonData, searchOptions )
		
		dataObservable : map( function( searchResult )
			local success, searchResultTable = pcall( KSJsonDecoder.decodeWebSafeJson, searchResult )
			local facetsArray = {}
			if success then
			
				local resultSets = searchResultTable.result_sets
				if resultSets then
					for _, resultSet in ipairs( resultSets ) do
						local name = resultSet.name
						if name == CURRENT_ONLINE_ASSET_SOURCE then
							
							facetsArray = resultSet.facets
							
							if onlineRequestLoggingNeeded then
								local startString = "\n----------------------------------\n"
								local facetsData = WFTableArchiving.tableToString( facetsArray )
								search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: facetsData response \n" .. facetsData .. "\n", 
										getFormattedTimeForSearchDebug(), "\n" )
							end
						end
					end
				end
			end
			
			self._searchResultSubject:onNext( facetsArray )
			self._searchResultSubject:onCompleted()
			
		end )
		:tapOnError( function( response, headers )
			
			if onlineRequestLoggingNeeded then
				local startString = "\n----------------------------------\n"
				local _, responseTable = pcall( KSJsonDecoder.decodeWebSafeJson, response )
				local searchResponseData = WFTableArchiving.tableToString( responseTable )

				search_debugPrint( "d", startString .. "Search : WFCatSearchOnlineRequest: facetsData response \n" .. searchResponseData .. "\n", 
						getFormattedTimeForSearchDebug(), "\n" )
			end

			self._searchResultSubject:onNext( {} )
			self._searchResultSubject:onCompleted()
			
		end)
		:call( WFRxRunUntilDone ) 
	end

	return self._searchResultSubject

end

--============== API's ================--
------------------- __init --------------------------------------
function WFCatSearchOnlineRequest:__init( userGivenSearchTerms )
	self._onlineSearchTerms = userGivenSearchTerms
	initializeCommonTerms( self )
	self._totalReceived = 0
end

--[[ wiki links : 
V2: https://wiki.corp.adobe.com/display/adobesearch/Lightroom+Search+V2+Request+and+Response

V3: https://wiki.corp.adobe.com/display/adobesearch/Lightroom+Facets+Design
    https://wiki.corp.adobe.com/display/adobesearch/Facet+Field+Info+Design

]]
---------------------- executeOnlineSearch ----------------------------------
function WFCatSearchOnlineRequest:executeOnlineSearch( catalog, catUser, url, searchOptions )
	self._catalog = catalog
	self._catUser = catUser
	self._url = url
	self._searchOptions = KSTableUtils.deepCopy( searchOptions )
	WFCatOnlineSearchResultTableHandler.clearSearchTable( self._catalog )
	self._isSearchActive = true
	
	_latestRequestCounter = _latestRequestCounter + 1
	return startNextSetSearch( self )
end

---------------------- executeFacetValuesSearch ----------------------------------
function WFCatSearchOnlineRequest:executeFacetValuesSearch( catalog, catUser, url, searchOptions )
	self._catUser = catUser
	self._url = url
	self._searchOptions = KSTableUtils.deepCopy( searchOptions )
	
	return startFacetValuesSearch( self )
end

-------------------------- closeRequest -------------------------------------
function WFCatSearchOnlineRequest:closeRequest(  )
	self._onlineSearchTerms = false
	self._currentStart = false
	self._limit = false
	self._totalAvailable = false
	self._totalReceived = 0
	self._url = false
	self._searchOptions = false
	self._searchResultSubject:dispose()
	self._searchResultSubject = false
end

-------------------------- getCurrentStatus -------------------------------------
function WFCatSearchOnlineRequest:getCurrentStatus(  )
	local status = {}
	status.latestRequestStart 	=  self._currentStart
	status.latestRequestEnd 	=  self._limit
	status.totalReceived 		=  self._totalReceived
	status.totalAvailable 		=  self._totalAvailable
	status.executionStatus		=  self._executionStatus
	return status
end

return Rx.Class 'WFCatSearchOnlineRequest'( WFCatSearchOnlineRequest )
