--[[----------------------------------------------------------------------------

WFCatSearchQuery.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require 'Rx'

local WFLogger = require 'WFLogger'

local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'
local WFRxScheduler = require 'WFRxScheduler'

local Observable = Rx.Observable

--------------------------------------------------------------------------------

local logger = WFLogger 'WFCatSearchQuery'
--	:setLevel 'debug'

local check = assert( WFArguments.check )
local derefOpt = KSTableUtils.derefOpt


--==============================================================================

local WFCatSearchQuery = {}

--------------------------------------------------------------------------------

local intOrAlphaOperators = {
	['<'] = true,
	['<='] = true,
	['=='] = true,
	['>='] = true,
	['>'] = true,
	['!='] = true,
}

local function intOrAlphaTerm( indexStore, searchDescription, predicate )

	local sql_table = searchDescription.sql_table
	local column = searchDescription.column
	local operator = predicate.operator
	assert( intOrAlphaOperators[ operator ], 'Operator "' .. tostring( operator ) .. '" is not a valid operator' )

	local queryParams = {
		sql = string.format( [[SELECT docId FROM %s WHERE %s %s ? ORDER BY docId]],
			sql_table,
			column,
			operator ),
	}

	for i, termKey in ipairs( searchDescription.query_params or { searchDescription.term } ) do
		queryParams[ i ] = predicate[ termKey ]
	end

	return WFCatSearchQuery.makeTerm( indexStore, queryParams )

end

--------------------------------------------------------------------------------

local function fullTextTerm( docStore, indexStore, sql, text )
	return WFCatSearchQuery.makeTerm( indexStore, {
		sql = sql,
		text
	} )
end


--------------------------------------------------------------------------------

local function reviewTerm( docStore, indexStore, sql, values )
	if type(values) == "table" then
		local n = #values
		
		return WFCatSearchQuery.makeTerm( indexStore, {
			sql = sql,
			 unpack( values, 1, n )
		} )
	else
		-- we can get the number or string equivlanet
		local numberEq  = tonumber(values)
		if numberEq == nil then
			numberEq = require 'WFCatSearchByReview'.getValueForReviewStatus(values)
		else
			numberEq = values
		end
		return WFCatSearchQuery.makeTerm( indexStore, {
			sql = sql,
			numberEq
			} )
	end
	
end


--------------------------------------------------------------------------------

local function ratingTerm( docStore, indexStore, searchDescription, predicate )

	local sql = searchDescription.sql
	--[[
	Rani : other operators such as <= and >= are also supported
	if predicate.operator and predicate.operator ~= '==' then
		assert( intOrAlphaOperators[ predicate.operator ], 'operator invalid for SQL' )
	end
	--]]
	
	sql = string.replace( sql, '%OPERATOR', predicate.operator or '==' )
		-- For data like ratings, favorites and reviews, we need to both
		-- include the user as part of the query, but may want to do
		-- logical operations on the other part (e.g. rating > some value).
		
	local valueKey = searchDescription.term
	local value = predicate[ valueKey ]
	
	return WFCatSearchQuery.makeTerm( indexStore, {
		sql = sql,
		value
	} )
	
end



--------------------------------------------------------------------------------

local function dateInfoDynamicTerm( docStore, indexStore, searchDescription, predicate )

	local columnInfo = predicate.columnInfo
	local sql_table = searchDescription.tableName
	
	assert ( columnInfo , 'ColumnInfo expected' )
	
	local n = 0
	local searchConditions = ""
	local values = {}
	for columnName, columnValue in pairs ( columnInfo ) do
	
		assert( searchDescription.allowedColumnNames[ columnName ], 'Column "' .. 
				columnName .. '" is not a valid for dateRangeOperations' )
		
		if n > 0 then
			searchConditions = searchConditions .. ' and '
		end
		
		table.insert( values, columnValue )
		searchConditions = searchConditions .. columnName .. " == ?"
				
		n = n + 1
	end
	
	local sql  = string.format( [[SELECT docId FROM %s WHERE %s  ORDER BY docId]],
			sql_table,
			searchConditions )
	
	return WFCatSearchQuery.makeTerm( indexStore, {
		sql = sql,
		unpack( values, 1, n ) 
	} )
	
end


--------------------------------------------------------------------------------

local function dateRangeTerm( docStore, indexStore, searchDescription, predicate )

	local sql = searchDescription.sql	
	local columnName = predicate.column
	assert( searchDescription.allowedColumnNames[ columnName ], 'Column "' .. columnName .. '" is not a valid for dateRangeOperations' )
	
	
	sql = string.replace( sql, '%COLUMN_NAME', columnName )
	local query_terms = searchDescription.query_params
	
	local values = {}
	local n = 0
	for _, term in ipairs ( query_terms ) do
		local termVal = predicate[ term ]
		if termVal then
			n = n + 1
			values [ n ] =  termVal
		end
	end
	
	return WFCatSearchQuery.makeTerm( indexStore, {
		sql = sql,
		unpack( values, 1, n ) 
	} )
	
end

--------------------------------------------------------------------------------

local function _someToManyTerm( docStore, indexStore, sql, ... )

	local firstValue = ( ... )
	if type( firstValue ) == 'number' then
		local xtra
		local nXtra = select( '#', ... ) - 1
		if nXtra > 0 then
			xtra = { select( 2, ... ) }
		end

		-- lookup local id for this document and switch param to guid based id.
		return docStore:docValue( firstValue )
			:flatMap( function( docValue )

				-- Note: if docValue is _empty we can't convert the id to a guid
				-- so _id will still be the local id, a number value.

				local value = docValue._id
				if xtra then
					return WFCatSearchQuery.makeTerm( indexStore, {
						sql = sql,
						value, unpack( xtra, 1, nXtra )
					} )
				else
					return WFCatSearchQuery.makeTerm( indexStore, {
						sql = sql,
						value
					} )
				end

			end )
	end

	return WFCatSearchQuery.makeTerm( indexStore, {
		sql = sql,
		...
	} )

end

--------------------------------------------------------------------------------

local function someToManyTerm( docStore, indexStore, searchDescription, predicate )

	local sql = searchDescription.sql
	if predicate.operator and predicate.operator ~= '==' then
		assert( intOrAlphaOperators[ predicate.operator ], 'operator invalid for SQL' )
	end
	sql = string.replace( sql, '%OPERATOR', predicate.operator or '==' )
		-- For data like ratings, favorites and reviews, we need to both
		-- include the user as part of the query, but may want to do
		-- logical operations on the other part (e.g. rating > some value).
	if searchDescription.query_params then
		local values = {}
		local n = #searchDescription.query_params
		local mappings = searchDescription.values or {}
		for index, key in ipairs( searchDescription.query_params ) do
			local value = predicate[ key ]
			if mappings[ key ] then
				value = mappings[ key ][ value ]
			end
			values[ index ] = value
		end
		return _someToManyTerm( docStore, indexStore, sql, unpack( values, 1, n ) )
	end
	local valueKey = searchDescription.term
	local value = predicate[ valueKey ]
	return _someToManyTerm( docStore, indexStore, sql, value )

end

--------------------------------------------------------------------------------

local syncStatusOperators = {
	['IN'] = true,
	['NOT IN'] = true,
}

--------------------------------------------------------------------------------

local WFCatSearchBySyncStatus = require 'WFCatSearchBySyncStatus'

local function syncStatusTerm( docStore, indexStore, searchDescription, predicate )
	
	if predicate.operator then
		assert( syncStatusOperators[ predicate.operator ], 'operator invalid for SQL' )
	end

	local sql = string.replace( searchDescription.sql, '%OPERATOR', predicate.operator or 'IN' )
	
	local syncFilterValue = searchDescription.query_params and predicate[ searchDescription.query_params[ 1 ] ]
	
	if WFCatSearchBySyncStatus.isMarkedForOfflineFilter( syncFilterValue ) then
	
		sql = WFCatSearchBySyncStatus.getMarkedForOfflineSql() 
	
	else
	
		local syncStatusValues = WFCatSearchBySyncStatus.getStatusIdsFromFilterValue( syncFilterValue )
		
		local syncStatus = " ( "
		
		for _, syncStatusId in ipairs( syncStatusValues ) do
			syncStatus = syncStatus .. tostring( syncStatusId )
			syncStatus = syncStatus .. ", "
		end

		if #syncStatusValues > 0 then
			syncStatus = syncStatus:sub( 1, -3 )  -- remove trailing ', '	
		end
		
		syncStatus = syncStatus .. " ) "
		
		sql = string.format( sql, syncStatus )
	
	end
	
	return WFCatSearchQuery.makeTerm( indexStore, {
		sql = sql,
	} )
	
	
end

--------------------------------------------------------------------------------

local function gatherTerms( docStore, indexStore, searchSchema, predicate )

	local searchType = predicate and predicate.type

	if searchType == 'all' or predicate == nil then
		return
	end

	local compoundTermFn
	if searchType == 'or' then
		compoundTermFn = WFCatSearchQuery.makeCompoundOrTerm
	end
	if searchType == 'and' then
		compoundTermFn = WFCatSearchQuery.makeCompoundAndTerm
	end
	if compoundTermFn then
		local terms = assert( predicate.terms, '"' .. searchType .. '" query type requires "terms" array' )
		local subTerms = {}
		for index, subpredicate in ipairs( terms ) do
			subTerms[ #subTerms + 1 ] = gatherTerms( docStore, indexStore, searchSchema, subpredicate )
		end
		if #subTerms > 1 then
			return compoundTermFn( indexStore, unpack( subTerms ) )
		else
			return subTerms[ 1 ]
		end
	end

	local searchDescription = searchSchema:getSearchDescription( searchType )
	if not searchDescription then
		return nil, 'Invalid search type: ' .. tostring( searchType or '< missing >' )
	end

	local dataType = searchDescription.dataType

	if dataType == 'noOp' then
		return
	elseif dataType == 'uint' or dataType == 'alpha' or dataType == 'int' then
		return intOrAlphaTerm( indexStore, searchDescription, predicate )
	end
	if dataType == 'manyToMany' or dataType == 'oneToMany' then
		return someToManyTerm( docStore, indexStore, searchDescription, predicate )
	end
	if dataType == 'fullTextSearch' then
		return fullTextTerm( docStore, indexStore, searchDescription.sql, predicate[ searchDescription.query_params ] )
	end
	if dataType == 'reviewTerm' then
		return reviewTerm( docStore, indexStore, searchDescription.sql, predicate[ searchDescription.query_params[1] ] )
	end
	if dataType == 'ratingTerm' then
		return ratingTerm( docStore, indexStore, searchDescription, predicate)
	end
	if dataType == 'dateRangeTerm' then
		return dateRangeTerm( docStore, indexStore, searchDescription, predicate)
	end
	if dataType == 'dateInfoDynamicTerm' then
		return dateInfoDynamicTerm( docStore, indexStore, searchDescription, predicate)
	end
	if dataType == 'syncStatusTerm' then
		return syncStatusTerm( docStore, indexStore, searchDescription, predicate )
	end
	
	
	return nil, 'unsupported dataType: ' .. tostring( dataType or '< unspecified >' )

end

--------------------------------------------------------------------------------

--[[ Search Description is a search as defined here: http://belt.sea.adobe.com/asset/search.html

	searchDesc = {
		predicates = {
			type = '< predicate type >',
			... other key/value pairs specific to this type
		}
	}

	@return an Observable with the search results
]]

function WFCatSearchQuery.getSearchTerms( docStore, indexStore, searchSchema, searchDescription )

	logger:infof( "getSearchTerms searchDescription = %s\n\nvia %s\n\n",
				  searchDescription, ':traceback:' )

	local predicates = searchDescription and searchDescription.predicates

	local terms, maybeError = gatherTerms( docStore, indexStore, searchSchema, predicates )

	if not terms and maybeError then
		error( maybeError )
	end

	return terms

end

--------------------------------------------------------------------------------

local function makeCompoundTerm( operator, ... )

	local n = select( '#', ... )
	return Observable.combineLatest( function( ... )
		return {
			operator = operator,
			n = n,
			...
		}
	end, ... )

end

--------------------------------------------------------------------------------

-- date test w 10,000 -> 0.022
--==============================================================================

function WFCatSearchQuery.makeTerm( indexStore, inQueryParams )

	local queryParams = KSTableUtils.shallowcopy( inQueryParams )

	return Observable.just( queryParams )

end

--------------------------------------------------------------------------------

function WFCatSearchQuery.makeCompoundOrTerm( indexStore, ... )

	return makeCompoundTerm( 'or', ... )

end

--------------------------------------------------------------------------------

function WFCatSearchQuery.makeCompoundAndTerm( indexStore, ... )

	return makeCompoundTerm( 'and', ... )

end

--------------------------------------------------------------------------------

local function virtualTableForCompoundTerm( indexStore, term )
	if not term then
		return
	end
	if term.sql then
		term = {
			operator = 'or',
			term,
		}
	end
	return indexStore:newVirtualTableForSetQuery( term )
end

--------------------------------------------------------------------------------

function WFCatSearchQuery.count( indexStore, searchTerm )

	check( 'WFRxIndexStore', '?table' )

	logger:infof( "count searchTerm = %s\n\nvia %s\n\n",
				  searchTerm, ':traceback:' )

	return ( searchTerm or Observable.just() )

		:debounce( 0.1, WFRxScheduler.sharedScheduler() )
		:map( function( predicate )
			if predicate then
				return indexStore:querySetLogic {
					predicate
				}
				:map( function( resultSet )
					return resultSet.length
				end )
			else
				return indexStore:query {
					sql = "SELECT count( docId ) FROM assets WHERE isMedia = 1"
				}
				:map( function( rows )
					return derefOpt( rows, 1, 1 )
				end )
			end
		end )
		:switch()

end

--------------------------------------------------------------------------------

--[[
	\brief Perform a query using ids from a result set.  This is much like
	index store query function, except the query may contain a special
	phrase '@JOIN' which will be replaced by the query system with a join
	based on doing virtual set queries.  The table should also include these keys:
		join: (a search term) from WFCatSearchQuery.getSearchTerms
		left: (string) the left side of the join
	
	\param indexStore (WFRxIndexStore) : The index store to query
	\param inQueryParams (table) : Same table as indexStore:query() but
		should also include 
--]]
function WFCatSearchQuery.query( indexStore, inQueryParams )

	check( 'WFRxIndexStore', 'table' )

	logger:infof( "query %s\n\nvia %s\n\n", inQueryParams, ':traceback:' )

	assert( type( inQueryParams.sql ) == 'string', 'SQL must be provided' )

	local queryParams = KSTableUtils.shallowcopy( inQueryParams )

	queryParams.calm = queryParams.calm or 2
		-- By default, calm all search queries so that the result _at most_ 
		-- refreshes once every couple of seconds.
		
	if queryParams.calm == 0 then
		queryParams.calm = nil
	end

	local searchTerm = inQueryParams.join
	local leftTableAndField = assert( type( inQueryParams.left ) == 'string' and inQueryParams.left, 'left must be a string' )

	local joinSql
		-- an Observable

	if searchTerm then
		joinSql = searchTerm
		:debounce( 0.1, WFRxScheduler.sharedScheduler() )
		:map( function( tableNameOrCompoundTerm )
			return virtualTableForCompoundTerm( indexStore, tableNameOrCompoundTerm )
		end )
		:switch()
		:map( function( tableName )
			if tableName then
				return string.format( 'JOIN %s AS vt ON %s == vt.id', tableName, leftTableAndField )
			else
				return ''
			end
		end )
	else
		joinSql = Observable.just( '' )
	end

	return joinSql:map( function( joinSql )
		local sql = queryParams.sql:replace( "@JOIN", joinSql )
		queryParams.sql = sql

		return indexStore:query( queryParams )
	end )

	:switch()

	:tapOnCompleted( function()
		searchTerm:close()
	end )

end

--==============================================================================

return WFCatSearchQuery
