--[[----------------------------------------------------------------------------

WFCatStorage.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'
local Observable = Rx.Observable
local BehaviorSubject = Rx.BehaviorSubject

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFLogger = require 'WFLogger'
local WFConfiguration = require 'WFConfiguration'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.oz_client.managed_catalog
local WFCatOzErrors = require 'WFCatOzErrors'
	
--------------------------------------------------------------------------------

local DEBUG_STORAGE_USED = WFConfiguration.Debug.debugStorageUsage
local debugStorageUsedStream

local check = assert( WFArguments.check )

local logger = WFLogger( 'WFCatStorage' )

local kLowOnStoragePercent = .9
local kOutOfStoragePercent = .99

--------------------------------------------------------------------------------

local function quota_exceeded_table( )
	
	return {
	
		sql = [[CREATE TABLE quota_exceeded( docType )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		
		map = function( docValue, emit )
			
			if WFCatOzErrors.hasStorageExceededError( docValue ) and not docValue._deleted then
				emit( docValue.type )				
			end
			
		end,
	}

end

--------------------------------------------------------------------------------

local function generateIndexSchema( _, indexStoreSchema )

	check( '_', 'WFRxIndexStoreSchema' )
	
	indexStoreSchema:addTable( quota_exceeded_table( ) )

end

--------------------------------------------------------------------------------

local function isAccountDocOverStorage( accountDoc )
	
	local maybeOzStorage = KSTableUtils.derefOpt( accountDoc, 'entitlement', 'storage' )
	return maybeOzStorage and ( maybeOzStorage.used > maybeOzStorage.limit )

end

--------------------------------------------------------------------------------

local function countOverOzStorageQuota( wfCatalog, optCalm )

	return wfCatalog:indexStore():query( {
		sql = "SELECT COUNT( docId ) FROM quota_exceeded",
		calm = optCalm or 60
	} )
	:map( function( results ) 
		
		return KSTableUtils.derefOpt( results, 1, 1 ) or 0
		
	end )
	
end

--------------------------------------------------------------------------------

local function docsOverOzStorageQuota( catalog, optCalm )
	
	return catalog:indexStore():query( { sql = [[SELECT docId FROM quota_exceeded]], calm = optCalm } )
		:map( function( docIdRows )
			
			local docIds = {}
			for _, row in ipairs( docIdRows ) do
				table.insert( docIds, row[ 1 ] )
			end
			
			return docIds
		end )

end

--------------------------------------------------------------------------------

local function resetQuotaExceeded( catalog )

	docsOverOzStorageQuota( catalog, 60 )
		:take( 1 )
		:flatMap( function( docIds )
			
			local bulkOverlay = {}

			for _, docId in ipairs( docIds ) do 
				table.insert( bulkOverlay, WFCatOzErrors.createOverlayToRemoveOzError( docId ) )
			end
			
			return catalog:docStore():bulkApplyOverlay( bulkOverlay )
			
		end )
		:call( WFRxRunUntilDone )
end

--------------------------------------------------------------------------------

local function storageEntitlement( wfCatalog )
	
	check( 'WFCatalog' )
	
	return wfCatalog:ozUser():entitlement()
		:filter( function( value ) return not not value end )
		:map( function ( value )
			return value
		end )
		:pluck( 'storage' )
end

--------------------------------------------------------------------------------

local function watchForQuotaReset( catalog )

	return Observable.combineLatest( storageEntitlement( catalog ), 
											countOverOzStorageQuota( catalog ) )
		:map( function( storage, docsQuotaExceeded )
			
			-- check local cache of account document to see whether we think there's storage 
			-- available.
			return storage.used < storage.limit and docsQuotaExceeded > 0
			
		end )
		:sample( 30, WFRxScheduler.sharedScheduler() ) -- emit no more than every 30s
		:flatMapLatest( function( shouldResetBasedOnLocalInfo )
			
			if shouldResetBasedOnLocalInfo then
				-- So far, we think we should reset the quota. But this is based on the local
				-- understanding of our storage usage on Oz. This might be out of date (NMB-6749)
				-- so get the latest account doc from Oz and see if it agrees in terms of having 
				-- available space.
				return catalog:ozUser():getAccountDocFromOz()
					:map( function( updatedAccountDoc )
					
						return not isAccountDocOverStorage( updatedAccountDoc )
					
					end )
			
			else
				return Observable.just( false )
			end
			
		end )
		:catch( function( maybeError )
			logger:errorf( "Unable to check storage space in the user account document: %s", maybeError )
			return Observable.just( false )
		end )			
		:subscribe( function( resetQuota )

			if resetQuota then
				resetQuotaExceeded( catalog )
			end
			
		end )

end

--------------------------------------------------------------------------------

local overStorageQuota = function( wfCatalog )

	return Observable.combineLatest( function( overQuotaCount, ozStorage )

		local isOverQuota = overQuotaCount > 0
		
		if not isOverQuota then
		
			if ozStorage and ozStorage.used and ozStorage.display_limit then
				isOverQuota = ozStorage.used >= ozStorage.display_limit
			end
			
		end

		return isOverQuota

	end,
	countOverOzStorageQuota( wfCatalog ), 
	storageEntitlement( wfCatalog ) )
	
end

--------------------------------------------------------------------------------

local function storageLimitPercentExceeded( storageUsed, storageLimit, limitPercent )

	local percent = 0

	if storageUsed and storageLimit and storageLimit > 0 then
		percent = storageUsed / storageLimit
	end

	return limitPercent and percent > limitPercent

end

--------------------------------------------------------------------------------

local function lowOnStorage( wfCatalog )

	return storageEntitlement( wfCatalog )
	
		:map( function( ozStorage )

			local storage_limit = ozStorage.display_limit
			local storage_warn_percent = kLowOnStoragePercent
			local storage_warn = ozStorage.warn
			
			if storage_limit > 0 and storage_warn > 0 and storage_warn < storage_limit then
				storage_warn_percent = storage_warn / storage_limit
			end

			return storageLimitPercentExceeded( ozStorage.used, ozStorage.display_limit, storage_warn_percent )

		end )

end
	
--------------------------------------------------------------------------------

if DEBUG_STORAGE_USED then

	-- Override storageEntitlement and countOverOzStorageQuota in order
	-- to simplify QE verification of cloud icon and sync activity popup.

	debugStorageUsedStream = BehaviorSubject( 0 )

	local storageEntitlementImp = storageEntitlement
	
	storageEntitlement = function( wfCatalog )

		return Observable.combineLatest( function( ozData, debugUsedAmount )
			ozData.storage_used = debugUsedAmount
			return ozData
		end, 
		storageEntitlementImp( wfCatalog ), 
		debugStorageUsedStream )

	end

	local function outOfStorage( wfCatalog )

		return storageEntitlement( wfCatalog )
		
			:map( function( ozStorage )
				return storageLimitPercentExceeded( ozStorage.used, ozStorage.display_limit, kOutOfStoragePercent )
			end )

	end

	local countOverOzStorageQuotaImp = countOverOzStorageQuota

	countOverOzStorageQuota = function( wfCatalog )

		return Observable.combineLatest( function( countOfAssetsOverStorage, storageLimitExceededByPercent )

			if storageLimitExceededByPercent and countOfAssetsOverStorage == 0 then
				countOfAssetsOverStorage = 1
			end

			return countOfAssetsOverStorage

		end, 
		countOverOzStorageQuotaImp( wfCatalog ), 
		outOfStorage( wfCatalog ) )

	end

end

--------------------------------------------------------------------------------

local function debugStorageUsed( amountUsed )

	if debugStorageUsedStream then
		debugStorageUsedStream:onNext( amountUsed )
	end
	
end

--------------------------------------------------------------------------------

local function makeStaticMember( fn )

	return function( _, ... ) return fn( ... ) end

end 

--------------------------------------------------------------------------------

local WFCatStorage = {
	
	_catalog = false,
	
	_watchQuotaForResetSubscription = false,
		-- (Subscription) watches for available space when docs are over quota to see if they can
		-- be resent
	
	_watchForDocsOverStorageSubscription = false,
		-- (Subscription) watches for StorageExceededErrors so that we can verify how much space is
		-- available
		
	_watchQuotaForOverageSubscription = false,
	
	_isOverStorage = Rx.ReplaySubject( 1 ),
		-- (ReplaySubject) simple behavior tracking whether the client is over storage on Oz or not
		
	class = {
		generateIndexSchema = generateIndexSchema,
		storageEntitlement = makeStaticMember( storageEntitlement ),
		countOverOzStorageQuota = makeStaticMember( countOverOzStorageQuota ),
		docsOverOzStorageQuota = makeStaticMember( docsOverOzStorageQuota ),
		overStorageQuota = makeStaticMember( overStorageQuota ),
		lowOnStorage = makeStaticMember( lowOnStorage ),
		debugStorageUsed = makeStaticMember( debugStorageUsed )
	},
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function WFCatStorage:__init( wfCatalog )

	check( 'WFCatStorage', 'WFCatalog' )
	self:setOwned {
		_catalog = wfCatalog:relayObject()
	}
	
	self._watchQuotaForResetSubscription = watchForQuotaReset( self._catalog )
	
	self._isOverStorage:onNext( false )
	
	self._watchForDocsOverStorageSubscription = Observable.combineLatest( storageEntitlement( self._catalog ),
																			countOverOzStorageQuota( self._catalog ) )
		:flatMapLatest( function( localCacheOfOzStorageInfo, countOverStorage )
			if countOverStorage > 0 then
				if ( localCacheOfOzStorageInfo.used <= localCacheOfOzStorageInfo.limit ) then
				
					return self._catalog:ozUser():getAccountDocFromOz()
						:catch( Observable.just() )
						:map( function( userAccountDoc )
							local isOverQuota = isAccountDocOverStorage( userAccountDoc )
							self._isOverStorage:onNext( isOverQuota )
						end )
						
				else
					self._isOverStorage:onNext( true )
				end
						
			else
				self._isOverStorage:onNext( false )
			end
			
			return Observable.just()
		end )
		:subscribe()
		
	self._watchQuotaForOverageSubscription = storageEntitlement( self._catalog )
		:map( function( storage )
			return storage.used > storage.limit
		end )
		:subscribe()
		
end

--------------------------------------------------------------------------------

function WFCatStorage:close()
	
	if self._watchQuotaForResetSubscription then
		self._watchQuotaForResetSubscription:dispose()
		self._watchQuotaForResetSubscription = nil
	end
	if self._watchForDocsOverStorageSubscription then
		self._watchForDocsOverStorageSubscription:dispose()
		self._watchForDocsOverStorageSubscription = nil
	end
	if self._watchQuotaForOverageSubscription then
		self._watchQuotaForOverageSubscription:dispose()
		self._watchQuotaForOverageSubscription = nil
	end
end

--------------------------------------------------------------------------------

function WFCatStorage:isOverStorage()
	return self._isOverStorage
end

--------------------------------------------------------------------------------

return  Rx.Class 'WFCatStorage'( WFCatStorage )
