--[[----------------------------------------------------------------------------

WFCatStyleBinaryFetcher.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local Rx = require 'Rx'
local WFCloseSupport = require 'WFCloseSupport'
local WFCatStyleBinaryDownloader = require 'WFCatStyleBinaryDownloader'
local WFRxScheduler = require 'WFRxScheduler'
local WFLogger = require 'WFLogger'
local KSTableUtils = require 'KSTableUtils'
local KSIdGenerator = require 'KSIdGenerator'
local WFRxArrayDiffer = require 'WFRxArrayDiffer'
local WFRxContinuousTask = require 'WFRxContinuousTask'

--------------------------------------------------------------------------------

local logger = WFLogger 'WFCatStyleBinaryFetcher'
	--:setLevel( 'DEBUG' )

local STYLE_QUERY_CALM_INTERVAL = 1 --In secs. Arbitrary value.
local STYLE_MAX_NUM_OF_IN_FLIGHT_REQUESTS = 20
local STYLE_ERROR_RETRY_DELAY = 60 * 10 -- In secs.
	-- If there are errors and there are no updates to
	-- the style table, then we need to trigger the
	-- download process to retry the error cases.

--------------------------------------------------------------------------------

-- Creates a separate function for downloading of 'v2' presets and profiles.
-- Per consults with Brian K, it is probably not wise to try to channel download
-- of presets and profiles through the pixel path provider state machinery,
-- as there is a 'baked in assumption' that the downloadable assets have a
-- docValue.type of 'image'. It might be feasible (though probably overkill)
-- to implement a separate state machine for managing profile/preset sync
--
-- Another possible path, per Brian is:
--[[

Actually thinking further, I think preset sync is very similar to `RenditionPerAssetPathFetcher.lua`

`RenditionPerAssetPathFetcher.lua` ensures we have at one least one local rendition per each `assetId` in the catalog. That's pretty much the same as ensuring we have a local preset file for each `presetId` in the catalog.

Here are the `RenditionPerAssetPathFetcher.lua` implementation details:

1. it sets up a `"allLocalAssetIds"` submodel which has a `:value()` stream that emits

```{ allAssetIds, {} }
{ addToAllAssetIds, removeFromAllAssetIds }
{ addToAllAssetIds, removeFromAllAssetIds }
...
```

2. it sets up a `"idlePixelPaths"` submodel. The submodel `:sendEvent( "fetchPaths" )` call drives the current work. The submodel `:value()` stream is the ongoing results from the `:sendEvent( "fetchPaths" )` calls.

3. The `"allLocalAssetIds"` submodel `:value()` stream is fed into a "needs rendition path" queue. That queue is popped on an `repeat ... until false` `AgEventLoup` task. The inner loop of that task spends most of its time in a wait state unless no current rendition downloads are pending and the "needs rendition path" queue is non empty.

When the wait is lifted, the queue is popped and the result is passed to `:sendEvent( "fetchPaths" )` on `"idlePixelPaths"` submodel.

4. As results come back, we update the current rendition downloads are pending list as needed

So if you copy `RenditionPerAssetPathFetcher.lua` to say, `LocalPresetSyncer.lua` file, make these changes plus implementing the daemon code to fulfill them, you'll be be pretty far along with preset sync:

1. In `LocalPresetSyncer.lua`, replace the `"allLocalAssetIds"' submodel  with a new `"needsPresetSyncIds"` submodel which should emit a stream of the following

```{ needsPresetDownloadPresetIds, {} }
{ addToNeedsPresetDownloadPresetIds, removeFromNeedsPresetDownloadPresetIds }
{ addToNeedsPresetDownloadPresetIds, removeFromNeedsPresetDownloadPresetIds }
...
```

2. On the daemon, implement `"needsPresetSyncIds"` so that it first get the list of all `presetIds` from the doc store and filters out all of them we either can't download or don't need to download. Then as updates come in (when presets are added, remove, or updated), emit the updates as arrays of adds and deletes. See `WFCatAssets.allLocalAssetIdsStream` for something very similar to that.

3. In `LocalPresetSyncer.lua`, replace `"idlePixelPaths"` submodel with `"ensurePresetIsLocal"` submodel and call `:sendEvent( "fetchPreset" )` on it.

4. On the daemon, implement that `:sendEvent( "fetchPreset" )` to accept one `presetId` at a time. Each time the daemon gets a new `presetId`, cancel the current download (if there is one) and start downloading the new `presetId` after checking the local path. In any case, when done with the `presetId` return `{ presetId, isLocal, maybeErrorString }` on the `:value()` stream so that the `AgEventLoup` task will unblock its wait.

5. In `LocalPresetSyncer.lua`, make whatever additional changes needed so that `"needsPresetSyncIds"` submodel `:value()` updates the work queue and that the `"ensurePresetIsLocal"` submodel `:value()` output is handled correctly.

--]]
--------------------------------------------------------------------------------

local function addToSet( t1, t2 )

	for _,v in ipairs( t2 ) do 
		t1[ v ] = true
	end
end

--------------------------------------------------------------------------------

local function removeFromSet( t1, t2 )

	for _,v in ipairs( t2 ) do 
		t1[ v ] = nil
	end
end

--------------------------------------------------------------------------------

local function matchesAssetRequest( assetId )

	return function( responseAssetId )

		return responseAssetId == assetId

	end

end
--------------------------------------------------------------------------------

local function makeResponseStreamForRequest( self, assetId )

	return self._binaryDownloader:getResponseStream()
		:where( matchesAssetRequest( assetId ) )
		:take( 1 )
end

--------------------------------------------------------------------------------

local function isVolumeOfflineOrOutOfSpace( self )

	return self._volumeStatus.offline or self._volumeStatus.outOfSpace

end

--------------------------------------------------------------------------------

local function doOneDownload( self, assetId )

	local savedSubscription
	
	return Rx.Observable.create(
		function( observer )

			local responseStream = makeResponseStreamForRequest( self, assetId )
			
			savedSubscription = 
					responseStream:subscribe( observer )

			self._binaryDownloader:requestStyleBinary( assetId )
			
			return Rx.Disposable.create( function()
			
				-- TODO [ashj 2018-03-09]: When the stream is
				-- complete, this will get called. Can we
				-- differentiate?
				-- Anyways cancelling will be a nop if the
				-- request has already completed.
				
				self._binaryDownloader:cancelStyleBinaryRequest( assetId )
				savedSubscription:dispose()

			end )
		
	end )
end

--------------------------------------------------------------------------------

local function getPendingStyles( self, indexStore )

	local pendingStylesQuery = indexStore:query {
		sql = [[ SELECT docId FROM presetsAndProfiles
				WHERE waitingForBinary = 1 and ozHasMaster = 1
				UNION
				SELECT docId FROM customCameraProfiles
				WHERE waitingForBinary = 1 and href NOT NULL ]],
				quiesceUntilUpToDate = true,
	}

	return pendingStylesQuery
		
		:calm( STYLE_QUERY_CALM_INTERVAL, WFRxScheduler.sharedScheduler() )
		
		:map( function( queryResult )
			
			local numItems = queryResult and #queryResult
			if numItems > 0 then
			
				local allPendingStyles = KSTableUtils.flatten( queryResult )
				return allPendingStyles
			end
			
			return {}
		end )
		
		:call( WFRxArrayDiffer )
		
end

--------------------------------------------------------------------------------

local function incrementNumOfInFlightRequests( self )

	self._numOfInFlighRequests = self._numOfInFlighRequests + 1
	self._numOfInFlightRequestsStream:onNext( self._numOfInFlighRequests )
	
	if self._test_numOfInFlightRequests then
		self._test_numOfInFlightRequests[ #self._test_numOfInFlightRequests + 1 ] = self._numOfInFlighRequests
	end
	logger:debugf( "Num of in-flight requests: %s", self._numOfInFlighRequests )
end

--------------------------------------------------------------------------------

local function decrementNumOfInFlightRequests( self )

	self._numOfInFlighRequests = self._numOfInFlighRequests - 1
	self._numOfInFlightRequestsStream:onNext( self._numOfInFlighRequests )
	
	if self._test_numOfInFlightRequests then
		self._test_numOfInFlightRequests[ #self._test_numOfInFlightRequests + 1 ] = self._numOfInFlighRequests
	end
	logger:debugf( "Num of in-flight requests: %s", self._numOfInFlighRequests )
end

--------------------------------------------------------------------------------

local function maybeDisposeInflightRequest( self, assetId )

	local requestInfo = self._inFlightRequests[ assetId ]
		if requestInfo and requestInfo.subcription then
			-- Count will be decremented in the
			-- finally function associated with the
			-- stream being cancelled.
			-- The inflight request entry will be
			-- removed after the dispose is complete
			-- in the finally fn of the subscription.
			
			requestInfo.subcription:dispose()
		end
end

--------------------------------------------------------------------------------
-- Forward declaration
local mayBeSendOneOrMore

--------------------------------------------------------------------------------

local function getDownloadSubscriptionForAsset( self, assetId, requestId )

	return doOneDownload( self, assetId )
					
		:tap( function( assetId, success, pathOrErr )
			
			if success then
				logger:infof( "Downloaded style asset %s : path=%s", assetId, pathOrErr )
				self._downloadResponseStream:onNext( true, assetId )
				
				-- Remove it from the errorStyles.
				self._errorStyles[ assetId ] = nil
			else
						
				logger:errorf( "Error for Asset %s = %s", assetId, pathOrErr )
				self._errorStyles[ assetId ] = true

			end
		end )
		
		:tapOnError( function( err )
			
			logger:infof( "Error for style asset %s = %s", assetId, err )
			self._errorStyles[ assetId ] = true
		end )
		
		:finally( function()
			
			-- Finally gets called even if the subscription
			-- is disposed.
			
			-- Count has to be decremented irrespective of
			-- whether the request is current, since we are
			-- expecting a response for each and every
			-- request, except for the requests which got
			-- cleared as part of stopping the continuous
			-- process. 
			-- This is needed so that we dont decrement
			-- the count, after the state is cleared.
			
			if self._inFlightRequestIds[ requestId ] then
				decrementNumOfInFlightRequests( self )
				self._inFlightRequestIds[ requestId ] = nil
			end
			
			-- Remove the asset from inflight request list,
			-- if we are processing the response for the
			-- latest request for that asset. 
			local requestInfo = self._inFlightRequests[ assetId ]
			
			if requestInfo and requestInfo.requestId == requestId then
				
				-- If a request for assetId is disposed
				-- and then raised again, then it is possible
				-- that the inFlight list is updated with the
				-- new entry. Hence remove it only, if the
				-- requestId matches.
				
				self._inFlightRequests[ assetId ] = nil
			end
			
			mayBeSendOneOrMore( self )
		end )
		
		:subscribe()
end
--------------------------------------------------------------------------------

mayBeSendOneOrMore = function( self )

	if isVolumeOfflineOrOutOfSpace( self ) then
		-- We cannot download since volume is
		-- not servicable. 
		return
	end
	
	if not self._isStarted then
		-- The continuous task is not running now.
		return
	end
	
	while( self._numOfInFlighRequests < self._maxNumOfFlightRequests 
		and next( self._pendingStyles ) ) do
		
		local nextElement = next( self._pendingStyles )
		
		-- Remove from the pending list and add to
		-- the inFlightList. 
		-- We are ensuring, at any point of time, 
		-- there are only _maxNumOfFlightRequests
		-- by checking the inFlightRequest count.
		
		self._pendingStyles[ nextElement ] = nil
		
		-- If a request for this asset is already
		-- in progress, need to cancel it and
		-- add a new request.
		-- Count will be decremented in the
		-- finally function associated with the
		-- stream being cancelled.
		
		maybeDisposeInflightRequest( self, nextElement )
		
		-- We are making sure that the subscribe and
		-- dispose is always called in order, by
		-- making both of them async. But if we
		-- call dispose and then subscribe in quick
		-- succession, then it is possible that
		-- the response of disposing off the subscription
		-- comes after the call to subscribe. Hence
		-- we need to make sure that we are updating
		-- the proper inflight request. For that we
		-- are using the requestId.
		
		-- Here the previous subscription, if any, is
		-- disposed and the new one is stored. I believe,
		-- in all cases, the old subscrition will be
		-- held untill it is disposed off and we will
		-- receive a call in finally. 
		local requestId = KSIdGenerator.uuidInLowercaseHex()
		
		self._inFlightRequestIds[ requestId ] = true
		
		self._inFlightRequests[ nextElement ] = {
			subcription = getDownloadSubscriptionForAsset( self, nextElement, requestId ),
			requestId = requestId,
		}
		incrementNumOfInFlightRequests( self )
		
	end
end

--------------------------------------------------------------------------------

local function cancelStaleRequests( self, assetIds )

	for i, assetId in ipairs( assetIds ) do
		
		maybeDisposeInflightRequest( self, assetId )
		
	end
end

--------------------------------------------------------------------------------

local function listenForPendingStyles( self, catalog )

	return getPendingStyles( self, catalog:indexStore() )
		
		:flatMapLatest( function( additions, removals )
			
			return self._continuousTaskForStyleChangeListening:throwIfNotRunning( additions, removals )
		
		end )
		
		:map( function( additions, removals )
		
			-- Update the pending list with
			-- the new items.
			addToSet( self._pendingStyles, additions )
			removeFromSet( self._pendingStyles, removals )
			
			-- Remove the deleted entries from the
			-- error list as well
			removeFromSet( self._errorStyles, removals )
			
			-- Cancel in-flight requests, if any
			-- for the removed assets.
			cancelStaleRequests( self, removals )
			
			-- Maybe some new styles have got added.
			-- Trigger the processing chain.
			mayBeSendOneOrMore( self )
			
		end )
	
		:catch( function( errorMessage )
		
			return self._continuousTaskForStyleChangeListening:throwIfNotStopped( errorMessage )
		end )
end


--------------------------------------------------------------------------------

local function mayBeRetryErrorRecords( self )

	return Rx.Observable.just()
		:map( function()
			
			if not next( self._errorStyles ) then
				-- No error records to process.
				return
			end
			
			if next( self._pendingStyles ) then
				-- There are pending records to process
				-- Lets not process error records in 
				-- this cycle.
				return
			end
			
			if next( self._inFlightRequests ) then
				-- Assume the download is taking more
				-- time as compared to the error
				-- interval. There is a chance that
				-- the same request will get added
				-- to the queue again. Avoiding this
				-- raise condition by waiting for all 
				-- the requests to finish
				return
			end
			
			-- Retry all error cases as if those
			-- are the normal pending ones.
			addToSet( self._pendingStyles, self._errorStyles )
			self._errorStyles = {}
			mayBeSendOneOrMore( self )
		end )
		
		:catch( function( errorMessage )
		
			return self._continuousTaskForErrorRetry:throwIfNotStopped( errorMessage )
		end )
	
end


--------------------------------------------------------------------------------

local function resetCurrentDownloadStates( self )

	self._inFlightRequests = {}
	self._inFlightRequestIds = {}
	self._pendingStyles = {}
	self._errorStyles = {}
	self._numOfInFlighRequests = 0
end

--------------------------------------------------------------------------------

local function disposeAllDownloads( self )

	for assetId, requestInfo in pairs( self._inFlightRequests or {} ) do
		if requestInfo.subscription then
			requestInfo.subcription:dispose()
		end
	end
	
end

--------------------------------------------------------------------------------

local WFCatStyleBinaryFetcher = {
	
	_catalog = false,
	
	_binaryDownloader = false,
	
	_downloadStreamSubscription = false,
	
	_downloadResponseStream = false,
	
	_inFlightRequests = false,
	
	_inFlightRequestIds = false,
	
	_numOfInFlighRequests = false,
	
	_numOfInFlightRequestsStream = false,
	
	_pendingStyles = false,
	
	_errorStyles = false,
	
	_continuousTaskForStyleChangeListening= false,
	
	_continuousTaskForErrorRetry = false,
	
	_maxNumOfFlightRequests = false,
	
	_test_numOfInFlightRequests = false,
	
	_styleVolumeGuid = false,
	
	_volumeStatus = false,
	
	_isStarted = false,
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:__init( catalog, optMaxFlightRequests )

	self._catalog = catalog
	self._styleVolumeGuid = catalog:fileSpec():pixelsPath().volumeGuid
	self._binaryDownloader = WFCatStyleBinaryDownloader( catalog )
	self._downloadResponseStream = Rx.Subject()
	self._maxNumOfFlightRequests = optMaxFlightRequests or STYLE_MAX_NUM_OF_IN_FLIGHT_REQUESTS
	self._numOfInFlightRequestsStream = Rx.BehaviorSubject( 0 )

	local downloadTaskDelays = {
		progressiveBackoff = false,
		errorDelay = 10,
		maxDelay = 10,
	}

	-- For style notification
	local continuousTaskForStyleListening = function() return listenForPendingStyles( self, catalog ) end
	self._continuousTaskForStyleChangeListening = WFRxContinuousTask( "StyleDownloadNotificationTask", continuousTaskForStyleListening, downloadTaskDelays )
	
	-- For error retries
	local errorRetryTaskDelays = {
		progressiveBackoff = false,
		errorDelay = STYLE_ERROR_RETRY_DELAY,
		maxDelay = STYLE_ERROR_RETRY_DELAY,
	}

	local continuousTaskForErrorRetry = function() return mayBeRetryErrorRecords( self ) end

	self._continuousTaskForErrorRetry = WFRxContinuousTask( "StyleErrorRetryTask", continuousTaskForErrorRetry, errorRetryTaskDelays )
	
	self._volumeStatus = {}
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:getResponseStream()
	
	return self._downloadResponseStream
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:start()
	
	self._isStarted = true
	resetCurrentDownloadStates( self )
	self._continuousTaskForErrorRetry:startContinuousTask()
	
	-- Set up the db listening subscription after
	-- setting up the listeners.
	self._continuousTaskForStyleChangeListening:startContinuousTask()
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:stop()

	self._isStarted = false
	self._continuousTaskForStyleChangeListening:stopContinuousTask()
	self._continuousTaskForErrorRetry:stopContinuousTask()
	
	disposeAllDownloads( self )
	resetCurrentDownloadStates( self )
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:reportStatusOfVolumes( statusOfVolumes )

	local offline = statusOfVolumes.offlineVolumes[ self._styleVolumeGuid ]
	local outOfSpace = statusOfVolumes.volumesOutOfSpace[ self._styleVolumeGuid ]
	local overIdleSpaceThreshold = statusOfVolumes.volumesOverIdleSpaceThreshold[ self._styleVolumeGuid ]
	
	local newVolumeStatus = {
		volumeGuid = self._styleVolumeGuid,
		offline = offline,
		outOfSpace = outOfSpace,
		overIdleSpaceThreshold = overIdleSpaceThreshold
	}
	
	if KSTableUtils.deepEquals( self._volumeStatus, newVolumeStatus ) then
		-- No change in status
		return
	end
	
	self._volumeStatus = newVolumeStatus
	
	-- Currently we are not cancelling the
	-- inFlight requests when volume is down.
	-- We will just stop any new requests.
	
	if self._isStarted then
		-- If we had pending requests before,
		-- we will need to add it for processing
		mayBeSendOneOrMore( self )
	end
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:close()
	
	self:stop()
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:isDownloadingStyles()
	
	return self._numOfInFlightRequestsStream
		:distinctUntilChanged()
		:map( function( num )
			return num > 0
		end )
end

--------------------------------------------------------------------------------

function WFCatStyleBinaryFetcher:setTestModeAndParams()
	
	self._test_numOfInFlightRequests = {}
end

--------------------------------------------------------------------------------

-- To be used only by tests.
function WFCatStyleBinaryFetcher:_test_getNumOfInFlightRequestArray()
	
	return self._test_numOfInFlightRequests
end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatStyleBinaryFetcher'( WFCatStyleBinaryFetcher )