--[[----------------------------------------------------------------------------

WFCatStyleSynchronizer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local Rx = require 'Rx'
local WFCloseSupport = require 'WFCloseSupport'
local WFRxScheduler = require 'WFRxScheduler'
local WFCatDelete = require 'WFCatDelete'
local WFCatStyleBinaryFetcher = require 'WFCatStyleBinaryFetcher'
local NmbCatPreset = require 'NmbCatPreset'
local NmbStyleFavorites = require 'NmbStyleFavorites'
	-- TODO [ashj 2018-04-18]: Need to remove
	-- calling of Nmb namespace from WFCat namespace.
	-- But not sure WFCatStyleSynchronizer should be
	-- converted to a Nmb only namespace.
local NmbLocalCameraProfileStatus = require 'NmbLocalCameraProfileStatus'

--------------------------------------------------------------------------------

local STYLE_REFRESH_CALM_INTERVAL = 2 --In secs. Arbitrary value.

--------------------------------------------------------------------------------

local function getStyleDeletionStream( catalog )

	return WFCatDelete.getDeletionStream()
		:where( function( success, path )
			
			-- Is there any way to identify otherthan
			-- path?
			
			return success and catalog:fileSpec():maybeStylePath( path )
		end )
end

--------------------------------------------------------------------------------

local WFCatStyleSynchronizer = {
	
	_styleBinaryFetcher = false,
	
	_styleRefreshStream = false,
	
	_cameraProfileStatus = false,
	
	_styleDownloadStreamSubscription = false,
	
	_ingestingStream = false,
	
	_cameraProfileLinkProcessor = false,
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:__init( catalog, workerThreadAgrs )

	self._styleBinaryFetcher = WFCatStyleBinaryFetcher( catalog )
	self._styleRefreshStream = Rx.Subject()
	self._ingestingStream = Rx.ReplaySubject( 1 )
	
	self._cameraProfileStatus = NmbLocalCameraProfileStatus( workerThreadAgrs )
	self._cameraProfileStatus:populateMissingProfileInfo()
	
	-- The preset refresh stream should fire when
	-- any new preset files are downloaded or
	-- anyone does an explicit call to trigger
	-- refresh
	
	self._styleDownloadStreamSubscription = 
		self._styleBinaryFetcher:getResponseStream()
			:merge( getStyleDeletionStream( catalog ) )
			:subscribe( function()
				self._styleRefreshStream:onNext()
			end )
	
	-- SPAM [matthewj 2018-04-06]: This maybe should be done differently with more ownership by WFCatStyleSynchronizer
	NmbCatPreset.maybeDoProfilePresetSlurp( catalog, self )
	NmbStyleFavorites.maybeSlurpFavorites( catalog )
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:initCameraProfileLinkProcessor( wfCatalog,
																profileCatalog )

	self._cameraProfileLinkProcessor = require 'NmbDCPDocumentProcessor'(
													wfCatalog, profileCatalog )

end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:getPresetProfileIngestInProgressStream()
	return self._ingestingStream
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:getStyleRefreshStream()
	
	return self._styleRefreshStream
		:calm( STYLE_REFRESH_CALM_INTERVAL, WFRxScheduler.sharedScheduler() )
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:isDownloadingStyles()
	return self._styleBinaryFetcher:isDownloadingStyles()
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:getCameraProfileRefreshStream()
	
	return self._cameraProfileStatus:getCameraProfileRefreshStream()
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:triggerStyleRefresh()
	
	self._styleRefreshStream:onNext( true )
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:start()

	self._styleBinaryFetcher:start()
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:stop()

	self._styleBinaryFetcher:stop()
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:close()

	if self._styleRefreshStream then
		self._styleRefreshStream:dispose()
		self._styleRefreshStream = nil
	end
	
	if self._cameraProfileStatus then
		self._cameraProfileStatus:close()
		self._cameraProfileStatus = nil
	end
	
	if self._ingestingStream then
		self._ingestingStream:dispose()
		self._ingestingStream = nil
	end

	if self._cameraProfileLinkProcessor then
		self._cameraProfileLinkProcessor:close()
		self._cameraProfileLinkProcessor = nil
	end
end

--------------------------------------------------------------------------------

function WFCatStyleSynchronizer:reportStatusOfVolumes( volumeStatuses )

	if self._styleBinaryFetcher then
		self._styleBinaryFetcher:reportStatusOfVolumes( volumeStatuses )
	end
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatStyleSynchronizer'( WFCatStyleSynchronizer )
