--[[----------------------------------------------------------------------------

WFCatSyncAutoDeleter.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFRxScheduler = require 'WFRxScheduler'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local KSFunctionUtils = require 'KSFunctionUtils'

	-- WF.oz_client.managed_catalog
local WFCatDelete = require 'WFCatDelete'

	-- WF.core.logger
local WFLogger = require 'WFLogger'

--------------------------------------------------------------------------------

local logger = WFLogger( 'WFCatSyncAutoDeleter' )

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

--============================================================================--

-- @mark WFCatSyncAutoDeleter class

--------------------------------------------------------------------------------

local WFCatSyncAutoDeleter = {

	_docStore = false,
		-- (WFRxDocStore)
	
	_indexStore = false,
		-- (WFRxIndexStore)

	_docIdsToProcess = false,
		-- (table)

	_deletedDocIdStreamSubscription = false,
		-- (Rx Subscription)
		
	_processDeletionsStreamSubscription = false,
		-- (Rx Subscription)

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local getDocIdsAffectedByDeletedAlbumSQL = KSFunctionUtils.memoizeFn( function( numberOfValues )
	
	return string.format( "SELECT docId FROM album_asset_v2 WHERE albumId IN ( ?%s )",
			string.rep( ", ?", numberOfValues - 1 ) )

end )

--------------------------------------------------------------------------------

local getDocIdsAffectedByDeletedAssetSQL = KSFunctionUtils.memoizeFn( function( numberOfValues )
	
	return string.format( "SELECT docId FROM album_asset_v2 WHERE assetId IN ( ?%s )",
			string.rep( ", ?", numberOfValues - 1 ) )

end )

--------------------------------------------------------------------------------

local function deleteRelatedDocuments( docStore, indexStore, docIdsToDelete, startOffset, chunkSize )

	for _, queryFunc in ipairs( { getDocIdsAffectedByDeletedAlbumSQL, getDocIdsAffectedByDeletedAssetSQL } ) do
	
		local query = queryFunc( chunkSize )
		
		logger:infof( "Proccessing deletion query: %s", query )
	
		indexStore:query{
			sql = query,
			quiesceUntilUpToDate = true,
			unpack( docIdsToDelete, startOffset, startOffset + chunkSize - 1  ) }

			:take( 1 )

			:flatMapLatest( function( docIdRows )
				return WFCatDelete.deleteDocsFromQueryResult( docStore, docIdRows, true )
			end )

			:call( runUntilDone )

	end

end

--------------------------------------------------------------------------------

local function processDeletedDocIds( self )

	local docIds = self._docIdsToProcess
	local numberOfDocsToProcess = #docIds
	
	local offset = 1
	
	while offset <= numberOfDocsToProcess do
	
		local chunkSize = 0
	
		if numberOfDocsToProcess >= 32 then
		
			chunkSize = 32
		
		elseif numberOfDocsToProcess >= 16 then
		
			chunkSize = 16
		
		elseif numberOfDocsToProcess >= 8 then
		
			chunkSize = 8
		
		elseif numberOfDocsToProcess >= 4 then
		
			chunkSize = 4
		
		elseif numberOfDocsToProcess >= 2 then
		
			chunkSize = 2
		
		else
		
			chunkSize = 1
		
		end
	
		logger:infof( "Now deleting %d docs", chunkSize )
	
		deleteRelatedDocuments( self._docStore, self._indexStore, docIds, offset, chunkSize )
	
		offset = offset + chunkSize
	
	end
	
	self._docIdsToProcess = {}
	
end

--------------------------------------------------------------------------------

function WFCatSyncAutoDeleter:__init( catalog, deletedDocIdStream, optArgs )

	check( 'WFCatSyncAutoDeleter', 'WFCatalog', 'table', '?table' )

	optArgs = optArgs or {}

	self._docStore = catalog:docStore()
	self._docStore:addCloseTarget( self )

	self._indexStore = catalog:indexStore()
	self._indexStore:addCloseTarget( self )

	self._docIdsToProcess = {}
	
	-- Queue up the deleted document IDs to process

	self._deletedDocIdStreamSubscription = deletedDocIdStream
		:subscribe( function( docId )
			table.insert( self._docIdsToProcess, docId )
		end )

	-- Every 15 seconds, process the deleted document IDs.

	self._processDeletionsStreamSubscription = Observable.timer( 0, optArgs.delay or 15, WFRxScheduler.sharedScheduler() )
		:subscribe( function()
			processDeletedDocIds( self )
		end )
	
end

--------------------------------------------------------------------------------

function WFCatSyncAutoDeleter:close()

	check( 'WFCatSyncAutoDeleter' )
	
	local deletedDocIdStreamSubscription = self._deletedDocIdStreamSubscription
	self._deletedDocIdStreamSubscription = nil
	
	if deletedDocIdStreamSubscription then
		deletedDocIdStreamSubscription:dispose()
	end

	local processDeletionsStreamSubscription = self._processDeletionsStreamSubscription
	self._processDeletionsStreamSubscription = nil
	
	if processDeletionsStreamSubscription then
		processDeletionsStreamSubscription:dispose()
	end

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatSyncAutoDeleter'( WFCatSyncAutoDeleter )
