--[[----------------------------------------------------------------------------

WFCatSyncStatus.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.managed_catalog
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatIdleDownloader = require 'WFCatIdleDownloader'
local WFCatOzErrors = require 'WFCatOzErrors'
local NmbFileManagementSettings = require 'NmbFileManagementSettings'

--------------------------------------------------------------------------------

local isAssetDocument = assert( WFCatIndexUtils.isAssetDocument )

local derefOpt = KSTableUtils.derefOpt

--------------------------------------------------------------------------------

local WFCatSyncStatus = {}

--------------------------------------------------------------------------------

-- !!! CAUTION [jandrew 2016-12-01]: When asset sync status enums are changed,
-- we must also bump the sync status version in the _map methods to force re-indexing. 
-- IF YOU ADD OR MODIFY THE SYNC STATUS ENUMERATION, YOU MUST INCREMENT THE SYNC
-- STATUS VERSION NUMBERS.

local kAssetNeedsMetadataUpload = 'NeedsMetadataUploaded'
local kAssetNeedsOriginalUpload = 'NeedsOriginalUploaded'
local kAssetNeedsOriginalDownload = 'NeedsOriginalDownloaded'
local kAssetNeedsOriginalDownloadForOffline = 'NeedsOriginalDownloadedForOffline'
local kAssetNeedsProxyDownloadForOffline = 'NeedsProxyDownloadedForOffline'
local kAssetNeedsProxyDownload = 'NeedsProxyDownloaded'
local kAssetNeedsDevelopXmpDownload = 'NeedsDevelopXmpDownloaded'
local kAssetComplete = 'Completed'
local kAssetCompleteNoMaster = 'CompletedNoMaster'
local kAssetStorageLimitExceeded = 'StorageLimitExceeded'
local kAssetInvalidOriginal = 'InvalidLocalOriginal'
local kAssetOriginalIsOnOfflineVolume = 'OriginalIsOnOfflineVolume'
local kAssetProxyIsOnOfflineVolume = 'ProxyIsOnOfflineVolume'
local kAssetNeedsOriginalDownloadToOfflineVolume = 'NeedsOriginalDownloadToOfflineVolume'
local kAssetNeedsProxyDownloadToOfflineVolume = 'NeedsProxyDownloadToOfflineVolume'
local kAssetNeedsOriginalUploadFromOfflineVolume = 'NeedsOriginalUploadedButOnOfflineVolume'
local kAssetNeedsOriginalMigration = 'NeedsOriginalMigration'
local kAssetNeedsProxyMigration = 'NeedsProxyMigration'
local kAssetOriginalDownloadError = 'OriginalDownloadError'
local kAssetProxyDownloadError = 'ProxyDownloadError'
local kAssetDevelopXmpDownloadError = 'DevelopXmpDownloadError'
local kAssetNeedsOriginalDownloadToOutOfSpaceVolume = 'NeedsOriginalDownloadToOutOfSpaceVolume'
local kAssetNeedsProxyDownloadToOutOfSpaceVolume = 'NeedsOriginalProxyToOutOfSpaceVolume'
local kAssetSyncPaused = 'SyncPaused'
local kAssetSyncNoConnection = 'SyncNoConnection'
local kAssetTrialOrSubscriptionExpired = 'TrialOrSubscriptionExpired'
local kAssetStackComplete = 'StackComplete'

local kAssetSyncStatusById = {
	kAssetNeedsMetadataUpload,						-- 1
	kAssetNeedsOriginalUpload,						-- 2
	kAssetNeedsOriginalDownload,					-- 3
	kAssetNeedsOriginalDownloadForOffline,			-- 4
	kAssetNeedsProxyDownload,						-- 5
	kAssetNeedsDevelopXmpDownload,					-- 6
	kAssetComplete,									-- 7
	kAssetCompleteNoMaster,							-- 8
	kAssetStorageLimitExceeded,						-- 9
	kAssetInvalidOriginal,							-- 10
	kAssetOriginalIsOnOfflineVolume,            	-- 11
	kAssetProxyIsOnOfflineVolume,            		-- 12
	kAssetNeedsOriginalDownloadToOfflineVolume, 	-- 13
	kAssetNeedsProxyDownloadToOfflineVolume,		-- 14
	kAssetNeedsOriginalMigration,					-- 15
	kAssetNeedsProxyMigration,						-- 16
	kAssetNeedsOriginalUploadFromOfflineVolume,		-- 17
	kAssetOriginalDownloadError, 					-- 18
	kAssetProxyDownloadError,						-- 19
	kAssetDevelopXmpDownloadError,					-- 20
	kAssetNeedsOriginalDownloadToOutOfSpaceVolume, 	-- 21
	kAssetNeedsProxyDownloadToOutOfSpaceVolume, 	-- 22
	kAssetSyncPaused,								-- 23
	kAssetSyncNoConnection,							-- 24
	kAssetNeedsProxyDownloadForOffline,				-- 25
	kAssetTrialOrSubscriptionExpired,				-- 26
	kAssetStackComplete,							-- 27
}

local kAssetSyncStatus = KSTableUtils.invert( kAssetSyncStatusById )

--------------------------------------------------------------------------------

local kAssetHasLocalOriginal = 'OriginalOnDisk'
local kAssetHasLocalOriginalForOffline = 'OriginalOnDiskForOffline'
local kAssetHasLocalProxy = 'ProxyOnDisk'
local kAssetHasNothing = 'NothingOnDisk'

local kAssetLocalFileTypeById = {
	kAssetHasLocalOriginal,				-- 1
	kAssetHasLocalOriginalForOffline,	-- 2
	kAssetHasLocalProxy,				-- 3
	kAssetHasNothing,					-- 4
}

local kAssetLocalFileType = KSTableUtils.invert( kAssetLocalFileTypeById )

--------------------------------------------------------------------------------

local kAlbumSyncing = 'syncing'
local kAlbumComplete = 'synced'
local kAlbumSyncPaused = 'paused'
local kAlbumSyncNoConnection = 'offline'

local kAlbumSyncStatusById = {
	kAlbumSyncing,			-- 1
	kAlbumComplete,			-- 2
	kAlbumSyncPaused,   	-- 3
	kAlbumSyncNoConnection, -- 4
}

local kAlbumSyncStatus = KSTableUtils.invert( kAlbumSyncStatusById )

--------------------------------------------------------------------------------

local kOzHasOriginal = 'OzHasOriginal'
local kOzHasProxy = 'OzHasSmartPreview'
local kOzHasNothing = 'OzHasNothing'

--------------------------------------------------------------------------------

local function docValueHasLocalOriginal( docValue )

	local localOriginalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
	local localOriginalRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue )
	return localOriginalVolumeGuid and localOriginalRelativePath

end

--------------------------------------------------------------------------------

local function docValueHasLocalProxy( docValue )

	local localProxyVolumeGuid = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
	local localProxyRelativePath = WFCatFileSpec.getProxyRelativePathFromDocValue( docValue )
	return localProxyVolumeGuid and localProxyRelativePath

end

--------------------------------------------------------------------------------

local function assetNeeded( docValue )

	local ozHasURIForOriginal = WFCatFileSpec.getOriginalLinkFromDocValue( docValue ) ~= nil
	local haveLocalPathForOriginal = docValueHasLocalOriginal( docValue )
	local needsOriginalDownload = not haveLocalPathForOriginal and ozHasURIForOriginal 

	local ozHasURIForProxy = WFCatFileSpec.getProxyLinkFromDocValue( docValue ) ~= nil
	local haveLocalPathForProxy = docValueHasLocalProxy( docValue )
	local needsProxyDownload = not haveLocalPathForOriginal and not haveLocalPathForProxy and ozHasURIForProxy
	
	local haveLocalPathForXmp = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue ) ~= nil
	local ozHasURIForXmp = WFCatFileSpec.getXmpDevelopLinkFromDocValue( docValue ) ~= nil
	local needsDevelopXmpDownload = not haveLocalPathForXmp and ozHasURIForXmp 
	
	local isMarkedForOffline = WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue )

	local sufficientRenditionFoundOnOz = WFCatFileSpec.checkOzForNonThumbnailRendition( docValue )

	local waitingForProxyOnOz = derefOpt( docValue, '_serverProvidedMetadata', 'startedWaitingForProxyTimestamp' )

	local keepAllOriginals = NmbFileManagementSettings.keepAllOriginals()
	local keepAllProxies = NmbFileManagementSettings.keepAllProxies()

	-- SPAM [jandrew 2017-09-12]: We pass `nil` for two of the parameters because they should be
	-- ignored when calculating what we NEED to sync as opposed to what we SHOULD try to sync;
	-- the difference between "need" and "should" in this context is that an asset that needs
	-- syncing still needs to be synced even if there's no disk space available. On the other hand,
	-- an asset that needs to be synced SHOULD NOT be synced if there's no disk space available.

	local binaryType, isBinaryRequired = WFCatIdleDownloader.whatToDownload(
		keepAllOriginals,
		keepAllProxies,
		needsOriginalDownload,
		haveLocalPathForOriginal,
		needsProxyDownload,
		haveLocalPathForProxy,
		needsDevelopXmpDownload,
		isMarkedForOffline,
		nil, -- originalVolumeOfflineOrFull.
		nil, -- proxyVolumeOfflineOrFull
		sufficientRenditionFoundOnOz,
		waitingForProxyOnOz )

	-- Sync status is not affected by purely idle downloads. It's only affected by idle
	-- downloads that are required. 
	
	if not isBinaryRequired then 
		binaryType = nil
	end

	return binaryType

end 

--------------------------------------------------------------------------------

local function isStack( docValue )
	
	return not not ( docValue.subtype == 'stack' )

end 

--------------------------------------------------------------------------------

local function doesAssetNeedOriginalUploaded( docValue )
	
	local ozMasterUrl, isOzOriginalInvalid = WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
	local hasLocalOriginal = docValueHasLocalOriginal( docValue )
	
	local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )
	local localOriginalSha256 = hasLocalOriginal and WFCatFileSpec.getLocalOriginalSha256( docValue )
	local isLocalOriginalValid = localOriginalSha256 == imageSha256
	
	local needsUpload = false
	
	if not not ( hasLocalOriginal and not ozMasterUrl ) then
	
		needsUpload = true
		
	elseif hasLocalOriginal and ozMasterUrl and isOzOriginalInvalid and isLocalOriginalValid then
	
		needsUpload = true
	
	end
	
	return needsUpload
	
end

--------------------------------------------------------------------------------

local function isMarkedForOfflineUse( docValue )
	
	return WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue )

end

--------------------------------------------------------------------------------

local function doesAssetNeedOriginalDownloaded( docValue )
	
	return assetNeeded( docValue ) == 'original'

end

--------------------------------------------------------------------------------

local function doesAssetNeedProxyDownloaded( docValue )

	return assetNeeded( docValue ) == 'proxy'
	
end

--------------------------------------------------------------------------------

local function doesAssetNeedDevelopXmpDownloaded( docValue )
	
	return assetNeeded( docValue ) == 'xmp_develop'
	
end

--------------------------------------------------------------------------------

local function isLatestMetadataUploaded( docValue )
	
	local spm = docValue and docValue._serverProvidedMetadata
	
	local links = KSTableUtils.derefOpt( spm, 'links' )
	
	-- Do we have nothing on Oz yet?
			
	if not links then
		return false
	end
	
	if spm then
	
		local revId = docValue._revId
		
		local revIds = spm.revision_ids
		
		if revIds then
		
			local cnt = #revIds
			
			if cnt == 0 then
				-- SPAM [matthewj 2016-09-02]: if the revision_ids table is present but 
				-- empty this is likely an older non-asset type doc (e.g. album) from 
				-- Oz which did not traditionally have revision ids. We will consider 
				-- these as synced for UI purposes as in most cases it will be. However,
				-- the first local change to the doc will not trigger the UI status to 
				-- move to 'syncing' as the revision id table will still be empty. Once
				-- this actually syncs to Oz and back the revision_ids array will no 
				-- longer be empty and all subsequent changes should correctly reflect
				-- in changes to the sync status for the UI
				return true
			end
			
			-- if local _revId can be found in revision_ids array tracked by Oz
			-- the latest metadata is uploaded
			
			for i = cnt, 1, -1 do
				
				local ozRevId = revIds[ i ]
				if revId == ozRevId then
					return true
				end
				
			end
			
		end
		
	end
	
	return false
	
end

--------------------------------------------------------------------------------

local function hasInvalidOriginalError( docValue )
	return KSTableUtils.derefOpt( docValue, '_localOnly', 'localOriginalError' )
end

--------------------------------------------------------------------------------

local function hasOriginalDownloadError( docValue )

	local originalHref = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
	
	local permanentOriginalDownloadErrorForCurrentHref = KSTableUtils.derefOpt( docValue, '_localOnly', 'binaryDownloadErrors', '/rels/master', originalHref )
	
	return not not permanentOriginalDownloadErrorForCurrentHref
	
end

--------------------------------------------------------------------------------

local function hasProxyDownloadError( docValue )
	
	local proxyRel
	if docValue.subtype == 'video' then	
		proxyRel = '/rels/rendition_type/720p'
	else
		proxyRel = '/rels/proxy_type/2560'
	end
	
	local proxyHref = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', proxyRel, 'href' )
	
	local permanentProxylDownloadErrorForCurrentHref = KSTableUtils.derefOpt( docValue, '_localOnly', 'binaryDownloadErrors', proxyRel, proxyHref )
	
	return not not permanentProxylDownloadErrorForCurrentHref
end

--------------------------------------------------------------------------------

local function hasXmpDevelopDownloadError( docValue )

	local xmpDevelopHref = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'href' )
	
	local permanentXmpDevelopDownloadErrorForCurrentHref = KSTableUtils.derefOpt( docValue, '_localOnly', 'binaryDownloadErrors', 'rels/xmp/develop', xmpDevelopHref )
	
	return not not permanentXmpDevelopDownloadErrorForCurrentHref
	
end

--------------------------------------------------------------------------------

-- !!! CAUTION [jandrew 2016-12-01]: When the following function, or functions within,
-- are changed and result in different outputs, we must bump the sync status version
-- numbers found in the _map methods to force re-indexing. 

function WFCatSyncStatus.getAssetStatus( docValue )

	local assetSyncStatusId

	if docValue._deleted then

		local links = KSTableUtils.derefOpt( docValue,
									 '_serverProvidedMetadata',
									 'links' )

		if not links then
			-- This gets executed if you delete an asset that hasn't been
			-- uploaded yet, it is called by the _map function in asset_status_table().
			return nil
		end

		assetSyncStatusId = kAssetNeedsMetadataUpload

	elseif isStack( docValue ) then
		
		if not isLatestMetadataUploaded( docValue ) then
			assetSyncStatusId = kAssetNeedsMetadataUpload
		else
			assetSyncStatusId = kAssetStackComplete
		end
	
	else
	
		if not isLatestMetadataUploaded( docValue ) then

			if WFCatOzErrors.hasStorageExceededError( docValue ) then
				assetSyncStatusId = kAssetStorageLimitExceeded
			elseif WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue ) then
				assetSyncStatusId = kAssetTrialOrSubscriptionExpired
			else
				assetSyncStatusId = kAssetNeedsMetadataUpload
			end

		elseif doesAssetNeedOriginalUploaded( docValue ) then

			if hasInvalidOriginalError( docValue ) then
				assetSyncStatusId = kAssetInvalidOriginal
			elseif WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue ) then
				assetSyncStatusId = kAssetTrialOrSubscriptionExpired
			else
				assetSyncStatusId = kAssetNeedsOriginalUpload
			end

		elseif doesAssetNeedOriginalDownloaded( docValue ) then
			
			if isMarkedForOfflineUse( docValue ) or
				NmbFileManagementSettings.keepAllOriginals() then
				assetSyncStatusId = kAssetNeedsOriginalDownloadForOffline
			else
				assetSyncStatusId = kAssetNeedsOriginalDownload
			end

		elseif doesAssetNeedProxyDownloaded( docValue ) then
			
			if isMarkedForOfflineUse( docValue ) or
				NmbFileManagementSettings.keepAllProxies() then
				assetSyncStatusId = kAssetNeedsProxyDownloadForOffline
			else
				assetSyncStatusId = kAssetNeedsProxyDownload
			end
			
		elseif doesAssetNeedDevelopXmpDownloaded( docValue ) then
			
			assetSyncStatusId = kAssetNeedsDevelopXmpDownload
		
		else
			
			if hasOriginalDownloadError( docValue ) then
				
				assetSyncStatusId = kAssetOriginalDownloadError
				
			elseif hasProxyDownloadError( docValue ) then
			
				assetSyncStatusId = kAssetProxyDownloadError
			
			elseif hasXmpDevelopDownloadError( docValue ) then
			
				assetSyncStatusId = kAssetDevelopXmpDownloadError
				
			elseif WFCatFileSpec.getOriginalLinkFromDocValue( docValue ) then
				
				assetSyncStatusId = kAssetComplete
			
			else
				
				assetSyncStatusId = kAssetCompleteNoMaster
			
			end

		end

	end

	return kAssetSyncStatus[ assetSyncStatusId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )

	local allVolumeAgents = wfCatalog:getAllVolumeAgents()

	local volumeAgentStatusStreams = {}
	
	local originalsPathInfo = wfCatalog:fileSpec():originalsPathInfo()
	local volumeGuidForOriginals = originalsPathInfo.volumeGuid
	
	local proxiesPathInfo = wfCatalog:fileSpec():proxiesPathInfo()
	local volumeGuidForProxies = proxiesPathInfo.volumeGuid
	
	-- Build up a table of volume agent status streams.

	for _, agent in pairs( allVolumeAgents ) do
		table.insert( volumeAgentStatusStreams, agent:getVolumeStatusStream() )
	end

	return Rx.Observable.combineLatest( unpack( volumeAgentStatusStreams ) )
	
		:map( function( ... )
		
			local knownVolumes = {}
			local onlineVolumes = {}
			local outOfSpaceVolumes = {}
			
			for _, volumeStatus in ipairs( arg ) do

				local volumeGuid = volumeStatus.volumeGuid
				local isOnline = volumeStatus.isOnline
				local outOfSpace = volumeStatus.isOutOfSpace

				knownVolumes[ volumeGuid ] = true
				
				onlineVolumes[ volumeGuid ] = isOnline
				
				outOfSpaceVolumes[ volumeGuid ] = outOfSpace
				
			end
			
			return volumeGuidForOriginals, volumeGuidForProxies,
					onlineVolumes, outOfSpaceVolumes, knownVolumes
		
		end )

end

--------------------------------------------------------------------------------

local function maybeGetOngoingDocValue( wfCatalog, docValues )

	if #docValues ~= 1 then
			
		return Rx.Observable.just( docValues )
	
	else

		local docValue = docValues[ 1 ] or {}

		return wfCatalog:docStore()
			:ongoingDocValue( docValue._id )
			:map( function( docValue )
				return { docValue }
			end )

	end

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedAssetStatusFromIdsAsStr( wfCatalog, assetIds )

	--[[ 
	
		True asset sync status is made up 3 separate data points.
		
		1. The current state of the asset as generated by "getAssetStatus". This
			is the same status that is stored in sync status index store table.
	
		2. The current state of the volumes where the assets are stored. If an asset
		 	needs to be downloaded, but the volume we need to download it to is
			offline, that's a new unique state.
		
		3. The current global state of sync; is sync paused and/or offline.

	]]--
	
	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	local volumeAgents = wfCatalog:getAllVolumeAgents()
	local volumeAgent = volumeAgents[ next( volumeAgents ) ]
	local syncPausedStream = volumeAgent and volumeAgent:isSyncPausedStream() or Rx.Observable.just( false )
	local internetConnectionStream = volumeAgent and volumeAgent:hasInternetConnectionStream() or Rx.Observable.just( true )
	
	return wfCatalog:docStore():bulkDocValue( assetIds )
	
		:flatMapLatest( function( docValues )
		
			local maybeOngoingDocValue = maybeGetOngoingDocValue( wfCatalog, docValues )
			
			return Rx.Observable.combineLatest( 
					maybeOngoingDocValue,
					mappedVolumeStatuses )

		end )
		
		:map( function( docValues, volumeGuidForOriginals, volumeGuidForProxies, 
				onlineVolumes, outOfSpaceVolumes, knownVolumes )
			
			local combinedSyncStatus
			
			for _, docValue in ipairs( docValues ) do
			
				local volumeGuidUsedByOriginal = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
				local volumeGuidUsedByProxy = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
				local isOriginalOnKnownVolume = knownVolumes[ volumeGuidUsedByOriginal ]
				local isProxyOnKnownVolume = knownVolumes[ volumeGuidUsedByProxy ]
				
				local isVolumeForOriginalsOnline = onlineVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOnline = onlineVolumes[ volumeGuidForProxies ]
			
				local isVolumeForOriginalsOutOfSpace = outOfSpaceVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOutOfSpace = outOfSpaceVolumes[ volumeGuidForProxies ]
				
				local isOriginalOnOnlineVolume = onlineVolumes[ volumeGuidUsedByOriginal ] 
				local isProxyOnOnlineVolume = onlineVolumes[ volumeGuidUsedByProxy ]
				
				-- Calculate the sync status.
				
				local syncStatus = WFCatSyncStatus.getAssetStatus( docValue )
				
				local currentDocSyncStatus = WFCatSyncStatus.calculateSyncStatusWithVolumeStatus( 
						syncStatus, volumeGuidUsedByOriginal, volumeGuidUsedByProxy,
						isOriginalOnKnownVolume, isProxyOnKnownVolume,
						isVolumeForOriginalsOnline, isVolumeForProxiesOnline,
						isVolumeForOriginalsOutOfSpace, isVolumeForProxiesOutOfSpace,
						isOriginalOnOnlineVolume, isProxyOnOnlineVolume )
				
				if not combinedSyncStatus then
					combinedSyncStatus = currentDocSyncStatus
				elseif combinedSyncStatus ~= currentDocSyncStatus then
					combinedSyncStatus = nil
					break
				end
					
			end
					
			return combinedSyncStatus
					
		end )
		
		:flatMap( function( syncStatus )
		
			return Rx.Observable.combineLatest( 
				Rx.Observable.just( syncStatus ),
				syncPausedStream,
				internetConnectionStream )
				
		end)
		
		:map( function( syncStatus, isSyncPaused, isOnline )
		
			return WFCatSyncStatus.calculateSyncStatusWithGlobalStatus( syncStatus, isSyncPaused, not isOnline )
		
		end )
		
		:map( WFCatSyncStatus.translateAssetSyncStatus )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedOzStatusFromIdsAsStr( wfCatalog, assetIds )
	
	return wfCatalog:docStore():bulkDocValue( assetIds )
	
		:flatMapLatest( function( docValues )
		
			return maybeGetOngoingDocValue( wfCatalog, docValues )

		end )
		
		:map( function( docValues )
			
			local combinedOzStatusStr
			
			local previousOzHasOriginal
			local previousOzHasProxy
			
			for _, docValue in ipairs( docValues ) do
			
				local ozHasOriginal = not not WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
				local ozHasProxy = not not WFCatFileSpec.getProxyLinkFromDocValue( docValue )
			
				if not combinedOzStatusStr then
				
					if ozHasOriginal then
					
						combinedOzStatusStr = kOzHasOriginal
					
					elseif ozHasProxy then
					
						combinedOzStatusStr = kOzHasProxy
					
					else
					
						combinedOzStatusStr = kOzHasNothing
					
					end
					
				elseif previousOzHasOriginal ~= ozHasOriginal
					or previousOzHasProxy ~= ozHasProxy then
				
					combinedOzStatusStr = nil
					break
					
				end
				
				previousOzHasProxy = ozHasProxy
				
				previousOzHasOriginal = ozHasOriginal
				
			end
		
			return combinedOzStatusStr
		
		end )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getLocalFileTypeForAsset( docValue, isOriginalOnOfflineVolume, isProxyOnOfflineVolume )

	local assetLocalFileTypeId = kAssetHasNothing

	if docValueHasLocalOriginal( docValue ) and not isOriginalOnOfflineVolume then

		if WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue ) then
			assetLocalFileTypeId = kAssetHasLocalOriginalForOffline
		else
			assetLocalFileTypeId = kAssetHasLocalOriginal
		end

	elseif docValueHasLocalProxy( docValue ) and not isProxyOnOfflineVolume then
	
		assetLocalFileTypeId = kAssetHasLocalProxy

	end

	return kAssetLocalFileType[ assetLocalFileTypeId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedLocalFileTypeFromIdsAsStr( wfCatalog, assetIds )

	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	local volumeAgents = wfCatalog:getAllVolumeAgents()
	local volumeAgent = volumeAgents[ next( volumeAgents ) ]
	
	return wfCatalog:docStore():bulkDocValue( assetIds )
	
		:flatMapLatest( function( docValues )
		
			local maybeOngoingDocValue = maybeGetOngoingDocValue( wfCatalog, docValues )
			
			return Rx.Observable.combineLatest( 
					maybeOngoingDocValue,
					mappedVolumeStatuses )

		end )

		:map( function( docValues, volumeGuidForOriginals, 
				volumeGuidForProxies, onlineVolumes )
					
			local combinedLocalFileType
			
			for _, docValue in pairs( docValues ) do
					
				local volumeGuidUsedByOriginal = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
				local volumeGuidUsedByProxy = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
			
				local isOriginalOnOfflineVolume = not onlineVolumes[ volumeGuidUsedByOriginal ] 
				local isProxyOnOfflineVolume = not onlineVolumes[ volumeGuidUsedByProxy ]
				
				local docLocalFileType = WFCatSyncStatus.getLocalFileTypeForAsset( docValue, isOriginalOnOfflineVolume, isProxyOnOfflineVolume )
				
				if not combinedLocalFileType then
				
					combinedLocalFileType = docLocalFileType
					
				elseif combinedLocalFileType ~= docLocalFileType then
				
					combinedLocalFileType = nil
					break
				
				end
				
			end
					
			return combinedLocalFileType
					
		end )

		:map( WFCatSyncStatus.translateAssetSyncLocalFileType )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedMarkedForOfflineUseFromIds( wfCatalog, assetIds )

	return wfCatalog:docStore():bulkDocValue( assetIds )
	
		:flatMapLatest( function( docValues )
		
			return maybeGetOngoingDocValue( wfCatalog, docValues )

		end )
		
		:map( function( docValues )

			local combinedMarkedForOfflineUse

			for _, docValue in pairs( docValues ) do

				local markedForOfflineUse =
					not not WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue )

				if combinedMarkedForOfflineUse == nil then
				
					combinedMarkedForOfflineUse = markedForOfflineUse
					
				elseif combinedMarkedForOfflineUse ~= markedForOfflineUse then
				
					combinedMarkedForOfflineUse = nil
					break
					
				end

			end

			return combinedMarkedForOfflineUse

		end )

end

--------------------------------------------------------------------------------

local function asset_status_table()

	local _map = function( docValue, emit )
	
		if isAssetDocument( docValue ) then
		
			local status = WFCatSyncStatus.getAssetStatus( docValue )
			
			if status then

				local originalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
				local proxyVolumeGuid = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )

				local assetId = docValue._id

				emit( assetId, status, originalVolumeGuid, proxyVolumeGuid )
				
			end
			
		end
		
	end

	-- SPAM [jandrew 2016-05-31]: 
	-- The contents of the asset sync status table are invalidated when switching 
	-- between file management strategies. However, the index store
	-- will automatically refresh any table whose mapping function changes between
	-- app runs. Here, we take advantage of this by providing different mapping functions
	-- based on the file management strategy being used. The only difference
	-- between the methods is the no-op third parameter to the _map method: it's enough
	-- to trigger the rebuild when the user changes modes.

	-- The 4th parameter (also a no-op parameter) is a table version value that is used
	-- to cause reindexing when the implementatin of _map or calls underneath change
	-- the status value.

	-- Note: The last two parameters are what make each mapping function unique.
	-- THEY SHOULD ALWAYS BE RAW NUMBERS, AND NOT VARIABLES.

	local defaultMap = function( docValue, emit )
		_map( docValue, emit, 1, 19 )
	end
	
	local keepAllOriginalsMap = function( docValue, emit )
		_map( docValue, emit, 2, 19 )
	end

	local keepAllProxiesMap = function( docValue, emit )
		_map( docValue, emit, 3, 19 )
	end
 
 	local keepEverythingMap = function( docValue, emit )
		_map( docValue, emit, 4, 19 )
	end
 
	local mapFn 
	
	if NmbFileManagementSettings.keepAllProxies() and NmbFileManagementSettings.keepAllOriginals() then
		mapFn = keepEverythingMap
	elseif NmbFileManagementSettings.keepAllProxies() then
		mapFn = keepAllProxiesMap
	elseif NmbFileManagementSettings.keepAllOriginals() then
		mapFn = keepAllOriginalsMap
	else
		mapFn = defaultMap
	end
	
	return {
		
		sql = [[CREATE TABLE asset_sync_status_ui( assetId, status INTEGER, volumeGuidUsedByOriginal, volumeGuidUsedByProxy )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		map = mapFn,
		docValueFilter = function( docValue )
			return isAssetDocument( docValue )
		end,
	}
	
end

--------------------------------------------------------------------------------

local function asset_status_by_assetId_index( )
	return [[CREATE INDEX asset_sync_status_ui_by_assetId_status ON asset_sync_status_ui( assetId, status )]]
end

--------------------------------------------------------------------------------

local function asset_status_by_status_index( )
	return [[CREATE INDEX asset_sync_status_ui_by_status ON asset_sync_status_ui( status )]]
end

--------------------------------------------------------------------------------

local function asset_local_file_type_table( )

	return {

		sql = [[CREATE TABLE asset_sync_local_file_type_ui( assetId, localFileType INTEGER )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		docValueFilter = function( docValue )
			return isAssetDocument( docValue )
		end,
		map = function( docValue, emit )
			if isAssetDocument( docValue ) then
			
				local TABLE_VERSION
				TABLE_VERSION = TABLE_VERSION or 1  -- a dummy change to force the reindexing

				local localFileType = WFCatSyncStatus.getLocalFileTypeForAsset( docValue )

				if localFileType then

					local assetId = docValue._id	
					emit( assetId, localFileType )

				end

			end
		end,

	}

end

--------------------------------------------------------------------------------

local function asset_local_file_type_by_assetId_index( )
	return [[CREATE INDEX asset_sync_local_file_type_ui_by_assetId_status ON asset_sync_local_file_type_ui( assetId, localFileType )]]
end

--------------------------------------------------------------------------------

local function asset_local_file_type_by_localFileType_index( )
	return [[CREATE INDEX asset_sync_local_file_type_ui_by_status ON asset_sync_local_file_type_ui( localFileType )]]
end

--------------------------------------------------------------------------------

local function album_status_table( )

	return {

		sql = [[CREATE TABLE album_sync_status_ui( albumId, status INTEGER )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		docValueFilter = function( docValue )
			return docValue.type == 'album'
		end,
		map = function( docValue, emit )
			if docValue and docValue.type == 'album' then

				local TABLE_VERSION
				TABLE_VERSION = TABLE_VERSION or 2  -- a dummy change to force the reindexing

				local albumId = docValue._id

				if isLatestMetadataUploaded( docValue ) then
					emit( albumId, kAlbumSyncStatus[ kAlbumComplete ] )
				else
					emit( albumId, kAlbumSyncStatus[ kAlbumSyncing ] )
				end

			end
		end,	
	}

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.generateIndexSchema( indexStoreSchema )
	
	indexStoreSchema:addTable( asset_status_table() )
	
	indexStoreSchema:addIndex( asset_status_by_assetId_index() )
	
	indexStoreSchema:addIndex( asset_status_by_status_index() )
	
	indexStoreSchema:addTable( asset_local_file_type_table() )
	
	indexStoreSchema:addIndex( asset_local_file_type_by_assetId_index() )
	
	indexStoreSchema:addIndex( asset_local_file_type_by_localFileType_index() )
	
	indexStoreSchema:addTable( album_status_table() )
	
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusForAlbums( wfCatalog, optIgnoreAlbumSyncStatus )
	
	local albumStatuses
	
	if optIgnoreAlbumSyncStatus then 
	
		-- Don't consider the sync status of the actual album document when determining 
		-- the status to show for the album. Essentially this means that we only consider
		-- the sync status of the underlying asset documents since this also ignores the 
		-- album_asset documents.
		albumStatuses = wfCatalog:indexStore():query {
			sql = [[SELECT albums.albumId FROM albums ]],
			calm = 0.25
		}
		:map( function( albums )
			local statusForAlbum = {}
			
			for _, album in ipairs( albums ) do
				local albumId = album[ 1 ]
				statusForAlbum[ albumId ] = kAlbumComplete -- mark them all complete
			end
			
			return statusForAlbum
		end )
		
	else 
	
		albumStatuses = wfCatalog:indexStore():query {
			sql = [[SELECT albums.albumId, album_sync_status_ui.status 
						FROM album_sync_status_ui
						JOIN albums USING( albumId ) ]],
			calm = 0.25
		}
		:map( function( albumStatuses )
			
			local statusForAlbum = {}
			
			for _, albumStatus in ipairs( albumStatuses ) do
				local albumId = albumStatus[ 1 ]
				local status = albumStatus[ 2 ]
				
				statusForAlbum[ albumId ] = kAlbumSyncStatusById[ status ]
			end
			
			return statusForAlbum
		end )
	
	end
	
	local getSyncingAlbumsQuerySQL = function( numberOfSyncingStatuses )
	
		return string.format( 
				[[SELECT DISTINCT albumId
					FROM album_asset_v2
					WHERE assetId IN
						(
							SELECT assetId FROM asset_sync_status_ui
							WHERE status IN ( ?%s )
						) ]], string.rep( ", ?", numberOfSyncingStatuses - 1 ) )
						
	end
	
	local getSyncingStatusIds = function( allSyncStatuses )
	
		-- Certain statuses (kAssetComplete, kAssetCompleteNoMaster, kAssetInvalidOriginal, etc.) would indicate complete 
		-- asset. All other assets are considered 'syncing'.
		
		local nonSyncingStateSet = 
			KSTableUtils.mapArrayToSet( {
				kAssetComplete,
				kAssetCompleteNoMaster,
				kAssetInvalidOriginal,
				kAssetOriginalDownloadError,
				kAssetProxyDownloadError,
				kAssetDevelopXmpDownloadError,
				kAssetStackComplete } )

		-- Copy status ids into a table that can be unpacked.
		
		local syncingStatusIds = {}

		for state, statusId in pairs( allSyncStatuses ) do

			if not nonSyncingStateSet[ state ] then

				table.insert( syncingStatusIds, statusId )

			end

		end
	
		return syncingStatusIds
	
	end
	
	local syncingStatusIds = getSyncingStatusIds( kAssetSyncStatus )
	
	local query = getSyncingAlbumsQuerySQL( #syncingStatusIds )
	
	local syncingAlbumsFromAssets = wfCatalog:indexStore():query {
			sql = query,
			calm = 3,
			unpack( syncingStatusIds )
		}
		
		:map( function( syncingAlbums )
			local statusForAlbum = {}
			
			for _, syncingAlbum in ipairs( syncingAlbums ) do
				statusForAlbum[ syncingAlbum[ 1 ] ] = kAlbumSyncing
			end
			
			return statusForAlbum
		end )
	
	local volumeAgents = wfCatalog:getAllVolumeAgents()
	local volumeAgent = volumeAgents[ next( volumeAgents ) ]
	local syncPausedStream = volumeAgent and volumeAgent:isSyncPausedStream() or Rx.Observable.just( false )
	local internetConnectionStream = volumeAgent and volumeAgent:hasInternetConnectionStream() or Rx.Observable.just( true )
	
	return Rx.Observable.combineLatest( function( 
			albumStatusMap, albumStatusFromAssetsMap, isSyncPaused, isSyncOnline )

			local combinedAlbumStatuses = {}
			
			for albumId, albumStatus in pairs( albumStatusMap ) do
				local combinedStatus
				
				local albumStatusFromAssets = albumStatusFromAssetsMap[ albumId ]
				
				if albumStatus == kAlbumSyncing or albumStatusFromAssets == kAlbumSyncing then
				
					if isSyncPaused then
					
						combinedStatus = kAlbumSyncPaused
						
					elseif not isSyncOnline then
					
						combinedStatus = kAlbumSyncNoConnection
						
					else
					
						combinedStatus = kAlbumSyncing
						
					end
					
				elseif albumStatus == kAlbumComplete or albumStatusFromAssets == kAlbumComplete then
					
					combinedStatus = kAlbumComplete
					
				else
				
					combinedStatus = albumStatus
				
				end
				
				combinedAlbumStatuses[ albumId ] = combinedStatus
			end
			
			return combinedAlbumStatuses
			
		end, 
	albumStatuses, syncingAlbumsFromAssets, syncPausedStream, internetConnectionStream )
	
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.syncPendingForAnyAsset( wfCatalog )

	return wfCatalog:indexStore():query {
			sql = [[SELECT status FROM asset_sync_status_ui WHERE status NOT IN ( ?, ?, ?, ?, ?, ?, ? ) LIMIT 1]],
			kAssetSyncStatus[ kAssetComplete ],
			kAssetSyncStatus[ kAssetCompleteNoMaster ],
			kAssetSyncStatus[ kAssetInvalidOriginal ],
			kAssetSyncStatus[ kAssetOriginalDownloadError ],
			kAssetSyncStatus[ kAssetProxyDownloadError ],
			kAssetSyncStatus[ kAssetDevelopXmpDownloadError ],
			kAssetSyncStatus[ kAssetStackComplete ]
		}
		:map( function( results )
			return KSTableUtils.derefOpt( results, 1, 1 ) ~= nil
		end )
		:distinctUntilChanged()

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.persistentSyncErrorForAnyAsset( wfCatalog )

	return wfCatalog:indexStore():query {
			sql = [[SELECT status FROM asset_sync_status_ui WHERE status IN ( ?, ?, ?, ? ) LIMIT 1]],
			kAssetSyncStatus[ kAssetInvalidOriginal ],
			kAssetSyncStatus[ kAssetOriginalDownloadError ],
			kAssetSyncStatus[ kAssetProxyDownloadError ],
			kAssetSyncStatus[ kAssetDevelopXmpDownloadError ],
		}
		:map( function( results )
			return KSTableUtils.derefOpt( results, 1, 1 ) ~= nil
		end )
		:distinctUntilChanged()

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusForAssets( wfCatalog )
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT status, count( status ) FROM asset_sync_status_ui GROUP BY status]],
		}
		:map( function( rows )
			local statusCounts = {}
			
			for _, row in ipairs( rows ) do
				local statusId = row[ 1 ]
				local statusCount = row[ 2 ]
				
				if statusId and statusCount then
					local status = kAssetSyncStatusById[ statusId ]
					if status then
						statusCounts[ status ] = tonumber( statusCount ) or 0
					end
				end
			end
			
			for _, status in pairs( kAssetSyncStatusById ) do
				if not statusCounts[ status ] then 
					statusCounts[ status ] = 0
				end
			end
			
			return statusCounts
		end )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.isDoingInitialSync( wfCatalog )

	return wfCatalog:docStore():ongoingLocalDocValue( 'Oz-isDoingInitialSync' )
	
		:map( function( docValue )
		
			local isDoingInitialSync = true
			
			if docValue.isDoingInitialSync ~= nil then
			 
				isDoingInitialSync = docValue.isDoingInitialSync
				
			end
			
			return isDoingInitialSync
			
		end )
		
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.translateAssetSyncStatus( status )
	
	return kAssetSyncStatusById[ status ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.translateAssetSyncLocalFileType( localFileType )
	
	return kAssetLocalFileTypeById[ localFileType ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.completeNoErrorSyncStatus( )
	
	return require 'WFCatSearchBySyncStatus'.completeNoErrorSyncStatus( )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.isAssetCompleted( statusOrId )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAssetSyncStatusById[ statusOrId ]
	end

	return statusOrId == kAssetComplete
		or statusOrId == kAssetCompleteNoMaster
		or statusOrId == kAssetInvalidOriginal
		or statusOrId == kAssetOriginalDownloadError
		or statusOrId == kAssetProxyDownloadError
		or statusOrId == kAssetDevelopXmpDownloadError
		or statusOrId == kAssetStackComplete

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.calculateSyncStatusWithVolumeStatus( 
		statusOrId, hasLocalOriginal, hasLocalProxy, 
		isOriginalOnKnownVolume, isProxyOnKnownVolume,
		isVolumeForOriginalsOnline, isVolumeForProxiesOnline,
		isVolumeForOriginalsOutOfSpace, isVolumeForProxiesOutOfSpace,
		isOriginalOnOnlineVolume, isProxyOnOnlineVolume )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAssetSyncStatusById[ statusOrId ]
	end

	if statusOrId == kAssetNeedsOriginalDownload
		or statusOrId == kAssetNeedsOriginalDownloadForOffline then
		
		if not isVolumeForOriginalsOnline then
		
			-- Original needs to be downloaded to an offline volume.
			statusOrId = kAssetNeedsOriginalDownloadToOfflineVolume
			
		elseif isVolumeForOriginalsOutOfSpace then
		
			-- Original needs to be downloaded to an offline volume.
			statusOrId = kAssetNeedsOriginalDownloadToOutOfSpaceVolume
		
		end
		
	elseif statusOrId == kAssetNeedsProxyDownload then
	
		if not isVolumeForProxiesOnline then
		
			-- Proxy needs to be downloaded to an offline volume.
			statusOrId = kAssetNeedsProxyDownloadToOfflineVolume
			
		elseif isVolumeForProxiesOutOfSpace then
		
			-- Original needs to be downloaded to an offline volume.
			statusOrId = kAssetNeedsProxyDownloadToOutOfSpaceVolume
		
		end
	
	elseif statusOrId == kAssetNeedsOriginalUpload then

		if hasLocalOriginal and not isOriginalOnKnownVolume then
		
			-- Original needs to be migrated to the original volume.
			statusOrId = kAssetNeedsOriginalMigration
		
		elseif hasLocalOriginal and not isOriginalOnOnlineVolume then
		
			-- Original needs to be uploaded, but is on an offline volume.
			statusOrId = kAssetNeedsOriginalUploadFromOfflineVolume
			
		end
		
	elseif statusOrId == kAssetComplete
		or statusOrId == kAssetCompleteNoMaster
		or statusOrId == kAssetInvalidOriginal 
		or statusOrId == kAssetOriginalDownloadError
		or statusOrId == kAssetProxyDownloadError
		or statusOrId == kAssetDevelopXmpDownloadError
		or statusOrId == kAssetStackComplete then
		
		if hasLocalProxy and not isProxyOnKnownVolume then
		
			-- Proxy needs to be migrated to the proxy volume.
			statusOrId = kAssetNeedsProxyMigration
			
		elseif hasLocalProxy and not isProxyOnOnlineVolume then
		
			-- Proxy is downloaded, but on an offline volume.
			statusOrId = kAssetProxyIsOnOfflineVolume
			
		end
		
		if hasLocalOriginal and not isOriginalOnKnownVolume then
		
			-- Original needs to be migrated to the original volume.
			statusOrId = kAssetNeedsOriginalMigration
		
		elseif hasLocalOriginal and not isOriginalOnOnlineVolume then
		
			-- Original is downloaded, but on an offline volume.
			statusOrId = kAssetOriginalIsOnOfflineVolume
			
		end
		
	end

	return kAssetSyncStatus[ statusOrId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.calculateSyncStatusWithGlobalStatus( 
		statusOrId, isSyncPaused, isOffline )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAssetSyncStatusById[ statusOrId ]
	end

	if isSyncPaused or isOffline then
	
		if statusOrId == kAssetNeedsMetadataUpload
			or statusOrId == kAssetNeedsOriginalUpload
			or statusOrId == kAssetNeedsOriginalDownload
			or statusOrId == kAssetNeedsOriginalDownloadForOffline
			or statusOrId == kAssetNeedsProxyDownload
			or statusOrId == kAssetNeedsDevelopXmpDownload
			or statusOrId == kAssetNeedsOriginalDownloadToOfflineVolume
			or statusOrId == kAssetNeedsProxyDownloadToOfflineVolume
			or statusOrId == kAssetNeedsOriginalUploadFromOfflineVolume
			or statusOrId == kAssetNeedsOriginalDownloadToOutOfSpaceVolume
			or statusOrId == kAssetNeedsProxyDownloadToOutOfSpaceVolume then
			
			statusOrId = isOffline and kAssetSyncNoConnection or kAssetSyncPaused
			
		end
		
	end

	return kAssetSyncStatus[ statusOrId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.offlineVolumesRequiringSyncing( wfCatalog )

	local syncStatusQuery = wfCatalog:indexStore():query {
		sql = [[
			SELECT DISTINCT status, volumeGuidUsedByOriginal, volumeGuidUsedByProxy
			FROM asset_sync_status_ui
			WHERE status IN ( ?, ?, ?, ? )
		]],
		kAssetSyncStatus[ kAssetNeedsOriginalDownload ],
		kAssetSyncStatus[ kAssetNeedsOriginalDownloadForOffline ],
		kAssetSyncStatus[ kAssetNeedsProxyDownload ],
		kAssetSyncStatus[ kAssetNeedsOriginalUpload ],
		calm = 1,
	}
	
	-- Combine the asset sync status query with the volume info status stream. We
	-- use this information to map the asset sync status to a new status based
	-- on the connection status of the volumes.
	
	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	local volumesRequiringSyncingStream 
		= Rx.Observable.combineLatest( syncStatusQuery, mappedVolumeStatuses )
		
			:map( function( rows, volumeGuidForOriginals, volumeGuidForProxies, 
					onlineVolumes, outOfSpaceVolumes, knownVolumes )
		
				local volumesRequiringSync = {}
			
				local isVolumeForOriginalsOnline = onlineVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOnline = onlineVolumes[ volumeGuidForProxies ]
			
				local isVolumeForOriginalsOutOfSpace = outOfSpaceVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOutOfSpace = outOfSpaceVolumes[ volumeGuidForProxies ]
			
				for _, row in ipairs( rows ) do
				
					local syncStatusId = row[ 1 ]
					local volumeGuidUsedByOriginal = row[ 2 ]
					local volumeGuidUsedByProxy = row[ 3 ]
					
					local isOriginalOnKnownVolume = knownVolumes[ volumeGuidUsedByOriginal ]
					local isProxyOnKnownVolume = knownVolumes[ volumeGuidUsedByProxy ]
					
					local isOriginalOnOnlineVolume = onlineVolumes[ volumeGuidUsedByOriginal ] 
					local isProxyOnOnlineVolume = onlineVolumes[ volumeGuidUsedByProxy ]
					
					-- Calculate the new sync status.
					
					syncStatusId = WFCatSyncStatus.calculateSyncStatusWithVolumeStatus( 
							syncStatusId, volumeGuidUsedByOriginal, volumeGuidUsedByProxy,
							isOriginalOnKnownVolume, isProxyOnKnownVolume,
							isVolumeForOriginalsOnline, isVolumeForProxiesOnline,
							isVolumeForOriginalsOutOfSpace, isVolumeForProxiesOutOfSpace,
							isOriginalOnOnlineVolume, isProxyOnOnlineVolume )
					
					local syncStatus = kAssetSyncStatusById[ syncStatusId ]
					
					if syncStatus == kAssetNeedsOriginalUploadFromOfflineVolume then
					
						volumesRequiringSync[ volumeGuidUsedByOriginal ] = true
					
					elseif syncStatus == kAssetNeedsOriginalDownloadToOfflineVolume then
						
						volumesRequiringSync[ volumeGuidForOriginals ] = true
						
					elseif syncStatus == kAssetNeedsProxyDownloadToOfflineVolume then
					
						volumesRequiringSync[ volumeGuidForProxies ] = true
					
					end
					
				end
			
				return volumesRequiringSync
			
			end )
			
	return volumesRequiringSyncingStream	
		
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getAssetSyncStatusList()

	return kAssetSyncStatus

end

--------------------------------------------------------------------------------

return WFCatSyncStatus
