
--[[----------------------------------------------------------------------------

WFCatSynchronizer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.oz_client.managed_catalog
local WFCatBinaryUploader = require 'WFCatBinaryUploader'
local WFCatConflictResolver = require 'WFCatConflictResolver'
local NmbErrorHandler = require 'NmbErrorHandler'
local WFCatIdleDownloader = require 'WFCatIdleDownloader'
local WFCatSyncClient = require 'WFCatSyncClient'
local WFCatStorage = require 'WFCatStorage'
local NmbFileManagementSettings = require 'NmbFileManagementSettings'
local WFCatSyncAutoDeleter = require 'WFCatSyncAutoDeleter'
local WFCatStaleBinaryRemover = require 'WFCatStaleBinaryRemover'
local WFCatStyleSynchronizer = require 'WFCatStyleSynchronizer'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFConfiguration = require 'WFConfiguration'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

--------------------------------------------------------------------------------
	
local debug_disableSync = WFConfiguration.Sync.disable or false

local check = assert( WFArguments.check )

--==============================================================================

local WFCatSynchronizerMethods = {

	_wfCatalog = false,
		-- (WFCatalog)

	_pixelProviderProcessor = false,
		-- (PixelProviderProcessor)

	_metadataSyncClient = false,
		-- (WFCatSyncClient)
		
	_ozStorageMonitor = false,
		-- (WFCatStorage)

	_conflictResolver = false,
		-- (WFCatConflictResolver)
		
	_autoDeleter = false,
		-- (WFCatSyncAutoDeleter)
		
	_profileSyncClient = false,
		-- (WFCatSyncClient)

	_idleDownloader = false,
		-- (WFCatIdleDownloader)

	_staleBinaryRemover = false,
		-- (WFCatStaleBinaryRemover)

	_binaryUploader = false,
		-- (WFCatBinaryUploader)
	
	_styleSynchronizer = false,
		-- (WFCatStyleSynchronizer)
		
	_errorHandler = false,
		-- (WFCatErrorHandler)
		
	_isSyncRunning = false,
		-- (boolean)
		
	_residualMode = false,
		-- (boolean)

	_restrictedCountryMode = false,
		-- (boolean)

	_idleDownloaderDelaySubscription = false,
		-- (Rx Timer Subscription)
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local function buildDownloadBinaryStream( pixelProviderProcessor, idlePixStreamIds )

	local pixStreamIdsLut = KSTableUtils.invert( idlePixStreamIds )

	return pixelProviderProcessor
			:relayStream( "pixelPathStream" )
			:filter( 

				function( pixStreamId )

					return not not pixStreamIdsLut[ pixStreamId ]

				end )
			:tap( 

				function( pixStreamId, localAssetId )

					pixelProviderProcessor
						:runUntilDone( 
							"removeRequest", 
								pixStreamId, localAssetId )

				end )

end

--------------------------------------------------------------------------------

local function buildDownloadBinaryFunc( pixelProviderProcessor, idlePixStreamIds )

	return function( localAssetId, binaryType, binarySize, isMarkedForOffline, hasLocalOriginal )

		local pixStreamId = idlePixStreamIds[ binaryType ]

		pixelProviderProcessor
			:runUntilDone( 
				"requestBinaryWhenIdle", 
					pixStreamId, localAssetId, binaryType, binarySize,
					isMarkedForOffline, hasLocalOriginal )
		
	end

end

--------------------------------------------------------------------------------

local function buildBatchCancelDownloadsFunc( pixelProviderProcessor, idlePixStreamIds )

	return function( binaryType, localAssetIds )

		local pixStreamId = idlePixStreamIds[ binaryType ]

		pixelProviderProcessor
			:runUntilDone( 
				"batchRemoveRequests", pixStreamId, localAssetIds )
		
	end

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:__init( wfCatalog, profileCatalog, pixelProviderProcessor, 
	workerThreadAgrs, syncEnabled, residualMode )

	check( 'WFCatSynchronizer', 'WFCatalog', '?WFCatalog', 'PixelProviderProcessor', 'table', '?boolean', '?boolean' )

	self._wfCatalog = wfCatalog
	self._pixelProviderProcessor = pixelProviderProcessor
	
	self._conflictResolver = WFCatConflictResolver( wfCatalog:docStore(), wfCatalog:indexStore() )
	self._ozStorageMonitor = WFCatStorage( wfCatalog )
	
	self._metadataSyncClient = WFCatSyncClient( wfCatalog )
	
	self._styleSynchronizer = WFCatStyleSynchronizer( wfCatalog, workerThreadAgrs )
	
	-- monitor quota
	self._ozStorageMonitor:isOverStorage()
		:distinctUntilChanged()
		:map( function( isOverStorage )
			self._metadataSyncClient:setUserIsOverQuota( isOverStorage )
		end )
		:call( WFRxRunUntilDone )
	
	-- monitor expiration
	wfCatalog:ozUser():isSubscriptionExpired()
		:distinctUntilChanged()
		:map( function( isExpired )
			self._metadataSyncClient:setUserIsExpired( isExpired )
		end )
		:call( WFRxRunUntilDone )
	
	self._autoDeleter = WFCatSyncAutoDeleter( wfCatalog, self._metadataSyncClient:getDeletedDocumentIdStream() )
	
	if profileCatalog then					  
		self._profileSyncClient = WFCatSyncClient( profileCatalog, {
																	metadataPullerArgs = {
																	 	progressiveBackoff = false,
																		noWorkDelay = 60 * 60 * 24, -- 1 day
																		errorDelay = 60 * 60, -- 1 hour
																		maxDelay = 60 * 60 * 24, -- 1 day 
																	}
																} )
		self._profileSyncClient:addCloseAction( function()
			profileCatalog:close()
		end )
	end
	
	local idlePixStreamBaseId = string.format( "%s:%s", 'WFCatSynchronizer', KSIdGenerator.uuidInLowercaseHex() )
	
	local idlePixStreamIds = {}
	idlePixStreamIds[ 'original' ] = string.format( "%s:%s", idlePixStreamBaseId, 'original' )
	idlePixStreamIds[ 'proxy' ] = string.format( "%s:%s", idlePixStreamBaseId, 'proxy' )
	idlePixStreamIds[ 'xmp_develop' ] = string.format( "%s:%s", idlePixStreamBaseId, 'xmp_develop' )
	
	local idleDownloadStream = buildDownloadBinaryStream( pixelProviderProcessor, idlePixStreamIds )
	local idleDownloadFunc = buildDownloadBinaryFunc( pixelProviderProcessor, idlePixStreamIds )
	local batchCancelDownloadsFunc = buildBatchCancelDownloadsFunc( pixelProviderProcessor, idlePixStreamIds ) 
	
	self._idleDownloader = WFCatIdleDownloader( 
			wfCatalog, idleDownloadFunc, idleDownloadStream, batchCancelDownloadsFunc,
			NmbFileManagementSettings.keepAllOriginals(),
			NmbFileManagementSettings.keepAllProxies() )

	self._binaryUploader = WFCatBinaryUploader( wfCatalog )

	self._staleBinaryRemover = WFCatStaleBinaryRemover( wfCatalog, self._idleDownloader )

	self._errorHandler = NmbErrorHandler( wfCatalog )
	
	self._residualMode = residualMode
	
	if syncEnabled then
		
		-- delay the start of the idle downloader a bit so that we're not overaggressively pulling binaries at the 
		-- start of sync when other things are likely more important to get through the sync process
		-- The purge process is setup to delay for 15 seconds on startup; it's good to have the idle downloader
		-- delay slightly longer than purge so purge can prevent the idle downloader from running if needed.
		
		self:startSyncing( 20 )
		
	end

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:close()

	check( 'WFCatSynchronizer' )

	self:stopSyncing()
	
	if self._profileSyncClient then 
		self._profileSyncClient:close()
	end
	
	self._autoDeleter:close()
	
	self._metadataSyncClient:close()
	self._ozStorageMonitor:close()
	self._idleDownloader:close()
	self._styleSynchronizer:close()
	self._binaryUploader:close()
	self._errorHandler:close()
	self._staleBinaryRemover:close()
	self._conflictResolver:close()
	
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:startSyncing( maybeIdleDownloaderDelay )
	
	local ozUserId = self._wfCatalog:ozUserId()
	
	if not ozUserId:match( '^0+$' ) and not debug_disableSync and not self._restrictedCountryMode then
		
		if self._isSyncRunning then
			return
		end
		
		self._isSyncRunning = true
		
		-- SPAM [matthewj 2016-06-30]: This ensures that search indexing is turned on for the user. This may
		-- not be the best place for enabling this, but it should work for now. Will need to move - or at least
		-- change it's logic if we allow users the ability to disable search indexing since this currently forces
		-- it to be on if it's not.
		
		-- [anmurark 2018-04-10]: NMB-7722 - Removing the explicit indexing
		-- since all accounts are now enabled for search by default

--[[	local runUntilDone = WFRxRunUntilDone
--		self._wfCatalog:ozUser():userInfoCache():docValue( ozUserId )
--		:flatMap( function( docValue )
--			local is_search_index = KSTableUtils.derefOpt( docValue, 'config', 'search_index' )
--			
--			if not is_search_index then
--				return self._wfCatalog:ozUser():enableFeature( "search_metadata" )
--			else 
--				return Observable.empty()
--			end
--		end )
--		:call( runUntilDone )]]

		if self._profileSyncClient then
			self._profileSyncClient:startSync()
		end
		
		if self._residualMode then
			self._metadataSyncClient:startSyncInResidualMode()
		else
			self._metadataSyncClient:startSync()
		end
		
		self._idleDownloaderDelaySubscription =
			Observable.timer( maybeIdleDownloaderDelay or 0.1, WFRxScheduler.sharedScheduler() )
				:map( function()
					self._idleDownloader:start()
				end )
				:subscribe()
				
		self._styleSynchronizer:start()

		if not self._residualMode then
			self._binaryUploader:startContinuousUploads()
		end

	end

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:stopSyncing()

	if self._profileSyncClient then 
		self._profileSyncClient:stopSync()
	end
	
	if self._residualMode then
		self._metadataSyncClient:stopSyncInResidualMode()
	else
		self._metadataSyncClient:stopSync()
	end
	
	if self._idleDownloaderDelaySubscription then
		self._idleDownloaderDelaySubscription:dispose()
		self._idleDownloaderDelaySubscription = nil
	end
	
	self._idleDownloader:stop()
	
	self._styleSynchronizer:stop()
	
	if not self._residualMode then
		self._binaryUploader:stopContinuousUploads()
	end
	
	self._isSyncRunning = false
	
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:enableResidualMode()
	
	self._residualMode = true

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:disableResidualMode()
	
	self._residualMode = false

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:enableRestrictedCountryMode()
	
	self._restrictedCountryMode = true

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isSyncRunningNow()
	
	return self._isSyncRunning

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isUploadingBinaries()
	return self._binaryUploader:isDoingWork()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isDownloadingRequiredOriginalsOrProxies()
	return self._idleDownloader:isDownloadingRequiredBinary()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isUploadingMetadata()
	return self._metadataSyncClient:isUploadingMetadata()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isDownloadingMetadata()
	return self._metadataSyncClient:isDownloadingMetadata()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isDownloadingStyles()
	return self._styleSynchronizer:isDownloadingStyles()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getLatestDownloadedAssetIdsStream()
	return self._metadataSyncClient:getLatestDownloadedAssetIdsStream()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isOzCatalogDeleted()
	return self._metadataSyncClient:isOzCatalogDeleted()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getIdleDownloader()
	return self._idleDownloader
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getBinaryUploader()
	return self._binaryUploader
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getStyleSynchronizer()
	return self._styleSynchronizer
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getRequiredUpdateStream()

	-- Listen on different threads, all of which make requests to Oz, for a required update error.

	return Rx.Observable.combineLatest( function( maybeRequiresUpdate1, maybeRequiresUpdate2, maybeRequiresUpdate3 )
			return maybeRequiresUpdate1 or maybeRequiresUpdate2 or maybeRequiresUpdate3
		end,
		self._metadataSyncClient:getRequiredUpdateStream(),
		self._profileSyncClient:getRequiredUpdateStream(),
		self._wfCatalog:ozUser():getRequiredUpdateStream() )
		
end

--------------------------------------------------------------------------------

return Rx.Class( 'WFCatSynchronizer', WFCatSynchronizerMethods )
