--[[----------------------------------------------------------------------------

WFConflictResolver.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger= require 'WFLogger'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFTableMerge = require 'WFTableMerge'

	-- WF.oz_client.index_store
local WFQueryDrivenWorkQueueScheduler = require 'WFQueryDrivenWorkQueueScheduler'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFConflictResolver'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFConflictResolver namespace

local WFConflictResolver = {}

--------------------------------------------------------------------------------

local function conflicts_table()

	return {
	
		sql = [[CREATE TABLE conflicts( status )]],
	
		map = function( docValue, emit )

			logger:debugf( 'conflict table reindexing\n\n%s\n\n', docValue )

			-- Do we have any conflicts from any documents?
			
			local syncStatusHash = KSTableUtils.derefOpt( docValue, '_syncStatus' )

			local hasConflicts = false

			if type( syncStatusHash ) == 'table' then
				for syncPeerName, peerStatus in pairs( syncStatusHash ) do
					if peerStatus.conflicts then
						hasConflicts = true
					end
				end
			end
			
			-- TODO_SNOWLESS: [scouten 2013-11-01]: Add a mechanism for distinguishing
			-- between those items that the app has determined require manual intervention
			-- and those that haven't been reviewed yet.

			if hasConflicts then
				emit 'conflict'
				logger:debug( "emit 'conflict'" )
			end
			
		end,
	
	}
	
end

--------------------------------------------------------------------------------

local function conflicts_index()

	return [[CREATE INDEX conflicts_status ON conflicts( status )]]

end

--------------------------------------------------------------------------------

function WFConflictResolver.filterTableDefinitions( tableDefinitions )
	
	tableDefinitions.ozSyncStatusForUI = conflicts_table()

	return tableDefinitions
	
end

--------------------------------------------------------------------------------

function WFConflictResolver.filterIndexDefinitions( indexDefinitions )
	
	indexDefinitions.ozSyncStatusForUI_status = conflicts_index()

	return indexDefinitions
	
end

--------------------------------------------------------------------------------

function WFConflictResolver.generateIndexSchema( indexStoreSchema )

	--check( 'WFIndexStoreSchema' )
		-- Ugh. Yet another reason why I hate mocks in unit tests.
		-- Mocks created at OzCatalogTestUtils.buildMocksForTestingCatalogService
		-- mean that the WFIndexStoreSchema mock object doesn't pass WFArguments
		-- smell test so now I have to disable this safety check. Only discovered
		-- this when I made a change in a yet another unrelated part of the app
		-- that introduced a dependency on this file, but only in Lightroom's
		-- test harness environment. (Thanks to differences how we configure
		-- OzCatalogIndexDefinitions via WFCapabilities, naturally.)

	indexStoreSchema:addTable( conflicts_table() )

	indexStoreSchema:addIndex( conflicts_index() )
	
end

--------------------------------------------------------------------------------

local builtInResolvers = {}

--------------------------------------------------------------------------------

function builtInResolvers.force( docValue )

	local syncStatusOverlay = {}

	for peerName, peerStatus in pairs( docValue._syncStatus ) do
		if peerStatus.conflicts then
			syncStatusOverlay[ peerName ] = { conflicts = { _delete = true } }
		end
	end

	if next( syncStatusOverlay ) then
		return { _syncStatus = syncStatusOverlay }
	end
	
end

--------------------------------------------------------------------------------

local function resolveSubtableByDate( conflictsTable, docValueTable )

	if type( conflictsTable ) ~= 'table' or type( docValueTable ) ~= 'table' then
		return
	end

	-- If either side flags an atomic merge and both sides provide a date, then
	-- we resolve at this level.
	
	if ( conflictsTable._merge == 'atomic' or docValueTable._merge == 'atomic' )
		and type( conflictsTable.date ) == 'string'
		and type( docValueTable.date ) == 'string'
	then

		if conflictsTable.date > docValueTable.date then
		
			local docValueOverlay = { _replace = true }
			for k, v in pairs( conflictsTable ) do
				docValueOverlay[ k ] = v
			end
			
			return docValueOverlay, nil, true
		
		else
		
			return nil, nil, true
		
		end

	end
	
	local overlay = {}
	local conflictsOverlay = {}
	local didResolveAll = true

	for k, v in pairs( conflictsTable ) do
	
		if type( v ) == 'table' then

			local overlayK, conflictsOverlayK, didResolveThis
				= resolveSubtableByDate( conflictsTable[ k ], docValueTable[ k ] )

			overlay[ k ] = overlayK
			
			if conflictsOverlayK then
				conflictsOverlay[ k ] = conflictsOverlayK
			else
				conflictsOverlay[ k ] = { _delete = true }
			end
			
			didResolveAll = didResolveAll and didResolveThis
		
		else
		
			conflictsOverlay[ k ] = v
			didResolveAll = false
			
		end
	
	end

	if didResolveAll then
		conflictsOverlay = nil
	end
	
	if not next( overlay ) then
		return nil, conflictsOverlay, didResolveAll
	else
		return overlay, conflictsOverlay, didResolveAll
	end

end

--------------------------------------------------------------------------------

function builtInResolvers.date( docValue )

	local mainOverlay
	local syncStatusOverlay = {}

	for peerName, peerStatus in pairs( docValue._syncStatus ) do

		if peerStatus.conflicts then

			local overlay, conflictsOverlay, didResolveFully
				= resolveSubtableByDate( peerStatus.conflicts, docValue )

			if conflictsOverlay then

				conflictsOverlay._replace = true
				syncStatusOverlay[ peerName ] = { conflicts = conflictsOverlay }

			elseif didResolveFully then

				syncStatusOverlay[ peerName ] = { conflicts = { _delete = true } }

			end

			if overlay then
				mainOverlay = overlay
				break
			end

		end

	end

	if next( syncStatusOverlay ) then
		mainOverlay = mainOverlay or {}
		mainOverlay._syncStatus = syncStatusOverlay
	end
	
	if mainOverlay and next( mainOverlay ) then
		return mainOverlay
	end
	
end

--------------------------------------------------------------------------------

function WFConflictResolver.makeSimpleResolver( docStore, indexStore )

	check( 'WFDocStore', 'WFIndexStore' )

	return WFConflictResolver.makeResolver( docStore, indexStore, { 'force' } )

end

--------------------------------------------------------------------------------

function WFConflictResolver.makeResolver( docStore, indexStore, strategies )

	check( 'WFDocStore', 'WFIndexStore', 'table' )

	for i = 1, #strategies do
		local ti = type( strategies[ i ] )
		if ti == 'string' then
			assert( builtInResolvers[ strategies[ i ] ], "Unknown built-in strategy" )
		else
			assert( ti == 'function', "Strategy must be known built-in string or function" )
		end
	end

	local queryB = indexStore:queryB [[SELECT docId FROM conflicts
										WHERE status == 'conflict'
										ORDER BY docId
										LIMIT 50]]

	local function resolveOneConflict( conflictQueryRow )
	
		local id = conflictQueryRow[ 1 ]
	
		return docStore:docValueP( id, 'main' )

			:mapDataP( function( _, docValue )
			
				local syncStatusHash = docValue._syncStatus
				if not syncStatusHash then return 'done', 'nothing to resolve' end

				local strategyIndex = 1
				local didChange = false

				while docValue._syncStatus do
				
					local strategyFn = strategies[ strategyIndex ]
					strategyIndex = strategyIndex + 1

					local overlay = {}
					
					if type( strategyFn ) == 'string' then
						strategyFn = builtInResolvers[ strategyFn ]
					end
		
					if strategyFn then
						overlay = strategyFn( docValue )
					else
						break
					end
					
					if type( overlay ) == 'table' and next( overlay ) then
						didChange = true
						docValue = WFTableMerge.applyOverlay( docValue, overlay )
					end
				
				end

				if didChange then
					return docStore:setDocValueP( docValue )
				end

			end )
		
	end

	return WFQueryDrivenWorkQueueScheduler( queryB,
											3, -- maxConcurrentTasks
											60, -- blackoutWindow
											resolveOneConflict )

end

--------------------------------------------------------------------------------

return WFConflictResolver
