--[[----------------------------------------------------------------------------

WFConflictResolverTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.self_tests
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.sync_engine
local WFConflictResolver = require 'WFConflictResolver'
local WFSyncAgent = require 'WFSyncAgent'
local WFTestSyncPeer = require 'WFTestSyncPeer'


--============================================================================--

local WFConflictResolverTest = {}

local docStore, indexStore, testSyncPeer, syncAgent, conflictResolver

--------------------------------------------------------------------------------

function WFConflictResolverTest.setUp()

	docStore = WFDocStore( ':memory:' )

	testSyncPeer = WFTestSyncPeer()
	syncAgent = WFSyncAgent( docStore, testSyncPeer, 'test' )

	indexStore = WFIndexStore( docStore, 'ConflictResolverTestIndexDefinitions.lua' )

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.tearDown( path )

	if conflictResolver then
		conflictResolver:close()
		conflictResolver = nil
	end

	if syncAgent then
		syncAgent:close()
		syncAgent = nil
	end

	if testSyncPeer then
		testSyncPeer:close()
		testSyncPeer = nil
	end

	if indexStore then
		if path then
			indexStore:DEBUG_dumpCoverageP( path )
		end
		indexStore:close()
		indexStore = nil
	end

	if docStore then
		docStore:close()
		docStore = nil
	end

end

--------------------------------------------------------------------------------

local function stripSyncStatusHashIfExists( docValue, syncPeerName )

	local syncStatusHash = KSTableUtils.derefOpt( docValue, '_syncStatus', syncPeerName )

	if docValue._idLocal then
		docValue._idLocal = '< local id stripped >'
	end

	if syncStatusHash then

		if type( syncStatusHash.accepted ) == 'string' then
			syncStatusHash.accepted = syncStatusHash.accepted:gsub(
				'^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d%d%dZ',
				"(valid ISO 8601 timestamp)" )
		end

		local errorHash = KSTableUtils.derefOpt( syncStatusHash, 'error' )

		if errorHash then

			if type( errorHash.message ) == 'string' then
				errorHash.message = errorHash.message:gsub( '^%[.*"%]:%d+: ', '' ) --"
					-- trailing quote fixes broken Halide syntax color
			end

			if type( errorHash.time ) == 'string' then
				errorHash.time = errorHash.time:gsub(
					'^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d%d%dZ',
					"(valid ISO 8601 timestamp)" )
			end

		end

	end

end

--------------------------------------------------------------------------------

local function mapChangedDocs_strip( _, changedDocs )

	for i = 1, #changedDocs do

		local docValue = changedDocs[ i ]

		docValue._hash = nil
		docValue._changeCounter = nil

		if docValue._idLocal then
			docValue._idLocal = '< local id stripped >'
		end

		stripSyncStatusHashIfExists( docValue, 'test' )
		stripSyncStatusHashIfExists( docValue, 'test2' )

	end

	changedDocs.continueFrom = nil

	return 'data', changedDocs

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.testSimpleResolver( test )

	WFAsyncTestCase( test, 5 )

		-- Start with a simple document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Send it to remote side.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_idLocal = '< local id stripped >',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- And the sync peer's :sendUpdatesP method should have been called once.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Change doc value on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )

		-- And change it locally, so as to introduce a conflict.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And, voila! We should have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_idLocal = '< local id stripped >',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
							conflicts = {
								foo = 45,
							},
						},
					},
					foo = 46,
				}
			} )

		-- Now, :sendUpdatesP should refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Now fire up the conflict resolver.
		
		:waitFor( function()
		
			conflictResolver = WFConflictResolver.makeSimpleResolver( docStore, indexStore )
			
			return docStore:docValueB( 'testDocId' )
				:mapB( function( docValue )
				
					if type( docValue ) == 'table' and WFUnknown.isKnown( docValue ) then

						docValue._hash = nil
						docValue._changeCounter = nil

						stripSyncStatusHashIfExists( docValue, 'test' )

					end

					return docValue
				
				end )
		
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_idLocal = '< local id stripped >',
				_slot = 'main',
				_syncStatus = {
					test = {
						accepted = "(valid ISO 8601 timestamp)",
						timestamp = 2,
					},
				},
				foo = 46,
			}

		-- Now, :sendUpdatesP should send the change through.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see new doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_idLocal = '< local id stripped >',
					_slot = 'main',
					foo = 46,
				}
			} )

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.testResolveAtomicByDate_happyPath1( test )

	WFAsyncTestCase( test, 5 )

		-- Fast forward to the conflict state.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						oz = {
							accepted = "2014-01-12T00:21:23.724Z",
							conflicts = {
								reviews = {
									[ "992266c9b920ae4b2d149c1666918ca9" ] = {
										_merge = 'atomic',
										date = "2014-01-13T20:45:01Z",
										flag = 'reject',
									},
								},
							},
							timestamp = "2014-01-13T20:45:01.125654Z",
						},
					},
					captureDate = "2013-06-05T12:40:36",
					changedAtTime = "2014-01-13T20:45:00.543Z",
					height = 2880,
					reviews = {
						[ "992266c9b920ae4b2d149c1666918ca9" ] = {
							_merge = 'atomic',
							date = "2014-01-13T20:45:00",
							flag = 'pick',
						},
					},
					type = 'asset',
					version = 1,
					width = 2160,
				}

		end )

		-- Now fire up the conflict resolver.
		
		:waitFor( function()
		
			conflictResolver = WFConflictResolver.makeResolver( docStore, indexStore,
																{ 'date' } )
			
			return docStore:docValueB( 'testDocId' )
				:mapB( function( docValue )
				
					if type( docValue ) == 'table' and WFUnknown.isKnown( docValue ) then

						docValue._hash = nil
						docValue._changeCounter = nil

						stripSyncStatusHashIfExists( docValue, 'test' )

					end

					return docValue
				
				end )
		
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_idLocal = '< local id stripped >',
				_slot = 'main',
				_syncStatus = {
					oz = {
						accepted = "2014-01-12T00:21:23.724Z",
						timestamp = "2014-01-13T20:45:01.125654Z",
					},
				},
				captureDate = "2013-06-05T12:40:36",
				changedAtTime = "2014-01-13T20:45:00.543Z",
				height = 2880,
				reviews = {
					[ "992266c9b920ae4b2d149c1666918ca9" ] = {
						_merge = 'atomic',
						date = "2014-01-13T20:45:01Z",
						flag = 'reject',
					},
				},
				type = 'asset',
				version = 1,
				width = 2160,
			}

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.testResolveAtomicByDate_happyPath2( test )

	WFAsyncTestCase( test, 5 )

		-- Fast forward to the conflict state.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						oz = {
							accepted = "2014-01-12T00:21:23.724Z",
							conflicts = {
								reviews = {
									[ "992266c9b920ae4b2d149c1666918ca9" ] = {
										_merge = 'atomic',
										date = "2014-01-13T20:45:01Z",
										flag = 'reject',
									},
								},
							},
							timestamp = "2014-01-13T20:45:01.125654Z",
						},
					},
					captureDate = "2013-06-05T12:40:36",
					changedAtTime = "2014-01-13T20:45:00.543Z",
					height = 2880,
					reviews = {
						[ "992266c9b920ae4b2d149c1666918ca9" ] = {
							_merge = 'atomic',
							date = "2014-01-13T20:45:44",
							flag = 'pick',
						},
					},
					type = 'asset',
					version = 1,
					width = 2160,
				}

		end )

		-- Now fire up the conflict resolver.
		
		:waitFor( function()
		
			conflictResolver = WFConflictResolver.makeResolver( docStore, indexStore,
																{ 'date' } )
			
			return docStore:docValueB( 'testDocId' )
				:mapB( function( docValue )
				
					if type( docValue ) == 'table' and WFUnknown.isKnown( docValue ) then

						docValue._hash = nil
						docValue._changeCounter = nil

						stripSyncStatusHashIfExists( docValue, 'test' )

					end

					return docValue
				
				end )
		
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_idLocal = '< local id stripped >',
				_slot = 'main',
				_syncStatus = {
					oz = {
						accepted = "2014-01-12T00:21:23.724Z",
						timestamp = "2014-01-13T20:45:01.125654Z",
					},
				},
				captureDate = "2013-06-05T12:40:36",
				changedAtTime = "2014-01-13T20:45:00.543Z",
				height = 2880,
				reviews = {
					[ "992266c9b920ae4b2d149c1666918ca9" ] = {
						_merge = 'atomic',
						date = "2014-01-13T20:45:44",
						flag = 'pick',
					},
				},
				type = 'asset',
				version = 1,
				width = 2160,
			}

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.testResolveAtomicByDate_partialResolution( test )

	WFAsyncTestCase( test, 5 )

		-- Fast forward to the conflict state.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						oz = {
							accepted = "2014-01-12T00:21:23.724Z",
							conflicts = {
								reviews = {
									[ "992266c9b920ae4b2d149c1666918ca9" ] = {
										_merge = 'atomic',
										date = "2014-01-13T20:45:01Z",
										flag = 'reject',
									},
								},
								blah = 42,
							},
							timestamp = "2014-01-13T20:45:01.125654Z",
						},
					},
					blah = 45,
					captureDate = "2013-06-05T12:40:36",
					changedAtTime = "2014-01-13T20:45:00.543Z",
					height = 2880,
					reviews = {
						[ "992266c9b920ae4b2d149c1666918ca9" ] = {
							_merge = 'atomic',
							date = "2014-01-13T20:45:00",
							flag = 'pick',
						},
					},
					type = 'asset',
					version = 1,
					width = 2160,
				}

		end )

		-- Now fire up the conflict resolver.
		
		:waitFor( function()
		
			conflictResolver = WFConflictResolver.makeResolver( docStore, indexStore,
																{ 'date' } )
			
			return docStore:docValueB( 'testDocId' )
				:mapB( function( docValue )
				
					if type( docValue ) == 'table' and WFUnknown.isKnown( docValue ) then

						docValue._hash = nil
						docValue._changeCounter = nil

						stripSyncStatusHashIfExists( docValue, 'test' )

					end

					return docValue
				
				end )
		
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_idLocal = '< local id stripped >',
				_slot = 'main',
				_syncStatus = {
					oz = {
						accepted = "2014-01-12T00:21:23.724Z",
						conflicts = {
							blah = 42,
						},
						timestamp = "2014-01-13T20:45:01.125654Z",
					},
				},
				blah = 45,
				captureDate = "2013-06-05T12:40:36",
				changedAtTime = "2014-01-13T20:45:00.543Z",
				height = 2880,
				reviews = {
					[ "992266c9b920ae4b2d149c1666918ca9" ] = {
						_merge = 'atomic',
						date = "2014-01-13T20:45:01Z",
						flag = 'reject',
					},
				},
				type = 'asset',
				version = 1,
				width = 2160,
			}

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.testResolveAtomicByDate_nestedPartialResolution( test )

	WFAsyncTestCase( test, 5 )

		-- Fast forward to the conflict state.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						oz = {
							accepted = "2014-01-12T00:21:23.724Z",
							conflicts = {
								reviews = {
									[ "992266c9b920ae4b2d149c1666918ca9" ] = {
										_merge = 'atomic',
										date = "2014-01-13T20:45:01Z",
										flag = 'reject',
									},
								},
								foo = {
									blah = 42,
								},
							},
							timestamp = "2014-01-13T20:45:01.125654Z",
						},
					},
					foo = {
						blah = 45,
					},
					captureDate = "2013-06-05T12:40:36",
					changedAtTime = "2014-01-13T20:45:00.543Z",
					height = 2880,
					reviews = {
						[ "992266c9b920ae4b2d149c1666918ca9" ] = {
							_merge = 'atomic',
							date = "2014-01-13T20:45:00",
							flag = 'pick',
						},
					},
					type = 'asset',
					version = 1,
					width = 2160,
				}

		end )

		-- Now fire up the conflict resolver.
		
		:waitFor( function()
		
			conflictResolver = WFConflictResolver.makeResolver( docStore, indexStore,
																{ 'date' } )
			
			return docStore:docValueB( 'testDocId' )
				:mapB( function( docValue )
				
					if type( docValue ) == 'table' and WFUnknown.isKnown( docValue ) then

						docValue._hash = nil
						docValue._changeCounter = nil

						stripSyncStatusHashIfExists( docValue, 'test' )

					end

					return docValue
				
				end )
		
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_idLocal = '< local id stripped >',
				_slot = 'main',
				_syncStatus = {
					oz = {
						accepted = "2014-01-12T00:21:23.724Z",
						conflicts = {
							foo = {
								blah = 42,
							},
						},
						timestamp = "2014-01-13T20:45:01.125654Z",
					},
				},
				foo = {
					blah = 45,
				},
				captureDate = "2013-06-05T12:40:36",
				changedAtTime = "2014-01-13T20:45:00.543Z",
				height = 2880,
				reviews = {
					[ "992266c9b920ae4b2d149c1666918ca9" ] = {
						_merge = 'atomic',
						date = "2014-01-13T20:45:01Z",
						flag = 'reject',
					},
				},
				type = 'asset',
				version = 1,
				width = 2160,
			}

end

--------------------------------------------------------------------------------

function WFConflictResolverTest.testResolveAtomicByDateAndForce( test )

	WFAsyncTestCase( test, 5 )

		-- Fast forward to the conflict state.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						oz = {
							accepted = "2014-01-12T00:21:23.724Z",
							conflicts = {
								reviews = {
									[ "992266c9b920ae4b2d149c1666918ca9" ] = {
										_merge = 'atomic',
										date = "2014-01-13T20:45:01Z",
										flag = 'reject',
									},
								},
								foo = {
									blah = 42,
								},
							},
							timestamp = "2014-01-13T20:45:01.125654Z",
						},
					},
					foo = {
						blah = 45,
					},
					captureDate = "2013-06-05T12:40:36",
					changedAtTime = "2014-01-13T20:45:00.543Z",
					height = 2880,
					reviews = {
						[ "992266c9b920ae4b2d149c1666918ca9" ] = {
							_merge = 'atomic',
							date = "2014-01-13T20:45:00",
							flag = 'pick',
						},
					},
					type = 'asset',
					version = 1,
					width = 2160,
				}

		end )

		-- Now fire up the conflict resolver.
		
		:waitFor( function()
		
			conflictResolver = WFConflictResolver.makeResolver( docStore, indexStore,
																{ 'date', 'force' } )
			
			return docStore:docValueB( 'testDocId' )
				:mapB( function( docValue )
				
					if type( docValue ) == 'table' and WFUnknown.isKnown( docValue ) then

						docValue._hash = nil
						docValue._changeCounter = nil

						stripSyncStatusHashIfExists( docValue, 'test' )

					end

					return docValue
				
				end )
		
		end )
		:continueWhenValueMatches
			{
				_id = 'testDocId',
				_idLocal = '< local id stripped >',
				_slot = 'main',
				_syncStatus = {
					oz = {
						accepted = "2014-01-12T00:21:23.724Z",
						timestamp = "2014-01-13T20:45:01.125654Z",
					},
				},
				foo = {
					blah = 45,
				},
				captureDate = "2013-06-05T12:40:36",
				changedAtTime = "2014-01-13T20:45:00.543Z",
				height = 2880,
				reviews = {
					[ "992266c9b920ae4b2d149c1666918ca9" ] = {
						_merge = 'atomic',
						date = "2014-01-13T20:45:01Z",
						flag = 'reject',
					},
				},
				type = 'asset',
				version = 1,
				width = 2160,
			}

end

--------------------------------------------------------------------------------

return WFConflictResolverTest
