--[[----------------------------------------------------------------------------

WFDocStore.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local KSDate = require 'KSDate'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'

	-- WF.core.threading
local WFWorkerThread = require 'WFWorkerThread'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.doc_store
local WFOzDocStorePath = require 'WFOzDocStorePath'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local WFDebugTraceback_traceback = require 'WFDebugTraceback'.traceback

local RECORD_TRANSACTION_SOURCES = false
local RECORD_API_PERFORMANCE = false

local kIgnoreType = {}


--============================================================================--

-- @mark WFDocStore class

--------------------------------------------------------------------------------
--[[!
	\class WFDocStore

	This class implements a generic document store. It can be thought of as
	a key-value database where the key is any arbitrary (non-empty) string and
	the value is a Lua table, where all values are of the following types:
	
		* boolean
		* number
		* string
		* table
	
	Table keys must be non-empty strings. (If strings, must not contain leading
	underscores.) There is one exception: A document may contain a top-level key
	named \c _localOnly, which is intended to store document content that is
	device-specific and not synchronized with other databases. (Note, however,
	that this is a convention: \c WFDocStore intentionally has no knowledge of
	synchronization.)
	
	Tables must honor the key and value restrictions recursively and may not
	contain metatables. Exception: Nested tables may use numeric values for keys.
	
	Tables must not contain reference cycles.
	
	It's probably worth reading http://guide.couchdb.org/editions/1/en/api.html#documents
	or other primers on CouchDB or Couchbase as a basis for understanding what's
	meant by our use of the term "document database." This database isn't
	implemented using CouchDB, but it is heavily influenced by CouchDB and
	you will understand many of the concepts behind this code if you understand
	CouchDB well.
	
	\c WFDocStore differs from CouchDB in that it allows individual documents
	to have named slots which represent alternative versions of the document
	content. Think of slots in the same sense as git branches, except that
	they apply to a single document at a time, instead of the entire database,
	and there is no revision history per slot. When unspecified, the branch name
	will be "main".
]]--

local docStoreMethods = {

	_path = false,
		-- (string) path to doc store DB

	_docStoreImp = false,
		-- (WFWorkerThread) reference to the worker thread
	
	_docChangeCounterB = false,
		-- (B< number >) most recent doc change counter value. Initially nil.
		-- Lazily instantiated.

	_docValueBehaviors = false,
		-- (fully weak table) document value behaviors that are currently active
			-- key: (string) doc ID + slot name
			-- value: (B< table >) behavior for that document value

	_fieldValueBehaviors = false,
		-- (fully weak table) field value behaviors that are currently active
			-- key: (string) doc ID + slot name + field path
			-- value: (B< table >) behavior for that field value

	_log = false,
		-- (WFLogger) logger for this doc store
	
	_recentErrorLog = false,
		-- (table) contains the most recent errors from the doc store for the
		-- diagnostic log
	
	_transactionSources = false,
		-- (table) used when RECORD_TRANSACTION_SOURCES is true
			-- key: (string) stack trace as string
			-- value: (number) # hits
	
	_perfLog = false,
		-- (table) used when RECORD_API_PERFORMANCE is true
			-- key (string) API call name (or doComplexLogic function name)
			-- value: (table) info on statement (count, min, max, total)
	
	WFCloseSupport.addCloseSupportMethods,

}
		
--------------------------------------------------------------------------------
--[[!
	\brief Constructs a new document database.

	\param path (string) path to the document database.
	
		For testing, you can use the special ID ":memory:" which
		creates a transient catalog in memory, or ":temp:"
		which creates a transient catalog on disk that is deleted
		when the DB is no longer needed. You may optionally add any
		suffix following ":memory:" or ":temp:" to create distinct
		instances of temporary document databases that can be addressed
		from multiple threads.
		
		As a shortcut, you may use a UUID instead of a full file-system
		path. This will construct an appropriate path for you.

		If no database exists at this path, it will be created.
		
		The path string must be non-empty.
]]--

function docStoreMethods:__init( --[[string]] path )

	check( 'WFDocStore', '+string' )

	WFCloseSupport.mustClose( self )
	
	-- Make sure :memory: and :temp: DBs don't conflict with each other.

	if path == ':memory:' or path == ':temp:' then
		path = path .. tostring( KSIdGenerator.localCounter() )
	end

	if #path == 32 and path:match( '^[0-9a-f]+$' ) then
		local catalogDir = WFOzDocStorePath.dirPathForCatalogId( path )
		path = KSPathUtils.child( catalogDir, 'database' )
	end

	self._path = path

	self:setOwned { _docStoreImp = WFWorkerThread( "WFDocStoreMessageProcessor.lua", path ) }

	self._docValueBehaviors = KSTableUtils.makeFullyWeak()
	self._fieldValueBehaviors = KSTableUtils.makeFullyWeak()
		
	self._log = WFLogger( tostring( self ) )

	self._recentErrorLog = {}
	self._perfLog = {}
	
	if RECORD_TRANSACTION_SOURCES then
		self._transactionSources = {}
	end
	
	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

function docStoreMethods:close()

	self._log:info( "doc store closing" )

end

--------------------------------------------------------------------------------

-- @mark shared utilities

--------------------------------------------------------------------------------

local function recordError( self, bucket, message )
	
	local errorLog = self._recentErrorLog
	
	local bucketLog = errorLog[ bucket ] or {}
	errorLog[ bucket ] = bucketLog
	
	bucketLog[ #bucketLog + 1 ] = {
		
		time = KSDate.timeInIso8601FormatWithTimeZone(),
		error = message,
		
	}
	
	if #bucketLog > 10 then
		table.remove( bucketLog, 1 )
	end
		
end

--------------------------------------------------------------------------------

local function recordTransaction( self, methodName )

	if RECORD_TRANSACTION_SOURCES then
	
		local traceback = WFDebugTraceback_traceback { 'WFDocStore.lua', 'Class.lua' }

		local key = string.format( "%s via %s", methodName, tostring( traceback ) )
		self._transactionSources[ key ] = ( self._transactionSources[ key ] or 0 ) + 1
	
	end

end

--------------------------------------------------------------------------------

-- @mark logging

--------------------------------------------------------------------------------
--[[!
	\brief Configure logging level for this doc store.
	
	\param level (string) desired log level (See WFLogger for values.)
	
	Set level to "info" to see API calls with their parameters in the log.
	
	Set level to "debug" to see API calls and the responses in the log.
	Note that logging at this time only tracks promise results. Behavior values
	are not currently logged.
]]--

function docStoreMethods:setLogLevel( --[[string]] level )

	check( 'WFDocStore', '+string' )
	
	self._log:setLevel( level )

	self._log:infof( "\n\ndocStore:setLogLevel( %s )", level )

end

--------------------------------------------------------------------------------
--[[
	If you are having trouble figuring out how an error is occurring, you can
	enable captureTracebackForFunc for the corresponding function. This will
	cause the calling stack trace to be included when the promise resolves
	in error status.
	
	Use this mode sparingly: It will dramatically decrease performance.
]]--


local captureTracebackForFunc = {
--	docValueP = true,
--	setDocValueP = true,
--	bulkSetDocValueP = true,
--	applyOverlayP = true,
--	copySlotToSlotP = true,
--	quietlyDeleteSlotP = true,
--	docChangeCounterP = true,
--	changedDocsP = true,
}

local captureTracebackForAny = false

for k, v in pairs( captureTracebackForFunc ) do
	if v then captureTracebackForAny = true end
end

--------------------------------------------------------------------------------

local loggedOnce = false -- always do this once for code coverage purposes

local function maybeLogResponseP( self, funcName, promise )

	if ( not loggedOnce )
		or captureTracebackForAny
		or self._log:will_debug()
		or RECORD_API_PERFORMANCE
	then
	
		local traceback = captureTracebackForFunc[ funcName ]
			and WFDebugTraceback_traceback { 'WFDocStore.lua' }
		
		local startTime = debug.timeNow()
		
		return promise:observeP( function( status, ... )
		
			loggedOnce = true
			
			local elapsedTime = debug.timeNow() - startTime

			if status == 'error' and traceback then
				self._log:errorf( "\n\ndocStore:%s called via %s\n\nresolved with %s, %s %s %s",
								  funcName, tostring( traceback ), status, ... )
			else
				self._log:debugf( "\n\ndocStore:%s resolved with %s, %s %s %s",
								  funcName, status, ... )
			end
			
			if RECORD_API_PERFORMANCE then

				local stat = self._perfLog[ funcName ]

				if stat then
					stat.count = stat.count + 1
					stat.min = math.min( stat.min, elapsedTime )
					stat.max = math.max( stat.max, elapsedTime )
					stat.total = stat.total + elapsedTime
				else
					stat = {
						funcName = funcName,
						count = 1,
						min = elapsedTime,
						max = elapsedTime,
						total = elapsedTime,
					}
					self._perfLog[ funcName ] = stat
				end
			
			end
		
		end )
	
	else
	
		return promise
	
	end

end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the value of a single document.

	\param docId (string or number) the document's unique ID (uuid) or the local
		id (integer)
	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	\return <code>P< table ></code> value for this document

	Note that the document content will be annotated with the following
	system-generated entries (at the top level of the value table):
	
		* \c _id (string): the document's unique ID
		* \c _slot (string): name of the slot that was fetched
		* \c _hash (string): hash value of the slot's contents
		* \c _empty (boolean): true if there isn't actually a value for this doc ID + slot
		* \c _changeCounter (number): a monotonically-increasing change counter
			that is increased whenever any branch of this document is revised
			(The functions \c docChangeCounterB and \c docChangeCounterP will
			return the maximum value of \c _changeCounter across the entire
			database.)
]]--

function docStoreMethods:docValueP( docId, optSlotName )

	check( 'WFDocStore', '+string|number', '?string' )
	
	self._log:infof( "\n\ndocStore:docValueP( %s, %s )", docId, optSlotName )
	
	return maybeLogResponseP( self, 'docValueP',
			self._docStoreImp:requestP( 'docValue', docId, optSlotName ) )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a document and assert that the docType is as expected.
	This is a wrapper around docValueP that asserts that the resulting document
	type matches the docType parameter. If it does not match, the promise
	resolves to an error.

	\param docType (string) the document's unique ID
	\param docId (string or number) the document's unique ID (uuid) or the local
		id (integer)
	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	\return <code>P< table ></code> value for this document, or an error
	if the docType does not match document.type field.
	
	@see docValueP
]]--

function docStoreMethods:docValueCheckTypeP( docType, docId, optSlotName )

	check( 'WFDocStore', '+string', '+string|number', '?string' )
	
	return self:docValueP( docId, optSlotName )
		:mapDataP( function( status, document )

			assert( document.type == docType
					or document._deleted
					or document._empty,
				"Document type invalid. Expected " .. docType
					.. " but got " .. tostring( document.type ) )

			return status, document

		end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves this value of a single document as a behavior.
	
	The behavior will be updated reasonably quickly as changes come in from
	this and other threads.
	
	\warning You should avoid watching many documents via \c docValueB or
		\c fieldValueB simultaneously for performance reasons. Console logs
		will be generated if the number of \c docValueB instances is large.
		(Note that \c fieldValueB uses \c docValueB in its implementation,
		but the performance concern is around the number of \b documents,
		not the number of fields, being watched.)

	\param docId (string or number) the document's unique ID (uuid) or the local
		id (integer)
	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	\return <code>B< table ></code> value for this document

	Note that the document content will be annotated with the following
	system-generated entries (at the top level of the value table):
	
		* \c _id (string): the document's unique ID
		* \c _slot (string): name of the slot that was fetched
		* \c _hash (string): hash value of the slot's contents
		* \c _empty (boolean): true if there isn't actually a value for this doc ID + slot
		* \c _changeCounter (number): a monotonically-increasing change counter
			that is increased whenever any branch of this document is revised
			(The functions \c docChangeCounterB and \c docChangeCounterP will
			return the maximum value of \c _changeCounter across the entire
			database.)

	In most cases (unless an existing doc value behavior can be reused), the
	value will be initially set to <code>WFUnknown.value()</code> until a
	correct value can be retrieved from the worker thread.
]]--

function docStoreMethods:docValueB( docId, optSlotName )
	
	check( 'WFDocStore', '+string|number', '?string' )

	self._log:infof( "\n\ndocStore:docValueB( %s, %s )", docId, optSlotName )

	local slotName = optSlotName or 'main'

	local docValueBehaviors = self._docValueBehaviors
	local key = string.format( '%s :: %s', tostring( docId ), slotName )
	
	local resultB = WFCloseSupport.fetchCached( docValueBehaviors, key )

	if not resultB then

		resultB = self._docStoreImp:requestE( 'docValueB', docId, optSlotName )
			:closingHoldB( WFUnknown.value() )

		docValueBehaviors[ key ] = resultB
		self:addCloseTarget( resultB )
		
	end
	
	return resultB:autoCloseRelayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a document and assert that the docType is as expected.
	This is a wrapper around docValueB that asserts that the resulting document
	type matches the docType parameter.  If it does not match, the behavior
	transitions to an error state.

	\param docType (string) the document's unique ID
	\param docId (string or number) the document's unique ID (uuid) or the local
		id (integer)
	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	\return <code>B< table ></code> value for this document, or an error
	if the docType does not match document.type field.
	
	@see docValueP
]]--

function docStoreMethods:docValueCheckTypeB( docType, docId, optSlotName )

	check( 'WFDocStore', '+string', '+string|number', '?string' )
	
	return self:docValueB( docId, optSlotName )
		:closingMapB( function( document )

			assert( document.type == docType
				or document._deleted
				or document._empty, "Document type invalid.  Expected " .. docType
				.. " but got " .. tostring( document.type ) )

			return document

		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the document value, replacing any existing value.
	
	\param value (table) new value for the entire document
	
	\c value \b must contain the following elements:
		* \c _id (string) the document's unique ID
		* \c _slot (string) the slot name to write to (often "main")
	
	\c value \b may contain the following elements:
		* \c _hash (string): if present, the current document value must
			have the given hash value or the request will fail
		* \c _changeCounter (number): if present, the current document value
			must have the given change counter value or the request will fail.
			The special value ':new:' can be used to assert that there is no
			existing value for this document+slot combination.
		* \c _deleted (boolean): if present, must be \c true and no other
			values may be present
		* \c _syncStatus (table): reserved for use by sync engine only

	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:setDocValueP( value )

	check( 'WFDocStore', 'table' )
	
	self._log:infof( "\n\ndocStore:setDocValueP( %s )\n\nvia %s", value, ':traceback:' )
	recordTransaction( self, 'setDocValueP' )
	
	return maybeLogResponseP( self, 'setDocValueP',
		self._docStoreImp:requestP( 'setDocValue', value ) )
		:observeP( function( status, message )
			
			if status ~= 'data' then
				recordError( self, 'setDocValueP', message )
			end
			
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets many document values at once, replacing any existing values.
	
	\param values (table) array of entire document values
	
	The array may contain zero or more document values. Each value entry
	looks like the \c value parameter to \c :setDocValue.
	
	If an error occurs while setting any one of these document values, then the
	entire set of document values will be discarded. In other words, this set
	of documents succeeds or fails atomically.

	\return <code>P< boolean ></code> resolves to \c true when done.
		You may safely ignore this promise.
]]--

function docStoreMethods:bulkSetDocValueP( values )

	check( 'WFDocStore', 'table' )
	
	self._log:infof( "\n\ndocStore:bulkSetDocValueP( %s )\n\nvia %s", values, ':traceback:' )
	recordTransaction( self, 'bulkSetDocValueP' )
	
	return maybeLogResponseP( self, 'bulkSetDocValueP',
		self._docStoreImp:requestP( 'bulkSetDocValue', values ) )
		:observeP( function( status, message )
			
			if status ~= 'data' then
				recordError( self, 'bulkSetDocValueP', message )
			end
			
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

-- @mark partial document values

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a behavior) the current value of a field in this document.
	
	\param docId (string or number) the document's unique ID (uuid) or the local
		id (integer)
	
	\param fieldPath (string) The field path, a slash-delimited string like
		a Unix paths, except that the root is implied. Use the path to
		identify values within tables or nested tables.	You can use number
		values to get the nth member of a table; for example, "tabA/1" to get
		the first member of an array in "tabA".

	\param optSlotName (optional, string) optional name of a slot for this
		document (default = "main")
	
	The behavior will be updated reasonably quickly as changes come in from
	this and other threads.
	
	\warning You should avoid watching many documents via \c docValueB or
		\c fieldValueB simultaneously for performance reasons. Console logs
		will be generated if the number of \c docValueB instances is large.
		(Note that \c fieldValueB uses \c docValueB in its implementation,
		but the performance concern is around the number of \b documents,
		not the number of fields, being watched.)
	
	\return <code>B< value ></code> A behavior whose value is
		the current value of the field.
]]--

function docStoreMethods:fieldValueB( docId, fieldPath, optSlotName )
	
	check( 'WFDocStore', '+string|number', '+string', '?+string' )

	return self:fieldValueCheckTypeB( kIgnoreType, docId, fieldPath, optSlotName )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a behavior) the current value of a field in this document
	and also checks that the document is of the expected type.
	
	\see fieldValueB
	
	\param docType (string) The document's type.
	
	\param docId (string or number) the document's unique ID (uuid) or the local
		id (integer)
	
	\param fieldPath (string) The field path, as in fieldValueB.

	\param optSlotName (optional, string) optional name of a slot.
		
	\return <code>B< value ></code> A behavior whose value is
		the current value of the field.
]]--

function docStoreMethods:fieldValueCheckTypeB( docType, docId, fieldPath, optSlotName )

	check( 'WFDocStore', '+string|table', '+string|number', '+string', '?string' )
	
	if type( docType ) == 'table' then
		assert( docType == kIgnoreType )
	end
	
	self._log:infof( "\n\ndocStore:fieldValueB( %s, %s, %s )", docId, fieldPath, optSlotName )

	local slotName = optSlotName or 'main'
	
	local key = string.format( '%s:%s:%s:%s',
							   docType == kIgnoreType and "ignored" or docType,
							   docId, slotName, fieldPath )
	
	local resultB = WFCloseSupport.fetchCached( self._fieldValueBehaviors, key )

	if not resultB then

		local fieldPathSegments = {}
		local nSegments = 0
		
		fieldPath:gsub( '([^/]+)',
			function( segment )
				nSegments = nSegments + 1
				if segment:match( '^%-?%d+$' ) then
					fieldPathSegments[ nSegments ] = tonumber( segment )
				else
					fieldPathSegments[ nSegments ] = segment
				end
			end )
		
		if nSegments == 0 then
			error( "WFDocStore:fieldValueB: fieldPath didn't contain a path", 2 )
		end
		
		resultB = self:docValueB( docId, slotName )
			:closingMapB( function( docValue )
	
				if WFUnknown.isUnknown( docValue ) then
					return WFUnknown.value()
				end
				
				if docType ~= kIgnoreType then
					if docValue.type ~= docType
						and not docValue._empty
						and not docValue._deleted
					then
						error( "Document has wrong type. Expected " .. docType
							   .. " but got " .. tostring( docValue.type ) )
					end
				end
	
				local t = docValue
				
				-- All parent nodes must be tables. If they are anything else,
				-- we just pretend they don't exist since we can't index
				-- into them. (Note that we don't support userdata values
				-- in documents since they don't persist or translate cleanly
				-- to JSON.)
				
				for i = 1, nSegments - 1 do
					t = t and t[ fieldPathSegments[ i ] ]
					if type( t ) ~= 'table' then
						t = nil
					end
				end
	
				return t and t[ fieldPathSegments[ nSegments ] ]
			
			end )

		self:addCloseTarget( resultB )

		self._fieldValueBehaviors[ key ] = resultB
		
	end
	
	return resultB:autoCloseRelayB()

end

--------------------------------------------------------------------------------
--[[!
	\brief Updates the document value, applying an overlay to any existing value
	
	\param overlay (table) delta to apply to the existing document value
	
	\c overlay \b must contain the following elements:
		* \c _id (string or number) the document's unique ID (uuid) or the local
			id (integer)
		* \c _slot (string) the slot name to write to (often "main")
	
	\c overlay \b may contain the following elements:
		* \c _hash (string) if present, the current document value must
			have the given hash value or the request will fail
		* \c _changeCounter (number) if present, the current document value
			must have the given change counter value or the request will fail.
			The special value ':new:' can be used to assert that there is no
			existing value for this document+slot combination.

	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:applyOverlayP( overlay )

	check( 'WFDocStore', 'table' )
	
	self._log:infof( "\n\ndocStore:applyOverlayP( %s )\n\nvia %s", overlay, ':traceback:' )
	recordTransaction( self, 'applyOverlayP' )

	return maybeLogResponseP( self, 'applyOverlayP',
		self._docStoreImp:requestP( 'applyOverlay', overlay )
		:runUntilDoneP() )

end

--------------------------------------------------------------------------------

-- @mark managing document slot values

--------------------------------------------------------------------------------
--[[!
	\brief Copy a value from one slot to another.

	\param docId (string) The document's unique ID. This must be the uuid; the
		integer local id is not supported.
	\param fromSlot (string) The name of the slot to copy from
	\param toSlot (string) The name of the new slot
	
	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:copySlotToSlotP( docId, fromSlot, toSlot )

	check( 'WFDocStore', '+string', '+string', '+string' )

	self._log:infof( "\n\ndocStore:copySlotToSlotP( %s, %s, %s )\n\nvia %s",
					 docId, fromSlot, toSlot, ':traceback:' )

	recordTransaction( self, 'copySlotToSlotP' )

	return maybeLogResponseP( self, 'copySlotToSlotP',
		self._docStoreImp:requestP( 'copySlotToSlot', docId, fromSlot, toSlot )
		:runUntilDoneP() )

end

--------------------------------------------------------------------------------
--[[!
	\brief Quietly delete slot.

	\param docId (string) The document's unique ID. This must be the uuid; the
		integer local id is not supported.
	\param slotName (string) The name of the slot to delete

	\warning This function is intended mostly for use by the sync engine. It
		\b quietly deletes the slot value. By that, we mean that the value
		simply disappears from the database without appearing in the change
		notifications from :changedDocs. (Any doc value behavior pointing to
		this slot will be updated.)
	
	It is not permitted to delete from the slot named "main".
	
	\return <code>P< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this promise.
]]--

function docStoreMethods:quietlyDeleteSlotP( docId, slotName )

	check( 'WFDocStore', '+string', '+string' )

	self._log:infof( "\n\ndocStore:quietlyDeleteSlotP( %s, %s )", docId, slotName )

	recordTransaction( self, 'quietlyDeleteSlotP' )

	return maybeLogResponseP( self, 'quietlyDeleteSlotP',
		self._docStoreImp:requestP( 'quietlyDeleteSlot', docId, slotName )
		:runUntilDoneP() )

end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------
--[[!
	\brief Tracks changes in this document by incrementing a counter for each
		change. Be aware that this value can jump, especially if many documents
		are changed at once.
	
	\return <code>B< number ></code> the current maximum value for
		\c _changeCounter across all documents in this database
]]--

function docStoreMethods:docChangeCounterB()

	check( 'WFDocStore' )

	self._log:info( "\n\ndocStore:docChangeCounterB()" )

	if not self._docChangeCounterB then
		self:setOwned {
			_docChangeCounterB = self._docStoreImp:requestE( 'docChangeCounterB' )
				:closingHoldB( WFUnknown.value() )
		}
	end
	
	return self._docChangeCounterB:relayB()
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a promise) the current change count for this document.
		This count includes every operation that has completed to this point.
 
	\return <code>P< number ></code> the current maximum value for
		\c _changeCounter across all documents in this database
]]--

function docStoreMethods:docChangeCounterP()

	check( 'WFDocStore' )

	self._log:info( "\n\ndocStore:docChangeCounterP()" )

	return maybeLogResponseP( self, 'docChangeCounterP',
		self._docStoreImp:requestP( 'docChangeCounter' ) )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the set of documents that have changed since a specified point.

	\param args (table) A table with these keys:
		* \c fromChangeCounter (number): A previous change-counter value. Documents
			whose change counter is equal to or greater than this value can match
			this query.
		* \c limit (optional, number): The maximum number of documents to return
		* \c idsOnly (optional, boolean): If true, only document IDs are returned
		* \c slowResponse (optional, boolean): If true, then don't rush to respond
		* \c slot (optional, string): If present, only documents whose value has
			changed in the specified slot will match this query. If not present,
			all document value changes will match this query.
		
	Note that regardless of any value specified in \c limit, a limit of 50 documents
	is always enforced. (The limit is raised to 1000 if \c idsOnly is true.)

	\return <code>P< table ></code> An array of fully-formed document values or IDs
		(as from \c :docValueP) that satisfy the changed-documents search criteria.
	
	This array/table also contains a top-level element named \c continueFrom
	whose value is the change-counter number to use	in the next \c :changedDocs call.
	That number may be less than the current value of \c :docChangeCounterB if
	the number of documents returned in this call was limited.
]]--

function docStoreMethods:changedDocsP( args )

	check( 'WFDocStore', 'table' )

	self._log:infof( "\n\ndocStore:changedDocsP( %s )", args )
	
	local enqueue = args.enqueue
	
	local selector
	
	if enqueue == 'slow' then
		selector = '-changedDocs'
	elseif enqueue == 'fast' then
		selector = '+changedDocs'
	elseif enqueue == 'normal' or enqueue == nil then
		selector = 'changedDocs'
	else
		selector = 'changedDocs'
		debug.print( "Invalid enqueue policy for changedDocsP: " .. tostring( enqueue ) )
	end
	
	return maybeLogResponseP( self, 'changedDocsP',
		self._docStoreImp:requestP( selector, args ) )

end

--------------------------------------------------------------------------------

-- @mark complex logic

--------------------------------------------------------------------------------
--[[!
	\brief Do complex logic in a single transaction.
	
	TODO [scouten 2014-01-16]: Document later. Until then, handle with care.
--]]

function docStoreMethods:doComplexLogicInTransactionP( namespaceAndFuncName, ... )

	check( 'WFDocStore', '+string' )

	self._log:infof( "\n\ndocStore:doComplexLogicInTransactionP( %q )",
					 namespaceAndFuncName )

	recordTransaction( self, namespaceAndFuncName )

	return maybeLogResponseP( self, namespaceAndFuncName,
		self._docStoreImp:requestP( 'doComplexLogicInTransaction',
									namespaceAndFuncName, ... ) )

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------
--[[!
	\brief Sends diagnostic data on synchronization to the log.
]]--

function docStoreMethods:sendDiagnosticLogData()

	-- Also include perf log as a separate section if we collected the data.
	
	if RECORD_API_PERFORMANCE then
	
		local rawReport = self:getApiPerformanceLog()
		
		local htmlReport = string.format( "<pre>%s</pre>",
										  KSStringUtils.escapeHtml( rawReport ) )
	
		local threadName = WFCurrentThread.DEBUG_getThreadName() or 'main'
	
		local idSafeThreadName = threadName
		idSafeThreadName = idSafeThreadName:replace( ' :: ', '+' )
	
		WFDiagnosticLog.addSectionToLog
			{
		
				id = string.format( 'WFDocStore-perf-%s-%s',
									self._path,
									idSafeThreadName ),
				
				name = string.format( "Doc Store API Performance: %s", threadName),
				weight = 401,
				
				body = htmlReport,
		
			}
	
	end
	
	return {

		name = 'WFDocStore :: ' .. self._path,
		id = 'WFDocStore-' .. self._path,
		weight = 400,
		
		body = {
			recentErrors = self._recentErrorLog,
			transactionSources = self._transactionSources,
		},

	}
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Gather statistics about doc store API performance.

	Only intended for debug / testing use.

	Only meaningful if RECORD_API_PERFORMANCE is true.
]]--

function docStoreMethods:getApiPerformanceLog()

	if not RECORD_API_PERFORMANCE then return "API performance log disabled" end
	
	local output = {
		"\nAPI PROFILE SUMMARY FOR DOC STORE ", self._path,
		"\n",
		"\n   Count       Min       Max       Total  API",
		"\n--------  --------  --------  ----------  -------------",
		"\n\n",
	}

	local sortedStats = {}
	local n = 0
	
	for k, v in pairs( self._perfLog ) do
		n = n + 1
		sortedStats[ n ] = v
	end
	
	table.sort( sortedStats, function( a, b )
	
		if a.total ~= b.total then
			return a.total > b.total
		else
			return a.funcName < b.funcName
		end
	
	end )

	for i = 1, #sortedStats do

		local stat = sortedStats[ i ]
		
		output[ #output + 1 ] =
			string.format(
				"%8.0f  %8.3f  %8.3f  %10.3f  %s\n",
					stat.count,
					stat.min,
					stat.max,
					stat.total,
					stat.funcName )
		
	end
	
	return table.concat( output )

end

--------------------------------------------------------------------------------

-- @mark miscellaneous

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the path to this document store.

	\return (string) The catalog's path.
]]--

function docStoreMethods:path()

	check( 'WFDocStore' )

	return self._path

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves this document store as a string.

	\return (string) The string representation.
]]--

function docStoreMethods:__tostring()

	check( 'WFDocStore' )

	return string.format( "WFDocStore( %s )", self._path )

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for unit test use only) Perform an integrity check
	of the doc store and report errors.

	\warning This function should only be used by unit tests.

	\return P< table > an array of failure messages (empty if none)
]]--

function docStoreMethods:performIntegrityCheckP()

	return self._docStoreImp:requestP( 'performIntegrityCheck' )

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for debugging use only) Generates code coverage statistics
		for the worker thread that implemented this doc store.

	\warning This function should only be used by WFDocStoreTest.lua.

	\return P< true > resolves when coverage data has been generated
]]--

function docStoreMethods:DEBUG_dumpCoverageP( path )

	return self._docStoreImp:requestP( 'DEBUG_dumpCoverage', path )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Permanently shuts down this doc store and deletes all related files.
]]--

function docStoreMethods:haltAndDeleteP()
	local _oldDocStoreImp = self._docStoreImp
	
	self._docStoreImp = nil
	local haltAndDeleteP =  _oldDocStoreImp:requestP( 'haltAndDelete' )
		:runUntilDoneP()
	
	self:close()
	return haltAndDeleteP
end

--------------------------------------------------------------------------------
--[[!
	\brief Return the number of SQLite transactions that have been executed
		thus far. Intended for debug / testing use only.
]]--

function docStoreMethods:nTransactionsP()

	return self._docStoreImp:requestP( 'nTransactions' )
		:mapDataP( function( _, nTransactions )

			return 'data', nTransactions, self._transactionSources

		end )

end

--------------------------------------------------------------------------------

return Rx.Class 'WFDocStore'( docStoreMethods )
