--[[----------------------------------------------------------------------------

WFDocStoreImp.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFDigest = require 'WFDigest'
local WFTableMerge = require 'WFTableMerge'

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'

	-- WF.sqlite.database
local WFSQLiteDatabase = require 'WFSQLiteDatabase'
local WFSQLiteDatabaseSchema = require 'WFSQLiteDatabaseSchema'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFDocStoreImp'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local assert = assert
local type = assert( type )
local pairs = assert( pairs )
local loadstring = assert( loadstring )
local setfenv = assert( setfenv )

local string_byte = assert( string.byte )
local string_format = assert( string.format )

local math_min = assert( math.min )

local tableToString = assert( WFTableArchiving.tableToString )

local check = assert( WFArguments.check )

local sha1 = assert( WFDigest.SHA1 )
local sha1_digest = assert( sha1.digest )


--============================================================================--

-- @mark database schema

local schema = WFSQLiteDatabaseSchema( '1.0' ):useWAL():setSynchronous( 1 ):setPageSize( 8192 )
	-- TODO [scouten 2014-01-31 via arahn]: Add a WFCapabilities flag to configure
	-- page size? 8K pages might be larger than we want for mobile.

--------------------------------------------------------------------------------
--[[
	Content in this document database is primarily stored via hash reference.
	In other words, the hash of the content is the primary key used to find
	the content.
	
	"Body" content here refers to primary key data that is likely to be synced
	with other sources. We define this as those top-level keys that do not start
	with underscore.
--]]
	
schema:addTable
	[[CREATE TABLE bodyContent( bodyContentHash UNIQUE NOT NULL,
								value NOT NULL )]]
	:onConflict 'replace'

--------------------------------------------------------------------------------
--[[
	Non-synced content is stored separately from body content on the theory that
	is is more volatile and that it's interesting to be able to test whether
	body content matches across slots. (See below for definition of "slot.")
--]]
	
schema:addTable
	[[CREATE TABLE annotationContent( annotationContentHash UNIQUE NOT NULL,
									  value NOT NULL )]]
	:onConflict 'replace'

--------------------------------------------------------------------------------
--[[
	Content hash is probably not the most user-friendly way to find content,
	so we introduce the concept of documents and value slots. The "slot" concept
	allows us to store multiple alternative values for the document (for instance,
	from external sync sources). We assert that the doc store will store identical
	content efficiently (i.e. not duplicate content) if multiple slots hold the
	same value, and also offer create copy-on-write semantics when one slot's value
	evolves.
	
	Slots are a little bit like named branches in git, but there is no assumption
	that we preserve history or support merging of values across slots, so the
	analogy is not pure.
--]]

schema:addTable
	[[CREATE TABLE documents(
						docId NOT NULL,
						slotName NOT NULL,
						changeCounter NOT NULL,
						bodyContentHash NOT NULL,
						annotationContentHash NOT NULL )]]
	:onConflict 'replace'

schema:addIndex
	[[CREATE UNIQUE INDEX documents_docId_slotName ON documents( docId, slotName )]]

schema:addIndex
	[[CREATE INDEX documents_bodyContentHash ON documents( bodyContentHash )]]

schema:addIndex
	[[CREATE INDEX documents_annotationContentHash ON documents( annotationContentHash )]]

schema:addIndex
	[[CREATE INDEX documents_changeCounter ON documents( changeCounter DESC, slotName )]]

--------------------------------------------------------------------------------
--[[
	This table is used in the implementation of the specific docs
	version of changedDocs. It lets us bind to a variable number of items.
]]--

schema:addTable
	[[CREATE TEMPORARY TABLE TEMP_docIds( docId )]]


--============================================================================--

-- @mark WFDocStoreImp class

--------------------------------------------------------------------------------
--[[!
	\class WFDocStoreImp

	This class is considered \b INTERNAL to the implementation of \c WFDocStore.
	Please do not access this class directly.
]]--

local docStoreImpMethods = {

	_db = false,
		-- (WFSQLiteDatabase) database file

	_exec = false,
		-- (function) cached reference to db:exec
	
	_gatherArrayOfArray = false,
		-- (function) cached reference to db:gatherArrayOfArray

	_path = false,
		-- (string) path to doc store DB

	_docChangeCounterB = false,
		-- (B< number >) most recent doc change counter value
		-- WARNING: This value is one lower than db:get/setChangeCounter().
		-- The semantics of the underlying WFSQLiteDatabase code are wrong, but it's
		-- not possible to change that in a way that doesn't break OzDocStoreImp.

	_docValueBehaviors = false,
		-- (KSReactive.makeWeakEventTable) document value behaviors that are currently active
			-- key: (string) doc ID + slot name

	_docValueBehaviors_eventExists = false,
		-- (function) cached reference to _docValueBehaviors:eventExists

	WFCloseSupport.addCloseSupportMethods,
	
	KSReactive.addReactiveSupportMethods,

}

--------------------------------------------------------------------------------

function docStoreImpMethods:__init( --[[string]] path )

	check( 'WFDocStoreImp', 'string' )

	-- Map :memory:(random) and :temp:(random) paths back to SQLite pseudo-paths.

	local dbPath = path

	if dbPath:match( '^:memory:' ) then

		dbPath = ':memory:'

	elseif dbPath:match( '^:temp:' ) then

		dbPath = ':temp:'

	elseif #dbPath > 0 then
		-- An empty path is legal and means create a temporary path on disk.
	
		local parentDir = KSPathUtils.parent( dbPath )
		WFFileUtils.mkdir( parentDir )
	
	end

	-- Create the actual database. Note that we disable query performance logging
	-- because the query manifest is very simple for this database and we've done
	-- what can be done IMHO.
	
	self._path = dbPath
	logger:infof( 'dbPath = %s', dbPath )

	local db = WFSQLiteDatabase( dbPath, schema, true )
	self:setOwned { _db = db }
	
	WFCloseSupport.mustCloseAtThreadTermination( self )

	db:setAutosaveInterval( 5 )
		-- TODO [scouten 2014-02-04]: Make this configurable?
	
	self._exec = assert( db.exec )
	self._gatherArrayOfArray = assert( db.gatherArrayOfArray )

	-- Set up change notification.

	local changeCounterSeedRow = db:execOnce(
		"SELECT max( changeCounter ) AS changeCounter FROM documents" )

	local currentChangeCounter = ( changeCounterSeedRow
		and changeCounterSeedRow.changeCounter or 0 )
	
	db:setChangeCounter( currentChangeCounter + 1 )

	self:createReactiveVariable( '_docChangeCounterB', currentChangeCounter )
	
	self:setOwned { _docValueBehaviors = KSReactive.makeWeakEventTable( nil ) }
	
	self._docValueBehaviors_eventExists = assert( self._docValueBehaviors.eventExists )

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

-- @mark shared utilities

--------------------------------------------------------------------------------

local function updateChangeCounter( self )

	local newChangeCounter = self._db:getChangeCounter()

	logger:infof( "newChangeCounter = %s", newChangeCounter )
	self._docChangeCounterB:set( newChangeCounter - 1 )
	
	return newChangeCounter
	
end

--------------------------------------------------------------------------------

local function insertChangeCounterPlaceholder_inTransaction( db, exec )

	local currentChangeCounter = db:getChangeCounter() - 1

	local newCCRow = exec( db, "SELECT max( changeCounter ) AS newChangeCounter FROM documents" )

	local newChangeCounter = newCCRow and newCCRow.newChangeCounter

	if not newChangeCounter or newCCRow.newChangeCounter < currentChangeCounter then
	
		-- INSERT OR REPLACE is known to change the rowid.  rowid is used as our id_local.  However,
		-- this (placeholder) document is not a real document, so it is okay that it gets a new
		-- id_local.

		exec( db,
			[[INSERT OR REPLACE INTO documents
				VALUES( '(placeholder)', '(ignore me)',
						?, '', '' )]], currentChangeCounter )
	
	end

end

--------------------------------------------------------------------------------

local function makeIntegrityCheckFunction( expectedEmptyCountQuery, messageFormat )

	expectedEmptyCountQuery = "select count(*) from " .. expectedEmptyCountQuery

	return function( db, exec )
	
			local _, count = next( exec( db, expectedEmptyCountQuery ) )

			if count == 0 then
				return true
			else
				return false, string.format( messageFormat, count )
			end

	end
	
end

local integrityCheckers = {
	makeIntegrityCheckFunction( [[bodyContent where bodyContentHash not
		in (select bodyContentHash from documents)]], "orphaned bodyContent: %d" ),

	makeIntegrityCheckFunction( [[annotationContent where annotationContentHash not
		in (select annotationContentHash from documents)]], "orphaned annotationContent: %d" ),
	
	makeIntegrityCheckFunction( [[documents where bodyContentHash not
		in (select bodyContentHash from bodyContent)]], "dangling bodyContent: %d" ),
	
	makeIntegrityCheckFunction( [[documents where annotationContentHash not
		in (select annotationContentHash from annotationContent)]], "dangling annotationContent: %d" ),
}

--------------------------------------------------------------------------------

local function makeMaybeGCOldContentValueFunc( selectContentHashSql,
											   deleteContentHashSql )

	return function( db, exec, oldContentHash, selector )

		-- At the point when this function is called, we have a content hash
		-- that may no longer be referenced. We want to get rid of the corresponding
		-- content unless some other reference still exists.
	
		-- No-op if old hash was nil.
	
		if not oldContentHash then
			return
		end
	
		-- Is there an existing reference?
		
		if exec( db, selectContentHashSql, oldContentHash ) then
			return
		end
		
		-- No existing reference: Delete.
		
		exec( db, deleteContentHashSql, oldContentHash )
		
	end

end

local maybeGCOldBodyContentValue = makeMaybeGCOldContentValueFunc(
	[[SELECT 1 FROM documents WHERE bodyContentHash == ? LIMIT 1]],
	[[DELETE FROM bodyContent WHERE bodyContentHash == ?]] )

local maybeGCOldAnnotationContentValue = makeMaybeGCOldContentValueFunc(
	[[SELECT 1 FROM documents WHERE annotationContentHash == ? LIMIT 1]],
	[[DELETE FROM annotationContent WHERE annotationContentHash == ?]] )

--------------------------------------------------------------------------------

local underscoreByte = string_byte( '_' )

local acceptableTypes = {
	[ 'boolean' ] = true,
	[ 'number' ] = true,
	[ 'string' ] = true,
	[ 'table' ] = true,
}

--------------------------------------------------------------------------------

local approvedKeys = {
	_id = true, -- since we get these from Oz all the time
	_deleted = true,
	_localOnly = true,
	_merge = true,
	_serverProvidedMetadata = true,
	_syncStatus = true,
}

--------------------------------------------------------------------------------

local kInfinity = 1/0
local kNegativeInfinity = -kInfinity

--------------------------------------------------------------------------------

local function assertDocValueIsValid( value, isSubtable )

	if getmetatable( value ) then
		error "table contains metatable"
	end
	
	local isDeleted = not isSubtable and value._deleted
	
	for k, v in pairs( value ) do
	
		local tk = type( k )
		
		if tk == 'string' then
		
			if not isSubtable and string_byte( k, 1 ) == underscoreByte and not approvedKeys[ k ]  then
				error( string_format( "value contains key with leading underscore (%s)", k ) )
			end
			
			if isDeleted and string_byte( k, 1 ) ~= underscoreByte then
				error( "body must be empty if _deleted is true" )
			end

		elseif tk == 'number' then

			if not isSubtable then
				error( "numeric keys not allowed at top level of doc value" )
			end

			if k == kInfinity or k == kNegativeInfinity then
				-- Note that Lua disallows NaN as key, so no need to test for that case here.
			
				error( string_format( "key %s is an invalid number",
									  tostring( k ) ) )
	
			end

		else
		
			error( string_format( "key %s is of type %s, must be string%s",
								  tostring( k ), tk,
								  isSubtable and " or number" or "" ) )

		end
		
		local tv = type( v )
		
		if v == kInfinity or v == kNegativeInfinity or v ~= v then
			error( string_format( "value %s is an invalid number",
								  tostring( v ) ) )
		end
		
		if not acceptableTypes[ tv ] then
			error( string_format( "value at key %s is of type %s (must be scalar or table)",
								  k, tv ) )
		end
		
		if tv == 'table' then
			assertDocValueIsValid( v, true )
		end
	
	end

end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------

local valueFnEnv = {} -- disallow all function calls

--------------------------------------------------------------------------------

local function contentFromFlattenedValue( value )

	-- Lightweight sandboxing. The one potentially malicious piece of code
	-- somebody could inject (and they'd have to do it directly into the
	-- SQLite database) would be to put an infinite loop here.
	-- Some more potentially valuable approaches we could take are
	-- described in http://lua-users.org/wiki/SandBoxes.

	local valueFn = loadstring( 'return ' .. value )
	assert( type( valueFn ) == 'function' )
	setfenv( valueFn, valueFnEnv ) -- disallow all function calls
	
	return valueFn()

end

--------------------------------------------------------------------------------

local function docValueFromContentRow( contentRow )

	-- WARNING: Care must be taken to ensure that the values in all queries
	-- that create contentRow tables that the columns line up the same.
	-- See :docValue, just below this, for the canonical column order.

	-- Convert flattened Lua to actual Lua values.

	local bodyContentValue = contentFromFlattenedValue( contentRow[ 4 ] )
	local docValue = contentFromFlattenedValue( contentRow[ 5 ] )

	-- We seed the docValue from the annotation in the (unexpected) case that
	-- both annotation and body contain the same key. If so, then we want the
	-- version in the bodyContent hash to dominate when we copy over.
	
	for k, v in pairs( bodyContentValue ) do
		docValue[ k ] = v
	end
	
	docValue._id = contentRow[ 1 ]
	docValue._changeCounter = contentRow[ 3 ]
	docValue._hash = contentRow[ 6 ]
	docValue._slot = contentRow[ 2 ]
	docValue._idLocal = contentRow[ 7 ]
	
	return docValue
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:docValue( docId, optSlotName )

	check( 'WFDocStoreImp', '+string|number', '?string' )

	local slotName = optSlotName or 'main'

	-- Find value.
	local idColumn = type( docId ) == 'number' and 'documents.rowid' or 'docId'
	
	local contentRows = self._gatherArrayOfArray( self._db,
		[[SELECT docId,
				 slotName,
				 changeCounter,
				 bodyContent.value,
				 annotationContent.value,
				 documents.bodyContentHash,
				 documents.rowid
			FROM documents
				INNER JOIN bodyContent USING( bodyContentHash )
				INNER JOIN annotationContent USING( annotationContentHash )
			WHERE ]] .. idColumn .. [[ == ? AND slotName == ?]], docId, slotName )
	
	-- If no such value, hack up an empty document.
	
	local contentRow = contentRows and contentRows[ 1 ]
	
	if not contentRow then
		return {
			_id = docId,
			_slot = slotName,
			_empty = true,
		}
	end

	return docValueFromContentRow( contentRow )
	
end

--------------------------------------------------------------------------------

local function docValueBehaviorKey( docId, slotName )

	return string_format( '%s :: %s', tostring( docId ), slotName )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:docValueB( docId, optSlotName )

	check( 'WFDocStoreImp', '+string|number', '?string' )

	local slotName = optSlotName or 'main'

	local docValueBehaviors = self._docValueBehaviors
	local key = docValueBehaviorKey( docId, slotName )

	local rawResultB, behaviorDidExist = docValueBehaviors:behaviorB( key )
	
	-- If this is a brand-new behavior, then post a value to it immediately.
	
	if not behaviorDidExist then
		docValueBehaviors:postToEvent( key, self:docValue( docId, slotName ) )
	end
	
	return rawResultB:autoCloseRelayB()

end

--------------------------------------------------------------------------------

local function valueToArchiveAndHash( value )

	local archive = tableToString( value )
	local contentHash = sha1_digest( archive )
	
	return archive, contentHash

end

--------------------------------------------------------------------------------

local function extractAndValidateDocValue( self, funcName, combinedValue )

	-- Extract doc ID from the value.
	
	local docId = combinedValue._id
	local idColumn = 'docId'
	combinedValue._id = nil
	
	if type( docId ) ~= 'string' or #docId < 1 then
	
		if type( docId ) == 'number' and docId > 0 then
			idColumn = 'documents.rowid'
		else
			error( string_format( "WFDocStoreImp:%s: _id must be non-empty string",
								  funcName ), 3 )
		end

	end

	-- Extract _slot from the value.
	
	local slotName = combinedValue._slot
	combinedValue._slot = nil
	
	if type( slotName ) ~= 'string' or #slotName < 1 then
		error( string_format( "WFDocStoreImp:%s: _slot must be non-empty string",
							  funcName ), 2 )
	end
	
	-- We allow _hash to be populated. If present, it must match the existing
	-- doc value. If not present, we'll replace any existing doc value.
	
	local mustMatchHash = combinedValue._hash
	combinedValue._hash = nil
	combinedValue._idLocal = nil
	
	-- We allow _changeCounter to be populated. If present, it must match the existing
	-- doc value. If not present, we'll replace any existing docValue.
	
	local mustMatchChangeCounter = combinedValue._changeCounter
	combinedValue._changeCounter = nil
	
	-- Verify that doc value is acceptable (does not contain illegal keys, etc.).

	assertDocValueIsValid( combinedValue )

	-- Split doc value into annotations (leading _ at top-level keys) and
	-- body content (everything else).
	
	local annotationContent = {}
	local bodyContent = {}
	
	for k, v in pairs( combinedValue ) do
		if string_byte( k, 1 ) == underscoreByte and k ~= '_deleted' then
			annotationContent[ k ] = v
		else
			bodyContent[ k ] = v
		end
	end
	
	-- Convert Lua to text and hash it.
	
	local bodyContentArchive, bodyContentHash = valueToArchiveAndHash( bodyContent )

	local annotationContentArchive, annotationContentHash
		= valueToArchiveAndHash( annotationContent )

	return
		{
			docId = docId,
			idColumn = idColumn,
			slotName = slotName,
			mustMatchHash = mustMatchHash,
			mustMatchChangeCounter = mustMatchChangeCounter,
			bodyContentHash = bodyContentHash,
			annotationContentHash = annotationContentHash,
			bodyContentArchive = bodyContentArchive,
			annotationContentArchive = annotationContentArchive,
		}

end

--------------------------------------------------------------------------------

local function setDocValueImp_inTransaction( self, db, exec, funcName, args )

	local idOrIdLocal = args.docId
	local idColumn = args.idColumn
	local slotName = args.slotName
	local mustMatchHash = args.mustMatchHash
	local mustMatchChangeCounter = args.mustMatchChangeCounter
	local bodyContentHash = args.bodyContentHash
	local annotationContentHash = args.annotationContentHash
	local bodyContentArchive = args.bodyContentArchive
	local annotationContentArchive = args.annotationContentArchive
	
	-- Look up existing body and annotationcontent hash values.
	-- Either or both of them might no longer be needed.

	local oldBodyContentHash, oldAnnotationContentHash
	local id_local, docId

	local existingRow = exec( db,
		[[SELECT rowid, docId, bodyContentHash, annotationContentHash, changeCounter
			FROM documents
			WHERE ]] .. idColumn .. [[ == ? AND slotName == ?]],
			idOrIdLocal, slotName )
	
	if existingRow ~= nil then
		id_local = existingRow.rowid
		docId = existingRow.docId
		oldBodyContentHash = existingRow.bodyContentHash
		oldAnnotationContentHash = existingRow.annotationContentHash
	else
		assert( type( idOrIdLocal ) == 'string', 'New documents cannot be created with local _id' )
		docId = idOrIdLocal
	end
	
	if mustMatchHash and oldBodyContentHash ~= mustMatchHash then
		error( string_format( "WFDocStoreImp:%s: _hash %s does not match existing document hash %s",
							  funcName, mustMatchHash, oldBodyContentHash ), 3 )
	end
	
	local oldChangeCounter = existingRow and existingRow.changeCounter
	
	if mustMatchChangeCounter
		and oldChangeCounter ~= mustMatchChangeCounter
		and not( mustMatchChangeCounter == ':new:' and existingRow == nil )
	then
		error( string_format( "WFDocStoreImp:%s: _changeCounter %s does not match existing document change counter %s",
							  funcName, mustMatchChangeCounter, oldChangeCounter ), 3 )
	end

	-- Short-circuit if new content == existing.
	
	if bodyContentHash == oldBodyContentHash
		and annotationContentHash == oldAnnotationContentHash
	then
		return bodyContentHash
	end
	
	-- Add new content blobs into DB if not already there.

	exec( db,
		[[INSERT OR IGNORE INTO bodyContent VALUES( ?, ? )]],
			bodyContentHash, bodyContentArchive )
	
	exec( db,
		[[INSERT OR IGNORE INTO annotationContent VALUES( ?, ? )]],
			annotationContentHash, annotationContentArchive )
	
	-- Update the reference in documents table.
	
	if id_local then
		exec( db,
			[[UPDATE documents
				SET changeCounter=incrementChangeCounter(), bodyContentHash=?, annotationContentHash=?
				WHERE docId=? AND slotName=?]],
				bodyContentHash, annotationContentHash, docId, slotName )
	else
		exec( db,
			[[INSERT INTO documents
				VALUES( ?, ?, incrementChangeCounter(), ?, ? )]],
				docId, slotName, bodyContentHash, annotationContentHash )
	end

	-- Bump local copy of change counter.
	
	updateChangeCounter( self )

	-- Do we need to update a doc value behavior?
	
	local docValueBehaviors = self._docValueBehaviors
	local key = docValueBehaviorKey( docId, slotName )
	
	if self._docValueBehaviors_eventExists( docValueBehaviors, key ) then
		docValueBehaviors:postToEvent( key, self:docValue( docId, slotName ) )
	end

	-- Delete old content if no longer referenced.
	
	maybeGCOldBodyContentValue( db, exec, oldBodyContentHash )
	maybeGCOldAnnotationContentValue( db, exec, oldAnnotationContentHash )

	return bodyContentHash
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:setDocValue( combinedValue )

	check( 'WFDocStoreImp', 'table' )
	
	logger:debugf( "setDocValue %s", combinedValue )

	local args = extractAndValidateDocValue( self, 'setDocValue', combinedValue )
	
	local db = self._db

	return db:withTransactionDo( function()

		return setDocValueImp_inTransaction( self, db, self._exec,
											 'setDocValue', args )

	end )
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:bulkSetDocValue( combinedValues )

	check( 'WFDocStoreImp', 'table' )
	
	logger:debugf( "bulkSetDocValue %s", combinedValues )

	local nCombinedValues = #combinedValues

	if nCombinedValues > 100 then
		logger:warnf( "WFDocStoreImp:bulkSetDocValue: %d values provided in this array, exceeds recommended maximum (100)",
					  nCombinedValues )
	end

	local argsPerDocument = {}
	
	for i = 1, nCombinedValues do
		argsPerDocument[ i ] = extractAndValidateDocValue( self, 'bulkSetDocValue',
														   combinedValues[ i ] )
	end
	
	local db = self._db

	db:withTransactionDo( function()

		for i = 1, nCombinedValues do
			setDocValueImp_inTransaction( self, db, self._exec,
										  'bulkSetDocValue', argsPerDocument[ i ] )
		end

	end )

end

--------------------------------------------------------------------------------

-- @mark partial document values

--------------------------------------------------------------------------------

local function applyOverlayPrepareValue( self, overlay )

	check( 'WFDocStoreImp', 'table' )

	-- Extract doc ID from the overlay.
	
	local docId = overlay._id
	overlay._id = nil
	local t = type( docId )

	if ( t ~= 'string' or #docId < 1 ) and ( t ~= 'number' or docId <= 0 ) then
		error( "WFDocStoreImp:applyOverlay: _id must be non-empty string", 2 )
	end

	-- Extract _slot from the overlay.
	
	local slotName = overlay._slot
	overlay._slot = nil
	
	if type( slotName ) ~= 'string' or #slotName < 1 then
		error( "WFDocStoreImp:applyOverlay: _slot must be non-empty string", 2 )
	end
	
	-- Get existing value, if any, apply overlay, write it back.
	
	local value = self:docValue( docId, slotName )

	if value and value.type and overlay and overlay.type then

		assert( value.type == overlay.type,
				"Once a document is created, its type cannot be changed" )

		-- WARNING [scouten 2014-01-03]: This restriction is not enforced
		-- by :setDocValueP or :bulkSetDocValueP. Doing so would mean we'd have to
		-- read and unpack the old doc body value, which we weren't doing otherwise.
		-- I'm concerned about the performance impact of adding that test.

	end

	value._idLocal = nil

	local empty = value._empty
	
	value._empty = nil
	
	WFTableMerge.applyOverlay( value, overlay )
	
	if overlay._deleted then
	
		for k in pairs( value ) do
		
			if not k:match( "^_" ) then
				value[ k ] = nil
			end
		
		end
		
	elseif empty then
		
		for k in pairs( value ) do
			if k ~= '_slot' and k ~= '_id' then
				empty = false
				break
			end
		end
		if empty then
			value = nil
		end
		
	end

	return value

end

--------------------------------------------------------------------------------

function docStoreImpMethods:applyOverlay( overlay )
	
	local value = applyOverlayPrepareValue( self, overlay )
	if value ~= nil then
		return self:setDocValue( value )
	end

end

--------------------------------------------------------------------------------

-- @mark slot management

--------------------------------------------------------------------------------

local function copySlotToSlotPrep( self, docId, fromSlot, toSlot )

	check( 'WFDocStoreImp', '+string', '+string', '+string' )
	
	-- Look up existing content hash values. Error if no such document/slot.

	local sourceRow = self._exec( self._db,
		[[SELECT bodyContentHash, annotationContentHash, changeCounter
			FROM documents
			WHERE docId == ? AND slotName == ?]],
			docId, fromSlot )
	
	if not sourceRow then
		error( string_format( "No such document/slot: %s/%s", docId, fromSlot ), 2 )
	end
	
	-- Short-circuit if dest row already has this value.

	local existingDestRow = self._exec( self._db,
		[[SELECT bodyContentHash, annotationContentHash
			FROM documents
			WHERE docId == ? AND slotName == ?]],
			docId, toSlot )

	if existingDestRow ~= nil then

		if existingDestRow.bodyContentHash == sourceRow.bodyContentHash
			and existingDestRow.annotationContentHash == sourceRow.annotationContentHash
		then
			return false, sourceRow.bodyContentHash
		end

	end
	
	return true, sourceRow, existingDestRow

end

--------------------------------------------------------------------------------

local function copySlotToSlot_inTransaction( self, docId, toSlot, sourceRow, optDestRow )

	local db = self._db
	local exec = self._exec

	local existingRow = exec( db,
		[[SELECT rowid FROM documents
			WHERE docId == ? AND slotName == ?]],
			docId, toSlot )

	-- Update the reference in documents table.
	
	if existingRow then
		exec( db,
			[[UPDATE documents
				SET changeCounter=incrementChangeCounter(), bodyContentHash=?, annotationContentHash=?
				WHERE docId=? AND slotName=?]],
				sourceRow.bodyContentHash, sourceRow.annotationContentHash, docId, toSlot )
	else
		exec( db,
			[[INSERT INTO documents
				VALUES( ?, ?, incrementChangeCounter(), ?, ? )]],
				docId, toSlot,
				sourceRow.bodyContentHash, sourceRow.annotationContentHash )
	end

	-- Do we need to update a doc value behavior?
	
	local docValueBehaviors = self._docValueBehaviors
	local key = docValueBehaviorKey( docId, toSlot )
	
	if docValueBehaviors:eventExists( key ) then
		docValueBehaviors:postToEvent( key, self:docValue( docId, toSlot ) )
	end

	if optDestRow then
		maybeGCOldBodyContentValue( self._db, self._exec, optDestRow.bodyContentHash )
		maybeGCOldAnnotationContentValue( self._db, self._exec, optDestRow.annotationContentHash )
	end

	-- Bump local copy of change counter.
	
	updateChangeCounter( self )
	
	return sourceRow.bodyContentHash

end

--------------------------------------------------------------------------------

function docStoreImpMethods:copySlotToSlot( docId, fromSlot, toSlot )

	local proceed, sourceRow, destRow = copySlotToSlotPrep( self, docId, fromSlot, toSlot )
	
	if proceed then
		return self._db:withTransactionDo( copySlotToSlot_inTransaction,
										   self, docId, toSlot, sourceRow, destRow )
	else
		return sourceRow -- is actually the hash value (didn't need to copy)
	end
	
end

--------------------------------------------------------------------------------

local function quietlyDeleteSlotPrep( self, docId, slotName )

	check( 'WFDocStoreImp', '+string', '+string' )
	
	if slotName == 'main' then
		error( "WFDocStoreImp:quietlyDeleteSlot: can not delete slot 'main'", 2 )
	end
	
	-- Look up existing body/annotation content hash values. There may or may not
	-- be other documents rows that share these contentHash values. If there are
	-- no longer any others, then we should also delete the content.

	local existingRow = self._exec( self._db,
		[[SELECT bodyContentHash, annotationContentHash
			FROM documents WHERE docId == ? AND slotName == ?]],
			docId, slotName )
	
	if existingRow == nil then
		return
	end
	
	local oldBodyContentHash = existingRow.bodyContentHash
	local oldAnnotationContentHash = existingRow.annotationContentHash

	return oldBodyContentHash, oldAnnotationContentHash

end

--------------------------------------------------------------------------------

local function quietlyDeleteSlot_inTransaction( self, docId, slotName,
												oldBodyContentHash,
												oldAnnotationContentHash )

	local db = self._db
	local exec = self._exec

	-- Remove this reference from the documents table.

	exec( db,
		  [[DELETE FROM documents WHERE docId == ? AND slotName == ?]],
		  docId, slotName )
	
	-- Do we need to update a doc value behavior?
	
	local docValueBehaviors = self._docValueBehaviors
	local key = docValueBehaviorKey( docId, slotName )
	
	if docValueBehaviors:eventExists( key ) then

		docValueBehaviors:postToEvent( key,
			{
				_id = docId,
				_slot = slotName,
				_empty = true,
			} )

	end

	-- Delete old content if no longer referenced.
	
	maybeGCOldBodyContentValue( db, exec, oldBodyContentHash )
	maybeGCOldAnnotationContentValue( db, exec, oldAnnotationContentHash )

	-- If needed, insert a placeholder document to ensure change counter
	-- doesn't fall backwards when DB is reopened.
	
	insertChangeCounterPlaceholder_inTransaction( db, self._exec )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:quietlyDeleteSlot( docId, slotName )

	local oldBodyContentHash, oldAnnotationContentHash
		= quietlyDeleteSlotPrep( self, docId, slotName )

	if oldBodyContentHash and oldAnnotationContentHash then
		
		return self._db:withTransactionDo( quietlyDeleteSlot_inTransaction,
										   self, docId, slotName,
										   oldBodyContentHash,
										   oldAnnotationContentHash )

	end
	
end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------

function docStoreImpMethods:docChangeCounterB()

	check( 'WFDocStoreImp' )

	return self._docChangeCounterB:relayB()
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:docChangeCounter()

	check( 'WFDocStoreImp' )

	return self._docChangeCounterB:valueNow()

end

--------------------------------------------------------------------------------

local changedDocs_maxLimit = 50
local changedDocs_maxLimitForIds = 1000
	-- Both of these are arbitrary; tune based on observed behavior (per platform?).

--------------------------------------------------------------------------------

local cachedChangedDocsSql = {}

local function changedDocsSqlForArgs( args )

	local sqlCacheIndexStr = 'default'
	
	if args.idsOnly or args.docs or not args.slot then
	
		sqlCacheIndexStr = string_format( '%s:%s:%s',
										  args.idsOnly and 'idsOnly' or 'x',
										  args.docs and 'docs' or 'x',
										  args.slot and 'x' or 'noSlot' )
		
	end
	
	local cachedSql = cachedChangedDocsSql[ sqlCacheIndexStr ]
	if cachedSql then
		return cachedSql
	end
	
	-- WARNING: Make sure the columns line up at the same indices as they do
	-- in :setDocValue if we are pulling full document values.
	
	local columns, innerJoinDocs, innerJoinContent, whereSlotName
	
	if args.idsOnly then
	
		columns = [[docId, slotName, changeCounter]]
	
	else

		columns = [[docId,
					 slotName,
					 changeCounter,
					 bodyContent.value AS bodyContent,
					 annotationContent.value as annotationContent,
					 documents.bodyContentHash,
					 documents.rowid]]

		innerJoinContent = [[INNER JOIN bodyContent USING( bodyContentHash )
							 INNER JOIN annotationContent USING( annotationContentHash )]]

	end

	if args.slot then
		whereSlotName = [[AND slotName == ?]]
	end

	if args.docs then
		innerJoinDocs = [[INNER JOIN TEMP_docIds USING( docId )]]
	end
	
	local sql = string_format(
		[[SELECT %s
			FROM documents
				%s
				%s
			WHERE changeCounter >= ?
				%s
			ORDER BY changeCounter
			LIMIT ? ]],
		columns,
		innerJoinDocs or '',
		innerJoinContent or '',
		whereSlotName or '' )

	cachedChangedDocsSql[ sqlCacheIndexStr ] = sql
	
	return sql

end

--------------------------------------------------------------------------------

local function changedDocsPrep( self, args )

	check( 'WFDocStoreImp', 'table' )
	
	local from = args.fromChangeCounter

	if not from then
		error( "WFDocStoreImp:changedDocs: fromChangeCounter argument missing", 2 )
	end
	
	if type( from ) ~= 'number'
		or from == kInfinity
		or from == kNegativeInfinity
		or from ~= from
		or from < 0
	then
		error( string_format( "WFDocStoreImp:changedDocs: fromChangeCounter illegal (%s)",
							  from ), 2 )
	end

	local maxLimit = args.idsOnly
						and changedDocs_maxLimitForIds
						or changedDocs_maxLimit
	
	local limit = args.limit or maxLimit
	if type( limit ) ~= 'number' then
		error( "WFDocStoreImp:changedDocs: limit must be a number or nil for default", 2 )
	end
	
	if limit < 1
		or limit == kInfinity
		or limit ~= limit
	then
		error( "WFDocStoreImp:changedDocs: limit must be a positive number", 2 )
	end
	
	if limit > maxLimit then

		logger:warnf( "changedDocs: limit is too high (%d), reducing to %d",
					  limit, maxLimit )
		
		limit = maxLimit

	end

	local docs = args.docs
	if docs and ( type( docs ) ~= 'table' or #docs == 0 ) then
		error( "WFDocStoreImp:changedDocs: docs must be a non-empty array of doc IDs or nil for all docs", 2 )
	end

	local idsOnly = args.idsOnly
	
	local slot = args.slot
	if slot and ( type( slot ) ~= 'string' or #slot == 0 ) then
		error( "WFDocStoreImp:changedDocs: slot must be a non-empty string or nil for all slots", 2 )
	end

	local sql = changedDocsSqlForArgs( args )

	return {
		docs = docs,
		slot = slot,
		from = from,
		limit = limit,
		idsOnly = idsOnly,
		sql = sql
	}

end

--------------------------------------------------------------------------------

local function changedDocsCore_inTransaction( self, parsedArgs )

	local db = self._db
	local docs = parsedArgs.docs
	
	if docs then

		if #docs > 100 then
			logger:warnf( "WARNING: doc inclusion list is awfully large %s", #docs )
		end

		db:exec [[DELETE FROM TEMP_docIds]]

		for i = 1, #docs do
			db:exec( [[INSERT INTO TEMP_docIds VALUES( ? )]], docs[ i ] )
		end
	
	end
	
	local slot = parsedArgs.slot
	local limit = parsedArgs.limit
	local from = parsedArgs.from
	local sql = parsedArgs.sql
	
	if slot then

		return db:gatherArrayOfArray( sql, from, slot, limit + 1 )

	else

		return db:gatherArrayOfArray( sql, from, limit + 1 )
	
	end

end

--------------------------------------------------------------------------------

local function postProcessChangedDocsRows( self, rawDocRows, parsedArgs )

	local result = {}
	
	local idsOnly = parsedArgs.idsOnly
	local limit = parsedArgs.limit
	
	if idsOnly then

		for i = 1, math_min( #rawDocRows, limit ) do
			result[ i ] = rawDocRows[ i ][ 1 ]
		end
	
	else
	
		for i = 1, math_min( #rawDocRows, limit ) do
			result[ i ] = docValueFromContentRow( rawDocRows[ i ] )
		end
	
	end

	if rawDocRows[ limit + 1 ] then
		result.continueFrom = rawDocRows[ limit + 1 ][ 3 ]
	else
		result.continueFrom = self._docChangeCounterB:valueNow() + 1
	end

	logger:debugf( "changedDocs is returning %s", result )

	return result

end

--------------------------------------------------------------------------------

function docStoreImpMethods:changedDocs( args )

	local parsedArgs = changedDocsPrep( self, args )

	local rawDocRows = self._db:withTransactionDo( changedDocsCore_inTransaction,
												   self, parsedArgs )
	
	return postProcessChangedDocsRows( self, rawDocRows, parsedArgs )

end

--------------------------------------------------------------------------------

-- @mark complex logic

--------------------------------------------------------------------------------

function docStoreImpMethods:doComplexLogicInTransaction( namespaceAndFuncName, ... )

	check( 'WFDocStoreImp', '+string' )
	
	logger:debugf( "doComplexLogicInTransaction" )

	local namespaceName, funcName = namespaceAndFuncName:match( "(.+)%.(.+)" )
	if not( namespaceName and funcName ) then
		error( "namespaceAndFuncName was not properly formatted", 2 )
	end
	
	local namespace = require( namespaceName )
	if not namespace then
		error( "no such namespace: " .. namespaceName, 2 )
	end
	
	local fn = namespace[ funcName ]
	if type( fn ) ~= 'function' then
		error( "no such function: " .. namespaceName .. "." .. funcName, 2 )
	end

	local args = { ... }
	args.n = select( '#', ... )

	local docStoreSelf = self
	local db = self._db

	local fakeDocStore = {}

	----------------------------------------------------------------------------
	
	function fakeDocStore:docValue( docId, optSlotName )
		return docStoreSelf:docValue( docId, optSlotName )
	end

	----------------------------------------------------------------------------

	function fakeDocStore:setDocValue( combinedValue )

		-- Unrolling WFDocStoreImp because we have a transaction boundary
		-- there that we don't need here.

		check( '_', 'table' )
	
		local args = extractAndValidateDocValue( docStoreSelf, 'setDocValue', combinedValue )
		
		return setDocValueImp_inTransaction( docStoreSelf, db, docStoreSelf._exec,
											 'setDocValue', args )
	
	end
	
	----------------------------------------------------------------------------

	function fakeDocStore:applyOverlay( overlay )
	
		local value = applyOverlayPrepareValue( docStoreSelf, overlay )
		if value then
			return fakeDocStore:setDocValue( value )
		end

	end
	
	----------------------------------------------------------------------------

	function fakeDocStore:copySlotToSlot( docId, fromSlot, toSlot )
	
		local proceed, sourceRow, destRow = copySlotToSlotPrep( docStoreSelf, docId, fromSlot, toSlot )
		
		if proceed then
			return copySlotToSlot_inTransaction( docStoreSelf, docId, toSlot, sourceRow, destRow )
		else
			return sourceRow -- is actually the hash value (didn't need to copy)
		end
		
	end

	----------------------------------------------------------------------------
	
	function fakeDocStore:quietlyDeleteSlot( docId, slotName )
	
		local oldBodyContentHash, oldAnnotationContentHash
			= quietlyDeleteSlotPrep( docStoreSelf, docId, slotName )
	
		if oldBodyContentHash and oldAnnotationContentHash then
			
			return quietlyDeleteSlot_inTransaction( docStoreSelf, docId, slotName,
													oldBodyContentHash,
													oldAnnotationContentHash )
	
		end
		
	end
	
	----------------------------------------------------------------------------
	
	function fakeDocStore:changedDocs( args )
	
		local parsedArgs = changedDocsPrep( docStoreSelf, args )
	
		local rawDocRows = changedDocsCore_inTransaction( docStoreSelf, parsedArgs )
		
		return postProcessChangedDocsRows( docStoreSelf, rawDocRows, parsedArgs )
	
	end

	----------------------------------------------------------------------------

	return db:withTransactionDo( function()

		return fn( fakeDocStore, unpack( args, 1, args.n ) )

	end )
	
end

--------------------------------------------------------------------------------

-- @mark miscellaneous

--------------------------------------------------------------------------------

function docStoreImpMethods:path()

	check( 'WFDocStoreImp' )

	return self._path

end

--------------------------------------------------------------------------------

function docStoreImpMethods:__tostring()

	check( 'WFDocStoreImp' )

	return string_format( "WFDocStoreImp( %s )", self._path )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:DEBUG_dumpCoverage( path )

	check( 'WFDocStoreImp', '+string' )

	require 'SetupCoverage'.dumpCoverage( path )

end

--------------------------------------------------------------------------------

function docStoreImpMethods:nTransactions()

	return self._db:nTransactions()

end

--------------------------------------------------------------------------------

function docStoreImpMethods:haltAndDelete()

	local path = self._path
	
	self:close() -- should cause DB object and friends to close

	if path then
		WFFileUtils.delete( path )
		WFFileUtils.delete( path .. '-shm' )
		WFFileUtils.delete( path .. '-wal' )
	end


end

--------------------------------------------------------------------------------

local function getCountQueryResult( db, sql )

	local rows = db:gatherArrayOfArray( sql )
	
	return rows and rows[ 1 ] and rows[ 1 ][ 1 ]

end

--------------------------------------------------------------------------------

function docStoreImpMethods:performIntegrityCheck()

	check( 'WFDocStoreImp' )
	
	local failures = { }

	for _, checker in ipairs( integrityCheckers ) do
		
		local result, message = checker( self._db, self._exec )
		
		if not result then
			failures[ #failures + 1 ] = message
		end
		
	end
	
	return failures
	
end

--------------------------------------------------------------------------------

function docStoreImpMethods:sendDiagnosticLogData()

	check( 'WFDocStoreImp' )

	-- Send information about index update performance.
	
	return {
	
		name = "WFDocStoreImp :: " .. self._path,
		id = 'WFDocStoreImp-' .. self._path,
		weight = 1600,
		
		body = {
		
			nDocsAllSlots = getCountQueryResult( self._db,
								"SELECT count() FROM documents" ),
			nDocsMainSlot = getCountQueryResult( self._db,
								"SELECT count() FROM documents WHERE slotName == 'main'" ),

			nTransactions = self._db:nTransactions(),
		
		},
	
	}

end

--------------------------------------------------------------------------------

return Rx.Class 'WFDocStoreImp'( docStoreImpMethods )
