--[[----------------------------------------------------------------------------

WFDocStore_idLocalTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'


--============================================================================--

local WFDocStore_idLocalTest = {}

local derefOpt = KSTableUtils.derefOpt

local docStore

local kSlot = '< slot >'
local kLocalId = '< local id >'

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.setUp()

	docStore = WFDocStore( ':memory:' )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.tearDown( coveragePath )

	if docStore then

		if coveragePath then

			local oldDocStore = docStore
			docStore:DEBUG_dumpCoverageP( coveragePath )
				:whenResolvedDo( function()
					oldDocStore:close()
				end )
				:runUntilDone()

		else

			docStore:close()

		end

		docStore = nil

	end

end

--------------------------------------------------------------------------------

local function extractIdLocal( doc )
	return derefOpt( doc, '_idLocal' ), derefOpt( doc, 'my_local_id' )
end

--------------------------------------------------------------------------------

local function addOneTuple( ... )
	return 1, ...
end

--------------------------------------------------------------------------------

local function id( i )
	return "test id " .. tostring( i )
end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testIdLocalIsResilient( test )

	test = WFAsyncTestCase( test )

	local function skipFn1( i )
		return ( i * 17 ) % 20 + 1
	end

	local function testIdLocals( ... )
		local t = { ... }
		local q = #t
		if q == 0 then
			q = 20
		end
		for n = 1, q do

			local i = t[ n ] or n
			-- Set doc value and capture rev ID.

			test:waitFor( function()

				return docStore:docValueP( id( i ), 'main' )
				:mapDataP_new( extractIdLocal )

			end )
			:continueWhenValueMatches( 'data', i, i )

		end
	end

	for i = 1, 20 do

		test:waitFor( function()
			return docStore:setDocValueP {
				_id = id( i ),
				_slot = 'main',
				my_local_id = i,
				test = 42,
			}
		end )

	end
	
	testIdLocals()

	-- Let's see if bulkSetDocValue disrupts the idLocal

	local skipIds = {}
	local bulk = {}

	for i = 1, 5 do

		local _idLocal = skipFn1( i )
		local _id = id( _idLocal )
		skipIds[ #skipIds + 1 ] = _idLocal
		bulk[ i ] = {
			_id = _id,
			_slot = 'main',
			my_local_id = _idLocal,
			test = "forty-two",
		}

	end

	test:waitFor( function()

		return docStore:bulkSetDocValueP( bulk )

	end )

	testIdLocals( unpack( skipIds ) )

	-- Let's see if setDocValueP disrupts the idLocal

	test:waitFor( function()

		return docStore:setDocValueP{
			_id = id( 13 ),
			_slot = 'main',
			my_local_id = 13,
			test = "two and forty",
		}

	end )

	testIdLocals( 13 )

	-- Let's see if applyOverlayP disrupts the idLocal

	test:waitFor( function()

		return docStore:applyOverlayP {
			_id = id( 7 ),
			_slot = 'main',
			test = "two and forty",
		}

	end )

	testIdLocals( 7 )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testSetDocValueP_failsNew_integerId( test )

	test = WFAsyncTestCase( test )

	test:waitFor( function()

		return docStore:setDocValueP {
			_id = 1,
			_slot = 'main',
			test = 42,
		}
	
	end )

	:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testSetDocValueP_integerId( test )

	test = WFAsyncTestCase( test )

	test:waitFor( function()

		return docStore:setDocValueP {
			_id = id( 1 ),
			_slot = 'main',
			test = 42,
		}
	
	end )
	
	test:waitFor( function()
	
		return docStore:setDocValueP {
			_id = 1,
			_slot = 'main',
			test = 'forty-two',
		}

	end )
	
	test:waitFor( function()
	
		return docStore:docValueP( 1, 'main' )
		:mapDataP_new( function( doc )
			return doc.test
		end )

	end )

	:continueWhenValueMatches( 'data', 'forty-two' )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testApplyOverlayP_integerId( test )

	test = WFAsyncTestCase( test )

	test:waitFor( function()

		return docStore:setDocValueP {
			_id = id( 1 ),
			_slot = 'main',
			test = 42,
		}
	
	end )
	
	test:waitFor( function()
	
		return docStore:applyOverlayP {
			_id = 1,
			_slot = 'main',
			test = 'forty-two',
		}

	end )
	
	test:waitFor( function()
	
		return docStore:docValueP( 1, 'main' )
		:mapDataP_new( function( doc )
			return doc.test
		end )

	end )

	:continueWhenValueMatches( 'data', 'forty-two' )

end

--------------------------------------------------------------------------------

local function getWithBehavior( test, methodName, mapFn, ... )

	local getArgs = { ... }
	
	for i, a in ipairs( getArgs ) do
		if a == kLocalId then
			getArgs[ i ] = 1
		elseif a == kSlot then
			getArgs[ i ] = 'main'
		end
	end

	WFAsyncTestCase( test )

	:waitFor( function()

		return docStore:setDocValueP {
			_id = id( 1 ),
			_slot = 'main',
			type = 'mydoc',
			my_local_id = 1,
			test = 42,
		}
		
	end )
	
	:waitFor( function()

		return docStore[ methodName ]( docStore, unpack( getArgs ) )
		:closingMapB( mapFn )

	end )

	:continueWhenValueMatches( 1, 1 )

end

--------------------------------------------------------------------------------

local function testGetWithPromise( test, methodName, mapFn, ... )

	local getArgs = { ... }
	getArgs[ #getArgs + 1 ] = 1
	getArgs[ #getArgs + 1 ] = 'main'

	WFAsyncTestCase( test )

	:waitFor( function()

		return docStore:setDocValueP {
			_id = id( 1 ),
			_slot = 'main',
			my_local_id = 1,
			type = 'mydoc',
			test = 42,
		}

	end )
	
	:waitFor( function()

		return docStore[ methodName ]( docStore, unpack( getArgs ) )
		:mapDataP_new( mapFn )

	end )

	:continueWhenValueMatches( 'data', 1, 1 )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testDocValueB_integerId( test )

	getWithBehavior( test, 'docValueB', extractIdLocal, kLocalId, kSlot )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testDocValueCheckTypeB_integerId( test )

	getWithBehavior( test, 'docValueCheckTypeB', extractIdLocal, 'mydoc', kLocalId, kSlot )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testFieldValueB_integerId( test )

	getWithBehavior( test, 'fieldValueB', addOneTuple, kLocalId, 'my_local_id', kSlot )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testFieldValueCheckTypeB_integerId( test )

	getWithBehavior( test, 'fieldValueCheckTypeB', addOneTuple, 'mydoc', kLocalId, 'my_local_id', kSlot )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testDocValueP_integerId( test )

	testGetWithPromise( test, 'docValueP', extractIdLocal )

end

--------------------------------------------------------------------------------

function WFDocStore_idLocalTest.testDocValueCheckTypeP_integerId( test )

	testGetWithPromise( test, 'docValueCheckTypeP', extractIdLocal, 'mydoc' )

end

--------------------------------------------------------------------------------

return WFDocStore_idLocalTest
