--[[----------------------------------------------------------------------------

WFIndexStore.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'
local KSReactiveTimer = require 'KSReactiveTimer'

	-- WF.core.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFWorkerThread = require 'WFWorkerThread'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'
local KSBucketBrigade = require 'KSBucketBrigade'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )

local DEBUG = _G.DEBUG


--============================================================================--
--[[!
	\class WFIndexStore
	\brief Tracks an \c OzDocStore and provides SQL-style indexing of its contents.
	
	\todo Write more narrative describing how it gets used.
--]]

-- @mark WFIndexStore

local indexStoreMethods = {

	_indexStoreWorkerThread = false,
		-- (WFWorkerThread) reference to the event processor that does the indexing

	_docStoreId = false,
		-- (string) doc store ID, used for debugging
		
	_queryResultsCache = false,
		-- (fully weak table) a mapping from cache keys to query results
		
	_queryResultsBrigade = false,
		-- (KSBucketBrigade) used to avoid having GC wipe things out too fast
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Creates an index store object that tracks the contents of an Oz document store.
	
	\param docStore (\c WFDocStore or \c OzDocStore) The document store object.
	\param indexSpecScript (string) The name of an index definition script for this index
--]]

function indexStoreMethods:__init( --[[WFDocStore or OzDocStore]] docStore,
								   --[[string]] indexSpecScript )

	check( 'WFIndexStore', 'WFDocStore|OzDocStore', '+string' )

	local docStoreId
	
	if WFDocStore:isClassOf( docStore ) then
		docStoreId = docStore:path()
	else
		docStoreId = 'OzDocStore ' .. docStore:catalogId()
	end

	self._docStoreId = docStoreId
	
	self._queryResultsCache = setmetatable( { }, { __mode = 'kv' } )
	
	local cacheBrigadeDepth = 10  -- depth 10 (10 GC cycles)
	
	-- FIXME: What we really want to test is whether we have "big memory"
	
	if require 'WFPlatformInfo'.isConstrainedMemoryDevice then
		cacheBrigadeDepth = 5
	end
	
	self._queryResultsBrigade = KSBucketBrigade.makeBrigade( cacheBrigadeDepth )
		
	self:setOwned {
		_indexStoreWorkerThread = WFWorkerThread( indexSpecScript, docStoreId )
	}
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Performs a single SQL query on the index store and returns the result
		as a Promise.

		Before performing the query, the method brings all relevant index tables
		up to date. If data is added to the doc store while this request
		is being processed, the newer data might or might not appear in the query results.
		
	\param sql (string) The SQL query to be performed. OR a table:
		- \c sql: (string) SQL query to be performed
		- \c tempIds: (optional, array of doc ids) if present, the table TEMP_ids will
			be populated with these ids for use in the query, e.g. as part of
			an join such as "INNER JOIN TEMP_ids ON TEMP_ids.docId = yourTable.docId"
			or via a where in clause such as "WHERE docId IN ( SELECT docId from TEMP_ids )"
			TEMP_ids is a temporary table with only one column, docId, and no special indexes.
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query
	\param ... (optional) The query arguments to be substituted for ? placeholders in the query.
		ONLY if sql was provided as a string, not as the table form.
	
	\return <code>P< table ></code> A Promise that resolves to the query results,
		as an array of arrays.
--]]

function indexStoreMethods:queryP( sql, ... )

	check( 'WFIndexStore', 'string|table' )
	
	if type( sql ) == 'string' then
		sql = {
			sql = sql,
			...
		}
	end
	
	return self._indexStoreWorkerThread:requestP( 'queryP', sql )

end

--------------------------------------------------------------------------------

local function getCachedResult( self, cacheKey )

	if not cacheKey then
		return
	end
	
	local entry = self._queryResultsCache[ cacheKey ]
	
	if not entry then
		return
	end
	
	self._queryResultsBrigade( entry )
	
	return entry.data

end

--------------------------------------------------------------------------------

local function storeCachedResult( self, cacheKey, value )

	if not cacheKey or not value or not WFUnknown.isKnown( value ) then
		return
	end
	
	local entry = self._queryResultsCache[ cacheKey ]
	
	if entry then
		entry.data = value
	else
		entry = { data = value }
		self._queryResultsCache[ cacheKey ] = entry
	end
	
	self._queryResultsBrigade( entry )

end

--------------------------------------------------------------------------------

local kUnknown = WFUnknown.value()
	
--------------------------------------------------------------------------------

local enqueue2selector = {
	normal = 'queryB',
	fast = '+queryB',
	slow = '-queryB'
}

--------------------------------------------------------------------------------

--[[!
	\brief Reports as a Behavior  the results of a SQL query on the index store.
	
	\param args (table) arguments for query
		- \c sql: (string) SQL query to be performed
		- \c calm: (optional, number) if present, calm query result updates to no more
			than one every _n_ seconds
		- \c cacheKey: (optional, string) if supplied, this will be used as a cache
			value to look up cached query results
		- \c enqueue: (optional, string) enqueue policy for this query. Default is 'normal'.
			other values are 'fast' and 'slow'. The former will push the query to the front
			of the list. The latter will cause even the initial response to be handled
			in order.
		- \c nArgs: (optional, number) if present, specifies number of arguments
			(overrides #args)
		- \c tempIds: (optional, array of doc ids) if present, the table TEMP_ids will
			be populated with these ids for use in the query, e.g. as part of
			an join such as "INNER JOIN TEMP_ids ON TEMP_ids.docId = yourTable.docId"
			or via a where in clause such as "WHERE docId IN ( SELECT docId from TEMP_ids )"
			TEMP_ids is a temporary table with only one column, docId, and no special indexes.
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query
	
	For compatibility with an older version of this API, this function also permits
	the first argument to be a string (interpreted as the SQL query string) followed
	by any number of query parameters as ... arguments. Future code should use the
	new named-argument syntax. Note that \c calm is inaccessible through the old syntax.
	
	You may use a behavior for \c sql and any of the query arguments. In this case
	the actual query performed will update in response to changes to the behavior(s).
	
	\return B< table > A Behavior whose value tracks query results, as an array of arrays.
	
	\b WARNING: Use this method sparingly. Creating many query behaviors will adversely
		affect the overall performance of the application.
	
	\b WARNING: This behavior may, at times, return information that is based on
		incomplete index table results. It will eventually catch up and update
		results if needed. (Out of date information is especially likely for the
		first value returned.)

--]]

function indexStoreMethods:queryB( args, ... )

	check( 'WFIndexStore', 'string|table' )

	-- Retrofit for old syntax.

	assert( not KSReactive.isBehavior( args ), 'SQL can no longer be specified via a behavior' )
	if type( args ) == 'string' then
		local nArgs = select( '#', ... )
		args = { sql = args, nArgs = nArgs, ... }
	end

	-- Fill in thread and traceback info if in debug build.

	args.threadName = DEBUG and WFCurrentThread.DEBUG_getThreadName() or nil
	args.traceback = DEBUG and debug.traceback( '', 2 ) or nil
	
	-- Are there any behaviors in params?
	
	local nArgs = args.nArgs or #args
	
	for i = 1, nArgs do
		assert( not KSReactive.isBehavior( args[ i ] ), 'parameters to SQL can no longer be specified via a behavior' )
	end
	
	-- Figure out the initial value. This will bias how we handle the
	-- enqueue logic.
	
	local cacheKey = args.cacheKey
	local initialValue = getCachedResult( self, cacheKey ) or kUnknown
	local haveCachedValue = initialValue ~= kUnknown
	
	-- Reconstruct the query as needed for behavior changes.
	
	local indexStoreWorkerThread = self._indexStoreWorkerThread
	
	local function makeQueryRequestE( sql, ... )

		local nArgs = select( '#', ... )
	
		local newArgs = {
			sql = sql,
			calm = args.calm,
			encoder = args.encoder,
			tempIds = args.tempIds,
			nArgs = nArgs,
			...
		}
		
		assert( not WFCloseSupport.hasBeenOrWillBeClosed( self ),
			    "queryB running after index store closed" )
			
		local enqueue = args.enqueue
		
		-- If we've got a value already, we don't need to rush to compute a new one
		-- But note that we only want to do this once.

		if haveCachedValue then
			enqueue = 'slow'
			haveCachedValue = false
		end
		
		if enqueue == nil then
			enqueue = 'normal'
		end
		
		local selector = enqueue2selector[ enqueue ]
		
		if not selector then
			debug.print( "Invalid enqueue value for index store: " .. tostring( enqueue ) )
			selector = enqueue2selector[ 'normal' ]
		end
		
		return indexStoreWorkerThread:requestE( selector, newArgs )
	
	end
	
	local function makeQueryRequestBWithInitialValue( initialValue, ... )
		local e = makeQueryRequestE( ... )
		local b = e:closingHoldB( initialValue )
		return b
	end
	
	local function makeResultQueryB( ... )
		
		-- If we have a cache key, then we want to cache results.
		
		if cacheKey then
			storeCachedResult( self, cacheKey, ( ... ) )
		end
		
		return makeQueryRequestBWithInitialValue( initialValue, ... )
		
	end
	
	local resultB
	
	if haveCachedValue then
	
		-- Wait 0.25 seconds
		
		local delayB = KSReactiveTimer.falseThenTrueB( 0.25 )
		
		-- Then wait for idle
			
		delayB = delayB:closingMapSwitchB( function( value )
			if value then
				return KSReactiveTimer.idleB()
			else
				return KSReactive.constantB( false )
			end
		end )
		
		-- Then do the query
		
		resultB = delayB:closingMapSwitchB( function( value )
			if value then
				return makeResultQueryB( args.sql, unpack( args, 1, nArgs ) )
			else
				return KSReactive.constantB( initialValue )
			end
		end )
		
	else

		resultB = makeResultQueryB( args.sql, unpack( args, 1, nArgs ) )

	end
	
	self:addCloseTarget( resultB )
	
	WFCloseSupport.mustClose( resultB )
		
	return resultB

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves (as a Behavior) the number of documents left to index.
	
	This can be used to approximate the time left until the index store
	catches up.
	
	\return (<code>B< number ></code>) A Behavior whose value is the number of documents.
--]]

function indexStoreMethods:sizeOfIndexQueueB()

	local resultB = self._indexStoreWorkerThread:requestE( 'sizeOfIndexQueueB' )
		:closingHoldB( 0 )
	
	self:addCloseTarget( resultB )
	
	return resultB

end

--------------------------------------------------------------------------------

function indexStoreMethods:indexingProgressB( ... )

	local resultB = self._indexStoreWorkerThread:requestE( 'indexingProgressB', ... )
		:closingHoldB( nil ):closingMapB( function( value )
			return value == nil and WFUnknown.value() or value
		end )
	
	self:addCloseTarget( resultB )
	
	return resultB

end

--------------------------------------------------------------------------------

function indexStoreMethods:shouldWaitForIndexingP( ... )

	return self._indexStoreWorkerThread:requestP( 'shouldWaitForIndexingP', ... )

end

--------------------------------------------------------------------------------
--[[!
	\brief Permanently shuts down this index store and deletes all related files.
--]]

function indexStoreMethods:haltAndDeleteP()

	return self._indexStoreWorkerThread:requestP( 'haltAndDelete' )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

function indexStoreMethods:halt()

	return self._indexStoreWorkerThread:requestP( 'halt' )
		:runUntilDoneP()
		
end

--------------------------------------------------------------------------------

function indexStoreMethods:DEBUG_dumpCoverageP( path )

	-- Should be used only by WFIndexStoreTest.lua.

	local resultP = self._indexStoreWorkerThread:requestP( 'DEBUG_dumpCoverageP', path )
	
	resultP:runUntilDone()
	
	return resultP

end

--------------------------------------------------------------------------------

return Rx.Class 'WFIndexStore'( indexStoreMethods )
