--[[----------------------------------------------------------------------------

WFIndexStoreImp.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua.threading
local WFCurrentThread = require 'WFCurrentThread'
local WFRequestDispatcher = require 'WFRequestDispatcher'

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.reactive
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFCapabilities = require 'WFCapabilities'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'
local WFOzDocStorePath = require 'WFOzDocStorePath'

	-- WF.oz_client.index_store
local WFIndexStoreMapServer = require 'WFIndexStoreMapServer'
local WFIndexStoreQueryServer = require 'WFIndexStoreQueryServer'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )


--============================================================================--

local indexStoreImpMethods = {

	_path = false,
		-- (string) path to DB file
	
	_mapServer = false,
		-- (WFIndexStoreMapServer) map server
	
	_queryServer = false,
		-- (WFIndexStoreQueryServer) query server

	_log = false,
		-- (WFLogger)
	
	_existingIndexStatus = false,
		-- (string) status of previous index file (for diagnostic log only)
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local hasCreatedAnIndexStore = false
	-- only permit one index store per message processor

--------------------------------------------------------------------------------
--[[!
	\brief Create a new index store implementation.
	
	Should only be used by index definition files.

	\param (\c WFIndexStoreSchema) index store schema
	\param (string) doc store ID (from WFIndexStore client thread code)
--]]

function indexStoreImpMethods:__init( indexStoreSchema, docStoreId )

	check( 'WFIndexStoreImp', 'WFIndexStoreSchema', 'string' )

	if hasCreatedAnIndexStore then
		error( "Can only create one WFIndexStoreImp per message processor", 2 )
	end
	
	hasCreatedAnIndexStore = true

	-- NOTE: Catalog ID should have been validated by WFIndexStore
	-- client-side code.

	self._log = WFLogger( string.format( 'WFIndexStoreImp %s', docStoreId ) )
		--:setLevel 'debug'

	self._log:debugf( "initializing with docStore = %s", docStoreId )

	local docStore, catalogId
	
	if docStoreId:match( "^OzDocStore " ) then

		catalogId = docStoreId:sub( 12 )
		docStore = require 'OzDocStore'( catalogId )

	else

		docStore = WFDocStore( docStoreId )

	end
	
	self:addCloseTarget( docStore )
	
	WFCloseSupport.mustCloseAtThreadTermination( self )
	
	WFDiagnosticLog.addLogProvider( self )
	
	-- Register event processor event handlers.
	
	WFRequestDispatcher.onMessageDo {
		message = 'queryB',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'queryP',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'sizeOfIndexQueueB',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'indexingProgressB',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'shouldWaitForIndexingP',
		target = self,
	}

	WFRequestDispatcher.onMessageDo {
		message = 'haltAndDelete',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'halt',
		target = self,
	}
	
	WFRequestDispatcher.onMessageDo {
		message = 'DEBUG_dumpCoverageP',
		target = self,
	}
	
	-- Create database in appropriate location.

	local dbPath
	local existingIndexStatus = 'testing'

	if WFDocStore:isClassOf( docStore ) then
	
		if docStoreId:sub( 1, 8 ) == ':memory:' then
	
			dbPath = ':memory:'
		
		elseif docStoreId:sub( 1, 6 ) == ':temp:' then
	
			dbPath = ':temp:'
		
		else
		
			dbPath = docStore:path():gsub( '%.[^./\\]*$', '' ) .. '.wfindex'
		
		end
	
	else

		if catalogId:sub( 1, 8 ) == ':memory:' then
	
			dbPath = ':memory:'
		
		elseif catalogId:sub( 1, 6 ) == ':temp:' then
	
			dbPath = ':temp:'
		
		else
	
			-- On 2012-05-23, we moved index store from caches to persistent
			-- storage. Try to reuse the file if possible.
			
			-- FIXME [scouten 2012-08-22]: How long before we simply ignore the
			-- old files and rebuild?
			
			local catalogDir = WFOzDocStorePath.dirPathForCatalogId( catalogId )
	
			dbPath = KSPathUtils.child( catalogDir, 'index' )
			
			if not WFFileUtils.exists( dbPath )
				and not WFFileUtils.exists( dbPath .. '-shm' )
				and not WFFileUtils.exists( dbPath .. '-wal' )
			then
			
				-- No index DB at new path. See if we have one at the old (caches)
				-- path that we can move.
				
				local cachesDir = KSPathUtils.getStandardFilePath 'carouselCaches'
				
				local oldIndexFileName = string.format( '%s.ozindex', catalogId )
				
				local maybeOldIndexPath = KSPathUtils.child( cachesDir,
															 oldIndexFileName )
	
				-- Yes, there is an existing index DB at old path. Move it
				-- to new location.
				
				if WFFileUtils.exists( maybeOldIndexPath ) then
				
					-- Move the -shm and -wal files if they exist. If not, ignore
					-- failure.
				
					WFFileUtils.move( maybeOldIndexPath, dbPath )
					WFFileUtils.move( maybeOldIndexPath .. '-shm',
									  dbPath .. '-shm' )
					WFFileUtils.move( maybeOldIndexPath .. '-wal',
									  dbPath .. '-wal' )
				
				end
			
			end
	
			existingIndexStatus = WFFileUtils.exists( dbPath )
									and 'exists' or 'missing'
	
		end
		
	end

	self._path = dbPath
		
	self._log:debugf( "WFIndexStoreImp.start, dbPath = %s", dbPath )

	-- If we fail trying to open the database, just trash it and start again.
	-- [scouten 2012-03-06, PR# 3107228]
	
	-- Or if the capability is set to trigger this, e.g. for testing how
	-- fast we are at doing a rebuild.

	local forceRebuild = "rebuild" == WFCapabilities.getCapability "WFIndexStoreImp.rebuildIndexStore"

	local db
	if forceRebuild or not pcall( function()

		db = indexStoreSchema:openOrCreateDatabaseAtPath( dbPath )

	end ) then
	
		if not forceRebuild then
			self._log:warn( "WFIndexStore was corrupt. Deleting and rebuilding." )
		end

		-- Trash the index DB and its possible sidecars.

		WFFileUtils.delete( dbPath )
		WFFileUtils.delete( dbPath .. '-shm' )
		WFFileUtils.delete( dbPath .. '-wal' )
		
		existingIndexStatus = 'corrupt'
		
		-- Try again. If we fail this time, that's it, man, game over, it's game over.
		
		db = indexStoreSchema:openOrCreateDatabaseAtPath( dbPath )

	end
	
	self:addCloseTarget( db )
	
	db:setAutosaveInterval( 15 )
		-- TODO [scouten 2014-02-04]: Make this configurable?

	-- Spin up map server and query server.
	
	self:setOwned { _mapServer = WFIndexStoreMapServer( docStore, db, indexStoreSchema ) }
	
	self:setOwned {
		_queryServer = WFIndexStoreQueryServer( self._mapServer, indexStoreSchema:tableNames() )
	}

	-- Wait for map server to be ready and report index status to Omniture.

	self._mapServer:isReadyP()
		:whenResolvedDo( function( status, maybeOutOfSync )
		
			if type( maybeOutOfSync ) == 'string' then
				existingIndexStatus = maybeOutOfSync
			end

			self._existingIndexStatus = existingIndexStatus
		
		end )
		:runUntilDone()

end

--------------------------------------------------------------------------------
--[[!
	\brief Create a behavior that tracks the results of a SQL query on the index store.
	
	\param args (table) arguments for query
		- \c sql: (string) SQL query to be performed
		- \c threadName: (optional, string) name of thread requesting the query
		- \c traceback: (optional, string) traceback of call site in requesting thread
		- \c calm: (optional, number) if present, calm query result updates to no more
			than one every _n_ seconds
		- \c outOfDateNotice: (optional, boolean) if true, query results will contain
			a member "outOfDate" (boolean) which will be true if query results are
			known to be lagging behind the doc store
		- \c nArgs: (optional, number) number of query arguments (helps with \c nil values
			in the middle of the list)
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query
	
	\return B< table > query results (array of arrays syntax)
	
	\b WARNING: Use this method sparingly. Creating many query behaviors will adversely
		affect the overall performance of the application.
	
	\b WARNING: This behavior may, at times, return information that is based on
		incomplete index table results. It will eventually catch up and update
		results if needed. (Out of date information is especially likely for the
		first value returned.)
--]]

function indexStoreImpMethods:queryB( --[[table]] args )

	check( 'WFIndexStoreImp', 'table' )

	return self._queryServer:queryB( args )

end

--------------------------------------------------------------------------------
--[[!
	\brief Create a promise that performs a single SQL query on the index store.

	This promise will first ensure that the relevant index tables are brought up
	to date as of the time this method is called before performing the query.
		
	\param args (table) table of args containing key 'sql' = SQL query to be performed
		and array of optional query arguments to be substituted for ? placeholders in query
	
	\return P< table > query results (array of arrays syntax)
--]]

function indexStoreImpMethods:queryP( args )

	-- FIXME [scouten 2011-03-31]: Need prioritization on these requests.

	check( 'WFIndexStoreImp', 'table' )

	return self._queryServer:queryP( args )

end

--------------------------------------------------------------------------------
--[[!
	\brief Return the number of documents left to index.
	
	This can be used to approximate the time left until the index store
	catches up.
	
	\return (<code>B< number ></code>) how many documents left to index?
--]]

function indexStoreImpMethods:sizeOfIndexQueueB()

	return self._mapServer:sizeOfIndexQueueB()

end

--------------------------------------------------------------------------------

function indexStoreImpMethods:shouldWaitForIndexingP( optBacklogThreshold, optTableSet )

	local backlogThreshold = optBacklogThreshold or 0
	local backlogB = self._mapServer:totalIndexBacklogB( optTableSet )
	
	local bigBacklogB = backlogB:closingMapB( function( value )
		if type( value ) == "number" then
			return value > backlogThreshold
		end
	end )
	
	local resultP = KSReactivePromiseBridge.behaviorToPromise( bigBacklogB, function( value )
		return type( value ) == "boolean"
	end ):observeP( function()
		bigBacklogB:close()
	end )

	return resultP

end

--------------------------------------------------------------------------------

function indexStoreImpMethods:indexingProgressB( optTableSet )

	local backlogB = self._mapServer:totalIndexBacklogB( optTableSet )
	local updateP = self._mapServer:bringTablesUpToDateP( optTableSet )
	
	updateP:runAttachedTo( backlogB )
	
	local startingValue
	return backlogB:closingMapB( function( value )
		if value == 0 then
			return 1 -- nothing to do!
		elseif type( value ) == "number" then
			if not startingValue then
				-- startingValue should be the largest number ever produced
				startingValue = value
			end

			-- ... therefore this output should map onto [0, 1]
			return ( startingValue - value ) / startingValue
		elseif WFUnknown.isUnknown( value ) then
			return nil -- cannot transit the unknown value
		else
			assert( false, string.format( "unexpected backlogB value: %s",
				tostring( value ) ) )
		end
	end )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Permanently shut down this index store and delete all related files.
--]]

function indexStoreImpMethods:haltAndDelete()

	local path = self._path
	
	self:close() -- should cause DB object and friends to close

	if path then
		WFFileUtils.delete( path )
		WFFileUtils.delete( path .. '-shm' )
		WFFileUtils.delete( path .. '-wal' )
	end

end

--------------------------------------------------------------------------------

function indexStoreImpMethods:halt()

	local path = self._path
	
	self:close() -- should cause DB object and friends to close
	
	return path or false
	
end

--------------------------------------------------------------------------------

function indexStoreImpMethods:sendDiagnosticLogData()

	-- Send information about query requests that are outstanding.

	local filteredQueryBRequests = self._queryServer:queryBStats()

	local nQueryRequestsCreated = filteredQueryBRequests.nQueryRequestsCreated
	filteredQueryBRequests.nQueryRequestsCreated = nil
	
	-- Send information about index update performance.
	
	return {
	
		name = "Index Store Performance :: " .. WFCurrentThread.DEBUG_getThreadName(),
		id = 'WFIndexStoreImp-' .. WFCurrentThread.DEBUG_getThreadName():replace( ' :: ', '+' ),
		weight = 1800,
		
		body = {
		
			path = self._path,
			
			currentQueryRequests = filteredQueryBRequests,
			
			tableUpdateProfile = self._mapServer:indexUpdatePerformanceStats(),
		
			nQueryRequestsCreated = nQueryRequestsCreated,
			nQueryRequestsActive = #filteredQueryBRequests,
			
			existingIndexStatus = self._existingIndexStatus,
		
		},
	
	}

end

--------------------------------------------------------------------------------

function indexStoreImpMethods:DEBUG_dumpCoverageP( path )

	require 'SetupCoverage'.dumpCoverage( path )

	return require 'KSPromises'.resolvedPromiseP( 'data', true )

end

--------------------------------------------------------------------------------

local WFIndexStoreImp_class = Rx.Class 'WFIndexStoreImp'( indexStoreImpMethods )

return {
	
	start = function( ... )
	
		local indexStore = WFIndexStoreImp_class( ... )
		rawset( _G, 'indexStore', indexStore )
	
	end,
	
}
