--[[----------------------------------------------------------------------------

WFIndexStoreMapServerMetricsTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'

	-- WF.oz_client.doc_store
local OzDocStore = require 'OzDocStore'

	-- WF.oz_client.index_store
local WFIndexStoreMapServer = require 'WFIndexStoreMapServer'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'


--============================================================================--

-- @mark WFIndexStoreMapServerMetricsTest

local WFIndexStoreMapServerMetricsTest = {}

--------------------------------------------------------------------------------

local docStore, indexStoreDb, mapServer
local testMetrics

--------------------------------------------------------------------------------

local function createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema:addTable {

		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",

		map = function( docValue, emit )
		
			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.
		
			if docValue._docType == 'asset' then
				emit( docValue.width / docValue.height,
					  docValue.captureDate:sub( 1, 10 ) )
			end
			
		end,

	}
	
	indexStoreSchema:addTable {

		sql = "CREATE TABLE keywords( keyword )",

		map = function( docValue, emit )
		
			if docValue.keywords then
			
				-- Intentionally unsafe access to docValue.keywords here.
				-- If keywords is non-nil, but not a table, we want the error.
				
				for k in pairs( docValue.keywords ) do
					emit( k )
				end
			
			end
		
		end,
	
	}
	
	-- Remember that indexStoreDb and mapServer are file-scoped locals.
	
	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	mapServer = WFIndexStoreMapServer( docStore, indexStoreDb, indexStoreSchema )

end

--------------------------------------------------------------------------------

local function generateTestAssets( numberOfTimes )

	local docIds = {}
	
	for i = 1, numberOfTimes do
		docIds[ i ] = KSIdGenerator.uuidInLowercaseHex()
	end
	
	table.sort( docIds ) -- so we have predictable results later
	
	local expectedCoreInfo = {}
	
	for i = 1, #docIds do
		expectedCoreInfo[ i ] = { docIds[ i ], i, 1, '2012-10-23' }
	end
	
	return docIds, expectedCoreInfo

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.setUp()

	docStore = OzDocStore ':memory:'
	
	indexStoreDb = nil

	mapServer = nil
	
	testMetrics = nil
	
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.tearDown()

	if docStore then
		docStore:close()
		docStore = nil
	end
	
	if indexStoreDb then
		indexStoreDb:close()
		indexStoreDb = nil
	end

	if mapServer then
		mapServer:close()
		mapServer = nil
	end
	
	testMetrics = nil

end

--------------------------------------------------------------------------------

local function bulkApplyOverlayP( docId )

	return docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docIds = { docId },
		overlay = {
			width = 1,
			height = 1,
			captureDate = '2012-10-23T11:43:29',
		},
	}
	
end

--------------------------------------------------------------------------------

local function bulkApplyOverlayAndUpdateTableP( docId )

	return docStore:applyOverlayToDocumentP {
		docType = 'asset',
		docIds = { docId },
		overlay = {
			width = 1,
			height = 1,
			captureDate = '2012-10-23T11:43:29',
		},
	}
	:mapDataP( function ()
	
		return mapServer:bringTableUpToDateP( 'coreInfo' )
	
	end )
		
end

--------------------------------------------------------------------------------

local function mapServerTest_bulkApplyOverlay( test, numberOfTimes )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()
	
	local docIds, expectedCoreInfo = generateTestAssets( numberOfTimes )
	
	-- Arbitrary wait time
	local TEST_WAIT_TIME = 0.01
	
	local waitTime = numberOfTimes * TEST_WAIT_TIME
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.
		
		:waitFor( function() -- Step 3
		
			return KSPromises.allP( bulkApplyOverlayP, docIds, #docIds )
			
		end )
		:continueWhenValueMatches( 'data', true )
		
		-- Bring coreInfo table up to date.
		
		:waitForPerf( function() -- Step 4
				
			return mapServer:bringTableUpToDateP( 'coreInfo' )
		
		end )
		
		-- Verify data and performance is correct and acceptable
		
		:waitFor( function() -- Step 5
		
			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]
			
		end )
		:continueWhenValueMatches( 'data', expectedCoreInfo )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServer_100_testPerformance( test )
	
	mapServerTest_bulkApplyOverlay( test, 100 )
		
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServer_1000_testPerformance( test )
	
	mapServerTest_bulkApplyOverlay( test, 1000 )
	
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServer_5000_testPerformance( test )
	
	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	mapServerTest_bulkApplyOverlay( test, 5000 )
			
end

--------------------------------------------------------------------------------

local function mapServerTest_bulkApplyOverlayAndUpdateTable( test, numberOfTimes )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()
	
	local docIds, expectedCoreInfo = generateTestAssets( numberOfTimes )
	
	-- Arbitrary wait time
	local TEST_WAIT_TIME = 0.02
	
	local waitTime = numberOfTimes * TEST_WAIT_TIME
	
	WFAsyncTestCase( test, waitTime )

		-- Spin up the map server.
	
		:waitFor( function() -- Step 1
		
			return mapServer:isReadyP()
		
		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.
		
		:waitForPerf( function() -- Step 3
					
			return KSPromises.allP( bulkApplyOverlayAndUpdateTableP, docIds, #docIds )
			
		end )
		:continueWhenValueMatches( 'data', true )
				
		-- Verify data and performance is correct and acceptable
		
		:waitFor( function() -- Step 4
			
			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo ORDER BY docId]]
			
		end )
		:continueWhenValueMatches( 'data', expectedCoreInfo )

end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServerAlternate_100_testPerformance( test )
	
	mapServerTest_bulkApplyOverlayAndUpdateTable( test, 100 )
			
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServerAlternate_1000_testPerformance( test )
	
	mapServerTest_bulkApplyOverlayAndUpdateTable( test, 1000 )
			
end

--------------------------------------------------------------------------------

function WFIndexStoreMapServerMetricsTest.testMapServerAlternate_5000_testPerformance( test )
	
	-- We don't want this long of a test on Android -- it's annoying:
	if ( WFPlatformInfo.isAndroid ) then
		test:notifyTestDidSucceed()
		return
	end

	mapServerTest_bulkApplyOverlayAndUpdateTable( test, 5000 )
	
end

--------------------------------------------------------------------------------

return WFIndexStoreMapServerMetricsTest
