--[[----------------------------------------------------------------------------

WFIndexStoreQueryServer_WFDocStore_Test.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local KSTestUtils = require 'KSTestUtils'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStoreMapServer = require 'WFIndexStoreMapServer'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'
local WFIndexStoreQueryServer = require 'WFIndexStoreQueryServer'


--============================================================================--

-- @mark WFIndexStoreQueryServer_WFDocStore_Test

local WFIndexStoreQueryServer_WFDocStore_Test = {}

--------------------------------------------------------------------------------

local docStore, indexStoreDb, mapServer, queryServer, queryB

local LONG_TEST_TIMEOUT = 5

--------------------------------------------------------------------------------

local function createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local indexStoreSchema = WFIndexStoreSchema()

	indexStoreSchema:addTable {

		sql = "CREATE TABLE coreInfo( aspectRatio, displayDate )",

		map = function( docValue, emit )

			-- WARNING: This version is grossly simplified and omits error checking.
			-- See OzCatalogIndexDefinitions.lua for a better example.

			emit( docValue.width / docValue.height,
				  docValue.captureDate:sub( 1, 10 ) )

		end,

	}

	indexStoreSchema:addTable {

		sql = "CREATE TABLE keywords( keyword )",

		map = function( docValue, emit )

			if docValue.keywords then

				-- Intentionally unsafe access to docValue.keywords here.
				-- If keywords is non-nil, but not a table, we want the error.

				for k in pairs( docValue.keywords ) do
					emit( k )
				end

			end

		end,

	}

	-- Remember that indexStoreDb and mapServer are file-scoped locals.

	indexStoreDb = indexStoreSchema:openOrCreateDatabaseAtPath( ':memory:' )

	mapServer = WFIndexStoreMapServer( docStore, indexStoreDb, indexStoreSchema )

	queryServer = WFIndexStoreQueryServer( mapServer, indexStoreSchema:tableNames() )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.setUp()

	docStore = WFDocStore ':memory:'

	indexStoreDb = nil

	mapServer = nil

	queryB = nil

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.tearDown()

	if docStore then
		docStore:close()
		docStore = nil
	end

	if indexStoreDb then
		indexStoreDb:close()
		indexStoreDb = nil
	end

	if mapServer then
		mapServer:close()
		mapServer = nil
	end

	if queryB then
		queryB:close()
		queryB = nil
	end

end

--------------------------------------------------------------------------------

-- @mark queryB

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_validateArgs_sql( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query with bogus args.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				return queryServer:queryB {
					sql = 42
				}

			end )

			return true

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_validateArgs_threadName( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query with bogus args.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				return queryServer:queryB {
					sql = [[SELECT * FROM coreInfo ORDER BY docId]],
					threadName = 42,
				}

			end )

			return true

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_validateArgs_traceback( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query with bogus args.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				return queryServer:queryB {
					sql = [[SELECT * FROM coreInfo ORDER BY docId]],
					threadName = "someThread.lua",
					traceback = false,
				}

			end )

			return true

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_validateArgs_calm_number( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query with bogus args.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				return queryServer:queryB {
					sql = [[SELECT * FROM coreInfo ORDER BY docId]],
					calm = "5",
				}

			end )

			return true

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_validateArgs_calm_negativeNumber( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query with bogus args.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				return queryServer:queryB {
					sql = [[SELECT * FROM coreInfo ORDER BY docId]],
					calm = -1,
				}

			end )

			return true

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_validateArgs_calm_zero( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query with bogus args.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				return queryServer:queryB {
					sql = [[SELECT * FROM coreInfo ORDER BY docId]],
					calm = 0,
				}

			end )

			return true

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_noRelevantTables( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a bogus query.

		:waitFor( function()

			return KSTestUtils.testFails( function()
				return queryServer:queryB {
					sql = [[SELECT * FROM noSuchTable ORDER BY docId]],
				}
			end )

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_empty( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that map server thinks both tables are
		-- not up to date (though they are empty, the a
		-- WFDocStore inserts a cross check document that
		-- bumps the counter and makes them out of date).

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query.

		:waitFor( function()

			return queryServer:queryB {
				sql = [[SELECT * FROM coreInfo ORDER BY docId]]
			}

		end )
		:continueWhenValueMatches( {} )

		-- The queryB on coreInfo will have triggered that
		-- table to be brought up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		-- Since keywords was not queried, it will still be
		-- out of date.

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_twoRows( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	queryB = queryServer:queryB { sql = [[SELECT * FROM coreInfo ORDER BY docId]] }

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that map server thinks coreInfo is
		-- up to date (since we have a query open, it will
		-- be re-indexed to be up to date eventually)

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		-- However, the keyword table will not be up to
		-- date because the indexStoreCrossCheck document's
		-- insertion will bump the change counter and we have
		-- no open query to force it to update.

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Verify that query result is empty.

		:waitFor( queryB )
		:continueWhenValueMatches( {} )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_multipleUpdates( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	queryB = queryServer:queryB { sql = [[SELECT * FROM coreInfo ORDER BY docId]] }

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that query result is empty.

		:waitFor( queryB )
		:continueWhenValueMatches( {} )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Add more data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '5eaefdf3a27445fa8ad10f1a42159aa7',
				_slot = 'main',
				width = 17,
				height = 10,
				captureDate = '2012-08-21',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 30,
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'5eaefdf3a27445fa8ad10f1a42159aa7',
					4,
					1.7,
					'2012-08-21',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					5,
					1,
					'2012-08-17',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_calm( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	queryB = queryServer:queryB {
		sql = [[SELECT * FROM coreInfo ORDER BY docId]],
		calm = 1,
	}

	local timeBeforeFirstQueryUpdate

	WFAsyncTestCase( test, LONG_TEST_TIMEOUT )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Remember the time now.

		:waitFor( function()

			timeBeforeFirstQueryUpdate = KSDate.timeNowForAnimation()
			return true

		end )

		-- Verify that query result is empty.

		:waitFor( queryB )
		:continueWhenValueMatches( {} )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Add more data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '5eaefdf3a27445fa8ad10f1a42159aa7',
				_slot = 'main',
				width = 17,
				height = 10,
				captureDate = '2012-08-21',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 30,
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'5eaefdf3a27445fa8ad10f1a42159aa7',
					4,
					1.7,
					'2012-08-21',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					5,
					1,
					'2012-08-17',
				},
			} )

		-- Verify that we didn't get here too soon.

		:waitFor( function()

			assert( KSDate.timeNowForAnimation() - timeBeforeFirstQueryUpdate >= 1,
				"Query updated too soon - violated 1 second calming rule" )

			return true

		end )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

-- Utility behavior function that collects the values of another behavior used
-- to look for state transitions via oneAfterTheOtherB. This is necessary
-- when one of the two states is transient so that waiting on it creates a race
-- condition that can result in sporadic timeouts.

local function valueAccumulatorB( otherB )
	local values = { }
	return otherB:mapB( function( value )
		values[ #values + 1 ] = value
		return unpack( values )
	end )
end

--------------------------------------------------------------------------------

local function findDeepEqual( t, v, start )

	for i = start or 1, t.n do
		if KSTableUtils.deepEquals( t[ i ], v ) then
			return i
		end
	end

end

--------------------------------------------------------------------------------

local function oneAfterTheOtherB( valuesB, before, after )

	return valuesB:mapB( function( first, ... )

		local values = { n = select( "#", ... ), ... }
		local foundBefore = findDeepEqual( values, before )
		local foundAfter = foundBefore and findDeepEqual( values, after, foundBefore + 1 )

		return foundAfter and true

	end )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_outOfDate( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	queryB = queryServer:queryB {
		sql = [[SELECT * FROM coreInfo ORDER BY docId]],
		calm = 1,
		outOfDateNotice = true,
	}

	local queryValuesB

	local timeBeforeFirstQueryUpdate

	WFAsyncTestCase( test, LONG_TEST_TIMEOUT )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that query result is empty.
		-- Remember the time now.

		:waitFor( function()

			timeBeforeFirstQueryUpdate = KSDate.timeNowForAnimation()
			return "ok"
		
		end )
		
		:waitFor( queryB )
		:continueWhenValueMatches( { outOfDate = false } )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
				outOfDate = false,
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Start accumulating query values.

		:waitFor( function()

			queryValuesB = valueAccumulatorB( queryB )

			return true

		end )

		-- Add more data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '5eaefdf3a27445fa8ad10f1a42159aa7',
				_slot = 'main',
				width = 17,
				height = 10,
				captureDate = '2012-08-21',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 30,
			}

		end )

		-- Wait for query behavior to provide out-of-date notice
		-- followed by new, up-to-date results.

		:waitFor( function()
			return oneAfterTheOtherB( queryValuesB, {
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
				outOfDate = true,
			}, {
				{
					'5eaefdf3a27445fa8ad10f1a42159aa7',
					4,
					1.7,
					'2012-08-21',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					5,
					1,
					'2012-08-17',
				},
				outOfDate = false,
			} )
		end )
		:continueWhenValueMatches( true )

		-- Verify that we didn't get here too soon.

		:waitFor( function()

			assert( KSDate.timeNowForAnimation() - timeBeforeFirstQueryUpdate >= 1,
				"Query updated too soon - violated 1 second calming rule" )

			return true

		end )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_join( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
				keywords = { foo = true, bar = true },
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
				keywords = { blue = true, xyz = true },
			}

		end )

		-- Verify that map server thinks both tables are
		-- not up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query that joins two tables.

		:waitFor( function()

			return queryServer:queryB {
				sql = [[SELECT coreInfo.docId, aspectRatio, keyword FROM coreInfo
					INNER JOIN keywords
					WHERE coreInfo.docId == keywords.docId
					AND keyword == 'foo'
					ORDER BY coreInfo.docId]]
			}

		end )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					1.5,
					'foo',
				},
			} )

		-- Verify that map server thinks both tables are
		-- now up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_outOfDate_join( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	local timeBeforeFirstQueryUpdate = KSDate.timeNowForAnimation()
	local timeBeforeSecondQueryUpdate

	local queryValuesB

	queryB = queryServer:queryB {
		sql = [[SELECT coreInfo.docId, aspectRatio, keyword FROM coreInfo
			INNER JOIN keywords
			WHERE coreInfo.docId == keywords.docId
			AND keyword == 'foo'
			ORDER BY coreInfo.docId]],
		calm = 1,
		outOfDateNotice = true,
	}

	WFAsyncTestCase( test, LONG_TEST_TIMEOUT )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that query result is empty.
		-- Remember the time now.

		:waitFor( queryB )
		:continueWhenValueMatches( { outOfDate = false } )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
				keywords = { foo = true, bar = true },
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
				keywords = { blue = true, xyz = true },
			}

		end )

		-- Wait for query behavior to update.

		:waitFor( queryB )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					1.5,
					'foo',
				},
				outOfDate = false,
			} )

		-- Verify that we didn't get here too soon.

		:waitFor( function()

			timeBeforeSecondQueryUpdate = KSDate.timeNowForAnimation()

			assert( KSDate.timeNowForAnimation() - timeBeforeFirstQueryUpdate >= .9,
				"Query updated too soon - violated 1 second calming rule" )

			return true

		end )

		-- Verify that map server thinks both tables are now
		-- up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( true )

		-- Start accumulating query values.

		:waitFor( function()

			queryValuesB = valueAccumulatorB( queryB )

			return true

		end )

		-- Add more data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '5eaefdf3a27445fa8ad10f1a42159aa7',
				_slot = 'main',
				width = 17,
				height = 10,
				captureDate = '2012-08-21',
				keywords = { blue = true, foo = true },
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 30,
			}

		end )

		-- Wait for query behavior to provide out-of-date notice
		-- followed by new, up-to-date results.

		:waitFor( function()
			return oneAfterTheOtherB( queryValuesB, {
				{
					'651d079c86bd4514aa0ea10874515af2',
					1.5,
					'foo',
				},
				outOfDate = true,
			}, {
				{
					'5eaefdf3a27445fa8ad10f1a42159aa7',
					1.7,
					'foo',
				},
				{
					'651d079c86bd4514aa0ea10874515af2',
					1.5,
					'foo',
				},
				outOfDate = false,
			} )
		end )
		:continueWhenValueMatches( true )

		-- Verify that we didn't get here too soon.

		:waitFor( function()

			local timeInterval = KSDate.timeNowForAnimation() - timeBeforeSecondQueryUpdate

			-- We don't have access to the exact time at which the query behavior fires
			-- in the worker thread, so we can't enforce 1.00 seconds precisely. Deal with
			-- this inevitable race condition by allowing a small amount of measurement slop.

			if timeInterval < 0.96 then
				error( string.format( "Query updated too soon, was %.2f seconds, expected > 1.00 seconds",
									  timeInterval ) )
			end

			return true

		end )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryB_queryParams( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that map server thinks both tables are
		-- not up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query.

		:waitFor( function()

			return queryServer:queryB {
				sql = [[SELECT * FROM coreInfo WHERE aspectRatio > ? ORDER BY docId]],
				1
			}

		end )
		:continueWhenValueMatches(
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

-- @mark queryP

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryP_empty( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that map server thinks both tables are
		-- not up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query.

		:waitFor( function()

			return queryServer:queryP {
			    sql = [[SELECT * FROM coreInfo ORDER BY docId]]
		    }

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryP_twoRows( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that map server thinks both tables are
		-- not up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query.

		:waitFor( function()

			return queryServer:queryP {
			    sql = [[SELECT * FROM coreInfo ORDER BY docId]]
			}

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
				{
					'd1ab970d2a2a4463982ee5530fc600ec',
					3,
					2/3,
					'2012-08-17',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryP_queryParams( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
			}

		end )

		-- Verify that coreInfo table is empty.

		:waitFor( function()

			return indexStoreDb:gatherArrayOfArray
				[[SELECT * FROM coreInfo]]

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Verify that map server thinks both tables are
		-- not up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query.

		:waitFor( function()

			return queryServer:queryP {
			    sql = [[SELECT * FROM coreInfo WHERE aspectRatio > ? ORDER BY docId]],
			    1
			}

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					2,
					1.5,
					'2012-08-15',
				},
			} )

		-- Verify that map server thinks coreInfo table is
		-- now up to date, but keywords table is not.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryP_join( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	local coreInfoUpToDateB = mapServer:isTableUpToDateB( 'coreInfo' )
	local keywordsUpToDateB = mapServer:isTableUpToDateB( 'keywords' )

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Start adding data to doc store.

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = '651d079c86bd4514aa0ea10874515af2',
				_slot = 'main',
				width = 30,
				height = 20,
				captureDate = '2012-08-15',
				keywords = { foo = true, bar = true },
			}

		end )

		:waitFor( function()

			return docStore:applyOverlayP {
				_id = 'd1ab970d2a2a4463982ee5530fc600ec',
				_slot = 'main',
				width = 20,
				height = 30,
				captureDate = '2012-08-17',
				keywords = { blue = true, xyz = true },
			}

		end )

		-- Verify that map server thinks both tables are
		-- not up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( false )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( false )

		-- Run a query that joins two tables.

		:waitFor( function()

			return queryServer:queryP {
			    sql = [[SELECT coreInfo.docId, aspectRatio, keyword FROM coreInfo
					INNER JOIN keywords
					WHERE coreInfo.docId == keywords.docId
					AND keyword == ?
					ORDER BY coreInfo.docId]],
				'foo'
			}

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					'651d079c86bd4514aa0ea10874515af2',
					1.5,
					'foo',
				},
			} )

		-- Verify that map server thinks both tables are
		-- now up to date.

		:waitFor( coreInfoUpToDateB )
		:continueWhenValueMatches( true )

		:waitFor( keywordsUpToDateB )
		:continueWhenValueMatches( true )

end

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryP_noRelevantTables( test )

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a bogus query.

		:waitFor( function()

			return KSTestUtils.testFails( function()
				return queryServer:queryP {
				    sql = [[SELECT * FROM noSuchTable ORDER BY docId]]
				}
			end )

		end )
		:continueWhenValueMatches( 'data', true )

end

--------------------------------------------------------------------------------

-- @mark :queryBStats

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryBStats_queryP( test )

	-- Verify that queryB stats doesn't include any hits from :queryP calls.

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Run a query that joins two tables.

		:waitFor( function()

			return queryServer:queryP {
			    sql = [[SELECT coreInfo.docId, aspectRatio, keyword FROM coreInfo
					INNER JOIN keywords
					WHERE coreInfo.docId == keywords.docId
					AND keyword == ?
					ORDER BY coreInfo.docId]],
				'foo'
			}

		end )

		-- Check stats.

		:waitFor( function()

			return queryServer:queryBStats()

		end )
		:continueWhenValueMatches( 'data',
			{
				nQueryBehaviorsCreated = 0
			} )

end

--------------------------------------------------------------------------------

local queryB2

--------------------------------------------------------------------------------

function WFIndexStoreQueryServer_WFDocStore_Test.testQueryBStats_twoQueries( test )

	-- Verify that queryB stats includes report from two queryB calls.

	createIndexStoreDbAndMapServer_coreInfoAndKeywordsExample()

	queryB = queryServer:queryB { sql = [[SELECT * FROM coreInfo ORDER BY docId]] }

	queryB2 = queryServer:queryB {
		sql = "SELECT coreInfo.docId, aspectRatio, keyword FROM coreInfo "
			.. "INNER JOIN keywords "
			.. "WHERE coreInfo.docId == keywords.docId AND keyword == ? "
			.. "ORDER BY coreInfo.docId",
		'foo',
		threadName = "testingThread.lua",
		traceback = 'stack traceback:\n\t[string "WFIndexStoreQueryServer_WFDocStore_Test.lua"]:1934: in function <[string "WFIndexStoreQueryServer_WFDocStore_Test.lua"]:1919>\n\t[C]: ?\n\t[C]: ?',
			-- exact contents don't matter, but need to match results in test data below
	}

	WFAsyncTestCase( test )

		-- Spin up the map server.

		:waitFor( function()

			return mapServer:isReadyP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Check stats.

		:waitFor( function()

			return queryServer:queryBStats()

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					nUpdatesSent = 1,
					relevantTables = {
						"coreInfo",
					},
					sql = "SELECT * FROM coreInfo ORDER BY docId",
				},
				{
					args = {
						"foo",
					},
					nUpdatesSent = 1,
					relevantTables = {
						"coreInfo",
						"keywords",
					},
					sql = "SELECT coreInfo.docId, aspectRatio, keyword FROM coreInfo INNER JOIN keywords WHERE coreInfo.docId == keywords.docId AND keyword == ? ORDER BY coreInfo.docId",
					threadName = "testingThread.lua",
					traceback = {
						"\t[string \"WFIndexStoreQueryServer_WFDocStore_Test.lua\"]:1934: in function <[string \"WFIndexStoreQueryServer_WFDocStore_Test.lua\"]:1919>",
						"\t[C]: ?",
						"\t[C]: ?",
					},
				},
				nQueryBehaviorsCreated = 2,
			} )

end

--------------------------------------------------------------------------------

return WFIndexStoreQueryServer_WFDocStore_Test
