--[[----------------------------------------------------------------------------

WFMediaCore.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.rxlua
local WFRxNativeObservable = require 'WFRxNativeObservable'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.mediacore
local WFMediaCoreImp = require 'WFMediaCoreImp'

--------------------------------------------------------------------------------

local logger = WFLogger 'WFMediaCore'
	--:setLevel( 'DEBUG' )
	
--------------------------------------------------------------------------------
local WFMediaCore = {}

--------------------------------------------------------------------------------
local _sessionInitDisposable


local _sessionInitStatus = Rx.ReplaySubject( 1 )

local _supportedFormatsSubject = Rx.ReplaySubject( 1 )

--------------------------------------------------------------------------------

local function getMediacoreExecutableBasePath()
	-- TODO [ ashj: 2016-03-29 ] If run as daemon, perhaps we can 
	-- reuse the dlms app which is packaged in the client. Licensing
	-- information too has to come from the client. During catalog
	-- creation or so, we should be able to accept certain parameters
	-- and store it as client info. And use that to fetch the path.
	-- What should be done if multiple clients are connected??

	-- SPAM [ ashj: 2016-05-25 ] The paths are almost hardcoded here.
	-- Maybe we can make it configurable or can fetch it from the
	-- client. If we rely totally on the client, to provide the
	-- path, then it will not work, for the situations where server
	-- is running independently. In that case, adding a config file
	-- to the build might work. But that again wont solve the
	-- case where there are multiple clients connected.
	-- The multiple client scenario needs to be thought through
	-- a little bit more.

	local function getValidBinaryPath( appContentsPath )
		local appFolderPath = KSPathUtils.child( appContentsPath, "Helpers" )
		local dlmsFolderPath = KSPathUtils.child( appFolderPath, "DynamicLinkMediaServer" )
		
		local dlmFolderPath = WFPlatformInfo.isMac and KSPathUtils.child( appFolderPath, "DynamicLinkSupport" )
			or  KSPathUtils.child( appFolderPath, "DynamicLinkMediaServer" )

		if WFFileUtils.exists( dlmsFolderPath ) == 'directory' and
			WFFileUtils.exists( dlmsFolderPath ) == 'directory' then

			return true, dlmsFolderPath, dlmFolderPath
		end

		return false
	end

	-- First check in app's own Helpers path
	local appFolderPath = KSPathUtils.parent( KSPathUtils.getStandardFilePath( "Resources" ) )
	local isValid, dlmsFolderPath, dlmFolderPath = getValidBinaryPath( appFolderPath )

	if isValid then
		return isValid, dlmsFolderPath, dlmFolderPath
	end

	-- In windows, the Helpers path is the only valid path
	if WFPlatformInfo.isWin then
		logger:warn( "DLMS executables not found" )
		return false
	end

	-- Maybe the executables are in the enclosing app; search
	-- there as well
	appFolderPath = KSPathUtils.parent( KSPathUtils.getStandardFilePath( "Resources" ) ) -- Contents
	appFolderPath = KSPathUtils.parent( appFolderPath ) -- client app
	appFolderPath = KSPathUtils.parent( appFolderPath ) -- MacOS
	appFolderPath = KSPathUtils.parent( appFolderPath ) -- Contents

	return getValidBinaryPath( appFolderPath )

end

--------------------------------------------------------------------------------
local function sessionReadyCallback( statusInfo )

	if not statusInfo then
		logger:warnf( "MediaCore statusInfo is nil." )
		return
	end

	if statusInfo.message == "status" then
		if statusInfo.status == "sessionReady" then
		
			logger:debug( "MediaCore session has been established successfully" )
			_sessionInitStatus:onNext( "success" )
		
		else
			logger:warnf( "MediaCore session init failed. Status:%s", tostring( statusInfo.status) )
			_sessionInitStatus:onNext( "failed" )
		end
		
	elseif statusInfo.message == "supported_formats" then
		logger:tracef( "MediaCore::SupportedFormats:: %s", tostring( statusInfo.supported_formats ) )
		_supportedFormatsSubject:onNext( statusInfo.supported_formats )
		
	elseif statusInfo.message == "log" then
		logger:debugf( "MediaCore::log:: %s", tostring( statusInfo.log ) )
	elseif statusInfo.message == "unknown" then
		logger:warnf( "MediaCore:: Unknown message: %s", tostring( statusInfo.code ) )
	end
end 
--------------------------------------------------------------------------------

local function newSession()
	
	if not _sessionInitDisposable then

		local isValid, dlmsFolderPath, dlmFolderPath = getMediacoreExecutableBasePath()

		if not isValid then
			-- Couldnt find the mediacore executables.
			-- Init will still succeed, if the executables
			-- are somehow running. Hence continuing with
			-- the init process after logging the error.
			logger:warnf( "Mediacore executables not found." )
		end

		if WFPlatformInfo.isMac then
			dlmsFolderPath = KSPathUtils.child( dlmsFolderPath, "application" )

			dlmFolderPath = KSPathUtils.child( dlmFolderPath, "support" )
			dlmFolderPath = KSPathUtils.child( dlmFolderPath, "common" )
			dlmFolderPath = KSPathUtils.child( dlmFolderPath, "dynamiclink" )

		else

			dlmsFolderPath = KSPathUtils.child( dlmsFolderPath, "dynamiclinkmediaserver" )
			dlmsFolderPath = KSPathUtils.child( dlmsFolderPath, "1.0" )
			
			dlmFolderPath = KSPathUtils.child( dlmFolderPath, "dynamiclinkmediaserver" )
			dlmFolderPath = KSPathUtils.child( dlmFolderPath, "1.0" )
			dlmFolderPath = KSPathUtils.child( dlmFolderPath, "32" )

		end
		
		local appName = WFPlatformInfo.getAppName and WFPlatformInfo.getAppName() or 
						WFPlatformInfo.getAppIdentifier and WFPlatformInfo.getAppIdentifier()
						or "WF"
		local version = WFPlatformInfo.getAppVersion and  WFPlatformInfo.getAppVersion() or "1"
		local buildNumber = WFPlatformInfo.getBuild and WFPlatformInfo.getBuild() or "1"

		_sessionInitDisposable = 
			WFRxNativeObservable( WFMediaCoreImp.newSession, dlmsFolderPath, 
				dlmFolderPath, appName, version, buildNumber )
				:subscribe( sessionReadyCallback )
	end
end

--------------------------------------------------------------------------------
function WFMediaCore.waitForSession()
	-- Initiate the new session request and wait for the session status
	newSession()
	
	return _sessionInitStatus
		:take( 1 )
		:map( function( status )
			if status == "success" then
				return "success"
			else
				error( "MediaCore session init failed" )
			end
		end)
end

--------------------------------------------------------------------------------

return WFMediaCore
