--[[----------------------------------------------------------------------------

WFMediaCoreContext.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.rxlua
local WFRxNativeObservable = require 'WFRxNativeObservable'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.mediacore
local WFMediaCore = require 'WFMediaCore'
local WFMediaCoreContextImp = require 'WFMediaCoreContextImp'
local WFRatioTime = require 'WFRatioTime'


--------------------------------------------------------------------------------

assert( WFPlatformInfo.hasMediaCore, "MediaCore integration not enabled" )

local check = assert( WFArguments.check )

local logger = WFLogger 'WFMediaCoreContext'
	--:setLevel( 'DEBUG' )

--------------------------------------------------------------------------------

local _frameTimeForRendering = WFRatioTime.makeRatioTime( 0, 1 )
--------------------------------------------------------------------------------

local function assertIfClosed( self )

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		error( "WFMediaCoreContext can not be used after it has been closed", 3 )
	end

end

--------------------------------------------------------------------------------

local function waitForClipInit( self )

	return WFMediaCore.waitForSession()
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFMediaCoreContextImp.initClip, self._imp )
		end)
		:map( function( clipInitStatusInfo )

			-- Log progress information
			if clipInitStatusInfo.message == "progress" then
				if clipInitStatusInfo.progress then
					logger:infof( "ClipInit progress: %s", tostring( clipInitStatusInfo.progress ) )
				else
					logger:infof( "ClipInit progress: Failure" )
				end
				return
			end

			-- Final result
			if clipInitStatusInfo.success then
				return true
			end
			
			if clipInitStatusInfo.offline then
				logger:errorf( "MediaCore: setClipPath returned offline: %s", tostring( clipInitStatusInfo.offline ))
				error( "MediaCore is offline" )
			end
			
			logger:errorf( "MediaCore: setClipPath returned unknown error: %s", tostring( clipInitStatusInfo.errorMessage ) )
			error( "MediaCore Unknown error" )
		end)

end

--------------------------------------------------------------------------------

local function combineInfoAndMetadata( info, xmpResult )

	if xmpResult and xmpResult.success then
		info.metadata = xmpResult.xmp
	end
	
	return info

end

--------------------------------------------------------------------------------

local function getDimensionsToRenderFromLongSide( longSide, actualWidth, actualHeight )

	local actualLongSide = math.max( actualWidth, actualHeight )
	local widthToRender, heightToRender
	if actualLongSide == actualWidth then
		widthToRender = math.min( longSide, actualLongSide )
		heightToRender = math.floor( ( actualHeight * widthToRender ) / actualWidth )
	else
		heightToRender = math.min( longSide, actualLongSide )
		widthToRender = math.floor( ( actualWidth * heightToRender ) / actualHeight )
	end

	return widthToRender, heightToRender
end

--============================================================================--

-- @mark WFMediaCoreContext class

--[[!
	\brief WFMediaCoreContext provides single entry point to all
	media core operations. 

	Mediacore initialization and usage of session and clip,
	caching of clip etc., are handled by this class, so that
	clients don't need to worry about all the intricacies. 
]]--

local WFMediaCoreContext = {

	_imp = false,
		-- ( WFMediaCoreContextImp )

	_filePath = false,
		-- (string) original file path

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new WFMediaCoreContext object for a single video file.

	\param filePath (string) path to video file

	
]]--

function WFMediaCoreContext:__init( filePath )

	check( 'WFMediaCoreContext', '+string' )

	self._imp = WFMediaCoreContextImp.make( filePath )
	
	self._filePath = filePath

end


--------------------------------------------------------------------------------

function WFMediaCoreContext:__tostring()

	return string.format( "WFMediaCoreContext( %s )", self._filePath )

end

--------------------------------------------------------------------------------
--[[!
	\brief Read metadata from this file.

]]--

function WFMediaCoreContext:readMetadata()

	check( 'WFMediaCoreContext' )

	assertIfClosed( self )
	
	return waitForClipInit( self )
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFMediaCoreContextImp.getInfo, self._imp )
				:combineLatest( combineInfoAndMetadata, 
					WFRxNativeObservable( WFMediaCoreContextImp.getXMP, self._imp ) )
		end )
end

--------------------------------------------------------------------------------
--[[!
	\brief Get the frame corresponding to the frame time

]]--

function WFMediaCoreContext:getFrame( frameTime, width, height, quality )

	check( 'WFMediaCoreContext' )

	assertIfClosed( self )
	
	return waitForClipInit( self )
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFMediaCoreContextImp.getFrame, self._imp,
				frameTime, width, height, quality )
		end)

end

--------------------------------------------------------------------------------
--[[!
	\brief Generate a preview jpeg for the video

]]--

function WFMediaCoreContext:render( longSide, quality )

	check( 'WFMediaCoreContext' )

	assertIfClosed( self )
	
	return waitForClipInit( self )
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFMediaCoreContextImp.getInfo, self._imp )
				:map( function( info )
					return info.width, info.height
				end)
				:flatMapLatest( function( width, height )

					local widthToRender, heightToRender = 
						getDimensionsToRenderFromLongSide( longSide, width, height )

					return WFRxNativeObservable( WFMediaCoreContextImp.getFrame, self._imp,
							_frameTimeForRendering, widthToRender, heightToRender, quality )
				end)

		end)

end

--------------------------------------------------------------------------------
--[[!
	\brief Closes the mediacore context and the 
	associated clip. 
	Make sure to call this function once you are
	done with the context. If any of the clip
	exists when the app is left open, app might
	crash inside mediacoreif while shutting down.
	
	TOFIX[ ashj: 2016-04-04 ] Can we do ref counting
	to close all the clips before the session
	is closed?

]]--

function WFMediaCoreContext:close()

	check( 'WFMediaCoreContext' )

	if self._imp then
		self._imp:close()
	end

	self._imp = nil

end

--------------------------------------------------------------------------------

return Rx.Class 'WFMediaCoreContext'( WFMediaCoreContext )
