--[[----------------------------------------------------------------------------

WFOzAlbumAssetAutoDeleter.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.lua.threading
local WFTimer = require 'WFTimer'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.sync_engine
local WFOzCatalogSpec = require 'WFOzCatalogSpec'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFOzAlbumAssetAutoDeleter class

--------------------------------------------------------------------------------
--[[!
	\class WFOzAlbumAssetAutoDeleter
	
	\brief An album_asset uploader is responsible for bulk uploading album_asset
	documents to Oz. We do this outside of the normal WFOzSyncPeer path because
	it is much more efficient this way.
	
	It can do this either automatically or manually, as requested via APIs.
]]--

local albumAssetAutoDeleterMethods = {

	_catalogSpec = false,
		-- (WFOzCatalogSpec) Oz catalog specification
	
	_docStore = false,
		-- (WFDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)

	_checkedChangesTrackerB = false,
		-- (B< number >) tracks the change counter values that have been checked
		-- in the doc store
	
	_checkedChangesIsBehindB = false,
		-- (B< boolean >) true if _checkedChangesTrackerB is behind the doc store

	_autoDeleteHandlerAction = false,
		-- (promise action) if we have an auto-delete checker action in flight, it's here

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local maybeStartAutoDelete -- forward declaration

--------------------------------------------------------------------------------
--[[!
	\brief Constructs an Oz album_asset uploader object.
	
	\param args (table) table containing the following arguments, all of which
		are required

		<ul>
			<li><code>catalogSpec</code>: (\c WFOzCatalogSpec) specifies the target catalog</li>
			<li><code>docStore</code>: (\c WFDocStore) local doc store for catalog</li>
			<li><code>indexStore</code>: (\c WFIndexStore) index store for catalog</li>
		</ul>
]]--

function albumAssetAutoDeleterMethods:__init( args )

	check( 'WFOzAlbumAssetAutoDeleter', 'table' )
	
	self._catalogSpec = assert( args.catalogSpec )
	assert( WFOzCatalogSpec:isClassOf( self._catalogSpec ) )
	
	self._docStore = assert( args.docStore )
	assert( WFDocStore:isClassOf( self._docStore ) )
	self._docStore:addCloseTarget( self )

	self._indexStore = assert( args.indexStore )
	assert( WFIndexStore:isClassOf( self._indexStore ) )
	self._indexStore:addCloseTarget( self )

	self:setOwned {
		_checkedChangesTrackerB = self._docStore:fieldValueB( '_checkedChangesTracker',
															  'continueFrom',
															  'WFOzAlbumAssetAutoDeleter' )
	}
	
	self:setOwned {
		_checkedChangesIsBehindB = KSReactive.callB(
			function( checkedChanges, changeCounter )
				
				if WFUnknown.isUnknown( checkedChanges )
					or WFUnknown.isUnknown( changeCounter )
				then
					return false
				end
				
				if type( checkedChanges ) ~= 'number' then
					return true
				else
					return checkedChanges < changeCounter
				end
				
			end,
			self._checkedChangesTrackerB,
			self._docStore:docChangeCounterB() )
	}
	
	self._checkedChangesIsBehindB:onDataB( function( isBehindNow )
		maybeStartAutoDelete( self )
	end )

end

--------------------------------------------------------------------------------

local function deleteRelatedDocumentsP( self, docId )

	return self._indexStore:queryP(
		[[SELECT docId FROM albumAssets WHERE album == ? OR asset == ?]],
		docId, docId )

		:mapDataP( function( _, docIdRows )
		
			if #docIdRows < 1 then
				return 'data', true
			end

			return KSPromises.allP(
				function( docIdRow )
				
					local docId = docIdRow[ 1 ]
					
					if docId then
					
						return self._docStore:applyOverlayP
							{
								_id = docId,
								_slot = 'main',
								_deleted = true,
								_syncStatus = {
									mute = 'related',
								},
							}
						
					end
				
				end, docIdRows )
		
		end )

end

--------------------------------------------------------------------------------

maybeStartAutoDelete = function( self )

	if self._autoDeleteHandlerAction then return end
	
	if not self._checkedChangesIsBehindB:valueNow() then return end
	
	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then return end
	
	self._autoDeleteHandlerAction = self._docStore:changedDocsP {
		fromChangeCounter = self._checkedChangesTrackerB:valueNow() or 0,
		slot = 'main',
	}
		:mapDataP( function( _, changedDocs )
			
			local relevantDeletedDocIds = {}
			local nRelevantDeletedDocIds = 0
			
			for i = 1, #changedDocs do
			
				local changedDoc = changedDocs[ i ]
				if changedDoc._deleted then
				
					local isMuted = KSTableUtils.derefOpt( changedDoc,
														   '_syncStatus',
														   'mute' )
					
					if isMuted ~= 'related' then
						nRelevantDeletedDocIds = nRelevantDeletedDocIds + 1
						relevantDeletedDocIds[ nRelevantDeletedDocIds ] = changedDoc._id
					end
				
				end
			
			end
			
			if nRelevantDeletedDocIds > 0 then

				return KSPromises.allP(
					function( id )
						return deleteRelatedDocumentsP( self, id )
					end,
					relevantDeletedDocIds, 5 )
					
					:mapP( function()
					
						-- Don't cargo-cult this code unless you are aware
						-- of the following: This intentionally ignores the
						-- error status from KSPromises.allP. In this case,
						-- I think it's OK to do that because we'll fall back
						-- to the old case of having dangling album_asset
						-- records that might get uploaded, resulting in
						-- 404 error noise. We shipped with that in Lightroom
						-- 5.4, so I think we can let it slide if this code
						-- fails.
					
						return 'data', changedDocs.continueFrom
					
					end )
			
			else
	
				return 'data', changedDocs.continueFrom
			
			end

		end )

		-- Update checkpoint.
		
		:mapDataP( function( _, continueFrom )

			return self._docStore:setDocValueP
				{
					_id = '_checkedChangesTracker',
					_slot = 'WFOzAlbumAssetAutoDeleter',
					continueFrom = continueFrom,
				}
																  		
		end )

		-- Wait a while before checking again.

		:mapP( function()
		
			return WFTimer.afterDelayP( 15 )
		
		end )
		
		:observeP( function()

			self._autoDeleteHandlerAction = false
			maybeStartAutoDelete( self )
		
		end )

		:whenResolvedDo( function() end ) -- no-op just so we can attach a lifetime
		:attachTo( self )
		:suppressCloseLogging()
			-- Since this runs on a timer, it might fire at any time.
			-- Getting closed mid-cycle is to be expected.

end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzAlbumAssetAutoDeleter'( albumAssetAutoDeleterMethods )
