--[[----------------------------------------------------------------------------

WFOzAlbumAssetAutoDeleterTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.reactive
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.sync_engine.metadata_sync
local WFOzAlbumAssetAutoDeleter = require 'WFOzAlbumAssetAutoDeleter'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzCatalogTestSession = require 'WFOzCatalogTestSession'


--============================================================================--

-- @mark WFOzAlbumAssetAutoDeleterTest

local WFOzAlbumAssetAutoDeleterTest = {}

--------------------------------------------------------------------------------

local TEST_WAIT_TIME = 240
local catalogTestSession, indexStore, wfOzSyncAgent, albumAssetAutoDeleter
local fakePhotoIdsByCaptureDate

--------------------------------------------------------------------------------

function WFOzAlbumAssetAutoDeleterTest.setUp()

	catalogTestSession = WFOzCatalogTestSession( 'WFOzAlbumAssetAutoDeleterTest' )

	indexStore = nil
	wfOzSyncAgent = nil
	albumAssetAutoDeleter = nil

end

--------------------------------------------------------------------------------

function WFOzAlbumAssetAutoDeleterTest.tearDown()

	if catalogTestSession then
		catalogTestSession:releaseTestUserP()
		catalogTestSession:close()
		catalogTestSession = nil
	end

end

--------------------------------------------------------------------------------

local function albumAssetAutoDeleterP()

	indexStore = WFIndexStore( catalogTestSession:docStore(),
							   'AlbumAssetUploaderTestIndexDefinitions.lua' )

	catalogTestSession:addCloseTarget( indexStore )

	return catalogTestSession:ozSyncAgentWithMetadataSyncCompletedP(
									{ suppressAlbumAssetUploads = true } )
		
		:mapDataP( function( _, syncAgent )
		
			return catalogTestSession:catalogSpecP()
		
		end )
		
		:mapDataP( function( _, catalogSpec )
		
			albumAssetAutoDeleter = WFOzAlbumAssetAutoDeleter
				{
					catalogSpec = catalogSpec,
					docStore = catalogTestSession:docStore(),
					indexStore = indexStore,
				}
			
			catalogTestSession:addCloseTarget( albumAssetAutoDeleter )
			
			fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()
			
			return 'data', albumAssetAutoDeleter
		
		end )

end

----------------------------------------------------------------------------

local function hasServerCreatedDate( docValue )

	return not not KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'created' )

end

----------------------------------------------------------------------------

local function hasBeenDeletedWithMuting( docValue )

	local isDeleted = KSTableUtils.derefOpt( docValue, '_deleted' )
	
	local isMuted = KSTableUtils.derefOpt( docValue, '_syncStatus', 'mute' ) == 'related'
	
	return not not ( isDeleted and isMuted )

end

----------------------------------------------------------------------------

function WFOzAlbumAssetAutoDeleterTest.testDeleteFromServer_album_testNightly( test )

	local docStore = catalogTestSession:docStore()

	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId1 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId2 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId3 = KSIdGenerator.uuidInLowercaseHex()

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the album asset uploader. Now the albumAssetAutoDeleter file variable
		-- will be defined.
		
		:waitFor( function()

			return albumAssetAutoDeleterP()

		end )

		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Create album and put all three assets in it.
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					type = 'album',
					name = 'test album',
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId1,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId2,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId3,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
					},
				}
			
		end )
		
		-- Now make sure album and all three album-asset records get uploaded.
		
		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId2 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId3 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		-- Now delete the album from server.
		
		:waitFor( function()
		
			local href
		
			-- Using docStore:docValueP would seem like the obvious thing to do
			-- here, but we might not have _serverProvidedMetadata filled in
			-- when we first get here.

			return KSReactivePromiseBridge.closingBehaviorToPromise(
				docStore:docValueB( albumId )
					:closingMapB( function( value )
					
						return KSTableUtils.derefOpt( value,
													  '_serverProvidedMetadata',
													  'links',
													  'self',
													  'href' )
	
					end ),
				function( value )
					return type( value ) == 'string'
				end )
			
				:mapDataP( function( _, _href )
					href = _href
					return catalogTestSession:catalogSpecP()
				end )
				
				:mapDataP( function( _, catalogSpec )
					return catalogSpec:getP( href, { requestMethod = 'DELETE' } )
				end )

		end )

		-- Wait for deletion to propagate back to client.
		
		:waitFor( function()
		
			return docStore:fieldValueB( albumId, '_deleted' )
			
		end )
		:continueWhenValueMatches( true )

		-- Wait for deletion (with muting) to show up in all three album_asset documents.
		
		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1 )
				:closingMapB( hasBeenDeletedWithMuting )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId2 )
				:closingMapB( hasBeenDeletedWithMuting )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId3 )
				:closingMapB( hasBeenDeletedWithMuting )
		
		end )
		:continueWhenValueMatches( true )
		
		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

----------------------------------------------------------------------------

function WFOzAlbumAssetAutoDeleterTest.testDeleteFromServer_asset_testNightly( test )

	local docStore = catalogTestSession:docStore()

	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId1 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId2 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId3 = KSIdGenerator.uuidInLowercaseHex()

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the album asset uploader. Now the albumAssetAutoDeleter file variable
		-- will be defined.
		
		:waitFor( function()

			return albumAssetAutoDeleterP()

		end )

		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Create album and put all three assets in it.
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					type = 'album',
					name = 'test album',
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId1,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId2,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId3,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
					},
				}
			
		end )
		
		-- Now make sure album and all three album-asset records get uploaded.
		
		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1 )
				:closingMapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId2 )
				:closingMapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId3 )
				:closingMapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		-- Now delete one of the assets from server.
		
		:waitFor( function()
		
			local href
		
			-- Using docStore:docValueP would seem like the obvious thing to do
			-- here, but we might not have _serverProvidedMetadata filled in
			-- when we first get here.

			return KSReactivePromiseBridge.closingBehaviorToPromise(
				docStore:docValueB( fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
					:closingMapB( function( value )
					
						return KSTableUtils.derefOpt( value,
													  '_serverProvidedMetadata',
													  'links',
													  'self',
													  'href' )
	
					end ),
				function( value )
					return type( value ) == 'string'
				end )
			
				:mapDataP( function( _, _href )
					href = _href
					return catalogTestSession:catalogSpecP()
				end )
				
				:mapDataP( function( _, catalogSpec )
					return catalogSpec:getP( href, { requestMethod = 'DELETE' } )
				end )

		end )

		-- Wait for deletion to propagate back to client.
		
		:waitFor( function()
		
			return docStore:fieldValueB( fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
										 '_deleted' )
			
		end )
		:continueWhenValueMatches( true )

		-- Wait for deletion (with muting) to show up in just one of the three album_asset documents.
		
		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1 )
				:closingMapB( hasBeenDeletedWithMuting )
		
		end )
		:continueWhenValueMatches( true )

		-- Wait a little while then make sure the other album_asset documents didn't
		-- get affected.

		:waitFor( 5 )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId2 )
				:closingMapB( hasBeenDeletedWithMuting )
		
		end )
		:continueWhenValueMatches( false )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId3 )
				:closingMapB( hasBeenDeletedWithMuting )
		
		end )
		:continueWhenValueMatches( false )
		
		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

local WFPlatformInfo = require 'WFPlatformInfo'

if WFPlatformInfo.isWindows then
	return {}
else
	return WFOzAlbumAssetAutoDeleterTest
end
