--[[----------------------------------------------------------------------------

WFOzAlbumAssetIndex.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'
local KSStringUtils = require 'KSStringUtils'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

local WFOzAlbumAssetIndex = {}

--------------------------------------------------------------------------------
-- @mark albums
--------------------------------------------------------------------------------

local function album_table()

	return {
	
		sql = [[CREATE TABLE albums( id, name, nameLC, updateTimeStamp, localTimeStamp )]],

		hasOneRowPerDocument = true,
	
		map = function( docValue, emit )
	
			if docValue and ( docValue._docType == 'album' or docValue.type == 'album' ) and not docValue._deleted then
			
				local name = docValue.name
				
				local id = docValue._id or docValue._docId
				local updateTimeStamp = docValue.updated
				local localTimeStamp = docValue._localOnly and docValue._localOnly.updateTimeStamp
				
				if id and name then
					emit( id, name, KSStringUtils.lower( name ), updateTimeStamp, localTimeStamp )
				end
			
			end
	
		end,
	
	}

end

--------------------------------------------------------------------------------

local function album_name_index()

	return [[CREATE INDEX albums_name ON albums( nameLC )]]

end

--------------------------------------------------------------------------------

local function album_asset_table()

	return {

		sql = [[CREATE TABLE album_asset( assetId, userId, albumId, updateTimeStamp, sortOrder, isCover )]],

		dontUpdateMapFunctionHash = true,
		
		map = function( docValue, emit )
	
			if docValue and not docValue._deleted then
			
				if docValue._docType == 'album_asset' or docValue.type == 'album_asset' then

					local albumId = docValue.album and docValue.album._id
					local assetId = docValue.asset and docValue.asset._id
					local updateTimeStamp = docValue.updated
					local sortOrder = docValue.order
					
					if not updateTimeStamp then
						updateTimeStamp = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata',
																	'updated' )
					end
						
					local isCover = docValue.cover or false

					local userId = docValue.added_by
					if userId and string.len( userId ) > 32 then
						userId = userId:sub( -32 )
					end

					if albumId and assetId then
						emit( assetId, userId, albumId, updateTimeStamp, sortOrder, isCover )
					end
					
				end
				
				-- Support for virtual albums of favorited assets.
	
				if ( docValue._docType == 'asset' or docValue.type == 'asset' )
					and not docValue._deleted
					and docValue.favorites and next( docValue.favorites ) ~= nil
				then
				
					-- pony has been unfavoriting by setting the favorites key value pair
					-- to have a value of false instead of removing the key so make
					-- sure we have a non false value

					local emitDocFavorite = true
					for k, v in pairs( docValue.favorites ) do
						if v then
							if emitDocFavorite then
								emit( docValue._docId, nil, 'favorites', nil )
								emitDocFavorite = false
							end

							emit( docValue._docId, nil, 'favorites' .. k, nil )
						end
					end
					
				end

			end

		end,

	}
end

--------------------------------------------------------------------------------

local function album_asset_index()

	return [[CREATE INDEX album_asset_index ON album_asset( albumId, assetId )]]
end

--------------------------------------------------------------------------------

local function album_asset_activity_date_userId_index()

	return [[CREATE INDEX album_asset_activity_date_userId ON album_asset( updateTimeStamp DESC, userId )]]
end

--------------------------------------------------------------------------------

function WFOzAlbumAssetIndex.filterIndexDefinitions( tableDefinitions, indexDefinitions )

	tableDefinitions.album = album_table()
	tableDefinitions.album_asset = album_asset_table()
	
	indexDefinitions.album_name_index = album_name_index()
	indexDefinitions.album_asset_index = album_asset_index()
		
	indexDefinitions.album_asset_activity_date_userId = album_asset_activity_date_userId_index()
		
end

--------------------------------------------------------------------------------

function WFOzAlbumAssetIndex.generateIndexSchema( indexSchema )

	check( 'WFIndexStoreSchema' )
	
	local tableDefinitions, indexDefinitions = {}, {}
	WFOzAlbumAssetIndex.filterIndexDefinitions( tableDefinitions, indexDefinitions )
	
	for _, tableDefn in pairs( tableDefinitions ) do
		indexSchema:addTable( tableDefn )
	end
	
	for _, index in pairs( indexDefinitions ) do
		indexSchema:addIndex( index )
	end

end

--------------------------------------------------------------------------------

return WFOzAlbumAssetIndex
