--[[----------------------------------------------------------------------------

WFOzAlbumAssetUploadUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger  = require 'WFLogger'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFDigest = require 'WFDigest'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFOzAlbumAssetUploaderIndex'
--	:setLevel 'debug'


--============================================================================--

local WFOzAlbumAssetUploadUtils = {}

--------------------------------------------------------------------------------

local function hasBeenUploaded( docValue )

	if KSTableUtils.derefOpt( docValue,
							  '_syncStatus',
							  'oz',
							  'timestamp' ) ~= nil
	then
		return true
	end

	if KSTableUtils.derefOpt( docValue,
							  '_syncStatus',
							  'oz_TEST',
							  'timestamp' ) ~= nil
	then
		return true
	end
	
	return false

end

--------------------------------------------------------------------------------

local function uploadHasFailed( docValue )

	if KSTableUtils.derefOpt( docValue,
							  '_syncStatus',
							  'oz',
							  'error',
							  'permanent' )
	then
		return true
	end

	if KSTableUtils.derefOpt( docValue,
							  '_syncStatus',
							  'oz_TEST',
							  'error',
							  'permanent' )
	then
		return true
	end
	
	return false

end

--------------------------------------------------------------------------------

-- TODO [matthewj 2014-12-01 via arahn]: It would be nice if this function could
-- be exported and re-used rather than having two implementations of this string
-- concatenation and MD5 algorithm. (It's also been implemented in WFOzSyncPeer.lua
-- on line 367). I realize it's a pretty simple method, but give its importance
-- because if this algorithm were updated in the future in one place but not the
-- other, 'needsUpdate' could get stuck permanently to TRUE which would introduce
-- a looper.

local function albumAssetPayloadHash( album_asset_id, order, cover )
	
	return WFDigest.MD5.digest( album_asset_id .. tostring( order ) .. tostring( cover ) )

end

--------------------------------------------------------------------------------

local function changedAfterUpgradeFor3954607( docValue )
	return not not KSTableUtils.derefOpt( docValue, 'hack', 'changedFor3954607' )
end

--------------------------------------------------------------------------------

function WFOzAlbumAssetUploadUtils.generateIndexSchema( indexStoreSchema )

	--check( 'WFIndexStoreSchema' )
		-- Ugh. Yet another reason why I hate mocks in unit tests.
		-- Mocks created at OzCatalogTestUtils.buildMocksForTestingCatalogService
		-- mean that the WFIndexStoreSchema mock object doesn't pass WFArguments
		-- smell test so now I have to disable this safety check. Only discovered
		-- this when I made a change in a yet another unrelated part of the app
		-- that introduced a dependency on this file, but only in Lightroom's
		-- test harness environment. (Thanks to differences how we configure
		-- OzCatalogIndexDefinitions via WFCapabilities, naturally.)

	indexStoreSchema:addTable {
	
		sql = [[CREATE TABLE albumAssetUploads( type, album, asset, sortOrder, needsUpdate, isCover )]],
	
		hasOneRowPerDocument = true,

		map = function( docValue, emit )

			logger:debugf( "indexing %s", docValue )

			local docType = docValue.type

			if docType == 'asset' or docType == 'album' then
			
				if hasBeenUploaded( docValue ) then
					logger:debugf( "emit uploaded %s", docType )
					emit( docType, nil, nil, nil)
				end
			
			elseif docType == 'album_asset' then

				local albumId = KSTableUtils.derefOpt( docValue, 'album', '_id' )
				local assetId = KSTableUtils.derefOpt( docValue, 'asset', '_id' )
				local order = KSTableUtils.derefOpt( docValue, 'order' )
				local isCover = KSTableUtils.derefOpt( docValue, 'cover' )
			
				local isAlreadyOnOz = hasBeenUploaded( docValue )
						
				local docHash = albumAssetPayloadHash( docValue._id, order, isCover )
				
				local ozHash = KSTableUtils.derefOpt( docValue,
													'_serverProvidedMetadata',
													'hash' )
								

				if not uploadHasFailed( docValue ) then
					
					-- If the album_asset is already on Oz then we need to check whether
					-- the order or cover state has changed. The hash stashed by WFOzSyncPeer
					-- on receipt of album_asset records is used for this purpose.
					if isAlreadyOnOz and docHash ~= ozHash then
						-- SPAM [matthewj]: Beware! On upgrade from before cover info existed in
						-- album_asset records to after, the hash is not in docs that have been
						-- synced to Oz. This means that the upgrade causes a re-index which
						-- would mark all album_assets as needing updates. However, since most
						-- contained no changes Oz would receive them silently and they wouldn't
						-- appear in the sync feed and so would be stuck needing to be updated.
						-- If there were too many of these, it would prevent other album_asset
						-- records from being sent. If the ozHash is nil only emit if cover or
						-- order information exists as Oz will respond with sync feed docs when
						-- we send these.
						if ozHash ~= nil or ( ( isCover or order )
							and changedAfterUpgradeFor3954607( docValue ) )
								-- 3954607 is related to switching to sending cover info in the album_asset documents
								-- and expecting to have a hash set in _serverProvidedMetadata when we receive these
								-- docs from Oz. On upgrade we won't have these hashes set and so we can't mark it
								-- as an update if we don't know it was changed by Lr6. This might drop a few album_asset
								-- *changes* (but not original docs) on the floor if there are any unsynced changes at
								-- the time of upgrade
							then
							
							
    						emit( 'album_asset', albumId, assetId, order, 1, isCover )
						end
												
					elseif ( not isAlreadyOnOz ) then
						logger:debugf( "emit not uploaded %s, %s, %s",
									   docType, albumId, assetId )
						emit( 'album_asset', albumId, assetId, order, nil, isCover )
					end
					
				end
			
			end

		end,
	
	}
	
	indexStoreSchema:addIndex
		[[CREATE INDEX albumAssetUploads_album ON albumAssetUploads( album, asset )]]

	-- Shame to add a second table that's effectively similar, but ... I think query
	-- performance for the uploader will suffer too much if we don't.

	indexStoreSchema:addTable {
	
		sql = [[CREATE TABLE albumAssets( album, asset )]],

		hasOneRowPerDocument = true,

		map = function( docValue, emit )

			if docValue.type == 'album_asset' then

				local albumId = KSTableUtils.derefOpt( docValue, 'album', '_id' )
				local assetId = KSTableUtils.derefOpt( docValue, 'asset', '_id' )
				emit( albumId, assetId )
			
			end

		end,
	
	}
	
	indexStoreSchema:addIndex
		[[CREATE INDEX albumAssets_album ON albumAssets( album, asset )]]

	indexStoreSchema:addIndex
		[[CREATE INDEX albumAssets_asset ON albumAssets( asset )]]

	-- TODO_SNOWLESS [scouten 2013-08-28]: Figure out how to handle album_asset deletion in bulk.
	-- Tracked in Lightroom Pivotal: https://www.pivotaltracker.com/story/show/60434452
	
end

--------------------------------------------------------------------------------

return WFOzAlbumAssetUploadUtils
