--[[----------------------------------------------------------------------------

WFOzAlbumAssetUploader.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'
local WFQueryDrivenWorkQueueScheduler = require 'WFQueryDrivenWorkQueueScheduler'

	-- WF.oz_client.sync_engine
local WFOzCatalogSpec = require 'WFOzCatalogSpec'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local logger = WFLogger 'WFOzAlbumAssetUploader'
--	:setLevel 'debug'

local derefOpt = KSTableUtils.derefOpt

local decodeOzJson = assert( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty )

--------------------------------------------------------------------------------

local uploadAssetsForAlbumP -- forward declaration


--============================================================================--

-- @mark WFOzAlbumAssetUploader class

--------------------------------------------------------------------------------
--[[!
	\class WFOzAlbumAssetUploader
	
	\brief An album_asset uploader is responsible for bulk uploading album_asset
	documents to Oz. We do this outside of the normal WFOzSyncPeer path because
	it is much more efficient this way.
	
	It can do this either automatically or manually, as requested via APIs.
]]--

local albumAssetUploaderMethods = {

	_catalogSpec = false,
		-- (WFOzCatalogSpec) Oz catalog specification
	
	_docStore = false,
		-- (WFDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)

	_workScheduler = false,
		-- (WFQueryDrivenWorkQueueScheduler)
		
	_recentErrorLog = false,
		-- (table) contains recent upload errors for album assets

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs an Oz album_asset uploader object.
	
	\param args (table) table containing the following arguments, all of which
		are required

		<ul>
			<li><code>catalogSpec</code>: (\c WFOzCatalogSpec) specifies the target catalog</li>
			<li><code>docStore</code>: (\c WFDocStore) local doc store for catalog</li>
			<li><code>indexStore</code>: (\c WFIndexStore) index store for catalog</li>
		</ul>
]]--

function albumAssetUploaderMethods:__init( args )

	check( 'WFOzAlbumAssetUploader', 'table' )
	
	self._catalogSpec = assert( args.catalogSpec )
	assert( WFOzCatalogSpec:isClassOf( self._catalogSpec ) )
	
	self._docStore = assert( args.docStore )
	assert( WFDocStore:isClassOf( self._docStore ) )
	self._docStore:addCloseTarget( self )

	self._indexStore = assert( args.indexStore )
	assert( WFIndexStore:isClassOf( self._indexStore ) )
	self._indexStore:addCloseTarget( self )

	local queryB = self._indexStore:queryB
		[[SELECT DISTINCT album FROM (
			SELECT a1.album, a1.asset, a1.docId
				FROM albumAssetUploads AS a1
					INNER JOIN albumAssetUploads AS a2
					INNER JOIN albumAssetUploads AS a3
						WHERE a1.album == a2.docId AND a1.asset == a3.docId
					ORDER BY a1.album
					LIMIT 200 )
			LIMIT 4]]
			
	-- SPAM [matthewj 2014-01-30]: Modifying the queue parameters here for bug #3701607
	-- For 5.4 this was initially modified to drop the maxConcurrentTasks to 1 and push
	-- the blackout window to 60 (from 20). This helped but didn't fix all the issues due
	-- to a bug in WFQueryDrivenWorkQueueScheduler related to #3716144. With that fixed we
	-- can bump the maxConcurrentTasks back up as well as drop the blackoutWindow a bit (to 40)
	-- I didn't drop it all the way back to 20 because there seem to still be instances with
	-- large collections and slower networks where all of the album_assets for a given bunch
	-- won't get processed in the blackout time and we'll resend some records needlessly. This
	-- should be fully rectified by a true bulk update mechanism in Oz where this uploader can
	-- go away.  Hopefully this comes ~6.0 timeframe.
	self:setOwned {
		_workScheduler = WFQueryDrivenWorkQueueScheduler(
			queryB,
			2, -- maxConcurrentTasks
			40, -- blackoutWindow
			function( albumId )
				return uploadAssetsForAlbumP( self, albumId[ 1 ] )
			end )
	}
	
	self._recentErrorLog = {}
	WFDiagnosticLog.addLogProvider( self )
end

--------------------------------------------------------------------------------

local function recordUploadError( self, errorRow )
	self._recentErrorLog[ #self._recentErrorLog + 1 ] = errorRow
	if #self._recentErrorLog > 20 then
		table.remove( self._recentErrorLog, 1 )
	end
end

--------------------------------------------------------------------------------

uploadAssetsForAlbumP = function( self, albumId )

	return self._indexStore:queryP(
		[[SELECT a1.asset, a1.docId, a1.sortOrder, a1.needsUpdate, a1.isCover FROM albumAssetUploads AS a1
				INNER JOIN albumAssetUploads AS a2
					WHERE a1.album == ? AND a1.asset == a2.docId
				ORDER BY a1.album, a1.asset
				LIMIT 500]], albumId )
		:mapDataP( function( _, assetAndDocIdsAndOrderAndUpdateAndIsCover )
		
			logger:infof( "uploadAssetsForAlbumP computed assetAndDocIdsAndOrder = %s",
						  assetAndDocIdsAndOrderAndUpdateAndIsCover )
										
			local posts = {}
			local puts = {}
			
			for _, item in pairs( assetAndDocIdsAndOrderAndUpdateAndIsCover ) do
				if item[ 4 ]  == 1 then
					-- Update
					table.insert( posts, item )
				else
					-- New
					table.insert( puts, item )
				end
			end
			
			local putsAndPosts = { puts, posts }

			-- Split into groups of 50 for upload.
			
			local groupedAssetAndDocIds = {}
			local nGroups = 0

			for _, putsOrPosts in ipairs( putsAndPosts ) do
						
				for i = 1, #putsOrPosts, 50 do
				
					nGroups = nGroups + 1
	
					local group = {}
					
					for j = 1, 50 do
	
						local assetAndDocIdAndOrderAndIsCover = putsOrPosts[ i + j - 1 ]
						
						if not assetAndDocIdAndOrderAndIsCover then
							break
						end
						
						group.isUpdate = assetAndDocIdAndOrderAndIsCover[ 4 ]
						
						-- SPAM: If we're sending the initial album_asset records up we
						-- don't want to send cover: false because it forces Oz to do extra
						-- work. For updates due to cover changes though we may need to
						-- send cover:false in order to avoid input validation errors for
						-- what looks like an empty update for the previous cover. Therefore
						-- we want to explicitly send a cover: false field even if it causes Oz
						-- to do a little more work than it needs to (it really just needs the
						-- new cover record only but I haven't really found an understandable
						-- to filter these out)
						local coverDefault
						if group.isUpdate then
							coverDefault = false
						end
										
						-- FIXME [matthewj 2014-12-01 via arahn]: The fact that this sometimes
						-- evaluates to false and sometimes to nil is a little non-obvious
						-- (even with the big SPAM above). An explicit if statement with
						-- assetAndDocIdAndOrderAndIsCover[ 5 ] == 1 and an else would be clearer.
						-- See suggested alternate code at
						-- https://git.corp.adobe.com/lua/wf/pull/307#discussion_r109784

						group[ j ] = {
							_id = assetAndDocIdAndOrderAndIsCover[ 1 ],
							album_asset_id = assetAndDocIdAndOrderAndIsCover[ 2 ],
							order = assetAndDocIdAndOrderAndIsCover[ 3 ],
							cover = assetAndDocIdAndOrderAndIsCover[ 5 ] == 1 and true or coverDefault,
						}
	
					end
					
					groupedAssetAndDocIds[ nGroups ] = group
				
				end
				
			end
			
			-- Send off as many requests as necessary to upload these groups.
			
			return KSPromises.mapArrayP(
				function( uploadGroup )
					
					-- Ensured above that all items in a group are either new or updates.
					local requestMethod = uploadGroup.isUpdate == 1 and 'POST' or 'PUT'
				
					-- remove the isUpdate field now that we've checked it since it's not
					-- a valid field in the Oz album asset record
					uploadGroup.isUpdate = nil
					
					local uploadContent = { assets = uploadGroup }
					
					logger:debugf( "Album %s add assets = ", uploadContent )

					local addAssetsToAlbumUrl = string.format( '%s/albums/%s/assets',
															   self._catalogSpec:catalogHref(),
															   albumId )

					local jsonData = KSJsonEncoder.encodeOzUploadJson( uploadContent )
					
					local putArgs = {
						requestHeaders = {
							{ header = 'Content-Type', value = 'application/json' },
							{ header = 'Accept', value = 'application/json' },
						},
						returnHttpErrorsAsData = true,
						requestMethod = requestMethod,
					}
			
					logger:debugf( "Will upload %s\n\nbody = %s\n\nargs = %s",
								   addAssetsToAlbumUrl, jsonData, putArgs  )

					return self._catalogSpec:putP( addAssetsToAlbumUrl, jsonData, putArgs )
					
						-- TODO_SNOWLESS [scouten 2013-12-18]: Add :observeP call to record
						-- Oz response status. Ideally this should tie into the common Oz error
						-- reporting infrastructure at WFOzSyncUtils.updateStatusFromHttpResultFn.
					
						:mapDataP( function( status, data, headers )
													
							logger:debugf( "Oz returned %s", headers.status_code )
					
							local result = decodeOzJson( data )

							logger:debugf( "Oz returned data %s", result )
							
							-- Process error notifications.
							
							local errors = result and result.errors
							if errors and not errors.code then
							
								return KSPromises.allP(
								
									function( errorRow )
										recordUploadError( self, errorRow )
										
										local errorCode = tonumber( errorRow.code )
										
										logger:errorf( "Processing error %s", errorRow )
										
										local errorOverlay = {
											_replace = true,
											permanent = true,
											code = errorRow.code,
											description = errorRow.description,
										}
										
										-- Scan backwards through the error list because
										-- (grrr) what Oz gives us is the asset ID that
										-- failed, not the album_asset ID that failed.
										-- Unit test happens to generate both in the same
										-- packet, so we cover the likely case of a duplicate
										-- by assuming last one loses.
										
										local errorAssetId = errorRow._id
										local docId

										for i = #uploadGroup, 1, -1 do
											local uploadItem = uploadGroup[ i ]
											if uploadItem._id == errorAssetId then
												docId = uploadItem.album_asset_id
												break
											end
										end
										
										if not docId then return end
										
										local docOverlay = {
											_id = docId,
											_slot = 'main',
											_syncStatus = {
												oz = {
													error = errorOverlay,
												},
											},
										}
										
										if errorCode == 1002 then
											
											-- Duplicate: Add annotation for corrected ID.
											local replacedById = KSTableUtils.derefOpt(
																			errorRow,
																			'album_asset',
																			'_id' )
											if replacedById ~= docId then
												docOverlay.replaced_by_id = replacedById
											end
										end
										
										return self._docStore:applyOverlayP( docOverlay )
									
									end, errors, 10 )

							end

							if derefOpt( errors, 'album', 1 )
								or derefOpt( errors, 'album_id', 1 )
							then
							
								-- The album to which we just attempted to upload
								-- no longer exists on Oz. Delete all corresponding
								-- album_asset records.
							
							end
						
						end )
			
				end, groupedAssetAndDocIds, 3)

		end )

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------
--[[!
	\brief Sends diagnostic data on synchronization to the log.
]]--

function albumAssetUploaderMethods:sendDiagnosticLogData()

	check( 'WFOzAlbumAssetUploader' )

	local catalogSpec = self._catalogSpec
	local wfOzUser = catalogSpec:wfOzUser()

	return {

		name = "WF Oz AlbumAsset Uploader",
		id = 'WFOzAlbumAssetUploader',
		weight = 505,

		body = {
		
			ozApiServer = catalogSpec:serverBaseUrl(),
			ozCatalogId = catalogSpec:catalogId(),
			ozCatalogHref = catalogSpec:catalogHref(),
			ozUserId = wfOzUser:userId(),
           
			recentErrors = self._recentErrorLog,
			
			-- TODO_SNOWLESS [scouten 2013-07-08]: Add quantity of uploads of each type.
			-- See uploadStatusReportBins in OzCentralUploadTask.lua.
			
		},
	
	}
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzAlbumAssetUploader'( albumAssetUploaderMethods )
