--[[----------------------------------------------------------------------------

WFOzAlbumAssetUploaderTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.lua
local WFTimer = require 'WFTimer'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.sync_engine.metadata_sync
local WFOzAlbumAssetUploader = require 'WFOzAlbumAssetUploader'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzCatalogTestSession = require 'WFOzCatalogTestSession'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'


--============================================================================--

-- @mark WFOzAlbumAssetUploaderTest

local WFOzAlbumAssetUploaderTest = {}
local TEST_WAIT_TIME = 240

--------------------------------------------------------------------------------

local catalogTestSession, indexStore, wfOzSyncAgent, albumAssetUploader
local fakePhotoIdsByCaptureDate

--------------------------------------------------------------------------------

function WFOzAlbumAssetUploaderTest.setUp()

	catalogTestSession = WFOzCatalogTestSession( 'WFOzAlbumAssetUploaderTest' )

	indexStore = nil
	wfOzSyncAgent = nil
	albumAssetUploader = nil

end

--------------------------------------------------------------------------------

function WFOzAlbumAssetUploaderTest.tearDown()

	if catalogTestSession then
		catalogTestSession:releaseTestUserP()
		catalogTestSession:close()
		catalogTestSession = nil
	end

end

--------------------------------------------------------------------------------

local function albumAssetUploaderP()

	indexStore = WFIndexStore( catalogTestSession:docStore(),
							   'AlbumAssetUploaderTestIndexDefinitions.lua' )

	catalogTestSession:addCloseTarget( indexStore )

	return catalogTestSession:ozSyncAgentWithMetadataSyncCompletedP(
									{ suppressAlbumAssetUploads = true } )
		
		:mapDataP( function( _, syncAgent )
		
			return catalogTestSession:catalogSpecP()
		
		end )
		
		:mapDataP( function( _, catalogSpec )
		
			albumAssetUploader = WFOzAlbumAssetUploader
				{
					catalogSpec = catalogSpec,
					docStore = catalogTestSession:docStore(),
					indexStore = indexStore,
				}
			
			catalogTestSession:addCloseTarget( albumAssetUploader )
			
			fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()
			
			return 'data', albumAssetUploader
		
		end )

end

----------------------------------------------------------------------------

local function hasServerCreatedDate( docValue )

	return not not KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'created' )

end

----------------------------------------------------------------------------

local function hasPermanentErrorAndRedirectedId( docValue )

	if KSTableUtils.derefOpt( docValue, '_syncStatus', 'oz', 'error', 'permanent' ) then
	
		return KSTableUtils.derefOpt( docValue, 'replaced_by_id' ), docValue
	
	else
	
		return false, docValue
	
	end

end

----------------------------------------------------------------------------

function WFOzAlbumAssetUploaderTest.testHappyPath_testNightly( test )

	local docStore = catalogTestSession:docStore()

	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId1 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId2 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId3 = KSIdGenerator.uuidInLowercaseHex()

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the album asset uploader. Now the albumAssetUploader file variable
		-- will be defined.
		
		:waitFor( function()

			return albumAssetUploaderP()

		end )

		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Create album and put all three assets in it.
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					type = 'album',
					name = 'test album',
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId1,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId2,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId3,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
					},
				}
			
		end )
		
		-- Now make sure all three album-asset records get uploaded.
		
		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId2 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId3 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

----------------------------------------------------------------------------

function WFOzAlbumAssetUploaderTest.testDuplicateHandling_testNightly( test )

	local docStore = catalogTestSession:docStore()

	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId1 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId1_dup = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId2 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId3 = KSIdGenerator.uuidInLowercaseHex()

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the album asset uploader. Now the albumAssetUploader file variable
		-- will be defined.
		
		:waitFor( function()

			return albumAssetUploaderP()

		end )

		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Create album and put all three assets in it.
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					type = 'album',
					name = 'test album',
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId1,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId2,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ],
					},
				}
			
		end )
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId3,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
					},
				}
			
		end )
		
		-- Kids, don't try this at home. But we're going to upload a second
		-- album_asset document for this and make sure we deal with it
		-- gracefully.
		
		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = albumAssetId1_dup,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					},
				}
			
		end )
		
		-- Now make sure the first three album-asset records get uploaded properly.
		
		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId2 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId3 )
				:mapB( hasServerCreatedDate )
		
		end )
		:continueWhenValueMatches( true )

		-- Now make sure the duplicate of the first album-asset record got
		-- tagged as an error properly.

		:waitFor( function()
		
			return docStore:docValueB( albumAssetId1_dup )
				:mapB( hasPermanentErrorAndRedirectedId )
		
		end )
		:continueWhenValueMatches( albumAssetId1, ':ignore:' )

		-- Clean up.

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

function WFOzAlbumAssetUploaderTest.testSortOrder_testNightly( test )

	local docStore = catalogTestSession:docStore()

	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId3 = KSIdGenerator.uuidInLowercaseHex()
	
	local docValueDidChangeToLast = false
	local docValueDidFlipBack = false
	local docValueDidFlipBackB

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the album asset uploader. Now the albumAssetUploader file variable
		-- will be defined.

		:waitFor( function()

			return albumAssetUploaderP()

		end )

		-- Turn on automatic sync.

		:waitFor( function()

			return catalogTestSession:ozSyncAgentP()

				:mapDataP( function( _, syncAgent )
					syncAgent:startAutomaticSync()
					return 'data', true
				end )

		end )

		-- Create album and put all three assets in it.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					type = 'album',
					name = 'test album',
				}

		end )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = albumAssetId3,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					order = 'original',
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
					},
				}

		end )

		-- Now make sure the album-asset record gets uploaded properly.

		:waitFor( function()

			return docStore:docValueB( albumAssetId3 )
				:mapB( hasServerCreatedDate )

		end )
		:continueWhenValueMatches( true )
		
		-- Set up an observer for sort order and make sure it never flips back.
		
		:waitFor( function()
		
			docValueDidFlipBackB = docStore:fieldValueB( albumAssetId3, 'order' )
				:mapB( function( value )
				
					debug.print( "order changed to ", value )
				
					if docValueDidFlipBack then
						return true
					elseif value == 'last' then
						docValueDidChangeToLast = true
					elseif value == 'first' then
						if docValueDidChangeToLast then
							docValueDidFlipBack = true
							return true
						end
					end
					
					return false
				
				end )
			
			return docValueDidFlipBackB
			
		end )
		:continueWhenValueMatches( false )

		-- Now change the album-asset info twice and make sure we get it back.

		:waitFor( function()
		
			-- This actually runs later: Change the sort order a second time.

			WFTimer.afterDelayDo( 10, function()

				debug.print( 'setting order to last' )

				docStore:setDocValueP
					{
						_id = albumAssetId3,
						_slot = 'main',
						type = 'album_asset',
						album = {
							_id = albumId,
						},
						order = 'last',
						asset = {
							_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
						},
					}

			end )

			debug.print( 'server create date verified, setting order to first' )

			-- This runs earlier. Change the sort order.

			return docStore:setDocValueP
				{
					_id = albumAssetId3,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					order = 'first',
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ],
					},
				}

		end )

		-- Verify that order doesn't flip back on us, even after a while.
		
		:waitFor( function()
			return docValueDidFlipBackB
		end )
		:continueWhenValueMatches( false )
		
		-- Wait a bit longer than a minute for the 60-second blackout window
		-- to expire and for us to upload the second value.
		
		:waitFor( 100 )
		
		:waitFor( function()
			return docValueDidFlipBackB
		end )
		:continueWhenValueMatches( false )
		
		-- Recheck the final doc value.

		:waitFor( function()
			return docStore:fieldValueB( albumAssetId3, 'order' )
		end )
		:continueWhenValueMatches( 'last' )

		-- Verify that correct doc value eventually made it to Oz.

		:waitFor( function()
		
			local href
		
			return docStore:docValueP( albumId )
			
				:mapDataP( function( _, value )

					return 'data', KSTableUtils.derefOpt( value,
														  '_serverProvidedMetadata',
														  'links',
														  '/rels/album_assets',
														  'href' )
				
				end )

				:mapDataP( function( _, _href )
					href = _href
					return catalogTestSession:catalogSpecP()
				end )
				
				:mapDataP( function( _, catalogSpec )
					return catalogSpec:getP( href )
				end )

				:mapDataP( function( _, data )
					return 'data', KSJsonDecoder.decodeWebSafeJson_stripNull( data )
				end )
				
				:mapDataP( function( _, value )
				
					return 'data', KSTableUtils.derefOpt( value,
														  'assets',
														  1,
														  'order' )
				
				end )

		end )
		:continueWhenValueMatches( 'data', 'last' )

		-- Clean up.

		:waitFor( function()

			return catalogTestSession:releaseTestUserP()

		end )

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function WFOzAlbumAssetUploaderTest.testSendDiagnosticLogData_testNightly( test )

	-- Since diagnostic log is expected to evolve frequently, we're not
	-- verifying specific content. The main concern is that it doesn't fall
	-- over and die when generating the log.

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return albumAssetUploaderP()

		end )
		
		:waitFor( function()
		
			return WFDiagnosticLog.processLogSectionP(
				albumAssetUploader:sendDiagnosticLogData() )
		
		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

local WFPlatformInfo = require 'WFPlatformInfo'

if WFPlatformInfo.isWindows then
	return {}
else
	return WFOzAlbumAssetUploaderTest
end
