--[[----------------------------------------------------------------------------

WFOzAlbumOrderingTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local WFOzAlbumOrdering = require 'WFOzAlbumOrdering'

local function addAsset( assets )
	local n = ( assets.n or 0 ) + 1
	local asset = {
		id = string.format( "asset %d", n ),
	}
	
	assets[ #assets + 1 ] = asset
	assets.n = n
end

--------------------------------------------------------------------------------

local function createAssets( n )
	local assets = { }
	
	for i = 1, n do
		addAsset( assets )
	end
	
	return assets
end

--------------------------------------------------------------------------------

local function emitorders( assets, optIndexSet )
	local indexSet = optIndexSet or { }

	local limit = #assets
	for i = 1, limit do
		local asset = assets[ i ]
		print( i, asset.order, asset.id, asset.touched and "*" or "",
			indexSet[ i ] and "X" or "" )
		
		if i < limit then
			local slope = "|"
			local next = assets[ i + 1 ].order
			if asset.order < next then
				slope = "\\"
			elseif asset.order > next then
				slope = "/"
			end
			print( "", slope )
		end
	end
end

--------------------------------------------------------------------------------

local function noop() end

--------------------------------------------------------------------------------

local function move( assets, src, dst )
	local asset = table.remove( assets, src )
	table.insert( assets, dst, asset )
	asset.touched = true
end

--------------------------------------------------------------------------------

local function reorder( assets, count, fn )
	fn = fn or noop

	for i = 1, count do
		local src = math.random( 1, #assets )
		local shift = math.random( 1, #assets - 1 )
		local dst = ( src + shift - 1 ) % #assets + 1 -- rotate
		assert( src ~= dst )
		fn( "move", src, dst )
		move( assets, src, dst )
	end
end

--------------------------------------------------------------------------------

local function remove( assets, count, fn )
	fn = fn or noop

	for i = 1, count do
		if #assets == 0 then
			return
		end
	
		local target = math.random( 1, #assets )
		fn( "remove", target )
		table.remove( assets, target )
	end
end

--------------------------------------------------------------------------------

local function add( assets, count, fn )
	fn = fn or noop

	for i = 1, count do
		local src = math.random( 1, #assets )
		addAsset( assets )
		move( assets, #assets, src )
		fn( "add", assets.n, src )
	end
end

--------------------------------------------------------------------------------

local function clearTouches( assets )
	for _, asset in ipairs( assets ) do
		asset.touched = nil
	end
end

--------------------------------------------------------------------------------

local function valueAfter( v )
	local result = v + 1
	assert( result > v, "no room to insert after" )
	return result
end

--------------------------------------------------------------------------------

local function valueBefore( v )
	local result = v - 1
	assert( result < v, "no room to insert before" )
	return result
end

--------------------------------------------------------------------------------

local function valueBetween( v1, v2 )
	local result = ( v1 + v2 ) / 2
	assert( result > v1 and result < v2, "no room to insert between" )
	return result
end

--==============================================================================

local orderGenerator = WFOzAlbumOrdering.orderGenerator(
	1, valueBefore, valueAfter, valueBetween
)

--------------------------------------------------------------------------------

local function countKeys( t )
	local count = 0
	for k in pairs( t ) do
		count = count + 1
	end
	
	return count
end

--------------------------------------------------------------------------------

local function verifyOrdering( input, changeCount, sortStrings )
	local generator = not sortStrings and orderGenerator
	local idToOrder = WFOzAlbumOrdering.adjustOrderFields( input, generator )

	if WFOzAlbumOrdering.findProblem( input, 1 ) ~= nil then
		emitorders( input )
		assert( false, "integrity check failure" )
	end
	
	local actual = countKeys( idToOrder )
	if changeCount and actual > changeCount then
		emitorders( input, idToOrder )
		assert( false, string.format( "too many changes: %d != %d", changeCount, actual ) )
	end
end

--==============================================================================

local WFOzAlbumOrderingTest = {}

-- FIXME [jandrew 2/5/2015]: This test fails on Android.
local WFPlatformInfo = require 'WFPlatformInfo'
if not WFPlatformInfo.isAndroid then

-- note this test is a non-deterministic property test. I've never seen it fail,
-- but in theory it could (and hopefully the log output provides enough clues as
-- to the cause if it ever does
function WFOzAlbumOrderingTest.testOrderings( test )

	local assets = createAssets( 5000 )
	for i = 1, 25 do
		reorder( assets, 100, noop )
		remove( assets, 250, noop )
		add( assets, 250, noop )
		verifyOrdering( assets, i == 1 and #assets or 350, true )
		
		clearTouches( assets )
	end
end

end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testEmpty( test )
	verifyOrdering( { } )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testTrivial( test )
	verifyOrdering( {
		{ id = "1", order = 4 },
	}, 0 )

	verifyOrdering( {
		{ id = "1", order = 2 },
		{ id = "2", order = 5 },
		{ id = "3", order = 5.1 },
	} )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testBlank( test )
	verifyOrdering( {
		{ id = "1" },
		{ id = "3" },
	}, 2 )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testMiddle( test )
	verifyOrdering( {
		{ id = "1" },
		{ id = "2", order = 5 },
		{ id = "3" },
	}, 2 )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testGaps( test )
	verifyOrdering( {
		{ id = "1" },
		{ id = "2", order = 5 },
		{ id = "3" },
		{ id = "4", order = 5.5 },
		{ id = "5" },
	}, 3 )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testTouchHints( test )
	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 5, touched = true },
		{ id = "2", order = 4, touched = true },
		{ id = "3", order = 2 },
		{ id = "4", order = 3 },
	}, 2 )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testNoTouch( test )
	-- without touch hints the 5 causes a cascade that updates 2 and 3 when
	-- it could have been avoided
	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 5 },
		{ id = "2", order = 4, touched = true },
		{ id = "3", order = 2 },
		{ id = "4", order = 3 },
	}, 3 )
end
--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testTouchGaps( test )
	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 2, touched = true }, -- false touch
		{ id = "3", order = 3 },
		{ id = "4", order = 5, touched = true },
		{ id = "5", order = 4 },
	}, 1 )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testInverted( test )
	verifyOrdering( {
		{ id = "1", order = 12 },
		{ id = "2", order = 10 },
		{ id = "3", order = 8 },
	}, 2 )
end

--------------------------------------------------------------------------------

function WFOzAlbumOrderingTest.testCollision( test )
	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 3, touched = true },
		{ id = "3", order = 1 },
		{ id = "3", order = 2 },
	}, 2 )

	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 1 },
		{ id = "3", order = 1 },
		{ id = "3", order = 1 },
	}, 3 )
	
	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 3 },
		{ id = "3", order = 1 },
	}, 1 )

	verifyOrdering( {
		{ id = "1", order = 1 },
		{ id = "2", order = 3, touched = true },
		{ id = "3", order = 1 },
	}, 2 )
end

--------------------------------------------------------------------------------

if WFPlatformInfo.isWindows then
	return {}
else
	return WFOzAlbumOrderingTest
end
