--[[----------------------------------------------------------------------------

WFOzBinaryProvider.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local KSTableUtils = require 'KSTableUtils'


--------------------------------------------------------------------------------

local binaryProviderLog = WFLogger 'WFOzBinaryProvider'
--	:setLevel( 'DEBUG' )

local derefOpt = KSTableUtils.derefOpt


--============================================================================--
-- @mark diagnostic log
--============================================================================--

local function addBinaryStatsToDiagnosticLog( self, bucket, time, size )
	
	local sizeInKB = size / 1000
	
	local bucketStats = self._binaryProviderStats.buckets[ bucket ]
	
	if not bucketStats then
		bucketStats = { count = 0,
		                timings = WFDiagnosticLog.makeRingBuffer { "Time (sec)" },
						sizes = WFDiagnosticLog.makeRingBuffer { "Size (KB)" },
					  }
		self._binaryProviderStats.buckets[ bucket ] = bucketStats
	
	end
	
	bucketStats.count = bucketStats.count + 1
	bucketStats.timings:add( time )
	bucketStats.sizes:add( size / 1000 )
	
	local totalBinarySize = self._binaryProviderStats.totalSize or 0
	
	self._binaryProviderStats.totalSize = totalBinarySize + sizeInKB
	
end


--============================================================================--
-- @mark WFOzBinaryUploader class
--============================================================================--

local binaryUploaderMethods = {

	WFCloseSupport.addCloseSupportMethods,

	_dataCache = false,
	
	_ozUserId = false,
	
	_ozCatalogId = false,
	
	_binaryProviderStats = false,
	
}

--------------------------------------------------------------------------------

function binaryUploaderMethods:__init( dataCache, userId, catalogId )

	self._ozUserId = userId
	
	self._ozCatalogId = catalogId
	
	self._binaryProviderStats = {
			buckets = {},
			totalSize = 0,
	}
	
	self:setOwned {
		_dataCache = dataCache,
	}
	
	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

local now  = require "KSDate".timeNowForAnimation

--------------------------------------------------------------------------------

function binaryUploaderMethods:getRenditionP( docValue, renditionType )

	binaryProviderLog:debugf( "getRenditionP, renditionType = %s, docValue = %s",
							   renditionType, docValue )

	if renditionType == '2048' or renditionType == '1024' then
		renditionType = 'preview'
	end
	
	local renditionMD5 = derefOpt( docValue, '_localOnly', 'renditions',
							renditionType, 'md5' )

	if not renditionMD5 then
		return KSPromises.resolvedPromiseP( 'data', nil )
	end
	
	local renditionDataP = self._dataCache:getP( renditionMD5 )
	
	return renditionDataP
	
end

--------------------------------------------------------------------------------

local function makeKSDataFromPathP( path )

	return KSPromises.resolvedPromiseP( 'data', KSData.memoryMappedContentsOfFile( path ) )
	
end
														
--------------------------------------------------------------------------------

local function makeProxyCreationLogger( self )

	local proxyStartTime = now()

	return function( status, maybeData )
				
		if status == 'data' and maybeData then

			local duration =  now() - proxyStartTime

			local size = KSData.length( maybeData )

			addBinaryStatsToDiagnosticLog( self, 'proxy', duration, size )

		end

	end

end

--------------------------------------------------------------------------------

function binaryUploaderMethods:getProxyP( docValue, proxyType )

	binaryProviderLog:debugf( "getProxyP, proxyType = %s, docValue = %s",
							   proxyType, docValue )

	local proxyFilePath = derefOpt( docValue, '_localOnly', 'proxy',
							proxyType, 'filePath' )
	
	local makeKSDataP
	if proxyFilePath and WFFileUtils.exists( proxyFilePath ) then
		makeKSDataP = makeKSDataFromPathP( proxyFilePath )
	else
		makeKSDataP = KSPromises.resolvedPromiseP( 'data', nil )
	end

	local maybeProxyAsKSDataP = makeKSDataP

	local logProxyCreation = makeProxyCreationLogger( self )

	local loggedMaybeProxyP = maybeProxyAsKSDataP:observeP( logProxyCreation )

	return loggedMaybeProxyP
	
end

--============================================================================--
-- @mark masters
--============================================================================--

function binaryUploaderMethods:getMasterP( docValue )

	binaryProviderLog:debugf( "getMasterP, docValue = %s", docValue )
	
	local localOnly = docValue._localOnly or {}
	local originalFilePath = localOnly.originalFilePath

	-- FIXME [scouten 2015-04-16]: Pending checkin from @arahn will remove the
	-- need for this hack. Don't want to disrupt his refactoring on import code,
	-- so we'll do this for now.
	
	if not( originalFilePath and WFFileUtils.exists( originalFilePath ) ) then
		originalFilePath = KSTableUtils.derefOpt( docValue, 'importSource', 'filePath' )
	end

	-- TODO [scouten 2015-04-21]: HACK HACK HACK ALERT!!! Sniff file name
	-- to obtain content-type. Ideally this should be part of the doc value.

	local HACK_contentType

	if originalFilePath then
	
		local fileType = originalFilePath:match( '%..*$' )
		
		if fileType then
			fileType = string.lower( fileType )
		end
		
		if fileType == '.jpg' then
			HACK_contentType = 'image/jpeg'
		elseif fileType == '.psd' then
			HACK_contentType = 'image/vnd.adobe.photoshop'
		elseif fileType == '.dng' then
			HACK_contentType = 'image/dng'
		elseif fileType == '.tif' or fileType == '.tiff' then
			HACK_contentType = 'image/tiff'
		else
			HACK_contentType = 'application/octet-stream'
		end
	
	end

	local resultP
	if originalFilePath and WFFileUtils.exists( originalFilePath ) then
		resultP = makeKSDataFromPathP( originalFilePath )
			:mapDataP( function( _, data )
				return 'data', data, HACK_contentType
			end )
	else
		resultP = KSPromises.resolvedPromiseP( 'data', nil )
	end
	
	return resultP
					
end

--------------------------------------------------------------------------------
--[[!
     \brief Enables logging of binary provider operations.
]]--

function binaryUploaderMethods:sendDiagnosticLogData()

	local finalStats = {}
	
	for bucket, stats in pairs( self._binaryProviderStats.buckets ) do
		finalStats[ bucket ] = { count = stats.count,
								 timings = { stats.timings:visualize "boxAndWhisker" },
								 size = { stats.sizes:visualize "boxAndWhisker" },
								}
	end
	
	WFDiagnosticLog.addSectionToLog {

		name = "WF Binary Provider",
		id = 'WFOzBinaryProvider' .. ' ::  ' .. self._ozUserId,
		weight = 500,

		body = {
			
			binaryStats = finalStats,
			totalBinarySize = self._binaryProviderStats.totalSize,
		}
	
	}
			
end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzBinaryProvider'( binaryUploaderMethods )
