--[[----------------------------------------------------------------------------

WFOzBinaryUploadUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCapabilities = require 'WFCapabilities'


--------------------------------------------------------------------------------

local binaryUploadLogger = WFLogger( 'WFOzBinaryUploadUtils' )
--	:setLevel( 'DEBUG' )

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

local WFOzBinaryUploadUtils = {}

--------------------------------------------------------------------------------

WFOzBinaryUploadUtils.kWFOzBinaryUploaderIgnoreError = "<ignore>"

--------------------------------------------------------------------------------

local function shouldUploadItem_imp( args )

	local docValue = assert( args.docValue )

	-- If this isn't an asset, nothing to upload. (And nowhere to upload it to!)
	
	if docValue.type ~= 'asset' then
		return false, 'notAsset'
	end
	
	-- If metadata isn't on Oz yet, can't upload binary.

	local links = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links' )
	
	local revisionUrlOnOz = KSTableUtils.derefOpt( links, 'self', 'href' )
	if not revisionUrlOnOz then
		return false, 'metadataNotUploaded'
	end

	-- If sync status error, don't send binaries until it's resolved
	local syncStatusError = KSTableUtils.derefOpt( docValue,
												   '_syncStatus',
												   'oz',
												   'error' )
	if syncStatusError then
		return false, 'itemSyncStatusError'
	end
	
	-- If we don't have any such binary, then we can't upload it.
	
	local binaryInfo = KSTableUtils.derefOpt( docValue,
											  '_localOnly',
											  unpack( args.infoPath ) )
	if binaryInfo and binaryInfo.missing then
		return false, 'itemMissingLocally'
	end
	
	-- Or maybe we had an error uploading it. Note that we can get
	-- an error of cancelled which happens in Lr if the render is
	-- preempted and we should retry in this case
	
	-- FIXME [matthewj 2014-12-01 via pkamin]: Who is supposed to set the
	-- message to 'canceled', the DataProvider? Can we use a constant here
	-- similar to WFOzBinaryUploadUtils.kWFOzBinaryUploaderIgnoreError
	-- that can be used by clients to provide specific error codes?
	
	if binaryInfo and binaryInfo.error and
		not ( binaryInfo.error.message == 'canceled' ) then
		return false, 'itemFailedToUpload'
	end
	
	-- If binary is missing, then we need to upload it.

	local binaryLinks = KSTableUtils.derefOpt( links, args.linkPath )
	
	local binaryUrlOnOz = KSTableUtils.derefOpt( binaryLinks, 'href' )
	if not binaryUrlOnOz then
		return true, 'itemNotUploaded'
	end
	
	-- If binary is present, but invalid, we still need to upload it.
	
	local binaryIsInvalid = KSTableUtils.derefOpt( binaryLinks, 'invalid' )
	if binaryIsInvalid then
		return true, 'itemIsInvalid'
	end
	
	-- Otherwise, it's good.
	
	return false, 'itemIsCurrent'

end

--------------------------------------------------------------------------------

function WFOzBinaryUploadUtils.shouldUploadRendition( docValue, renditionType )

	check( 'table', '+string' )

	return shouldUploadItem_imp
		{
			docValue = docValue,
			infoPath = { 'renditions', renditionType },
			linkPath = '/rels/rendition_type/' .. renditionType,
		}

end

--------------------------------------------------------------------------------

function WFOzBinaryUploadUtils.shouldUploadProxy( docValue, proxyType )

	check( 'table', '+string' )

	return shouldUploadItem_imp
		{
			docValue = docValue,
			infoPath = { 'proxy', proxyType },
			linkPath = '/rels/proxy_type/' .. proxyType,
		}

end

--------------------------------------------------------------------------------

function WFOzBinaryUploadUtils.shouldUploadMaster( docValue )

	check( 'table' )
	
	-- TODO [scouten 2015-04-16]: The need for this should go away when
	-- WF managed import code comes online.

	if not KSTableUtils.derefOpt( docValue, '_localOnly', 'originalFilePath' ) then

		-- Hack file path into doc value for now and clear "missing" flag.
		
		local altOriginalPath = KSTableUtils.derefOpt( docValue, 'importSource', 'filePath' )
		
		if altOriginalPath and require 'WFFileUtils'.exists( altOriginalPath ) then
			
			if not docValue._localOnly then
				docValue._localOnly = {}
			end
			
			docValue._localOnly.originalFilePath = altOriginalPath
			
			if docValue._localOnly.master then
				docValue._localOnly.master.missing = nil
			end
		
		end
	
	end
	
	-- END HACK
	
	local willNotUploadMaster = KSTableUtils.derefOpt( docValue,
													   'importSource',
													   'willNotUploadMaster' )

	if willNotUploadMaster then

		return false, 'willNotUploadMaster'

	else

		return shouldUploadItem_imp
			{
				docValue = docValue,
				infoPath = { 'master' },
				linkPath = '/rels/master',
			}
		
	end

end

--------------------------------------------------------------------------------

local function isAssetMasterMissingOnOz( docValue )

	if docValue.type ~= 'asset' then

		return false

	end

	local syncStatusError = KSTableUtils.derefOpt( docValue, '_syncStatus',
													'oz', 'error' )

	if syncStatusError then

		-- if the asset is in error state, the presence of the remote master is
		-- irrelevant
	
		return false

	end
	
	local willNotUploadMaster = KSTableUtils.derefOpt( docValue,
													   'importSource',
													   'willNotUploadMaster' )

	if willNotUploadMaster then

		binaryUploadLogger:infof( "If asset %s originated remotely, we will never be able to download a master on this client",
									docValue._id )

		return false

	end


	local masterURLInfo = KSTableUtils.derefOpt( docValue,
													'_serverProvidedMetadata',
													'links', '/rels/master' )

	if not masterURLInfo or not masterURLInfo.href or
			masterURLInfo.invalid then

		return true

	end

	return false

end

--------------------------------------------------------------------------------

function WFOzBinaryUploadUtils.getRenditionTypes()

	local renditionTypesStr =
		WFCapabilities.getCapability( "WFOzBinaryUploader.renditionTypes" )
			or "thumbnail2x,2048"
	
	return KSStringUtils.split( renditionTypesStr, ',' )

end

--------------------------------------------------------------------------------

function WFOzBinaryUploadUtils.getProxyTypes()

	local proxyTypesStr =
		WFCapabilities.getCapability( "WFOzBinaryUploader.proxyTypes" )
			or "2560"

	return KSStringUtils.split( proxyTypesStr, ',' )

end

--------------------------------------------------------------------------------

local shouldUploadRendition = WFOzBinaryUploadUtils.shouldUploadRendition
local shouldUploadProxy = WFOzBinaryUploadUtils.shouldUploadProxy
local shouldUploadMaster = WFOzBinaryUploadUtils.shouldUploadMaster

--------------------------------------------------------------------------------

local function makeNameTable( arr )

	local nameTable = {}

	for _, name in ipairs( arr ) do

		nameTable[ name ] = name

	end

	return nameTable

end

--------------------------------------------------------------------------------

local _uploadableBinaryTypesArray = { "rendition", "proxy", "master" }

--------------------------------------------------------------------------------

local _uploadableBinaryTypes = makeNameTable( _uploadableBinaryTypesArray )

--------------------------------------------------------------------------------

local _nonUploadableBinaryTypesArray = { "missingRemoteMaster" }

--------------------------------------------------------------------------------

local _nonUploadableBinaryTypes =
		makeNameTable( _nonUploadableBinaryTypesArray )
	-- FIXME [bkruse 2015-01-28 via arahn]: Is this extra obfuscation worth it?
	-- More detailed comment at
	-- https://git.corp.adobe.com/lua/wf/pull/318#discussion_r121184.

--------------------------------------------------------------------------------

function WFOzBinaryUploadUtils.generateIndexSchema( indexStoreSchema )

	-- TODO_SNOWLESS [scouten 2013-06-20]: This logic doesn't yet handle Revel-style
	-- upload wave prioritization nor does it handle freemium embargoing.

	--check( 'WFIndexStoreSchema' )
		-- Ugh. Yet another reason why I hate mocks in unit tests.
		-- Mocks created at OzCatalogTestUtils.buildMocksForTestingCatalogService
		-- mean that the WFIndexStoreSchema mock object doesn't pass WFArguments
		-- smell test so now I have to disable this safety check. Only discovered
		-- this when I made a change in a yet another unrelated part of the app
		-- that introduced a dependency on this file, but only in Lightroom's
		-- test harness environment. (Thanks to differences how we configure
		-- OzCatalogIndexDefinitions via WFCapabilities, naturally.)

	local renditionTypes = WFOzBinaryUploadUtils.getRenditionTypes()
	local proxyTypes = WFOzBinaryUploadUtils.getProxyTypes()

	local nRenditionTypes = #renditionTypes
	local nProxyTypes = #proxyTypes

	indexStoreSchema:addTable {
	
		sql = [[CREATE TABLE missingRenditionsOnOz( captureDate, binaryType, renditionType, uploadSequence )]],
	
		map = function( docValue, emit )

			-- We only care about assets. Other doc types are metadata only
			-- and are thus handled by WFSyncAgent.

			if docValue.type ~= 'asset'
				or KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'subtype' ) ~= 'jpeg'
			then
				return
			end
	
			binaryUploadLogger:debugf( 'missingRenditionsOnOz reindexing\n\n%s\n\n', docValue )

			local captureDate = docValue.captureDate
			
			-- Which rendition types need to be uploaded?

			for i = 1, nRenditionTypes do
			
				local renditionType = renditionTypes[ i ]
			
				if shouldUploadRendition( docValue, renditionType ) then
					emit( captureDate, _uploadableBinaryTypes.rendition, renditionType, i )
				end
			
			end
			
			-- Which proxy types need to be uploaded?
			
			for i = 1, nProxyTypes do
			
				local proxyType = proxyTypes[ i ]
				
				if shouldUploadProxy( docValue, proxyType ) then
					emit( captureDate, _uploadableBinaryTypes.proxy, proxyType, nRenditionTypes + i )
				end
			
			end
			
			-- And how 'bout a master?
			
			if shouldUploadMaster( docValue ) then
				emit( captureDate, _uploadableBinaryTypes.master, '', nRenditionTypes + nProxyTypes + 1 )
			end
			
			if isAssetMasterMissingOnOz( docValue ) then

				binaryUploadLogger:infof( "Asset %s may have originated remotely and it does not have a master uploaded by remote client",
											docValue._id )

				emit( captureDate, _nonUploadableBinaryTypes.missingRemoteMaster, '', nRenditionTypes + nProxyTypes + 1 )

			end

		end,
	
	}
	
	indexStoreSchema:addIndex
		[[CREATE INDEX missingRenditionsOnOz_uploadSequence ON missingRenditionsOnOz(
					captureDate, docId, uploadSequence DESC, binaryType )]]
	
	indexStoreSchema:addIndex
		[[CREATE INDEX missingRenditionsOnOz_missingBinaryType ON missingRenditionsOnOz(
					binaryType, docId )]]
	
end

--------------------------------------------------------------------------------

local _isUploadableBinaryTypeStringList =
		string.format( "'%s'", table.concat( _uploadableBinaryTypesArray,
												"','" ) )

--------------------------------------------------------------------------------

local _uploadSequenceOrdering = "uploadSequence, captureDate"

--------------------------------------------------------------------------------

local _captureDateAssetIdOrdering = "captureDate, docId, uploadSequence DESC"

--------------------------------------------------------------------------------

WFOzBinaryUploadUtils.getUploadableAssetsSQL = function( uploadSequence )

	local ordering =
			( uploadSequence == 'renditionsFirst' or not uploadSequence ) and
			_uploadSequenceOrdering or _captureDateAssetIdOrdering
				
	return string.format(
				[[SELECT docId, binaryType, renditionType
					FROM missingRenditionsOnOz
					WHERE binaryType IN ( %s )
					ORDER BY %s
					LIMIT 20]],
				_isUploadableBinaryTypeStringList, ordering )
		
end

--------------------------------------------------------------------------------

WFOzBinaryUploadUtils.getMissingRemoteMasterAssetsB =
function( indexStore, maybeFrequency )

	local calmValue = tonumber( maybeFrequency ) or 1

	calmValue = calmValue > 0 and calmValue or 1

	return indexStore:queryB{

				calm = calmValue,

				sql = [[SELECT docId FROM missingRenditionsOnOz
							WHERE binaryType = "missingRemoteMaster"
							ORDER BY binaryType, docId]],

			}
		
end

--------------------------------------------------------------------------------

return WFOzBinaryUploadUtils
