--[[----------------------------------------------------------------------------

WFOzBinaryUploader.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local WFCapabilities = require 'WFCapabilities'

local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'
local WFQueryDrivenWorkQueueScheduler = require 'WFQueryDrivenWorkQueueScheduler'

	-- WF.oz_client.sync_engine
local WFOzCatalogSpec = require 'WFOzCatalogSpec'
local WFOzBinaryUploadUtils = require 'WFOzBinaryUploadUtils'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local binaryUploaderLog = WFLogger 'WFOzBinaryUploader'
	:setLevel( 'INFO' )

local now = assert( KSDate.timeNowForAnimation )

--------------------------------------------------------------------------------

local function addBinaryStatsToDiagnosticLog( self, bucket, time, size )
	
	local sizeInKB = size / 1000
	
	local bucketStats = self._uploadSpeedStats.buckets[ bucket ]
	local uploadSpeed = sizeInKB / time
	
	if not bucketStats then

		bucketStats = {
			count = 0,
			timings = WFDiagnosticLog.makeRingBuffer { "Time (sec)" },
			sizes = WFDiagnosticLog.makeRingBuffer { "Size (KB)" },
			speed = WFDiagnosticLog.makeRingBuffer { "Upload speed (Kbps)" },
		}
									
		self._uploadSpeedStats.buckets[ bucket ] = bucketStats
	
	end
	
	bucketStats.count = bucketStats.count + 1
	bucketStats.timings:add( time )
	bucketStats.sizes:add( sizeInKB )
	bucketStats.speed:add( uploadSpeed )
	
	local totalBinarySize = self._uploadSpeedStats.totalSize or 0
	self._uploadSpeedStats.totalSize = totalBinarySize + sizeInKB
	
end

--------------------------------------------------------------------------------
	
local function uploadBinaryDataP( args, applyStatusOverlayP )

	local self = args.self
	local data = args.data

	local docValue = args.docValue
	
	if data == nil then
		return applyStatusOverlayP( 'missing', true )
	end
	
	-- Where will we put this item?
	
	local revisionUrl = KSTableUtils.derefOpt( docValue,
											   '_serverProvidedMetadata',
											   'links',
											   '/rels/latest_revision',
											   'href' )
	
	if not revisionUrl then

		binaryUploaderLog:errorf( "No latest_revision URL in doc value\n\n%s",
								   docValue )
		
		error( "No latest_revision URL; can not upload" )
	
	end
	
	local binaryDataId = KSIdGenerator.uuidInLowercaseHex()
	local putUrl = string.format( args.putUrlTemplate, revisionUrl, binaryDataId )

	local requestHeaders = args.requestHeaders
	
	requestHeaders[ #requestHeaders + 1 ] = { header = 'Content-Type',
											  value = args.contentType }
	
	requestHeaders[ #requestHeaders + 1 ] = { header = 'Accept', value = 'application/json' }
	
	local startTime = now()

	local size
	if KSData.isData( data ) then
		size = KSData.length( data )
	elseif type( data ) == 'string' then
		size = #data
	else
		size = 0
	end
	
	-- Override Oz endpoint if so requested via WFCapabilities.

	local baseUrl = WFCapabilities.getCapability( 'WFOzBinaryUploader.binaryUploadBaseUrl' )
	if baseUrl then
		putUrl = baseUrl .. putUrl
	end
	
	binaryUploaderLog:infof( "PUT %s, headers %s", putUrl, requestHeaders )

	return self._catalogSpec:putP( putUrl, data, { requestHeaders = requestHeaders } )

		:mapP( function( status, responseData, headers, ... )
		
			local statusCode = type( headers ) == 'table' and headers.status_code
		
			-- Release the data object ASAP.
		
			if KSData.isData( data ) then
				data:close()
			end

			-- Calculate elapsed time.
			
			local elapsedTime = now() - startTime
			local roundedElapsedTime = math.floor( elapsedTime + .5 )
		
			-- Only a 201 response is acceptable.
		
			if status == 'data' and statusCode == '201' then

				addBinaryStatsToDiagnosticLog( self, args.bucket, elapsedTime, size )

				return status, responseData, headers, ...

			end
			
			-- Special case for HTTP transit failure.
			
			if not headers then
			
				local annotatedReason = string.format( "%s (%d seconds)",
													   tostring( responseData ),
													   roundedElapsedTime )
			
				local logFmtStr = "Unable to communicate with Oz URL %s\n\nreason = %s"
			
				binaryUploaderLog:errorf( logFmtStr, putUrl, annotatedReason )
				
				return 'error', annotatedReason
			
			end

			-- We were able to communicate with Oz, but Oz didn't like the response. Log this.
		
			if not headers then
				headers = {}
			end
			
			local logFmtStr =  "Got status %s back from Oz\n\n"
			local logParams = { headers.status_code or 'nil' }

			if putUrl then
				logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
				logParams[ #logParams + 1 ] = putUrl
			end

			logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = responseData
			
			binaryUploaderLog:errorf( logFmtStr, unpack( logParams ) )
			
			-- If we got a 2xx, 3xx, or 4xx status code, we probably won't
			-- succeed if we try again with the same data. Flag this asset
			-- document so that we don't retry without explicitly clearing
			-- the error flag.
			
			if type( statusCode ) == 'string' and statusCode < '500' then
			
				return applyStatusOverlayP( 'error', statusCode )
			
			end
		
		end )

	-- TODO_SNOWLESS [scouten 2013-07-08]: Ask Oz to generate a full-res
	-- rendition if this is sub-res. See OzCentralUploadTask.lua, lines 812++.
	
	-- TODO_SNOWLESS [scouten 2013-07-09]: Track how long it takes to complete
	-- upload. Search for timeStampAtFirstUpload in OzCentralUploadTask.lua.
	
	-- TODO_SNOWLESS [scouten 2013-07-08]: Provide a hook for binary provider
	-- to move/purge content upon successful upload.
	
end

--------------------------------------------------------------------------------

local function handleBinaryResultsP( args )
	
	binaryUploaderLog:debugf( "handleBinaryResultsP status = %s, data = %s, docValue = %s",
							   args.status, tostring( args.data ), args.docValue )
	
	local self = args.self
	local status = args.status
	local data = args.data
	
	local docValue = args.docValue
	
	local function applyStatusOverlayP( key, value )
	
		local statusOverlay = {
			_id = docValue._id,
			_slot = 'main',
			_localOnly = {},
		}
		
		local t = statusOverlay._localOnly
	
		for i = 1, #args.infoPath do
			local k = args.infoPath[ i ]
			t[ k ] = {}
			t = t[ k ]
		end
		
		t[ key ] = value
		
		return self._docStore:applyOverlayP( statusOverlay )
			:mapDataP( function()
				return 'data', nil
			end )

	end
	
	if status == 'error' then
		
		if data == WFOzBinaryUploadUtils.kWFOzBinaryUploaderIgnoreError then
			return 'data', nil
		else
			return applyStatusOverlayP( 'error', data )
		end
	else
		
		return uploadBinaryDataP( args, applyStatusOverlayP )
	
	end
	
end

--------------------------------------------------------------------------------

local function uploadOneRenditionP( self, docValue, renditionType )

	binaryUploaderLog:debugf( "uploadOneRendition renditionType = %s, docValue = %s",
							   renditionType, docValue )

	return self._binaryProvider:getRenditionP( docValue, renditionType )
		:mapP( function( status, data, contentType )
		
			return handleBinaryResultsP {

				self = self,
				docValue = docValue,
				status = status,
				data = data,
				bucket = 'rendition/' .. renditionType,
				
				infoPath = { 'renditions', renditionType },
				
				contentType = contentType or 'image/jpeg',

				requestHeaders = {
					{
						header = 'X-Rendition-Type',
						value = renditionType,
					},
				},
				
				putUrlTemplate = '%s/renditions/%s',
			
			}
		
		end )

end

--------------------------------------------------------------------------------

local function uploadOneProxyP( self, docValue, proxyType )

	binaryUploaderLog:debugf( "uploadOneProxyP proxyType = %s, docValue = %s",
							   proxyType, docValue )

	return self._binaryProvider:getProxyP( docValue, proxyType )
	
		:mapP( function( status, data, contentType )
		
			return handleBinaryResultsP {

				self = self,
				docValue = docValue,
				status = status,
				data = data,
				bucket = 'proxy/' .. proxyType,

				infoPath = { 'proxy', proxyType },
				
				contentType = contentType or 'image/dng',

				requestHeaders = {
					{
						header = 'X-Proxy-Type',
						value = proxyType,
					},
				},
				
				putUrlTemplate = '%s/proxies/%s',
			
			}

		end )

end

--------------------------------------------------------------------------------

local function getMasterRequestHeaders()

	local headers = {}

	if WFCapabilities.getCapability( 'WFOzBinaryUploader.requestOzRenditionGeneration' ) then
		
		headers[ #headers + 1 ] = {
			header = 'X-Generate-Renditions',
			value = 'all',
		}
	
	end
	
	if WFCapabilities.getCapability( 'WFOzBinaryUploader.requestOzProxyGeneration' ) then
	
		headers[ #headers + 1 ] = {
			header = 'X-Generate-Proxies',
			value = table.concat( WFOzBinaryUploadUtils.getProxyTypes(), ',' ),
		}
	
	end

	return headers
	
end

--------------------------------------------------------------------------------

local function uploadMasterP( self, docValue )

	-- TODO_SNOWLESS [scouten 2013-06-21]: Provide support for upload via copy
	-- (for virtual copies). See function uploadAssetOriginalViaCopyP in
	-- OzCentralUploadTask.lua.

	binaryUploaderLog:debugf( "uploadMasterP docValue = %s", docValue )

	return self._binaryProvider:getMasterP( docValue )
	
		:mapP( function( status, data, contentType )
		
			return handleBinaryResultsP {

				self = self,
				docValue = docValue,
				status = status,
				data = data,
				bucket = 'master',

				infoPath = { 'master' },
				
				contentType = contentType or 'image/jpeg',

				requestHeaders = getMasterRequestHeaders(),
				
				putUrlTemplate = '%s/master',
			
			}

		end )

end

--------------------------------------------------------------------------------

local function makeUploadOneBinaryFunc( self )

	return function( uploadQueryRow )

		local assetId = uploadQueryRow[ 1 ]
		local binaryType = uploadQueryRow[ 2 ]
		local renditionType = uploadQueryRow[ 3 ]
		
		binaryUploaderLog:infof( 'uploadOneP asset = %s, binaryType = %s, renditionType = %s',
								  assetId, binaryType, renditionType )
	
		local uploadsInProgress = self._uploadsInProgress
		local existingUploadCount = ( uploadsInProgress[ assetId ] or 0 )
		uploadsInProgress[ assetId ] = existingUploadCount + 1
		
		if existingUploadCount == 0 then
			self._uploadStatusBPerAsset:postToEvent( assetId, true )
		end
		
		return self._docStore:docValueP( assetId )
		
			:mapDataP( function( status, docValue )
				
				if binaryType == 'rendition' then
					return uploadOneRenditionP( self, docValue, renditionType )
				elseif binaryType == 'proxy' then
					return uploadOneProxyP( self, docValue, renditionType )
				else
					return uploadMasterP( self, docValue )
				end
			
			end )
			:observeP( function( status, result )
			
				-- Decrement uploads-in-progress count for this asset.
				-- Zero count elides to nil so we don't take up space in
				-- the table.
	
				local existingUploadCount = ( uploadsInProgress[ assetId ] or 1 )
				
				if existingUploadCount <= 1 then
					uploadsInProgress[ assetId ] = nil
					self._uploadStatusBPerAsset:postToEvent( assetId, false )
				else
					uploadsInProgress[ assetId ] = existingUploadCount - 1
				end
				
				local uploadStatusBinName = binaryType == 'master' and 'master'
											or binaryType .. '/' .. renditionType
				
				local uploadStatusBin = self._uploadStatusBins[ uploadStatusBinName ]
				if not uploadStatusBin then
					
					uploadStatusBin = {
						nUploads = 0,
						nFails = 0,
					}
					
				 	self._uploadStatusBins[ uploadStatusBinName ] = uploadStatusBin

				end
					
				uploadStatusBin.nUploads = uploadStatusBin.nUploads + 1
					
				if status ~= 'data' then
					
					uploadStatusBin.nFails = uploadStatusBin.nFails + 1
					
					local errorLog = self._uploadErrorLog
					
					errorLog[ #errorLog + 1 ] = {
							
						time = require 'KSDate'.timeInIso8601FormatWithTimeZone(),
						
						id = assetId,
						binaryType = binaryType,
						renditionType = renditionType,
						error = result,
					
					}
					
					if #errorLog > 10 then
						table.remove( errorLog, 1 )
					end
					
				end
				
			end )
	
	end

end


--============================================================================--

-- @mark WFOzBinaryUploader class

--------------------------------------------------------------------------------
--[[!
	\class WFOzBinaryUploader
	
	\brief A binary uploader is responsible for uploading binary content
	(renditions, proxies, and master images) to Oz.
	
	It can do this either automatically or manually, as requested via APIs.
]]--

local binaryUploaderMethods = {

	_catalogSpec = false,
		-- (WFOzCatalogSpec) Oz catalog specification
	
	_docStore = false,
		-- (WFDocStore)
	
	_indexStore = false,
		-- (WFIndexStore)
	
	_binaryProvider = false,
		-- (object) this object provides binary data upon request
	
	_uploadsInProgress = false,
		-- (table)
			-- key (string): asset ID
			-- value (number): number of uploads currently in progress for this asset
	
	_uploadStatusBPerAsset = false,
		-- (KSReactive.makeWeakEventTable) what photos are uploading now
	
	_workScheduler = false,
		-- (WFQueryDrivenWorkQueueScheduler)
	
	_uploadQueueOptions = false,
		-- options to control the upload queue behavior
		
	_uploadOneBinaryFn = false,
		-- (function) upload a single binary object (used primarily for testing)
		
	_uploadStatusBins = false,
		-- (table) record of upload status/counts for each binary rendition type
			
	_uploadErrorLog = false,
		-- (table) record of recent upload errors for the diagnostic log
	
	_uploadSpeedStats = false,
		-- (table) record of recent upload speeds
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs an Oz binary uploader object.
	
	\param args (table) table containing the following arguments, all of which
		are required

		<ul>
			<li><code>catalogSpec</code>: (\c WFOzCatalogSpec) specifies the target catalog</li>
			<li><code>docStore</code>: (\c WFDocStore) local doc store for catalog</li>
			<li><code>indexStore</code>: (\c WFIndexStore) index store for catalog</li>
			<li><code>binaryProvider</code>: binary provider for peer
		</ul>
]]--

function binaryUploaderMethods:__init( args )

	check( 'WFOzBinaryUploader', 'table' )
	
	self._catalogSpec = assert( args.catalogSpec )
	assert( WFOzCatalogSpec:isClassOf( self._catalogSpec ) )
	
	self._docStore = assert( args.docStore )
	assert( WFDocStore:isClassOf( self._docStore ) )

	self._indexStore = assert( args.indexStore )
	assert( WFIndexStore:isClassOf( self._indexStore ) )
	self._indexStore:addCloseTarget( self )

	self._binaryProvider = assert( args.binaryProvider )
	assert( type( self._binaryProvider ) == 'table' )
	
	local uploadSequence = args.uploadSequence or 'renditionsFirst'
	assert( uploadSequence == 'renditionsFirst'
		 or uploadSequence == 'photosFirst' )
	
	-- SPAM [matthewj 2014-01-29]: These options were added to allow Lr to try some performance
	-- troubleshooting/tuning via config flags by adjusting the concurrent tasks and the
	-- blackout window duration. uploadQueueOptions.concurrentTasks and .blackout can be
	-- passed in on creation to control how the WFQueryDrivenWorkQueueScheduler is configured.

	self._uploadQueueOptions = args.uploadQueueOptions or {}
	self._uploadsInProgress = {}
	self._uploadStatusBPerAsset = KSReactive.makeWeakEventTable( false )

	local uploadOneItemFn = makeUploadOneBinaryFunc( self )
	self._uploadOneBinaryFn = uploadOneItemFn
	
	-- Typically, we'll create our work scheduler right away,
	-- but in some test scenarios we want to control its creation.
	
	if not args.noScheduler then
		self:startUploading( uploadSequence )
	end
	
	self._uploadStatusBins = {}
	self._uploadErrorLog = {}

	self._uploadSpeedStats = {
		buckets = {},
		totalSize = 0,
	}
	
	WFCloseSupport.mustClose( self )
	
	WFDiagnosticLog.addLogProvider( self )
	
end

--------------------------------------------------------------------------------
--[[
	\brief Returns true if we are currently uploading binaries for a given asset.
	
	\param assetId (string) asset iD
	
	\return <code>B< boolean ></code> true if currently uploading binaries;
		false if not
]]--

function binaryUploaderMethods:currentlyUploadingBinariesForAssetB( assetId )

	local uploadStatusBPerAsset = self._uploadStatusBPerAsset

	local eventExistedAlready = uploadStatusBPerAsset:eventExists( assetId )

	local result = uploadStatusBPerAsset:behaviorB( assetId )
	
	if not eventExistedAlready then
	
		local uploadsInProgress = self._uploadsInProgress[ assetId ]

		if uploadsInProgress and uploadsInProgress > 0 then
			uploadStatusBPerAsset:postToEvent( assetId, true )
		end
	
	end
	
	return result

end

--------------------------------------------------------------------------------
--[[
	\brief Upload a single binary file. Used primarily for testing.
	
	\param assetId (string) UUID of the asset
	\param binaryType (string) what type of binary should be uploaded
	\param renditionType (string) named subtype of rendition or proxy
	
	\return (promise) resolves when upload is complete
]]--

function binaryUploaderMethods:uploadOneBinaryP( assetId, binaryType, renditionType )

	check( 'WFOzBinaryUploader', '+string', '+string', '?string' )
	
	return self._uploadOneBinaryFn { assetId, binaryType, renditionType }

end

--------------------------------------------------------------------------------
--[[!
	\brief Start automatically uploading binaries. Intended only for testing.
]]--

function binaryUploaderMethods:startUploading( uploadSequence )

	if not self._workScheduler then
	
		local sql = WFOzBinaryUploadUtils.getUploadableAssetsSQL( uploadSequence )
		
		local queryB = self._indexStore:queryB( sql )
	
		self._workScheduler = WFQueryDrivenWorkQueueScheduler( queryB,
															   self._uploadQueueOptions.concurrentTasks or 2,
															   self._uploadQueueOptions.blackout or 60,
															   self._uploadOneBinaryFn )

	end

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------
--[[!
	\brief Sends diagnostic data on synchronization to the log.
]]--

function binaryUploaderMethods:sendDiagnosticLogData()

	local catalogSpec = self._catalogSpec
	local wfOzUser = catalogSpec:wfOzUser()
	
	local finalStats = {}
	
	for bucket, stats in pairs( self._uploadSpeedStats.buckets ) do
		finalStats[ bucket ] = {
			count = stats.count,
			timings = { stats.timings:visualize 'boxAndWhisker' },
			size = { stats.sizes:visualize 'boxAndWhisker' },
			speed = { stats.speed:visualize 'boxAndWhisker' },
		}
	end

	return {

		name = "WF Oz Binary Uploader",
		id = 'WFOzBinaryUploader',
		weight = 510,

		body = {
		
			ozApiServer = catalogSpec:serverBaseUrl(),
			ozCatalogId = catalogSpec:catalogId(),
			ozCatalogHref = catalogSpec:catalogHref(),
			
			ozUserId = wfOzUser:userId(),
			
			binaryRenditionStatus = self._uploadStatusBins,
			
			recentErrors = self._uploadErrorLog,
			
			uploadSpeedStats = finalStats,
			totalBinarySize = self._uploadSpeedStats.totalSize,
			
			-- TODO_SNOWLESS [scouten 2013-07-08]: Add upload time tracking.
			-- ([scouten 2014-01-14]: Is this different from what I've
			-- implemented with uploadSpeedStats?)
			
			-- TODO_SNOWLESS [scouten 2013-07-08]: Add memory usage tracking.
			-- See ENABLE_MEMORY_POLLING_FOR_METRICS in OzCentralUploadTask.lua.

		},
	
	}
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzBinaryUploader'( binaryUploaderMethods )
