--[[----------------------------------------------------------------------------

WFOzBinaryUploaderTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFCapabilities = require 'WFCapabilities'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.sync_engine.binary_upload
local WFOzBinaryUploader = require 'WFOzBinaryUploader'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzCatalogTestSession = require 'WFOzCatalogTestSession'
local WFOzTestBinaryProviders = require 'WFOzTestBinaryProviders'


--============================================================================--

-- @mark WFOzBinaryUploaderTest

local WFOzBinaryUploaderTest = {}

--------------------------------------------------------------------------------

local TEST_WAIT_TIME = 240
local catalogTestSession, indexStore, wfOzSyncAgent, wfOzBinaryUploader
local fakePhotoIdsByCaptureDate

--------------------------------------------------------------------------------

function WFOzBinaryUploaderTest.setUp()

	WFCapabilities.setCapability( 'WFOzBinaryUploader.renditionTypes', 'thumbnail2x,2048' )
	WFCapabilities.setCapability( 'WFOzBinaryUploader.proxyTypes', '2560' )

	catalogTestSession = WFOzCatalogTestSession( 'WFOzBinaryUploaderTest' )

	indexStore = nil
	wfOzSyncAgent = nil
	wfOzBinaryUploader = nil

end

--------------------------------------------------------------------------------

function WFOzBinaryUploaderTest.tearDown( path )

	if indexStore then
		if path then
			indexStore:DEBUG_dumpCoverageP( path )
		end
		indexStore:close()
		indexStore = nil
	end

	if catalogTestSession then
		catalogTestSession:releaseTestUserP()
		catalogTestSession:close()
		catalogTestSession = nil
	end

end

--------------------------------------------------------------------------------

local function wfOzBinaryUploaderP( binaryProvider )

	indexStore = WFIndexStore( catalogTestSession:docStore(),
							   'BinaryUploaderTestIndexDefinitions.lua' )

	return catalogTestSession:ozSyncAgentWithMetadataSyncCompletedP()
		
		:mapDataP( function()
		
			return catalogTestSession:catalogSpecP()
		
		end )
		
		:mapDataP( function( _, catalogSpec )
		
			wfOzBinaryUploader = WFOzBinaryUploader
				{
					catalogSpec = catalogSpec,
					docStore = catalogTestSession:docStore(),
					indexStore = indexStore,
					binaryProvider = binaryProvider,
				}
			
			catalogTestSession:addCloseTarget( wfOzBinaryUploader )
			
			fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()
			
			return 'data', wfOzBinaryUploader
		
		end )

end

--------------------------------------------------------------------------------

local function stripRevisionAndSelfLinks( linksTable )

	linksTable[ "/rels/first_revision" ] = nil
	linksTable[ "/rels/latest_revision" ] = nil
	linksTable[ "self" ] = nil
	
	return linksTable

end

--------------------------------------------------------------------------------

local function mapStripAllButRenditionInfo( docValue )

	if type( docValue ) ~= 'table' or WFUnknown.isUnknown( docValue ) then
		return docValue
	end

	return
		{
			_localOnly = {
				renditions = KSTableUtils.derefOpt( docValue, '_localOnly', 'renditions' )
			},
			_serverProvidedMetadata = {
				links = stripRevisionAndSelfLinks( KSTableUtils.derefOpt( docValue,
									'_serverProvidedMetadata', 'links' ) or {} ),
			},
		},
		docValue

end

--------------------------------------------------------------------------------

function WFOzBinaryUploaderTest.testMissingRendition_testNightly( test )

	local missingBinaryProvider = WFOzTestBinaryProviders.missingBinaryProvider()
	catalogTestSession:setTestName( test:testName() )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the binary uploader. Now the wfOzBinaryUploader file variable
		-- will be defined.

		:waitFor( function()

			return wfOzBinaryUploaderP( missingBinaryProvider )

		end )

		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Wait a while.
		
		:waitFor( 60 )

		-- Confirm that missing value placeholders were written into
		-- all three documents.
		
		:waitFor( function()
			return catalogTestSession:docStore():docValueB(
				fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ] )
				:mapB( mapStripAllButRenditionInfo )
		end )
		:continueWhenValueMatches(
			{
				_localOnly = {
					renditions = {
						[ "thumbnail2x" ] = {
							missing = true,
						},
						[ "2048" ] = {
							missing = true,
						},
					},
				},
				_serverProvidedMetadata = {
					links = {},
				},
			}, ':ignore:' )
		
		:waitFor( function()
			return catalogTestSession:docStore():docValueB(
				fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ] )
				:mapB( mapStripAllButRenditionInfo )
		end )
		:continueWhenValueMatches(
			{
				_localOnly = {
					renditions = {
						[ "thumbnail2x" ] = {
							missing = true,
						},
						[ "2048" ] = {
							missing = true,
						},
					},
				},
				_serverProvidedMetadata = {
					links = {},
				},
			}, ':ignore:' )
		
		:waitFor( function()
			return catalogTestSession:docStore():docValueB(
				fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
				:mapB( mapStripAllButRenditionInfo )
		end )
		:continueWhenValueMatches(
			{
				_localOnly = {
					renditions = {
						[ "thumbnail2x" ] = {
							missing = true,
						},
						[ "2048" ] = {
							missing = true,
						},
					},
				},
				_serverProvidedMetadata = {
					links = {},
				},
			}, ':ignore:' )
		
		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

function WFOzBinaryUploaderTest.testBogusRendition_testNightly( test )

	local badContentBinaryProvider = WFOzTestBinaryProviders.badContentBinaryProvider()
	catalogTestSession:setTestName( test:testName() )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the binary uploader. Now the wfOzBinaryUploader file variable
		-- will be defined.
		
		:waitFor( function()

			return wfOzBinaryUploaderP( badContentBinaryProvider )

		end )

		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Wait a while.
		
		:waitFor( 60 )

		-- Confirm that error placeholders were written into
		-- all three documents.
		
		:waitFor( function()
			return catalogTestSession:docStore():docValueB(
				fakePhotoIdsByCaptureDate[ "2001-02-03T04:05:06" ] )
				:mapB( mapStripAllButRenditionInfo )
		end )
		:continueWhenValueMatches(
			{
				_localOnly = {
					renditions = {
						[ "thumbnail2x" ] = {
							error = "415",
						},
						[ "2048" ] = {
							error = "415",
						},
					},
				},
				_serverProvidedMetadata = {
					links = {},
				},
			}, ':ignore:' )
		
		:waitFor( function()
			return catalogTestSession:docStore():docValueB(
				fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ] )
				:mapB( mapStripAllButRenditionInfo )
		end )
		:continueWhenValueMatches(
			{
				_localOnly = {
					renditions = {
						[ "thumbnail2x" ] = {
							error = "415",
						},
						[ "2048" ] = {
							error = "415",
						},
					},
				},
				_serverProvidedMetadata = {
					links = {},
				},
			}, ':ignore:' )
		
		:waitFor( function()
			return catalogTestSession:docStore():docValueB(
				fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
				:mapB( mapStripAllButRenditionInfo )
		end )
		:continueWhenValueMatches(
			{
				_localOnly = {
					renditions = {
						[ "thumbnail2x" ] = {
							error = "415",
						},
						[ "2048" ] = {
							error = "415",
						},
					},
				},
				_serverProvidedMetadata = {
					links = {},
				},
			}, ':ignore:' )
		
		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

local function hasBinaryLinks( docValue )

	local links = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links' )
	if type( links ) ~= 'table' then return nil end
	
	for k, linksStanza in pairs( links ) do
	
		linksStanza.created_by = nil
		
		if linksStanza.href then
		
			if linksStanza.href:match( "^/v1%.0/catalogs/" ) then
				linksStanza.href = "valid href link"
			end
		
		end
	
	end
	
	return links

end

--------------------------------------------------------------------------------

function WFOzBinaryUploaderTest.testStartAutomaticSync_testNightly( test )

	local uploadingForTestAssetB
	catalogTestSession:setTestName( test:testName() )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Spin up the binary uploader. Now the wfOzBinaryUploader file variable
		-- will be defined.
		
		:waitFor( function()

			return wfOzBinaryUploaderP( WFOzTestBinaryProviders.testBinaryProvider() )

		end )
		
		-- Verify that we're not already reporting "currently uploading"
		-- status for the last of the photos to be uploaded.
		
		:waitFor( function()

			uploadingForTestAssetB
				= wfOzBinaryUploader:currentlyUploadingBinariesForAssetB(
					fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
			
			return uploadingForTestAssetB

		end )
		:continueWhenValueMatches( false )
		
		-- Turn on automatic sync.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', syncAgent

				end )
		
		end )

		-- Then we should show "currently uploading" for that photo.
		
		:waitFor( function()
		
			return uploadingForTestAssetB

		end )
		:continueWhenValueMatches( true )

		-- And, once uploads are complete, we should set that back to false.
		
		:waitFor( function()
		
			return uploadingForTestAssetB

		end )
		:continueWhenValueMatches( false )
	
		-- Just for completeness, let's look at the document value for one of the
		-- photos and ensure that it has all the correct links.

		:waitFor( function()
		
			return catalogTestSession:docStore()
				:docValueB( fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
				:mapB( hasBinaryLinks )
		
		end )
		:continueWhenValueMatches
			{
				[ "/rels/first_revision" ] = {
					href = "valid href link",
				},
				[ "/rels/latest_revision" ] = {
					href = "valid href link",
				},
				[ "self" ] = {
					href = "valid href link",
				},
				[ "/rels/rendition_type/2048" ] = {
					href = "valid href link",
				},
				[ "/rels/rendition_type/1280" ] = {
					href = "valid href link",
				},
				[ "/rels/rendition_type/thumbnail2x" ] = {
					href = "valid href link",
				},
				[ "/rels/rendition_type/640" ] = {
					href = "valid href link",
				},
				[ "/rels/proxy_type/2560" ] = {
					href = "valid href link",
				},
				[ "/rels/master" ] = {
					content_type = "image/jpeg",
					href = "valid href link",
				},
				["/rels/xmp"] = {
					href = "valid href link",
				},
			}

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function WFOzBinaryUploaderTest.testSendDiagnosticLogData_testNightly( test )

	-- Since diagnostic log is expected to evolve frequently, we're not
	-- verifying specific content. The main concern is that it doesn't fall
	-- over and die when generating the log.

	catalogTestSession:setTestName( test:testName() )

	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()

			return wfOzBinaryUploaderP( WFOzTestBinaryProviders.testBinaryProvider() )

		end )
		
		:waitFor( function()
		
			return WFDiagnosticLog.processLogSectionP(
				wfOzBinaryUploader:sendDiagnosticLogData() )
		
		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

local WFPlatformInfo = require 'WFPlatformInfo'

if WFPlatformInfo.isWindows then
	return {}
else
	return WFOzBinaryUploaderTest
end
