--[[----------------------------------------------------------------------------

WFOzCatalogIndexDefinitions.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFCapabilities = require 'WFCapabilities'
local WFDigest = require 'WFDigest'

	-- WF.oz_client.catalog
local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'
local WFOzContentTypes = require 'WFOzContentTypes'
local WFOzDocIndexUtils = require 'WFOzDocIndexUtils'

	-- WF.oz_client.index_store
local WFIndexStoreImp = require 'WFIndexStoreImp'
local WFIndexStoreSchema = require 'WFIndexStoreSchema'

	-- WF.oz_client.sync_agent
local WFOzSyncConstants = require 'WFOzSyncConstants'

	-- WF.oz_client.sync_engine.binary_upload
local WFOzBinaryUploadUtils = require 'WFOzBinaryUploadUtils'

	-- WF.oz_client.sync_engine.metadata_sync
local WFConflictResolver = require 'WFConflictResolver'
local WFOzAlbumAssetUploadUtils = require 'WFOzAlbumAssetUploadUtils'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFOzCatalogIndexDefinitions'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local derefOpt = assert( KSTableUtils.derefOpt )

--------------------------------------------------------------------------------

local INCLUDE_NO_TAG_CASE = false
local EXTRACT_HASH_TAGS = true
local GENERATE_FULL_PLACE_HIERARCHY = true

--============================================================================--

local tableDefinitions = {}
local indexDefinitions = {}

--------------------------------------------------------------------------------

local function assetHasThumbnail( docValue )

	-- Per discussion on bug #2942278, this filter should be applied to all
	-- indexes that drive user-visible content. (Note, for instance, that we
	-- exclude it from the debug-only upload status filter.)

	-- Exclude documents other than assets.
	-- WFDocStore uses type; OzDocStore uses _docType. We try to be compatible
	-- with both here.

	if docValue._docType then
		if docValue._docType ~= 'asset' then return false end
	else
		if docValue.type ~= 'asset' then return false end
	end
	
	-- Include if we have a local thumbnail.
	
	if derefOpt( docValue, '_localOnly', 'renditions', 'thumbnail', 'md5' ) then
		return true
	end
	
	-- Include if we can get a thumbnail from Oz.
	
	if derefOpt( docValue, '_serverProvidedMetadata', 'links',
				 '/rels/rendition_type/144h', 'href' )
	then
		return true
	end
	
	-- Include if we can get a 2xthumbnail from Oz.
	
	if derefOpt( docValue, '_serverProvidedMetadata', 'links',
				 '/rels/rendition_type/thumbnail2x', 'href' )
	then
		return true
	end
	
	-- No thumbnail? Pretend it doesn't exist yet.
	
	return false

end

--------------------------------------------------------------------------------

local coreInfoFromDocValue = WFOzDocIndexUtils.coreInfoFromDocValue

--------------------------------------------------------------------------------
-- @mark coreInfo
--------------------------------------------------------------------------------

tableDefinitions.coreInfo = {
	
	-- Instead of specifying CREATE TABLE SQL, in order to use
	-- niceToHave, you must specify the field names and let the index
	-- store generate the SQL
	tableName = "coreInfo",
	
	fields = {
		"aspectRatio", "captureDate", "displayDate", "timestamp",
		"copyTimestamp", "thumbnailMD5", "eventName", "eventNameLC",
		"favoritedCount", "importedBy", "contentType", "duration", "favoritedUsers", "hasOriginal" },
	
	COLLATE = {
		eventName = 'i18n_collate',
	},
	
	-- queryB won't send updates for changes that only affect niceToHave
	-- columns
	niceToHave = {
		"thumbnailMD5",
	},

	map = function( docValue, emit )

		if docValue._deleted or not assetHasThumbnail( docValue ) then return end

		local aspectRatio, thumbnailMD5, favoritedCount, favoritedUsers = coreInfoFromDocValue( docValue )

		local displayDate
		
		-- Capture date may have a "Z" appended to the end, signifying that the
		-- timestamp is in UTC (20 chars). Otherwise, the timestamp is treated as
		-- local time (19 chars). Either way, we just care about getting the
		-- YYYY-MM-DD from the beginning of the string.

		if docValue.captureDate and string.len( docValue.captureDate ) > 10 then
			displayDate = string.replace( string.sub( docValue.captureDate, 1, 10 ), ":", "-" )
		end

		local timestamp = derefOpt( docValue, 'importSource', 'importTimestamp' )

		local copyTimestamp = derefOpt( docValue, 'importSource', 'copyTimestamp' )
			
		local eventName = docValue.eventName or ''
		
		local importedBy = derefOpt( docValue, 'importSource', 'importedBy' ) or ''
		
		local contentTypeNumber = 0
		
		local contentType = derefOpt( docValue, 'importSource', 'contentType' )
			or derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'content_type' )
		if contentType then
			contentTypeNumber = WFOzContentTypes.getContentTypeNumberForContentType( contentType )
		else
			local subtype = derefOpt( docValue, '_serverProvidedMetadata', 'subtype' )
			contentTypeNumber = WFOzContentTypes.getContentTypeNumberRangeForSubType( subtype )
		end
		
		local hasOriginalPath = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
		local hasOriginal = hasOriginalPath ~= nil
		
		local durationRatio = derefOpt( docValue, 'video', 'duration' )
		local duration = nil
		if type( durationRatio ) == 'table' then
			duration = math.ceil( ( durationRatio[ 1 ] or 0 ) / ( durationRatio[ 2 ] or 1 ) )
		end
		
		if aspectRatio then
			emit( aspectRatio, docValue.captureDate, displayDate, timestamp,
				  copyTimestamp, thumbnailMD5, eventName, KSStringUtils.lower( eventName ),
				  favoritedCount, importedBy, contentTypeNumber, duration, favoritedUsers, hasOriginal )
		end
			
	end,

}

-- We do many extractions based on events plus a date range. This index covers
-- those cases.

indexDefinitions.coreInfo_eventDateTrackList =
	[[CREATE INDEX coreInfo_eventDateTrackList ON coreInfo( eventNameLC, displayDate )]]
	
-- We seed the date and event views based on extracting display date, event name,
-- and capture date. This index covers all of those elements. We shouldn't actually
-- need the lowercase event name since it is derivable but we include it here so
-- that we can ensure consistency.

indexDefinitions.coreInfo_eventDateTrackContents =
	[[CREATE INDEX coreInfo_eventDateTrackContents ON coreInfo( displayDate, eventName,
																eventNameLC, captureDate )]]

-- New index for imported by index.

indexDefinitions.coreInfo_importedBy =
	[[CREATE INDEX coreInfo_importedBy ON coreInfo( importedBy, captureDate, docId )]]
	
-- Index for track view "Added By" which sorts by timestamp

indexDefinitions.coreInfo_eventDateTrackContents =
	[[CREATE INDEX coreInfo_addedByTimestamp ON coreInfo( importedBy, timestamp )]]


--------------------------------------------------------------------------------
-- @mark coreMD5
--------------------------------------------------------------------------------

local WFOzRenditionServerUtils = require 'WFOzRenditionServerUtils'

WFOzRenditionServerUtils.filterIndexDefinitions( tableDefinitions, indexDefinitions )
	
--------------------------------------------------------------------------------
-- @mark selectedPhotos
--------------------------------------------------------------------------------

tableDefinitions.selectedPhotos = {

	sql = [[CREATE TABLE selectedPhotos( isSelected )]],

	map = function( docValue, emit )

		local isSelected = docValue
			and docValue._localOnly
			and docValue._localOnly.selected
			and not docValue._deleted

		if isSelected then
			emit( true )
		end
			
	end,

}


--------------------------------------------------------------------------------
-- @mark tags: keywords, people, places
--------------------------------------------------------------------------------

local function hashTagContentsPattern( illegal )

	return '[^%s' .. illegal .. '%c,\"][^%s' .. illegal .. '%c,]*'
		-- fix Halide "

end

--------------------------------------------------------------------------------

local function hasNumericPrefix( text )
	
	return text:match( "^[%d.,]*[%d]+" )
	
end
			
--------------------------------------------------------------------------------

tableDefinitions.tags = {

	sql = [[CREATE TABLE tags( tag, lcTag, tagKind )]],

	map = function( docValue, emit )

		if not assetHasThumbnail( docValue ) then return end

		if ( docValue._docType == 'asset' or docValue.type == 'asset') and not docValue._deleted then
		
			local cases = {
				keywords = 'k',
				people = 'p',
				-- [{'place', 'details', values=true}] = 'g'
			}
			
			local textSources = {
				'caption',	-- Caption first because it is more likely to be specific
				'eventName'
			}
			
			local function stripTrailingPunctuationAndWhiteSpace( str )
			
				if not str or type( str ) ~= 'string' or str == '' then
					return str
				end
				
				return str:match( '(.-)[%s%p%c]*$' )
			
			end
		
			local extracted = { } -- The set of extracted tags of various kinds
			
			local function emitTag( tagKindKey, tagName, tagKey )
			
				if not tagName then
					return
				end
				
				tagName = tostring( tagName )
			
				local tags = extracted[ tagKindKey ]
				
				if not tags then
					tags = { }
					extracted[ tagKindKey ] = tags
				end
				
				if tagKey then
					tagKey = tostring( tagKey )
				else
					tagKey = KSStringUtils.lower( tagName )
				end
				
				if not tags[ tagKey ] then
					tags[ tagKey ] = true
					emit( tagName, tagKey, tagKindKey )
				end
			
			end
			
			for subpath, tagKindKey in pairs( cases ) do
				if type( subpath ) == 'string' then
					subpath = { subpath }
				end
				
				local tags = derefOpt( docValue, unpack( subpath ) )
				
				if type( tags ) == 'table' and next( tags ) then
						-- protect against invalid data in document
						-- (e.g. a comma-delimited string instead of an array)
						
					for tagName, value in pairs( tags ) do

						if subpath.values then
							tagName = value
						end
						
						emitTag( tagKindKey, tostring( tagName ) )

					end
				
				end
				
			end
			
			-- Extract hash tags from captions and event names
			
			if EXTRACT_HASH_TAGS then
			
				local function extractPrefixedTags( prefix, illegal, tagKindKey, text )
				
					-- FIXME: We need a better hash tag extractor. I'm just disallowing a few characters.
						
					-- Disallow the case where we start with a double quote
				
					for hashTag in string.gmatch( text, prefix .. '(' .. hashTagContentsPattern( illegal ) .. ')' ) do
					
						-- Strip trailing punctuation
						
						hashTag = stripTrailingPunctuationAndWhiteSpace( hashTag )
						
						if hashTag == "" then
							-- Skip the empty string
						elseif hasNumericPrefix( hashTag ) then
							-- Skip numbers
						else
							emitTag( tagKindKey, hashTag )
						end
						
					end
					
					-- Handle the case where we start with a double quote. We now look for a matching quote
					-- and essentially allow everything in between other than control characters. Note that
					-- embedded hash tags also get extracted by the code above.
					
					for hashTag in string.gmatch( text, prefix .. '\"([^%c\"]-)\"' ) do
						-- " fix Halide syntax hiliting for unbalanced quote
						if hashTag == "" then
							-- skip the empty string
						else
							emitTag( tagKindKey, hashTag )
						end
					end
				
				end
				
				for _, path in ipairs( textSources ) do
			
					local text = docValue[ path ]
					
					if type( text ) == 'string' then
					
						extractPrefixedTags( '#', '#@+', 'k', text )
						
						extractPrefixedTags( '+', '#@+', 'p', text )

					end
					
				end
			
			end
			
			-- Generate no tag entries
			
			if INCLUDE_NO_TAG_CASE then
			
				for _, tagKindKey in pairs( cases ) do
					if not extracted[ tagKindKey ] then
						emit( "", "NO TAG", tagKindKey )
							-- using uppercase chars for the no-tag case
							-- to avoid possibility of conflicting with user
							-- tag by that value.
					end
				end
			
			end
			
			-- Generate place tags
			
			local function makePlaceTagFromDetails( details, n )
				-- Generate a tag name and key based on details[ 1 .. n ]
				
				-- Don't bother if the entry is empty
								
				if type( details[ n ] ) ~= 'string' then
					return nil, nil
				end
				
				-- Compact out bad values. The type test above basically reflects the fact
				-- that when compacted we would get the same thing as a previous value if
				-- this didn't match.
				
				local pieces = { }
				local pieceCount = 0
				
				for i = 1, n do
					local x = details[ i ]
					if type( x ) == 'string' and x ~= "" then
						pieceCount = pieceCount + 1
						pieces[ pieceCount ] = x
					end
				end
				
				-- The tag name will be based on these in reverse order. The tag key will
				-- be based on removing case info.
				
				local tagNamePieces = { }
				local tagKeyPieces = { }
				
				for i = 1, pieceCount do
					tagNamePieces[ i ] = pieces[ pieceCount + 1 - i ]
					tagKeyPieces[ i ] = KSStringUtils.lower( pieces[ i ] )
				end
				
				return table.concat( tagNamePieces, ', ' ), table.concat( tagKeyPieces, '<>' )
				
			end
			
			local function emitTagsForPlaceDetails( placeDetails )
			
				if type( placeDetails ) ~= 'table' then
					return
				end
				
				local n = table.maxn( placeDetails )
				
				if GENERATE_FULL_PLACE_HIERARCHY then
				
					for i = 1, n do
						local tagName, tagKey = makePlaceTagFromDetails( placeDetails, i )
						emitTag( 'g', tagName, tagKey )
					end
					
				else
				
					local tagName, tagKey = makePlaceTagFromDetails( placeDetails, n )
					emitTag( 'g', tagName, tagKey )

				end
			
			end
			
			emitTagsForPlaceDetails( docValue.place and docValue.place.details )
			
			-- If we didn't generate a place that way, try the text sources
			
			if EXTRACT_HASH_TAGS and not extracted[ 'g' ] then
			
				local string_match = string.match
			
				local function trimTrailingTags( text, prefix, illegal )
				
					if type( text ) ~= 'string' then
						return text
					end
					
					-- Find the first quoted hash tag and remove it
					
					local start = string.find( text, prefix .. '\"([^%c\"]-)\"' )
						-- Halide fix "
						
					if start then
						text = string.sub( text, 1, start - 1 )
					end
				
					-- Find the first hash tag if any and truncate there
					
					start = 1
					local finish
					
					local tagPattern = prefix .. hashTagContentsPattern( illegal )
					
					while true do
					
						start, finish = string.find( text, tagPattern, start )
							
						if not start then
							break
						end
						
						if hasNumericPrefix( string.sub( text, start + #prefix, finish ) ) then
							start = finish + 1
						else
							text = string.sub( text, 1, start - 1 )
							break
						end
						
					end
					
					-- Find an instance of the hash symbol followed by whitespace and
					-- truncate there
					
					start = string.find( text, prefix .. '%s' )
					
					if start then
						text = string.sub( text, 1, start - 1 )
					end
					
					return text
				
				end
			
				local function findPlaceName( text )
				
					if type( text ) ~= 'string' then
						return nil
					end
					
					-- Try looking for text surrounded by double-quotes preceded by an at sign
					-- and a space.
					
					-- GRUNGE: The space is only required below to avoid collision with Twitter
					-- at-tags which mean something totally different. It's required here for
					-- consistency. If we wanted to make it optional, we would shift to %s* in
					-- the two patterns below.
					
					local extracted = string_match( text, '@%s+\"([^%c\"]-)\"' )
						-- Fix Halide "
					
					if extracted then
						return extracted
					end
					
					-- Otherwise just look for a long run of text that starts with an at sign
					
					extracted = string_match( text, "@%s+([^%c]*)" )
						-- @
						-- One or more space characters
						-- Greedily grab characters other than control characters
						
					-- We stop at the first tag because this is probably a transition in the type
					-- of data.
						
					extracted = trimTrailingTags( extracted, '#', '#@+' )
					
					extracted = trimTrailingTags( extracted, '+', '#@+' )
						
					return stripTrailingPunctuationAndWhiteSpace( extracted )
						-- But then strip trailing punctuation and white space
				
				end
				
				local function findPlaceDetails( placeName )
				
					if type( placeName ) ~= 'string' then
						return nil
					end
					
					local placeDetails = { }
					
					local notDividerPattern = '%s*([^,]*[^%s,])'
						-- Skip whitespace
						-- Greedily grab non comma characters
						-- Make sure we can end with something that isn't a comma or a space
											
					for piece in string.gmatch( placeName, notDividerPattern ) do
						if piece == "" then
							-- Skip empty pieces
						else
							placeDetails[ #placeDetails + 1 ] = piece
						end
					end
					
					local n = #placeDetails
					
					if n == 0 then
						return nil
					end
					
					-- Now, we want to invert the place details
					
					local result = { }
					
					for i = 1, n do
						result[ i ] = placeDetails[ n + 1 - i ]
					end
					
					return result
				
				end
			
				for _, path in ipairs( textSources ) do
			
					local text = docValue[ path ]
					
					local placeName = findPlaceName( text )
					
					local placeDetails = findPlaceDetails( placeName )
					
					emitTagsForPlaceDetails( placeDetails )
					
					if extracted[ 'g' ] then
						break
					end
					
				end
			
			end
			
			-- Generate the no place tag if need be
			
			if INCLUDE_NO_TAG_CASE then
				if not extracted[ 'g' ] then
					emit( "", "NO TAG", 'g' )
				end
			end
		
		end
	
	end,

}

-- We need an index to find the photos for a keyword. This index may also help
-- with extracting unique keywords since it is sorted conveniently.

indexDefinitions.tags_index =
	[[CREATE INDEX tags_index ON tags( lcTag, tagKind )]]

--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- @mark ozUploadStatusForView
--------------------------------------------------------------------------------

tableDefinitions.ozUploadStatusForView =  {

	sql = [[CREATE TABLE ozUploadStatusForView( ozUploadStatus, docType, captureDate, deviceName, uploadStatusWithDeviceId )]],

	map = function( docValue, emit )

		logger:debugf( 'ozUploadStatusForView reindexing %s' )

		if ( docValue._docType == 'asset' or docValue.type == 'asset' )
		  and assetHasThumbnail( docValue ) and not docValue._deleted
		then

			local result = OzCatalogUploadStatus.calculateUploadStatus( docValue )
			
			local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.originalFilePath
				
			local localPath_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.filePath
				
			local isInCaches = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.isInCacheFolder
				
			local isInCaches_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.isInCacheFolder

			local ozOriginalUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href

			local ozThumbnailUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/144h" ]
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/144h" ].href
				
			local ozThumbnail2xUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/thumbnail2x" ]
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/thumbnail2x" ].href

			local ozPreviewUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/1024" ]
				and docValue._serverProvidedMetadata.links[ "/rels/rendition_type/1024" ].href
			
			-- Is this a new photo not uploaded or a changed photo?

			if WFOzSyncConstants.isNotUploaded( result ) then

				-- If we don't have any renditions uploaded, it is a new photo.

				if ozOriginalUrl and ozThumbnailUrl and ozThumbnail2xUrl and ozPreviewUrl then
					result = 'notUploadedChanged'
				else
					result = 'notUploadedNew'
				end

			end

			local deviceName = ''

			-- If the status is uploaded, check originals.

			if result == 'uploaded' then
	
				if not ozOriginalUrl then
					
					if docValue.importSource and docValue.importSource.originalStatus == 'lost' then
						result = 'originalLost'
					else
						result = 'uploadingOtherDevice'
						deviceName = ( docValue and docValue.importSource and docValue.importSource.importedOnDevice ) or 'unknown device'
					end

				else

					-- We need to check for originals (base and/or latest) that are either not present locally
					-- or present in caches so that the user can see them in the 'Downloading Originals' status
					-- view and know that these are not in robust storage yet.

					if OzCatalogUploadStatus.assetOriginalIsValid( docValue ) then
						if not localPath or ( localPath and isInCaches ) then
							result = 'downloadingOriginal'
						end
					else
						if ( not localPath_latestOriginal or ( localPath_latestOriginal and isInCaches_latestOriginal ) ) or
						( not localPath or ( localPath and isInCaches ) ) then
							result = 'downloadingOriginal'
						end
					end

				end

			end

			logger:debugf( "docValue = %s\n\nuploadStatus = %s\n\n",
						   docValue, result )

			-- Create a unique id for this row, status + device name.

			local uploadStatusWithDeviceId = result..deviceName
			
			if result ~= 'uploaded' and result ~= 'deleted' then
				emit( result, docValue._docType or docValue.type, docValue.captureDate, deviceName, uploadStatusWithDeviceId )
			end
		
		end

	end,

}

indexDefinitions.ozUploadStatusForView_trackContents =
	[[CREATE INDEX ozUploadStatusForView_trackContents ON ozUploadStatusForView(
					uploadStatusWithDeviceId, captureDate DESC, docId )]]
					
--------------------------------------------------------------------------------
-- @mark importSource
--------------------------------------------------------------------------------

tableDefinitions.importSource =  {

	sql = [[CREATE TABLE importSource( sha256, localUrl )]],

	map = function( docValue, emit )

		if ( docValue._docType == 'asset' or docValue.type == 'asset' ) and not docValue._deleted then
		
			local sha256 = docValue.importSource
								and docValue.importSource.sha256
			
			local localUrl = docValue._localOnly
								and docValue._localOnly.localUrl
			
			if sha256 or localUrl then
				emit( sha256, localUrl )
			end
		
		end

	end,

}

indexDefinitions.importSource_sha256 =
	[[CREATE INDEX importSource_sha256 ON importSource( sha256 )]]

indexDefinitions.importSource_localUrl =
	[[CREATE INDEX importSource_localUrl ON importSource( localUrl )]]

--------------------------------------------------------------------------------
-- @mark autoImportLog
--------------------------------------------------------------------------------

tableDefinitions.autoImportLog =  {

	sql = [[CREATE TABLE autoImportLog( sha256, localUrl )]],

	map = function( docValue, emit )

		if ( docValue._docType == 'autoImport' or docValue.type == 'autoImport' ) and not docValue._deleted then
		
			local sha256 = docValue._localOnly
					   and docValue._localOnly.sha256
			
			local localUrl = docValue._localOnly
					     and docValue._localOnly.localUrl
			
			if sha256 or localUrl then
				emit( sha256, localUrl )
			end
		
		end

	end,

}

indexDefinitions.autoImportLog_sha256 =
	[[CREATE INDEX autoImportLog_sha256 ON autoImportLog( sha256 )]]

indexDefinitions.autoImportLog_localUrl =
	[[CREATE INDEX autoImportLog_localUrl ON autoImportLog( localUrl )]]

--------------------------------------------------------------------------------
-- @mark cacheReferences
--------------------------------------------------------------------------------

-- provided by WFOzRenditionServerUtils.filterIndexDefinitions

--------------------------------------------------------------------------------
-- @mark nonOriginalsCacheReferences
--------------------------------------------------------------------------------

tableDefinitions.nonOriginalsCacheReferences = {

	sql = [[CREATE TABLE nonOriginalsCacheReferences( md5 )]],

	map = function( docValue, emit )
		
		if docValue._deleted or ( docValue._docType ~= 'asset' and docValue.type ~= 'asset' ) then
			return
		end
		
		local localOnly = docValue._localOnly
	
		local function emitForRenditionName( renditionName )
		
			local renditionInfo = derefOpt( localOnly, 'renditions', renditionName )
			if not renditionInfo then return end
			
			local md5 = renditionInfo.md5

			if md5 then
				emit( md5 )
			end
		
		end

		emitForRenditionName( 'thumbnail' )
		emitForRenditionName( 'thumbnail2x' )
		emitForRenditionName( 'preview' )

	end,

}

indexDefinitions.nonOriginalsCacheReferences_md5 =
	[[CREATE INDEX nonOriginalsCacheReferences_md5 ON nonOriginalsCacheReferences( md5 )]]

--------------------------------------------------------------------------------
-- @mark missingOriginals
--------------------------------------------------------------------------------

tableDefinitions.missingOriginals = {
	
	sql = [[CREATE TABLE missingOriginals( photoId, captureDate, missingInOz )]],
	
	map = function( docValue, emit )
	
		if ( docValue._docType == 'asset' or docValue.type == 'asset' ) and not docValue._deleted then
	
			local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.originalFilePath
								
			local localPath_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.filePath
				
			local ozUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
				
			local ozUrl_firstMaster = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/first_master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/first_master" ].href
				
			local isInCaches = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.isInCacheFolder
				
			local isInCaches_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.isInCacheFolder
				
			-- If we've tried to fetch this original and the request returns with a 404, mark this asset as missing its original in Oz
			-- and place it lower in the queue.
			
			local missingInOz_original = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.missingInOz and 1 or 0
				
			local missingInOz_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.missingInOz and 1 or 0
				
			-- In older documents, 'isInCacheFolder' can be missing.
			-- In that case if there is a local path, the original must be in the documents folder.
			
			-- If there are server urls for the original and (possibly) the latest original
			-- and no local paths or if these are in caches, then these are "missing originals."

			if localPath then
				logger:debugf( "scanning %s", docValue )
			end
			
			if OzCatalogUploadStatus.assetOriginalIsValid( docValue ) then
			
				-- If there have been no destructive edits and the original has not changed
				-- check if we have the original.
				
				if ozUrl and ( not localPath or isInCaches ) then
					logger:debugf( "flagged %s for download because original missing", docValue._id or docValue._docId )
					emit( docValue._id or docValue._docId, docValue.captureDate, missingInOz_original )
				end
				
			else
			
				-- If destructive edits have been made, check if we need either the base original
				-- or the latest original.
				
				if ( ozUrl and ( not localPath_latestOriginal or isInCaches_latestOriginal ) )
					or ( ozUrl_firstMaster and ( not localPath or isInCaches ) )
				then
					logger:debugf( "flagged %s for download because original out of date", docValue._id )
					emit( docValue._id or docValue._docId, docValue.captureDate, missingInOz_latestOriginal )
				end
				
			end
	
		end
	
	end,
	
}

indexDefinitions.missingOriginals_index =
	[[CREATE INDEX missingOriginals_index ON missingOriginals( missingInOz, captureDate, docId )]]

--------------------------------------------------------------------------------
-- @mark currentOriginalsInDocuments
--------------------------------------------------------------------------------

-- This index rolls up assets for which we have a current ( for some definition of current )
-- base original or latest original already downloaded and present in the documents folder.
-- This is used by WFOzPurgeAgent to move these originals from Documents to Caches if
-- 'Store Originals' is off

tableDefinitions.currentOriginalsInDocuments = {
	
	sql = [[CREATE TABLE currentOriginalsInDocuments( photoId, originalFilePath, latestOriginalFilePath )]],
	
	map = function( docValue, emit )
	
		if ( docValue._docType == 'asset' or docValue.type == 'asset' ) and not docValue._deleted then
	
			local localPath = docValue
				and docValue._localOnly
				and docValue._localOnly.originalFilePath
				
			local localPath_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.filePath
				
			local isInCaches = docValue
				and docValue._localOnly
				and docValue._localOnly.original
				and docValue._localOnly.original.isInCacheFolder
				
			local isInCaches_latestOriginal = docValue
				and docValue._localOnly
				and docValue._localOnly.latestOriginal
				and docValue._localOnly.latestOriginal.isInCacheFolder
				
			local ozUrl = docValue
				and docValue._serverProvidedMetadata
				and docValue._serverProvidedMetadata.links
				and docValue._serverProvidedMetadata.links[ "/rels/master" ]
				and docValue._serverProvidedMetadata.links[ "/rels/master" ].href
	
			-- Add a check to see if this original is valid ( current )
			-- If isInCaches is nil, that means this is an older document and
			-- we assume that the original is present in Documents
			
			local originalFilePath = nil
			local latestOriginalFilePath = nil
			
			if ozUrl and localPath_latestOriginal and OzCatalogUploadStatus.assetLatestOriginalIsValid( docValue ) and not isInCaches_latestOriginal then
				latestOriginalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath_latestOriginal )
			end
			-- We check for the master url and the first revision url here to ensure that the upload
			-- of this original has completed before moving it to Caches
			
			if ozUrl and localPath and not isInCaches then
				-- Temporary check to correct the local file path and replace
				-- 'Adobe Carousel Library.cr' with 'Adobe Revel Library.cr'
		
				-- FIXME [ngautam 2011-11-15] This should really be done via an upgrade process
				-- that updates the database with the correct file paths but this is not
				-- feasible in the current time frame.
				originalFilePath = OzCatalogUploadStatus.checkAndCorrectLocalPath( localPath )
			end
			if latestOriginalFilePath or originalFilePath then
				emit( docValue._id or docValue._docId, originalFilePath, latestOriginalFilePath )
			end
		end
	
	end,
	
}

indexDefinitions.currentOriginalsInDocuments_index =
	[[CREATE INDEX currentOriginalsInDocuments_index ON currentOriginalsInDocuments( photoId )]]

--------------------------------------------------------------------------------
-- @mark albums
--------------------------------------------------------------------------------

-- All albums tables now generated separately. (See WFOzAlbumAssetIndex.lua.)

--------------------------------------------------------------------------------
-- @mark Sharing
--------------------------------------------------------------------------------

tableDefinitions.space_asset = {
	
	-- Instead of specifying CREATE TABLE SQL, in order to use
	-- niceToHave, you must specify the field names and let the index
	-- store generate the SQL
	tableName = "space_asset",
	
	fields = {
		"spaceId", "assetId", "updateTimeStamp" },

	map = function( docValue, emit )

		if docValue and not docValue._deleted
		  and ( docValue._docType == 'space_asset' or docValue.type == 'space_asset' )
		then

			local assetId = docValue.asset and docValue.asset._id
			local spaceId = docValue.space and docValue.space._id
			local updateTimeStamp = docValue.updated
			
			if spaceId and assetId then
				
				emit( spaceId, assetId, updateTimeStamp )
				
			end
				
		end
	end,

}

--------------------------------------------------------------------------------

tableDefinitions.space_album = {

	sql = [[CREATE TABLE space_album( spaceId, albumId, updateTimeStamp )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted
		  and ( docValue._docType == 'space_album' or docValue.type == 'space_album' )
		then

			local albumId = docValue.album and docValue.album._id
			local spaceId = docValue.space and docValue.space._id
			local updateTimeStamp = docValue.updated
			
			if spaceId and albumId then
				
				emit( spaceId, albumId, updateTimeStamp )
				
			end
				
		end
					
	end,

}

indexDefinitions.space_album_index =
	[[CREATE INDEX space_album_index ON space_album( spaceId, albumId )]]
	
--------------------------------------------------------------------------------

tableDefinitions.space_user = {

	sql = [[CREATE TABLE space_user( spaceId, userId, userEmail )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted
		  and ( docValue._docType == 'space_user' or docValue.type == 'space_user' )
		then
			
			-- All users will have an associated email, users in the Revel system
			-- will have an ID as well.
				
			local spaceId = docValue.space and docValue.space._id
			local userId = docValue.user and docValue.user._id
			local userEmail = docValue.user and docValue.user.email
			
			if spaceId and userEmail then
				
				emit( spaceId, userId, userEmail )
			
			end
				
		end
					
	end,

}

indexDefinitions.space_user_index =
	[[CREATE INDEX space_user_index ON space_user( spaceId, userId, userEmail )]]

--------------------------------------------------------------------------------
-- @mark Comments
--------------------------------------------------------------------------------

tableDefinitions.comment = {

	sql = [[CREATE TABLE comment( assetId, spaceId, body, created, author_first_name, author_last_name, author_full_name, author_id, author_links_hash )]],

	map = function( docValue, emit )

		if docValue and not docValue._deleted
		  and ( docValue._docType == 'comment' or docValue.type == 'comment' )
		then
			
			local localOnly = docValue._localOnly or {}
			local spaceId = ( docValue.space and docValue.space._id ) or ( localOnly.spaceId )
			local assetId = ( docValue.asset and docValue.asset._id ) or ( localOnly.assetId )
			local created = docValue.created or derefOpt( docValue, '_serverProvidedMetadata', 'created' ) or  ( localOnly.created )
				-- check _localOnly.created, to allow indexing of comments that we've created locally
				-- but haven't yet gotten synced back from Oz

			local body = docValue.body
			local authorInfo = docValue.author or ( localOnly.author )
			local author_first_name, author_last_name, author_full_name, author_id, author_links_hash
			if authorInfo then
				author_first_name, author_last_name, author_full_name, author_id =
					authorInfo.first_name,
					authorInfo.last_name,
					authorInfo.full_name,
					authorInfo._id
				
				author_links_hash = WFDigest.SHA1.digest( authorInfo.links and
															WFTableArchiving.tableToString( authorInfo.links )
															or "nil" )
			end
			
			if assetId and created then
				
				if not spaceId then
					
					spaceId = WFOzSyncConstants.kGlobalSpace
				
				end
				
				emit( assetId, spaceId, body, created, author_first_name, author_last_name, author_full_name, author_id, author_links_hash )
				
			end
				
		end
					
	end,

}

indexDefinitions.comment_asset_space =
	[[CREATE INDEX comment_asset_space ON comment( assetId, spaceId )]]

--------------------------------------------------------------------------------
-- @mark Space Favorites
--------------------------------------------------------------------------------

tableDefinitions.spacefavorite = {
	
	sql = [[CREATE TABLE spacefavorite( assetId, spaceId, created, author_first_name, author_last_name, author_full_name, author_id, author_links_hash )]],
	
	map = function( docValue, emit )
		
		if docValue and not docValue._deleted
		  and ( docValue._docType == 'spacefavorite' or docValue.type == 'spacefavorite' )
		then
			local localOnly = docValue._localOnly or {}
			
			local spaceId = ( docValue.space and docValue.space._id ) or ( localOnly.spaceId )
			local assetId = ( docValue.asset and docValue.asset._id ) or ( localOnly.assetId )
			local created = docValue.created or ( localOnly.created ) or derefOpt( docValue, '_serverProvidedMetadata', 'created' )
			
			local authorInfo = docValue.author or ( localOnly.author )
			local author_first_name, author_last_name, author_full_name, author_id, author_links_hash
			if authorInfo then
				author_first_name, author_last_name, author_full_name, author_id =
					authorInfo.first_name,
					authorInfo.last_name,
					authorInfo.full_name,
					authorInfo._id
				
				author_links_hash = WFDigest.SHA1.digest( authorInfo.links and
															WFTableArchiving.tableToString( authorInfo.links )
															or "nil" )
			end
			
			emit( assetId, spaceId, created, author_first_name, author_last_name, author_full_name, author_id, author_links_hash )
		
		end
		
	end,

}

indexDefinitions.favorite_asset_space =
	[[CREATE INDEX favorite_asset_space ON spacefavorite( assetId, spaceId )]]


--==============================================================================

local indexStoreSchema = WFIndexStoreSchema()

require 'WFOzAlbumAssetIndex'.filterIndexDefinitions( tableDefinitions, indexDefinitions )
require 'WFOzSyncStatus'.filterIndexDefinitions( tableDefinitions, indexDefinitions )

tableDefinitions = WFCapabilities.callFilters( "WFOzCatalogIndexDefinitions.tableDefinitions", tableDefinitions )

for k, definition in pairs( tableDefinitions ) do
	indexStoreSchema:addTable( definition )
end

--------------------------------------------------------------------------------

indexDefinitions = WFCapabilities.callFilters( "WFOzCatalogIndexDefinitions.indexDefinitions", indexDefinitions )

for k, definition in pairs( indexDefinitions ) do
	indexStoreSchema:addIndex( definition )
end

--------------------------------------------------------------------------------

WFCapabilities.callFilters( "WFOzCatalogIndexDefinitions.modifyIndexSchema", indexStoreSchema )

--------------------------------------------------------------------------------

WFConflictResolver.generateIndexSchema( indexStoreSchema )
WFOzAlbumAssetUploadUtils.generateIndexSchema( indexStoreSchema )
WFOzBinaryUploadUtils.generateIndexSchema( indexStoreSchema )

--------------------------------------------------------------------------------

local function initMessageProcessor( ... )
	WFIndexStoreImp.start( indexStoreSchema, ... )
end

--============================================================================--

return {
	test_indexStoreSchema = indexStoreSchema,
	initMessageProcessor = initMessageProcessor,
	INCLUDE_NO_TAG_CASE = INCLUDE_NO_TAG_CASE,
}
