--[[----------------------------------------------------------------------------

WFOzCatalogSpec.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFArguments = require 'WFArguments'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFOzCatalogSpec class

--------------------------------------------------------------------------------
--[[!
	\class WFOzCatalogSpec
	
	\brief A \c WFOzCatalogSpec gathers together very basic information about an Oz
		catalog so that other classes that do the work of communicating with that
		catalog can share a common vocabulary for accessing the catalog.
	
	Please help keep this class simple and STATELESS!
]]--

local ozCatalogMethods = {

	_catalogId = false,
		-- (string) UUID for the catalog
	
	_baseHref = false,
		-- (string) relative HREF for the Oz API
		
	_catalogHref = false,
		-- (string) relative HREF for the catalog
	
	_ozUser = false,
		-- (WFOzUser) user account for the catalog

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the catalog.
	
	\param ozUser (WFOzUser) user account for the catalog
	\param catalogId (string) relative HREF or UUID for the catalog
]]--

function ozCatalogMethods:__init( ozUser, catalogId )

	check( 'WFOzCatalogSpec', 'WFOzUser', '+string' )

	self._baseHref = "/v1.0"

	if catalogId:match( '^/' ) then

		self._catalogHref = catalogId
		self._catalogId = catalogId:sub( -32 )
	
	else
		
		self._catalogHref = self._baseHref .. "/catalogs/" .. catalogId
		self._catalogId = catalogId
	
	end
	
	if not self._catalogHref:match( "^/v[0-9.]+/catalogs/[0-9a-f]+$" ) then
		error( "WFOzCatalogSpec: catalog HREF malformed: " .. self._catalogHref, 4 )
	end
	
	if #self._catalogId ~= 32 or
		not self._catalogId:match( "^[0-9a-f]+$" )
	then
		error( "WFOzCatalogSpec: catalog ID malformed: " .. self._catalogId, 4 )
	end
	
	self._ozUser = ozUser

end

--------------------------------------------------------------------------------

function ozCatalogMethods:__tostring()

	return string.format( "WFOzCatalogSpec( id = %s, user = %s )",
						  self._catalogId,
						  tostring( self._ozUser ) )

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the base URL for the server.
	
	\return (string) URL for the catalog
]]--

function ozCatalogMethods:serverBaseUrl()

	check( 'WFOzCatalogSpec' )

	return self._ozUser:serverBaseUrl()

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the UUID for the catalog.
	
	\return (string) UUID for the catalog
]]--

function ozCatalogMethods:catalogId()

	check( 'WFOzCatalogSpec' )

	return self._catalogId

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the base URL for OZ api.
	
	\return (string) base URL for the OZ api (i.e. "/v1.0/")
]]--

function ozCatalogMethods:baseHref()

	check( 'WFOzCatalogSpec' )

	return self._baseHref

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the base URL for the catalog.
	
	\return (string) base URL for the catalog (i.e. "/v1.0/catalogs/(uuid)")
]]--

function ozCatalogMethods:catalogHref()

	check( 'WFOzCatalogSpec' )

	return self._catalogHref

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the user's Oz ID.
]]--

function ozCatalogMethods:ozUserId()
	
	check( 'WFOzCatalogSpec' )

	return self._ozUser:userId()

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the WFOzUser object for this user.
]]--

function ozCatalogMethods:wfOzUser()
	
	check( 'WFOzCatalogSpec' )

	return self._ozUser

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP GET request, signing the request with the current
		access token of the current user and handles refreshing the access
		token if needed. Returns the response via a promise.

	\param url (string) The URL to fetch. If it is a server-relative URL
		(e.g. "/v1.0/..."), the URL will be assumed to be relative to the core
		server URL for this user.
	\param optArgs (optional, table) Options to modify the request.
	\param optSequencing (optional, string) 'LIFO' or 'FIFO'. FIFO by default.

	\return <code>P< data, headers ></code> A Promise that resolves to the response
		from HTTP server

	\sa WFHttpClient.getP
]]--

function ozCatalogMethods:getP( url, optArgs, optSequencing )

	check( 'WFOzCatalogSpec', '+string', '?table', '?string' )

	return self._ozUser:getP( url, optArgs, optSequencing )

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP PUT request, signing the request with the current
		access token of the current user and handles refreshing the access
		token if needed. Returns the response via a promise.

	\param url (string) The URL to fetch. If it is a server-relative URL
		(e.g. "/v1.0/..."), the URL will be assumed to be relative to the core
		server URL for this user.
	\param data (\c KSData or table) data to send with the request
		(use a Lua table for form data)
	\param optArgs (optional, table) options to modify the request

	\return <code>P< data, headers ></code> A Promise that resolves to the response
		from HTTP server
	
	\sa HttpClient.putP
]]--

function ozCatalogMethods:putP( url, data, optArgs )

	check( 'WFOzCatalogSpec', '+string', '_', '?table' )

	return self._ozUser:putP( url, data, optArgs )

end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzCatalogSpec'( ozCatalogMethods )
