--[[----------------------------------------------------------------------------

WFOzCatalogTestSession.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.lua.threading
local WFTimer = require 'WFTimer'

	-- WF.core.os
local KSKeychainService = require 'KSKeychainService'
local KSPrefs = require 'KSPrefs'
local WFFileUtils = require 'WFFileUtils'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.rxlua
local WFRxPromiseBridge = require 'WFRxPromiseBridge'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFDebugTraceback = require 'WFDebugTraceback'
local WFTableMerge = require 'WFTableMerge'

	-- WF.oz.client.account
local WFOzUserTestSession = require 'WFOzUserTestSession'

	-- WF.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.sync_engine
local WFOzCatalogSpec = require 'WFOzCatalogSpec'
local WFOzSyncPeer = require 'WFOzSyncPeer'
local WFOzUser = require 'WFOzUser'
local WFOzUserSyncPeer = require 'WFOzUserSyncPeer'
local WFSyncAgent = require 'WFSyncAgent'

	-- WF.web.json
local KSJsonEncoder = require 'KSJsonEncoder'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local asKSPromise = WFRxPromiseBridge


--============================================================================--

-- @mark WFOzCatalogTestSession class

--------------------------------------------------------------------------------
--[[
	\class WFOzCatalogTestSession

	A \c WFOzCatalogTestSession is testing infrastructure code that sets up
	and destroys (when done) a temporary catalog on the debug server ("ozstage").
]]--

local WFOzCatalogTestSession = {

	_traceback = false,
		-- (WFDebugTraceback) who created this object?

	_testName = false,
		-- (string) name of test we're running, if available

	_userTestSession = false,
		-- (WFOzUserTestSession)

	_docStore = false,
		-- (WFDocStore)

	_ozUserIdP = false,
		-- (P< string >) will resolve to an Oz user ID when ready

	_wfOzUserP = false,
		-- (P< WFOzUser >) will resolve to a WFOzUser object when ready

	_wfOzCatalogSpecP = false,
		-- (P< WFOzCatalogSpec >) will resolve to a WFOzCatalogSpec object when ready

	_wfOzCatalogSpecWithPhotosP = false,
		-- (P< WFOzCatalogSpec >) will resolve to a WFOzCatalogSpec object when ready
		-- (same as above, but with photos imported)

	_wfOzSyncAgentP = false,
		-- (P< WFSyncAgent >) will resolve to a WFSyncAgent object configured to
		-- talk to Oz when ready

	_wfOzSyncAgentWithMetadataSyncCompletedP = false,
		-- (P< WFSyncAgent >) same as above, but having waited for metadata
		-- sync round trip to complete

	_wfOzUserSyncAgentP = false,
		-- (P< WFSyncAgent >) will resolve to a WFSyncAgent object configured to
		-- talk to the Oz user account when ready

	_wfOzUser = false,
		-- (WFOzUser) designated user for the designated test object class

	_fakePhotoIdsByCaptureDate = false,
		-- (table) mapping designated capture date of test photos to their
		-- runtime-generated IDs

	_userInfoSyncAgent = false,
		-- (WFSyncAgent) the sync agent for the user info cache

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:__init( testName )

	check( 'WFOzCatalogTestSession', 'string' )

	self._traceback = WFDebugTraceback.traceback
						{ 'WFOzCatalogTestSession.lua', 'Class.lua' }

	self._testName = testName

	self:setOwned { _docStore = WFDocStore( ':memory:' ) }

	self._userTestSession = WFOzUserTestSession( testName )

	self._fakePhotoIdsByCaptureDate = {}

	WFCloseSupport.mustClose( self )

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:close()

	if self._ozUserIdP
		or self._wfOzCatalogSpecP
		or self._wfOzCatalogSpecWithPhotosP
		or self._wfOzSyncAgentP
		or self._wfOzSyncAgentWithMetadataSyncCompletedP
	then

		self:releaseTestUserP()

		error( "WFOzCatalogTestSession object missed call to :releaseTestUserP\n\ncreated at "
			   .. tostring( self._traceback ) )

	end

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:setTestName( testName )

	check( 'WFOzCatalogTestSession', '+string' )

	self._testName = testName

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:docStore()

	return self._docStore

end

--------------------------------------------------------------------------------
--[[!
	\brief Acquire a temporary user account for use with this test catalog.
	
	\return <code>P< OzUser ></code> a user account that has been logged in
]]--

function WFOzCatalogTestSession:ozUserIdP()

	if not self._ozUserIdP then

		self._ozUserIdP = self._userTestSession:getCleanOzAccount()
			:call( asKSPromise )
		
		self._ozUserIdP:whenResolvedDo( function() end ):attachTo( self )
	
	end
	
	return self._ozUserIdP

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:releaseTestUserP()

	if self._ozUserIdP then

		local resultP

		if self._userInfoSyncAgent then

			-- Wait for sync to be fully stopped.

			resultP = self._userInfoSyncAgent:stopAutomaticSyncP()
				:mapDataP( function()
					self._userInfoSyncAgent:close()
					self._userInfoSyncAgent = false
				end )

		else

			resultP = KSPromises.resolvedPromiseP( 'data', true )

		end

		-- Once we have the user info sync shut down, close the rest
		-- of this object's infrastructure.

		resultP = resultP
			:mapDataP( function()

				self._ozUserIdP = false
				self._wfOzCatalogSpecP = false
				self._wfOzCatalogSpecWithPhotosP = false
				self._wfOzSyncAgentP = false
				self._wfOzSyncAgentWithMetadataSyncCompletedP = false

				local ozUserPath
				local userId
				if self._wfOzUser then
					userId = self._wfOzUser:userId()
					ozUserPath = self._wfOzUser:userInfoCachePath()
				end

				self:close()

				return 'data', userId, ozUserPath

			end )
			:mapDataP( function( status, userId, ozUserPath )

				return self._userTestSession:releaseUser()
					:call( asKSPromise )
					:mapDataP( function()

						return 'data', userId, ozUserPath

					end )

			end )
			:mapDataP( function( status, userId, ozUserPath )
				
				if ozUserPath then
					WFFileUtils.delete( ozUserPath )
					WFFileUtils.delete( ozUserPath .. '-wal' )
					WFFileUtils.delete( ozUserPath .. '-shm' )
				end
				
				if WFPlatformInfo.isWindows then
					local KSPrefsImpLua = require 'KSPrefsImpLua'
					KSPrefsImpLua.deletePrefsFile()
				end

				if userId then
					return KSKeychainService.deleteP( userId )
				end

			end )

		return resultP:runUntilDoneP()

	else

		return KSPromises.resolvedPromiseP( 'data', true )

	end

end

--------------------------------------------------------------------------------

local function startAutoSyncingUserInfoCache( self, wfOzUser )

	KSPrefs.currentUserId = nil

	if self._userInfoSyncAgent then
		self._userInfoSyncAgent:stopAutomaticSync()
		self._userInfoSyncAgent:close()
		self._userInfoSyncAgent = false
	end

	local currentUserSyncPeer = WFOzUserSyncPeer( wfOzUser )
	local docStore = wfOzUser:userInfoCache()

	self._userInfoSyncAgent = WFSyncAgent( docStore, currentUserSyncPeer, 'ozUser' )

	self._userInfoSyncAgent:startAutomaticSync()
	wfOzUser:addCloseTarget( self._userInfoSyncAgent )

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:wfOzUserP()

	if self._wfOzUserP then
		return self._wfOzUserP
	end

	-- Spin up the basic user infrastructure.

	self._wfOzUserP = self:ozUserIdP()

		:mapDataP( function( _, ozUserId )

			-- Set up WFOzUser object for this test user account.

			self:setOwned { _wfOzUser = WFOzUser( OzCatalogUrls.baseServerUrl, ozUserId ) }

			return 'data', self._wfOzUser

		end )

	return self._wfOzUserP

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:catalogSpecP()

	if self._wfOzCatalogSpecP then
		return self._wfOzCatalogSpecP
	end

	-- Spin up the basic user infrastructure.

	local wfOzUserIdP = self:wfOzUserP()

		:mapDataP( function( _, wfOzUser )

			return startAutoSyncingUserInfoCache( self, wfOzUser )

		end )

	-- Now that we've done that, create a new test catalog.

	local catalogId = KSIdGenerator.uuidInLowercaseHex()

	local newCatalogP = wfOzUserIdP:mapDataP( function()

		local newCatalogUrl = string.format( '/v1.0/catalogs/%s', catalogId )

		local catalogData = {
			name = 'WFOzCatalogTestSession',
			subtype = 'lightroom',
		}

--		if args.subtype then
--			catalogData[ 'subtype' ] = args.subtype
--		end

		local newCatalogData = KSJsonEncoder.encodeOzUploadJson( catalogData )

		local newCatalogPutArgs = {
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
			},
		}

		return self._wfOzUser:putP( newCatalogUrl, newCatalogData, newCatalogPutArgs )
			:mapDataP( function( status, data, headers )

				assert( type( headers ) == 'table' )
				assert( headers.status_code == '201' )

				local location = headers.Location
				assert( type( location ) == 'string' )

				return 'data', location

			end )

	end )

	self._wfOzCatalogSpecP = newCatalogP
		:mapDataP( function()

			return 'data', WFOzCatalogSpec( self._wfOzUser, catalogId )

		end )

	self._wfOzCatalogSpecP:whenResolvedDo( function() end ):attachTo( self )

	return self._wfOzCatalogSpecP

end

--------------------------------------------------------------------------------

-- Note that we're intentionally skirting the issue of where renditions are
-- stored for now.

local baseImageInfo = {
	captureDate = "2004-11-02T11:10:09",
	height = 16,
	importSource = {
		fileCreationDate = "2004-11-02T08:37:24",
		fileName = "fakeFile01.jpg",
		fileSize = 13493,
		folder = "/fake/path/to/fake",
		folderPath = "/fake/path/to/fake",
		importTimestamp = "2013-06-17T15:06:46.643",
		--importedBy = kDesignatedTestUserId,
			-- filled in later (see addFakePhotoP)
		importedOnDevice = "WFOzCatalogTestSession.lua",
		sha256 = "36844083843640f92e1d477cca9b3df773c2512151bfe65dd7458a865e8206a1",
	},
	type = 'asset',
	version = 1,
	width = 16,
}

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:fakePhotoIdsByCaptureDate()

	return self._fakePhotoIdsByCaptureDate

end

--------------------------------------------------------------------------------

local function addFakePhotoP( self, uniqueInfo )

	return self:ozUserIdP()

		:mapDataP( function( _, ozUserId )

			local imageInfo = KSTableUtils.deepCopy( baseImageInfo )
			if uniqueInfo then
				WFTableMerge.applyOverlay( imageInfo, uniqueInfo )
			end

			imageInfo._id = KSIdGenerator.uuidInLowercaseHex()
			imageInfo._slot = 'main'
			imageInfo.importSource.importedBy = ozUserId

			self._fakePhotoIdsByCaptureDate[ imageInfo.captureDate ] = imageInfo._id

			return self._docStore:setDocValueP( imageInfo )

		end )

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:catalogSpecWithThreePhotosP()

	if self._wfOzCatalogSpecWithPhotosP then
		return self._wfOzCatalogSpecWithPhotosP
	end

	self._wfOzCatalogSpecWithPhotosP = self:catalogSpecP()
		:mapDataP( function( _, catalogSpec )
			return addFakePhotoP( self,
				{
					captureDate = "2004-11-02T11:10:09"
				} )
		end )
		:mapDataP( function()
			return addFakePhotoP( self,
				{
					captureDate = "2012-05-02T08:07:06"
				} )
		end )
		:mapDataP( function()
			return addFakePhotoP( self,
				{
					captureDate = "2001-02-03T04:05:06"
				} )
		end )
		:mapDataP( function()
			return self:catalogSpecP()
		end )

	return self._wfOzCatalogSpecWithPhotosP

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:ozSyncAgentP( options )

	if self._wfOzSyncAgentP then
		return self._wfOzSyncAgentP
	end

	self._wfOzSyncAgentP = self:catalogSpecWithThreePhotosP()
		:mapDataP( function( _, catalogSpec )

			local wfOzSyncPeer = WFOzSyncPeer( catalogSpec, options )

			local syncAgent = WFSyncAgent( self:docStore(),
										   wfOzSyncPeer,
										   'oz_TEST' )

			self:addCloseTarget( syncAgent )

			return 'data', syncAgent

		end )

	return self._wfOzSyncAgentP

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:ozSyncAgentWithMetadataSyncCompletedP( options )

	if self._wfOzSyncAgentWithMetadataSyncCompletedP then
		return self._wfOzSyncAgentWithMetadataSyncCompletedP
	end

	self._wfOzSyncAgentWithMetadataSyncCompletedP = self:ozSyncAgentP( options )
		:mapDataP( function( _, syncAgent )

			return syncAgent:sendUpdatesP()

				-- Wait for Oz to think about what we've told it.

				:mapDataP( function()

					return WFTimer.afterDelayP( 10 )

				end )
				:mapDataP( function()

					return syncAgent:getChangesP()

				end )

		end )

	return self._wfOzSyncAgentWithMetadataSyncCompletedP

end

--------------------------------------------------------------------------------

function WFOzCatalogTestSession:ozUserSyncAgentP( options )

	if self._wfOzUserSyncAgentP then
		return self._wfOzUserSyncAgentP
	end

	self._wfOzUserSyncAgentP = self:wfOzUserP()
		:mapDataP( function( _, wfOzUser )

			local wfOzUserSyncPeer = WFOzUserSyncPeer( wfOzUser )

			local wfDocStore = wfOzUser:userInfoCache()
			local syncAgent = WFSyncAgent( wfDocStore,
										   wfOzUserSyncPeer,
										   'ozUser_TEST' )

			self:addCloseTarget( syncAgent )

			return 'data', syncAgent

		end )

	return self._wfOzUserSyncAgentP

end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzCatalogTestSession'( WFOzCatalogTestSession )
