--[[----------------------------------------------------------------------------

WFOzFavorites.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFLogger = require "WFLogger"
local WFArguments = require "WFArguments"
local KSTableUtils = require "KSTableUtils"
local WFUnknown = require "WFUnknown"

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--==============================================================================

local favoritesLog = WFLogger 'WFOzFavorites'

-- @mark WFOzFavorites class

local WFOzFavorites = {
}

--------------------------------------------------------------------------------

-- @mark favorites

--------------------------------------------------------------------------------
local kFavoriteFields = "created, author_first_name, author_last_name, author_full_name, author_id, author_links_hash"
local nextIndex
do
	local index = 0
	nextIndex = function() index = index + 1 return index end
end

local kDocIdIndex				= nextIndex()
local kCreatedIndex				= nextIndex()
local kAuthorFirstNameIndex		= nextIndex()
local kAuthorLastNameIndex		= nextIndex()
local kAuthorFullNameIndex		= nextIndex()
local kAuthorIdIndex			= nextIndex()
local kAuthorLinksHashIndex 	= nextIndex()

local kFavoriteOnSpaceByAlbumClause = [[
		FROM
			spacefavorite
		WHERE
			spacefavorite.assetId == ? AND spaceId
		IN (
			SELECT
				spaceId
			FROM
				space_album
			WHERE
				space_album.albumId == ?
		)
		ORDER BY
			created]]

--------------------------------------------------------------------------------

local function formatFavorite( idRow )
	local favorite = {}

	favorite.id = idRow[ kDocIdIndex ]
	
	favorite.created = idRow[ kCreatedIndex ]
	
	favorite.author = {
		firstName = idRow[ kAuthorFirstNameIndex ],
		fullName = idRow[ kAuthorFullNameIndex ],
		lastName = idRow[ kAuthorLastNameIndex ],
		id = idRow[ kAuthorIdIndex ],
	}
	
	return favorite
end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves table of avatar urls associated with a favorite
	
	\param wfOzCatalog WFOzCatalog
	\param commentId (string) The unique ID of the favorite
	
	\return <code>B< table ></code> A promise object whose result is a table of
		avatar link URLs. This returns all avatar links currently referenced
		in the doc store.
]]--

function WFOzFavorites.getAvatarUrlsForFavoriteP( wfOzCatalog, favoriteId )
	
	WFArguments.check( 'WFOzCatalog', 'string' )
	
	return wfOzCatalog:docStore():docValueP( favoriteId )
			:mapDataP( function( status, favoriteDoc )
				
				local links
				if type( favoriteDoc ) == 'table' then
					links = KSTableUtils.derefOpt( favoriteDoc, 'author', 'links' )
				end
				
				return status, links
			end )

end

--------------------------------------------------------------------------------

function WFOzFavorites.countFavoritesOnAlbumAssetB( --[[WFOzCatalog]] wfOzCatalog, --[[string]] albumId, --[[string]] assetId )
	
	check( 'WFOzCatalog', 'uuid', 'uuid' )
	
	return wfOzCatalog:indexStore():queryB( [[
		SELECT
			COUNT( spacefavorite.docId ) ]] .. kFavoriteOnSpaceByAlbumClause, assetId, albumId )
	:closingMapB( function( idRows )
		local result = idRows and idRows[ 1 ]
		if result then
			return unpack( result, 1 )
		end
	end )
end

--------------------------------------------------------------------------------

function WFOzFavorites.getFavoritesOnAlbumAssetB( --[[WFOzCatalog]] wfOzCatalog, --[[string]] albumId, --[[string]] assetId )
	WFArguments.check( 'WFOzCatalog', 'string', 'string' )

	return wfOzCatalog:indexStore():queryB( [[
		SELECT
			spacefavorite.docId, ]] .. kFavoriteFields .. [[
		]] .. kFavoriteOnSpaceByAlbumClause, assetId, albumId )
	:closingMapB( function( idRows )
		if WFUnknown.isUnknown( idRows ) then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatFavorite( idRow )
		end
		return result
	end )

end

--==============================================================================

return WFOzFavorites
