--[[----------------------------------------------------------------------------

WFOzSyncPeer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua.threading
local WFIdle = require 'WFIdle'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFTableMerge = require 'WFTableMerge'

	-- WF.oz_client.sync_engine
local WFOzSyncUtils = require 'WFOzSyncUtils'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

	-- third_party.yajl
local yajl = require 'yajl'


--------------------------------------------------------------------------------

local uploadLogger = WFLogger 'WFOzSyncPeer :: upload'
--	:setLevel 'debug'

local downloadLogger = WFLogger 'WFOzSyncPeer :: download'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFOzSyncPeer class

--------------------------------------------------------------------------------
--[[!
	\class WFOzSyncPeer

	This class implements a sync peer that can be used for syncing between
    a WFDocStore and an Oz catalog.
]]--

local syncPeerMethods = {
	
	_catalogSpec = false,
		-- (WFOzCatalogSpec) Oz user and catalog references

	_suppressAlbumAssetUploads = false,
		-- (boolean) true to suppress album_asset uploads

	_currentHostErrorB = false,
	_hostIsReachableB = false,
	_forcedUpgradeLinkB = false,
	_minSyncDelayIntervalB = false,
	_maxSyncDelayIntervalB = false,
	_triggerSyncE = false,
	_rawUploadCompletedE = false,
		-- (see WFOzSyncUtils.syncPeerSharedInit)
	
	_prefetchNextChangesP = false,
		-- (P< JSON >) prefetch of next part of changes feed

	_testForceUpgradeLink = false,
		-- (boolean) true to trigger force upgrade test

	_retryAssetIds = false,
		-- (table) asset IDs that we may need to reload
	
	_deletedAlbumIds = false,
		-- (table) set of album IDs that have been deleted
	
	_deletedAssetIds = false,
		-- (table) set of asset IDs that have been deleted
	
	WFCloseSupport.addCloseSupportMethods,
	KSReactive.addReactiveSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the WF <-> Oz sync peer for the given user and catalog.
]]--

function syncPeerMethods:__init( catalogSpec, options )

	check( 'WFOzSyncPeer', 'WFOzCatalogSpec', '?table' )

	self._catalogSpec = catalogSpec

	if options and options.suppressAlbumAssetUploads then
		self._suppressAlbumAssetUploads = true
	end

	self._retryAssetIds = {}

	self._deletedAlbumIds = {}
	self._deletedAssetIds = {}
	
	WFOzSyncUtils.syncPeerSharedInit( self, catalogSpec:serverBaseUrl() )

end

--------------------------------------------------------------------------------

function syncPeerMethods:minSyncDelayIntervalB()

	return self._minSyncDelayIntervalB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:maxSyncDelayIntervalB()

	return self._maxSyncDelayIntervalB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:triggerSyncE()

	return self._triggerSyncE:relayE()

end

--------------------------------------------------------------------------------

function syncPeerMethods:hostIsReachableB()

	return self._hostIsReachableB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:currentHostErrorB()

	return self._currentHostErrorB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:forcedUpgradeLinkB()

	return self._forcedUpgradeLinkB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:testForceUpgradeLink()

	self._testForceUpgradeLink = true

end

--------------------------------------------------------------------------------

local syncHistory = WFDiagnosticLog.makeRingBuffer {

	accumulate = true,

	mapFn = function( updates )

		local deletions, assets = 0, 0

		for i, update in ipairs( updates ) do

			if update.deleted then
				deletions = deletions + 1
			elseif update.type == 'asset' then
				assets = assets + 1
			end

		end

		return #updates, deletions, assets

	end,

	"updates", "deletions", "assets",

}

--------------------------------------------------------------------------------

local function updateHistoricDiagnosticData( updates )

	syncHistory:add( updates )

end

--------------------------------------------------------------------------------

local function buildSyncHistorySparkline()

	return {
		syncHistory:visualize( 'sparkline' ),
	}

end

--------------------------------------------------------------------------------

local syncErrors = {}

--------------------------------------------------------------------------------

local function updateErrorDiagnosticData( error )
	
	if not error then
		return
	end
	
	syncErrors[ #syncErrors + 1 ] = error
	
	if #syncErrors > 30 then
		table.remove( syncErrors, 1 )
	end

end

--------------------------------------------------------------------------------

-- @mark :changedDocsP

--------------------------------------------------------------------------------

local function remapOzDocumentDataWithPayload( ozDocument )

	-- Invert the relationship between payload and server-provided metadata.

	local result = ozDocument.payload
	
	result._serverProvidedMetadata = ozDocument
	ozDocument.payload = nil
	
	result._id = ozDocument._id
	ozDocument._id = nil
	
	result.type = ozDocument.type
	ozDocument.type = nil
	
	result._timestamp = ozDocument.updated
	
	if result.type == 'asset' then

		result._serverProvidedMetadata.wfOzPayloadHash
			= WFOzSyncUtils.hashDocValue( result )

		result._serverProvidedMetadata.shaHash
			= KSTableUtils.derefOpt( result, 'importSource', 'sha256' )
				or 'none'

	end

	return result
	
end

--------------------------------------------------------------------------------

local fieldsToPutInServerProvidedMetadata = {

	created = true,
	created_by = true,
	deleted_by = true,
	id = true,
	links = true,
	updated = true,

}

--------------------------------------------------------------------------------

local function stripPayloadLinkNoise( noisyTable )

	noisyTable.id = nil
	noisyTable.links = nil

end

--------------------------------------------------------------------------------

local function remapOzDocumentDataWithoutPayload( ozDocument )

	local spm = {}
	
	for k in pairs( fieldsToPutInServerProvidedMetadata ) do
		spm[ k ] = ozDocument[ k ]
		ozDocument[ k ] = nil
	end
	
	ozDocument._serverProvidedMetadata = spm
	ozDocument._timestamp = spm.updated

	-- If this document was deleted on Oz, create a suitable
	-- tombstone for use here in WFDocStore. If we are setting
	-- _deleted to true, all body content must be removed or
	-- WFDocStore will flag an error.

	if ozDocument.deleted then

		ozDocument._deleted = true
		ozDocument.deleted = nil

		local keysToClear = {}
		for k in pairs( ozDocument ) do
			if not k:match( '^_' ) then
				keysToClear[ k ] = true
			end
		end
		
		for k in pairs( keysToClear ) do
			ozDocument[ k ] = nil
		end

	end
	
	if ozDocument.type == 'album_asset' then
	
		ozDocument.id = nil
		spm.links = nil
		spm.id = nil
		
		-- FIXME [matthewj 2014-12-01 via arahn]: Here is the other place this
		-- crypto-hash is computed. See discussion at
		-- https://git.corp.adobe.com/lua/wf/pull/307#discussion_r109777

		local md5 = require 'WFDigest'.MD5.digest
		ozDocument._serverProvidedMetadata.hash = md5( ozDocument._id .. tostring( ozDocument.order ) .. tostring( ozDocument.cover ) )
		
		stripPayloadLinkNoise( ozDocument.album )
		stripPayloadLinkNoise( ozDocument.asset )
	
	end
	
	if ozDocument.type == 'album' then
	
		if ozDocument.cover then
			stripPayloadLinkNoise( ozDocument.cover )
		end
	
	end
	
	return ozDocument

end

--------------------------------------------------------------------------------

local function remapOzDocumentData( self, ozDocument )

	-- If document is deleted, we might want to save off the doc ID to avoid
	-- unwanted deletion attempts for related album_asset documents.

	if ozDocument.deleted then
	
		if ozDocument.type == 'album' then
			self._deletedAlbumIds[ ozDocument._id ] = true
		elseif ozDocument.type == 'asset' then
			self._deletedAssetIds[ ozDocument._id ] = true
		end
	
	end
	
	-- Rearrange document to fit WF conventions instead of Oz conventions.
	-- This is a destructive transformation, but that's OK because the original
	-- document is no longer used after this call from changedDocs_imp.

	if ozDocument.payload and not ozDocument.deleted then
	
		return remapOzDocumentDataWithPayload( ozDocument )
	
	else
	
		return remapOzDocumentDataWithoutPayload( ozDocument )
	
	end

end

--------------------------------------------------------------------------------

local function reloadAssets( self, updates, retryAssetIds )

	-- We don't expect the array to get large, so not worrying about
	-- inefficient array construction here.

	local assetIdArray = {}
	for assetId in pairs( retryAssetIds ) do
		assetIdArray[ #assetIdArray + 1 ] = assetId
	end

	local function reloadOneAsset( assetId )
	
		local assetUrl = string.format( '%s/assets/%s',
										self._catalogSpec:catalogHref(), assetId )
	
		return self._catalogSpec:getP( assetUrl )
			:observeP( function( ... )
				WFOzSyncUtils.updateStatusFromHttpResultFn( self, ... )
			end )
			:mapDataP( function( _, data )
				local decodedData = KSJsonDecoder.decodeWebSafeJson_stripNull( data )
				decodedData.type = 'asset' -- ugh!
				decodedData.payload.type = nil
				return 'data', remapOzDocumentData( self, decodedData )
			end )
	
	end

	return KSPromises.mapArrayP( reloadOneAsset, assetIdArray, 5 )
		:mapDataP( function( _, reloadedAssets )
		
			for i = 1, #reloadedAssets do
				updates[ #updates + 1 ] = reloadedAssets[ i ]
			end
			
			return 'data', updates
		
		end )

end

--------------------------------------------------------------------------------

function syncPeerMethods:changedDocsP( since )

	check( 'WFOzSyncPeer', '?number|string' )

	-- If the sync bookmark doesn't include the catalog ID, then it's probably
	-- bogus. This, I believe, is the root cause of bug 3702952. Fall back to
	-- just starting over from scratch.

	if since and
		( type( since ) ~= 'string'
			or not since:match( self._catalogSpec:catalogId() ) )
	then
		since = nil
	end

	local url = since or ( self._catalogSpec:catalogHref() .. '/updates' )

	local changesFeedContentP = self._prefetchNextChangesP
	self._prefetchNextChangesP = false
	
	if changesFeedContentP then
	
		downloadLogger:debugf( "changedDocsP did prefetch %s", url )
	
	else

		if self._testForceUpgradeLink then
			url = url .. '?force_upgrade_error=1'
		end
	
		downloadLogger:debugf( "changedDocsP will fetch %s", url )
	
		changesFeedContentP = self._catalogSpec:getP( url )

	end

	return changesFeedContentP

		:observeP( function( ... )

			WFOzSyncUtils.updateStatusFromHttpResultFn( self, ... )

		end )

		:mapP( function( status, data, headers )

			local statusCode = KSTableUtils.derefOpt( headers, 'status_code' )

			if status ~= 'data' or statusCode ~= '200' then

				downloadLogger:errorf( "changedDocsP GET %s failed with status = %s"
									   .. "\n\ndata = %s\n\nheaders = %s\n\n",
									   status, data, headers )
				
				return 'error', statusCode or data

			end

			local decodedData = KSJsonDecoder.decodeWebSafeJson_stripNull( data )
			
			downloadLogger:debugf( "changedDocsP GET %s returned\n\n%s\n\n", url, decodedData )

			local result = {
				continueFrom = KSTableUtils.derefOpt( decodedData,
													  'links', 'next', 'href' )
			}
			
			local updates = KSTableUtils.derefOpt( decodedData, 'updates' ) or {}
			
			WFIdle.doLater( function()
				updateHistoricDiagnosticData( updates )
			end )
						
			for i = 1, #updates do
				result[ i ] = remapOzDocumentData( self, updates[ i ] )
			end

			return 'data', result

		end )

		:mapDataP( function( status, updates )
		
			if #updates > 0 and updates.continueFrom then
			
				self._prefetchNextChangesP = self._catalogSpec:getP( updates.continueFrom )
				
				self._prefetchNextChangesP:whenResolvedDo( function() end )
					:attachTo( self )
					:suppressCloseLogging()
				
			end
		
		end )

		:mapDataP( function( status, updates )
		
			-- Reload any assets for which we've received "not the head revision"
			-- errors, but that weren't already covered in the changes feed. (Ugh.)
		
			if next( self._retryAssetIds ) then

				local retryAssetIds = self._retryAssetIds
				self._retryAssetIds = {}
				
				for i = 1, #updates do
					local id = updates[ i ]._id
					if id then
						retryAssetIds[ id ] = nil
					end
				end
				
				if next( retryAssetIds ) then
					return reloadAssets( self, updates, retryAssetIds )
				end
			
			end
		
		end )
		
end

--------------------------------------------------------------------------------

-- @mark :sendUpdatesP

--------------------------------------------------------------------------------

local function replaceDeleteWithJsonNull( t )

	local didReplaceAnything = false

	for k, v in pairs( t ) do
	
		if type( v ) == 'table' then
			
			if v._delete == true then
				didReplaceAnything = true
				t[ k ] = yajl.null
			else
				didReplaceAnything = replaceDeleteWithJsonNull( v ) or didReplaceAnything
			end

		end
	
	end
	
	return didReplaceAnything

end

--------------------------------------------------------------------------------

local function commonDeleteP( self, args )

	local uploadUrl = assert( args.uploadUrl )
	local requestMethod = args.requestMethod or 'DELETE'
	local acceptStatus = args.acceptStatus or
		{
			[ '204' ] = true,
			[ '404' ] = true,
		}

	local uploadHeaders =
		{
			requestMethod = requestMethod,
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
				{ header = 'Accept', value = 'application/json' },
			},
			returnHttpErrorsAsData = true,
		}

	uploadLogger:debugf( "\%s %s\n\n%s\n\n",
						 requestMethod,
						 uploadUrl, uploadHeaders )
	
	return self._catalogSpec:getP( uploadUrl, uploadHeaders )

		:observeP( function( ... )

			WFOzSyncUtils.updateStatusFromHttpResultFn( self, ... )

		end )

		:mapP( function( status, data, headers )

			if status == 'error'
				or ( not acceptStatus[ headers and headers.status_code or 'nil' ] )
			then

				uploadLogger:errorf( "\n%s %s failed with\n\n%s\n\nbody = %s\n\n",
									 requestMethod, uploadUrl, headers )
				
				updateErrorDiagnosticData( {
						time = KSDate.timeInIso8601FormatWithTimeZone(),
						method = requestMethod,
						url = uploadUrl,
						headers = headers,
						message = tostring( data ),
				} )
				
				return 'error', headers.status_code

			else

				uploadLogger:debugf( "\n%s %s succeeded with\n\n%s\n\n",
									 requestMethod, uploadUrl, headers )

				return 'data', uploadUrl

			end

		end )

end

--------------------------------------------------------------------------------

local function commonUploadP( self, args )

	local uploadUrl = assert( args.uploadUrl )
	local uploadContent = assert( args.uploadContent )
	local requestMethod = args.requestMethod or 'PUT'
	local acceptStatus = args.acceptStatus or { [ '201' ] = true }
	
	if not next( uploadContent ) then
		-- 4006288: We might see empty updates sent to Oz which would 
		-- return an error that would be unnecessarily handled. If the 
		-- payload is empty, skip it.
		return KSPromises.resolvedPromiseP( 'data', true )
	end
	
	if uploadContent._deleted == true then
		return commonDeleteP( self, args )
	end
	
	local didReplaceAnyNulls = replaceDeleteWithJsonNull( uploadContent )

	local uploadHeaders =
		{
			requestMethod = requestMethod,
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
				{ header = 'Accept', value = 'application/json' },
			},
			returnHttpErrorsAsData = true,
		}

	local uploadJson = KSJsonEncoder.encodeOzUploadJson( uploadContent )

	uploadLogger:debugf( "\%s %s\n\n%s\n\n%s\n\n",
						 requestMethod,
						 uploadUrl, uploadHeaders,
						 didReplaceAnyNulls and uploadJson or uploadContent )
	
	return self._catalogSpec:putP( uploadUrl, uploadJson, uploadHeaders )

		:observeP( function( ... )

			WFOzSyncUtils.updateStatusFromHttpResultFn( self, ... )

		end )

		:mapP( function( status, data, headers )

			if status == 'error'
				or ( not acceptStatus[ headers and headers.status_code or 'nil' ] )
			then

				uploadLogger:errorf( "\n%s %s failed with\n\n%s\n\nbody = %s\n\n",
									 requestMethod, uploadUrl, headers,
									 didReplaceAnyNulls and uploadJson or uploadContent )
				
				local statusCode = headers and headers.status_code
				
				if statusCode and statusCode < '500' then
				
					local success, decodedData = pcall( KSJsonDecoder.decodeWebSafeJson,
														data )

					-- Special case: Result code '1005' means we didn't have the head
					-- revision when we attempted to push our change -- but only if
					-- errors.parent_revision_id is non-nil. This is not a permanent
					-- error and should be cleared when we receive changes again.
					
					-- Result code '1002' means we're attempting to push a new document
					-- when one already exists. Handle that similarly.
					
					local resultCode = tonumber( KSTableUtils.derefOpt( decodedData, 'code' ) )

					local resultParentRevId = KSTableUtils.derefOpt( decodedData,
																	  'errors',
																	  'parent_revision_id' )

					local permanent = resultCode and resultCode ~= 1002
						and not ( resultCode == 1005 and resultParentRevId ~= nil )
					
					-- Special case for space objects referencing albums that don't
					-- (yet?) exist. See 3835852.
					
					if resultCode == 1000 then
						
						if KSTableUtils.derefOpt( decodedData, 'errors', 'album' ) then
							permanent = false
						end
					
					end
					
					if permanent then

						updateErrorDiagnosticData( {
								time = KSDate.timeInIso8601FormatWithTimeZone(),
								method = requestMethod,
								url = uploadUrl,
								headers = headers,
								body = didReplaceAnyNulls and uploadJson or uploadContent,
						} )
					
					end
						
					local errorBlock =
						{
							message = "Upload rejected",
							resultCode = headers and headers[ "X-Result-Code" ],
							resultData = success and decodedData or nil,
							permanent = permanent,
						}
					
					return 'error', errorBlock
					
				else
				
					updateErrorDiagnosticData( {
							time = KSDate.timeInIso8601FormatWithTimeZone(),
							method = requestMethod,
							url = uploadUrl,
							headers = headers,
							body = didReplaceAnyNulls and uploadJson or uploadContent,
					} )

					return 'error', headers and headers.status_code or data
				
				end

			else

				uploadLogger:debugf( "\n%s %s succeeded with\n\n%s\n\n",
									 requestMethod, uploadUrl, headers )

				return 'data', uploadUrl

			end

		end )

end

--------------------------------------------------------------------------------

local function deleteAssetP( self, delta )

	local assetId = delta._id

	uploadLogger:debugf( "deleteAssetP %s", assetId )
	
	self._deletedAssetIds[ assetId ] = true
	
	local uploadUrl = string.format( '%s/assets/%s',
									 self._catalogSpec:catalogHref(), assetId )
				
	return commonDeleteP( self,
		{
			uploadUrl = uploadUrl,
		} )
	
end

--------------------------------------------------------------------------------

local function convertSubSubTablesToNonOverlay( payload, ignoreDevelop, correspondingFullDocValue )

	for k, v in pairs( payload ) do
		
		if ( not ignoreDevelop or k ~= 'develop' )
			and type( v ) == 'table'
			and not v._delete
		then
			
			for k2, v2 in pairs( v ) do
				if type( v2 ) == 'table' and not v2._delete then
					v[ k2 ] = correspondingFullDocValue[ k ][ k2 ]
				end
			end

		end

	end

end

--------------------------------------------------------------------------------

local function uploadAssetMetadataP( self, delta )

	if delta._deleted == true then
		return deleteAssetP( self, delta )
	end

	local assetId = delta._id

	uploadLogger:debugf( "uploadAssetMetadataP %s", assetId )
	
	local fullDocValue = delta._fullDocValue or {}
	
	-- Create asset revision ID.

	local revisionId = KSIdGenerator.uuidInLowercaseHex()
	
	-- Warn if we're uploading a delta but think this is new. Possible bug
	-- in the sync agent that's causing this.
	
	if delta._timestamp == ':new:' and not delta.width then
	
		uploadLogger:errorf( "Attempt to upload incomplete doc value as new\n\n%s", delta )
		
		-- Mark this document for reload. Per bug #3703831, I think the underlying
		-- problem was we had most of the document value already uploaded to Oz already
		-- and we somehow generated a delta before we were ready for it.
		
		self._retryAssetIds[ assetId ] = true

		return KSPromises.resolvedPromiseP( 'error',
			{
				message = "Apparently incomplete delta -- reloading",
				permanent = false,
			} )
	
	end
	
	-- Is this a new revision or child?
	
	local childRevisionTemplate = KSTableUtils.derefOpt( delta,
														 '_fullDocValue',
														 '_serverProvidedMetadata',
														 'link_templates',
														 '/rels/child_revision',
														 'href' )

	-- Transform WF deltas into Oz-style document for uploading.
	-- Note: It's OK that this is a destructive transformation.

    local payload = KSTableUtils.deepCopy( delta )

	payload._id = nil
	payload._localOnly = nil
	payload._fullDocValue = nil
	payload._previouslyAgreedValue = nil
	payload._serverProvidedMetadata = nil
	payload._timestamp = nil
	payload._lightroomId = nil
	
	-- Oz only does delta application for tables at the top level of the payload hash.
	-- If we have overlays in sub-subtables, we have to copy from full doc value.
	-- Exception: develop hash gets one more level of overlayable behavior.
	
	convertSubSubTablesToNonOverlay( payload, true, fullDocValue )

	-- Special handling: Oz treats develop section as a "full-featured" hash.
	-- WF treats it with atomic merge rules. So we go back and hack the develop
	-- settings to match Oz's rules.
	
	if payload.develop then
	
		-- Re-establish a "normal" delta for develop, overriding the
		-- WF convention of using atomic merge for this hash. (Alternative
		-- considered: Stomp out the _merge = 'atomic' keyword in the develop
		-- hash. Decided not to do that because:
		
			-- 1: We actually *want* _merge = 'atomic' behavior when reading
			-- changes *from* Oz. IOW, we don't want to merge some parts of
			-- the XMP and discard others. We want to accept one version or
			-- the other in entirety.
			
			-- 2: Making sure that the deletion of the _merge semantics was
			-- fully propagated at all the right locations in the code was
			-- considered risy at best.
			
		-- Enforcing that rule here was deemed the lesser of evils.
	
		local newDevelop = KSTableUtils.derefOpt( delta,
												  '_fullDocValue',
												  'develop' ) or {}
	
		local oldDevelop = KSTableUtils.derefOpt( delta,
												  '_previouslyAgreedValue',
												  'develop' ) or {}

		newDevelop._merge = nil
		oldDevelop._merge = nil
		
		local developOverlay = WFTableMerge.createOverlay( oldDevelop, newDevelop )
		
		-- Just for even more fun, subtables within develop are given overlay
		-- but second-level tables inside of that are not.
		
		convertSubSubTablesToNonOverlay( developOverlay, false, newDevelop )
		
		-- Reassert _merge mode if necessary.
		
		if KSTableUtils.derefOpt( oldDevelop, 'develop', '_merge' ) ~= 'atomic' then
			developOverlay._merge = 'atomic'
		end
		
		payload.develop = developOverlay

	end

	-- Oz complains if you upload a partial location, so we go back and fill
	-- in the whole deal if needed.

	local location = payload.location
	
	if type( location ) == 'table' and not location._delete then
		
		local fdvLocation = fullDocValue.location or {}
		
		location.latitude = location.latitude or fdvLocation.latitude
		location.longitude = location.longitude or fdvLocation.longitude
		location.altitude = location.altitude or fdvLocation.altitude
		
		if not location.latitude or not location.longitude then
			location.latitude = nil
			location.longitude = nil
		end
		
		if not next( location ) then
			payload.location = nil
		end
	
	end

	local subType = 'jpeg'
		-- All still images are 'jpeg' on Oz even if they're not JPEG.

	if childRevisionTemplate then
		subType = nil
	end
	
	local uploadContent =
		{
			subtype = subType,
			user_id = self._catalogSpec:ozUserId(),
			payload = payload,
		}

	-- Note: If the asset already exists, we will get an error condition, which
	-- is to be expected. We'll reload the document, re-merge, and try again later.
	
	local uploadUrl = string.format( '%s/assets/%s/revisions/%s',
									 self._catalogSpec:catalogHref(), assetId, revisionId )
				
	if childRevisionTemplate then
		uploadUrl = childRevisionTemplate:gsub( '{.*}', revisionId )
			-- FIXME [scouten 2013-06-19]: Be more careful about placeholder
			-- search here.
	end
	
	return commonUploadP( self,
		{
			uploadUrl = uploadUrl,
			uploadContent = uploadContent,
		} )
		:observeP( function( status, message )
		
			-- If we got a "not the head revision" error, signal that we might need to
			-- manually reload the doc from Oz.
		
			if status == 'error' and type( message ) == 'table'
				and ( message.resultCode == '1002' or message.resultCode == '1005' )
			then
				self._retryAssetIds[ assetId ] = true
			end
		
		end )
	
end

--------------------------------------------------------------------------------

local function isAlbumDeleted( delta )

	return not not delta._deleted

end

--------------------------------------------------------------------------------

local function isNewAlbum( fullDocValue )

	local alreadyExisted =
			type( fullDocValue._serverProvidedMetadata ) == 'table'
			and ( not not fullDocValue._serverProvidedMetadata.created )

	return not alreadyExisted

end

--------------------------------------------------------------------------------

local function uploadAlbumP( self, delta )

	local albumId = delta._id

	uploadLogger:debugf( "uploadAlbumP %s", albumId )
	
    local payload = KSTableUtils.deepCopy( delta )

	local albumIsDeleted = isAlbumDeleted( delta )
	
	if albumIsDeleted then
		self._deletedAlbumIds[ albumId ] = true
	end

	local fullDocValue = delta._fullDocValue or {}
	payload._id = nil
	payload._fullDocValue = nil
	payload._timestamp = nil
	payload._previouslyAgreedValue = nil
	payload.revision_ids = nil
	payload.type = nil
	
	local uploadUrl = string.format( '%s/albums/%s',
									 self._catalogSpec:catalogHref(), albumId )

	return commonUploadP( self,
		{

			uploadUrl = uploadUrl,
			uploadContent = payload,

			requestMethod = albumIsDeleted and 'DELETE'
							or isNewAlbum( fullDocValue ) and 'PUT'
							or 'POST',

			acceptStatus = {
				[ '201' ] = true,
				[ '204' ] = true,
			},

		} )

end

--------------------------------------------------------------------------------

local function deleteAlbumAssetP( self, delta )

	local albumId = KSTableUtils.derefOpt( delta,
										   '_previouslyAgreedValue',
										   'album',
										   '_id' )

	local assetId = KSTableUtils.derefOpt( delta,
										   '_previouslyAgreedValue',
										   'asset',
										   '_id' )
	
	if self._deletedAlbumIds[ albumId ] or self._deletedAssetIds[ assetId ] then
		return KSPromises.resolvedPromiseP( 'data', true )
	end
	
	uploadLogger:debugf( "deleteAlbumAssetP %s %s" )

	if not( albumId and assetId ) then
	
		uploadLogger:warnf( "unable to find asset ID or album ID in previously agreed value %s",
							delta )
	
		return KSPromises.resolvedPromiseP( 'data', true )

	end

	local deleteAssetFromAlbumUrl = string.format( '%s/albums/%s/assets?asset_ids=%s',
												   self._catalogSpec:catalogHref(),
												   albumId, assetId )

	return commonDeleteP( self,
		{
			uploadUrl = deleteAssetFromAlbumUrl,
		} )

end

--------------------------------------------------------------------------------

local function uploadAlbumAssetP( self, delta )

	if delta._deleted == true then
		return deleteAlbumAssetP( self, delta )
	end
	
	-- TODO_SNOWLESS [scouten 2013-12-18]: Remove this code in favor of
	-- WFOzAlbumAssetUploader?
	
	if self._suppressAlbumAssetUploads then
		return KSPromises.resolvedPromiseP( 'data', ':skip:' )
	end
	
	local albumId = delta.album._id
	local assetId = delta.asset._id

	uploadLogger:debugf( "uploadAlbumAssetP %s %s", albumId, assetId )

	-- Special case: Oz likes to annotate the album_asset record with lots
	-- of extra junk that causes the sync agent to re-send the doc. So if
	-- we're not sending anything new here, just squelch this send.

	if not( albumId and assetId ) then
		return KSPromises.resolvedPromiseP( 'data', true )
	end
	
	local addAssetsToAlbumUrl = string.format( '%s/albums/%s/assets',
											   self._catalogSpec:catalogHref(),
											   albumId )

	local uploadContent =
		{
			assets = {
				{
					_id = assetId,
					album_asset_id = delta._id,
				},
			},
		}
	
	return commonUploadP( self,
		{
			uploadUrl = addAssetsToAlbumUrl,
			uploadContent = uploadContent,
			acceptStatus = {
				[ '201' ] = true,
				[ '204' ] = true,
			},
		} )

	-- TODO_SNOWLESS [scouten 2013-08-15]: Make sure we include the same duplicate-avoidance
	-- code that we have in old sync code. See GRUNGE [ngautam 2012-05-18] comment
	-- near WFOzSyncCycle.lua, line 503.
	
	-- TODO_SNOWLESS [scouten 2013-08-15]: Handle album deleted status.

end

--------------------------------------------------------------------------------

local function deleteSpaceP( self, delta )
		
	local spaceId = delta._id
	
	local deleteSpaceUrl = string.format( '%s/spaces/%s',
											   self._catalogSpec:catalogHref(),
											   spaceId )
	
	uploadLogger:debugf( "deleteSpaceP %s", spaceId )

	return commonDeleteP( self,
		{
			uploadUrl = deleteSpaceUrl,
		} )

end

--------------------------------------------------------------------------------

local function uploadSpaceP( self, delta )
	
	if delta._deleted == true then
		return deleteSpaceP( self, delta )
	end
	
	-- SPAM [matthewj]: When a space is created it needs to specify the album/asset that is in
	-- that space. The old style sharing code in OzCatSharingHelper sends this information via
	-- the _localOnly section which we don't receive here (but the old OzCentralUploadTask does).
	-- Until that's changed and it sends the 'space' doc with all the data we need to create
	-- the correct document to send to Oz we will ignore space docs that aren't deletions and
	-- instead wait for the space_album doc which has all the data we need (see uploadSpaceAlbumP)

	if delta._timestamp == ':new:' then
		return KSPromises.resolvedPromiseP( 'data', true )
	end
	
	local updateSpaceUrl = string.format( '%s/spaces/%s',
												self._catalogSpec:catalogHref(),
												delta._id )
	
	local payload = KSTableUtils.deepCopy( delta )

	payload._id = nil
	payload._localOnly = nil
	payload._fullDocValue = nil
	payload._previouslyAgreedValue = nil
	payload._serverProvidedMetadata = nil
	payload._timestamp = nil
	payload._lightroomId = nil
	
	return commonUploadP( self,
		{
			requestMethod = 'POST',
			uploadUrl = updateSpaceUrl,
			uploadContent = payload,
			acceptStatus = {
				[ '204' ] = true,
			},
		} )
	
end

--------------------------------------------------------------------------------

local function deleteSpaceAlbumP( self, delta )
	-- The space_album will be deleted from Oz when the space is and that comes to the sync peer
	-- first so there's no use sending anything at this point
	return KSPromises.resolvedPromiseP( 'data', true )
end

--------------------------------------------------------------------------------

local function uploadSpaceAlbumP( self, delta )

	if delta._deleted == true then
		return deleteSpaceAlbumP( self, delta )
	end
	
	-- SPAM [matthewj]: When we get a space album record we haven't yet created the space itself on
	-- Oz as they are two documents that go up in 1 step and we didn't have enough information in
	-- uploadSpaceP to send the right record to Oz (see comments in uploadSpaceP). So now we're going
	-- to take the space_album doc and rework it into the doc necessary to create a space that
	-- includes the album
	
	local spaceAlbumId = delta._id
	local spaceId = KSTableUtils.derefOpt( delta, 'space', '_id' )
	local albumId = KSTableUtils.derefOpt( delta, 'album', '_id' )

	uploadLogger:debugf( "uploadSpaceAlbumP %s", spaceAlbumId )
	
	local addSpaceUrl = string.format( '%s/spaces/%s',
											   self._catalogSpec:catalogHref(),
											   spaceId )
	
	local payload = {
		download = false, -- UGH. It would be really nice if we had these values from
		private = false,  -- the space doc. These are different defaults than Revel uses
		type = 'space',
		resources = {
			{
				_id = albumId,
				type = 'album',
				space_album_id = spaceAlbumId,
			},
		},
	}
	
	return commonUploadP( self,
		{
			uploadUrl = addSpaceUrl,
			uploadContent = payload,
			acceptStatus = {
				[ '201' ] = true,
				[ '204' ] = true,
			},
		} )
end

--------------------------------------------------------------------------------

local function deleteCommentP( self, delta )
	
	local commentId = delta._id
	
	local spaceId = KSTableUtils.derefOpt( delta,
										   '_previouslyAgreedValue',
										   'space',
										   '_id' )

	local assetId = KSTableUtils.derefOpt( delta,
										   '_previouslyAgreedValue',
										   'asset',
										   '_id' )
	
	if not( spaceId and assetId ) then
	
		uploadLogger:warnf( "unable to find asset ID or space ID in previously agreed value %s",
							delta )
	
		return KSPromises.resolvedPromiseP( 'data', true )

	end

								
	local deleteCommentUrl = string.format( '%s/spaces/%s/assets/%s/comments/%s',
											   self._catalogSpec:baseHref(),
											   spaceId, assetId, commentId )
	
	uploadLogger:debugf( "deleteCommentP %s", commentId )

	return commonDeleteP( self,
		{
			uploadUrl = deleteCommentUrl,
		} )

end

--------------------------------------------------------------------------------

local kMaxCommentLength = 1024 - 1 -- must be LESS THAN 1024

local function uploadCommentP( self, delta )
	
	if delta._deleted == true then
		return deleteCommentP( self, delta )
	end
	
	local commentId = delta._id
	local spaceId = KSTableUtils.derefOpt( delta, 'space', '_id' )
	local assetId = KSTableUtils.derefOpt( delta, 'asset', '_id' )

	uploadLogger:debugf( "uploadCommentP %s", commentId )
	
	local addCommentUrl = string.format( '%s/spaces/%s/assets/%s/comments/%s',
											self._catalogSpec:baseHref(),
											spaceId, assetId, commentId )
	
	local payload = {
		body = string.sub( delta.body, 1, kMaxCommentLength )
		-- author_name (do we need to support override?)
	}
	
	return commonUploadP( self,
		{
			uploadUrl = addCommentUrl,
			uploadContent = payload,
		} )


end

--------------------------------------------------------------------------------

local uploadFunctionsByDocType = {

	asset = uploadAssetMetadataP,
	album = uploadAlbumP,
	album_asset = uploadAlbumAssetP,
	space = uploadSpaceP,
	space_album = uploadSpaceAlbumP,
	comment = uploadCommentP,
	catalog = 'nop',
	autoImport = 'nop',

}

--------------------------------------------------------------------------------

function syncPeerMethods:sendUpdatesP( updates )
	
	check( 'WFOzSyncPeer', 'table' )

	return WFOzSyncUtils.sendUpdatesP_imp( self, updates, uploadFunctionsByDocType )

end

--------------------------------------------------------------------------------

function syncPeerMethods:maxUpdatesPerSend()

	return 4  -- depends on remote peer

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function syncPeerMethods:sendDiagnosticLogData()

	local catalogSpec = self._catalogSpec
	local wfOzUser = catalogSpec:wfOzUser()
	
	local ozUser = require 'OzUser'.make( wfOzUser:userId() )
	
	local ozUserInfoP = ozUser:userInfoP()
		:observeP( function()
			ozUser:close()
		end )

	return {

		name = "WF Oz Sync Agent",
		id = 'WFOzSyncAgent',
		weight = 500,

		body = {
		
			ozApiServer = catalogSpec:serverBaseUrl(),
			ozCatalogId = catalogSpec:catalogId(),
			ozCatalogHref = catalogSpec:catalogHref(),

			ozHostReachable = self._hostIsReachableB,

			ozHostErrorStatus = self._currentHostErrorB
				:mapB( function( value ) return tostring( value ) end ),

			ozUserId = wfOzUser:userId(),
            ozUserInfo = ozUserInfoP:runUntilDoneP(),
			
			syncHistory = buildSyncHistorySparkline(),

			recentOzErrors = syncErrors,
		
		},
	
	}
	
end

--------------------------------------------------------------------------------
--[[
	-- TODO_SNOWLESS [scouten 2013-07-08]:
	
		- "internal" / "external" progress tracking (see updateInternalProgress,
			download bookmark, and friends in OzCentralSyncAgent.lua)
		- failed upload tracking for diagnostic log (see _statusReport and _recentErrors
			in OzCentralSyncAgent.lua)
		- # cycles tracking (see _nCycles* in OzCentralSyncAgent.lua)
		- delete thumbnails, etc. from local storage when corresponding documents
			deleted from Oz (see lines 280++ in receiveOneAssetP in WFOzDownloadTask.lua)
		- document-type specific refresh-once logic? (see refreshDocumentTypeOnceP
			in WFOzDownloadTask.lua) - for Revel only; may no longer be necessary
			even for Revel
		- capture download errors (see lines 906++ in WFOzDownloadTask.lua)
]]--

--------------------------------------------------------------------------------

return Rx.Class 'WFOzSyncPeer'( syncPeerMethods )
