--[[----------------------------------------------------------------------------

WFOzSyncPeerTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'
local WFOzTestUtils = require 'WFOzTestUtils'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'

	-- WF.oz_client.sync_engine
local WFOzCatalogTestSession = require 'WFOzCatalogTestSession'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'


--============================================================================--

local WFOzSyncPeerTest = {}

--------------------------------------------------------------------------------

local catalogTestSession
local TEST_WAIT_TIME = 240

--------------------------------------------------------------------------------

function WFOzSyncPeerTest.setUp()
	
	catalogTestSession = WFOzCatalogTestSession( 'WFOzSyncPeerTest' )

end

--------------------------------------------------------------------------------

function WFOzSyncPeerTest.tearDown()

	if catalogTestSession then
		catalogTestSession:releaseTestUserP()
		catalogTestSession:close()
		catalogTestSession = nil
	end

end

--------------------------------------------------------------------------------

local function sortSubtable( keypath, t )
	
	local WFDigest = require 'WFDigest'
	local WFTableArchiving = require 'WFTableArchiving'
	
	local updates = t
	
	for _, key in ipairs( keypath ) do
		updates = updates[ key ] or {}
	end
	
	local hashes = {}
	
	local result = {}
	
	for i, update in ipairs( updates ) do
		
		result[ i ] = update
		hashes[ update ] = update.type
			.. '+' .. WFDigest.SHA1.digest( WFTableArchiving.tableToString( update ) )
		
	end
	
	table.sort( result, function( a, b )
		return hashes[ a ] > hashes[ b ]
	end )
	
	return result
	
end

--------------------------------------------------------------------------------

local function sortUpdateByTypeAndHash( t )
	return sortSubtable( { 'updates' }, t )
end

--------------------------------------------------------------------------------

local function sortByTypeAndHash( t )
	return sortSubtable( { }, t )
end

--------------------------------------------------------------------------------

local function mapAwayUuidsAndTimestamps( t )

	if type( t._changeCounter ) == 'number' and t._changeCounter > 0 then
		t._changeCounter = "(change counter)"
	end
	
	if type( t.continueFrom ) == 'number' and t.continueFrom > 0 then
		t.continueFrom = "(change counter)"
	end
	
	if type( t._idLocal ) == 'number' and t._idLocal > 0 then
		t._idLocal = "(id local)"
	end
	
	-- map revision_ids away as this information is not important for v1 sync clients
	
	t.revision_ids = nil
	
	for k, v in pairs( t ) do
	
		if k == 'first_name' or k == 'full_name' or k == 'last_name' then

			t[ k ] = k

		elseif type( v ) == 'string' then
		
			-- Hash codes and UUIDs.
			
			if #v == 32 or #v == 40 or #v == 64 then
				v = v:gsub( '^[0-9a-f]+$', string.format( "hash of %d chars", #v ) )
			end
			
			-- URN-style IDs should be consistent in use of "carousel" and not "revel".
			
			v = v:replace( ":revel:", ":carousel:" )
			
			-- URN-style IDs.
			
			v = v:gsub( "account:[0-9a-f]+", "account:(account ID)" )
			v = v:gsub( "catalog:[0-9a-f]+", "catalog:(catalog ID)" )
			v = v:gsub( "album:[0-9a-f]+", "asset:(album ID)" )
			v = v:gsub( "asset:[0-9a-f]+", "asset:(asset ID)" )
			
			-- IDs in URLs.
			
			v = v:gsub( "accounts/[0-9a-f]+", "accounts/(account ID)" )
			v = v:gsub( "catalogs/[0-9a-f]+", "catalogs/(catalog ID)" )
			v = v:gsub( "albums/[0-9a-f]+", "albums/(album ID)" )
			v = v:gsub( "assets/[0-9a-f]+", "assets/(asset ID)" )
			v = v:gsub( "revisions/[0-9a-f]+", "revisions/(revision ID)" )
			v = v:gsub( "libraries/[0-9a-f]+", "libraries/(library ID)" )
			
			-- ISO-8601 timestamps.
			
			v = v:gsub( "^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d+$", "(client timestamp)" )
			v = v:gsub( "^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d+Z$", "(Oz timestamp)" )
			
			-- Updates link.
			
			v = v:gsub( "updated_since=.+Z$", "updated_since=(Oz update timestamp)" )
			
			t[ k ] = v
			
		end
		
		if type( v ) == 'table' then
			mapAwayUuidsAndTimestamps( v )
		end
	
	end
	
	t = WFOzTestUtils.v1.removeFragileFieldsFromCatalog( t )
	
	return t

end

--------------------------------------------------------------------------------

function WFOzSyncPeerTest.testSimpleSendAndDelete_testNightly( test )

	local syncAgent, albumId, albumAssetId, photoId1, photoId2, catalogSpec, nextUpdatesLink
	catalogTestSession:setTestName( test:testName() )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalogTestSession:catalogSpecP()
				:mapDataP( function( _, _catalogSpec )

					catalogSpec = _catalogSpec

					nextUpdatesLink = catalogSpec:catalogHref() .. "/updates"
					return 'data', true
				
				end )
		
		end )

		-- Load up our three documents and add an album and album_asset records.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, _syncAgent )
				
					syncAgent = _syncAgent
				
					-- Create an album.
				
					albumId = KSIdGenerator.uuidInLowercaseHex()
					
					local photoIds = catalogTestSession:fakePhotoIdsByCaptureDate()
					photoId1 = photoIds[ "2004-11-02T11:10:09" ]
					photoId2 = photoIds[ "2001-02-03T04:05:06" ]
				
					return catalogTestSession:docStore():setDocValueP
						{
							_id = albumId,
							_slot = 'main',
							type = 'album',
							name = "Test Album",
						}
				
				end )
				:mapDataP( function()
				
					-- Attach one photo to the album.
					
					albumAssetId = KSIdGenerator.uuidInLowercaseHex()

					return catalogTestSession:docStore():setDocValueP
						{
							_id = albumAssetId,
							_slot = 'main',
							type = 'album_asset',
							album = {
								_id = albumId,
							},
							asset = {
								_id = photoId1,
							},
						}
					
				end )
				:mapDataP( function()
				
					-- Attach another photo to the same album.
					
					return catalogTestSession:docStore():setDocValueP
						{
							_id = KSIdGenerator.uuidInLowercaseHex(),
							_slot = 'main',
							type = 'album_asset',
							album = {
								_id = albumId,
							},
							asset = {
								_id = photoId2,
							},
						}
					
				end )
		
		end )
		
		-- TODO [scouten 2013-08-15]: Need to test a second album.

		-- Send the documents over to Oz.
		
		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )
	
		-- Wait a few moments (real-time hack, anyone?)
		
		:waitFor( 10 )

		-- Access Oz directly to verify that changes are present.
		
		:waitFor( function()
		
			return catalogSpec:getP( nextUpdatesLink )
				:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
				:mapDataP( function( _, updates )
				
					nextUpdatesLink = KSTableUtils.derefOpt( updates,
															 'links',
															 'next',
															 'href' )
					
					return 'data', sortUpdateByTypeAndHash( mapAwayUuidsAndTimestamps( updates ) )
				
				end )
			
		end )
		:continueWhenValueMatches( 'data',
			sortUpdateByTypeAndHash {
				base = "https://api.ozstage.com",
				links = {
					next = {
						href = "/v1.0/catalogs/(catalog ID)/updates?updated_since=(Oz update timestamp)",
					},
				},
				updates = {
					{
						_id = "hash of 32 chars",
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						link_templates = {
							[ "/rels/child_revision" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
							},
						},
						links = {
							[ "/rels/first_revision" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							[ "/rels/latest_revision" ] = {
								created_by = "urn:carousel:account:(account ID)",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							[ "self" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
							},
						},
						payload = {
							captureDate = "2004-11-02T11:10:09",
							height = 16,
							importSource = {
								fileCreationDate = "2004-11-02T08:37:24",
								fileName = "fakeFile01.jpg",
								fileSize = 13493,
								folder = "/fake/path/to/fake",
								folderPath = "/fake/path/to/fake",
								importTimestamp = "(client timestamp)",
								importedBy = "hash of 32 chars",
								importedOnDevice = "WFOzCatalogTestSession.lua",
								sha256 = "hash of 64 chars",
							},
							type = "asset",
							version = 1,
							width = 16,
						},
						subtype = "jpeg",
						type = "asset",
						updated = "(Oz timestamp)",
					},
					{
						_id = "hash of 32 chars",
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						link_templates = {
							[ "/rels/child_revision" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
							},
						},
						links = {
							[ "/rels/first_revision" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							[ "/rels/latest_revision" ] = {
								created_by = "urn:carousel:account:(account ID)",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							[ "self" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
							},
						},
						payload = {
							captureDate = "2012-05-02T08:07:06",
							height = 16,
							importSource = {
								fileCreationDate = "2004-11-02T08:37:24",
								fileName = "fakeFile01.jpg",
								fileSize = 13493,
								folder = "/fake/path/to/fake",
								folderPath = "/fake/path/to/fake",
								importTimestamp = "(client timestamp)",
								importedBy = "hash of 32 chars",
								importedOnDevice = "WFOzCatalogTestSession.lua",
								sha256 = "hash of 64 chars",
							},
							type = "asset",
							version = 1,
							width = 16,
						},
						subtype = "jpeg",
						type = "asset",
						updated = "(Oz timestamp)",
					},
					{
						_id = "hash of 32 chars",
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						link_templates = {
							[ "/rels/child_revision" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
							},
						},
						links = {
							[ "/rels/first_revision" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							[ "/rels/latest_revision" ] = {
								created_by = "urn:carousel:account:(account ID)",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							[ "self" ] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
							},
						},
						payload = {
							captureDate = "2001-02-03T04:05:06",
							height = 16,
							importSource = {
								fileCreationDate = "2004-11-02T08:37:24",
								fileName = "fakeFile01.jpg",
								fileSize = 13493,
								folder = "/fake/path/to/fake",
								folderPath = "/fake/path/to/fake",
								importTimestamp = "(client timestamp)",
								importedBy = "hash of 32 chars",
								importedOnDevice = "WFOzCatalogTestSession.lua",
								sha256 = "hash of 64 chars",
							},
							type = "asset",
							version = 1,
							width = 16,
						},
						subtype = "jpeg",
						type = "asset",
						updated = "(Oz timestamp)",
					},
					{
						_id = "hash of 32 chars",
						created = "(Oz timestamp)",
						created_by = "urn:carousel:account:(account ID)",
						id = "urn:carousel:catalog:(catalog ID):asset:(album ID)",
						links = {
							[ "/rels/album_assets" ] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)/assets",
							},
							[ "/rels/album_assets_with_videos" ] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)/assets?subtype=jpeg%3Bvideo",
							},
							["/rels/web_album"] = {
								href = "https://stage.adobelr.com/libraries/(library ID)/albums/(album ID)",
							},
							[ "self" ] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)",
							},
						},
						name = "Test Album",
						type = "album",
						updated = "(Oz timestamp)",
					},
					{
						_id = "hash of 32 chars",
						added_by = "urn:carousel:account:(account ID)",
						album = {
							_id = "hash of 32 chars",
							id = "urn:carousel:catalog:(catalog ID):asset:(album ID)",
							links = {
								[ "self" ] = {
									href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)",
								},
							},
						},
						asset = {
							_id = "hash of 32 chars",
							id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
							links = {
								[ "self" ] = {
									href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
								},
							},
						},
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(album ID):asset:(asset ID)",
						links = {
							[ "self" ] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)/assets/(asset ID)",
							},
						},
						type = "album_asset",
						updated = "(Oz timestamp)",
					},
					{
						_id = "hash of 32 chars",
						added_by = "urn:carousel:account:(account ID)",
						album = {
							_id = "hash of 32 chars",
							id = "urn:carousel:catalog:(catalog ID):asset:(album ID)",
							links = {
								[ "self" ] = {
									href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)",
								},
							},
						},
						asset = {
							_id = "hash of 32 chars",
							id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
							links = {
								[ "self" ] = {
									href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
								},
							},
						},
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(album ID):asset:(asset ID)",
						links = {
							[ "self" ] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)/assets/(asset ID)",
							},
						},
						type = "album_asset",
						updated = "(Oz timestamp)",
					},
				},
			} )
		
		-- Then try to read the docs back from Oz.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )
					return syncAgent:getChangesP()
				end )
		
		end )
		:continueWhenValueMatches( 'data', true )
		
		-- Now verify contents of our local DB have been updated
		-- to reflect Oz links.
		
		:waitFor( function()
		
			return catalogTestSession:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( function( _, changes )
				
					return 'data', sortByTypeAndHash( mapAwayUuidsAndTimestamps( changes ) )
				
				end )
			
		end )
		:continueWhenValueMatches( 'data',
			sortByTypeAndHash {
				{
					_changeCounter = "(change counter)",
					_hash = "hash of 40 chars",
					_id = "hash of 32 chars",
					_idLocal = "(id local)",
					_serverProvidedMetadata = {
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						link_templates = {
							["/rels/child_revision"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
							},
						},
						links = {
							["/rels/first_revision"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							["/rels/latest_revision"] = {
								created_by = "urn:carousel:account:(account ID)",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							["self"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
							},
						},
						subtype = "jpeg",
						updated = "(Oz timestamp)",
						wfOzPayloadHash = "hash of 40 chars",
					},
					_slot = "main",
					_syncStatus = {
						oz_TEST = {
							accepted = "(Oz timestamp)",
							timestamp = "(Oz timestamp)",
						},
					},
					captureDate = "2012-05-02T08:07:06",
					height = 16,
					importSource = {
						fileCreationDate = "2004-11-02T08:37:24",
						fileName = "fakeFile01.jpg",
						fileSize = 13493,
						folder = "/fake/path/to/fake",
						folderPath = "/fake/path/to/fake",
						importTimestamp = "(client timestamp)",
						importedBy = "hash of 32 chars",
						importedOnDevice = "WFOzCatalogTestSession.lua",
						sha256 = "hash of 64 chars",
					},
					type = "asset",
					version = 1,
					width = 16,
				},
				{
					_changeCounter = "(change counter)",
					_hash = "hash of 40 chars",
					_id = "hash of 32 chars",
					_idLocal = "(id local)",
					_serverProvidedMetadata = {
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						link_templates = {
							["/rels/child_revision"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
							},
						},
						links = {
							["/rels/first_revision"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							["/rels/latest_revision"] = {
								created_by = "urn:carousel:account:(account ID)",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							["self"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
							},
						},
						subtype = "jpeg",
						updated = "(Oz timestamp)",
						wfOzPayloadHash = "hash of 40 chars",
					},
					_slot = "main",
					_syncStatus = {
						oz_TEST = {
							accepted = "(Oz timestamp)",
							timestamp = "(Oz timestamp)",
						},
					},
					captureDate = "2004-11-02T11:10:09",
					height = 16,
					importSource = {
						fileCreationDate = "2004-11-02T08:37:24",
						fileName = "fakeFile01.jpg",
						fileSize = 13493,
						folder = "/fake/path/to/fake",
						folderPath = "/fake/path/to/fake",
						importTimestamp = "(client timestamp)",
						importedBy = "hash of 32 chars",
						importedOnDevice = "WFOzCatalogTestSession.lua",
						sha256 = "hash of 64 chars",
					},
					type = "asset",
					version = 1,
					width = 16,
				},
				{
					_changeCounter = "(change counter)",
					_hash = "hash of 40 chars",
					_id = "hash of 32 chars",
					_idLocal = "(id local)",
					_serverProvidedMetadata = {
						created = "(Oz timestamp)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						link_templates = {
							["/rels/child_revision"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
							},
						},
						links = {
							["/rels/first_revision"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							["/rels/latest_revision"] = {
								created_by = "urn:carousel:account:(account ID)",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
							},
							["self"] = {
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
							},
						},
						subtype = "jpeg",
						updated = "(Oz timestamp)",
						wfOzPayloadHash = "hash of 40 chars",
					},
					_slot = "main",
					_syncStatus = {
						oz_TEST = {
							accepted = "(Oz timestamp)",
							timestamp = "(Oz timestamp)",
						},
					},
					captureDate = "2001-02-03T04:05:06",
					height = 16,
					importSource = {
						fileCreationDate = "2004-11-02T08:37:24",
						fileName = "fakeFile01.jpg",
						fileSize = 13493,
						folder = "/fake/path/to/fake",
						folderPath = "/fake/path/to/fake",
						importTimestamp = "(client timestamp)",
						importedBy = "hash of 32 chars",
						importedOnDevice = "WFOzCatalogTestSession.lua",
						sha256 = "hash of 64 chars",
					},
					type = "asset",
					version = 1,
					width = 16,
				},
				{
					_changeCounter = "(change counter)",
					_hash = "hash of 40 chars",
					_id = "hash of 32 chars",
					_idLocal = "(id local)",
					_serverProvidedMetadata = {
						created = "(Oz timestamp)",
						created_by = "urn:carousel:account:(account ID)",
						id = "urn:carousel:catalog:(catalog ID):asset:(album ID)",
						links = {
							["/rels/album_assets"] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)/assets",
							},
							[ "/rels/album_assets_with_videos" ] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)/assets?subtype=jpeg%3Bvideo",
							},
							["/rels/web_album"] = {
								href = "https://stage.adobelr.com/libraries/(library ID)/albums/(album ID)",
							},
							["self"] = {
								href = "/v1.0/catalogs/(catalog ID)/albums/(album ID)",
							},
						},
						updated = "(Oz timestamp)",
					},
					_slot = "main",
					_syncStatus = {
						oz_TEST = {
							accepted = "(Oz timestamp)",
							timestamp = "(Oz timestamp)",
						},
					},
					name = "Test Album",
					type = "album",
				},
				{
					_changeCounter = "(change counter)",
					_hash = "hash of 40 chars",
					_id = "hash of 32 chars",
					_idLocal = "(id local)",
					_serverProvidedMetadata = {
						created = "(Oz timestamp)",
						hash = "hash of 32 chars",
						updated = "(Oz timestamp)",
					},
					_slot = "main",
					_syncStatus = {
						oz_TEST = {
							accepted = "(Oz timestamp)",
							timestamp = "(Oz timestamp)",
						},
					},
					added_by = "urn:carousel:account:(account ID)",
					album = {
						_id = "hash of 32 chars",
					},
					asset = {
						_id = "hash of 32 chars",
					},
					type = "album_asset",
				},
				{
					_changeCounter = "(change counter)",
					_hash = "hash of 40 chars",
					_id = "hash of 32 chars",
					_idLocal = "(id local)",
					_serverProvidedMetadata = {
						created = "(Oz timestamp)",
						hash = "hash of 32 chars",
						updated = "(Oz timestamp)",
					},
					_slot = "main",
					_syncStatus = {
						oz_TEST = {
							accepted = "(Oz timestamp)",
							timestamp = "(Oz timestamp)",
						},
					},
					added_by = "urn:carousel:account:(account ID)",
					album = {
						_id = "hash of 32 chars",
					},
					asset = {
						_id = "hash of 32 chars",
					},
					type = "album_asset",
				},
				continueFrom = "(change counter)",
			} )

		-- Delete one photo from album and watch that flow through.
		
		:waitFor( function()
		
			return catalogTestSession:docStore():setDocValueP
				{
					_id = albumAssetId,
					_slot = 'main',
					_deleted = true,
				}
		
		end )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )
	
		:waitFor( 10 )
		
		:waitFor( function()
			
			return syncAgent:syncUntilDoneP()
		
		end )

		:waitFor( function()
			
			return syncAgent:syncUntilDoneP()
		
		end )
		
		:waitFor( function()
		
			return catalogSpec:getP( nextUpdatesLink )
				:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
				:mapDataP( function( _, updates )
				
					nextUpdatesLink = KSTableUtils.derefOpt( updates,
															 'links',
															 'next',
															 'href' )
					
					return 'data', sortUpdateByTypeAndHash( mapAwayUuidsAndTimestamps( updates ) )
				
				end )
			
		end )
		:continueWhenValueMatches( 'data',
			sortUpdateByTypeAndHash {
				base = "https://api.ozstage.com",
				links = {
					next = {
						href = "/v1.0/catalogs/(catalog ID)/updates?updated_since=(Oz update timestamp)",
					},
				},
				updates = {
					{
						_id = "hash of 32 chars",
						deleted = true,
						deleted_by = "urn:carousel:account:(account ID)",
						id = "urn:carousel:catalog:(catalog ID):asset:(album ID):asset:(asset ID)",
						type = "album_asset",
						updated = "(Oz timestamp)",
					},
				},
			} )

		-- Delete the album and one photo and watch those changes flow through.
		
		:waitFor( function()
		
			return catalogTestSession:docStore():setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					_deleted = true,
				}
		
		end )

		:waitFor( function()
		
			return catalogTestSession:docStore():setDocValueP
				{
					_id = photoId1,
					_slot = 'main',
					_deleted = true,
				}
		
		end )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )
	
		:waitFor( 10 )
		
		:waitFor( function()
			
			return syncAgent:syncUntilDoneP()
		
		end )
		:continueWhenValueMatches( 'data', true )
		
		:waitFor( function()
		
			return catalogSpec:getP( nextUpdatesLink )
				:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
				:mapDataP( function( _, updates )
				
					nextUpdatesLink = KSTableUtils.derefOpt( updates,
															 'links',
															 'next',
															 'href' )
					
					return 'data', sortUpdateByTypeAndHash( mapAwayUuidsAndTimestamps( updates ) )
		
		end )

		end )
		:continueWhenValueMatches( 'data',
			sortUpdateByTypeAndHash {
				base = "https://api.ozstage.com",
				links = {
					next = {
						href = "/v1.0/catalogs/(catalog ID)/updates?updated_since=(Oz update timestamp)",
					},
				},
				updates = {
					{
						_id = "hash of 32 chars",
						deleted = true,
						deleted_by = "urn:carousel:account:(account ID)",
						id = "urn:carousel:catalog:(catalog ID):asset:(album ID)",
						type = "album",
						updated = "(Oz timestamp)",
					},
					{
						_id = "hash of 32 chars",
						deleted = true,
						deleted_by = "urn:carousel:account:(account ID)",
						id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
						type = "asset",
						updated = "(Oz timestamp)",
					},
				},
			} )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

function WFOzSyncPeerTest.testSecondLevelOverlays_testNightly( test )

	local syncAgent, photoId1, catalogSpec, nextUpdatesLink, userId
	
	catalogTestSession:setTestName( test:testName() )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalogTestSession:catalogSpecP()
				:mapDataP( function( _, _catalogSpec )

					catalogSpec = _catalogSpec

					nextUpdatesLink = catalogSpec:catalogHref() .. "/updates"
					return 'data', true
				
				end )
		
		end )
	
		:waitFor( function()
		
			return catalogTestSession:ozUserIdP()
				:mapDataP( function( _, _userId )
					userId = _userId
				end )
		
		end )

		-- Load up a document
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:observeP( function( _, _syncAgent )
				
					syncAgent = _syncAgent
				
					-- Create an asset.
				
					local photoIds = catalogTestSession:fakePhotoIdsByCaptureDate()
					photoId1 = photoIds[ "2004-11-02T11:10:09" ]
				
				end )
		
		end )
		
		:waitFor( function()

			-- Upload the assets
			return syncAgent:syncUntilDoneP()
		
		end )
		
		:waitFor( 5 )
			-- wait for realtime hack
		
		:waitFor( function()

			-- Download Oz's version
			return syncAgent:syncUntilDoneP()
		
		end )
		
		:waitFor( function()
		
			-- Set a pick flag
			
			return catalogTestSession:docStore():applyOverlayP
				{
					_id = photoId1,
					_slot = 'main',
					reviews = {
						[ userId ] = {
							_merge = 'atomic',
							flag = "pick",
							date = "2013-07-10T08:32:13Z",
						},
					},
				}
			
		end )
		
		:waitFor( function()

			return syncAgent:syncUntilDoneP()
			
		end )
		
		:waitFor( 5 )
		
		:waitFor( function()
		
			-- modify just one part of a sub-sub table
			-- by making that pick a reject
			
			return catalogTestSession:docStore():applyOverlayP
				{
					_id = photoId1,
					_slot = 'main',
					reviews = {
						[ userId ] = {
							flag = "reject",
						},
					},
				}
			
		end )
		
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
	
		-- Wait a few moments (real-time hack, anyone?)
		
		:waitFor( 10 )

		-- Access Oz directly to verify that changes are present.
		
		:waitFor( function()
		
			return catalogSpec:getP( nextUpdatesLink )
				:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
				:mapDataP( function( _, updates )
					local assetDoc
					for _, doc in ipairs( updates.updates ) do
						if doc._id == photoId1 then
							assetDoc = doc
						end
					end
					return 'data', mapAwayUuidsAndTimestamps( assetDoc )
				
				end )
			
		end )
		:continueWhenValueMatchesFunction( function()
			return 'data',
			{
				_id = "hash of 32 chars",
				created = "(Oz timestamp)",
				id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
				link_templates = {
					[ "/rels/child_revision" ] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
					},
				},
				links = {
					[ "/rels/first_revision" ] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
					},
					[ "/rels/latest_revision" ] = {
						created_by = "urn:carousel:account:(account ID)",
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
					},
					[ "self" ] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
					},
				},
				payload = {
					captureDate = "2004-11-02T11:10:09",
					height = 16,
					importSource = {
						fileCreationDate = "2004-11-02T08:37:24",
						fileName = "fakeFile01.jpg",
						fileSize = 13493,
						folder = "/fake/path/to/fake",
						folderPath = "/fake/path/to/fake",
						importTimestamp = "(client timestamp)",
						importedBy = "hash of 32 chars",
						importedOnDevice = "WFOzCatalogTestSession.lua",
						sha256 = "hash of 64 chars",
					},
					reviews = {
						[ userId ] = {
							_merge = 'atomic',
							flag = "reject",
							date = "2013-07-10T08:32:13Z",
						},
					},
					type = "asset",
					version = 1,
					width = 16,
				},
				subtype = "jpeg",
				type = "asset",
				updated = "(Oz timestamp)",
			}
		end )
		
		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

function WFOzSyncPeerTest.testDevelopExceptionToSecondLevelOverlays_testNightly( test )

	-- See Watson bug 3968220: Disabling lens profile corrections in Lightroom
	-- does not cause the lens profile record to be removed from Oz.

	local syncAgent, photoId1, catalogSpec, nextUpdatesLink, userId
	
	catalogTestSession:setTestName( test:testName() )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalogTestSession:catalogSpecP()
				:mapDataP( function( _, _catalogSpec )

					catalogSpec = _catalogSpec

					nextUpdatesLink = catalogSpec:catalogHref() .. "/updates"
					return 'data', true
				
				end )
		
		end )
	
		:waitFor( function()
		
			return catalogTestSession:ozUserIdP()
				:mapDataP( function( _, _userId )
					userId = _userId
				end )
		
		end )

		-- Create some assets.
		
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:observeP( function( _, _syncAgent )
				
					syncAgent = _syncAgent
				
					local photoIds = catalogTestSession:fakePhotoIdsByCaptureDate()
					photoId1 = photoIds[ "2004-11-02T11:10:09" ]
				
				end )
		
		end )
		
		-- Take one asset and add a lens profile to it.
		
		:waitFor( function()
		
			return catalogTestSession:docStore():applyOverlayP {
			
				_id = photoId1,
				_slot = 'main',
				develop = {
					_merge = "atomic",
					croppedWidth = 3264,
					croppedHeight = 2448,
					xmpCameraRaw = [[<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="Adobe XMP Core 5.6-c011 79.156380, 2014/05/21-23:38:37 "> <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"> <rdf:Description rdf:about="" xmlns:crs="http://ns.adobe.com/camera-raw-settings/1.0/" crs:Version="9.0" crs:ProcessVersion="6.7" </rdf:Description> </rdf:RDF> </x:xmpmeta>]],
						-- Can we get away with a vastly stripped version of XMP
					userOrientation = 3,
					processingModel = "lightroom",
					profiles = {
						camera = {
							filename = "Nikon D3S Camera Neutral.dcp",
							href = "/v1.0/catalogs/f270b1c513494be08836d593cd1d6363/assets/654e0985f94c466ba6cee28f75188e41/revisions/7ba83c92d69342f0ad4f59d2eea39fb2/master",
						},
						lens = {
							filename = "iPhone 5s.lcp",
							--href = "/v1.0/catalogs/f270b1c513494be08836d593cd1d6363/assets/4d6971f47dd64bc2999ed3b144e2f1ce/revisions/870aa8c2e4dc488a94206bb5da05fe95/master",
						},
							-- Yeah, I know: Unlikely combination. But it's close enough for our purposes here.
					},
				},
			
			}
		
		end )
		
		-- Upload the assets.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()
		
		end )

		-- Wait for Oz to include assets in sync feed.
		
		:waitFor( 6 )
		
		-- Download from Oz.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()
		
		end )
		
		-- Access Oz directly to confirm that the profile is there.
		
		:waitFor( function()
		
			return catalogSpec:getP( nextUpdatesLink )
				:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
				:mapDataP( function( _, updates )
					local assetDoc
					for _, doc in ipairs( updates.updates ) do
						if doc._id == photoId1 then
							assetDoc = doc
						end
					end
					return 'data', mapAwayUuidsAndTimestamps( assetDoc )
				
				end )
			
		end )
		:continueWhenValueMatchesFunction( function()
			return 'data',
			{
				_id = "hash of 32 chars",
				created = "(Oz timestamp)",
				id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
				link_templates = {
					["/rels/child_revision"] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
					},
				},
				links = {
					["/rels/first_revision"] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
					},
					["/rels/latest_revision"] = {
						created_by = "urn:carousel:account:(account ID)",
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
					},
					["/rels/profiles/camera"] = {
						filename = "Nikon D3S Camera Neutral.dcp",
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/master",
					},
					["/rels/profiles/lens"] = {
						filename = "iPhone 5s.lcp",
					},
					["self"] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
					},
				},
				payload = {
					captureDate = "2004-11-02T11:10:09",
					develop = {
						_merge = "atomic",
						croppedHeight = 2448,
						croppedWidth = 3264,
						processingModel = "lightroom",
						profiles = {
							camera = {
								filename = "Nikon D3S Camera Neutral.dcp",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/master",
							},
							lens = {
								filename = "iPhone 5s.lcp",
							},
						},
						userOrientation = 3,
						xmpCameraRaw = "<x:xmpmeta xmlns:x=\"adobe:ns:meta/\" x:xmptk=\"Adobe XMP Core 5.6-c011 79.156380, 2014/05/21-23:38:37 \"> <rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"> <rdf:Description rdf:about=\"\" xmlns:crs=\"http://ns.adobe.com/camera-raw-settings/1.0/\" crs:Version=\"9.0\" crs:ProcessVersion=\"6.7\" </rdf:Description> </rdf:RDF> </x:xmpmeta>",
					},
					height = 16,
					importSource = {
						fileCreationDate = "2004-11-02T08:37:24",
						fileName = "fakeFile01.jpg",
						fileSize = 13493,
						folder = "/fake/path/to/fake",
						folderPath = "/fake/path/to/fake",
						importTimestamp = "(client timestamp)",
						importedBy = "hash of 32 chars",
						importedOnDevice = "WFOzCatalogTestSession.lua",
						sha256 = "hash of 64 chars",
					},
					type = "asset",
					version = 1,
					width = 16,
				},
				subtype = "jpeg",
				type = "asset",
				updated = "(Oz timestamp)",
			}

		end )
		
		-- Remove only the lens profile locally, leaving the camera profile.
		
		:waitFor( function()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = photoId1,
					_slot = 'main',
					develop = {
						profiles = {
							lens = {
								_delete = true,
							},
						},
					},
				}
			
		end )
		
		-- Push these changes and wait for Oz to include in the sync feed.
		
		:waitFor( function()

			return syncAgent:syncUntilDoneP()
			
		end )
		
		:waitFor( 10 )
		
		-- Access Oz directly to verify that changes are present.
		
		:waitFor( function()
		
			return catalogSpec:getP( nextUpdatesLink )
				:mapDataP( KSJsonDecoder.decodeWebSafeJson_stripNullP )
				:mapDataP( function( _, updates )
					local assetDoc
					for _, doc in ipairs( updates.updates ) do
						if doc._id == photoId1 then
							assetDoc = doc
						end
					end
					return 'data', mapAwayUuidsAndTimestamps( assetDoc )
				
				end )
			
		end )
		:continueWhenValueMatchesFunction( function()
			return 'data',
			{
				_id = "hash of 32 chars",
				created = "(Oz timestamp)",
				id = "urn:carousel:catalog:(catalog ID):asset:(asset ID)",
				link_templates = {
					["/rels/child_revision"] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/children/{revision_id}",
					},
				},
				links = {
					["/rels/first_revision"] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
					},
					["/rels/latest_revision"] = {
						created_by = "urn:carousel:account:(account ID)",
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)",
					},
					["/rels/profiles/camera"] = {
						filename = "Nikon D3S Camera Neutral.dcp",
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/master",
					},
					["self"] = {
						href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)",
					},
				},
				payload = {
					captureDate = "2004-11-02T11:10:09",
					develop = {
						_merge = "atomic",
						croppedHeight = 2448,
						croppedWidth = 3264,
						processingModel = "lightroom",
						profiles = {
							camera = {
								filename = "Nikon D3S Camera Neutral.dcp",
								href = "/v1.0/catalogs/(catalog ID)/assets/(asset ID)/revisions/(revision ID)/master",
							},
						},
						userOrientation = 3,
						xmpCameraRaw = "<x:xmpmeta xmlns:x=\"adobe:ns:meta/\" x:xmptk=\"Adobe XMP Core 5.6-c011 79.156380, 2014/05/21-23:38:37 \"> <rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"> <rdf:Description rdf:about=\"\" xmlns:crs=\"http://ns.adobe.com/camera-raw-settings/1.0/\" crs:Version=\"9.0\" crs:ProcessVersion=\"6.7\" </rdf:Description> </rdf:RDF> </x:xmpmeta>",
					},
					height = 16,
					importSource = {
						fileCreationDate = "2004-11-02T08:37:24",
						fileName = "fakeFile01.jpg",
						fileSize = 13493,
						folder = "/fake/path/to/fake",
						folderPath = "/fake/path/to/fake",
						importTimestamp = "(client timestamp)",
						importedBy = "hash of 32 chars",
						importedOnDevice = "WFOzCatalogTestSession.lua",
						sha256 = "hash of 64 chars",
					},
					type = "asset",
					version = 1,
					width = 16,
				},
				subtype = "jpeg",
				type = "asset",
				updated = "(Oz timestamp)",
			}
		end )
		
		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function WFOzSyncPeerTest.testSendDiagnosticLogData_testNightly( test )

	-- Since diagnostic log is expected to evolve frequently, we're not
	-- verifying specific content. The main concern is that it doesn't fall
	-- over and die when generating the log.

	catalogTestSession:setTestName( test:testName() )
	WFAsyncTestCase( test, TEST_WAIT_TIME )
	
		:waitFor( function()
		
			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )
					return WFDiagnosticLog.processLogSectionP(
						syncAgent:sendDiagnosticLogData() )
				end )
		
		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

-- FIXME [ppaurav > jandrew 2013-07-16]
-- Watson Bug #3578247
if WFPlatformInfo.isWindows then
	return {}
else
	return WFOzSyncPeerTest
end
