--[[----------------------------------------------------------------------------

WFOzSyncStatus.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

You may want to configure the rendition definitions via something like:

	WFCapabilities_setCapability( "WFOzBinaryUploader.renditionTypes",
								  "thumbnail2x,2048" );

	WFCapabilities_setCapability( "WFOzBinaryUploader.proxyTypes", "2560" );

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.sync_engine
local WFOzBinaryUploadUtils = require 'WFOzBinaryUploadUtils'
local WFOzSyncUtils = require 'WFOzSyncUtils'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFOzSyncStatus'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local shouldUploadRendition = WFOzBinaryUploadUtils.shouldUploadRendition
local shouldUploadProxy = WFOzBinaryUploadUtils.shouldUploadProxy
local shouldUploadMaster = WFOzBinaryUploadUtils.shouldUploadMaster


--============================================================================--

-- @mark WFOzSyncStatus namespace

local WFOzSyncStatus = {}

--------------------------------------------------------------------------------

local function ozSyncStatusForUI_table()

	local renditionTypes = WFOzBinaryUploadUtils.getRenditionTypes()
	local proxyTypes = WFOzBinaryUploadUtils.getProxyTypes()

	local nRenditionTypes = #renditionTypes
	local nProxyTypes = #proxyTypes
	
	local uploadErrorPaths = {}
	
	for i = 1, nRenditionTypes do
		uploadErrorPaths[ i ] = { 'renditions', renditionTypes[ i ] }
	end
	
	for i = 1, nProxyTypes do
		uploadErrorPaths[ nRenditionTypes + i ] = { 'proxy', proxyTypes[ i ] }
	end
	
	uploadErrorPaths[ nRenditionTypes + nProxyTypes + 1 ] = { 'master' }
	
	local nUploadErrorPaths = #uploadErrorPaths
	
	return {
	
		sql = [[CREATE TABLE ozSyncStatusForUI( status )]],
	
		hasOneRowPerDocument = true,

		map = function( docValue, emit )

			-- We only care about assets. Other doc types are metadata only
			-- and are thus handled by WFSyncAgent.

			if docValue.type ~= 'asset' then
				return
			end

			local subtype = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'subtype' )
			if subtype and subtype ~= 'jpeg' then
				return
			end
	
			logger:debugf( 'ozSyncStatusForUI reindexing\n\n%s\n\n', docValue )

			-- Did we encounter an error syncing to Oz?
			
			local syncStatusHash = KSTableUtils.derefOpt( docValue,
														  '_syncStatus' )

			if type( syncStatusHash ) == 'table' then
				for syncPeerName, peerStatus in pairs( syncStatusHash ) do
					if peerStatus.error and peerStatus.error.permanent then
						logger:debugf( "ozSyncStatusForUI emit 'error'" )
						emit( 'error' )
						return
					end
				end
			end
			
			-- If this asset is deleted and has also been deleted on Oz,
			-- then there can be no status to report. None of the other tests
			-- below are relevant.
			
			local links = KSTableUtils.derefOpt( docValue,
												 '_serverProvidedMetadata',
												 'links' )

			if docValue._deleted then

				if links then
					logger:debugf( "ozSyncStatusForUI emit 'metadata'" )
					emit( 'metadata' )
					return
				end
				
				return
				
			end
			
			-- Did we encounter an error uploading binaries to Oz?
			
			local willNotUploadMaster = KSTableUtils.derefOpt( docValue,
															   'importSource',
															   'willNotUploadMaster' )
			
			local localOnlyHash = KSTableUtils.derefOpt( docValue, '_localOnly' )
			if localOnlyHash then
			
				for i = 1, nUploadErrorPaths do
				
					-- Disregard master error status if we're not going to upload
					-- master. Some older versions of WF would record a missing status
					-- here.

					local binaryInfoHash = KSTableUtils.derefOpt( localOnlyHash,
																  unpack( uploadErrorPaths[ i ] ) )

					if not( uploadErrorPaths[ i ][ 1 ] == 'master' and willNotUploadMaster ) then

						if binaryInfoHash and binaryInfoHash.missing then
							logger:debugf( "ozSyncStatusForUI emit 'missing'" )
							emit( 'missing' )
							return
						end
					
						if binaryInfoHash and binaryInfoHash.error then
							logger:debugf( "ozSyncStatusForUI emit 'error'" )
							emit( 'error' )
							return
						end
						
					end

				end
			
			end
			
			-- Do we have nothing on Oz yet?
			
			if not links then
				logger:debugf( "ozSyncStatusForUI emit 'metadata'" )
				emit( 'metadata' )
				return
			end
			
			-- Does Oz have the same metadata that we do?
			
			local ozHash = KSTableUtils.derefOpt( docValue,
												  '_serverProvidedMetadata',
												  'wfOzPayloadHash' )
			
			local myHash = WFOzSyncUtils.hashDocValue( docValue )
			
			if ozHash ~= myHash then
				logger:debugf( "ozSyncStatusForUI emit 'metadata'" )
				emit( 'metadata' )
				return
			end
			
			-- Which rendition types need to be uploaded?

			for i = 1, nRenditionTypes do
			
				local renditionType = renditionTypes[ i ]
			
				if shouldUploadRendition( docValue, renditionType ) then
					logger:debugf( "ozSyncStatusForUI emit 'renditions' for type %s",
								   renditionType )
					emit( 'renditions' )
					return
				end
			
			end
			
			-- And how 'bout a master?
			
			if ( not willNotUploadMaster ) and shouldUploadMaster( docValue ) then
				logger:debug( "ozSyncStatusForUI emit 'master'" )
				emit( 'master' )
				return
			end

			-- Which proxy types need to be uploaded?
			
			for i = 1, nProxyTypes do
			
				local proxyType = proxyTypes[ i ]
				
				if shouldUploadProxy( docValue, proxyType ) then
					logger:debugf( "ozSyncStatusForUI emit 'proxies' for type %s",
								   proxyType )
					emit( 'proxies' )
					return
				end
			
			end
			
			logger:debug( "ozSyncStatusForUI emit 'done'" )
			emit( 'done' )

		end,
	
	}
	
end

--------------------------------------------------------------------------------

local function ozSyncStatusForUI_index()

	return [[CREATE INDEX ozSyncStatusForUI_status ON ozSyncStatusForUI( status )]]

end

--------------------------------------------------------------------------------

function WFOzSyncStatus.filterIndexDefinitions( tableDefinitions, indexDefinitions )
	
	tableDefinitions.ozSyncStatusForUI = ozSyncStatusForUI_table()
	
	indexDefinitions.ozSyncStatusForUI_status = ozSyncStatusForUI_index()
	
	return tableDefinitions
	
end

--------------------------------------------------------------------------------

function WFOzSyncStatus.generateIndexSchema( indexStoreSchema )

	--check( 'WFIndexStoreSchema' )
		-- Ugh. Yet another reason why I hate mocks in unit tests.
		-- Mocks created at OzCatalogTestUtils.buildMocksForTestingCatalogService
		-- mean that the WFIndexStoreSchema mock object doesn't pass WFArguments
		-- smell test so now I have to disable this safety check. Only discovered
		-- this when I made a change in a yet another unrelated part of the app
		-- that introduced a dependency on this file, but only in Lightroom's
		-- test harness environment. (Thanks to differences how we configure
		-- OzCatalogIndexDefinitions via WFCapabilities, naturally.)

	indexStoreSchema:addTable( ozSyncStatusForUI_table() )
	
	indexStoreSchema:addIndex( ozSyncStatusForUI_index() )
	
end

--------------------------------------------------------------------------------

local memoizedFunctions = {}

--------------------------------------------------------------------------------

local function makeMemoizedUnparamaterizedFunction( sql, mapFn )

	memoizedFunctions[ sql ] = WFCloseSupport.memoizeFn(
		
		function( indexStore )
	
			return indexStore:queryB( sql )
				:mapB( mapFn )
		
		end )

	return function( indexStore )

		check( 'WFIndexStore' )
		return memoizedFunctions[ sql ]( indexStore ):relayB()
	
	end

end

--------------------------------------------------------------------------------

local function makeMemoizedCountQueryFunction( sql )

	return makeMemoizedUnparamaterizedFunction( sql,
		function( value )
		
			if WFUnknown.isUnknown( value ) then
				return value
			end
			
			local count = KSTableUtils.derefOpt( value, 1, 1 )
			return count or WFUnknown.value()
		
		end )

end

--------------------------------------------------------------------------------

local function makeMemoizedParameterizedFunction( sql, mapFn )

	memoizedFunctions[ sql ] = WFCloseSupport.memoizeFn(
		
		function( indexStore )
	
			return WFCloseSupport.memoizeFn(
				
				function( param )
			
					return indexStore:queryB( sql, param )
						:mapB( mapFn )
				
				end )
		
		end )

	return function( indexStore, param )

		check( 'WFIndexStore', '+string' )  -- 'uuid'?
		return memoizedFunctions[ sql ]( indexStore )( param ):relayB()
	
	end
	
end

--------------------------------------------------------------------------------

local function makeOwningMemoizedParameterizedFunction( params )

	local mapFn = assert( params.map )
	params.map = nil

	local sql = assert( params.sql )

	memoizedFunctions[ sql ] = WFCloseSupport.memoizeFn(
		
		function( indexStore )
	
			return WFCloseSupport.memoizeFn(
				
				function( param )
			
					local newParams = KSTableUtils.shallowcopy( params )
					
					newParams[ 1 ] = param
			
					return indexStore:queryB( newParams )
						:closingMapB( mapFn )
				
				end )
		
		end )

	return function( indexStore, param )

		check( 'WFIndexStore', '+string' )  -- 'uuid'?
		return memoizedFunctions[ sql ]( indexStore )( param ):autoCloseRelayB()
	
	end
	
end

--------------------------------------------------------------------------------

WFOzSyncStatus.countAssetsToBeUploadedB = makeMemoizedCountQueryFunction(
	[[SELECT count( rowid ) FROM ozSyncStatusForUI
		WHERE status NOT IN ( 'missing', 'error', 'done' )]] )

--------------------------------------------------------------------------------

WFOzSyncStatus.countAssetsCompletedB = makeMemoizedCountQueryFunction(
	[[SELECT count( rowid ) FROM ozSyncStatusForUI
		WHERE status IN ( 'missing', 'error', 'done' )]] )

--------------------------------------------------------------------------------

WFOzSyncStatus.countAssetsB = makeMemoizedCountQueryFunction(
	[[SELECT count( rowid ) FROM ozSyncStatusForUI]] )

--------------------------------------------------------------------------------

WFOzSyncStatus.errorAssetsB = makeMemoizedUnparamaterizedFunction(

	[[SELECT docId FROM ozSyncStatusForUI
		WHERE status IN ( 'error', 'missing' )]],

	function( value )

		if type( value ) ~= 'table' or WFUnknown.isUnknown( value ) then
			return value
		end
		
		local result = {}
		
		for i = 1, #value do
			result[ i ] = value[ i ][ 1 ]
		end
		
		return result

	end )

--------------------------------------------------------------------------------

WFOzSyncStatus.syncStatusForAssetB = makeMemoizedParameterizedFunction(
	[[SELECT status FROM ozSyncStatusForUI WHERE docId == ?]],
	function( value )
					
		if WFUnknown.isUnknown( value ) then
			return value
		end
		
		return KSTableUtils.derefOpt( value, 1, 1 )
	
	end )

--------------------------------------------------------------------------------

-- @mark per-album sync status

-- WARNING: The album_asset index table must be included for these functions
-- to work properly. (See WFOzAlbumAssetIndex.lua.)

-- TODO [scouten 2014-03-18]: Want to make that happen automatically without
-- requiring separate action and also without creating a duplicate table.

--------------------------------------------------------------------------------

WFOzSyncStatus.albumSizeB = makeOwningMemoizedParameterizedFunction {

	sql = [[SELECT count( rowid ) FROM album_asset WHERE albumId == ?]],

	map = function( value )
					
		if WFUnknown.isUnknown( value ) then
			return value
		end
		
		local status = KSTableUtils.derefOpt( value, 1, 1 )
		return status
	
	end,
	
	calm = 1,
	
}

--------------------------------------------------------------------------------

WFOzSyncStatus.uploadedAssetsForAlbumB = makeOwningMemoizedParameterizedFunction {

	sql = [[SELECT count( ozSyncStatusForUI.docId ) FROM ozSyncStatusForUI
		INNER JOIN album_asset ON album_asset.assetId == ozSyncStatusForUI.docId
		WHERE album_asset.albumId == ?
			AND ozSyncStatusForUI.status == 'done']],

	map = function( value )
					
		if WFUnknown.isUnknown( value ) then
			return value
		end
		
		local status = KSTableUtils.derefOpt( value, 1, 1 )
		return status
	
	end,
	
	calm = 1.5,
	
}

--------------------------------------------------------------------------------

return WFOzSyncStatus
