--[[----------------------------------------------------------------------------

WFOzSyncStatusTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local WFCapabilities = require 'WFCapabilities'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.oz_client.sync_engine.binary_upload
local WFOzBinaryUploader = require 'WFOzBinaryUploader'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzCatalogTestSession = require 'WFOzCatalogTestSession'
local WFOzSyncStatus = require 'WFOzSyncStatus'
local WFOzTestBinaryProviders = require 'WFOzTestBinaryProviders'


--============================================================================--

-- @mark WFOzSyncStatusTest

local WFOzSyncStatusTest = {}

--------------------------------------------------------------------------------

local catalogTestSession, docStore, indexStore, wfOzSyncAgent, wfOzBinaryUploader
local fakePhotoIdsByCaptureDate
local TEST_WAIT_TIME = 360

--------------------------------------------------------------------------------

local function addTestSessionCloseTarget( obj )

	catalogTestSession:addCloseTarget( obj )

end

--------------------------------------------------------------------------------

function WFOzSyncStatusTest.setUp( coveragePath )

	if coveragePath then
		TEST_WAIT_TIME = TEST_WAIT_TIME * 2
	end

	WFCapabilities.setCapability( 'WFOzBinaryUploader.renditionTypes', 'thumbnail2x,2048' )
	WFCapabilities.setCapability( 'WFOzBinaryUploader.proxyTypes', '2560' )

	catalogTestSession = WFOzCatalogTestSession( 'WFOzSyncStatusTest' )

	indexStore = nil
	wfOzSyncAgent = nil
	wfOzBinaryUploader = nil

end

--------------------------------------------------------------------------------

function WFOzSyncStatusTest.tearDown()

	if docStore then
		docStore = nil
	end

	if catalogTestSession then
		catalogTestSession:releaseTestUserP()
		catalogTestSession:close()
		catalogTestSession = nil
	end

end

--------------------------------------------------------------------------------

local function getOrCreateIndexStore()

	if not docStore then
		docStore = catalogTestSession:docStore()
	end

	if not indexStore then

		indexStore = WFIndexStore( catalogTestSession:docStore(),
								   'SyncStatusTestIndexDefinitions.lua' )

		addTestSessionCloseTarget( indexStore )

	end

	return indexStore

end

--------------------------------------------------------------------------------

local function wfOzBinaryUploaderP( binaryProvider )

	getOrCreateIndexStore()

	return catalogTestSession:ozSyncAgentWithMetadataSyncCompletedP()

		:mapDataP( function()

			return catalogTestSession:catalogSpecP()

		end )

		:mapDataP( function( _, catalogSpec )

			wfOzBinaryUploader = WFOzBinaryUploader
				{
					catalogSpec = catalogSpec,
					docStore = catalogTestSession:docStore(),
					indexStore = indexStore,
					binaryProvider = binaryProvider,
					noScheduler = true,
				}

			addTestSessionCloseTarget( wfOzBinaryUploader )

			fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			return 'data', wfOzBinaryUploader

		end )

end

--------------------------------------------------------------------------------

function WFOzSyncStatusTest.testSyncStatusBehaviors_testNightly( test )

	-- We test most of the WFOzSyncStatus APIs in one combined test because
	-- the various uploads take a considerable amount of time.

	getOrCreateIndexStore()

	local countAssetsB = WFOzSyncStatus.countAssetsB( indexStore )
	addTestSessionCloseTarget( countAssetsB )

	local countAssetsToBeUploadedB = WFOzSyncStatus.countAssetsToBeUploadedB( indexStore )
	addTestSessionCloseTarget( countAssetsToBeUploadedB )

	local errorAssetsB = WFOzSyncStatus.errorAssetsB( indexStore )
	addTestSessionCloseTarget( errorAssetsB )

	local expectedErrorManifest = {}

	local syncStatusForAssetB

	catalogTestSession:setTestName( test:testName() )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Initial condition: # photos to upload should be zero since we haven't
		-- added anything to the catalog yet.

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 0 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 0 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		-- Enable automatic metadata sync for this catalog.

		:waitFor( function()

			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', true

				end )

		end )

		-- Now add the three fake photos. We should see those three awaiting upload.

		:waitFor( catalogTestSession:catalogSpecWithThreePhotosP() )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 3 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			expectedErrorManifest[ 1 ] = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ]

			return 'ok'

		end )

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			syncStatusForAssetB = WFOzSyncStatus.syncStatusForAssetB( indexStore,
									fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
			addTestSessionCloseTarget( syncStatusForAssetB )

			return syncStatusForAssetB

		end )
		:continueWhenValueMatches( 'metadata' )

		-- Now push metadata for the photos to Oz. Count should be unchanged.

		:waitFor( catalogTestSession:ozSyncAgentWithMetadataSyncCompletedP() )

		:waitFor( 10 )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 3 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'renditions' )

		-- Put an error into one of the documents. We know Oz will reject an asset
		-- with a location hash with invalid lat/long, so we'll use that as our test case.

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					_slot = 'main',
					location = {
						latitude = -121247.734290,
						longitude = 122.646336,
					},
				}

		end )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 2 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( expectedErrorManifest )

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			syncStatusForAssetB = WFOzSyncStatus.syncStatusForAssetB( indexStore,
									fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ] )
			addTestSessionCloseTarget( syncStatusForAssetB )

			return syncStatusForAssetB

		end )
		:continueWhenValueMatches( 'error' )

		-- Now fix the error and verify that we go back to normal state.

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					_slot = 'main',
					location = {
						latitude = -47.734290,
						longitude = 122.646336,
					},
					_syncStatus = {
						oz_TEST = {
							_delete = true,
						},
					},
				}

		end )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 3 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'renditions' )

		-- Spin up the binary uploader. Now the wfOzBinaryUploader file variable
		-- will be defined.

		:waitFor( function()

			return wfOzBinaryUploaderP( WFOzTestBinaryProviders.testBinaryProvider() )

		end )

		-- Now upload one set of renditions. Count should be unchanged.

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'rendition', 'thumbnail2x' )

		end )

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'rendition', '2048' )

		end )

		:waitFor( 10 )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 3 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'master' )

		-- Now upload master for same image. Photo we're watching should now require
		-- a proxy upload. (Historical note: It used to be that we could upload proxy
		-- first, followed by master. Revel took advantage of that workflow. As of
		-- early January 2016, that scenario will invalidate the previously-uploaded
		-- proxy.)

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'master' )

		end )

		:waitFor( 10 )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'proxies' )

		-- Now upload a proxy for same image. Count should decrement since we've now
		-- uploaded the full complement of binaries.

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'proxy', '2560' )

		end )

		:waitFor( 10 )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 2 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'done' )

		-- Finish uploading all binaries. Count should (eventually) fall to zero.

		:waitFor( function()

			wfOzBinaryUploader:startUploading()
			return "ok"

		end )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 0 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'done' )

		-- Introduce a metadata change. Should cause pending-upload count to bounce back to 1.

		:waitFor( function()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					_slot = 'main',
					foo = 57,
				}

		end )

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 1 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'metadata' )

		-- And when we sync this metadata change, it should fall back to zero.

		:waitFor( countAssetsB )
		:continueWhenValueMatches( 3 )

		:waitFor( countAssetsToBeUploadedB )
		:continueWhenValueMatches( 0 )

		:waitFor( errorAssetsB )
		:continueWhenValueMatches( {} )

		:waitFor( function()
			return syncStatusForAssetB
		end )
		:continueWhenValueMatches( 'done' )

		-- We're done: Release test user account.

		:waitFor( function()

			return catalogTestSession:releaseTestUserP()

		end )

end

--------------------------------------------------------------------------------

function WFOzSyncStatusTest.testAlbumStatusBehaviors_testNightly( test )

	getOrCreateIndexStore()

	local albumId = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId1 = KSIdGenerator.uuidInLowercaseHex()
	local albumAssetId2 = KSIdGenerator.uuidInLowercaseHex()

	local fakePhotoIdsByCaptureDate

	local albumSizeB = WFOzSyncStatus.albumSizeB( indexStore, albumId )
	addTestSessionCloseTarget( albumSizeB )

	local uploadedAssetsForAlbumB = WFOzSyncStatus.uploadedAssetsForAlbumB( indexStore, albumId )
	addTestSessionCloseTarget( uploadedAssetsForAlbumB )

	catalogTestSession:setTestName( test:testName() )

	WFAsyncTestCase( test, TEST_WAIT_TIME )

		-- Initial condition: # photos to upload should be zero since we haven't
		-- added anything to the catalog yet.

		:waitFor( albumSizeB )
		:continueWhenValueMatches( 0 )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Enable automatic metadata sync for this catalog.

		:waitFor( function()

			return catalogTestSession:ozSyncAgentP()
				:mapDataP( function( _, syncAgent )

					syncAgent:startAutomaticSync()
					return 'data', true

				end )

		end )

		-- Now add the three fake photos. We should still see # uploaded == 0.

		:waitFor( catalogTestSession:catalogSpecWithThreePhotosP() )

		:waitFor( function()

			fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()
			return 'ok'

		end )

		:waitFor( albumSizeB )
		:continueWhenValueMatches( 0 )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Create album and add assets to it one-by-one.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = albumId,
					_slot = 'main',
					type = 'album',
					name = 'test album',
				}

		end )

		:waitFor( albumSizeB )
		:continueWhenValueMatches( 0 )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = albumAssetId1,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					},
				}

		end )

		:waitFor( albumSizeB )
		:continueWhenValueMatches( 1 )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = albumAssetId2,
					_slot = 'main',
					type = 'album_asset',
					album = {
						_id = albumId,
					},
					asset = {
						_id = fakePhotoIdsByCaptureDate[ "2004-11-02T11:10:09" ],
					},
				}

		end )

		:waitFor( albumSizeB )
		:continueWhenValueMatches( 2 )

		-- Now push metadata for the photos to Oz. Count should be unchanged.

		:waitFor( catalogTestSession:ozSyncAgentWithMetadataSyncCompletedP() )

		:waitFor( 10 )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Put an error into one of the documents. We know Oz will reject an asset
		-- with an incomplete location hash, so we'll use that as our test case.

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					_slot = 'main',
					location = {
						latitude = 47.734290,
						bogus = 'whatever',
					},
				}

		end )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Now fix the error and verify that we go back to normal state.

		:waitFor( function()

			local fakePhotoIdsByCaptureDate = catalogTestSession:fakePhotoIdsByCaptureDate()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					_slot = 'main',
					location = {
						longitude = 122.646336,
						bogus = { _delete = true },
					},
					_syncStatus = {
						oz_TEST = {
							_delete = true,
						},
					},
				}

		end )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Spin up the binary uploader. Now the wfOzBinaryUploader file variable
		-- will be defined.

		:waitFor( function()

			return wfOzBinaryUploaderP( WFOzTestBinaryProviders.testBinaryProvider() )

		end )

		-- Now upload one set of renditions. Count should be unchanged.

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'rendition', 'thumbnail2x' )

		end )

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'rendition', '2048' )

		end )

		:waitFor( 10 )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Now upload master for same image. Count should still be unchanged.

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'master' )

		end )

		:waitFor( 10 )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 0 )

		-- Now upload proxy for same image. Count should increment.
		-- We do master first, because the proxy we just uploaded would get
		-- marked invalid otherwise.

		:waitFor( function()

			return wfOzBinaryUploader:uploadOneBinaryP(
							fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
							'proxy', '2560' )

		end )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 1 )

		-- Finish uploading all binaries. Count should (eventually) increase to two.

		:waitFor( function()

			wfOzBinaryUploader:startUploading()
			return "ok"

		end )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 2 )

		-- Make a metadata change, but don't sync it. Count should fall to one.

		:waitFor( function()

			return catalogTestSession:docStore():applyOverlayP
				{
					_id = fakePhotoIdsByCaptureDate[ "2012-05-02T08:07:06" ],
					_slot = 'main',
					foo = 57,
				}

		end )

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 1 )

		-- And when we sync this metadata change, it should revert to two.

		:waitFor( uploadedAssetsForAlbumB )
		:continueWhenValueMatches( 2 )

		:waitFor( function()

			return catalogTestSession:releaseTestUserP()

		end )

end

--------------------------------------------------------------------------------

return WFOzSyncStatusTest
