--[[----------------------------------------------------------------------------

WFOzSyncUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFTimer = require 'WFTimer'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFDigest = require 'WFDigest'
local WFUnknown = require 'WFUnknown'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'

	-- WF.web.http_client
local WFReachability = require 'WFReachability'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local sha1 = assert( WFDigest.SHA1 )

local hashDocValueLogger = WFLogger 'WFOzSyncUtils :: hashDocValue'
--	:setLevel 'debug'

local ozUploadLogger = WFLogger 'WFOzSyncUtils :: upload'
--	:setLevel 'debug'


--============================================================================--

-- @mark WFOzSyncUtils namespace

local WFOzSyncUtils = {}

--------------------------------------------------------------------------------
--[[!
	\brief Return a hash code from the syncable portions of a doc value.
	
	\param (table) doc value
	
	\return (string) hash code
]]--

function WFOzSyncUtils.hashDocValue( docValue )

	check( 'table' )

	docValue = KSTableUtils.deepCopy( docValue )

	docValue._serverProvidedMetadata = nil
	docValue._id = nil
	docValue._idLocal = nil
	docValue._slot = nil
	docValue._hash = nil
	docValue._changeCounter = nil
	docValue._localOnly = nil
	docValue._timestamp = nil
	docValue._syncStatus = nil

	hashDocValueLogger:infof( "stripped doc value = %s\n\n", docValue )

	local archive = WFTableArchiving.tableToString( docValue )
	return sha1.digest( archive )

end

--------------------------------------------------------------------------------

local function noOpUploadP()

	return KSPromises.resolvedPromiseP( 'data', true )

end

--------------------------------------------------------------------------------

-- FIXME [scouten 2012-04-17]: This should move out to a string utils file.

local function split( str, pat )

	local t = {}  -- NOTE: use {n = 0} in Lua-5.0
	local fpat = "(.-)" .. pat
	local last_end = 1

	local s, e, cap = str:find( fpat, 1 )

	while s do

		if s ~= 1 or cap ~= "" then
			table.insert( t, cap )
		end

		last_end = e+1
		s, e, cap = str:find( fpat, last_end )

	end

		if last_end <= #str then
			cap = str:sub( last_end )
			table.insert( t, cap )
		end

	return t

end

--------------------------------------------------------------------------------

function WFOzSyncUtils.updateStatusFromHttpResultFn( self, status, message, headers )

	local newStatus = nil
	
	local existingHostError = self._currentHostErrorB:valueNow()
	
	-- Certain errors are sticky. Once we get there, we stay there.
	
	if existingHostError == 'enterpriseUsageUnauthorized'
		or existingHostError == 'forcedUpgrade'
	then
		return
	end

	if status ~= 'data' then

		-- TODO_SNOWLESS [scouten 2013-12-10]: Capture error data for diagnostic log.
		-- See args.errorCapture.downloadLatestFromOzP in WFOzDownloadTask.lua.

		if headers == nil and type( message ) == 'string'
			and message ~= 'Logged out' then

			newStatus = 'networkFailure'
		
		elseif type( headers ) == 'table' then
		
			local statusCode = tonumber( headers.status_code )
			
			if statusCode == 503 then

				newStatus = 'serviceUnavailable'

			elseif statusCode == 404 then
			
				newStatus = 'itemDeleted'
			
			elseif statusCode == 403 then

				-- TODO [scouten 2014-02-20]: Switch this over to using the new
				-- WFSyncAgent kill switch mechanism for Lightroom 5.5.
				
				if headers[ 'X-Result-Code' ] == '4999' then
					newStatus = 'forcedUpgrade'
				elseif headers[ 'X-Result-Code' ] == '1046' then
					newStatus = 'enterpriseUsageUnauthorized'
				else
					newStatus = 'itemUnavailable'
				end

			end

			-- If forced upgrade, try to find upgrade link.
			
			if newStatus == 'forcedUpgrade' then

				-- TODO [scouten 2014-02-20]: Switch this over to using the new
				-- WFSyncAgent kill switch mechanism for Lightroom 5.5.
				
				-- First cut: Parse message as JSON and extract.
				-- Known good as of January 2014.
				
				local _, messageJson = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull,
											  message )
				
				local upgradeLink = KSTableUtils.derefOpt( messageJson,
														   'links',
														   '/rels/client_app',
														   'href' )
				
				if upgradeLink then

					self._forcedUpgradeLinkB:set( upgradeLink )

				else
				
					-- First cut: Inherited from older Revel code. Not sure
					-- if this is still valid.
		
					local errorData = message
								 		and type( message ) == 'string'
										and split( message, ',' )
				
					if errorData and errorData[ 2 ] == '4999' then
						self._forcedUpgradeLinkB:set( errorData[ 3 ] )
					end
				
				end
				
			end
				
		end
	
	end
	
	local currentHostErrorB = self._currentHostErrorB
	
	if not WFCloseSupport.hasBeenOrWillBeClosed( currentHostErrorB ) then
		currentHostErrorB:set( newStatus )
	end
	
end

--------------------------------------------------------------------------------

local function uploadDocumentP( self, delta, uploadFunctionsByDocType )

	local docType = KSTableUtils.derefOpt( delta, '_fullDocValue', 'type' )
		or KSTableUtils.derefOpt( delta, '_previouslyAgreedValue', 'type' )

	local uploadFn = docType and uploadFunctionsByDocType[ docType ]
	
	if docType == nil and delta._deleted == true then
		uploadFn = noOpUploadP
	else
		assert( type( docType ) == 'string',
				"Can not upload documents without a 'type' member" )
	end
	
	-- Oz doesn't like documents with the ID 'pending' since it's not a proper
	-- UUID, so we head those off here.
	
	if delta._id == 'pending' then
		return KSPromises.resolvedPromiseP( 'error',
			{
				message = "Illegal document ID",
				permanent = true,
			} )
	end
	
	if uploadFn == 'nop' then
		uploadFn = noOpUploadP
	end

	if not uploadFn then
	
		ozUploadLogger:errorf( "No upload function for doc type %s", docType )

		uploadFn = noOpUploadP

	end
	
	return uploadFn( self, delta )

		:observeP( function( ... )

			WFOzSyncUtils.updateStatusFromHttpResultFn( self, ... )
			
		end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Shared code for :sendUpdatesP for catalog and user sync peers.
]]--

function WFOzSyncUtils.sendUpdatesP_imp( self, updates, uploadFunctionsByDocType )

	check( 'WFOzSyncPeer|WFOzUserSyncPeer', 'table', 'table' )
	
	local skippedDocs = {}
	
	return KSPromises.mapArrayP(
		function( document )

			return uploadDocumentP( self, document, uploadFunctionsByDocType )
	
				:mapP( function( status, message )
	
					-- Track :skip: results and make sure they get relayed
					-- properly back to WFSyncAgent.
					
					if status == 'data' and message == ':skip:' then
						skippedDocs[ document._id ] = ':skip:'
					end
	
					-- We want to convert errors to data so that mapArrayP
					-- handles all documents and returns an array of results.
					-- We can then unwrap the results and report back to the
					-- caller.
	
					return 'data', { id = document._id, status = status, message = message }
	
				end )
				
		end, updates )

		:mapP( function( status, results )
		
			if type( results ) ~= 'table' then
			
				if status == 'data' and next( skippedDocs ) then
					return status, skippedDocs
				else
					return status, results
				end

			end

			-- Results should be an array of results where each result is a table with
			-- an ID (the document ID), status ('data' or 'error') and in the case of an
			-- error, a message. Convert this into a single array that we can pass back
			-- to sync engine.

			local errorResults = skippedDocs
			local hadAnyErrors = false
			
			for i = 1, #results do
			
				local result = results[ i ]

				if result.status == 'error' then
			
					hadAnyErrors = true
			
					ozUploadLogger:warnf( "Document ID %s failed with message %s",
										  result.id,
										  result.message )

					errorResults[ result.id ] = result.message or "unknown error"
			
				else
				
					assert( result.status == 'data' )

				end
			
			end

			if hadAnyErrors then
				return 'error', errorResults
			end
			
			if next( skippedDocs ) then
				return 'data', skippedDocs
			else
				return 'data', 'all docs succeeded'
			end
			
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

local minDownloadPollingIntervals = {
	unreachable = 30,
	wifi = 2,
	wwan = 10,
	serviceUnavailable = 60,
	default = 2 -- backup value
}

local maxDownloadPollingIntervals = {
	unreachable = 90,
	wifi = 30,
	wwan = 60,
	serviceUnavailable = 60,
	default = 60 -- backup value
}

--------------------------------------------------------------------------------

local function pollingIntervalB( reachableB, hostErrorB, pollingIntervalMap )

	return KSReactive.dataCallB(
		function( reachable, hostError )
		
			if not reachable or WFUnknown.isUnknown( reachable ) then
				reachable = 'default'
			end
			
			if hostError then
				reachable = 'serviceUnavailable'
			end
		
			return pollingIntervalMap[ reachable ]
				or pollingIntervalMap.default
		
		end,
		reachableB, hostErrorB )

end

--------------------------------------------------------------------------------

function WFOzSyncUtils.syncPeerSharedInit( self, serverBaseUrl )

	-- Create behavior to track systemic errors (Oz host says service unavailable,
	-- catalog is 404'd, etc.).
	
	self:createReactiveVariable( '_currentHostErrorB', nil )
	
	self:createReactiveVariable( '_forcedUpgradeLinkB', nil )
	
	-- Reachability for Oz server governs maximum sync cycle time.
	-- We lengthen those intervals if it's the user info.

	-- TODO [scouten 2013-12-10]: Temporary hack. WFReachability doesn't
	-- work in Lightroom's main thread. Just substitute a constantB in that case.
	-- Remove this when Oz sync code moves off main thread.
	
	local hostIsReachableB
	
	if require 'WFCurrentThread'.DEBUG_getThreadName() ~= nil then
		hostIsReachableB = WFReachability.hostIsReachableB( serverBaseUrl )
	else
		hostIsReachableB = KSReactive.constantB( 'wifi' )
	end
	
	self:setOwned {
	
		_hostIsReachableB = hostIsReachableB,
	
		_minSyncDelayIntervalB = pollingIntervalB(
			hostIsReachableB,
			self._currentHostErrorB,
			minDownloadPollingIntervals ),
			
		_maxSyncDelayIntervalB = pollingIntervalB(
			hostIsReachableB,
			self._currentHostErrorB,
			maxDownloadPollingIntervals )
	
	}

	-- If class of service improves, we trigger an immediate sync.
	
	local oldPollingInterval = 0

	self:setOwned { _triggerSyncE = KSReactive.sourceE() }
	
	self._maxSyncDelayIntervalB:ownedOnDataB( self, function( newPollingInterval )

		-- If it's not a number, then we're probably not ready yet.

		if type( newPollingInterval ) ~= 'number' then return end
		
		if newPollingInterval < oldPollingInterval then
			self._triggerSyncE:post( true )
		end
	
	end )
	
	-- Trigger download of changes from Oz every so often following
	-- upload attempts.
	
	self:setOwned { _rawUploadCompletedE = KSReactive.sourceE() }
	
	self._rawUploadCompletedE:calmE( 4 ):ownedOnDataE( self, function()
	
		WFTimer.afterDelayDo( 7, function()
			if not WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				self._triggerSyncE:post( true )
			end
		end )
	
	end )

end

--------------------------------------------------------------------------------

return WFOzSyncUtils
