--[[----------------------------------------------------------------------------

WFOzTestBinaryProviders.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

Binary rendition providers to be used for unit testing.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local KSStringUtils = require 'KSStringUtils'


--============================================================================--

-- @mark WFOzTestBinaryProviders namespace

local WFOzTestBinaryProviders = {}

--------------------------------------------------------------------------------

-- Notice the one %s in this template, which is in the middle of the
-- "software" field in the Exif data for the tiny jpeg. We can insert
-- any base64 valid char we want there and it will still decode into
-- a valid JPEG (but with a new, unique SHA256).

local kTinyJpegBase64Template =
	[[/9j/4QAoRXhpZgAASUkqAAgAAAABAJiCAgADAAAANDIAAAAAAAA0MgAAAAD/7AARRHVja3]] ..
	[[kAAQAEAAAAAAAA/+0ALFAAAAAAA%sAAAAAAAAAAOEJJTQQlAAAAAAAQAAAAAAAAAAAAAAAA]] ..
	[[AAAAAP/uAA5BZG9iZQBkwAAAAAH/2wCEABsaGikdKUEmJkFCLy8vQkc/Pj4/R0dHR0dHR0]] ..
	[[dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0cBHSkpNCY0PygoP0c/NT9H]] ..
	[[R0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR0dHR//AAB]] ..
	[[EIAEAAQAMBIgACEQEDEQH/xABOAAEBAQAAAAAAAAAAAAAAAAAABQMBAQEBAAAAAAAAAAAA]] ..
	[[AAAAAAADBRABAAAAAAAAAAAAAAAAAAAAABEBAAAAAAAAAAAAAAAAAAAAAP/aAAwDAQACEQ]] ..
	[[MRAD8A1ARaYAAAAACGAszFwBFpgAIYCzMAAXAEWmAAhgLMwABcARaYACGAszH/2Q==]]

local colors = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"

--------------------------------------------------------------------------------

local function makeBinaryWithRandomSeed( seedNumber )

	assert( type( seedNumber ) == 'number' and seedNumber >= 0 )

	seedNumber = ( seedNumber % 62 ) + 1
	local base64SeedChar = colors:sub( seedNumber, seedNumber )

	local binarySeed = string.format( kTinyJpegBase64Template, base64SeedChar )
	local binaryString = KSStringUtils.decodeBase64( binarySeed )
	
	return KSData.make( binaryString )

end

--------------------------------------------------------------------------------

local function makeBinaryForDocValueP( docValue, seedNumberBias )

	local captureDate = docValue.captureDate
	local strippedCaptureDate = captureDate:gsub( '[-+ZT:.]', '' )

	local seedNumber = tonumber( strippedCaptureDate ) + seedNumberBias
	
	return KSPromises.resolvedPromiseP( 'data', makeBinaryWithRandomSeed( seedNumber ) )
	
end

--------------------------------------------------------------------------------

WFOzTestBinaryProviders.testBinaryProvider = Rx.Class 'testBinaryProvider' {

	getRenditionP = function( self, docValue, renditionType )

		if renditionType == 'thumbnail2x' then
			return makeBinaryForDocValueP( docValue, 1 )
		elseif renditionType == '2048' then
			return makeBinaryForDocValueP( docValue, 2 )
		end
	
	end,
	
	getProxyP = function( self, docValue, proxyType )
	
		if proxyType == '2560' then
			return makeBinaryForDocValueP( docValue, 3 )
		end
	
	end,
	
	getMasterP = function( self, docValue )

		return makeBinaryForDocValueP( docValue, 4 )
	
	end,

}

--------------------------------------------------------------------------------

WFOzTestBinaryProviders.missingBinaryProvider = Rx.Class 'missingBinaryProvider' {

	getRenditionP = function( self, docValue, renditionType )

		return KSPromises.resolvedPromiseP( 'data', nil )

	end,
	
	getProxyP = function( self, docValue, proxyType )
	
		return KSPromises.resolvedPromiseP( 'data', nil )
	
	end,
	
	getMasterP = function( self, docValue )

		return KSPromises.resolvedPromiseP( 'data', nil )
	
	end,

}

--------------------------------------------------------------------------------

WFOzTestBinaryProviders.badContentBinaryProvider = Rx.Class 'badContentTypeBinaryProvider' {

	getRenditionP = function( self, docValue, renditionType )

		return KSPromises.resolvedPromiseP( 'data', 'bogus', 'text/plain' )

	end,
	
	getProxyP = function( self, docValue, proxyType )
	
		return KSPromises.resolvedPromiseP( 'data', 'bogus', 'text/plain' )
	
	end,
	
	getMasterP = function( self, docValue )

		return KSPromises.resolvedPromiseP( 'data', 'bogus', 'text/plain' )
	
	end,

}

--------------------------------------------------------------------------------

return WFOzTestBinaryProviders
