--[[----------------------------------------------------------------------------

WFOzUser.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2011 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local KSPrefs = require 'KSPrefs'
local WFFileUtils = require 'WFFileUtils'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCapabilities = require 'WFCapabilities'
local WFCloseSupport = require 'WFCloseSupport'
local WFObjectProxy = require 'WFObjectProxy'

	-- WF.oz_client.index_store
local WFIndexStore = require 'WFIndexStore'

	-- WF.web.oauth2_client
local KSOAuth2Session = require 'KSOAuth2Session'

	-- WF.web.utils
local WFURI = require 'WFURI'

--------------------------------------------------------------------------------

local logger = WFLogger 'WFOzUser'
--	:setLevel 'debug'

local check = assert( WFArguments.check )


--------------------------------------------------------------------------------

if ( require 'WFCapabilities'.getCapability 'WFOzUser.disabled' == 'disabled' ) then

	logger:warnf( 'WFOzUser has been disabled, please use WFCatUser\n%s', ':traceback:' )

end


--==============================================================================

-- @mark WFOzUser class

--------------------------------------------------------------------------------
--[[!
	\class WFOzUser
	
	\brief A \c WFOzUser gathers together the very basic information about an
		Oz account so that other classes that do the work of communicating with
		that account can share a common vocabulary for accessing that user's
		information.
	
	Please help keep this class simple and STATELESS!
]]--

local ozUserMethods = {

	_serverBase = false,
		-- (string) base URL for Oz server being used (e.g. "https://api.ozstage.com")

	_ozUserId = false,
		-- (string) Oz user ID (UUID)
		
	_catalogSubtype = false,
		-- (string) specify catalog subtype

	_oAuth2Session = false,
		-- (WFOAuth2Session) OAuth2 session for talking to Oz via IMS
		-- created lazily

	_userInfoCache = false,
		-- (WFDocStore) this doc store retains info about this user.

	_userInfoIndex = false,
		-- (WFIndexStore) index store for user info doc store

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the user.
	
	\param serverBase (string) base URL for Oz server being used
		(e.g. "https://api.ozstage.com")
	\param userId (string) UUID for the user on Oz
]]--

function ozUserMethods:__init( serverBase, ozUserId )

	check( 'WFOzUser', '+string', '+string' )

	self._serverBase = serverBase
	self._catalogSubtype = WFCapabilities.getCapability( "WFOzCatalog.subtype" )
	
	if ( not ozUserId:match( "^[0-9a-f]+$" ) ) or #ozUserId ~= 32 then
		if string.find( ozUserId, "^:memory:" ) then
			-- used for unit testing
		else
			error( "WFOzUser: user UUID malformed: " .. ozUserId, 4 )
		end
	end
	
	self._ozUserId = ozUserId
	
	WFCloseSupport.mustClose( self )

end

--------------------------------------------------------------------------------

function ozUserMethods:__tostring()

	return string.format( "WFOzUser( id = %s, server = %s )",
						   self._ozUserId,
						   self._serverBase )

end

--------------------------------------------------------------------------------

-- @mark basic info about user

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the unique ID of this Oz user.

	\return (string) The ID.
]]--

function ozUserMethods:userId()

	check( 'WFOzUser' )

	return self._ozUserId

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the base URL for the server.
	
	\return (string) URL for the catalog
]]--

function ozUserMethods:serverBaseUrl()

	check( 'WFOzUser' )

	return self._serverBase

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the catalog subtypes the user is interested in.
	
	\return (string) subtype of the catalog
]]--

function ozUserMethods:catalogSubtype()

	check( 'WFOzUser' )
	
	return self._catalogSubtype

end


--------------------------------------------------------------------------------
--[[!
	\brief Get the path to the user info doc store.
	
	\return (string) path
]]--

function ozUserMethods:userInfoCachePath()
	
	local userId = self._ozUserId

	if string.find( userId, "^:memory:" ) then
		return self._ozUserId
	end
	
	local userInfoCachePath = KSPathUtils.child( KSPathUtils.getStandardFilePath 'caches',
											'Adobe' )
	WFFileUtils.mkdir( userInfoCachePath )
	userInfoCachePath = KSPathUtils.child( userInfoCachePath,
										   WFPlatformInfo.getAppIdentifier() )

	WFFileUtils.mkdir( userInfoCachePath )
	userInfoCachePath = KSPathUtils.child( userInfoCachePath,
										   'oz-user-' .. userId )
	return userInfoCachePath
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Construct a doc store and index store to cache information about
		this user.
	
	\return doc store, index store
]]--

function ozUserMethods:userInfoCache()

	check( 'WFOzUser' )

	if not self._userInfoCache or
		WFCloseSupport.hasBeenOrWillBeClosed( self._userInfoCache )
	then

		local userInfoCachePath = self:userInfoCachePath()

		self:setOwned {
			_userInfoCache = require 'WFDocStore'( userInfoCachePath )
		}
		
	end

	if not self._userInfoIndex or
		WFCloseSupport.hasBeenOrWillBeClosed( self._userInfoIndex )
	then

		self:setOwned {
			_userInfoIndex = WFIndexStore( self._userInfoCache,
											'OzUserInfoCacheIndexDefinitions.lua' )
		}
	
	end
	
	return self._userInfoCache, self._userInfoIndex

end

--------------------------------------------------------------------------------

local function oAuth2SessionForSelf( self )

	local result = self._oAuth2Session

	if not result then
		result = KSOAuth2Session( self._ozUserId )
		self:setOwned { _oAuth2Session = result }
	end
	
	return result

end

--------------------------------------------------------------------------------
--[[!
	\brief Simulates logout from the current authorization session.
]]--

function ozUserMethods:simulateForcedLogout()

	return oAuth2SessionForSelf( self ):simulateForcedLogoutP()

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports (as a behavior) whether this user has a current valid
		connection to the Oz server.
	
	\return <code>B< boolean ></code> A behavior whose value is
		\c true while there is a valid connection, and becomes \c false
		if the OAuth2 refresh token for this user is ever rejected.
]]--

function ozUserMethods:hasValidConnectionToOzB()

	return oAuth2SessionForSelf( self ):hasValidLoginB()

end

--------------------------------------------------------------------------------

local function mapPartialUrlToCompleteUrl( self, url )

	if url:match( "^/" ) then
		return WFURI.resolve( self._serverBase, url )
	else
		return url
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP GET request, signing the request with the current
		access token of the current user and handles refreshing the access
		token if needed. Returns the response via a promise.

	\param url (string) The URL to fetch. If it is a server-relative URL
		(e.g. "/v1.0/..."), the URL will be assumed to be relative to the core
		server URL for this user.
	\param optArgs (optional, table) Options to modify the request.
	\param optSequencing (optional, string) 'LIFO' or 'FIFO'. FIFO by default.

	\return <code>P< data, headers ></code> A Promise that resolves to the response
		from HTTP server

	\sa WFHttpClient.getP
]]--

function ozUserMethods:getP( url, optArgs, optSequencing )

	check( 'WFOzUser', '+string', '?table', '?string' )

	return oAuth2SessionForSelf( self ):getP( mapPartialUrlToCompleteUrl( self, url ),
											 optArgs,
											 optSequencing )
end

--------------------------------------------------------------------------------
--[[!
	\brief Starts an HTTP PUT request, signing the request with the current
		access token of the current user and handles refreshing the access
		token if needed. Returns the response via a promise.

	\param url (string) The URL to fetch. If it is a server-relative URL
		(e.g. "/v1.0/..."), the URL will be assumed to be relative to the core
		server URL for this user.
	\param data (\c KSData or table) data to send with the request
		(use a Lua table for form data)
	\param optArgs (optional, table) options to modify the request

	\return <code>P< data, headers ></code> A Promise that resolves to the response
		from HTTP server
	
	\sa HttpClient.putP
]]--

function ozUserMethods:putP( url, data, optArgs )

	check( 'WFOzUser', '+string', '_', '?table' )

	return oAuth2SessionForSelf( self ):putP( mapPartialUrlToCompleteUrl( self, url ),
											 data,
											 optArgs )

end

--------------------------------------------------------------------------------
--[[!
	\brief Logs this user out of the current Oz session.
]]--

function ozUserMethods:logoutP()

	check( 'WFOzUser' )

	-- About the { logout = true } option here: If we lack both access and refresh
	-- tokens, then we should just quietly say we're done rather than firing an error
	-- that we don't have this information. "The item could not be deleted because
	-- it could not be found." And all that.

	logger:infof( "attempting to log out user %s", self._ozUserId )

	-- Delete the cached user information for this user.

	-- TODO [arahn 2014-05-14]: We need a way to do this safely that doesn't impede
	-- our ability to relogin in the same app session, or later.
	-- Things we've tried that don't work:
	--	1) Deleting the user from the doc store.  This leaves behind a tombstone
	--	which causes errors upon relogin because the user is going from deleted->
	--	non deleted, which the doc store does not like.
	--	2) Calling haltAndDelete on the doc and index stores for the user.  This
	--	is problematic because OzUser is cached and shared via WFObjectProxy.  Enough
	--	places in the code "leak" user documents, that upon logout, they aren't all
	--	closed.  Thus upon re-login, the same oz user is recycled, and it's member
	--	user info cache objects are dead.  (A new oz user object would not have this problem
	--	because new doc store and index store would be created for it.)
	--	3) Simply closing and nil-ing out the info caches.
	--	4) Maybe we can try 2 again, but update the object proxy cache mechanism to
	--	not recycle a user whose info cache is dead...  No that doesn't work because
	--	there are other places in the code we create WFOzUser objects. They share the same
	--	docstore and indexstore message processors.  Those MP are effectively dead
	--	if we call haltAndDeleteP -- the only way to "restart" them is if all references
	--	to them are closed, at which point the MPs are shut down.  Then creating a new
	--	user would be okay because upon the messaging processor getting initialized it
	--	would set up new DBs for the docstore/indexstore.  But finding and closing
	--	all WFOzUser objects would be hard, and this would be a fragile solution should new
	--	code be added which creates a WFOzUser and doesn't close it, we'd be in a bad
	--	way again.  What we really need is for the operations that currently happen
	--	in init() for doc store and index store to be made restartable, via a message
	-- 	that can be sent after haltAndDeleteP which puts it back into working form.

	local resultP = oAuth2SessionForSelf( self ):logoutP()
	
	-- Optimistic logout: Clear out current user ID now; hope we can also convince
	-- Oz to log out as well.

	KSPrefs.currentUserId = nil
	
	-- Now return the logout promise.
	
	return resultP

end

--------------------------------------------------------------------------------

local WFOzUser_class = Rx.Class 'WFOzUser'( ozUserMethods )

local validator = function( cache, baseUrl, userId )
	
	return userId
	
end

local WFOzUser = WFObjectProxy( 'WFOzUser', WFOzUser_class, validator, true )

--------------------------------------------------------------------------------

return WFOzUser
