--[[----------------------------------------------------------------------------

WFOzUserModel.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFCloseSupport = require 'WFCloseSupport'
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.catalog
local OzCatalogUrls = require 'OzCatalogUrls'

	-- WF.oz_client.sync_engine.user
local WFOzUser = require 'WFOzUser'
local WFOzUserLogin = require 'WFOzUserLogin'
local WFOzUserSynchronizer = require 'WFOzUserSynchronizer'


--------------------------------------------------------------------------------

local userModelLog = WFLogger 'WFOzUserModel'
--	:setLevel( 'DEBUG' )


--============================================================================--

-- @mark WFOzUserModel

local WFOzUserModel = {}

--------------------------------------------------------------------------------

function WFOzUserModel.makeUserModel( optServerBase, processId )

	local KSPrefs = require 'KSPrefs'

	local lastUser = KSPrefs.lastDaemonUser
	local lastUserId
	local prevProcessId
	
	if lastUser and processId then
		local equalsIndex = string.find( lastUser, '=' )
		if equalsIndex and equalsIndex > 0 then
			prevProcessId = string.sub( lastUser, 1, equalsIndex - 1 )
			lastUserId = string.sub( lastUser, equalsIndex + 1 )
		end
	end
	
	local serverBaseUrlB = KSReactive.variableB( optServerBase or OzCatalogUrls.baseServerUrl )
	
	local userIdB  = KSReactive.variableB( nil )
	
	if prevProcessId == processId then
		userIdB:set( lastUserId )
	end
	
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local previousUser

	local currentUserB = KSReactive.callB(
		function( serverBaseUrl, userId )

			local newUser

			if serverBaseUrl and userId then
				newUser = WFOzUser( serverBaseUrl, userId )
			end

			if previousUser and previousUser ~= newUser then
				previousUser:close()
			end

			return newUser

		end,
		serverBaseUrlB, userIdB )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local userInfoCacheB = currentUserB
		:mapB( function( wfOzUser )

			if wfOzUser then
				return wfOzUser:userInfoCache(), wfOzUser:userId()
			else
				return nil
			end

		end )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local userInfoB = userInfoCacheB
		:mapSwitchB( function( userInfoCache, userId )

			if userInfoCache and userId then
				return userInfoCache:docValueB( userId )
					:mapB( WFUnknown.mapUnknownToNil )
			else
				return KSReactive.constantB( nil )
			end

		end )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local currentUserFieldValueBForPath = WFCloseSupport.memoizeFn(
		function( memberFieldPath )

			return userInfoCacheB
				:mapSwitchB( function( userInfoCache, userId )

					if userInfoCache and userId then
						return userInfoCache:fieldValueB( userId, memberFieldPath )
							:mapB( WFUnknown.mapUnknownToNil )
					else
						return KSReactive.constantB( nil )
					end

				end )

		end )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local currentUserInfoCachePathB = currentUserB
		:mapB( function( wfOzUser )

			if wfOzUser then
				return wfOzUser:userInfoCachePath()
			else
				return nil
			end

		end )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local hasValidConnectionToOzB = currentUserB
		:mapSwitchB( function( currentUser, userId )

			if currentUser then
				return currentUser:hasValidConnectionToOzB()
					:mapB( WFUnknown.mapUnknownToNil )
			else
				return KSReactive.constantB( false )
			end

		end )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local shouldSyncUserInfoB = KSReactive.variableB( false )

	local currentUserSyncAgent

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local userInfoSyncAgentB = KSReactive.callB(

		function( shouldSync, wfOzUser )

			userModelLog:infof( "new user info sync agent for ozUser = %s shouldSync = %s",
								 wfOzUser and wfOzUser:userId() or "", shouldSync )

			if currentUserSyncAgent then
				currentUserSyncAgent:close()
				currentUserSyncAgent = nil
			end

			if shouldSync and wfOzUser then

				currentUserSyncAgent = WFOzUserSynchronizer( wfOzUser:serverBaseUrl(),
															   wfOzUser:userId() )

				return currentUserSyncAgent

			end

		end,
		shouldSyncUserInfoB, currentUserB )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function extractBFromSyncPeerB( syncPeerMethodName )

		return userInfoSyncAgentB
			:mapSwitchB( function( currentUserSyncAgent )

				if currentUserSyncAgent then
					return currentUserSyncAgent[ syncPeerMethodName ]( currentUserSyncAgent )
						:mapB( WFUnknown.mapUnknownToNil )
				else
					return KSReactive.constantB( false )
				end

			end )

	end

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local trialEndDateB = userInfoB
		:mapB( function( userInfo )

			local status = KSTableUtils.derefOpt( userInfo,
												   'entitlement',
												   'status' )

			local trialEnd = KSTableUtils.derefOpt( userInfo,
													 'entitlement',
													 'trial',
													 'end' )

			if status == 'trial' then
				return trialEnd
			else
				return nil
			end

		end )


	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local syncCollectionsByDefaultB = currentUserFieldValueBForPath(
											'config/lrd_sync_collections_by_default' )
		:mapB( function( value )

			if WFUnknown.isUnknown( value ) then
				return true
			else
				return not not value
			end

		end )
		
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local firstCollectionAutoSyncB = currentUserFieldValueBForPath( 
											'config/disable_lrd_auto_sync_collection' )
		:mapB( function( value )
			
			if not value or WFUnknown.isUnknown( value ) then
				return 2 -- Auto create first collection and mark for sync
			else
				return value 
			end
		end )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function mapFalseToNil( value )

		if value == false then
			return nil
		else
			return value
		end

	end

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local loginStatusB = KSReactive.variableB( nil )

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function loginPromiseDidResolve( status, wfOzUser )
	
		if status == 'data' then

			userModelLog:infof( "Successfully logged in to Oz with user ID %s",
								 wfOzUser:userId() )

			userIdB:set( wfOzUser:userId() )
			
			if processId then
				KSPrefs.lastDaemonUser = processId .. "=" .. wfOzUser:userId()
			end
			
			loginStatusB:set( nil )

			if currentUserB:valueNow() ~= wfOzUser then
				debug.print( "Huh ... didn't set currentUserB as expected" )
			end

		else

			userModelLog:errorf( "Failed to log in to Oz: %s", wfOzUser )

			userIdB:set( nil )
			KSPrefs.lastDaemonUser = nil
			loginStatusB:set( 'error' )

		end
	
	end

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function loginFromUsernameAndPassword( model, baseServerUrl, email, password )
	
		loginStatusB:set( 'pending' )
	
		WFOzUserLogin.loginFromEmailAndPasswordP( baseServerUrl, email, password )
			:whenResolvedDo( loginPromiseDidResolve )
			:runUntilDone()
	
	end
	
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function loginFromTokens( model,
									 baseServerUrl,
									 accessToken,
									 refreshToken,
									 optDeviceToken )
	
		loginStatusB:set( 'pending' )
	
		WFOzUserLogin.loginFromTokensP( baseServerUrl, accessToken, refreshToken, optDeviceToken )
			:whenResolvedDo( loginPromiseDidResolve )
			:runUntilDone()
	
	end
	
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function activateOfflineMode( model )
	
		userModelLog:infof( "Offline Mode. Using %s", lastUserId )
		userIdB:set( lastUserId )

	end

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	return {

		value = {

			-- basic user info

			ozUserId = userIdB,
			isLoggedIn = userIdB:toBooleanB(),
			email = currentUserFieldValueBForPath 'email',
			fullName = currentUserFieldValueBForPath 'full_name',
			firstName = currentUserFieldValueBForPath 'first_name',
			lastName = currentUserFieldValueBForPath 'last_name',
			userInfoCachePath = currentUserInfoCachePathB,

			-- user info synchronization status

			shouldSyncUserInfo = shouldSyncUserInfoB,
			isSyncingUserInfo = userInfoSyncAgentB:toBooleanB(),

			-- host system status

			hasValidConnectionToOz = hasValidConnectionToOzB,
			hostIsReachable = extractBFromSyncPeerB 'hostIsReachableB',
			currentHostError = extractBFromSyncPeerB 'currentHostErrorB'
									:mapB( mapFalseToNil ),
			forcedUpgradeLink = extractBFromSyncPeerB 'forcedUpgradeLinkB'
									:mapB( mapFalseToNil ),

			-- account/trial status

			accountStatus = currentUserFieldValueBForPath 'entitlement/status',
			trialEndDate = trialEndDateB,
			assetLimit = currentUserFieldValueBForPath 'entitlement/asset_limit/enforce',
			assetLimitForDisplay = currentUserFieldValueBForPath 'entitlement/asset_limit/display',

			-- user config flags

			syncCollectionsByDefault = syncCollectionsByDefaultB,
			firstCollectionAutoSync = firstCollectionAutoSyncB,

		},

		events = {

			activateOfflineMode = activateOfflineMode,

			loginFromUsernameAndPassword = loginFromUsernameAndPassword,
			
			loginFromTokens = loginFromTokens,

			logout = function()
				local wfOzUser = currentUserB:valueNow()
				if wfOzUser then
					wfOzUser:logoutP()
				end
				userIdB:set( nil )
				loginStatusB:set( nil )
			end,

			startSync = function()
				shouldSyncUserInfoB:set( true )
			end,

			stopSync = function()
				shouldSyncUserInfoB:set( false )
			end,
			
			simulateForcedLogout = function( self )

				local wfOzUser = currentUserB:valueNow()
				if wfOzUser then
					wfOzUser:simulateForcedLogout()
				end
			end,
			
			DEBUG_dumpCoverage = function( self, path )
				if path then
					require 'SetupCoverage'.dumpCoverage( path )
				end
			end,

		},
		
		close = function()

			userModelLog:info( "Closing" )

			userIdB:close()

		end,

	}

end

--------------------------------------------------------------------------------

return WFOzUserModel
