--[[----------------------------------------------------------------------------

WFOzUserSyncPeer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local KSReactive = require 'KSReactive'

	-- WF.oz_client.sync_engine
local WFOzSyncUtils = require 'WFOzSyncUtils'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local WFURI = require 'WFURI'


--------------------------------------------------------------------------------

local logger = WFLogger 'WFOzUserSyncPeer'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local tinsert = table.insert

local kDawnOfTime = "0000-00-00T00:00:00.00000Z"

local syncCatalogUrls = not ( require 'WFCapabilities'.getCapability(
								'WFOzUserSyncPeer.suppressCatalogFetch' ) )


--============================================================================--

-- @mark WFOzUserSyncPeer class

--------------------------------------------------------------------------------
--[[!
	\class WFOzUserSyncPeer

	This class implements a sync peer that can be used for syncing information
	about an Oz user and a local WFDocStore dedicated to that user's information.
]]--

local syncPeerMethods = {
	
	_wfOzUser = false,
		-- (OzUser)
	
	_otherUserUrls = false,
	_newUserIds = false,
	
	_currentHostErrorB = false,
	_hostIsReachableB = false,
	_forcedUpgradeLinkB = false,
	_minSyncDelayIntervalB = false,
	_maxSyncDelayIntervalB = false,
	_triggerSyncE = false,
	_rawUploadCompletedE = false,
		-- (see WFOzSyncUtils.syncPeerSharedInit)

	_calmedChangeCounterB = false,

	WFCloseSupport.addCloseSupportMethods,
	KSReactive.addReactiveSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the WF <-> Oz sync peer for the given user.
	
	\param wfOzUser (WFOzUser) the user to sync
]]--

function syncPeerMethods:__init( wfOzUser )

	check( 'WFOzUserSyncPeer', 'WFOzUser' )
	
	self:setOwned {
		_wfOzUser = wfOzUser:relayObject(),
	}
	self._otherUserUrls = {}
	self._newUserIds = {}

    WFOzSyncUtils.syncPeerSharedInit( self, wfOzUser:serverBaseUrl() )

end

------------------------------------------------------------------------------

function syncPeerMethods:docStoreChangeCounter( docStore )

	check( 'WFOzUserSyncPeer', 'WFDocStore' )

	if not self._calmedChangeCounterB then
		local calmedChangeCounterB = docStore:docChangeCounterB():closingCalmB( 15 )
		self:setOwned { _calmedChangeCounterB = calmedChangeCounterB }
	end
	
	return self._calmedChangeCounterB:relayB()
	
end

--------------------------------------------------------------------------------

function syncPeerMethods:minSyncDelayIntervalB()

	return self._minSyncDelayIntervalB
		:mapB( function( interval )
			if interval < 15 then
				interval = 15
			end
			return interval
		end )

end

--------------------------------------------------------------------------------

function syncPeerMethods:maxSyncDelayIntervalB()

	return self._maxSyncDelayIntervalB
		:mapB( function( interval )
			if interval < 15 then
				interval = 15
			end
			return interval
		end )

end

--------------------------------------------------------------------------------

function syncPeerMethods:triggerSyncE()

	return self._triggerSyncE:relayE()

end

--------------------------------------------------------------------------------

function syncPeerMethods:hostIsReachableB()

	return self._hostIsReachableB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:currentHostErrorB()

	return self._currentHostErrorB:relayB()

end

--------------------------------------------------------------------------------

function syncPeerMethods:forcedUpgradeLinkB()

	return self._forcedUpgradeLinkB:relayB()

end

--------------------------------------------------------------------------------

-- @mark tracking other user

--------------------------------------------------------------------------------

function syncPeerMethods:watchUser( userId )

	local accountUrl = string.format( '/v1.0/accounts/%s', userId )

	for _, url in ipairs( self._otherUserUrls ) do

		if url == accountUrl then
			-- already present
			return
		end

	end

	-- Ensure we get one update before doing update date filtering.

	self._newUserIds[ userId ] = true
	table.insert( self._otherUserUrls, accountUrl )
	
end

--------------------------------------------------------------------------------

local function stopWatchingUserUrl( self, accountUrl )

	for i, url in ipairs( self._otherUserUrls ) do
		if url == accountUrl then
			table.remove( self._otherUserUrls, i )
			
			local userId = string.match( accountUrl, ".+/(.*)$" )
			self._newUserIds[ userId ] = nil

			return true
		end
	end

	return false
end

--------------------------------------------------------------------------------

-- @mark :changedDocsP

--------------------------------------------------------------------------------

local function addDeletedCatalogs( changesFeed, localIds, curIds )
	
	for docId in pairs( curIds ) do
		localIds[ docId ] = nil
	end
	for docId in pairs( localIds ) do
		table.insert( changesFeed, {
			_id = docId,
			_deleted = true,
			_timestamp = ':new:'
		} )
	end

	return changesFeed
	
end

--------------------------------------------------------------------------------

local function addUpdatedCatalogs( changesFeed, since, malformedLibraryIds, data )

	local origCatalogs = data and data.catalogs or {}
	local nCatalogs = #origCatalogs
	
	-- TO DO: Save off catalog list in user info cache?
	
	for i = 1, nCatalogs do

		local origCatalog = origCatalogs[ i ]

		-- Check if this catalog's doc store has data, if not, write the information
		-- received now to the doc store to prevent a delay in fetching this information
		-- when it is accessed.

		local updated = origCatalog.updated

		if updated > since or malformedLibraryIds[ origCatalog._id ] then
			origCatalog.type = 'catalog'
			origCatalog.base = data.base
			origCatalog._timestamp = updated
			tinsert( changesFeed, origCatalog )
		end
	end

end

--------------------------------------------------------------------------------

-- OPTIMIZE_ME: Actually, we would rather do this computation only when the
-- debug log is active. And if not that, then we might well want a caching version
-- of this function.

local function extendKeyPath( keyPath, k )
	if keyPath == "" then
		return k
	else
		return keyPath .. '.' .. k
	end
end

--------------------------------------------------------------------------------

local function compareDocs( a, b, keyPath )

	keyPath = keyPath or ""

	for k in pairs( b ) do
		if a[ k ] == nil then
			logger:debugf( "Doc 1 missing value at %s.%s", keyPath, k )
			return false
		end
	end

	for k, aa in pairs( a ) do
		local bb = b[ k ]
		if aa == bb then
			-- Do nothing. Happy case.
		elseif bb == nil then
			logger:debugf( "Doc 2 missing value at %s.%s", keyPath, k )
			return false
		elseif type( aa ) ~= type( bb ) then
			logger:debugf( "Different types of value at %s.%s", keyPath, k )
			return false
		elseif type( aa ) == 'table' then
			-- Both tables. We're happy unless the tables don't match.
			if not compareDocs( aa, bb, extendKeyPath( keyPath, k ) ) then
				return false
			end
		else
			logger:debugf( "Different values of same type at %s.%s", keyPath, k )
			return false
		end
	end

	return true

end

--------------------------------------------------------------------------------

local addUpdatedUser = function( changesFeed, since, userInfo, curUserDocs )

	if not userInfo.updated then
		local theSame = false
		local oldDoc = curUserDocs and curUserDocs[ userInfo._id ]
		if oldDoc then
			for k,v in pairs( oldDoc ) do
				if k:sub( 1, 1 ) == '_' and k ~= '_id' then
					oldDoc[ k ] = nil
				end
			end
			oldDoc.updated, oldDoc.type = nil, nil
			
			theSame = compareDocs( oldDoc, userInfo )
		end
		if theSame then
			return
		end

		-- HACK : Oz doesn't provide an updated field for user docs
		-- that aren't the current user
		-- So spoof an updated date that is 1 millisecond newer.

		local _, _, year, month, day, hour, minute, seconds = string.find( since,
							"(%d%d%d%d)%-(%d%d)%-(%d%d)T(%d%d):(%d%d):(%d%d%.%d+)" )

		if seconds then

			year = tonumber( year )
			month = tonumber( month )
			day = tonumber( day )
			hour = tonumber( hour )
			minute = tonumber( minute )
			seconds = tonumber( seconds )
			seconds = seconds + 0.000001

			if seconds >= 60 then
				minute = minute + 1
				seconds = seconds - 60
			end

			if minute >= 60 then
				hour = hour + 1
				minute = minute - 60
			end

			if hour >= 24 then
				day = day + 1
				hour = hour - 1
			end

			-- If we've been unlucky enough that by adding 1 ms we are now on a day of
			-- the month that is > number of days in this month, then oh well.

			userInfo.updated = string.format( "%04d-%02d-%02dT%02d:%02d:%09.6fZ",
												year, month, day, hour, minute, seconds )

		else

			userInfo.updated = kDawnOfTime

		end

	end
	
	if userInfo.updated > since then
		
		logger:debugf( "getUserInfoDocumentP received new info %s", userInfo )
		
		userInfo._timestamp = userInfo.updated
		userInfo.type = 'user'

		tinsert( changesFeed, userInfo )

	end

end

--------------------------------------------------------------------------------

local function addIdsFromCatalogJson( outIdSet, doc )

	local catalogs = doc and doc.catalogs or {}

	for _, catalog in ipairs( catalogs ) do

		local id = catalog._id
		if id then
			outIdSet[ id ] = true
		end

	end

end

--------------------------------------------------------------------------------

local userInfoGetOptions = {
	requestHeaders = {
		{ header = 'Accept', value = 'application/json' },
	},
}

--------------------------------------------------------------------------------

local hack_kickCollaborators = 0

local function getUserInfoDocumentP( self, combinedSince )
	
	-- We wish we had one feed for user and catalog documents
	-- But there is none provided by oz.
	-- Futhermore, the "updated" field on user and catalog documents do not
	-- appear to track synchronously related to one another -- e.g. if we update
	-- our continueFrom timestamp to the maximum of all updated fields on
	-- user and catalog documents, we can miss updates to the user because
	-- the user *will* change, but the updated user document will receive
	-- and "updated" timestamp OLDER than the newest catalog "updated" timestamp
	
	local ozUserId = self._wfOzUser:userId()
	local accountUrl = string.format( '/v1.0/accounts/%s', ozUserId )
	local catalogInfoUrl = '/v1.0/catalogs'
	
	local catalogSubtype = self._wfOzUser:catalogSubtype()
	local encodedSubtype
	
	if catalogSubtype then
		encodedSubtype = 'subtype=' .. WFURI.escape( catalogSubtype, 'http_query_key' )
		catalogInfoUrl = string.format( '%s?%s', catalogInfoUrl, encodedSubtype )
	end

	local _, _, userSince, catalogSince = string.find( combinedSince or "", "(.+)\t(.+)" )
	userSince = userSince or kDawnOfTime
	catalogSince = catalogSince or kDawnOfTime
	
	local relevantUrls
	
	if syncCatalogUrls then
		relevantUrls = { accountUrl, catalogInfoUrl }
	else
		relevantUrls = { accountUrl }
	end

	local userAndLibraryGetP = KSPromises.mapArrayP( function( url )

		if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
			return 'error', 'closed'
		end
		
		return self._wfOzUser:getP( url, userInfoGetOptions )
			:mapP( function ( status, data, headers )

				local accessForbidden = status == 'error'
						and headers
						and headers.status_code == '403'
						and headers[ 'X-Result-Code' ] == '4300'
						
				if accessForbidden and stopWatchingUserUrl( self, url ) then
					return 'data', "{}", headers -- map this error to an empty table to keep processing
				end
				
				return status, data, headers
			end )

			:observeP( function( ... )
				
				if not WFCloseSupport.hasBeenOrWillBeClosed( self ) then
					WFOzSyncUtils.updateStatusFromHttpResultFn( self, ... )
				end
	
			end )
	
			:mapP( KSJsonDecoder.decodeWebSafeJson_stripNullP )

	end, relevantUrls )
	
	local curIdSetP = userAndLibraryGetP
		:mapDataP( function( _, userInfoDocs )
			local curIdSet = {}
			addIdsFromCatalogJson( curIdSet, userInfoDocs[ 2 ] )
			addIdsFromCatalogJson( curIdSet, userInfoDocs[ 3 ] )
			return 'data', curIdSet
		end )

	curIdSetP:runUntilDone():suppressCloseLogging()

	-- Fetch the current list of catalogs.

	local docStore = self._wfOzUser:userInfoCache()
	local curLibraryIds = {}
	local curUserDocs = {}
	local malformedLibraryIds = {}
	
	local function nextBatchP( continueFrom )

		if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
			return 'error', 'closed'
		end
		
		return docStore:changedDocsP {
			slot = 'main',
			fromChangeCounter = continueFrom,
		}

		:mapDataP( function( _, changedDocs )

			for _, doc in ipairs( changedDocs ) do

				if doc.type == 'catalog' then
					curLibraryIds[ doc._id ] = true
				end
				if doc.type == 'user' then
					curUserDocs[ doc._id ] = doc
				end

				if doc.links == nil and doc.link_templates == nil then
					-- This must be a prior version of the doc,
					-- from a time when we only stored a few fields from
					-- it instead of the whole thing.
					-- We never shipped a version with this, but enough
					-- testing was done that QE machines are probably contaminated.
					malformedLibraryIds[ doc._id ] = true
				end

			end
			if changedDocs[ 1 ] then
				return nextBatchP( changedDocs.continueFrom )
			else
				return 'data', curLibraryIds, malformedLibraryIds, curUserDocs
			end

		end )

	end
	
	-- This is grotesque hackery -- every sync cycle we are polling the entire contents
	-- of the user info cache docStore just so we can find out which documents have been
	-- deleted, and so that we can compare the shared catalog info to what Oz is now
	-- providing.
	local currentDocIdsP = nextBatchP( 0 )

	local createFeedP = KSPromises.callP( function( userInfoDocs, curLibraryIds, malformedLibraryIds, curUserDocs )

			local changesFeed = {}
		
			addUpdatedUser( changesFeed, userSince, userInfoDocs[ 1 ] )
			
			-- This is a hack -- injecting the catalog and shared catalog info into
			-- the user, but Oz doesn't provide a single document for this stuff.
			addUpdatedCatalogs( changesFeed, catalogSince, malformedLibraryIds, userInfoDocs[ 2 ] )
			addUpdatedCatalogs( changesFeed, catalogSince, malformedLibraryIds, userInfoDocs[ 3 ] )
			
			for i = 4, #userInfoDocs do

				local doc = userInfoDocs[ i ]
				if doc._id then
					local thisUserSince = self._newUserIds[ doc._id ] and "" or userSince
					self._newUserIds[ doc._id ] = nil
					addUpdatedUser( changesFeed, thisUserSince, userInfoDocs[ i ], curUserDocs )
				end
			end
			
			for _, doc in ipairs( changesFeed ) do

				local timestamp = doc._timestamp
				local continueFrom

				if doc.type == 'user' then
					continueFrom = userSince
				else
					continueFrom = catalogSince
				end

				if timestamp and ( continueFrom == nil or continueFrom < timestamp ) then
					if doc.type == 'user' then
						userSince = timestamp
					else
						catalogSince = timestamp
					end
				end

			end
			
			changesFeed.continueFrom = userSince .. "\t" .. catalogSince
			
			if #changesFeed == 0 then
				hack_kickCollaborators = hack_kickCollaborators - 1
			else
				hack_kickCollaborators = 2
			end
			
			return changesFeed
	
		end, userAndLibraryGetP, currentDocIdsP )
		
		:runUntilDoneP()
	
	return KSPromises.callP( addDeletedCatalogs, createFeedP, currentDocIdsP, curIdSetP )

end

--------------------------------------------------------------------------------

function syncPeerMethods:changedDocsP( since )

	check( 'WFOzUserSyncPeer', '?number|string' )

	since = tostring( since or '' )
		-- Earlier versions of this code simply returned the number 0.
		-- Make sure the comparison below doesn't fail in that case.

	return getUserInfoDocumentP( self, since )

end

--------------------------------------------------------------------------------

-- @mark :sendUpdatesP

--------------------------------------------------------------------------------

local canSafelyIgnore = {
	_fullDocValue = true,
	_id = true,
	_previouslyAgreedValue = true,
	_timestamp = true,
	base = true,
	config = true,
	country = true,
	email = true,
	entitlement = true,
	first_name = true,
	full_name = true,
	id = true,
	last_name = true,
	limits = true,
	links = true,
	resources = true,
	role = true,
	status = true,
	trial = true,
	type = true,
	updated = true,
	usage = true,
	settings = true,
}

--------------------------------------------------------------------------------

local function uploadUserP( self, delta )

	for k in pairs( delta ) do

		if not canSafelyIgnore[ k ] then

			logger:errorf( "uploadUserP not implemented: key %s\n\ndelta was %s",
						   k, delta )

--			return KSPromises.resolvedPromiseP( 'error',
--												"User info updates not yet implemented" )

		end

	end

	return KSPromises.resolvedPromiseP( 'data', true )

end

--------------------------------------------------------------------------------

local uploadFunctionsByDocType = {

	catalog = 'nop',
	user = uploadUserP,

}

--------------------------------------------------------------------------------

function syncPeerMethods:sendUpdatesP( updates )
	
	check( 'WFOzUserSyncPeer', 'table' )
	
	return WFOzSyncUtils.sendUpdatesP_imp( self, updates, uploadFunctionsByDocType )
	
end

--------------------------------------------------------------------------------

function syncPeerMethods:maxUpdatesPerSend()

	return 4  -- depends on remote peer

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function syncPeerMethods:sendDiagnosticLogData()

	local wfOzUser = self._wfOzUser

	local ozUser = require 'OzUser'.make( wfOzUser:userId() )

	return {

		name = "WF Oz User Info Sync Agent",
		id = 'WFOzUserSyncAgent',
		weight = 510,

		body = {
		
			ozApiServer = wfOzUser:serverBaseUrl(),

			ozUserId = wfOzUser:userId(),
            ozUserInfo = ozUser:userInfoP():runUntilDoneP(),

			ozHostReachable = self._hostIsReachableB,

			ozHostErrorStatus = self._currentHostErrorB
				:mapB( function( value ) return tostring( value ) end ),

		},
	
	}
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFOzUserSyncPeer'( syncPeerMethods )
