--[[----------------------------------------------------------------------------

WFOzUserSyncPeerTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local WFUnknown = require 'WFUnknown'

	-- WF.oz_client.sync_engine.oz_utils
local WFOzCatalogTestSession = require 'WFOzCatalogTestSession'


--============================================================================--

local WFOzUserSyncPeerTest = {}

--------------------------------------------------------------------------------

local catalogTestSession
local TEST_WAIT_TIME = 240

--------------------------------------------------------------------------------

function WFOzUserSyncPeerTest.setUp()
	
	catalogTestSession = WFOzCatalogTestSession( 'WFOzUserSyncPeerTest' )

end

--------------------------------------------------------------------------------

function WFOzUserSyncPeerTest.tearDown()

	if catalogTestSession then
		catalogTestSession:releaseTestUserP()
		catalogTestSession:close()
		catalogTestSession = nil
	end

end

--------------------------------------------------------------------------------

local function mapAwayUuidsAndTimestamps( t )

	if type( t._changeCounter ) == 'number' and t._changeCounter > 0 then
		t._changeCounter = "(change counter)"
	end
	
	if type( t.continueFrom ) == 'number' and t.continueFrom > 0 then
		t.continueFrom = "(change counter)"
	end

	for k, v in pairs( t ) do
	
		if type( v ) == 'string' then
		
			-- Hash codes and UUIDs.
			
			if #v == 32 or #v == 40 or #v == 64 then
				v = v:gsub( '^[0-9a-f]+$', string.format( "hash of %d chars", #v ) )
			end
			
			-- URN-style IDs.
			
			v = v:gsub( "account:[0-9a-f]+", "account:(account ID)" )
			v = v:gsub( "catalog:[0-9a-f]+", "catalog:(catalog ID)" )
			v = v:gsub( "album:[0-9a-f]+", "asset:(album ID)" )
			v = v:gsub( "asset:[0-9a-f]+", "asset:(asset ID)" )
			
			-- IDs in URLs.
			
			v = v:gsub( "catalogs/[0-9a-f]+", "catalogs/(catalog ID)" )
			v = v:gsub( "albums/[0-9a-f]+", "albums/(album ID)" )
			v = v:gsub( "assets/[0-9a-f]+", "assets/(asset ID)" )
			v = v:gsub( "revisions/[0-9a-f]+", "revisions/(revision ID)" )
			
			-- ISO-8601 timestamps.
			
			v = v:gsub( "^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d+$", "(client timestamp)" )
			v = v:gsub( "^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d+Z$", "(Oz timestamp)" )
			
			-- Updates link.
			
			v = v:gsub( "updated_since=.+Z$", "updated_since=(Oz update timestamp)" )
			
			t[ k ] = v
			
		end
		
		if type( v ) == 'table' then
			mapAwayUuidsAndTimestamps( v )
		end
	
	end
	
	return t

end

--------------------------------------------------------------------------------

local userInfoWhitelist = {

	_changeCounter = true,
	_hash = true,
	_id = true,
	_slot = true,

	base = true,
	country = true,
	created = true,
	email = "^.+%d+%@gmail%.com$",
	first_name = "^%w+",
	full_name = "^%w+.* %w+.*",
	id = true,
	last_name = "^%w+",
	status = true,
	trial = true,
	updated = true,
	type = true,
	wcd_guid = "^[0-9A-F]+$",

}

--------------------------------------------------------------------------------

local function mapUsingWhitelist( t )

	local result = {}
	
	if type( t ) ~= 'table' or WFUnknown.isUnknown( t ) then
		return t
	end
	
	for k, pattern in pairs( userInfoWhitelist ) do

		local v = t[ k ]
		
		if v then
		
			if type( pattern ) == 'string' and type( v ) == 'string' then
				if v:match( pattern ) then
					result[ k ] = pattern
				else
					result[ k ] = t[ k ]
				end
			else
				result[ k ] = t[ k ]
			end
		
		end

	end

	return result, t

end

--------------------------------------------------------------------------------

function WFOzUserSyncPeerTest.testGetUserInfo_testNightly( test )

	local syncAgent, ozUserId
	catalogTestSession:setTestName( test:testName() )
	
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalogTestSession:ozUserIdP()
				:mapDataP( function( _, _ozUserId )
				
					ozUserId = _ozUserId
					return 'data', true
				
				end )
			
		end )

		:waitFor( function()
		
			return catalogTestSession:ozUserSyncAgentP()
				:mapDataP( function( _, _syncAgent )

					syncAgent = _syncAgent
					syncAgent:startAutomaticSync()
					
					return 'data', true
				
				end )
		
		end )
		
		:waitFor( function()
		
			-- FIXME [scouten 2014-03-18]: This seems unnecessarily complicated.
			-- Can we simplify? (Maybe we need to use multiple steps here.)

			return catalogTestSession:wfOzUserP()
				:mapDataP( function( status, wfOzUser )
					local userInfoCache = wfOzUser:userInfoCache()
					return 'data', userInfoCache
				end)
				:untilResolvedB( 'data', nil )
				:mapSwitchB( function( userInfoCache )
					if userInfoCache then
						return userInfoCache:docValueB( ozUserId )
					end
					local KSReactive = require 'KSReactive'
					return KSReactive.constantB( nil )
				end )
				:mapB( mapAwayUuidsAndTimestamps )
				:mapB( mapUsingWhitelist )
			
		end )
		:continueWhenValueMatches(
			{
				_changeCounter = "(change counter)",
				_hash = "hash of 40 chars",
				_id = "hash of 32 chars",
				_slot = "main",
				base = "https://api.ozstage.com",
				country = "US",
				created = "(Oz timestamp)",
				email = "^.+%d+%@gmail%.com$",
				first_name = "^%w+",
				full_name = "^%w+.* %w+.*",
				id = "urn:carousel:account:(account ID)",
				last_name = "^%w+",
				status = "created",
				trial = {
					eligible = "yes",
				},
				type = "user",
				updated = "(Oz timestamp)",
				wcd_guid = "^[0-9A-F]+$",
			},
			':ignore:' )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end
	
--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function WFOzUserSyncPeerTest.testSendDiagnosticLogData_testNightly( test )

	-- Since diagnostic log is expected to evolve frequently, we're not
	-- verifying specific content. The main concern is that it doesn't fall
	-- over and die when generating the log.

	catalogTestSession:setTestName( test:testName() )
	WFAsyncTestCase( test, TEST_WAIT_TIME )

		:waitFor( function()
		
			return catalogTestSession:ozUserSyncAgentP()
				:mapDataP( function( _, syncAgent )
					return WFDiagnosticLog.processLogSectionP(
						syncAgent:sendDiagnosticLogData() )
				end )
				:mapDataP( function( status, result )
					if result and result.name then
						return status, true
					end
				end )
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()
		
			return catalogTestSession:releaseTestUserP()
		
		end )

end

--------------------------------------------------------------------------------

-- FIXME [ppaurav > jandrew 2013-07-16]
-- Watson Bug #3578247
if WFPlatformInfo.isWindows then
	return {}
else
	return WFOzUserSyncPeerTest
end
