--[[----------------------------------------------------------------------------

WFOzUserSynchronizer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFArguments_uuid = require 'WFArguments_uuid'
local WFCloseSupport = require 'WFCloseSupport'
local WFObjectProxy = require 'WFObjectProxy'

	-- WF.oz_client.sync_engine.metadata_sync
local WFConflictResolver = require 'WFConflictResolver'
local WFSyncAgent = require 'WFSyncAgent'

	-- WF.oz_client.sync_engine.user
local WFOzUser = require 'WFOzUser'
local WFOzUserSyncPeer = require 'WFOzUserSyncPeer'


--------------------------------------------------------------------------------

local check = WFArguments_uuid and assert( WFArguments.check )
local WFCloseSupport_fetchCached = assert( WFCloseSupport.fetchCached )

--------------------------------------------------------------------------------

local function forceAcceptOzChanges( docValue )

	local mainOverlay
	local syncStatusOverlay = {}

	if docValue._syncStatus then
	
		for peerName, peerStatus in pairs( docValue._syncStatus ) do
			
			local conflicts = peerStatus.conflicts
			mainOverlay = mainOverlay or {}
			for k,v in pairs( conflicts ) do
				mainOverlay[ k ] = v
			end
			
		end
		
	end
	
	if next( syncStatusOverlay ) then
		mainOverlay = mainOverlay or {}
		mainOverlay._syncStatus = syncStatusOverlay
	end
	
	if mainOverlay and next( mainOverlay ) then
		return mainOverlay
	end
	
end

--==============================================================================

local WFOzUserSynchronizerMethods = {

	_userSyncPeer = false,
		-- the WFOzUserSyncPeer; owned by the sync agent
	
	_conflictResolver = false,
		-- the owned WFConflictResolver
	
	_syncAgent = false,
		-- the owned WFSyncAgent
		
	_serverBase = false,
		-- the oz endpoint, a string
		
	_wfOzUser = false,
		-- the owned WFOzUser
	
	WFCloseSupport.addCloseSupportMethods,
}

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:__init( serverBase, ozUserId )
	
	check( 'WFOzUserSynchronizerImp', '+string', 'uuid' )
	
	self._serverBase = serverBase
	
	local wfOzUser = WFOzUser( serverBase, ozUserId )
	local userSyncPeer = WFOzUserSyncPeer( wfOzUser )
	local docStore, indexStore = wfOzUser:userInfoCache()
	
	local syncAgent = WFSyncAgent( docStore, userSyncPeer, 'ozUser' )
	local conflictResolver = WFConflictResolver.makeResolver( docStore, indexStore, { forceAcceptOzChanges } )

	syncAgent:startAutomaticSync()
	
	self._userSyncPeer = userSyncPeer
		-- sync peer is owned by the sync agent.
	
	self:setOwned {
		_conflictResolver = conflictResolver,
		_syncAgent = syncAgent,
		_wfOzUser = wfOzUser,
	}

end

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:wfSyncAgent( )
	return self._syncAgent
end

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:hostIsReachableB( ... )
	return self._userSyncPeer:hostIsReachableB( ... )
end

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:currentHostErrorB( ... )
	return self._userSyncPeer:currentHostErrorB( ... )
end

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:forcedUpgradeLinkB( ... )
	return self._userSyncPeer:forcedUpgradeLinkB( ... )
end

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:syncUntilDoneP()
	return self._syncAgent:syncUntilDoneP()
	:runUntilDoneP()
end

--------------------------------------------------------------------------------

function WFOzUserSynchronizerMethods:watchUser( otherUserId )
	self._userSyncPeer:watchUser( otherUserId )
end

--------------------------------------------------------------------------------

local WFOzUserSynchronizerImp = Rx.Class( 'WFOzUserSynchronizerImp', WFOzUserSynchronizerMethods )

--------------------------------------------------------------------------------

local function validateCached( cache, userId, serverBase )

	local cachedUserSyncAgent = WFCloseSupport_fetchCached( cache, userId )
	
	local oldServerBase = cachedUserSyncAgent and cachedUserSyncAgent._serverBase
	
	if cachedUserSyncAgent
			and ( oldServerBase ~= serverBase ) then
			
			-- Don't reuse a user that doesn't match all arguments
		cache[ userId ] = nil
		cachedUserSyncAgent = nil
	end
	
	return userId, cachedUserSyncAgent
	
end

--------------------------------------------------------------------------------

local WFOzUserSynchronizer = WFObjectProxy( 'WFOzUserSynchronizer', WFOzUserSynchronizerImp, validateCached, true )

--==============================================================================

return WFOzUserSynchronizer
