--[[----------------------------------------------------------------------------

WFRxDocStoreImp.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local BehaviorSubject = Rx.BehaviorSubject
local Observable = Rx.Observable
local ReplaySubject = Rx.ReplaySubject

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.rxlua
local WFRxWeakObservableTable = require 'WFRxWeakObservableTable'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

	-- WF.core.utils
local KSDate = require 'KSDate'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFTableMerge = require 'WFTableMerge'
local WFPerformanceUtils = require 'WFPerformanceUtils'
local WFPerformanceReport = require 'WFPerformanceReport'
local WFConfiguration = require 'WFConfiguration'

	-- WF.sqlite.database
local WFSQLiteDatabase = require 'WFSQLiteDatabase'
local WFSQLiteDatabaseSchema = require 'WFSQLiteDatabaseSchema'

--------------------------------------------------------------------------------

local docStoreImpLog = WFLogger 'WFRxDocStoreImp'

local changedDocsLog = WFLogger 'WFRxDocStoreImp :: changedDocs'

local compactionLog = WFLogger 'WFRxDocStore :: compact'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local string_byte = assert( string.byte )
local string_format = string.format

local math_min = assert( math.min )
local math_floor = assert( math.floor )

local tableToString = assert( WFTableArchiving.tableToString )

local WFPerformanceUtils_wrapInLogLongExecutionInDiagnosticLogWithDynamicDescription =
		assert( WFPerformanceUtils.wrapInLogLongExecutionInDiagnosticLogWithDynamicDescription,
				 "No WFPerformanceUtils.wrapInLogLongExecutionInDiagnosticLogWithDynamicDescription" )
				 
--------------------------------------------------------------------------------

-- SPAM [ashj 2017-08-02]: This string is being referred elsewhere
-- to handle the specific error. If you are changing here, please
-- do handle the usages as well.
local ERROR_REVISION_WITHOUT_PARENT = "revision_without_parent"

--------------------------------------------------------------------------------

local function errorf( ... )
	error( string_format( ... ), 2 )
end

--------------------------------------------------------------------------------

local makeOngoingDocValueForDocId -- forward declaration

local makeOngoingLocalDocValueForDocName -- forward declaration

--============================================================================--

-- @mark database schema

local pageSize = WFConfiguration.DocStore.SQLITE_PAGE_SIZE or 8192

local schema = WFSQLiteDatabaseSchema( '1.0' )
	:useWAL()
	:useForeignKeys()
	:setSynchronous( 1 )
	:setPageSize( pageSize )

--------------------------------------------------------------------------------
--[[
	Map public document IDs to private (integer) document IDs for
	faster reference through the rest of the schema.
]]--

schema:addTable
	[[CREATE TABLE docs( localDocId INTEGER PRIMARY KEY,
						 fullDocId TEXT UNIQUE NOT NULL,
						 winningRevSequence INTEGER REFERENCES revs( sequence ) ON DELETE SET NULL,
						 mostRecentRevSequence INTEGER REFERENCES revs( sequence ) ON DELETE SET NULL,
						 deleted BOOLEAN DEFAULT 0,
						 hasConflicts BOOLEAN DEFAULT 0,
						 annotation BLOB )]]

schema:addIndex
	[[CREATE INDEX docs_mostRecentRevSequence ON docs( mostRecentRevSequence )]]

schema:addIndex
	[[CREATE INDEX docs_fullDocId ON docs( fullDocId )]]

-------------------------------------------------------------------------------
-- Store actual revision content.

schema:addTable
	[[CREATE TABLE revs( sequence INTEGER PRIMARY KEY AUTOINCREMENT,
						 localDocId INTEGER NOT NULL REFERENCES docs( localDocId ) ON DELETE CASCADE,
						 revId TEXT NOT NULL COLLATE BINARY,
						 parentRevSequence INTEGER REFERENCES revs( sequence ) ON DELETE SET NULL,
						 current BOOLEAN,
						 deleted BOOLEAN DEFAULT 0,
						 content BLOB )]]
	:onConflict 'replaceAll'

	-- TODO [scouten 2015-08-10]: Remove this provision in a month or so. See also
	-- corresponding test in WFRxDocStoreImpTest.

	-- TODO [scouten 2015-05-13]: Consider moving content out to its own field?

schema:addIndex
	[[CREATE INDEX revs_localDocIdAndCurrent on revs(localDocId, current)]]

schema:addIndex
	[[CREATE INDEX revs_localDocIdAndRevIdAndSequence ON revs( localDocId, revId, sequence )]]

schema:addIndex
	[[CREATE INDEX revs_sequences ON revs( sequence )]]

--------------------------------------------------------------------------------
-- Store labels.

schema:addTable
	[[CREATE TABLE labels( localDocId INTEGER REFERENCES docs( localDocID ) ON DELETE CASCADE,
						   revSequence INTEGER REFERENCES revs( sequence ) ON DELETE CASCADE,
						   name STRING,
						   UNIQUE( localDocId, name ) )]]

schema:addIndex
	[[CREATE INDEX labels_revSequence on labels(revSequence)]]

--------------------------------------------------------------------------------
-- Store local documents that do not sync.

schema:addTable
	[[CREATE TABLE localDocs( name TEXT UNIQUE NOT NULL,
							  content BLOB )]]

--============================================================================--

-- @mark WFRxDocStoreImp class

--------------------------------------------------------------------------------
--[[!
	\class WFRxDocStoreImp

	TO DO: Write new documentation for this.
]]--

local WFRxDocStoreImp = {

	_db = false,
		-- (WFSQLiteDatabase) database file

	_exec = false,
		-- (function) cached reference to db:exec

	_gatherArrayOfArray = false,
		-- (function) cached reference to db:gatherArrayOfArray

	_path = false,
		-- (string) path to doc store DB

	_changeCounter = false,
		-- (BehaviorSubject< number >) most recent rev change counter value

	_deletedDocStream = false,
		-- (ReplaySubject< table>) A stream that fires with deleted doc values

	_ongoingDocValues = false,
		-- (WFRxWeakObservableTable) document value subjects that are currently active
			-- key: (string) doc ID
			-- value: (Rx.ReplaySubject) current document value

	_ongoingLocalDocValues = false,
		-- (WFRxWeakObservableTable) local document value subjects that are currently active
			-- key: (string) doc name
			-- value: (Rx.ReplaySubject) current document value
			
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local function openAndObtainDatabase( self, dbPath )

	self._path = dbPath
	docStoreImpLog:infof( 'dbPath = %s', dbPath )

	local db = WFSQLiteDatabase( dbPath, schema )
	self:setOwned { _db = db }

	self._exec = assert( db.exec )
	self._gatherArrayOfArray = assert( db.gatherArrayOfArray )

	return db

end

--------------------------------------------------------------------------------

local function maybeCloseAndClearDatabase( self )

	if not self._db then

		return

	end

	self._db:close()

	self._db = nil

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:__init( path )

	check( 'WFRxDocStoreImp', 'string' )

	-- Map :memory:(random) and :temp:(random) paths back to SQLite pseudo-paths.

	local dbPath = path

	if dbPath:match( '^:memory:' ) then

		dbPath = ':memory:'

	elseif dbPath:match( '^:temp:' ) then

		dbPath = ':temp:'

	elseif #dbPath > 0 then

		-- An empty path is legal and means create a temporary path on disk.

		local parentDir = KSPathUtils.parent( dbPath )
		WFFileUtils.mkdir( parentDir )

	end

	-- Create the actual database. Note that we disable query performance logging
	-- because the query manifest is very simple for this database and we've done
	-- what can be done IMHO.

	local db = openAndObtainDatabase( self, dbPath )

	-- Set up change notification.

	local changeCounterSeedRow = db:execOnce(
		"SELECT max( sequence ) AS changeCounter FROM revs" )

	local changeCounter = ( changeCounterSeedRow
		and changeCounterSeedRow.changeCounter or 0 )

	self._changeCounter = BehaviorSubject( changeCounter )

	self._deletedDocStream = ReplaySubject( 1 )

	self._ongoingDocValues = WFRxWeakObservableTable(
		function( docId )
			return makeOngoingDocValueForDocId( self, docId )
		end )

	self._ongoingLocalDocValues = WFRxWeakObservableTable(
		function( name )
			return makeOngoingLocalDocValueForDocName( self, name )
		end )

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:close()

	maybeCloseAndClearDatabase( self )

	if self._changeCounter then
		self._changeCounter:onCompleted()
		self._changeCounter = nil
	end

	if self._ongoingDocValues then
		self._ongoingDocValues:dispose()
		self._ongoingDocValues = nil
	end

	if self._ongoingLocalDocValues then
		self._ongoingLocalDocValues:dispose()
		self._ongoingLocalDocValues = nil
	end

	if self._deletedDocStream then
		self._deletedDocStream:onCompleted()
		self._deletedDocStream = nil
	end

end

--------------------------------------------------------------------------------

local function assertNotClosed( self, methName )

	if self._db then

		return

	end

	local errMsg = string.format( 
					"WFRxDocStoreImp:%s: doc store has been closed", 
					type( methName ) == "string" and 
						methName or "<unknown method>" )

	docStoreImpLog:error( errMsg )

	error( errMsg )

end


--------------------------------------------------------------------------------

-- @mark shared utilities

--------------------------------------------------------------------------------

local underscoreByte = string_byte( '_' )

--------------------------------------------------------------------------------

local function assertDocIdIsValid( docId )

	check( '+string|number' )

	if type( docId ) == 'number' then
		assert( docId > 0 and math_floor( docId ) == docId,
				"local id must be positive integer, was " .. tostring( docId ) )
	else
		assert( docId:byte( 1 ) ~= underscoreByte,
				"document ID must not start with _" )

		assert( docId:match( "^[0-9a-zA-Z_.%-]+$" ),
				"document ID must only contain letters, numbers, and _.-" )
	end

	return docId

end

--------------------------------------------------------------------------------

local function createLocalDocId_inTransaction( db, exec, fullDocId )

	exec( db, [[INSERT INTO docs( fullDocId ) VALUES( ? )]], fullDocId )

	return db:getLastInsertRowId()

end

--------------------------------------------------------------------------------
-- Returns four values:
	-- 1: winning revision ID for this document ID
	-- 2: what is its sequence number?
	-- 3: is this revision deleted?
	-- 4: are there any conflicting revisions?

local function winningRevIdForLocalDocId_inTransaction( self, localDocId, favorSequence )

	assert( localDocId > 0 )

	local revRows = self._db:gatherArrayOfArray(
		[[SELECT revId, sequence, deleted
			FROM revs
				WHERE localDocId = ? AND current = 1
				ORDER BY ( sequence == ? ) DESC, deleted ASC, revId DESC
				LIMIT 2]],
		localDocId, favorSequence )

	local winningRow = revRows[ 1 ]

	if winningRow then

		local deleted = winningRow[ 3 ] == 1
		local conflict = false

		if not deleted then

			local maybeConflictingRow = revRows[ 2 ]
			conflict = not not( maybeConflictingRow and maybeConflictingRow[ 3 ] == 0 )

		end

		return winningRow[ 1 ], winningRow[ 2 ], deleted, conflict

	end

end

--------------------------------------------------------------------------------

local acceptableTypes = {
	[ 'boolean' ] = true,
	[ 'number' ] = true,
	[ 'string' ] = true,
	[ 'table' ] = true,
}

--------------------------------------------------------------------------------

local approvedTopLevelKeys = {
	_id = true,
	_revId = true,
	_revIds = true,
	_localId = true,
	_deleted = true,
	_localOnly = true,
	_labels = true,
	_serverProvidedMetadata = true,
	_newRevId = true,
}

--------------------------------------------------------------------------------

local kInfinity = 1/0
local kNegativeInfinity = -kInfinity

--------------------------------------------------------------------------------

local function assertDocValueIsValid( value, nestingLevel )

	nestingLevel = nestingLevel or 1

	if nestingLevel > 20 then
		error "table nested too deeply"
	end

	if getmetatable( value ) then
		error "table contains metatable"
	end

	if nestingLevel == 1 and value._deleted ~= nil and value._deleted ~= true then
		errorf( "value contains illegal _deleted value: %s (must be nil or true)",
				tostring( value._deleted ) )
	end

	if nestingLevel == 1 and value._localOnly ~= nil and type( value._localOnly ) ~= 'table' then
		errorf( "_localOnly must be a table" )
	end

	local hasStringKeys, hasNumericKeys

	for k, v in pairs( value ) do

		local tk = type( k )

		if tk == 'string' then

			if nestingLevel == 1 then

				if string_byte( k, 1 ) == underscoreByte and not approvedTopLevelKeys[ k ] then
					errorf( "value contains top-level key with leading underscore (%s)", k )
				end

			end

			hasStringKeys = true

		elseif tk == 'number' then

			if nestingLevel == 1 then
				error( "numeric keys not allowed at top level of doc value" )
			end

			if k <= 0 or math_floor( k ) ~= k then
				error( "numeric keys must be positive integers" )
			end

			if k == kInfinity then
				-- Note that Lua disallows NaN as key, so no need to test for that case here.

				errorf( "key %s is an invalid number", tostring( k ) )

			end

			hasNumericKeys = true

		else

			errorf( "key %s is of type %s, must be string%s",
					tostring( k ), tk,
					nestingLevel > 1 and " or number" or "" )

		end

		local tv = type( v )

		if v == kInfinity or v == kNegativeInfinity or v ~= v then
			errorf( "value %s is an invalid number", tostring( v ) )
		end

		if not acceptableTypes[ tv ] then
			errorf( "value at key %s is of type %s (must be scalar or table)", k, tv )
		end

		if nestingLevel == 1 and k == '_labels' then

			if tv ~= 'table' then
				error( "_labels must be a table if present" )
			end

			for label, labelTrue in pairs( v ) do

				if type( label ) ~= 'string' or #label == 0 then
					error( "label name must be non-empty string" )
				end

				if labelTrue ~= true then
					error( "label value must be true" )
				end

			end

		elseif nestingLevel == 1 and k == '_revIds' then

			if tv ~= 'table' then
				error( "_revIds must be a table if present" )
			end

			for i, revId in pairs( v ) do

				if i == 'force' and revId == true then

					-- OK: whitelisted

				else

					if type( i ) ~= 'number' then
						error( "_revIds must be an array of revision IDs" )
					end

					if type( revId ) ~= 'string' then
						error( "_revIds must be an array of (string) revision IDs" )
					end

				end

			end

		elseif tv == 'table' then

			assertDocValueIsValid( v, nestingLevel + 1 )

		end

	end

	if hasStringKeys and hasNumericKeys then
		error( "tables must not have both string and numeric keys" )
	end

end

--------------------------------------------------------------------------------

local function documentRevisionExists_inTransaction( db, exec, localDocId, revId )

	local revisionRow = exec( db,
			[[SELECT sequence FROM revs WHERE localDocId = ? AND revId = ?
				ORDER BY sequence DESC LIMIT 1]],
			localDocId, revId )

	return not not revisionRow

end

--------------------------------------------------------------------------------

local function getSequenceForDocumentAndRevision_inTransaction( db,
																exec,
																localDocId,
																revId,
																onlyCurrent )

	-- Choose the appropriate SQL query. The "easy" way would be to do
	-- string.format to assemble the query, but this is likely to be called
	-- frequently enough that the formatting overhead will be noticable.
	-- So we'll spell out the variations explicitly.

	local sql

	if onlyCurrent then

		sql = [[SELECT sequence FROM revs
					WHERE localDocId = ? AND revId = ? AND current = 1
					ORDER BY sequence DESC LIMIT 1]]

	else

		sql = [[SELECT sequence FROM revs
					WHERE localDocId = ? AND revId = ?
					ORDER BY sequence DESC LIMIT 1]]

	end

	local sequenceRow = exec( db, sql, localDocId, revId )

	return sequenceRow and sequenceRow.sequence

end

--------------------------------------------------------------------------------

-- @mark whole document values

--------------------------------------------------------------------------------

local valueFnEnv = {} -- disallow all function calls

--------------------------------------------------------------------------------

local function contentFromFlattenedValue( value )

	-- Lightweight sandboxing. The one potentially malicious piece of code
	-- somebody could inject (and they'd have to do it directly into the
	-- SQLite database) would be to put an infinite loop here.
	-- Some more potentially valuable approaches we could take are
	-- described in http://lua-users.org/wiki/SandBoxes.

	local valueFn = loadstring( 'return ' .. value )
	assert( type( valueFn ) == 'function' )
	setfenv( valueFn, valueFnEnv ) -- disallow all function calls

	return valueFn()

end

--------------------------------------------------------------------------------

local function onlyFieldsInSpec( spec, t )
	if not spec then
		return t
	end

	local res = { }
	
	for k, subspec in pairs( spec ) do
		local v = t[ k ]
		if type( subspec ) == "table" then
			if type( v ) == "table" then
				res[ k ] = onlyFieldsInSpec( subspec, v ) -- apply recursively
			end
			
			-- if the value is not a table, it is not a match, so exclude it
		elseif subspec == true then
			res[ k ] = v -- include the whole value, regardless of type
		else
			docStoreImpLog:warnf( "invalid field spec (type %s) for key: %s", 
				type( subspec ), tostring( k ) )
		end
	end
	
	return res
end

--------------------------------------------------------------------------------

local function docValue_inTransaction( _, db, exec, docId, revId, ignoreNonMatch, addDeletionMarker, optFieldSpec )

	local revisionRow

	local swapIds = type( docId ) == 'number'
		-- after querying, swap localId and docId
	local whereIdCol = swapIds and 'docs.localDocId' or 'fullDocId'
	local selectIdCol = swapIds and 'fullDocId' or 'docs.localDocId'

	if revId then

		revisionRow = exec( db,
			[[SELECT ]] .. selectIdCol .. [[ AS selectIdCol, revId, sequence, content, annotation, revs.deleted, hasConflicts
				FROM revs
					INNER JOIN docs ON docs.localDocId = revs.localDocId
				WHERE
					]] .. whereIdCol .. [[ = ? AND revId = ?
				ORDER BY sequence DESC LIMIT 1]], docId, revId )

			-- EXPLAIN QUERY PLAN says this uses automatic indexes on both
			-- revs and docs tables.

		if not revisionRow then

			-- If we don't find a match on rev ID, try matching it as a label.

			local labelRow = exec( db,
				[[SELECT revSequence
					FROM labels
						INNER JOIN docs ON docs.localDocId = labels.localDocId
					WHERE
						]] .. whereIdCol .. [[ = ? AND name = ?]], docId, revId )

			if labelRow and labelRow.revSequence then
				revisionRow = exec( db,
					[[SELECT ]] .. selectIdCol .. [[ AS selectIdCol, revId, sequence, content, annotation, revs.deleted, hasConflicts
						FROM revs
							INNER JOIN docs ON docs.localDocId = revs.localDocId
						WHERE
							]] .. whereIdCol .. [[ = ? AND sequence = ?]], docId, labelRow.revSequence )
			end

		end

		if not revisionRow then
			if ignoreNonMatch then
				return
			else
				errorf( "WFRxDocStoreImp:docValue: no such revision or label %s for %s", revId, docId ) end
			--end  -- funky formatting because coverage views this as an unreachable line :-(
		end

	else

		revisionRow = exec( db,
			[[SELECT ]] .. selectIdCol .. [[ AS selectIdCol, revId, sequence, content, annotation, hasConflicts,
					docs.deleted AS deleted,
					docs.mostRecentRevSequence AS mostRecentRevSequence
				FROM revs
					INNER JOIN docs ON docs.winningRevSequence = revs.sequence
				WHERE
					]] .. whereIdCol .. [[ = ?]], docId )

			-- EXPLAIN QUERY PLAN says this uses automatic index on docs and
			-- integer primary key on revs.

		if not revisionRow or revisionRow.deleted == 1 then
			return {
				_id = docId,
				_empty = true,
				_conflict = ( revisionRow and revisionRow.hasConflicts ~= 0 ) or nil,
				_deleted = ( addDeletionMarker and revisionRow and revisionRow.deleted == 1 ) or nil
			}
		end

	end

	if revisionRow.content == nil then
		error "Revision has been purged"
	end

	local docValue = onlyFieldsInSpec( optFieldSpec, contentFromFlattenedValue( revisionRow.content ) )
	local annotation = onlyFieldsInSpec( optFieldSpec, contentFromFlattenedValue( revisionRow.annotation ) )

	-- Body content is per-revision; annotation content is unversioned
	-- (i.e. per document ID). Merge them together here.

	for k, v in pairs( annotation ) do
		docValue[ k ] = v
	end

	if swapIds then
		docValue._id = revisionRow.selectIdCol
		docValue._localId = docId
	else
		docValue._id = docId
		docValue._localId = revisionRow.selectIdCol
	end

	docValue._revId = revisionRow.revId
	docValue._changeCounter = revisionRow.mostRecentRevSequence or revisionRow.sequence

	if revisionRow.hasConflicts == 1 then
		docValue._conflict = true
	end

	-- Merge in any labels that may be associated with this revision.

	local labelRows = db:gatherArrayOfArray( [[SELECT labels.name FROM labels WHERE revSequence = ?]],
											 revisionRow.sequence )

	if #labelRows > 0 then

		local labels = {}

		for _, labelRow in ipairs( labelRows ) do
			labels[ labelRow[ 1 ] ] = true
		end

		docValue._labels = labels

	end

	return docValue

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the value of a single document once.

	\param docId (string) the document's unique ID
	\param revId (optional, string) revision to retrieve (can also be a label name)
	\param optFieldSpec (optional, table of arrays of strings) paths of fields to include.
		For example, for the purpose of retrieving a thumbnails, this optFieldSpec might be used:
		{
			{ "_id" },
			{ "_revId" },
			{ "_serverProvidedMetadata", "base" },
			{ "_serverProvidedMetadata", "links", "/rels/rendition_type/thumbnail2x", "href" },
		}
	}

	\return <code>Observable< table ></code> value for this document

	Note that the document content will be annotated with the following
	system-generated entries (at the top level of the value table):

		* \c _id (string): the document's unique ID
		* \c _empty (boolean): true if there isn't actually a value for this doc ID
		* \c _deleted (boolean): true if this revision is deleted (only used when
			\c revId is present
		* \c _changeCounter (number): a monotonically-increasing change counter
			that is increased whenever any branch of this document is revised
			(The functions \c changeCounter and \c changeCounterOnce will
			return the maximum value of \c _changeCounter across the entire
			database.)
		* \c _labels (table): if present, this table is the set of labels that
			have been applied to this revision.
]]--

function WFRxDocStoreImp:docValue( docId, revId, optFieldSpec )

	check( 'WFRxDocStoreImp', 'string|number', '?+string' )

	assertNotClosed( self, 'docValue' )

	docStoreImpLog:debugf( "docValue %s revId = %s", docId, revId )

	assertDocIdIsValid( docId )

	local db = self._db

	-- Wrap this in a transaction to ensure no other thread / process changes
	-- values between the various queries that are performed.

	local docValue = db:withTransactionDo( function()

		return docValue_inTransaction( self, db, self._exec, docId, revId, false, true, optFieldSpec)

	end )

	return Observable.just( docValue )

end

--------------------------------------------------------------------------------

local function bulkDocValueImp( self, docIds, optFieldSpec )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'bulkDocValue' )

	docStoreImpLog:debugf( "bulkDocValue docIds = %s", docIds )

	for i = 1, #docIds do
		assertDocIdIsValid( docIds[ i ] )
	end

	local db = self._db

	local label = docIds.label
	if label ~= nil then
		assert( type( label ) == 'string' and #label > 0, "label, if present, must be a non-empty string" )
	end

	-- Wrap this in a transaction to ensure no other thread / process changes
	-- values between the various queries that are performed.

	return db:withTransactionDo( function()

		local docValues = {}

		local nDocs = 0

		for i = 1, #docIds do

			local docValue = docValue_inTransaction( self, db, self._exec, docIds[ i ], label, true, true )

			if docValue then
				nDocs = nDocs + 1
				docValues[ nDocs ] = onlyFieldsInSpec( optFieldSpec, docValue )
			end

		end

		return docValues

	end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Gets many document values at once.

	\param docIds (table) array of document IDs
	\param optFieldSpec (optional, table of arrays of strings) paths of fields to include.
		For example, for the purpose of retrieving a thumbnails, this optFieldSpec might be used:
		{
			{ "_id" },
			{ "_revId" },
			{ "_serverProvidedMetadata", "base" },
			{ "_serverProvidedMetadata", "links", "/rels/rendition_type/thumbnail2x", "href" },
		}
	}

	\return <code>Observable< table ></code> resolves to an array of document
		values when done. The items in the array will be the document values
		in the same sequence as requested on input. Each item will contain the
		same results as would be returned for an equivalent \c :docValue call.

	It is not possible to request specific revision IDs via this API.
]]--

function WFRxDocStoreImp:bulkDocValue( docIds, optFieldSpec )

	return Observable.just( bulkDocValueImp( self, docIds, optFieldSpec ) )

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:benchmarkedBulkDocValue( docIds )

	local start = KSDate.timeNowForAnimation()

	local result = bulkDocValueImp( self, docIds )

	return Observable.just( result, KSDate.timeNowForAnimation() - start )

end

--------------------------------------------------------------------------------

makeOngoingDocValueForDocId = function( self, docId )

	local result = ReplaySubject( 1 )

	local db = self._db

	local docValue = db:withTransactionDo( function()

		return docValue_inTransaction( self, db, self._exec, docId )

	end )

	result:onNext( docValue )

	return result

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:ongoingDocValue( docId )

	check( 'WFRxDocStoreImp', 'string|number' )

	assertNotClosed( self, 'ongoingDocValue' )

	docStoreImpLog:debugf( "ongoingDocValue %s", docId )

	assertDocIdIsValid( docId )

	return self._ongoingDocValues:observableForKey( docId )

end

--------------------------------------------------------------------------------

makeOngoingLocalDocValueForDocName = function( self, name )

	local result = ReplaySubject( 1 )

	local db = self._db

	local localDocValueRow = self._exec( db,
										 [[SELECT content FROM localDocs WHERE name == ?]],
										 name )

	local docValue = localDocValueRow and contentFromFlattenedValue( localDocValueRow.content )

	if docValue then
		docValue._id = name
	end
	
	result:onNext( docValue or { _id = name, _empty = true } )

	return result

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:ongoingLocalDocValue( name )

	check( 'WFRxDocStoreImp', 'string' )

	assertNotClosed( self, 'ongoingLocalDocValue' )

	docStoreImpLog:debugf( "ongoingLocalDocValue %s", name )

	return self._ongoingLocalDocValues:observableForKey( name )

end

--------------------------------------------------------------------------------

local function validateAndExtractDocValue( _, docValue )

	-- Verify that doc value is acceptable (does not contain illegal keys, etc.).

	assertDocValueIsValid( docValue )

	local annotationContent = {}
	local bodyContent = {}

	for k, v in pairs( docValue ) do
		if string_byte( k, 1 ) == underscoreByte and k ~= '_deleted' then
			annotationContent[ k ] = v
		else
			bodyContent[ k ] = v
		end
	end

	return bodyContent, annotationContent

end

--------------------------------------------------------------------------------

local function validateAndArchiveDocValue( _, _, docValue, annotation )

	-- Extract doc ID from the value.

	local docId = assertDocIdIsValid( annotation._id )
	annotation._id = nil
	annotation._localId = nil

	-- Extract the new revId from the value
	
	local newRevId = annotation._newRevId
	annotation._newRevId = nil
	
	if newRevId ~= nil and type( newRevId ) ~= 'string' then
		error( "_newRevId must be a string if present" )
	end

	-- Extract parent rev ID from the value.

	local parentRevId = annotation._revId

	if parentRevId ~= nil and type( parentRevId ) ~= 'string' then
		error( "_revId must be a string if present" )
	end

	annotation._revId = nil

	-- Or extract parent rev IDs array.

	local parentRevIds = annotation._revIds

	if parentRevIds ~= nil
		and ( type( parentRevIds ) ~= 'table' or #parentRevIds == 0 )
	then
		error( "_revIds must be a non-empty array if present" )
	end

	annotation._revIds = nil

	-- Can't have both.

	if parentRevId and parentRevIds then
		error( "Must not specify both _revId and _revIds" )
	end

	-- Extract labels from the value.

	local labels = annotation._labels
	annotation._labels = nil

	-- Convert Lua to text.

	local bodyContentArchive = tableToString( docValue )
	local annotationContentArchive = tableToString( annotation )

	return
		{
			docId = docId,
			parentRevId = parentRevId,
			parentRevIds = parentRevIds,
			newRevId = newRevId,
			bodyContentArchive = bodyContentArchive,
			annotationContentArchive = annotationContentArchive,
			docValue = docValue,
			annotation = annotation,
			labels = labels,
		}

end

--------------------------------------------------------------------------------

local function extractAndValidateDocValue( self, funcName, docValue )

	local bodyContent, annotationContent = validateAndExtractDocValue( self, docValue )

	return validateAndArchiveDocValue( self, funcName, bodyContent, annotationContent )

end

--------------------------------------------------------------------------------

local function validateDocValue_inTransaction() -- self, db )

	-- TODO [scouten 2014-08-19]: Implement validation.
	-- For now, everything validates.

end

--------------------------------------------------------------------------------

local function setDocValueImp_inTransaction( self, db, exec, _, args )

	local docId = args.docId
	local parentRevId = args.parentRevId
	local parentRevIds = args.parentRevIds
	local bodyContentArchive = args.bodyContentArchive
	local annotationContentArchive = args.annotationContentArchive
	local docValue = args.docValue
	local annotation = args.annotation
	local labels = args.labels

	local swapIds = type( docId ) == 'number'
		-- after querying, swap localId and docId
	local whereIdCol = swapIds and 'docs.localDocId' or 'fullDocId'
	local selectIdCol = swapIds and 'fullDocId' or 'docs.localDocId'

	-- Get the doc's local numeric ID and its current winning revision.

	local docIdRow = exec( db,
						   [[SELECT ]] .. selectIdCol .. [[ AS selectId, winningRevSequence,
						   		deleted, annotation
								FROM docs WHERE ]] .. whereIdCol .. [[ = ?]],
						   docId )

		-- EXPLAIN QUERY PLAN says this uses auto index on docs table.

	local localDocId
	if swapIds then
		localDocId = docId
		docId = docIdRow and docIdRow.selectId
	else
		localDocId = docIdRow and docIdRow.selectId
	end

	if not docId then
		errorf( "Cannot create a new document with setDocValue and localId (%d).", localDocId )
	end

	local oldWinningRevSequence = docIdRow and docIdRow.winningRevSequence
	local oldRevIsDeleted = docIdRow and ( docIdRow.deleted == 1 )
	local oldDocExisted = not not docIdRow

	-- Special case: If no _localOnly value is provided, then carry over any
	-- previous value.

	if not( annotation and annotation._localOnly ) then

		local priorAnnotationArchive = docIdRow and docIdRow.annotation

		local priorAnnotation = priorAnnotationArchive
			and contentFromFlattenedValue( priorAnnotationArchive )
			or {}

		local priorLocalOnly = priorAnnotation._localOnly

		if priorLocalOnly then
			annotation._localOnly = priorLocalOnly
			annotationContentArchive = tableToString( annotation )
		end

	end

	-- Identify parent revision.

	local parentSequence

	if parentRevIds and localDocId then

		-- We think we're replacing some existing revision. In this case, we allow
		-- the revision to be non-current, but we always lose if a conflict would be
		-- created.

		for i = #parentRevIds, 1, -1 do

			parentSequence = getSequenceForDocumentAndRevision_inTransaction( db, exec,
											localDocId, parentRevIds[ i ], false )

			if parentSequence then
				break
			end

		end

		if oldRevIsDeleted and not parentSequence then

			-- If current rev is a deletion, then we'll ignore the fact that
			-- we didn't get a _revId parameter and plunk it in.

			local revIdRow = exec( db, [[SELECT revId FROM revs WHERE sequence = ?]],
								   oldWinningRevSequence )

			parentRevId = revIdRow and revIdRow.revId
			parentSequence = oldWinningRevSequence

		end

		-- The search for a parent sequence may fail to come up with the right parentSequence
		-- if Oz compressed away revisions. If we find a revision in the docStore with a higher
		-- sequence number that bears the same label as this new doc, use that sequence number
		-- as parentSequence. If the feed sets more then one label, use the highest sequence number
		-- for any of the labels.
		
		if not parentRevIds.force then
			-- Optimize away the common case, that is that there are no local edits.
			if labels and (not parentSequence or parentSequence ~= oldWinningRevSequence) then
				for label in pairs( labels ) do
					local  revSequenceRow = exec( db, [[SELECT revSequence FROM labels
													WHERE localDocId = ? AND name = ?
													ORDER BY revSequence DESC LIMIT 1]],
												localDocId, label )
					-- EXPLAIN QUERY PLAN says this uses auto index on labels table (created by UNIQUE constraint)

					local labeledSequence = revSequenceRow and revSequenceRow.revSequence

					if not parentSequence or (labeledSequence and labeledSequence > parentSequence) then
						parentSequence = labeledSequence
					end
				end
			end
		end

		-- Validate doc value.

		validateDocValue_inTransaction( self, db )

	elseif parentRevId then

		-- We're replacing an existing revision. Make sure it's current
		-- and find its sequence number.

		-- TODO [scouten 2014-08-25]: Optimize for the common case where we
		-- already got the parent sequence ID from previous query.

		parentSequence = getSequenceForDocumentAndRevision_inTransaction( db, exec,
											localDocId, parentRevId, true )

		if not parentSequence then

			if documentRevisionExists_inTransaction( db, exec, localDocId, parentRevId ) then

				errorf( "_rev %s is not current (would create conflict)", parentRevId )

			else

				errorf( "can not find _rev %s", parentRevId )

			end

		end

		-- Validate doc value.

		validateDocValue_inTransaction( self, db )

	else

		-- Validate doc value.

		validateDocValue_inTransaction( self, db )

		-- Does this doc exist?

		if localDocId then

			if oldRevIsDeleted then

				-- If current rev is a deletion, then we'll ignore the fact that
				-- we didn't get a _revId parameter and plunk it in.

				local revIdRow = exec( db, [[SELECT revId FROM revs WHERE sequence = ?]],
									   oldWinningRevSequence )

				parentRevId = revIdRow and revIdRow.revId
				parentSequence = oldWinningRevSequence

			elseif oldWinningRevSequence then

				docStoreImpLog:errorf( "Creating new revision without specifying a parent ID: %s",
									   docValue )

				errorf( "%s, can not create new revision (would create conflict)", ERROR_REVISION_WITHOUT_PARENT ) end

			--end  (funky formatting because code coverage doesn't see this line as reachable)

		else

			-- Doc didn't previously exist: Create a local ID for it now.

			localDocId = createLocalDocId_inTransaction( db, exec, docId )

		end

	end

	local parentRevIsWinning = ( parentRevId or parentRevIds ) and ( parentSequence == oldWinningRevSequence )

	-- Generate new rev ID.

	local newRevId

	if parentRevIds then

		newRevId = parentRevIds[ #parentRevIds ]

		-- If we're force-deleting (i.e. deletion from Oz without revision_ids provided),
		-- graft this deletion revision on top of the previous parent.

		if docValue._deleted and not parentSequence and parentRevIds.force then
			parentSequence = oldWinningRevSequence
			parentRevIsWinning = true
		end

	else

		-- The change is originating locally (i.e. via typical :setDocValue API).

		-- Special case: If we're not changing the body content of the document, don't
		-- change the revision ID.

		local suppressNewRevId = false

		if parentSequence and parentRevIsWinning and not oldRevIsDeleted then

			local existingParentRow = exec( db,
								[[SELECT content FROM revs WHERE sequence == ?]],
								parentSequence )

			suppressNewRevId = bodyContentArchive == ( existingParentRow and existingParentRow.content )

		end

		if suppressNewRevId then
			newRevId = parentRevId
		else
			newRevId = args.newRevId or KSIdGenerator.uuidInLowercaseHex()
		end

	end

	-- We need to know if parent rev is current. If not, we're creating a conflict.

	local parentRevWasCurrent = parentRevIsWinning

	if not parentRevIsWinning and parentSequence then
		-- A winning rev is inherently current. In this (common) case, we should
		-- optimize away the following query.

		local parentRevIsCurrentRow = exec( db,
							[[SELECT current FROM revs WHERE sequence == ?]],
							parentSequence )

		if parentRevIsCurrentRow.current == 1 then
			parentRevWasCurrent = true
		end

	end

	-- Actually insert the new revision.

	exec( db,
		  [[INSERT INTO revs( localDocId, revId, parentRevSequence, current, deleted, content )
				VALUES( ?, ?, ?, ?, ?, ? )]],
		  localDocId,
		  newRevId,
		  parentSequence,
		  args.markNonCurrent == true and 0 or 1,
		  docValue._deleted == true and 1 or 0,
		  bodyContentArchive )

	-- Get new change counter, but don't update yet.

	local newRevSequence = self._db:getLastInsertRowId()

	-- If we replaced a revision, make it no longer current.

	if parentSequence then
		exec( db, [[UPDATE revs SET current = 0 WHERE sequence = ?]], parentSequence )
	end

	-- Update document summary information.

	local newWinningRevId, newWinningRevSequence, newRevIsDeleted, newRevIsInConflict

	if oldDocExisted then

		local favorRevSequence
		if parentRevIsWinning and parentSequence == oldWinningRevSequence then
			favorRevSequence = newRevSequence
		elseif parentRevIds then
			if parentRevIds.force then
				favorRevSequence = newRevSequence
			else
				favorRevSequence = oldWinningRevSequence
			end
		end

		newWinningRevId, newWinningRevSequence, newRevIsDeleted, newRevIsInConflict
			= winningRevIdForLocalDocId_inTransaction( self, localDocId, favorRevSequence )

		-- Any derivation from a non-head revision is a conflict, even if we didn't
		-- detect it above. (Reasoning: We need to know about and preserved a deletion
		-- from Oz that conflicts with a locally-originated edit. But we also rely on
		-- deleting a current revision as the means of resolving a conflict, so a rule
		-- that simply says creating a non-winning revision flags a conflict is
		-- incorrect.)

		if not parentRevWasCurrent then
			newRevIsInConflict = true
		end

	else

		newWinningRevId = newRevId
		newWinningRevSequence = newRevSequence
		newRevIsDeleted = docValue._deleted == true
		newRevIsInConflict = false

	end

	if newRevIsInConflict then
		docStoreImpLog:debugf( "Creating a conflicting revision %s", docValue )
	end

	exec( db, [[UPDATE docs
					SET winningRevSequence = ?,
						mostRecentRevSequence = ?,
						deleted = ?,
						hasConflicts = ?,
						annotation = ?
					WHERE localDocId = ?]],
		  newWinningRevSequence,
		  newRevSequence,
		  newRevIsDeleted and 1 or 0,
		  newRevIsInConflict and 1 or 0,
		  annotationContentArchive,
		  localDocId )

		-- EXPLAIN QUERY PLAN says this uses integer primary key on docs.

	-- Update labels.

	if labels then

		for label in pairs( labels ) do
			exec( db, [[INSERT OR REPLACE INTO labels VALUES( ?, ?, ? )]],
				  localDocId, newRevSequence, label )

		end

	end

	local buildFinalDocValue = function()

		local dv = {}

		for k, v in pairs( docValue ) do
			dv[ k ] = v
		end

		for k, v in pairs( annotation ) do
			dv[ k ] = v
		end

		dv._localId = localDocId
		dv._id = docId
		dv._revId = newRevId
		dv._changeCounter = newRevSequence

		if newRevIsInConflict then
			dv._conflict = true
		end

		return dv

	end

	-- If the new revision is deleted, signal the deletion via the deleted doc stream

	local finalDocValue = nil

	if newRevIsDeleted then

		finalDocValue = buildFinalDocValue()
		self._deletedDocStream:onNext( finalDocValue )

	end

	-- Do we need to update an observable doc value?

	local ongoingDocValues = self._ongoingDocValues

	if ongoingDocValues:hasObservableForKey( docId ) then

		local ongoingDocValue = ongoingDocValues:observableForKey( docId )

		if newWinningRevId == newRevId then

			if not finalDocValue then
				finalDocValue = buildFinalDocValue()
			end

			if finalDocValue._deleted then
				ongoingDocValue:onNext( { _id = docId, _empty = true } )
			else
				ongoingDocValue:onNext( finalDocValue )
			end

		else

			local docValueNow = docValue_inTransaction( self, db, self._exec, docId )

			ongoingDocValue:onNext( docValueNow )

		end

	end

	-- Return rev ID that we assigned and also the new change counter,
	-- which must be updated outside of the transaction.

	return newRevId, newRevSequence, localDocId

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the document value, replacing any existing value.

	\param value (table) new value for the entire document

	\c value \b must contain the following elements:
		* \c _id (string) the document's unique ID. If the document already
			exists and you know the _localId, you may supply an integer
			local ID instead of the unique ID.
		* \c _revId (string) the document's current revision ID
			(unless the document is new, in which case it may be omitted)
		* \c _revIds (array of strings) intended for use in sync; if provided:
			* the array must be non-empty
			* the new revision's ID will be the last rev ID in the array
			* the new revision's ID will be the last member of this array
			* the new revision's parent ID will be the \b LAST rev ID in the
				array that matches an existing ID
			* any conflict that might arise will be resolved by making the
				new revision the \b losing revision (separate code will resolve
				the conflict)
			* in this case, \c _revId must \b not be specified
		* \c _labels (table) revision labels to apply to this revision
			(Any existing labels named here will be moved to this revision.)

	\c value \b may contain the following elements:
		* \c _deleted (boolean): if present, must be \c true and no other
			values may be present

	\b Important: The top-level key \c _localOnly gets special handling. If
		no value is provided for this key, any previous value is left in place.
		This behavior does not apply to any other key. This does imply that you
		can not delete any existing \c _localOnly entry in the document.

	\return <code>Observable< string ></code> resolves with the new value's
		revision ID when the write completes. You may safely ignore this Observable.
]]--

function WFRxDocStoreImp:setDocValue( docValue )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'setDocValue' )

	docStoreImpLog:debugf( "setDocValue %s", docValue )

	local args = extractAndValidateDocValue( self, 'setDocValue', docValue )

	local db = self._db

	local revId, newRevSequence = db:withTransactionDo( function()

		return setDocValueImp_inTransaction( self, db, self._exec,
											 'setDocValue', args )

	end )

	if newRevSequence then
		self._changeCounter:onNext( newRevSequence )
	end

	return Observable.just( revId )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets many document values at once, replacing any existing values.

	\param values (table) array of entire document values

	The array may contain zero or more document values. Each value entry
	looks like the \c value parameter to \c :setDocValue.

	The array may also contain a <code>label = "name"</code> entry. If this exists,
	that label will be applied to all documents in this batch.

	If an error occurs while setting any one of these document values, then the
	entire set of document values will be discarded. In other words, this set
	of documents succeeds or fails atomically.

	\return <code>Observable< table ></code> resolves to (what?) when done.
		You may safely ignore this promise.

	TODO [scouten 2015-05-07]: Describe results table.
]]--

function WFRxDocStoreImp:bulkSetDocValue( docValues )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'bulkSetDocValue' )

	docStoreImpLog:debugf( "bulkSetDocValue %s", docValues )

	local nDocValues = #docValues

	if nDocValues == 0 then
		error( "WFRxDocStoreImp:bulkSetDocValue: no values in this array", 2 )
	end

	if nDocValues > 100 then
		docStoreImpLog:warnf( string_format( "WFRxDocStoreImp:bulkSetDocValue: "
											 .. "%d values provided in this array, "
											 .. "exceeds recommended maximum (100)",
							  nDocValues ), 2 )
	end

	local label = docValues.label

	if label and ( type( label ) ~= 'string' or #label == 0 ) then
		error( "WFRxDocStoreImp:bulkSetDocValue: label must be a non-empty string if provided", 2 )
	end

	local db = self._db

	local revResults, finalRevSequence = db:withTransactionDo( function()

		local results = {}
		local latestRevSequence

		for i = 1, nDocValues do

			local args = extractAndValidateDocValue( self, 'bulkSetDocValue', docValues[ i ] )

			if label then
				args.labels = args.labels or {}
				args.labels[ label ] = true
			end

			local revId, newRevSequence, localId = setDocValueImp_inTransaction(
								self, db, self._exec, 'bulkSetDocValue', args )

			if newRevSequence then
				latestRevSequence = newRevSequence
			end

			results[ i ] = {
				id = args.docId,
				ok = true,
				revId = revId,
				localId = localId,
			}

		end

		return results, latestRevSequence

	end )

	if finalRevSequence then
		self._changeCounter:onNext( finalRevSequence )
	end

	return Observable.just( revResults )

end

--------------------------------------------------------------------------------

local function currentRevisionsForDoc_inTransaction( _, db, _, docId, includeDeleted )

	local revisionRows

	if includeDeleted then

		revisionRows = db:gatherArrayOfArray(
			[[SELECT revId, sequence, content, revs.deleted, hasConflicts, annotation, docs.localDocId
				FROM revs
					INNER JOIN docs ON docs.localDocId = revs.localDocId
				WHERE
					fullDocId == ? AND current == 1
				ORDER BY ( sequence == docs.winningRevSequence ) DESC,
						 revs.deleted ASC, revId DESC]],
				docId )

	else

		revisionRows = db:gatherArrayOfArray(
			[[SELECT revId, sequence, content, revs.deleted, hasConflicts, annotation, docs.localDocId
				FROM revs
					INNER JOIN docs ON docs.localDocId = revs.localDocId
				WHERE
					fullDocId == ? AND current == 1 AND revs.deleted == 0
				ORDER BY ( sequence == docs.winningRevSequence ) DESC,
						 revId DESC]],
				docId )

	end

	return revisionRows

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns all current revisions of a document.

	\param docId (string) the document's unique ID
	\param includeDeleted (boolean) true to include revisions that are current
		but have been deleted

	\return <code>Observable< table ></code> array of revisions

	Each revision in the array will look like the response from \c :docValue.

	If there are no conflicts (no resolved conflicts, in the case of \c includeDeleted),
	then this function will look like the single response from \c :docValue.

	The winning revision will always be first in the array.
]]--

function WFRxDocStoreImp:currentRevisionsForDoc( docId, includeDeleted )

	check( 'WFRxDocStoreImp', 'string', '?boolean' )

	assertNotClosed( self, 'currentRevisionsForDoc' )

	docStoreImpLog:debugf( "currentRevisionsForDoc %s %s", docId, includeDeleted )

	assertDocIdIsValid( docId )

	local db = self._db

	local revisionRows = db:withTransactionDo( function()

		return currentRevisionsForDoc_inTransaction( self, db, self._exec, docId, includeDeleted )

	end )

	local result = {}

	for i = 1, #revisionRows do

		local revisionRow = revisionRows[ i ]

		local docValue = contentFromFlattenedValue( revisionRow[ 3 ] )
		local annotation = contentFromFlattenedValue( revisionRow[ 6 ] )

		if annotation then
			for k, v in pairs( annotation ) do
				docValue[ k ] = v
			end
		end

		docValue._id = docId
		docValue._localId = revisionRow[ 7 ]
		docValue._revId = revisionRow[ 1 ]
		docValue._changeCounter = revisionRow[ 2 ]

		if revisionRow[ 5 ] == 1 then
			docValue._conflict = true
		end

		if revisionRow[ 4 ] == 1 then
			docValue._deleted = true
		end

		result[ i ] = docValue

	end

	return Observable.just( result )

end

--------------------------------------------------------------------------------

local function deleteDoc_imp( self, docId, optRevId, markNonCurrent )

	-- Extract doc ID from the value.

	assertDocIdIsValid( docId )

	local db = self._db

	local result, newRevSequence = db:withTransactionDo( function()

		local exec = self._exec

		local docValue = docValue_inTransaction( self, db, exec, docId, optRevId )

		if docValue._empty then
			return true
		end

		if docValue._deleted then
			return docValue._revId
		end

		docValue._deleted = true
		docValue._conflict = nil
		docValue._changeCounter = nil
		docValue._labels = nil
		
		local args = extractAndValidateDocValue( self, 'applyOverlay', docValue )
		
		args.markNonCurrent = markNonCurrent
		
		return setDocValueImp_inTransaction( self, db, exec,
											 'applyOverlay', args )

	end )

	if newRevSequence then
		self._changeCounter:onNext( newRevSequence )
	end

	return Observable.just( result )

end

--------------------------------------------------------------------------------
--[[!
	\brief Deletes a document or a specific revision of a document.

	\param docId (string|number) the document's unique or local ID
	\param revId (optional, string) revision to delete
		
	\return <code>Observable< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this Observable.

	This replaces the revision in question with a new value that contains
	the value <code>_deleted = true</code>. This signals that the previous
	revision is now deleted.

	Among other things, this can be used to resolve a conflict. A conflict
	exists only when multiple current revisions for a document exist. If you
	delete revisions until there is only one current revision, the conflict
	is resolved.

	The revision being deleted must be current.
]]--

function WFRxDocStoreImp:deleteDoc( docId, optRevId )

	check( 'WFRxDocStoreImp', 'string|number', '?+string' )

	assertNotClosed( self, 'deleteDoc' )

	docStoreImpLog:debugf( "deleteDoc %s %s", docId, optRevId )

	return deleteDoc_imp( self, docId, optRevId )

end

--------------------------------------------------------------------------------

--[[!
	\brief Deletes a specific revision of a document without creating the new revision as
		a current revision. This is necessary when resolving conflicts so that, once
		processed, the deletion of the conflicting revision does not create a current 
		revision (see bug NMB-2952)
		
	\param docId (string|number) the document's unique or local ID
	\param revId (string) revision to delete
	
	\return <code>Observable< string ></code> resolves with the new value's hash code
		when the write completes. You may safely ignore this Observable.
]]--
function WFRxDocStoreImp:deleteRevisionAsNonCurrent( docId, revId )
	
	check( 'WFRxDocStoreImp', 'string|number', 'string' )

	assertNotClosed( self, 'deleteRevisionAsNonCurrent' )

	return deleteDoc_imp( self, docId, revId, true )

end

--------------------------------------------------------------------------------

-- @mark revision labels

--------------------------------------------------------------------------------
--[[!
	\brief Attaches a label to a given revision of a document.

	\param docId (string) document ID
	\param label (string) label name (any non-empty string)
	\param revId (string) revision ID

	Think of a label as a movable tag that can be used later to search for
	particular document revisions. Like git tags, they are simply named aliases
	for another entity in the system (in this case document revisions).
	Unlike git tags, they apply only to a single document.

	Labeled revisions are not purged when the database is compacted.

	\return <code>Observable< empty ></code> resolves when done
]]--

function WFRxDocStoreImp:setLabel( docId, label, revId )

	check( 'WFRxDocStoreImp', '+string', '+string', '+string' )

	assertNotClosed( self, 'setLabel' )

	docStoreImpLog:debugf( "setLabel %s %s %s", docId, label, revId )

	local db = self._db

	db:withTransactionDo( function()

		local exec = self._exec

		-- Get the doc's local numeric ID.

		local docIdRow = exec( db, [[SELECT localDocId FROM docs WHERE fullDocId = ?]], docId )

		local localDocId = docIdRow and docIdRow.localDocId
		if not localDocId then
			error( string.format( "WFRxDocStoreImp:setLabel: no such document %s", docId ), 4 )
		end

		-- Get the rev's local numeric ID (change counter).

		local revIdRow = exec( db, [[SELECT sequence FROM revs WHERE revId = ? ORDER BY sequence DESC LIMIT 1]], revId )

		local revSequence = revIdRow and revIdRow.sequence
		if not revSequence then
			error( string.format( "WFRxDocStoreImp:setLabel: no such revision %s", revId ), 4 )
		end

		-- Update the label.

		exec( db, [[INSERT OR REPLACE INTO labels VALUES( ?, ?, ? )]],
			  localDocId, revSequence, label )

		-- Do we need to update an observable doc value?

			-- TODO [scouten 2015-05-13]: Unimplemented.

	end )

	return Observable.empty()

end

--------------------------------------------------------------------------------
--[[!
	\brief Get the revision for a specific label of a document.

	\param docId (string) document ID
	\param label (string) label name (any non-empty string)

	\return <code>Observable< string ></code> tagged revision
]]--

function WFRxDocStoreImp:getRevisionForLabel( docId, label )

	check( 'WFRxDocStoreImp', '+string', '+string' )

	assertNotClosed( self, 'getRevisionForLabel' )

	docStoreImpLog:debugf( "getRevisionForLabell %s %s", docId, label )

	local db = self._db

	local result = db:withTransactionDo( function()

		local exec = self._exec

		-- Get the doc's local numeric ID.

		local docIdRow = exec( db, [[SELECT localDocId FROM docs WHERE fullDocId = ?]], docId )

		local localDocId = docIdRow and docIdRow.localDocId
		if not localDocId then
			error( string.format( "WFRxDocStoreImp:getRevisionLabel: no such document %s", docId ), 4 )
		end

		-- Get the tagged revs local sequenceId.

		local revSequenceRow = exec( db, [[SELECT revSequence FROM labels WHERE localDocId = ? AND name = ? ORDER BY revSequence DESC LIMIT 1]], 
										localDocId, label )

		local revSequence = revSequenceRow and revSequenceRow.revSequence
		if not revSequence then
			error( string.format( "WFRxDocStoreImp:getRevisionLabel: no such label %s", label ), 4 )
		end

		-- Retrieve the revision.

		local revIdRow = exec( db, [[SELECT revId FROM revs WHERE sequence = ? ORDER BY revId DESC LIMIT 1]], revSequence )
		
		local revId = revIdRow and revIdRow.revId

		if not revId then
			error( string.format( "WFRxDocStoreImp:getRevisionForLabel: no such local revision %s", revSequence ), 4 )
		end
		
		return revId

	end )

	return Observable.just( result )

end

--------------------------------------------------------------------------------
--[[!
	\brief Update a specific label for a doc from a revision list.

	Updates the specified label of the specified docId to the latest revId 
	in the revIdList which also exists in the local docStore under the condition
	that the local sequencenumber is higher than the current tagged sequence number.

	\param docId (string) document ID
	\param label (string) label name (any non-empty string)
	\param revList (table) revision list.

	\return <code>Observable< empty ></code> resolves when done
]]--

function WFRxDocStoreImp:updateLabelFromRevisionList( docId, label, revIdList )

	check( 'WFRxDocStoreImp', '+string', '+string', 'table' )

	assertNotClosed( self, 'updateLabelFromRevisonList' )

	docStoreImpLog:debugf( "updateLabelFromRevisonList %s %s %s", docId, label, revIdList )

	local db = self._db

	db:withTransactionDo( function()

		local exec = self._exec

		-- Get the doc's local numeric ID.

		local docIdRow = exec( db, [[SELECT localDocId FROM docs WHERE fullDocId = ?]], docId )

		local localDocId = docIdRow and docIdRow.localDocId
		if not localDocId then
			error( string.format( "WFRxDocStoreImp:updateLabelFromRevisonList: no such document %s", docId ), 4 )
		end

		-- Get the tagged revs local sequenceId.

		local revTaggedSequenceRow = exec( db, [[SELECT revSequence FROM labels WHERE localDocId = ? ORDER BY revSequence DESC LIMIT 1]], localDocId )

		local revTaggedSequence = revTaggedSequenceRow and revTaggedSequenceRow.revSequence
		if not revTaggedSequence then
			error( string.format( "WFRxDocStoreImp:updateLabelFromRevisonList: no such label %s", label ), 4 )
		end
		
		-- Get the local revisions which have a higher sequence number than the tagged one
		
		local revSequenceRows = db:gatherArrayOfArray(
				[[SELECT sequence, revId FROM revs WHERE localDocId = ? AND sequence > ? ORDER BY sequence DESC]], 
				localDocId, revTaggedSequence )
				
		local found = false

		local toBeLabeledSequence
		local revId
		
		for i = 1, #revSequenceRows do

			local row = revSequenceRows[ i ]

			toBeLabeledSequence = row[1]
			revId = row[2]
			
			for j = #revIdList, 1, -1 do
        		if revIdList[j] == revId then
					found = true
					break
				end
			end
			
			if found == true then
				break
			end
			
		end
		
		-- If we found a local revision in the revIdList, update the label.
		if found then

			exec( db, [[INSERT OR REPLACE INTO labels VALUES( ?, ?, ? )]],
			  	localDocId, toBeLabeledSequence, label )

		end
			
	end )

	return Observable.empty()

end

--------------------------------------------------------------------------------

-- @mark partial document values

--------------------------------------------------------------------------------

local function assertValidOverlay( overlay )

	-- It's not allowed to specify parent _revId or _revIds.

	if overlay._revId ~= nil then
		error( "WFRxDocStoreImp:applyOverlay: _revId is not permitted", 2 )
	end

	if overlay._revIds ~= nil then
		error( "WFRxDocStoreImp:applyOverlay: _revIds is not permitted", 2 )
	end

	-- It's not allowed to update labels.

	if overlay._labels ~= nil then
		error( "WFRxDocStoreImp:applyOverlay: _labels is not permitted", 2 )
	end

	-- Verify that doc value is acceptable (does not contain illegal keys, etc.).

	assertDocValueIsValid( overlay )

end

--------------------------------------------------------------------------------

local function applyOverlayImp_inTransaction( self, db, docId, overlay, labels )

	local exec = self._exec

	local oldDocValue = docValue_inTransaction( self, db, exec, docId, nil, false, true )

	if not next( overlay ) then
		return oldDocValue._revId
	end
	
	if oldDocValue.type and overlay.type and oldDocValue.type ~= overlay.type then
		errorf( "Cannot change type of document via applyOverlay once set before" )
	end
	
	if oldDocValue.subtype and overlay.subtype and oldDocValue.subtype ~= overlay.subtype then
		errorf( "Cannot change subtype of document via applyOverlay once set before" )
	end

	local newDocValue = KSTableUtils.deepCopy( oldDocValue )

	WFTableMerge.applyOverlay( newDocValue, overlay )

	if KSTableUtils.deepEquals( oldDocValue, newDocValue ) then
		return oldDocValue._revId
	end

	newDocValue._empty = nil
	newDocValue._conflict = nil
	newDocValue._changeCounter = nil
	newDocValue._labels = nil

	local args = extractAndValidateDocValue( self, 'applyOverlay', newDocValue )

	if labels then
		args.labels = labels
	end
	
	return setDocValueImp_inTransaction( self, self._db, exec,
										 'applyOverlay', args )

end

--------------------------------------------------------------------------------
--[[!
	\brief Updates the document value, applying an overlay to any existing value.

	\param overlay (table) delta to apply to the existing document value

	\c overlay \b must contain the following elements:
		* \c _id (string) the document's unique ID or the document's
			integer local ID.

	\c overlay <b>MUST NOT</b> contain a \c _rev entry. This function can only
		be applied to the current winning revision of a document.

	\c overlay <b>MUST NOT</b> contain a \c _labels entry. This function can not
		be used to create or update labels.

	\return <code>Observable< string ></code> resolves with the new value's
		revision ID when the write completes. You may safely ignore this promise.
]]--

function WFRxDocStoreImp:applyOverlay( overlay )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'applyOverlay' )

	docStoreImpLog:debugf( "applyOverlay %s", overlay )

	-- Extract doc ID from the value.

	local docId = assertDocIdIsValid( overlay._id )
	overlay._id = nil

	assertValidOverlay( overlay )

	local db = self._db

	local result, newRevSequence = db:withTransactionDo( function( myDb)
		return applyOverlayImp_inTransaction( self, myDb, docId, overlay )
	end, db )

	if newRevSequence then
		self._changeCounter:onNext( newRevSequence )
	end

	return Observable.just( result )

end

--------------------------------------------------------------------------------
--[[!
	\brief Updates multiple document values, applying a different overlay to
		each existing value, if any.

	\param overlayArray (array) an array of deltas, just like those supplied to
		applyOverlay() above.
		
		The array may also contain a <code>label = "name"</code> entry. If this exists,
		that label will be applied to all documents in this batch.

	\return <code>Observable< string ></code> resolves with an array of new value's
		revision IDs when the write completes. You may safely ignore this.
]]--

function WFRxDocStoreImp:bulkApplyOverlay( overlayArray )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'applyOverlay' )

	if #overlayArray == 0 then
		return Observable.throw( "Empty overlay array" )
	end

	local db = self._db

	local label = overlayArray.label

	if label and ( type( label ) ~= 'string' or #label == 0 ) then
		error( "WFRxDocStoreImp:bulkApplyOverlay: label must be a non-empty string if provided", 2 )
	end
	
	local labels
	if label then
		labels = {}
		labels[ label ] = true
	end
	
	local result, finalRevSequence = db:withTransactionDo( function( myDb )

		local result = {}
		local latestRevSequence

		for i, overlay in ipairs( overlayArray ) do

			-- Extract doc ID from the value.

			local docId = assertDocIdIsValid( overlay._id )
			overlay._id = nil

			assertValidOverlay( overlay )

			local newRevSequence

			result[ i ], newRevSequence = applyOverlayImp_inTransaction( self, myDb, docId, overlay, labels )

			if newRevSequence then
				latestRevSequence = newRevSequence
			end

		end

		return result, latestRevSequence

	end, db )

	if finalRevSequence then
		self._changeCounter:onNext( finalRevSequence )
	end

	return Observable.just( result )

end

--------------------------------------------------------------------------------

-- @mark change notification

--------------------------------------------------------------------------------
--[[!
	\brief Tracks changes in this database by incrementing a counter for each
		change. Be aware that this value can jump, especially if many revisions
		are changed at once.

	\return <code>Subject< number ></code> the current maximum value for
		\c _changeCounter across all revisions in this database
]]--

function WFRxDocStoreImp:changeCounter()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'changeCounter' )

	return self._changeCounter

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports the current change count for this revision.
		This count includes every operation that has completed to this point.

	\return <code>Observable< number ></code> the current maximum value for
		\c _changeCounter across all revisions in this database
]]--

function WFRxDocStoreImp:changeCounterOnce()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'changeCounterOnce' )

	return self._changeCounter:take( 1 )

end

--------------------------------------------------------------------------------

local function changedDocsPrep( _, args )

	local from = args.fromChangeCounter

	if not from then
		error( "WFRxDocStoreImp:changedDocs: fromChangeCounter argument missing", 3 )
	end

	if type( from ) ~= 'number'
		or from == kInfinity
		or from == kNegativeInfinity
		or from ~= from
		or from < 0
	then
		error( string_format( "WFRxDocStoreImp:changedDocs: fromChangeCounter illegal (%s)",
							  from ), 2 )
	end

	local maxLimit = 100

	local limit = args.limit or maxLimit
	if type( limit ) ~= 'number' then
		error( "WFRxDocStoreImp:changedDocs: limit must be a number or nil for default", 2 )
	end

	if limit < 1
		or limit == kInfinity
		or limit ~= limit
	then
		error( "WFRxDocStoreImp:changedDocs: limit must be a positive number", 2 )
	end

	if limit > maxLimit then

		docStoreImpLog:warnf( "changedDocs: limit is too high (%d), reducing to %d",
							  limit, maxLimit )

		limit = maxLimit

	end

	return {
		from = from,
		limit = limit,
	}

end

--------------------------------------------------------------------------------

local function changedDocsCore_inTransaction( self, parsedArgs )

	local db = self._db

	local limit = parsedArgs.limit
	local from = parsedArgs.from

	return db:gatherArrayOfArray(
		[[SELECT fullDocId, revId, mostRecentRevSequence,
				content, revs.deleted, hasConflicts, docs.localDocId, annotation
			FROM docs
				INNER JOIN revs ON docs.winningRevSequence == revs.sequence
			WHERE mostRecentRevSequence >= ?
			ORDER BY mostRecentRevSequence
			LIMIT ? ]],
		from, limit + 1 )

end

--------------------------------------------------------------------------------

local function postProcessChangedDocsRows( self, rawDocRows, parsedArgs )

	local result = {}

	local limit = parsedArgs.limit

	for i = 1, math_min( #rawDocRows, limit ) do

		local currentRevRow = rawDocRows[ i ]

		local docValue = contentFromFlattenedValue( currentRevRow[ 4 ] )
		local annotation = contentFromFlattenedValue( currentRevRow[ 8 ] )

		for k, v in pairs( annotation ) do
			docValue[ k ] = v
		end

		docValue._id = currentRevRow[ 1 ]
		docValue._localId = currentRevRow[ 7 ]
		docValue._revId = currentRevRow[ 2 ]
		docValue._changeCounter = currentRevRow[ 3 ]

		if currentRevRow[ 5 ] == 1 then
			docValue._deleted = true
		end

		if currentRevRow[ 6 ] == 1 then
			docValue._conflict = true
		end

		result[ i ] = docValue

	end

	if rawDocRows[ limit + 1 ] then
		result.continueFrom = rawDocRows[ limit + 1 ][ 3 ]
	else
		result.continueFrom = self._changeCounter:getValue() + 1
	end

	changedDocsLog:debugf( "changedDocs is returning %s", result )

	return Observable.just( result )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the set of documents that have changed since a specified point.

	\param args (table) A table with these keys:
		* \c fromChangeCounter (number): A previous change-counter value. Documents
			whose change counter is equal to or greater than this value can match
			this query.
		* \c limit (optional, number): The maximum number of documents to return

	Note that regardless of any value specified in \c limit, a limit of 100 documents
	is always enforced. (The limit is raised to 1000 if \c idsOnly is true.)

	\return <code>Observable< table ></code> An array of fully-formed document values
		(as from \c :docValue) that satisfy the changed-documents search criteria.

	This array/table also contains a top-level element named \c continueFrom
	whose value is the change-counter number to use in the next \c :changedDocs call.
	That number may be less than the current value of \c :changeCounter if
	the number of documents returned in this call was limited.
]]--

function WFRxDocStoreImp:changedDocs( args )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'changedDocs' )

	local parsedArgs = changedDocsPrep( self, args )

	local rawDocRows = self._db:withTransactionDo( changedDocsCore_inTransaction,
													 self, parsedArgs )

	return postProcessChangedDocsRows( self, rawDocRows, parsedArgs )

end

--------------------------------------------------------------------------------
--[[!
	\brief Watches the database for document value changes and reports them
		as they happen.

	\param optFromChangeCounter (optional, number): A previous change-counter value.
		When provided, any changes >= optFromChangeCounter will be sent, even if
		they have already occurred. It is an error to specify optFromChangeCounter
		that is more than 100 behind the current change counter. If not provided,
		the observations will start from the current change counter.

	\return <code>Observable< table ></code> A series of fully-formed document
		values (as from \c :docValue) as they are saved to the database.
]]--

function WFRxDocStoreImp:_raw_changesFeed( optFromChangeCounter )

	check( 'WFRxDocStoreImp', '?number' )

	assertNotClosed( self, '_raw_changesFeed' )

	-- We use a local behavior subject to track what we've reported to date.
	-- This is either the initial starting point

	local continueFromB = Rx.BehaviorSubject( optFromChangeCounter or -1 )

	-- This first part of the Observable chain should fire only if
	-- continueFromB is behind the doc store's change counter.

	return Observable.combineLatest( function( a, b ) return a, b end,
									 continueFromB, self._changeCounter )

		:filter( function( continueFrom, changeCounter )

			-- Special (ugly?) case: If we didn't get a starting point,
			-- swap in the current change counter. Obviously, that implies
			-- we're caught up now.

			if continueFrom == -1 then
				continueFromB:onNext( changeCounter )
				return false
			end

			return continueFrom <= changeCounter
				-- or is it < ?

		end )

		-- We have new documents to report: Use :changedDocs to get that info now.

		:flatMapLatest( function( continueFrom )

			return self:changedDocs { fromChangeCounter = continueFrom }

				-- Convert the :changedDocs result array into a series of
				-- individual onNext notifications.

				:flatMapLatest( function( changedDocs )

					local result

					-- #changedDocs > 0 covers the (unlikely?) case that
					-- we get no documents back. Not sure what Observable.from
					-- does in that case.

					if #changedDocs > 0 then
						result = Observable.from( changedDocs )
					else
						result = Observable.empty()
					end

					-- Once we've sent out all of those values, bump
					-- continueFromB so we can do this all again
					-- once there is more to report (which might well be
					-- right now).

					return result:tapOnCompleted( function()
						continueFromB:onNext( changedDocs.continueFrom )
					end )

				end )

		end )

end

--------------------------------------------------------------------------------

local function docRevisionHistory_inTransaction( self, docId, optRevId )

	local db = self._db

	local exec = self._exec

	local revSequenceRow

	if optRevId then

		revSequenceRow = exec( db,
			[[SELECT sequence
				FROM revs
					INNER JOIN docs ON docs.localDocId = revs.localDocId
				WHERE
					fullDocId = ? AND revId = ?
				ORDER BY sequence DESC LIMIT 1]], docId, optRevId )

	else

		revSequenceRow = exec( db,
			[[SELECT winningRevSequence AS sequence FROM docs WHERE fullDocId = ?]], docId )

	end

	local nextRevSequence = revSequenceRow and revSequenceRow.sequence

	local historyRows = db:gatherArrayOfArray(
		[[SELECT parentRevSequence, sequence, revId, revs.deleted, content ISNULL
			FROM revs
				INNER JOIN docs ON docs.localDocId == revs.localDocId
			WHERE
				fullDocId = ?
			ORDER BY sequence DESC]], docId )

	local result = {}
	local nResults = 0

	for i = 1, #historyRows do

		local row = historyRows[ i ]

		local rowSequence = row[ 2 ]

		if rowSequence == nextRevSequence then

			local rowRevId = row[ 3 ]

			local status = 'available'

			if row[ 4 ] == 1 then
				status = 'deleted'
			end

			-- NOTE: The sequence of deleted and missing here is based on
			-- an original goal of compatibility with Apache CouchDB. A rev
			-- that is a deletion and has since been purged will be reported
			-- as "missing", not "deleted".

			if row[ 5 ] == 1 then
				status = 'missing'
			end

			nResults = nResults + 1
			result[ nResults ] = {
				revId = rowRevId,
				status = status,
			}

			nextRevSequence = row[ 1 ]

		end

	end

	return result

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns additional history about the document revisions leading to
		a specific revision.

	\param docId (string) the document's unique ID

	\param optRevId (optional, string) the revision to trace (if not specified,
		defaults to the current winning revision)

	\param optStartingFromAnyOf (optional, table) if any of the revision IDs
		specified in this array occur, stop tracing history beyond the first
		matching rev ID

	\return <code>Observable< table ></code> array of matching revision IDs in history

	An example return might be:
		<code>
			{
				"rev": "8-6f5ad8db0f34af24a6e0984cd1a6cfb9",
				"status": "available"
			},
			{
				"rev": "7-77fba3a059497f51ec99b9b478b569d2",
				"status": "deleted"
			},
			{
				"rev": "6-136813b440a00a24834f5cb1ddf5b1f1",
				"status": "available"
			},
			{
				"rev": "5-fd96acb3256302bf0dd2f32713161f2a",
				"status": "missing"
			},
			{
				"rev": "4-874985bc28906155ba0e2e0538f67b05",
				"status": "missing"
			},
			{
				"rev": "3-0de77a37463bf391d14283e626831f2e",
				"status": "missing"
			},
			{
				"rev": "2-d795d1b924777732fdea76538c558b62",
				"status": "missing"
			},
			{
				"rev": "1-917fa2381192822767f010b95b45325b",
				"status": "missing"
			}
		</code>
]]--

function WFRxDocStoreImp:docRevisionHistory( docId, optRevId, optStartingFromAnyOf )

	check( 'WFRxDocStoreImp', 'string', '?+string', '?table' )

	assertNotClosed( self, 'docRevisionHistory' )

	local result = self._db:withTransactionDo( docRevisionHistory_inTransaction,
											   self, docId, optRevId, optStartingFromAnyOf )

	return Observable.just( result )

end

--------------------------------------------------------------------------------

-- @mark local (non-syncing) documents

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the value of a single local (non-syncing) document.

	\param docId (string) the local document's unique ID

	\return <code>Observable< table ></code> value for this document

	Note that the document content will be annotated with the following
	system-generated entries (at the top level of the value table):

		* \c _empty (boolean): true if there isn't actually a value for this doc ID
]]--

function WFRxDocStoreImp:localDocValue( name )

	check( 'WFRxDocStoreImp', 'string' )

	assertNotClosed( self, 'localDocValue' )

	local localDocValueRow = self._exec( self._db,
										 [[SELECT content FROM localDocs WHERE name == ?]],
										 name )

	local docValue = localDocValueRow and contentFromFlattenedValue( localDocValueRow.content )

	if docValue then
		docValue._id = name
	end
	
	return Observable.just( docValue or { _id = name, _empty = true } )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the value for a local (non-syncing) document, replacing any
		existing value.

	Local documents are intended for use as state bookmarks. They differ from
	traditional syncing documents in the following ways:

		* Local document IDs are stored in a separate table and a separate
		  namespace from traditional syncing documents. You may reuse the same
		  document ID among the document types and the values will not cross-pollute.
		* Local documents are not versioned. (For that reason, the \c _rev
		  parameter must not be specified in a local document value.)
		* Local documents can not have conflicts.
		* There is no change notification for local documents.
		* Local documents are not replicated to other document databases.

	\param value (table) new value for the entire document

	\c value \b must contain the following elements:
		* \c _id (string) the document's unique ID

	\return <code>Observable< empty ></code> Resolves when the write completes.
		(You may safely ignore this promise.)
]]--

function WFRxDocStoreImp:setLocalDocValue( docValue )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'setLocalDocValue' )

	local localDocId = assertDocIdIsValid( docValue._id )
	docValue._id = nil

	assertDocValueIsValid( docValue )

	if docValue._revId then
		error( "WFRxDocStoreImp:setLocalDocValue does not accept _revId", 2 )
	end

	if docValue._deleted then
		error( "WFRxDocStoreImp:setLocalDocValue does not accept _deleted", 2 )
	end

	local serializedDocValue = tableToString( docValue )

	self._exec( self._db,
				[[INSERT OR REPLACE INTO localDocs VALUES( ?, ? )]],
				localDocId,
				serializedDocValue )

	-- Do we need to update an observable doc value

	local ongoingLocalDocValues = self._ongoingLocalDocValues

	if ongoingLocalDocValues:hasObservableForKey( localDocId ) then

		local ongoingLocalDocValue = ongoingLocalDocValues:observableForKey( localDocId )

		if docValue then
			docValue._id = localDocId
		end
		
		ongoingLocalDocValue:onNext( docValue )

	end

	return Observable.empty()

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:applyOverlayLocalDocValue( overlay )

	check( 'WFRxDocStoreImp', 'table' )

	assertNotClosed( self, 'applyOverlayLocalDocValue' )

	local localDocId = assertDocIdIsValid( overlay._id )
	overlay._id = nil

	assertDocValueIsValid( overlay )

	if overlay._revId then
		error( "WFRxDocStoreImp:applyOverlayLocalDocValue does not accept _revId", 2 )
	end

	if overlay._deleted then
		error( "WFRxDocStoreImp:applyOverlayLocalDocValue does not accept _deleted", 2 )
	end
	
	local localDocValueRow = self._exec( self._db,
										 [[SELECT content FROM localDocs WHERE name == ?]],
										 localDocId )

	local docValue = localDocValueRow and contentFromFlattenedValue( localDocValueRow.content ) or {}
	
	WFTableMerge.applyOverlay( docValue, overlay )
	
	docValue._id = localDocId
	
	return self:setLocalDocValue( docValue )
	
end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

local performIntegrityCheck -- forward declaration

--------------------------------------------------------------------------------

local function getCountQueryResult( db, sql )

	local rows = db:gatherArrayOfArray( sql )

	return rows and rows[ 1 ] and rows[ 1 ][ 1 ]

end

--------------------------------------------------------------------------------

local function computeAveSize( db, sql )

	local rows = db:gatherArrayOfArray( sql )

	if #rows == 0 then

		return

	end

	return string.format( " %s (%s records)", unpack( rows[ 1 ] ) )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sends diagnostic data on database performance to the log.
]]--

function WFRxDocStoreImp:sendDiagnosticLogData()

	check( 'WFRxDocStoreImp' )

	-- Send information about doc store performance.
	local aveAssetRevContentSize, aveAssetDocAnnotationSize
	if WFConfiguration.DocStore.DiagnosticLogDocStats then
		
		aveAssetRevContentSize = computeAveSize(
											self._db,
											[[SELECT  SUM( LENGTH ( content ) ) / COUNT( * ), COUNT(  * )
												FROM revs INNER JOIN docs 
														ON docs.winningRevSequence = revs.sequence
												WHERE docs.deleted != 1 AND content LIKE '%type = "asset"%';]]
											)
	
		aveAssetDocAnnotationSize = computeAveSize(
											self._db,
											[[SELECT  SUM( LENGTH ( annotation ) ) / COUNT( * ), COUNT(  * )
												FROM revs INNER JOIN docs 
														ON docs.winningRevSequence = revs.sequence
												WHERE docs.deleted != 1 AND content LIKE '%type = "asset"%';]]
											)
	end
			
	return {

		name = "WFRxDocStoreImp :: " .. self._path,
		id = 'WFRxDocStoreImp-' .. self._path,
		weight = 1600,

		body = {

			nDocs = getCountQueryResult( self._db,
										 "SELECT count() FROM docs" ),

			nRevs = getCountQueryResult( self._db,
										 "SELECT count() FROM revs" ),

			aveAssetRevContentSize = aveAssetRevContentSize, 
			aveAssetDocAnnotationSize = aveAssetDocAnnotationSize,
			nTransactions = self._db:nTransactions(),

			integrityCheck = performIntegrityCheck( self ),

		},

	}

end

--------------------------------------------------------------------------------

-- @mark miscellaneous

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves the path to this document store.

	\return (string) The catalog's path.
]]--

function WFRxDocStoreImp:path()

	check( 'WFRxDocStoreImp' )

	return self._path

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves this document store as a string.

	\return (string) The string representation.
]]--

function WFRxDocStoreImp:__tostring()

	check( 'WFRxDocStoreImp' )

	return string_format( "WFRxDocStoreImp( %s )", self._path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrieves a stream of deleted doc values.

	\return <code>Observable< table ></code> Fires with doc values as documents are
		deleted.
]]--

function WFRxDocStoreImp:deletedDocStream( optFieldSpec )

	check( 'WFRxDocStoreImp' )
	
	if optFieldSpec then
		return self._deletedDocStream
			:map( function( docValue )
				return onlyFieldsInSpec( optFieldSpec, docValue )
			end )
	else
		return self._deletedDocStream
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Free disk space used to describe old document revisions.

	\warning This function should be used cautiously as it may take significant
		time. All other doc store features will be queued up and will not
		complete until the compaction has completed.

	\return <code>Observable< empty ></code> Resolves when compaction has completed.
		(You may safely ignore this Observable.)
]]--

function WFRxDocStoreImp:compact()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'compact' )

	compactionLog:info( "Deleting content from old revisions" )

	self._exec( self._db,
					[[UPDATE revs SET content = NULL
						WHERE current == 0
							AND sequence NOT IN ( SELECT revSequence FROM labels )]] )

	local path = self._path
	if path and path ~= ':temp:' and path ~= ':memory:' and path ~= '' then

		-- Close and reopen the database.

		maybeCloseAndClearDatabase( self )

		local db = openAndObtainDatabase( self, path )

		compactionLog:info( "Flushing SQLite WAL" )

		db:exec( [[PRAGMA wal_checkpoint( RESTART )]] )

		compactionLog:info( "Vaccuming SQLite database" )

		db:exec( [[VACUUM]] )

	end

	return Observable.empty()

end

--------------------------------------------------------------------------------
--[[!
	\brief Close the underlying sqlite db

	\warning This function should only be used at shutdown because all other
		docStore features that require an open db will fail. This applies to
		all threads that have a docStoreImp instance

	\return <code>Observable< empty ></code> Resolves when close has completed.
		Caller must subscribe to this Observable or closeDb call might not run.
]]--

function WFRxDocStoreImp:closeDb()

	check( 'WFRxDocStoreImp' )

	maybeCloseAndClearDatabase( self )

	return Observable.empty()

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for debugging use only) Generates code coverage statistics
		for the worker thread that implemented this doc store.

	\warning This function should only be used by WFRxDocStoreTest.lua.

	\return <code>Observable< empty ></code> Resolves when compaction has completed.
		generated. (You may safely ignore this Observable.)
]]--

WFRxDocStoreImp.DEBUG_dumpCoverage = function( _, path )

	check( 'WFRxDocStoreImp', '+string' )

	require 'SetupCoverage'.dumpCoverage( path )

	return Observable.empty()

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for performance testing only) Enables a mode where the doc
		store will collect performance statistics on its SQLite database.

	\return <code>Observable< empty ></code> Resolves when compaction has completed.
		been enabled. (You may safely ignore this Observable.)
]]--

function WFRxDocStoreImp:PERF_collectQueryPerformanceStats()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'PERF_collectQueryPerformanceStats' )

	self._db:PERF_collectQueryPerformanceStats()

	return Observable.empty()

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for performance testing only) Generates a string suitable
		for logging to a text file or window describing performance results.

	\return <code>Observable< string ></code> query statistics report
]]--

function WFRxDocStoreImp:PERF_generateQueryStatsReport()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'PERF_generateQueryStatsReport' )

	return Observable.just( self._db:PERF_generateQueryStatsReport() )

end

--------------------------------------------------------------------------------

--[[
local function makeIntegrityCheckFunction( expectedEmptyCountQuery, messageFormat )

	expectedEmptyCountQuery = "SELECT count(*) FROM " .. expectedEmptyCountQuery

	return function( db, exec )

		local _, count = next( exec( db, expectedEmptyCountQuery ) )

		if count == 0 then
			return true
		else
			return false, string.format( messageFormat, count )
		end

	end

end
]]--

--------------------------------------------------------------------------------

local integrityCheckers = {

	-- TODO [scouten 2015-05-06]: Update with current integrity checks.
	-- We're not using hash any more, so these are not relevant.

--	makeIntegrityCheckFunction( [[bodyContent WHERE bodyContentHash NOT
--		IN ( SELECT bodyContentHash FROM documents )]], "orphaned bodyContent: %d" ),
--
--	makeIntegrityCheckFunction( [[annotationContent WHERE annotationContentHash NOT
--		IN (SELECT annotationContentHash FROM documents)]], "orphaned annotationContent: %d" ),
--
--	makeIntegrityCheckFunction( [[documents WHERE bodyContentHash NOT
--		IN (SELECT bodyContentHash FROM bodyContent)]], "dangling bodyContent: %d" ),
--
--	makeIntegrityCheckFunction( [[documents WHERE annotationContentHash NOT
--		IN (SELECT annotationContentHash FROM annotationContent)]], "dangling annotationContent: %d" ),

}

--------------------------------------------------------------------------------

performIntegrityCheck = function( self )

	local failures = {}

	for _, checker in ipairs( integrityCheckers ) do

		local result, message = checker( self._db, self._exec )

		if not result then
			failures[ #failures + 1 ] = message
		end

	end

	return failures

end

--------------------------------------------------------------------------------
--[[!
	\brief (Intended for unit test use only) Perform an integrity check
		of the doc store and report errors.

	\warning This function should only be used by unit tests.

	\return <code>Observable</code> signals onCompleted if successful;
		signals onError with an array of failure messages if any problems are found
]]--

function WFRxDocStoreImp:performIntegrityCheck()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'performIntegrityCheck' )

	local result = performIntegrityCheck( self )

	if #result == 0 then
		return Observable.empty()
	else
		return Observable.throw( result )
	end

end

--------------------------------------------------------------------------------

function WFRxDocStoreImp:halt()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'halt' )
	
	self:close() -- should cause DB object and friends to close
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Permanently shuts down this doc store and deletes all related files.

	\return <code>Observable< empty ></code> Resolves when done.
		(You may safely ignore this Observable.)
]]--

function WFRxDocStoreImp:haltAndDelete()

	check( 'WFRxDocStoreImp' )

	assertNotClosed( self, 'haltAndDelete' )

	local path = self._path

	self:close() -- should cause DB object and friends to close

	if path and #path > 0 then
		WFFileUtils.delete( path )
		WFFileUtils.delete( path .. '-shm' )
		WFFileUtils.delete( path .. '-wal' )
	end

	return Observable.empty()

end

--------------------------------------------------------------------------------

local _perfWarningThreshold = 1 / 10 -- seconds

--------------------------------------------------------------------------------

-- SPAM [jandrew 2017-05-24]: The following metrics are hidden behind config flags. 
-- They're usually too noisey and result in very large stamina logs.

--------------------------------------------------------------------------------

local _apisThatRequireExplicitStaminaWhitelist = {
		'changedDocs', 
		'changeCounter' }

--------------------------------------------------------------------------------

local function isAPIBlacklistedFromStamina( apiName )

	if not _apisThatRequireExplicitStaminaWhitelist[ apiName ] then

		return false

	end

	-- the API must have a config.lua entry in order to be enabled

	if WFConfiguration.Stamina[ string.format( "%sMetrics", apiName ) ] then

		return false

	end

	return true

end

--------------------------------------------------------------------------------

local function monitorPerformanceImp( apiName, descFunc )

	-- Logger Warnings

	WFRxDocStoreImp[ apiName ] =
		WFPerformanceUtils_wrapInLogLongExecutionInDiagnosticLogWithDynamicDescription(
			"DocStore", _perfWarningThreshold, apiName, descFunc,
				WFRxDocStoreImp[ apiName ] )

	if isAPIBlacklistedFromStamina( apiName ) then

		return WFRxDocStoreImp[ apiName ]

	end

	-- Stamina Monitoring

	return WFPerformanceReport.wrapObservableFunction( 
				'WFRxDocStoreImp:' .. apiName, 
				WFRxDocStoreImp[ apiName ] )

end 

--------------------------------------------------------------------------------

local function justReturn( v )

	return v

end

--------------------------------------------------------------------------------

local function monitorPerformance1( apiName )

	return monitorPerformanceImp( apiName, justReturn )

end

--------------------------------------------------------------------------------

local function makeArrayCountDesc( desc, _, array )

	return string.format( "%s( %s )", 
							desc, 
							type( array ) == "table" and 
								#array or 
								"not an array!" )

end

--------------------------------------------------------------------------------

local function monitorPerformanceN( apiName )

	return monitorPerformanceImp( apiName, makeArrayCountDesc )

end

--------------------------------------------------------------------------------

WFRxDocStoreImp.localDocValue = monitorPerformance1( 'localDocValue' )

WFRxDocStoreImp.applyOverlayLocalDocValue = monitorPerformance1( 'applyOverlayLocalDocValue' )

WFRxDocStoreImp.setLocalDocValue = monitorPerformance1( 'setLocalDocValue' )

WFRxDocStoreImp.changedDocs = monitorPerformance1( 'changedDocs' )

WFRxDocStoreImp.changeCounter = monitorPerformance1( 'changeCounter' )

WFRxDocStoreImp.bulkApplyOverlay = monitorPerformance1( 'bulkApplyOverlay' )

WFRxDocStoreImp.applyOverlay = monitorPerformance1( 'applyOverlay' )

WFRxDocStoreImp.deleteDoc = monitorPerformance1( 'deleteDoc' )

WFRxDocStoreImp.deleteRevisionAsNonCurrent = monitorPerformance1( 'deleteRevisionAsNonCurrent' )

WFRxDocStoreImp.bulkSetDocValue = monitorPerformanceN( 'bulkSetDocValue' )

WFRxDocStoreImp.setDocValue = monitorPerformance1( 'setDocValue' )

WFRxDocStoreImp.bulkDocValue = monitorPerformanceN( 'bulkDocValue' )

WFRxDocStoreImp.docValue = monitorPerformance1( 'docValue' )
	
--------------------------------------------------------------------------------

return Rx.Class 'WFRxDocStoreImp'( WFRxDocStoreImp )
