--[[----------------------------------------------------------------------------

WFSyncAgent.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'
local WFDiagnosticStatAccumulator = require 'WFDiagnosticStatAccumulator'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua.threading
local WFTimer = require 'WFTimer'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local logger = WFLogger 'WFSyncAgent :: init'
--	:setLevel 'debug'


--============================================================================--

-- @mark WFSyncAgent class

--------------------------------------------------------------------------------
--[[!
	\class WFSyncAgent
	
	\brief A sync agent brings together one doc store and one external service.
	It provides the code for processing updates from the external service
	and determining what updates to send to that service.
	
	An "external service" could be any number of things, such as:
	
		* Oz server
		* another catalog within the host app (for instance, .lrcat file)
		* another instance of the host app on another device, though that use
			case is not tested or supported today
	
	See \c WFTestSyncPeer class as an example implementation of "external
	service" that is used for unit testing this sync agent code.
]]--

local syncAgentMethods = {

	_docStore = false,
		-- (WFDocStore) this client's doc store
	
	_docStoreChangeCounterB = false,
		-- (B< number >) docStore's change counter
	
	_syncPeer = false,
		-- (object) external sync peer
	
	_name = false,
		-- (string) name of external service (used to construct doc store slot names)

	_syncStatusDocName = false,
		-- (string) name of sync status doc to use

	_inboxSlotName = false,
		-- (string) name of doc store slot we use for incoming docs

	_outboxSlotName = false,
		-- (string) name of doc store slot we use for outgoing docs

	_transientBlackoutSlotName = false,
		-- (string) name of doc store slot we use for recording transient blackouts

	_resolvedSlotName = false,
		-- (string) name of doc store slot we use for resolved doc values
		-- "resolved" = the last point where local last successfully merged
		-- remote's values
	
	_getChangesP = false,
		-- (KSPromise) receive request in progress (if any)

	_sendUpdatesP = false,
		-- (KSPromise) send request in progress (if any)

	_syncUntilDoneP = false,
		-- (KSPromise) sync until done request in progress (if any)

	_hasReconciledInbox = false,
		-- (boolean) true if we've called reconcileInboxP at least once

	_hasProcessedOutbox = false,
		-- (boolean) true if we've called processOutboxP at least once

	_syncActivityLog = false,
		-- (WFLogger) activity log for sync traffic; WARNING: VERY HIGH TRAFFIC!
	
	_conflictDiagnosticLog = false,
		-- (table) holds conflict information for the diagnostic log
	
	_isDoingAutomaticSync = false,
		-- (boolean) true doing automatic sync
	
	_shouldStopAutomaticSync = false,
		-- (boolean) set to true when automatic sync should be stopped

	_isSyncingAutomaticallyB = false,
		-- (B< boolean | 'pendingStop' >) true if automatic sync is engaged
	
	_killSwitchB = false,
		-- (B< false | string >) if string, sync peer says to shut down permanently
			-- string will be reason for shut down
	
	_activeSyncPumpE = false,
		-- (E< whatever >) true if we have a sync pump engaged
	
	_triggerSyncE = false,
		-- (B< boolean >) sync trigger (either from sync peer or local doc store)
	
	_triggerSyncCount = 0,
		-- (number) how many times has _triggerSyncE fired?
	
	_triggerPushCount = 0,
		-- (number) how many times has docChangeCounterB fired?
	
	_lastResponseWasErrorB = false,
		-- (B< boolean >) true if most recent sync cycle ended in error

	_minSyncDelayIntervalB = false,
		-- (B< number >) minimum # seconds to wait between sync cycles

	_maxSyncDelayIntervalB = false,
		-- (B< number >) maximum # seconds to wait between sync cycles
	
	_lastResponseWasErrorB = false,
		-- (B< boolean >) true if most recent sync cycle ended in error

	_attachedObjects = false,
		-- (table) objects that have been attached to this sync agent
	
	_recentSyncFailures = false,
		-- (table) array of recent sync failures
	
	_didClearInitialSyncFlag = false,
		-- (boolean) true if we have cleared the initial sync flag at least once
	
	_nChangesReceived = false,
		-- (WFDiagnosticStatAccumulator) how many changed documents have we received per cycle?
	
	_nChangesSent = false,
		-- (WFDiagnosticStatAccumulator) how many changed documents have we sent per cycle?

	_receivedDocumentsSummary = false,
		-- (table) summary of each batch of documents received from sync peer

	_previousReceivedTimestamp = false,
		-- (number) time of previous received batch
	
	_getChangesCycleTime = false,
		-- (WFDiagnosticStatAccumulator) how long does it take to run the get changes cycle?

	_sendUpdatesCycleTime = false,
		-- (WFDiagnosticStatAccumulator) how long does it take to run the get changes cycle?
	
	_syncIdleCycleTime = false,
		-- (WFDiagnosticStatAccumulator) how long do we spend idle between sync cycles?

	WFCloseSupport.addCloseSupportMethods,
	
	KSReactive.addReactiveSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Constructs the sync agent.
	
	\param docStore (\c WFDocStore) local doc store to be kept in sync with external service
	\param syncPeer (\c object) wrapper that can communicate with external service
	\param name (\c string) name of external service
]]--

function syncAgentMethods:__init( docStore, syncPeer, name )

	check( 'WFSyncAgent', 'WFDocStore', 'table', '+string' )

	WFCloseSupport.mustClose( self )

	logger:infof( "Creating sync agent named %s in thread %s\n\nvia %s\n\n",
							    name, ':thread:', ':traceback:' )

	self._docStore = docStore
	docStore:addCloseTarget( self )
	
	local changeCounterB
	if syncPeer:hasMethod( 'docStoreChangeCounter' ) then
		changeCounterB = syncPeer:docStoreChangeCounter( docStore )
	else
		changeCounterB = docStore:docChangeCounterB()
	end
	
	self:setOwned { _docStoreChangeCounterB = changeCounterB }

	self:setOwned { _syncPeer = syncPeer }

	self._nChangesReceived = WFDiagnosticStatAccumulator()
	self._nChangesSent = WFDiagnosticStatAccumulator()
	self._getChangesCycleTime = WFDiagnosticStatAccumulator()
	self._sendUpdatesCycleTime = WFDiagnosticStatAccumulator()
	self._syncIdleCycleTime = WFDiagnosticStatAccumulator()
	
	self._recentSyncFailures = {}
	self._receivedDocumentsSummary = {}
	
	local syncPeerTrigger = syncPeer:hasMethod( 'triggerSyncE' )
							and syncPeer:triggerSyncE()
	
	if syncPeerTrigger then

		self:setOwned {
			_triggerSyncE = syncPeerTrigger:calmE( 2 )
		}
	
		self._triggerSyncE:ownedOnDataE( self, function()
			self._triggerSyncCount = self._triggerSyncCount + 1
		end )

	else

		self:setOwned {
			_triggerSyncE = KSReactive.zeroE()
		}

	end
	
	self._docStoreChangeCounterB:ownedOnDataB( self, function()
		self._triggerPushCount = self._triggerPushCount + 1
	end )
	
	self:createReactiveVariable( '_lastResponseWasErrorB', false )
	
	if syncPeer:hasMethod( 'minSyncDelayIntervalB' ) then

		self:setOwned { _minSyncDelayIntervalB = syncPeer:minSyncDelayIntervalB() }
			
		assert( KSReactive.isBehavior( self._minSyncDelayIntervalB ) )

	else

		-- If no min delay interval is provided, use arbitrary defaults.
		
		self:setOwned {
			_minSyncDelayIntervalB = self._lastResponseWasErrorB
				:mapB( function( lastResponseWasError )
					return lastResponseWasError and 15 or 2
				end )
		}
	
	end

	if syncPeer:hasMethod( 'maxSyncDelayIntervalB' ) then

		self:setOwned { _maxSyncDelayIntervalB = syncPeer:maxSyncDelayIntervalB() }
			
		assert( KSReactive.isBehavior( self._maxSyncDelayIntervalB ) )

	else

		-- If no max delay interval is provided, use arbitrary defaults.
		
		self:setOwned {
			_maxSyncDelayIntervalB = self._lastResponseWasErrorB
				:mapB( function( lastResponseWasError )
					return lastResponseWasError and 60 or 10
				end )
		}
	
	end
		
	self._killSwitchB = KSReactive.variableB( false )
	self:addCloseTarget( self._killSwitchB )
		-- TODO [scouten 2014-02-20]: Remember to switch this to
		-- createReactiveVariable when merging with latest WF.
	
	self._name = name
	
	self._syncStatusDocName = string.format( 'syncPeer-%s', name )
	self._inboxSlotName = string.format( '%s-inbox', name )
	self._outboxSlotName = string.format( '%s-outbox', name )
	self._transientBlackoutSlotName = string.format( '%s-transientBlackout', name )
	self._resolvedSlotName = string.format( '%s-resolved', name )

	self._attachedObjects = {}

	self._syncActivityLog = WFLogger( string.format( 'WFSyncAgent :: %s :: activity',
													 name ) )
--		:setLevel 'debug'
			-- WARNING: VERY HIGH TRAFFIC if you set this to DEBUG level!!!

--		:setLevel 'info'
			-- Use this level to observe interactions with the sync peer.

	self._conflictDiagnosticLog = {}
	
	self._isSyncingAutomaticallyB = KSReactive.variableB( false )

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

function syncAgentMethods:close()

	check( 'WFSyncAgent' )

	self:stopAutomaticSync()

end

--------------------------------------------------------------------------------

function syncAgentMethods:setLogLevel( level )

	check( 'WFSyncAgent' )
	
	self._syncActivityLog:setLevel( level )

end

--------------------------------------------------------------------------------
--[[!
	\brief Returns the sync peer attached to this sync agent.
]]--

function syncAgentMethods:syncPeer()

	check( 'WFSyncAgent' )

	return self._syncPeer

end

--------------------------------------------------------------------------------
--[[!
	\brief Attach a helper object to this sync agent.
	
	Attached objects become close targets of this sync agent and will also be
	strongly referenced from it. Use this for auxiliary sync helpers related to
	this sync agent.
	
	\param object (\c WFCloseTarget) any valid close target object
]]--

function syncAgentMethods:attachHelperObject( object )

	check( 'WFSyncAgent', 'table' )

	self:addCloseTarget( object )

	self._attachedObjects[ object ] = true

end

--------------------------------------------------------------------------------
--[[!
	\brief Remove a helper object from this sync agent.
	
	Remove an object previously added via \c :attachHelperObject.
	
	\param object (\c WFCloseTarget) any valid close target object
]]--

function syncAgentMethods:removeHelperObject( object )

	check( 'WFSyncAgent', 'table' )

	object:close()

	self._attachedObjects[ object ] = nil
	
	-- We don't need to remove it as a close target since close targets
	-- won't do anything if they have already been closed.

end

--------------------------------------------------------------------------------

local function killSwitchDeployed( self, reason )

	if type( reason ) ~= 'string' or #reason == 0 then
		reason = "no reason provided"
	end
	
	self._killSwitchB:set( reason )
	self:stopAutomaticSync()

end

--------------------------------------------------------------------------------

local function recordSyncFailure( self, failRecord )

	failRecord.time = KSDate.timeInIso8601Format( nil, true )
	
	table.insert( self._recentSyncFailures, 1, failRecord )
	
	for i = 21, #self._recentSyncFailures do
		self._recentSyncFailures[ i ] = nil
	end

end

--------------------------------------------------------------------------------

-- @mark :getChangesP

--------------------------------------------------------------------------------

local function reconcileInboxP( self )
	
	self._hasReconciledInbox = true
	
	return self._docStore:doComplexLogicInTransactionP(
		'WFSyncAgentInboxHandler.reconcileInbox',
		{
			syncPeerName = self._name,
			inboxSlotName = self._inboxSlotName,
			resolvedSlotName = self._resolvedSlotName,
		} )
	
end

--------------------------------------------------------------------------------

local approvedKeys = {
	_deleted = true,
	_serverProvidedMetadata = true,
}

--------------------------------------------------------------------------------

local function getChangesP_imp( self )

	-- Where were we before?
	
	local since
	
	local resultP = self._docStore:docValueP( self._syncStatusDocName, 'wfSyncAgent' )

		-- Extract the last changed marker from previous status (if any).

		:mapDataP( function( _, docValue )
		
			local since = docValue and docValue.continueFrom
		
			if since == nil then
			
				return self._docStore:applyOverlayP
					{
						_id = 'isDoingInitialSync',
						_slot = 'local',
						syncPeer = {
							[ self._name ] = true,
						},
					}
					:mapDataP( function()
						return 'data', since
					end )
			
			else
			
				return 'data', since
				
			end
		
		end )

		-- Ask external service for updates since that marker.
		
		:mapDataP( function( _, _since )
		
			if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				return 'error', "Closed"
			end
		
			since = _since
			return self._syncPeer:changedDocsP( since )
		
		end )
		:observeP( function( status, updates )
		
			if status ~= 'data' then
			
				self._syncActivityLog:infof( "syncPeer:changedDocsP( %s ) resolved with non data status %s\n\nmessage = %s\n\n",
											 since, status, tostring( updates ) )
				
				recordSyncFailure( self,
					{
						what = 'get',
						since = since,
						status = status,
						message = updates,
					} )

			elseif #updates == 0 then

				self._syncActivityLog:debugf( "syncPeer:changedDocsP( %s ) resolved with status %s\n\nupdates = %s\n\n",
											  since, status, updates )

			else
			
				self._nChangesReceived:add( #updates )

				self._syncActivityLog:infof( "syncPeer:changedDocsP( %s ) resolved with status %s\n\nupdates = %s\n\n",
											 since, status, updates )

			end
			
		end )

		-- Process those updates.
		
		:mapDataP( function( _, updates )
		
			if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
				return 'error', "Closed"
			end
		
			-- Shared error function for misbehaving sync peer code.
			
			local function gripe( message, ... )
				
				local fullMessage = string.format( "sync peer %s: changedDocsP " .. message,
												   self._name, ... )
				
				self._syncActivityLog:errorf( "%s", fullMessage )
				error( fullMessage, 2 )

			end
		
			-- Make sure sync peer is behaving properly.
		
			if type( updates ) ~= 'table' then
				gripe "did not return a table"
			end
			
			-- Check for kill switch.
			
			if updates.killSwitch then
				killSwitchDeployed( self, updates.killSwitch )
				return 'error', "Kill switch: " .. tostring( updates.killSwitch )
			end
			
			-- Check response contents.
			
			local cfType = type( updates.continueFrom )
			if cfType ~= 'string' and cfType ~= 'number' then
				gripe "did not contain a valid continueFrom entry (must be number or string)"
			end
			
			for i = 1, #updates do
			
				local updateItem = updates[ i ]
				
				if type( updateItem ) ~= 'table' then
					gripe( "item %d in updates response was not a table", i )
				end
				
				if type( updateItem._id ) ~= 'string' or #updateItem._id == 0 then
					gripe( "item %d in updates response: _id must be non-empty string",
						   i )
				end
				
				if updateItem._id == 'pending' then
					gripe( "item %d in updates response: illegal _id 'pending'", i )
				end
				
				local timestampType = type( updateItem._timestamp )
				if timestampType ~= 'string' and timestampType ~= 'number' then
					gripe( "item %d in updates response: _timestamp must be string or number",
						   i )
				end
			
			end
		
			if #updates > 0 then

				-- Write all of the received document values into inbox
				-- without interpretation.
				
				local inboxDocValues = {}
				
				local timeOfThisBatch = KSDate.timeNowForAnimation()
				local timeSincePreviousBatch

				if self._previousReceivedTimestamp then
					timeSincePreviousBatch = math.floor( ( timeOfThisBatch - self._previousReceivedTimestamp ) * 10 ) / 10
				end
				
				self._previousReceivedTimestamp = timeOfThisBatch
				
				local receivedDocumentsBatch = {
					time = KSDate.timeInIso8601Format( nil, true ),
					timeSincePreviousBatch = timeSincePreviousBatch,
					totalDocsReceived = #updates,
					types = {},
				}
				
				local summaryByType = receivedDocumentsBatch.types
				
				for i = 1, #updates do

					local updateItem = updates[ i ]

					local inboxDocValue = {
						_id = updateItem._id,
						_slot = self._inboxSlotName,
						_syncStatus = {
							[ self._name ] = {
								timestamp = updateItem._timestamp,
								deleted = updateItem._deleted or nil,
							},
						},
						_serverProvidedMetadata = updateItem._serverProvidedMetadata
					}
					
					local docType = tostring( updateItem.type or '(unknown)' )
					summaryByType[ docType ] = ( summaryByType[ docType ] or 0 ) + 1
					
					for k, v in pairs( updateItem ) do
						if ( not k:match( "^_" ) ) or approvedKeys[ k ] then
							inboxDocValue[ k ] = v
						end
					end
					
					inboxDocValues[ i ] = inboxDocValue
				
				end
				
				self._receivedDocumentsSummary[ #self._receivedDocumentsSummary + 1 ] = receivedDocumentsBatch
				
				if #self._receivedDocumentsSummary > 1000 then
					table.remove( self._receivedDocumentsSummary, 1 )
				end
				
				return self._docStore:bulkSetDocValueP( inboxDocValues )

				-- Then remember that we have received these values
				-- so we don't ask remote side for them again.

				:mapDataP( function()

					return self._docStore:applyOverlayP {
						_id = self._syncStatusDocName,
						_slot = 'wfSyncAgent',
						continueFrom = updates.continueFrom,
					}
				
				end )
				
				-- And then attempt to merge these values with what
				-- we have locally.
			
				:mapDataP( function()
				
					return reconcileInboxP( self )
				
				end )

			else
			
				-- Remember that we got nothing. This is our signal that we
				-- completely caught up (at least once) with the remote content.
				-- Do this only once so we don't create lots of doc-store
				-- traffic when the doc store is idle.

				local clearInitialSyncFlagP

				if self._didClearInitialSyncFlag then

					clearInitialSyncFlagP = KSPromises.resolvedPromiseP( 'data', true )

				else

					-- First time through (this app session): Set doing initial sync
					-- flag to false and also record the continueFrom value so we don't
					-- run the risk of leaving continueFrom as nil (which would cause
					-- us to set isDoingInitialSync to true again).

					self._didClearInitialSyncFlag = true

					clearInitialSyncFlagP = self._docStore:applyOverlayP
						{
							_id = 'isDoingInitialSync',
							_slot = 'local',
							syncPeer = {
								[ self._name ] = false,
							},
						}
						:mapDataP( function()
		
							return self._docStore:applyOverlayP {
								_id = self._syncStatusDocName,
								_slot = 'wfSyncAgent',
								continueFrom = updates.continueFrom,
							}
						
						end )
				

				end
				
				return clearInitialSyncFlagP
					:mapDataP( function()
			
						-- We still run the reconcile inbox step at least once
						-- just in case we had received documents in a previous
						-- session that we failed to process for some reason.
					
						if self._hasReconciledInbox and not updates.forceReconcile then
		
							return 'data', false
		
						else
		
							return reconcileInboxP( self )
								:mapDataP( function()
									return 'data', false
								end )
							
						end
					
					end )
			
			end
		
		end )
		
	-- Make sure we only have one receive request in flight at a time.
	-- Also make sure we see this request through until completion.
	
	local startTime = KSDate.timeNowForAnimation()
	
	self._getChangesP = resultP:observeP( function()
	
		self._syncActivityLog:debug( "getChangesP finished" )

		self._getChangesP = false
		
		self._getChangesCycleTime:add( KSDate.timeNowForAnimation() - startTime )
	
	end )
	
	self._getChangesP:whenResolvedDo( function() end )
		:attachTo( self )
		:suppressCloseLogging()

	self._syncActivityLog:debug( "getChangesP starting" )

	return self._getChangesP

end

--------------------------------------------------------------------------------
--[[!
	\brief Request documents from external source and try to process them through
		to "main" slot.
	
	\return P< boolean > resolves to true if any new document values were received;
		false if nothing new from external service
]]--

function syncAgentMethods:getChangesP()

	check( 'WFSyncAgent' )

	-- Don't do this more than once at a time.
	
	if self._getChangesP then
		return self._getChangesP
	end
	
	-- Also serialize get and send requests. Too confusing to do both at once.
	
	if self._sendUpdatesP then
		return self._sendUpdatesP
			:mapP( function()
				return getChangesP_imp( self )
			end )
			:runUntilDoneP()
	else
		return getChangesP_imp( self )
	end
		
end


--------------------------------------------------------------------------------

-- @mark :sendUpdatesP

--------------------------------------------------------------------------------

local function sendOutboundChangesP( self, doAddChangedDocsToOutbox )

	local docStore = self._docStore

	-- Find out how many documents we're allowed to send at once.
	
	local limit = self._syncPeer:maxUpdatesPerSend()

	if type( limit ) ~= 'number' or limit < 1 or math.floor( limit ) ~= limit then

		return KSPromises.resolvedPromiseP( 'error',
			"WFSyncAgent: sync peer returned invalid response to :maxUpdatesPerSend call: "
			   .. tostring( limit ) .. " (must be a positive integer)" )

	end

	-- Pump the outbox and see what we need to send.
		
	return docStore:doComplexLogicInTransactionP(
		'WFSyncAgentOutboxPrep.getUpdatesToSend',
		{
			addChangedDocsToOutbox = doAddChangedDocsToOutbox,
			syncPeerName = self._name,
			syncStatusDocName = self._syncStatusDocName,
			outboxSlotName = self._outboxSlotName,
			transientBlackoutSlotName = self._transientBlackoutSlotName,
			resolvedSlotName = self._resolvedSlotName,
			limit = limit,
		} )
		
		:mapDataP( function( _, updatePackage )
	
			local conflictLogEntries = updatePackage.conflictLogEntries
			local docIdsThatDidntNeedUpdate = updatePackage.docIdsThatDidntNeedUpdate
			
			local updates = updatePackage.updates
			local nUpdates = #updates
			
			self._nChangesSent:add( nUpdates )

			-- Remember any conflict log info we got for diagnostic log.
			
			local conflictLog = self._conflictDiagnosticLog

			for i = 1, #conflictLogEntries do

				conflictLog[ #conflictLog + 1 ] = conflictLogEntries[ i ]

				if #conflictLog > 20 then
					table.remove( conflictLog, 1 )
				end

			end
		
			-- Do we need to continue? Note that we always run the process
			-- outbox step at least once just in case we had marked documents
			-- to send in a previous session that we failed to process for
			-- some reason.

			if nUpdates == 0 then
				return 'data', false
			end
			
			-- Call syncPeer:sendUpdatesP and process response.
			
			self._syncActivityLog:infof( "syncPeer:sendUpdatesP %s\n\n", updates )
			
			return self._syncPeer:sendUpdatesP( updates )

				:mapP( function( status, perDocStatus )
				
					self._syncActivityLog:debugf( "syncPeer:sendUpdatesP responded with status = %s\n\nperDocStatus = %s\n\n",
												  status, perDocStatus )
				
					if status == 'data' or type( perDocStatus ) == 'table' then
					
						-- If there was a partial error response, record that for
						-- diagnostic log.
					
						if type( perDocStatus ) == 'table' then
						
							recordSyncFailure( self,
								{
									what = "send",
									status = status,
									message = perDocStatus,
									updates = updates,
								} )
		
							-- Check for kill switch.
							
							if perDocStatus.killSwitch then
								killSwitchDeployed( self, perDocStatus.killSwitch )
								return 'error',
									   "Kill switch: " .. tostring( perDocStatus.killSwitch )
							end
				
						end
						
						-- Process response from peer.
					
						return self._docStore:doComplexLogicInTransactionP(
							'WFSyncAgentOutboxPostProc.handleSendUpdatesResponse',
							{
								status = status,
								perDocStatus = perDocStatus,
								outboxDocs = updatePackage.outboxDocs,
								syncPeerName = self._name,
								outboxSlotName = self._outboxSlotName,
								transientBlackoutSlotName = self._transientBlackoutSlotName,
								resolvedSlotName = self._resolvedSlotName,
								docIdsThatDidntNeedUpdate = docIdsThatDidntNeedUpdate,
								syncStatusDocName = self._syncStatusDocName,
							} )
		
							:mapDataP( function( _, status, message )
							
								self._hasProcessedOutbox = true
		
								-- Continue until we empty the outbox.
							
								if status == 'data' and message then
									return sendOutboundChangesP( self, false )
										:mapDataP( function( _, nestedHadUploads )
											return 'data', not not ( nestedHadUploads
																	 or nUpdates > 0 )
										end )
		
								else
									return status, message
								end
		
							end )
						
					else
		
						-- If we didn't get a per-document status, just treat
						-- this as an overall failure.
		
						recordSyncFailure( self,
							{
								what = "send",
								status = status,
								message = perDocStatus,
								updates = updates,
							} )
					
						return status, perDocStatus
					
					end
				
				end )
				
			end )
	
end

--------------------------------------------------------------------------------

local function sendUpdatesP_imp( self )

	-- Make sure we only have one send request in flight at a time.
	-- Also make sure we see this request through until completion.
	
	local startTime = KSDate.timeNowForAnimation()

	self._sendUpdatesP = sendOutboundChangesP( self, true )
		:observeP( function( status )
	
			self._syncActivityLog:debug( "sendUpdatesP finished" )
	
			-- If we have an error, then we need to flag the outbox
			-- for retry next time.
	
			if status ~= 'data' then
				self._hasProcessedOutbox = false
			end
		
			self._sendUpdatesP = false

			self._sendUpdatesCycleTime:add( KSDate.timeNowForAnimation() - startTime )
				
		end ):runUntilDoneP()
	
	self._syncActivityLog:debug( "sendUpdatesP starting" )

	return self._sendUpdatesP

end

--------------------------------------------------------------------------------
			
function syncAgentMethods:sendUpdatesP()

	check( 'WFSyncAgent' )

	-- Don't do this more than once at a time.
	
	if self._sendUpdatesP then
		return self._sendUpdatesP
	end
	
	-- Also serialize get and send requests. Too confusing to do both at once.
	
	if self._getChangesP then
		return self._getChangesP
			:mapP( function()
				return sendUpdatesP_imp( self )
			end )
			:runUntilDoneP()
	else
		return sendUpdatesP_imp( self )
	end
		
end

--------------------------------------------------------------------------------

-- @mark :syncUntilDoneP

--------------------------------------------------------------------------------

local function cancelAutomaticSync( self )

	if self._shouldStopAutomaticSync then

		self._isSyncingAutomaticallyB:set( false )
		self._shouldStopAutomaticSync = false

		return true

	end

end

--------------------------------------------------------------------------------

local function syncUntilDoneP_imp( self, anyChanges )

	return self:getChangesP()

		:mapDataP( function( _, receivedAnyChanges )

			if cancelAutomaticSync( self ) then
				return 'data', 'stopped'
			end
		
			return self:sendUpdatesP()

				:mapDataP( function( _, sentAnyChanges )
				
					if cancelAutomaticSync( self ) then
						return 'data', 'stopped'
					end
				
					if receivedAnyChanges or sentAnyChanges then

						return syncUntilDoneP_imp( self, true )
							-- WARNING [scouten 2014-10-27]: Please avoid
							-- the temptation to attach a mapDataP or similar
							-- here. This was found to lead to an undesirable
							-- accumulation of promises. (See
							-- https://git.corp.adobe.com/lua/wf/pull/283
							-- for details.

					else
						return 'data', anyChanges or false
					end
				
				end )
		
		end )
		
end

--------------------------------------------------------------------------------

function syncAgentMethods:syncUntilDoneP()

	check( 'WFSyncAgent' )

	self._syncActivityLog:debugf( "syncUntilDoneP starting via %s", ':traceback:' )

	-- Don't do this more than once at a time.
	
	if not self._syncUntilDoneP then

		self._syncUntilDoneP = syncUntilDoneP_imp( self )
			:observeP( function( status, message )
			
				if not WFCloseSupport.hasBeenOrWillBeClosed( self ) then
					self._lastResponseWasErrorB:set( status == 'error' )
				end
			
				self._syncUntilDoneP = false

				self._syncActivityLog:debugf( "syncUntilDoneP finished with %s, %s",
											  status, message )
			
			end )
		
	end
		
	return self._syncUntilDoneP

end

--------------------------------------------------------------------------------

-- @mark kill switch

--------------------------------------------------------------------------------

function syncAgentMethods:killSwitchB()

	return self._killSwitchB:relayB()
	
end

--------------------------------------------------------------------------------

-- @mark automatic sync

--------------------------------------------------------------------------------

function syncAgentMethods:isSyncingAutomaticallyB()

	check( 'WFSyncAgent' )

	return self._isSyncingAutomaticallyB:relayB()

end

--------------------------------------------------------------------------------

local function pumpAutomaticSync( self, delayInterval, pushOnly )

	assert( not self._killSwitchB:valueNow(),
			"Can't continue syncing when kill switch has deployed" )

	local oldTriggerSyncCount = self._triggerSyncCount
	local oldTriggerPushCount = self._triggerPushCount

	if self._activeSyncPumpE then
		return
	end
	
	self._activeSyncPumpE = true

	local syncP
	
	if pushOnly and pushOnly > 0 then
		syncP = self:sendUpdatesP()
	else
		syncP = self:syncUntilDoneP()
	end

	return syncP
	
		:whenResolvedDo( function( status, didAnything )
		
			-- Did we get shut down?
		
			if status == nil or didAnything == 'stopped' or cancelAutomaticSync( self ) then
				self._activeSyncPumpE = false
				return
			end
			
			-- If sync trigger fired while we were here, just start over again.
			
			if self._triggerSyncCount > oldTriggerSyncCount then

				self._activeSyncPumpE = false
				pumpAutomaticSync( self, 0 )
				return

			elseif self._triggerPushCount > oldTriggerPushCount then

				self._activeSyncPumpE = false
				pumpAutomaticSync( self, 0, pushOnly and ( pushOnly - 1 ) or 4 )
				return

			end
			
			-- Calculate delay before next sync attempt.
			-- If we're failing, escalate delay rapidly.
	
			local minDelay = self._minSyncDelayIntervalB:valueNow()
			local maxDelay = self._maxSyncDelayIntervalB:valueNow()
	
			if status == 'data' and didAnything then
				delayInterval = minDelay
			else
				delayInterval = delayInterval * 1.5
			end

            delayInterval = math.max( delayInterval, minDelay )
            delayInterval = math.min( delayInterval, maxDelay )

			local startTime = KSDate.timeNowForAnimation()

            local timerE = KSReactive.timerE( delayInterval )

			-- Restart cycle if any of the following happens:
				-- Delay interval expires.
				-- Auto-sync status changes. (Will cause cancel.)
				-- External trigger (if any) fires.

			self:setOwned {
				_activeSyncPumpE = KSReactive.anyE(
					timerE,
					self._isSyncingAutomaticallyB,
					self._triggerSyncE,
					self._docStoreChangeCounterB )
			}

            local eventDidFire = false -- extra insurance against multiple triggers

            self._activeSyncPumpE:ownedOnDataE( self, function()

				if eventDidFire then return end
				eventDidFire = true
				
				self._syncIdleCycleTime:add( KSDate.timeNowForAnimation() - startTime )

				self:setOwned { _activeSyncPumpE = false }

				if not cancelAutomaticSync( self ) then
					pumpAutomaticSync( self, delayInterval )
				end
			
			end )

		end )
		:attachTo( self )
		:suppressCloseLogging()

end

--------------------------------------------------------------------------------

function syncAgentMethods:startAutomaticSync()

	check( 'WFSyncAgent' )

	self._syncActivityLog:debugf( "startAutomaticSync starting via %s", ':traceback:' )

	self._shouldStopAutomaticSync = false

	assert( not self._killSwitchB:valueNow(),
			"Can't restart syncing after kill switch has deployed" )

	if self._isSyncingAutomaticallyB:valueNow() ~= true then

		self._isSyncingAutomaticallyB:set( true )
		pumpAutomaticSync( self, 2 )

	end

end

--------------------------------------------------------------------------------

function syncAgentMethods:stopAutomaticSync()

	check( 'WFSyncAgent' )

	self._syncActivityLog:debugf( "stopAutomaticSync starting via %s", ':traceback:' )

	self._shouldStopAutomaticSync = true
	
	if self._isSyncingAutomaticallyB:valueNow() == true then
		self._isSyncingAutomaticallyB:set( 'pendingStop' )
	end

end

--------------------------------------------------------------------------------

function syncAgentMethods:stopAutomaticSyncP()

	check( 'WFSyncAgent' )

	self._syncActivityLog:debugf( "stopAutomaticSyncP starting via %s", ':traceback:' )

	self:stopAutomaticSync()

	local function hasAnyOutstandingTraffic()
	
		self._syncActivityLog:debugf( "stopAutomaticSyncP get = %s send = %s sync = %s",
									  not not self._getChangesP,
									  not not self._sendUpdatesP,
									  not not self._syncUntilDoneP )
	
		return not ( self.getChangesP or self._sendUpdatesP or self._syncUntilDoneP )
	
	end

	local function tryAgainSoonP()
	
		return WFTimer.afterDelayP( .25 )
			:mapDataP( function()
			
				if hasAnyOutstandingTraffic() then
					return tryAgainSoonP()
				else
					return 'data', true
				end
			
			end )
		
	end
	
	if hasAnyOutstandingTraffic() then
		return tryAgainSoonP()
	else
		return KSPromises.resolvedPromiseP( 'data', true )
	end

end

--------------------------------------------------------------------------------

-- @mark clear permanent errors

--------------------------------------------------------------------------------

--[[!
	\brief Requests that any so-called "permanent" errors on a document be
		retried.
	
	\param docId (string) document ID to retry

	\return P< true > resolves to true when permanent error flag has been cleared
]]--

function syncAgentMethods:clearPermanentErrorsP( docId )

	check( 'WFSyncAgent', '+string' )

	return self._docStore:docValueP( docId )
		:mapDataP( function( _, docValue )
		
			local overlay = {}
			
			for k, v in pairs( docValue._syncStatus or {} ) do
			
				if v.error and v.error.permanent then

					if not overlay._syncStatus then
						overlay._syncStatus = {}
					end

					overlay._syncStatus[ k ] = {
						error = {
							permanent = false
						}
					}

				end
			
			end
			
			if next( overlay ) then

				overlay._id = docId
				overlay._slot = 'main'

				return self._docStore:applyOverlayP( overlay )
			
			else
			
				return 'data', true
			
			end
		
		end )
		:runUntilDoneP()

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------
--[[!
	\brief Sends diagnostic data on synchronization to the log.
]]--

function syncAgentMethods:sendDiagnosticLogData()

	check( 'WFSyncAgent' )

	-- If the sync peer implements diagnostic log method,
	-- include its report here. If not, quietly swallow
	-- any errors that result from calling a non-existent
	-- method.

	local success, syncPeerItems = pcall( function()
		return self._syncPeer:sendDiagnosticLogData().body
	end )
	
	local result = {

		name = "WF Sync Agent :: " .. self._name,
		id = 'WFSyncAgent-' .. self._name,
		weight = 500,

		body = {
		
			syncStatusDoc = self._docStore:docValueP( self._syncStatusDocName,
													  'wfSyncAgent' )
										  :runUntilDoneP(),
			
			inboxDocs = self._docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = self._inboxSlotName
				}:runUntilDoneP(),
		
			isSyncingAutomaticallyB = self._isSyncingAutomaticallyB,
			
			killSwitch = self._killSwitchB,
		
			hasReconciledInbox = self._hasReconciledInbox,
			hasProcessedOutbox = self._hasProcessedOutbox,

			recentErrors = self._recentSyncFailures,
			
			receivedDocumentsSummary = self._receivedDocumentsSummary,
			
			perf = {
				getChanges = self._getChangesCycleTime:dump(),
				sendUpdates = self._sendUpdatesCycleTime:dump(),
				idle = self._syncIdleCycleTime:dump(),
				changesReceivedPerCycle = self._nChangesReceived:dump(),
				changesSentPerCycle = self._nChangesSent:dump(),
			},
			
		},
	
	}
	
	if #self._conflictDiagnosticLog > 0 then
		
		result.body.conflicts = self._conflictDiagnosticLog
	
	end
	
	if success and type( syncPeerItems ) == 'table' then
	
		local resultBody = result.body
	
		for k, v in pairs( syncPeerItems ) do
		
			if resultBody[ k ] then
				resultBody[ 'syncPeer_' .. k ] = v
			else
				resultBody[ k ] = v
			end
		
		end
	
	end
	
	return result

end

--------------------------------------------------------------------------------

return Rx.Class 'WFSyncAgent'( syncAgentMethods )
