--[[----------------------------------------------------------------------------

WFSyncAgentInboxHandler.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

These functions implement key performance-sensitive parts of the sync agent.
Intended for use only by WFSyncAgent.lua.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local KSPCallLoops = require 'KSPCallLoops'
local KSTableUtils = require 'KSTableUtils'
local WFTableMerge = require 'WFTableMerge'


--------------------------------------------------------------------------------

local inboxLog = WFLogger 'WFSyncAgent :: inbox'
--	:setLevel( 'DEBUG' )


--============================================================================--

-- @mark WFSyncAgentInboxHandler

local WFSyncAgentInboxHandler = {}

--------------------------------------------------------------------------------

local function revertConflictingValues( inboxValue, oldResolvedValue, conflicts )

	-- Everywhere there's a conflict, we want to smash the old resolved value
	-- back into the inbox.
	
	-- If we had something other than a table at the old resolved value, just
	-- pretend it's an empty table. This will cause the pre-conflict values to
	-- all be considered as nil.
	
	if type( oldResolvedValue ) ~= 'table' then
		oldResolvedValue = {}
	end

	for k, v in pairs( conflicts ) do
	
		if k ~= '_replace' then

			inboxValue = inboxValue or {}
			
			if type( v ) == 'table' then
			
				inboxValue[ k ] = revertConflictingValues( inboxValue[ k ],
														   oldResolvedValue[ k ],
														   v )
			
			else
	
				inboxValue[ k ] = oldResolvedValue[ k ]
			
			end
		
		end
	
	end
	
	return inboxValue

end

--------------------------------------------------------------------------------

local function reconcileInboxDocumentWithValues( args, inboxDoc,
												 existingMainSlotValue,
												 existingResolvedSlotValue )

	inboxLog:debugf( "reconcileInboxDocumentWithValues\n\ninboxDoc = %s\n\nexistingMainSlotValue = %s\n\nexistingResolvedSlotValue = %s\n\n",
					 inboxDoc, existingMainSlotValue, existingResolvedSlotValue )
	
	local syncPeerName = args.syncPeerName
	
	-- Do a three-way merge, but ensure that we don't create false
	-- conflicts on system-generated values.

	inboxDoc._changeCounter = existingMainSlotValue._changeCounter
	inboxDoc._hash = nil
	inboxDoc._idLocal = nil
	inboxDoc._slot = 'main'
	
	local existingMainSlotIsDeleted = existingMainSlotValue._deleted
	
	local inboxIsDeleted = KSTableUtils.derefOpt( inboxDoc,
												  '_syncStatus',
												  syncPeerName,
												  'deleted' )

	existingResolvedSlotValue._changeCounter = existingMainSlotValue._changeCounter
	existingResolvedSlotValue._hash = nil
	existingResolvedSlotValue._idLocal = nil
	existingResolvedSlotValue._slot = 'main'

	-- Don't allow sync status from one sync peer to erase or invalidate sync status
	-- from other peers.
	
	local existingMainSlotSyncStatus = existingMainSlotValue._syncStatus
	local inboxDocSyncStatus = inboxDoc._syncStatus
	
	if inboxIsDeleted then
		inboxDocSyncStatus[ syncPeerName ].deleted = nil
	end
	
	if existingMainSlotSyncStatus then
	
		inboxDocSyncStatus = inboxDocSyncStatus or {}
		
		for k, v in pairs( existingMainSlotSyncStatus ) do
			if k ~= syncPeerName then
				inboxDocSyncStatus[ k ] = v
			end
		end
		
		if next( inboxDocSyncStatus ) then
			inboxDoc._syncStatus = inboxDocSyncStatus
		end
	
	end

	-- Disallow conflicts on _serverProvidedMetadata and _localOnly.
	
	local function autoResolveConflicts( key )
	
		if existingMainSlotValue[ key ]
			and existingResolvedSlotValue[ key ]
			and not inboxDoc[ key ]
		then
			inboxDoc[ key ] = existingMainSlotValue[ key ]
		end
	
	end
	
	autoResolveConflicts( '_localOnly' )
	autoResolveConflicts( '_serverProvidedMetadata' )

	-- Now merge the main document body.

	local merged, conflicts = WFTableMerge.threeWayMerge( existingResolvedSlotValue,
														  existingMainSlotValue,
														  KSTableUtils.deepCopy( inboxDoc ) )

	if conflicts and conflicts._syncStatus then
		conflicts._syncStatus = nil
		if not next( conflicts ) then
			conflicts = nil
		end
	end
	
	-- If sync peer is signaling deletion, this is a big hammer.
	-- Updates always lose to deletion.
	
	if merged._deleted or inboxIsDeleted or existingMainSlotIsDeleted then
	
		merged._deleted = true
		
		-- Remove all keys that don't have the underscore prefix.
		
		for k in pairs( merged ) do

			-- NOTE: Below test checks for a leading underscore (95 = underscore).
			-- We use this format because we think :byte is much faster than :match
			-- or :sub, both of which would have to perform more complex string
			-- manipulation. This being in a tight performance loop, it was deemed
			-- preferable to use this approach.
		
			if k:byte( 1 ) ~= 95 then
				merged[ k ] = nil
			end
		
		end
	
	end
	
	inboxLog:debugf( "reconcileInboxDocumentWithValues\n\nRESULT merged = %s\n\nconflicts = %s\n\n",
					 merged, conflicts )

	if not conflicts then

		WFTableMerge.applyOverlay( merged,
			{
				_syncStatus = {
					[ syncPeerName ] = {
						timestamp = KSTableUtils.derefOpt( inboxDoc,
														   '_syncStatus',
														   syncPeerName,
														   'timestamp' ),
						conflicts = { _delete = true },
					},
				},
				_deleted = inboxIsDeleted or nil,
			} )

		-- If we're receiving the same document that we've been attempting
		-- to send (it happens!), clear any error flag. (See PR# 3724615.)

		local syncError = KSTableUtils.derefOpt( merged,
												 '_syncStatus',
												 syncPeerName,
												 'error' )
		
		if syncError then
		
			local overlay, hasConflicts = WFTableMerge.createOverlay( merged, inboxDoc )
			local overlayIsEmpty = true

			if not hasConflicts then
			
				for k in pairs( overlay ) do
					if type( k ) ~= 'string' or not k:match( '^_' ) then
						overlayIsEmpty = false
						break
					end
				end
				
				if overlayIsEmpty then
					merged._syncStatus[ syncPeerName ].error = nil
				end
			
			end
		
		end
		
		if not next( merged._syncStatus[ syncPeerName ] ) then
			merged._syncStatus[ syncPeerName ] = nil
		end
		
		if not next( merged._syncStatus ) then
			merged._syncStatus = nil
		end

	else

		inboxLog:infof( "**** CONFLICT DETECTED ****" )
		inboxLog:infof( "existingResolvedSlotValue = %s", existingResolvedSlotValue )
		inboxLog:infof( "existingMainSlotValue = %s", existingMainSlotValue )
		inboxLog:infof( "inboxDoc = %s", inboxDoc )
		inboxLog:infof( "conflicts = %s", conflicts )
		inboxLog:infof( "**** END OF CONFLICT LOG ****" )

		local oldConflicts = conflicts
		if inboxIsDeleted then
			conflicts = {}
		end

		conflicts._replace = true
		
		WFTableMerge.applyOverlay( merged,
			{
				_syncStatus = {
					[ syncPeerName ] = {
						timestamp = KSTableUtils.derefOpt( inboxDoc,
														   '_syncStatus',
														   syncPeerName,
														   'timestamp' ),
						conflicts = conflicts,
					},
				},
			} )

		if inboxIsDeleted then
			merged._syncStatus[ syncPeerName ].conflicts._deleted = true
		end

		-- Advance resolved slot as much as possible for any values that
		-- are not in conflict.
		
		if not inboxIsDeleted then
			inboxDoc = revertConflictingValues( inboxDoc, existingResolvedSlotValue,
												oldConflicts )
		end
		
		-- Remember the new values from the remote side so we can retry
		-- merge later.
		
		WFTableMerge.applyOverlay( inboxDoc,
			{
				_syncStatus = {
					[ syncPeerName ] = {
						conflicts = conflicts,
					},
				},
			} )

	end

	inboxDoc._slot = args.resolvedSlotName
	inboxDoc._changeCounter = nil

	return merged, inboxDoc

end

--------------------------------------------------------------------------------

local function reconcileOneInboxDocument( args, inboxDoc )

	local docStore = args.docStore

	local id = assert( inboxDoc._id )

	local existingMainSlotValue = docStore:docValue( id, 'main' )
	local existingResolvedSlotValue = docStore:docValue( id, args.resolvedSlotName )
	
	return reconcileInboxDocumentWithValues( args,
											 inboxDoc,
											 existingMainSlotValue,
											 existingResolvedSlotValue )

end

--------------------------------------------------------------------------------

local function reconcileBatchOfInboxDocuments( args, inboxDocs )

	local docStore = args.docStore

	local function actionFn( i, inboxDoc )
	
		local docId = inboxDoc._id

		local newDocValue, newResolvedSlotValue = reconcileOneInboxDocument( args, inboxDoc )

		-- Successful reconciliation: Write updates for this document and clear it from inbox.
		
		docStore:setDocValue( newDocValue )
		docStore:setDocValue( newResolvedSlotValue )
		docStore:quietlyDeleteSlot( docId, args.inboxSlotName )

	end
	
	local function errorFn( i, message, inboxDoc )
	
		-- TODO [scouten 2014-01-17]: How might we get here? Not sure.
		-- Not sure what to do, but we should definitely avoid getting stuck.

		local docId = inboxDoc._id
		
		inboxLog:errorf( "**** UNEXPECTED ERROR ****" )
		inboxLog:errorf( "docId = %s", docId )
		inboxLog:errorf( "inboxDoc = %s", inboxDoc )
		inboxLog:errorf( "message = %s", message )
		inboxLog:errorf( "Inbox update dropped on floor" )
		inboxLog:errorf( "**** END OF ERROR LOG ****" )

		docStore:quietlyDeleteSlot( docId, args.inboxSlotName )

	end

	KSPCallLoops.pdoWithIterator( actionFn, errorFn, ipairs( inboxDocs ) )
	
end

--------------------------------------------------------------------------------

function WFSyncAgentInboxHandler.reconcileInbox( docStore, args )

	-- Save off the doc store for enclosed functions.

	args.docStore = docStore

	-- Repeat this loop until we've cleared all inbox documents.

	while true do

		-- Note that we delete items out of the inbox as we process it, so the
		-- volume of data here should be relatively small.
	
		local inboxDocs = docStore:changedDocs
			{
				fromChangeCounter = 0,
				slot = args.inboxSlotName,
			}
		
		if #inboxDocs == 0 then
			return true
		end
		
		reconcileBatchOfInboxDocuments( args, inboxDocs )

	end

end

--------------------------------------------------------------------------------

return WFSyncAgentInboxHandler
