--[[----------------------------------------------------------------------------

WFSyncAgentOutboxPostProc.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

These functions implement key performance-sensitive parts of the sync agent.
Intended for use only by WFSyncAgent.lua.

------------------------------------------------------------------------------]]

	-- WF.core.os
local KSDate = require 'KSDate'


--============================================================================--

-- @mark WFSyncAgentOutboxPostProc

local WFSyncAgentOutboxPostProc = {}

--------------------------------------------------------------------------------

local function peerDidAcceptChangedDoc( args, outboxDoc )

	local docStore = args.docStore

	-- Since server has accepted our change, de facto, it is now the
	-- resolved value.

	docStore:copySlotToSlot( outboxDoc._id,
							 args.outboxSlotName,
							 args.resolvedSlotName,
							 false )
	
	-- And we should now remove it from the outbox.
	
	docStore:quietlyDeleteSlot( outboxDoc._id, args.outboxSlotName )
	
	-- And remove any previous error status.
	
	docStore:applyOverlay
		{
			_id = outboxDoc._id,
			_slot = 'main',
			_syncStatus = {
				[ args.syncPeerName ] = {
					accepted = KSDate.timeInIso8601Format(),
					error = {
						_delete = true,
					},
				},
			},
		}

end

--------------------------------------------------------------------------------

local function peerDidRejectChangedDoc( args, outboxDoc, perDocStatus )

	local docStore = args.docStore

	-- Remove doc from the outbox.
	
	docStore:quietlyDeleteSlot( outboxDoc._id, args.outboxSlotName )
	
	-- And flag the error. If the error block contains { permanent = true } we
	-- won't retry unless we get an explicit client code request to do so.
	
	local errorBlock = perDocStatus
	
	if type( perDocStatus ) ~= 'table' then
		errorBlock = { message = perDocStatus }
	end
	
	errorBlock._replace = true
	errorBlock.time = KSDate.timeInIso8601Format()

	local permanent = not not( errorBlock.permanent )

	docStore:applyOverlay
		{
			_id = outboxDoc._id,
			_slot = 'main',
			_syncStatus = {
				[ args.syncPeerName ] = {
					_replace = true,
					accepted = { _delete = true },
					error = errorBlock,
				},
			},
		}
	
	-- Is this a transient failure? If so, flag it as such so we remember
	-- to try again in a little while. (For now, "little while" means 30
	-- seconds, but we may make that configurable later.)
	
	if not permanent then
	
		local blackoutExpirationTime = KSDate.timeNow() + 30
		local blackoutExpirationTimeStr = KSDate.timeInIso8601Format( blackoutExpirationTime )
	
		docStore:setDocValue
			{
				_id = outboxDoc._id,
				_slot = args.transientBlackoutSlotName,
				expires = blackoutExpirationTimeStr,
			}
	
	end

end

--------------------------------------------------------------------------------

function WFSyncAgentOutboxPostProc.handleSendUpdatesResponse( docStore, args )

	args.docStore = docStore
	
	local status = args.status
	local perDocStatus = args.perDocStatus
	local outboxDocs = args.outboxDocs
	local docIdsThatDidntNeedUpdate = args.docIdsThatDidntNeedUpdate

	if status == 'data' and type( perDocStatus ) ~= 'table' then
		
		-- If we succeeded, then remove all documents
		-- from outbox and advance the "continue from"
		-- pointer accordingly.
		
		for i = 1, #outboxDocs do

			local outboxDoc = outboxDocs[ i ]
			if not docIdsThatDidntNeedUpdate[ outboxDoc._id ] then
				peerDidAcceptChangedDoc( args, outboxDoc )
			end

		end

		return 'data', #outboxDocs > 0
		
	elseif type( perDocStatus ) == 'table' then

		-- We have a per-document status. Need to figure out
		-- what succeeded and what failed.
		
		local didHaveAnyFailures
		local nSkips = 0

		for i = 1, #outboxDocs do
		
			local outboxDoc = outboxDocs[ i ]

			local docStatus = perDocStatus[ outboxDoc._id ]
			
			if docStatus == ':skip:' then
			
				-- Sync peer wants to quietly skip this document. Remove from
				-- outbox and pretend nothing happened.
				
				-- TODO [scouten 2014-03-06]: This is an ugly ugly hack. When
				-- we have the full bulk document API, please remove this!
			
				docStore:quietlyDeleteSlot( outboxDoc._id, args.outboxSlotName )
				nSkips = nSkips + 1
			
			elseif docStatus then
				
				didHaveAnyFailures = true
				
				peerDidRejectChangedDoc( args, outboxDoc,
										 perDocStatus[ outboxDoc._id ] )
			
			else
				
				if not docIdsThatDidntNeedUpdate[ outboxDoc._id ] then
					peerDidAcceptChangedDoc( args, outboxDoc )
				end
			
			end
		
		end

		if not didHaveAnyFailures then
		
			if nSkips == #outboxDocs then
			
				return 'data', true
			
			else
		
				return 'error', "sync peer returned a table that did not have any errors"
			
			end
			
		else
		
			return 'error', "sync peer rejected one or more documents"

		end
	
	else
	
		return 'error', "incorrect arguments to outbox post-proc"

	end

end

--------------------------------------------------------------------------------

return WFSyncAgentOutboxPostProc
