--[[----------------------------------------------------------------------------

WFSyncAgentOutboxPrep.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

These functions implement key performance-sensitive parts of the sync agent.
Intended for use only by WFSyncAgent.lua.

------------------------------------------------------------------------------]]

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFTableMerge = require 'WFTableMerge'


--------------------------------------------------------------------------------

local outboxPrepLog = WFLogger 'WFSyncAgent :: outbox Prep'
--	:setLevel( 'DEBUG' )


--============================================================================--

-- @mark WFSyncAgentOutboxPrep

local WFSyncAgentOutboxPrep = {}

--------------------------------------------------------------------------------

local function putOneDocumentInOutbox( args, possibleOutboundDoc )

	local docStore = args.docStore

	-- If we have an error that is marked permanent, then skip this doc.

	local hasPermanentError = KSTableUtils.derefOpt( possibleOutboundDoc,
													 '_syncStatus',
													 args.syncPeerName,
													 'error',
													 'permanent' )
	
	if hasPermanentError then
		return
	end
	
	-- Ignore document if it has been deleted with muting.
	
	if possibleOutboundDoc._deleted
		and KSTableUtils.derefOpt( possibleOutboundDoc,
								   '_syncStatus',
								   'mute' )
	then
		return
	end

	-- Does this document have any conflicts?

	local conflicts = KSTableUtils.derefOpt( possibleOutboundDoc,
											 '_syncStatus',
											 args.syncPeerName,
											 'conflicts' )

	outboxPrepLog:debugf( "putOneDocumentInOutbox %s\n\nconflicts = %s",
						  possibleOutboundDoc, conflicts )

	if conflicts then
		
		-- Let's see if we can auto-resolve any conflicts before sending.
		
		local outboundDocCopy = KSTableUtils.deepCopy( possibleOutboundDoc )
		
		WFTableMerge.applyOverlay( outboundDocCopy, conflicts )
		
		if KSTableUtils.deepEquals( possibleOutboundDoc, outboundDocCopy ) then

			-- Success! Go ahead and put this in the outbox, even with the
			-- conflicts. convertOutboxDocToUpdate will use the conflict
			-- information to generate an appropriate overlay.

			docStore:copySlotToSlot( possibleOutboundDoc._id,
									 'main', args.outboxSlotName )

			-- And erase the _conflicts stanza from the main slot value.
			
			docStore:applyOverlay
				{
					_id = possibleOutboundDoc._id,
					_slot = 'main',
					_syncStatus = {
						[ args.syncPeerName ] = {
							conflicts = { _delete = true },
						},
					},
				}
		
		else

			-- Can't resolve conflicts: Don't put the document in the outbox.
		
			args.conflictLogEntries[ #args.conflictLogEntries + 1 ] =
				{
					docId = possibleOutboundDoc._id,
					docType = possibleOutboundDoc.type,
					time = KSDate.timeInIso8601FormatWithTimeZone(),
					newDoc = possibleOutboundDoc,
					conflicts = conflicts,
				}
			
		end
		
	else

		docStore:copySlotToSlot( possibleOutboundDoc._id,
								 'main', args.outboxSlotName )
	
	end

end

--------------------------------------------------------------------------------

local function addChangedDocsToOutbox( docStore, args )
	
	-- We're adding changed docs to the outbox. Where did we leave off last time?

	local syncStatusDocValue = docStore:docValue( args.syncStatusDocName, 'wfSyncAgent' )
	local outboundContinueFrom = syncStatusDocValue.outboundContinueFrom or 0
	
	-- Ask local doc store for changes in 'main' slot since then.
	
	outboxPrepLog:debugf( "sendUpdatesP_imp outboundContinueFrom = %s", outboundContinueFrom )
	
	local possibleOutboundDocs = docStore:changedDocs
		{
			fromChangeCounter = outboundContinueFrom,
			slot = 'main',
		}
		
	outboxPrepLog:debugf( "sendUpdatesP_imp possibleOutboundDocs = %s", possibleOutboundDocs )

	outboxPrepLog:debugf( "sendUpdatesP changedDocs.continueFrom = %s",
						  possibleOutboundDocs.continueFrom )

	-- Add each possible document to outbox.

	if #possibleOutboundDocs > 0 then

		for i = 1, #possibleOutboundDocs do
			putOneDocumentInOutbox( args, possibleOutboundDocs[ i ] )
		end

		-- Remember where we should continue from the next time we
		-- want to prime the outbox pump.

		outboxPrepLog:debugf( "sendOutboundChanges new outboundContinuefrom = %s",
							  possibleOutboundDocs.continueFrom )
	
		docStore:applyOverlay
			{
				_id = args.syncStatusDocName,
				_slot = 'wfSyncAgent',
				outboundContinueFrom = possibleOutboundDocs.continueFrom,
			}
		
	end
	
	return #possibleOutboundDocs

end

--------------------------------------------------------------------------------

local function convertOutboxDocToUpdate( args, outboxDoc )

	local docStore = args.docStore

	local resolvedDocValue = docStore:docValue( outboxDoc._id, args.resolvedSlotName )

	local overlay
	
	outboxPrepLog:debugf( "convertOutboxDocToUpdate\n\noutboxDoc = %s\n\nresolvedDocValue = %s\n\n",
						  outboxDoc, resolvedDocValue )
	
	local blackoutDocument = docStore:docValue( outboxDoc._id, args.transientBlackoutSlotName )
	
	if not blackoutDocument._empty then
		outboxPrepLog:debugf( "outboxDoc ID %s is embargoed; ignoring for now", outboxDoc._id )
		return nil
	end
	
	local rdvCopy = KSTableUtils.deepCopy( resolvedDocValue )
	
	if resolvedDocValue._empty then
	
		overlay = outboxDoc
		overlay._timestamp = ':new:'
	
	else

		-- Did we manage to auto-resolve any conflicts? If so,
		-- advance the resolved doc value before creating overlay.

		local conflicts = KSTableUtils.derefOpt( outboxDoc,
												 '_syncStatus',
												 args.syncPeerName,
												 'conflicts' )

		if conflicts then
		
			-- Before auto-resolving, make sure we're using the
			-- fully-updated remote value.

			local rdvConflicts = KSTableUtils.derefOpt( resolvedDocValue,
														'_syncStatus',
														args.syncPeerName,
														'conflicts' )
			
			if rdvConflicts then
				WFTableMerge.applyOverlay( resolvedDocValue, rdvConflicts )
			end

			WFTableMerge.applyOverlay( resolvedDocValue, conflicts )
			outboxDoc._syncStatus[ args.syncPeerName ].conflicts = nil

		end
	
		overlay = WFTableMerge.createOverlay( resolvedDocValue, outboxDoc )

		local isEmptyUpload = true

		for k in pairs( overlay ) do
			if not k:match( '^_' ) and k ~= 'revision_ids' then
				isEmptyUpload = false
				break
			end
		end
		
		if isEmptyUpload then
			return nil
		end

		overlay._timestamp = KSTableUtils.derefOpt( resolvedDocValue,
													'_syncStatus',
													args.syncPeerName,
													'timestamp' )
		
		if not overlay._timestamp then

			-- It's possible we might never have had a resolved value
			-- because we had conflicts early on and have since resolved
			-- them. If so, then we can look for timestamp in the outbox
			-- document.
			
			overlay._timestamp = KSTableUtils.derefOpt( outboxDoc,
														'_syncStatus',
														args.syncPeerName,
														'timestamp' )
		
		end
		
		if not overlay._timestamp then
		
			-- If still no timestamp, we probably just sent a doc change
			-- and haven't had it read back to us via the remote
			-- site's changes feed. We'll stomp a ':new:' flag in there
			-- which will probably cause an error and resync.
		
			overlay._timestamp = ':new:'

		end

	end

	overlay._changeCounter = nil
	overlay._id = outboxDoc._id
	overlay._index = outboxDoc._index
	overlay._idLocal = nil
	overlay._hash = nil
	overlay._slot = nil
	overlay._localOnly = nil
	overlay._syncStatus = nil
	
	local strippedFullDocValue = {}
	for k, v in pairs( outboxDoc ) do
		if not k:match( '^_' ) or k == '_serverProvidedMetadata' then
			strippedFullDocValue[ k ] = v
		end
	end
	
	overlay._fullDocValue = KSTableUtils.deepCopy( strippedFullDocValue )
		-- Do a deep copy because debug logging and transit choke on value if we don't.
	
	if not rdvCopy._empty then
		overlay._previouslyAgreedValue = rdvCopy
		overlay._previouslyAgreedValue._slot = nil
		overlay._previouslyAgreedValue._hash = nil
		overlay._previouslyAgreedValue._idLocal = nil
		overlay._previouslyAgreedValue._changeCounter = nil
		overlay._previouslyAgreedValue._syncStatus = nil
	end
	
	return overlay

end

--------------------------------------------------------------------------------

local function convertOutboxToUpdates( args )

	local docStore = args.docStore

	-- Find out how many documents we're allowed to send at once.
	
	local limit = args.limit

	-- Note that we delete items out of the outbox as we process it, so the
	-- volume of data here should be relatively small.

	local docIdsThatDidntNeedUpdate = {}
	args.docIdsThatDidntNeedUpdate = docIdsThatDidntNeedUpdate
		-- Set of doc IDs for which convertOutboxDocToUpdate generated a
		-- null update.

	local outboxDocs = docStore:changedDocs
		{
			fromChangeCounter = 0,
			slot = args.outboxSlotName,
		}
	
	local nOutboxDocs = #outboxDocs
	if nOutboxDocs == 0 then
		return
	end

	-- Convert each possible outbox doc to an actual update to send.
	-- This will be the payload for the call to syncPeer:sendUpdatesP.
	
	local updates = {}
	local nUpdates = 0
	
	local stopIndex
	
	for i = 1, #outboxDocs do
	
		local outboxDoc = outboxDocs[ i ]

		local updateDoc = convertOutboxDocToUpdate( args, outboxDoc )

		outboxPrepLog:debugf( "convertOutboxDocToUpdate generated updateDoc = %s",
							  updateDoc )

		if updateDoc then

			nUpdates = nUpdates + 1
			updates[ nUpdates ] = updateDoc
			
			if nUpdates == limit then
				stopIndex = i
				break
			end
			
		else
		
			-- Avoid feedback loop of processing "accepted"
			-- notifications touching documents. See bug 3605785.
			
			docIdsThatDidntNeedUpdate[ outboxDoc._id ] = true
			
			docStore:quietlyDeleteSlot( outboxDoc._id,
										outboxDoc._slot )

		end
		
	end
	
	-- Truncate outbox array so we don't pretend further updates
	-- got accepted in post-proc.

	if stopIndex then
	
		for i = stopIndex + 1, #outboxDocs do
			outboxDocs[ i ] = nil
		end
	
	end
	
	args.outboxDocs = outboxDocs
	
	return updates
	
end

--------------------------------------------------------------------------------

local function collectAndClearTransientBlackouts( args )

	local docStore = args.docStore

	local possibleBlackoutDocs = docStore:changedDocs
		{
			fromChangeCounter = 0,
			slot = args.transientBlackoutSlotName,
		}
	
	local timeNow = KSDate.timeInIso8601Format()
	
	for i = 1, #possibleBlackoutDocs do

		local possibleBlackoutDoc = possibleBlackoutDocs[ i ]
		local blackoutExpiration = possibleBlackoutDoc.expires
		
		if type( blackoutExpiration ) ~= 'string'
			or blackoutExpiration < timeNow
		then
		
			-- Remove the transient blackout sentinel.
		
			docStore:quietlyDeleteSlot( possibleBlackoutDoc._id,
										args.transientBlackoutSlotName )

			-- Also tweak this doc so as to make it appear
			-- in subsequent changedDocs report for main slot.

			docStore:applyOverlay
				{
					_id = possibleBlackoutDoc._id,
					_slot = 'main',
					_syncStatus = {
						[ args.syncPeerName ] = {
							error = {
								retry = timeNow,
							},
						},
					},
				}
		
		end

	end

end

--------------------------------------------------------------------------------

function WFSyncAgentOutboxPrep.getUpdatesToSend( docStore, args )
	
	outboxPrepLog:debugf( "getUpdatesToSend\n\nargs = %s\n\n", args )

	args.docStore = docStore
	args.conflictLogEntries = {}
	
	collectAndClearTransientBlackouts( args )
	
	local nPossibleOutboundDocs = 0
	
	if args.addChangedDocsToOutbox then
		nPossibleOutboundDocs = addChangedDocsToOutbox( docStore, args )
	end
	
	local updates = convertOutboxToUpdates( args )

	return {
		nPossibleOutboundDocs = nPossibleOutboundDocs,
		conflictLogEntries = args.conflictLogEntries,
		updates = updates or {},
		outboxDocs = args.outboxDocs,
		docIdsThatDidntNeedUpdate = args.docIdsThatDidntNeedUpdate or {},
	}

end

--------------------------------------------------------------------------------

return WFSyncAgentOutboxPrep
