--[[----------------------------------------------------------------------------

WFSyncAgentTest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.self_test
local WFAsyncTestCase = require 'WFAsyncTestCase'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local KSTestUtils = require 'KSTestUtils'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.doc_store
local WFDocStore = require 'WFDocStore'

	-- WF.oz_client.sync_agent
local WFSyncAgent = require 'WFSyncAgent'

	-- WF.oz_client.sync_agent.test
local WFTestSyncPeer = require 'WFTestSyncPeer'


--============================================================================--

local WFSyncAgentTest = {}

local docStore, testSyncPeer, syncAgent, testSyncPeer2, syncAgent2

local allChangedDocs = {}
local allChangedDocsTotal = 0
local allChangedDocIds = {}
local allChangedDocIdsTotal = 0
local tempFromChangeCounter = 0

--------------------------------------------------------------------------------

-- Essentially, we want this integrity check appended to every test without actually
-- having to duplicate it on every test.  There should really be a better way.

local RealWFAsyncTestCase = WFAsyncTestCase
 
local function decoratedAsyncTestCase(...)
	return RealWFAsyncTestCase(...)
		:beforeRunning( function( test )

			if docStore then
				test:waitFor( function()
					
					return docStore:performIntegrityCheckP()
					
				end )
				
				test:continueWhenValueMatches( 'data', { } )
			end
			
		end )
end

WFAsyncTestCase = decoratedAsyncTestCase

--------------------------------------------------------------------------------

function WFSyncAgentTest.setUp()

	allChangedDocs = {}
	allChangedDocsTotal = 0
	allChangedDocIds = {}
	allChangedDocIdsTotal = 0
	tempFromChangeCounter = 0

	docStore = WFDocStore( ':memory:' )

	testSyncPeer = WFTestSyncPeer()
	syncAgent = WFSyncAgent( docStore, testSyncPeer, 'test' )

	-- Sync peer #2 is not used in most cases, but there are a few interop
	-- tests that cover this.

	testSyncPeer2 = WFTestSyncPeer()
	syncAgent2 = WFSyncAgent( docStore, testSyncPeer2, 'test2' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.tearDown()

	if syncAgent then
		syncAgent:close()
		syncAgent = nil
	end

	if syncAgent2 then
		syncAgent2:close()
		syncAgent2 = nil
	end

	if testSyncPeer then
		testSyncPeer:close()
		testSyncPeer = nil
	end

	if testSyncPeer2 then
		testSyncPeer2:close()
		testSyncPeer2 = nil
	end

	if docStore then
		docStore:close()
		docStore = nil
	end

end

--------------------------------------------------------------------------------

local function stripSyncStatusHashIfExists( docValue, syncPeerName )

	local syncStatusHash = KSTableUtils.derefOpt( docValue, '_syncStatus', syncPeerName )

	if syncStatusHash then

		if type( syncStatusHash.accepted ) == 'string' then
			syncStatusHash.accepted = syncStatusHash.accepted:gsub(
				'^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d%d%dZ',
				"(valid ISO 8601 timestamp)" )
		end

		local errorHash = KSTableUtils.derefOpt( syncStatusHash, 'error' )

		if errorHash then

			if type( errorHash.message ) == 'string' then
				errorHash.message = errorHash.message:gsub( '^%[.*"%]:%d+: ', '' ) --"
					-- trailing quote fixes broken Halide syntax color
			end

			if type( errorHash.time ) == 'string' then
				errorHash.time = errorHash.time:gsub(
					'^%d%d%d%d%-%d%d%-%d%dT%d%d:%d%d:%d%d%.%d%d%dZ',
					"(valid ISO 8601 timestamp)" )
			end

		end

	end

end

--------------------------------------------------------------------------------

local function mapChangedDocs_strip( _, changedDocs )

	for i = 1, #changedDocs do

		local docValue = changedDocs[ i ]

		docValue._idLocal = nil
		docValue._hash = nil
		docValue._changeCounter = nil

		stripSyncStatusHashIfExists( docValue, 'test' )
		stripSyncStatusHashIfExists( docValue, 'test2' )

	end

	tempFromChangeCounter = changedDocs.continueFrom
	changedDocs.continueFrom = nil

	return 'data', changedDocs

end

--------------------------------------------------------------------------------

local function mapChangedDocs_stripAndSort( _, changedDocs )

	for i = 1, #changedDocs do

		local docValue = changedDocs[ i ]

		docValue._idLocal = nil
		docValue._hash = nil
		docValue._changeCounter = nil

		stripSyncStatusHashIfExists( docValue, 'test' )
		stripSyncStatusHashIfExists( docValue, 'test2' )

	end

	tempFromChangeCounter = changedDocs.continueFrom
	changedDocs.continueFrom = nil
	
	table.sort( changedDocs,
		function( a, b )
		
			return a._id < b._id
		
		end )

	return 'data', changedDocs

end

--------------------------------------------------------------------------------

local function mapChangedDocs_stripToIdOnly( _, changedDocs )

	for i = 1, #changedDocs do
		allChangedDocIds[changedDocs[i]._id] = changedDocs[i]._id
	end

	tempFromChangeCounter = changedDocs.continueFrom

	return 'data', allChangedDocIds
end

--------------------------------------------------------------------------------

-- @mark misc

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSyncPeer()

	assert( syncAgent:syncPeer() == testSyncPeer )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSetLogLevel()

	syncAgent:setLogLevel 'debug'
		-- Hard to verify that it actually does anything...

end

--------------------------------------------------------------------------------

-- @mark :attachHelperObject

--------------------------------------------------------------------------------

local helperClass = Rx.Class "helperClass" { WFCloseSupport.addCloseSupportMethods }

--------------------------------------------------------------------------------

function WFSyncAgentTest.testAttachHelperObject()

	local helperObject = helperClass()

	syncAgent:attachHelperObject( helperObject )
	
	syncAgent:close()
	syncAgent = nil
	
	assert( WFCloseSupport.hasBeenOrWillBeClosed( helperObject ) )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testRemoveHelperObject()

	local helperObject = helperClass()

	syncAgent:attachHelperObject( helperObject )
	
	assert( not WFCloseSupport.hasBeenOrWillBeClosed( helperObject ) )

	syncAgent:removeHelperObject( helperObject )
	
	assert( WFCloseSupport.hasBeenOrWillBeClosed( helperObject ) )

	syncAgent:close()
	syncAgent = nil

end

--------------------------------------------------------------------------------

-- @mark :getChangesP

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_noop( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_noopTwice( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_overlappingRequestsCoalesced( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:delayNextChangedDocsP( .25 )

			local p1 = syncAgent:getChangesP()
			local p2 = syncAgent:getChangesP()

			assert( p1 == p2, "requests weren't coalesced" )

			return p1

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_oneDoc( test )

	WFAsyncTestCase( test )

		-- Get changes before there are any changes.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Verify "since" argument to :changedDocs call on sync peer.

		:waitFor( function()

			return KSPromises.resolvedPromiseP( 'data',
					testSyncPeer:sinceArgToLastChangedDocsPCall() )

		end )
		:continueWhenValueMatches( 'data', nil )

		-- Now introduce a change.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :getChangesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify "since" argument to :changedDocs call on sync peer.

		:waitFor( function()

			return KSPromises.resolvedPromiseP( 'data',
					testSyncPeer:sinceArgToLastChangedDocsPCall() )

		end )
		:continueWhenValueMatches( 'data', 1 )

		-- And we should see doc values in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Now change the doc on remote side a second time.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 53,
				}

		end )

		-- :getChangesP() should signal activity again.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify "since" argument to :changedDocs call on sync peer.

		:waitFor( function()

			return KSPromises.resolvedPromiseP( 'data',
					testSyncPeer:sinceArgToLastChangedDocsPCall() )

		end )
		:continueWhenValueMatches( 'data', 2 )

		-- And we should see doc values in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
						},
					},
					foo = 53,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_doesntAskForSameChangesAgain( test )

	WFAsyncTestCase( test )

		-- Introduce a change on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :getChangesP() should signal activity.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Next call to :getChangesP() should not signal activity.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And should have remembered where it was after first call.

		:waitFor( function()

			return KSPromises.resolvedPromiseP( 'data',
					testSyncPeer:sinceArgToLastChangedDocsPCall() )

		end )
		:continueWhenValueMatches( 'data', 2 )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_errorThenOneDoc( test )

	WFAsyncTestCase( test )

		-- Introduce a change on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Make sure :getChangesP() can handle failure from sync peer
		-- successfully.

		:waitFor( function()

			testSyncPeer:delayNextChangedDocsP( .5 )
			testSyncPeer:overrideNextChangedDocsPResponse( 'error', "simulated failure" )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- At this point we should not doc values in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
			} )

		-- On retry, :getChangesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify "since" argument to :changedDocs call on sync peer.

		:waitFor( function()

			return KSPromises.resolvedPromiseP( 'data',
					testSyncPeer:sinceArgToLastChangedDocsPCall() )

		end )
		:continueWhenValueMatches( 'data', nil )

		-- And we should see doc values in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseNotTable( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data', 42 )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseMissingContinueFrom( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data', {} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseInvalidContinueFrom( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
														   { continueFrom = false } )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseItemNotTable( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					continueFrom = 99,
					"not a table",
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseItemMissingId( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					continueFrom = 99,
					{},
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseItemIdEmptyString( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					continueFrom = 99,
					{ _id = '' },
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseItemIdNotString( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					continueFrom = 99,
					{ _id = 56 },
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseItemTimestampMissing( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					continueFrom = 99,
					{ _id = 'blah' },
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetChangesP_gripeResponseItemTimestampNotStringOrNumber( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					continueFrom = 99,
					{ _id = 'blah', _timestamp = false },
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

-- @mark :sendUpdatesP

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_noop( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_noopTwice( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_overlappingRequestsCoalesced( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			-- testSyncPeer:delayNextChangedDocsP( .25 )

			local p1 = syncAgent:sendUpdatesP()
			local p2 = syncAgent:sendUpdatesP()

			assert( p1 == p2, "requests weren't coalesced" )

			return p1

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_gripeMaxUpdatesResponseNotNumber( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextMaxUpdatesPerSendResponse( 'bogus' )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_gripeMaxUpdatesResponseNegative( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextMaxUpdatesPerSendResponse( -42 )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_gripeMaxUpdatesResponseZero( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextMaxUpdatesPerSendResponse( 0 )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_gripeMaxUpdatesResponseNotInteger( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:overrideNextMaxUpdatesPerSendResponse( 3.14159 )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- :sendUpdatesP() should signal no activity.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- Now tweak a document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we got exactly one call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- :sendUpdatesP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And doc values should be unchanged in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we didn't get another call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_deleteWithMute( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- :sendUpdatesP() should signal no activity.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- Now tweak a document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we got exactly one call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- :sendUpdatesP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And doc values should be unchanged in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we didn't get another call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Now delete the document using our magic muting flag.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = { mute = true },
					_deleted = true,
				}

		end )

		-- :sendUpdatesP() should not signal activity again.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And doc values should be unchanged in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we didn't get another call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_errorPeerFailureThenSuccess( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with failure.

		:waitFor( function()

			testSyncPeer:overrideNextSendUpdatesPResponse( 'error', 'failfailfail' )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Local doc store should have the intended sent document still in
		-- the outbox and should not retry putting the document in the outbox
		-- (unless it changes, of course).

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				},
				{
					_id = 'testDocId',
					_slot = 'test-outbox',
					foo = 42,
				},
				{
					_id = 'syncPeer-test',
					_slot = 'wfSyncAgent',
					outboundContinueFrom = 2,
				},
			} )

		-- :sendUpdatesP() succeeds this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And then we should see change markers in
		-- main doc and outbox should be empty.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'syncPeer-test',
					_slot = 'wfSyncAgent',
					outboundContinueFrom = 2,
				},
				{
					_id = 'testDocId',
					_slot = 'test-resolved',
					foo = 42,
				},
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
						},
					},
					foo = 42,
				},
			} )

		-- :sendUpdatesP() should now say there's nothing to send.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_errorPeerFailureThenLocalChange( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with failure.

		:waitFor( function()

			testSyncPeer:overrideNextSendUpdatesPResponse( 'error', 'failfailfail' )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Local doc store should have the intended sent document still in
		-- the outbox and should not retry putting the document in the outbox
		-- (unless it changes, of course).

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				},
				{
					_id = "testDocId",
					_slot = "test-outbox",
					foo = 42,
				},
				{
					_id = "syncPeer-test",
					_slot = "wfSyncAgent",
					outboundContinueFrom = 2,
				},
			} )

		-- Then we make another change locally while we have something lurking
		-- in the outbox.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
					blah = 'yawn',
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- :sendUpdatesP() succeeds this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Remote doc store should have the final
		-- updated value.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
					blah = 'yawn',
				}
			} )

		-- And then we should see change markers in
		-- main doc and outbox should be empty.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "syncPeer-test",
					_slot = "wfSyncAgent",
					outboundContinueFrom = 5,
				},
				{
					_id = "testDocId",
					_slot = "test-resolved",
					foo = 45,
					blah = 'yawn',
				},
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
						},
					},
					foo = 45,
					blah = 'yawn',
				},
			} )

		-- :sendUpdatesP() should now say there's nothing to send.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_errorPerDocFailure_permanent( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with per-document failure.

		:waitFor( function()

			testSyncPeer:partiallyFailNextSendUpdatesPResponse
				{
					testDocId = {
						message = "totally failed, dude!",
						permanent = true,
					},
				}

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- We should see error status written back to our document.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					_syncStatus = {
						test = {
							error = {
								message = "totally failed, dude!",
								time = "(valid ISO 8601 timestamp)",
								permanent = true,
							},
						},
					},
				},
			} )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Since this error was marked with permanent flag,
		-- we should not attempt to send it again.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testClearPermanentErrorsP( test )

	WFAsyncTestCase( test, 60 )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with permanent failure.

		:waitFor( function()

			testSyncPeer:partiallyFailNextSendUpdatesPResponse
				{
					testDocId = {
						message = "totally failed, dude!",
						permanent = true,
					},
				}

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- We should see error status written back to our document.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					_syncStatus = {
						test = {
							error = {
								message = "totally failed, dude!",
								time = "(valid ISO 8601 timestamp)",
								permanent = true,
							},
						},
					},
				},
			} )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Since this error was marked with permanent flag,
		-- we should not attempt to send it again.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- And we should not see doc value in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Now clear the permanent error flag.
		
		:waitFor( function()
			
			return syncAgent:clearPermanentErrorsP( 'testDocId' )
		
		end )

		-- Since this error is no longer marked with permanent flag,
		-- we should attempt to send it again.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return testSyncPeer:argsToLastSendUpdatesPCall()

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_fullDocValue = {
						foo = 42,
					},
					_id = 'testDocId',
					_timestamp = ':new:',
					foo = 42,
				}
			} )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '2' )

		-- And now we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_errorPerDocFailure_transient_testNightly( test )

	WFAsyncTestCase( test, 60 )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with per-document failure.

		:waitFor( function()

			testSyncPeer:partiallyFailNextSendUpdatesPResponse
				{
					testDocId = "sorry, I'm not feeling well right now",
				}

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- And we should not see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Since this error was not marked with permanent flag,
		-- we should attempt to send it again, but not until
		-- 30-second blackout window has expired.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Wait out the blackout window and verify that we do now
		-- send the change.

		:waitFor( 30 )
		
		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return testSyncPeer:argsToLastSendUpdatesPCall()

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_fullDocValue = {
						foo = 42,
					},
					_id = 'testDocId',
					_timestamp = ':new:',
					foo = 42,
				}
			} )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )

		end )
		:continueWhenValueMatches( 'data', '2' )

		-- And now we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				},
			} )


end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_gripePerDocFailureTableEmpty( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with per-document failure.

		:waitFor( function()

			testSyncPeer:overrideNextSendUpdatesPResponse( 'error', {} )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Can't test for this, but error message should say something
		-- about empty error table response.

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDoc_gripePerDocFailureTableMalformed( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() responds with per-document failure.

		:waitFor( function()

			testSyncPeer:overrideNextSendUpdatesPResponse( 'error', { 'abc', 'def' } )
				-- Error response table *should* be [id] = message mapping.

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Can't test for this, but error message should say something
		-- about empty error table response.

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_twoDocs( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocTwo',
					_slot = 'main',
					blah = 79,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				},
				{
					_id = 'testDocTwo',
					_slot = 'main',
					blah = 79,
				}
			} )

		-- :sendUpdatesP() should now say there's nothing to send.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_twoDocs_partialFailureThenSuccess_testNightly( test )

	WFAsyncTestCase( test, 60 )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocTwo',
					_slot = 'main',
					blah = 79,
				}

		end )

		-- :sendUpdatesP() configured to succeed on one doc, fail the other.

		:waitFor( function()

			testSyncPeer:partiallyFailNextSendUpdatesPResponse
				{
					testDocId = "totally failed, dude!",
				}

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- And we should see the one doc value in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocTwo',
					_slot = 'main',
					blah = 79,
				}
			} )

		-- :sendUpdatesP() configured to succeed, but nothing will happen
		-- because of the 30-second failure after transient failures.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And we should still see only the one doc value in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocTwo',
					_slot = 'main',
					blah = 79,
				}
			} )

		-- Wait out the blackout period.

		:waitFor( 30 )

		-- :sendUpdatesP() should now succeed.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should now see both doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocTwo',
					_slot = 'main',
					blah = 79,
				},
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				},
			} )

		-- :sendUpdatesP() should now say there's nothing to send.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDocTwice_receiveBeforeSecondChange( test )

	WFAsyncTestCase( test )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- :sendUpdatesP() should signal no activity.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- Now tweak a document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we got exactly one call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- :sendUpdatesP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And doc values should be unchanged in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we didn't get another call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- So we go back and get updates.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see updated sync status in the local document.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						},
					},
				}
			} )

		-- Now tweak same document a second time.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 45,
				}

		end )

		-- Now that we're up to date, we should be able to
		-- call :sendUpdatesP() and it should signal activity
		-- this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 45,
					foo = 42,
				}
			} )

		-- Verify that only the diffs were sent to sync peer.

		:waitFor( function()

			return testSyncPeer:argsToLastSendUpdatesPCall()

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDocId",
					_timestamp = 1,
					_fullDocValue = {
						bar = 45,
						foo = 42,
					},
					_previouslyAgreedValue = {
						_id = "testDocId",
						foo = 42,
					},
					bar = 45,
					-- foo = 42, -- foo should NOT be retransmitted!
				},
			} )

		-- Verify that we got a second call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '2' )

		-- Verify that document value didn't get tampered with.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 45,
					foo = 42,
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						},
					},
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendUpdatesP_oneDocTwice_receiveAfterSecondChange_testNightly( test )

	WFAsyncTestCase( test, 60 )

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- :sendUpdatesP() should signal no activity.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )

		-- Now tweak a document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we got exactly one call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- :sendUpdatesP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And doc values should be unchanged in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Verify that we didn't get another call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Now tweak same document a second time.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 45,
				}

		end )

		-- :sendUpdatesP() should fail because this time we don't know
		-- the remote timestamp of the document we're overriding.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- So we go back and get updates.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see updated sync status in the local document.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 45,
					foo = 42,
					_syncStatus = {
						test = {
							error = {
								message = "WFDocStoreImp:setDocValue: _changeCounter :new: does not match existing document change counter 1",
								time = "(valid ISO 8601 timestamp)",
							},
							timestamp = 1,
						},
					},
				}
			} )

		-- We can send changes now, but they'll be blacked out
		-- for a while.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( 30 )

		-- Now that we're up to date, we should be able to
		-- call :sendUpdatesP() and it should signal activity
		-- this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 45,
					foo = 42,
				}
			} )

		-- Verify that only the diffs were sent to sync peer.

		:waitFor( function()

			return testSyncPeer:argsToLastSendUpdatesPCall()

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDocId",
					_timestamp = 1,
					_fullDocValue = {
						bar = 45,
						foo = 42,
					},
					_previouslyAgreedValue = {
						_id = "testDocId",
						foo = 42,
					},
					bar = 45,
					-- foo = 42, -- foo should NOT be retransmitted!
				},
			} )

		-- Verify that we got a second call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '3' )


end

--------------------------------------------------------------------------------

-- @mark automatic sync

--------------------------------------------------------------------------------

function WFSyncAgentTest.testStartStopAutomaticSync_testNightly( test )

	WFAsyncTestCase( test, 60 )

		-- Verify precondition: Default state is auto-sync is off.

		:waitFor( function()
			return syncAgent:isSyncingAutomaticallyB()
		end )
		:continueWhenValueMatches( false )

		-- Queue up some changes on each side.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}

		end )

		-- Wait a bit and ensure that the docs did NOT propagate.

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}
			} )

		-- Now turn on automatic sync and validate that the docs
		-- DID propagate both ways.

		:waitFor( function()

			syncAgent:startAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( true )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
			} )

		-- Wait a little while and trigger some more activity on both sides.
		-- Make sure that traffic syncs without any external input.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc2',
					_slot = 'main',
					yaddaYadda = 'not so much',
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc2_fromRemote',
					_slot = 'main',
					bread = 'ciabatta',
				}

		end )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
				{
					_id = 'testDoc2_fromRemote',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 3,
						},
					},
					bread = 'ciabatta',
				},
				{
					_id = 'testDoc2',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					yaddaYadda = 'not so much',
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
				{
					_id = 'testDoc2_fromRemote',
					_slot = 'main',
					bread = 'ciabatta',
				},
				{
					_id = 'testDoc2',
					_slot = 'main',
					yaddaYadda = 'not so much',
				},
			} )

		-- Turn off automatic sync again. Make more changes and validate
		-- that the doc stores are islands again.

		:waitFor( function()

			syncAgent:stopAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( false )
			-- May go through a 'pendingStop' phase before settling on false.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
					oh = 'hello',
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
					henning = 'tv',
				}

		end )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc2_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 3,
						},
					},
					bread = "ciabatta",
				},
				{
					_id = "testDoc2",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					yaddaYadda = "not so much",
				},
				{
					_id = "testDoc",
					_slot = "main",
					oh = "hello",
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc",
					_slot = "main",
					yaddaYadda = true,
				},
				{
					_id = "testDoc2_fromRemote",
					_slot = "main",
					bread = "ciabatta",
				},
				{
					_id = "testDoc2",
					_slot = "main",
					yaddaYadda = "not so much",
				},
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					eis = "yum",
					henning = "tv",
				},
			} )

end

--------------------------------------------------------------------------------

-- @mark automatic sync

--------------------------------------------------------------------------------

function WFSyncAgentTest.testStartStopAutomaticSync_testNightly( test )

	WFAsyncTestCase( test, 60 )

		-- Verify precondition: Default state is auto-sync is off.

		:waitFor( function()
			return syncAgent:isSyncingAutomaticallyB()
		end )
		:continueWhenValueMatches( false )

		-- Queue up some changes on each side.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}

		end )

		-- Wait a bit and ensure that the docs did NOT propagate.

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}
			} )

		-- Now turn on automatic sync and validate that the docs
		-- DID propagate both ways.

		:waitFor( function()

			syncAgent:startAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( true )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
			} )

		-- Wait a little while and trigger some more activity on both sides.
		-- Make sure that traffic syncs without any external input.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc2',
					_slot = 'main',
					yaddaYadda = 'not so much',
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc2_fromRemote',
					_slot = 'main',
					bread = 'ciabatta',
				}

		end )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
				{
					_id = 'testDoc2_fromRemote',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 3,
						},
					},
					bread = 'ciabatta',
				},
				{
					_id = 'testDoc2',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					yaddaYadda = 'not so much',
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
				{
					_id = 'testDoc2_fromRemote',
					_slot = 'main',
					bread = 'ciabatta',
				},
				{
					_id = 'testDoc2',
					_slot = 'main',
					yaddaYadda = 'not so much',
				},
			} )

		-- Turn off automatic sync again. Make more changes and validate
		-- that the doc stores are islands again.

		:waitFor( function()

			syncAgent:stopAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( false )
			-- May go through a 'pendingStop' phase before settling on false.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
					oh = 'hello',
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
					henning = 'tv',
				}

		end )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc2_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 3,
						},
					},
					bread = "ciabatta",
				},
				{
					_id = "testDoc2",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					yaddaYadda = "not so much",
				},
				{
					_id = "testDoc",
					_slot = "main",
					oh = "hello",
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc",
					_slot = "main",
					yaddaYadda = true,
				},
				{
					_id = "testDoc2_fromRemote",
					_slot = "main",
					bread = "ciabatta",
				},
				{
					_id = "testDoc2",
					_slot = "main",
					yaddaYadda = "not so much",
				},
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					eis = "yum",
					henning = "tv",
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.DISABLE_testStartStopAutomaticSync_recoversFromFailure_testNightly( test )

	-- TODO [scouten 2013-12-10]: Figure out why this test is failing in Lr test runner.

	WFAsyncTestCase( test, 60 )

		-- Queue up some changes on each side.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}

		end )

		-- Configure the sync peer to fail the first sync attempt.
		-- Then turn on automatic sync and validate that the docs
		-- do eventually propagate both ways. The next request should
		-- succeed. The automatic sync code should retry after a
		-- brief pause.

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'error', "simulated failure" )

			syncAgent:startAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( true )

		:waitFor( 15 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.DISABLED_testAutomaticSyncViaTriggerSyncE_testNightly( test )

	-- TODO [scouten 2013-12-10]: Logic for triggerSyncE changed.
	-- It's no longer the case that triggerSyncE supercedes the built-in
	-- timer. Now it merely supplants it. Not sure how to test for that.

	if syncAgent then
		syncAgent:close()
	end
	
	testSyncPeer = WFTestSyncPeer( true )
	syncAgent = WFSyncAgent( docStore, testSyncPeer, 'test' )

	WFAsyncTestCase( test, 60 )

		-- Turn off automatic sync. The

		:waitFor( function()

			syncAgent:stopAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( true )

		:waitFor( 5 )

		-- Queue up some changes on each side. These should not propagate
		-- automatically because we're using the sync-now trigger.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}

		end )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}
			} )

		-- Tickle the sync-now trigger via sync peer.
		-- Wait a few seconds and ensure that the docs did propagate.
		
		:waitFor( function()

			testSyncPeer._triggerSyncE:post( true )
			
			return 5
		
		end )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
			} )

		-- Turn off automatic sync again. Make more changes and validate
		-- that the doc stores are islands again, even if we tickle the
		-- trigger event stream.

		:waitFor( function()

			syncAgent:stopAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( false )
			-- May go through a 'pendingStop' phase before settling on false.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
					oh = 'hello',
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
					henning = 'tv',
				}

		end )

		:waitFor( function()

			testSyncPeer._triggerSyncE:post( true )
			
			return 5
		
		end )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					yaddaYadda = true,
					oh = 'hello',
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
					henning = 'tv',
				},
			} )

end

--------------------------------------------------------------------------------

-- @mark kill switch

--------------------------------------------------------------------------------

function WFSyncAgentTest.testKillSwitch_get( test )

	WFAsyncTestCase( test )

		-- Get changes before there are any changes.

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					killSwitch = "testing kill switch on get",
				} )

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Verify kill switch set.

		:waitFor( function()
		
			return syncAgent:killSwitchB()
		
		end )
		:continueWhenValueMatches( "testing kill switch on get" )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testKillSwitch_send( test )

	WFAsyncTestCase( test )

		-- Queue up some changes to send.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}

		end )

		-- :sendUpdatesP() should signal no activity.

		:waitFor( function()

			testSyncPeer:overrideNextSendUpdatesPResponse( 'data',
				{
					killSwitch = "testing kill switch on send",
				} )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )

		-- Verify kill switch set.

		:waitFor( function()
		
			return syncAgent:killSwitchB()
		
		end )
		:continueWhenValueMatches( "testing kill switch on send" )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testKillSwitch_disableAutoSync_testNightly( test )

	WFAsyncTestCase( test, 60 )

		-- Queue up some changes on each side.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}

		end )

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}

		end )

		-- Wait a bit and ensure that the docs did NOT propagate.

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				}
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				}
			} )

		-- Now turn on automatic sync and validate that the docs
		-- DID propagate both ways.

		:waitFor( function()

			syncAgent:startAutomaticSync()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( true )

		:waitFor( 10 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
			} )

		-- Wait a little while and trigger the kill switch response.
		-- Make sure that auto sync disables without any external input.

		:waitFor( function()

			testSyncPeer:overrideNextChangedDocsPResponse( 'data',
				{
					killSwitch = "testing kill switch",
				} )

			return docStore:setDocValueP
				{
					_id = 'testDoc2',
					_slot = 'main',
					yaddaYadda = 'not so much',
				}

		end )

		:waitFor( 5 )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDoc_fromRemote",
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					eis = "yum",
				},
				{
					_id = "testDoc",
					_slot = "main",
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					yaddaYadda = true,
				},
				{
					_id = 'testDoc2',
					_slot = 'main',
					yaddaYadda = 'not so much',
				},
			} )

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDoc_fromRemote',
					_slot = 'main',
					eis = 'yum',
				},
				{
					_id = 'testDoc',
					_slot = 'main',
					yaddaYadda = true,
				},
			} )
			
		-- Make sure automatic sync got disabled.
		
		:waitFor( function()

			return syncAgent:isSyncingAutomaticallyB()

		end )
		:continueWhenValueMatches( false )
		
		-- Make sure kill switch reason is available.
		
		:waitFor( function()
		
			return syncAgent:killSwitchB()
		
		end )
		:continueWhenValueMatches( "testing kill switch" )

		-- Try turning on automatic sync again. Should be disallowed.

		:waitFor( function()

			KSTestUtils.assertFails( function()

				syncAgent:startAutomaticSync()

			end )

			syncAgent:stopAutomaticSync()

			return 'ok'

		end )

end

--------------------------------------------------------------------------------

-- @mark complex get/send sequences

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_doesntSendUpdateBackAfterReceive( test )

	WFAsyncTestCase( test )

		-- Verify precondition: No changes to receive or send.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		-- Now introduce a change on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Watch it flow through from remote to local.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 45,
				}
			} )

		-- Should be nothing to send back to remote peer.

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_serverChangesBetweenSendAndReadback( test )

	WFAsyncTestCase( test )

		-- Send a document change to remote side.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Now that it's been sent, change value on server before we read it back.
		-- This simulates the quick changes scenario that led Andy to add the
		-- _uploadedValues code to the old OzDocStore.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Now read back changes from remote.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Since we know that the server accepted our change, the change to
		-- "foo" should not be a conflict.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_localChangesBetweenSendAndReadback( test )

	WFAsyncTestCase( test )

		-- Send a document change to remote side.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Now that it's been sent, change value on local side before we read it back.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Now read back changes from remote.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Since we know that the server accepted our change, the change to
		-- "foo" should not be a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 45,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_twoLocalChangesWithSkip( test )

	-- This test verifies that we can handle changes being sent to remote site
	-- by other back-channel (as is, unfortunately, necessary for WFOzAlbumAssetUploader).
	-- Specifically tests a race condition that we've observed in the field when
	-- dealing with sort order.

	WFAsyncTestCase( test )

		-- Make a change locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Somehow, magically, that change also appears remotely.
		
		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )
		
		-- What happens if (hack alert!!!) we stomp this value into resolved slot?
		
		:waitFor( function()
			
			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'test-resolved',
					foo = 42,
				}
			
		end )

		-- Sync peer receives the change but ignores it because there's a back-channel.
		
		:waitFor( function()

			testSyncPeer:overrideNextSendUpdatesPResponse( 'data', --nil )
				{
					testDocId = ':skip:',
				} )

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be no sync status at this point.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Now that it's been sent, change value on local side before we read it back.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Now read back changes from remote. Since we haven't sent the second change
		-- yet, remote side should only see the first.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Remote side reads back the first value, which appears as a conflict since
		-- we never recorded it as having been sent.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								foo = 42,
							},
							timestamp = 1,
						},
					},
					foo = 45,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_getThenSendSequence( test )

	local getP, sendP
	local getPHasResolved, sendPHasResolved

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:delayNextChangedDocsP( .5 )

			getP = syncAgent:getChangesP()
				:observeP( function()
					getPHasResolved = true
					assert( not sendPHasResolved )
				end )
				:runUntilDoneP()

			sendP = syncAgent:sendUpdatesP()
				:observeP( function()
					sendPHasResolved = true
					assert( getPHasResolved )
				end )
				:runUntilDoneP()

			return getP

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return sendP
		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_sendAndThenGetSequence( test )

	local getP, sendP
	local getPHasResolved, sendPHasResolved

	WFAsyncTestCase( test )

		:waitFor( function()

			testSyncPeer:delayNextChangedDocsP( .5 )

			sendP = syncAgent:sendUpdatesP()
				:observeP( function()
					sendPHasResolved = true
					assert( not getPHasResolved )
				end )
				:runUntilDoneP()

			getP = syncAgent:getChangesP()
				:observeP( function()
					getPHasResolved = true
					assert( sendPHasResolved )
				end )
				:runUntilDoneP()

			return sendP

		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return getP
		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_localChangeWhileWaitingForReadback_testNightly( test )

	-- This simulates the five-second rule that we get from Oz.
	-- (IOW changes do not appear in the changes feed until five seconds
	-- after they were committed by Oz.)

	WFAsyncTestCase( test, 60 )

		-- Create initial document.

		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
		
		end )
		
		-- Send it to remote side.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that we have sync status.
		
		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
						},
					},
					foo = 42,
				},
			} )

		-- And we should see this doc value in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Now change locally *before* we read back from remote side.
		
		:waitFor( function()
		
			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 59,
				}
			
		end )
		
		-- Send this update through, too. This should fail because we
		-- haven't yet read back the change from remote side.
		
		:waitFor( function()
		
			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'error', ':ignore:' )
		
		-- Remote database should be unchanged.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )
	
		-- Now read changes from remote.
		
		:waitFor( function()
		
			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- Verify that document value did not change.
		
		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							error = {
								message = "WFDocStoreImp:setDocValue: _changeCounter :new: does not match existing document change counter 1",
								time = "(valid ISO 8601 timestamp)",
							},
							timestamp = 1,
						},
					},
					foo = 59,
				},
			} )
		
		-- Send changes. Nothing should happen because of the 30-second
		-- blackout rule on transient failures.
		
		:waitFor( function()
		
			return syncAgent:sendUpdatesP()
		
		end )
		:continueWhenValueMatches( 'data', false )
		
		-- Now wait out the blackout period.
		
		:waitFor( 30 )
		
		-- Now send changes again. This time we should have merged
		-- changes successfully and the push should work.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Verify that we have updated (non-error) sync status.
		
		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1, -- ???
						},
					},
					foo = 59,
				},
			} )

		-- And we should see this doc value in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 59,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testGetAndSend_reconcileRemoteDeleteWithoutResolved( test )

	-- This simulates what we believe is the root cause for the sync looper
	-- bug discovered this week. Here's what we *think* happened:
	
		-- User created an album in LrD.
		-- LrD uploaded that album and got an ACK for it.
		-- Oz failed to include that album document in the changes feed,
			-- presumably due to high load.
		-- For that reason, any assets included in that album did not get
			-- added to the album on Oz. (The _serverProvidedMetadata block
			-- was missing.)
		-- The user later deleted this album via LrM.
		-- BUG: LrD failed to accept the deletion of the album document.

	WFAsyncTestCase( test, 5 )

		-- Create an approximation of the relevant portion of the doc store
		-- as we received it from the user in question.

		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							error = {
								message = "Upload rejected",
								resultCode = "1002",
								resultData = {
									code = 1002,
									description = "Resource Already Exists",
									errors = {
										album = {
											"already exists",
										},
									},
								},
							retry = "2014-07-31T21:40:18.723Z",
							time = "2014-07-31T21:39:45.820Z",
							},
						},
					},
					cover = {
						_id = "60b86e1134ba19564da98539ba100cbd",
					},
					name = "December 2011",
					type = "album",
				}
			
		end )

		:waitFor( function()
		
			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'test-inbox',
					_serverProvidedMetadata = {
						created = "2014-04-08T17:46:26.477484Z",
						created_by = "urn:carousel:account:decbf139ed3080c1c91aaf9c856f8629",
						deleted_by = "urn:carousel:account:decbf139ed3080c1c91aaf9c856f8629",
						id = "urn:revel:catalog:b96c9532b4a446fd845492783f2c8506:album:b979c9d2b353dac98b1dcc43729e39fd",
						updated = "2014-06-19T04:42:51.329916Z",
					},
					_syncStatus = {
						test = {
							deleted = true,
							timestamp = "2014-06-19T04:42:51.329916Z",
						},
					},
					_deleted = true,
				}
			
		end )

		-- IMPORTANT DETAIL: There is no "-resolved" slot for this document.
		-- We think that is because Oz failed to send the document back in the
		-- changes feed.

		-- Do a sync cycle. This should self-resolve without sending any data.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Verify that we have sync status and that local document is now also deleted.
		
		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_deleted = true,
					_id = "testDocId",
					_serverProvidedMetadata = {
						created = "2014-04-08T17:46:26.477484Z",
						created_by = "urn:carousel:account:decbf139ed3080c1c91aaf9c856f8629",
						deleted_by = "urn:carousel:account:decbf139ed3080c1c91aaf9c856f8629",
						id = "urn:revel:catalog:b96c9532b4a446fd845492783f2c8506:album:b979c9d2b353dac98b1dcc43729e39fd",
						updated = "2014-06-19T04:42:51.329916Z",
					},
					_slot = "main",
					_syncStatus = {
						test = {
							timestamp = "2014-06-19T04:42:51.329916Z",
						},
					},
				},
			} )

		-- We should no longer have anything in test-inbox.

		:waitFor( function()

			return docStore:docValueP( 'testDocId', 'test-inbox' )

		end )
		:continueWhenValueMatches( 'data',
			{
				_empty = true,
				_id = "testDocId",
				_slot = "test-inbox",
			} )

		-- And the sync peer's :sendUpdatesP method should not have been called.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '0' )


end

--------------------------------------------------------------------------------

-- @mark conflict detection

--------------------------------------------------------------------------------

local function simpleConflictDetectionPreamble( test, waitTime )

	return WFAsyncTestCase( test, waitTime )

		-- Start with a simple document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Send it to remote side.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- And the sync peer's :sendUpdatesP method should have been called once.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Change doc value on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )

		-- And change it locally, so as to introduce a conflict.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And, voila! We should have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
							conflicts = {
								foo = 45,
							},
						},
					},
					foo = 46,
				}
			} )

		-- Now, :sendUpdatesP should refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_conflictRemainsOnSecondChange( test )

	simpleConflictDetectionPreamble( test )

		-- Change doc value a second time on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 48,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should still have a conflict, but with new remote side values.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
							conflicts = {
								foo = 48,
							},
						},
					},
					foo = 46,
				}
			} )

		-- Now, :sendUpdatesP should again refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_conflictAutoResolvedByRemoteChange( test )

	simpleConflictDetectionPreamble( test, 5 )

		-- Change doc value a second time on remote side.
		-- This time the remote side happens to choose the same value
		-- as the local conflicting change, so it should auto-resolve.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should no longer have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
								-- is coming up as 2 right now, but that fails below
						},
					},
					foo = 46,
				}
			} )

		-- Now, :sendUpdatesP should report a no-op because there's
		-- nothing new to upload.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Now change value locally.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 53,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- We should still not have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
						},
					},
					foo = 53,
				}
			} )

		-- Now, :sendUpdatesP should report traffic since it needs
		-- to send the new local doc value.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- The sync peer's :sendUpdatesP method should have been called
		-- a second time.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '2' )

		-- And we should see new doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 53,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testLocalOnlyConflictAutoResolved( test, waitTime )

	return WFAsyncTestCase( test, waitTime )

		-- Start with a simple document that has some _localOnly
		-- content.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					_localOnly = {
						blah = 45,
					},
				}

		end )

		-- Send it to remote side.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database,
		-- but without the _localOnly block.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Change doc value on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
				}

		end )

		-- And make the matching change locally, but change _localOnly
		-- content while we're at it. This could throw off the reconciled
		-- value logic, but it should be fixed automatically.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
					_localOnly = {
						blah = 47,
					},
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should not have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					foo = 46,
					_localOnly = {
						blah = 47,
					},
				}
			} )

end

--------------------------------------------------------------------------------

local function conflictDetectionPartialResolutionPreamble( test, waitTime )

	return WFAsyncTestCase( test, waitTime )

		-- Start with a simple document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 145,
				}

		end )

		-- Send it to remote side.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 145,
				}
			} )

		-- And the sync peer's :sendUpdatesP method should have been called once.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Change doc value on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
					bar = 155,
				}

		end )

		-- And change it locally, so as to introduce a conflict.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
					bar = 170,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And, voila! We should have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
							conflicts = {
								foo = 45,
								bar = 155,
							},
						},
					},
					foo = 46,
					bar = 170,
				}
			} )

		-- Resolved doc value should have advanced the sync-time counter
		-- but not the underlying content.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'test-resolved',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDocId",
					_slot = "test-resolved",
					_syncStatus = {
						test = {
							timestamp = 2,
							conflicts = {
								bar = 155,
								foo = 45,
							},
						},
					},
					bar = 145,
					foo = 42,
				},
			} )

		-- Now, :sendUpdatesP should refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Change doc value a second time on remote side.
		-- New value partially matches local value.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should still have a conflict, but with only the one remote
		-- value that conflicts, not both.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
							conflicts = {
								bar = 155,
							},
						},
					},
					foo = 46,
					bar = 170,
				}
			} )

		-- Resolved doc value should have advanced for the one field
		-- that we successfully merged.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'test-resolved',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'test-resolved',
					_syncStatus = {
						test = {
							timestamp = 3,
							conflicts = {
								bar = 155,
							},
						},
					},
					foo = 46,
					bar = 145,
				}
			} )

		-- Now, :sendUpdatesP should still refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_partialResolution1( test )

	conflictDetectionPartialResolutionPreamble( test )

		-- Now tweak local document to fully match remote value.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 155,
				}

		end )

		-- :sendUpdatesP() should signal no activity this time.
		-- (Local now matches remote so conflict should just go away.)

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- The conflict should have gone away locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
						},
					},
					foo = 46,
					bar = 155,
				}
			} )

		-- And we should see unchanged doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
					bar = 155,
				}
			} )

		-- Verify that we got no new call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_partialResolution2( test )

	conflictDetectionPartialResolutionPreamble( test )

		-- Now tweak local document to match the conflicting remote value
		-- but also adding a new value.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = 155,
					stop = true,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.
		-- The conflict should be removed and the new local value "stop"
		-- should be sent.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- The conflict should have gone away locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
						},
					},
					foo = 46,
					bar = 155,
					stop = true,
				}
			} )

		-- We should get new values back from remote side (confirming
		-- the change).

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- The local sync timestamp should be updated.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					foo = 46,
					bar = 155,
					stop = true,
				}
			} )

		-- And we should see changed doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 46,
					bar = 155,
					stop = true,
				}
			} )

end

--------------------------------------------------------------------------------

local function conflictDetectionNestedPartialResolutionPreamble( test, waitTime )

	return WFAsyncTestCase( test, waitTime )

		-- Start with a simple document.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 42 },
					bar = { b = 145 },
				}

		end )

		-- Send it to remote side.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 42 },
					bar = { b = 145 },
				}
			} )

		-- And the sync peer's :sendUpdatesP method should have been called once.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Change doc value on remote side.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 45 },
					bar = { b = 155 },
				}

		end )

		-- And change it locally, so as to introduce a conflict.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 46 },
					bar = { b = 170 },
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And, voila! We should have a conflict.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
							conflicts = {
								foo = { a = 45 },
								bar = { b = 155 },
							},
						},
					},
					foo = { a = 46 },
					bar = { b = 170 },
				}
			} )

		-- Resolved doc value should have advanced the sync-time counter
		-- but not the underlying content.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'test-resolved',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = "testDocId",
					_slot = "test-resolved",
					_syncStatus = {
						test = {
							timestamp = 2,
							conflicts = {
								bar = { b = 155 },
								foo = { a = 45 },
							},
						},
					},
					bar = { b = 145 },
					foo = { a = 42 },
				},
			} )

		-- Now, :sendUpdatesP should refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

		-- Change doc value a second time on remote side.
		-- New value partially matches local value.

		:waitFor( function()

			return testSyncPeer:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 46 },
				}

		end )

		-- Now read changes back from remote side.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should still have a conflict, but with only the one remote
		-- value that conflicts, not both.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
							conflicts = {
								bar = { b = 155 },
							},
						},
					},
					foo = { a = 46 },
					bar = { b = 170 },
				}
			} )

		-- Resolved doc value should have advanced for the one field
		-- that we successfully merged.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'test-resolved',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'test-resolved',
					_syncStatus = {
						test = {
							timestamp = 3,
							conflicts = {
								bar = { b = 155 },
							},
						},
					},
					foo = { a = 46 },
					bar = { b = 145 },
				}
			} )

		-- Now, :sendUpdatesP should still refuse to operate.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And the sync peer's :sendUpdatesP method should not have been called again.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_nestedPartialResolution1( test )

	conflictDetectionNestedPartialResolutionPreamble( test )

		-- Now tweak local document to fully match remote value.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = { b = 155 },
				}

		end )

		-- :sendUpdatesP() should signal no activity this time.
		-- (Local now matches remote so conflict should just go away.)

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- The conflict should have gone away locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
						},
					},
					foo = { a = 46 },
					bar = { b = 155 },
				}
			} )

		-- And we should see unchanged doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 46 },
					bar = { b = 155 },
				}
			} )

		-- Verify that we got no new call to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '1' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_nestedPartialResolution2( test )

	conflictDetectionNestedPartialResolutionPreamble( test )

		-- Now tweak local document to match the conflicting remote value
		-- but also adding a new value.

		:waitFor( function()

			return docStore:applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					bar = { b = 155 },
					stop = true,
				}

		end )

		-- :sendUpdatesP() should signal activity this time.
		-- The conflict should be removed and the new local value "stop"
		-- should be sent.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- The conflict should have gone away locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 3,
						},
					},
					foo = { a = 46 },
					bar = { b = 155 },
					stop = true,
				}
			} )

		-- We should get new values back from remote side (confirming
		-- the change).

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- The local sync timestamp should be updated.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					foo = { a = 46 },
					bar = { b = 155 },
					stop = true,
				}
			} )

		-- And we should see changed doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { a = 46 },
					bar = { b = 155 },
					stop = true,
				}
			} )

end

----------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndPeerInSync( test )

	-- Create asset local, on one sync peer, but in sync.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Create same doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Sync up.

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be no conflict. Should see unchanged doc in sync peer ...

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- ... and locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndTwoPeersInSync( test )

	-- Create asset local, both peers, all in sync.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Now create a doc in second sync peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Create same doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be no conflict. Should see unchanged on peer 1 ...

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- ... and peer 2 ...

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- ... and locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndPeerNotInSync( test )

	-- Create asset local, on one sync peer, not in sync.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Create conflict doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 52,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								foo = 42,
							},
							timestamp = 1,
						},
					},
					foo = 52,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndTwoPeersNotInSync( test )

	-- Create asset local, both peers, none in sync.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Now create a conflicting doc in second sync peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 52,
				}

		end )

		-- Create a conflicting doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 62,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								foo = 42,
							},
							timestamp = 1,
						},
						test2 = {
							conflicts = {
								foo = 52,
							},
							timestamp = 1,
						},
					},
					foo = 62,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndTwoPeersOneInSync( test )

	-- Create asset local, both peers, one peer in sync.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Now create a conflicting doc in second sync peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 52,
				}

		end )

		-- Create a conflicting doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 52,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								foo = 42,
							},
							timestamp = 1,
						},
						test2 = {
							timestamp = 1,
						},
					},
					foo = 52,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndPeerPartiallyInSync( test )

	-- Create asset local, on one sync peer, partially in sync.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 52,
				}

		end )

		-- Create a conflicting doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 62,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								bar = 52,
							},
							timestamp = 1,
						},
					},
					foo = 42,
					bar = 62,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createLocalAndTwoPeersPartiallyInSync( test )

	-- Create asset local, both peers, partially in sync (diff fields diff peers).

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 52,
				}

		end )

		-- Now create a conflicting doc in second sync peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 62,
					bar = 72,
				}

		end )

		-- Create a conflicting doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 72,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								bar = 52,
							},
							timestamp = 1,
						},
						test2 = {
							conflicts = {
								foo = 62,
							},
							timestamp = 1,
						},
					},
					foo = 42,
					bar = 72,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createTwoPeersNotInSync( test )

	-- Create conflicting assets on both sync peers only.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Now create a conflicting doc in second sync peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 52,
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.
		-- Synced first document, so no conflict with 1st peer.
		-- Only conflict with second peer.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							conflicts = {
								foo = 52,
							},
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_createTwoPeersPartiallyInSync( test )

	-- Create conflicting assets on both sync peers only.

	WFAsyncTestCase( test )

		-- Now create a doc in the sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 52,
				}

		end )

		-- Now create a conflicting doc in second sync peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = "not 52",
				}

		end )

		-- Sync up.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should be a conflict now in local doc store.
		-- Synced first document, so no conflict with 1st peer.
		-- Only partial conflict with second peer.
		
		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							conflicts = {
								bar = "not 52",
							},
							timestamp = 1,
						},
					},
					foo = 42,
					bar = 52,
				}
			} )

end

--------------------------------------------------------------------------------
-- Update conflicts with single peer covered in previous test cases.

function WFSyncAgentTest.testConflictDetection_updateTwoPeersInSync(test)

	-- Create asset local, on one sync peer, but in sync.

	WFAsyncTestCase( test )

		-- Make sure we are in sync.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Create a doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 5,
				}

		end )

		-- Sync up to both peers.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- For brevity, assume value made it to peers.
		-- If it didn't, this test will fail anyway.

		-- Update first peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 62,
					bar = 72,
				}

		end )

		-- Update second peer with same values.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 62,
					bar = 72,
				}

		end )

		-- Sync up to both peers.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Ensure that no conflicts exist and doc is updated locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					foo = 62,
					bar = 72,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_updateTwoPeersNotInSync( test )

	-- Create asset local, on one sync peer, but in sync.

	WFAsyncTestCase( test )

		-- Make sure we are in sync.
		
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Create a doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 52,
				}

		end )

		-- Sync up to both peers.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- For brevity, assume value made it to peers.
		-- If it didn't, tests will fail anyway.

		-- Update local doc.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = "not foo",
					bar = "not bar",
				}

		end )

		-- Update first peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 62,
					bar = 72,
				}

		end )

		-- Update second peer with different values.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 82,
					bar = 92,
				}

		end )

		-- Sync up to both peers.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Ensure that conflicts now exist and doc is updated locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								foo = 62,
								bar = 72,
							},
							timestamp = 2,
						},
						test2 = {
							conflicts = {
								foo = 82,
								bar = 92,
							},
							timestamp = 2,
						},
					},
					foo = "not foo",
					bar = "not bar",
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testConflictDetection_updateTwoPeersPartiallyInSync( test )

	-- Create asset local, on one sync peer, but in sync.

	WFAsyncTestCase( test )

		-- Make sure we are in sync.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Create a doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = 52,
				}

		end )

		-- Sync up to both peers.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- For brevity, assume value made it to peers.
		-- If it didn't, tests will fail anyway.

		-- Update local doc.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = "not foo",
					bar = "not bar",
				}

		end )

		-- Update first peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = "not foo",
					bar = 72,
				}

		end )

		-- Update second peer with different values.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 82,
					bar = "not bar",
				}

		end )

		-- Sync up to both peers.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Ensure that conflicts now exist and doc is updated locally.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							conflicts = {
								bar = 72,
							},
							timestamp = 2,
						},
						test2 = {
							conflicts = {
								foo = 82,
							},
							timestamp = 2,
						},
					},
					foo = "not foo",
					bar = "not bar",
				}
			} )

end

--------------------------------------------------------------------------------

-- @mark multi-peer sequences

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultipleSyncPeers( test )

	WFAsyncTestCase( test )

		-- Verify precondition: No changes to receive or send in either direction.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		-- Now introduce a change on one side.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Watch it flow through from peer #1 to "local" doc store.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Should still be nothing to get from peer #2.

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		-- And we should see this content in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 45,
				}
			} )

		-- Should be nothing to send to peer #1.

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		-- Now send the doc across to peer #2.

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Peer #2 should now have this doc.

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}
			} )

		-- Should get same value back from remote side.

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Our local doc should now have the sync status changes.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						},
					},
					foo = 45,
				}
			} )

		-- And now there should be no further traffic on either side.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultipleSyncPeers_simultaneous_noConflict( test )

	local getFromPeer1P, getFromPeer2P

	WFAsyncTestCase( test )

		-- Now introduce same change on both sides.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 45,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Now kick off sync with both peers simultaneously.

		:waitFor( function()

			getFromPeer1P = syncAgent:getChangesP():runUntilDoneP()
			getFromPeer2P = syncAgent2:getChangesP():runUntilDoneP()

			return getFromPeer1P

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()
			return getFromPeer2P
		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see this content in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							timestamp = 1,
						},
					},
					foo = 45,
				}
			} )

		-- And now there should be no further traffic on either side.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultipleSyncPeers_baselessMerge( test )

	-- Brian Kruse discovered a case in which a baseless merge could cause
	-- a nil table dereference. Exercise that case here.

	WFAsyncTestCase( test )

		-- Create same document on both remote peers without syncing between
		-- them first.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					captureDate = "2013-06-11T19:30:09",
					develop = {
						croppedHeight = 1920,
						croppedWidth = 2560,
						userOrientation = 3,
						xmpBase = "mumble",
					},
					height = 2448,
					importSource = {
						fileCreationDate = "2013-06-11T19:30:09",
						fileName = "es-0001-003.jpg",
						filePath = "/Users/foo/Desktop/m0815/2013-06-11/es-0001-003.jpg",
						fileSize = 1609118,
						folder = "2013-06-11",
						folderPath = "/Users/foo/Desktop/m0815/2013-06-11",
						importTimestamp = "2013-06-11T19:30:09",
						importedBy = "placeholderUserId",
						importedOnDevice = "importedOnDevice_LR_computer",
						sha256 = "f3af352e74573188ccbfcd6cd6ed725d7c9af9d3deaa771410b66272dc8297d7",
					},
					location = {
						altitude = 35,
						latitude = 53.545333,
						longitude = 9.938,
					},
					rating = 3,
					version = 1,
					width = 3264,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					captureDate = "2013-06-11T19:30:09",
					develop = {
						croppedHeight = 1905,
						croppedWidth = 2540,
						userOrientation = 3,
						xmpBase = "mumble",
					},
					height = 2448,
					importSource = {
						fileCreationDate = "2013-06-11T19:30:09",
						fileName = "es-0001-003.jpg",
						filePath = "/Users/foo/Desktop/m0815/2013-06-11/es-0001-003.jpg",
						fileSize = 1609118,
						folder = "/Users/foo/Desktop/m0815/2013-06-11",
						folderPath = "/Users/foo/Desktop/m0815/2013-06-11",
						importTimestamp = "2013-07-11T12:47:58",
						importedBy = "importedBy_LR",
						importedOnDevice = "importedOnDevice_LR_computer",
						sha256 = "f3af352e74573188ccbfcd6cd6ed725d7c9af9d3deaa771410b66272dc8297d7",
					},
					location = {
						altitude = 35,
						latitude = 53.545333,
						longitude = 9.938,
					},
					rating = 3,
					type = "asset",
					version = 1,
					width = 3264,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		-- Watch it flow through from peer #1 to "local" doc store.

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see this content in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					captureDate = "2013-06-11T19:30:09",
					develop = {
						croppedHeight = 1920,
						croppedWidth = 2560,
						userOrientation = 3,
						xmpBase = "mumble",
					},
					height = 2448,
					importSource = {
						fileCreationDate = "2013-06-11T19:30:09",
						fileName = "es-0001-003.jpg",
						filePath = "/Users/foo/Desktop/m0815/2013-06-11/es-0001-003.jpg",
						fileSize = 1609118,
						folder = "2013-06-11",
						folderPath = "/Users/foo/Desktop/m0815/2013-06-11",
						importTimestamp = "2013-06-11T19:30:09",
						importedBy = "placeholderUserId",
						importedOnDevice = "importedOnDevice_LR_computer",
						sha256 = "f3af352e74573188ccbfcd6cd6ed725d7c9af9d3deaa771410b66272dc8297d7",
					},
					location = {
						altitude = 35,
						latitude = 53.545333,
						longitude = 9.938,
					},
					rating = 3,
					version = 1,
					width = 3264,
				}
			} )

		-- Now pull from peer #2. Should get same doc (with some slight variations
		-- that cause a conflict).

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see this conflict in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							conflicts = {
								develop = {
									croppedHeight = 1905,
									croppedWidth = 2540,
								},
								importSource = {
									folder = "/Users/foo/Desktop/m0815/2013-06-11",
									importTimestamp = "2013-07-11T12:47:58",
									importedBy = "importedBy_LR",
								},
							},
							timestamp = 1,
						},
					},
					captureDate = "2013-06-11T19:30:09",
					develop = {
						croppedHeight = 1920,
						croppedWidth = 2560,
						userOrientation = 3,
						xmpBase = "mumble",
					},
					height = 2448,
					importSource = {
						fileCreationDate = "2013-06-11T19:30:09",
						fileName = "es-0001-003.jpg",
						filePath = "/Users/foo/Desktop/m0815/2013-06-11/es-0001-003.jpg",
						fileSize = 1609118,
						folder = "2013-06-11",
						folderPath = "/Users/foo/Desktop/m0815/2013-06-11",
						importTimestamp = "2013-06-11T19:30:09",
						importedBy = "placeholderUserId",
						importedOnDevice = "importedOnDevice_LR_computer",
						sha256 = "f3af352e74573188ccbfcd6cd6ed725d7c9af9d3deaa771410b66272dc8297d7",
					},
					location = {
						altitude = 35,
						latitude = 53.545333,
						longitude = 9.938,
					},
					rating = 3,
					type = "asset",
					version = 1,
					width = 3264,
				}
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultipleSyncPeers_serverProvidedMetadata_bug3584425( test )

	-- Test special-case code that ensures that _serverProvidedMetadata hash
	-- doesn't get stomped by changes from another sync peer.

	WFAsyncTestCase( test, 5 )

		-- Peer #1 introduces a doc.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = 3,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Now send the doc across to peer #2.

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Peer #2 should now have this doc.

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = 3,
				}
			} )

		-- Peer #1 changes the rating.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = 5,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Peer #2 annotates the doc.

		:waitFor( function()

			return testSyncPeer2:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					_serverProvidedMetadata = {
						-- obviously highly abbreviated from the original!
						updated = "2013-06-28T18:38:04.632976Z",
					},
				}

		end )

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Confirm that updated doc is in "local" doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:04.632976Z",
					},
					rating = 5,
				}
			} )

		-- Bring peer #1 up to date.

		:waitFor( function()
			return syncAgent:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', false )

		-- Send new rating to peer #2.

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = 5,
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:04.632976Z",
					},
				}
			} )

		-- Peer #2 annotates the doc again.

		:waitFor( function()

			return testSyncPeer2:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					_serverProvidedMetadata = {
						-- obviously highly abbreviated from the original!
						updated = "2013-06-28T18:38:12.165612Z",
					},
				}

		end )

		-- Peer #1 changes the rating again.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = 1,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		-- Read back previous value for rating from peer #2.

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 3,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 4,
						},
					},
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:12.165612Z",
					},
					rating = 1,
				}
			} )

		-- Send new rating to peer #2.

		:waitFor( function()
			return syncAgent2:sendUpdatesP()
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = 1,
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:12.165612Z",
					},
				}
			} )

		-- Peer #2 annotates the doc again.

		:waitFor( function()

			return testSyncPeer2:docStore():applyOverlayP
				{
					_id = 'testDocId',
					_slot = 'main',
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:20.281587Z",
					},
				}

		end )

		:waitFor( function()
			return syncAgent2:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 3,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 6,
						},
					},
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:20.281587Z",
					},
					rating = 1,
				}
			} )

		-- Now change rating again at peer #1 and receive changes.
		-- When bug #3584425 occurs, the result is _serverProvidedMetadata
		-- gets stomped.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					rating = nil,
				}

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

		:waitFor( function()
			return syncAgent:getChangesP()
		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 4,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 6,
						},
					},
					_serverProvidedMetadata = {
						updated = "2013-06-28T18:38:20.281587Z",
					},
				}
			} )

end

--------------------------------------------------------------------------------

-- @mark diagnostic log

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSendDiagnosticLogData_testNightly( test )

	-- Since diagnostic log is expected to evolve frequently, we're not
	-- verifying specific content. The main concern is that it doesn't fall
	-- over and die when generating the log.

	WFAsyncTestCase( test, 30 )

		:waitFor( function()

			return WFDiagnosticLog.processLogSectionP( syncAgent:sendDiagnosticLogData() )

		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

end

--------------------------------------------------------------------------------

-- @mark syncUntilDoneP tests

----------------------------------------------------------------------------------

function WFSyncAgentTest.testSyncUntilDone_whenAlreadySynced( test )

	WFAsyncTestCase( test )

		-- Create a doc in peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Sync peer doc.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true)

		-- Get changes in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Should be nothing to sync.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Should be all synced up.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = tempFromChangeCounter,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSyncUntilDone_repeatedSyncs( test )

	WFAsyncTestCase( test )

		-- Create a doc in peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Sync peer doc.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Get changes in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Should be nothing to sync.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Should be nothing to sync, many times over.

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		:waitFor( function()
			return syncAgent:syncUntilDoneP()
		end )
		:continueWhenValueMatches( 'data', false )

		-- Nothing new should have happened in doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = tempFromChangeCounter,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSyncUntilDone_whenAlreadySynced_compare( test )

	WFAsyncTestCase( test, 10 )

		-- Create a doc in peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Sync peer doc.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Get changes in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Sync again. Nothing should happen.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- No change should happen in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = tempFromChangeCounter,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testSyncUntilDone_afterAlreadySynced( test )

	WFAsyncTestCase( test, 10 )

		-- Create a doc in peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- Sync peer doc.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Get changes in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_stripAndSort )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Should be all synced up.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = tempFromChangeCounter,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

		-- Should be nothing to sync.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Should be no new activity in doc store after no-op sync.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = tempFromChangeCounter,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

-- @mark stress-test cases

--------------------------------------------------------------------------------

local function testDocIdNumber( i )

	return string.format( "testDocId-%04d", i )

end

--------------------------------------------------------------------------------

local function setupDocBatchChanged( totalDocs, omitAccepted )

	local batch = {}

	for i = 1, totalDocs do

		batch[ i ] =
			{
				_id = testDocIdNumber( i ),
				_slot = "main",
				_syncStatus = {
					test = {
						accepted = ( not omitAccepted )
							and "(valid ISO 8601 timestamp)" or nil,
						timestamp = i,
					},
				},
				foo = i,
			}
	end

	return batch

end

--------------------------------------------------------------------------------

local function createSyncPeerDocs( x )

	return testSyncPeer:docStore():setDocValueP
		{
			_id = testDocIdNumber( x ),
			_slot = 'main',
			foo = x,
		}

end

--------------------------------------------------------------------------------

local function createLocalDocs( x )

	return docStore:setDocValueP
		{
			_id = testDocIdNumber( x ),
			_slot = 'main',
			foo = x,
		}

end

--------------------------------------------------------------------------------

local function createSecondSyncPeerDocs( x )

	return testSyncPeer2:docStore():setDocValueP
		{
			_id = testDocIdNumber( x ),
			_slot = 'main',
			bar = x,
		}

end

--------------------------------------------------------------------------------

local function getSimpleIntegerArray( size )

	local a = {}
	for i = 1, size do
		a[ i ] = i
	end

	return a

end

--------------------------------------------------------------------------------

local function getBatchIntegerArray( startAt, size )

	local a = {}
	for i = 1, size do
		a[ i ] = startAt
		startAt = startAt + 1
	end

	return a

end

--------------------------------------------------------------------------------

local function mapChangedDocsId_stripAndSort( _, changedDocs )

	local ids = {}

	for i = 1, #changedDocs do
		ids[ i ] = changedDocs[ i ]
	end

	tempFromChangeCounter = changedDocs.continueFrom

    table.sort( ids )

	return 'data', ids

end

--------------------------------------------------------------------------------

local function setupDocBatchIds( startAt, endAt )

	local ids = {}

	for i = startAt, endAt do
		ids[ i ] = testDocIdNumber( i )
	end

	return ids

end

--------------------------------------------------------------------------------

local function setupDocBatchIdsOnly( size )

	local ids = {}

	for i = 1, size do
		ids[ testDocIdNumber( i ) ] = testDocIdNumber( i )
	end

	return ids

end

--------------------------------------------------------------------------------

local function allChangedDocsP( docStore, changedDocsArgs, previousChanges )

	return docStore:changedDocsP( changedDocsArgs )
		:mapDataP( function( _, changes )
		
			if #changes == 0 then

				return 'data', previousChanges or changes

			else
			
				changedDocsArgs.fromChangeCounter = changes.continueFrom
				
				if previousChanges then
				
					local nPreviousChanges = #previousChanges
					
					for i = 1, #changes do
						previousChanges[ i + nPreviousChanges ] = changes[ i ]
					end
					
					previousChanges.continueFrom = changes.continueFrom
					
				end
				
				return allChangedDocsP( docStore, changedDocsArgs,
										previousChanges or changes )
			
			end
		
		end )

end

--------------------------------------------------------------------------------

local function testMultiDocLoad_common( test, n, time, maxTransactions )

	local expectedChanged = setupDocBatchChanged( n, true )

	local testCase = WFAsyncTestCase( test, time or 10 )

		-- Now introduce n changes in remote DB.

		:waitFor( function()

			return KSPromises.allP( createSyncPeerDocs, getSimpleIntegerArray( n ) )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Bring all changes into main doc store.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Check that we have received all the changes we expected.

		:waitFor( function()

			return allChangedDocsP( docStore,
				{
					fromChangeCounter = 0,
					slot = 'main',
				} )
				:mapDataP( mapChangedDocs_stripAndSort )

		end )
		:continueWhenValueMatches( 'data', expectedChanged )

		-- Sync should report no further activity.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )
		
	if maxTransactions then
	
		testCase:waitFor( function()
		
			return docStore:nTransactionsP()
				:mapDataP( function( _, nTransactions, maybeTransactionSources )
				
					return 'data',
						   nTransactions,
						   nTransactions <= maxTransactions,
						   maybeTransactionSources
				
				end )
		
		end )
		:continueWhenValueMatches( 'data', ':ignore:', true, ':ignore:' )
	
	end

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_51docs( test )

	testMultiDocLoad_common( test, 51, 10, 16 )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_101docs_testNightly( test )

	testMultiDocLoad_common( test, 101, 20 )

end

--------------------------------------------------------------------------------

local function setupDocBatch_send( startAt, endAt )

	local batch = {}
	local batchCount = 1

	for i = startAt, endAt do

		batch[ batchCount ] =
			{
				_id = testDocIdNumber( i ),
				_slot = "main",
				foo = i,
			}

		batchCount = batchCount + 1

	end

	return batch

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_sendUpdatesP_50docs( test )

	local firstFiftyDocReturn = setupDocBatch_send( 1, 50 )

	WFAsyncTestCase( test, 10 )

		-- Create a bunch of documents.

		:waitFor( function()

			return KSPromises.allP( createLocalDocs, getSimpleIntegerArray( 50 ) )

		end )
		:continueWhenValueMatches( 'data', true )

		-- :sendUpdatesP() should signal activity this time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', firstFiftyDocReturn )

		-- Verify that we got exactly 13 calls to sync peer's :sendUpdatesP.
		-- Note that the sync peer is configured to ask for no more than 4 updates
		-- per call to :sendUpdatesP. ceil( 50 / 4 ) => 13

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '13' )

		-- :sendUpdatesP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:sendUpdatesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- And doc values should be unchanged in remote database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data', firstFiftyDocReturn )

		-- Verify that we didn't get any more calls to sync peer's :sendUpdatesP.

		:waitFor( function()

			return tostring( testSyncPeer:countSendUpdatesPCalls() )
				-- hack to get around interpretation of numbers as delay intervals

		end )
		:continueWhenValueMatches( 'data', '13' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_syncUntilDoneP_51docs_fromPeer_testNightly( test )

	local expectedChanged = setupDocBatchChanged( 51, true )

	WFAsyncTestCase( test, 10 )

		-- :syncUntilDoneP signal no activity.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Now tweak 51 documents on sync peer's doc store.

		:waitFor( function()

			local a = getSimpleIntegerArray( 51 )
			return KSPromises.allP( createSyncPeerDocs, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- :syncUntilDoneP() should signal activity this time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should see doc values in local doc store.

		:waitFor( function()

			return allChangedDocsP( docStore,
				{
					fromChangeCounter = 0,
					slot = 'main',
				} )
				:mapDataP( mapChangedDocs_stripAndSort )

		end )
		:continueWhenValueMatches( 'data', expectedChanged )

		-- Should be no more sync activity.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_syncUntilDoneP_51docs_toPeer_testNightly( test )

	WFAsyncTestCase( test, 10 )

		-- :syncUntilDoneP signal no activity.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Now tweak 51 docs in the local doc store.

		:waitFor( function()

			local a = getBatchIntegerArray( 1, 51 )
			return KSPromises.allP( createLocalDocs, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Send those to remote site.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			return allChangedDocsP( testSyncPeer:docStore(),
				{
					fromChangeCounter = 0,
					slot = 'main',
				} )
				:mapDataP( mapChangedDocs_stripAndSort )

		end )
		:continueWhenValueMatches( 'data', setupDocBatch_send( 1, 51 ) )

		-- :syncUntilDoneP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- We should see no further activity in remote DB.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = tempFromChangeCounter,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_stripAndSort )

		end )
		:continueWhenValueMatches( 'data', {} )

end

--------------------------------------------------------------------------------

local function testMultiDocLoad_idOnly( test, n, time )

	local batchIds = setupDocBatchIds( 1, n )

	WFAsyncTestCase( test, time )

		-- Create n documents in remote doc store.

		:waitFor( function()

			return KSPromises.allP( createSyncPeerDocs, getSimpleIntegerArray( n ) )

		end )
		:continueWhenValueMatches( 'data', true )

        -- Grab all changes via :syncUntilDoneP.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should see all of these new doc values in main database.
		-- Since we used .allP to create the documents above, there is no
		-- guarantee of ordering, so we sort the document IDs below.

		:waitFor( function()

			return allChangedDocsP( docStore,
				{
					fromChangeCounter = 0,
					slot = 'main',
					idsOnly = true,
				} )
				:mapDataP( mapChangedDocsId_stripAndSort )

		end )
		:continueWhenValueMatches( 'data', batchIds )

		-- Should not signal activity this time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Get API performance log if there is one.
		
		:waitFor( function()
			
			return docStore:getApiPerformanceLog()
		
		end )
		:continueWhenValueMatches( 'data', ':ignore:' )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_1000docs_idOnly_testNightly( test )

	testMultiDocLoad_idOnly( test, 1000, 60 )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_1001docs_idOnly_testNightly( test )

	testMultiDocLoad_idOnly( test, 1001, 60 )

end

--------------------------------------------------------------------------------

local function getAndStoreChangedLocalDocIds( x )

	return docStore:changedDocsP
		{
			fromChangeCounter = tempFromChangeCounter,
			slot = 'main',
		}
		:mapDataP( mapChangedDocs_stripToIdOnly )

end

--------------------------------------------------------------------------------

local function getAndStoreChangedPeerDocIds( x )

	return testSyncPeer:docStore():changedDocsP
		{
			fromChangeCounter = tempFromChangeCounter,
			slot = 'main',
		}
		:mapDataP( mapChangedDocs_stripToIdOnly )

end

--------------------------------------------------------------------------------

local function getAndStoreChangedPeer2DocIds( x )

	return testSyncPeer2:docStore():changedDocsP
		{
			fromChangeCounter = tempFromChangeCounter,
			slot = 'main',
		}
		:mapDataP( mapChangedDocs_stripToIdOnly )

end

---------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_fromTwoSyncPeers_testNightly( test )

	-- 101 documents from each sync peer.

	local expectedDocIds = setupDocBatchIdsOnly( 202 )

	WFAsyncTestCase( test, 20 )

		-- Now introduce changes.

		:waitFor( function()

			local a = getSimpleIntegerArray( 101 )
			return KSPromises.allP( createSyncPeerDocs, a )

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			local a = getBatchIntegerArray( 102, 101 )
			return KSPromises.allP( createSecondSyncPeerDocs, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should signal activity this time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Should signal activity from other peer as well.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in main database.

		:waitFor( function()

			-- We will need to get five batches of documents ceil(202/50).

			local a = getSimpleIntegerArray( 5 )
			return KSPromises.allP( getAndStoreChangedLocalDocIds, a )

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return allChangedDocIds

		end )
		:continueWhenValueMatches( 'data', expectedDocIds )

		-- Verify next call to getChangesP doesn't have any changes.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Verify next call to getChangesP doesn't have any changes for 2nd peer.

		:waitFor( function()

			return syncAgent2:getChangesP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_toTwoSyncPeers_testNightly( test )

	local expectedDocIds = setupDocBatchIdsOnly( 101 )

	WFAsyncTestCase( test, 20 )

		-- :syncUntilDoneP signal no activity.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- :syncUntilDoneP signal no activity.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Now tweak 50 docs in the local doc store.

		:waitFor( function()

			local a = getSimpleIntegerArray( 101 )
			return KSPromises.allP( createLocalDocs, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- :syncUntilDoneP() should signal activity this time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- As should the second sync agent.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in remote database.

		:waitFor( function()

			-- We will need to get three batches of documents ceil(101/50).

			local a = getSimpleIntegerArray( 3 )
			return KSPromises.allP( getAndStoreChangedPeerDocIds, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- All IDs should have been received.

		:waitFor( function()

			return allChangedDocIds

		end )
		:continueWhenValueMatches( 'data', expectedDocIds )

		-- And we should see doc values in second peer as well.

		:waitFor( function()

			-- Reset allChangedDocIds.

			allChangedDocIds = {}
			tempFromChangeCounter = 0

			-- We will need to get three batches of documents ceil(101/50).

			local a = getSimpleIntegerArray( 3 )
			return KSPromises.allP( getAndStoreChangedPeer2DocIds, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- All IDs should have been received.

		:waitFor( function()

			return allChangedDocIds

		end)
		:continueWhenValueMatches( 'data', expectedDocIds )

		-- :syncUntilDoneP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Same for second peer.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testMultiDocLoad_betweenTwoSyncPeers_testNightly( test )

	local expectedDocIds = setupDocBatchIdsOnly( 101 )

	WFAsyncTestCase( test, 20 )

		-- :syncUntilDoneP signal no activity.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- :syncUntilDoneP signal no activity.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Now tweak 101 docs in the first sync peer.

		:waitFor( function()

			local a = getSimpleIntegerArray( 101 )
			return KSPromises.allP( createSyncPeerDocs, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- :syncUntilDoneP() should get docs from first sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And send to second sync peer.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in second sync peer.

		:waitFor( function()

			-- We will need to get three batches of documents ceil(101/50).

			local a = getSimpleIntegerArray( 3 )
			return KSPromises.allP( getAndStoreChangedPeer2DocIds, a )

		end )
		:continueWhenValueMatches( 'data', true )

		-- :syncUntilDoneP() should not signal activity the second time.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- Same for second peer.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', false )

		-- All IDs should have been received.

		:waitFor( function()

			return allChangedDocIds

		end )
		:continueWhenValueMatches( 'data', expectedDocIds )

end

--------------------------------------------------------------------------------

-- @mark delete document scenarios

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSync_fromSingleSyncPeer( test )

	WFAsyncTestCase( test )
		
		-- Now create a doc on remote peer.
		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
					bar = { baz = 42 },
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in main database.
		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
					bar = { baz = 42 },
				}
			})
			
		-- Now let's delete from sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- :syncUntilDoneP() should sync deleted doc.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see value deleted from local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
					_syncStatus = {
						test = {
							timestamp = 2,
						},
					},
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSync_toSingleSyncPeer( test )

	WFAsyncTestCase( test )

		-- Now create a doc locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { bar = 42 },
				}

		end )

		-- :syncUntilDoneP() should send doc to sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see doc values in peer database.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = { bar = 42 },
				}
			} )

		-- Now let's delete from local doc store.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- :syncUntilDoneP() should sync deleted doc.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see value deleted from peer doc store.

		:waitFor( function()

			return testSyncPeer:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_deleted = true,
					_id = 'testDocId',
					_slot = 'main',
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSync_fromMultipleSyncPeers( test )

	WFAsyncTestCase( test )

		-- Create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should see doc values in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- And we should see doc values in second peer.

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}
			} )

		-- Now let's delete from sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- :syncUntilDoneP() should sync deleted doc

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see value deleted from local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
					_syncStatus = {
						test = {
							timestamp = 2,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
				},
			} )

		-- And we should see value deleted from second peer.

		:waitFor( function()

			return testSyncPeer2:docStore():changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_fromPeer( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Now let's delete from sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And update locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 1,
				}

		end )

		-- Note that we're using :getChangesP here because we know
		-- the attempt to :sendUpdatesP would fail with sync peer rejected
		-- documents.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see a conflict in local doc store, *but*
		-- the deletion takes effect.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
							conflicts = {
								_deleted = true,
							},
						},
					},
					_deleted = true,
				},
			} )

end
--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_toPeer( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Now let's delete from local peer.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And update on remote peer

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 1,
				}

		end )

		-- Note that we're using :getChangesP here because we know
		-- the attempt to :sendUpdatesP would fail with sync peer rejected
		-- documents.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see a conflict in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
							conflicts = {
								foo = 1,
							},
						},
					},
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_noConflictBothDeleted( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
					},
					foo = 42,
				}
			} )

		-- Now let's delete from local peer.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And also from remote peer

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- Note that we're using :getChangesP here because we know
		-- the attempt to :sendUpdatesP would fail with sync peer rejected
		-- documents.

		:waitFor( function()

			return syncAgent:getChangesP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- And we should see a conflict in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
						},
					},
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_multiplePeers( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Sync to second peer.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						}
					},
					foo = 42,
				}
			} )
			
		-- Now let's delete from second peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And update in first sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 1,
				}

		end )

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- No conflict. Deletion is the big hammer that wins.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
						},
					},
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_multiplePeers2( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Sync to second peer.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						}
					},
					foo = 42,
				}
			} )
			
		-- Now let's update second peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 1,
				}

		end )

		-- And delete in first sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- And we should see a conflict in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 2,
							conflicts = {
								foo = 1
							},
						},
					},
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_multiplePeersBothConflict( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.

		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		-- Sync to second peer.

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						}
					},
					foo = 42,
				}
			} )
			
		-- Now let's delete on second peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And delete in first sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And update locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 99,
				}

		end )
		
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- No conflict. Deletion is the big hammer that wins.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
						},
						test2 = {
							timestamp = 2,
						},
					},
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

function WFSyncAgentTest.testDeleteSyncConflict_multiplePeersBothConflict2( test )

	WFAsyncTestCase( test )

		-- Now create a doc on remote peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 42,
				}

		end )

		-- :syncUntilDoneP() should get doc from sync peer.
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		--sync to second peer
		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- We should see doc value in main database.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 1,
						},
						test2 = {
							accepted = "(valid ISO 8601 timestamp)",
							timestamp = 1,
						}
					},
					foo = 42,
				}
			} )
			
		-- Now let's delete second peer.

		:waitFor( function()

			return testSyncPeer2:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )

		-- And update first sync peer.

		:waitFor( function()

			return testSyncPeer:docStore():setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					foo = 98,
				}

		end )

		-- And delete locally.

		:waitFor( function()

			return docStore:setDocValueP
				{
					_id = 'testDocId',
					_slot = 'main',
					_deleted = true,
				}

		end )
		
		:waitFor( function()

			return syncAgent:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )

		:waitFor( function()

			return syncAgent2:syncUntilDoneP()

		end )
		:continueWhenValueMatches( 'data', true )
		
		-- And we should see a conflict in local doc store.

		:waitFor( function()

			return docStore:changedDocsP
				{
					fromChangeCounter = 0,
					slot = 'main',
				}
				:mapDataP( mapChangedDocs_strip )

		end )
		:continueWhenValueMatches( 'data',
			{
				{
					_id = 'testDocId',
					_slot = 'main',
					_syncStatus = {
						test = {
							timestamp = 2,
							conflicts = {
								foo = 98
							},
						},
						test2 = {
							timestamp = 2,
						},
					},
					_deleted = true,
				},
			} )

end

--------------------------------------------------------------------------------

return WFSyncAgentTest
